/*
 *  Bounce.java
 *
 *  Monday, June 19, 2000
 *
 *   Copyright 2000 Motorola, Inc.  All Rights Reserved.
 *  This notice does not imply publication.
 */

package com.mot.j2me.midlets.bounce;

import javax.microedition.lcdui.*;
import javax.microedition.midlet.*;

/**
 * Bounces rectangles on the screen. Demonstrates the application
 * framework and animation using multiple threads.
 *
 * @see MIDlet
 */
public class Bounce extends MIDlet {

    /**
     * The screen for this application
     */
    private BounceCanvas myCanvas;

    /**
     * The <code>RectThreads</code> managed by this application
     */
    private RectThread rectThreads[];

    /**
     * Set to true when its time to end this MIDlet
     */
    private boolean done = false;

    /**
     * Reference to the Display
     */
    private Display myDisplay;

    /**
     * Create the Bounce midlet
     */
    Bounce() {

        myDisplay = Display.getDisplay(this);
        myCanvas = new BounceCanvas();
        rectThreads = new RectThread[4];

        /*
         * Create four bouncing rectangles
         */
        rectThreads[0] = new RectThread(2, 2, 30, 30, 8, myCanvas);
        rectThreads[1] = new RectThread(-2, -2, 40, 40, 4, myCanvas);
        rectThreads[2] = new RectThread(-1, -1, 50, 10, 10,myCanvas);
        rectThreads[3] = new RectThread(4, 4, 10, 10, 2, myCanvas);
    }

    /**
     * Starts the animation threads
     */
    public void startApp() throws MIDletStateChangeException {

        myDisplay.setCurrent(myCanvas);
        for (int i = 0; i < rectThreads.length; i++) {
            rectThreads[i].start();
        }
    }

    /**
     * Stop the animation threads
     */
    public void pauseApp() {

        done = true;
        for(int i = 0; i < rectThreads.length; i++) {
            rectThreads[i].stopThread();
        }
    }

    /**
     * Called by the framework before the application is unloaded
     */
    public void destroyApp(boolean unconditional) {
    }

    /**
     * The frame class for this MIDlet
     */
    class BounceCanvas extends Canvas {

        /**
         * Background color (i.e. the color of the screen)
         */
        public static final int BACKGROUND_COLOR = 0xFFFFFF; // white

        /**
         * Foreground color (i.e. the color of the rectangles)
         */
        public static final int FOREGROUND_COLOR = 0x000000; // black

        /**
         * Paint a white background in the frame
         */
        public void paint(Graphics g) {

            synchronized (g) {

                /*
                 * Clear the screen
                 */
                g.setColor(BACKGROUND_COLOR);
                g.fillRect(0, 0, getWidth(), getHeight());

                /*
                 * Paint all the rectangles
                 */
                for (int i = 0; i < rectThreads.length; i++) {
                    rectThreads[i].paint(g);
                }
            }
        }
    }

    /**
     * A thread class which draws a bouncing rectangle in a frame
     */
    class RectThread extends Thread {

        /**
         * Time (ms) between each animation frame
         */
        private final int FRAME_DELAY = 100;

        /**
         * Current x-coordinate of the rectangle
         */
        private int x;

        /**
         * Current y-coordinate of the rectangle
         */
        private int y;

        /**
         * Change in <code>x</code> per animation cycle
         */
        private int dx;

        /**
         * Change in <code>y</code> per animation cycle
         */
        private int dy;

        /**
         * Coordinate of the right edge of the screen
         */
        private int rightEdge;

        /**
         * Coordinate of the bottom edge of the screen
         */
        private int bottomEdge;

        /**
         * Length and width of this rectangle
         */
        private int size;

        /**
         * The frame to draw in
         */
        private BounceCanvas myCanvas;

        /**
         * Flag which stops this thread
         */
        private volatile boolean stopThread = false;

        /**
         * Creates a <code>RectThread</code>
         *
         * @param   dx          change in "x" per animation cycle
         * @param   dy          change in "y"
         * @param   startX      starting x-coordinate of the rectangle
         * @param   startY      starting y-coordinate
         * @param   size        length and width of the square
         * @param   myCanvas    the canvas to draw in
         */
        RectThread(int dx, int dy, int startX, int startY, int size,
                BounceCanvas myCanvas) {

            x = startX;
            y = startY;
            this.dx = dx;
            this.dy = dy;
            this.size = size;

            this.myCanvas = myCanvas;

            rightEdge = myCanvas.getWidth() - 1;
            bottomEdge = myCanvas.getHeight() - 1;
        }

        /**
         * Stops this thread
         */
        void stopThread() {

            stopThread = false;
        }

        /**
         * Paint the rectangle
         */
        public void paint(Graphics g) {

            g.setColor(BounceCanvas.FOREGROUND_COLOR);
            g.fillRect(x, y, size, size);
        }

        /**
         * Animates a rectangle in a frame
         */
        public void run(){

            while (!stopThread) {

                int oldX = x;
                int oldY = y;

                try {
                    sleep(FRAME_DELAY);
                } catch (InterruptedException ie) {
                }

                /*
                 * Calculate next position
                 */
                x += dx;
                y += dy;

                /*
                 * Did we hit a horizontal edge?
                 */
                if ((x + size) > rightEdge) {
                    dx *= -1;               // reverse direction
                    x = rightEdge - size;   // put us back in bounds
                } else if (x < 0) {
                    dx *= -1;
                    x = 0;
                }

                /*
                 * Did we hit a vertical edge?
                 */
                if ((y + size) > bottomEdge) {
                    dy *= -1;
                    y = bottomEdge - size;
                } else if (y < 0) {
                    dy *= -1;
                    y = 0;
                }

                myCanvas.repaint(Math.min(oldX, x), Math.min(oldY, y),
                                 Math.abs(x-oldX) + size,
                                 Math.abs(y-oldY) + size);
                myCanvas.serviceRepaints();
            }
        }
    }
}