/*
 * @(#)Manager.java	1.26 01/09/04
 *
 *  Copyright (c) 2001 Sun Microsystems, Inc., 901 San Antonio Road,
 *  Palo Alto, CA 94303, U.S.A.  All Rights Reserved.
 *
 *  Sun Microsystems, Inc. has intellectual property rights relating
 *  to the technology embodied in this software.  In particular, and
 *  without limitation, these intellectual property rights may include
 *  one or more U.S. patents, foreign patents, or pending
 *  applications.  Sun, Sun Microsystems, the Sun logo, Java, KJava,
 *  and all Sun-based and Java-based marks are trademarks or
 *  registered trademarks of Sun Microsystems, Inc.  in the United
 *  States and other countries.
 *
 *  This software is distributed under licenses restricting its use,
 *  copying, distribution, and decompilation.  No part of this
 *  software may be reproduced in any form by any means without prior
 *  written authorization of Sun and its licensors, if any.
 *
 *  FEDERAL ACQUISITIONS:  Commercial Software -- Government Users
 *  Subject to Standard License Terms and Conditions
 */
package com.sun.midp.dev;

import javax.microedition.io.*;
import java.util.*;
import java.io.*;

import javax.microedition.midlet.MIDlet;

import javax.microedition.lcdui.*;

import javax.microedition.rms.*;

import com.sun.midp.lcdui.Resource;

import com.sun.midp.midlet.MIDletInfo;
import com.sun.midp.midlet.MIDletSuite;
import com.sun.midp.midlet.Scheduler;

import com.sun.midp.midlet.MIDletSuite;

import com.sun.midp.midletsuite.*;

import com.sun.midp.security.SecurityDomain;
import com.sun.midp.security.Actions;

import com.sun.midp.Configuration;

import javax.microedition.io.*;

import com.sun.midp.io.j2me.storage.*;


/**
 * The Graphical MIDlet suite manager.
 * <p>
 * Starts with a selector that provides a list of MIDlet suites and
 * a set of commands to perform. It displays the MIDlet names for a suite
 * under the MIDlet suite name, except if there is only one suite then
 * instead of display the suite name, MIDlet-1 name and icon are used.
 * <p>
 * The commands are:</p>
 * <ul>
 * <li><b>Install</b>: Let the user install a suite from a list suites
 * obtained using an HTML URL given by the user. This list is derived by
 * extracting the links with hrefs that are in quotes and end with ".jad" from
 * the HTML page. An href in an extracted link is assumed to be an absolute
 * URL for a MIDP application descriptor.</li>
 * <li><b>Launch</b>: Launch the suite the user selected.
 * <li><b>Remove</b>: Remove the suite (with confirmation) the user selected.
 * </li>
 * <li><b>Update</b>: Update the suite the user selected.</li>
 * <li><b>Info</b>: Show the user general information of the selected suite.
 * <li><b>Settings</b>: Let the user change the manager's settings.
 * </ul>
 */
public class Manager extends MIDlet implements CommandListener {
    /** Standard timeout for our alerts. */
    private static final int ALERT_TIMEOUT = 1500;
    /** Standard name for the settings record store. */
    private static final String SETTINGS_STORE = "settings";

    /** This class has a different security domain than the MIDlet suite. */
    private static SecurityDomain classSecurityDomain;
    /** Cache of the suite icon. */
    private static Image suiteIcon;
    /** Cache of the empty icon. */
    private static Image emptyIcon;
    /** Cache of the single suite icon. */
    private static Image singleSuiteIcon;
    /** So the static method can know if there is a color display. */
    private static boolean colorDisplay;

    /** The installer that is being used to install or update a suite. */
    private Installer installer;
    /** List of all the MIDlet suites. */
    private List mlist;
    /** True until startApp is called the first time. */
    private static boolean first = true;
    /** Display for this MIDlet. */
    private Display display;    
    /** Number of midlets in minfo. */
    private int mcount;
    /** MIDlet suite information, class, name, icon; one per MIDlet suite. */
    private MIDletSuiteInfo[] minfo; 
    /** Currently selected suite. */
    private int selectedSuite;
    /** Contains the new default URL the user typed in. */
    private TextBox settingsTextBox;
    /** Contains the default URL for the install list. */
    private String defaultInstallListUrl = "http://";
    /** Contains the URL the user typed in. */
    private TextBox urlTextBox;
    /** Lets the enter a password and username. */
    private Form passwordForm;
    /** Contains the username for installing. */
    private TextField usernameField;
    /** Contains the password for installing. */
    private TextField passwordField;
    /** Background installer that holds state for the current install. */
    private BackgroundInstaller backgroundInstaller;
    /** Displays the progress of the install. */
    private Form conForm;
    /** Keeps track of when the display last changed. UCT in millseconds. */
    private long lastDisplayChange;
    /** What to display to the user when the current action is cancelled. */
    private String cancelledMessage;
    /** What to display to the user when the current action is finishing. */
    private String finishingMessage;
    /** Displays a list of suites to install to the user. */
    private List installListBox;
    /** Contains a list of suites to install. */
    private Vector installList;
    /** Command object for "Launch". */
    private Command launchCmd = new Command(Resource.getString("Launch"), 
                                           Command.ITEM, 1);
    /** Command object for "Info". */
    private Command infoCmd = new Command(Resource.getString("Info"), 
                                           Command.ITEM, 2);
    /** Command object for "Remove". */
    private Command removeCmd = new Command(Resource.getString("Remove"), 
                                            Command.ITEM, 3);
    /** Command object for "Update". */
    private Command updateCmd = new Command(Resource.getString("Update"), 
                                           Command.ITEM, 4);
    /** Command object for "Install". */
    private Command installCmd = new Command(Resource.getString("Install"), 
                                           Command.SCREEN, 1);
    /** Command object for "Cancel" command for the remove form. */
    private Command cancelInstallCmd = new Command(
                                           Resource.getString("Cancel"), 
                                           Command.CANCEL, 1);
    /** Command object for "Settings". */
    private Command settingsCmd = new Command(Resource.getString("Settings"), 
                                           Command.SCREEN, 2);
    /** Command object for "About". */
    private Command aboutCmd = new Command(Resource.getString("About"), 
                                           Command.SCREEN, 3);
    /** Command object for settings screen to save the settings. */
    private Command saveSettingsCmd =
        new Command(Resource.getString("Save"), Command.SCREEN, 1);
    /** Command object for URL screen to go and get URL's for suites. */
    private Command selectSuiteCmd =
        new Command(Resource.getString("Go"), Command.SCREEN, 1);
    /** Command object for "Back" command for the info form. */
    private Command backCmd = new Command(Resource.getString("Back"), 
                                           Command.BACK, 1);
    /** Command object for "Cancel" command for the remove form. */
    private Command cancelCmd = new Command(Resource.getString("Cancel"), 
                                           Command.CANCEL, 1);
    /** Command object for "Remove" command for the remove form. */
    private Command removeOkCmd = new Command(Resource.getString("Remove"), 
                                           Command.SCREEN, 1);
    /** Command object for "Stop" command for update and install forms. */
    private Command stopCmd = new Command(Resource.getString("Stop"), 
                                           Command.STOP, 1);
    /** Command object for "Next" command for password form. */
    private Command nextCmd = new Command(Resource.getString("Next"), 
                                           Command.SCREEN, 1);
    /** Command object for "OK" command for warning form. */
    private Command okCmd = new Command(Resource.getString("OK"), 
                                           Command.OK, 1);
    /** Translated small copyright string. */
    private static final String SMALL_COPYRIGHT = Resource.getString(
        "Copyright 2000-2001 Sun Microsystems Inc. All Rights Reserved.");
    /** Translated long copyright string. */
    private static final String COPYRIGHT = Resource.getString(
        "Copyright 2000-2001 Sun Microsystems Inc. All Rights Reserved.\n"
      + "Sun, Sun Microsystems, the Sun logo, Java, and all Sun-based and "
      + "Java-based marks are trademarks or registered trademarks of Sun "
      + "Microsystems, Inc. in the United States and other countries.\n"
      + "\n"
      + "This product meets the Wireless Profile Compatible designation "
      + "requirement, containing tested and compatible Wireless Profile "
      + "software from Sun Microsystems, Inc.\n"
      + "\n"
      + "Use is subject to license terms and limited to demonstration only. "
      + "Sun Microsystems, Inc.  has intellectual property rights relating "
      + "to the technology embodied in this software.  In particular, and "
      + "without limitation, these intellectual property rights may include "
      + "one or more U.S. patents, foreign patents, or pending applications.\n"
      + "U.S. Government approval required when exporting the product.\n"
      + "\n"
      + "FEDERAL ACQUISITIONS\n"
      + "Commercial Software -- Government Users Subject to Standard "
      + "License Terms and Conditions.");

    /**
     * Initializes the security domain for this class, so it can
     * perform actions that a normal MIDlet Suite cannot.
     *
     * @param theDomain SecurityDomain for this class.
     */
    public static void initSecurityDomain(SecurityDomain theDomain) {
        if (classSecurityDomain != null) {
            return;
        }

        classSecurityDomain = theDomain;
    }

    /**
     * Set up the initial state for the Manager.
     *
     * @param logoDisplayed true if the logo has been displayed
     */
    public static void initialize(boolean logoDisplayed) {
        first = !logoDisplayed;
    }

    /**
     * Translate an InvalidJadException into a message for the user.
     *
     * @param exception exception to translate
     * @param name name of the MIDlet suite to insert into the message
     * @param jadUrl URL of a JAD, can be null
     *
     * @return message to display to the user
     */
    private static String translateJadException(
            InvalidJadException exception, String name, String jadUrl) {
        String[] values = {name, jadUrl, exception.getExtraData()};
        String key;

        switch (exception.getReason()) {
        case InvalidJadException.MISSING_PROVIDER_CERT:
        case InvalidJadException.MISSING_SUITE_NAME:
        case InvalidJadException.MISSING_VENDOR:
        case InvalidJadException.MISSING_VERSION:
        case InvalidJadException.MISSING_JAR_URL:
        case InvalidJadException.MISSING_JAR_SIZE:
        case InvalidJadException.MISSING_JAR_DIGEST:
            key = "%1 cannot be installed because critical information is " +
                  "missing from the application file (.jad).";
            break;

        case InvalidJadException.MISSING_CONFIGURATION:
        case InvalidJadException.MISSING_PROFILE:
            key = "%1 cannot be installed because critical information is " +
                  "missing from the archive file (.jar).";
            break;

        case InvalidJadException.SUITE_NAME_MISMATCH:
        case InvalidJadException.VERSION_MISMATCH:
        case InvalidJadException.VENDOR_MISMATCH:
        case InvalidJadException.JAR_SIZE_MISMATCH:
            key = "%1 cannot be installed because critical information " +
                  "between the website and the application file does not " +
                  "match.";
            break;

        case InvalidJadException.OLD_VERSION:
            key = "A newer version of %1 is already installed on this " +
                  "device.";
            break;

        case InvalidJadException.NEW_VERSION:
            key = "An older version of %1 is already installed on this " +
                  "device. Do you want to update the older version?";
            break;

        case InvalidJadException.ALREADY_INSTALLED:
            key = "The same version of %1 is already installed on this " +
                  "device. Do you want to update the existing version?";
            break;

        case InvalidJadException.JAD_SERVER_NOT_FOUND:
        case InvalidJadException.JAD_NOT_FOUND:
            key = "%1 cannot be found at this URL. " +
                  "Contact the application provider for more information.";
            break;

        case InvalidJadException.JAR_SERVER_NOT_FOUND:
        case InvalidJadException.JAR_NOT_FOUND:
            key = "The archive file (.jar) for %1 cannot be found at its " + 
                  "URL. Contact the application provider for more " +
                  "information.";
            break;

        case InvalidJadException.INVALID_KEY:
        case InvalidJadException.INVALID_VALUE:
        case InvalidJadException.INVALID_VERSION:
            key = "%1 cannot be installed because critical " + 
                  "information is not formatted correctly and is invalid.";
            break;

        case InvalidJadException.INSUFFICIENT_STORAGE:
            key = "There is not enough room to install %1 (%3K is needed " +
                  "for installation.). Try removing other items to free " +
                  "up space.";
            break;

        case InvalidJadException.INVALID_JAR_TYPE:
            key = "The archive file (.jar) for %1 does not appear to be " +
                  "the correct type. Contact the application provider " +
                  "for more information.";
            break;

        case InvalidJadException.INVALID_JAD_TYPE:
            key = "The application file (.jad) for %1 does not appear to be " +
                  "the correct type. Contact the application provider " +
                  "for more information.";
            break;

        case InvalidJadException.CORRUPT_JAR:
            key = "The archive file (.jar) for %1 appears to be corrupt. " +
                  "Contact the application provider " +
                  "for more information.";
            break;

        case InvalidJadException.JAD_MOVED:
            key = "The new version of %1 is not from the same provider " +
                  "as the old version. The download URLs do not match. " +
                  "Do you want to install the new version?" +
                  "\n\nOld URL: %3\nNew URL: %2";
            break;

        case InvalidJadException.CANNOT_AUTH:
            key = "The server has requested an authentication of the user " +
                  "with a scheme that this device does not support.";
            break;

        case InvalidJadException.DEVICE_INCOMPATIBLE:
            key = "%1 is not designed to work with this device and cannot " +
                  "be installed.";
            break;

        default:
            return exception.getMessage();
        }

        return Resource.getString(key, values);
    }

    /**
     * Gets the MIDlet suite icon from storage.
     *
     * @return icon image
     */
    private static Image getSuiteIcon() {
        if (suiteIcon != null) {
            return suiteIcon;
        }

        suiteIcon = getIconFromStorage(
                        (colorDisplay ? "_suite_8.png" : "_suite_2.png"));
        return suiteIcon;
    }

    /**
     * Gets the empty icon from storage.
     *
     * @return icon image
     */
    private static Image getEmptyIcon() {
        if (emptyIcon != null) {
            return emptyIcon;
        }

        emptyIcon = getIconFromStorage("_empty.png");
        return emptyIcon;
    }

    /**
     * Gets the single MIDlet suite icon from storage.
     *
     * @return icon image
     */
    private static Image getSingleSuiteIcon() {
        if (singleSuiteIcon != null) {
            return singleSuiteIcon;
        }

        singleSuiteIcon = getIconFromStorage(
                        (colorDisplay ? "_single8.png" : "_single2.png"));
        return singleSuiteIcon;
    }

    /**
     * Gets an icon from storage.
     *
     * @param iconName name without a path
     * @return icon image
     */
    private static Image getIconFromStorage(String iconName) {
        String iconFilename;
        RandomAccessStream stream;
        byte[] rawPng;

        iconFilename = File.getStorageRoot() + iconName;
        stream = new RandomAccessStream(classSecurityDomain);
        try { 
            stream.connect(iconFilename, Connector.READ);
            rawPng = new byte[stream.getSizeOf()];
            stream.readBytes(rawPng, 0, rawPng.length);
            stream.disconnect();
            return Image.createImage(rawPng, 0, rawPng.length);
        } catch (java.io.IOException noImage) {
        }

        return null;
    }

    /**
     * Create and initialize a new Manager MIDlet.
     * The Display is retreived and the list of MIDlets read
     * from the descriptor file.
     */
    public Manager() {
        installer = Installer.getInstaller(classSecurityDomain);
        display = Display.getDisplay(this);
        colorDisplay = display.isColor();
        mcount = 0;
        minfo = new MIDletSuiteInfo[20];
        readMIDletSuiteInfo();
        restoreSettings();
    }

    /**
     * Start puts up a List of the MIDlets found in the descriptor file.
     */
    public void startApp() {
        setupList();

        if (mcount > 0) {
            mlist.addCommand(launchCmd);
            mlist.addCommand(infoCmd);
            mlist.addCommand(removeCmd);
            mlist.addCommand(updateCmd);
        }

        mlist.addCommand(installCmd);
        mlist.addCommand(settingsCmd);
        mlist.addCommand(aboutCmd);
        mlist.setCommandListener(this); // Listen for the selection

        if (first) {
            first = false;
            Alert splash = new Alert(null);
            if (display.numColors() > 2) {
                splash.setImage(getIconFromStorage((colorDisplay ?
                    "_logo_8.png" : "_logo_2.png")));
            }

            splash.setString(SMALL_COPYRIGHT);
            display.setCurrent(splash, mlist);
        } else {
            display.setCurrent(mlist);
        }
    }

    /**
     * Pause; there are no resources that need to be released.
     */
    public void pauseApp() {
    }

    /**
     * Destroy cleans up.
     *
     * @param unconditional is ignored; this object always
     * destroys itself when requested.
     */
    public void destroyApp(boolean unconditional) {
        /*
         * The backgroundInstaller could be waiting for an update of the 
         * install state.
         */
        cancelBackgroundInstall();
    }

    /**
     * Respond to a command issued on any Screen.
     * The commands on list is Select and About.
     * Select triggers the creation of the MIDlet of the same name.
     * About puts up the copyright notice.
     *
     * @param c command activiated by the user
     * @param s the Displayable the command was on.
     */
    public void commandAction(Command c, Displayable s) {
        if ((s == mlist && c == List.SELECT_COMMAND) || (c == launchCmd)) {
            launchSuite(minfo[mlist.getSelectedIndex()]);
        } else if (c == infoCmd) {
            displaySuiteInfo(minfo[mlist.getSelectedIndex()]);
        } else if (c == removeCmd) {
            // save the selected suite for the removeOk command
            selectedSuite = mlist.getSelectedIndex();
            confirmRemove(minfo[selectedSuite]);
        } else if (c == removeOkCmd) {
            removeSuite(minfo[selectedSuite]);
        } else if (c == updateCmd) {
            updateSuite(minfo[mlist.getSelectedIndex()]);
        } else if (s == mlist && c == installCmd) {
            // user wants a list of suites to install
            getUrl();
        } else if (c == selectSuiteCmd) {
            // user wants to select a suite to install
            selectSuiteToInstall(urlTextBox.getString());
        } else if (s == installListBox &&
                  (c == List.SELECT_COMMAND || c == installCmd)) {
            // the user has selected a suite
            installSuite(installListBox.getSelectedIndex());
        } else if (c == nextCmd) {
            // the user has entered a username and password
            resumeInstallWithPassword();
        } else if (c == okCmd) {
            resumeInstallAfterWarning();
        } else if (c == settingsCmd) {
            getSettings();
        } else if (c == saveSettingsCmd) {
            saveSettings();
        } else if (c == aboutCmd) {
            Alert a = new Alert(Resource.getString("About Wireless Profile"));
            a.setString(COPYRIGHT);
            display.setCurrent(a, mlist);
        } else if (c == stopCmd) {
            if (installer != null) {
                /*
                 * backgroundInstaller may be displaying
                 * the "Finishing" message
                 */
                if (installer.stopInstalling()) {
                    displayCancelledMessage(cancelledMessage);
                }
            } else {
                display.setCurrent(mlist);
            }
        } else if (c == cancelInstallCmd) {
            cancelBackgroundInstall();
        } else if (c == backCmd || c == cancelCmd) {
            // goto back to the main list of suites
            display.setCurrent(mlist);
        }
    }

    /**
     * Lauches a suite.
     *
     * @param suiteInfo information for suite to launch
     */
    private void launchSuite(MIDletSuiteInfo suiteInfo) {
        try {
            // Create an instance of the MIDlet class
            // All other initialization happens in MIDlet constructor
            if (installer.execute(suiteInfo.storageName,
                                  suiteInfo.displayName)) {
                /*
                 * Give the new MIDlet the screen by destroy our self,
                 * because we are running in a limited VM and must
                 * restart the VM let the select suite run.
                 */
                destroyApp(false);
                notifyDestroyed();
            } else {
                // Give the new MIDlet the screen by pausing and
                // asking to be resumed.
                notifyPaused();
                resumeRequest();
            }
        } catch (Exception ex) {
            StringBuffer sb = new StringBuffer()
                .append(suiteInfo.displayName)
                .append("\n")
                .append(Resource.getString("Error"))
                .append(ex.toString());

            Alert a = new Alert(Resource.getString("Cannot start: "), 
                                sb.toString(), null, AlertType.ERROR);
            a.setTimeout(Alert.FOREVER);
            display.setCurrent(a, mlist);
        }
    }

    /**
     * Display the information for a suite.
     *
     * @param suiteInfo information for suite to display
     */
    private void displaySuiteInfo(MIDletSuiteInfo suiteInfo) {
        Form infoForm;
        String name;
        Image icon;

        try {
            infoForm = new Form(null);

            if (suiteInfo.singleMidlet) {
                name = suiteInfo.displayName;
                icon = suiteInfo.icon;
            } else {
                name = suiteInfo.midletSuite.getProperty(
                          Installer.SUITE_NAME_PROP);
                icon = getSuiteIcon();
            }

            infoForm.setTitle(Resource.getString("Info") + ": " + name);

            infoForm.append(
                new ImageItem(null, icon, ImageItem.LAYOUT_NEWLINE_BEFORE +
                              ImageItem.LAYOUT_CENTER +
                              ImageItem.LAYOUT_NEWLINE_AFTER, null));

            // round up the size to a Kilobyte
            infoForm.append(Resource.getString("Size"));
            infoForm.append(": ");
            infoForm.append(
                Integer.toString((suiteInfo.midletSuite.getStorageUsed() +
                    1023) / 1024));
            infoForm.append(" K");
            infoForm.append("\n");

            infoForm.append(Resource.getString("Version"));
            infoForm.append(": ");
            infoForm.append(
                suiteInfo.midletSuite.getProperty(Installer.VERSION_PROP));
            infoForm.append("\n");

            infoForm.append(Resource.getString("Vendor"));
            infoForm.append(": ");
            infoForm.append(
                suiteInfo.midletSuite.getProperty(Installer.VENDOR_PROP));
            infoForm.append("\n");

            if (!suiteInfo.singleMidlet) {
                infoForm.append(Resource.getString("Contents"));
                infoForm.append(": ");
                infoForm.append("\n");
                appendMIDletsToForm(suiteInfo.midletSuite, infoForm);
            }

            infoForm.append(Resource.getString("Website"));
            infoForm.append(": ");
            infoForm.append(suiteInfo.midletSuite.getJadUrl());
        } catch (Exception ex) {
            StringBuffer sb = new StringBuffer()
                .append(suiteInfo.displayName)
                .append("\n")
                .append(Resource.getString("Exception"))
                .append(": ")
                .append(ex.toString());

            Alert a = new Alert(Resource.getString("Cannot access: "), 
                                sb.toString(), null, AlertType.ERROR);
            a.setTimeout(Alert.FOREVER);
            display.setCurrent(a, mlist);
            return;
        }

        infoForm.addCommand(backCmd);
        infoForm.setCommandListener(this);
        display.setCurrent(infoForm);
    }

    /**
     * Confirm the removal of a suite.
     *
     * @param suiteInfo information for suite to remove
     */
    private void confirmRemove(MIDletSuiteInfo suiteInfo) {
        Form confirmForm;
        String extraConfirmMsg;

        try {
            confirmForm = new Form(null);

            confirmForm.setTitle(Resource.getString("Confirmation"));

            confirmForm.append(Resource.getString(
                "Are you sure you want to remove "));
            confirmForm.append("\"");

            if (suiteInfo.singleMidlet) {
                confirmForm.append(suiteInfo.displayName);
            } else {
                confirmForm.append(suiteInfo.midletSuite.getProperty(
                                   Installer.SUITE_NAME_PROP));
            }

            confirmForm.append("\"?");
            confirmForm.append("\n");
            confirmForm.append("\n");

            extraConfirmMsg = 
                suiteInfo.midletSuite.getProperty("MIDlet-delete-confirm");
            if (extraConfirmMsg != null) {
                confirmForm.append(extraConfirmMsg);
                confirmForm.append("\n");
                confirmForm.append("\n");
            }

            if (!suiteInfo.singleMidlet) {
                confirmForm.append(Resource.getString("This suite contains"));
                confirmForm.append(": ");
                confirmForm.append("\n");
                appendMIDletsToForm(suiteInfo.midletSuite, confirmForm);
                confirmForm.append("\n");
            }

            confirmForm.append(Resource.getString(
                "Removing will erase all data and cannot be undone."));
        } catch (Exception ex) {
            StringBuffer sb = new StringBuffer()
                .append(suiteInfo.displayName)
                .append("\n")
                .append(Resource.getString("Exception"))
                .append(": ")
                .append(ex.toString());

            Alert a = new Alert(Resource.getString("Cannot access: "), 
                                sb.toString(), null, AlertType.ERROR);
            a.setTimeout(Alert.FOREVER);
            display.setCurrent(a, mlist);
            return;
        }

        confirmForm.addCommand(cancelCmd);
        confirmForm.addCommand(removeOkCmd);
        confirmForm.setCommandListener(this);
        display.setCurrent(confirmForm);
    }

    /**
     * Remove a suite.
     *
     * @param suiteInfo information for suite to remove
     */
    private void removeSuite(MIDletSuiteInfo suiteInfo) {
        installer.remove(suiteInfo.storageName);
        destroyApp(false);
        notifyDestroyed();
    }

    /**
     * Update a suite.
     *
     * @param suiteInfo information for suite to update
     */
    private void updateSuite(MIDletSuiteInfo suiteInfo) {
        String name;

        if (suiteInfo.singleMidlet) {
            name = suiteInfo.displayName;
        } else {
            name = 
                suiteInfo.midletSuite.getProperty(Installer.SUITE_NAME_PROP);
        }

        cancelledMessage = Resource.getString("Update cancelled.");
        finishingMessage = Resource.getString("Finishing update.");
        installSuiteCommon("Updating", name,
            suiteInfo.midletSuite.getJadUrl(),
            name + Resource.getString(" was successfully updated"),
            true);
    }

    /**
     * Ask the user for the URL.
     */
    private void getUrl() {
        try {
            if (urlTextBox == null) {
                urlTextBox = new TextBox(Resource.getString(
                                         "Enter a website to Install From:"),
                                         defaultInstallListUrl, 1024,
                                         TextField.ANY);
                urlTextBox.addCommand(backCmd);
                urlTextBox.addCommand(selectSuiteCmd);
                urlTextBox.setCommandListener(this);
            }

            display.setCurrent(urlTextBox);
        } catch (Exception ex) {
            Alert a = new Alert(Resource.getString("Exception"), 
                                ex.toString(), null, AlertType.ERROR);
            a.setTimeout(Alert.FOREVER);
            display.setCurrent(a, mlist);
        }
    }

    /**
     * Ask the user for the settings.
     */
    private void getSettings() {
        if (settingsTextBox == null) {
            settingsTextBox = new TextBox(Resource.getString(
                                 "Set Website URL"), null,
                                 1024, TextField.ANY);
            settingsTextBox.addCommand(backCmd);
            settingsTextBox.addCommand(saveSettingsCmd);
            settingsTextBox.setCommandListener(this);
        }

        settingsTextBox.setString(defaultInstallListUrl);
        display.setCurrent(settingsTextBox);
    }

    /**
     * Save the settings the user entered.
     */
    private void saveSettings() {
        String temp;
        ByteArrayOutputStream bas;
        DataOutputStream dos;
        byte[] data;
        RecordEnumeration records;
        RecordStore settings;
        
        temp = settingsTextBox.getString();
        if (temp == null) {
            display.setCurrent(mlist);
            return;
        }

        try {
            bas = new ByteArrayOutputStream();
            dos = new DataOutputStream(bas);
            dos.writeUTF(temp);
            data = bas.toByteArray();
            settings = RecordStore.openRecordStore(SETTINGS_STORE, true);
            records = settings.enumerateRecords(null, null, false);
            if (records.numRecords() == 0) {
                settings.addRecord(data, 0, data.length);
            } else {
                settings.setRecord(records.nextRecordId(), data, 0,
                                   data.length);
            }

            settings.closeRecordStore();

            defaultInstallListUrl = temp;
            if (urlTextBox != null) {
                urlTextBox.setString(defaultInstallListUrl);
            }

            displaySuccessMessage(Resource.getString("Saved!"));
        } catch (Exception ex) {
            Alert a = new Alert(Resource.getString("Exception"), 
                                ex.toString(), null, AlertType.ERROR);
            a.setTimeout(Alert.FOREVER);
            display.setCurrent(a, mlist);
        }
    }

    /**
     * Get the settings the Manager saved for the user.
     */
    private void restoreSettings() {
        ByteArrayInputStream bas;
        DataInputStream dis;
        byte[] data;
        RecordEnumeration records;
        RecordStore settings = null;
        
        try {
            settings = RecordStore.openRecordStore(SETTINGS_STORE, false);
            records = settings.enumerateRecords(null, null, false);

            if (records.numRecords() == 0) {
                return;
            }

            data = records.nextRecord();
            bas = new ByteArrayInputStream(data);
            dis = new DataInputStream(bas);
            
            defaultInstallListUrl = dis.readUTF();
            return;
        } catch (RecordStoreException e) {
            // ignore
        } catch (IOException e) {
            // ignore
        } finally {
            if (settings != null) {
                try {
                    settings.closeRecordStore();
                } catch (RecordStoreException e) {
                    // ignore
                }
            }
        }
    }

    /**
     * Let the user select a suite to install. The suites that are listed
     * are the links on a web page that end with .jad.
     *
     * @param url where to get the list of suites to install.
     */
    private void selectSuiteToInstall(String url) {
        new Thread(new BackgroundInstallListGetter(this, url)).start();
    }

    /**
     * Install a suite.
     *
     * @param selectedSuite index into the installList
     */
    private void installSuite(int selectedSuite) {
        SuiteDownloadInfo suite;

        suite = (SuiteDownloadInfo)installList.elementAt(selectedSuite);
        cancelledMessage = Resource.getString("Installation cancelled.");
        finishingMessage = Resource.getString("Finishing installation.");
        installSuiteCommon("Installing", suite.label, suite.url,
            suite.label + Resource.getString(" was successfully installed"),
            false);
    }

    /**
     * Install a suite.
     *
     * @param action action to put in the form's title
     * @param name name to in the form's title
     * @param url URL of a JAD
     * @param successMessage message to display to user upon success
     * @param overwriteFlag if true the current suite should be
     *                      overwritten without asking the user.
     */
    private void installSuiteCommon(String action, String name, String url,
            String successMessage, boolean overwriteFlag) {
        Form installForm;

        try {
            installForm = displayConnectingForm(action, name, url);
            installForm.addCommand(stopCmd);
            installForm.setCommandListener(this);
            backgroundInstaller = new BackgroundInstaller(this, url, name,
                                      successMessage, overwriteFlag);
            new Thread(backgroundInstaller).start();
        } catch (Exception ex) {
            StringBuffer sb = new StringBuffer()
                .append(name)
                .append("\n")
                .append(Resource.getString("Error"))
                .append(": ")
                .append(ex.toString());

            Alert a = new Alert(Resource.getString("Cannot access: "), 
                                sb.toString(), null, AlertType.ERROR);
            a.setTimeout(Alert.FOREVER);
            display.setCurrent(a, mlist);
            return;
        }
    }

    /**
     * Ask a username and password.
     */
    private void getUsernameAndPassword() {
        if (passwordForm == null) {
            passwordForm = new Form(null);

            usernameField = new TextField(
                            Resource.getString("Please Enter ID"), null, 40,
                            TextField.ANY);
            passwordForm.append(usernameField);

            passwordField = new TextField(
                            Resource.getString("Password"), null, 40,
                            TextField.PASSWORD);
            passwordForm.append(passwordField);
            passwordForm.addCommand(cancelInstallCmd);
            passwordForm.addCommand(nextCmd);
            passwordForm.setCommandListener(this);
        }

        passwordField.setString("");
        display.setCurrent(passwordForm);
    }

    /**
     * Resume the install of the suite with a password and username.
     */
    private void resumeInstallWithPassword() {
        String username;
        String password;


        username = usernameField.getString();
        password = passwordField.getString();
        if (username == null || username.length() == 0) {
            Alert a = new Alert(Resource.getString("Error"), 
                                Resource.getString(
                                "The ID has not been entered."),
                                null, AlertType.ERROR);
            a.setTimeout(ALERT_TIMEOUT);
            display.setCurrent(a, passwordForm);
            return;
        }

        if (password == null || password.length() == 0) {
            Alert a = new Alert(Resource.getString("Error"), 
                                Resource.getString(
                                "The password has not been entered."),
                                null, AlertType.ERROR);
            a.setTimeout(ALERT_TIMEOUT);
            display.setCurrent(a, passwordForm);
            return;
        }

        // redisplay the progress form
        display.setCurrent(conForm);

        backgroundInstaller.installState.setUsername(username);
        backgroundInstaller.installState.setPassword(password);

        backgroundInstaller.continueInstall = true;
        backgroundInstaller.installState.notify();
    }

    /**
     * Give the user a chance to act on warning during an installation.
     *
     * @param name name of the MIDlet suite to insert into the message
     * @param jadUrl URL of a JAD, can be null
     * @param e last exception from the installer
     */
    private void warnUser(String name, String jadUrl, InvalidJadException e) {
        Form warningForm;
        String extraConfirmMsg;

        warningForm = new Form(null);
        warningForm.setTitle(Resource.getString("Warning"));
        warningForm.append(translateJadException(e, name, jadUrl)); 
        warningForm.addCommand(cancelInstallCmd);
        warningForm.addCommand(okCmd);
        warningForm.setCommandListener(this);
        display.setCurrent(warningForm);
    }

    /**
     * Update the status form.
     *
     * @param status current status of the install.
     */
    private void updateStatus(int status) {
        if (status != Installer.STORING_SUITE) {
            return;
        }

        displaySuccessMessage(finishingMessage);
    }

    /**
     * Resume the install after a the user overrides a warning.
     */
    private void resumeInstallAfterWarning() {
        // redisplay the progress form
        display.setCurrent(conForm);

        backgroundInstaller.continueInstall = true;
        backgroundInstaller.installState.notify();
    }

    /** Cancel an install (if there is one) waiting for user input. */
    private void cancelBackgroundInstall() {
        if (backgroundInstaller != null &&
                backgroundInstaller.installState != null) {
            backgroundInstaller.installState.notify();
        }
    }

    /**
     * Display the connecting form to the user.
     *
     * @param action action to put in the form's title
     * @param name name to in the form's title
     * @param url URL of a JAD
     *
     * @return displayed form
     */
    private Form displayConnectingForm(String action, String name,
                                       String url) {
        Image icon;

        conForm = new Form(null);

        conForm.setTitle(Resource.getString(action) + " " + name);

        icon = getIconFromStorage(
                                  (colorDisplay ? "_dt_8.png" : "_dt_2.png"));
        conForm.append(new ImageItem(null, icon, ImageItem.LAYOUT_CENTER +
                                     ImageItem.LAYOUT_NEWLINE_BEFORE +
                                     ImageItem.LAYOUT_NEWLINE_AFTER, null));

        conForm.append(action);
        conForm.append(Resource.getString(" from"));
        conForm.append(": ");
        conForm.append(url);

        display.setCurrent(conForm);

        return conForm;
    }

    /**
     * Alert the user that an action was successful.
     * @param successMessage message to display to user
     */
    private void displaySuccessMessage(String successMessage) {
        Image icon;
        Alert successAlert;

        icon = getIconFromStorage(
                       (colorDisplay ? "_dukeok8.png" : "_dukeok2.png"));

        successAlert = new Alert(null, successMessage, icon, null);

        successAlert.setTimeout(ALERT_TIMEOUT);

        // We need to prevent "flashing" on fast development platforms.
        while (System.currentTimeMillis() - lastDisplayChange < ALERT_TIMEOUT);

        lastDisplayChange = System.currentTimeMillis();

        display.setCurrent(successAlert, mlist);
    }

    /**
     * Alert the user that an action was canceled. The backgroundInstaller
     * will hide the message.
     * @param message message to display to user
     */
    private void displayCancelledMessage(String message) {
        Form form;
        Image icon;

        form = new Form(null);

        icon = getIconFromStorage(
                       (colorDisplay ? "_ack_8.png" : "_ack_2.png"));
        form.append(new ImageItem(null, icon, ImageItem.LAYOUT_CENTER +
                                     ImageItem.LAYOUT_NEWLINE_BEFORE +
                                     ImageItem.LAYOUT_NEWLINE_AFTER, null));

        form.append(message);

        display.setCurrent(form);
    }

    /**
     * Appends a names of all the MIDlets in a suite to a Form, one per line.
     *
     * @param midletSuite suite of MIDlets
     * @param form form to append to
     */
    private void appendMIDletsToForm(MIDletSuite midletSuite, Form form) {
        int numberOfMidlets;
        MIDletInfo midletInfo;

        numberOfMidlets = midletSuite.getNumberOfMIDlets();
        for (int i = 1; i <= numberOfMidlets; i++) {
            midletInfo = new MIDletInfo(
                             midletSuite.getProperty("MIDlet-" + i));
            form.append(midletInfo.name);
            form.append("\n");
        }
    }

    /**
     * Read the set of MIDlet names, icons and classes
     * Fill in the list.
     */
    private void setupList() {
        if (mlist == null) {
            mlist = new List(Resource.getString("Main Screen"), 
                             Choice.IMPLICIT);

            // Add each midlet
            for (int i = 0; i < mcount; i++) {
                mlist.append(" " + minfo[i].displayName, minfo[i].icon);
            }
        }
    }
        
    /**
     * Read in and create a MIDletInfo for each MIDlet suite.
     */
    private void readMIDletSuiteInfo() {
        String[] suiteNames;
        MIDletSuite midletSuite;
        int i;
        int numberOfMidlets;
        String nth;
        String attr;
        byte[] rawPng;

        suiteNames = installer.list();

        for (i = 0; i < suiteNames.length; i++) {
            try {
                midletSuite = installer.getMIDletSuite(suiteNames[i]);

                numberOfMidlets = midletSuite.getNumberOfMIDlets();

                if (numberOfMidlets > 1) {
                    addMIDletSuite(new MIDletSuiteInfo(suiteNames[i],
                                                       midletSuite));
                }

                for (int j = 1; j <= numberOfMidlets; j++) {
                    nth = "MIDlet-"+ j;
                    attr = midletSuite.getProperty(nth);
                    if (attr == null || attr.length() == 0) {
                        break;
                    }
                    
                    addMIDletSuite(new MIDletSuiteInfo(suiteNames[i],
                                                       midletSuite, attr));
                }

                if (numberOfMidlets == 1) {
                    minfo[mcount - 1].singleMidlet = true;
                    minfo[mcount - 1].icon = getSingleSuiteIcon();
                }
            } catch (Exception e) {
                // move on to the next suite
            }
        }
    }

    /**
     * Add a MIDlet suite to the list.
     * @param info MIDlet suite information to add to MIDlet
     */
    private void addMIDletSuite(MIDletSuiteInfo info) {
        if (mcount >= minfo.length) {
            MIDletSuiteInfo[] n = new MIDletSuiteInfo[mcount+4];
            System.arraycopy(minfo, 0, n, 0, mcount);
            minfo = n;
        }

        minfo[mcount++] = info;
    }

    /**
     * Simple attribute storage for MIDlet suites
     */
    private class MIDletSuiteInfo {
        /** Midlet suite interface for displaying its info. */
        MIDletSuite midletSuite;
        /** Storage name of the MIDlet suite. */
        String storageName;
        /** Display name of the MIDlet suite. */
        String displayName;
        /** Name of the MIDlet to run. */
        String midletToRun;
        /** Icon for this suite. */
        Image icon;
        /** Icon name for the first midlet. */
        String iconName;
        /** Is this single MIDlet MIDlet suite. */
        boolean singleMidlet = false;

        /**
         * Constructs a MIDletSuiteInfo object for a suite
         *
         * @param theStorageName name the installer has for this suite
         * @param theMidletSuite MIDletSuite object for this suite
         */
        MIDletSuiteInfo(String theStorageName, MIDletSuite theMidletSuite) {
            midletSuite = theMidletSuite;
            displayName = midletSuite.getProperty(Installer.SUITE_NAME_PROP);
            if (displayName == null) {
                displayName = theStorageName;
            }

            icon = getSuiteIcon();
            storageName = theStorageName;
        }

        /**
         * Constructs a MIDletSuiteInfo object for a MIDlet of a suite
         *
         * @param theStorageName name the installer has for this suite
         * @param theMidletSuite MIDletSuite object for this suite
         * @param attr an MIDlet-<n> value for the MIDlet
         */
        MIDletSuiteInfo(String theStorageName, MIDletSuite theMidletSuite,
                        String attr) {
            MIDletInfo midletInfo = new MIDletInfo(attr);

            midletSuite = theMidletSuite;

            displayName = midletInfo.name;
            icon = getEmptyIcon();

            iconName = midletInfo.icon;
            storageName = theStorageName;
        }
    }

    /** A class to install a suite in a background thread. */
    private class BackgroundInstaller implements Runnable, InstallListener {
        /** Parent manager. */
        private Manager parent;
        /** JAD URL to install from. */
        private String jadUrl;
        /** Name of MIDlet suite. */
        private String name;
        /**
         * Message for the user after the current install completes
         * successfully.
         */
        private String successMessage;
        /** Flag to overwrite the current suite. */
        private boolean overwrite;
        /** State of the install. */
        InstallState installState;
        /** Signals that the user wants the install to continue. */
        boolean continueInstall;

        /**
         * Construct a BackgroundInstaller.
         *
         * @param theParent parent Mangager of this object
         * @param theJadUrl where to get the JAD.
         * @param theName name of the MIDlet suite
         * @param theSuccessMessage message to display to user upon success
         * @param overwriteFlag if true the current suite should be
         *                      overwritten without asking the user.
         */
        private BackgroundInstaller(Manager theParent, String theJadUrl,
                String theName, String theSuccessMessage,
                boolean overwriteFlag) {
            parent = theParent;
            jadUrl = theJadUrl;
            name = theName;
            successMessage = theSuccessMessage;
            overwrite = overwriteFlag;
        }

        /**
         * Run the installer.
         */
        public void run() {
            try {
                parent.installer.install(jadUrl, overwrite, this);
                parent.displaySuccessMessage(successMessage);

                /*
                 * We need to prevent "flashing" on fast development
                 * platforms.
                 */
                while ((System.currentTimeMillis() -
                        parent.lastDisplayChange) < parent.ALERT_TIMEOUT);

                parent.destroyApp(false);
                parent.notifyDestroyed();
                return;
            } catch (Throwable ex) {
                String msg;

                if (parent.installer != null &&
                        parent.installer.wasStopped()) {

                    /*
                     * We need to prevent "flashing" on fast development
                     * platforms.
                     */
                    while ((System.currentTimeMillis() -
                        parent.lastDisplayChange) < parent.ALERT_TIMEOUT);

                    parent.display.setCurrent(parent.mlist);
                    return;
                }

                if (ex instanceof InvalidJadException) {
                    msg = translateJadException((InvalidJadException)ex,
                                                name, jadUrl);
                } else if (ex instanceof IOException) {
                    msg = Resource.getString(
                          "The connection dropped and the installation " +
                          "did not complete. Please try installing again.");
                } else {
                    msg = ex.getMessage();
                }
            
                Alert a = new Alert(Resource.getString("Install Error"), 
                                    msg, null, AlertType.ERROR);
                a.setTimeout(Alert.FOREVER);
                parent.display.setCurrent(a, parent.mlist);
            }
        }

        /**
         * Called with the current state of the install so the user can be
         * asked to override the warning. Calls the parent to display the
         * warning to the user and then waits on the state object for
         * user's response.
         *
         * @param state current state of the install.
         *
         * @return true if the user wants to continue,
         *         false to stop the install
         */
        public boolean warnUser(InstallState state) {
            boolean temp;

            installState = state;

            InvalidJadException e = installState.getLastException();

            if (e.getReason() == InvalidJadException.UNAUTHORIZED) {
                parent.getUsernameAndPassword();
            } else {
                parent.warnUser(name, jadUrl, e);
            }

            synchronized (installState) {
                try {
                    installState.wait();
                } catch (InterruptedException ie) {
                    // ignore
                }
            }

            installState = null;
            temp = continueInstall;
            continueInstall = false;
            return temp;
        }

        /**
         * Called with the current status of the install.
         * Changes the status alert box text based on the status.
         *
         * @param status current status of the install.
         */
        public void updateStatus(int status) {
            parent.updateStatus(status);
        }
    }

    /** A class to get the install list in a background thread. */
    private class BackgroundInstallListGetter implements Runnable {
        /** Parent manager. */
        private Manager parent;
        /** URL of the list. */
        private String url;

        /**
         * Construct a BackgroundInstallListGetter.
         *
         * @param theParent parent Mangager of this object
         * @param theUrl where to get the list of suites to install.
         */
        private BackgroundInstallListGetter(Manager theParent,
                                            String theUrl) {
            parent = theParent;
            url = theUrl;
        }

        /**
         * Get the list of suites for the user to install.
         * The suites that are listed
         * are the links on a web page that end with .jad.
         */
        public void run() {
            StreamConnection conn = null;
            InputStreamReader in = null;
            String errorMessage;
            long startTime;

            startTime = System.currentTimeMillis();

            try {
                conn = (StreamConnection)Connector.open(url, Connector.READ);
                in = new InputStreamReader(conn.openInputStream());
                try {
                    displayConnectingForm("Getting Install List", "", url);
                    parent.installList =
                        SuiteDownloadInfo.getDownloadInfoFromPage(in);
                    
                    if (parent.installList.size() > 0) {
                        parent.installListBox = new List(Resource.getString(
                                                "Select one to install:"),
                                                Choice.IMPLICIT);

                        // Add each suite
                        for (int i = 0; i < parent.installList.size(); i++) {
                            SuiteDownloadInfo suite =
                                (SuiteDownloadInfo)installList.elementAt(i);
                            parent.installListBox.append(suite.label,
                                                         (Image)null);
                        }

                        parent.installListBox.addCommand(backCmd);
                        parent.installListBox.addCommand(installCmd);
                        parent.installListBox.setCommandListener(parent);

                        /*
                         * We need to prevent "flashing" on fast development
                         * platforms. 
                         */
                        while (System.currentTimeMillis() - startTime < 
                            parent.ALERT_TIMEOUT);

                        display.setCurrent(parent.installListBox);
                        return;
                    }

                    errorMessage = Resource.getString(
                        "No MIDlet Suites found. " +
                        "Check the URL to make sure it is correct.");
                } catch (Exception ex) {
                    errorMessage = ex.getMessage();
                }
            } catch (Exception ex) {
                errorMessage = Resource.getString(
                    "The connection failed. Please check the website " +
                    "URL and try again.");
            } finally {
                try {
                    conn.close();
                    in.close();
                } catch (Exception e) {
                    // ignore
                }
            }

            Alert a = new Alert(Resource.getString("Error"), 
                                errorMessage, null, AlertType.ERROR);
            a.setTimeout(Alert.FOREVER);
            display.setCurrent(a, parent.mlist);
        }
    }
}
