/*
 * @(#)Protocol.java	1.72 01/08/29
 * Copyright (c)s 1999,2001 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

package com.sun.midp.io.j2me.http;

/**
 * A class representing a http connection. An http connection consists of
 * stream connection as well as input and output streams for read/write data to
 * and from a web server. This version supports HTTP1.1 persistent connections
 * allowing connects to be shared from a connection pool. This pool and the
 * maximum number of connections can be configured for a particular platform.
 * Proxy connections are also allowed through this interface.
 *
 * <p> The actual connection to the web server does not take place until the
 * application needs an, (1) input stream, (2) flush data, (3)request some 
 * header info or closes the connection (with outstanding data in the 
 * output stream). Because of this issue the state transition must allow for
 * some flexability to move backwards for WRITE state conditions.
 *
 * <p> Persistent connections are provided through the use of a connection
 * pool that tracks the connect status. There are maximum threshold values
 * defined and these values can be overiden using property key/value pars.
 * The connection pool provides a synchronized interface for managing the
 * maximum configurable connections. Persistent connections will only be 
 * supported for HTTP1.1 connections - otherwise the connections will be
 * closed and disgarded after its done (HTTP1.0 behavior).
 *
 * <p> This class extends the ConnectionBaseAdapter where Connector type
 * objects (like this) use various features. Output and Input streams are
 * created and managed in the adapter class.
 *
 * <p> The reading and writing of data through the input and output streams
 * are configured, buffered and managed depending on the ability of a 
 * platform to read/write on those streams. 
 *
 */

import java.io.IOException;
import java.io.InputStream;
import java.io.InterruptedIOException;
import java.io.OutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.ByteArrayOutputStream;

import java.util.Enumeration;
import java.util.Hashtable;

import javax.microedition.io.StreamConnection;
import javax.microedition.io.HttpConnection;
import javax.microedition.io.Connector;
import javax.microedition.io.Connection;
import javax.microedition.io.ConnectionNotFoundException;

import com.sun.midp.io.InternalConnector;
import com.sun.midp.io.HttpUrl;

import com.sun.cldc.io.DateParser;

import com.sun.midp.io.j2me.http.StreamConnectionPool;
import com.sun.midp.io.j2me.http.StreamConnectionElement;
import com.sun.midp.Configuration;
import com.sun.midp.io.ConnectionBaseAdapter;
import com.sun.midp.io.Properties;

/**
 * This class implements the necessary functionality
 * for an HTTP connection.
 *
 */
public class Protocol extends ConnectionBaseAdapter 
    implements HttpConnection  {

    /** Saved copy of URL string to be processed. */
    protected String saved_url;
    /** Protocol for this class or subclass. */
    protected String protocol;
    /** Default port number for this protocol. */
    protected int default_port;
    /** Parsed Url. */
    protected HttpUrl url;
    /** url.host + ":" + url.port. */
    protected String hostAndPort;

    /** Numeric code returned from HTTP response header. */
    protected int responseCode;
    /** Message string from HTTP response header. */
    protected String responseMsg;

    /** Collection of request headers as name/value pairs. */
    protected Properties reqProperties;

    /** Collection of response headers as name/value pairs. */
    protected Properties headerFields;

    /** HTTP method type for the current request. */
    protected String method;
    
    /** 
     * Type of read or write connection desired from the generic connection 
     * framework.
     */
    protected int mode;
    
    /** Reconnect int variable to limit the number of interations. */
    protected int reconnects;

    /** Maximum number of reconnect tries for a session. */
    private static final int MAX_NUMBER_OF_RECONNECTS = 2;

    /*
     * In/Out Streams used to buffer input and output.
     */
    /** Input stream associated with the current HTTP connection. */
    private InputStream in;
    /** Output stream associated with the current HTTP connection. */
    private OutputStream out;

    /*
     * The streams from the underlying socket connection.
     */
    /** Low level socket connection used for the HTTP requests. */
    private StreamConnection streamConnection;
    /** Low level socket output stream. */
    private DataOutputStream streamOutput;
    /** Low level socket input stream. */
    private DataInputStream streamInput;

    /**
     * A shared temporary buffer. 
     */
    private StringBuffer stringbuffer;

    /** The "host:port" value to use for HTTP proxied requests. */
    private String http_proxy;

    /** HTTP version string to use with all outgoing HTTP requests. */
    protected static final String HTTP_VERSION = "HTTP/1.1";

    /** Persistent connection poo.l */
    private static StreamConnectionPool connectionPool; 
    /** Stream connection container. */
    private StreamConnectionElement connectionElement = null;
    /** HTTP version string set with all incoming HTTP responses. */
    private String httpVer = null;
    /** Used when appl calls setRequestProperty("Connection", "close"). */
    private boolean ConnectionCloseFlag;
    /**
     * Total number of bytes in the current chunk or content-length when
     * data is sent as one big chunk.
     */
    private int chunksize;
    /**
     * Number of bytes read from the stream for non-chunked data or
     * the bytes read from the current chunk.
     */ 
    private int totalbytesread;
    /** True if Transfer-Encoding: chunkedIn. */
    private boolean chunkedIn;	   
    /** True if Transfer-Encoding: chunkedOut. */
    private boolean chunkedOut;	   
    /** True if eof seen. */
    private boolean eof;	   
    /** Internal stream buffer to minimize the number of TCP socket reads. */
    private byte[] readbuf;         
    /** Number of bytes left in internal input stream buffer. */
    private int bytesleft;
    /** Number of bytes read from the internal input stream buffer. */ 
    private int bytesread;     
    /** Buffered data output for content length calculation. */
    private ByteArrayOutputStream byteArrayOutput;
    /** Maximum number of persistent connections. */
    private int maxNumberOfPersistentConnections = 1;
    /** Default size for input buffer. */     
    private final int HTTP_BUFFER_SIZE = 256;
    /** Default size for output buffer. */
    private final int HTTP_OUTPUT_BUFFER_SIZE = 2048;

    /** Initial state for http state transition. */
    private static final int INIT_STATE    = 0;
    /** Write state for http state transition. */
    private static final int WRITE_STATE   = 1;
    /** Open state for http state transition. */
    private static final int OPEN_STATE    = 2;
    /** Chunk state for http state transition. */
    private static final int CHUNK_STATE   = 3;
    /** Read state for http state transition. */
    private static final int READ_STATE    = 4;
    /** Close state for http state transition. */
    private static final int CLOSE_STATE   = 5;
    /** Close state for http state transition. */
    private static final int FLUSH_STATE   = 6;
    /** Done state for http state transition. */
    private static final int CONNECT_STATE = 7;
    /** The http current state flag. */
    private int state = INIT_STATE;

    /** Collection of "Proxy-" headers as name/value pairs. */
    private Properties proxyHeaders = new Properties();
    /** Last handshake error. */
    private byte handshakeError;
    

    /**
     * Create a new instance of this class and intialize variables.
     * Initially an http connection is unconnected to the network.
     */
    public Protocol() {

        reqProperties = new Properties();
        headerFields = new Properties();
	stringbuffer = new StringBuffer(32);

        method = GET;
        responseCode = -1;
        protocol = "http";
        default_port = 80;
        maxNumberOfPersistentConnections = 1;

        /*
         * Initialize state
         */
        state = INIT_STATE;

        /*
         * bug#4455443 - allows for configuration options to shut off 
         * the persistent connection feature for http 
         */
        String flag = 
            Configuration.getProperty(
                                "com.sun.midp.io.http.force_non_persistent");
        if ((flag != null) && (flag.equals("true"))) {
            ConnectionCloseFlag = true;
        }

        /*
         * Get the  maximum number of persistent connections
         * from the configuration file.
         */
         String num = 
         Configuration.getProperty(
                            "com.sun.midp.io.http.max_persistent_connections");
         if (num != null) {
             try {
                 maxNumberOfPersistentConnections = Integer.parseInt(num);
             } catch (NumberFormatException nfe) { }
             if (maxNumberOfPersistentConnections <= 0) {
                 maxNumberOfPersistentConnections = 1;
             }
         }

         readbuf = new byte[HTTP_BUFFER_SIZE];

    }
    /**
     * Provides the connect() method that sets up the connection, but
     * does not actually connect to the server until there's something
     * to do.
     *
     * @param name             The URL for the connection, without the
     *                         without the protcol part.
     * @param mode             The access mode
     * @param timeouts         A flag to indicate that the called wants
     *                         timeout exceptions
     *
     * @exception IllegalArgumentException If a parameter is invalid.
     * @exception ConnectionNotFoundException If the connection cannot be
     *             found.
     * @exception IOException  If some other kind of I/O error occurs.
     */
    protected void connect(String name, int mode, boolean timeouts)
        throws IOException, IllegalArgumentException,
	ConnectionNotFoundException {
                
        if (mode != Connector.READ && mode != Connector.WRITE
            && mode != Connector.READ_WRITE) {
            throw new IllegalArgumentException("illegal mode: " + mode);
        }

        this.saved_url = name;
        this.mode = mode;

        url = new HttpUrl(protocol, name);

        if (url.port == -1) {
            url.port = default_port;
        }

        if (url.host == null) {
            // avoid NullPointerExceptions
            url.host = "";
        }

        hostAndPort = url.host + ":" + url.port;
    }
    
    /**
     * Open the input stream if it has not already been opened.
     *
     * @exception IOException is thrown if it has already been opened.
     * @return input stream for the current connection
     */
    public InputStream openInputStream() throws IOException {
        
        if (in != null) {
	    throw new IOException("input stream already open");
	}
        
	/* Check that the connection was opened for reading */
        if (mode != Connector.READ && mode != Connector.READ_WRITE) {
            throw new IOException("write-only connection");
        }

        /*
         * Send a request to the web server if there wasn't one
         * sent already - now moving to READ state
         */
        sendRequest(READ_STATE);

        /*
         * Initialize and set the current input stream variables
         */
        bytesleft = 0;
        chunksize = 0;
        bytesread = 0;
        totalbytesread = 0;
        chunkedIn = false;
        eof = false;
        /*
         * Determine if this is a chunked datatransfer and setup
         */
        String te = (String)getHeaderField("transfer-encoding");
        if (te != null && te.equals("chunked")) {
            chunkedIn = true;
            chunksize = readChunkSize();
        } else {
            // treat non chunked data of known length as one big chunk
            chunksize = (int)getLength();
        }

        if (chunksize == 0) {
            eof = true;
        }

        /*
         * Call into parent to create input stream passed back to the user
         */
        in = super.openInputStream();
	
        return in;
    }
       
    /**
     * Open the output stream if it has not already been opened.
     *
     * @exception IOException is thrown if it has already been opened.
     * @return output stream for the current connection
     */
    public OutputStream openOutputStream() throws IOException {

        if (mode != Connector.WRITE && mode != Connector.READ_WRITE) {
            throw new IOException("read-only connection");
        }

        /*
         * Create a byte array output stream for output buffering
         * once the user calls flush() this gets written to stream
         */
        byteArrayOutput = new ByteArrayOutputStream(HTTP_OUTPUT_BUFFER_SIZE);

        /*
         * call into parent to create output stream passed back to the user
         */
        out = super.openOutputStream();

        return out;
    }
    
    /**
     * Reads up to <code>len</code> bytes of data from the input stream into
     * an array of bytes.
     * This method reads NonChunked http connection input streams.
     * This method can only be called after the InputStream setup is complete.
     *
     * @param      b     the buffer into which the data is read.
     * @param      off   the start offset in array <code>b</code>
     *                   at which the data is written.
     * @param      len   the maximum number of bytes to read.
     * @return     the total number of bytes read into the buffer, or
     *             <code>-1</code> if there is no more data because the end of
     *             the stream has been reached.
     * @exception  IOException  if an I/O error occurs.
     */
    protected int readBytes(byte b[], int off, int len)
        throws IOException {

        int rc;
        
        /*
         * Be consistent about returning EOF once encountered.
         */
        if (eof) {
            return (-1);
        }
        
        /*
         * If the http connection is chunked call the readBytesChunked method
         */
        if (chunkedIn) {
            /*
             * "smooth" out (minmize the calls to this method) the chunks at
             * this level when a caller gives a large buffer. This is a minor
             * optimation, since readBytesChucked optimizes the the number of
             * native TCP socket calls. This also allow reuse of
             * readBytesChunked for unchucked data with a known content length.
             */
            int total = 0;
            do {
                rc = readBytesChunked(b, off, len);
                if (rc == -1) {
                    if (total == 0) {
                        total = -1;
                    }

                    break;
                }

                total += rc;
                off += rc;
                len -= rc;
            } while (len > 0);

            return total;
        } else if (chunksize > 0) {
            /*
             * Non-chunked data of known length is treated as one big chunk
             */

            return readBytesChunked(b, off, len);
        }

        /*
         * Non-chunked unknown length
         */

        if (bytesleft == 0) {
            /*
             * the internal input stream buffer is empty, read from the stream
             */

            if (len >= HTTP_BUFFER_SIZE) {
                /*
                 * No need to buffer, if the caller has given a big buffer.
                 */
                rc = streamInput.read(b, off, len);
            } else {
                rc = streamInput.read(readbuf, 0, HTTP_BUFFER_SIZE);
                bytesleft = rc;
                bytesread = 0;
            }

            if (rc == -1) {
                /*
                 * The next call to this method should not read.
                 */
                eof = true;
                return -1;
            }

            totalbytesread += rc;

            if (bytesleft == 0) {
                /*
                 * The data was read directly into the caller's buffer.
                 */
                return rc;
            }
        }

        rc = readFromBuffer(b, off, len);

        /*
         * Move to READ state - if there is a problem - raise an exception.
         */
        if (state != CONNECT_STATE) {
            setState(READ_STATE);
        }
        
        return rc;
    }
    
    /**
     * Reads up to <code>len</code> bytes of data from the internal buffer into
     * an array of bytes.
     *
     * @param      b     the buffer into which the data is read.
     * @param      off   the start offset in array <code>b</code>
     *                   at which the data is written.
     * @param      len   the maximum number of bytes to read.
     * @return     the total number of bytes read into the buffer, or
     *             <code>-1</code> if there is no more data because the end of
     *             the stream has been reached.
     * @exception  IOException  if an I/O error occurs.
     */
    private int readFromBuffer(byte b[], int off, int len)
        throws IOException {
        
        /*
         * copy http buffer data into user buffer, then
         * increment and decrement counters
         */
        
        int rc;

        if (len > bytesleft) {
            rc = bytesleft;
        } else {
            rc = len;            
        }

        System.arraycopy(readbuf, bytesread, b, off, rc);

        bytesleft -= rc;
        bytesread += rc;
        
        return rc;
    }

    /**
     * Reads up to <code>len</code> bytes of data from the input stream into
     * an array of bytes.
     * This method reads Chunked and known length non-chunked http connection
     * input streams. For non-chunked set the field <code>chunkedIn</code>
     * should be false.
     *
     * @param      b     the buffer into which the data is read.
     * @param      off   the start offset in array <code>b</code>
     *                   at which the data is written.
     * @param      len   the maximum number of bytes to read.
     * @return     the total number of bytes read into the buffer, or
     *             <code>-1</code> if there is no more data because the end of
     *             the stream has been reached.
     * @exception  IOException  if an I/O error occurs.
     */
    protected int readBytesChunked(byte b[], int off, int len)
        throws IOException {

        int rc;

        if (bytesleft == 0) {
            /*
             * the internal input stream buffer is empty, read from the stream
             */

            if (totalbytesread == chunksize) {
                /*
                 * read the end of the chunk and get the size of the
                 * the next if there is one
                 */

                if (!chunkedIn) {
                    /*
                     * non-chucked data is treated as one big chunk so there
                     * is no more data so just return as if there are no
                     * more chunks
                     */
                    eof = true;
                    return -1;
                }


                readCRLF();

                chunksize = readChunkSize();
                if (chunksize == 0) {
                    eof = true;

                    /*
                     * REFERENCE: HTTP1.1 document 
                     * SECTION: 3.6.1 Chunked Transfer Coding
                     * in some cases there may be an OPTIONAL trailer
                     * containing entity-header fields. since we don't support
                     * the available() method for TCP socket input streams and
                     * for performance and reuse reasons we do not attempt to
                     * clean up the current connections input stream. 
                     * check readResponseMessage() method in this class for
                     * more details
                     */
                    return -1;
                }

                /*
                 * we have not read any bytes from this new chunk
                 */
                totalbytesread = 0;
            }

            int bytesToRead = chunksize - totalbytesread;

            if (len >= bytesToRead) {
                /*
                 * No need to buffer, if the caller has given a big buffer.
                 */
                rc = streamInput.read(b, off, bytesToRead);

            } else if (len >= HTTP_BUFFER_SIZE) {
                /*
                 * No need to buffer, if the caller has given a big buffer.
                 */
                rc = streamInput.read(b, off, len);
            } else {
                if (HTTP_BUFFER_SIZE >= bytesToRead) {
                    rc = streamInput.read(readbuf, 0, bytesToRead);
                } else {
                    rc = streamInput.read(readbuf, 0, HTTP_BUFFER_SIZE);
                }

                bytesleft = rc;
                bytesread = 0;
            }

            if (rc == -1) {
                /*
                 * Network problem or the wrong length was sent by the server.
                 */
                eof = true;
                throw new IOException("unexpected end of stream");
            }

            totalbytesread += rc;

            if (bytesleft == 0) {
                /*
                 * The data was read directly into the caller's buffer.
                 */
                return rc;
            }
        }

        rc = readFromBuffer(b, off, len);

        return rc;
    }

    /** 
     * Read the chunk size from the input.
     * It is a hex length followed by optional headers (ignored).
     * and terminated with CRLF.
     *
     * @return size of the buffered read
     */
    private int readChunkSize() throws IOException {

        int size = -1;
        try {
            String chunk = null;
            try {
                chunk = readLine(streamInput);
            } catch (IOException ioe) {
                /* throw new IOException(ioe.getMessage()); */
            }
            if (chunk == null) {
                throw new IOException("No Chunk Size");
            }
            int i;
            for (i = 0; i < chunk.length(); i++) {
                char ch = chunk.charAt(i);
                if (Character.digit(ch, 16) == -1)
                    break;
            }
            /* look at extensions?.... */
            size = Integer.parseInt(chunk.substring(0, i), 16);
        } catch (NumberFormatException e) {
            throw new IOException("invalid chunk size number format");
        }
        
        return size;
    }
    
    /**
     * Read CRLF from the InputStream.
     *
     * @exception IOException is thrown if either CR or LF
     * is missing.
     */
    private void readCRLF() throws IOException { 

        int ch;
        ch = streamInput.read();
        if (ch != '\r') 
            throw new IOException("missing CRLF"); 
        ch = streamInput.read();
        if (ch != '\n')
            throw new IOException("missing CRLF");
    } 

    /**
     * Writes <code>len</code> bytes from the specified byte array
     * starting at offset <code>off</code> to this output stream.
     * <p>
     * Polling the native code is done here to allow for simple
     * asynchronous native code to be written. Not all implementations
     * work this way (they block in the native code) but the same
     * Java code works for both.
     * <p>
     * This method can only be called after an OutputStream setup has be
     * done.
     *
     * @param      b     the data.
     * @param      off   the start offset in the data.
     * @param      len   the number of bytes to write.
     * @return     int   number of bytes written to stream
     * @exception  IOException  if an I/O error occurs. In particular,
     *             an <code>IOException</code> is thrown if the output
     *             stream is closed.
     */
    protected int writeBytes(byte b[], int off, int len)
        throws IOException {
                       
        /*
         * Here's the only time the state can transition backwards -because
         * of the unusual way http connection sends a request to the server
         * (please reference the class description for more details).
         */
        if (state < OPEN_STATE) {
            setState(WRITE_STATE);
        }
        /*
         * If the length of the paramter buffer area is larger than the 
         * internal buffer area (HTTP_OUTPUT_BUFFER_SIZE) - then write
         *  the param buffer area directly to the output stream.
         */
        if (len > HTTP_OUTPUT_BUFFER_SIZE) {

            /*
             * once the output buffer size exceeds the HTTP_OUTPUT_BUFFER_SIZE
             * the session becomes a HTTP1.1 type session
             */
            chunkedOut = true;
            
            /*
             * Set the Transfer-Encoding property to "chunked".
             */
            setRequestProperty("Transfer-Encoding", "chunked");
            
            if (byteArrayOutput.size() > 0) {
                flush();
            }
            
            byteArrayOutput.reset();
            
            sendRequest(CHUNK_STATE);
            
            sendRequestWritePayloadDirect(b, off, len);
            
            return len;
        }
        if (byteArrayOutput.size()+len > (HTTP_OUTPUT_BUFFER_SIZE)) {
            /*
             * once the output buffer size exceeds the HTTP_OUTPUT_BUFFER_SIZE
             * the session becomes a HTTP1.1 type session
             */
            chunkedOut = true;
            flush();
        }
        
        byteArrayOutput.write(b, off, len);
                
        return len;
 
    }
    /**
     * Flush any output data to the output stream.
     *
     * @exception IOException if an I/O error occurs
     */
    public void flush() throws IOException {

        chunkedOut = true;
        
        if (byteArrayOutput.size() > 0) {
                        
            /*
             * set Transfer-Encoding to "chunked".
             */
            setRequestProperty("Transfer-Encoding", "chunked");
            
            /*
             * move to chunk state at the end of the sendRequest()
             */
            sendRequest(CHUNK_STATE);

        } else {
            streamOutput.flush();
        }

    }

    /** 
     * Get the original URL used to open the HTTP connection.
     *
     * @return HTTP URL used in the the current connection
     */
    public String getURL() {
	/*
         * RFC:  Add back protocol stripped by Content Connection.
         */
        
        return protocol + ":" + saved_url;
    }

    /** 
     * Get the protocol scheme parsed from the URL.
     *
     * @return protocol scheme is "http" 
     */
    public String getProtocol() {
        return protocol;
    }

    /** 
     * Get the host name parsed from the URL.
     *
     * @return host name from the parsed URL
     */
    public String getHost() {
        return (url.host.length() == 0 ? null : url.host);
    }

    /** 
     * Get the file path name parsed from the URL.
     *
     * @return file path name from the parsed URL
     */
    public String getFile() {
        return url.path;
    }

    /** 
     * Get the fragment identifier  parsed from the URL.
     *
     * @return reference component from the parsed URL
     */
    public String getRef() {
        return url.fragment;
    }

    /** 
     * Get the query string  parsed from the URL.
     *
     * @return query string from the parsed URL
     */
    public String getQuery() {
        return url.query;
    }

    /** 
     * Get the query string  parsed from the URL.
     *
     * @return query string from the parsed URL
     */
    public int getPort() {
        return url.port;
    }

    /** 
     * Get the request method of the current connection.
     *
     * @return request method is GET, HEAD or POST
     * @see #setRequestMethod
     */
    public String getRequestMethod() {

        return method;
    }

    /** 
     * Set the request method of the current connection.
     *
     * @param method request method is GET, HEAD or POST
     * @exception IOException is thrown if the connection is already open
     * @see #getRequestMethod
     */
    public void setRequestMethod(String method) throws IOException {

        if (state == CONNECT_STATE)
            throw new IOException("connection already open");
        
        if (!method.equals(HEAD) && 
            !method.equals(GET) && 
            !method.equals(POST)) {
            throw new IOException("unsupported method: " + method);
        }
        
        this.method = method;
    }

    /** 
     * Get the request header value for the named property.
     *
     * @param key property name of specific HTTP 1.1 header field
     * @return value of the named property, if found, null otherwise.
     * @see #setRequestProperty
     */
    public String getRequestProperty(String key) {
        
        /* https handles the proxy fields in a different way */
        if (key.startsWith("Proxy-")) {
            return proxyHeaders.getProperty(key);
        }

        return reqProperties.getProperty(key);
    }

    /**
     * Set the request header name/value of specific HTTP 1.1 header field.
     *
     * @param key property name 
     * @param value property value
     * @exception IllegalArgumentException is thrown if value contains CRLF.
     * @exception IOException  If some other kind of I/O error occurs.
     * @see #getRequestProperty
     */
    public void setRequestProperty(String key, String value) throws
            IOException {
        
        if (state == CONNECT_STATE)
            throw new IOException("connection already open");

        if ((value.indexOf("\r\n")) != -1) {
            
            // illegal values passed for properties - raise an exception       
            throw new IllegalArgumentException("illegal value found");
        }

        setRequestField(key, value);
    }

    /**
     * Add the named field to the list of request fields.
     * This method is where a subclass should override properties.
     *
     * @param key key for the request header field.
     * @param value the value for the request header field.
     */
    protected void setRequestField(String key, String value) {

        /* https handles the proxy fields in a different way */
        if (key.startsWith("Proxy-")) {
            proxyHeaders.setProperty(key, value);
            return;
        }

        /*
         * if application setRequestProperties("Connection", "close")
         * then we need to know this & take appropriate default close action
         */
        if ((key.equals("Connection")) && (value.equals("close"))) {
            ConnectionCloseFlag = true;
        }

        if ((key.equals("Transfer-Encoding")) && (value.equals("chunked"))) {
            chunkedOut = true;
        }

        reqProperties.setProperty(key, value);
    }

    /** 
     * Get the response code of the current request.
     *
     * @return numeric value of the parsed response code
     * @exception IOException is thrown if a network error ocurrs
     */   
    public int getResponseCode() throws IOException {

        /*
         * move to READ state at the end of the sendRequest()
         */
        sendRequest(READ_STATE);

        return responseCode;
    }

    /** 
     * Get the response message of the current request.
     *
     * @return message associated with the current response header
     * @exception IOException is thrown if a network error ocurrs
     */   
    public String getResponseMessage() throws IOException {

        /*
         * move to READ state at the end of the sendRequest()
         */
        sendRequest(READ_STATE);

        return responseMsg;
    }

    /** 
     * Get the Content-Length for the current response.
     *
     * @return length of data to be transmitted after the response headers
     */   
    public long getLength() {
        
        try {
            /*
             * move to READ state at the end of the sendRequest()
             */
            sendRequest(READ_STATE);
        }
        catch (IOException x) {
            return -1;
        }
	return getHeaderFieldInt("content-length", -1);
    }

    /** 
     * Get the Content-Type for the current response.
     *
     * @return MIME type of data to be transmitted after the response header
     */   
    public String getType() {
        
        try {
            /*
             * move to READ state at the end of the sendRequest()
             */
            sendRequest(READ_STATE);
        }
        catch (IOException x) {
            return null;
        }
	return getHeaderField("content-type");
    }

    /** 
     * Get the Content-Encoding for the current response.
     *
     * @return encoding type of data to be transmitted after the 
     *         response headers
     */   
    public String getEncoding() {

        try {
            /*
             * move to READ state at the end of the sendRequest()
             */
            sendRequest(READ_STATE);
        }
        catch (IOException x) {
            return null;
        }
	return getHeaderField("content-encoding");
    }

    /** 
     * Get the Expires header for the current response.
     *
     * @return expiration data for the transmitted data
     */   
    public long getExpiration() {
	return getHeaderFieldDate("expires", 0);
    }

    /** 
     * Get the Date header for the current response.
     *
     * @return timestamp for the data transmission event
     */   
    public long getDate() {
	return getHeaderFieldDate("date", 0);
    }

    /** 
     * Get the Last-Modified date header for the current response.
     *
     * @return timestamp for the transmitted data last modification
     */   
    public long getLastModified() {
	return getHeaderFieldDate("last-modified", 0);
    }

    /** 
     * Get the named header field  for the current response.
     *
     * @param name header field to be examined
     * @return value of requested header, if found, otherwise null
     */   
    public String getHeaderField(String name) {
        try {
            /*
             * move to READ state at the end of the sendRequest()
             */
            sendRequest(READ_STATE);
        }
        catch (IOException x) {
            return null;
        }
        return (headerFields.getProperty(toLowerCase(name)));

    }
    
    /** 
     * Get the indexed header field  for the current response.
     *
     * @param index header field offset to be examined
     * @return key name of requested header, if found, otherwise null
     */   
    public String getHeaderField(int index) {
        try {
            /*
             * move to READ state at the end of the sendRequest()
             */
            sendRequest(READ_STATE);
        }
        catch (IOException x) {
            return null;
        }
        
        if (index >= headerFields.size())
            return null;
        
        return (headerFields.getValueAt(index));
    }

    /** 
     * Get the indexed header field value for the current response.
     *
     * @param index header field value offset to be examined
     * @return value of requested header, if found, otherwise null
     */   
    public String getHeaderFieldKey(int index) {

        try {
            /*
             * move to READ state at the end of the sendRequest()
             */
            sendRequest(READ_STATE);
        }
        catch (IOException x) {
            return null;
        }

        if (index >= headerFields.size())
            return null;
        
        return ((String)(headerFields.getKeyAt(index)));
    }
    
    /** 
     * Get the named header field  for the current response and return a 
     * numeric value for the parsed field, with a supplied default value  
     * if the field does not exist or can not be parsed cleanly.
     *
     * @param name of the field to be examined
     * @param def default value to use, if field is not parsable
     * @return numeric value of requested header, if found, otherwise 
     *         supplied default is returned
     */   
    public int getHeaderFieldInt(String name, int def) {
	try {
            /*
             * move to READ state at the end of the sendRequest()
             */
            sendRequest(READ_STATE);
        }
	catch (IOException x) {
            return def;
        }
	try {
	    return Integer.parseInt(getHeaderField(name));
	} 
        catch (IllegalArgumentException iae) { }
        catch (NullPointerException npe) { }

	return def;
    }

    /** 
     * Get the named header field for the current response and return a date
     * value for the parsed field,with a supplied default value if the field
     * does not exist or can not be parsed cleanly.
     *
     * @param name of the field to be examined
     * @param def default value to use, if field is not parsable
     * @return date value of requested header, if found, otherwise 
     *         supplied default is returned
     */   
    public long getHeaderFieldDate(String name, long def) {
        
        try {
            /*
             * move to READ state at the end of the sendRequest()
             */
            sendRequest(READ_STATE);
        } catch (IOException x) {
            return def;
        }

        try {
            return DateParser.parse(getHeaderField(name));
        } 
        catch (NumberFormatException nfe) { }
        catch (IllegalArgumentException iae) { }
        catch (NullPointerException npe) { }
        
        return def;
    }

    /** 
     * Connect to the underlying socket transport.
     * If an http_proxy was specified the socket connection will be made to 
     * the proxy server and the requested URL will include the full http URL.
     * <P> 
     * On output the Content-Length header is included in the request based
     * on the size of the buffered output array.
     * <P> 
     * This routine inserts the Host header needed for HTTP 1.1 virtual host
     * addressing.
     * <P>
     * This routine also receives the reply response and parses the headers
     * for easier access. After the headers are parsed the application has
     * access to the raw data from the socket stream connection.
     *
     * @param     next_state  transition to this state when finished
     * @exception IOException is thrown if the connection cannot be opened
     */
    protected void sendRequest(int next_state) throws IOException {

        if (state == CONNECT_STATE) {
            return;
        }

        if (state > OPEN_STATE) {
            
            /*
             * if moved to CHUNK state all we need to do is write payload
             */
            if (next_state == CHUNK_STATE) {
                sendRequestWritePayload(next_state);
            }
            /*
             * if current state is CHUNK and next state is CLOSE 
             * or READ STATE's - time to end
             */
            if ((state == CHUNK_STATE) && ((next_state == CLOSE_STATE) ||
                                           (next_state == READ_STATE))) {
                if (byteArrayOutput.size() > 0)
                    sendRequestWritePayload(state);
                    
                sendRequestEndSession(next_state);
            } else {            
                setState(next_state);
            }

        } else {

            /*
             * before (creating) and getting a connection set the state OPEN.
             */
            setState(OPEN_STATE);

            streamConnection = connect();

            streamOutput = streamConnection.openDataOutputStream();
            streamInput = streamConnection.openDataInputStream();
            
            sendRequestGeneric(next_state);
            
            /* 
             * write the payload to the output stream
             */
            sendRequestWritePayload(next_state);
            
            /*
             * if next state not CHUNK write - time to end.
             */
            if (next_state != CHUNK_STATE) {
                sendRequestEndSession(next_state);
            } else {
            
                setState(next_state);                
            }

        }
        /*
	 * If the connection is a proxy connection and the proxy 
	 * server returns an error, raise an exception.
	 */
        if (http_proxy != null) {
            if (responseCode >= HTTP_INTERNAL_ERROR) {
                throw new IOException("server error ["+responseCode+"]");
            }
        }

        return;
    }
    /**
     * Simplifies the sendRequest() method functionality into one method
     * this is extremely helpful for persistent connection support and
     * retries.
     *
     * @param     next_state  the next state to transition to
     * @exception IOException is thrown if the connection cannot be opened
     */
    protected void sendRequestGeneric(int next_state) throws IOException {
        int numberOfKeys;

        if (reconnects > MAX_NUMBER_OF_RECONNECTS) {
            throw new IOException("ERROR: too many reconnections");
        }
        reconnects++;

	// HTTP 1.0 requests must contain content length for proxies
	if (getRequestProperty("Content-Length") == null) {
	    setRequestProperty("Content-Length",
			     "" + (out == null ? 0 : byteArrayOutput.size()));
	}

        /*
         * if in OPEN state - then begin to format http request headers.
         */
        if (state != OPEN_STATE) {
            return;
        }
        
        String reqLine;
        
        /*
         * HTTP RFC and bug#4402149,
         * if there is no ("/") slash in the path then add one.
         */
        String filename = url.path;
        if (filename == null) {
            filename = "/";
        }

        /*
         * Note: the "ref" or fragment, is not sent to the server.
         */
        if (http_proxy == null) {
            reqLine = method + " " + filename
                + (url.query == null ? "" : "?" + url.query)
                + " " + HTTP_VERSION + "\r\n";
        } else {
            reqLine = method + " "  
                + protocol + "://" + hostAndPort
                + filename
                + (url.query == null ? "" : "?" + url.query)
                + " " + HTTP_VERSION + "\r\n";
        }
        
        /*
         *  Persistent connections:
         *  lets try and write to the output stream - if it fails
         *  then the connection is dead - re-connect and come
         *  back through this method again (until max amount of times).
         */
        try {
            streamOutput.write((reqLine).getBytes());
        } catch (IOException ioe) {
            
            if (chunkedOut) {
                throw (ioe);
            }
            reconnectServer();
            sendRequestGeneric(next_state);
            return;
        }
        /*
         * HTTP 1/1 requests require the Host header to distinguish
         * virtual host locations.
         */
        
        setRequestProperty("Host", hostAndPort);

        /*
         * Setup the various http header field/values defined and/or
         * required.
         */
        numberOfKeys = reqProperties.size();
        for (int i = 0; i < numberOfKeys; i++) {
            String key = (String)reqProperties.getKeyAt(i);
            String reqPropLine = key + ": " + reqProperties.getValueAt(i) +
                "\r\n";
            
            if (key.equals("Content-Length")) {
                /*
                 * If its CHUNK state - no content-length: size required.
                 */
                if ((next_state == CHUNK_STATE) || (chunkedOut)) {
                    continue;                         
                } else {
                    /*
                     * Check that the output stream has been opened.
                     */
                    if (byteArrayOutput == null) {
                        reqPropLine = "Content-Length: 0" + "\r\n";
                    } else {
                        reqPropLine = "Content-Length: "+
                            byteArrayOutput.size() + "\r\n";
                    }
                }
            }
            
            /*
             *  Persistent connections:
             *  lets try and write to the output stream - if it fails
             *  then the connection is dead - re-connect and come
             *  back through this method again (until max amount of times)
             */
            try {
                streamOutput.write((reqPropLine).getBytes());
            } catch (IOException ioe) {

                if (chunkedOut)
                    throw (ioe);
                reconnectServer();
                sendRequestGeneric(next_state);
                return;
            }
            
        }
        
        /*
         *  Persistent connections:
         *  lets try and write to the output stream - if it fails
         *  then the connection is dead - re-connect and come
         *  back through this method again (until max amount of times).
         */
        try {
            streamOutput.write("\r\n".getBytes());
            
        } catch (IOException ioe) {

            if (chunkedOut)
                throw (ioe);
            reconnectServer();
            sendRequestGeneric(next_state);
            return;
        }
        
        return;
        
    }

    /**
     * Ends the http session output stream.
     *
     * @param     next_state  the next state to transition to - not done here
     * @exception IOException is thrown, if an I/O error occurs for final stream
     *              output or on reading the response message line
     */
    protected void sendRequestEndSession(int next_state) throws IOException {

        /*
         * if this is a CHUNKed session write out the last set of CRLF
         */
        if ((state == CHUNK_STATE) || (chunkedOut)) {

            /*
             * reference: RFC2616 - section 3 protocol parameters
             * 3.6 transfer coding:
             * 3.6.1 chunked transfer coding:
             * chunk-body = chunk / last chunk / trailer / CRLF
             *         chunk =      chunk-size / chunk-data / CRLF
             *       * last_chunk = "0" / CRLF
             *       * trailer    = " " / CRLF
             * * indicates its done here.
             */
            
            /*
             * write the last chunk (size=0 / CRLF) and the dummy trailer 
             */
            streamOutput.write("0\r\n\r\n".getBytes()); 
            streamOutput.flush();
        } else {
            streamOutput.flush();
        }

        readResponseMessage(streamInput);
        readHeaders(streamInput);
        
        /*
         * Ignore a continuation header and read the true headers again.
         * (Bug# 4382226 discovered with Jetty HTTP 1.1 web server.
         */
        if (responseCode == 100) {
            readResponseMessage(streamInput);
            readHeaders(streamInput);
        }
        
        setState(CONNECT_STATE);
        
        return;
    }
    /**
     * Write the http message payload bytes to the output stream.
     *
     * @param      b     the data.
     * @param      off   the start offset in the data.
     * @param      len   the number of bytes to write.
     * @exception  IOException 
     */
    protected void sendRequestWritePayloadDirect(byte b[], int off, int len) 
        throws IOException {
                
        String temp = Integer.toHexString(len);

        streamOutput.write(temp.getBytes());
        streamOutput.write("\r\n".getBytes());             
        
        streamOutput.write(b, off, len);
        streamOutput.write("\r\n".getBytes());         
        
    }

    /**
     * Write the http message payload bytes to the output stream.
     *
     * @param     next_state  the next state to transition to - not done here
     * @exception IOException 
     */
    protected void sendRequestWritePayload(int next_state) 
        throws IOException {
                     
        if ((byteArrayOutput == null) || (byteArrayOutput.size() == 0)) {
            return;
        }

        /*
         *  Persistent connections:
         *  lets try and write to the output stream - if it fails
         *  then the connection is dead - re-connect and come
         *  back through this method again (until max amount of times).
         */
        
        try {
            /*
             * If a CHUNKed session then write out the chunk size first 
             * with a trailing CRLF.
             *
             * reference: RFC2616 - section 3 protocol parameters
             * 3.6 transfer coding:
             * 3.6.1 chunked transfer coding:
             * chunk-body = chunk / last chunk / trailer / CRLF
             *       * chunk =      chunk-size / chunk-data / CRLF
             *         last_chunk = "0" / CRLF
             *         trailer    = " " / CRLF
             * * indicates its done here.
             */
            if ((next_state == CHUNK_STATE) || (chunkedOut)) {

                /*
                 * For CHUNKed write out the chunk size with CRLF.
                 */
                String temp = Integer.toHexString(byteArrayOutput.size());
                streamOutput.write(temp.getBytes());
                streamOutput.write("\r\n".getBytes());  
            }           
            streamOutput.write(byteArrayOutput.toByteArray());
            byteArrayOutput.reset();
            
            /*
             * If a CHUNKed session then write out another CRLF and flush().
             */
            if ((next_state == CHUNK_STATE) || (chunkedOut)) {
                streamOutput.write("\r\n".getBytes()); 
            }
        } catch (IOException ioe) {
            if (chunkedOut)
                throw (ioe);
            reconnectServer();
            sendRequestGeneric(next_state);
            return;
        }
        
    }

    /**
     * Disconnect/Reconnect to server - somehow the connection was lost.
     *
     * @exception IOException is thrown if the connection cannot be opened
     */
    protected void reconnectServer() throws IOException {

        boolean old_ConnectionCloseFlag = ConnectionCloseFlag;
        String old_httpVer = httpVer;

        ConnectionCloseFlag = true; // need to force disconnect!!
        httpVer = "HTTP/1.0";       // need to force disconnect!!

        disconnect();
        streamConnection = connect();
        
        ConnectionCloseFlag = old_ConnectionCloseFlag;
        httpVer = old_httpVer;

        return;
    }

    /**
     * Connect to the underlying socket transport.
     * If an http_proxy was specified the socket connection will be made to 
     * the proxy server.
     *
     * @return TCP/IP stream connection
     * @exception IOException is thrown if the connection cannot be opened
     */
    protected StreamConnection connect() throws IOException {

        StreamConnection conn;

        /*
         * Get the proxy here instead of the connector,
         * so when this method subclassed by HTTPS http_proxy will be null
         * and the proxy will not be added into the request.
         */
	http_proxy = Configuration.getProperty("com.sun.midp.io.http.proxy");
        
        /*
         * if first time intialize the connection pool and create
         * a connection (you could create more)
         */
        if (connectionPool == null) {
            connectionPool = new StreamConnectionPool(
                                          maxNumberOfPersistentConnections);
        }
        
        /*
         * Each subsequent time call into the connection pool
         * and either grab an available one or create a new one
         */
        if (http_proxy == null) {
            conn = connectionPool.getConnection(saved_url, 
                                                url.host, 
                                                url.port, 
                                                hostAndPort);
            streamOutput = conn.openDataOutputStream();
            streamInput = conn.openDataInputStream();
        } else {
            conn = connectionPool.getConnection(saved_url, 
                                                url.host, 
                                                url.port, 
                                                http_proxy);
            streamOutput = conn.openDataOutputStream();
            streamInput = conn.openDataInputStream();

            try {
                doTunnelHandshake(streamOutput, streamInput);
            } catch (IOException ioe) {
                connectionElement = (StreamConnectionElement)conn;
                boolean rc = connectionPool.remove(connectionElement);
                streamOutput.close();
                streamInput.close();
                String response = ioe.getMessage();
                if (response.indexOf(" 500 ") > -1) {
                    throw new ConnectionNotFoundException(ioe.getMessage());
                } else {
                    throw ioe;
                }
            }
        }
        
        connectionElement = (StreamConnectionElement)conn;
        
        return conn;
    }

    /**
     * Connects to the SSL tunnel and completes the intialization of the
     * tunnel (handshake). The handshake based on the Internet-Draft
     *	"A. Luotonen, Tunneling TCP based protocols through Web proxy servers,
     *  February 1999".
     * @param os output stream for secure handshake
     * @param is input stream for secure handshake
     * @exception IOException is thrown if an error occurs in the SSL handshake
     */ 
    protected void doTunnelHandshake(OutputStream os, InputStream is) throws
            IOException {
	String required;
        String optional;
        String endOfLine = "\r\n";
        String emptyLine = endOfLine;
        int numberOfKeys;
        StringBuffer temp;
	boolean newline;
	String response;

        /*
         * request = required *optional emptyLine
         * required = "CONNECT" SP HOST ":" PORT SP HTTP_VERSION endOfLine
         * optional = HTTP_HEADER endOfLine ; proxy dependent: most likely
         *                                  ; used for authorization.
         * emptyLine = endOfLine
         * endOfLine = *1CR LF
         *
         * example:
         * CONNECT home.acme.com:443 HTTP/1.0
         * Proxy-Authorization: Basic QWxhZGRpbjpvcGVuIHNlc2FtZQ==
         *
         */
	required = "CONNECT " + hostAndPort + " " + HTTP_VERSION + endOfLine;
	os.write(required.getBytes());

        numberOfKeys = proxyHeaders.size();
        for (int i = 0; i < numberOfKeys; i++) {
	    optional = proxyHeaders.getKeyAt(i) + ": " +
                       proxyHeaders.getValueAt(i) + endOfLine;
	    os.write(optional.getBytes());
	}

        os.write(emptyLine.getBytes());
	os.flush();

	/*
         * response = status *optional emptyLine
         * status = HTTP_VERSION SP STATUSCODE STATUS_MESSAGE *1CR LF
         * optional = HTTP_HEADER *1CR LF
         * emptyLine = *1CR LF
         *
         * example:
         * HTTP/1.0 200 Connection established
         *
         */

	// Read in response until an empty line is found (1*CR LF 1*CR LF)
        temp = new StringBuffer();
	newline = false;
	while (true) {
	    int c = is.read();
            if (c == -1) {
                break;
	    } else if (c == '\n') {
		if (newline) {
		    break;
		}
		newline = true;
	    } else if (c != '\r') {
		newline = false;
	    }

	    temp.append((char)c);
	}

        if (temp.length() == 0) {
            temp.append("none");
        }

	response = temp.toString();

        if (response.indexOf(" 200 ") == -1) {
	    throw new
                IOException("Error initializing HTTP tunnel connection: \n"
                            + response);
	}
    }
    
    /** 
     * Check the initial response message looking for the 
     * appropriate HTTP version string. Parse the response 
     * code for easy application branching on condition codes.
     *
     * @param in input stream where the response headers are read
     * @exception IOException  is thrown if the header response can 
     *                         not be parsed
     */
    private void readResponseMessage(InputStream in) throws IOException {

        String line = null;
        responseCode = -1;
        responseMsg = null;
        try {
            line = readLine(in);
        } catch (IOException ioe) {
            /* throw new IOException(ioe.getMessage()); */
        }

        /*
         * REFERENCE: HTTP1.1 document 
         * SECTION: 3.6.1 Chunked Transfer Coding
         * in some cases there may be an OPTIONAL trailer containing 
         * entity-header fields. since we don't support the available()
         * method for inputstreams and for performance reasons we
         * do not attempt to clean up the previous connections input
         * stream. the first thing we do here is read the stream and 
         * discard it.
         */
        if (line != null && line.length() == 0) {
            try {
                line = readLine(in);
            } catch (IOException ioe) {
                /* throw new IOException(ioe.getMessage()); */
            }
        }
            
	int httpEnd, codeEnd;
        
	responseCode = -1;
	responseMsg = null;

    malformed: {
	    if (line == null)
		break malformed;
	    httpEnd = line.indexOf(' ');
	    if (httpEnd < 0)
		break malformed;
    
	    httpVer = line.substring(0, httpEnd);
	    if (!httpVer.startsWith("HTTP"))
		break malformed;
	    if (line.length() <= httpEnd)
		break malformed;
    
	    codeEnd = line.substring(httpEnd + 1).indexOf(' ');
	    if (codeEnd < 0)
		break malformed;
	    codeEnd += (httpEnd + 1);
	    if (line.length() <= codeEnd) 
		break malformed;
    
	    try {
                responseCode = Integer.parseInt(line.substring(httpEnd+1, 
                                                               codeEnd));
            }
	    catch (NumberFormatException nfe) {
		break malformed;
	    }
    
	    responseMsg = line.substring(codeEnd + 1);
	    return;
	}  

	throw new IOException("malformed response message");
    }

    /** 
     * Read the response message headers.
     * Parse the response headers name value pairs for easy application use.
     *
     * @param in input stream where the response headers are read
     * @exception IOException  is thrown if the response headers cannot 
     *                         be parsed
     */
    private void readHeaders(InputStream in) throws IOException {
	String line, key, value;
	int index;
        
	for (;;) {
            try {
                line = readLine(in);
            } catch (IOException ioe) {
                throw new IOException(ioe.getMessage());
            }

	    if (line == null || line.equals(""))
		return;

	    index = line.indexOf(':');
	    if (index < 0) {
		throw new IOException("malformed header field");
            }

	    key = line.substring(0, index);
	    if (key.length() == 0) {
		throw new IOException("malformed header field");
            }

	    if (line.length() <= index + 1) 
                value = "";
	    else 
                value = line.substring(index + 1).trim();

            /**
             * Check the response header to see if the server would like
             * to close the connection.
             * BUG#4492849
             */
            if ((key.equals("Connection")) && (value.equals("close"))) {
                ConnectionCloseFlag = true;
            }

            headerFields.addProperty(toLowerCase(key), value);
	}
    }

    /**
     * Uses the shared stringbuffer to read a line terminated by CRLF 
     * and return it as string.
     *
     * @param     in  InputStream to read the data
     * @return    one line of input header 
     * @exception IOException if EOF encountered while reading headers
     */
    private String readLine(InputStream in) throws IOException {
	int c;

	stringbuffer.setLength(0);
	for (;;) {
	    try {
		c = in.read();
		if (c < 0) {
                    /* throw new IOException("EOF"); */
		    return null;
		}
		if (c == '\r') {
		    continue;
		}
	    } catch (IOException ioe) {
                /* throw new IOException(ioe.getMessage()); */
		return null;
	    }
	    if (c == '\n') {
		break;
	    }
	    stringbuffer.append((char)c);
	}

	return stringbuffer.toString();
    }

    /** 
     * Disconnect from the underlying socket transport.
     * Closes the low level socket connection and the input and 
     * output streams used by the socket.
     * <P>
     * Resets the flags associated with the last transaction and the 
     * connectedness of the current HTTP connection.
     *
     * @exception IOException is thrown if the connection or 
     *                        associated streams cannot be closed
     */
    protected void disconnect() throws IOException {
        disconnect(streamConnection, streamInput, streamOutput);
    }

    /**
     * Disconnect the current low level socket connection. If the connection
     * is an HTTP1.1 connection that connection will be put back in the pool
     * for another session to use to connect.
     *
     * @param connection connection return from {@link #connect()}
     * @param inputStream input stream opened from <code>connection</code>
     * @param outputStream output stream opened from <code>connection</code>
     * @exception IOException if an I/O error occurs while
     *                  the connection is terminated.
     */
    protected void disconnect(StreamConnection connection,
           InputStream inputStream, OutputStream outputStream)
           throws IOException {

        /** 
         * if the current state is CHUNK then close the outputstream 
         */
        if ((state == CHUNK_STATE) || (state == WRITE_STATE)) {
            sendRequest(CLOSE_STATE);
        }

        if ((streamConnection != null) && (httpVer != null)) {
	    streamOutput.flush();
            
            /*
             * if this is NOT HTTP1.1 or the application specifically
             * wants to close the stream - do this close everything down.
             */
            if ((httpVer.equals("HTTP/1.0")) || (ConnectionCloseFlag)) {
                boolean rc = connectionPool.remove(connectionElement);
                streamOutput.close();
                streamInput.close();
                streamConnection.close();
                streamConnection = null;

            } else {
                /*
                 * if this is a HTTP1.1 connection we need to check for an
                 * abort condition (if bytesleft != 0 or more chunks) we
                 * need to close and throw this connection away.
                 */

                if ((bytesleft > 0) || (eof != true)) {
                    boolean rc = connectionPool.remove(connectionElement);
                    streamOutput.close();
                    streamInput.close();
                    streamConnection.close();
                    streamConnection = null;
                }
            }
        } else {
            
            if (ConnectionCloseFlag) {
                if (connectionPool != null) {
                    boolean rc = connectionPool.remove(connectionElement);
                }
                if (streamOutput != null) {
                    streamOutput.close();
                }
                if (streamInput != null) {
                    streamInput.close();
                }
                if (streamConnection != null) {
                    streamConnection.close();
                }
                streamConnection = null;
                return;
            }

        }

        /*
         * return the connection to the connection pool
         */
        if (streamConnection != null) {
            try {
                connectionPool.returnConnection(streamConnection);
            } catch (Exception e) {
                throw new IOException(e.getMessage());
            }
        }
        
	responseCode = -1;
        reconnects = 0;
	responseMsg = null;

    }

    /** 
     * Convert a string to lowercase. Used to make uniform 
     * name for header fields in the response message for
     * simple comparisons.
     *
     * @param string text to be converted to lowercase
     * @return lowercase version of input string
     */   
    private String toLowerCase(String string) {
	// Uses the shared stringbuffer to create a lower case string.
	stringbuffer.setLength(0);
	for (int i = 0; i < string.length(); i++) {
	    stringbuffer.append(Character.toLowerCase(string.charAt(i)));
	}

	return stringbuffer.toString();
    }

    /**
     * Set the current state of the http connection - the state can
     * never move backwards (only if retured to OPEN_STATE).
     *
     * @param  pstate the next state to move to.
     * @exception IOException thrown if invalid state transition
     */
    private void setState(int pstate) throws IOException {

        if ((pstate < INIT_STATE) || (pstate > CONNECT_STATE)) {
            throw new RuntimeException("invalid state transition");
        }
        if ((pstate < state) && 
            ((state != READ_STATE) && (pstate != OPEN_STATE)) && 
            ((state != FLUSH_STATE) && (pstate != OPEN_STATE))) {
            throw new RuntimeException("invalid state transition");
        }

        state = pstate;
        return;
    }

}
