/*
 * @(#)DefaultInputMethodHandler.java	1.46 01/08/12
 *
 * Copyright (c) 2000-2001 Sun Microsystems, Inc., 901 San Antonio Road,
 * Palo Alto, CA 94303, U.S.A.  All Rights Reserved.
 *
 * Sun Microsystems, Inc. has intellectual property rights relating
 * to the technology embodied in this software.  In particular, and
 * without limitation, these intellectual property rights may include
 * one or more U.S. patents, foreign patents, or pending
 * applications.  Sun, Sun Microsystems, the Sun logo, Java, KJava,
 * and all Sun-based and Java-based marks are trademarks or
 * registered trademarks of Sun Microsystems, Inc.  in the United
 * States and other countries.
 *
 * This software is distributed under licenses restricting its use,
 * copying, distribution, and decompilation.  No part of this
 * software may be reproduced in any form by any means without prior
 * written authorization of Sun and its licensors, if any.
 *
 * FEDERAL ACQUISITIONS:  Commercial Software -- Government Users
 * Subject to Standard License Terms and Conditions
 */
package com.sun.midp.lcdui;

import java.util.TimerTask;
import java.util.Timer;

import javax.microedition.lcdui.Canvas;
import javax.microedition.lcdui.TextBox;
import javax.microedition.lcdui.TextField;
import javax.microedition.lcdui.Display;
import javax.microedition.lcdui.Displayable;
import javax.microedition.lcdui.Graphics;
import javax.microedition.lcdui.Font;

/**
 * A Default implementation of the InputMethodHandler interface.
 */
public class DefaultInputMethodHandler extends InputMethodHandler {
    /**
     * InputMethodClient client that is implemented by TextBox.
     */
    protected InputMethodClient imc;
    /**
     * Text buffer.
     */
    protected char buf[] = new char[2];
    /**
     * Number of characters in the buffer.
     */
    protected int bufLen;
    /**
     * Committed length of editted buffer.
     */
    protected int committedLen;
    /**
     * Max length to fit the client's max length.
     */
    protected int maxLen;
    /**
     * Depository for an event.
     */
    protected int savedEvent;
    /*
     * IDs for supported input modes
     */
    /** The 'none' input mode. */
    protected static final int IM_NONE        = 0;
    /** The 'roman caps' input mode. */
    protected static final int IM_ROMAN_CAPS  = 1;
    /** The 'lowercase roman' input mode. */
    protected static final int IM_ROMAN_SMALL = 2;
    /** The 'numeric' input mode. */
    protected static final int IM_NUMERIC     = 3;
    /** The 'symbol' input mode. */
    protected static final int IM_SYMBOL      = 4;

    /**
     * Default input mode.
     */
    protected int defaultMode;
    /**
     * The number of allowed input modes.
     */
    protected int allowedModesNum;
    /**
     * The set of allowed input modes.
     */
    protected int allowedModes[] = new int[5];
    /**
     * Currrent input mode.
     */
    protected int inputMode;
    /**
     * Next input mode.
     */
    protected int nextInputMode;
    /**
     * InputMode values for supported input modes.
     */
    protected String supportedInputModes[] = {
        "LATIN",
        "LATIN_DIGITS",
    };

    /**
     * Conversion table between this input modes and InputMode's input modes.
     */
    protected Object inputModeConvTable[][] = {
        {"LATIN",              new Integer(IM_ROMAN_CAPS)},
        {"LATIN",              new Integer(IM_ROMAN_SMALL)},
        {"LATIN_DIGITS",       new Integer(IM_NUMERIC)},
        {"LATIN",              new Integer(IM_SYMBOL)},
    };

    /**
     * Current constraint.
     */
    protected int constraint;

    // Key definition for key maps

    /** A special no-op key. */
    protected static final int KEY_NONE   = -1;
    /** The '0' key. */
    protected static final int KEY_NUM0   = 0;
    /** The '1' key. */
    protected static final int KEY_NUM1   = 1;
    /** The '2' key. */
    protected static final int KEY_NUM2   = 2;
    /** The '3' key. */
    protected static final int KEY_NUM3   = 3;
    /** The '4' key. */
    protected static final int KEY_NUM4   = 4;
    /** The '5' key. */
    protected static final int KEY_NUM5   = 5;
    /** The '6' key. */
    protected static final int KEY_NUM6   = 6;
    /** The '7' key. */
    protected static final int KEY_NUM7   = 7;
    /** The '8' key. */
    protected static final int KEY_NUM8   = 8;
    /** The '9' key. */
    protected static final int KEY_NUM9   = 9;
    /** The '*' key. */
    protected static final int KEY_STAR   = 10;
    /** The '#' key. */
    protected static final int KEY_POUND  = 11;
    /** The 'CLR' key. */
    protected static final int KEY_CLEAR  = 100;

    /**
     * Current input table that can be one of the following.
     */
    protected char keyMap[][];

    /**
     * The uppercase roman key map.
     */
    protected char upperRomanKeyMap[][] = {
	{'0'},
	{'1'},
	{'A', 'B', 'C', '2'},
	{'D', 'E', 'F', '3'},
	{'G', 'H', 'I', '4'},
	{'J', 'K', 'L', '5'},
	{'M', 'N', 'O', '6'},
	{'P', 'Q', 'R', 'S', '7'},
	{'T', 'U', 'V', '8'},
	{'W', 'X', 'Y', 'Z', '9'},
	{'\0'},
	{' '}
    };

    /**
     * The lowercase roman key map.
     */
    protected char lowerRomanKeyMap[][] = {
	{'0'},
	{'1'},
	{'a', 'b', 'c', '2'},
	{'d', 'e', 'f', '3'},
	{'g', 'h', 'i', '4'},
	{'j', 'k', 'l', '5'},
	{'m', 'n', 'o', '6'},
	{'p', 'q', 'r', 's', '7'},
	{'t', 'u', 'v', '8'},
	{'w', 'x', 'y', 'z', '9'},
	{'\0'},
	{' '}
    };

    /**
     * The numeric key map.
     * Includes 0-9,*, and space.
     */
    protected char numericKeyMap[][] = {
	{'0'}, 
        {'1'}, 
        {'2'}, 
        {'3'}, 
        {'4'}, 
        {'5'},
	{'6'}, 
        {'7'}, 
        {'8'}, 
        {'9'},
	{'*'},
	{' '}
    };

    /**
     * The phone numeric key map.
     * Includes 0-9,*,#,+.
     */
    protected char phoneNumericKeyMap[][] = {
	{'0'}, 
        {'1'}, 
        {'2'}, 
        {'3'}, 
        {'4'}, 
        {'5'},
	{'6'}, 
        {'7'}, 
        {'8'}, 
        {'9'},
	{'*'},
	{'#', '+'}
    };


    /**
     * Symbol table.
     */
    protected SymbolTable st = new SymbolTable();

    /**
     * The number of symbol_table is designed to be 25 for 5x5 matrix,
     * starting the selection at 12.  But if you have more, the total
     * must be under 36 for 6x6 matrix.  
     */
    protected char symbolTableChars[] = {
        '_',	'$',	'(',	')',	'\\',

        '~',	'"',	'\'',	'/',	'&',

        '*',	'@',	'.',	'?',	'!',

        '#',	'-',	',',	':',	';',

        '%',	'=',	'+',	'<',	'>'};

    /**
     * Default cursor position for the Symbol table.
     */
    protected int defaultSymbolCursorPos = 12;

    /**
     * Boolean indicating whether the current client is returned from
     * the Symbol table or not.
     */
    protected boolean hasSymbolTable = false;

    /**
     * Set the focus of this handler.
     * Overrides InputMethodHandler.setFocus
     *
     * @param focus If True, this handler should take focus
     */
    public void setFocus(boolean focus) {
        /*
         * Note that setFocus() is always called when Display is
         * changed.  
         */
        if (imc == null) {
            return;
        }
        if (focus) {
            if (hasSymbolTable) {
                int mode = nextInputMode();
                if (bufLen > 0) {
                    handleCharInput();
                }
                setInputMode(mode);
		hasSymbolTable = false;
            } else {
                setInputMode(defaultMode);
            }
        } else {
            if (!hasSymbolTable) {
                setInputMode(IM_NONE);
            }
        }
    }

    /**
     * Establish the input method client for this handler.
     * Overrides InputMethodHandler.setInputMethodClient
     *
     * @param imc The input method client
     */
    public void setInputMethodClient(InputMethodClient imc) {
        /*
         * return if the same imc.  this is called when the Symbol
         * table returns.  
         */
        if (this.imc == imc) {
            return;
        }

        // initialize variables
        lastKey = -1;
        lastPos = 0;
        incomposing = false;

        // set the imc specific stuff
        this.imc = imc;

        constraint = imc.getConstraints() & TextField.CONSTRAINT_MASK;

        setConstraint();

        // initialize inputMode and keyMap in case TextBox is
        // invoked from TextField.
        inputMode = defaultMode;

	switch (inputMode) {
	case IM_ROMAN_CAPS:
	    keyMap = upperRomanKeyMap;
	    break;
	case IM_ROMAN_SMALL:
	    keyMap = lowerRomanKeyMap;
	    break;
	case IM_NUMERIC:
	    keyMap = (constraint == TextField.PHONENUMBER) ?
                phoneNumericKeyMap : numericKeyMap;
	    break;
        }
    }

    /**
     * Clear the input method client.
     * Overrides InputMethodHandler.clearInputMethodClient.
     *
     * @param imcToBeCleared The input method client to be cleared
     */
    public void clearInputMethodClient(InputMethodClient imcToBeCleared) {
        if (!hasSymbolTable && imc == imcToBeCleared) {
            imc = null;
        }
    }

    /**
     * Translate the given key code to its index in the key map.
     *
     * @param keyCode The key code of the key pressed
     * @return int The index of the given key code in the key map
     */
    protected int toKeyMapIndex(int keyCode) {
	switch (keyCode) {
        case Canvas.KEY_NUM0:
            return KEY_NUM0;
        case Canvas.KEY_NUM1:
            return KEY_NUM1;
        case Canvas.KEY_NUM2:
            return KEY_NUM2;
        case Canvas.KEY_NUM3:
            return KEY_NUM3;
        case Canvas.KEY_NUM4:
            return KEY_NUM4;
        case Canvas.KEY_NUM5:
            return KEY_NUM5;
        case Canvas.KEY_NUM6:
            return KEY_NUM6;
        case Canvas.KEY_NUM7:
            return KEY_NUM7;
        case Canvas.KEY_NUM8:
            return KEY_NUM8;
        case Canvas.KEY_NUM9:
            return KEY_NUM9;
        case Canvas.KEY_STAR:
            return KEY_STAR;
        case Canvas.KEY_POUND:
            return KEY_POUND;
        case -8:
            return KEY_CLEAR;
	default:
	    return KEY_NONE;
	}
    }

    /**
     * Handle a key pressed event.
     * Overrides InputMethodHandler.keyPressed.
     *
     * @param keyCode The code of the key that was pressed
     * @return boolean If True, this handler has handled the key press
     */
    public boolean keyPressed(int keyCode) {
        int idx = toKeyMapIndex(keyCode);

        if (idx == KEY_NONE) {
            return false;
        }

        if (idx == KEY_CLEAR) {
            setTimer(TM_CLEAR_BUFFER, 1500);
            return true;
        }

        switch (constraint) {
        case TextField.NUMERIC:
            if (idx == KEY_STAR) {
                toggleSign();
                return true;
            }
            if (idx == KEY_POUND) {
                return false;
            }
            break;
        case TextField.PHONENUMBER:
            break;
        case TextField.EMAILADDR:
        case TextField.URL:
        case TextField.ANY:
            if (idx == KEY_STAR) {
                setTimer(TM_INPUT_MODE, 1000);
                return true;
            }
            break;
        }
        
        if (processChar(idx)) {
            handleCharInput();
            return true;
        }

        return false;
    }

    /**
     * Handle a key released event.
     * Overrides InputMethodHandler.keyReleased.
     *
     * @param keyCode The code of the key that was released
     * @return boolean If True, this handler has handled the key release
     */
    public boolean keyReleased(int keyCode) {
        int idx = toKeyMapIndex(keyCode);

        if (idx == KEY_NONE) {
            return false;
        }

        if (idx == KEY_CLEAR) {
            if (timerType == TM_CLEAR_BUFFER) {
                endComposition();
                int pos = imc.getCaretPosition();
                if (pos > 0) {
                    imc.replace(null, 0, 1, pos-1, pos);
                }
            }
            return true;
        }

        if (idx == KEY_STAR) {
            if (timerType == TM_INPUT_MODE) {
                endComposition();
                setInputMode(nextInputMode());
            }
            return true;
        }

        return false;
    }

    /**
     * Handle a key repeated event.
     * Overrides InputMethodHandler.keyRepeated.
     *
     * @param keyCode The code of the key that was repeated
     * @return boolean If True, this handler has handled the key repeat
     */
    public boolean keyRepeated(int keyCode) {
        return false;
    }

    /**
     * End the composition of this handler.
     * Overrides InputMethodHandler.endComposition.
     */
    public void endComposition() {
        if (timerType != TM_NONE) {
            cancelTimer();
            timerType = TM_NONE;
        }
        if (incomposing) {
            imc.setCaretVisible(true);
            incomposing = false;
        }
	end();
    }

    /**
     * Get the list of supported input modes of this handler.
     * Overrides InputMethodHandler.supportedInputModes.
     *
     * @return String[] The array of supported input modes
     */
    public String[] supportedInputModes() {
        return supportedInputModes;
    }

    /**
     * Insert a character.
     *
     * @param c The character to insert
     */
    protected void insertChar(char c) {
        if (imc.getSize() < imc.getMaxSize()) {
            char[] buf = {c};
            imc.insert(buf, 0, 1, imc.getCaretPosition());
        }
    }

    /**
     * Toggle the sign (+/-).
     */
    protected void toggleSign() {
        char[] tmp = new char[imc.getSize() + 1];
        int len = imc.getChars(tmp);
        if (len > 0 && tmp[0] == '-') {
            --len;
            if (len == 0) {
                imc.setChars(null, 0, 0);
            } else {
                System.arraycopy(tmp, 1, tmp, 0, len);
                int pos = imc.getCaretPosition();
                imc.setChars(tmp, 0, len);
                imc.setCaretPosition(--pos);
            }
        } else {
            if (len >= imc.getMaxSize()) {
                return;
            }
            if (len != 0) {
                System.arraycopy(tmp, 0, tmp, 1, len);
            }
            ++len;
            tmp[0] = '-';
            int pos = imc.getCaretPosition();
            imc.setChars(tmp, 0, len);
            imc.setCaretPosition(++pos);
        }
    }

    /** Flag indicating this handler is composing text. */
    protected boolean incomposing;
    /** The text input position. */
    protected int inputPos; // keep track of text input position

    /**
     * Handle character input.
     */
    protected void handleCharInput() {
        if (bufLen == 0) {
            return;
        }

        if (incomposing == false) {
            inputPos = imc.getCaretPosition();
        }

        if (committedLen >= 1) {
            if (incomposing == true) {
                if ((imc.getSize() + committedLen) >= imc.getMaxSize()) {
                    committedLen = imc.getMaxSize() - imc.getSize();
                }
                if (committedLen > 0) {
                    imc.replace(buf, 0, committedLen, inputPos-1, inputPos);
                }
            } else {
                for (int i = 0; i < committedLen; i++) {
                    if (imc.getSize() < imc.getMaxSize()) {
                        imc.insert(buf, i, i+1, inputPos+i);
                    }
                }
            }
            incomposing = false;
        }

        if (bufLen > 0 && bufLen > committedLen) {
            if (incomposing == true) {
                imc.replace(buf, committedLen, 1, inputPos-1, inputPos);
            } else {
                if (imc.getSize() < imc.getMaxSize()) {
                    imc.insert(buf, committedLen, 1, inputPos++);
                    incomposing = true;
                }
            }
        }

        if (incomposing == true) {
            setTimer(TM_IN_COMPOSING, 1600);
        } else {
            imc.setCaretVisible(true);
        }
    }

    /** The '0' Timer. */
    protected final static int TM_NONE         = 0;
    /** The input mode timer. */
    protected final static int TM_INPUT_MODE   = 1;
    /** The 'in composition' timer. */
    protected final static int TM_IN_COMPOSING = 2;
    /** The 'clear buffer' timer. */
    protected final static int TM_CLEAR_BUFFER = 3;
    /** The type of timer to set. */
    protected int timerType;

    /**
     * Set a new timer.
     *
     * @param type The type of Timer to set
     * @param delay The length of delay for the Timer
     */
    protected void setTimer(int type, long delay) {
        if (type != timerType) {
            if (type == TM_IN_COMPOSING) {
                imc.setCaretVisible(false);
            }
            timerType = type;
        }
        cancelTimer();
        try {
            timerClient = new TimerClient();
            timerService.schedule(timerClient, delay);
        } catch (IllegalStateException e) {
            e.printStackTrace();
            cancelTimer();
        }
    }

    /** The Timer to service TimerTasks. */
    protected Timer timerService = new Timer();
    /** A TimerTask. */
    protected TimerTask timerClient = null;

    /**
     * A special TimerTask class
     */
    class TimerClient extends TimerTask {
        /**
         * Simply calls timerWentOff()
         */
        public final void run() {
            timerWentOff();
        }
    }

    /**
     * Cancel any running Timer.
     */
    protected void cancelTimer() {
        if (timerClient != null) {
            timerClient.cancel();
            timerClient = null;
        }
    }

    /**
     * Called by the TimerTask when a Timer executes.
     */
    protected void timerWentOff() {
        switch (timerType) {
        case TM_INPUT_MODE:
            endComposition();
            nextInputMode = inputMode;
            setInputMode(IM_SYMBOL);
            break;
        case TM_IN_COMPOSING:
            endComposition();
            break;
        case TM_CLEAR_BUFFER:
            endComposition();
            imc.setChars(null, 0, 0);
            break;
        }
    }

    /** The last key and last position of the input. */
    protected int lastKey, lastPos;

    /**
     * Process a character.
     *
     * @param idx The index into the keymap
     * @return boolean
     */
    protected boolean processChar(int idx) {
	/*
	 * 1. To start from the initial state.  */
	if (lastKey == -1) {
	    if (inputMode == IM_NONE) {
		buf[0] = 0;
		bufLen = committedLen = 0;
            } else {
                buf[0] = keyMap[idx][0];
                if (buf[0] == '\0') {
                    return false;
                }
                if (keyMap[idx].length == 1) {
		    bufLen = committedLen = 1;
                } else {
                    bufLen = 1;
		    lastKey = idx;
		    committedLen = 0;
		    lastPos = 0;
                }
            }
	    return true;
	} else

	/*
	 * 2. To commit the previous key and send the current key.
	 */
        if (lastKey != idx) {
            char c = keyMap[idx][0];
	    if (c == '\0') {
		return false;
	    }
	    buf[0] = keyMap[lastKey][lastPos];
            buf[1] = c;
            bufLen = 2;
            lastPos = 0;
            if (keyMap[idx].length == 1) {
	        committedLen = 2;
	        lastKey = -1;
            } else {
	        committedLen = 1;
	        lastKey = idx;
            }
	    return true;
	} else 

	/*
	 * 3. To send the next candidate at the same key position.
	 */
        if (lastKey == idx) {
	    int len = keyMap[idx].length;
	    if (len == (lastPos+1)) {
		lastPos = 0;
	    } else {
		lastPos += 1;
	    }
	    buf[0] = keyMap[idx][lastPos];
	    bufLen = 1;
	    committedLen = 0;
	    return true;
	}

	return false;
    }

    /**
     * End the composition.
     */
    protected void end() {
	if (lastKey != -1) {
	    buf[0] = keyMap[lastKey][lastPos];
	    bufLen = 1;
	    committedLen = 1;
	    lastKey = -1;
	} else {
	    bufLen = 0;
	    committedLen = 0;
	}
    }

    /**
     * Get the next input mode.
     *
     * @return int The next input mode
     */
    protected int nextInputMode() {
        if (nextInputMode != IM_NONE) {
            int tmp = nextInputMode;
            nextInputMode = IM_NONE;
            return tmp;
        }

	int n = 0;

	// get the current input mode
	for (; n < allowedModesNum; n++) {
	    if (allowedModes[n] == inputMode) {
		break;
	    }
	}

        // advance the input mode
        n++;
        if (n >= allowedModesNum) {
            n = 0;
        }

	return allowedModes[n];
    }

    /**
     * Set the input mode.
     *
     * @param mode The new input mode to use
     * @return boolean True, if the mode was valid
     */
    protected boolean setInputMode(int mode) {
	switch (mode) {
	case IM_ROMAN_CAPS:
	    keyMap = upperRomanKeyMap;
	    break;
	case IM_ROMAN_SMALL:
	    keyMap = lowerRomanKeyMap;
	    break;
	case IM_NUMERIC:
	    keyMap = (constraint == TextField.PHONENUMBER) ?
                phoneNumericKeyMap : numericKeyMap;
	    break;
	case IM_SYMBOL:
            st.invokeSYM();
	    break;
	case IM_NONE:
	    break;
        default:
            return false;
	}

        imc.setInputMode(inputMode = mode);

        return true;
    }

    /**
     * Establish the constraints of this handler.
     */
    protected void setConstraint() {
        if (constraint == TextField.NUMERIC ||
            constraint == TextField.PHONENUMBER) {
            defaultMode = IM_NUMERIC;
            allowedModesNum = 1;
            allowedModes[0] = IM_NUMERIC;
        } else if (constraint == TextField.EMAILADDR ||
            constraint == TextField.URL) {
            defaultMode = IM_ROMAN_SMALL;
            allowedModesNum = 4;
            allowedModes[0] = IM_ROMAN_SMALL;
            allowedModes[1] = IM_ROMAN_CAPS;
            allowedModes[2] = IM_NUMERIC;
            allowedModes[3] = IM_SYMBOL;
        } else {
            String dMode = null; // default mode
            String aMode[] = null;  // allowed modes

            if (constraint == TextField.ANY) {
                Object obj = imc.getClientObject();
                if (obj instanceof TextBox) {
                    dMode = InputMode.getMode((TextBox)obj);
                    aMode = InputMode.getAllowedModes((TextBox)obj);
                } else if (obj instanceof TextField) {
                    dMode = InputMode.getMode((TextField)obj);
                    aMode = InputMode.getAllowedModes((TextField)obj);
                }
            }
            
            allowedModesNum = 0;

            if (dMode == null && aMode == null) {
                for (int i = 0; i < inputModeConvTable.length; i++) {
                    allowedModes[allowedModesNum++] = 
                        ((Integer)inputModeConvTable[i][1]).intValue();
                }
                defaultMode = allowedModes[0];
            } else {
                if (dMode == null && aMode != null) {
                    dMode = aMode[0];
                }
                if (dMode != null && aMode == null) {
                    aMode = new String[supportedInputModes.length];
                    System.arraycopy(supportedInputModes, 0, 
                                     aMode, 0, supportedInputModes.length);
                }
                
                int i, j;

                for (i = 0; i < inputModeConvTable.length; i++) {
                    if (((String)inputModeConvTable[i][0]).
                        equals(dMode)) {
                        defaultMode = 
                            ((Integer)inputModeConvTable[i][1]).intValue();
                        break;
                    }
                }

                for (j = 0; j < inputModeConvTable.length; j++) {
                    for (i = 0; i < aMode.length; i++) {
                        if (((String)inputModeConvTable[j][0]).
                            equals(aMode[i])) {
                            allowedModes[allowedModesNum++] = 
                                ((Integer)inputModeConvTable[j][1]).intValue();
                        }
                    }
                }

                if (defaultMode == IM_NONE) {
                    defaultMode = allowedModes[0];
                } else {
                    if (defaultMode != allowedModes[0]) {
                        for (i = 0; i < allowedModesNum; i++) {
                            if (defaultMode == allowedModes[i]) {
                                break;
                            }
                        }
                        int t = allowedModes[0];
                        allowedModes[i] = t;
                        allowedModes[0] = defaultMode;
                    }
                }

            }
        }
    }

    /**
     * A special Canvas to display a symbol table.
     */
    protected class SymbolTable extends Canvas {
        /** The margin size */
        private final int MARGIN = 1;
        /** The margin size */
        private final int DMARGIN = 2;
        /** Cell size */
        private int cc;
        /** Height margin */
        private int hmargin;
        /** Width margin */
        private int wmargin;
        /** Margin for the cursor */
        private int margin;
        /** Window x position */
        private int wx;
        /** Window y position */
        private int wy;
        /** Window width */
        private int ww;
        /** Window height */
        private int wh;
        /** Number of columns */
        private int cols;
        /** Number of rows */
        private int rows;
        /** Current cursor position */
        private int pos;
        /** New cursor position */
        private int newpos;
        /** Font */
        private Font font;
        /** Flag */
        private boolean firstTime = false;
        /** The current Display object */
        private Display currentDisplay;
        /** The previous Displayable */
        private Displayable previousScreen;

        /**
         * Initialize the symbol table
         */
        void init() {
            if (symbolTableChars.length <= 25) {
                rows = cols = 5;
            } else {
                rows = 6;
                cols = 6;
            }

            int w = getWidth() / cols;
            int h = getHeight() / rows;

            cc = (w > h) ? h : w;

            int cw = 0, ch = 0;
            int[] fs = {Font.SIZE_LARGE, Font.SIZE_MEDIUM, Font.SIZE_SMALL};
            for (int i = 0; i < fs.length; i++) {
                font = Font.getFont(Font.FACE_SYSTEM,
                                    Font.STYLE_BOLD,
                                    fs[i]);
                cw = font.charWidth('M');
                ch = font.getHeight();
                if (cw <= cc && ch <= cc) {
                    break;
                }
            }

            ww = cols * cc;
            wh = rows * cc;

            wx = (getWidth() - ww) / 2;
            wy = getHeight() - wh;

            hmargin = (cc - ch) / 2;
            wmargin = cc / 2;
            margin = hmargin + 2;
        }

        /**
         * Invoke this symbol table.
         */
        public void invokeSYM() {
            if (font == null) {
                init();
            }
	    currentDisplay = imc.getDisplay();
            previousScreen = currentDisplay.getCurrent();
            currentDisplay.setCurrent(this);
            hasSymbolTable = true;
        }

        /**
         * Notify this symbol table its being shown on the screen.
	 * Overrides Canvas.showNotify.
         */
        protected void showNotify() {
            pos = newpos = defaultSymbolCursorPos;
        }

        /**
         * Notify this symbol table its being hidden.
	 * Overrides Canvas.hideNotify.
         */
        protected void hideNotify() {
            firstTime = false;
        }

        /**
         * Paint this symbol table.
	 * Overrides Canvas.paint.
         *
         * @param g The Graphics object to paint to
         */
        protected void paint(Graphics g) {
            if (!firstTime) {
                paintPanel(g);
                firstTime = true;
            }
            showCursor(g, pos, false);
            showCursor(g, pos = newpos, true);
        }

        /**
         * Paint the symbol table panel
         *
         * @param g The Graphics object to paint to
         */
        void paintPanel(Graphics g) {
            g.setFont(font);

            g.setGrayScale(255);
            g.setClip(0, 0, getWidth(), getHeight());
            g.fillRect(0, 0, getWidth(), getHeight());

            g.setGrayScale(0);
            g.drawRect(wx+1, wy+1, ww-2, wh-2);

            for (int r = 0; r < rows; r++) {
                for (int c = 0; c < cols; c++) {
                    int i = r * cols + c;
                    if (i == symbolTableChars.length) {
                        break;
                    }
                    drawChar(g, symbolTableChars[i], r, c, false);
                }
            }
        }

        /**
         * Draw a character
         *
         * @param g The Graphics object to paint to
         * @param c The character to draw
         * @param row The row the character is in
         * @param col The column the character is in
         * @param reverse A flag to draw the character in inverse
         */
        void drawChar(Graphics g, char c, int row, int col, boolean reverse) {
            int h = font.charWidth(c);
            int y = wy + row * cc + hmargin;
            int x = wx + col * cc + wmargin;

            g.setFont(font);

            if (reverse) {
                g.setGrayScale(255);
            } else {
                g.setGrayScale(0);
            }

            g.drawChar(c, x, y, Graphics.HCENTER | Graphics.TOP);
        }

        /**
         * Show the cursor of this symbol table
         *
         * @param g The Graphics object to paint to
         * @param pos The position of the cursor
         * @param show A flag indicating the visibility of the cursor
         */
        void showCursor(Graphics g, int pos, boolean show) {
            int row = pos / cols;
            int col = pos % cols;
            int y = wy + row * cc;
            int x = wx + col * cc;

            if (show) {
                g.setGrayScale(0);
            } else {
                g.setGrayScale(255);
            }
            g.fillRect(x+margin, y+margin, cc-margin-1, cc-margin-1);

            drawChar(g, symbolTableChars[pos], row, col, show);
        }

        /**
         * Handle a key press event on this symbol table.
	 * Overrides Canvas.keyPressed.
         *
         * @param keyCode The key that was pressed
         */
        protected void keyPressed(int keyCode) {
            if ((keyCode > 0) && (((char)keyCode) == '*')) {
                currentDisplay.setCurrent(previousScreen);
            } else {
                switch (getGameAction(keyCode)) {
                case Canvas.RIGHT:
                    if ((pos + 1) < symbolTableChars.length) {
                        newpos = pos + 1;
                        repaint();
                    }
                    break;

                case Canvas.LEFT:
                    if (pos > 0) {
                        newpos = pos - 1;
                        repaint();
                    }
                    break;

                case Canvas.UP: {
                    int p = pos - cols;
                    if (p >= 0) {
                        newpos = p;
                        repaint();
                    }
                    break;
                }

                case Canvas.DOWN: {
                    int p = pos + cols;
                    if (p < symbolTableChars.length) {
                        newpos = p;
                        repaint();
                    }
                    break;
                }

                case Canvas.FIRE:
                    buf[0] = symbolTableChars[pos];
                    bufLen = 1;
                    committedLen = 1;
                    currentDisplay.setCurrent(previousScreen);
                    break;
                }
            }
        }
    }
}
