/**
 *
 * @(#)Installer.java	1.31 01/08/22
 *
 * Copyright 2001 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 */

package com.sun.midp.midletsuite;

import java.util.Vector;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.ByteArrayOutputStream;
import java.io.ByteArrayInputStream;
import java.io.DataOutputStream;
import java.io.DataInputStream;

import java.lang.String;
import java.lang.IllegalArgumentException;

import javax.microedition.io.Connector;
import javax.microedition.io.Connection;
import javax.microedition.io.HttpConnection;
import javax.microedition.io.ConnectionNotFoundException;

import com.sun.midp.security.SecurityDomain;
import com.sun.midp.security.Actions;

import com.sun.midp.Configuration;
import com.sun.midp.midlet.Scheduler;
import com.sun.midp.midlet.MIDletSuite;

import com.sun.midp.io.Base64;
import com.sun.midp.io.Properties;
import com.sun.midp.io.HttpUrl;

import com.sun.midp.io.j2me.storage.RandomAccessStream;
import com.sun.midp.io.j2me.storage.File;

/**
 * An Installer manages MIDlet suites and libraries
 * present in a Java application environment.  An MIDlet suite
 * or extension library is distributed as a descriptor and JAR pair.
 * The descriptor identifies the configuration and contains security
 * information and the manifest of the JAR describes the contents.
 * The implementation of an Installer is
 * specific to the platform and provides access to 
 * procedures that make an MIDlet suite visible to users. 
 * <P>
 * Each installed package is uniquely identified by a storage name
 * constructed from the combination
 * of the values of the <code>MIDlet-Name</code> and
 * <code>MIDlet-Vendor</code> attributes.
 * The syntax and content of the strings used to identify
 * installed packages are implementation dependent.
 * Only packages installed or upgraded using this API appear 
 * in the list of known packages.
 *
 */
public class Installer {
    /**
     * Status code for local writing of the verified MIDlet suite.
     * Stopping the install at this point has no effect, so there user
     * should not be given a chance to stop the install.
     */
    public static final int STORING_SUITE = 1;

    /** Filename of Manifest inside the application archive. */
    public static final String JAR_MANIFEST       = "META-INF/MANIFEST.MF";

    /** MIDlet property for the size of the application data. */
    public static final String DATA_SIZE_PROP     = "MIDlet-Data-Size";

    /** MIDlet property for the size of the application archive. */
    public static final String JAR_SIZE_PROP      = "MIDlet-Jar-Size";

    /** MIDlet property for the application archive URL. */
    public static final String JAR_URL_PROP       = "MIDlet-Jar-URL";

    /** MIDlet property for the suite name. */
    public static final String SUITE_NAME_PROP    = "MIDlet-Name";

    /** MIDlet property for the suite vendor. */
    public static final String VENDOR_PROP        = "MIDlet-Vendor";

    /** MIDlet property for the suite version. */
    public static final String VERSION_PROP       = "MIDlet-Version";

    /** MIDlet property for the install notify URL. */
    public static final String NOTIFY_PROP        = "MIDlet-Install-Notify";

    /** MIDlet property for the microedition configuration. */
    public static final String CONFIGURATION_PROP =
        "MicroEdition-Configuration";

    /** MIDlet property for the profile. */
    public static final String PROFILE_PROP       = "MicroEdition-Profile";

    /** Max number of bytes to download at one time. */
    private static final int MAX_DL_SIZE = 1024;

    /** Tag that signals that the HTTP server supports basic auth. */
    private static final String BASIC_TAG = "basic";

    /** Tag that marks the version in cookie. */
    private static final String VERSION_TAG = "version";

    /** Tag that marks the domain in cookie. */
    private static final String DOMAIN_TAG = "domain";

    /** Tag that marks the path in cookie. */
    private static final String PATH_TAG = "path";

    /** Media-Type for valid application descriptor files. */
    static final String JAD_MT = "text/vnd.sun.j2me.app-descriptor";

    /** Media-Type for valid Jar file. */
    static final String JAR_MT_1 = "application/java";

    /** Media-Type for valid Jar file. */
    static final String JAR_MT_2 = "application/java-archive";

    /** Filename to save the domain owner of the suite. */
    static final String DOMAIN_OWNER_FILENAME  = "domainOwner.utf";

    /** Filename to save the URL of the JAD. */
    static final String JAD_URL_FILENAME  = "jadUrl.utf";

    /** Filename to save the encoding of the JAD. */
    static final String JAD_ENCODING_FILENAME  = "jadEncoding.utf";

    /** Filename to suite JAR. */
    static final String JAR_FILENAME = "suite.jar";

    /** Filename to save the suite application descriptor. */
    static final String JAD_FILENAME = "suite.jad";

    /**
     * Filename to save the JAR of the suite temporarily. This is used
     * to avoid overwriting an existing JAR prior to
     * verification.
     */
    static final String TMP_FILENAME      = "installer.tmp";

    /** Filename to save the application manifest. */
    static final String MANIFEST_FILENAME = "suite.mf";

    /** Filename of the list of installed suites. */
    static final String SUITE_LIST_FILENAME = "suites.utf";

    /** Success message for the suite provider. */
    static final String SUCCESS_MSG = "900 Success";

    /** Error message for the suite provider. */
    static final String INSUFFICIENT_MEM_MSG = "901 Insufficient Memory";

    /** Error message for the suite provider. */
    static final String USER_CANCELLED_MSG = "902 User Cancelled";

    /** Error message for the suite provider. */
    static final String JAR_SIZE_MISMATCH_MSG = "904 JAR Size Mismatch";

    /** Error message for the suite provider. */
    static final String ATTRIBUTE_MISMATCH_MSG = "905 Attribute Mismatch";

    /** Error message for the suite provider. */
    static final String INVALID_JAD_MSG = "906 Invalid Descriptor";

    /** Error message for the suite provider. */
    static final String INVALID_JAR_MSG = "907 Invalid JAR";

    /** Error message for the suite provider. */
    static final String INCOMPATIBLE_MSG =
        "908 Incompatible Configration or Profile";

    /** Private reference to the singleton Installer class. */
    private static Installer myInstaller;

    /** This class has a different security domain than the MIDlet suite. */
    private static SecurityDomain classSecurityDomain;

    /** The unique storage name of the next MIDlet suite to run. */
    private String nextMidletSuiteToRun;

    /** The of the next MIDlet to run. */
    private String nextMidletToRun; 

    /** Cache of installed MIDlet suites. */
    private Vector midletSuiteList;

    /** HTTP connection to close when we stop the installation. */
    private HttpConnection httpConnection;

    /** HTTP stream to close when we stop the installation. */
    private InputStream httpInputStream;

    /** Receives warnings and status. */
    private InstallListener listener;

    /** Holds the install state. */
    private InstallStateImpl state;

    /**
     * Initializes the security domain for this class, so it can
     * perform actions that a normal MIDlet Suite cannot.
     *
     * @param theDomain SecurityDomain for this class.
     */
    public static void initSecurityDomain(SecurityDomain theDomain) {
        if (classSecurityDomain != null) {
            return;
        }

        classSecurityDomain = theDomain;
        MIDletSuiteImpl.initSecurityDomain(classSecurityDomain);
    }

    /**
     * Return a reference to the singleton Installer object.
     *
     * @return the installer reference.
     *
     * @exception SecurityException if the caller does not have permission
     *   to install software.
     */
    public static Installer getInstaller() throws SecurityException {
        MIDletSuite midletSuite = Scheduler.getScheduler().getMIDletSuite();

        // if a MIDlet suite is not scheduled, assume the JAM is calling.
        if (midletSuite != null) {
            midletSuite.checkIfPermitted(Actions.LIFECYCLE_MANAGEMENT);
        }

        return getInstallerCommon();
    }

    /**
     * Return a reference to the singleton Installer object.
     *
     * @param securityDomain security domain of the calling class
     *
     * @return the installer reference.
     *
     * @exception SecurityException if the caller does not have permission
     *   to install software.
     */
    public static Installer getInstaller(SecurityDomain securityDomain)
           throws SecurityException {
        securityDomain.checkIfPermitted(Actions.LIFECYCLE_MANAGEMENT);

        return getInstallerCommon();
    }

    /**
     * Return a reference to the singleton Installer object.
     *
     * @return the installer reference.
     */
    private static Installer getInstallerCommon() {
        String nameOfInstaller;

        if (myInstaller != null) {
            return myInstaller;
        }

        /*
         * In internal.config an alternate installer, such as a
         * SecureInstaller can be specified.
         */
        nameOfInstaller =
            Configuration.getProperty("com.sun.midp.midletsuite.installer");
        if (nameOfInstaller == null) {
            myInstaller = new Installer();
            return myInstaller;
        }
        
        try {
            myInstaller = 
                (Installer)Class.forName(nameOfInstaller).newInstance();
            return myInstaller;
        } catch (Exception e) {
            throw new RuntimeException(e.getMessage());
        }
    }

    /**
     * This method returns a storage identifier of an application.
     *
     * @param vendorName name of the vendor that created the application, as
     *          given in a JAD file.
     * @param appName name of the application, as given in a JAD file.
     *
     * @return the platform-specific storage name of the application
     *          given by vendorName and appName.
     */
    public static String makeStorageName(String vendorName, String appName) {
        return File.unicodeToAsciiFilename(vendorName) + '_' +
            File.unicodeToAsciiFilename(appName) + '_';
    }

    /**
     * Constructor to prevent a second instantiation of the Installer.
     */
    protected Installer() {
        if (myInstaller != null) {
            throw new
                SecurityException("Illegal instantiation of the Installer");
        }
    }

    /**
     * Install a software package from the given URL. The URL MUST
     * refer to an application descriptor.
     * <p>
     * If the component to be installed is the same as an existing
     * component (by comparing the <code>MIDlet-Name</code>,
     * <code>MIDlet-Vendor</code> attributes)
     * then this install is an upgrade if the version number is greater 
     * than the current version.  If so, the new version replaces in its
     * entirety the current version.
     * <p>
     * It is implementation dependent when the upgraded component is
     * made available for use.
     * <p>
     * The implementation of install must be robust in the presence
     * of failures such as running out of memory.  If this method
     * throws an exception then the package must <em>not</em> be installed
     * and any previous version of the component must be left intact
     * and operational.
     * <p>
     * To receive status updates and installer warnings, provide an install
     * listener. If no listener is provided all warnings will be thrown
     * as exceptions.
     *
     * @param location the URL from which the application descriptor can be
     *        updated.
     * @param force if <code>true</code> the MIDlet suite components to be
     *              installed will overwrite any existing components without
     *              any version comparison.
     * @param installListener object to receive status updates and install
     *     warnings, can be null
     *
     * @return the unique name of the installed package. 
     *
     * @exception ConnectionNotFoundException if JAD URL is invalid
     * @exception IOException is thrown if any error prevents the installation
     *   of the MIDlet suite, including being unable to access the application
     *   descriptor or JAR.
     * @exception InvalidJadException if the downloaded application descriptor
     *   is invalid.
     * @exception SecurityException if the caller does not have permission
     *   to install software.
     */
    public String install(String location, boolean force,
            InstallListener installListener)
            throws IOException, InvalidJadException, SecurityException {

        state = new InstallStateImpl();

        state.jadUrl = location;
        state.force = force;
        state.nextStep = 1;

        listener = installListener;

        return performInstall();
    }

    /**
     * Same as {@link #install(String, boolean, InstallListener)} but starting
     * with a downloaded JAD and other associated information.
     *
     * @param location the URL from which the application descriptor can be
     *        updated.
     * @param jad  application descriptor to install from
     * @param jadEncoding character encoding for the JAD or null for the
     *                    default encoding
     * @param jadCookie cookie that came with the JAD, can be null
     *                  will be sent with the JAR request
     * @param username for HTTP authentication, can be null
     * @param password for HTTP authentication, can be null if username is null
     * @param force if <code>true</code> the MIDlet suite components to be
     *              installed will overwrite any existing components without
     *              any version comparison.
     * @param installListener object to receive status updates and install
     *     warnings, can be null
     *
     * @return the unique name of the installed package. 
     *
     * @exception IOException is thrown if any error prevents the installation
     *   of the MIDlet suite, including being unable to access the application
     *   descriptor or JAR.
     * @exception InvalidJadException if the downloaded application descriptor
     *   is invalid.
     * @exception SecurityException if the caller does not have permission
     *   to install software.
     */
    public String install(String location, byte[] jad,
            String jadEncoding, String jadCookie, String username,
            String password, boolean force, InstallListener installListener)
            throws IOException, InvalidJadException, SecurityException {

        state = new InstallStateImpl();

        state.jadUrl = location;
        state.force = force;
        state.nextStep = 2;

        state.jad = jad;
        state.jadEncoding = jadEncoding;
        state.cookie = jadCookie;
        state.username = username;
        state.password = password;

        listener = installListener;

        return performInstall();
    }

    /**
     * Same as {@link #install(String, boolean, InstallListener)}
     * but warnings are converted to exceptions.
     *
     * @param location the URL from which the application descriptor can be
     *     retrieved.
     * @param force if <code>true</code> the MIDlet suite components to be
     *              installed will overwrite any existing components without
     *              any version comparison.
     *
     * @return the unique name of the installed package. 
     *
     * @exception IOException is thrown if any error prevents the installation
     *   of the MIDlet suite, including being unable to access the application
     *   descriptor or JAR.
     * @exception InvalidJadException if the downloaded application descriptor
     *   is invalid.
     * @exception SecurityException if the caller does not have permission
     *   to install software.
     */
    public String install(String location, boolean force)
            throws IOException, InvalidJadException, SecurityException {
        return install(location, force, null);
    }

    /**
     * Perform an install.
     *
     * @return the unique name of the installed package. 
     *
     * @exception IOException is thrown, if an I/O error occurs during
     * descriptor or jar file download
     * @exception InvalidJadException is thrown, if the descriptor file is not 
     * properly formatted or does not contain the required 
     * information.
     */
    private synchronized String performInstall()
            throws IOException, InvalidJadException {

        state.startTime = System.currentTimeMillis();

        while (state.nextStep < 6) {
            /*
             * clear the previous warning, so we can tell if another has
             * happened
             */
            state.exception = null;

            if (state.stopInstallation) {
                postStatusBackToProvider(USER_CANCELLED_MSG);
                throw new IOException("stopped");
            }

            switch (state.nextStep) {
            case 1:
                installStep1();
                break;

            case 2:
                installStep2();
                break;

            case 3:
                installStep3();
                break;

            case 4:
                installStep4();
                break;

            case 5:
                installStep5();
                break;
            }

            if (state.exception != null) {
                if (listener == null) {
                    throw state.exception;
                }

                if (!listener.warnUser(state)) {
                    state.stopInstallation = true;
                    postStatusBackToProvider(USER_CANCELLED_MSG);
                    throw state.exception;
                }
            }
        }

        return state.storageName;
    }

    /**
     * Download the JAD,
     * save it and any cookies in the install state.
     *
     * @exception IOException is thrown, if an I/O error occurs during
     * descriptor or jar file download
     * @exception InvalidJadException is thrown, if the descriptor file is not 
     * properly formatted or does not contain the required 
     */
    private void installStep1() throws IOException, InvalidJadException {
        byte[] jad;

        if (state.jadUrl == null || state.jadUrl.length() == 0) {
            throw new IOException("Must specify URL of .jad file");
        }

        try {
            jad = downloadJAD();

            state.jad = jad;
            state.nextStep++;
            return;
        } catch (InvalidJadException ije) {
            if (ije.getReason() == InvalidJadException.UNAUTHORIZED) {
                state.exception = ije;
                return;
            }

            throw ije;
        }
    }

    /**
     * Parse the JAD, make sure it has
     * the required properties, and save them in the install state.
     *
     * @exception IOException is thrown, if an I/O error occurs during
     * descriptor or jar file download
     * @exception InvalidJadException is thrown, if the descriptor file is not 
     * properly formatted or does not contain the required 
     */
    private void installStep2()
            throws IOException, InvalidJadException {
        MIDletSuite midletSuite;
        int cmpResult;

        state.jadProps = new JadProperties();
        try {
            parseJad(new ByteArrayInputStream(state.jad), state.jadEncoding,
                state.jadProps);
        } catch (InvalidJadException ije) {
            postStatusBackToProvider(INVALID_JAD_MSG);
            throw ije;
        }

        state.suiteName = state.jadProps.getProperty(SUITE_NAME_PROP);
        if (state.suiteName == null || state.suiteName.length() == 0) {
            postStatusBackToProvider(INVALID_JAD_MSG);
            throw new
                InvalidJadException(InvalidJadException.MISSING_SUITE_NAME);
        }

        state.vendor = state.jadProps.getProperty(VENDOR_PROP);
        if (state.vendor == null || state.vendor.length() == 0) {
            postStatusBackToProvider(INVALID_JAD_MSG);
            throw new
                InvalidJadException(InvalidJadException.MISSING_VENDOR);
        }

        state.version = state.jadProps.getProperty(VERSION_PROP);
        if (state.version == null || state.version.length() == 0) {
            postStatusBackToProvider(INVALID_JAD_MSG);
            throw new
                InvalidJadException(InvalidJadException.MISSING_VERSION);
        }

        // Check if app already exists
        state.storageName = makeStorageName(state.vendor, state.suiteName);
        midletSuite = getMIDletSuite(state.storageName);

        try {
            checkVersionFormat(state.version);
            state.nextStep++;

            if (midletSuite == null) {
                // there is not previous version
                return;
            }

            state.isPreviousVersion = true;

            // get this now so we do not have to save the whole suite object
            state.previousJadUrl = midletSuite.getJadUrl();

            // If it does, check version information
            cmpResult = vercmp(state.version,
                               midletSuite.getProperty(VERSION_PROP));

            if (cmpResult < 0) {
                // never install an older version, stop install
                throw new
                    InvalidJadException(InvalidJadException.OLD_VERSION);
            }

            // Only check version if we aren't forcing an install (update)
            if (state.force) {
                return;
            }

            if (cmpResult > 0) {
                // new version, warn user
                state.exception = new InvalidJadException(
                                  InvalidJadException.NEW_VERSION);
                return;
            }

            // already installed, warn user
            state.exception = new InvalidJadException(
                              InvalidJadException.ALREADY_INSTALLED);
            return;
        } catch (NumberFormatException nfe) {
            postStatusBackToProvider(INVALID_JAD_MSG);
            throw new
                InvalidJadException(InvalidJadException.INVALID_VERSION);
        }
    }

    /**
     * Verify the JAD, and if the
     * JAD belongs to an installed suite, check the URL against the
     * installed one.
     *
     * @exception IOException is thrown, if an I/O error occurs during
     * descriptor or jar file download
     * @exception InvalidJadException is thrown, if the descriptor file is not 
     * properly formatted or does not contain the required 
     */
    private void installStep3()
            throws IOException, InvalidJadException {
        try {
            state.domainOwner = verifyJad(state.jad, state.jadProps);
        } catch (InvalidJadException ije) {
            postStatusBackToProvider(INVALID_JAD_MSG);
            throw ije;
        }

        state.nextStep++;

        /*
         * the presence of a domain owner means a secure installer
         * subclass verified this JAD's signature. So checking the URL is
         * a should not be done.
         */

        // perform a domain check not a straight compare
        if (state.domainOwner == null && state.previousJadUrl != null) {
            HttpUrl old = new HttpUrl(state.previousJadUrl);
            HttpUrl current = new HttpUrl(state.jadUrl);

            if ((current.domain != null && old.domain == null) ||
                (current.domain == null && old.domain != null) ||
                (current.domain != null && old.domain != null &&
                 !current.domain.regionMatches(true, 0, old.domain, 0,
                                           old.domain.length()))) {
                /*
                 * The jad is at new location, could be bad,
                 * let the user decide
                 */
                state.exception = new InvalidJadException(
                                  InvalidJadException.JAD_MOVED,
                                  (state.previousJadUrl == null ? "none" :
                                  state.previousJadUrl));
                return;
            }
        }
    }

    /**
     * Make sure the suite can fit in storage.
     *
     * @exception IOException is thrown, if an I/O error occurs during
     * descriptor or jar file download
     * @exception InvalidJadException is thrown, if the descriptor file is not 
     * properly formatted or does not contain the required 
     */
    private void installStep4()
            throws IOException, InvalidJadException {
        String sizeString;
        int dataSize;
        int suiteSize;

        sizeString = state.jadProps.getProperty(JAR_SIZE_PROP);
        if (sizeString == null) {
            postStatusBackToProvider(INVALID_JAD_MSG);
            throw new
                InvalidJadException(InvalidJadException.MISSING_JAR_SIZE);
        }
        
        try {
            state.expectedJarSize = Integer.parseInt(sizeString);
        } catch (NumberFormatException e) {
            postStatusBackToProvider(INVALID_JAD_MSG);
            throw new
                InvalidJadException(InvalidJadException.INVALID_VALUE);
        }

        sizeString = state.jadProps.getProperty(DATA_SIZE_PROP);
        if (sizeString == null) {
            dataSize = 0;
        } else {
            try {
                dataSize = Integer.parseInt(sizeString);
            } catch (NumberFormatException e) {
                postStatusBackToProvider(INVALID_JAD_MSG);
                throw new
                    InvalidJadException(InvalidJadException.INVALID_VALUE);
            }
        }

        /*
         * A suite is a jad + jar + manifest + url + data size.
         * lets say the manifest is the same size as the jad
         * since we do know at this point. the size is in bytes,
         * UTF-8 chars can be upto 3 bytes
         */
        suiteSize = state.expectedJarSize + (state.jad.length * 2) +
                    (state.jadUrl.length() * 3) + dataSize;
        
        state.file = new File(classSecurityDomain);

        if (suiteSize > state.file.getBytesAvailableForFiles()) {
            postStatusBackToProvider(INSUFFICIENT_MEM_MSG);

            // the size reported to the user should be in K and rounded up
            throw new
                InvalidJadException(InvalidJadException.INSUFFICIENT_STORAGE,
                    Integer.toString((suiteSize + 1023)/ 1024));
        }

        state.nextStep++;
    }

    /**
     * Download the JAR, make sure it is the correct size, make sure
     * the required attributes match the JAD's, and then store the
     * application.
     *
     * @exception IOException is thrown, if an I/O error occurs during
     * descriptor or jar file download
     * @exception InvalidJadException is thrown, if the descriptor file is not 
     * properly formatted or does not contain the required 
     */
    private void installStep5()
            throws IOException, InvalidJadException {
        String jarUrl;
        String tempFilename;
        int bytesDownloaded;
        RandomAccessStream myStorage = null;
        byte[] manifest;
        ManifestProperties jarProps;
        String configuration;
        String temp;
        String profile;
        String storageRoot;
        DataOutputStream storageStream;

        // Save jar file to temp name; we need to do this to read the
        // manifest entry, but, don't want to overwrite an exisiting
        // application in case there are problems with the manifest
        jarUrl = state.jadProps.getProperty(JAR_URL_PROP);
        if (jarUrl == null || jarUrl.length() == 0) {
            postStatusBackToProvider(INVALID_JAD_MSG);
            throw new
                InvalidJadException(InvalidJadException.MISSING_JAR_URL);
        }

        tempFilename = File.getStorageRoot() + TMP_FILENAME;

        try {
            bytesDownloaded = downloadJAR(jarUrl, tempFilename);
        } catch (InvalidJadException ije) {
            if (ije.getReason() == InvalidJadException.UNAUTHORIZED) {
                state.exception = ije;
                return;
            }

            throw ije;
        }

        try {
            if (bytesDownloaded != state.expectedJarSize) {
                postStatusBackToProvider(JAR_SIZE_MISMATCH_MSG);
                throw new
                    InvalidJadException(InvalidJadException.JAR_SIZE_MISMATCH);
            }

            myStorage = new RandomAccessStream(classSecurityDomain);

            /*
             * if the domain owner is null then we do not have to verify
             * the JAR
             */
            if (state.domainOwner != null) {
                verifyJar(state.jadProps, myStorage, tempFilename);
            }

            // Create JAR Properties (From .jar file's MANIFEST
            try {
                manifest = JarReader.readJarEntry(classSecurityDomain,
                                                  tempFilename, JAR_MANIFEST);
            } catch (IOException ioe) {
                postStatusBackToProvider(INVALID_JAR_MSG);
                throw new
                    InvalidJadException(InvalidJadException.CORRUPT_JAR);
            }

            jarProps = new ManifestProperties();
            jarProps.load(new ByteArrayInputStream(manifest));

            configuration = jarProps.getProperty(CONFIGURATION_PROP);
            if (configuration == null || configuration.length() == 0) {
                postStatusBackToProvider(INVALID_JAR_MSG);
                throw new InvalidJadException(
                    InvalidJadException.MISSING_CONFIGURATION);
            }

            // JAD overrides the manifest
            temp = state.jadProps.getProperty(CONFIGURATION_PROP);
            if (temp != null) {
                configuration = temp;
            }

            /*
             * Since there is only one version to support, do a simple compare
             * of a harded coded value, so this code will break in
             * later versions. In later versions the configuration file
             * should be read and parsed and compared to the parsed
             * application value.
             */
            if (!configuration.equals("CLDC-1.0")) {
                postStatusBackToProvider(INCOMPATIBLE_MSG);
                throw new InvalidJadException(
                    InvalidJadException.DEVICE_INCOMPATIBLE);
            }

            profile = jarProps.getProperty(PROFILE_PROP);
            if (profile == null || profile.length() == 0) {
                postStatusBackToProvider(INVALID_JAR_MSG);
                throw new
                    InvalidJadException(InvalidJadException.MISSING_PROFILE);
            }

            // JAD overrides the manifest
            temp = state.jadProps.getProperty(PROFILE_PROP);
            if (temp != null) {
                profile = temp;
            }

            /*
             * Since there is only one version to support, do a simple compare
             * of a harded coded value, so this code will break in
             * later versions. In later versions the configuration file
             * should be read and parsed and compared to the parsed
             * application value.
             */
            if (!profile.equals("MIDP-1.0")) {
                postStatusBackToProvider(INCOMPATIBLE_MSG);
                throw new InvalidJadException(
                    InvalidJadException.DEVICE_INCOMPATIBLE);
            }

            // Check Manifest entries against .jad file
            if (!state.suiteName.equals(
                    jarProps.getProperty(SUITE_NAME_PROP))) {
                postStatusBackToProvider(ATTRIBUTE_MISMATCH_MSG);
                throw new InvalidJadException(
                    InvalidJadException.SUITE_NAME_MISMATCH);
            }

            if (!state.version.equals(jarProps.getProperty(VERSION_PROP))) {
                postStatusBackToProvider(ATTRIBUTE_MISMATCH_MSG);
                throw new
                    InvalidJadException(InvalidJadException.VERSION_MISMATCH);
            }

            if (!state.vendor.equals(jarProps.getProperty(VENDOR_PROP))) {
                postStatusBackToProvider(ATTRIBUTE_MISMATCH_MSG);
                throw new
                    InvalidJadException(InvalidJadException.VENDOR_MISMATCH);
            }

            storageRoot = File.getStorageRoot() + state.storageName;

            // make sure at least 1 second has passed
            while (System.currentTimeMillis() - state.startTime < 1000);

            synchronized (state) {
                // this is the point of no return, one last check
                if (state.stopInstallation) {
                    postStatusBackToProvider(USER_CANCELLED_MSG);
                    throw new IOException("stopped");
                }

                if (listener != null) {
                    listener.updateStatus(STORING_SUITE);
                }

                state.writingSuite = true;
            }
        } catch (Exception e) {
            state.file.delete(tempFilename);
            if (e instanceof IOException) {
                throw (IOException)e;
            }

            throw (RuntimeException)e;
        }

        try {
            state.file.rename(tempFilename, storageRoot + JAR_FILENAME);
            
            myStorage.connect(storageRoot + MANIFEST_FILENAME,
                              RandomAccessStream.READ_WRITE_TRUNCATE);
            myStorage.writeBytes(manifest, 0, manifest.length);
            myStorage.disconnect();

            // Save JAD, JAR & manifest files
            myStorage.connect(storageRoot + JAD_FILENAME, 
                              RandomAccessStream.READ_WRITE_TRUNCATE);
            myStorage.writeBytes(state.jad, 0, state.jad.length);
            myStorage.disconnect();

            // convert the JAD URL to UTF8 and write it to storage
            myStorage.connect(storageRoot + JAD_URL_FILENAME,
                              RandomAccessStream.READ_WRITE_TRUNCATE);
            storageStream = myStorage.openDataOutputStream();
            storageStream.writeUTF(state.jadUrl);
            storageStream.close();
            myStorage.disconnect();

            if (state.jadEncoding != null) {
                // convert the JAD encoding to UTF8 and write it to storage
                myStorage.connect(storageRoot + JAD_ENCODING_FILENAME,
                                  RandomAccessStream.READ_WRITE_TRUNCATE);
                storageStream = myStorage.openDataOutputStream();
                storageStream.writeUTF(state.jadEncoding);
                storageStream.close();
                myStorage.disconnect();
            } else if (state.file.exists(storageRoot +
                                         JAD_ENCODING_FILENAME)) {
                // erase the old encoding since we did not get new one
                state.file.delete(storageRoot + JAD_ENCODING_FILENAME);
            }

            // todo get the jad encoding and write to a file
            if (state.domainOwner != null) {
                // convert the domain owner to UTF8 and write it to storage
                myStorage.connect(storageRoot + DOMAIN_OWNER_FILENAME,
                                  RandomAccessStream.READ_WRITE_TRUNCATE);
                storageStream = myStorage.openDataOutputStream();
                storageStream.writeUTF(state.domainOwner);
                storageStream.close();
                myStorage.disconnect();
            }

            addToSuiteList(state.storageName);

            state.nextStep++;

            try {
                postStatusBackToProvider(SUCCESS_MSG);
            } catch (RuntimeException re) {
                /*
                 * The suite is successfully installed, but the post of the
                 * status message failed. Do not let this failure prevent 
                 * the suite from being used.
                 */
            }

            return;
        } catch (Exception e) {
            try {
                myStorage.disconnect();
            } catch (IOException ex) {
                // ignore
            }

            remove(state.storageName);

            if (e instanceof IOException) {
                throw (IOException)e;
            }

            throw (RuntimeException)e;
        }
    }

    /**
     * Download an application descriptor file from the given URL.
     *
     * @return a byte array representation of the file or null if not found.
     *
     * @exception IOException is thrown if any error prevents the download
     *   of the JAD.
     */
    private byte[] downloadJAD() throws IOException {
        String[] encoding = new String[1];
        ByteArrayOutputStream bos = new ByteArrayOutputStream(MAX_DL_SIZE);
        String[] acceptableTypes = {JAD_MT};
        String[] extraFieldKeys = new String[3];
        String[] extraFieldValues = new String[3];
        String locale;

        extraFieldKeys[0] = "User-Agent";
        extraFieldValues[0] = "Profile/" +
                             System.getProperty("microedition.profiles") +
                             " Configuration/" +
                             System.getProperty("microedition.configuration");

        extraFieldKeys[1] = "Accept-Charset";
        extraFieldValues[1] = "UTF-8, " +
                              System.getProperty("microedition.encoding");

        /* locale can be null */
        locale = System.getProperty("microedition.locale");
        if (locale != null) {
            extraFieldKeys[2] = "Accept-Language";
            extraFieldValues[2] = locale;
        }

        downloadResource(state.jadUrl, extraFieldKeys, extraFieldValues,
                         acceptableTypes, bos, encoding,
                         InvalidJadException.JAD_SERVER_NOT_FOUND,
                         InvalidJadException.JAD_NOT_FOUND,
                         InvalidJadException.INVALID_JAD_TYPE);

        state.jadEncoding = encoding[0];
        return bos.toByteArray();
    }

    /**
     * Download an application archive file from the given URL into the
     * given file. Automatically handle re-trys.
     *
     * @param jarUrl the URL from which the JAR can be retrieved.
     *
     * @param filename name of the file to write. This file resides
     *          in the storage area of the given application.
     *
     * @return size of the JAR
     *
     * @exception IOException is thrown if any error prevents the download
     *   of the JAR.
     */
    private int downloadJAR(String jarUrl, String filename)
            throws IOException {
        HttpUrl parsedUrl;
        String url;
        String[] acceptableTypes = {JAR_MT_1, JAR_MT_2};
        int jarSize;

        RandomAccessStream jarOutputStream = null;
        OutputStream outputStream = null;

        parsedUrl = new HttpUrl(jarUrl);
        if (parsedUrl.authority == null && state.jadUrl != null) {
            // relative URL, add the JAD URL as the base
            try {
                parsedUrl.addBaseUrl(state.jadUrl);
            } catch (IOException e) {
                postStatusBackToProvider(INVALID_JAD_MSG);
                throw new InvalidJadException(
                     InvalidJadException.JAR_SERVER_NOT_FOUND);
            }

            url = parsedUrl.toString();
        } else {
            url = jarUrl;
        }

        jarOutputStream = new RandomAccessStream(classSecurityDomain);
        jarOutputStream.connect(filename,
                                RandomAccessStream.READ_WRITE_TRUNCATE);
        outputStream = jarOutputStream.openOutputStream();

        try {
            jarSize = downloadResource(url, null, null,
                         acceptableTypes, outputStream, null,
                         InvalidJadException.JAR_SERVER_NOT_FOUND,
                         InvalidJadException.JAR_NOT_FOUND,
                         InvalidJadException.INVALID_JAR_TYPE);
            return jarSize;
        } catch (InvalidJadException ije) {
            switch (ije.getReason()) {
            case InvalidJadException.JAR_SERVER_NOT_FOUND:
            case InvalidJadException.JAR_NOT_FOUND:
                postStatusBackToProvider(INVALID_JAD_MSG);
                break;

            case InvalidJadException.INVALID_JAR_TYPE:
                postStatusBackToProvider(INVALID_JAR_MSG);
            }

            throw ije;
        } finally {
            try {
                jarOutputStream.disconnect();
            } catch (Exception e) {
                // ignore
            }
        }
    }

    /**
     * Download an resource from the given URL into the output stream.
     *
     * @param url location of the resource to download
     * @param extraFieldKeys keys to the extra fields to put in the request
     * @param extraFieldValues values to the extra fields to put in the request
     * @param acceptableTypes list of acceptable media types for this resource,
     *                        there must be at least one
     * @param output output stream to write the resource to
     * @param encoding an array to receive the character encoding of resource,
     *                 can be null
     * @param serverNotFoundCode reason code to use when the server is not
     *     found
     * @param resourceNotFoundCode reason code to use when the resource is not
     *     found on the server
     * @param invalidMediaTypeCode reason code to use when the media type of
     *     the resource is not valid
     *
     * @return size of the resource
     *
     * @exception IOException is thrown if any error prevents the download
     *   of the resource.
     */
    private int downloadResource(String url, String[] extraFieldKeys,
            String[] extraFieldValues, String[] acceptableTypes,
            OutputStream output, String[] encoding,
            int serverNotFoundCode, int resourceNotFoundCode,
            int invalidMediaTypeCode) throws IOException {
        Connection conn = null;
        StringBuffer acceptField;
        int responseCode;
        String retryAfterField;
        int retryInterval;
        String mediaType;
        String temp;

        try {
            for (; ; ) {
                try {
                    conn = Connector.open(url, Connector.READ);
                } catch (Exception e) {
                    throw new InvalidJadException(serverNotFoundCode);
                }

                if (!(conn instanceof HttpConnection)) {
                    throw new InvalidJadException(serverNotFoundCode);
                }

                httpConnection = (HttpConnection)conn;

                if (extraFieldKeys != null) { 
                    for (int i = 0; i < extraFieldKeys.length &&
                            extraFieldKeys[i] != null; i++) {
                        httpConnection.setRequestProperty(extraFieldKeys[i],
                                                          extraFieldValues[i]);
                    }
                }
                    
                // 256 is given to avoid resizing without adding lengths
                acceptField = new StringBuffer(256);

                // there must be one or more acceptable media types
                acceptField.append(acceptableTypes[0]);
                for (int i = 1; i < acceptableTypes.length; i++) {
                    acceptField.append(", ");
                    acceptField.append(acceptableTypes[i]);
                }

                httpConnection.setRequestProperty("Accept",
                                                  acceptField.toString());

                httpConnection.setRequestMethod(HttpConnection.GET);

                if (state.username != null &&
                        state.password != null) {
                    httpConnection.setRequestProperty("Authorization",
                        formatAuthCredentials(state.username,
                                              state.password));
                }

                if (state.cookie != null) {
                    httpConnection.setRequestProperty("Cookie",
                        formatReturnCookie(state.cookie));
                }

                try {
                    responseCode = httpConnection.getResponseCode();
                } catch (IOException ioe) {
                    throw new InvalidJadException(serverNotFoundCode);
                }

                responseCode = httpConnection.getResponseCode();
                if (responseCode != HttpConnection.HTTP_UNAVAILABLE) {
                    break;
                }

                retryAfterField = httpConnection.getHeaderField("Retry-After");
                if (retryAfterField == null) {
                    break;
                }

                try {
                    /*
                     * see if the retry interval is in seconds, and
                     * not an absolute date
                     */
                    retryInterval = Integer.parseInt(retryAfterField);
                    if (retryInterval > 0) {
                        if (retryInterval > 60) {
                            // only wait 1 min
                            retryInterval = 60;
                        }

                        Thread.sleep(retryInterval * 1000);
                    }
                } catch (InterruptedException ie) {
                    // ignore thread interrupt
                    break;
                } catch (NumberFormatException ne) {
                    // ignore bad format
                    break;
                }

                httpConnection.close();

                if (state.stopInstallation) {
                    postStatusBackToProvider(USER_CANCELLED_MSG);
                    throw new IOException("stopped");
                }
            } // end for

            if (responseCode == HttpConnection.HTTP_NOT_FOUND) {
                throw new InvalidJadException(resourceNotFoundCode);
            }

            if (responseCode == HttpConnection.HTTP_NOT_ACCEPTABLE) {
                throw new InvalidJadException(invalidMediaTypeCode);
            }

            if (responseCode == HttpConnection.HTTP_UNAUTHORIZED) {
                // automatically throws the correct exception
                checkIfBasicAuthSupported(
                     httpConnection.getHeaderField("WWW-Authenticate"));

                throw new
                    InvalidJadException(InvalidJadException.UNAUTHORIZED);
            }

            if (responseCode != HttpConnection.HTTP_OK) {
                throw new IOException("Failed to download " + url +
                    " HTTP response code: " + responseCode);
            }

            mediaType = getMediaType(httpConnection.getType());
            if (mediaType != null) {
                boolean goodType = false;

                for (int i = 0; i < acceptableTypes.length; i++) {
                    if (mediaType.equals(acceptableTypes[i])) {
                        goodType = true;
                        break;
                    }
                }

                if (!goodType) {
                    throw new InvalidJadException(invalidMediaTypeCode,
                                                  mediaType);
                }
            }

            if (encoding != null) {
                encoding[0] = getCharset(httpConnection.getType());
            }

            temp = httpConnection.getHeaderField("Set-Cookie");
            if (temp != null) {
                state.cookie = temp;
            }

            httpInputStream = httpConnection.openInputStream();
            return transferData(httpInputStream, output);
        } finally {
            // Close the streams or connections this method opened.
            try {
                httpInputStream.close();
            } catch (Exception e) {
                // ignore
            }
                    
            try {
                conn.close();
            } catch (Exception e) {
                // ignore
            }
        }
    }

    /**
     * Check to make sure the HTTP server will support Basic authtication.
     *
     * @param wwwAuthField WWW-Authenticate field from the response header
     *
     * @exception InvalidJadException if server does not support Basic
     *                                authentication
     */
    private void checkIfBasicAuthSupported(String wwwAuthField) 
            throws InvalidJadException {
        wwwAuthField = wwwAuthField.trim();

        if (!wwwAuthField.regionMatches(true, 0, BASIC_TAG, 0,
                                        BASIC_TAG.length())) {
            throw new InvalidJadException(InvalidJadException.CANNOT_AUTH);
        }
    }

    /**
     * Parse out the media-type from the content-type field and convert
     * to lower case.
     * The media-type everything for the ';' that marks the parameters.
     *
     * @param contentType value of the content-type field
     *
     * @return media-type in lower case
     */
    private static String getMediaType(String contentType) {
        int semiColon;

        if (contentType == null) {
            return null;
        }

        semiColon = contentType.indexOf(';');
        if (semiColon < 0) {
            return contentType.toLowerCase();
        }

        return contentType.substring(0, semiColon).toLowerCase();
    }

    /**
     * Parse out the charset from the content-type field.
     * The charset parameter is after the ';' in the content-type field.
     *
     * @param contentType value of the content-type field
     *
     * @return charset
     */
    private static String getCharset(String contentType) {
        int start;
        int end;

        if (contentType == null) {
            return null;
        }

        start = contentType.indexOf("charset");
        if (start < 0) {
            return null;
        }

        start = contentType.indexOf('=', start);
        if (start < 0) {
            return null;
        }

        // start past the '='
        start++;

        end = contentType.indexOf(';', start);
        if (end < 0) {
            end = contentType.length();
        }
            
        return contentType.substring(start, end).trim();
    }

    /**
     * Format the username and password for HTTP basic authentication
     * according RFC 2617.
     *
     * @param username for HTTP authentication
     * @param password for HTTP authentication
     *
     * @return properly formated basic authentication credential
     */
    private static String formatAuthCredentials(String username,
                                                String password) {
        byte[] data = new byte[username.length() + password.length() + 1];
        int j = 0;

        for (int i = 0; i < username.length(); i++, j++) {
            data[j] = (byte)username.charAt(i);
        }

        data[j] = (byte)':';
        j++;

        for (int i = 0; i < password.length(); i++, j++) {
            data[j] = (byte)password.charAt(i);
        }

        return "Basic " + Base64.encode(data, 0, data.length);
    }

    /**
     * Format the cookie to be returned to the HTTP server
     * according to RFC 2109.
     *
     * @param cookieField Set-Cookie field from the response header
     *
     * @return Cookie field to send back to the server
     */
    String formatReturnCookie(String cookieField) {
        Vector params = new Vector();
        int i;
        char currentChar;
        String param;
        int start = 0;
        StringBuffer cookie;
        String version = null;
        String domain = null;
        String path = null;
        

        try {
            /*
             * get the parameters from only the first cookie,
             * so upto the first ","
             */
            for (i = 0; i < cookieField.length(); i++) {
                currentChar = cookieField.charAt(i);
                if (currentChar == ',') {
                    break;
                }

                // parameters are separated by a ";" and can have white space
                if (currentChar == ';') {
                    params.addElement(cookieField.substring(start, i).trim());
                    start = i + 1;
                }
            }

            params.addElement(cookieField.substring(start, i).trim());

            cookie = new StringBuffer((String)params.elementAt(0));

            for (i = 1; i < params.size(); i++) {
                param = (String)params.elementAt(i);

                if (param.regionMatches(true, 0, VERSION_TAG, 0,
                                        VERSION_TAG.length())) {
                    version = param;
                    continue;
                }

                if (param.regionMatches(true, 0, DOMAIN_TAG, 0,
                                        DOMAIN_TAG.length())) {
                    domain = param;
                }

                if (param.regionMatches(true, 0, PATH_TAG, 0,
                                        PATH_TAG.length())) {
                    path = param;
                }
            }        
        } finally {}

        cookie.append(";$");
        cookie.append(version);

        if (domain != null) {
            cookie.append(";$");
            cookie.append(domain);
        }

        if (path != null) {
            cookie.append(";$");
            cookie.append(path);
        }

        return cookie.toString();
    }

    /**
     * Post a status message back to the provider's URL in JAD.
     *
     * @param message status message to post
     */
    private void postStatusBackToProvider(String message) {
        String url;
        HttpConnection transaction;
        OutputStream out;

        if (state.jadProps == null) {
            return;
        }

        url = state.jadProps.getProperty(NOTIFY_PROP);
        if (url == null) {
            return;
        }

        try {
            transaction = (HttpConnection)Connector.open(url, Connector.WRITE);
            try {
                transaction.setRequestMethod(HttpConnection.POST);

                if (state.cookie != null) {
                    transaction.setRequestProperty("Cookie",
                        formatReturnCookie(state.cookie));
                }

                out = transaction.openOutputStream();
                try {
                    out.write(message.getBytes());
                } finally {
                    out.close();
                }
            } finally {
                transaction.close();
            }
        } catch (IOException ioe) {
            // ignore IOExceptions
        } catch (IllegalArgumentException iae) {
            // ignore IllegalArgumentExceptions
        } catch (ClassCastException cce) {
            // ignore ClassCastExceptions
        }
    }

    /**
     * Function that actually does the work of transfering file data.
     * <p>
     * If the amount of data to be read is larger than <code>maxDLSize</code>
     * we will break the input into chunks no larger than
     * <code>maxDLSize</code>. This prevents the VM from running out of
     * memory when processing large files.
     *
     * @param in the input stream to read from.
     * @param out the output stream to write to.
     *
     * @return number of bytes written to the output stream
     *
     * @exception IOException if any exceptions occur during transfer
     * of data.
     */
    private int transferData(InputStream in, OutputStream out)
            throws IOException {
        byte[] buffer = new byte[MAX_DL_SIZE];
        int bytesRead;
        int totalBytesWritten = 0;

        try {
            for (; ; ) {
                bytesRead = in.read(buffer);
                if (bytesRead == -1) {
                    return totalBytesWritten;
                }

                out.write(buffer, 0, bytesRead);
                totalBytesWritten += bytesRead;
            }
        } catch (IOException ioe) {
            if (state.stopInstallation) {
                postStatusBackToProvider(USER_CANCELLED_MSG);
                throw new IOException("stopped");
            } else {
                throw ioe;
            }
        }
    }

    /**
     * Stops the installation. If installer is not installing then this
     * method has no effect. This will cause the install method to
     * throw an IOException if the install is not writting the suite
     * to storage which is the point of no return.
     *
     * @return true if the install will stop, false if it is too late
     */
    public boolean stopInstalling() {
        if (state == null) {
            return false;
        }

        synchronized (state) {
            if (state.writingSuite) {
                return false;
            }
        
            state.stopInstallation = true;

            try {
                httpInputStream.close();
            } catch (Exception e) {
                // ignore
            }

            try {
                httpConnection.close();
            } catch (Exception e) {
                // ignore
            }

            return true;
        }
    }

    /**
     * Tell if the installation was stopped by another thread.
     * @return true if the installation was stopped by another thread.
     */
    public boolean wasStopped() {
        if (state == null) {
            return false;
        }

        return state.stopInstallation;
    }

    /**
     * Gets the security domain name for this MIDlet Suite from storage.
     *
     * @param storageName given to the suite by the installer when it was
     *        downloaded.
     * @return name of the security domain for the MIDlet Suite
     */
    protected String getSecurityDomainName(String storageName) {
        // This is a method that is overridden by a secure installer.
        return "untrusted";
    }

    /**
     * Parses a JAD.
     * @param jad raw bytes of the jad
     * @param enc character encoding used
     * @param jadProps Properties specified in .jad file
     * @exception IOException if any error prevents the reading
     *   of the JAD
     * @exception InvalidJadException if the JAD is not valid
     */
    public void parseJad(InputStream jad, String enc,
                         JadProperties jadProps)
            throws IOException, InvalidJadException {
        // This is a method that is overridden by a secure installer.

        jadProps.load(jad, enc);
    }

    /**
     * Verifies a JAD.
     * @param jad raw bytes of the jad
     * @param jadProps Properties specified in .jad file
     *
     * @return name of the authorizing domain owner or null if not signed
     *
     * @exception IOException if any error prevents the reading
     *   of the JAD
     * @exception InvalidJadException if the JAD is not valid
     */
    protected String verifyJad(byte[] jad, Properties jadProps)
            throws IOException, InvalidJadException {
        // This is a place holder for a secure installer.
        return null;
    }

    /**
     * Verifies a Jar.
     *
     * @param jadProps Properties specified in .jad file
     * @param jarStorage System store for applications
     * @param jarFilename name of the jar to read.
     *
     * @exception IOException if any error prevents the reading
     *   of the JAR
     * @exception InvalidJadException if the JAR is not valid
     */
    protected void verifyJar(Properties jadProps,
                             RandomAccessStream jarStorage,
                             String jarFilename)
        // This is a place holder for a secure installer.
        throws IOException, InvalidJadException {
    }

    /**
     * Compares two version strings. The return values are very similar to
     * that of strcmp() in 'C'. If the first version is less than the second
     * version, a negative number will be returned. If the first version is
     * greater than the second version, a positive number will be returned.
     * If the two versions are equal, zero is returned.
     * <p>
     * Versions must be in the form <em>xxx.yyy.zzz</em>, where:
     * <pre>
     *     <em>xxx</em> is the major version
     *     <em>yyy</em> is the minor version
     *     <em>zzz</em> is the micro version
     * </pre>
     * It is acceptable to omit the micro and possibly the minor versions.
     * If these are not included in the version string, the period immediately
     * preceding the number must also be removed. So, the versions
     * <em>xxx.yyy</em> or <em>xxx</em> are also valid.
     * <p>
     * Version numbers do not have to be three digits wide. However, you may
     * pad versions with leading zeros if desired.
     * <p>
     * If a version number is omitted, its value is assumed to be zero. All
     * tests will be based on this assumption.
     * <p>
     * For example:
     * <pre>
     *    1.04  >  1.
     *    1.04  <  1.4.1
     *    1.04  =  1.4.0
     * </pre>
     * <p>
     *
     * @param ver1 the first version to compare.
     * @param ver2 the second versoin to compare.
     *
     * @return  1 if <code>ver1</code> is greater than <code>ver2</code>
     *          0 if <code>ver1</code> is equal to <code>ver2</code>
     *         -1 if <code>ver1</code> is less than <code>ver2</code>
     *
     * @exception NumberFormatException if either <code>ver1</code> or 
     * <code>ver2</code> contain characters that are not numbers or periods.
     */
    private int vercmp(String ver1, String ver2)
            throws NumberFormatException {
        String strVal1;
        String strVal2;
        int    intVal1;
        int    intVal2;
        int    idx1 = 0;
        int    idx2 = 0;
        int    newidx;

        if ((ver1 == null) && (ver2 == null)) {
            return 0;
        }

        if (ver1 == null) {
            return -1;
        }

        if (ver2 == null) {
            return 1;
        }

        for (int i = 0; i < 3; i++) {
            strVal1 = "0"; // Default value
            strVal2 = "0"; // Default value
            if (idx1 >= 0) {
                newidx = ver1.indexOf('.', idx1);
                if (newidx < 0) {
                    strVal1 = ver1.substring(idx1);
                } else {
                    strVal1 = ver1.substring(idx1, newidx);
                    newidx++; // Idx of '.'; need to go to next char
                }

                idx1 = newidx;
            }

            if (idx2 >= 0) {
                newidx = ver2.indexOf('.', idx2);
                if (newidx < 0) {
                    strVal2 = ver2.substring(idx2);
                } else {
                    strVal2 = ver2.substring(idx2, newidx);
                    newidx++;
                }

                idx2 = newidx;
            }

            intVal1 = Integer.parseInt(strVal1); // May throw NFE
            intVal2 = Integer.parseInt(strVal2); // May throw NFE

            if (intVal1 > intVal2) {
                return 1;
            }

            if (intVal1 < intVal2) {
                return -1;
            }
        }

        return 0;
    }

    /**
     * Checks the format of a version string.
     * <p>
     * Versions must be in the form <em>xxx.yyy.zzz</em>, where:
     * <pre>
     *     <em>xxx</em> is the major version
     *     <em>yyy</em> is the minor version
     *     <em>zzz</em> is the micro version
     * </pre>
     * It is acceptable to omit the micro and possibly the minor versions.
     * If these are not included in the version string, the period immediately
     * preceding the number must also be removed. So, the versions
     * <em>xxx.yyy</em> or <em>xxx</em> are also valid.
     * <p>
     * Version numbers do not have to be three digits wide. However, you may
     * pad versions with leading zeros if desired.
     *
     * @param ver the version to check.
     *
     * @exception NumberFormatException if <code>ver</code> 
     *     contains any characters that are not numbers or periods.
     */
    private void checkVersionFormat(String ver) throws NumberFormatException {
        int length;
        int start = 0;
        int end;

        length = ver.length();
        for (int i = 0; ; i++) {
            // check for more than 3 parts or a trailing '.'
            if (i == 3 || start == length) {
                throw new NumberFormatException();
            }

            end = ver.indexOf('.', start);
            if (end == -1) {
                end = length;
            }

            // throws NFE if the substring is not all digits
            Integer.parseInt(ver.substring(start, end));

            if (end == length) {
                // we are done
                return;
            }

            // next time around start after the index of '.'
            start = end + 1;
        }
    }

    /**
     * Gets the MIDlet Suite from storage.
     *
     * @param storageName given to the suite by the installer
     *                   when it was downloaded.
     * @return MIDlet Suite for use by the Scheduler
     */
    public MIDletSuite getMIDletSuite(String storageName) {
        return getMIDletSuite(storageName, null);
    }

    /**
     * Gets the MIDlet Suite from storage, and selects one midlet to be run.
     *
     * @param storageName the unique name of the suite given
     *        by the installer when it was downloaded.
     * @param midletToRun the name of the initial MIDlet in this suite to run,
     *        can be null.
     *
     * @return MIDlet Suite for use by the Scheduler
     */
    public MIDletSuite getMIDletSuite(String storageName,
                                      String midletToRun) {
        File file = new File(classSecurityDomain);
        String storageRoot = File.getStorageRoot() + storageName;

        if (!suiteExists(storageName)) {
            return null;
        }

        return new MIDletSuiteImpl(classSecurityDomain,
                                   Scheduler.getScheduler(),
                                   storageRoot,
                                   getSecurityDomainName(storageName),
                                   midletToRun);
    }

    /**
     * Remove a software package given its storage name
     * <p>
     * If the component is in use it must continue to be available
     * to the other components that are using it.  The resources it
     * consumes must not be released until it is not in use.
     * 
     * @param storageName storage name for the installed package.
     *
     * @exception SecurityException if the caller does not have permission
     *   to remove the installed package.
     */
    public void remove(String storageName) {
        File file = new File(classSecurityDomain);
        String storageRoot = File.getStorageRoot() + storageName;
        Vector files;

        /*
         * remove the files without checking the suite list so that this
         * method can remove files from midlets run from the local
         * classpath.
         */
        synchronized (myInstaller) {
            files = file.filenamesThatStartWith(storageRoot);
            for (int i = 0; i < files.size(); i++) {
                try {
                    file.delete((String)files.elementAt(i));
                } catch (IOException ioe) {
                }
            }

            removeFromSuiteList(storageName);
        }
    }

    /**
     * Queue the execution of the named Application suite to run.
     * The current application suite should terminate itself normally
     * to make resources available to the new application suite. Only
     * one package and set of MIDlets can be queued in this manner.
     * If multiple calls to execute are made, the package and MIDlets
     * specified during the <em>last</em> invokation will be executed
     * when the current application is terminated.
     *
     * @param storageName of an installed package
     * @param midlet name of MIDlet to invoke
     *
     * @return true if the current application MUST exit before
     * the new application suite can be executed.
     *
     * @exception SecurityException if the caller does not have permission
     *   to execute the installed package.
     */
    public boolean execute(String storageName, String midlet) 
            throws SecurityException {
        nextMidletSuiteToRun = storageName;
        nextMidletToRun = midlet; 
        return true;
    }

    /**
     * Gets the unique storage name of the next MIDlet suite to run.
     *
     * @return storage name of a MIDlet suite
     */
    public String getNextMIDletSuiteToRun() {
        return nextMidletSuiteToRun;
    }

    /**
     * Gets the name of the next MIDlet to run.
     *
     * @return storage name of a MIDlet
     */
    public String getNextMIDletToRun() {
        return nextMidletToRun; 
    }

    /**
     * List all installed software packages by storage name.
     *
     * @return an array of Strings of the storage names for the
     *     installed packages.
     * @exception SecurityException if the caller does not have permission
     *     to see what software is installed.
     */
    public String[] list() {
        Vector suiteList = getSuiteList();

        String[] array = new String[suiteList.size()];
        for (int i = 0; i < suiteList.size(); i++) {
            array[i] = (String)suiteList.elementAt(i);
        }

        return array;
    }

    /**
     * Tell if a suite exists.
     *
     * @param storageName storage name of a suite.
     *
     * @return true if a suite of the given storage name
     *          already exists on the system.
     */
    public boolean suiteExists(String storageName) {
        Vector suiteList;

        suiteList = getSuiteList();
        for (int i = 0; i < suiteList.size(); i++) {
            if (suiteList.elementAt(i).equals(storageName)) {
                return true;
            }
        }

        return false;
    }

    /**
     * Retrieve the list of MIDlet suites and store them into a Vector
     * object. Each element in the Vector is the storage name
     * of an installed application.
     *
     * @return the list of installed suites stored in a Vector.
     */
    private Vector getSuiteList() {
        RandomAccessStream storage;
        DataInputStream in;
        String line;

        if (midletSuiteList != null) {
            return midletSuiteList;
        }

        midletSuiteList = new Vector();
        try {
            storage = new RandomAccessStream(classSecurityDomain);

            storage.connect(File.getStorageRoot() +
                           SUITE_LIST_FILENAME, Connector.READ);
            try {
                in = storage.openDataInputStream();
                for (; ; ) {
                    line = in.readUTF();
                    if (line == null) {
                        break;
                    }

                    midletSuiteList.addElement(line);
                }
            } finally {
                storage.disconnect();
            }
        } catch (IOException e) {
            /* will happen the first time, because it will not exist */
        }

        return midletSuiteList;
    }

    /**
     * Adds storage name into the list of MIDlet suites.
     *
     * @param storageName ID of a suite
     */
    private void addToSuiteList(String storageName) {
        Vector suiteList;

        if (suiteExists(storageName)) {
            return;
        }

        suiteList = getSuiteList();
        suiteList.addElement(storageName);
        saveSuiteList(suiteList);
    }

    /**
     * Remove the suite from the list of installed suites.
     *
     * @param storageName a suite
     */
    private void removeFromSuiteList(String storageName) {
        Vector suiteList;

        if (!suiteExists(storageName)) {
            return;
        }

        suiteList = getSuiteList();
        suiteList.removeElement(storageName);
        saveSuiteList(suiteList);
    }

    /**
     * Write the given MIDlet suite list to disk.
     *
     * @param list Vector in the same format as one returned by getSuiteList().
     */
    private void saveSuiteList(Vector list) {
        File file;
        RandomAccessStream storage;
        String filename;
        DataOutputStream out;
        String line;

        file = new File(classSecurityDomain);
        storage = new RandomAccessStream(classSecurityDomain);
        filename = File.getStorageRoot() + SUITE_LIST_FILENAME;
        try {
            file.delete(filename);
        } catch (Exception e) {
            // ignore the exception that happen when the file does not exist
        }

        try {
            storage.connect(filename, RandomAccessStream.READ_WRITE_TRUNCATE);
            try {
                out = storage.openDataOutputStream();
                for (int i = 0; i < list.size(); i++) {
                    line = (String)list.elementAt(i);
                    out.writeUTF(line);
                }
            } finally {
                storage.disconnect();
            }
        } catch (IOException e) {
            /* ignored. Shouldn't happen */
        }
    }

    /**
     * Gets the domain owner of an installed package.
     *
     * @param storageRoot storage root name for the installed package
     *                    it should have any file separators need.
     * @return A string with the value of the domain owner.
     *          <code>null</code> is returned if no value is available
     */
    String getDomainOwner(String storageRoot) {
        RandomAccessStream myStorage;
        DataInputStream dataStream;
        String domainOwner;

        myStorage = new RandomAccessStream(classSecurityDomain);

        try {
            myStorage.connect(storageRoot + DOMAIN_OWNER_FILENAME,
                              Connector.READ);
            try {
                dataStream = myStorage.openDataInputStream();
                domainOwner = dataStream.readUTF();
            } finally {
                myStorage.disconnect();
            }

            return domainOwner;
        } catch (IOException ioe) {
            return null;
        }
    }

    /**
     * Holds the state of an installation, so it can restarted after it has
     * been stopped.
     */
    private class InstallStateImpl implements InstallState {

        /** When the install started, in milliseconds */
        private long startTime;

        /** What to do next. */
        private int nextStep;

        /** Signals the installation to stop. */
        private boolean stopInstallation;

        /** Signals that installation is writting the suite. */
        private boolean writingSuite;

        /** exception that stopped the installation */
        private InvalidJadException exception;

        /** URL of the JAD */
        private String jadUrl;

        /** where the JAD was download from last time */
        private String previousJadUrl;

        /**
         * where or not to force an overwrite of existing components without
         * any version comparison
         */
        private boolean force;

        /** raw JAD */
        private byte[] jad;

        /** character encoding of the JAD */
        private String jadEncoding;

        /** HTTP cookie */
        private String cookie;

        /** parsed JAD */
        private JadProperties jadProps;

        /** cached File object */
        private File file;

        /** user name for authentication */
        private String username;

        /** password for authentication */
        private String password;

        /** name of the suite */
        private String suiteName;

        /** vendor of the suite */
        private String vendor;

        /** version of the suite */
        private String version;

        /** what name the installed suite is stored by */
        private String storageName;

        /** domain owner authorizing the suite, for secure installing */
        private String domainOwner;

        /** how big the JAD says the JAR is */
        private int expectedJarSize;

        /** signals there is a previous version that will be overwritten */
        private boolean isPreviousVersion;

        /**
         * Get the last recoverable exception that stopped the install.
         * Non-recoverable exceptions are thrown and not saved in the state.
         *
         * @return last exception that stopped the install
         */
        public InvalidJadException getLastException() {
            return exception;
        }

        /**
         * Get the unique name that the installed suite was stored with.
         *
         * @return storage name that can be used to load the suite
         */
        public String getStorageName() {
            return storageName;
        }

        /**
         * Set the username to be used for HTTP authentication.
         *
         * @param username 8 bit username, cannot contain a ":"
         */
        public void setUsername(String username) {
            this.username = username;
        }

        /**
         * Set the password to be used for HTTP authentication.
         *
         * @param password 8 bit password
         */
        public void setPassword(String password) {
            this.password = password;
        }
    }
}
