/*
 * @(#)Canvas.java	1.20 01/08/15
 * Copyright (c) 1999-2001 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package javax.microedition.lcdui;

/**
 * <p> The Canvas class is a base class for writing applications that need to
 * handle low-level events and to issue graphics calls for drawing to the
 * display. Game applications will likely make heavy use of the Canvas class.
 * From an application development perspective, the Canvas class is
 * interchangeable with standard Screen classes, so an application may mix and
 * match Canvas with high-level screens as needed. For example, a List screen
 * may be used to select the track for a racing game, and a Canvas subclass
 * would implement the actual game.</P>
 *
 * <P>The Canvas provides the developer with methods to handle game actions,
 * key events, and
 * pointer events (if supported by the device).  Methods are
 * also provided to identify the device's capabilities and keyboard mapping.
 * The key events are reported with respect to <em>key codes</em>, which
 * are directly bound to concrete keys on the device, use of which may hinder
 * portability.  Portable applications should use game actions instead of key 
 * codes.</p>
 * 
 * <p> Like other subclasses of Displayable, the Canvas class allows the
 * application to register a listener for commands.  Unlike other Displayables,
 * however, the Canvas class requires applications to subclass it in order to
 * use it. The paint() method is declared <code>abstract</code>, and so the
 * application <em>must</em> provide an implementation in its subclass. Other
 * event-reporting methods are not declared <code>abstract,</code> and their
 * default implementations are empty (that is, they do nothing). This allows
 * the application to override only the methods that report events in which the
 * application has interest.  </p>
 *
 * <p> This is in contrast to the {@link Screen Screen} classes, which allow
 * the application to define listeners and to register them with instances of
 * the Screen classes. This style is not used for the Canvas class, because
 * several new listener interfaces would need to be created, one for each kind
 * of event that might be delivered. An alternative would be to have fewer
 * listener interfaces, but this would require listeners to filter out events
 * in which they had no interest. </p>
 *
 * <a name="keyevents"></a>
 * <h3>Key Events</h3>
 *
 * <p> Applications receive keystroke events in which the individual keys are
 * named within a space of <em>key codes</em>. Every key for which events are
 * reported to MIDP applications is assigned a key code.
 * The key code values are unique for each hardware key unless two keys are
 * obvious synonyms for each other.
 * MIDP defines the following key codes:
 * {@link #KEY_NUM0 KEY_NUM0},
 * {@link #KEY_NUM1 KEY_NUM1},
 * {@link #KEY_NUM2 KEY_NUM2},
 * {@link #KEY_NUM3 KEY_NUM3},
 * {@link #KEY_NUM4 KEY_NUM4},
 * {@link #KEY_NUM5 KEY_NUM5},
 * {@link #KEY_NUM6 KEY_NUM6},
 * {@link #KEY_NUM7 KEY_NUM7},
 * {@link #KEY_NUM8 KEY_NUM8},
 * {@link #KEY_NUM9 KEY_NUM9},
 * {@link #KEY_STAR KEY_STAR}, and
 * {@link #KEY_POUND KEY_POUND}.
 * 
 * (These key codes correspond to keys on a ITU-T standard telephone keypad.)
 * Other keys may be present on the keyboard, and they will generally have key
 * codes distinct from those list above.  In order to guarantee portability,
 * applications should use only the standard key codes. </p>
 * 
 * <p>The standard key codes' values are equal to the Unicode encoding for the
 * character that represents the key.  If the device includes any other keys
 * that have an obvious correspondence to a Unicode character, their key code
 * values should equal the Unicode encoding for that character.  For keys that
 * have no corresponding Unicode character, the implementation must use
 * negative values.  Zero is defined to be an invalid key code.  It is thus
 * possible for an application to convert a keyCode into a Unicode character
 * using the following code: </p>
 * 
 * <pre>
 *    if (keyCode > 0) {
 *        char ch = (char)keyCode;
 *        // ...
 *    }
 * </pre>
 * 
 * <p>This technique is useful only in certain limited cases.  In particular,
 * it is not sufficient for full textual input, because it does not handle
 * upper and lower case, keyboard shift states, and characters that require
 * more than one keystroke to enter.  For textual input, applications should
 * always use {@link TextBox TextBox} or {@link TextField TextField}
 * objects.</p>
 *
 * <p> It is sometimes useful to find the <em>name</em> of a key in order to
 * display a message about this key. In this case the application may use the
 * {@link #getKeyName(int) getKeyName()} method to find a key's name. </p>
 *
 * <a name="gameactions"></a>
 * <h3>Game Actions</h3>
 * 
 * Portable applications that need arrow key events and gaming-related events 
 * should use <em>game actions</em> in preference to key codes and key names.  
 * MIDP defines the following game actions:
 * {@link #UP UP},
 * {@link #DOWN DOWN},
 * {@link #LEFT LEFT},
 * {@link #RIGHT RIGHT},
 * {@link #FIRE FIRE},
 * {@link #GAME_A GAME_A},
 * {@link #GAME_B GAME_B},
 * {@link #GAME_C GAME_C}, and
 * {@link #GAME_D GAME_D}.
 * 
 * <P> Each key code may be mapped to at most one game action.  However, a game
 * action may be associated with more than one key code.  The application can
 * translate a key code into a game action using the {@link #getGameAction(int)
 * getGameAction(int keyCode)} method, and it can translate a key code into a
 * game action using the {@link #getKeyCode(int) getKeyCode(int gameAction)}
 * method.  The implementation is not allowed to change the mapping of game
 * actions and key codes during execution of the application.</P>
 *
 * <p> Portable applications that are interested in using game actions should
 * translate every key event into a game action by calling the {@link
 * #getGameAction getGameAction()} method and then testing the result.  For
 * example, on some devices the game actions UP, DOWN, LEFT and RIGHT may be
 * mapped to 4-way navigation arrow keys. In this case, getKeyCode(UP) would
 * return a device-dependent code for the up-arrow key.  On other devices, a
 * possible mapping would be on the number keys 2, 4, 6 and 8. In this case,
 * getKeyCode(UP) would return KEY_NUM2.  In both cases, the getGameAction()
 * method would return the LEFT game action when the user presses the key that
 * is a "natural left" on her device.  </P>
 *
 * <a name="commands"></a>
 * <h3>Commands</h3>
 * 
 * <p> It is also possible for the user to issue {@link Command commands} when
 * a canvas is current. Commands are mapped to keys and menus in a
 * device-specific fashion. For some devices the keys used for commands may
 * overlap with the keys that will deliver key code events to the canvas. If
 * this is the case, the device will provide a means transparent to the
 * application that enables the user to select a mode that determines whether
 * these keys will deliver commands or key code events to the application. The
 * set of key code events available to a canvas will not change depending upon
 * the number of commands that are present on the canvas.  Game developers 
 * should be aware that access to commands will vary greatly across devices, 
 * and that requiring the user to issue commands during game play may have a 
 * great impact on the ease with which the game can be played. </p>
 *
 * <a name="eventdelivery"></a>
 * <h3>Event Delivery</h3>
 *
 * <P> The Canvas object defines several methods that are called by the
 * implementation. These methods are primarily for the purpose of delivering
 * events to the application, and so they are referred to as
 * <em>event delivery</em> methods. The set of methods is: </p>
 *
 * <ul>
 * <li> showNotify() </li>
 * <li> hideNotify() </li>
 * <li> keyPressed() </li>
 * <li> keyRepeated() </li>
 * <li> keyReleased() </li>
 * <li> pointerPressed() </li>
 * <li> pointerDragged() </li>
 * <li> pointerReleased() </li>
 * <li> paint() </li>
 * <li> the CommandListener's commandAction() method </li>
 * </ul>
 *
 * <p> These methods are all called serially. That is, the implementation will
 * never call an event delivery method before a prior call
 * to <em>any</em> of the
 * event delivery methods has returned. (But see the note below.) This
 * property enables applications to be assured that processing of a previous
 * user event will have completed before the next event is delivered. </p>
 *
 * <p> Calls to the run() method of Runnable objects passed to
 * Display.callSerially() will also be serialized along with calls to the event
 * delivery methods. See {@link Display#callSerially(Runnable) callSerially()}
 * for further information. </p>
 *
 * <p><b>Note: </b> The serviceRepaints() method is an exception to this rule,
 * as it blocks until paint() is called and returns. This will occur even if
 * the application is in the midst of one of the event delivery methods and it
 * calls serviceRepaints(). </p>
 *
 * <p> The key-related, pointer-related, paint(), and commandAction() methods
 * will only be called while the Canvas is actually visible on the output
 * device. These methods will therefore only be called on this Canvas object
 * only after a call to showNotify() and before a call to hideNotify(). After
 * hideNotify() has been called, none of the key, pointer, paint, and
 * commandAction() methods will be called until after a subsequent call to
 * showNotify() has returned.  A call to a sun() method resulting from
 * callSerially() may occur irrespective of calls to showNotify() and
 * hideNotify().  </p>
 *
 * <p> The {@link #showNotify() showNotify()} method is called prior to the
 * Canvas actually being made visible on the display, and the {@link
 * #hideNotify() hideNotify()} method is called after the Canvas has been
 * removed from the display.  The visibility state of a Canvas (or any other
 * Displayable object) may be queried through the use of the {@link
 * Displayable#isShown() Displayable.isShown()} method.  The change in
 * visibility state of a Canvas may be caused by the application management
 * software moving MIDlets between foreground and background states, or by the
 * system obscuring the Canvas with system screens.  Thus, the calls to
 * showNotify() and hideNotify() are not under the control of the MIDlet and
 * may occur fairly frequently.  Application developers are encouraged to
 * perform expensive setup and teardown tasks outside the showNotify() and
 * hideNotify() methods in order to make them as lightweight as possible. </p>
 */
public abstract class Canvas extends Displayable {
    
    /**
     * <P>Constant for the UP game action.</P>
     *
     * <P>Constant value 1 is set to UP.</P>
     */
    public static final int UP = 1; 
    
    /**
     * <P>Constant for the DOWN game action.</P>
     *
     * <P>Constant value 6 is set to DOWN.</P>
     */
    public static final int DOWN = 6;
    
    
    /**
     * <P>Constant for the LEFT game action.</P>
     *
     * <P>Constant value 2 is set to LEFT.</P>
     */
    public static final int LEFT = 2;
    
    /**
     * <P>Constant for the RIGHT game action.</P>
     *
     * <P>Constant value 5 is set to RIGHT.</P>
     */
    public static final int RIGHT = 5;
    
    /**
     * <P>Constant for the FIRE game action.</P>
     *
     * <P>Constant value 8 is set to FIRE.</P>
     */
    public static final int FIRE = 8; 

    
    /**
     * <P>Constant for the general purpose "A" game action.</P>
     *
     * <P>Constant value 9 is set to GAME_A.</P>
     */
    public static final int GAME_A = 9; 
    
    /**
     * <P>Constant for the general purpose "B" game action.</P>
     *
     * <P>Constant value 10 is set to GAME_B.</P>
     */
    public static final int GAME_B = 10; 
    
    /**
     * <P>Constant for the general purpose "C" game action.</P>
     *
     * <P>Constant value 11 is set to GAME_C.</P>
     */
    public static final int GAME_C = 11;
    
    /**
     * <P>Constant for the general purpose "D" game action.</P>
     *
     * <P>Constant value 12 is set to GAME_D.</P>
     */
    public static final int GAME_D = 12; 
    
    /**
     * <P>keyCode for ITU-T key 0.</P>
     *
     * <P>Constant value 48 is set to KEY_NUM0.</P>
     */
    public static final int KEY_NUM0  =  48;
    
    /**
     * <P>keyCode for ITU-T key 1.</P>
     *
     * <P>Constant value 49 is set to KEY_NUM1.</P>
     */
    public static final int KEY_NUM1  =  49;
    
    /**
     * <P>keyCode for ITU-T key 2.</P>
     *
     * <P>Constant value 50 is set to KEY_NUM2.</P>
     */
    public static final int KEY_NUM2  =  50;
    
    /**
     * <P>keyCode for ITU-T key 3.</P>
     *
     * <P>Constant value 51 is set to KEY_NUM3.</P>
     */
    public static final int KEY_NUM3 = 51;
    
    /**
     * <P>keyCode for ITU-T key 4.</P>
     *
     * <P>Constant value 52 is set to KEY_NUM4.</P>
     */
    public static final int KEY_NUM4 = 52;
    
    /**
     * <P>keyCode for ITU-T key 5.</P>
     *
     * <P>Constant value 53 is set to KEY_NUM5.</P>
     */
    public static final int KEY_NUM5 = 53;
    
    /**
     * <P>keyCode for ITU-T key 6.</P>
     *
     * <P>Constant value 54 is set to KEY_NUM6.</P>
     */
    public static final int KEY_NUM6 = 54;
    
    /**
     * <P>keyCode for ITU-T key 7.</P>
     *
     * <P>Constant value 55 is set to KEY_NUM7.</P>
     */
    public static final int KEY_NUM7 = 55;
    
    /**
     * <P>keyCode for ITU-T key 8.</P>
     *
     * <P>Constant value 56 is set to KEY_NUM8.</P>
     */
    public static final int KEY_NUM8 = 56;
    
    /**
     * <P>keyCode for ITU-T key 9.</P>
     *
     * <P>Constant value 57 is set to KEY_NUM09.</P>
     */
    public static final int KEY_NUM9 = 57;
    
    /**
     * <P>keyCode for ITU-T key "star" (*).</P>
     *
     * <P>Constant value 42 is set to KEY_STAR.</P>
     */
    public static final int KEY_STAR = 42;
    
    /**
     * <P>keyCode for ITU-T key "pound" (#).</P>
     *
     * <P>Constant value 35 is set to KEY_POUND.</P>
     */
    public static final int KEY_POUND = 35;

    /**
     * Constructs a new Canvas object.
     */
    protected Canvas()  { }
    
    /**
     * Gets width of the displayable area in pixels. The value is unchanged
     * during the execution of the application and all Canvases will 
     * have the same value.
     * @return width of the displayable area
     */
    public int getWidth() {
        // SYNC NOTE: return of atomic value, no locking necessary
        return Display.WIDTH;
    }

    
    /**
     * Gets height of the displayable area in pixels. The value is unchanged
     * during the execution of the application and all Canvases will
     * have the same value.
     * @return height of the displayable area
     */
    public int getHeight() {
        // SYNC NOTE: return of atomic value, no locking necessary
        return Display.HEIGHT;
    }
    
    /**
     * Checks if the Graphics is double buffered by the implementation.
     * @return true if double buffered, false otherwise.
     */
    public boolean isDoubleBuffered() {
        // SYNC NOTE: return of atomic value, no locking necessary
        return Display.IS_DOUBLE_BUFFERED;
    }

    /**
     * Checks if the platform supports pointer press and release events.
     * @return true if the device supports pointer events
     */
    public boolean hasPointerEvents() {
        // SYNC NOTE: return of atomic value, no locking necessary
        return Display.POINTER_SUPPORTED;
    }
    
    /**
     * Checks if the platform supports pointer motion events (pointer dragged).
     * Applications may use this method to determine if the platform is capable
     * of supporting motion events.
     * @return true if the device supports pointer motion events
     */
    public boolean hasPointerMotionEvents() {
        // SYNC NOTE: return of atomic value, no locking necessary
        return Display.MOTION_SUPPORTED;
    }
    
    /**
     * Checks if the platform can generate repeat events when key
     * is kept down.
     * @return true if the device supports repeat events
     */
    public boolean hasRepeatEvents() { 
        // SYNC NOTE: return of atomic value, no locking necessary
        return Display.REPEAT_SUPPORTED;
    }
    
    /**
     * <p>Gets a key code that corresponds to the specified game action on the
     * device.  The implementation is required to provide a mapping for every
     * game action, so this method will always return a valid key code for
     * every game action.  See <a href="#gameactions">above</a> for further
     * discussion of game actions.</p>
     * 
     * <p>Note that a key code is associated with at most one game action,
     * whereas a game action may be associated with several key codes.  Then,
     * supposing that <code>g</code> is a valid game action and <code>k</code>
     * is a valid key code for a key associated with a game action, consider
     * the following expressions:</p>
     *
     * <pre>
     *    g == getGameAction(getKeyCode(g))     // (1)
     *    k == getKeyCode(getGameAction(k))     // (2)
     * </pre>
     *
     * <p>Expression (1) is <em>always</em> true.  However, expression (2) might
     * be true but is <em>not necessarily</em> true.</p>
     *
     * <P>The mapping between key codes and game actions
     * will not change during the execution of the application.</P>
     *
     * @param gameAction the game action
     * @return a key code corresponding to this game action
     * @throws IllegalArgumentException if gameAction is not a valid game action
     */
    public int getKeyCode(int gameAction) {
        // SYNC NOTE: no locking necessary as we are doing a static
        // table lookup and getKeyCode() is implemented natively
        int n = Display.getKeyCode(gameAction);

        if (n == 0) {
            throw new IllegalArgumentException();
        }

        return n;
    }
    
    /**
     * <p> Gets an informative key string for a key. The string returned will
     * resemble the text physically printed on the key.  This string is 
     * suitable for displaying to the user.  For example, on a device
     * with function keys F1 through F4, calling this method on the keyCode for
     * the F1 key will return the string "F1". A typical use for this string
     * will be to compose help text such as "Press F1 to proceed." </p>
     *
     * <p> This method will return a non-empty string for every valid key code.
     * </p>
     *
     * <p> There is no direct mapping from game actions to key names. To get
     * the string name for a game action GAME_A, the application must call </p>
     *
     * <pre>
     *    getKeyName(getKeyCode(GAME_A))
     * </pre>
     *
     * @param keyCode the key code being requested
     * @return a string name for the key
     * @throws IllegalArgumentException if keyCode is not a valid key code
     */
    public String getKeyName(int keyCode) {
        // SYNC NOTE: no locking necessary as we are doing a static
        // table lookup and getKeyName() is implemented natively
        String s = Display.getKeyName(keyCode);

        if (s == null) {
            throw new IllegalArgumentException();
        }

        return s;
    }

    /**
     * <p>Gets the game action associated with the given key code of the
     * device.  Returns zero if no game action is associated with this key
     * code.  See <a href="#gameactions">above</a> for further discussion of
     * game actions.  </p>
     *
     * <P>The mapping between key codes and game actions
     * will not change during the execution of the application.</P>
     * 
     * @param keyCode the key code
     * @return the game action corresponding to this key, or 0 if none
     * @throws IllegalArgumentException if keyCode is not a valid key code
     */
    public int getGameAction(int keyCode) {
        // SYNC NOTE: no locking necessary as we are doing a static
        // table lookup and getGameAction() is implemented natively
        int n = Display.getGameAction(keyCode);

        if (n == -1) {
            throw new IllegalArgumentException();
        }

        return n;
    }

    /**
     * <P>Called when a key is pressed.</P>
     *
     * <P>The getGameAction() method can be called to
     * determine what game action, if any, is mapped to the key.
     * Class Canvas has an empty implementation of this method, and
     * the subclass has to redefine it if it wants to listen this method.
     * @param keyCode The key code of the key that was pressed.
     */
    protected void keyPressed(int keyCode) {
    }

    /**
     * Called when a key is repeated (held down). The getGameAction() method can
     * be called to determine what game action,
     * if any, is mapped to the key.
     * Class Canvas has an empty implementation of this method, and
     * the subclass has to redefine it if it wants to listen this method.
     * @param keyCode The key code of the key that was repeated
     * @see #hasRepeatEvents()
     */
    protected void keyRepeated(int keyCode) {
    }

    /**
     * Called when a key is released. The getGameAction() method
     * can be called to
     * determine what game action, if any, is mapped to the key.
     * Class Canvas has an empty implementation of this method, and
     * the subclass has to redefine it if it wants to listen this method.
     * @param keyCode The key code of the key that was released
     */
    protected void keyReleased(int keyCode) {
    }
    
    /**
     * Called when the pointer is pressed. The
     *  {@link #hasPointerEvents() hasPointerEvents()}
     * method may be called to determine if the device supports pointer events.
     * Class Canvas has an empty implementation of this method, and
     * the subclass has to redefine it if it wants to listen this method.
     * @param x The horizontal location where the pointer was pressed (relative
     * to the Canvas)
     * @param y The vertical location where the pointer was pressed (relative to
     * the Canvas)
     */
    protected void pointerPressed(int x, int y) {
    }
    
    /**
     * Called when the pointer is released. The
     * {@link #hasPointerEvents() hasPointerEvents()}
     * method may be called to determine if the device supports pointer events.
     * Class Canvas has an empty implementation of this method, and
     * the subclass has to redefine it if it wants to listen this method.
     * @param x The horizontal location where the pointer was released (relative
     * to the Canvas)
     * @param y The vertical location where the pointer was released
     * (relative to the Canvas)
     */
    protected void pointerReleased(int x, int y) {
    }
    
    /**
     * Called when the pointer is dragged. The
     * {@link #hasPointerMotionEvents() hasPointerMotionEvents()}
     * method may be called to determine if the device supports pointer events.
     * Class Canvas has an empty implementation of this method, and
     * the subclass has to redefine it if it wants to listen this method.
     * @param x The horizontal location where the pointer was dragged (relative
     * to the Canvas)
     * @param y The vertical location where the pointer was dragged (relative to
     * the Canvas)
     */
    protected void pointerDragged(int x, int y) {
    }
    
    /**
     * <p> Requests a repaint for the specified region of the Screen. Calling
     * this method may result in subsequent call to paint(), where the passed
     * Graphics object's clip region will include at least the specified region.
     * </p>
     *
     * <p> If the canvas is not visible, or if width and height are zero
     * or less, or if the rectangle does not specify a visible region
     * of the display, this call has no effect. </p>
     *
     * <p> The call to paint() occurs independently of the call to repaint().
     * That is, repaint() will not block waiting for paint() to finish. The
     * paint() method will either be called after the caller of repaint()
     * returns
     * to the implementation (if the caller is a callback) or on another thread
     * entirely. </p>
     *
     * <p> To synchronize with its paint() routine, applications can use either
     * {@link Display#callSerially(Runnable) Display.callSerially()} or
     * {@link #serviceRepaints() serviceRepaints()}, or they can code explicit
     * synchronization into their paint() routine. </p>
     *
     * <p> The origin of the coordinate system is above and to the left of the
     * pixel in the upper left corner of the displayable area of the Canvas. The
     * X-coordinate is positive right and the Y-coordinate is positive
     * downwards.
     * </p>
     *
     * @param x the x coordinate of the rectangle to be repainted
     * @param y the y coordinate of the rectangle to be repainted
     * @param width the width of the rectangle to be repainted
     * @param height the height of the rectangle to be repainted
     *
     * @see Display#callSerially(Runnable)
     * @see #serviceRepaints()
     */
    public final void repaint(int x, int y, int width, int height) {
        synchronized (Display.LCDUILock) {
            super.repaint(x, y, width, height);
        }
    }

    /**
     * Requests a repaint for the entire Canvas. The effect is identical to
     * <p> <code> repaint(0, 0, getWidth(), getHeight()); </code>
     */
    public final void repaint() {
        synchronized (Display.LCDUILock) {
            super.repaint();
        }
    }
    
    /**
     * <p> Forces any pending repaint requests to be serviced immediately. This
     * method blocks until the pending requests have been serviced. If there are
     * no pending repaints, or if this canvas is not visible on the display,
     * this call does nothing and returns immediately. </p>
     *
     * <p> <b> WARNING: </b> This method blocks until the call to the
     * application's paint() method returns. The application has no control over
     * which thread calls paint(); it may vary from implementation to
     * implementation. If the caller of serviceRepaints() holds a lock that the
     * paint() method acquires, this may result in deadlock. Therefore, callers
     * of serviceRepaints() <em>must not</em> hold any locks that might be
     * acquired within the paint() method. The
     * {@link Display#callSerially(Runnable) Display.callSerially()}
     * method provides a facility where an application can be called back after
     * painting has completed, avoiding the danger of deadlock.
     * </p>
     *
     * @see Display#callSerially(Runnable)
     */
    public final void serviceRepaints() {
        // SYNC NOTE: unlike most public API methods, no locking is done
        // here.  This is necessary because Display.serviceRepaints()
        // needs to handle its own locking.  We avoid locking by making
        // a copy of currentDisplay -- an atomic operation -- before
        // testing and using it.

        Display d = currentDisplay;

        if (d != null) {
            d.serviceRepaints();
        }
    }


    /**
     * The implementation calls showNotify()
     * immediately prior to this Canvas being made visible on the display.
     * Canvas subclasses may override
     * this method to perform tasks before being shown, such
     * as setting up animations, starting timers, etc.
     * The default implementation of this method in class Canvas is empty.
     */
    protected void showNotify() {
    }
    
    /**
     * The implementation calls hideNotify() shortly after the Canvas has been 
     * removed from the display.
     * Canvas subclasses may override this method in order to pause animations, 
     * revoke timers, etc.  The default implementation of this 
     * method in class Canvas is empty.
     */
    protected void hideNotify()  {
    }

    /**
     * Renders the Canvas. The application must implement this method
     * in order to paint any graphics.<p>
     *
     * The Graphics object's clip region defines the area of the screen that is
     * considered to be invalid. A correctly-written paint() routine must paint
     * <em>every</em> pixel within this region. Applications <em>must not</em>
     * assume that
     * they know the underlying source of the paint() call and use this
     * assumption
     * to paint only a subset of the pixels within the clip region. The
     * reason is
     * that this particular paint() call may have resulted from multiple
     * repaint()
     * requests, some of which may have been generated from outside the
     * application. An application that paints only what it thinks is
     * necessary to
     * be painted may display incorrectly if the screen contents had been
     * invalidated by, for example, an incoming telephone call. <p>
     *
     * Operations on this graphics object after the paint() call returns are
     * undefined. Thus, the application <em>must not</em> cache this Graphics
     * object for later use or use by another thread. It must only be
     * used within
     * the scope of this method. <p>
     *
     * The implementation may postpone visible effects of
     * graphics operations until the end of the paint method.<p>
     *
     * <p> The contents of the Canvas are never saved if it is hidden and then
     * is made visible again. Thus, shortly after showNotify() is called,
     * paint() will always be called with a Graphics object whose clip region
     * specifies the entire displayable area of the Canvas.  Applications
     * <em>must not</em> rely on any contents being preserved from a previous
     * occasion when the Canvas was current. This call to paint() will not
     * necessarily occur before any other key, pointer, or commandAction()
     * methods are called on the Canvas.  Applications whose repaint
     * recomputation is expensive may create an offscreen Image, paint into it,
     * and then draw this image on the Canvas when paint() is called. </p>
     *
     * <P>The application code must never call paint(); it is called only by
     * the implementation.</P>
     *
     * <P>The Graphics object passed to the paint() method has the following 
     * properties:</P>
     * <UL>
     * <LI>the destination is the actual display, or if double buffering is in 
     * effect, a back buffer for the display;</LI>
     * <LI>the clip region includes at least one pixel
     * within this Canvas;</LI>
     * <LI>the current color is black;</LI>
     * <LI>the font is the same as the font returned by
     * {@link Font#getDefaultFont() Font.getDefaultFont()};</LI>
     * <LI>the stroke style is {@link Graphics#SOLID SOLID};</LI>
     * <LI>the origin of the coordinate system is located at the upper-left
     * corner of the Canvas; and</LI>
     * <LI>the Canvas is visible, that is, a call to isShown() will return
     * <code>true</code>.</LI>
     * </UL>
     *
     * @param g the Graphics object to be used for rendering the Canvas
     */
    protected abstract void paint(Graphics g);
    
}
