/*
 *  @(#)ChoiceGroup.java	1.94 01/08/21
 * Copyright (c) 1999-2001 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

package javax.microedition.lcdui;

/**
 * <p> A ChoiceGroup is a group of selectable elements intended to be
 * placed within a
 * {@link Form}. The group may be created with a mode that requires a
 * single choice to be made or that allows multiple choices. The
 * implementation is responsible for providing the graphical representation of
 * these modes and must provide visually different graphics for different
 * modes. For example, it might use "radio buttons" for the single choice
 * mode and "check boxes" for the multiple choice mode. </p>
 *
 * <p> <b>Note:</b> most of the essential methods have been specified in
 * the {@link Choice Choice} interface.</p>
 */
public class ChoiceGroup extends Item implements Choice {

    /**
     * The type of this ChoiceGroup
     */
    private int choiceType;

    /**
     * The number of elements in this ChoiceGroup
     */
    private int numOfEls;

    /**
     * The currently selected index of this ChoiceGroup (-1 by default)
     */
    private int selectedIndex = -1;

    /**
     * The currently highlighted index of this ChoiceGroup (-1 by default)
     */
    private int hilightedIndex = -1;

    /**
     * The array of selected elements of this ChoiceGroup (in the case
     * of a multi-select type)
     */
    private boolean selEls[];

    /**
     * The pixel height of this ChoiceGroup
     */
    private int height;
    
    /**
     * Creates a new, empty ChoiceGroup, specifying its title and its type.
     * The type must be one of EXCLUSIVE or MULTIPLE. The IMPLICIT
     * choice type is not allowed within a ChoiceGroup. <p>
     *
     * @param label the item's label (see {@link Item Item})
     * @param choiceType either EXCLUSIVE or MULTIPLE
     * @throws IllegalArgumentException if choice type is neither
     * EXCLUSIVE nor MULTIPLE
     * @see Choice#EXCLUSIVE
     * @see Choice#MULTIPLE
     * @see Choice#IMPLICIT
     */
    public ChoiceGroup(String label, int choiceType) {
        this(label, choiceType, new String[] {}, null);
    }
    
    /**
     * <p>Creates a new ChoiceGroup, specifying its title, the type of the
     * ChoiceGroup, and an array of Strings and Images to be used as its
     * initial contents. </p>
     *
     * <p>The type must be one of EXCLUSIVE or MULTIPLE.  The IMPLICIT type is 
     * not allowed for ChoiceGroup.</p>
     *
     * <p>The stringElements array must be non-null and every array element
     * must also be non-null.  The length of the stringElements array
     * determines the number of elements in the ChoiceGroup.  The 
     * imageElements array
     * may be null to indicate that the ChoiceGroup elements have no 
     * images.  If the
     * imageElements array is non-null, it must be the same length as the
     * stringElements array.  Individual elements of the imageElements array
     * may be null in order to indicate the absence of an image for the
     * corresponding ChoiceGroup element.  Any elements present in the
     * imageElements array must refer to immutable images.</p>
     * 
     * @param label the item's label (see {@link Item Item})
     * @param choiceType EXCLUSIVE or MULTIPLE
     * @param stringElements set of strings specifying the string parts of the
     * ChoiceGroup elements
     * @param imageElements set of images specifying the image parts of
     * the ChoiceGroup elements
     *
     * @throws NullPointerException if stringElements is null
     * @throws NullPointerException if the stringElements array contains
     * any null elements
     * @throws IllegalArgumentException if the imageElements array is non-null
     * and has a different length from the stringElements array
     * @throws IllegalArgumentException if choiceType is neither
     * EXCLUSIVE nor MULTIPLE
     * @throws IllegalArgumentException if any image in the imageElements
     * array is mutable
     *
     * @see Choice#EXCLUSIVE
     * @see Choice#MULTIPLE
     * @see Choice#IMPLICIT
     */
    public ChoiceGroup(String label, int choiceType, 
                       String[] stringElements, Image[] imageElements) {

        this(label, choiceType, stringElements, imageElements, false);
    }
    
    /**
     * Special constructor used by List
     *
     * @param label the item's label (see {@link Item Item})
     * @param choiceType EXCLUSIVE or MULTIPLE
     * @param stringElements set of strings specifying the string parts of the
     * ChoiceGroup elements
     * @param imageElements set of images specifying the image parts of
     * the ChoiceGroup elements
     * @param implicitAllowed Flag to allow implicit selection
     *
     * @throws NullPointerException if stringElements is null
     * @throws NullPointerException if the stringElements array contains
     * any null elements
     * @throws IllegalArgumentException if the imageElements array is non-null
     * and has a different length from the stringElements array
     * @throws IllegalArgumentException if choiceType is neither
     * EXCLUSIVE nor MULTIPLE
     * @throws IllegalArgumentException if any image in the imageElements
     * array is mutable
     *
     * @see Choice#EXCLUSIVE
     * @see Choice#MULTIPLE
     * @see Choice#IMPLICIT
     */
    ChoiceGroup(String label, int choiceType, String[] stringElements,
            Image[] imageElements, boolean implicitAllowed) {

        // If stringElements is null NullPointerException will be thrown
        // as expected
        super(label, stringElements.length + 1);

        if (imageElements != null &&
            stringElements.length != imageElements.length) {
            throw new IllegalArgumentException();
        }

        if (!((choiceType == Choice.MULTIPLE) ||
                (choiceType == Choice.EXCLUSIVE) ||
                ((choiceType == Choice.IMPLICIT) && implicitAllowed))) {
            throw new IllegalArgumentException();
        }

        synchronized (Display.LCDUILock) {
            this.choiceType = choiceType;
            numOfEls = stringElements.length;

            switch (choiceType) {
                case Choice.MULTIPLE:
                    selEls = new boolean[numOfEls];
                    for (int i = 0; i < numOfEls; i++) {
                        selEls[i] = false;
                    }
                    break;
                case Choice.IMPLICIT:
                case Choice.EXCLUSIVE:
                    if (numOfEls > 0) {
                        selectedIndex = 0;
                    }
                    break;
                }

            // layout array and its 0 element (label) are allocated in
            // the super class (Item)
            for (int i = 0; i < numOfEls; i++) {
                Image img = imageElements == null ? null : imageElements[i];
                checkNull(stringElements[i], img);

                boolean offset = choiceType == Choice.IMPLICIT && 
                                 img == null;
                                  
                layouts[i + 1] = new IaSLayout(stringElements[i], img,
                                               Screen.CONTENT_FONT,
                                               offset);
            }
        } // synchronized
    }

    /**
     * @return the number of elements in the ChoiceGroup
     */
    public int size() {
        // SYNC NOTE: return of atomic value, no locking necessary
        return numOfEls;
    }

    /**
     * @param elementNum the index of the element to be queried
     * @return the string part of the element
     * @throws IndexOutOfBoundsException if elementNum is invalid
     * @see #getImage(int)
     */
    public String getString(int elementNum) {
        synchronized (Display.LCDUILock) {
            checkIndex(elementNum);
            return ((IaSLayout)layouts[elementNum + 1]).getString();
        }
    }

    /**
     * @param elementNum the number of the element to be queried
     * @return the image part of the element, or null if there is no image
     * @throws IndexOutOfBoundsException if elementNum is invalid
     * @see #getString(int)
     */
    public Image getImage(int elementNum) {
        synchronized (Display.LCDUILock) {
            checkIndex(elementNum);
            return ((IaSLayout)layouts[elementNum + 1]).getImage();
        }
    }

    /**
     * @param stringPart the string part of the element to be added
     * @param imagePart the image part of the element to be added, or null if
     * there is no image part
     * @return the assigned index of the element
     * @throws IllegalArgumentException if the image is mutable
     * @throws NullPointerException if stringPart is null
     */
    public int append(String stringPart, Image imagePart) {
        synchronized (Display.LCDUILock) {
            checkNull(stringPart, imagePart);
            return insertImpl(numOfEls, stringPart, imagePart);
        }
    }

    /**
     * @param elementNum the index of the element where insertion is to occur
     * @param stringElement the string part of the element to be inserted
     * @param imageElement the image part of the element to be inserted,
     * or null if there is no image part
     * @throws IndexOutOfBoundsException if elementNum is invalid
     * @throws IllegalArgumentException if the image is mutable
     * @throws NullPointerException if stringPart is null
     */
    public void insert(int elementNum, String stringElement,
                       Image imageElement) {

        synchronized (Display.LCDUILock) {
            if (elementNum < 0 || elementNum > numOfEls) {
                throw new IndexOutOfBoundsException();
            }
            checkNull(stringElement, imageElement);
            insertImpl(elementNum, stringElement, imageElement);
        }
    }

    /**
     * @param elementNum the index of the element to be deleted
     * @throws IndexOutOfBoundsException if elementNum is invalid
     */
    public void delete(int elementNum) {
        synchronized (Display.LCDUILock) {
            checkIndex(elementNum);

            int deletedHeight = 0;
            if (initLayoutDone()) {
                deletedHeight = -layouts[elementNum + 1].getHeight();
                height += deletedHeight;
            }

            if (elementNum != numOfEls - 1) {
                System.arraycopy(layouts, elementNum + 2, layouts,
                                 elementNum + 1, numOfEls - elementNum - 1);

                if (choiceType == ChoiceGroup.MULTIPLE) {
                    System.arraycopy(selEls, elementNum + 1, selEls,
                                     elementNum, numOfEls - elementNum - 1);
                }
            }

            layouts[numOfEls] = null;
            if (choiceType == ChoiceGroup.MULTIPLE) {
                selEls[numOfEls - 1] = false;
            }

            numOfEls--;

            if (numOfEls == 0) {
                hilightedIndex = selectedIndex = -1;
            } else {
                // adjust hilighted index
                if (elementNum < hilightedIndex) {
                    hilightedIndex--;
                } else if (elementNum == hilightedIndex &&
                        hilightedIndex == numOfEls) {
                    hilightedIndex = numOfEls - 1;
                }

                // adjust selected index if choiceGroup is not MULTIPLE
                if (choiceType != ChoiceGroup.MULTIPLE) {
                    if (elementNum < selectedIndex) {
                        selectedIndex--;
                    } else if (elementNum == selectedIndex &&
                            selectedIndex == numOfEls) {
                        selectedIndex = numOfEls - 1;
                    }
                }
                }

                int firstIndexY = layouts[0].getHeight();
                int hilightedIndexY = hilightedIndex == -1 ? 0 : getYpos(
                0, firstIndexY, hilightedIndex);
                Screen owner = getOwner();
                if (owner != null) {
                    owner.contentChanged(null, 0, hilightedIndexY,
					 deletedHeight);
                }
        } // synchronized
    }

    /**
     * @param elementNum the index of the element to be set
     * @param stringPart the string part of the new element
     * @param imagePart the image part of the element, or null if there is
     * no image part
     * @throws IndexOutOfBoundsException if elementNum is invalid
     * @throws IllegalArgumentException if the image is mutable
     * @throws NullPointerException if stringPart is null
     */
    public void set(int elementNum, String stringPart, Image imagePart) {
        synchronized (Display.LCDUILock) {
            checkIndex(elementNum);
            checkNull(stringPart, imagePart);
            setImpl(elementNum, stringPart, imagePart);
        }
    }

    /**
     * @param elementNum the index of the element to be queried
     *
     * @return selection state of the element
     * 
     * @throws IndexOutOfBoundsException if elementNum is invalid
     */
    public boolean isSelected(int elementNum) {
        synchronized (Display.LCDUILock) {
            checkIndex(elementNum);
            return (choiceType == Choice.MULTIPLE ? selEls[elementNum] :
                                            (selectedIndex == elementNum));
        }
    }

    /**
     * <p>Returns the index number of an element in the ChoiceGroup that is
     * selected. For ChoiceGroup objects of type EXCLUSIVE
     * there is at most one element selected, so
     * this method is useful for determining the user's choice.  Returns -1 if
     * there are no elements in the ChoiceGroup. </p>
     *
     * <p>For ChoiceGroup objects of type MULTIPLE, this always
     * returns -1 because no
     * single value can in general represent the state of such a ChoiceGroup.
     * To get the complete state of a MULTIPLE Choice, see {@link
     * #getSelectedFlags getSelectedFlags}.</p>
     *
     * @return index of selected element, or -1 if none
     *
     * @see #setSelectedIndex
     */
    public int getSelectedIndex() { 
        // SYNC NOTE: return of atomic value, no locking necessary
        return selectedIndex;
    }

    /**
     * <p>Queries the state of a ChoiceGroup and returns the state
     * of all elements in the
     * boolean array selectedArray_return. NOTE: this is a result parameter.
     * It must be at least as long as the size
     * of the ChoiceGroup as returned by size().
     * If the array is longer, the extra
     * elements are set to false.</p>
     * 
     * <p>For ChoiceGroup objects of type MULTIPLE, any
     * number of elements may be selected and set to true in the result
     * array.  For ChoiceGroup objects of type EXCLUSIVE,
     * exactly one element will be selected, unless there are
     * zero elements in the ChoiceGroup. </p>
     *
     * @return the number of selected elements in the ChoiceGroup
     *
     * @param selectedArray_return array to contain the results.
     * @throws IllegalArgumentException if selectedArray_return is shorter
     * than the size of the ChoiceGroup.
     * @throws NullPointerException if selectedArray_return is null.
     * 
     * @see #setSelectedFlags
     */
    public int getSelectedFlags(boolean[] selectedArray_return) {
        synchronized (Display.LCDUILock) {
            checkFlag(selectedArray_return);
            int selectedNum = 0;
            if (choiceType == Choice.MULTIPLE) {
                System.arraycopy(selEls, 0, selectedArray_return, 0, numOfEls);
                for (int i = 0; i < numOfEls; i++) {
                    if (selEls[i]) selectedNum++;
                }
                for (int i = numOfEls; i < selectedArray_return.length; i++) {
                    selectedArray_return[i] = false;
                }
            } else {
                for (int i = 0; i < selectedArray_return.length; i++) {
                    selectedArray_return[i] = false;
                }
                if (selectedIndex != -1) {
                    selectedArray_return[selectedIndex] = true;
                    selectedNum = 1;
                }
            }
            return selectedNum;
        }
    }

    /**
     * <P>For ChoiceGroup objects of type MULTIPLE, this simply sets an
     * individual element's selected state. </P>
     *
     * <P>For ChoiceGroup objects of type EXCLUSIVE, this can be used only to
     * select an element.  That is, the <code> selected </code> parameter must
     * be <code> true </code>. When an element is selected, the previously
     * selected element is deselected. If <code> selected </code> is <code>
     * false </code>, this call is ignored.</P>
     *
     * <p>For both list types, the elementNum parameter must be within the range
     * [0..size()-1], inclusive. </p>
     * 
     * @param elementNum the number of the element. Indexing of the elements
     * is zero-based.
     * @param selected the new state of the element true=selected, 
     * false=not selected.
     * @throws IndexOutOfBoundsException if elementNum is invalid
     *
     * @see #getSelectedIndex
     */
    public void setSelectedIndex(int elementNum, boolean selected) {
        synchronized (Display.LCDUILock) {
            checkIndex(elementNum);
            Screen owner = getOwner();

            switch (choiceType) {
                case Choice.IMPLICIT:
                    // hilight changes as a result of selection only
                    // if it is an implicit choice
                    hilightedIndex = elementNum;
                    /* falls through */
                case Choice.EXCLUSIVE:
                    // selected item cannot be deselected
                    if (selectedIndex == elementNum && !selected) {
                        return;
                    }
                    int oldSelected = selectedIndex;
                    selectedIndex = elementNum;
                    if (owner != null && owner.isShown()) {
                        repaintElements(oldSelected, selectedIndex);
                    }
                    break;
                case Choice.MULTIPLE:
                    selEls[elementNum] = selected;
                    if (owner != null && owner.isShown()) {
                        repaintElements(elementNum, -1);
                    }
                    break;
            }
        } // synchronized
    }

    /**
     * Attempts to set the selected state of every element in the 
     * ChoiceGroup. The array
     * must be at least as long as the size of the ChoiceGroup. If the array is
     * longer, the additional values are ignored. <p>
     *
     * For ChoiceGroup objects of type MULTIPLE, this sets the
     * selected state of every
     * element in the Choice. An arbitrary number of elements may be selected.
     * <p>
     *
     * For ChoiceGroup objects of type EXCLUSIVE, exactly one array
     * element must have the value <code>true</code>. If no element
     * is true, the first element
     * in the Choice will be selected. If two or more elements are true, the
     * implementation will choose the first true element and select it. <p>
     *
     * @param selectedArray an array in which the method collect
     *  the selection status
     * @throws IllegalArgumentException if selectedArray is shorter than the
     * size of the ChoiceGroup.
     * @throws NullPointerException if the selectedArray is null.
     *
     * @see #getSelectedFlags
     */
    public void setSelectedFlags(boolean[] selectedArray) {
        synchronized (Display.LCDUILock) {
            checkFlag(selectedArray);
            if (numOfEls == 0) {
                return;
            }

            if (choiceType == Choice.MULTIPLE) {
                System.arraycopy(selectedArray, 0, selEls, 0, numOfEls);

                // repaint if ChoiceGroup was added to Form or List
                repaint(0, 0, Display.WIDTH, height);
            } else {
                int i = 0;
                for (; i < numOfEls; i++) {
                    if (selectedArray[i]) {
                        break;
                    }
                }
                if (i == numOfEls) {
                    i = 0;
                }
                setSelectedIndex(i, true);
            }
        } // synchronized
    }

    /**
     * Sets the label of the Item. If label is null, specifies that this
     * item has no label.
     * @param label the label string
     */
    public void setLabel(String label) {
        synchronized (Display.LCDUILock) {
            super.setLabel(label);

            int deltaHeight = ((StringLayout)layouts[0]).setString(label);
            height += deltaHeight;
            contentChanged(0, 0, deltaHeight);
        }
    }

    // package private

    /**
     * Signal this ChoiceGroup that its focus state has changed
     */
    void focusChanged() {
        repaintElements(hilightedIndex, -1);
    }

    /**
     * Traverse this ChoiceGroup
     *
     * @param dir The direction of traversal, UP or DOWN
     * @param top
     * @param bottom
     * @param traversingIn
     * @param allowTraverseOut
     * @return int  The number of pixels traversed
     */
    int traverse(int dir, int top, int bottom, 
                 boolean traversingIn, boolean allowTraverseOut) {

        boolean down = dir == Canvas.DOWN;
 
        // If there are no items scolling is done only inside the label
        if (numOfEls == 0) {
            return Form.traverse(down, top, bottom, getHeight(), 
                                 Screen.CONTENT_HEIGHT, 
                                 traversingIn, false);
        }

        // When we are traversingIn
        // we should scroll by as much as we could
        if (traversingIn) {
            int s = Form.traverse(down, top, bottom, getHeight(), 
                                  Screen.CONTENT_HEIGHT, traversingIn, true);
            if (down) {
                // first item will not be made visible after scroll
                if (bottom + s < layouts[1].getHeight()) {
                    return s;
                }
                hilightedIndex = 0;
            } else {
                hilightedIndex = numOfEls-1;
            }

            if (s == 0) {
                repaintElements(hilightedIndex, -1);
            }
            return s;
        }

        // top and bottom in the coordinate system of hilightedIndex
        int hilightedY = getYpos(0, layouts[0].getHeight(), hilightedIndex);
        int hilTop = top - hilightedY;
        int hilBottom = bottom - hilightedY;

        // see if traversal within current item is needed
        int s = Form.traverse(down, hilTop, hilBottom,
                              layouts[hilightedIndex+1].getHeight(),
                              layouts[hilightedIndex+1].getLineHeight(),
                              false, true);

        // move to the next item
        if (s < 0) {

            // traversal within bottom most item was done
            if (down && hilightedIndex == numOfEls-1) {
                return s;
            }

            if (!down && hilightedIndex == 0) {
                // traversal within top most item was done                
                return Form.traverse(down, top, bottom, getHeight(), 
				     Screen.CONTENT_HEIGHT, 
				     traversingIn, true);
            }

            // move hilight to the next item
            int oldHilight = hilightedIndex;
            
            if (down) {
                hilightedY += layouts[hilightedIndex+1].getHeight();
                hilightedIndex++;
            } else {
                hilightedIndex--;
                hilightedY -= layouts[hilightedIndex+1].getHeight();
            }
            
            hilTop = top - hilightedY;
            hilBottom = bottom - hilightedY;
            
            // see if traversal within current item is needed
            s = Form.traverse(down, hilTop, hilBottom,
                              layouts[hilightedIndex+1].getHeight(),
                              layouts[hilightedIndex+1].getLineHeight(),
                              true, true);

            if (choiceType == Choice.IMPLICIT) {
                // in IMPLICIT list hilight and selection go together
                selectedIndex = hilightedIndex;
            }
            
            // no scrolling is needed 
            // (entire new hilighted element is visible)
            if (hilightedIndex != -1 &&
                top <= hilightedY && 
                bottom >= hilightedY+layouts[hilightedIndex+1].getHeight()) {
                repaintElements(oldHilight, hilightedIndex);
                return 0;
            }
        }

        return s;
    }

    /**
     * Determine if this ChoiceGroup can take the focus or not
     *
     * @return boolean  Wether this ChoiceGroup has enough elements to
     *                  take the focus
     */
    boolean takesFocus() {
        return numOfEls > 0;
    }

    /**
     * Initialize the highlight of this ChoiceGroup
     *
     * @param vpY
     * @param vpH
     * @return int Always returns 0
     */
    int initHilight(int vpY, int vpH) {
        if (numOfEls > 0) {
            hilightedIndex = 0;
            if (choiceType == List.IMPLICIT) {
                selectedIndex = 0;
            }
        }

        return 0;
    }

    /**
     * Perform a selection on this ChoiceGroup
     *
     * @return boolean Flag of wether the selection succeeded or not
     */
    boolean select() {
        if (numOfEls == 0) {
            return false;
        }

        switch (choiceType) {
            case Choice.IMPLICIT:
                return getOwner() instanceof List;
            case Choice.EXCLUSIVE:
                if (hilightedIndex == selectedIndex) {
                   return false;
                }
                setSelectedIndex(hilightedIndex, true);
                return getOwner() instanceof Form;
            case Choice.MULTIPLE:
                setSelectedIndex(hilightedIndex, !selEls[hilightedIndex]);
               return getOwner() instanceof Form;
            default:
                return false;
        }
    }

    /**
     * Paint this ChoiceGroup given the supplied Graphics context
     *
     * @param g The Graphics object to paint to
     */
    void paint(Graphics g) {
        int clipY1 = g.getClipY();
        int clipY2 = clipY1 + g.getClipHeight();

        int translatedY = 0;

        if (translatedY < clipY2) {
            layouts[0].paint(g, true, false);
            int h = layouts[0].getHeight();
            g.translate(0, h);
            translatedY += h;
        }

        int imgWidth = 0;
        if (choiceType == Choice.EXCLUSIVE) {
            imgWidth = RD_WIDTH;
        } else if (choiceType == Choice.MULTIPLE) {
            imgWidth = CKBX_WIDTH;
        }

        boolean itemIsInFocus = getOwner() instanceof List ?
                                true : hasFocus();

        for (int i = 0; i < numOfEls && translatedY < clipY2; i++) {
            Layout layout = layouts[i + 1];
            int h = layout.getHeight();

            if (translatedY + h >= clipY1) {
                boolean inverted = (i == hilightedIndex && itemIsInFocus);
                if (choiceType == Choice.IMPLICIT) {
                   layout.paint(g, true, inverted);
                } else {
                    Image img = null;
                    if (choiceType == Choice.EXCLUSIVE) {
                        img = (i == selectedIndex) ? RD_ON_IMG : RD_OFF_IMG;
                    } else {
                        img = (selEls[i]) ? CKBX_ON_IMG : CKBX_OFF_IMG;
                    }

                    g.setColor(Display.ERASE_COLOR);
                    g.fillRect(0, 0, imgWidth + 2, h);
               
                    g.drawImage(img, 1, (layout.getLineHeight() + 1) / 2,
                                Graphics.LEFT | Graphics.VCENTER);

                    g.translate(imgWidth + 2, 0);
                    layout.paint(g, true, inverted);
                    g.translate(-imgWidth-2, 0);
                }
            }
            
            g.translate(0, h);
            translatedY += h;
        }

        g.translate(0, -translatedY);
    }

    /**
     * Get the height of this ChoiceGroup
     *
     * @return int The height of this ChoiceGroup
     */
    int getHeight() {
        return height;
    }

    /**
     * Set the width for this ChoiceGroup to lay out with.
     *
     * @param width The allowable width for this ChoiceGroup
     * @return int  The height required by this ChoiceGroup to
     *              display given the allowable width
     */
    int setWidth(int width) {
        if (width > 0) {
            height = layouts[0].setWidth(width);

            if (choiceType == Choice.EXCLUSIVE) {
                width -= RD_WIDTH + 2;
            } else if (choiceType == Choice.MULTIPLE) {
                width -= CKBX_WIDTH + 2;
            }

            for (int i = 0; i < numOfEls; i++) {
                height += layouts[i + 1].setWidth(width);
            }
        }
        return height;
    }

    // private

    /**
     * Set a particular element of this ChoiceGroup
     *
     * @param elementNum The index of the element to establish
     * @param stringPart The string part of the element to establish
     * @param imagePart The image part of the element to establish
     */
    private void setImpl(int elementNum,
                         String stringPart, Image imagePart) {
        // indent 2nd line onward only for IMPLICIT list AND
        // there is not image
        boolean offset = choiceType == Choice.IMPLICIT && imagePart == null;

        if (initLayoutDone()) {
            int deltaHeight = 0;
            if (layouts[elementNum + 1] != null) {
                deltaHeight = -layouts[elementNum + 1].getHeight();
            }
            
            layouts[elementNum + 1] = new IaSLayout(stringPart, imagePart,
                                                    Screen.CONTENT_FONT,
                                                    offset);
            int w = Display.WIDTH;
            if (choiceType == Choice.EXCLUSIVE) {
                w -= RD_WIDTH + 2;
            } else if (choiceType == Choice.MULTIPLE) {
                w -= CKBX_WIDTH + 2;
            }
            deltaHeight += layouts[elementNum + 1].setWidth(w);
            height += deltaHeight;

            contentChanged(0, getYpos(0, layouts[0].getHeight(), elementNum),
                           deltaHeight);
        } else {
            layouts[elementNum + 1] = new IaSLayout(stringPart, imagePart,
                                                    Screen.CONTENT_FONT,
                                                    offset);
        }
    }

    /**
     * Insert a particular element of this ChoiceGroup
     *
     * @param elementNum The index to insert the element
     * @param stringPart The string part of the element to insert
     * @param imagePart The image part of the element to insert
     * @return int  The index of the newly inserted element
     */
    private int insertImpl(int elementNum, String stringPart,
                           Image imagePart) {

        if (numOfEls + 1 == layouts.length) {
            Layout[] newLayouts = new Layout[layouts.length + 4];
            System.arraycopy(layouts, 0, newLayouts, 0, elementNum + 1);
            System.arraycopy(layouts, elementNum + 1, newLayouts,
                             elementNum + 2, numOfEls - elementNum);
            layouts = newLayouts;
        } else {
            System.arraycopy(layouts, elementNum + 1, layouts, elementNum + 2,
                             numOfEls - elementNum);
        }

        if (choiceType == Choice.MULTIPLE) {
            if (selEls.length == numOfEls) {
                boolean newSelEls[] = new boolean[numOfEls + 4];
                System.arraycopy(selEls, 0, newSelEls, 0, elementNum);
                System.arraycopy(selEls, elementNum, newSelEls, elementNum + 1,
                                 numOfEls - elementNum);
                selEls = newSelEls;
            } else {
                System.arraycopy(selEls, elementNum, selEls, elementNum + 1,
                                 numOfEls - elementNum);
            }
            selEls[elementNum] = false;
        }

        layouts[elementNum + 1] = null;
        numOfEls++;

        if (choiceType != Choice.MULTIPLE &&
                (elementNum < selectedIndex || selectedIndex == -1)) {
            selectedIndex++;
        }

        if (elementNum < hilightedIndex || hilightedIndex == -1) {
            hilightedIndex++;
        }

        setImpl(elementNum, stringPart, imagePart);

        return elementNum;
    }

    /**
     * Repaint the elements of this ChoiceGroup in the range of
     * first and second.
     *
     * @param first The start index of the elements to repaint
     * @param second The end index of the elements to repaint
     */
    private void repaintElements(int first, int second) {
        if (first == - 1 && second == -1) {
            return;
        }

        if ((first > second && second != -1) || first == -1) {
            int tmp = first; first = second; second = tmp;
        }

        int translatedY = layouts[0].getHeight();

        Screen owner = getOwner();
        int itemY = owner instanceof Form ? ((Form)owner).getItemY(this) : 0;

        if (first >= 0 && first < numOfEls) {
            translatedY = getYpos(0, translatedY, first);

            // Short-cut when two items are next to each other
            if (first + 1 == second) {
                owner.repaintContent(0, itemY + translatedY, Display.WIDTH,
                                     layouts[first + 1].getHeight() +
                                     layouts[second + 1].getHeight());
                return;
            }
            owner.repaintContent(0, itemY + translatedY, Display.WIDTH,
                                 layouts[first + 1].getHeight());
        }

        if (second >= 0 && second < numOfEls) {
            translatedY = getYpos(first, translatedY, second);
            owner.repaintContent(0, itemY + translatedY, Display.WIDTH,
                                 layouts[second + 1].getHeight());
        }
    }

    /**
     * Get the y coordinate of the element at the given index with the
     * given start index and y coordinate starting position of the element
     * at that start index
     *
     * @param startIndex The index of the starting element
     * @param yPosOfStart The y coordinate position of the starting element
     * @param elementNum The index of the element to get the y position for
     * @return int The y coordinate position of the element at index elementNum
     */
    private int getYpos(int startIndex, int yPosOfStart, int elementNum) {
        int y = yPosOfStart < 0 ? 0 : yPosOfStart;

        for (int i = startIndex; i < elementNum; i++) {
            y += layouts[i + 1].getHeight();
        }
        return y;
    }

    /**
     * Check the validity of a given element index
     *
     * @param elementNum The index to check
     * @throws IndexOutOfBoundsException If no element exists at the
     *                                   that index
     */
    private void checkIndex(int elementNum) {
        if (elementNum < 0 || elementNum >= numOfEls) {
            throw new IndexOutOfBoundsException();
        }
    }

    /**
     * Check the given values for null
     *
     * @param stringPart The string part of the element
     * @param imagePart The image part of the element
     * @throws NullPointerException If the string part is null
     * @throws IllegalArgumentException If the image part is null
     *                                  or mutable
     */
    private void checkNull(String stringPart, Image imagePart) {
        if (stringPart == null) {
            throw new NullPointerException();
        }

        if (imagePart != null && imagePart.isMutable()) {
            throw new IllegalArgumentException();
        }
    }

    /**
     * Check the validity of the selection array
     *
     * @param flag  The array of boolean flags representing the
     *              selected state of the elements
     * @throws NullPointerException If the flag array is null
     * @throws IllegalArgumentException If the flag array is not
     *                                  the same size as the element array
     */
    private void checkFlag(boolean[] flag) {
        if (flag == null) {
            throw new NullPointerException();
        }

        if (flag.length < numOfEls) {
            throw new IllegalArgumentException();
        }
    }

    /**
     * Image to represent an unselected checkbox
     */
    private static final Image CKBX_OFF_IMG;
    /**
     * Image to represent a selected checkbox
     */
    private static final Image CKBX_ON_IMG;
    /**
     * Width of a checkbox image (both selected and unselected)
     */
    private static final int CKBX_WIDTH = 10;
    /**
     * Image to represent an unselected radio button
     */
    private static final Image RD_OFF_IMG;
    /**
     * Image to represent a selected radio button
     */
    private static final Image RD_ON_IMG;
    /**
     * Width of a radio button image (both selected and unselected)
     */
    private static final int RD_WIDTH = 11;

    static {
        /**
         * Byte array used to create an unselected checkbox
         */
        byte ckbx_off[] = {
            (byte)0x89, (byte)0x50, (byte)0x4e, (byte)0x47,
            (byte)0x0d, (byte)0x0a, (byte)0x1a, (byte)0x0a,
            (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x0d,
            (byte)0x49, (byte)0x48, (byte)0x44, (byte)0x52,
            (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x0a,
            (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x0b,
            (byte)0x02, (byte)0x03, (byte)0x00, (byte)0x00,
            (byte)0x00, (byte)0x3b, (byte)0x00, (byte)0xf4,
            (byte)0x8b, (byte)0x00, (byte)0x00, (byte)0x00,
            (byte)0x06, (byte)0x50, (byte)0x4c, (byte)0x54,
            (byte)0x45, (byte)0xff, (byte)0xff, (byte)0xff,
            (byte)0xbb, (byte)0xbb, (byte)0xbb, (byte)0xc7,
            (byte)0x4b, (byte)0xdf, (byte)0xf8, (byte)0x00,
            (byte)0x00, (byte)0x00, (byte)0x1e, (byte)0x49,
            (byte)0x44, (byte)0x41, (byte)0x54, (byte)0x78,
            (byte)0xda, (byte)0x63, (byte)0x60, (byte)0x60,
            (byte)0x60, (byte)0x60, (byte)0x0c, (byte)0x65,
            (byte)0xf8, (byte)0xcd, (byte)0xe0, (byte)0xc0,
            (byte)0x20, (byte)0xc0, (byte)0xc4, (byte)0xc0,
            (byte)0xc0, (byte)0x80, (byte)0x0b, (byte)0x33,
            (byte)0x86, (byte)0x32, (byte)0xfc, (byte)0x06,
            (byte)0x00, (byte)0x3f, (byte)0x74, (byte)0x03,
            (byte)0x01, (byte)0x35, (byte)0x94, (byte)0x27,
            (byte)0xe6, (byte)0x00, (byte)0x00, (byte)0x00,
            (byte)0x00, (byte)0x49, (byte)0x45, (byte)0x4e,
            (byte)0x44, (byte)0xae, (byte)0x42, (byte)0x60,
            (byte)0x82,
        };

        /**
         * Byte array used to create a selected checkbox
         */
        byte ckbx_on[] = {
            (byte)0x89, (byte)0x50, (byte)0x4e, (byte)0x47,
            (byte)0x0d, (byte)0x0a, (byte)0x1a, (byte)0x0a,
            (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x0d,
            (byte)0x49, (byte)0x48, (byte)0x44, (byte)0x52,
            (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x0a,
            (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x0b,
            (byte)0x02, (byte)0x03, (byte)0x00, (byte)0x00,
            (byte)0x00, (byte)0x3b, (byte)0x00, (byte)0xf4,
            (byte)0x8b, (byte)0x00, (byte)0x00, (byte)0x00,
            (byte)0x0c, (byte)0x50, (byte)0x4c, (byte)0x54,
            (byte)0x45, (byte)0xff, (byte)0xff, (byte)0xff,
            (byte)0xbb, (byte)0xbb, (byte)0xbb, (byte)0x6a,
            (byte)0x6a, (byte)0x6a, (byte)0x02, (byte)0x02,
            (byte)0x02, (byte)0x6a, (byte)0x99, (byte)0x0e,
            (byte)0xd1, (byte)0x00, (byte)0x00, (byte)0x00,
            (byte)0x35, (byte)0x49, (byte)0x44, (byte)0x41,
            (byte)0x54, (byte)0x78, (byte)0xda, (byte)0x63,
            (byte)0x60, (byte)0x60, (byte)0x7c, (byte)0xc0,
            (byte)0xf0, (byte)0xff, (byte)0xb3, (byte)0x03,
            (byte)0xd3, (byte)0x41, (byte)0x91, (byte)0x0f,
            (byte)0x4c, (byte)0x0c, (byte)0x6c, (byte)0x0c,
            (byte)0x4c, (byte)0x0c, (byte)0xfc, (byte)0x0c,
            (byte)0x4c, (byte)0x0c, (byte)0x12, (byte)0x0c,
            (byte)0x4c, (byte)0xbc, (byte)0x7f, (byte)0x18,
            (byte)0x98, (byte)0x7e, (byte)0x2d, (byte)0x60,
            (byte)0x60, (byte)0xf9, (byte)0xf5, (byte)0x93,
            (byte)0x81, (byte)0xe9, (byte)0xbf, (byte)0x03,
            (byte)0x03, (byte)0xe3, (byte)0x7f, (byte)0x86,
            (byte)0x8f, (byte)0x00, (byte)0x05, (byte)0x25,
            (byte)0x0c, (byte)0xde, (byte)0x01, (byte)0x9d,
            (byte)0xad, (byte)0x64, (byte)0x00, (byte)0x00,
            (byte)0x00, (byte)0x00, (byte)0x49, (byte)0x45,
            (byte)0x4e, (byte)0x44, (byte)0xae, (byte)0x42,
            (byte)0x60, (byte)0x82,
        };

        /**
         * Byte array used to create an unselected radio button
         */
        byte rad_off[] = {
            (byte)0x89, (byte)0x50, (byte)0x4e, (byte)0x47,
            (byte)0x0d, (byte)0x0a, (byte)0x1a, (byte)0x0a,
            (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x0d,
            (byte)0x49, (byte)0x48, (byte)0x44, (byte)0x52,
            (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x0b,
            (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x0b,
            (byte)0x02, (byte)0x03, (byte)0x00, (byte)0x00,
            (byte)0x00, (byte)0xd4, (byte)0xc2, (byte)0x9f,
            (byte)0xb5, (byte)0x00, (byte)0x00, (byte)0x00,
            (byte)0x09, (byte)0x50, (byte)0x4c, (byte)0x54,
            (byte)0x45, (byte)0xff, (byte)0xff, (byte)0xff,
            (byte)0xbb, (byte)0xbb, (byte)0xbb, (byte)0x6a,
            (byte)0x6a, (byte)0x6a, (byte)0xd4, (byte)0x0d,
            (byte)0xff, (byte)0xe8, (byte)0x00, (byte)0x00,
            (byte)0x00, (byte)0x2a, (byte)0x49, (byte)0x44,
            (byte)0x41, (byte)0x54, (byte)0x78, (byte)0xda,
            (byte)0x63, (byte)0x60, (byte)0x5a, (byte)0xc5,
            (byte)0xc0, (byte)0x20, (byte)0xc1, (byte)0x30,
            (byte)0x81, (byte)0x41, (byte)0x81, (byte)0x41,
            (byte)0x81, (byte)0x29, (byte)0x81, (byte)0xe1,
            (byte)0x05, (byte)0x13, (byte)0x03, (byte)0x03,
            (byte)0x03, (byte)0x32, (byte)0x66, (byte)0x50,
            (byte)0x60, (byte)0x50, (byte)0x60, (byte)0xfa,
            (byte)0xc1, (byte)0x50, (byte)0xc0, (byte)0xc0,
            (byte)0xb4, (byte)0x8a, (byte)0x01, (byte)0x00,
            (byte)0x71, (byte)0x18, (byte)0x05, (byte)0x3d,
            (byte)0x49, (byte)0xbc, (byte)0x9c, (byte)0x62,
            (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00,
            (byte)0x49, (byte)0x45, (byte)0x4e, (byte)0x44,
            (byte)0xae, (byte)0x42, (byte)0x60, (byte)0x82,
        };

        /**
         * Byte array used to create a selected radio button
         */
        byte rad_on[] = {
            (byte)0x89, (byte)0x50, (byte)0x4e, (byte)0x47,
            (byte)0x0d, (byte)0x0a, (byte)0x1a, (byte)0x0a,
            (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x0d,
            (byte)0x49, (byte)0x48, (byte)0x44, (byte)0x52,
            (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x0b,
            (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x0b,
            (byte)0x02, (byte)0x03, (byte)0x00, (byte)0x00,
            (byte)0x00, (byte)0xd4, (byte)0xc2, (byte)0x9f,
            (byte)0xb5, (byte)0x00, (byte)0x00, (byte)0x00,
            (byte)0x0c, (byte)0x50, (byte)0x4c, (byte)0x54,
            (byte)0x45, (byte)0xff, (byte)0xff, (byte)0xff,
            (byte)0xbb, (byte)0xbb, (byte)0xbb, (byte)0x6a,
            (byte)0x6a, (byte)0x6a, (byte)0x02, (byte)0x02,
            (byte)0x02, (byte)0x6a, (byte)0x99, (byte)0x0e,
            (byte)0xd1, (byte)0x00, (byte)0x00, (byte)0x00,
            (byte)0x32, (byte)0x49, (byte)0x44, (byte)0x41,
            (byte)0x54, (byte)0x78, (byte)0xda, (byte)0x05,
            (byte)0xc1, (byte)0x41, (byte)0x0d, (byte)0x00,
            (byte)0x20, (byte)0x0c, (byte)0x04, (byte)0xc1,
            (byte)0xed, (byte)0xf5, (byte)0x8b, (byte)0x01,
            (byte)0x12, (byte)0x44, (byte)0x9c, (byte)0x0a,
            (byte)0x34, (byte)0x13, (byte)0x5e, (byte)0x48,
            (byte)0x40, (byte)0x11, (byte)0x65, (byte)0x86,
            (byte)0x2c, (byte)0xd4, (byte)0xe3, (byte)0x62,
            (byte)0xcc, (byte)0x59, (byte)0x4d, (byte)0x9a,
            (byte)0x28, (byte)0x02, (byte)0x55, (byte)0xa1,
            (byte)0xb7, (byte)0xc1, (byte)0x98, (byte)0xc1,
            (byte)0x25, (byte)0x8b, (byte)0x0f, (byte)0xd8,
            (byte)0x77, (byte)0x0a, (byte)0x29, (byte)0x4e,
            (byte)0x1a, (byte)0xb1, (byte)0x3b, (byte)0x00,
            (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x49,
            (byte)0x45, (byte)0x4e, (byte)0x44, (byte)0xae,
            (byte)0x42, (byte)0x60, (byte)0x82,
        };

        CKBX_OFF_IMG = Image.createImage(ckbx_off, 0, ckbx_off.length);
        CKBX_ON_IMG  = Image.createImage(ckbx_on, 0, ckbx_on.length);
        RD_OFF_IMG   = Image.createImage(rad_off, 0, rad_off.length);
        RD_ON_IMG    = Image.createImage(rad_on, 0, rad_on.length);
    }

    /**
     * This class handles the layout for a choice group item, which consists
     * of an image (optionally) and a string. When the image is less the half
     * the display width, the string appears on the same line as the image.
     * When the image is more than half the display width, the string appears
     * below the image, indented by 3 white spaces.
     *
     * This class extends the StringLayout class to allow the superclass to
     * handle the string layout.
     */
    class IaSLayout extends StringLayout {
        /**
         * Image of the element
         */
        private Image img;
        /**
         * Height and width of the image of the element
         */
        private int imgHeight, imgWidth;

        /**
         * This is the total height and width of the item. This is different
         * from the sum of the image and string height and width, because
         * depending on the width of the image, they could be side-by-side,
         * or on-top of each other.
         */
        private int layoutedHeight, layoutedWidth;

        /**
         * Construct a new IaSLayout
         *
         * @param str The string part of the element
         * @param img The image part of the element
         * @param font The Font used to render the string of the element
         * @param offset The offset of the text from the image of the element
         */
        IaSLayout(String str, Image img, Font font, boolean offset) {
            super(str, font);

            this.img = img;
            if (img != null) {
                imgHeight = img.getHeight();
                // pad the image with same spaces before the string
                imgWidth  = img.getWidth() + 2;
            }
            setOffset(offset);
        }

        /**
         * Get the image for this element
         *
         * @return Image The image to render for this element
         */
        Image getImage() {
            return img;
        }

        /**
         * Get the width of this element
         *
         * @return int  The width of this element
         */
        public int getWidth() {
            return layoutedWidth;
        }

        /**
         * Set the allowable width for this element
         *
         * @param w The allowable width for this element
         * @return int  The height required to display this element
         *              given the allowable width
         */
        public int setWidth(int w) {
            if ((layoutedWidth == w) || (w < 0)) {
                return layoutedHeight;
            }

            layoutedWidth = w;

            if (imgWidth > (w / 2)) {
                // height of the item is the sum of the image and 
                // string heights if the image is too wide.
                layoutedHeight = super.setWidth(w - offsetWidth) + imgHeight;
            } else {
                // return the taller of the image or the text
                int superHeight = super.setWidth(w - imgWidth);
                layoutedHeight = imgHeight;
                if (superHeight > layoutedHeight) {
                    layoutedHeight = superHeight;
                }
            }

            return layoutedHeight;
        }

        /**
         * Get the height of this element
         *
         * @return int The height of this element
         */
        public int getHeight() {
            return layoutedHeight;
        }

        /**
         * Paint this element
         *
         * @param g The Graphics context to paint to
         * @param eraseBgrnd    A flag to erase the background before painting
         * @param inverted  A flag to signal the element should be painted in
         *                  inverted colors
         */
        public void paint(Graphics g,
                          boolean eraseBgrnd, boolean inverted) {
            // clear the background
            if (eraseBgrnd) {
                g.setColor(inverted ?
                    Display.BG_H_COLOR : Display.ERASE_COLOR);
                g.fillRect(0, 0, layoutedWidth, layoutedHeight);
            }

            if (img != null) {
                // center the image vertically in a line, only
                // when the image is less than a line height, and there is
                // only one line of text.
                if (imgWidth < (Display.WIDTH / 2) &&
                        layoutedHeight <= lineHeight) {
                    g.drawImage(img, 0, lineHeight / 2,
                                Graphics.VCENTER | Graphics.LEFT);
                } else {
                    g.drawImage(img, 0, 0, Graphics.TOP | Graphics.LEFT);
                }
            }

            // depending on the width of the image, the string appears to the
            // right or below (and indented) of the image.
            int tx = 0;
            int ty = 0;
            if (imgWidth > (Display.WIDTH / 2)) {
                ty = imgHeight;
                tx = offsetWidth;
            } else {
                tx = imgWidth;
            }
            g.translate(tx, ty);
            super.paint(g, false, inverted);
            g.translate(-tx, -ty);
        }
    } // end of IaStringLayout
}

