/*
 * @(#)DateField.java	1.98 01/09/17
 * Copyright (c) 1999-2001 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package javax.microedition.lcdui;

import java.util.Calendar;
import java.util.Date;
import java.util.TimeZone;
import com.sun.midp.lcdui.Resource;

/**
 * A DateField is an editable component for presenting date and time (calendar)
 * information that may be placed into a Form. Value for this field can be
 * initially set or left unset. If value is not set then the UI for the field
 * shows this clearly. The field value for "not initialized state" is not valid
 * value and <code>getDate()</code> for this state returns null.
 * <p>
 * Instance of a DateField can be configured to accept date or time information
 * or both of them. This input mode configuration is done by DATE, TIME or
 * DATE_TIME static fields of this class. DATE input mode allows to set only
 * date information and TIME only time information (hours, minutes). DATE_TIME
 * allows to set both clock time and date values.
 * <p>
 * In TIME input mode the date components of Date object
 * must be set to the "zero epoch" value of January 1, 1970.
 * <p>
 * Calendar calculations in this field are based on default locale and defined
 * time zone. Because of the calculations and different input modes date object
 * may not contain same millisecond value when set to this field and get back
 * from this field.
 */
public class DateField extends Item {
    /**
     * Flag to signal the clock representation uses AM and PM notation
     */
    private static final boolean CLOCK_USES_AM_PM = true;

    /**
     * <P>Input mode for date information (day, month, year). With
     * this mode this
     * DateField presents and allows only to modify date value. The time
     * information of date object is ignored.</P>
     *
     * <P>Value 1 is assigned to DATE.</P>
     */
    public static final int DATE = 1;

    /**
     * <P>Input mode for time information (hours and minutes). With
     * this mode this
     * DateField presents and allows only to modify time. The date components
     * should be set to the "zero epoch" value of January 1, 1970 and
     * should not be
     * accessed.</P>
     *
     * <P>Value 2 is assigned to TIME.</P>
     */
    public static final int TIME = 2;

    /**
     * <P>Input mode for date (day, month, year) and time (minutes, hours)
     * information. With this mode this DateField presents and allows to modify
     * both time and date information.</P>
     *
     * <P>Value 3 is assigned to DATE_TIME.</P>
     */
    public static final int DATE_TIME = 3;

    /**
     * Creates a DateField object with the specified label and mode. This call
     * is identical to <code>DateField(label, mode, null)</code>.
     *
     * @param label item label
     * @param mode the input mode, one of DATE, TIME or DATE_TIME
     * @exception IllegalArgumentException if the input mode's value is invalid
     */
    public DateField(String label, int mode) {
        this(label, mode, null);
    }

    /**
     * Creates a date field in which calendar calculations are based on specific
     * TimeZone object and the default calendaring system for the current
     * locale.
     * The value of the DateField is initially in the "uninitialized" state.
     * If timeZone is null, the system's default time zone is used.
     *
     * @param label item label
     * @param mode the input mode, one of DATE, TIME or DATE_TIME
     * @param timeZone a specific time zone, or null for the default time zone
     * @exception IllegalArgumentException if the input mode's value is invalid
     */
    public DateField(String label, int mode, java.util.TimeZone timeZone) {
        super(label, 3);

        synchronized (Display.LCDUILock) {
            if ((mode != DATE) && (mode != TIME) && (mode != DATE_TIME)) {
                throw new IllegalArgumentException("Invalid input mode");
            }

            this.mode = mode;

            if (timeZone == null) {
                timeZone = TimeZone.getDefault();
            }

            this.currentDate = Calendar.getInstance(timeZone);

            layouts[1] = new StringLayout(
                toString((mode == TIME) ? TIME : DATE), Screen.CONTENT_FONT);
            layouts[2] = new StringLayout(
                (mode == DATE_TIME) ? toString(TIME) : null,
                Screen.CONTENT_FONT);
        } // synchronized
    }

    /**
     * Returns date value of this field. Returned value is null if
     * field value is
     * not initialized. The date object is constructed according the rules of
     * locale specific calendaring system and defined time zone.
     *
     * In TIME mode field the date components are set to the "zero
     * epoch" value of January 1, 1970. If a date object that presents time
     * beyond one day from this "zero epoch" then this field is in "not
     * initialized" state and this method returns null.
     *
     * In DATE mode field the time component of the calendar is set to zero when
     * constructing the date object.
     *
     * @return date object representing time or date depending on input mode
     * @see #setDate
     */
    public java.util.Date getDate() {
        synchronized (Display.LCDUILock) {
            // NOTE: 
            // defensive copy of the Date object is necessary 
            // because CLDC's Calendar returns a reference to an internal, 
            // shared Date object.  See bugID: 4479408.
            return (initialized ? 
                new java.util.Date(currentDate.getTime().getTime()) : null);

        } // synchronized
    }

    /**
     * Sets a new value for this field. Null can be passed to set the field
     * state to "not initialized" state. The input mode of this field defines
     * what components of passed Date object is used.<p>
     *
     * In TIME input mode the date components must be set to the "zero
     * epoch" value of January 1, 1970. If a date object that presents time
     * beyond one day then this field is in "not initialized" state.
     * In TIME input mode the date component of Date object is ignored and time
     * component is used to precision of minutes.<p>
     *
     * In DATE input mode the time component of Date object is ignored.<p>
     *
     * In DATE_TIME input mode the date and time component of Date are used but
     * only to precision of minutes.
     *
     * @param date new value for this field
     * @see #getDate
     */
    public void setDate(java.util.Date date) {
        synchronized (Display.LCDUILock) {
            if (date == null) {
                initialized = false;
            } else {
                currentDate.setTime(date);

                if (mode == TIME) {
                    // NOTE:
                    // It is unclear from the spec what should happen 
                    // when DateField with TIME mode is set to
                    // a value that is on a day other than 1/1/1970.
                    //
                    // Two possible interpretations of the spec are:
                    // 1. DateField is put into the "uninitialized" state; or
                    // 2. The time portion of the DateField is set to the 
                    //    time-of-day portion of the Date object passed in, 
                    //    and the date portion of the DateField is set 
                    //    to 1/1/1970.
                    //
                    // Currently we are using the first approach.
                    initialized =
                        (currentDate.get(Calendar.YEAR)  == 1970) &&
                        (currentDate.get(Calendar.MONTH) == Calendar.JANUARY)
                        && (currentDate.get(Calendar.DATE)   == 1);
                } else {
                    // Currently spec does not prohibit from losing
                    // irrelevant for that mode information
                    // so we always zero out hours and minutes

                    // NOTE: the specification doesn't prohibit 
                    // the loss of information irrelevant to 
                    // the current input mode, so we always zero out the
                    // hours and minutes.
                    if (mode == DATE) {
                        currentDate.set(Calendar.HOUR, 0);
                        currentDate.set(Calendar.MINUTE, 0);
                    }
                    initialized = true;
                }

                // always ignore seconds and milliseconds
                currentDate.set(Calendar.SECOND, 0);
                currentDate.set(Calendar.MILLISECOND, 0);
            }
            contentChanged();
        } // synchronized
    }

    /**
     * Gets input mode for this date field. Valid input modes are
     * DATE, TIME and DATE_TIME.
     *
     * @return input mode of this field
     * @see #setInputMode
     */
    public int getInputMode() {
        // SYNC NOTE: return of atomic value, no locking necessary
            return mode;
    }

    /**
     * Set input mode for this date field. Valid input modes are
     * DATE, TIME and DATE_TIME.
     *
     * @param mode the input mode, must be one of DATE, TIME or DATE_TIME
     * @exception IllegalArgumentException if an invalid value is specified
     * @see #getInputMode
     */
    public void setInputMode(int mode) {
        if ((mode != DATE) && (mode != TIME) && (mode != DATE_TIME)) {
            throw new IllegalArgumentException("Invalid input mode");
        }

        synchronized (Display.LCDUILock) {
            if (this.mode != mode) {
		int oldMode = this.mode;
                this.mode = mode;

                // While the input mode is changed
                // some irrelevant values for new mode could be lost. 
                // Currently that is allowed by the spec.

                // So for TIME mode we make sure that time is set
                // on a zero epoch date
                // and for DATE mode we zero out hours and minutes
                if (mode == TIME) {
                    currentDate.set(Calendar.YEAR, 1970);
                    currentDate.set(Calendar.MONTH, Calendar.JANUARY);
                    currentDate.set(Calendar.DATE, 1);
                } else if (mode == DATE) {
                    currentDate.set(Calendar.HOUR, 0);
                    currentDate.set(Calendar.MINUTE, 0);
                }

		// set layouts to correct date/time information
		// propogate height changes to the form height changed
		int deltaHeight = ((StringLayout)layouts[1]).setString(
                                  toString(mode == TIME ? TIME : DATE));
		if (mode == DATE_TIME) {
		    deltaHeight += ((StringLayout)layouts[2]).setString(
                                     toString(TIME));
		} else if (oldMode == DATE_TIME) {
		    deltaHeight += ((StringLayout)layouts[2]).setString(
                                     null);
		    // if both (DATE and TIME) fields were present and
		    // highlight was on the TIME field and
		    // move the highlight to be on the DATE since
		    // TIME does not exist any more with the mode change
		    if (hasFocus() && highlight == 2) {
			highlight = 1;
		    }
		}
		
		height += deltaHeight;
		contentChanged(0, 0, deltaHeight);
            }
        } // synchronized
    }

    /**
     * Sets the label of the Item. If label is null, specifies that this
     * item has no label.
     * @param label the label string
     */
    public void setLabel(String label) {
        synchronized (Display.LCDUILock) {
            super.setLabel(label);

            int deltaHeight = ((StringLayout)layouts[0]).setString(label);
            height += deltaHeight;
            contentChanged(0, 0, deltaHeight);
        }
    }

    // package private

    /**
     * Called from the Date Editor to save the selected Date.
     *
     * @param date The Date object to which current date should be set.
     */
    void saveDate(java.util.Date date) {
        initialized = true;
        
        currentDate.setTime(date);

        contentChanged();
    }

    /**
     * Paint this DateField
     *
     * @param g The Graphics object to paint to
     */
    void paint(Graphics g) {
        boolean inverted = hasFocus();
        if (layouts != null) {
            int translatedY = 0;
            for (int i = 0; i < layouts.length; i++) {
                int h = layouts[i].getHeight();
                layouts[i].paint(g, true,
                    inverted && (i > 0) && (i == highlight));
                g.translate(0, h);
                translatedY += h;
            }
            g.translate(0, -translatedY);
        }
    }

    /**
     * Get the am/pm text given a Calandar
     *
     * @param calendar The Calendar object to retrieve the time from
     * @return String The am/pm text based on the time in the calendar
     */
    static String ampmString(Calendar calendar) {
        int hour   = calendar.get(Calendar.HOUR_OF_DAY);
        int minute = calendar.get(Calendar.MINUTE);

        if (hour >= 12) {
            return ((minute == 0) && (hour == 12)) ? "noon" : "PM";
        } else {
            return ((minute == 0) && (hour == 00)) ? "mid." : "AM";
        }
    }

    /**
     * Get the day of the week text given a Calendar
     *
     * @param calendar The Calendar object to retrieve the date from
     * @return String The day of the week text based on the date in the
     *                  calendar
     */
    static String dayOfWeekString(Calendar calendar) {
        String str;
        switch (calendar.get(Calendar.DAY_OF_WEEK)) {
        case Calendar.SUNDAY:    str = "Sun"; break;
        case Calendar.MONDAY:    str = "Mon"; break;
        case Calendar.TUESDAY:   str = "Tue"; break;
        case Calendar.WEDNESDAY: str = "Wed"; break;
        case Calendar.THURSDAY:  str = "Thu"; break;
        case Calendar.FRIDAY:    str = "Fri"; break;
        case Calendar.SATURDAY:  str = "Sat"; break;
        default: 
            str = Integer.toString(calendar.get(Calendar.DAY_OF_WEEK));
        }
        return str;
    }

    /**
     * Translate the mode of a DateField into a readable string
     *
     * @param mode The mode to translate
     * @return String A human readable string representing the mode of the
     *              DateField
     */
    String toString(int mode) {
        if (mode == DATE) {
            if (!initialized) {
                    return Resource.getString("<date>");
            }

            return Resource.getDateString(
                dayOfWeekString(currentDate),
                twoDigits(currentDate.get(Calendar.DATE)),
                MONTH_NAMES[currentDate.get(Calendar.MONTH)].substring(0, 3),
                Integer.toString(currentDate.get(Calendar.YEAR)));

        } else if (mode == TIME) {
            if (!initialized) {
                    return Resource.getString("<time>");
            }

            if (CLOCK_USES_AM_PM) {
                return Resource.getTimeString(
                    twoDigits(currentDate.get(Calendar.HOUR)),
                    twoDigits(currentDate.get(Calendar.MINUTE)),
                    twoDigits(currentDate.get(Calendar.SECOND)),
                    ampmString(currentDate));
            } else {
                return Resource.getTimeString(
                    twoDigits(currentDate.get(Calendar.HOUR_OF_DAY)),
                    twoDigits(currentDate.get(Calendar.MINUTE)),
                    twoDigits(currentDate.get(Calendar.SECOND)),
                    null);
            }
        } else {
            if (!initialized) {
                    return Resource.getString("<date/time>");
            }

            return Resource.getDateTimeString(
                dayOfWeekString(currentDate),
                twoDigits(currentDate.get(Calendar.DATE)),
                MONTH_NAMES[currentDate.get(Calendar.MONTH)].substring(0, 3),
                Integer.toString(currentDate.get(Calendar.YEAR)),
                twoDigits(currentDate.get(Calendar.HOUR_OF_DAY)),
                twoDigits(currentDate.get(Calendar.MINUTE)),
                twoDigits(currentDate.get(Calendar.SECOND)),
                ampmString(currentDate));
        }
    }

    /**
     * Get the height of this DateField
     *
     * @return int  The height of this DateField
     */
    int getHeight() {
        return height;
    }

    /**
     * Set the allowable width of this DateField
     *
     * @param width The allowable width of this DateField
     * @return int The height of this DateField based on the
     *              allowable width
     */
    int setWidth(int width) {
        height = 0;
        if (layouts != null) {
            for (int i = 0; i < layouts.length; i++) {
                if (layouts[i] != null) {
                    height += layouts[i].setWidth(width);
                }
            }
        }
        return height;
    }

    /**
     * Initialize the highlight of this DateField
     *
     * @param vpY
     * @param vpH
     * @return int Always returns 0;
     */
    int initHilight(int vpY, int vpH) {
        // Label has focus only if it is taller than the viewPort
        if (layouts[0].getHeight() > vpH) {
            highlight = 0;
        } else {
            highlight = 1;
        }
        return 0;
    }

    /**
     * Perform a selection on this DateField
     *
     * @return boolean Always returns false
     */
    boolean select() {
        Screen returnScreen = getOwner();

        if (editor == null) {
            editor = new EditScreen(returnScreen, this);
        }

        switch (mode) {
        case DATE:
            if (!initialized) {
                currentDate.set(Calendar.HOUR, 0);
                currentDate.set(Calendar.MINUTE, 0);
                currentDate.set(Calendar.SECOND, 0);
                currentDate.set(Calendar.MILLISECOND, 0);
            }
            editor.setDateTime(currentDate.getTime(), DATE);
            break;

        case TIME:
            editor.setDateTime(initialized ? currentDate.getTime() : EPOCH,
                               TIME);
            break;

        case DATE_TIME:
            editor.setDateTime(currentDate.getTime(),
                               (highlight < 2) ? DATE : TIME);
        }

        editor.layoutChanged();

        returnScreen.currentDisplay.setCurrent(editor);

        return false;
    }

    /**
     * Traverse this DateField
     *
     * @param dir
     * @param top
     * @param bottom
     * @param traversingIn
     * @param allowTraverseOut
     * @return int
     */
    int traverse(int dir, int top, int bottom, 
                 boolean traversingIn,  boolean allowTraverseOut) {

        int s = Form.traverse(dir == Canvas.DOWN, top, bottom, 
                              getHeight(), Screen.CONTENT_HEIGHT,
                              traversingIn, takesFocus());

        int labelHeight = layouts[0].getHeight();

        // NOTE: We assume that each Date and Time are
        // represented each by one line

        // Traversing into this item (focus of that item was set to false)
        if (traversingIn) {
            if (dir == Canvas.DOWN) {
                highlight = labelHeight > (bottom - top) ? 0 : 1;
            } else {
                highlight = (mode == DATE_TIME ? 2 : 1);
            }
            return s;
        }

        // Item already has focus (further traversal)
        if (dir == Canvas.DOWN) {
            if (highlight == 0) { // currently label is shown
                // after scrolling by height = s is done
                // portion of Date or Time (depending on the mode)
                // will be displayed and it should be highlighted
                if (labelHeight < bottom + s) {
                    highlight = 1;
                }
            } else if (highlight == 1) { // currently first field has focus
                int endOfField = labelHeight + layouts[1].getHeight();
                // possibly only portion of this field was shown
                if (endOfField > bottom) {
                    s = endOfField - bottom;
                } else if (mode == DATE_TIME) {
                    highlight = 2;
                    // possibly no scrolling is needed to transfer focus
                    if (s < 0) {
                        s = 0;
                    }
                }
            }
        } else {
            if (highlight == 2) {
                highlight = 1;
                // possibly no scrolling is needed to transfer focus
                if (s < 0 || (top < labelHeight)) {
                    s = 0;
                }
            } else if (highlight == 1) {
                // after scrolling by height = s is done
                // only label will be displayed
                if (bottom - s < labelHeight) {
                    highlight = 0;
                }
            }
        }

        if (s == 0 && highlight != 0) {
            repaint(0, labelHeight, Display.WIDTH, height - labelHeight);
        }
        return s;
    }

    /**
     * Static zero epoch Date - used as a default value for
     * uninitialized DateFields with TIME mode.
     */
    static final java.util.Date EPOCH = new java.util.Date(0);

    // REMIND: Needs to be localized?
    /**
     * Static array holding the names of the 12 months
     */
    static final String MONTH_NAMES[] = {
        "January", "February", "March",     "April",   "May",      "June",
        "July",    "August",   "September", "October", "November", "December"
    };

    /**
     * table of trigonometric functions, in 16.16 fixed point
     */
    static final int TRIG_TABLE[] = {
        65535, // cos 0
        65525, // cos 1
        65495, // cos 2
        65445, // cos 3
        65375, // cos 4
        65285, // cos 5
        65175, // cos 6
        65046, // cos 7
        64897, // cos 8
        64728, // cos 9
        64539, // cos 10
        64330, // cos 11
        64102, // cos 12
        63855, // cos 13
        63588, // cos 14
        63301, // cos 15
        62996, // cos 16
        62671, // cos 17
        62327, // cos 18
        61964, // cos 19
        61582, // cos 20
        61182, // cos 21
        60762, // cos 22
        60325, // cos 23
        59869, // cos 24
        59394, // cos 25
        58902, // cos 26
        58392, // cos 27
        57863, // cos 28
        57318, // cos 29
        56754, // cos 30
        56174, // cos 31
        55576, // cos 32
        54962, // cos 33
        54330, // cos 34
        53683, // cos 35
        53018, // cos 36
        52338, // cos 37
        51642, // cos 38
        50930, // cos 39
        50202, // cos 40
        49459, // cos 41
        48701, // cos 42
        47929, // cos 43
        47141, // cos 44
        46340, // cos 45
        45524, // cos 46
        44694, // cos 47
        43851, // cos 48
        42994, // cos 49
        42125, // cos 50
        41242, // cos 51
        40347, // cos 52
        39439, // cos 53
        38520, // cos 54
        37589, // cos 55
        36646, // cos 56
        35692, // cos 57
        34728, // cos 58
        33753, // cos 59
        32767, // cos 60
        31771, // cos 61
        30766, // cos 62
        29752, // cos 63
        28728, // cos 64
        27696, // cos 65
        26655, // cos 66
        25606, // cos 67
        24549, // cos 68
        23485, // cos 69
        22414, // cos 70
        21336, // cos 71
        20251, // cos 72
        19160, // cos 73
        18063, // cos 74
        16961, // cos 75
        15854, // cos 76
        14742, // cos 77
        13625, // cos 78
        12504, // cos 79
        11380, // cos 80
        10251, // cos 81
        9120,  // cos 82
        7986,  // cos 83
        6850,  // cos 84
        5711,  // cos 85
        4571,  // cos 86
        3429,  // cos 87
        2287,  // cos 88
        1143,  // cos 89
        0      // cos 90
    };

    /**
     * Utility method to return the cosine of an angle
     *
     * @param angle The angle to compute the cosine of
     * @return int The cosine of the angle
     */
    static int cos(int angle) {
        angle += 360000;
        angle %= 360;

        if (angle >= 270) {
            return TRIG_TABLE[360 - angle];
        } else if (angle >= 180) {
            return -TRIG_TABLE[angle - 180];
        } else if (angle >= 90) {
            return -TRIG_TABLE[180 - angle];
        } else {
            return TRIG_TABLE[angle];
        }
    }

    /**
     * Utility method to return the sin of an angle
     *
     * @param angle The angle to compute the sin of
     * @return int The sin of the angle
     */
    static int sin(int angle) {
        return cos(angle - 90);
    }

    // private

    /**
     * A utility method to do all the necessary height calculations
     * when content was changed.
     */
    private void contentChanged() {
        int deltaHeight = ((StringLayout)layouts[1]).setString(
                                  toString(mode == TIME ? TIME : DATE));
        if (mode == DATE_TIME) {
            deltaHeight += ((StringLayout)layouts[2]).setString(
                                  mode == DATE_TIME ? toString(TIME) : null);
        }
        height += deltaHeight;
        contentChanged(0, 0, deltaHeight);
    }

    /**
     * A utility method to return a numerical digit as two digits
     * if it is less than 10
     *
     * @param n The number to convert
     * @return String The String representing the number in two digits
     */
    private static String twoDigits(int n) {
        if (n == 0) {
            return "00";
        } else if (n < 10) {
            return "0" + n;
        } else {
            return "" + n;
        }
    }

    /**
     * The highlight of this DateField
     */
    private int highlight;
    /**
     * A flag indicating the initialization state of this DateField
     */
    private boolean initialized; // = false;
    /**
     * The mode of this DateField
     */
    private int mode;
    /**
     * The editor for this DateField
     */
    private EditScreen editor = null;
    /**
     * The height of this DateField
     */
    private int height;         //  = 0;
    /**
     * The last saved date.
     * This is used for making the last saved date bold.
     */
    private Calendar currentDate;
    /**
     * The image representing an up arrow
     */
    private static final Image ARROW_UP;
    /**
     * The image representing an down arrow
     */
    private static final Image ARROW_DOWN;
    /**
     * The image representing an left arrow
     */
    private static final Image ARROW_LEFT;
    /**
     * The image representing an right arrow
     */
    private static final Image ARROW_RIGHT;

    static {
        /**
         * The byte array to use for the up arrow
         */
        byte up_bytes[] = {
            (byte)0x89, (byte)0x50, (byte)0x4e, (byte)0x47,
            (byte)0x0d, (byte)0x0a, (byte)0x1a, (byte)0x0a,
            (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x0d,
            (byte)0x49, (byte)0x48, (byte)0x44, (byte)0x52,
            (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x07,
            (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x04,
            (byte)0x08, (byte)0x02, (byte)0x00, (byte)0x00,
            (byte)0x00, (byte)0xcd, (byte)0xa4, (byte)0xb2,
            (byte)0x2a, (byte)0x00, (byte)0x00, (byte)0x00,
            (byte)0x04, (byte)0x67, (byte)0x41, (byte)0x4d,
            (byte)0x41, (byte)0x00, (byte)0x00, (byte)0x58,
            (byte)0xc7, (byte)0xfc, (byte)0x47, (byte)0xe0,
            (byte)0x02, (byte)0x00, (byte)0x00, (byte)0x00,
            (byte)0x20, (byte)0x63, (byte)0x48, (byte)0x52,
            (byte)0x4d, (byte)0x00, (byte)0x00, (byte)0x7a,
            (byte)0x25, (byte)0x00, (byte)0x00, (byte)0x80,
            (byte)0x83, (byte)0x00, (byte)0x00, (byte)0xf9,
            (byte)0xff, (byte)0x00, (byte)0x00, (byte)0x80,
            (byte)0xe8, (byte)0x00, (byte)0x00, (byte)0x75,
            (byte)0x30, (byte)0x00, (byte)0x00, (byte)0xea,
            (byte)0x60, (byte)0x00, (byte)0x00, (byte)0x3a,
            (byte)0x97, (byte)0x00, (byte)0x00, (byte)0x17,
            (byte)0x6f, (byte)0x97, (byte)0xa9, (byte)0x99,
            (byte)0xd4, (byte)0x00, (byte)0x00, (byte)0x00,
            (byte)0x28, (byte)0x49, (byte)0x44, (byte)0x41,
            (byte)0x54, (byte)0x78, (byte)0x9c, (byte)0x62,
            (byte)0xf8, (byte)0x0f, (byte)0x03, (byte)0x0c,
            (byte)0x0c, (byte)0x0c, (byte)0x70, (byte)0x36,
            (byte)0x40, (byte)0x00, (byte)0x31, (byte)0xc0,
            (byte)0x85, (byte)0x20, (byte)0x00, (byte)0xc2,
            (byte)0x05, (byte)0x08, (byte)0x20, (byte)0x06,
            (byte)0x64, (byte)0x21, (byte)0xb8, (byte)0x04,
            (byte)0x40, (byte)0x00, (byte)0x31, (byte)0x60,
            (byte)0x05, (byte)0x00, (byte)0x01, (byte)0x06,
            (byte)0x00, (byte)0xbc, (byte)0x0c, (byte)0x23,
            (byte)0xdd, (byte)0x02, (byte)0x54, (byte)0x80,
            (byte)0xac, (byte)0x00, (byte)0x00, (byte)0x00,
            (byte)0x00, (byte)0x49, (byte)0x45, (byte)0x4e,
            (byte)0x44, (byte)0xae, (byte)0x42, (byte)0x60,
            (byte)0x82,
        };

        /**
         * The byte array to use for the down arrow
         */
        byte down_bytes[] = {
            (byte)0x89, (byte)0x50, (byte)0x4e, (byte)0x47,
            (byte)0x0d, (byte)0x0a, (byte)0x1a, (byte)0x0a,
            (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x0d,
            (byte)0x49, (byte)0x48, (byte)0x44, (byte)0x52,
            (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x07,
            (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x04,
            (byte)0x08, (byte)0x02, (byte)0x00, (byte)0x00,
            (byte)0x00, (byte)0xcd, (byte)0xa4, (byte)0xb2,
            (byte)0x2a, (byte)0x00, (byte)0x00, (byte)0x00,
            (byte)0x04, (byte)0x67, (byte)0x41, (byte)0x4d,
            (byte)0x41, (byte)0x00, (byte)0x00, (byte)0x58,
            (byte)0xc7, (byte)0xfc, (byte)0x47, (byte)0xe0,
            (byte)0x02, (byte)0x00, (byte)0x00, (byte)0x00,
            (byte)0x20, (byte)0x63, (byte)0x48, (byte)0x52,
            (byte)0x4d, (byte)0x00, (byte)0x00, (byte)0x7a,
            (byte)0x25, (byte)0x00, (byte)0x00, (byte)0x80,
            (byte)0x83, (byte)0x00, (byte)0x00, (byte)0xf9,
            (byte)0xff, (byte)0x00, (byte)0x00, (byte)0x80,
            (byte)0xe8, (byte)0x00, (byte)0x00, (byte)0x75,
            (byte)0x30, (byte)0x00, (byte)0x00, (byte)0xea,
            (byte)0x60, (byte)0x00, (byte)0x00, (byte)0x3a,
            (byte)0x97, (byte)0x00, (byte)0x00, (byte)0x17,
            (byte)0x6f, (byte)0x97, (byte)0xa9, (byte)0x99,
            (byte)0xd4, (byte)0x00, (byte)0x00, (byte)0x00,
            (byte)0x26, (byte)0x49, (byte)0x44, (byte)0x41,
            (byte)0x54, (byte)0x78, (byte)0x9c, (byte)0x62,
            (byte)0x60, (byte)0xc0, (byte)0x06, (byte)0x00,
            (byte)0x02, (byte)0x88, (byte)0xe1, (byte)0xff,
            (byte)0xff, (byte)0xff, (byte)0x98, (byte)0x22,
            (byte)0x00, (byte)0x01, (byte)0x04, (byte)0xc2,
            (byte)0xc8, (byte)0x12, (byte)0x10, (byte)0x2e,
            (byte)0x40, (byte)0x00, (byte)0x41, (byte)0x29,
            (byte)0x88, (byte)0x04, (byte)0x9c, (byte)0x0d,
            (byte)0x10, (byte)0x60, (byte)0x00, (byte)0x98,
            (byte)0xe9, (byte)0x23, (byte)0xdd, (byte)0x30,
            (byte)0xaf, (byte)0x47, (byte)0x9e, (byte)0x00,
            (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x49,
            (byte)0x45, (byte)0x4e, (byte)0x44, (byte)0xae,
            (byte)0x42, (byte)0x60, (byte)0x82,
        };

        /**
         * The byte array to use for the left arrow
         */
        byte left_bytes[] = {
            (byte)0x89, (byte)0x50, (byte)0x4e, (byte)0x47,
            (byte)0x0d, (byte)0x0a, (byte)0x1a, (byte)0x0a,
            (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x0d,
            (byte)0x49, (byte)0x48, (byte)0x44, (byte)0x52,
            (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x04,
            (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x07,
            (byte)0x08, (byte)0x02, (byte)0x00, (byte)0x00,
            (byte)0x00, (byte)0xa0, (byte)0x07, (byte)0x7b,
            (byte)0x87, (byte)0x00, (byte)0x00, (byte)0x00,
            (byte)0x04, (byte)0x67, (byte)0x41, (byte)0x4d,
            (byte)0x41, (byte)0x00, (byte)0x00, (byte)0x58,
            (byte)0xc7, (byte)0xfc, (byte)0x47, (byte)0xe0,
            (byte)0x02, (byte)0x00, (byte)0x00, (byte)0x00,
            (byte)0x20, (byte)0x63, (byte)0x48, (byte)0x52,
            (byte)0x4d, (byte)0x00, (byte)0x00, (byte)0x7a,
            (byte)0x25, (byte)0x00, (byte)0x00, (byte)0x80,
            (byte)0x83, (byte)0x00, (byte)0x00, (byte)0xf9,
            (byte)0xff, (byte)0x00, (byte)0x00, (byte)0x80,
            (byte)0xe8, (byte)0x00, (byte)0x00, (byte)0x75,
            (byte)0x30, (byte)0x00, (byte)0x00, (byte)0xea,
            (byte)0x60, (byte)0x00, (byte)0x00, (byte)0x3a,
            (byte)0x97, (byte)0x00, (byte)0x00, (byte)0x17,
            (byte)0x6f, (byte)0x97, (byte)0xa9, (byte)0x99,
            (byte)0xd4, (byte)0x00, (byte)0x00, (byte)0x00,
            (byte)0x32, (byte)0x49, (byte)0x44, (byte)0x41,
            (byte)0x54, (byte)0x78, (byte)0x9c, (byte)0x62,
            (byte)0xf8, (byte)0x0f, (byte)0x03, (byte)0x0c,
            (byte)0x0c, (byte)0x0c, (byte)0x00, (byte)0x01,
            (byte)0xc4, (byte)0x80, (byte)0x60, (byte)0x31,
            (byte)0x30, (byte)0x00, (byte)0x04, (byte)0x10,
            (byte)0x03, (byte)0x82, (byte)0xc5, (byte)0xc0,
            (byte)0x00, (byte)0x10, (byte)0x40, (byte)0x28,
            (byte)0x08, (byte)0x20, (byte)0x80, (byte)0x50,
            (byte)0x64, (byte)0x00, (byte)0x02, (byte)0x08,
            (byte)0x45, (byte)0x0f, (byte)0x40, (byte)0x00,
            (byte)0x31, (byte)0x20, (byte)0x9b, (byte)0x06,
            (byte)0x10, (byte)0x60, (byte)0x00, (byte)0xb9,
            (byte)0xf5, (byte)0x23, (byte)0xdd, (byte)0xb2,
            (byte)0x55, (byte)0x3d, (byte)0xf7, (byte)0x00,
            (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x49,
            (byte)0x45, (byte)0x4e, (byte)0x44, (byte)0xae,
            (byte)0x42, (byte)0x60, (byte)0x82,
        };

        /**
         * The byte array to use for the right arrow
         */
        byte right_bytes[] = {
            (byte)0x89, (byte)0x50, (byte)0x4e, (byte)0x47,
            (byte)0x0d, (byte)0x0a, (byte)0x1a, (byte)0x0a,
            (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x0d,
            (byte)0x49, (byte)0x48, (byte)0x44, (byte)0x52,
            (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x04,
            (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x07,
            (byte)0x08, (byte)0x02, (byte)0x00, (byte)0x00,
            (byte)0x00, (byte)0xa0, (byte)0x07, (byte)0x7b,
            (byte)0x87, (byte)0x00, (byte)0x00, (byte)0x00,
            (byte)0x04, (byte)0x67, (byte)0x41, (byte)0x4d,
            (byte)0x41, (byte)0x00, (byte)0x00, (byte)0x58,
            (byte)0xc7, (byte)0xfc, (byte)0x47, (byte)0xe0,
            (byte)0x02, (byte)0x00, (byte)0x00, (byte)0x00,
            (byte)0x20, (byte)0x63, (byte)0x48, (byte)0x52,
            (byte)0x4d, (byte)0x00, (byte)0x00, (byte)0x7a,
            (byte)0x25, (byte)0x00, (byte)0x00, (byte)0x80,
            (byte)0x83, (byte)0x00, (byte)0x00, (byte)0xf9,
            (byte)0xff, (byte)0x00, (byte)0x00, (byte)0x80,
            (byte)0xe8, (byte)0x00, (byte)0x00, (byte)0x75,
            (byte)0x30, (byte)0x00, (byte)0x00, (byte)0xea,
            (byte)0x60, (byte)0x00, (byte)0x00, (byte)0x3a,
            (byte)0x97, (byte)0x00, (byte)0x00, (byte)0x17,
            (byte)0x6f, (byte)0x97, (byte)0xa9, (byte)0x99,
            (byte)0xd4, (byte)0x00, (byte)0x00, (byte)0x00,
            (byte)0x32, (byte)0x49, (byte)0x44, (byte)0x41,
            (byte)0x54, (byte)0x78, (byte)0x9c, (byte)0x62,
            (byte)0x60, (byte)0x60, (byte)0x60, (byte)0xf8,
            (byte)0x0f, (byte)0x03, (byte)0x00, (byte)0x01,
            (byte)0xc4, (byte)0x00, (byte)0x01, (byte)0x10,
            (byte)0x0e, (byte)0x40, (byte)0x00, (byte)0x31,
            (byte)0xc0, (byte)0x01, (byte)0x90, (byte)0x03,
            (byte)0x10, (byte)0x40, (byte)0x0c, (byte)0xc8,
            (byte)0x00, (byte)0x20, (byte)0x80, (byte)0x50,
            (byte)0x64, (byte)0x00, (byte)0x02, (byte)0x08,
            (byte)0x45, (byte)0x0f, (byte)0x40, (byte)0x00,
            (byte)0x21, (byte)0x58, (byte)0x40, (byte)0x00,
            (byte)0x10, (byte)0x60, (byte)0x00, (byte)0x06,
            (byte)0xa9, (byte)0x23, (byte)0xdd, (byte)0xcf,
            (byte)0x35, (byte)0x80, (byte)0x1e, (byte)0x00,
            (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x49,
            (byte)0x45, (byte)0x4e, (byte)0x44, (byte)0xae,
            (byte)0x42, (byte)0x60, (byte)0x82,
        };

        ARROW_UP    = Image.createImage(up_bytes, 0, up_bytes.length);
        ARROW_DOWN  = Image.createImage(down_bytes, 0, down_bytes.length);
        ARROW_LEFT  = Image.createImage(left_bytes, 0, left_bytes.length);
        ARROW_RIGHT = Image.createImage(right_bytes, 0, right_bytes.length);
    }

    /**
     * The EditScreen class is a special editor for a DateField.
     * It can edit both date and time.
     */
    class EditScreen extends Screen implements CommandListener {
        /**
         * Flag to signal whether a select action transfers focus or not
         * (false by default)
         */
        private static final boolean SELECT_TRANSFERS_FOCUS = false;

        /**
         * Special command to go "back" from the editor to the DateField
         */
        Command Back = new Command(
            Resource.getString("Back"), Command.BACK, 0);
        /**
         * Special command to "ok" the changes done in the editor
         */
        Command OK   = new Command
            (Resource.getString("Save"), Command.OK, 1);

        /**
         * Flag to format am/pm string
         */
        private boolean ampmAfterTime = Resource.isAMPMafterTime();
        /**
         * The width and height of the editor
         */
        private int width, height;

        /**
         * Create a new EditScreen
         *
         * @param returnScreen The Screen to return to from the editor
         * @param field The DateField this EditScreen is editing
         */
        EditScreen(Screen returnScreen, DateField field) {
            super(field.getLabel());

            this.returnScreen = returnScreen;
            this.field = field;
            this.calendar = Calendar.getInstance();

            addCommand(OK);
            addCommand(Back);
            setCommandListener(this);
        }

        /**
         * Handle a command action
         *
         * @param cmd The Command to handle
         * @param s   The Displayable with the Command
         */
        public void commandAction(Command cmd, Displayable s) {
            Form form = null;
            Item item = null;
            ItemStateListener isl = null;

            synchronized (Display.LCDUILock) {
                if (cmd == OK) {
                    field.saveDate(calendar.getTime());
                    item = field;
                    form = (Form)item.getOwner();
                    isl = form.getItemStateListener();
                }
                currentDisplay.setCurrent(returnScreen);

            } // synchronized

            // SYNC NOTE: Move the call to the application's
            // ItemStateChangedListener outside the lock
            if (form != null) {
                form.itemStateChanged(item, isl);
            }
        }

        /**
         * Set the current date and time
         *
         * @param currentValue The Date to set the editor to
         * @param mode The operating mode of the DateField
         */
        void setDateTime(Date currentValue, int mode) {
            calendar.setTime(currentValue);

            this.mode = mode;
            if (CLOCK_USES_AM_PM && !ampmAfterTime) {
                timeSel = Calendar.AM_PM;
            } else {
                timeSel = Calendar.HOUR;
            }
            // highlight = calendar.get(Calendar.DATE);
            highlight = -1; // highlight the year to start with

            repaint();
        }

        /**
         * Paint the content of this editor
         *
         * @param g The Graphics object to paint to
         */
        void paintContent(Graphics g) {
            if (mode == TIME) {
                paintClock(g);
            } else {
                paintCalendar(g);
            }
        }

        /**
         * Layout the content of this editor given the width/height
         *
         * @param w The allowable width of this editor
         * @param h The allowable height of this editor
         * @return int The height needed to fully display this editor
         *              given the allowable width
         */
        int layoutContent(int w, int h) {
            this.width = w;
            this.height = h;

            if (mode == TIME) {
                return (50 + large.getHeight());
            }
            return (regular.getHeight() - 1 +
                    7 * regular.getBaselinePosition() + 7);
        }

        /**
         * Initialize the highlight of this editor
         *
         * @param vpY
         * @param vpH
         * @return int Always returns 0
         */
        int initHilight(int vpY, int vpH) {
            if (mode == DATE) {
                int hilightBottom = highlightY(true);
                if (hilightBottom > vpH) {
                    return (hilightBottom - vpH);
                }
            }
            return 0;
        }

        /**
         * Traverse this editor
         *
         * @param action
         * @param top
         * @param bottom
         * @return int
         */
        int traverse(int action, int top, int bottom) {
            if (mode == DATE) {
                return traverseDate(action, top, bottom);
            }

            return traverseClock(action, top, bottom);
        }

        /**
         * Handle a keypress
         *
         * @param keyCode The integer code of the key that was pressed
         */
        void keyPressed(int keyCode) {
            synchronized (Display.LCDUILock) {
                super.keyPressedImpl(keyCode);
                if (Display.getGameAction(keyCode) == Canvas.FIRE) {
                    selectFired();
                }
            }
        }

        /**
         * Handle a selection
         */
        void selectFired() {
            if (!SELECT_TRANSFERS_FOCUS) {
                return;
            }

            if (mode == DATE) {
                if (highlight > 0) {
                    highlight = -1;
                } else if (highlight == -1) {
                    highlight = 0;
                } else {
                    highlight = calendar.get(Calendar.DATE);
                }
                if (highlight > 0) {
                    calendar.set(Calendar.DATE, highlight);
                }
            } else {
                if (timeSel == Calendar.MINUTE) {
                    timeSel = Calendar.HOUR;
                } else {
                    timeSel = Calendar.MINUTE;
                }
            }

            repaint();
        }

        /**
         * Paint the clock
         *
         * @param g The Graphics to paint to
         */
        void paintClock(Graphics g) {
            int hour   = calendar.get(Calendar.HOUR) % 12;
            int minute = calendar.get(Calendar.MINUTE);

            g.setColor(Display.ERASE_COLOR);
            g.fillRect(0, 0, width, height);

            int digits_height = large.getHeight() 
                              + ARROW_UP.getHeight()
                              + ARROW_DOWN.getHeight()
                              + 2;

            int clockSize = height - digits_height;
            if (width < clockSize) {
                clockSize = width;
            }
            if (60 < clockSize) {
                clockSize = 60;
            }

            // For reference, the above if statements are replacing
            // this Math() call below.
            // Math.min(60, Math.min(width, height - digits_height));

            g.translate((width - clockSize) / 2, 
                        (height - (clockSize + digits_height)) / 2);

            g.setColor(Display.FG_COLOR);

            g.drawRoundRect(0, 0, clockSize, clockSize, 
                            clockSize / 2, clockSize / 2);
            g.drawLine(clockSize / 2, 0, clockSize / 2, 5);
            g.drawLine(clockSize / 2, clockSize, 
                       clockSize / 2, clockSize - 5);
            g.drawLine(0, clockSize / 2, 5, clockSize / 2);
            g.drawLine(clockSize, clockSize / 2, 
                       clockSize - 5, clockSize / 2);

            int minuteAngle = 90 - (minute * 6);
            int hourAngle   = 90 - (hour * 30 + (minute / 2));

            g.translate(clockSize / 2, clockSize / 2);
            g.drawLine(0, 0,
                        (cos(hourAngle)*clockSize / 4) >> 16,
                       -(sin(hourAngle)*clockSize / 4) >> 16);

            g.drawLine(0, 0,
                        (cos(minuteAngle)*(clockSize / 2 - 10)) >> 16,
                       -(sin(minuteAngle)*(clockSize / 2 - 10)) >> 16);

            g.translate(0, clockSize / 2 + 2 + ARROW_UP.getHeight());

            g.setFont(large);

            if (CLOCK_USES_AM_PM) {
                String ampm = Resource.getString(ampmString(calendar));

                if (hour == 0) {
                    hour = 12;
                }

                String timeString;

                if (ampmAfterTime) {
                    timeString = 
                        twoDigits(hour) + ":" + twoDigits(minute) + " " + ampm;
                } else {
                    timeString = 
                        ampm + " " + twoDigits(hour) + ":" + twoDigits(minute);
                }

                g.translate(-large.stringWidth(timeString) / 2, 0);

                g.drawString(timeString,
                             0, 0, Graphics.LEFT | Graphics.TOP);

                int dX;
                int w;
                int h = large.getBaselinePosition() + 1;
                int offset;
                int len;

                if (ampmAfterTime) {
                    if (timeSel == Calendar.HOUR) {
                        offset = 0;
                        len = 2;
                    } else if (timeSel == Calendar.MINUTE) {
                        offset = 3;
                        len = 2;
                    } else {
                        offset = 6;
                        len = timeString.length() - offset;
                    }
                } else {
                    offset = ampm.length();
                    if (timeSel == Calendar.HOUR) {
                        offset += 1;
                        len = 2;
                    } else if (timeSel == Calendar.MINUTE) {
                        offset += 4;
                        len = 2;
                    } else {
                        len = offset;
                        offset = 0;
                    }
                }

                dX = large.substringWidth(timeString, 0, offset);
                w = large.substringWidth(timeString, offset, len);

                g.fillRect(dX, 1, w, h -1);

                g.setColor(Display.FG_H_COLOR);
                g.drawSubstring(timeString, offset, len,
                                dX, 0, Graphics.LEFT | Graphics.TOP);

                if (ampmAfterTime) {
                    if (timeSel != Calendar.HOUR) {
                        g.drawImage(ARROW_LEFT, 
                                    -1, h / 2 + 2,
                                    Graphics.RIGHT | Graphics.VCENTER);
                    }
                    if (timeSel != Calendar.AM_PM) {
                        g.drawImage(ARROW_RIGHT,
                                    large.stringWidth(timeString) + 1,
                                    h / 2 + 2,
                                    Graphics.LEFT | Graphics.VCENTER);
                    }
                } else {
                    if (timeSel != Calendar.AM_PM) {
                        g.drawImage(ARROW_LEFT,
                                    -1, h / 2 + 2,
                                    Graphics.RIGHT | Graphics.VCENTER);
                    }
                    if (timeSel != Calendar.MINUTE) {
                        g.drawImage(ARROW_RIGHT,
                                    large.stringWidth(timeString) + 1,
                                    h / 2 + 2,
                                    Graphics.LEFT | Graphics.VCENTER);
                    }
                }

                g.drawImage(ARROW_UP,
                            dX + w / 2, 0, Graphics.HCENTER | Graphics.BOTTOM);
                g.drawImage(ARROW_DOWN,
                            dX + w / 2, h + 1,
                            Graphics.HCENTER | Graphics.TOP);
            } else {
                g.drawString(":", 0, 0, Graphics.LEFT | Graphics.TOP);

                hour = calendar.get(Calendar.HOUR_OF_DAY);

                String str = hour + "";
                int w = large.stringWidth(str);
                int h = large.getBaselinePosition() + 1;

                g.translate(-1, 0);
                if (timeSel == Calendar.HOUR) {
                    g.setColor(Display.BG_H_COLOR);
                    g.fillRect(-w, 1, w + 1, h - 1);

                    g.setColor(Display.FG_H_COLOR);
                    g.drawImage(ARROW_UP, -w / 2, 0, 
                                Graphics.HCENTER | Graphics.BOTTOM);
                    g.drawImage(ARROW_DOWN,
                                -w / 2, h + 1,
                                Graphics.HCENTER | Graphics.TOP);
                } else {
                    g.setColor(Display.FG_COLOR);
                    g.drawImage(ARROW_LEFT, -(w + 1), h / 2, 
                                Graphics.RIGHT | Graphics.VCENTER);
                }

                g.drawString(str, 0, 0,
                             Graphics.RIGHT | Graphics.TOP);

                str = twoDigits(minute);
                w   = large.stringWidth(str);

                g.translate(1 + large.charWidth(':'), 0);

                if (timeSel == Calendar.MINUTE) {
                    g.setColor(Display.BG_H_COLOR);
                    g.fillRect(0, 1, w + 1, h - 1);

                    g.setColor(Display.FG_H_COLOR);
                    g.drawImage(ARROW_UP, w / 2, 0, 
                                Graphics.HCENTER | Graphics.BOTTOM);
                    g.drawImage(ARROW_DOWN, w / 2,  h + 1,
                                Graphics.HCENTER | Graphics.TOP);
                } else {
                    g.setColor(Display.FG_COLOR);
                    g.drawImage(ARROW_RIGHT, w + 2, h / 2, 
                                Graphics.LEFT | Graphics.VCENTER);

                }
                g.drawString(twoDigits(minute), 0, 0,
                             Graphics.LEFT | Graphics.TOP);
            }
        }

        /**
         * The selected time element
         */
        int timeSel = 0;

        /**
         * Traverse the clock
         *
         * @param action
         * @param top
         * @param bottom
         * @return int
         */
        int traverseClock(int action, int top, int bottom) {
            int hrInc = 1;

            switch (action) {
                case Canvas.LEFT:
                    if (timeSel == Calendar.MINUTE) {
                        timeSel = Calendar.HOUR;
                        return 0;
                    } else if (CLOCK_USES_AM_PM) {
                        if (timeSel == Calendar.AM_PM) {
                            timeSel = Calendar.MINUTE;
                        } else if (timeSel == Calendar.HOUR) {
                            timeSel = Calendar.AM_PM;
                        }
                        return 0;
                    }
                            return -1;

                case Canvas.RIGHT:
                    if (timeSel == Calendar.HOUR) {
                        timeSel = Calendar.MINUTE;
                        return 0;
                    } else if (CLOCK_USES_AM_PM) {
                        if (timeSel == Calendar.MINUTE) {
                            timeSel = Calendar.AM_PM;
                        } else if (timeSel == Calendar.AM_PM) {
                            timeSel = Calendar.HOUR;
                        }
                        return 0;
                    }
                    return -1;

                case Canvas.UP:
                    if (CLOCK_USES_AM_PM && (timeSel == Calendar.AM_PM)) {
                        hrInc = 12;
                    } else if (timeSel == Calendar.MINUTE) {
                        int m = calendar.get(Calendar.MINUTE);
                        if (m == 59) {
                            calendar.set(Calendar.MINUTE, 0);
                        } else {
                            calendar.set(Calendar.MINUTE, m + 1);
                            hrInc = 0;
                        }
                    }
                    calendar.set(Calendar.HOUR_OF_DAY,
                        (hrInc + calendar.get(Calendar.HOUR_OF_DAY)) % 24);
                    return 0;

                case Canvas.DOWN:
                    if (CLOCK_USES_AM_PM && (timeSel == Calendar.AM_PM)) {
                        hrInc = 12;
                    } else if (timeSel == Calendar.MINUTE) {
                        int m = calendar.get(Calendar.MINUTE);
                        if (m == 0) {
                            calendar.set(Calendar.MINUTE, 59);
                        } else {
                            calendar.set(Calendar.MINUTE, m - 1);
                            hrInc = 0;
                        }
                    }
                    hrInc = 24 - hrInc;
                    calendar.set(Calendar.HOUR_OF_DAY,
                        (hrInc + calendar.get(Calendar.HOUR_OF_DAY)) % 24);
                    return 0;
                }
            return -1;
        } // traverseClock()

        /**
         * Traverse the Date editor
         *
         * @param action
         * @param top
         * @param bottom
         * @return int
         */
        int traverseDate(int action, int top, int bottom) {
            int scrollHeight = 0;

            switch (action) {
                case Canvas.LEFT:
                    if (highlight == 1) {
                        return -1;
                    }
                    if (highlight > 1) {
                        highlight--;
                        if (calendar.get(Calendar.DAY_OF_WEEK) ==
                                Calendar.SUNDAY) {
                            int topOfHighlight = highlightY(false);
                            if (topOfHighlight < top) {
                                scrollHeight = top - topOfHighlight;
                            }
                        }
                        break;
                    }
                    int year  = calendar.get(Calendar.YEAR);
                    int month = calendar.get(Calendar.MONTH);
                    if (highlight == 0) {
                        if (month > Calendar.JANUARY) {
                            int day = calendar.get(Calendar.DATE);
                            lastDay = daysInMonth(month - 1, year);
                            if (day > lastDay) {
                                calendar.set(Calendar.DATE, lastDay);
                            }
                            calendar.set(Calendar.MONTH, month - 1);
                        } else {
                            lastDay = 31;
                            calendar.set(Calendar.MONTH, Calendar.DECEMBER);
                            calendar.set(Calendar.YEAR,  year - 1);
                        }
                    } else if (highlight == -1) {
                        calendar.set(Calendar.YEAR, year - 1);
                        lastDay = daysInMonth(month, year);
                    }
                    setDayOffset();
                    break;

                case Canvas.RIGHT:
                    if (highlight == lastDay) {
                        return -1;
                    }
                    if ((highlight > 0) && (highlight < lastDay)) {
                        highlight++;
                        if (calendar.get(Calendar.DAY_OF_WEEK) ==
                                Calendar.SATURDAY) {
                            int bottomOfHighlight = highlightY(true);
                            if (bottomOfHighlight > bottom) {
                                scrollHeight = bottomOfHighlight - bottom;
                            }
                        }
                        break;
                    }
                    year  = calendar.get(Calendar.YEAR);
                    month = calendar.get(Calendar.MONTH);
                    if (highlight == 0) {
                        if (month < Calendar.DECEMBER) {
                            int day = calendar.get(Calendar.DATE);
                            lastDay = daysInMonth(month + 1, year);
                            if (day > lastDay) {
                                calendar.set(Calendar.DATE, lastDay);
                            }
                            calendar.set(Calendar.MONTH, month + 1);
                        } else {
                            calendar.set(Calendar.MONTH, Calendar.JANUARY);
                            calendar.set(Calendar.YEAR, year + 1);
                        }
                    } else if (highlight == -1) {
                        calendar.set(Calendar.YEAR, year + 1);
                        lastDay = daysInMonth(month, year);
                    }
                    setDayOffset();
                    break;

                case Canvas.UP:
                    if (highlight == -1) {
                        return -1;
                    }
                    if (highlight == 0) {
                        highlight = -1;
                    } else if (highlight <= 7) {
                        highlight = 0;
                    } else {
                        highlight -= 7;
                    }
                    int topOfHighlight = highlightY(false);
                    if (topOfHighlight < top) {
                        scrollHeight = top - topOfHighlight;
                    }
                    break;

                case Canvas.DOWN:
                    if (highlight == lastDay) {
                        return -1;
                    }
                    if (highlight == -1) {
                        highlight = 0;
                    } else if (highlight == 0) {
                        highlight = 1;
                    } else if ((highlight + 7) < lastDay) {
                        highlight += 7;
                    } else if ((highlight + 7) > lastDay) {
                        highlight = lastDay;
                    }
                    int bottomOfHighlight = highlightY(true);
                    if (bottomOfHighlight > bottom) {
                        scrollHeight = bottomOfHighlight - bottom;
                    }
                    break;

                default:
                    return -1;
            }

            if (highlight > 0) {
                calendar.set(Calendar.DATE, highlight);
            }
                return scrollHeight;
        } // traverseDate()

        /**
         * The initial highlight
         */
        private int highlight;
        /**
         * The last day of the month
         */
        private int lastDay;
        /**
         * The day offset
         */
        private int dayOffset;

        /**
         * Utility method to calculate the number of days
         * in a month
         *
         * @param month  The month to use
         * @param year  The year the month occurs in
         * @return int  The number of days in the month
         */
        private int daysInMonth(int month, int year) {
            switch (month) {
                case Calendar.JANUARY:
                case Calendar.MARCH:
                case Calendar.MAY:
                case Calendar.JULY:
                case Calendar.AUGUST:
                case Calendar.OCTOBER:
                case Calendar.DECEMBER:
                    return 31;
                case Calendar.FEBRUARY:
                    if (((year % 400) == 0)
                        || (((year & 3) == 0) && ((year % 100) != 0))) {
                        return 29;
                    }
                    return 28;
                case Calendar.APRIL:
                case Calendar.JUNE:
                case Calendar.SEPTEMBER:
                case Calendar.NOVEMBER:
                default:
                    return 30;
            }
        }

        /**
         * The "large" Font
         */
        Font large   = Font.getFont(Font.FACE_SYSTEM,
                                    Font.STYLE_PLAIN,
                                    Font.SIZE_LARGE);

        /**
         * The "regular" font
         */
        Font regular = Font.getFont(Font.FACE_SYSTEM, 
                                    Font.STYLE_PLAIN, 
                                    Font.SIZE_SMALL);

        /**
         * The "bold" font
         */
        Font bold    = Font.getFont(Font.FACE_SYSTEM, 
                                    Font.STYLE_BOLD,
                                    Font.SIZE_SMALL);

        /**
         * Set the highlight
         *
         * @param addLineHeight
         * @return int
         */
        int highlightY(boolean addLineHeight) {
            if (highlight == -1) {
                return addLineHeight ? regular.getBaselinePosition() : 0;
            } else if (highlight == 0) {
                return regular.getBaselinePosition()
                    + (addLineHeight ? regular.getHeight() : 0);
            } else {
                int line = 1 + (highlight + dayOffset - 2) / 7;
                return (regular.getBaselinePosition() + 1) * line +
                       regular.getHeight() + 1 +
                       (addLineHeight ? regular.getBaselinePosition() +2 : 0);
            }
        }

        /**
         * Paint the Calendar
         *
         * @param g The Graphics context to paint to
         */
        void paintCalendar(Graphics g) {

            g.setColor(Display.ERASE_COLOR);
            g.fillRect(g.getClipX(), g.getClipY(), 
                       g.getClipWidth(), g.getClipHeight());

            boolean currentDateOnDisplay = false;
            int year      = calendar.get(Calendar.YEAR);
            int month     = calendar.get(Calendar.MONTH);
            int selection = calendar.get(Calendar.DATE);

            // check if currentDate is on display
            int currYear  = currentDate.get(Calendar.YEAR);
            int currMonth = currentDate.get(Calendar.MONTH);
            int currDate  = currentDate.get(Calendar.DATE);
            if ((currYear == year) && (currMonth == month)) {
                currentDateOnDisplay = true;
            }

            g.setFont(regular);
            int w = regular.stringWidth("0000");
            int h = regular.getBaselinePosition();
            g.translate(0, -1);

            if (highlight == -1) {
                g.setColor(Display.BG_H_COLOR);
                g.fillRect(((width - w) / 2) - 1, 1, w + 1, h);
                g.setColor(Display.FG_H_COLOR);
            } else {
                g.setColor(Display.FG_COLOR);
            }

            g.drawString("" + year, width / 2, 0,
                         Graphics.TOP | Graphics.HCENTER);
            g.drawImage(ARROW_LEFT, (width - w) / 2 - 2, h / 2,
                        Graphics.VCENTER | Graphics.RIGHT);
            g.drawImage(ARROW_RIGHT, (width + w) / 2 + 2, h / 2,
                        Graphics.VCENTER | Graphics.LEFT);

            g.translate(0, h + 1);
            w = regular.stringWidth(Resource.getString(MONTH_NAMES[month]));
            h = regular.getHeight();

            if (highlight == 0) {
                g.setColor(Display.BG_H_COLOR);
                g.fillRect(((width - w) / 2) - 1, 1, w + 2, h);
                g.setColor(Display.FG_H_COLOR);
            } else {
                g.setColor(Display.FG_COLOR);
            }

            g.setFont(regular);
            g.drawString(Resource.getString(MONTH_NAMES[month]), width / 2, 0,
                         Graphics.TOP | Graphics.HCENTER);
            g.drawImage(ARROW_LEFT, (width - w) / 2 - 2, h / 2,
                        Graphics.VCENTER | Graphics.RIGHT);
            g.drawImage(ARROW_RIGHT, (width + w) / 2 + 2, h / 2,
                        Graphics.VCENTER | Graphics.LEFT);

            g.translate(0, h);

            int o = width / 14;
            int rem = width % 14 / 2;
            int x = o * (dayOffset * 2 - 1) + rem;
            int y = 0;
            int lastCol = 14 * o;
            h = regular.getBaselinePosition() + 1;
            
            for (int i = 1; i <= lastDay; ++i) {
                String str = "" + i;
                if (i == highlight) {
                    w = regular.stringWidth(str);

                    g.setColor(Display.BG_H_COLOR);
                    g.fillRect(x - w / 2 - 1, y + 1, w + 1, h - 1);
                    g.setColor(Display.FG_H_COLOR);
                } else {
                    g.setColor(Display.FG_COLOR);
                }

                g.setFont(((currentDateOnDisplay) &&
                        (i == currDate) &&
                        (i != highlight)) ? bold : regular);
                g.drawString(str, x, y, Graphics.TOP | Graphics.HCENTER);
                x += 2 * o;
                if (x > lastCol) {
                    x = o + rem;
                    y += h;
                }
            }
        }

        /**
         * Set the day offset
         */
        private void setDayOffset() {
            Date save = calendar.getTime();
            calendar.set(Calendar.DATE, 1);
            dayOffset = calendar.get(Calendar.DAY_OF_WEEK);

            if (Resource.getFirstDayOfWeek() != Calendar.SUNDAY) {
                dayOffset = (dayOffset == 1) ? 7 : (dayOffset - 1);
            }
            calendar.setTime(save);
        }

        /**
         * Notify this editor it is being shown
         *
         * @param d The Display showing this editor
         */
        void showNotifyImpl(Display d) {
            super.showNotifyImpl(d);
            setDayOffset();
            lastDay = daysInMonth(calendar.get(Calendar.MONTH),
            calendar.get(Calendar.YEAR));
        }

        /**
         * Determine if this editor is the edit screen for the
         * given Displayable.
         *
         * @param d The Displayable to check
         * @return boolean True if the Displayable is equal to this editor's
         *                  return screen (ie, the screen containing the
         *                  originating field being edited)
         */
        boolean isEditScreen(Displayable d) {
          return d == returnScreen;
        }

        /**
         * The calendar holding the date/time for this editor
         */
        Calendar calendar;
        /**
         * The mode of this editor (Date, Time, Date & Time)
         */
        int mode;
        /**
         * The DateField being edited by this editor
         */
        DateField field;
        /**
         * The encapsulating Screen holding the DateField being
         * edited by this editor (allows us to return to this screen
         * when we are done editing).
         */
        Screen returnScreen;

    } // EditScreen
}
