/*
 * @(#)Form.java	1.134 01/08/29
 * Copyright (c) 1999-2001 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package javax.microedition.lcdui;

/**
 * A Form is a Screen that contains an arbitrary mixture of items: images,
 * read-only text fields, editable text fields, editable date fields, gauges,
 * and choice groups. In general, any subclass of the
 * {@link Item Item} class may be contained within a form.
 * The implementation handles layout, traversal, and scrolling.
 * None of the components contained
 * within has any internal scrolling; the entire contents
 * scrolls together. Note that this differs
 * from the behavior of other classes, the List for example,
 * where only the interior scrolls. <P>
 *
 * <p> The items contained within a Form may be edited using append, delete,
 * insert, and set methods.  Items within a Form are referred to by their
 * indexes, which are consecutive integers in the range from zero to size()-1,
 * with zero referring to the first item and size()-1 to the last item.  </p>
 *
 * <p> An item may be placed within at most one Form. If the application
 * attempts to place an item into a Form, and the item is already owned by this
 * or another Form, an IllegalStateException is thrown. The application must
 * remove the item from its currently containing Form before inserting it into
 * the new Form. </p>
 *
 * <p> As with other screens, the layout policy in most devices is
 * vertical. In forms this applies to items involving user
 * input. So, a new line is always started for focusable items like
 * TextField, DateField, Gauge or ChoiceGroup. </p>
 *
 * <P>Strings and images, which do not involve user interactions,
 * behave differently; they are
 * filled in horizontal lines, unless newline is embedded in the
 * string or layout directives of the {@link ImageItem ImageItem}
 * force a new line.
 * Contents will be wrapped (for text) or clipped (for
 * images) to fit the width of the display, and scrolling will
 * occur vertically as necessary. There will be no horizontal
 * scrolling. </P>
 *
 * <p> If the Form is visible on the display when changes to its contents are
 * requested by the application, the changes take place immediately. That is,
 * applications need not take any special action to refresh a Form's display
 * after its contents have been modified. </p>
 *
 * <p> When a Form is present on the display the user can interact
 * with it and its Items indefinitely (for instance, traversing
 * from Item to Item and possibly
 * scrolling). These traversing and scrolling operations do not cause
 * application-visible events. The system notifies
 * the application when the user modifies the state of an interactive Item 
 * contained within the Form.  This notification is accomplished by calling the
 * {@link ItemStateListener#itemStateChanged itemStateChanged()}
 * method of the listener declared to the Form with the
 * {@link #setItemStateListener setItemStateListener()} method. </p>
 *
 * <p> As with other Displayable objects, a Form can declare
 * {@link Command commands} and declare a command listener with the
 * {@link Displayable#setCommandListener setCommandListener()} method.
 * {@link CommandListener CommandListener}
 * objects are distinct from
 * {@link ItemStateListener ItemStateListener} objects, and they are declared 
 * and invoked separately. </p>
 * 
 * <P><B>Notes for application developers:</B></P>
 * <UL>
 * <LI>Although this class allows creation of arbitrary combination 
 * of components
 * the application developers should keep the small screen size in mind.
 * Form is designed to contain a <em>small number of closely related</em>
 * UI elements. </LI>
 *
 * <LI>If the number of items does not fit on the screen, the
 * implementation may choose to make it scrollable or to fold some components
 * so that a new screen is popping up when the element is edited.</LI>
 * </UL>
 *
 * @see Item
 */
public class Form extends Screen {

    /**
     * This is the rate at wich the internal array of Items grows if
     * it gets filled up
     */
    private static final int GROW_SIZE = 4;

    /** Array of Items that were added to this form. */
    private Item items[];
    /** The number of actual Items added is numOfItems. */
    private int  numOfItems; // = 0;

    /** Item currently displayed (not -1 if there is at least 1 item) */
    private int curItemIndex = -1;
    /** Y position of the currently selected item */
    private int curItemY;   // = 0; in coordinate system of Form's child

    /** itemStateListener that has to be notified of any state changes */
    private ItemStateListener itemStateListener;

    /**
     * Creates a new, empty Form.
     *
     * @param title the Form's title, or null for no title
     */
    public Form(String title) {
        this(title, null);
    }
    
    /**
     * Creates a new Form with the specified contents. This is identical to
     * creating an empty Form and then using a set of <code>append</code>
     * methods.  The 
     * items array may be null, in which case the Form is created empty.  If 
     * the items array is non-null, each element must be a valid Item not 
     * already contained within another Form.
     *
     * @param title the Form's title string
     * @param items the array of items to be placed in the Form, or null if 
     * there are no items
     * @throws IllegalStateException if one of the items is already owned by
     * another container
     * @throws NullPointerException if an element of the items array is null
     */
    public Form(String title, Item[] items) {
        super(title);

        synchronized (Display.LCDUILock) {

            if (items == null) {
                this.items = new Item[GROW_SIZE];
		// numOfItems was initialized to 0 
		// so there is no need to update it
                return;

            } else {
                this.items = new Item[items.length > GROW_SIZE ?
                                      items.length : GROW_SIZE];
            }

	    // We have to check all items first so that some
	    // items would not be added to a form that was not 
	    // instanciated
	    for (int i = 0; i < items.length; i++) {
                // NullPointerException will be thrown by
                // items[i].getOwner() if items[i] == null;
		if (items[i].getOwner() != null) {
		    throw new IllegalStateException();
		}
	    }

	    numOfItems = items.length;

            TaGLayout lastTagLayout = null;

            for (int i = 0; i < numOfItems; i++) {
                items[i].setOwner(this);

                // Group adjacent StringItems and ImageItems into TaGLayouts
                if (items[i].isGroupable()) {
                    if (lastTagLayout == null) {
                        lastTagLayout = new TaGLayout(Screen.CONTENT_FONT);
                    }
                    lastTagLayout.insert(null, items[i]);
                } else {
                    lastTagLayout = null;
                }
                this.items[i] = items[i];
            }
        } // synchronized
    }


    /**
     * Adds an Item into the Form. Strings are filled so that current line
     * is continued if possible. If the text width is greater that the remaining
     * horizontal space on the current line, the implementation inserts
     * a new line
     * and appends the rest of the text. Whenever possible the implementation
     * should avoid breaking words into two lines. Instead, occurrences of white
     * space (space or tab) should be used as potential places for splitting the
     * lines. Also, a newline character in the string causes starting of a new
     * line. <P>
     *
     * Images are laid out in the same manner as strings, unless the layout
     * directives of {@link ImageItem ImageItem} specify otherwise. Focusable
     * items (TextField, ChoiceGroup, DateField, and Gauge) are placed on their
     * own horizontal lines. <p>
     *
     * @param item the {@link Item Item} to be added.
     * @return the assigned index of the Item
     * @throws IllegalStateException if the item is already owned by
     * a container
     * @throws NullPointerException if item is null
     */
    public int append(Item item) {
        synchronized (Display.LCDUILock) {
            // NullPointerException will be thrown
            // by item.getOwner() if item == null
            if (item.getOwner() != null) {
                throw new IllegalStateException();
            }

            return insertImpl(numOfItems, item);
        }
    }

    /**
     * <p> Adds an item consisting of one String to the form. The effect visible
     * to the application is identical to </p>
     *
     * <p> <code>
     * append(new StringItem(null, str))
     * </code> </p>
     *
     * @param str the String to be added
     * @return the assigned index of the Item
     * @throws NullPointerException if str is null
     */
    public int append(String str) {
        if (str == null) {
            throw new NullPointerException();
        }

        synchronized (Display.LCDUILock) {
            return insertImpl(numOfItems, new StringItem(null, str));
        }
    }
    
    /**
     * <p> Adds an item consisting of one Image to the form. The effect visible
     * to the application is identical to </p>
     *
     * <p> <code>
     * append(new ImageItem(null, img, ImageItem.LAYOUT_DEFAULT, null))
     * </code> </p>
     *
     * @param img the image to be added
     * @return the assigned index of the Item
     * @throws IllegalArgumentException if the image is mutable
     * @throws NullPointerException if img is null
     */
    public int append(Image img) {
        if (img == null) {
            throw new NullPointerException();
        }

        synchronized (Display.LCDUILock) {
            return insertImpl(numOfItems,
                new ImageItem(null, img, ImageItem.LAYOUT_DEFAULT, null));
        }
    }

    /**
     * <p> Inserts an item into the Form just prior to the item specified.
     * The size of the Form grows by one.  The itemNum parameter must be 
     * within the range [0..size()], inclusive.
     * The index of the last item is size()-1, and 
     * so there is actually no item whose index is size(). If this value
     * is used for itemNum, the new item is inserted immediately after
     * the last item. In this case, the effect is identical to
     * {@link #append(Item) append(Item)}. </p>
     *
     * <p> The semantics are otherwise identical to
     * {@link #append(Item) append(Item)}. </p>
     *
     * @param itemNum the index where insertion is to occur
     * @param item the item to be inserted
     * @throws IndexOutOfBoundsException if itemNum is invalid
     * @throws IllegalStateException if the item is already owned by
     * a container
     * @throws NullPointerException if item is null
     */
    public void insert(int itemNum, Item item) {
        synchronized (Display.LCDUILock) {
            // NullPointerException will be thrown
            // by item.getOwner() if item == null
            if (item.getOwner() != null) {
                throw new IllegalStateException();
            }

            if (itemNum < 0 || itemNum > numOfItems) {
                throw new IndexOutOfBoundsException();
            }
            insertImpl(itemNum, item);
        }
    }

    /**
     * <p> Deletes the Item referenced by itemNum. The size of the Form
     * shrinks by one. It is legal to delete all items from a Form.
     * The itemNum parameter must be 
     * within the range [0..size()-1], inclusive. </p>
     *
     * @param itemNum the index of the item to be deleted
     * @throws IndexOutOfBoundsException if itemNum is invalid
     */
    public void delete(int itemNum) {
        synchronized (Display.LCDUILock) {
            if (itemNum < 0 || itemNum >= numOfItems) {
                throw new IndexOutOfBoundsException();
            }

            Item deletedItem = items[itemNum];
            int deletedHeight =
                initLayoutDone() ? -deletedItem.getHeight() : 0;

            if (deletedItem.isGroupable()) {
                
                // delete this item from the layout
                deletedHeight =
                    ((TaGLayout)deletedItem.layouts[0]).delete(deletedItem);

            } else if (itemNum > 0 && itemNum < numOfItems - 1) {

                // it is possible that deleted item was surrounded
                // StringItems and/or ImageItems 
                // that now have to be grouped
                if (items[itemNum-1].isGroupable() && 
                    items[itemNum+1].isGroupable()) {

                    TaGLayout tag = (TaGLayout)items[itemNum-1].layouts[0];
                    deletedHeight -= items[itemNum+1].getHeight();
                    for (int i = itemNum+1; i < numOfItems; i++) {
                        if (items[i].isGrouped(items[itemNum+1])) {
                            deletedHeight += tag.insert(null, items[i]);  
                        } else {
                            break;
                        }
                    }
                }
            }

            if (numOfItems > 1 && curItemIndex == itemNum) {
                if (curItemIndex > 0) {
                    moveCurItem(curItemIndex-1);
                } else {
                    moveCurItem(curItemIndex+1);
                }
            }

            if (curItemIndex >= itemNum) {
                curItemIndex--;
            }

            deletedItem.setOwner(null);

            numOfItems--;

            if (itemNum < numOfItems) {
                System.arraycopy(items, itemNum + 1, items, itemNum,
                                 numOfItems - itemNum);
            }

            // Delete reference to the last item 
            // that was left after array copy
            items[numOfItems] = null;

            contentChanged(null, 0, 0, deletedHeight);
        } // synchronized
    }

    /**
     * <p> Sets the item referenced by itemNum to the specified item,
     * replacing the previous item. The previous item is removed
     * from this Form.
     * The itemNum parameter must be 
     * within the range [0..size()-1], inclusive. </p>
     *
     * The end result is equal to<P>
     * <code>insert(n, item); delete(n+1);</code><br>
     * although the implementation may optimize the repainting
     * and usage of the array that stores the items. <P>
     *
     * @param itemNum the index of the item to be replaced
     * @param item the new item to be placed in the Form
     *
     * @throws IndexOutOfBoundsException if itemNum is invalid
     * @throws IllegalStateException if the item is already owned by
     * a container
     * @throws NullPointerException if item is null
     */
    public void set(int itemNum, Item item) {
        synchronized (Display.LCDUILock) {
            // NullPointerException will be thrown
            // by item.getOwner() if item == null
            if (item.getOwner() != null) {
                throw new IllegalStateException();
            }

            if (itemNum < 0 || itemNum >= numOfItems) {
                throw new IndexOutOfBoundsException();
            }

            setImpl(itemNum, item);
        }
    }
     
    /**
     * <p> Gets the item at given position.  The contents of the Form are left 
     * unchanged.
     * The itemNum parameter must be 
     * within the range [0..size()-1], inclusive. </p>
     *
     * @param itemNum the index of item
     *
     * @return the item at the given position
     *
     * @throws IndexOutOfBoundsException if itemNum is invalid
     */
    public Item get(int itemNum) {
        synchronized (Display.LCDUILock) {
            if (itemNum < 0 || itemNum >= numOfItems) {
                throw new IndexOutOfBoundsException();
            }

            return items[itemNum];
        }
    }

    /**
     * Sets the ItemStateListener for the Form, replacing any previous
     * ItemStateListener. If iListener is null, simply removes the previous
     * ItemStateListener.
     * @param iListener the new listener, or null to remove it
     */
    public void setItemStateListener(ItemStateListener iListener) {
        synchronized (Display.LCDUILock) {
            itemStateListener = iListener;
        }
    }

    /**
     * Gets the number of items in the Form.
     * @return the number of items
     */
    public int size() {
        // SYNC NOTE: return of atomic value, no locking necessary
        return numOfItems;
    }

    /**
     * Paint this Form's content
     *
     * @param g The Graphics object to paint to
     */
    void paintContent(Graphics g) {
        int clipY1 = g.getClipY();
        int clipY2 = clipY1 + g.getClipHeight();

        int translatedY = 0;
        Item prevItem = null;

        for (int i = 0; i < numOfItems && translatedY < clipY2; i++) {
            if (!items[i].isGrouped(prevItem)) {
                int h = items[i].getHeight();
                if (translatedY + h >= clipY1) {
                    items[i].paint(g);
                }

                translatedY += h;
                g.translate(0, h);
                prevItem = items[i];
            }
        }

        // Form content is shorter than the clip;
        // fill the remaining area with erase color
        // Graphics was already translated by translatedY.
        // That is why we start painting from 0.
        if (translatedY < clipY2) {
            g.setColor(Display.ERASE_COLOR);
            g.fillRect(g.getClipX(), 0, 
                       g.getClipWidth(), clipY2 - translatedY);
        }
    }

    /**
     * Layout this Form's content
     *
     * @param w The width of the Display
     * @param h The height of the Display
     * @return int The height required to fully display this Form
     */
    int layoutContent(int w, int h) {
        if (numOfItems == 0) {
            return 0;
        }

        int height = items[0].setWidth(w);
        for (int i = 1; i < numOfItems; i++) {
            if (!items[i].isGrouped(items[i-1])) {
                height += items[i].setWidth(w);
            }
        }
        return height;
    }

    /**
     * Initializes hilight when form is shown.
     * @param vpY y coordinate of viewport location
     * @param vpH viewport height
     * @return new y coordinate of viewport location
     *         child (all items) coordinate system
     */
    int initHilight(int vpY, int vpH) {
        if (numOfItems == 0) {
            return 0;
        }

        // Last time that Form was displayed an item from this form
        // had focus; To init focus to a different item we have
        // to set focus to be false on the current one.
        if (curItemIndex != -1 && items[curItemIndex].hasFocus()) {
            items[curItemIndex].setFocus(false);
        }

        curItemIndex = curItemY = 0;

        // if first item can take focus and it is shorter than the
        // viewPort height and it is not grouped with the last item
        // focus should be initialized to some other item.
        if (!items[0].takesFocus() && 
            items[0].getHeight() < vpH &&
            !items[0].isGrouped(items[numOfItems-1])) {

            for (int i = 1; i < numOfItems; i++) {
                if (!items[i].isGrouped(items[0])) {
                    curItemIndex = i;
                    curItemY = items[i-1].getHeight();
                    break;
                }
            }
        }

        if (items[curItemIndex].takesFocus()) { 
            items[curItemIndex].setFocus(true);
            items[curItemIndex].initHilight(vpY-curItemY, vpH);
        }

        return 0;
    }

    /**
     * Retrieve the ItemStateListener for this Form.
     * NOTE: calls to this method should only occur from within
     * a lock on LCDUILock.
     *
     * @return ItemStateListener The ItemStateListener of this Form,
     *                           null if there isn't one set
     */
    ItemStateListener getItemStateListener() {
        return itemStateListener;
    }

    /**
     * Insert an Item into this Form
     *
     * @param itemNum The index into the Item array to insert this Item
     * @param item The Item to insert
     * @return int The index at which the newly inserted Item can be found
     */
    private int insertImpl(int itemNum, Item item) {
        boolean forceFocus = false;

        if (items.length == numOfItems) {
           Item newItems[] = new Item[numOfItems + GROW_SIZE];
            System.arraycopy(items, 0, newItems, 0, itemNum);
            System.arraycopy(items, itemNum, newItems, itemNum + 1,
                             numOfItems - itemNum);
            items = newItems;
        } else {
            System.arraycopy(items, itemNum, items, itemNum + 1,
                             numOfItems - itemNum);
        }

        numOfItems++;

        if (initLayoutDone()) {
            if (curItemIndex == -1) {
                curItemIndex = 0;

		// This is the first item added to a form; give it focus
		forceFocus = true;
            }
            if (curItemIndex > itemNum) {
                curItemIndex++;
            }
        }

        items[itemNum] = null;
        setImpl(itemNum, item);

	if (forceFocus && items[curItemIndex].takesFocus()) {
	    items[curItemIndex].setFocus(true);
	}

        return itemNum;
    }

    /**
     * Set a specific Item index to be a new Item
     *
     * @param itemNum The Item index to change
     * @param item The new Item to set
     */
    private void setImpl(int itemNum, Item item) {
        int deltaHeight = 0;
        boolean itemHadFocus = false;
        Item oldItem = items[itemNum];

        // delete oldItem height
        if (oldItem != null) {
            if (initLayoutDone()) {
                deltaHeight = -oldItem.getHeight();
            }
            itemHadFocus = oldItem.hasFocus();
        }

        // Set owner, layout and focus

        if (item.isGroupable()) {
            if (oldItem != null && oldItem.isGroupable()) {
                TaGLayout tag = (TaGLayout)oldItem.layouts[0];
                deltaHeight = tag.set(oldItem, item);

                // try merge with a tag layout before next item
            } else if (itemNum+1 < numOfItems &&
                       items[itemNum+1].isGroupable()) {

                TaGLayout tag = (TaGLayout)items[itemNum+1].layouts[0];
                deltaHeight = tag.insert(items[itemNum+1], item);

                if (curItemIndex == itemNum+1 && tag.get(item) == 0) {
                    curItemIndex = itemNum;
                }

                // try merge with a tag layout after previous item
            } else if (itemNum > 0 && items[itemNum-1].isGroupable()) {

                TaGLayout tag = (TaGLayout)items[itemNum-1].layouts[0];
                deltaHeight = tag.insert(null, item);

                if (curItemIndex == itemNum-1 && tag.getSize() > 2)  {
                    curItemIndex = itemNum;
                }

                // item is surrounded by items that are not
                // StringItems or ImageItems                
            } else {
                
                TaGLayout tag = new TaGLayout(Screen.CONTENT_FONT);
                tag.insert(null, item);

                if (initLayoutDone()) {
                    deltaHeight = tag.setWidth(Display.WIDTH);
                }
            }
            
        } else { // not grouped

            if (oldItem != null && oldItem.isGroupable()) {
                TaGLayout tag = (TaGLayout)oldItem.layouts[0];
                deltaHeight = tag.delete(oldItem);
            }

            // possibly new item is inserted in the middle of grouped items
            if (itemNum > 0 && itemNum+1 < numOfItems &&
                items[itemNum-1].isGrouped(items[itemNum+1])) {

                TaGLayout prevTag = (TaGLayout)items[itemNum-1].layouts[0];
                TaGLayout nextTag = new TaGLayout(Screen.CONTENT_FONT);

                // move all items after the inserted item to a new group
                // from the group where the items before the inserted item were
                for (int i = itemNum+1; i < numOfItems; i++) {
                    if (items[i].layouts[0] == prevTag) {
                        deltaHeight += prevTag.delete(items[i]);
                        // note that we do not add to the delta height on
                        // the insert since we created a new layout
                        // height will be added at once when all items
                        // are moved
                        nextTag.insert(null, items[i]);
                    } else {
                        break;
                    }
                }
                if (initLayoutDone()) {
                    deltaHeight += nextTag.setWidth(Display.WIDTH);
                }
            }
            
            if (initLayoutDone()) {
                deltaHeight += item.setWidth(Display.WIDTH);
            }
        }
        
        item.setOwner(this);
        item.setFocus(itemHadFocus);
        
        items[itemNum] = item;
        
        if (oldItem != null) {
            oldItem.setOwner(null);
        }
        
        contentChanged(items[itemNum], 0, 0, deltaHeight);
    }
    
    /**
     * Get the y coordinate location of the specified item
     *
     * @param item The Item to calculate the y position for
     * @return int The y coordinate location of the Item
     */
    int getItemY(Item item) {

        if (numOfItems == 1) {
            return item == items[0] ? 0 : -1;
        }

        for (int i = 1, y = 0; i < numOfItems; i++) {
            if (!items[i].isGrouped(items[i-1])) {
                y += items[i-1].getHeight();
            }

            if (items[i] == item) {
                return y;
            }
        }
        return -1;
    }

    /**
     * Handle a key press
     *
     * @param keyCode The key that was pressed
     */
    void keyPressed(int keyCode) {
        Item curItem = null;
        ItemStateListener isl = null;

        synchronized (Display.LCDUILock) {
            if (numOfItems == 0) {
                return;
            }

            int gameKeyCode = Display.getGameAction(keyCode);
            if (gameKeyCode == Canvas.UP ||
                    gameKeyCode == Canvas.DOWN) {
                super.keyPressedImpl(keyCode);
            } else if (curItemIndex != -1) {
                curItem = items[curItemIndex];

                if ((gameKeyCode == Canvas.FIRE && curItem.select()) ||
                        (curItem.keyPressed(keyCode))) {
                    isl = itemStateListener;
                }
            }
        } // synchronized

        if (isl != null) {
            // Protect from any unexpected application exceptions
            try {
                // SYNC NOTE: We lock on calloutLock around any calls
                // into application code
                synchronized (Display.calloutLock) {
                    isl.itemStateChanged(curItem);
                }
            } catch (Throwable thr) {
                Display.handleThrowable(thr);
            }
        }
    }

    /**
     * Traverse this Form
     *
     * @param down
     * @param top
     * @param bottom
     * @param height
     * @param lineHeight
     * @param traversingIn
     * @param takesFocus
     * @return int
     */
    static int traverse(boolean down, int top, int bottom,
                        int height, int lineHeight, 
                        boolean traversingIn, boolean takesFocus) {
        
        // if going down the bottom of the unit is already visible or
        // if going up the top of the unit is already visible
        // then we can either return 0 meaning that traversal was done
        // or return -1 meaning that we should traverse to the next unit.
        if ((down && (bottom >= height))  || (!down && (top <= 0))) {
            return (traversingIn && takesFocus) ? 0 : -1;
        }

        int vpHeight = bottom - top;
        int heightToEnd = down ? height - bottom : top;

        // when traversing in show as much of the unit as you can
        // while scrolling as little as possible
        if (traversingIn) {
            int s = vpHeight;
            if (down && top < 0) {
                s = -top;
            } else if (!down && height < bottom) {
                s = bottom - height;
            }
            if (heightToEnd < s) {
                s = heightToEnd;
            }
            return s;
        }

        // if what was not displayed is shorter than the viewport height
        // than scroll just by this amount
        if (heightToEnd <= vpHeight) {
            return heightToEnd;
        }

        // other wise we will show some context 
        // (the smallest context could be half a line)
        int blockIncr = vpHeight - lineHeight/2;
        if (blockIncr < lineHeight) {
            blockIncr = lineHeight;
        } else {
            blockIncr -= (blockIncr % lineHeight);
        }

        return heightToEnd < blockIncr ? heightToEnd : blockIncr;
    }

    /**
     * Traverse this Form
     *
     * @param gameKeyCode
     * @param top
     * @param bottom
     * @return -1 if there is no need to scroll and 
     *            there is no need to repaint;
     *          0 if there is no need to scroll but 
     *            content should be repainted; or
     *          scrollHeight by which the viewPort y location to be adjusted
     */
    int traverse(int gameKeyCode, int top, int bottom) {
        if (gameKeyCode != Canvas.UP && gameKeyCode != Canvas.DOWN) {
            return -1;
        }

        int dir = gameKeyCode == Canvas.UP ? -1 : 1;
        Item curItem = items[curItemIndex];

        int itemNum = curItemIndex;
        int y = curItemY;

        for (int i = curItemIndex; i >= 0 && i < numOfItems; i += dir) {
            boolean traversingIn = i != curItemIndex;
            if (traversingIn && items[i].isGrouped(items[itemNum])) {
                continue;
            }
            
            // update y and itemNum
            if (traversingIn) {
                int itemHeight = items[i].getHeight();
                if (dir < 0) {
                    y -= itemHeight;
                } else {
                    y += items[itemNum].getHeight();
                }
                itemNum = i;
            }

            int vpY1 = top - y;
            int vpY2 = bottom - y;
            boolean allowExit = dir < 0 ? i != 0 : i != numOfItems-1;
            int scrollHeight = items[i].traverse(gameKeyCode, vpY1, vpY2, 
                                                 traversingIn, allowExit);

            if (scrollHeight >= 0) {
                // traversal was done within item
                if (traversingIn) {
                    moveCurItem(i);
                    curItemY = y;
                }
                return scrollHeight;
            }
        }
        return -1;
    }

    /**
     * Move the currently selected item to a new index
     *
     * @param newCurItem The index of the new location for the Item
     */
    private void moveCurItem(int newCurItem) {
        if (!initLayoutDone()) {
            return;
        }

        if (curItemIndex != -1 && items[curItemIndex] != null &&
                items[curItemIndex].hasFocus()) {
            items[curItemIndex].setFocus(false);
        }
        curItemIndex = newCurItem;

        if (curItemIndex != -1 && items[curItemIndex].takesFocus()) {
            items[curItemIndex].setFocus(true);
        }
    }

    /**
     * Signal this Form that its height has changed
     *
     * @param vpY
     * @param vpH
     * @param deltaHeight
     * @return int
     */
    int heightChanged(int vpY, int vpH, int deltaHeight) {

        vpY = super.heightChanged(vpY, vpH, deltaHeight);

        if (curItemIndex != -1 && numOfItems > 1) {

            // find the item that is currently at the very top
            curItemY = 0;
            for (int i = 1; i < numOfItems; i++) {
                if (!items[i].isGrouped(items[i-1])) {
                    int h = items[i-1].getHeight();
                    // The following can happen only when
                    // item displayed is taller that the 
                    // viewport area
                    if (curItemY + h > vpY + vpH) {
                        moveCurItem(i);
                        return vpY;
                    }
                    curItemY += h;
                }

                if (curItemY >= vpY) {
                    // currentItem is visible and it can take focus
                    if (items[i].takesFocus()) {
                        moveCurItem(i);
                        return vpY;
                    }
                }
            }

            // current item is not entirely visible and it is hard to
            // init focus
            return initHilight(vpY, vpH);
        }
        return vpY;
    }

    /**
     * Handle a typed key from a keyboard
     *
     * @param c The char typed from the keyboard
     */
    void keyTyped(char c) {
        synchronized (Display.LCDUILock) {
            if (curItemIndex != -1) {
                items[curItemIndex].keyTyped(c);
            }
        }
    }

    /** 
     * This method should be used by items of this form 
     * when they need to notify itemStateListener of its changed state
     *
     * @param item The Item which state was changed
     * @param isl The itemStateListener to notify of the change
     */
    static final void itemStateChanged(Item item, ItemStateListener isl) {
        if (isl == null || item == null) {
            return;
        }

        // Protect from any unexpected application exceptions
        try {
            // SYNC NOTE: We lock on calloutLock around any calls
            // into application code
            synchronized (Display.calloutLock) {
                isl.itemStateChanged(item);
            }
        } catch (Throwable thr) {
            Display.handleThrowable(thr);
        }
    }
}
