/*
 * @(#)Gauge.java	1.73 01/08/09
 * Copyright (c) 1999-2001 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package javax.microedition.lcdui;

/**
 * <P>The Gauge class implements a bar graph display of a value intended for
 * use in a form. Gauge is optionally
 * interactive. The values accepted by the object are small integers in the
 * range zero through a maximum value established by the application. The
 * application is expected to normalize its values into this range. The
 * device is expected to normalize this range into a smaller set of values for
 * display purposes. Doing so will not change the actual value contained
 * within the object. The range of values specified by the application may be
 * larger than the number of distinct visual states possible on the device, so
 * more than one value may have the same visual representation. </p>
 *
 * <P>For example, consider a Gauge object that has a range of values from zero
 * to 99, running on a device that displays the Gauge's approximate value
 * using a set of one to ten bars. The device might show one bar for values
 * zero through nine, two bars for values ten through 19, three bars for
 * values 20 through 29, and so forth. </p>
 *
 * <P>A Gauge may be interactive or non-interactive. Applications may set or
 * retrieve the Gauge's value at any time regardless of the interaction mode.
 * The implementation may change the visual appearance of the bar graph
 * depending on whether the object is created in interactive mode. </p>
 *
 * <P>In interactive mode, the user is allowed to modify the value. The user
 * will always have the means to change the value up or down by one and
 * may also have the means to change the value in greater increments.
 * The user is
 * prohibited from moving the value outside the established range. The
 * expected behavior is that the application sets the initial value and then
 * allows the user to modify the value thereafter. However, the application
 * is not prohibited from modifying the value even while the user is
 * interacting with it. </p>
 *
 * <p> In many cases the only means for the user to modify the value will be
 * to press a button to increase or decrease the value by one unit at a time.
 * Therefore, applications should specify a range of no more than a few dozen
 * values. </p>
 *
 * <P>In non-interactive mode, the user is prohibited from modifying the value.
 * An expected use of the non-interactive mode is as a "progress indicator" to
 * give the user some feedback as progress occurs during a long-running
 * operation. The application is expected to update the value periodically
 * using the setValue() method. An application using the Gauge as a progress
 * indicator should typically also attach a
 * {@link Command#STOP STOP}
 * command to the Form containing the Gauge to allow the user to halt the
 * operation in progress.
 * </p>
 */
public class Gauge extends Item {

    /** The current value of this gauge */
    private int value;

    /** The maximum possible value of this gauge */
    private int maxValue;

    /** Wether this gauge is interactive or not */
    private boolean interactive;

    /** The number of blocks making up this gauge */
    private int blockCount;

    /**
     * The buffer of horizontal space to split on
     * either side of this gauge (creating left/right margins)
     */
    private static final int BLOCK_SPACE    = 8;

    /**
     * Greatest height, in pixels, this gauge will
     * occupy on screen
     */
    private static final int GAUGE_HEIGHT   = 25;

    /**
     * The numerical value assigned to each block.
     * The number of blocks (blockCount) times the
     * value of each block (valueOfEachBlock) shall
     * equal the maximum value of this gauge (maxValue)
     */
    private int valueOfEachBlock;

    /** Gauge overall height */
    private int height;

    /**
     * <p>Creates a new Gauge object with the given label, in interactive or
     * non-interactive mode, with the given maximum and initial values. The
     * maximum value must be greater than zero, otherwise an exception is
     * thrown.  The initial value must be within the range zero to maxValue,
     * inclusive.  If the initial value is less than zero, the value is set to
     * zero.  If the initial value is greater than maxValue, it is set to
     * maxValue.</p>
     *
     * @param label the Gauge's label
     * @param interactive tells whether the user can change the value
     * @param maxValue the maximum value
     * @param initialValue the initial value in the range [0..maxValue]
     * @throws IllegalArgumentException if maxValue is invalid
     */
    public Gauge(String label, boolean interactive, int maxValue,
                 int initialValue)
    {
        // This super.constructor call guarantees that our layouts[] is
        // non-null and that index 0 contains a StringLayout
        super(label, 1);

        synchronized (Display.LCDUILock) {
            this.interactive = interactive;
            this.value = initialValue;
            setMaxValueImpl(maxValue);
        }
    }

    // public implementation

    /**
     * <p>Sets the current value of this Gauge object. If the value is less
     * than zero, zero is used. If the current value is greater than the
     * maximum value, the current value is set to be equal to the maximum
     * value. </p>
     *
     * @param value the new value
     *
     * @see #getValue
     */
    public void setValue(int value) {
        synchronized (Display.LCDUILock) {
            this.value = value;

            checkValue();
            contentChanged(0, height - (GAUGE_HEIGHT + 5), 0);
        }
    }

    /**
     * <p>Gets the current value of this Gauge object.</p>
     *
     * @return current value of the Gauge
     *
     * @see #setValue
     */
    public int getValue() {
        // SYNC NOTE: return of atomic value, no locking necessary
        return value;
    }

    /**
     * <p>Sets the maximum value of this Gauge object.  The new maximum value
     * must be greater than zero, otherwise an exception is thrown.  If the
     * current value is greater than new maximum value, the current value is
     * set to be equal to the new maximum value.</p>
     *
     * @param maxValue the new maximum value
     *
     * @throws IllegalArgumentException if maxValue is invalid
     * 
     * @see #getMaxValue
     */
    public void setMaxValue(int maxValue) {
        synchronized (Display.LCDUILock) {
            setMaxValueImpl(maxValue);
        }
    }

    /**
     * <p>Gets the maximum value of this Gauge object.</p>
     *
     * @return the maximum value of the Gauge
     *
     * @see #setMaxValue
     */
    public int getMaxValue() {
        // SYNC NOTE: return of atomic value, no locking necessary
        return maxValue;
    }

    /**
     * <p>Tells whether the user is allowed to change the value of the Gauge.
     * </p>
     *
     * @return a boolean indicating whether the Gauge is interactive
     */
    public boolean isInteractive() {
        // SYNC NOTE: return of atomic value, no locking necessary
        return interactive;
    }

    /**
     * Sets the label of the Item. If label is null, specifies that this
     * item has no label.
     * @param label the label string
     */
    public void setLabel(String label) {
        synchronized (Display.LCDUILock) {
            super.setLabel(label);

            int deltaHeight = ((StringLayout)layouts[0]).setString(label);
            height += deltaHeight;
            contentChanged(0, 0, deltaHeight);
        }
    }

    // package private implementation

    /**
     * Paint this Gauge
     *
     * @param g The Graphics to paint to
     */
    void paint(Graphics g) {

        g.setColor(Display.ERASE_COLOR);
        g.fillRect(g.getClipX(), 0, g.getClipWidth(), height);

        // We need to check the clip to determine if
        // the label needs painting
        int layoutHeight = layouts[0].getHeight();
        if (g.getClipY() < layoutHeight) {
            layouts[0].paint(g, false, false);
        }
        g.translate(0, layoutHeight);

        // Depending on the clip, we may need to repaint
        // only some or all of the blocks.
        int startBlock = 0;
        int stopBlock = blockCount;

        if (g.getClipX() > 0) {
            startBlock = (g.getClipX() - 6) / BLOCK_SPACE;
        }

        if (g.getClipWidth() == BLOCK_SPACE - 3) {
            // We just need to paint 1 bar
            stopBlock = startBlock + 1;
            // NOTE: We're a little careless here. It could
            // be possible that the clip contains several
            // bars rather than one. If this is the case,
            // we paint from the start bar to the end of
            // the gauge, rather than figure the more precise
            // end bar. Its not possible with the current
            // emulator but it may be on some platforms.
        }

        if (hasFocus()) {
            g.setColor(Display.FG_COLOR);
        } else {
            g.setColor(0x00606060); // dark gray
            // Keep in mind that on a 1-bit display, dark gray will
            // round to black
        }

        //
        // these values are used in calculating the blockHeight
        // they only need to be calculated once every paint
        //
        int bc  = blockCount - 1;
        int bc2 = bc * bc;

        for (int i = startBlock; i < stopBlock; i++) {

            int blockHeight;
            if (interactive && blockCount > 1) {

                //
                // this implements a quadratic spline from
                // p0 to p2 using p1 as a control point to
                // affect the curve
                //
                // int p0 = 5,                 // minimum height
                //     p1 = 1,                 // control point
                //     p2 = GAUGE_HEIGHT - p0; // maximum height
                // 
                // the original equation is below, but the one used
                // has been expanded to use the p values above
                //
                // blockHeight = p0 +
                //     (2 * p1 * i * (bc - i) / bc2) + 
                //     ((p2 * (i * i)) / bc2);
                //

                blockHeight = 5 + 
                              (((i * (bc - i)) << 1) / bc2) + 
                              (((GAUGE_HEIGHT - 5) * (i * i)) / bc2);
            } else {
                blockHeight = GAUGE_HEIGHT;
            }

            if ((valueOfEachBlock * (i + 1)) > value && value != maxValue)
            {
                if (hasFocus()) {
                    g.drawRect((i * BLOCK_SPACE) + 6, 
                               GAUGE_HEIGHT - blockHeight,
                               BLOCK_SPACE - 3, blockHeight);
                } else {
                    g.setStrokeStyle(Graphics.DOTTED);

                    g.drawRect((i * BLOCK_SPACE) + 6, 
                               GAUGE_HEIGHT - blockHeight,
                               BLOCK_SPACE - 3, blockHeight);
                    g.setStrokeStyle(Graphics.SOLID);
                }
            } else {
                g.fillRect((i * BLOCK_SPACE) + 6, 
                           GAUGE_HEIGHT - blockHeight,
                           BLOCK_SPACE - 3, blockHeight);
            }
        }

        g.translate(0, -layoutHeight);

        setHorizontalScroll();
    }

    /**
     * Signal this Gauge that its focus has changed
     * (either gained or lost focus)
     */
    void focusChanged() {
        if (!hasFocus()) {
            // We have lost the focus, so we need to re-set
            // the right/left arrows, as the next component
            // to receive focus may not use the left/right
            // arrows at all, and Screen no longer re-sets
            // the horizontal arrows for us
            getOwner().setHorizontalScroll(0, 100);
        }
        super.focusChanged();
    }

    /**
     * Set the width for this Guage
     *
     * @param width The width to set this Gauge to
     * @return int The height required to display this Gauge in the
     *              given width
     */
    int setWidth(int width) {
        height = GAUGE_HEIGHT + 5;
        height += layouts[0].setWidth(width);
        return height;
    }

    /**
     * Get the height of this Gauge
     *
     * @return int The height of this Gauge
     */
    int getHeight() {
        return height;
    }

    /**
     * Handle a key press
     *
     * @param keyCode The key that was pressed
     * @return boolean A flag indicating the key pressed was handled
     *                  by this Gauge
     */
    boolean keyPressed(int keyCode) {
        if (!interactive) {
            return false;
        }

        if ((Display.getGameAction(keyCode) ==  Canvas.LEFT) ||
            (Display.getGameAction(keyCode) ==  Canvas.RIGHT))
        {

            // There are two cases where we need to repaint a
            // block when a user changes a Gauge's value:
            // 1. If the user is increasing the Gauge's value
            //    and the value increases such that a new bar
            //    needs to be filled in. The easy test for this
            //    is: value % valueOfEachBlock == 0, indicating
            //    the value has increased precisely enough to
            //    fill in an entire block.
            //
            // 2. If the user is decreasing the Gauge's value
            //    and the value decreases such that a currently
            //    filled in bar needs to be un-filled. The easy
            //    test for this is:
            //    (value + 1) % valueOfEachBlock == 0, indicating
            //    we have just decreased 1 unit below a full
            //    block.

            int blockToRepaint = -1;
            int oldValue = value;

            if (Display.getGameAction(keyCode) == Canvas.LEFT) {
                if (value % valueOfEachBlock == 0) {
                    blockToRepaint = (value / valueOfEachBlock) - 1;
                } else if (value == maxValue) {
                    blockToRepaint = blockCount - 1;
                }
                value--;
            } else {
                value++;
                if (value % valueOfEachBlock == 0) {
                    blockToRepaint = (value / valueOfEachBlock) - 1;
                } else if (value == maxValue) {
                    blockToRepaint = blockCount - 1;
                }
            }
            checkValue();

            // There's no need to cause a repaint if none of the
            // blocks either need to be filled in or need to be
            // un-filled.
            if (blockToRepaint < 0 || blockToRepaint >= blockCount) {
                setHorizontalScroll();
                return value != oldValue;
            }

            // We need to remember to offset by the height of our
            // label if there is one.
            int blockY = 0;
            blockY = layouts[0].getHeight();

            // Cause a repaint of only the bar which needs updating
            repaint(blockToRepaint * BLOCK_SPACE + 6, blockY,
                    BLOCK_SPACE - 2, GAUGE_HEIGHT + 2);

            return value != oldValue;
        }
        return false;
    }

    // private implementation

    /**
     * Set the number of blocks rendered by this Gauge. This
     * will use the width of the display and calculate the
     * blockCount value.
     */
    private void setBlockCount() {
        this.blockCount = (Display.WIDTH - 8) / BLOCK_SPACE;
        if (blockCount > maxValue) {
            blockCount = maxValue;
        }
    }

    /**
     * Set the value of each block. This will use the
     * blockCount value and maxValue to assign a value
     * to each block.
     */
    private void setBlockValue() {
        valueOfEachBlock = 1;

        if (maxValue <= blockCount) {
            return;
        }

        valueOfEachBlock = maxValue / blockCount;
        if (maxValue % blockCount != 0) {
            valueOfEachBlock++;
        }

        // There's a case when our max is close to
        // a multiple of our value per block (ie 50
        // as a max value, 4 or 5 as a value per block).
        // In this case, we decrease the number of blocks
        // and make each block an even value
        // (5, using the above example).
        blockCount = maxValue / valueOfEachBlock;
        if (maxValue % valueOfEachBlock != 0) {
            blockCount++;
        }
        // If the max was a value like 52, then there
        // would be the full number of blocks (11 with
        // the RI screen width) and the last block
        // would have 2 units of value. The last block
        // in the gauge always as as many or less units
        // than the other blocks in the gauge.
    }

    /**
     * Set the horizontal "scroll" for this Gauge. Note that
     * "scroll" is a bit of a misnomer, but we are turning on
     * the left/right indicator arrows on the display to signal
     * the user the Gauge can be incremented/decremented.
     */
    private void setHorizontalScroll() {
            // If we have focus and are interactive, we check
            // to see if we need to add the right or left arrow
            if (hasFocus() && interactive) {
                // NOTE: I couldn't think offhand how to do an
                // accurate proportion value, so I just use 50
                int scrollPosition = value * 100 / maxValue;
                // Fix the rounding error
                if (value > 0 && scrollPosition == 0) {
                    scrollPosition = 1;
            }
                getOwner().setHorizontalScroll(scrollPosition, (50));
            }
    }

    /**
     * Utility method to ensure the value of the Gauge is always
     * in a range of 0 to maxValue
     *
     * @return int The corrected version of the current value
     */
    private int checkValue() {
        if (value < 0) {
            value = 0;
        } else if (value > maxValue) {
            value = maxValue;
        }
        return value;
    }

    /**
     * Set the max value of this Gauge
     *
     * @param maxValue The maximum value to set for this Gauge
     */
    private void setMaxValueImpl(int maxValue) {
        if (maxValue <= 0) {
            throw new IllegalArgumentException();
        }

        this.maxValue = maxValue;
        checkValue();

        // If the max value changes, the block count needs
        // to be re-calculated
        setBlockCount();
        setBlockValue();

        contentChanged(0, height - (GAUGE_HEIGHT + 5), 0);
    }
}

