/*
 * @(#)Graphics.java	1.49 01/08/09
 * Copyright (c) 1999-2001 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package javax.microedition.lcdui;

/**
 * Provides simple 2D geometric rendering capability.
 *
 * NOTE: all calls are unsynchronized, because multiple threads
 * should not be accessing the (lone) screen graphics context.
 * See ImageGraphics for the synchronized versions.
 *
 * Drawing primitives are provided for text, images, lines, rectangles,
 * and arcs. Rectangles and arcs may also be filled with a solid color.
 * Rectangles may also be specified with rounded corners. <p>
 *
 * The only drawing operation provided is pixel replacement. The
 * destination pixel value is simply replaced by the current
 * pixel value specified in the graphics object being used for
 * rendering. No facility for combining pixel values, such as
 * raster-ops or alpha blending, is provided. <p>
 *
 * A 24-bit color model is provided, with 8 bits for each of red, green, and
 * blue components of a color. Not all devices support a full 24 bits' worth
 * of color and thus they will map colors requested by the application into
 * colors available on the device. Facilities are provided in the {@link
 * Display Display} class for obtaining device characteristics, such as
 * whether color is available and how many distinct gray levels are available.
 * This enables applications to adapt their behavior to a device without
 * compromising device independence. <p>
 *
 * Graphics may be rendered directly to the display or to an off-screen
 * image buffer. The destination of rendered graphics depends on the
 * provenance of the graphics object. A graphics object for rendering
 * to the display is passed to the Canvas object's
 * {@link Canvas#paint(Graphics) paint()}
 * method. This is the only means by which a graphics object may
 * be obtained whose destination is the display. Furthermore, applications
 * may draw using this graphics object only for the duration of the
 * paint() method. <p>
 *
 * A graphics object for rendering to an off-screen image buffer may
 * be obtained by calling the
 * {@link Image#getGraphics() getGraphics()}
 * method on the desired image.
 * A graphics object so obtained may be held indefinitely
 * by the application, and requests may be issued on this graphics
 * object at any time. <p>
 *
 * The default coordinate system's origin is at the
 * upper left-hand corner of the destination. The X-axis direction is
 * positive towards the right, and the Y-axis direction is positive
 * downwards. Applications may assume that horizontal and vertical
 * distances in the coordinate system represent equal distances on the
 * actual device display, that is, pixels are square. A facility is provided
 * for translating the origin of the coordinate system.
 * All coordinates are specified as integers. <p>
 *
 * The coordinate system represents locations between pixels, not the
 * pixels themselves. Therefore, the first pixel in the upper left corner
 * of the display lies in the square bounded by coordinates
 * (0,0) , (1,0) , (0,1) , (1,1). <p>
 *
 * Under this definition, the semantics for fill operations are clear.
 * Since coordinate grid lines lie between pixels, fill operations
 * affect pixels that lie entirely within the region bounded by the
 * coordinates of the operation. For example, the operation
 * <pre>
 *    g.fillRect(0, 0, 3, 2)
 * </pre>
 * paints exactly six pixels.  (In this example, and in all subsequent 
 * examples, the variable <code>g</code> is assumed to contain a reference to a 
 * Graphics object.) <p>
 *
 * Each character of a font contains a set of pixels that forms the shape of
 * the character.  When a character is painted, the pixels forming the
 * character's shape are filled with the Graphics object's current color, and
 * the pixels not part of the character's shape are left untouched.
 * The text drawing calls
 * {@link #drawChar drawChar()},
 * {@link #drawChars drawChars()},
 * {@link #drawString drawString()}, and
 * {@link #drawSubstring drawSubstring()}
 * all draw text in this manner. <p>
 *
 * Lines, arcs, rectangles, and rounded rectangles may be drawn with either a
 * SOLID or a DOTTED stroke style, as set by the {@link #setStrokeStyle
 * setStrokeStyle()} method.  The stroke style does not affect fill, text, and
 * image operations. <p>
 * 
 * For the SOLID stroke style,
 * drawing operations are performed with a one-pixel wide pen that fills
 * the pixel immediately
 * below and to the right of the specified coordinate. Drawn lines
 * touch pixels at both endpoints. Thus, the operation
 * <pre>
 *    g.drawLine(0, 0, 0, 0)
 * </pre>
 * paints exactly one pixel, the first pixel in the upper left corner
 * of the display. <p>
 *
 * Drawing operations under the DOTTED stroke style will touch a subset of 
 * pixels that would have been touched under the SOLID stroke style.  The 
 * frequency and length of dots is implementation-dependent.  The endpoints of 
 * lines and arcs are not guaranteed to be drawn, nor are the corner points of 
 * rectangles guaranteed to be drawn.  Dots are drawn by painting with the 
 * current color; spaces between dots are left untouched. <p>
 *
 * An artifact of the coordinate system is that the area affected by a fill
 * operation differs slightly from the area affected by a draw operation given
 * the same coordinates. For example, consider the operations
 * <pre>
 *    g.fillRect(x, y, w, h); // 1
 *    g.drawRect(x, y, w, h); // 2
 * </pre>
 * Statement (1) fills a rectangle w pixels wide and h pixels high.
 * Statement (2) draws a rectangle whose left and top
 * edges are within the area filled by statement (1). However, the
 * bottom and right edges lie one pixel outside the filled area.
 * This is counterintuitive, but it preserves the invariant that
 * <pre>
 *    g.drawLine(x, y, x+w, y);
 *    g.drawLine(x+w, y, x+w, y+h);
 *    g.drawLine(x+w, y+h, x, y+h);
 *    g.drawLine(x, y+h, x, y);
 * </pre>
 * has an effect identical to statement (2) above. <p>
 *
 * The exact pixels painted by drawLine() and drawArc() are not
 * specified. Pixels touched by a fill operation must either
 * exactly overlap or directly abut pixels touched by the
 * corresponding draw operation. A fill operation must never leave
 * a gap between the filled area and the pixels touched by the
 * corresponding draw operation, nor may the fill operation touch
 * pixels outside the area bounded by the corresponding draw operation. <p>
 *
 * There is a single clipping rectangle. Operations are provided for
 * intersecting the current clip rectangle with a given rectangle and for
 * setting the current clip rectangle outright. The only pixels touched by
 * graphics operations are those that lie entirely within the clip rectangle.
 * Pixels outside the clip rectangle are not affected by any graphics
 * operations. It is legal to specify a clipping rectangle whose width or
 * height is zero or negative. In this case the clipping rectangle is
 * considered to be empty, that is, no pixels are contained within it.
 * Therefore, if any graphics operations are issued under such a clipping
 * rectangle, no pixels will be modified. <p>
 *
 * If a graphics operation is affected by the clip rectangle, the pixels
 * touched by that operation must be the same ones that would be touched
 * as if the clip rectangle did not affect the operation. For example,
 * consider a clip rectangle (cx, cy, cw, ch) and a point (x1, y1) that
 * lies outside this rectangle and a point (x2, y2) that lies within this
 * rectangle. In the following code fragment,
 * <pre>
 *    g.setClip(0, 0, canvas.getWidth(), canvas.getHeight());
 *    g.drawLine(x1, y1, x2, y2); // 3
 *    g.setClip(cx, cy, cw, ch);
 *    g.drawLine(x1, y1, x2, y2); // 4
 * </pre>
 * 
 * The pixels touched by statement (4) must be identical to the pixels
 * within (cx, cy, cw, ch) touched by statement (3). <p>
 *
 * <a name="anchor"></a>
 * <h3>Anchor Points</h3> <p>
 *
 * The drawing of text is based on "anchor points".
 * Anchor points are used to minimize the amount of
 * computation required when placing text.
 * For example, in order to center a piece of text,
 * an application needs to call stringWidth() or
 * charWidth() to get the width and then perform a
 * combination of subtraction and division to
 * compute the proper location.
 * The method to draw text is defined as follows:
 * <pre>
 *    public void drawString(String text, int x, int y, int anchor);
 * </pre>
 * This method draws text in the current color,
 * using the current font
 * with its anchor point at (x,y). The definition
 * of the anchor point must be one of the
 * horizontal constants (LEFT, HCENTER, RIGHT)
 * combined with one of the vertical constants
 * (TOP, BASELINE, BOTTOM) using the logical OR operator. <p>
 *
 * Vertical centering of the text is not specified since it is not considered
 * useful, it is hard to specify, and it is burdensome to implement. Thus,
 * the VCENTER value is not allowed in the anchor point parameter of text
 * drawing calls. <p>
 *
 * The actual position of the bounding box
 * of the text relative to the (x, y) location is
 * determined by the anchor point. These anchor
 * points occur at named locations along the
 * outer edge of the bounding box. Thus, if <code>f</code>
 * is <code>g</code>'s current font (as returned by
 * <code>g.getFont()</code>, the following calls will all have
 * identical results:
 * <pre>
 *    g.drawString(str, x, y, TOP|LEFT);
 *    g.drawString(str, x + f.stringWidth(str)/2, y, TOP|HCENTER);
 *    g.drawString(str, x + f.stringWidth(str), y, TOP|RIGHT);
 *
 *    g.drawString(str, x,
 *        y + f.getBaselinePosition(), BASELINE|LEFT);
 *    g.drawString(str, x + f.stringWidth(str)/2,
 *        y + f.getBaselinePosition(), BASELINE|HCENTER);
 *    g.drawString(str, x + f.stringWidth(str),
 *        y + f.getBaselinePosition(), BASELINE|RIGHT);
 *
 *    drawString(str, x,
 *        y + f.getHeight(), BOTTOM|LEFT);
 *    drawString(str, x + f.stringWidth(str)/2,
 *        y + f.getHeight(), BOTTOM|HCENTER);
 *    drawString(str, x + f.stringWidth(str),
 *        y + f.getHeight(), BOTTOM|RIGHT);
 * </pre> <p>
 *
 * For text drawing, the inter-character and inter-line spacing (leading)
 * specified by the font designer are included as part of the values returned
 * in the {@link Font#stringWidth(java.lang.String) stringWidth()} 
 * and {@link Font#getHeight() getHeight()} 
 * calls of class {@link Font Font}.
 * For example, given the following code:
 * <pre>
 *    // (5)
 *    g.drawString(string1+string2, x, y, TOP|LEFT);
 *
 *    // (6)
 *    g.drawString(string1, x, y, TOP|LEFT);
 *    g.drawString(string2, x + f.stringWidth(string1), y, TOP|LEFT);
 * </pre>
 * Code fragments (5) and (6) behave identically. This
 * occurs because f.stringWidth()
 * includes the inter-character spacing.
 * Similarly, reasonable vertical spacing may be
 * achieved simply by adding the font height
 * to the Y-position of subsequent lines. For example:
 * <pre>
 *    g.drawString(string1, x, y, TOP|LEFT);
 *    g.drawString(string2, x, y + f.fontHeight(), TOP|LEFT);
 * </pre>
 * draws string1 and string2 on separate lines with
 * an appropriate amount of inter-line spacing. <p>
 *
 * The stringWidth() of the string and the fontHeight() of the font in which
 * it is drawn define the size of the bounding box of a piece of text. As
 * described above, this box includes inter-line and inter-character spacing.
 * The implementation is required to put this space below and to right of the
 * pixels actually belonging to the characters drawn. Applications that wish
 * to position graphics closely with respect text (for example, to paint a
 * rectangle around a string of text) may assume that there is space below and
 * to the right of a string and that there is <em>no</em> space above and to the
 * left of the string. <p>
 *
 * Anchor points are also used for positioning of images. Similar to text
 * drawing, the anchor point for an image specifies the point on the bounding
 * rectangle of the destination that is to positioned at the (x,y) location
 * given in the graphics request. Unlike text, vertical centering of images
 * is well-defined, and thus the VCENTER value may be used within the anchor
 * point parameter of image drawing requests. Because images have no notion
 * of a baseline, the BASELINE value may not be used within the anchor point
 * parameter of image drawing requests.
 */
public class Graphics {

    // SYNC NOTE: the main Graphics class is entirely unlocked.  There is only
    // one instance of Graphics (created in Display) and it is only ever legal
    // for the application to use it when the system calls the paint() method
    // of a Canvas.  Since paint() calls are serialized, and these methods
    // modify only instance state, no locking is necessary.  (If any method
    // were to read or modify global state, locking would need to be done in
    // those cases.)

    /**
     * <P>Constant for centering text and images horizontally
     * around the anchor point</P>
     *
     * <P>Value 1 is assigned to HCENTER.</P>
     */
    public static final int HCENTER = 1;

    /**
     * <P>Constant for centering images vertically
     * around the anchor point.</P>
     *
     * <P>Value 2 is assigned to VCENTER.</P>
     */
    public static final int VCENTER = 2;

    /**
     * <P>Constant for positioning the anchor point of text and images
     * to the left of the text or image.</P>
     *
     * <P>Value 4 is assigned to LEFT.</P>
     */
    public static final int LEFT = 4;

    /**
     * <P>Constant for positioning the anchor point of text and images
     * to the right of the text or image.</P>
     *
     * <P>Value 8 is assigned to RIGHT.</P>
     */
    public static final int RIGHT = 8;

    /**
     * <P>Constant for positioning the anchor point of text and images
     * above the text or image.</P>
     *
     * <P>Value 16 is assigned to TOP.</P>
     */
    public static final int TOP = 16;

    /**
     * <P>Constant for positioning the anchor point of text and images
     * below the text or image.</P>
     *
     * <P>Value 32 is assigned to BOTTOM.</P>
     */
    public static final int BOTTOM = 32;

    /**
     * <P>Constant for positioning the anchor point at the baseline of text.</P>
     *
     * <P>Value 64 is assigned to BASELINE.</P>
     */
    public static final int BASELINE = 64;

    /**
     * <P>Constant for the SOLID stroke style.</P>
     *
     * <P>Value 0 is assigned to SOLID.</P>
     */
    public static final int SOLID = 0;

    /**
     * <P>Constant for the DOTTED stroke style.</P>
     *
     * <P>Value 1 is assigned to DOTTED.</P>
     */
    public static final int DOTTED = 1;

    /**
     * Intialize the native peer of this Graphics context
     */
    private native void init();

    /**
     * Create a Graphics object with the given width and height
     *
     * @param w The maximum width of the new Graphics object
     * @param h The maximum height of the new Graphics object
     */
    Graphics(int w, int h) {
        destination = null;

        maxWidth  = w;
        maxHeight = h;

        init();
        reset();
    }

    /**
     * Retrieve the Graphics context for the given Image
     *
     * @param img The Image to get a Graphics context for
     * @return Graphics Will return a new ImageGraphics object if
     *                  the Image is non-null, otherwise will return
     *                  a new Graphics object with Display.WIDTH
     *                  maximum width and Display.HEIGHT maximum
     *                  height
     */
    static Graphics getGraphics(Image img) {
        if (img == null) {
            return new Graphics(Display.WIDTH, Display.HEIGHT);
        } else {
            return new ImageGraphics(img);
        }
    }

    /**
     * Translates the origin of the graphics context to the point
     * (x, y) in the current coordinate system. All coordinates
     * used in subsequent rendering operations on this graphics
     * context will be relative to this new origin.<p>
     *
     * The effect of calls to translate() are cumulative. For example, calling
     * translate(1, 2) and then translate(3, 4) results in a translation of
     * (4, 6). <p>
     *
     * The application can set an absolute origin (ax, ay) using the following
     * technique:<p>
     * <code>
     * g.translate(ax - g.getTranslateX(), ay - g.getTranslateY())
     * </code><p>
     *
     * @param x the x coordinate of the new translation origin
     * @param y the y coordinate of the new translation origin
     * @see #getTranslateX()
     * @see #getTranslateY()
     */
    public void translate(int x, int y) {
        transX += x;
        transY += y;
    }

    /**
     * Gets the X coordinate of the translated origin of this graphics context.
     * @return X of current origin
     */
    public int getTranslateX() {
        return transX;
    }

    /**
     * Gets the Y coordinate of the translated origin of this graphics context.
     * @return Y of current origin
     */
    public int getTranslateY() {
        return transY;
    }

    /**
     * Gets the current color.
     * @return an integer in form 0x00RRGGBB
     * @see #setColor(int, int, int)
     */
    public int getColor() {
        return rgbColor;
    }

    /**
     * Gets the red component of the current color.
     * @return integer value in range 0-255
     * @see #setColor(int, int, int)
     */
    public int getRedComponent() {
        return (rgbColor >> 16) & 0xff;
    }

    /**
     * Gets the green component of the current color.
     * @return integer value in range 0-255
     * @see #setColor(int, int, int)
     */
    public int getGreenComponent() {
        return (rgbColor >> 8) & 0xff;
    }

    /**
     * Gets the blue component of the current color.
     * @return integer value in range 0-255
     * @see #setColor(int, int, int)
     */
    public int getBlueComponent() {
        return rgbColor & 0xff;
    }

    /**
     * Gets the current grayscale value of the color being used for rendering
     * operations. If the color was set by setGrayScale(), that value is simply
     * returned. If the color was set by one of the methods that allows setting
     * of the red, green, and blue components, the value returned is 
     * computed from
     * the RGB color components (possibly in a device-specific fashion) 
     * that best
     * approximates the brightness of that color.
     *
     * @return integer value in range 0-255
     * 
     * @see #setGrayScale
     */
    public int getGrayScale() {
        return gray;
    }

    /**
     * Get a gray value given the RGB values
     *
     * @param red The Red pixel value
     * @param green The Green pixel value
     * @param blue The Blue pixel value
     * @return int The grayscale value corresponding to the RGB color
     */
    private static int grayVal(int red, int green, int blue) {
        /* CCIR Rec 601 luma (nonlinear rgb to nonlinear "gray") */
        return (red*76 + green*150 + blue*29) >> 8;
    }

    /**
     * Get a specific pixel value
     *
     * @param rgb
     * @param gray
     * @param isGray
     * @return int
     */
    private native int getPixel(int rgb, int gray, boolean isGray);

    /**
     * Sets the current color to the specified RGB values. All subsequent
     * rendering operations will use this specified color.
     * @param red The red component of the color being set in range 0-255.
     * @param green The green component of the color being set in range 0-255.
     * @param blue The blue component of the color being set in range 0-255.
     * @throws IllegalArgumentException if any of the color components
     * are outside of range 0-255.
     *
     * @see #getColor
     */
    public void setColor(int red, int green, int blue) {
        if ((red < 0)   || (red > 255) 
            || (green < 0) || (green > 255)
            || (blue < 0)  || (blue > 255)) {
            throw new IllegalArgumentException("Value out of range");
        }

        rgbColor = (red << 16) | (green << 8) | blue;
        gray = grayVal(red, green, blue);
        pixel = getPixel(rgbColor, gray, false);
    }

    /**
     * Sets the current color to the specified RGB values. All subsequent
     * rendering operations will use this specified color. The RGB value
     * passed in is interpreted with the least significant eight bits
     * giving the blue component, the next eight more significant bits
     * giving the green component, and the next eight more significant
     * bits giving the red component. That is to say, the color component
     * is specified in the form of 0x00RRGGBB. The high order byte of this value
     * is ignored.
     *
     * @param RGB the color being set
     *
     * @see #getColor
     */
    public void setColor(int RGB) {
        int red   = (RGB >> 16) & 0xff;
        int green = (RGB >> 8)  & 0xff;
        int blue  = (RGB)  & 0xff;

        rgbColor = RGB & 0x00ffffff;
        gray = grayVal(red, green, blue);
        pixel = getPixel(rgbColor, gray, false);
    }

    /**
     * Sets the current grayscale to be used for all subsequent
     * rendering operations. For monochrome displays, the behavior
     * is clear. For color displays, this sets the color for all
     * subsequent drawing operations to be a gray color equivalent
     * to the value passed in. The value must be in the range 0-255.
     * @param value the desired grayscale value
     * @throws IllegalArgumentException if the gray value is out of range
     *
     * @see #getGrayScale
     */
    public void setGrayScale(int value) {
        if ((value < 0) || (value > 255)) {
            throw new IllegalArgumentException("Gray value out of range");
        }

        rgbColor = (value << 16) | (value << 8) | value;
        gray = value;
        pixel = getPixel(rgbColor, gray, true);
    }

    /**
     * Gets the current font.
     * @return current font
     * @see javax.microedition.lcdui.Font
     * @see #setFont(javax.microedition.lcdui.Font)
     */
    public Font getFont() {
        return currentFont;
    }

    /**
     * Sets the stroke style used for drawing lines, arcs, rectangles, and 
     * rounded rectangles.  This does not affect fill, text, and image 
     * operations.
     * @param style can be SOLID or DOTTED
     * @throws IllegalArgumentException if the style is illegal
     *
     * @see #getStrokeStyle
     */
    public void setStrokeStyle(int style) {
        if ((style != SOLID) && (style != DOTTED)) {
            throw new IllegalArgumentException("Invalid line style");
        }

        this.style = style;
    }

    /**
     * Gets the stroke style used for drawing operations.
     * @return stroke style, SOLID or DOTTED
     *
     * @see #setStrokeStyle
     */
    public int getStrokeStyle() {
        return style;
    }

    /**
     * Sets the font for all subsequent text rendering operations.  If font is 
     * null, it is equivalent to
     * <code>setFont(Font.getDefaultFont())</code>.
     * 
     * @param font the specified font
     * @see javax.microedition.lcdui.Font
     * @see #getFont()
     * @see #drawString(java.lang.String, int, int, int)
     * @see #drawChars(char[], int, int, int, int, int)
     */
    public void setFont(Font font) {
        currentFont = (font == null) ? Font.getDefaultFont() : font;
    }
 
    /**
     * Gets the X offset of the current clipping area, relative
     * to the coordinate system origin of this graphics context.
     * Separating the getClip operation into two methods returning
     * integers is more performance and memory efficient than one
     * getClip() call returning an object.
     * @return X offset of the current clipping area
     * @see #clipRect(int, int, int, int)
     * @see #setClip(int, int, int, int)
     */
    public int getClipX() {
        return clip[0] - transX;
    }

    /**
     * Gets the Y offset of the current clipping area, relative
     * to the coordinate system origin of this graphics context.
     * Separating the getClip operation into two methods returning
     * integers is more performance and memory efficient than one
     * getClip() call returning an object.
     * @return Y offset of the current clipping area
     * @see #clipRect(int, int, int, int)
     * @see #setClip(int, int, int, int)
     */
    public int getClipY() {
        return clip[1] - transY;
    }

    /**
     * Gets the width of the current clipping area.
     * @return width of the current clipping area.
     * @see #clipRect(int, int, int, int)
     * @see #setClip(int, int, int, int)
     */
    public int getClipWidth() {
        return clip[2];
    }


    /**
     * Gets the height of the current clipping area.
     * @return height of the current clipping area.
     * @see #clipRect(int, int, int, int)
     * @see #setClip(int, int, int, int)
     */
    public int getClipHeight()	{
        return clip[3];
    }

    /**
     * Intersects the current clip with the specified rectangle.
     * The resulting clipping area is the intersection of the current
     * clipping area and the specified rectangle.
     * This method can only be used to make the current clip smaller.
     * To set the current clip larger, use the setClip method.
     * Rendering operations have no effect outside of the clipping area.
     * @param x the x coordinate of the rectangle to intersect the clip with
     * @param y the y coordinate of the rectangle to intersect the clip with
     * @param width the width of the rectangle to intersect the clip with
     * @param height the height of the rectangle to intersect the clip with
     * @see #setClip(int, int, int, int)
     */
    public void clipRect(int x, int y, int width, int height) {
        x += transX;
        y += transY;

	/*
	 * short values are safe because we guarantee positive numbers
	 * and a width and height which don't exceed the displayable area.
	 */
        if (x > clip[0]) {
            clip[0] = (short) (x & 0x7fff);
            clipped = true;
        }

        if (y > clip[1]) {
            clip[1] = (short) (y & 0x7fff);
            clipped = true;
        }

        x += width;
        y += height;

        if (x < clipX2) {
            clipX2 = x;
            clipped = true;
        }

        if (y < clipY2) {
            clipY2 = y;
            clipped = true;
        }

        if (clipped) {
	    clip[2] = (short) (clipX2 - clip[0]);
	    clip[3] = (short) (clipY2 - clip[1]);

            if ((clip[2] < 0) || (clip[3] < 0)) {
                clip[2] = 0;
                clip[3] = 0;
            }
        }
    }

    /**
     * Sets the current clip to the rectangle specified by the
     * given coordinates.
     * Rendering operations have no effect outside of the clipping area.
     * @param x the x coordinate of the new clip rectangle
     * @param y the y coordinate of the new clip rectangle
     * @param width the width of the new clip rectangle
     * @param height the height of the new clip rectangle
     * @see #clipRect(int, int, int, int)
     */
    public void setClip(int x, int y, int width, int height) {
        x += transX;
        y += transY;

        short clipX1 = (x > 0) ? ((short)(x & 0x7fff)) : 0;
        short clipY1 = (y > 0) ? ((short)(y & 0x7fff)) : 0;

        if ((x >= maxWidth)  || (width <= 0)
            || (y >= maxHeight) || (height <= 0)) {
            clip[0] = clip[1] = clip[2] = clip[3] = 0;
                clipped = true;
            return;
        }

            clipX2 = x + width;
        if (clipX2 > maxWidth) {
            clipX2 = maxWidth;
        }

            clipY2 = y + height;
        if (clipY2 > maxHeight) {
            clipY2 = maxHeight;
        }

        /* safe because we already guaranteed values are small positive */
        clip[0] = clipX1;
        clip[1] = clipY1;
        clip[2] = (short) (clipX2 - clipX1);
        clip[3] = (short) (clipY2 - clipY1);

        if (clip[2] < 0) {
            clip[2] = 0;
        }

        if (clip[3] < 0) {
            clip[3] = 0;
        }

        if ((clipX1 > 0) || (clipY1 > 0)
            || (clipX2 < maxWidth) || (clipY2 < maxHeight)) {
            clipped = true;
        }
    }

    /**
     * Draws a line between the coordinates (x1,y1) and (x2,y2) using
     * the current color and stroke style.
     * @param x1 the x coordinate of the start of the line
     * @param y1 the y coordinate of the start of the line
     * @param x2 the x coordinate of the end of the line
     * @param y2 the y coordinate of the end of the line
     */
    public native void drawLine(int x1, int y1, int x2, int y2);

    /**
     * Fills the specified rectangle with the current color.
     * If either width or height is zero or less,
     * nothing is drawn.
     * @param x the x coordinate of the rectangle to be filled
     * @param y the y coordinate of the rectangle to be filled
     * @param width the width of the rectangle to be filled
     * @param height the height of the rectangle to be filled
     * @see #drawRect(int, int, int, int)
     */
    public native void fillRect(int x, int y, int width, int height);
 
    /**
     * Draws the outline of the specified rectangle using the current
     * color and stroke style.
     * The resulting rectangle will cover an area (width + 1)
     * pixels wide by (height + 1) pixels tall.
     * If either width or height is less than
     * zero, nothing is drawn.
     * @param x the x coordinate of the rectangle to be drawn
     * @param y the y coordinate of the rectangle to be drawn
     * @param width the width of the rectangle to be drawn
     * @param height the height of the rectangle to be drawn
     * @see #fillRect(int, int, int, int)
     */
    public native void drawRect(int x, int y, int width, int height);

    /**
     * Draws the outline of the specified rounded corner rectangle
     * using the current color and stroke style.
     * The resulting rectangle will cover an area (width + 1) pixels wide
     * by (height + 1) pixels tall.
     * If either width or height is less than
     * zero, nothing is drawn.
     * @param x the x coordinate of the rectangle to be drawn
     * @param y the y coordinate of the rectangle to be drawn
     * @param width the width of the rectangle to be drawn
     * @param height the height of the rectangle to be drawn
     * @param arcWidth the horizontal diameter of the arc at the four corners
     * @param arcHeight the vertical diameter of the arc at the four corners
     * @see #fillRoundRect(int, int, int, int, int, int)
     */
    public native void drawRoundRect(int x, int y, int width, int height,
                                     int arcWidth, int arcHeight);
 
    /**
     * Fills the specified rounded corner rectangle with the current color.
     * If either width or height is zero or less,
     * nothing is drawn.
     * @param x the x coordinate of the rectangle to be filled
     * @param y the y coordinate of the rectangle to be filled
     * @param width the width of the rectangle to be filled
     * @param height the height of the rectangle to be filled
     * @param arcWidth the horizontal diameter of the arc at the four
     * corners
     * @param arcHeight the vertical diameter of the arc at the four corners
     * @see #drawRoundRect(int, int, int, int, int, int)
     */
    public native void fillRoundRect(int x, int y, int width, int height,
                                     int arcWidth, int arcHeight);
                          
    /**
     * Fills a circular or elliptical arc covering the specified rectangle.
     * <p>
     * The resulting arc begins at <code>startAngle</code> and extends
     * for <code>arcAngle</code> degrees.
     * Angles are interpreted such that 0 degrees
     * is at the 3 o'clock position.
     * A positive value indicates a counter-clockwise rotation
     * while a negative value indicates a clockwise rotation.
     * <p>
     * The center of the arc is the center of the rectangle whose origin
     * is (<em>x</em>,&nbsp;<em>y</em>) and whose size is specified by the
     * <code>width</code> and <code>height</code> arguments.
     * <p>
     * If either width or height is zero or less,
     * nothing is drawn.
     *
     * <p> The filled region consists of the "pie wedge" region bounded
     *  by the arc
     * segment as if drawn by drawArc(), the radius extending from the
     * center to
     * this arc at <code>startAngle</code> degrees, and radius
     * extending from the
     * center to this arc at <code>startAngle + arcAngle</code> degrees. </p>
     *
     * <p> The angles are specified relative to the non-square extents of
     * the bounding rectangle such that 45 degrees always falls on the
     * line from the center of the ellipse to the upper right corner of
     * the bounding rectangle. As a result, if the bounding rectangle is
     * noticeably longer in one axis than the other, the angles to the
     * start and end of the arc segment will be skewed farther along the
     * longer axis of the bounds. </p>
     *
     * @param x the <em>x</em> coordinate of the upper-left corner of
     * the arc to be filled.
     * @param y the <em>y</em> coordinate of the upper-left corner of the
     * arc to be filled.
     * @param width the width of the arc to be filled
     * @param height the height of the arc to be filled
     * @param startAngle the beginning angle.
     * @param arcAngle the angular extent of the arc,
     * relative to the start angle.
     * @see #drawArc(int, int, int, int, int, int)
     */
    public native void fillArc(int x, int y, int width, int height,
                               int startAngle, int arcAngle);

    /**
     * Draws the outline of a circular or elliptical arc
     * covering the specified rectangle,
     * using the current color and stroke style.
     * <p>
     * The resulting arc begins at <code>startAngle</code> and extends
     * for <code>arcAngle</code> degrees, using the current color.
     * Angles are interpreted such that 0&nbsp;degrees
     * is at the 3&nbsp;o'clock position.
     * A positive value indicates a counter-clockwise rotation
     * while a negative value indicates a clockwise rotation.
     * <p>
     * The center of the arc is the center of the rectangle whose origin
     * is (<em>x</em>,&nbsp;<em>y</em>) and whose size is specified by the
     * <code>width</code> and <code>height</code> arguments.
     * <p>
     * The resulting arc covers an area
     * <code>width&nbsp;+&nbsp;1</code> pixels wide
     * by <code>height&nbsp;+&nbsp;1</code> pixels tall.
     * If either width or height is less than zero,
     * nothing is drawn.
     *
     * <p> The angles are specified relative to the non-square extents of
     * the bounding rectangle such that 45 degrees always falls on the
     * line from the center of the ellipse to the upper right corner of
     * the bounding rectangle. As a result, if the bounding rectangle is
     * noticeably longer in one axis than the other, the angles to the
     * start and end of the arc segment will be skewed farther along the
     * longer axis of the bounds. </p>
     *
     * @param x the <em>x</em> coordinate of the upper-left corner of
     * the arc to be drawn.
     * @param y the <em>y</em> coordinate of the upper-left corner of
     * the arc to be drawn.
     * @param width the width of the arc to be drawn
     * @param height the height of the arc to be drawn
     * @param startAngle the beginning angle
     * @param arcAngle the angular extent of the arc, relative to
     * the start angle.
     * @see #fillArc(int, int, int, int, int, int)
     */
    public native void drawArc(int x, int y, int width, int height,
                               int startAngle, int arcAngle);

    /**
     * Draws the specified String using the current font and color.
     * The x,y position is the position of the anchor point.
     * See <a href="#anchor">anchor points</a>.
     * @param str the String to be drawn
     * @param x the x coordinate of the anchor point
     * @param y the y coordinate of the anchor point
     * @param anchor the anchor point for positioning the text
     * @throws NullPointerException if str is null
     * @throws IllegalArgumentException if anchor is not a legal value
     * @see #drawChars(char[], int, int, int, int, int)
     */
    public native void drawString(java.lang.String str,
                                  int x, int y, int anchor);

    /**
     * Draws the specified String using the current font and color.
     * The x,y position is the position of the anchor point.
     * See <a href="#anchor">anchor points</a>.
     * @param str the String to be drawn
     * @param offset zero-based index of first character in the substring
     * @param len length of the substring
     * @param x the x coordinate of the anchor point
     * @param y the y coordinate of the anchor point
     * @param anchor the anchor point for positioning the text
     * @see #drawString(String, int, int, int).
     * @throws StringIndexOutOfBoundsException if offset and length do not
     * specify a valid range within the String str
     * @throws IllegalArgumentException if anchor is not a legal value
     * @throws NullPointerException if str is null
     */
    public native void drawSubstring(String str, int offset, int len,
                                     int x, int y, int anchor);

    /**
     * Draws the specified character using the current font and color.
     * @param character the character to be drawn
     * @param x the x coordinate of the anchor point
     * @param y the y coordinate of the anchor point
     * @param anchor the anchor point for positioning the text; see
     * <a href="#anchor">anchor points</a>
     *
     * a valid range within the data array
     * @throws IllegalArgumentException if anchor is not a legal value
     *
     * @see #drawString(java.lang.String, int, int, int)
     * @see #drawChars(char[], int, int, int, int, int)
     */
    public native void drawChar(char character, int x, int y, int anchor);


    /**
     * Draws the specified characters using the current font and color.
     * @param data the array of characters to be drawn
     * @param offset the start offset in the data
     * @param length the number of characters to be drawn
     * @param x the x coordinate of the anchor point
     * @param y the y coordinate of the anchor point
     * @param anchor the anchor point for positioning the text; see
     * <a href="#anchor">anchor points</a>
     *
     * @throws ArrayIndexOutOfBoundsException if offset and length do not
     * specify a valid range within the data array
     * @throws IllegalArgumentException if anchor is not a legal value
     * @throws NullPointerException if data is null
     *
     * @see #drawString(java.lang.String, int, int, int)
     */
    public native void drawChars(char[] data, int offset, int length,
                                 int x, int y, int anchor);
 
    /**
     * Draws the specified image by using the anchor point.
     * The image can be drawn in different positions relative to
     * the anchor point by passing the appropriate position constants.
     * See <a href="#anchor">anchor points</a>.
     * @param img the specified image to be drawn
     * @param x the x coordinate of the anchor point
     * @param y the y coordinate of the anchor point
     * @param anchor the anchor point for positioning the image
     * @throws IllegalArgumentException if anchor is not a legal value
     * @throws NullPointerException if img is null
     * @see Image
     */
    public native void drawImage(Image img, int x, int y, int anchor);


    // private implementation //

    /** Translated x,y coordinates */
    private int transX, transY;
    /** Array to hold the clip values */
    private short clip[] = new short[4];     /* format is -- x y w h */
    /** A flag indicating the clipping state */
    private boolean clipped = false;
    /** Clip coordinates */
    private int clipX2, clipY2;
    /** Pixel values */
    private int rgbColor, gray, pixel;
    /** Line stroke style */
    private int style;          // line stroke style
    /** The current Font */
    private Font currentFont;
    /** The maximum width and height */
    private int maxWidth, maxHeight;
    /** The 'destination' Image to paint to */
    Image destination;

    /**
     * Reset this Graphics context with the given coordinates
     *
     * @param x1 The upper left x coordinate
     * @param y1 The upper left y coordinate
     * @param x2 The lower right x coordinate
     * @param y2 The lower right y coordinate
     */
    void reset(int x1, int y1, int x2, int y2) {
        transX = transY = 0;
        short clipX1 = (x1 > 0) ? ((short) (x1 & 0x7fff)) : 0;
        short clipY1 = (y1 > 0) ? ((short) (y1 & 0x7fff)) : 0;

        clipX2  = x2;
        clipY2  = y2;

        clip[0] = clipX1;
        clip[1] = clipY1;
        clip[2] = (short) (clipX2 - clipX1);
        clip[3] = (short) (clipY2 - clipY1);

        if (clip[2] < 0) {
            clip[2] = 0;
        }

        if (clip[3] < 0) {
            clip[3] = 0;
        }

        if ((clipX1 > 0) || (clipY1 > 0)
            || (clipX2 < maxWidth) || (clipY2 < maxHeight)) {
            clipped = true;
        } else {
            clipped = false;
        }

        currentFont = Font.getDefaultFont();

        style = SOLID;
        rgbColor = gray = 0;
        pixel = getPixel(rgbColor, gray, true);
    }

    /**
     * Reset this Graphics context to its default dimensions
     * (same as reset(0, 0, maxWidth, maxHeight)
     */
    void reset() {
        reset(0, 0, maxWidth, maxHeight);
    }

}

/**
 * <p>SYNC NOTE: Unlike the Graphics class, multiple threads may issue calls
 * on ImageGraphics instances at any time.  It is therefore the responsibility
 * of the ImageGraphics class to ensure that all public methods are MT-safe.
 * </p>
 * 
 * <p>Some methods are implicitly safe already, namely those that are native
 * methods and those that make a single atomic access to instance state.
 * Those method implementations are simply inherited from the superclass.  The
 * inherited methods are included below, within comments, to indicate this
 * fact.  If the superclass' implementation of a method changes, this class
 * may need to override it and provide locking. </p>
 * 
 * <p>Any method that has multiple accesses to shared state must be protected
 * with a synchronized block.  As for the Graphics class, all access is to the
 * state of this instance, and so locking is done on 'this' instead of on
 * Display.LCDUILock.</p>
 */
class ImageGraphics extends Graphics {

    /**
     * Constuct a new ImageGraphics
     *
     * @param img The Image to use to construct the ImageGraphics
     */
    ImageGraphics(Image img) {
        super(img.getWidth(), img.getHeight());
        destination = img;
    }

    /**
     * Translate this Graphics context by the given offset
     *
     * @param x The x coordinate offset
     * @param y The y coordinate offset
     */
    public void translate(int x, int y) {
        synchronized (this) {
            super.translate(x, y);
        }
    }

    /**
     * Set the Color of this Graphics context
     *
     * @param red The Red color component
     * @param green The Green color component
     * @param blue The Blue color component
     */
    public void setColor(int red, int green, int blue) {
        synchronized (this) {
            super.setColor(red, green, blue);
        }
    }

    /**
     * Set the Color of this Graphics context
     *
     * @param RGB The RGB composite color value
     */
    public void setColor(int RGB) {
        synchronized (this) {
            super.setColor(RGB);
        }
    }

    /**
     * Set the gray scale value of this Graphics context
     *
     * @param value The gray scale value
     */
    public void setGrayScale(int value) {
        synchronized (this) {
            super.setGrayScale(value);
        }
    }

    /**
     * Get the x coordinate of the clip
     *
     * @return int The x coordinate of the clip
     */
    public int getClipX() {
        synchronized (this) {
            return super.getClipX();
        }
    }

    /**
     * Get the y coordinate of the clip
     *
     * @return int The y coordinate of the clip
     */
    public int getClipY() {
        synchronized (this) {
            return super.getClipY();
        }
    }

    // SYNC NOTE: the superclass implementations of getClipWidth() and
    // getClipHeight() read a single value atomically, so no locking is 
    // necessary.  We simply inherit the implementations from the superclass.

    // public int getClipWidth();
    // public int getClipHeight();

    /**
     * Add the clip rectangle of this Graphics context
     *
     * @param x The x coordinate of the rectangle
     * @param y The y coordinate of the rectangle
     * @param w The width of the rectangle
     * @param h The height of the rectangle
     */
    public void clipRect(int x, int y, int w, int h) {
        synchronized (this) {
            super.clipRect(x, y, w, h);
        }
    }

    /**
     * Set the clip of this Graphics context
     *
     * @param x The x coordinate of the clip
     * @param y The y coordinate of the clip
     * @param w The width of the clip
     * @param h The height of the clip
     */
    public void setClip(int x, int y, int w, int h) {
        synchronized (this) {
            super.setClip(x, y, w, h);
        }
    }

    // SYNC NOTE: all of the following are native methods, which run
    // atomically, so no synchronization is necessary.  We therefore simply
    // inherit them from the superclass.

    // public native void drawLine(int x1, int y1, int x2, int y2);

    // public native void fillRect(int x, int y, int width, int height);

    // public native void drawRect(int x, int y, int width, int height);

    // public native void drawRoundRect(int x, int y, int width, int height,
    //                                  int arcWidth, int arcHeight);

    // public native void fillRoundRect(int x, int y, int width, int height,
    //                                  int arcWidth, int arcHeight);

    // public native void fillArc(int x, int y, int width, int height,
    //                            int startAngle, int arcAngle);

    // public native void drawArc(int x, int y, int width, int height,
    //                            int startAngle, int arcAngle);

    // public native void drawString(java.lang.String str,
    //                               int x, int y, int anchor);

    // public native void drawSubstring(String str, int offset, int len,
    //                                  int x, int y, int anchor);

    // public native void drawChar(char character, int x, int y, int anchor);

    // public native void drawChars(char[] data, int offset, int length,
    //                              int x, int y, int anchor);

    // public native void drawImage(Image img, int x, int y, int anchor);
}
