/*
 * @(#)Screen.java	1.124 01/08/21 
 * Copyright (c) 1999-2001 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package javax.microedition.lcdui;

import java.util.TimerTask;
import java.util.Timer;

/**
 * <P>The common superclass of all high-level user interface classes.  Adds
 * optional title and ticker-tape output to the Displayable class.  The
 * contents displayed and their interaction with the user are defined by 
 * subclasses.</P>
 *
 * <P>Using subclass-defined methods, the application may change the contents
 * of a Screen object while it is shown to the user.  If this occurs, and the
 * Screen object is visible, the display will be updated automatically.  That
 * is, the implementation will refresh the display in a timely fashion without
 * waiting for any further action by the application.  For example, suppose a
 * List object is currently displayed, and every element of the List is
 * visible.  If the application inserts a new element at the beginning of the
 * List, it is displayed immediately, and the other elements will be
 * rearranged appropriately.  There is no need for the application to call
 * another method to refresh the display.</P>
 *
 * <P>It is recommended that applications change the contents of a Screen only
 * while it is not visible (that is, while another Displayable is current).
 * Changing the contents of a Screen while it is visible may result in
 * performance problems on some devices, and it may also be confusing if the
 * Screen's contents changes while the user is interacting with it.</P>
 */
public abstract class Screen extends Displayable  {

    /** The Ticker which may be set on this Screen */
    private Ticker ticker;
    /** A Timer which will handle firing repaints of the TickerPainter */
    private final static Timer  tickerTimer;
    /** A TimerTask which will repaint the Ticker on a repeated basis */
    private TickerPainter       tickerPainter;

    /** Renders the title for this Screen */
    private Layout screenTitle;

    /**
     * Whether or not this Screen is currently displayed on the physical
     * device
     */
    private boolean visible;         // = false;
    /**
     * Whether or not the Ticker (if there is one) is part of the display
     * of this Screen (ie, not being optimized out of the layout for space)
     */
    private boolean tickerIsVisible; // = false;
    /** this variable is set to true once layout was done */
    private boolean layoutDoneOnce;  // = false;

    /** Special title font */
    private final static  Font TITLE_FONT =
        Font.getFont(Font.FACE_SYSTEM, Font.STYLE_BOLD, Font.SIZE_MEDIUM);
    /** Special title height */
    private final static int TITLE_HEIGHT = TITLE_FONT.getHeight();

    /** Special content font, shared amongst the LCDUI package */
    final static Font CONTENT_FONT  = Font.getDefaultFont();
    /** Special content height, shared amongst the LCDUI package */
    final static int CONTENT_HEIGHT = CONTENT_FONT.getHeight();

    /** height of the child that is shown in the viewport */
    private int childHeight;        // = 0;

    /**
     * When hasBorder is true there is a pixel border painted around
     * content and viewPort is reduced by 4 pixels. This is used by
     * subclasses such as TextBox which need a pixel border around them.
     */
    boolean hasBorder;             // = false;

    /** view port X position (in Sreen coordinates) */
    private int globalViewPortX;
    /** view port Y position (in Sreen coordinates) */
    private int globalViewPortY;

    /** view port position (in child coordinates) */
    private int viewPortY;
    /** view port width */
    int viewPortWidth;
    /** view port height */
    int viewPortHeight;

    /**
     * Creates a new Screen object with no title and no ticker.
     */
    Screen() {
        this(null);
    }

    /**
     * Creates a new Screen object with the given title and with no ticker.
     *
     * @param title the Screen's title, or null for no title
     */
    Screen(String title) {
        // SYNC NOTE: since subclasses can't protect their call to super()
        // we'll sync it here
        synchronized (Display.LCDUILock) {
            if (title != null) {
                screenTitle = new StringLayout(title, TITLE_FONT);
            }
        }
    }

    /**
     * Gets the title of the Screen. 
     * 
     * @return a String representing the title value, or null if there 
     *         is no title
     * 
     * @see #setTitle     
     */
    public String getTitle() {
        synchronized (Display.LCDUILock) {
            return getTitleImpl();
        }
    }

    /**
     * Sets the title of the Screen. If null is given, removes the title. <p>
     *
     * If the Screen is physically visible, the visible effect
     * should take place no later than immediately
     * after the callback or
     * {@link javax.microedition.midlet.MIDlet#startApp startApp}
     * returns back to the implementation.
     *
     * @param s the new title, or null for no title
     *
     * @see #getTitle
     */
    public void setTitle(String s) {
        synchronized (Display.LCDUILock) {
            String oldTitle = getTitleImpl();
            if ((oldTitle == null && s == null) ||
                    (oldTitle != null && s != null && oldTitle.equals(s))) {
                return;
            }

	    int deltaHeight = 0;

            if (screenTitle != null) {
                deltaHeight = ((StringLayout)screenTitle).setString(s);
            } else if (s != null) {
                screenTitle = new StringLayout(s, TITLE_FONT);
                if (layoutDoneOnce) {
                    deltaHeight = screenTitle.setWidth(Display.WIDTH);
                }
            }

	    // if layout was done yet there is no need to repaint
	    // nor to re-layout
	    if (!layoutDoneOnce) {
		return;
	    }

	    // repaint only the title area if title height did not change
            if (deltaHeight == 0) {
                if (visible) {
                    int y = 0;
                    if (tickerIsVisible) {
                        y = ticker.PREFERRED_HEIGHT;
                    }
                    repaint(0, y, Display.WIDTH, screenTitle.getHeight());
                }
                return;
            }

            // layoutChanged will take care of layout
            // then do repaint if this screen is visible
            layoutChanged();
        } // sychronized
    }

    /**
     * Set a ticker for use with this Screen, replacing any previous ticker.
     * If null, removes the ticker object
     * from this screen. The same ticker is may be shared by several Screen
     * objects within an application. This is done by calling setTicker() on
     * different screens with the same Ticker object.
     * If the Screen is physically visible, the visible effect
     * should take place no later than immediately
     * after the callback or
     * {@link javax.microedition.midlet.MIDlet#startApp startApp}
     * returns back to the implementation.
     * @param ticker the ticker object used on this screen
     * 
     * @see #getTicker
     */
    public void setTicker(Ticker ticker) {
        synchronized (Display.LCDUILock) {
            // Return early if there's nothing to do
            if (this.ticker == ticker) {
                return;
            }

            // Stop any currently running Ticker.
            stopTicker();

            // We initialize the new ticker to start its message display
            // from the right side of the screen. This is important for new
            // tickers as well as tickers which may have already been running
            // and are making a subsequent appearance to the screen.
            // (Easier to add this here than attempt to fold into the
            // if-else below)
            if (ticker != null) {
                ticker.reset();
            }

            // CASES:
            // 1. Had an invisible non-null ticker, setting a null ticker
            //    - We need to set the new ticker. There's no need to re-layout
            //      or start the new ticker
            // 2. Had an invisible non-null ticker, setting a non-null ticker
            //    - We need to set the new ticker. There's no need to re-layout
            //      or start the new ticker
            // 3. Had a visible non-null ticker, setting a null ticker
            //    - We need to set the new ticker and re-layout. There's no
            //      need to start the new ticker.
            // 4. Had a null ticker, setting a non-null ticker
            //    - We need to set the new ticker, re-layout, and possibly
            //      start up the new ticker
            // 5. Had a visible non-null ticker, setting a non-null ticker
            //    - We need to set the new ticker. There's no need to re-layout
            //      but we do have to start the new ticker

            Ticker oldTicker = this.ticker;
            this.ticker = ticker;

            // This 'if' test covers conditions 3 & 4. Note we don't need the
            // test for case 4 to be 'this.ticker == null && ticker != null'
	    // because
            // we're guaranteed that 'ticker != null' above where we test for
            // 'this.ticker == ticker'; so 'this.ticker == null' is enough to
            // satisfy case 4.
            if ((tickerIsVisible && ticker == null) ||
                oldTicker == null)
            {
                layoutChanged(); // takes care of setting tickerIsVisible
            }

            // We know from cases 4 & 5 that we may need to start the new
            // ticker, we do the 'if' test to avoid a blanket call to
            // startTicker(), ie in cases 1 & 3. The added test for
            // tickerIsVisible shortcircuits the method call to startTicker()
            // which will simply return anyway (case 2).
            if (ticker != null && tickerIsVisible) {
                startTicker();
            }
        } // synchronized
    }

    /**
     * Gets the ticker used by this Screen.
     * @return ticker object used, or null if no ticker is present
     *
     * @see #setTicker     
     */
    public Ticker getTicker() {
        // SYNC NOTE: return of atomic value, no locking necessary
        return ticker;
    }

    // package private ********************************************

    /**
     * Paint the content of this Screen
     *
     * @param g The Graphics to paint to
     */
    abstract void paintContent(Graphics g);
    /**
     * Layout the content of this Screen
     *
     * @param w The width of the layout
     * @param h The height of the Layout
     * @return int The new height of the Screen
     */
    abstract int layoutContent(int w, int h);

    /**
     * Get the minimum height for this Screen
     *
     * @param width The maximum width
     * @param height The maximum height
     * @return int The minimum content height to display this Screen
     */
    int getMinimumContentHeight(int width, int height) {
        return (hasBorder ? (CONTENT_HEIGHT * 2 + 8) : CONTENT_HEIGHT * 2);
    }

    /**
     * Signal this Screen that its layout has changed
     */
    void layoutChanged() {
        if (layoutDoneOnce) {
            doLayout();
            if (visible) {
                repaint(0, 0, Display.WIDTH, Display.HEIGHT);
            }
        }
    }

    /**
     * Layout the contents of this Screen
     */
    final void doLayout() {
        int tickerHeight = 0;
        if (ticker != null) {
            // By default, the ticker is visible. Below will calculate
            // whether we optimize out the display of the ticker or not
            tickerIsVisible = true;
            tickerHeight = ticker.PREFERRED_HEIGHT;
        }

        int titleHeight = 0;
        int titleBorder = 0;

        String title = getTitleImpl();
        if (title != null && !title.equals("")) {
            screenTitle.setMaxHeight(2 * TITLE_HEIGHT);
            titleHeight = screenTitle.setWidth(Display.WIDTH);
            titleBorder = TITLE_IMG.getHeight() + 1;
        }

        int h = tickerHeight + titleHeight + titleBorder +
                getMinimumContentHeight(Display.WIDTH, 
                                        Display.HEIGHT - TITLE_HEIGHT);

        // reset the global viewport
        globalViewPortY = 0;

        // If there is not enough space to fit everything:
        // 1. ticker should not be displayed if it is present
        // 2. two line title should be reduced to 1 line if it is present
        // 3. content should be modified

        if (tickerHeight > 0 && h > Display.HEIGHT) {
            h -= tickerHeight;
            tickerIsVisible = false;
        } else {
            globalViewPortY = tickerHeight;
        }

        if (titleHeight != 0) {
            if (h > Display.HEIGHT && titleHeight > TITLE_HEIGHT) {
                h = h - titleHeight + TITLE_HEIGHT;
                titleHeight = TITLE_HEIGHT;
                screenTitle.setMaxHeight(titleHeight);
            }
        }

        globalViewPortY += titleHeight + titleBorder;
        viewPortWidth = Display.WIDTH;
        viewPortHeight = Display.HEIGHT - globalViewPortY;
        if (hasBorder) {
            globalViewPortX  = 4;
            globalViewPortY += 4;
            viewPortWidth   -= 8;
            viewPortHeight  -= 8;
        }

        childHeight = layoutContent(viewPortWidth, viewPortHeight);
        viewPortY = initHilight(viewPortY, viewPortHeight);
        layoutDoneOnce = true;
    }

    /**
     * Set the child height of this Screen
     *
     * @param height The childHeight
     */
    void setChild(int height) {
        childHeight = height;
    }

    /**
     * Initializes hilight when form is shown.
     * @param vpY y coordinate of viewport location
     * @param vpH viewport height
     * @return new y coordinate of viewport location
     *         child (all items) coordinate system
     */
    int initHilight(int vpY, int vpH) {
        return 0;
    }

    /**
     * Signal this Screen that its height has changed
     *
     * @param vpY
     * @param vpH
     * @param deltaHeight
     * @return int
     */
    int heightChanged(int vpY, int vpH, int deltaHeight) {
        childHeight += deltaHeight;

        if (vpY + vpH > childHeight) {
            vpY = childHeight - vpH;
            if (vpY < 0) {
                vpY = 0;
            }
        }
        return vpY;
    }

    /**
     * This method is called to change child height - must be 
     * called by any operation that can change the height of the screens
     * child.
     *
     * @param item The Item which changed
     * @param x The x coordinate of the change
     * @param y The y coordinate of the change
     * @param deltaHeight The change in height caused by the change
     */
    void contentChanged(Item item, int x, int y, int deltaHeight) {
        if (!initLayoutDone()) {
            return;
        }

        // Only call heightChanged if there's really a height delta
        // Fix of bug 4366597
        if (deltaHeight != 0) {
            viewPortY = heightChanged(viewPortY, viewPortHeight, deltaHeight);
        }

        // Repaint if this screen is visible
        if (isShown()) {
            setVerticalScroll();
            if (deltaHeight == 0 && item != null) {
                item.repaint(x, y, viewPortWidth, item.getHeight() - y);
                return;
            }
            repaintContent();
        }
    }

    /**
     * Set the vertical scroll indicators for this Screen
     */
    void setVerticalScroll() {

        // Previously, Screen was always resetting the left/right
        // scroll indicators to 0 when it set the up/down indicators,
        // now it sets the up/down indicators as appropriate, but
        // maintains the left/right indicators as they were, and it
        // is the component's responsibility (ie Gauge) to update
        // the right/left scroll indicators if necessary

        if (childHeight <= viewPortHeight) {
            super.setVerticalScroll(0, 100);
        } else {
            super.setVerticalScroll(
                (viewPortY * 100 / (childHeight - viewPortHeight)),
                (viewPortHeight * 100 / childHeight));
        }
    }

    /**
     * Determine if the initial layout has been done
     *
     * @return boolean True if the initial layout has been done
     */
    final boolean initLayoutDone() {
        return layoutDoneOnce;
    }


    /**
     * Repaint the content of this Screen
     */
    final void repaintContent() {
        repaint(globalViewPortX, globalViewPortY,
                viewPortWidth, viewPortHeight);
    }

    /**
     * Repaint the content of this Screen in the given clip rect
     *
     * @param x The x coordinate of the clip
     * @param y The y coordinate of the clip
     * @param width The width of the clip
     * @param height The height of the clip
     */
    final void repaintContent(int x, int y, int width, int height) {
        // do not repaint content that is entirely not visible
        if (y < viewPortY + viewPortHeight && y + height > viewPortY) {
            repaint(x + globalViewPortX, y + globalViewPortY - viewPortY,
                    width, height);
        }
    }

    /**
     * Paint this Screen
     *
     * @param g The Graphics context to paint to
     */
    final void paint(Graphics g) {
        int clipRectY1 = g.getClipY();
        int clipRectY2 = clipRectY1 + g.getClipHeight();

        int translatedY = 0;
        int titleTextHeight = 0;
        int titleHeight = 0;

        synchronized (Display.LCDUILock) {
            String title = getTitleImpl();
            if (title != null && !title.equals("") && screenTitle != null) {
                titleTextHeight = screenTitle.getHeight() + 2;
                titleHeight = titleTextHeight + TITLE_IMG.getHeight();
            }

            // paint ticker
            if (tickerIsVisible) {
                if (clipRectY2 > 0 && clipRectY1 < ticker.PREFERRED_HEIGHT) {
                    ticker.paintContent(g);
                }

                translatedY += ticker.PREFERRED_HEIGHT;
                g.translate(0, ticker.PREFERRED_HEIGHT);
            }

            // paint title
            if (titleHeight != 0) {
                if (clipRectY2 > translatedY &&
                    clipRectY1 < translatedY + titleHeight) {
                    g.setColor(Display.ERASE_COLOR);
                    g.fillRect(0, 0, Display.WIDTH, titleHeight);

                    screenTitle.paint(g, false, false);

                    // NOTE: We test to see if the width of the Display is
                    // wider than the width of our image, if so, we re-draw
                    // the image offset to the right as many times as necessary
                    // to fill the width of the Display. Specific ports would
                    // probably want to optimize this for the specific width
                    // of their device.
                    for (int imgLoc = 0; imgLoc < Display.WIDTH; imgLoc += 96)
                    {
                        g.drawImage(TITLE_IMG, imgLoc, titleTextHeight,
                                    Graphics.LEFT | Graphics.BOTTOM);
                    }
                }
                translatedY += titleHeight;
                g.translate(0, titleHeight);
            }

            // paint content
            // doLayout has already worked out the size of the content
            if (clipRectY2 > translatedY && clipRectY1 < Display.HEIGHT) {
                if (hasBorder) {
                    g.setColor(Display.ERASE_COLOR);
                    g.drawRect(0, 0, viewPortWidth + 7, viewPortHeight + 7);
                    g.drawRect(1, 1, viewPortWidth + 5, viewPortHeight + 5);
                    g.drawRect(3, 3, viewPortWidth + 1, viewPortHeight + 1);
                    g.setColor(Display.FG_COLOR);
                    g.drawRect(2, 2, viewPortWidth + 3, viewPortHeight + 3);
                }
                g.translate(globalViewPortX,
                    globalViewPortY - translatedY - viewPortY);
                g.clipRect(0, viewPortY, viewPortWidth, viewPortHeight);
                paintContent(g);
            }
        } // synchronized
    }

    /**
     * Traverse this Screen's contents
     *
     * @param dir The direction of traversal
     * @param top
     * @param bottom
     * @return -1 if there is no need to scroll and 
     *            there is no need to repaint;
     *          0 if there is no need to scroll but 
     *            content should be repainted
     *          scrollHeight by which the viewPortY has to be adjusted
     */
    int traverse(int dir, int top, int bottom) {
        // if we are at the top of the screen and we are scrolling up or
        // if we are scrolling down and we are at the very bottom,
        // then don't scroll.
        if (dir == Canvas.UP) {
            if (top <= 0) {
                return -1;
            }
        } else if (dir == Canvas.DOWN) {
            if (bottom >= childHeight) {
                return -1;
            }
        } else {
            return -1;
        }

        // calculate the amount to scroll a screenful,
        // with the minimum being a line of content
        int blockIncr = (((bottom - top + (CONTENT_HEIGHT / 2)) /
                           CONTENT_HEIGHT) - 1) * CONTENT_HEIGHT;
        if (blockIncr < CONTENT_HEIGHT) {
            blockIncr = CONTENT_HEIGHT;
        }

        if ((dir == Canvas.DOWN) && ((childHeight - bottom) < blockIncr)) {
            return (childHeight - bottom);
        } else if ((dir == Canvas.UP) && (top < blockIncr)) {
            return top;
        }

        return blockIncr;
    }

    /**
     * Handle a key pressed event
     *
     * @param keyCode The key that was pressed
     */
    void keyPressed(int keyCode) {
        synchronized (Display.LCDUILock) {
            keyPressedImpl(keyCode);
        } // synchronized
    }

    /**
     * Handle a key repeated event
     *
     * @param keyCode The key that was pressed
     */
    void keyRepeated(int keyCode) {
        // SYNC NOTE: Locking handled by keyPressed()
        keyPressed(keyCode);
    }

    // SYNC NOTE: These three empty methods should include a
    // synchronized implementation (similar to keyPressedImpl())
    // when/if they are needed
    /**
     * Handle a pointer pressed event
     *
     * @param x The x coordinate of the press
     * @param y The y coordinate of the press
     */
    void pointerPressedImpl(int x, int y) { }

    /**
     * Handle a pointer released event
     *
     * @param x The x coordinate of the release
     * @param y The y coordinate of the release
     */
    void pointerReleasedImpl(int x, int y) { }

    /**
     * Handle a pointer dragged event
     *
     * @param x The x coordinate of the drag
     * @param y The y coordinate of the drag
     */
    void pointerDraggedImpl(int x, int y) { }

    /**
     * This method is used to determine if that screen instance
     * is used as an edit screen for the screen passed in.
     *
     * @param d The Displayable to check
     * @return boolean True, if this Screen is the "edit" screen
     *                  for the Displayable
     */
    boolean isEditScreen(Displayable d) {
        return false;
    }

    /**
     * Set the content rectangle for this Screen
     *
     * @param rect An array holding the x,y,width,height
     *              values of the rectangle
     */
    void getContentRect(int rect[]) {
        if (rect != null && rect.length >= 4) {
            rect[0] = globalViewPortX;
            rect[1] = globalViewPortY;
            rect[2] = viewPortWidth;
            rect[3] = viewPortHeight;

            if (hasBorder) {
                rect[0] -= 2;
                rect[1] -= 2;
                rect[2] += 4;
                rect[3] += 4;
            }
        }
    }

    /**
     * Handle a key pressed event
     *
     * @param keyCode The key that was pressed
     */
    void keyPressedImpl(int keyCode) {
        int gameKeyCode = Display.getGameAction(keyCode);

        switch (gameKeyCode) {
            case Canvas.UP:
            case Canvas.DOWN:
            case Canvas.LEFT:
            case Canvas.RIGHT:
                int scrollHeight = traverse(gameKeyCode, viewPortY,
                                            viewPortY + viewPortHeight);

                // there is no need to scroll and
                // there is no need to repaint content
                if (scrollHeight < 0) {
                    return;
                }

                if (scrollHeight > 0) {
                    if (gameKeyCode == Canvas.UP ||
                            gameKeyCode == Canvas.LEFT) {
                        viewPortY -= scrollHeight;
                    } else if (gameKeyCode == Canvas.DOWN ||
                            gameKeyCode == Canvas.RIGHT) {
                        viewPortY += scrollHeight;
                    }
                }

                // there is a need to repaint content
                repaintContent();
                setVerticalScroll();
                break;
        }
    }

    /**
     * Handle a key repeated event
     *
     * @param keyCode The key that was pressed
     */
    void keyRepeatedImpl(int keyCode) {
        keyPressedImpl(keyCode);
    }

    // SYNC NOTE: These two empty methods should include a
    // synchronized implementation (similar to keyPressedImpl())
    // when/if they are needed
    /**
     * Handle a key released event from the keyboard
     *
     * @param keyCode The key that was released
     */
    void keyReleasedImpl(int keyCode) { }

    /**
     * Handle a key typed event from the keyboard
     *
     * @param keyCode The key that was typed
     */
    void keyTypedImpl(int keyCode) { }

    /**
     * Notify this Screen it is being hidden
     *
     * @param d The Display hiding this Screen
     */
    void hideNotifyImpl(Display d) {
        super.hideNotifyImpl(d);

        if (!visible) {
            return;
        }

        visible = false;
        stopTicker();
    }

    /**
     * Notify this Screen it is being shown on the display
     *
     * @param d The Display showing this Screen
     */
    void showNotifyImpl(Display d) {
        super.showNotifyImpl(d);

        if (visible) {
            return;
        }

        if (layoutDoneOnce) {
            Displayable currentCopy = d.getCurrent();
            // We only want to initialize the highlight if we are
            // not returning from another screen and we are not
            // returning from an editor
            if (currentCopy != this && (currentCopy instanceof Screen) &&
                    !((Screen)currentCopy).isEditScreen(this)) {
                viewPortY = initHilight(viewPortY, viewPortHeight);
            }
        } else {
            doLayout();
        }

        visible = true;
        startTicker();

        repaint(0, 0, Display.WIDTH, Display.HEIGHT);
        setVerticalScroll();
    }

    /** The image for this Screen */
    private static final Image TITLE_IMG;

    static {
        /** The data to construct the image for this Screen */
        byte title_data[] = {
            (byte)0x89, (byte)0x50, (byte)0x4e, (byte)0x47,
            (byte)0x0d, (byte)0x0a, (byte)0x1a, (byte)0x0a,
            (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x0d,
            (byte)0x49, (byte)0x48, (byte)0x44, (byte)0x52,
            (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x60,
            (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x02,
            (byte)0x02, (byte)0x03, (byte)0x00, (byte)0x00,
            (byte)0x00, (byte)0xda, (byte)0xca, (byte)0x31,
            (byte)0xfa, (byte)0x00, (byte)0x00, (byte)0x00,
            (byte)0x06, (byte)0x50, (byte)0x4c, (byte)0x54,
            (byte)0x45, (byte)0xbb, (byte)0xbb, (byte)0xbb,
            (byte)0x6a, (byte)0x6a, (byte)0x6a, (byte)0x0f,
            (byte)0x3b, (byte)0xef, (byte)0x7a, (byte)0x00,
            (byte)0x00, (byte)0x00, (byte)0x10, (byte)0x49,
            (byte)0x44, (byte)0x41, (byte)0x54, (byte)0x78,
            (byte)0xda, (byte)0x63, (byte)0x60, (byte)0xc0,
            (byte)0x01, (byte)0x18, (byte)0x43, (byte)0x71,
            (byte)0x48, (byte)0x00, (byte)0x00, (byte)0x08,
            (byte)0x43, (byte)0x00, (byte)0x57, (byte)0xe8,
            (byte)0x2f, (byte)0x93, (byte)0x38, (byte)0x00,
            (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x49,
            (byte)0x45, (byte)0x4e, (byte)0x44, (byte)0xae,
            (byte)0x42, (byte)0x60, (byte)0x82,
        };

        TITLE_IMG = Image.createImage(title_data, 0, title_data.length);
        tickerTimer = new Timer();

        // NOTE: The width of the default TITLE_IMG is
        // 96 pixels and the height is 2 pixels
    }

    /**
     * A special helper class to repaint the Ticker of this Screen
     * if one exists
     */
    private class TickerPainter extends TimerTask {
        /**
         * Repaint the ticker area of this Screen
         */
        public final void run() {
            synchronized (Display.LCDUILock) {
                // Repaint with a clip of just the text message
                Screen.this.repaint(0, 2, Display.WIDTH, CONTENT_HEIGHT);
            }
        }
    }

    /**
     * Start the ticker running.
     * Here we ensure that a tickerPainter can only ever be scheduled
     * if the Screen is visible and the ticker is visible. The visibility
     * of the ticker (tickerIsVisible) is only determined in doLayout()
     */
    private void startTicker() {
        if (visible && tickerIsVisible) {
            tickerPainter = new TickerPainter();
            tickerTimer.schedule(tickerPainter, 0, Ticker.TICK_RATE);
        }
    }

    /**
     * Stop the ticker from running
     * Here we only have to check for the existence of a tickerPainter
     */
    private void stopTicker() {
        if (tickerPainter != null) {
            tickerPainter.cancel();
            tickerPainter = null;
        }
    }

    /**
     * Get the title of this Screen
     *
     * @return String The title of this Screen
     */
    private String getTitleImpl() {
        return (screenTitle == null ?
            null : ((StringLayout)screenTitle).getString());
    }

}


