/*
 * @(#)TaGLayout.java	1.85 01/08/21
 * Copyright (c) 1999-2001 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR tITS DERIVATIVES.
 */
package javax.microedition.lcdui;

/**
 * This class allows to group together multiple StringItems and ImageItems
 * (or in other words items which isGroupable() returns true.
 *
 *  Here is current LAYOUT POLICY:
 * 
 * Interactive items with null, empty or non-null label)
 * are always laid out on a new line. Interactive items are:
 * TextField, DateField, Gauge, ChoiceGroup
 *      
 * StringItems and ImageItems with NON-NULL and NOT EMPTY labels 
 * are laid out on a new line as well.
 * 
 * \n is treated as a line break.
 * 
 * ImageItem with NULL or EMPTY label is attempted to be placed
 * next to the content of the previous StringItem or ImageItem.
 * ImageItem is placed at the beginning of next line if one of the
 * following is true:
 * 1) its image is too wide to fit on this line
 * 2) that is the first item in the form
 * 3) its layout has LAYOUT_NEWLINE_BEFORE set
 * 4) there is an ImageItem before this one and its layout 
 *    has LAYOUT_NEWLINE_AFTER set
 *    
 * StringItem's text is attempted to be placed next to its label
 * (if label is NON-NULL and NOT EMPTY) or next to the content of
 * the previous StringItem (text) or ImageItem (image)
 * (if label is NULL or EMPTY).
 * - Text is placed at the beginning of the next line if one of the
 *   following is true:
 *   1) that is the first item in the form
 *   2) there is an ImageItem before this StringItem and its layout
 *      has LAYOUT_NEWLINE_AFTER set
 *   3) it is too wide to fit on this line and it cannot be broken
 *      into words (there are no spaces within text).
 * - If text is too wide to fit on this line and it can be broken
 *   into words then as many words as possible are put on this line
 *   and then as many words as possible go on the next line, etc.
 *   
 *   NOTE: There is no word concatenation with previous labels and texts.
 *   Label and Text part of StringItem imply implicit word break.
 * 
 * StringItems and ImageItems that are placed next to each other 
 * in a Form are scrolled together as a unit. 
 *
 * Form is also the one that actually controls the grouping and 
 * asks TaGLayout to insert/set/delete a groupable Item in the
 * corresponding TaGLayout. While TaGLayout makes is responsible to
 * the actual "contract" between groupable Item and Form 
 * (item.layout[0] points to correct TaGLayout while TaGLayout has 
 * item in its list of items).
 */
class TaGLayout implements Layout {

    /** The size to grow the internal Item array by when necessary */
    private static final int GROW_SIZE = 4;
    /** The set of Items in this layout */
    private Item items[];
    /** The number of Items in the set */
    private int  numItems; // = 0;

    /**
     * This array holds information about line wrapping.
     * A lot of information is encoded in that array.
     * Each physical line on the screen has 2 entries in the
     * lineBreaks array to represent a line break:
     * 1) index of an item in the items array multiplied by 2 
     *    if that is a label part of the item, or
     *    index of an item in the items array multiplied byt 2 plus 1
     *    if that is the content part of the item.
     *    Content could be String for StringItem or Image for ImageItem.
     * 2) offset inside that part (label or content) of the item
     *
     * As you can see offset could be in a label or in the content.
     * Even or oddness of the first element in the pair determines 
     * if offset applies to the content or the label.
     * 
     * For example,
     *
     *      Form f = new Form();
     *      f.append(new StringItem("Label1: ", 
     *                              "Here is some text "));
     *      f.append("And more text");
     *      f.append(new StringItem("Label2: ", "The end"); 
     *
     * Suppose on the screen it would look like:
     *
     *
     *      --------------------
     *      Label1: Here is some    <--- only item 0 is on that line
     *      text And more text      <--- part of item 0 and all of item 1
     *      Label2: The end         <--- all of item 2
     *
     * When layout is done (setWidth() is called on created
     * TaGLayout for 3 StringItems)
     * lineBreaks array will look the following way
     * (note that line breaks represent first char (or image part) on
     * next line or in other word until where chars or image appear 
     * still on this line):
     * 
     * for line 0:
     *   lineBreaks[0] = 1      <--- odd => content; 1/2=0=>index of item = 0
     *   lineBreaks[1] = 13     <--- offset inside content (12th char is last
     *                               on that line)
     *
     * for line 1:
     *   lineBreaks[2] = 4      <--- even => label; 4/2=2=>index of item = 2 
     *   lineBreaks[3] = 0      <--- offset inside label 
     *                               (next line starts at the beg. of item 2)
     *
     * for line 2:
     *   lineBreaks[4] = 6      <--- the last line break points to the
     *   lineBreaks[5] = 0           number of items and to 0 offset
     *
     */
    private short[] lineBreaks = {0, 0};
    /** The number of lines in this layout */
    private int numLines;  // = 0;
    /** The width of this layout */
    private int width = -1; // -1 when layout was not done yet
    /** The height of this layout */
    private int height;     // = 0;
    /** The minimum height of this layout */
    private int minHeight = -1;
    /** The maximum height of this layout */
    private int maxHeight = -1;
    /** The Font used to render this layout */
    private Font f;
    /** The height (in pixels) of one line of this layout */
    private int lineHeight;

    /**
     * Creates a new, empty TaGLayout object.
     *
     * @param font The Font to use when rendering
     */
    TaGLayout(Font font) {
        f = font;
        lineHeight = font.getHeight();
        items = new Item[GROW_SIZE];
    }

    /**
     * Returns the number of elements contained in the layout
     *
     * @return int The number of Items in this layout
     */
    int getSize() { 
        return numItems;
    }

    /**
     * Sets the slot where oldItem is to contain the given item, 
     * replacing whatever was previously there.  
     * If <code> item </code> is null, 
     * a null element is placed into this slot instead of an item.
     *
     * @param oldItem the Item to replace
     * @param item the new Item.
     * @return delta of the height that was changed
     */
    int set(Item oldItem, Item item) {
            
        int index = get(oldItem);
        item.layouts[0] = this;
        items[index] = item;        

        return updateContent(index);            
    }
    
    /**
     * Inserts an item  just prior to the indicated item.  All
     * subsequent elements are pushed to higher positions, 
     * and the element count increases by one.
     * 
     * @param nextItem indicates the slot at which a new item is to be 
     *        inserted; if nextItem is null item will be appended.
     * @param item The new Item to insert
     * @return delta of the height that was changed
     */
    int insert(Item nextItem, Item item) {

        int index = nextItem == null ? numItems : get(nextItem);

        if (numItems == items.length) {
            Item[] newItems = new Item[items.length + GROW_SIZE];
            System.arraycopy(items, 0, newItems, 0, index);
            System.arraycopy(items, index, newItems, index + 1, 
                             numItems - index);
            items = newItems;
            
        } else {
            System.arraycopy(items, index, items, index + 1, 
                             numItems - index);
        }

        item.layouts[0] = this;
        items[index] = item;

        numItems++;

        return updateContent(index);
    }
    
    
    /**
     * Deletes the element at the indicated position.  Elements beyond this
     * position are moved down to a lower position, and the element count
     * decreases by one.
     *
     * @param item The Item to be deleted.
     * @return delta of the height that was changed
     */
    int delete(Item item) {

        item.layouts[0] = null;

        int index = get(item);
        if (numItems > 1 && index != numItems - 1) {
            System.arraycopy(items, index + 1, 
                             items, index, numItems - index - 1);
        }

        numItems--;
        items[numItems] = null;

        return updateContent(index);
    }
    
    /**
     * Returns the index of the item within the layout. 
     *
     * @param item which index is to be retrieved.
     * @return int The index of the Item
     */
    int get(Item item) {

        for (int i = 0; i < numItems; i++) {
            if (items[i] == item) {
                return i;
            }
        }
        
        return -1;
    }
    
    /**
     * Paints layout content to the provided graphics 
     * at the specified location.
     *
     * @param g graphics where TaGLayout should be painted.
     * @param eraseBgrnd If True, erase the background before painting
     * @param inverted If True, render with the colors inverted
     */
    public void paint(Graphics g, boolean eraseBgrnd, boolean inverted) {
        if (numLines <= 0) {
            return;
        } 

        int from = 0;
        int y = g.getClipY();

        // if the start of the clip rect is outside of this layout 
        // object, no painting is needed.
        if (y > height) {
            return;
        }
        int yEnd = y + g.getClipHeight();

        // if the height of the layout to be painted is less than the
        // end of the clip rect, then paint up till the height only
        if (yEnd > height) {
          yEnd = height;
        }

        if (y < 0) {
            y = 0;
        } else if (y > 0) {
            from = y/lineHeight;
            if (from >= numLines) {
                return;
            }
            y = from * lineHeight;
            if (y > yEnd) {
                return;
            }
        }

        if (eraseBgrnd) {
            g.setColor(inverted ? Display.BG_H_COLOR : Display.ERASE_COLOR);
            g.fillRect(0, y, width, yEnd);
        }

        g.setFont(f);
        g.setColor(inverted ? Display.FG_H_COLOR : Display.FG_COLOR);

        int nextIndex  = 0;
        int nextOffset = 0;
        int next = 0;

        if (from > 0) {
            next       = 2*(from-1);
            nextIndex  = lineBreaks[next++];
            nextOffset = lineBreaks[next++];
        }

        // if the first line contains an image that might
        // be offset from the left we should start drawing 
        // from the first line of that image. Otherwise
        // we do not know by how much to offset from the left
        if ((items[nextIndex/2] instanceof ImageItem) && 
            (nextIndex != 0) && ((nextIndex % 2) != 0)) {
            y -= nextOffset;
            from -= nextOffset/lineHeight;

            if (from == 0) {
                next = nextIndex = nextOffset = 0;
            } else {
                next = 2*(from-1);
                nextIndex  = lineBreaks[next++];
                nextOffset = lineBreaks[next++];
            }
        }

        int xOffset = 0;

        // to compensate for the fact that in the loop the first thing 
        // that will be done - move to a new line (which includes 
        // addition of lineHeight 
        y -= lineHeight;

        int j = nextIndex;
        int offset = nextOffset;

        while(j < 2*numItems) {

            // move to the next item if the end of current item was painted
            if (j < nextIndex) {
                j++;
                
                if (j >= 2*numItems) {
                    break;
                }

                offset = 0;
            }
            
            // move to next line if the end of current line was reached
            if (j == nextIndex && offset == nextOffset) {
                
                if (y > yEnd || next > 2*numLines) {
                    break;
                }

                xOffset = 0;
                y += lineHeight;

                nextIndex = lineBreaks[next++];
                nextOffset = lineBreaks[next++];
            }
            
            Item item = items[j/2];
            Object obj = null;

            // Get the next object that has to be painted;
            // even numbers correspond to item's labels
            // while odd numbers are Item's content (text or image)
            if ((j % 2) == 0) {
                obj = item.getLabel();
            } else {                   
                if (item instanceof StringItem) {
                    obj = ((StringItem)item).getText();
                } else if (item instanceof ImageItem) {
                    obj = ((ImageItem)item).getImage();
                }
            }

            if (obj == null) continue;
            
            // Next object is an Image
            if (obj instanceof Image) {

                Image img    = (Image)obj;
                int imgWidth  = img.getWidth();
                int imgHeight = img.getHeight();
                int layout    = ((ImageItem)item).getLayout();

                int rowsInImage = (imgHeight - offset + lineHeight - 1) /
                                  lineHeight;

                if (((layout & ImageItem.LAYOUT_NEWLINE_BEFORE) != 0) &&
                    ((layout & ImageItem.LAYOUT_NEWLINE_AFTER) != 0)) {
                    layout &= ImageItem.LAYOUT_CENTER;
                    if (layout == ImageItem.LAYOUT_CENTER) {
                        xOffset += Math.max(0, (width - xOffset - imgWidth)/2);
                    } else if (layout == ImageItem.LAYOUT_RIGHT) {
                        xOffset += Math.max(0, width - xOffset - imgWidth);
                    }
                }

                int vPadding = ((imgHeight+lineHeight-1)/
                                lineHeight*lineHeight-imgHeight)/2;

                g.drawImage(img, xOffset, y - offset + vPadding, 
                            Graphics.TOP | Graphics.LEFT);

                if (rowsInImage > 1) {
                    next += 2*(rowsInImage-2);

                    y += (rowsInImage-1)*lineHeight;

                    nextIndex = lineBreaks[next++];
                    nextOffset = lineBreaks[next++];
                }

                xOffset += img.getWidth();
                offset = 0;

            } else if (obj instanceof String) {
                // Next object is a String

                String str = (String)obj;                
                int subStrLen = 0;
    
                // if next line starts with a different Item
                // paint till the end of that item's label or text,
                // otherwise paint till the next offset inside that
                // item
                if (next < 2*numLines && j == nextIndex) {
                    subStrLen = nextOffset - offset;
                } else {
                    subStrLen = str.length() - offset;
                }
                
                // paint only if there are are characters other 
                // than '\n'
                if (subStrLen > 0 && str.charAt(offset) != '\n') {
                    int subStrLenWithLf = subStrLen; 
                    // dont' paint the last character '\n' and '\r'
                    // on this line
                    for (int last = offset+subStrLen-1; last >= offset;
                        last--) {
                        if (str.charAt(last) == '\n' ||
                            str.charAt(last) == '\r') {
                            subStrLenWithLf--;
                        } else {
                            break;
                        }
                    }
                    
                    if (subStrLenWithLf > 0) {
                        g.drawSubstring(str, offset, subStrLenWithLf,
                                        xOffset, y, 
                                        Graphics.TOP | Graphics.LEFT);
                        
                        xOffset += f.substringWidth(str, offset, 
                                                    subStrLenWithLf);
                    }
                }
                
                offset += subStrLen;
            }
        }
    }

    /**
     * Get the line height of this layout
     *
     * @return int The height (in pixels) of one line of this layout
     */
    public int getLineHeight() {
        return lineHeight;
    }

    /**
     * Get the width of this layout
     *
     * @return int The width of this layout
     */
    public int getWidth() {
        return width;
    }

    /**
     * Get the height of this layout
     *
     * @return int The height of this layout
     */
    public int getHeight() {
        return height;
    }

    /**
     * Set the width of this layout
     *
     * @param w The new width of this layout
     * @return int The height of this layout after setting the new width
     */
    public int setWidth(int w) {

        if (width != w && w >= 0) {
            width = w;

            if (w == 0) {
                numLines = 0;
            } else if (w > 0) {                
                updateLineBreaks(w);
            }
            
            height = numLines*lineHeight;
            
            if (minHeight != -1 && height < minHeight) {
                height = minHeight;
            }
            
            if (maxHeight != -1 && height > maxHeight) {
                height = maxHeight;
            }
        }

        return height;
    }

    /**
     * Set the minimum height for this layout
     *
     * @param h The new minimum height for this layout
     */
    void setMinHeight(int h) {
        if (h < 0) {
            h = 0;
        }
        minHeight = h;
        if (height < minHeight) {
            height = minHeight;
        }
    }

    /**
     * Set the maximum height for this layout
     *
     * @param h The new maximum height for this layout
     */
    public void setMaxHeight(int h) {
        if (h < 0) {
            h = 0;
        }
        maxHeight = h;
        if (height > maxHeight) {
            height = maxHeight;
        }
    }

    /**
     * Signal this layout that an item has changed
     *
     * @param item The Item which has changed
     * @return int The new height based on the Item change
     */
    int itemContentChanged(Item item) {
        return updateContent(get(item));
    }

    /**
     * Add a line break in this layout
     *
     * @param index The index to insert the break
     * @param offset
     * @param last
     * @return int
     */
    private int addBreak(short index, short offset, int last) {
        if (last+1 < lineBreaks.length) {
            short newBreaks[] = new short[lineBreaks.length + GROW_SIZE];
            System.arraycopy(lineBreaks, 0, newBreaks, 0, lineBreaks.length);
            lineBreaks = newBreaks;
        }
        lineBreaks[last++] = index;
        lineBreaks[last++] = offset;

        return last;
    }

    /**
     * Update the line breaks of this layout with the given width
     *
     * @param width The new width to use to calculate line breaks
     */
    private void updateLineBreaks(int width) {

        if (numItems == 0 || width <= 0) {
            height = maxHeight > 0 ? maxHeight : 0;
            numLines = lineBreaks[0] = lineBreaks[1] = 0;
            return;
        } 

        numLines = 0;

        int curWidth   = 0;
        int lastIndex  = 0;

        for (int i = 0; i < 2*numItems; ++i) {

            Item item = items[i/2];
            Object obj = null;

            // Get the next object that has to be laid out;
            // even numbers correspond to item's labels
            // while odd numbers are Item's content (text or image)
            if ((i % 2) == 0) {
                obj = item.getLabel();
                
                // Insert a new line if Label does not start
                // on a new line
                if (obj != null && !((String)obj).equals("") 
                    && curWidth != 0) {

                    lastIndex = addBreak((short)i, (short)0, lastIndex);
                    numLines++;
                    
                    curWidth = 0; 
                }

            } else {                  
                if (item instanceof StringItem) {
                    obj = ((StringItem)item).getText();
                } else if (item instanceof ImageItem) {
                    obj = ((ImageItem)item).getImage();
                }
            }
            if (obj == null) {
                continue;
            }

            // Next object to be laid out is an Image
            if (obj instanceof Image) {
                Image img      = (Image)obj;
                int imgWidth   = img.getWidth();
                int imgHeight  = img.getHeight();
                int layout     = ((ImageItem)item).getLayout();

                // new line before image should be added if image does not
                // fit on this line or if LAYOUT_NEWLINE_BEFORE was set
                // in Item's layout
                boolean addNewLineBefore = 
                    (((layout & ImageItem.LAYOUT_NEWLINE_BEFORE) != 0) ||
                         (curWidth + imgWidth > width));

                // new line before image should be added 
                // if LAYOUT_NEWLINE_AFTER was set in Item's layout and this
                // is not the last image on this layout
                boolean addNewLineAfter =
                    (((layout & ImageItem.LAYOUT_NEWLINE_AFTER) != 0) && 
                     (i < 2*numItems-1));
                if (!addNewLineBefore && !addNewLineAfter) {
                    curWidth += imgWidth;
                }

                // add line breaks if new line has to be added before, or
                // after, or if image is taller than the lineHeight
                if (addNewLineBefore || addNewLineAfter ||
                    imgHeight > lineHeight) {
                    
                    int heightOffset = 0;
                    if ((addNewLineBefore && curWidth == 0) ||
                        !addNewLineBefore) {
                        heightOffset = lineHeight;
                    }

                    for (; heightOffset < imgHeight;
                         heightOffset += lineHeight) {
                        lastIndex = addBreak((short)i, (short)heightOffset,
                                             lastIndex);
                        numLines++;
                    }

                    if (addNewLineAfter) {
                        lastIndex = addBreak((short)(i+1), (short)0,
                                             lastIndex);
                        
                        numLines++;
                        curWidth = 0;

                    } else if (addNewLineBefore) {

                        curWidth = imgWidth;
                    }
                }

            } else if (obj instanceof String) { // Text

                // Next object to be laid out is an Image
                String str = (String)obj;
                int strLength = str.length();

                if (strLength == 0) continue;

                int wordWidth  = 0;
                int spaceIndex = -1;

                /*
                int curStart = fromLine == numLines-1 ? 
                               lineBreaks[lastInfo-1] : 0;
                               */
                for (int j = 0; j < strLength; j++) {

                    char ch = str.charAt(j);
                    int charWidth = f.charWidth(ch);

                    // add a line break if next character is '\n';
                    // if this is the last character in this string
                    // then first element on the next line will be
                    // with a 0 offset, otherwise it will be next character
                    if (ch == '\n') {
                        if (j == strLength-1) {
                            lastIndex = addBreak((short)(i+1), (short)0, 
                                                 lastIndex);
                        } else {
                            lastIndex = addBreak((short)i, (short)(j+1),
                                                 lastIndex);
                        }
                        numLines++;
                        curWidth = wordWidth = 0;
                        spaceIndex = -1;
                        
                        continue;
                    }
                    
                    if (ch == ' ') {
                        spaceIndex = j;
                        wordWidth = 0;
                    } else {
                        wordWidth += charWidth;
                    }
                    curWidth += charWidth;

                    // add a line break at the last recoreded word break
                    // if next character does not fit on this line
                    if (curWidth > width) { 
                        
                        if (wordWidth > width) { // long word
                            lastIndex = addBreak((short)i, (short)j, 
                                                 lastIndex);
                            curWidth = wordWidth = charWidth;
                        } else {              // do word wrap
                            lastIndex = addBreak((short)i, 
                                                 (short)(spaceIndex+1),
                                                 lastIndex);
                            if (j == spaceIndex) {
                                wordWidth = 0;
                            }
                            curWidth = wordWidth;
                        }
                        numLines++;
                        spaceIndex = -1;
                    }
                }
            }
        }
        
        // always add the last break (for the last line)
        addBreak((short)(2*numItems), (short)0, lastIndex);
        numLines++;

        height = numLines * lineHeight;
        if (minHeight != -1 && height < minHeight) {
            height = minHeight;
        }

        if (maxHeight != -1 && height > maxHeight) {
            height = maxHeight;
        }

        // printLines();
    }

    /**
     * Update the content of this layout
     *
     * @param index The index of the Item to update
     * @return int The delta height change caused by the update
     */
    private int updateContent(int index) {

        // update lines array if it was set up already
        if (width <= 0) {
            return 0;
        }

        int oldHeight = height;

        if (index == 0 || numLines == 0) {

            updateLineBreaks(width);
            return height - oldHeight;

        }

        for (int line = 0; line < numLines; line++) {
            int nextIndex = line+1 < numLines ? 
                lineBreaks[line+1] : numItems;
            if (index > nextIndex) {
                if (line > 0) line--;
            }
            
            if (index >= nextIndex) {
                if (line > 0) line--;
                updateLineBreaks(/* line, */width);
                return height - oldHeight;
            }
        }

        updateLineBreaks(/* numLines-1, */ width);
        return height - oldHeight;
    }
    /*
    void printLines() {
        
        int next = 0;
        int nextIndex  = 0;
        int nextOffset = 0;

        System.err.println("\nPrintlines numLines=" +numLines+
                           " lineBreaks.length=" +lineBreaks.length+ 
                           " numItems=" +numItems+ 
                           " items.length=" +items.length);

        for (int j = nextIndex, offset = nextOffset; j < 2*numItems; ) {

            // move to the next subItem if the end of current item was painted
            if (j < nextIndex) {
                j++;
                if (j >= 2*numItems) {
                    break;
                }
                
                offset = 0;
            }

            // move to next line if the end of current line was reached
            if (j == nextIndex && offset == nextOffset) {
                if (next > 2*numLines) {
                    break;
                }

                nextIndex = lineBreaks[next++];
                nextOffset = lineBreaks[next++];
                
                System.err.println("Line: " +(next/2-1)+ 
                                   " index=" +(j/2)+
                                   " offset=" +offset+
                                   " nextIndex=" +nextIndex+
                                   " nextOffset=" +nextOffset);
            }
            
            Item item = items[j/2];            
            boolean isLabel = ((j%2) == 0);

            Object obj = null;

            if (isLabel) {
                obj = item.getLabel();
            } else {                  
                if (item instanceof StringItem) {
                    obj = ((StringItem)item).getText();
                } else if (item instanceof ImageItem) {
                    obj = ((ImageItem)item).getImage();
                }
            }
            
            if (obj == null) continue;

            if (obj instanceof Image) {
                Image img = (Image)obj;
                int imgHeight = img.getHeight();
                int layout = ((ImageItem)item).getLayout();
                String layoutStr = "(";
                if ((layout & ImageItem.LAYOUT_NEWLINE_BEFORE) != 0) {
                    layoutStr += "LAYOUT_NEWLINE_BEFORE ";
                }
                if ((layout & ImageItem.LAYOUT_NEWLINE_AFTER) != 0) {
                    layoutStr += "LAYOUT_NEWLINE_AFTER ";
                }                
                if ((layout & ImageItem.LAYOUT_DEFAULT) != 0) {
                    layoutStr += "LAYOUT_DEFAULT ";
                }
                if ((layout & ImageItem.LAYOUT_CENTER) != 0) {
                    layoutStr += "LAYOUT_CENTER ";
                }
                if ((layout & ImageItem.LAYOUT_LEFT) != 0) {
                    layoutStr += "LAYOUT_LEFT ";
                }
                if ((layout & ImageItem.LAYOUT_RIGHT) != 0) {
                    layoutStr += "LAYOUT_RIGHT ";
                }
                layoutStr += ")";
                System.err.println("\t Image: layout=" +layout+
                                   " j=" +j+ " index=" +(j/2)+ 
                                   " offset=" +offset+
                                   " layout=" +layoutStr);
                
                offset += lineHeight;

            } else if (obj instanceof String) {
                String str = (String)obj;
                int strLen = str.length();
                
                int endIndex = (j == nextIndex ? nextOffset : strLen); 

                System.err.print("\t" + (isLabel ? "Label" : "Text"));
                System.err.println(" (j=" +j+ " index=" +(j/2)+
                                   " offset=" +offset+ 
                                   " endIndex=" +endIndex+")='" +
                                   str.substring(offset, endIndex)+ 
                                   "'");
                offset = endIndex;
            }
        }
    } */
}
