/*
 * @(#)TextBasic.java	1.37 01/09/17 
 * Copyright (c) 1999-2001 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

package javax.microedition.lcdui;

import com.sun.midp.lcdui.InputModeHelper;
import com.sun.midp.lcdui.InputModeClient;

/**
 * The basic support for a text component. This class is used
 * by TextField, TextBox, and TextEditor
 */
class TextBasic {

    /** Buffer holding the data */
    char buffer[];
    /** The number of characters in the buffer */
    int numChars;    // = 0;
    /** The current position in the buffer */
    int curPos;      // = 0;
    /** The constraints applied to this text object */
    int constraints; // = TextField.ANY;
    /** The policy applied to this text object */
    Policy policy;
    /** The default mode of this text object */
    String defaultMode;
    /** The allowed modes of this text object */
    String allowedMode[];

    /**
     * Construct a new TextBasic
     *
     * @param constraints The constraints to apply to this text object
     * @param maxSize The maximum size of this text object
     * @param text The initial text to set in this text object
     */
    TextBasic(int constraints, int maxSize, String text) {
        if (text == null) {
            buffer = new char[0];
        } else {
            buffer = text.toCharArray();
            numChars = curPos = text.length();
        }

        // getPolicy throws an IllegalArgumentException
        // if the constraints parameter is invalid
        policy = Policy.getPolicy(constraints);
        this.constraints = constraints;
        if (!policy.validateContent(buffer, 0, numChars)) {
            numChars = curPos = 0;
            throw new IllegalArgumentException();
        }

        if (numChars > maxSize) {
            throw new IllegalArgumentException();
        }
        setMaxSize(maxSize);
    }


    // util methods

    /**
     * Set the constraints of this text object
     *
     * @param constraints The new constraints
     * @return int The delta height caused by the constraint change
     */
    int setConstraints(int constraints)  {
        if (constraints == this.constraints) {
            return 0;
        }
        this.constraints = constraints;

        // policy is never null, as its set initially in the constructor
        int w = policy.width;
        int h = policy.height;

        // getPolicy throws an IllegalArgumentException
        // if the constraints parameter is invalid
        policy = Policy.getPolicy(constraints);

        if (!policy.validateContent(buffer, 0, numChars)) {
            numChars = curPos = 0;
        }

        // If the old policy hadn't been layed out yet, simply return
        if (w == -1) {
            return 0;
        }

        // return the delta height between the old and new policies
        return h - policy.setWidth(w, buffer, numChars, curPos);
    }

    /**
     * Set the maximum size of this text object (in chars)
     *
     * @param maxSize The maximum number of characters this text object
     *                  can hold
     * @return int The delta height change caused by this size change
     */
    int setMaxSize(int maxSize) { 
        if (maxSize <= 0) {
            throw new IllegalArgumentException();
        }

        maxSize = policy.constrainedSize(maxSize, buffer);
        if (maxSize != buffer.length) {
            char[] newbuf = new char[maxSize];
            if (maxSize < numChars) {
                numChars = maxSize;
                if (curPos > numChars) {
                    curPos = numChars;
                }
            }
            System.arraycopy(buffer, 0, newbuf, 0, numChars);
            buffer = newbuf;
        }

        return policy.contentChanged(buffer, numChars, 0, curPos);
    }	
    
    /**
     * Insert a sub string into a given position
     *
     * @param src The String to insert
     * @param position The index to insert the string
     * @return int The delta height change caused by this insertion
     */
    int insert(String src, int position) {
        if (position < 0) {
            position = 0;
        } else if (position > numChars) {
            position = numChars;
        }
        // If src == null,
        // NullPointerException will be thrown in src.length() by VM
        int length = src.length();
        if (numChars + length > buffer.length) {
            throw new IllegalArgumentException();
        }

        System.arraycopy(buffer, position, buffer, position + length,
                         numChars - position);
        src.getChars(0, length, buffer, position);

        numChars += length;

        if (!policy.validateContent(buffer, position, length)) {
            numChars -= length;
            System.arraycopy(buffer, position + length, buffer, position,
                             numChars - position);
            throw new IllegalArgumentException();
        }

        if (position <= curPos) {
            curPos += length;
        }
        return policy.contentChanged(buffer, numChars, position, curPos);
    }

    /**
     * Insert characters into this text object
     *
     * @param data The array of characters to insert
     * @param offset The start point of the array to insert
     * @param length The number of characters to insert out of the 'data'
     *               array
     * @param position The position in this text object to insert the
     *                  new characters
     * @return int The delta height change caused by this insertion
     */
    int insert(char[] data, int offset, int length, int position) {

        if (position < 0) {
            position = 0;
        } else if (position > numChars) {
            position = numChars;
        }

        if (numChars + length > buffer.length) {
            throw new IllegalArgumentException();
        }
        // If data == null,
        // NullPointerException will be thrown in  data.length by VM
        if ((offset < 0) || (length < 0) || (offset + length > data.length)) {
            throw new ArrayIndexOutOfBoundsException();
        }

        System.arraycopy(buffer, position, buffer, position + length,
                         numChars - position);

        System.arraycopy(data, offset, buffer, position, length);

        numChars += length;

        if (!policy.validateContent(buffer, position, length)) {
            numChars -= length;
            System.arraycopy(buffer, position + length, buffer, position,
                             numChars - position);
            throw new IllegalArgumentException();
        }

        if (position <= curPos) {
            curPos += length;
        }
        return policy.contentChanged(buffer, numChars, position, curPos);
    }

    /**
     * Delete characters from this text object
     *
     * @param offset The offset into the buffer to start deletion
     * @param length The number of characters to delete out of the buffer
     * @return int The delta height change caused by this deletion
     */
    int deleteChars(int offset, int length) {

        if ((offset < 0 || length < 0) || (offset + length > numChars)) {
            throw new StringIndexOutOfBoundsException();
        }

        System.arraycopy(buffer, offset + length, buffer, offset,
                         numChars - offset - length);

        numChars -= length;

        if (offset < curPos) {
            curPos -= length;
        } else {
            if (curPos > numChars) {
                curPos = numChars;
            }
        }

        return policy.contentChanged(buffer, numChars, offset, curPos);
    }

    /**
     * Replace a section of characters in the buffer
     *
     * @param data The new characters
     * @param offset The offset into the buffer to start to replace
     * @param length The number of characters to replace
     * @return int The delta height change caused by this replacement
     */
    int setChars(char[] data, int offset, int length)  {
        if (data == null) {
            numChars = curPos = 0;
            return policy.contentChanged(buffer, numChars, 0, curPos);
        }

        if ((offset < 0 || length < 0) || (offset + length > data.length)) {
            throw new ArrayIndexOutOfBoundsException();
        }

        if (length > buffer.length) {
            throw new IllegalArgumentException();
        }


        if (!policy.validateContent(data, offset, length)) {
            throw new IllegalArgumentException();
        }

        System.arraycopy(data, offset, buffer, 0, length);
        numChars = length;
        curPos = numChars;
        return policy.contentChanged(buffer, numChars, 0, curPos);
    }

    /**
     * Set the string for this text object
     *
     * @param text The new text for this text object
     * @return int The delta height change caused by this set
     */
    int setString(String text) {
        // System.out.println("inside TextBasics setString for "+text);
        if (text == null) {
            numChars = 0;
            curPos = 0;
        } else {
            int length = text.length();
            if (length > buffer.length) {
                throw new IllegalArgumentException();
            }
            text.getChars(0, length, buffer, 0);

            if (!policy.validateContent(buffer, 0, length)) {
                throw new IllegalArgumentException();
            }

            numChars = length;
            curPos = length;
        }

        return policy.contentChanged(buffer, numChars, 0, curPos);
    }

    /**
     * Retrieve this text object's characters into an array
     *
     * @param data The array to fill with characters
     * @return int The number of characters placed into the array
     */
    int getChars(char[] data)  {
        if (numChars > data.length) {
            throw new ArrayIndexOutOfBoundsException();
        }

        if ((buffer != null) && (numChars > 0)) {
            System.arraycopy(buffer, 0, data, 0, numChars);
            return numChars;
        } else {
            return 0;
        }
    }

    /**
     * Traverse this text object
     *
     * @param dir The direction of traversal
     * @param top
     * @param bottom
     * @return int
     */
    int traverse(int dir, int top, int bottom) {
        // No traversal is possible if dir is not a traversal direction
        // and if numChars is 0
        if ((numChars == 0) ||
                ((dir != Canvas.UP) && (dir != Canvas.DOWN) &&
                (dir != Canvas.LEFT) && (dir != Canvas.RIGHT))) {
            return -1;
        }

        int oldCurPos = curPos;
        curPos = policy.moveCursor(dir, curPos, buffer, numChars);
        if (curPos != oldCurPos) {
            int cursorY1 = policy.cursorY;
            int cursorY2 = cursorY1 + policy.lineHeight;
            if (top <= cursorY1 && bottom >= cursorY2) {
                return 0;
            }
            int t1 = cursorY2 - bottom;
            int t2 = top - cursorY1;
            return t1 >= t2 ? t1 : t2;
        }
        return -1;
    }

    /**
     * Set the width of this text object
     * @param width The width of this object
     * @return new height after width was set
     */
    int setWidth(int width) {
        return policy.setWidth(width, buffer, numChars, curPos);
    }

    /**
     * Set the horizontal scroll indicators for this text object
     *
     * @param parent The Screen object who is the parent of this
     *                  text object
     */
    void setHorizontalScroll(Screen parent) {
        if (parent instanceof TextBox) {
            if (numChars == 0) {
                parent.setHorizontalScroll(0, 100);
            } else {
                parent.setHorizontalScroll(curPos * 100 / numChars,
                                           100 / (numChars + 1));
            }
        }
    }

    /**
     * Get the maximum width of this text object
     *
     * @param allowedWidth The maximum allowed width of this object
     * @return int The actual width of this text object
     */
    int getMaxWidth(int allowedWidth) {
        return policy.getMaxWidth(allowedWidth, buffer.length);
    }

    static {

        /**
         * A special input mode client to handle multiple input modes
         * on text components.
         */
        InputModeClient imc = new InputModeClient() {
            /**
             * Set the mode of this input client
             *
             * @param tb The TextBox this imc belongs to
             * @param defaultMode The default mode for this imc
             */
            public void setMode(TextBox tb, String defaultMode) {
                tb.sD.defaultMode = defaultMode;
            }
            /**
             * Get the mode of this input client
             *
             * @param tb The TextBox this imc belongs to
             * @return String The default mode of the TextBox
             */
            public String getMode(TextBox tb) {
                return tb.sD.defaultMode;
            }
            /**
             * Set the allowed input modes of the TextBox
             *
             * @param tb The TextBox to set the allowed modes on
             * @param allowedMode The set of allowed input modes
             */
            public void setAllowedModes(TextBox tb, String[] allowedMode) {
                tb.sD.allowedMode = allowedMode;
            }
            /**
             * Get all the allowed modes of this textbox
             *
             * @param tb The TextBox to get the allowed modes for
             * @return String[] The allowed input modes of the TextBox
             */
            public String[] getAllowedModes(TextBox tb) {
                return tb.sD.allowedMode;
            }
            /**
             * Set the input mode for the TextField to the given mode
             *
             * @param tf The TextField to set the input mode
             * @param defaultMode The input mode to set on the TextField
             */
            public void setMode(TextField tf, String defaultMode) {
                tf.sD.defaultMode = defaultMode;
            }
            /**
             * Get the mode of the given TextField
             *
             * @param tf The TextField to get the input mode of
             * @return String The input mode of the TextField
             */
            public String getMode(TextField tf) {
                return tf.sD.defaultMode;
            }
            /**
             * Set the allowed input modes of the given TextField
             *
             * @param tf The TextField to set the allowed input modes of
             * @param allowedMode An array of allowed input modes
             */
            public void setAllowedModes(TextField tf, String[] allowedMode) {
                tf.sD.allowedMode = allowedMode;
            }
            /**
             * Get the allowed input modes of the given TextField
             *
             * @param tf The TextField to get the allowed input modes of
             * @return String[] An array of allowed input modes
             */
            public String[] getAllowedModes(TextField tf) {
                return tf.sD.allowedMode;
            }
        };

        InputModeHelper.setInputModeHelper(imc);
    }
}

