/*
 * @(#)TextField.java	1.164 01/08/09
 * Copyright (c) 1999-2001 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

package javax.microedition.lcdui;

import com.sun.midp.lcdui.EventHandler;

/**
 * <P>A TextField is an editable text component that may be placed into
 * a {@link Form Form}. It can be
 * given a piece of text that is used as the initial value.</p>
 *
 * <P>A TextField has a maximum size, which is the maximum number of characters
 * that can be stored in the object at any time (its capacity). This limit is
 * enforced when the TextField instance is constructed,
 * when the user is editing text within the TextField, as well as
 * when the application program calls methods on the TextField that modify its
 * contents. The maximum size is the maximum stored capacity and is unrelated
 * to the number of characters that may be displayed at any given time.
 * The number of characters displayed and their arrangement into rows and
 * columns are determined by the device. </p>
 *
 * <p>The implementation may place a boundary on the maximum size, and the 
 * maximum size actually assigned may be smaller than the application had 
 * requested.  The value actually assigned will be reflected in the value
 * returned by {@link #getMaxSize() getMaxSize()}.  A defensively-written 
 * application should compare this value to the maximum size requested and be 
 * prepared to handle cases where they differ.</p>
 *
 * <a name="constraints"></a>
 * <h3>Input Constraints</h3>
 * 
 * <P>The TextField shares the concept of <em>input
 * constraints</em> with the {@link TextBox TextBox} object. The different
 * constraints allow the application to request that the user's input be
 * restricted in a variety of ways. The implementation is required to
 * restrict the user's input as requested by the application. For example, if
 * the application requests the NUMERIC constraint on a TextField, the
 * implementation must allow only numeric characters to be entered. </p>
 *
 * <P>The implementation is not required to do any syntactic validation
 * of the contents of the text object. Applications must be prepared
 * to perform such checking themselves.</P>
 *
 * <P>The implementation may provide special formatting for the value entered.
 * For example, a PHONENUMBER field may be separated and punctuated as
 * appropriate for the phone number conventions in use, grouping the digits
 * into country code, area code, prefix, etc. Any spaces or punctuation
 * provided are not considered part of the text field's value. For
 * example, a TextField with the PHONENUMBER constraint might display as
 * follows: <pre>
 * (408) 555-1212
 * </pre>
 * but the value of the field visible to the application would be
 * a string representing a legal phone number like "4085551212". Note that
 * in some networks a '+' prefix is part of the number and returned as a part
 * of the string.</p>
 */
public class TextField extends Item {

    /**
     * The user is allowed to enter any text.
     *
     * <P>Constant 0 is assigned to ANY.</P>
     */
    public final static int ANY = 0;

    /**
     * The user is allowed to enter an e-mail address.
     *
     * <P>Constant 1 is assigned to EMAILADDDR.</P>
     */
    public final static int EMAILADDR = 1;

    /**
     * The user is allowed to enter only an integer value. The implementation
     * must restrict the contents to consist of an optional minus sign
     * followed by an optional string of numerals.
     *
     * <P>Constant 2 is assigned to NUMERIC.</P>
     */
    public final static int NUMERIC = 2;

    /**
     * <P>The user is allowed to enter a phone number. The phone
     * number is a special
     * case, since a phone-based implementation may be linked to
     * the native phone
     * dialing application. The implementation may automatically start a phone
     * dialer application that is initialized so that pressing a single key
     * would be enough to make a call. The call must not made automatically
     * without requiring user's confirmation.  The exact set of characters
     * allowed is specific to the device and to the device's network and may
     * include non-numeric characters.</P>
     *
     * <P>Constant 3 is assigned to PHONENUMBER.</P>
     */
    public final static int PHONENUMBER = 3;

    /**
     * The user is allowed to enter a URL.
     *
     * <P>Constant 4 is assigned to URL.</P>
     */
    public final static int URL = 4;

    /**
     * <P>The text entered must be masked so that the characters typed are not
     * visible. The actual contents of the text field are not affected, but each
     * character is displayed using a mask character such as "*". The character
     * chosen as the mask character is implementation-dependent. This is useful
     * for entering confidential information such as passwords or PINs (personal
     * identification numbers). </P>
     *
     * <P>The PASSWORD modifier can be combined with other input constraints by
     * using the logical OR operator (|). However, The PASSWORD modifier is
     * nonsensical with some constraint values such as EMAILADDR, 
     * PHONENUMBER, and URL.</P>
     *
     * <P>Constant 0x10000 is assigned to PASSWORD.</P>
     */
    public final static int PASSWORD = 0x10000;

    /**
     * The mask value for determining the constraint mode. The 
     * application should
     * use the logical AND operation with a value returned by 
     * getConstraints() and
     * CONSTRAINT_MASK in order to retrieve the current constraint mode,
     * in order to remove any modifier flags such as the PASSWORD flag.
     *
     * <P>Constant 0xFFFF is assigned to CONSTRAINT_MASK.</P>
     */
    public final static int CONSTRAINT_MASK = 0xFFFF;

    /**
     * The TextEditor to handle edits to this (and all) TextFields
     */
    final static TextEditor textEditor = new TextEditor();
    /** The internal text handler */
    TextBasic sD; /* shared data */
    /** The overall width of this TextField */
    private int width;
    /** The label width of the label for this TextField */
    private int labelWidth;
    /** The width of the TextField */
    private int tfWidth;
    /** The content height of this TextField */
    private int contentHeight;
    /** The overall height of this TextField */
    private int height;
    /** A special spacer value */
    private final static int SPACER = 8;
    /** A flag indicating a horizontal layout */
    private boolean horLayout;

    /**
     * Creates a new TextField object with the given label, initial
     * contents, maximum size in characters, and constraints.
     * If the text parameter is null, the TextField is created empty.
     * The maxSize parameter must be greater than zero.
     *
     * @param label item label
     * @param text the initial contents, or null if the TextField is to be empty
     * @param maxSize the maximum capacity in characters
     * @param constraints see <a href="#constraints">input constraints</a>
     *
     * @throws IllegalArgumentException if maxSize is zero or less
     * @throws IllegalArgumentException if the value of the constraints
     * parameter is invalid
     * @throws IllegalArgumentException if text is illegal
     * for the specified constraints
     * @throws IllegalArgumentException if the length of the string exceeds
     * the requested maximum capacity
     * or the maximum capacity actually assigned
     */
    public TextField(String label, String text, int maxSize, int constraints) {
        super(label, 1);
        synchronized (Display.LCDUILock) {
            sD = new TextBasic(constraints, maxSize, text);
        }
    }

    /**
     * Gets the contents of the TextField as a string value.
     *
     * @return the current contents
     *
     * @see #setString
     */
    public String getString() {
        synchronized (Display.LCDUILock) {
                return new String(sD.buffer, 0, sD.numChars);
        }
    }

    /**
     * Sets the contents of the TextField as a string value, replacing the
     * previous contents.
     *
     * @param text the new value of the TextField, or null if the TextField is
     * to be made empty
     * @throws IllegalArgumentException if the text is illegal
     * for the current <a href="#constraints">input constraints</a>
     * @throws IllegalArgumentException if the text would exceed the current
     * maximum capacity
     *
     * @see #getString
     */
    public void setString(String text) {
        synchronized (Display.LCDUILock) {
            int deltaHeight = sD.setString(text);
            contentHeight += deltaHeight;
            height        += deltaHeight;
            contentChanged(0, 0, deltaHeight);
        }
    }

    /**
     * Copies the contents of the TextField into a character array starting at
     * index zero. Array elements beyond the characters copied are left
     * unchanged.
     *
     * @param data the character array to receive the value
     * @return the number of characters copied
     * @throws ArrayIndexOutOfBoundsException if the array is too short for the
     * contents
     * @throws NullPointerException if <code>data</code> is <code>null</code>
     *
     * @see #setChars
     */
    public int getChars(char[] data)  {
        synchronized (Display.LCDUILock) {
            return sD.getChars(data);
        }
    }

    /**
     * Sets the contents of the TextField from a character array, replacing the
     * previous contents. Characters are copied from the region of the
     * <code>data</code> array
     * starting at array index <code>offset</code> and running for
     * <code>length</code> characters.
     * If the data array is null, the TextField
     * is set to be empty and the other parameters are ignored.
     *
     * @param data the source of the character data
     * @param offset the beginning of the region of characters to copy
     * @param length the number of characters to copy
     * @throws ArrayIndexOutOfBoundsException if offset and length do not
     * specify a valid range within the data array
     * @throws IllegalArgumentException if the text is illegal
     * for the current <a href="#constraints">input constraints</a>
     * @throws IllegalArgumentException if the text would exceed the current
     * maximum capacity
     *
     * @see #getChars
     */
    public void setChars(char[] data, int offset, int length)  {
        synchronized (Display.LCDUILock) {
            int deltaHeight = sD.setChars(data, offset, length);
            contentHeight += deltaHeight;
            height        += deltaHeight;
            contentChanged(0, 0, deltaHeight);
        }
    }

    /**
     * <p>Inserts a string into the contents of the TextField.  The string is
     * inserted just prior to the character indicated by the
     * <code>position</code> parameter, where zero specifies the first
     * character of the contents of the TextField.  If <code>position</code> is
     * less than or equal to zero, the insertion occurs at the beginning of
     * the contents, thus effecting a prepend operation.  If
     * <code>position</code> is greater than or equal to the current size of
     * the contents, the insertion occurs immediately after the end of the
     * contents, thus effecting an append operation.  For example,
     * <code>text.insert(s, text.size())</code> always appends the string 
     * s to the current contents.</p>
     * 
     * <p>The current size of the contents is increased by the number of
     * inserted characters. The resulting string must fit within the current
     * maximum capacity. </p>
     *
     * <p>If the application needs to simulate typing of characters it can
     * determining the location of the current insertion point ("caret")
     * using the with {@link #getCaretPosition() getCaretPosition()} method.  
     * For example,
     * <code>text.insert(s, text.getCaretPosition())</code> inserts the string 
     * s at the current caret position.</p>
     *
     * @param src the String to be inserted
     * @param position the position at which insertion is to occur
     * 
     * @throws IllegalArgumentException if the resulting contents are illegal
     * for the current <a href="TextField.html#constraints">input
     * constraints</a>
     * @throws IllegalArgumentException if the insertion would exceed 
     * the current maximum capacity
     * @throws NullPointerException if <code>src</code> is <code>null</code>
     */
    public void insert(String src, int position) {
        synchronized (Display.LCDUILock) {
            int deltaHeight = sD.insert(src, position);
            contentHeight += deltaHeight;
            height        += deltaHeight;
            contentChanged(0, 0, deltaHeight);
        }
    }

    /**
     * <p>Inserts a subrange of an array of characters into the contents of
     * the TextField.  The offset and length parameters indicate the subrange
     * of the data array to be used for insertion. Behavior is otherwise
     * identical to {@link #insert(String, int) insert(String, int)}. </p>
     * 
     * @param data the source of the character data
     * @param offset the beginning of the region of characters to copy
     * @param length the number of characters to copy
     * @param position the position at which insertion is to occur
     * 
     * @throws ArrayIndexOutOfBoundsException if offset and length do not
     * specify a valid range within the data array
     * @throws IllegalArgumentException if the resulting contents are illegal
     * for the current <a href="TextField.html#constraints">input
     * constraints</a>
     * @throws IllegalArgumentException if the insertion would exceed the
     * current maximum capacity
     * @throws NullPointerException if <code>data</code> is <code>null</code>
     */
    public void insert(char[] data, int offset, int length, int position)  {
        synchronized (Display.LCDUILock) {
            sD.insert(data, offset, length, position);
        }
    }

    /**
     * <p>Deletes characters from the TextField.</p>
     *
     * @param offset the beginning of the region to be deleted
     * @param length the number of characters to be deleted
     *
     * @throws StringIndexOutOfBoundsException if offset and length do not
     * specify a valid range within the contents of the TextField
     */
    public void delete(int offset, int length) {
        synchronized (Display.LCDUILock) {
            int deltaHeight = sD.deleteChars(offset, length);
            contentHeight += deltaHeight;
            height        += deltaHeight;
            contentChanged(0, 0, deltaHeight);
        }
    }

    /**
     * Returns the maximum size (number of characters) that can be
     * stored in this
     * TextField.
     * @return the maximum size in characters
     *
     * @see #setMaxSize
     */
    public int getMaxSize() {
        synchronized (Display.LCDUILock) {
            return sD.buffer.length;
        }
    }

    /**
     * Sets the maximum size (number of characters) that can be
     * contained in this
     * TextField. If the current contents of the TextField are larger than
     * maxSize, the contents are truncated to fit.
     *
     * @param maxSize the new maximum size
     *
     * @return assigned maximum capacity - may be smaller than requested.
     * @throws IllegalArgumentException if maxSize is zero or less.
     *
     * @see #getMaxSize
     */
    public int setMaxSize(int maxSize) {
        synchronized (Display.LCDUILock) {
            int oldHeight = height;
            sD.setMaxSize(maxSize);
            layoutChanged(oldHeight);
            return sD.buffer.length;
        }
    }

    /**
     * Gets the number of characters that are currently stored in this
     * TextField.
     * @return number of characters in the TextField
     */
    public int size() {
        synchronized (Display.LCDUILock) {
            if (sD.buffer == null) {
                return 0;
            } else {
                return sD.numChars;
            }
        }
    }

    /**
     * Gets the current input position.
     * For some UIs this may block some time and ask the user about
     * the intended caret 
     * position, on some UIs may just return the caret position.
     * @return the current caret position, 0 if in the beginning.
     */
    public int getCaretPosition() {
        synchronized (Display.LCDUILock) {
            return sD.numChars;
        }
    }

    /**
     * Sets the input constraints of the TextField. If the the current contents
     * of the TextField do not match the new constraints, the contents are
     * set to empty.
     *
     * @param constraints see <a href="#constraints">input constraints</a>
     *
     * @throws IllegalArgumentException if constraints is not any of the ones
     * specified in <a href="TextField.html#constraints">input constraints</a>
     *
     * @see #getConstraints
     */
    public void setConstraints(int constraints)  {
        synchronized (Display.LCDUILock) {
            int oldHeight = height;
            sD.setConstraints(constraints);
            sD.setMaxSize(sD.buffer.length);
            layoutChanged(oldHeight);
        }
    }

    /**
     * Get the current input constraints of the TextField.
     *
     * @return the current constraints value (see
     * <a href="#constraints">input constraints</a>)
     *
     * @see #setConstraints
     */
    public int getConstraints() {
        synchronized (Display.LCDUILock) {
            return sD.constraints;
        }
    }

    /**
     * Sets the label of the Item. If label is null, specifies that this
     * item has no label.
     * @param label the label string
     */
    public void setLabel(String label) {
        synchronized (Display.LCDUILock) {
            super.setLabel(label);

            int oldHeight = height;
            ((StringLayout)layouts[0]).setString(label);
            layoutChanged(oldHeight);
        }
    }

    // package private

    /**
     * setWidth allows the width of the TextField to be set
     *
     * @param width The width to set the TextField to
     * @return int The new height based on the new width
     */
    int setWidth(int width) {
        if (this.width != width) {
            this.width = width;
            doLayout();
        }
        return height;
    }

    /**
     * Get the height of this TextField
     *
     * @return int Return the height in pixels of this TextField
     */
    int getHeight() {
        return height;
    }

    /**
     * Paint this TextField
     *
     * @param g The Graphics object to paint to
     */
    void paint(Graphics g) {
        int translateX = 0;
        int translateY = 0;

        g.setColor(Display.ERASE_COLOR);
        g.fillRect(0, 0, width, height);

        // draw label
        if (horLayout) {
            g.translate(0, 4);
            layouts[0].paint(g, false, false);
            translateX += labelWidth;
            g.translate(labelWidth, -4);
        } else {
            layouts[0].paint(g, false, false);
            int labelHeight = layouts[0].getHeight();
            g.translate(0, labelHeight);
            translateY += labelHeight;
        }

        // draw box
        if (hasFocus()) {
            g.setColor(0x00000000);
        } else {
            g.setColor(0x00AFAFAF);
        }
        g.drawRect(2, 2, tfWidth + 4, contentHeight + 4);

        g.translate(4, 4);
        translateX += 4;
        translateY += 4;

        // need to special case password paint
        // to show last char during input (otherwise ECHO char will be drawn)
        sD.policy.paint(g, sD.buffer, sD.numChars, false, 0,
                        (sD.policy instanceof PassTextPolicy));

        g.translate(-translateX, -translateY);
    }
    
    /**
     * ONLY used by TextEditor. Content of text field might have changed;
     * height and contentHeight variables have to be updated. It is the
     * caller's responsibility to notify the Form's ItemStateChangeListener.
     *
     * @param newContentHeight The new height of the content
     */
    void saveHeight(int newContentHeight) {
        if (!horLayout) {
            int deltaHeight = newContentHeight - contentHeight;
            contentHeight = newContentHeight;
            height += deltaHeight;
            contentChanged(0, 0, deltaHeight);
        }
    }

    /**
     * Select this TextField
     *
     * @return boolean Always returns false, invokes the TextEditor
     */
    boolean select() {
        textEditor.invoke(getOwner(), this, 0, false);
        return false;
    }

    /**
     * Handle a key pressed event
     *
     * @param keyCode The code of the key which was pressed
     * @return boolean Always returns false, may invoke the editor
     */
    boolean keyPressed(int keyCode) {
        // System.out.println("inside textfields keyPressed for "+keyCode);
        if (Display.getSystemKey(keyCode) == EventHandler.SYSTEM_KEY_SEND) {
            if (sD.policy instanceof PhonePolicy) {
                PhoneDial.call(getString());
            }
            return false;
        }

        // fix for delete as delete cause problems on invoking textbox.
        // If delete is pressed in TextField then it deletes one character
        // and then invokes the TextBox. Doesn't do anything for long delete.
        if (Display.getSystemKey(keyCode) == EventHandler.SYSTEM_KEY_CLEAR) {
            // delete case
            if (sD.numChars > 0) {
                delete(sD.numChars - 1, 1);
            }
            keyCode = 0; // non action
        }

        textEditor.invoke(getOwner(), this, keyCode, false);

        // NOTE: The text editor will notify the item state listener
        // itself upon a 'SAVE' command, so we always return false
        // from here
        return false;
    }

    /**
     * Method to allow for keyboard entry
     *
     * @param c The character entered from the keyboard
     */
    void keyTyped(char c) {
        textEditor.invoke(getOwner(), this, c, true);
    }

    // private

    /**
     * Layout the contents of this TextField
     */
    private void doLayout() {
        labelWidth = 0;
        height     = layouts[0].setWidth(width);

        tfWidth       = sD.getMaxWidth(width - SPACER);
        contentHeight = sD.setWidth(tfWidth);

        horLayout = false;

        // Horizontal layout could be possible only if label is one line high
        // and content is one line hight (the same font is used in both cases)
        if (height <= Screen.CONTENT_HEIGHT &&
                contentHeight <= Screen.CONTENT_HEIGHT) {
            String label = getLabel();
            if (label != null) {
                labelWidth = Screen.CONTENT_FONT.stringWidth(label);
            }

            // See if label and text field itself will fit on one line
            if (labelWidth + tfWidth + SPACER < width) {
                horLayout = true;
                // contentHeight is at least of a lineHeight
                height = contentHeight + SPACER;
                return;
            }
        }

        height += (contentHeight + SPACER);
    }

    /**
     * Notify this TextField that its layout has changed
     *
     * @param oldHeight The old height of this TextField
     */
    private void layoutChanged(int oldHeight) {
        if (initLayoutDone()) {
            doLayout();
            contentChanged(0, 0, height - oldHeight);
        }
    }

}
