/* 
 * @(#)configuration.c	1.14 01/07/23 @(#)
 * Copyright (c) 2001 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 * 
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 * 
 */

/*=========================================================================
 * MIDP native configuration property implementation
 *=========================================================================
 * SYSTEM:    MIDP
 * SUBSYSTEM: Configuration properties 
 * FILE:      configprop.c
 * OVERVIEW:  This file defines the functions that access runtime configuration
 *            properties that are predefined in configuration files or
 *            overridden on the command line or environmental variables.
 * ASSUMPTIONS:
 *            The complete property mechanism will not be needed when 
 *            MIDP is ported to a physical device where there is no
 *            need for command line arguments or environment variable overrides.
 *            This module is intended to simplify the porting effort in identifying
 *            runtime switches that can be optimized when shrinking the 
 *            the complete system by hard wiring specific components in the system.
 *
 * DataStructure:
 *           Two configuration files are supported. First, the lib/system.config
 *           contains the properties visible to the application code via the
 *           System.getProperty() interface. Second, the lib/internal.config
 *           contains implementation specific properties that are not intended
 *           to be exposed to the MIDlet application.
 * 
 *           A configuration file contains "key: value\n" lines. Carriage
 *           returns are ignored, and lines beginning with the pound sign 
 *           are skipped as comment lines. White space after the colon are
 *           trimmed before the kay and value are recorded.
 *
 *           The system will continue to run even if the configuration files
 *           can not be read or contain parsing errors in the data.
 *
 * AUTHOR:    Gary Adams
 *=======================================================================*/

/*=========================================================================
 * Include files
 *=======================================================================*/
#include <stdio.h>
#include <stdlib.h>

#ifdef WIN32
#include <string.h>
#else
#include <strings.h>
#endif

#include <ctype.h>

#include "configuration.h"
#include "storage.h"

#include "global.h"
#include "kvmUtil.h"

#ifdef UNIX
#define IMPL_PROPERTY_FILE "lib/internal.config"
#define APPL_PROPERTY_FILE "lib/system.config"
#else /* WINDOWS */
#define IMPL_PROPERTY_FILE "lib\\internal.config"
#define APPL_PROPERTY_FILE "lib\\system.config"
#endif

/* buffer size for line parsing */
#define MAX_LINE 512

/*========================================================================
 * Variables
 *========================================================================*/
typedef struct _configproperty {
    struct _configproperty *next ;
    char *key;
    char *value;
} Property ;
 
/* 
 * Use a simple linked list of name value pairs for application level and
 * implementation targeted properties. The space is partitioned to provide 
 * some separation of values that can be protected.
 */
static Property * applicationProperties = NULL ;
static Property * implementationProperties = NULL ;

/* Forward references */
extern void initProps(Property **, char *, char *);

/*=========================================================================
 * FUNCTION:      initializeConfig()
 * TYPE:          public operation
 * OVERVIEW:      Initialize the configuration data.
 * INTERFACE:
 *=======================================================================*/
/* a default for the configuration property is required by the CLDC spec */
#define DEFAULT_CONFIGURATION "microedition.configuration"
#define DEFAULT_CLDC "CLDC-1.0"

/* a default for the encoding property is required by the CLDC spec */
#define ENCODING_PROP_NAME "microedition.encoding"
#define DEFAULT_CHARACTER_ENCODING "ISO-8859-1"

/* a default for profiles property is required by the MIDP spec */
#define PROFILES_PROP_NAME "microedition.profiles"
#define DEFAULT_PROFILE "MIDP-1.0"

void initializeConfig (char *midp_home) {
    initProps(&implementationProperties, IMPL_PROPERTY_FILE, midp_home);
    initProps(&applicationProperties, APPL_PROPERTY_FILE, midp_home);

    /* 
     * Make sure the configuration was specified, because
     * some older code requires it in the CLDC classes.
     */
    if (getSystemProp(DEFAULT_CONFIGURATION) == NULL)
	setSystemProp(DEFAULT_CONFIGURATION, DEFAULT_CLDC);

    if (getSystemProp(PROFILES_PROP_NAME) == NULL) {
	setSystemProp(PROFILES_PROP_NAME, DEFAULT_PROFILE);
    }

    if (getSystemProp(ENCODING_PROP_NAME) == NULL) {
	setSystemProp(ENCODING_PROP_NAME, DEFAULT_CHARACTER_ENCODING);
    }
}

/*=========================================================================
 * FUNCTION:      finalizeConfig()
 * TYPE:          public operation
 * OVERVIEW:      Finalize the configuration subsystem by releasing
 *                all the allocating memory buffers
 * INTERFACE:
 *=======================================================================*/
void finalizeConfig () {
    Property *prop, *tmp;

    for (prop = implementationProperties ; prop ; prop = tmp ){
	tmp = prop->next;
	jam_free(prop->key);
	jam_free(prop->value);
	jam_free(prop);
    }
    for (prop = applicationProperties ; prop ; prop = tmp ){
	tmp = prop->next;
	jam_free(prop->key);
	jam_free(prop->value);
	jam_free(prop);
    }
}

/*=========================================================================
 * FUNCTION:      parseConfig()
 * TYPE:          public operation
 * OVERVIEW:      Seed the implementation properties from 
 *                a well known property file 
 * INTERFACE:
 *=======================================================================*/
static void parseConfig (int fd, Property ** props) {
    char buffer[MAX_LINE+1];
    int i;
    int j;
    int key_index ;
    int value_index;
    int line = 0;
    int len; 
    char *errStr = NULL ;
    Property * prop;

    /* Read a line at a time */
    for (i= 0 ; i < MAX_LINE ; i++) {
	len = storageRead(&errStr, fd, &buffer[i], 1);
	if (errStr != NULL) {
	    fprintf (stderr, "Warning: could not read config file: %s\n", errStr);
	    storageFreeError(errStr);
	    return;
	}
	/* End of File */
	if (len <= 0)
	    break ; 

	/* End of Line */
	if ((buffer[i] == '\n') || (i == MAX_LINE )) {
	    line++;
	    buffer[i] = 0;

	    /* Skip comment lines which begin  with '#'*/
	    if (buffer[0] != '#'){
	      
		/* Parse the line */
		key_index= 0;
		for (j = key_index; buffer[j] ; j++){
		    if (buffer[j] == ':') {
			buffer[j] = 0;
			value_index = ++j;
			while (isspace(buffer[value_index])) value_index++ ;
		  
			if (prop = (Property *) jam_malloc (sizeof(Property))){
			    prop->next = *props ;
			    prop->key = jam_strdup(buffer+key_index);
			    prop->value = jam_strdup(buffer+value_index);
    
			    *props = prop ;
			}
			/* DEBUG : fprintf (stderr, "key(%s) value(%s)\n",
				 &buffer[key_index],&buffer[value_index]);  */
			break;
		    }
		}
	    }
	    
	    /* If we exceded the line buffer size, flush til end of line */
	    while (i  == MAX_LINE  && buffer[i] != '\n' && len > 0)
		len = storageRead(&errStr, fd, &buffer[i], 1);
	    
	    /* Reset to the beginning of line for the next property value*/
	    i = -1;
	}
	/* Ignore carriage returns */
	if (buffer[i] == '\r') {
	    i-- ;
	}

    }
}

/*=========================================================================
 * FUNCTION:      initProps(Property **, char*)
 * TYPE:          public operation
 * OVERVIEW:      Seed the implementation properties from 
 *                a well known property file 
 * INTERFACE:
 *   parameters:  Property ** props - the current pool of properties
 *                char * name - base filename for configuration values
 *=======================================================================*/
void initProps (Property ** props, char *name, char *midp_home) {
    char pathname[MAX_FILENAME_LENGTH + 1]; 
    int fd = -1 ;
    char * errStr;

    /* Property file can be relative or at midp_home variable */
    (void)strcpy(pathname, midp_home);
    (void)strcat(pathname, getFileSeparator());
    (void)strcat(pathname, name);

    if ((fd = storageOpen(&errStr, pathname, OPEN_READ)) != -1){
	/* Read through the file one line at a time */
	parseConfig (fd, props);

	/* Close the storage handle */
	storageClose (&errStr, fd);
    } else {
	if (errStr != NULL) {
	    fprintf (stderr, "Warning: could not open config file(%s): %s\n",
		     pathname, errStr);
	    storageFreeError(errStr);
	    return;
	}

    }
    
}

/*=========================================================================
 * FUNCTION:      findProp(Property * , char *)
 * TYPE:          public operation
 * OVERVIEW:      Return a value of a property key
 * INTERFACE:
 *   parameters:  Property *props
 *                char *key
 *   returns:     char *result
 *=======================================================================*/
static char * findProp (Property * prop, char* key) {
    Property * p= prop;
    while (p) {
	if (strcmp(key, p->key) == 0 )
	    return (p->value);
	p = p->next;
    }
    return(NULL);
}



/*=========================================================================
 * FUNCTION:      setProp(Property *, char * , char *)
 * TYPE:          public operation
 * OVERVIEW:      Set a value of a property key, value pair
 * INTERFACE:
 *   parameters:  Property *props
 *                char *key
 *                char *value
 *=======================================================================*/
static void setProp (Property **props, char* key , char* value ) {
    /* Try to find the property in the current pool. */
    Property *p;
    for (p= *props; p; p = p->next){
	if (strcmp(key, p->key) == 0 ){
	    jam_free (p->value);
	    p->value = jam_strdup(value);
	    return ;
	}
    }

    /* If the value is not defined, add it now */
    if (p = (Property *) jam_malloc (sizeof(Property))){
	p->next = *props ;
	p->key = jam_strdup(key);
	p->value = jam_strdup(value);
    
	*props = p ;
    }
  
}


/*=========================================================================
 * FUNCTION:      setInternalProp(char * , char *)
 * TYPE:          public operation
 * OVERVIEW:      Set a value of a property key, value pair
 * INTERFACE:
 *   parameters:  char *key
 *                char *value
 *=======================================================================*/
void setInternalProp (char* key , char* value ) {
    setProp(&implementationProperties, key, value);
}

/*=========================================================================
 * FUNCTION:      getInternalProp(char *)
 * TYPE:          public operation
 * OVERVIEW:      Return a value of a property key
 * INTERFACE:
 *   parameters:  char *key
 *   returns:     char *result
 *=======================================================================*/
char * getInternalProp (char* key) {
    char *result = findProp(implementationProperties, key) ;
    return (result? result : findProp(applicationProperties, key));
}

/*=========================================================================
 * FUNCTION:      getInternalProp(char *)
 * TYPE:          public operation
 * OVERVIEW:      Return a value of a property key or the supplied default.
 * INTERFACE:
 *   parameters:  char *key
 *                char *def 
 *   returns:     char *result
 *=======================================================================*/
char * getInternalPropDefault (char* key, char * def) {
    /* check the implementation specific configuration properties. */
    char *result = findProp(implementationProperties, key) ;
    if (result == NULL) {
	/* check the apllication properties, as well. */
	result = findProp(applicationProperties, key);
    }
    /* If the key was not found return the supplied default value. */
    return (result? result : def);
}

/*=========================================================================
 * FUNCTION:      getSystemProp(char *)
 * TYPE:          public operation
 * OVERVIEW:      Return a value of a property key
 * INTERFACE:
 *   parameters:  char *key
 *   returns:     char *result
 *=======================================================================*/
char * getSystemProp (char* key ) {
    return findProp(applicationProperties, key);
}


/*=========================================================================
 * FUNCTION:      setSystemProp(char*, char*)
 * TYPE:          public operation
 * OVERVIEW:      Set a value of a property key, value pair
 * INTERFACE:
 *   parameters:  char *key
 *                char *value
 *=======================================================================*/
void setSystemProp (char* key , char* value) {
    setProp(&applicationProperties, key, value);
}

/*=========================================================================
 * FUNCTION:      getProperty0(Ljava.lang.String)Ljava.lang.String (STATIC)
 * CLASS:         com.sun.midp.Configuration
 * TYPE:          static native function
 * OVERVIEW:      Return a property
 * INTERFACE (operand stack manipulation):
 *   parameters:  A string
 *   returns:     A string
 *=======================================================================*/
void Java_com_sun_midp_Configuration_getProperty0(void)
{
    STRING_INSTANCE string = topStackAsType(STRING_INSTANCE);
    STRING_INSTANCE result;
    char* key = getStringContents(string);
    char* value = NULL;

    /* Look up the property value */
    value = getInternalProp(key);

    result = value == NULL ? NULL : instantiateString(value, strlen(value));
    topStackAsType(STRING_INSTANCE) = result;
}
