/*
 *  @(#)MEKeyTool.java	1.3 01/08/22
 * 
 *  Copyright (c) 2001 Sun Microsystems, Inc., 901 San Antonio Road,
 *  Palo Alto, CA 94303, U.S.A.  All Rights Reserved.
 *
 *  Sun Microsystems, Inc. has intellectual property rights relating
 *  to the technology embodied in this software.  In particular, and
 *  without limitation, these intellectual property rights may include
 *  one or more U.S. patents, foreign patents, or pending
 *  applications.  Sun, Sun Microsystems, the Sun logo, Java, KJava,
 *  and all Sun-based and Java-based marks are trademarks or
 *  registered trademarks of Sun Microsystems, Inc.  in the United
 *  States and other countries.
 *
 *  This software is distributed under licenses restricting its use,
 *  copying, distribution, and decompilation.  No part of this
 *  software may be reproduced in any form by any means without prior
 *  written authorization of Sun and its licensors, if any.
 *
 *  FEDERAL ACQUISITIONS:  Commercial Software -- Government Users
 *  Subject to Standard License Terms and Conditions
 */
package com.sun.midp.mekeytool;

import java.util.*;
import java.io.*;

import java.security.*;
import java.security.cert.*;
import java.security.interfaces.RSAPublicKey;
import java.math.BigInteger;

import com.sun.midp.publickeystore.PublicKeyStoreBuilderBase;
import com.sun.midp.publickeystore.PublicKeyInfo;

/**
 * Manages the initial public keystore needed to bootstrap the MIDP
 * security RI. It provides both a Java and a command line interface.
 * <p>
 * The anchor of trust on an ME (mobile equipment) are the public keys
 * loaded on it by the manufacturer, in RI this is known as the
 * <i>ME keystore</i>. This tool does for the RI what the manufacturer must
 * do for the ME so that trusted MIDP applications can be authenticated.
 * @see #main(String[])
 */
public class MEKeyTool {
    /** default MIDP application directory, see Utility.c getStorageRoot() */
    private final static String defaultAppDir = "appdb";

    /** default ME keystore filename, see com.sun.midp.Main.java */
    private final static String defaultKeystoreFilename = "_main.ks";

    /** read-writable ME keystore that does not depend on KSSL */
    private PublicKeyStoreBuilderBase keystore;

    /** an enumeration for getFirstKey and getNextKey */
    private Enumeration owners;

    /**
     * Performs the command specified in the first argument.
     * <p><pre>
     *MEKeyTool supports the following commands:
     *
     *  no args  - same has -help
     *  -import  - import a public key from a JCE keystore
     *              into a ME keystore
     *  -delete  - delete a key from a ME keystore
     *  -help    - print a usage summary
     *  -list    - list the owner and validity period of each
     *              key in a ME keystore
     *
     *Parameters for (commands):
     *
     *  -MEkeystore &lt;filename of the ME keystore&gt; (optional for all)
     *  -keystore   &lt;filename of the JCA keystore&gt; (optional import)
     *  -storepass  &lt;password for the JCA keystore&gt; (optional import)
     *  -alias      &lt;short string ID of a key in a JCA keystore&gt; (import)
     *  -domain     &lt;security domain of the ME key&gt; (optional import)
     *  -owner      &lt;name of the owner of a ME key&gt; (delete)
     *
     *Defaults:
     *
     *  -MEkeystore appdir/main.ks
     *  -keystore   &lt;user's home dir&gt;/.keystore
     *  -domain     untrusted
     * </pre>
     * @param args command line arguments
     */
    public static void main(String[] args) {
        File meKeystoreFile = null;

        if (args.length == 0 || args[0].equalsIgnoreCase("-help")) {
            // user just needs help with the arguments
            System.out.println("\n  MEKeyTool argument combinations:\n\n" +
                "    -delete [-MEkeystore <filename>] " +
                "-owner <owner name>\n" +
                "    -help\n" +
                "    -import [-MEkeystore <filename>] " +
                "[-keystore <filename>]\n" +
                "            [-storepass <password>] -alias <key alias> " +
                "[-domain <domain>]\n" +
                "    -list [-MEkeystore <filename>]\n");
            return;
        }


        // start with the default keystore file
        meKeystoreFile = new File(defaultAppDir, defaultKeystoreFilename);

        try {
            if (args[0].equalsIgnoreCase("-import")) {
                importCommand(meKeystoreFile, args);
                return;
            }
            
            if (args[0].equalsIgnoreCase("-delete")) {
                deleteCommand(meKeystoreFile, args);
                return;
            }

            if (args[0].equalsIgnoreCase("-list")) {
                listCommand(meKeystoreFile, args);
                return;
            }

            System.out.println("  Invalid command: " + args[0]);
            return;
        } catch (Exception e) {
            System.out.println("  Error: " + e.getMessage());
        }
    }

    /**
     * Process the command line arguments for the import command and
     * then imports a public key from a JCA keystore to ME keystore.
     * This method assumes the first argument is the import command
     * and skips it.
     * @param meKeystoreFile ME keystore abstract file name
     * @param args command line arguments
     * @exception Exception if an unrecoverable error occurs
     */
    private static void importCommand(File meKeystoreFile, String[] args)
            throws Exception {
        String jcaKeystoreFilename = null;
        String keystorePassword = null;
        String alias = null;
        String domain = "untrusted";
        MEKeyTool keyTool;

        for (int i = 1; i < args.length; i++) {
            if (args[i].equalsIgnoreCase("-MEkeystore")) {
                i++;
                meKeystoreFile = new File(args[i]); 
            } else if (args[i].equalsIgnoreCase("-keystore")) {
                i++;
                jcaKeystoreFilename = args[i]; 
            } else if (args[i].equalsIgnoreCase("-storepass")) {
                i++;
                keystorePassword = args[i]; 
            } else if (args[i].equalsIgnoreCase("-alias")) {
                i++;
                alias = args[i];
            } else if (args[i].equalsIgnoreCase("-domain")) {
                i++;
                domain = args[i];
            } else {
                throw new Exception("Invalid argument for import command: " +
                                    args[i]);
            }
        }


        if (jcaKeystoreFilename == null) {
            jcaKeystoreFilename = System.getProperty("user.home") +
                                  File.separator + ".keystore";
        }
        
        if (alias == null) {
            throw new Exception("SE key alias not specified");
        }

        try {
            keyTool = new MEKeyTool(meKeystoreFile);
        } catch (FileNotFoundException fnfe) {
            keyTool = new MEKeyTool();
        }

        keyTool.importKeyFromJcaKeystore(jcaKeystoreFilename,
                                      keystorePassword,
                                      alias, domain);
        keyTool.saveKeystore(meKeystoreFile);
    }

    /**
     * Process the command line arguments for the delete command and
     * then delete a public key from a ME keystore.
     * This method assumes the first argument is the delete command
     * and skips it.
     * @param meKeystoreFile ME keystore abstract file name
     * @param args command line arguments
     * @exception Exception if an unrecoverable error occurs
     */
    private static void deleteCommand(File meKeystoreFile, String[] args)
            throws Exception {
        String owner = null;
        MEKeyTool keyTool;

        for (int i = 1; i < args.length; i++) {
            if (args[i].equalsIgnoreCase("-MEkeystore")) {
                i++;
                meKeystoreFile = new File(args[i]); 
            } else if (args[i].equalsIgnoreCase("-owner")) {
                i++;
                owner = args[i];
            } else {
                throw new Exception("Invalid argument for delete command: " +
                                    args[i]);
            }
        }

        keyTool = new MEKeyTool(meKeystoreFile);

        if (owner == null) {
            throw new Exception("Key owner not specified");
        }

        keyTool.deleteKey(owner);
        keyTool.saveKeystore(meKeystoreFile);
    }

    /**
     * Process the command line arguments for the list command and
     * then list the public keys of a ME keystore.
     * This method assumes the first argument is the list command
     * and skips it.
     * @param meKeystoreFile ME keystore abstract file name
     * @param args command line arguments
     * @exception Exception if an unrecoverable error occurs
     */
    private static void listCommand(File meKeystoreFile, String[] args)
            throws Exception {
        MEKeyTool keyTool;
        PublicKeyInfo key;

        for (int i = 1; i < args.length; i++) {
            if (args[i].equalsIgnoreCase("-MEkeystore")) {
                i++;
                meKeystoreFile = new File(args[i]); 
            } else {
                throw new Exception("Invalid argument for list command: " +
                                    args[i]);
            }
        }

        System.out.println("");

        keyTool = new MEKeyTool(meKeystoreFile);
        key = keyTool.getFirstKey();
        for (int i = 1; key != null; i++) {
            System.out.println("Key " + Integer.toString(i));
            System.out.println(formatKeyInfo(key));
            key = keyTool.getNextKey();
        }

        System.out.println("");
    }

    /**
     * Constructs a MEKeyTool with an empty keystore.
     */
    public MEKeyTool() {
        keystore = new PublicKeyStoreBuilderBase();
    }

    /**
     * Constructs a MEKeyTool and loads its keystore using a filename.
     * @param meKeystoreFilename serialized keystore file
     * @exception FileNotFoundException if the file does not exist, is a
     * directory rather than a regular file, or for some other reason
     * cannot be opened for reading.
     * @exception IOException if the key storage was corrupted
     */
    public MEKeyTool(String meKeystoreFilename)
        throws FileNotFoundException, IOException {

        FileInputStream input;

        input = new FileInputStream(new File(meKeystoreFilename));

        try {
            keystore = new PublicKeyStoreBuilderBase(input);
        } finally {
            input.close();
        }
    };

    /**
     * Constructs a MEKeyTool and loads its keystore from a file.
     * @param meKeystoreFile serialized keystore file
     * @exception FileNotFoundException if the file does not exist, is a
     * directory rather than a regular file, or for some other reason
     * cannot be opened for reading.
     * @exception IOException if the key storage was corrupted
     */
    public MEKeyTool(File meKeystoreFile)
        throws FileNotFoundException, IOException {

        FileInputStream input;

        input = new FileInputStream(meKeystoreFile);

        try {
            keystore = new PublicKeyStoreBuilderBase(input);
        } finally {
            input.close();
        }
    };

    /**
     * Constructs a MEKeyTool and loads its keystore from a stream.
     * @param meKeystoreStream serialized keystore stream
     * @exception IOException if the key storage was corrupted
     */
    public MEKeyTool(InputStream meKeystoreStream)
            throws IOException {
        keystore = new PublicKeyStoreBuilderBase(meKeystoreStream);
    };

    /**
     * Copies a key from a Standard Edition keystore into the ME keystore.
     * @param jcakeystoreFilename name of the serialized keystore
     * @param keystorePassword password to unlock the keystore
     * @param alias the ID of the key in the SE keystore
     * @param domain security domain of any application authorized
     *               with the corresponding private key
     */
    public void importKeyFromJcaKeystore(String jcakeystoreFilename,
                                      String keystorePassword,
                                      String alias, String domain)
            throws IOException, GeneralSecurityException {
        FileInputStream keystoreStream;
        KeyStore jcaKeystore;
        X509Certificate cert;
        RSAPublicKey rsaKey;
        String owner;
        long notBefore;
        long notAfter;
        byte[] rawModulus;
        int i;
        int keyLen;
        byte[] modulus;
        byte[] exponent;

        // Load the keystore
        keystoreStream = new FileInputStream(new File(jcakeystoreFilename));

        try {
            jcaKeystore = KeyStore.getInstance(KeyStore.getDefaultType());
            jcaKeystore.load(keystoreStream, keystorePassword.toCharArray());
        } finally {
            keystoreStream.close();
        }

        // get the cert from the keystore
        cert = (X509Certificate)jcaKeystore.getCertificate(alias);
        if (cert == null) {
            throw new CertificateException("Certificate not found");
        }

        // ME stores a condensed owner DN for a key
        owner = condenseDistinguishedName(cert.getSubjectDN().getName());
        notBefore = cert.getNotBefore().getTime();
        notAfter = cert.getNotAfter().getTime();

        // get the key from the cert
        rsaKey = (RSAPublicKey)cert.getPublicKey();

        // get the key parameters from the key
        rawModulus = rsaKey.getModulus().toByteArray();

        /*
         * the modulus is given as the minimum positive integer,
         * will not padded to the bit size of the key, or may have a extra
         * pad to make it positive. KSSL expect the key to be signature
         * bit size. but we cannot get that from the key, so we should
         * remove any zero pad bytes and then pad out to a multiple of 8 bytes
         */
        for (i = 0; i < rawModulus.length && rawModulus[i] == 0; i++);

        keyLen = rawModulus.length - i;
        keyLen = (keyLen + 7) / 8 * 8;
        modulus = new byte[keyLen];

        int k, j;
        for (k = rawModulus.length - 1, j = keyLen - 1;
                 k >= 0 && j >= 0; k--, j--) {
            modulus[j] = rawModulus[k];
        }

        exponent = rsaKey.getPublicExponent().toByteArray();

        // add the key
        if (!keystore.addKey(new PublicKeyInfo(owner, notBefore, notAfter,
                                          modulus, exponent, domain))) {
            throw new CertificateException(
                "Owner already has a key in the ME keystore");
        }
    }

    /**
     * Deletes a public key using the owner's distinguished name.
     * @param owner name of the key's owner
     * @return true, if the key was deleted, else false
     */
    public boolean deleteKey(String owner) {
        return keystore.deleteKey(owner);
    };

    /**
     * Gets the first key in the keystore.
     * @return all the information related to the first key
     */
    protected PublicKeyInfo getFirstKey() {
        owners = keystore.getOwners();
        return getNextKey();
    };

    /**
     * Gets the next key after the previous one returned by
     * {@link #getFirstKey} or this method. If getFirstKey is not called
     * before the first call to this method, null will be returned.
     * @return all the information related to the next key, or null if
     *        there are no more keys
     */
    protected PublicKeyInfo getNextKey() {
        String owner;

        if (owners == null || !owners.hasMoreElements()) {
            return null;
        }

        owner = (String)owners.nextElement();
        return keystore.findKey(owner);
    };

    /**
     * Saves the keystore to a file.
     * @param meKeystoreFile serialized keystore file
     */
    public void saveKeystore(File meKeystoreFile) throws IOException {
        FileOutputStream output;

        output = new FileOutputStream(meKeystoreFile);

        keystore.serialize(output);
        output.close();
    }

    /**
     * Get an enumeration of owner names from the keystore.
     * @return enumeration of owner names
     */
    public Enumeration getOwners() {
	return keystore.getOwners();
    }

    /**
     * Gets the read-write keystore this tool is manipulating.
     * For advanced users.
     * @return read-write keystore 
     */
    public PublicKeyStoreBuilderBase getKeystore() {
        return keystore;
    }

    /**
     * Creates a string representation of a key that is displayed to a
     * user during a list command. The string does not include the modulus
     * and exponent.
     * @param keyInfo key to display
     * @return printable representation of the key
     */
    public static String formatKeyInfo(PublicKeyInfo keyInfo) {
        return "  Owner: " + keyInfo.getOwner() +
            "\n  Valid from " +
            (new Date(keyInfo.getNotBefore())).toString() +
            " to " + (new Date(keyInfo.getNotAfter())).toString() +
            "\n  Domain: " + keyInfo.getDomain();
    };

    /**
     * Condenses a X.500 distinguished name by removing any quotes,
     * removing a leading spaces before attribute tags,
     * and changing any "," separating attributes to a ";"
     * to match KSSL.
     * @param dn distinguished name to condense
     * @return condensed name
     */ 
    public static String condenseDistinguishedName(String dn) {
        char[] name = new char[dn.length()];
        int len = 0;
        char current;
        int ATTR = 1;
        int QUOTE = 2;
        int END_OF_QUOTE = 3;
        int END_OF_ATTR = 4;
        int state = ATTR;

        for (int i = 0; i < dn.length(); i++) {
            current = dn.charAt(i);

            /*
             * "," or ";" separates attributes,
             * except when they appear in quotes
             */
            if (current == '"') {
                if (state == QUOTE) {
                    state = END_OF_QUOTE;
                } else {
                    state = QUOTE;
                }

                continue;
            } else if (state != QUOTE) {
                if (current == ',' || current == ';') {
                    state = END_OF_ATTR;
                } else if (state == END_OF_ATTR) {
                    if (current == ' ') {
                        // skip any spaces between attributes
                        continue;
                    } else {
                        state = ATTR;
                    }
                }
            }

            // KSSL separates with a ";", so we must to match
            if (state == END_OF_ATTR) {
                name[len] = ';';
            } else {
                name[len] = current;
            }
            
            len++;
        }

        return new String(name, 0, len);
    }
}

