/*
 * @(#)ChoiceView.java	1.27 98/11/11 SMI
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package sun.awt.touchable;

import java.awt.*;
import sun.awt.otk.*;

class ChoiceView extends TouchableView implements ChoiceOwner {
    protected Dimension minSize   = null;
    protected int minButtonWidth  = 0;
    protected int minButtonHeight = 0;

    /**
     * The popup implementation is delegated to a MenuView.
     */
    protected MenuView menu;
    protected boolean selectionInProgress = false;
    protected MenuItemOperations selectedItem;

    public ChoiceView(ChoiceModel model,
		      ObjectToolkit toolkit,
		      ContainerView parent,
		      ColorPolicy colors) {
	this(model, toolkit, parent, colors, NULL_BOUNDS);
    }

    public ChoiceView(ChoiceModel model,
		      ObjectToolkit toolkit,
		      ContainerView parent,
		      ColorPolicy colors,
		      Rectangle bounds) {
	super(model, toolkit, parent, colors, bounds);

	/* MenuView needs a model, but we can't use the ChoiceModel */
	/* because that will make the menu appear to the toolkit as */
	/* a Choice component.  So we create a special empty model. */
	Model nullModel = new NullModel(model);

	/* Create a MenuView that will notify us if it is dismissed  */
	menu = new MenuView(nullModel, toolkit, Palette.MENU_COLORS) {
	    public void setSelectedItem(MenuItemOperations item) {
		selectionInProgress = false;
		super.setSelectedItem(item);
	    }

	    public void dismiss(boolean doAction, int modifiers) {
		selectionInProgress = false;

		if (!doAction) {
		    selectChoice(null);
		}

		super.dismiss(doAction, modifiers);
	    }
	};

	nullModel.setDefaultView(menu);
	menu.addController(MenuSoundController.getController());
	menu.addController(new MenuController());
    }

    public void setVisible(boolean b) {
	if (!b) {
	    menu.setVisible(false);
	}

	super.setVisible(b);
    }

    protected void paintView(Model m, Graphics g) {
	boolean enabled    = m.isEnabled();

	ColorPolicy colors = getColorPolicy();
	Color bgrndColor   = colors.getBackground((ComponentModel)m,
						  selectionInProgress,
						  enabled, false);
	Color fgrndColor   = colors.getForeground((ComponentModel)m,
						  selectionInProgress,
						  enabled, false);
	g.setColor(bgrndColor);
	Rectangle r = win.getBounds(false);
	g.fillRect(r.x = DrawUtils.CONTROL_BUTTON_W, 
		   r.y = 0, 
		   r.width -= DrawUtils.CONTROL_BUTTON_W, 
		   r.height);

	r.x = 0;
	r.width = DrawUtils.CONTROL_BUTTON_W;
	DrawUtils.paintArrow(g, new Rectangle(r), Palette.CTRLS_COLORS,
			     selectionInProgress, enabled,
			     DrawUtils.DOWN, Label.CENTER);

	if ((menu != null) && (menu.countItems() > 0)) {
	    g.setFont(getFont());
	    g.setColor(fgrndColor);
	    g.drawString(selectedItem.getLabel(), 
			 r.width + DrawUtils.TEXT_PAD_W,
			 DrawUtils.getBaseLine(getMetrics(), r.height));
	}

	drawFocus();
    }

    public synchronized Dimension getPreferredSize(int rows, int cols) {
	return new Dimension(minSize);
    }

    public synchronized Dimension getMinimumSize(int rows, int cols) {
	return new Dimension(minSize);
    }

    public void updateDependent(Subject m, Object arg) {
	if (    (arg != null) 
	    && !(arg instanceof AWTEvent)  ) {
	    /* a repaint for the Choice should not go to the menu. */
	    /* And AWT events are redirected as necessary, so they */
	    /* shouldn't be forwarded right now, either.           */
	    menu.updateDependent(m, arg);
	}

	if (arg instanceof OrderedList) {
	    calculateSize();
	    return;
	} else if (arg instanceof Integer) {
	    if (selectionInProgress) return;

	    menu.setSelectedItem(((Integer)arg).intValue());
	    selectChoice(menu.getSelectedItem());
	    return;
	} else if (arg instanceof MenuItemOperations) {
	    if (selectionInProgress) return;

	    MenuItemOperations item = (MenuItemOperations)arg;
	    menu.setSelectedItem(item);
	    selectChoice(item);
	    return;
	}

	super.updateDependent(m, arg);
    }

    protected synchronized boolean handleAWTEvent(Model m, AWTEvent e) {
	if (! super.handleAWTEvent(m, e)) {
	    if (selectionInProgress) {
		return menu.handleAWTEvent(m, e);
	    } else {
		return false;
	    }
	} else {
	    return true;
	}
    }

    /* ChoiceOwner interface */
    public void initiateChoice() {
	selectionInProgress = true;

	//repaint arrow to show that menu is up
	repaintAll(getModel(), null);

	menu.setSelectedItem(selectedItem);

	Rectangle r = win.getBounds(true);

	// horizontal adjustment
	r.x += DrawUtils.CONTROL_BUTTON_W;

	menu.showAt(r.x, r.y);
    }

    /* ChoiceOwner interface */
    public boolean selectionInProgress() {
	return this.selectionInProgress;
    }

    /* ChoiceOwner interface */
    public void selectChoice(MenuItemOperations item) {
	if (item != null) {
	    selectedItem = item;
	}

	selectionInProgress = false;
	menu.setVisible(false);

	repaintAll(getModel(), null);
    }

    private void calculateSize() {
	/* recalculate button's preferred size */
	minSize = menu.getSize();

	minSize.width -= minButtonWidth;
	minSize.width += 2*DrawUtils.TEXT_PAD_W;
	minSize.width += DrawUtils.CONTROL_BUTTON_W;

	minSize.height = minButtonHeight;
    }

    protected synchronized void fontChanged(Model m, Font f, FontMetrics fm) {
	Dimension b     = DrawUtils.getTextMinimumSize("", fm);
	minButtonWidth  = b.width;
	minButtonHeight = b.height;

	menu.fontChanged(m, f, fm);
	calculateSize();
	repaintAll(m, null);
    }
   /**
     * repaint the view after clearing out its background
     */
    protected void repaintAll(Model m, Rectangle r){

	/* Dont bother doing the repaint if the component
	 * is not visible.
	 */
	if (!isVisible()) {
	    return;
	}

	Graphics g = null;
	try {
	    g = getGraphics();
	    paintView(m, g);
	} finally {
	    if (g != null) {
		g.dispose();
	    }
	}	
    }
} 

class NullModel implements Model, MenuItemOwner {
    private View view;
    ChoiceModel choice;

    public NullModel(ChoiceModel c) {
	choice = c;
    }

    public View getDefaultView() {
	return view;
    }

    public void initComplete() {
    }

    public boolean isDefaultView(View v) {
	return view.equals(v);
    }

    /**
     * Set the default view.  This also means initializing the view
     * via calls to updateDependents.
     */
    public void setDefaultView(View newView) {
	view = newView;
    }

    public void dispose() {
	view = null;
    }

    public boolean isEnabled() {
	return true;
    }

    public void addDependent(Dependent d) {
    }

    public void removeDependent(Dependent d) {
    }

    public void removeDependents() {
    }

    public void notifyDependents() {
    }

    public void notifyDependents(Object arg) {
    }

    public int countDependents() {
	return 1;
    }

    public void itemChanged(MenuItemOperations item) {
	choice.itemChanged(item);
    }

    public Font getFont() {
	return choice.getFont();
    }

    public void postMenuShowEvent(AWTEvent e) {
	choice.postMenuShowEvent(e);
    }
	
}
