/*
 * @(#)DialogView.java	1.25 98/11/30 SMI
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package sun.awt.touchable;

import java.awt.*;
import sun.awt.otk.*;
import sun.porting.windowsystem.WindowSystem;
import sun.porting.windowsystem.WindowFactory;

class DialogView extends TouchableContainerView {

    /**
     * Creates a DialogView object as a view on the given Model, residing in the
     * specified toolkit and using the supplied ColorPolicy to do its painting.
     * @param model the model associated with this view.
     * @param toolkit the toolkit to use.
     * @param colors the ColorPolicy to use.
     */
    public DialogView(DialogModel model,
		      ObjectToolkit toolkit,
		      ColorPolicy colors) {
	this(model, toolkit, colors, NULL_BOUNDS);
    }

    /**
     * Creates a DialogView object as a view on the given Model, residing in the
     * specified toolkit and using the supplied ColorPolicy to do its painting, 
     * with initial bounds as indicated.
     * @param model the model associated with this view.
     * @param toolkit the toolkit to use.
     * @param colors the ColorPolicy to use.
     * @param bounds to use when creating the underlying window.
     */
    public DialogView(DialogModel model,
		      ObjectToolkit toolkit,
		      ColorPolicy colors,
		      Rectangle bounds) {
	super(model, toolkit, null, colors, bounds,
	      WindowFactory.TOP_LEVEL_DIALOG);

	model.setInsets(this, DrawUtils.DIALOG_INSETS.top, 
			DrawUtils.DIALOG_INSETS.left,
			DrawUtils.DIALOG_INSETS.bottom,
			DrawUtils.DIALOG_INSETS.right);
	setBounds(bounds.x, bounds.y, bounds.width, bounds.height);
    }

    /**
     * Sets the bounds of the view, relative to the view's parent. The x and y coordinates
     * for the view's top left edge are calculated, based on the indicated width and
     * height, such that the view will be flush with the bottom of the screen, and
     * horizontally centered.
     * @param x not used
     * @param y not used
     * @param width The new width of the view
     * @param height The new height of the view
     */
    public void setBounds(int x, int y, int width, int height) {
	// Dialogs are always flush with bottom of screen, centered
	Dimension screenSize = getToolkit().getScreenSize();
	x = (screenSize.width - width) >> 1;
	y = screenSize.height - height;
 	super.setBounds(x, y, width, height);
    }

    /**
     * Gets a Graphics object that has been configured to paint into this View.
     * The clip area of this Graphics object will be set to honor the dialog borders.
     * @return A graphics object that can be used to paint into this View.
     */
     public Graphics getGraphics() {
   	Rectangle bounds = getBounds();
   	Insets insets = ((DialogModel)getModel()).getInsets();
   	Graphics g = super.getGraphics();

   	g.setClip(insets.left, insets.top,
   		  bounds.width - insets.left - insets.right,
   		  bounds.height - insets.top - insets.bottom);
   	return g;
     }

    /**
     * Paints the DialogView.
     * @param m The model which generated the update
     * @param g The Graphics object that should be used for painting
     */
    protected void paintView(Model m, Graphics g) {
	Rectangle bounds = getBounds();
	Insets insets = ((DialogModel)m).getInsets();

	// need to get superclass graphics context, 
	// since DialogView.getGraphics enforces clipping
	// to the insets
 	g = super.getGraphics();

	// draw borders - left, top, then right 
	GroupingColorPolicy colors = (GroupingColorPolicy)getColorPolicy();
 	g.setColor(colors.getGroupingColor());
 	g.fillRect(0, 0, DrawUtils.DIALOG_BORDER_THICK,
		   bounds.height);
 	g.fillRect(0, 0, bounds.width,
		   DrawUtils.DIALOG_BORDER_THICK);
 	g.fillRect(bounds.width - DrawUtils.DIALOG_BORDER_THICK,
		   0,
 		   DrawUtils.DIALOG_BORDER_THICK, bounds.height);
	g.dispose();
    }

    private volatile boolean dialogVisible;

    /**
     * Sets the visible state of this view. For a modal dialog, grabs all input when
     * showing the dialog and releases the grab when hiding the dialog.
     * @param show The desired visibility; true for visible, false for invisible.
     * @exception if dialog is non-modal
     */
    public void setVisible(boolean show) {
 	if (show) {
 	    if (((DialogModel)getModel()).isModal()) {
 		/*
 		 * By convention, all toolkit peer sets do a modal show on
 		 * its own thread so that the call stack does not reference
 		 * any user classes.  (It appears there is a security risk
 		 * otherwise.)
 		 */
  		Thread modalThread = new ModalThread(this);

  		try {
  		    modalThread.start();
		    Sounds.play(Sounds.SWISH);
  		    modalThread.join();
  		} catch (InterruptedException e) {
  		    // REMIND this should eventually be caught silently
  		    System.err.println("Interrupted -- " +e);
  		    e.printStackTrace();
  		}
 	    } else {
 		throw new RuntimeException("Non-modal dialogs are not supported");
 	    }
 	} else {	// hide
  	    synchronized (this) {
  		sun.porting.windowsystem.Window win = getContainerWindow();

  		if (dialogVisible) {
		    win.releaseGrab();
		    Sounds.play(Sounds.UNSWISH);
 		    super.setVisible(false);
  		    dialogVisible = false;
  		    notifyAll();
 		}
 	    }
 	}
    }

    synchronized void modalShow() {
  	sun.porting.windowsystem.Window win = getContainerWindow();

	win.grabInput(win.DISCARD_NOISY);
 	dialogVisible = true;
 	super.setVisible(true);

	// Modal dialogs have to get focus
	requestFocus();

 	while (dialogVisible) {
 	    try {
 		wait();
 	    } catch (InterruptedException e) {
		// REMIND this should eventually be caught silently
		System.err.println("Interrupted -- " +e);
		e.printStackTrace();
 	    }
 	}
    }
}

/*
  * We must use a separate thread to post the modal dialog, else
  * there is the potential for the modal loop to do event handling
  * with unsecure classes on the call stack.
  */
class ModalThread extends Thread {
    DialogView d;

    public ModalThread(DialogView d) {
 	super("AWT-ModalDialog");
 	this.d = d;
    }

    public void run() {
 	d.modalShow();
    }
}

