/*
 * @(#)IMForOldEvents.java	1.2 98/12/04 SMI
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

package sun.awt.touchable;
import sun.awt.otk.*;
import java.awt.*;
import java.awt.event.*;
import sun.awt.otk.ComponentModel;
import java.util.Hashtable;
import com.sun.awt.*;


/**
 * <p>
 * IMForOldEvents provides input method support for non text 
 * components contained in apps and applets that use the 
 * 1.0 event model. This class is not built by default since
 * all licensees may not care about this support.  For those
 * that do, the steps to use it are given below.
 * <p>
 * In the JDK, for non text components, input method support is 
 * only provided to components that have key listeners attached 
 * to them.  Given that listeners are a 1.1 event model feature,  
 * apps that use the 1.0 event model are unsupported.
 * <p> 
 * It is important to note that the virtual keyboard support in truffle
 * is provided via an input method.  And therefore not using this class
 * would mean no virtual keyboard support for non text components
 * contained in apps that use the old event model.
 * <p>
 * To provide this support, this class examines the target and 
 * determines if the 1.0 event model is being used.  If so, it adds 
 * the appropriate listeners to enable the input method support.  When
 * a listener callback is invoked, it creates an old style event and 
 * forwards it the app/applet.  Strictly speaking, to provide input 
 * method support all that is needed is a key listener.  However, 
 * once any listener is added, java.awt.Component assumes that 
 * the component entirely supports the 1.1 event model and hence 
 * does not deliver any events via the old mechanism.  It is to work 
 * around this that when a decision is made to add a key listner,
 * all relevant listeners are added - and when invoked, they all create an
 * equivalent 1.0 event (where one exists) and forward the 
 * event to the component.  In cases where there is no 
 * matching 1.0 event, the event is ignored.
 *
 * <p>
 * <u>Steps to use this class</u><br>
 * <br>
 * <ol>
 * <li> Add IMForOldEvents to your makefile. (solaris && win32).
 *      This file is not built by default.
 *
 * <li> Add src/share/native/sun/awt/awt_listeners_access.c to your
 *      makefiles.  Again this is not built by default.
 *
 * <li> Update TouchableToolkit to use this class.  One way to
 *      to this would be to:
 *      <ul>
 *      <li>  Construct this class in the constructor for 
 *            TouchableToolkit. ie. imSupport = new IMForOldEvents(this).
 *      <li>  In the mapComponent() method call 
 *            imSupport.addListeners(target, model);
 *      <li>  In the disposedPeer() method call,
 *            imSupport.removeListeners((Component)target, (ComponentModel)peer);
 *      </ul>
 * <li> And thats it!
 * </ol>
 */
public class IMForOldEvents {

    private Hashtable targetListenerTable;
    private boolean keyboardSupportRequired;

    /**
     * IMForOldEvents creates an object that provides keyboard support
     * if required for components using the 1.0 event model.
     */
    public IMForOldEvents(ObjectToolkit tk) {
	targetListenerTable = new Hashtable(11);
	keyboardSupportRequired = (tk.isVirtualKeyboardRequired() 
				   ||  tk.getRealInputMethodAdapter() != null);
    }

    private static native void initJNI();
    static {
	initJNI();
    }
    private native boolean newEventsOnly(Component target);
    private native KeyListener getKeyListeners(Component target);
    private native MouseListener getMouseListeners(Component target);
    private native MouseMotionListener getMouseMotionListeners(Component target);
    private native FocusListener getFocusListeners(Component target);
    private native ComponentListener getComponentListeners(Component target);
    private native ActionListener getActionListeners(Component target);
    private native ItemListener getItemListeners(Component target);
    private native WindowListener getWindowListeners(Component target);


    /**
     *
     */
    public void addListeners(Component target, ComponentModel model) {

	if (target instanceof TextComponent ||
	    !keyboardSupportRequired || 
	    newEventsOnly(target)) {
	    return;
	}

	if ((target instanceof KeyboardInputPreferred) ||
	    ((target instanceof Canvas || target instanceof Panel) &&
	     (!(target instanceof ActionInputPreferred ||
		target instanceof PositionalInputPreferred ||
		target instanceof NoInputPreferred)))) {
		
	    EventTranslator eventTranslator =  new EventTranslator(target, model);
	    targetListenerTable.put(target, eventTranslator);
	    target.addKeyListener(eventTranslator);
	    target.addMouseListener(eventTranslator);
	    target.addMouseMotionListener(eventTranslator);
	    target.addFocusListener(eventTranslator);
	    target.addComponentListener(eventTranslator);
	    
	    if (target instanceof Window) {
		((Window) target).addWindowListener(eventTranslator);
	    } else if (target instanceof List) {
		List l = (List)target;
		l.addActionListener(eventTranslator);
		l.addItemListener(eventTranslator);
	    } else if (target instanceof Choice) {
		((Choice) target).addItemListener(eventTranslator);
	    } else if (target instanceof Checkbox) {
		((Checkbox)target).addItemListener(eventTranslator);
	    } else if (target instanceof Button) {
		((Button)target).addActionListener(eventTranslator);
	    } else if (target instanceof TextField) {
		((TextField)target).addActionListener(eventTranslator);
	    }
	}
    }


    void removeListeners(Component target, ComponentModel model) {
	/*
	 * If any listeners where added to the target 
	 * by the peer, ensure that they are removed.
	 */
	EventTranslator eventTranslator = (EventTranslator)targetListenerTable.get(target);
	if (eventTranslator != null) {
	    target.removeKeyListener(eventTranslator);
	    target.removeMouseListener(eventTranslator);
	    target.removeMouseMotionListener(eventTranslator);
	    target.removeFocusListener(eventTranslator);
	    target.removeComponentListener(eventTranslator);

    
	    if (target instanceof Window) {
		((Window) target).removeWindowListener(eventTranslator);
	    } else if (target instanceof List) {
		List l = (List)target;
		l.removeActionListener(eventTranslator);
		l.removeItemListener(eventTranslator);
	    } else if (target instanceof Choice) {
		((Choice) target).removeItemListener(eventTranslator);
	    } else if (target instanceof Checkbox) {
		((Checkbox)target).removeItemListener(eventTranslator);
	    } else if (target instanceof Button) {
		((Button)target).removeActionListener(eventTranslator);
	    } else if (target instanceof TextField) {
		((TextField)target).removeActionListener(eventTranslator);
	    }
	}
    }

    /**
     *<p>
     * EventTranslator is a class that implements all the listeners
     * that are necessary to successfully deliver all possible 1.0
     * events to a component using the old event model.
     *<p>
     * If a listener is invoked and there is no matching 1.0 event 
     * then the event is ignored.  Otherwise an old style event
     * created and posted.
     *<p>
     * The decision about whether the component supports the old
     * or new event model is done when the peer is constructed.
     * To guard against the possibility that the app/applet added
     * the listener later, we check the listeners added to the component
     * each time before delivering the event.  If the listener is of
     * type AWTEventMulticaster, then we know that another listener
     * exists, and so we do not deliver this event via this listener.
     * This protects against duplicate event delivery.
     *
     */
    private class EventTranslator extends KeyAdapter implements MouseListener, 
								MouseMotionListener,
								FocusListener,
								ComponentListener,
								ActionListener,
								ItemListener,
								WindowListener {
	Component target;
	ComponentModel model;
	
	public EventTranslator(Component target, ComponentModel model) {
	    this.target = target;
	    this.model = model;
	}
	
	public void mouseClicked(MouseEvent e) {
	// We ignore this event since there is no concept of
	// a mouse click in the 1.0 event model.
	}
    
	public void mousePressed(MouseEvent e) {
	    MouseListener mouseListener = getMouseListeners(target);
		
	    /**
	     * We only add this listener if at construction
	     * time, no listener of this type has been added.
	     * But to guard against the case where they are
	     * added later, we check again before event delivery.
	     */
	    if (mouseListener instanceof AWTEventMulticaster) {
		return;
	    }
	    Event evt = new Event(e.getSource(), 0, Event.MOUSE_DOWN, e.getX(), e.getY(), 
				  0, e.getModifiers(), null);
	    target.postEvent(evt);
	}
	
	public void mouseReleased(MouseEvent e) {
	    MouseListener mouseListener = getMouseListeners(target);
	    
	    /**
	     * We only add this listener if at construction
	     * time, no listener of this type has been added.
	     * But to guard against the case where they are
	     * added later, we check again before event delivery.
	     */
	    if (mouseListener instanceof AWTEventMulticaster) {
		return;
	    }
	    Event evt = new Event(e.getSource(), 0, Event.MOUSE_UP, e.getX(), e.getY(), 
				  0, e.getModifiers(), null);
	    target.postEvent(evt);
	}
	
	public void mouseEntered(MouseEvent e) {
	    MouseListener mouseListener = getMouseListeners(target);
	    
	    /**
	     * We only add this listener if at construction
	     * time, no listener of this type has been added.
	     * But to guard against the case where they are
	     * added later, we check again before event delivery.
	     */
	    if (mouseListener instanceof AWTEventMulticaster) {
		return;
	    }
	    Event evt = new Event(e.getSource(), 0, Event.MOUSE_ENTER, e.getX(), e.getY(), 
				  0, e.getModifiers(), null);
	    target.postEvent(evt);
	}
	
	public void mouseExited(MouseEvent e) {  
	    MouseListener mouseListener = getMouseListeners(target);
	    
	    /**
	     * We only add this listener if at construction
	     * time, no listener of this type has been added.
	     * But to guard against the case where they are
	     * added later, we check again before event delivery.
	     */
	    if (mouseListener instanceof AWTEventMulticaster) {
		return;
	    }
	    Event evt = new Event(e.getSource(), 0, Event.MOUSE_EXIT, e.getX(), e.getY(), 
				  0, e.getModifiers(), null);
	    target.postEvent(evt);
	}
	
	
	public void mouseDragged(MouseEvent e) {
	    MouseMotionListener mouseMotionListener = getMouseMotionListeners(target);
	    
	    /**
	     * We only add this listener if at construction
	     * time, no listener of this type has been added.
	     * But to guard against the case where they are
	     * added later, we check again before event delivery.
	     */
	    if (mouseMotionListener instanceof AWTEventMulticaster) {
		return;
	    }
	    Event evt = new Event(e.getSource(), 0, Event.MOUSE_DRAG, e.getX(), e.getY(), 
				  0, e.getModifiers(), null);
	    target.postEvent(evt);
	}
	
	public void mouseMoved(MouseEvent e) {
	    MouseMotionListener mouseMotionListener = getMouseMotionListeners(target);
	    
	    /**
	     * We only add this listener if at construction
	     * time, no listener of this type has been added.
	     * But to guard against the case where they are
	     * added later, we check again before event delivery.
	     */
	    if (mouseMotionListener instanceof AWTEventMulticaster) {
		return;
	    }
	    Event evt = new Event(e.getSource(), 0, Event.MOUSE_MOVE, e.getX(), e.getY(), 
				  0, e.getModifiers(), null);
	    target.postEvent(evt);
	}
	
	
	/**
	 * Note: keyPressed()  and keyReleased() methods are deliberately no-ops
	 * given that this listener is used only in the case where key events are
	 * being buffered.  And in the case of buffering, these events are delivered
	 * to the component as a sequence of KeyTyped events.  The key presses and 
	 * key releases are consumed and never passed on.
	 */
	public void keyTyped(KeyEvent e) {
	    KeyListener keyListener = getKeyListeners(target);
	    
	    /**
	     * We only add this key listener if at construction
	     * time, no key listeners have been added.
	     * But to guard against the case where they have 
	     * been added later, we check again before 
	     * event delivery.
	     */
	    if (keyListener instanceof AWTEventMulticaster) {
		return;
	    }
	    
	    /** 
	     * Each KeyTyped java.awt.AWTEvent, gets mapped to a 
	     * a KeyPress java.awt.Event and a KeyRelease 
	     * java.awt.Event since KeyTyped events do not
	     * exist in the 1.0 event model.
	     */
	    Event evt = new Event(e.getSource(), 0, Event.KEY_PRESS, 0, 0, 
				  (int)e.getKeyChar(), 0, null);
	    target.postEvent(evt);
		
	    evt = new Event(e.getSource(), 0, Event.KEY_RELEASE, 0, 0, 
			    (int)e.getKeyChar(), 0, null);
	    target.postEvent(evt);
	}
	
	/**
	 * Invoked when a component gains the keyboard focus.
	 */
	public void focusGained(FocusEvent e) {
	    FocusListener focusListener = getFocusListeners(target);
	    
	    /**
	     * We only add this focus listener if at construction
	     * time, no focus listeners have been added.
	     * But to guard against the case where they have 
	     * been added later, we check again before 
	     * event delivery.
	     */
	    if (focusListener instanceof AWTEventMulticaster) {
		return;
	    }
	    Event evt = new Event(e.getSource(), 0, Event.GOT_FOCUS, 0, 0,
				  0, 0, null);
	    target.postEvent(evt);
	}
	
	/**
	 * Invoked when a component loses the keyboard focus.
	 */
	public void focusLost(FocusEvent e) {
	    FocusListener focusListener = getFocusListeners(target);
	    
	    /**
	     * We only add this focus listener if at construction
	     * time, no focus listeners have been added.
	     * But to guard against the case where they have 
	     * been added later, we check again before 
	     * event delivery.
	     */
	    if (focusListener instanceof AWTEventMulticaster) {
		return;
	    }
	    Event evt = new Event(e.getSource(), 0, Event.LOST_FOCUS, 0, 0,
			      0, 0, null);
	    target.postEvent(evt);
	}
	
	/**
	 * Invoked when component has been resized.
	 * Do nothing, coz there is no corresponding 1.0 event.
	 */
	public void componentResized(ComponentEvent e){}
	
	/**
	 * Invoked when component has been moved.
	 */    
	public void componentMoved(ComponentEvent e){
	    ComponentListener componentListener = getComponentListeners(target);
	    
	    /**
	     * We only add this focus listener if at construction
	     * time, no focus listeners have been added.
	     * But to guard against the case where they have 
	     * been added later, we check again before 
	     * event delivery.
	     */
	    if (componentListener instanceof AWTEventMulticaster) {
		return;
	    }
	    Event evt = new Event(e.getSource(), 0, Event.WINDOW_MOVED, 0, 0,
				  0, 0, null);
	    target.postEvent(evt);
	    
	}
	
	/**
	 * Invoked when component has been shown.
	 * Do nothing, coz there is no corresponding 1.0 event.
	 */
	public void componentShown(ComponentEvent e){}
	
	/**
	 * Invoked when component has been hidden.
	 * Do nothing, coz there is no corresponding 1.0 event.
	 */
	public void componentHidden(ComponentEvent e){}

	
	/**
	 * Invoked when an item's state has been changed.
	 */
	public void itemStateChanged(ItemEvent e) {
	    ItemListener itemListener = getItemListeners(target);
	    
	    /**
	     * We only add this focus listener if at construction
	     * time, no focus listeners have been added.
	     * But to guard against the case where they have 
	     * been added later, we check again before 
	     * event delivery.
	     */
	    if (itemListener instanceof AWTEventMulticaster) {
		return;
	    }
	    int id = ((e.getStateChange() == ItemEvent.SELECTED) ? Event.LIST_SELECT : Event.LIST_DESELECT);
	    Event evt = new Event(e.getSource(), 0, id, 0, 0,
				  0, 0, null);
	    target.postEvent(evt);	
	}
	
	
	/**
	 * Invoked when an action occurs.
	 */
	public void actionPerformed(ActionEvent e) {
	    ActionListener actionListener = getActionListeners(target);
	    
	    /**
	     * We only add this focus listener if at construction
	     * time, no focus listeners have been added.
	     * But to guard against the case where they have 
	     * been added later, we check again before 
	     * event delivery.
	     */
	    if (actionListener instanceof AWTEventMulticaster) {
		return;
	    }
	    
	    Event evt = new Event(e.getSource(), 0, Event.ACTION_EVENT, 0, 0,
				  0, 0, null);
	    target.postEvent(evt);	
	}
	
	/**
	 * Invoked when a window is in the process of being closed.
	 * The close operation can be overridden at this point.
	 */
	public void windowClosing(WindowEvent e) {
	    WindowListener windowListener = getWindowListeners(target);
	    
	    /**
	     * We only add this focus listener if at construction
	     * time, no focus listeners have been added.
	     * But to guard against the case where they have 
	     * been added later, we check again before 
	     * event delivery.
	     */
	    if (windowListener instanceof AWTEventMulticaster) {
		return;
	    }
	    
	    Event evt = new Event(e.getSource(), 0, Event.WINDOW_DESTROY, 0, 0,
				  0, 0, null);
	    target.postEvent(evt);	
	}
	
	/**
	 * Invoked when a window is iconified.
	 */
	public void windowIconified(WindowEvent e) {
	    WindowListener windowListener = getWindowListeners(target);
	    
	    /**
	     * We only add this focus listener if at construction
	     * time, no focus listeners have been added.
	     * But to guard against the case where they have 
	     * been added later, we check again before 
	     * event delivery.
	     */
	    if (windowListener instanceof AWTEventMulticaster) {
		return;
	    }
	    
	    Event evt = new Event(e.getSource(), 0, Event.WINDOW_ICONIFY, 0, 0,
				  0, 0, null);
	    target.postEvent(evt);	
	}
	
	/**
	 * Invoked when a window is de-iconified.
	 */
	public void windowDeiconified(WindowEvent e) {
	    WindowListener windowListener = getWindowListeners(target);
	 
	    /**
	     * We only add this focus listener if at construction
	     * time, no focus listeners have been added.
	     * But to guard against the case where they have 
	     * been added later, we check again before 
	     * event delivery.
	     */
	    if (windowListener instanceof AWTEventMulticaster) {
		return;
	    }
	    
	    Event evt = new Event(e.getSource(), 0, Event.WINDOW_DEICONIFY, 0, 0,
				  0, 0, null);
	    target.postEvent(evt);	
	}
	
	/**
	 * Invoked when a window is activated.
	 */
	public void windowActivated(WindowEvent e) {
	    WindowListener windowListener = getWindowListeners(target);
	    
	    /**
	     * We only add this focus listener if at construction
	     * time, no focus listeners have been added.
	     * But to guard against the case where they have 
	     * been added later, we check again before 
	     * event delivery.
	     */
	    if (windowListener instanceof AWTEventMulticaster) {
		return;
	    }
	    
	    Event evt = new Event(e.getSource(), 0, Event.WINDOW_EXPOSE, 0, 0,
				  0, 0, null);
	    target.postEvent(evt);	
	    
	}
	
	/**
	 * Invoked when a window is de-activated.
	 * Do nothing, coz there is no corresponding 1.0 event.
	 */
	public void windowDeactivated(WindowEvent e){}
	
	
	/**
	 * Invoked when a window has been opened.
	 * Do nothing, coz there is no corresponding 1.0 event.
	 */
	public void windowOpened(WindowEvent e) {}
	
	
	/**
	 * Invoked when a window has been closed.
	 * Do nothing, coz there is no corresponding 1.0 event.
	 */
	
	public void windowClosed(WindowEvent e) {}
	
    }
}    

