/*
 * @(#)MenuView.java	1.45 98/11/19 SMI
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package sun.awt.touchable;

import java.awt.*;
import java.awt.event.AdjustmentEvent;
import java.awt.event.AdjustmentListener;
import java.awt.event.MouseEvent;
import sun.awt.otk.*;
import sun.porting.toolkit.ToolkitEventHandler;

/**
 * Note: because this view is used to implement menus in several different
 * circumstances, it is implemented with no explicit knowledge of the model.
 */
class MenuView extends BasicView implements DragMenu, Adjustable, Scrollable {
    static int BORDER = DrawUtils.MENU_BORDER;

    protected MenuItemOperations items[] = null;
    protected Rectangle item_bounds[] = null;
    protected Rectangle menu_bounds = null;

    protected int pressed_item = -1;
    protected int width  = 0;
    protected int height = 0;
    protected boolean visible;

    private boolean doEventGrab;

    public MenuView(Model model,
		    ObjectToolkit toolkit,
		    ColorPolicy colors) {
	this(model, toolkit, colors, NULL_BOUNDS, true);
    }

    public MenuView(Model model,
		    ObjectToolkit toolkit,
		    ColorPolicy colors,
		    Rectangle bounds,
		    boolean doEventGrab) {
	super(model, toolkit, null, colors, bounds);

	menu_bounds   = new Rectangle(bounds.x, bounds.y,
				    bounds.width, bounds.height);

	Color bgColor = colors.getBackground(null, false,
					     true, false);
	win.setBackgroundColor(bgColor);
	this.doEventGrab = doEventGrab;

	/* first real controller must be the scroll controller */
	addController(Scroll.controller);
	addController(new SimpleController() {
	    public boolean handleEvent(View v,
				       int id, long when, int modifiers,
				       int x, int y, int arg1, int arg2) {
		if (   (id == ToolkitEventHandler.POINTER_PRESSED)
		    || (id == ToolkitEventHandler.POINTER_RELEASED)) {

		    Rectangle r = v.getBounds(true);
		    x -= r.x;
		    y -= r.y;

		    /* gobble any events that land on the scroll buttons */
		    return (buttons != null &&
		      ((buttons[0].isShowing() && buttons[0].inButton(x, y))
		    || (buttons[1].isShowing() && buttons[1].inButton(x, y))));
		} else {
		    return false;
		}
	    }
	});
    }

    protected sun.porting.windowsystem.Window getContainerWindow() {
	return super.getContainerWindow();
    }

    public void setSelectedItem(MenuItemOperations item) {
	for (int i = 0; i < items.length; ++i) {
	    if (item.equals(items[i])) {
		setSelectedItem(i);
		break;
	    }
	}
    }

    protected void setSelectedItem(int itemIndex) {
	int old_item = pressed_item;
	pressed_item = itemIndex;

	if (!visible) return;

	if (old_item != pressed_item) {
	    Graphics g = null;
	    try {
		g = getGraphics();
		if (old_item >= 0) {
		    itemChanged(items[old_item], g);
		}

		if (pressed_item >= 0) {
		    itemChanged(items[pressed_item], g);
		}
	    } finally {
		if (g != null) {
		    g.dispose();
		}
	    }
	}
    }

    
    /* Scrollable interface */
    public boolean startScrolling(int x, int y, int modifiers) {
	if (scroller == null) {
	    return false;
	}

	return scroller.startScrolling(x - menu_bounds.x,
				       y - menu_bounds.y,
				       modifiers);
    }

    public boolean stopScrolling() {
	if (scroller == null) {
	    return false;
	}

	return scroller.stopScrolling();
    }

    /* DragMenu interface, for benefit of Controller */
    public MenuItemOperations getItem(int x, int y) {
	if (!visible) {
	    return null;
	}

	int i = locateItem(x, y);
	return (i < 0) ? null : items[i];
    }

    public MenuItemOperations getSelectedItem() {
	return (pressed_item < 0) ? null : items[pressed_item];
    }

    public void dismiss(boolean doAction, int modifiers) {
	if (doAction) {
	    MenuItemOperations m = getSelectedItem();

	    if (m != null) {
		m.actionOccurred(modifiers);
	    }
	}

	setVisible(false);
    }

    public boolean insideMenu(int x, int y) {
	return visible && menu_bounds.contains(x, y);
    }


    /* convenience routine for Choice implementation. */
    public void showAt(int x, int y) {
	if (item_bounds.length == 0) {
	    return;
	}

	disableScrolling();
	Rectangle lastItem = item_bounds[item_bounds.length - 1];
	setBounds(x, y, 
		  2*BORDER + lastItem.width, 
		  BORDER + lastItem.y + lastItem.height);

	/**
	 * Scroll so that pre-selected item can be positioned at x, y if possible.
	 */

	Rectangle r = item_bounds[pressed_item];

	if ((scroller != null) && scroller.isEnabled() && (pressed_item >= 0)) {
	    Dimension d = getToolkit().getScreenSize();
	    int menu_height = menu_bounds.height;

	    setValue(0);	// start by scrolling to top

	    /* 
	     * If the location is 25% down from the top, attempt to scroll the
	     * preselected item so it is 25% from the top of the menu, etc.
	     */
	    int targetPosition = y*menu_height/d.height;
	    targetPosition = Math.min(menu_height - 2*buttonHeight, targetPosition);
	    targetPosition = Math.max(2*buttonHeight, targetPosition);

	    while ((r.y - currentScrollPosition) > targetPosition) {
		int inc = getUnitIncrement();
		int oldScrollPos = currentScrollPosition;
		setValue(currentScrollPosition + inc);
		if (!buttons[1].isEnabled()) {
		    buttons[1].setShowing(false);
		}
		if (currentScrollPosition == oldScrollPos) {
		    break;
		}
	    }
	}
	setPosition(x, y - r.y + currentScrollPosition);

	setVisible(true);
    }

    /**
     * AWT's expectation is that the menu peer is going to grab
     * input events.  Were this not the case, the component which
     * got the mouse down would have to forward all drag events
     * to the menu, because the normal definition for delivery says
     * that the recipient of the down receives all drags and the
     * matching up.  
     *
     * The view is responsible for performing the grab when the 
     * menu is made visible.
     */
    public void setVisible(boolean visible) {
	if (visible) setStackOrder(0);

 	if (this.visible == visible) {
 	    return;
	}

	if (visible && (scroller != null)) {
	    scroller.updateEnabled(getMinimum(), getMaximum(), menu_bounds.height);
	}

	super.setVisible(visible);

	Model m = getModel();
	if (m instanceof MenuItemOwner) {
	    long when = System.currentTimeMillis();
	    AWTEvent e = new MenuShowEvent(this, when, visible);
	    ((MenuItemOwner)m).postMenuShowEvent(e);
	}

	if (visible && doEventGrab) {
	    win.grabInput(win.DELIVER_ALL);
	} else {
	    pressed_item = -1;
	}

	this.visible = visible;
    }

    protected void handleViewEvent(Model m, ViewEvent e) {
	handleAWTEvent(m, e);
    }

    protected boolean handleAWTEvent(Model m, AWTEvent e) {
 	if (e instanceof MouseEvent) {
 	    /* convert to global coordinates */
 	    MouseEvent me = (MouseEvent) e;
 	    Component c = (Component) me.getSource();
 	    Point p = c.getLocationOnScreen();
 	    me.translatePoint(p.x, p.y);
 	}
	
	return super.handleAWTEvent(getModel(), e);
    }

    protected boolean handleEvent(View v,
				  int id, long when, int modifiers,
				  int x, int y, int arg1, int arg2) {

	/* Use global event coords for menus */
	Rectangle r = (v == this) ? menu_bounds : v.getBounds(true);

	return super.handleEvent(this, id, when, modifiers,
				 x + r.x, y + r.y, arg1, arg2);
    }
			       

    public void updateDependent(Subject m, Object arg) {
	if (arg instanceof OrderedList) {
	    setMenuItems((OrderedList)arg);
	    return;
	} else if (arg instanceof MenuItemOperations) {
	    itemChanged((MenuItemOperations)arg);
	    return;
	} else if (arg instanceof Event) {
	    Event e = (Event) arg;
	    try {
		Component c = (Component) e.target;
 		Point p = c.getLocationOnScreen();
 		setPosition(p.x + e.x, p.y + e.y);
	    } catch (Throwable t) {
		setPosition(e.x, e.y);
	    }
	    return;
	}

	super.updateDependent(m, arg);
    }

    ScrollbarView scroller = null;

    protected boolean enableScrolling() {
	if (scroller == null) {
	    int buttonWidth;

	    Dimension minSize = MenuScrollButton.minimumSize();

	    buttonWidth  = menu_bounds.width - 2*BORDER;
	    buttonHeight = minSize.height + BORDER;

	    buttons = new MenuScrollButton[2];

	    buttons[0] = 
		new MenuScrollButton(this,
				     Palette.MENU_ARROW_COLORS,
				     AdjustmentEvent.BLOCK_DECREMENT,
				     false,
				     BORDER,
				     BORDER,
				     buttonWidth,
				     minSize.height,
				     DrawUtils.UP);

	    buttons[1] =
		new MenuScrollButton(this, 
				     Palette.MENU_ARROW_COLORS,
				     AdjustmentEvent.UNIT_INCREMENT,
				     true,
				     BORDER,
				     menu_bounds.height - buttonHeight,
				     buttonWidth,
				     minSize.height,
				     DrawUtils.DOWN);

	    scroller = new ScrollbarView(this, this, true, buttons);

	    buttons[0].setScrollbar(scroller);
	    buttons[1].setScrollbar(scroller);

	    scroller.setEnabled(false);
	} else {
	    buttons[0].setShowing(false);
	    buttons[1].setShowing(true);
	}

	if (!scroller.isEnabled()) {
	    buttons[0].setEnabled(false);
	    buttons[1].setEnabled(true);

	    currentScrollPosition = 0;
	    topVisibleIndex = 0;
	    bottomVisibleIndex = 0; 
	    while (bottomVisibleIndex < item_bounds.length) {
		Rectangle r = item_bounds[bottomVisibleIndex];
		if ((r.y + r.height) > (menu_bounds.height - buttonHeight)) {
		    --bottomVisibleIndex;
		    break;
		}

		++bottomVisibleIndex;
	    }

	    scroller.setEnabled(true);
	    return true;
	}


	return false;
    }

    protected boolean disableScrolling() {
	if (scroller != null) {
	    buttons[0].setShowing(true);
	    buttons[1].setShowing(false);
	    scroller.setEnabled(false);

	    topVisibleIndex = 0;
	    bottomVisibleIndex = 0x3fffffff;

	    return true;
	}


	return false;
    }

    public void setBounds(int x, int y, int width, int height) {
	boolean repaint = false;
	boolean menuClipped = false;

	Dimension d = getToolkit().getScreenSize();
	int MAX_HEIGHT = (d.height * 2) / 3;

	if (width > d.width) {
	    width = d.width;
	}

	if (height > MAX_HEIGHT) {
	    height = MAX_HEIGHT;
	    menuClipped = true;
	}

	if ((x + width) > d.width) {
	    x = d.width - width;
	}

	if ((y + height) > d.height) {
	    y = d.height - height;
	}

	if (x < 0) {
	    x = 0;
	}

	if (y < 0) {
	    y = 0;
	}

	menu_bounds.x = x;
	menu_bounds.y = y;
	menu_bounds.width  = width;
	menu_bounds.height = height;

	super.setBounds(x, y, width, height);

	if (menuClipped) {
	    repaint = enableScrolling();

	    if (height != menu_bounds.height) {
		Dimension minSize = ScrollButton.minimumSize();

		buttons[1].setPosition((menu_bounds.width - minSize.width) / 2,
				       menu_bounds.height - minSize.height);

		repaint = true;
	    }
	} else if (height != menu_bounds.height) {
	    repaint = disableScrolling();
	}

	if (visible && repaint) {
	    Graphics g = null;
	    try {
		g = getGraphics();
		if (scroller == null) {
		    paintView(getModel(), g);
		} else {
		    scroller.updateEnabled(getMinimum(), getMaximum(), 
					   menu_bounds.height);
		    scroller.paintScrollbar(g);
		}
	    } finally {
		if (g != null) {
		    g.dispose();
		}
	    }
	}
    }

    protected void setPosition(int x, int y) {
	setBounds(x, y, menu_bounds.width, menu_bounds.height);
    }

    protected void setSize(int width, int height) {
	setBounds(menu_bounds.x, menu_bounds.y, width, height);
    }

    protected Dimension getSize() {
	return new Dimension(menu_bounds.width, menu_bounds.height);
    }

    protected int countItems() {
	return (item_bounds == null) ? 0 : item_bounds.length;
    }

    protected int locateItem(int posX, int posY) {
	if (countItems() == 0) {
	    return -1;
	}

	posX -= menu_bounds.x;
	posY -= menu_bounds.y;
	posY += currentScrollPosition;

	for (int i = 0; i < items.length; ++i) {
	    if (item_bounds[i].contains(posX, posY)) {
		return i;
	    }
	}

	return -1;
    }

    protected void setupClip(Graphics g) {
	int yStart = (buttons[0].isShowing()) ? buttonHeight : BORDER;
	int yEnd   = menu_bounds.height
	    - (buttons[1].isShowing() ? buttonHeight : BORDER);

	g.clipRect(BORDER, yStart, 
		   menu_bounds.width - 2*BORDER, yEnd - yStart);
    }

    protected void paintView(Model m, Graphics g) {
	/* REMIND: right now there is no title */

	if (countItems() == 0) {
	    return;
	}

	// draw border
        ColorPolicy colors = getColorPolicy();
	g.setColor(colors.getBackground(null, true, true, false));
	for (int i = 1; i <= BORDER; i++)
	    g.drawRect(i-1, i-1, menu_bounds.width - (2*i - 1),
		       menu_bounds.height - (2*i - 1));

	if (scroller != null) {
	    scroller.paintScrollbar(g);

	    setupClip(g);
	}

	for (int i = Math.max(0, topVisibleIndex - 1); 
	     i < Math.min(bottomVisibleIndex + 2, items.length); ++i) {
	    paintItem(g, item_bounds[i], i);
	}
    }

    protected void paintItem(Graphics g, Rectangle r, int item_index) {

	g.translate(0, -currentScrollPosition);

	MenuItemOperations item = items[item_index];
	boolean selected = false;
	boolean active   = true;
        ColorPolicy colors = getColorPolicy();

	if (item.isSeparator()) {
	    g.setColor(colors.getBackground(null, selected, active, false));
	    g.fillRect(r.x, r.y, r.width, r.height);

	    int y = r.y + (r.height >> 1);
	    g.setColor(colors.getForeground(null, selected, active, false));
	    g.drawLine(r.x, y, r.x + r.width, y);
	    g.translate(0, currentScrollPosition);
	    return;
	}

	String s          = item.getLabel();
	Font f            = item.getFont();
	active            = item.isEnabled();
	selected          = item_index == pressed_item;

	DrawUtils.drawButton(g, r.x, r.y, r.width, r.height, 
			     s, f, null, colors, null, selected, active,
			     false, Label.LEFT);

	g.translate(0, currentScrollPosition);
    }

    protected void fontChanged(Model m, Font f, FontMetrics fm) {
	recalculateSizes();
    }

    protected void setMenuItems(OrderedList v) {
	if ((items == null) || (items.length != v.size())) {
	    items = new MenuItemOperations[v.size()];
	    item_bounds = new Rectangle[v.size()];
	}

	v.copyInto(items);
	recalculateSizes();
    }

    private void recalculateSizes() {
	if (items == null) return;

	/* Now recompute the size based on the new list of elements. */
	Font lastFont  = null;
	FontMetrics fm = null;

	int w = 0;
	int h = BORDER;

	/* REMIND: We must account for a title, if there is one! */
	for (int i = 0; i < items.length; ++i) {
	    if (items[i].isSeparator()) {
		item_bounds[i] = new Rectangle(BORDER, h,
					       w, DrawUtils.MENU_SEP_HEIGHT);

		h += DrawUtils.MENU_SEP_HEIGHT;

		continue;
	    }

	    String s = items[i].getLabel();
	    Font f   = items[i].getFont();

	    if (f != lastFont) {
		fm = getToolkit().getFontMetrics(f);
	    }

	    Dimension d = DrawUtils.getTextMinimumSize(s, fm);
	    item_bounds[i] = new Rectangle(BORDER, h, 
					   w, d.height);

	    w = Math.max(w, d.width);
	    h += d.height;
	}

	for (int i = 0; i < items.length; ++i) {
	    item_bounds[i].width = w;
	}

	w += 2*BORDER;
	h += 2*BORDER;

	setSize(w, h);
    }

    protected void itemChanged(MenuItemOperations item) {
	itemChanged(item, null);
    }

    protected void itemChanged(MenuItemOperations item,
			       Graphics g) {

	if (item.isSeparator()) {
	    return;
	}

	/* check for a size change */
	for (int i = 0; i < items.length; ++i) {
	    if (items[i].equals(item)) {
		String s = item.getLabel();
		FontMetrics fm = getToolkit().getFontMetrics(item.getFont());
		Dimension d = DrawUtils.getTextMinimumSize(s, fm);

		int new_width  = menu_bounds.width;
		int new_height = menu_bounds.height;

		if (d.width > item_bounds[i].width) {
		    new_width = width + d.width - item_bounds[i].width;
		    for (int j = 0; j < items.length; ++j) {
			item_bounds[j].width = d.width;
		    }
		}

		if (d.height != item_bounds[i].height) {
		    new_height = height + d.height - item_bounds[i].height;
		    item_bounds[i].height = d.height;
		}

		if (   (new_width != menu_bounds.width) 
		    || (new_height != menu_bounds.height)) {
		    setSize(new_width, new_height);

		    if (g != null) {
			/* items have moved, so we need to repaint all */
			paintView(getModel(), g);
		    }
		} else if (g != null) {
		    paintItem(g, item_bounds[i], i);
		}

		return;
	    }
	}
    }

    // This is the Adjustable interface.  It is only partially implemented
    // because we are only expecting the scrollbar to interact with it.
    // All of the "set" methods except setValue are programmed to throw
    // an exception if called.

    /* the first item that is FULLY visible */
    int topVisibleIndex = 0;

    /* the last item that is FULLY visible */
    int bottomVisibleIndex = 0x3fffffff;

    /* current scrolling position */
    int currentScrollPosition = 0;

    /* the height of one scroll button */
    int buttonHeight = 0;

    MenuScrollButton buttons[] = null;

    public int getOrientation() {
	return Adjustable.VERTICAL;
    }

    public int getMinimum() {
	return buttons[0].isShowing() ? BORDER - buttonHeight : 0;
    }

    public int getMaximum() {
	if (countItems() == 0) {
	    return 0;
	}

	Rectangle lastItem = item_bounds[item_bounds.length - 1];
	return lastItem.y + lastItem.height +
	    (buttons[1].isShowing() ? buttonHeight - BORDER : 0);
    }

    public int getUnitIncrement() {
	if (countItems() == 0) {
	    return 0;
	}

	/**
	 * Implementation trick, of sorts.  By defining one of the
	 * buttons to be in "unit increment" mode and the other one
	 * to be in "block decrement mode," and returning appropriate
	 * values for the unit and block increments, we can make the
	 * buttons scroll by unequal amounts, taking into account
	 * the varying heights of the items to be scrolled.
	 *
	 * The unit increment is used to make new elements visible at
	 * the bottom of the menu; on a normal desktop this is called
	 * scrolling "down" and corresponds to UNIT_INCREMENT or
	 * BLOCK_INCREMENT.
	 */
	if (bottomVisibleIndex > item_bounds.length) {
	    return 0;
	} else if ((bottomVisibleIndex + 2) >= item_bounds.length) {
	    /* scroll until last element is at bottom */
	    Rectangle lastItem = item_bounds[item_bounds.length - 1];
	    return getMaximum() - getVisibleAmount()
		- currentScrollPosition;
	} else {
	    /*
	     * If the next element is very tall we may have to lose more 
	     * than one element in order to make it visible...hence the
	     * loop.  And because the next element could also be very short,
	     * we start by assuming that we don't need to lose any elements
	     * off the other end.
	     */
	    int visibleAmount = menu_bounds.height - 2*buttonHeight;
	    Rectangle newLastItem  = item_bounds[bottomVisibleIndex + 1];
	    if (   (visibleAmount >= 3*buttonHeight)
		&& (topVisibleIndex == 0) 
		&& (bottomVisibleIndex + 2 < items.length)) {
		newLastItem = item_bounds[bottomVisibleIndex + 2];
	    }
	    for (int i = topVisibleIndex; i < bottomVisibleIndex; ++i) {
		Rectangle newFirstItem = item_bounds[i];

		/*
		 * Calculate the distance required to center the elements
		 * within the menu.
		 */

		int requiredVisibleArea = 
		    newLastItem.y + newLastItem.height - newFirstItem.y;

		int space = (visibleAmount - requiredVisibleArea) / 2;
		if (space >= 0) {
		    int pos = newFirstItem.y - buttonHeight - space;
		    return pos - currentScrollPosition;
		}
	    }

	    /* we have a *very* big item.  Center it within the menu. */
	    int n = newLastItem.y - buttonHeight
		+ ((newLastItem.height - visibleAmount) / 2)
		- currentScrollPosition;
	    return n;
	}
    }

    public int getBlockIncrement() {
	if (countItems() == 0) {
	    return 0;
	}

	/**
	 * Implementation trick, of sorts.  By defining one of the
	 * buttons to be in "unit increment" mode and the other one
	 * to be in "block decrement mode," and returning appropriate
	 * values for the unit and block increments, we can make the
	 * buttons scroll by unequal amounts, taking into account
	 * the varying heights of the items to be scrolled.
	 *
	 * The block increment is used to make new elements visible at
	 * the top of the menu; on a normal desktop this is called
	 * scrolling "up" and corresponds to UNIT_DECREMENT or 
	 * BLOCK_DECREMENT.
	 */
	if (topVisibleIndex <= 1) {
	    return currentScrollPosition - getMinimum();
	} else {
	    /*
	     * If the next element is very tall we may have to lose more 
	     * than one element in order to make it visible...hence the
	     * loop.  And because the next element could also be very short,
	     * we start by assuming that we don't need to lose any elements
	     * off the other end.
	     */
	    int visibleAmount = menu_bounds.height - 2*buttonHeight;
	    Rectangle newFirstItem = item_bounds[topVisibleIndex - 1];
	    if ((bottomVisibleIndex == item_bounds.length - 1) && (topVisibleIndex > 1)) {
		newFirstItem = item_bounds[topVisibleIndex - 2];
	    }
	    for (int i = bottomVisibleIndex; i > topVisibleIndex; --i) {
		Rectangle newLastItem  = item_bounds[i];

		/*
		 * Calculate the distance required to center the elements
		 * within the menu.
		 */

		int requiredVisibleArea = 
		    newLastItem.y + newLastItem.height - newFirstItem.y;

		int space = (visibleAmount - requiredVisibleArea) / 2;

		if (space >= 0) {
		    int pos = newFirstItem.y - buttonHeight - space;
		    return currentScrollPosition - pos;
		}
	    }

	    /* we have a *very* big item.  Center it within the menu. */
	    int n = newFirstItem.y - buttonHeight
		+ ((newFirstItem.height - visibleAmount) / 2)
		- currentScrollPosition;
	    return -n;
	}
    }

    /**
     * Gets the length of the visible amount.  In our case this
     * will depend on whether the buttons are visible.
     */
    public int getVisibleAmount() {
	int n = menu_bounds.height;

	if (buttonHeight == 0) {
	    return n;
	}

	if (buttons[0].isShowing()) {
	    n -= buttonHeight;
	}

	if (currentScrollPosition >= getMaximum() - menu_bounds.height) {
	    return n;
	}

	if (buttons[1].isShowing()) {
	    n -= buttonHeight;
	}
	
	return n;
    }

    protected void buttonVisibilityChange(MenuScrollButton b) {
	setValue(currentScrollPosition, true);
    }

    public void setValue(int v) {
	setValue(v, false);
    }

    public void setValue(int v, boolean justUpdateVisible) {
	v = Math.max(v, getMinimum());
	v = Math.min(v, getMaximum() - menu_bounds.height);
	int delta = v - currentScrollPosition;

	if ((delta == 0) && !justUpdateVisible) {
	    return;
	}

	/* calculate the new values for top and bottom visible index */
	int yBound = menu_bounds.height + v;
	if (buttons[1].isShowing()) {
	    yBound -= buttonHeight;
	}

	if (delta >= 0) {
	    while (bottomVisibleIndex < (item_bounds.length - 1)) {
		Rectangle r = item_bounds[bottomVisibleIndex];
		if (r.y + r.height > yBound) {
		    --bottomVisibleIndex;
		    break;
		}
		++bottomVisibleIndex;
	    }
	} else {
	    while (bottomVisibleIndex > 0) {
		Rectangle r = item_bounds[bottomVisibleIndex];
		if (r.y + r.height <= yBound) {
		    break;
		}
		--bottomVisibleIndex;
	    }
	}

	yBound = v;
	if (buttons[0].isShowing()) {
	    yBound += buttonHeight;
	}

	if (delta > 0) {
	    while (topVisibleIndex < (item_bounds.length - 1)) {
		Rectangle r = item_bounds[topVisibleIndex];
		if (r.y >= yBound) {
		    break;
		}
		++topVisibleIndex;
	    }
	} else {
	    while (topVisibleIndex > 0) {
		Rectangle r = item_bounds[topVisibleIndex];
		if (r.y < yBound) {
		    ++topVisibleIndex;
		    break;
		}
		--topVisibleIndex;
	    }
	}

	if (justUpdateVisible) {
	    return;
	}

	currentScrollPosition = v;

	scroller.updateEnabled(getMinimum(), getMaximum(), menu_bounds.height);

	if (!visible) {
	    return;
	}

	Graphics g = null;
	try {
	    g = getGraphics();

	    /*
	     * copyArea is not working, probably because the emulator
	     * is having some trouble generating the right expose
	     * events.  So for the time being, we call paintView
	     * instead.  
	     */
	    paintView(getModel(), g);

	    // 
	    // setupClip(g);
	    // g.copyArea(BORDER, yStart, 
	    //            menu_bounds.width - 2*BORDER, yEnd - yStart, 
	    //            0, -delta);
	} finally {
	    if (g != null) {
		g.dispose();
	    }
	}
    }

    public int getValue() {
	return currentScrollPosition;
    }

    public void setMinimum(int min) {
	throw new RuntimeException("Who is calling this?");
    }

    public void setMaximum(int max) {
	throw new RuntimeException("Who is calling this?");
    }

    public void setUnitIncrement(int u) {
	throw new RuntimeException("Who is calling this?");
    }

    public void setBlockIncrement(int b) {
	throw new RuntimeException("Who is calling this?");
    }

    public void setVisibleAmount(int v){
	throw new RuntimeException("Who is calling this?");
    }

    public void addAdjustmentListener(AdjustmentListener l) {
	throw new RuntimeException("Who is calling this?");
    }

    public void removeAdjustmentListener(AdjustmentListener l) {
	throw new RuntimeException("Who is calling this?");
    }
}

class MenuScrollButton extends ScrollButton {
    protected MenuView menu;
    private boolean currentlyVisible;

    MenuScrollButton(MenuView menu, ColorPolicy colors,
		     int mode, boolean enabled,
		     int x, int y, int width, int height, 
		     int arrowType) {
	super(null, colors,
	      mode, enabled, x, y, width, height, arrowType);
	this.menu = menu;
	setShowing(enabled);
    }

    protected boolean isShowing() {
	return currentlyVisible;
    }

    protected void setShowing(boolean b) {
	currentlyVisible = b;
    }

    protected void setPosition(int x, int y) {
	setBounds(x, y, bounds.width, bounds.height);
    }

    protected void setEnabled(boolean enabled) {
	boolean wasEnabled = isEnabled();

	super.setEnabled(enabled);

	if (!wasEnabled && enabled) {
	    menu.buttonVisibilityChange(this);
	    currentlyVisible = true;
	}
    }

    protected void paintButton(Graphics g, boolean pressed, int align) {
	if (!currentlyVisible) {
	    return;
	}
	super.paintButton(g, pressed, Label.RIGHT);
    }
}

