/*
 * @(#)ScrollPaneView.java	1.18 99/07/29 SMI
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package sun.awt.touchable;

import java.awt.*;
import java.awt.event.AdjustmentEvent;
import sun.awt.otk.*;
import sun.porting.windowsystem.Window;

/**
 * A class to implement a scrolling view; this class creates and manages the associated
 * scrollbars as necessary.
 */
class ScrollPaneView extends TouchableContainerView implements Scrollable {

    protected Window clipwin;

    protected ScrollbarView hsb, vsb;
    protected int hsbSpace = 0;		// space currently used by horizontal scrollbar
    protected int vsbSpace = 0;		// space currently used by vertical scrollbar
    protected static final int MARGIN = DrawUtils.INTER_ELEM_PAD;

    /**
     * Creates a ScrollPaneView object as a view on the given Model, residing in the
     * specified toolkit and using the supplied ColorPolicy to do its painting.
     * Creates the underlying window as a child of the given parent.
     * @param model associated with this view.
     * @param toolkit to use.
     * @param parent to use when creating the underlying window.
     * @param colors the ColorPolicy to use.
     */
    public ScrollPaneView(ScrollPaneModel model, ObjectToolkit toolkit,
			  ContainerView parent, ColorPolicy colors) {
	this(model, toolkit, parent, colors, NULL_BOUNDS);
    }

    /**
     * Creates a ScrollPaneView object as a view on the given Model, residing in the
     * specified toolkit and using the supplied ColorPolicy to do its painting.
     * Creates the underlying window as a child of the given parent, with initial
     * bounds as indicated.
     * @param model associated with this view.
     * @param toolkit to use.
     * @param parent to use when creating the underlying window.
     * @param colors the ColorPolicy to use.
     * @param bounds to use when creating the underlying window.
     */
    public ScrollPaneView(ScrollPaneModel model, ObjectToolkit toolkit,
			  ContainerView parent, ColorPolicy colors,
			  Rectangle bounds) {

	super(model, toolkit, parent, colors, bounds);

 	clipwin = ((TouchableToolkit)toolkit).makeWindow(this.win, 0, 0, 0, 0);
 	clipwin.map();

	if (model.getScrollbarDisplayPolicy() != ScrollPane.SCROLLBARS_NEVER) {
	    // start with scrollbars; we'll lose them if we can
	    vsb = new ScrollbarView(this, Palette.CTRLS_COLORS, 
				    model.getVAdjustable(), true,
				    DrawUtils.INTER_ELEM_PAD,
				    AdjustmentEvent.BLOCK_INCREMENT);
	    hsb = new ScrollbarView(this, Palette.CTRLS_COLORS, 
				    model.getHAdjustable(), true,
				    DrawUtils.INTER_ELEM_PAD,
				    AdjustmentEvent.BLOCK_INCREMENT);

	    vsbSpace = vsb.getSize().width;
	    hsbSpace = hsb.getSize().height;
	}

 	model.setInsets(this, MARGIN, MARGIN, 
			hsbSpace + MARGIN, vsbSpace + MARGIN);
    }

    /**
     * Returns the container window used to implement scrolling.
     * @return the container window
     */    
    public sun.porting.windowsystem.Window getContainerWindow() {
	return clipwin;
    }
    
    /**
     * Paints the ScrollPaneView, including any associated scrollbars.
     * @param m The model which generated the update
     * @param g The Graphics object that should be used for painting
     */
    protected void paintView(Model m, Graphics g) {
 	if (vsbSpace != 0)
 	    vsb.paintScrollbar(g);
 	if (hsbSpace != 0)
 	    hsb.paintScrollbar(g);

	drawFocus();
    }

    /**
     * Called when bounds have been changed on the ScrollPaneView.
     * Adjusts the size of the scrollbars as necessary.
     * @param m The model that originated the change.
     * @param r Rectangle which describes the new bounds.
     */
    protected void boundsChanged(Model m, Rectangle r) {
	ScrollPaneModel pm = (ScrollPaneModel)m;
	Dimension sbSize;
	Insets insets = pm.getInsets();
	int clipWidth = r.width - insets.left - insets.right;
	int clipHeight = r.height - insets.top - insets.bottom;

	clipwin.setBounds(insets.left, insets.top, clipWidth, clipHeight);

	if (vsbSpace != 0) {
	    sbSize = vsb.getSize();
	    vsb.setBounds(r.width - sbSize.width, 
			  (clipHeight - sbSize.height)/2, 
			  sbSize.width, sbSize.height);
	}

	if (hsbSpace != 0) {
	    sbSize = hsb.getSize();
	    hsb.setBounds((clipWidth - sbSize.width)/2, 
			  r.height - sbSize.height, 
			  sbSize.width, sbSize.height);
	}
	repaintAll(m, r);
 	super.boundsChanged(m, r);
    }

    /**
     * The notification method for the Dependent interface.  Indicates that
     * an observable change has occurred on Subject m; arg is an optional 
     * argument which may help indicate what has changed.
     *
     * ScrollPaneView handles the following arg types.  Unrecognized types are
     * passed to super.
     * <pre>
     *     Adjustable - indicates that the value of one of the scrollbars has changed.
     *     A call is made to the appropriate scrollbar view in case its scroll buttons
     *     need to be enabled or disabled as a result of the new value.
     *
     *     Dimension - indicates that the ScrollPane child has been resized. If this
     *     results in a change to the scrollbar layout, a call is made to boundsChanged()
     *     to advise theView.
     * </pre>
     * @param m The Subject (model) whose state has changed
     * @param arg An argument indicating the substance of the change.  May be null
     */
    public void updateDependent(Subject m, Object arg) {

	if (arg instanceof Adjustable) {
	    // this is from ScrollPaneModel.setValue - have scrollbar check 
	    // if buttons need to be enabled or disabled as a result of new value
	    switch (((Adjustable)arg).getOrientation()) {
	    case Adjustable.VERTICAL:
		if (vsb != null) vsb.updateEnabled();
		break;
	    case Adjustable.HORIZONTAL:
		if (hsb != null) hsb.updateEnabled();
		break;
	    default:
		break;
	    }
	    return;
	} else if (arg instanceof Dimension) {
	    if (sbLayoutChanged((ScrollPaneModel)m, (Dimension)arg))
		boundsChanged((Model)m, getBounds());	
	    return;
	}
	super.updateDependent(m, arg);
    }

    private boolean sbLayoutChanged(ScrollPaneModel pm, 
				    Dimension childSize) {

	ScrollPaneModel model = (ScrollPaneModel)getModel();
	int sbDisplayPolicy = model.getScrollbarDisplayPolicy();
	Dimension spSize = model.getSize();
	int vsbWidth = getVScrollbarWidth();
	int hsbHeight = getHScrollbarHeight();
	int oldHSpace = hsbSpace;
	int oldVSpace = vsbSpace;

	if (sbDisplayPolicy == ScrollPane.SCROLLBARS_AS_NEEDED) {
	    hsbSpace = 
		childSize.width <= (spSize.width  - 2*MARGIN) ? 0 : hsbHeight;
	    vsbSpace = 
		childSize.height <= (spSize.height - 2*MARGIN) ? 0 : vsbWidth;

	    if (hsbSpace == 0 && vsbSpace != 0) {
		hsbSpace = 
		    (childSize.width <= (spSize.width - vsbWidth - 2*MARGIN)) ?
		    0 : hsbHeight;
	    }
	    if (vsbSpace == 0 && hsbSpace != 0) {
		vsbSpace = 
		    (childSize.height <= 
		     (spSize.height - hsbHeight - 2*MARGIN)) ? 0 : vsbWidth;
	    }
	} else if (sbDisplayPolicy == ScrollPane.SCROLLBARS_ALWAYS) {
	    vsbSpace = vsbWidth;
	    hsbSpace = hsbHeight;
	} else {
	    vsbSpace = hsbSpace = 0;
	}

	// If we hid either of the scrollbars, but left ourselves
	// scrolled off of the top and/or right of the pane, we need
	// to scroll to the top and/or right of to make it visible.
	if (vsbSpace == 0)
	    model.getVAdjustable().setValue(0);
	if (hsbSpace == 0)
	    model.getHAdjustable().setValue(0);

	pm.setInsets(this, MARGIN, MARGIN, 
		     hsbSpace + MARGIN, vsbSpace + MARGIN);

	// have scrollbar check if buttons need to be enabled or 
	// disabled as a result of changes to child
	// window size (max) or clipwin size (vis),
	// by calling updateEnabled(min, max, vis
	if (vsbSpace != 0)
	    vsb.updateEnabled(0, childSize.height, 
			      spSize.height - hsbSpace - 2*MARGIN);
	if (hsbSpace != 0)
	    hsb.updateEnabled(0, childSize.width, 
			      spSize.width - vsbSpace - 2*MARGIN);

	// set background colour here, since ScrollPane child may not 
	// exist yet when initialize is called
	BasicColorPolicy colors = (BasicColorPolicy)getColorPolicy();
	model.setChildBackground(colors.getForeground(pm, 
						 pm.isEnabled(),
						 false, false));
	return ((oldHSpace != hsbSpace) || (oldVSpace != vsbSpace));
    }

    /**
     * Start scrolling the view.
     * @param x the x coordinate associated with the event
     * @param y the y coordinate associated with the event
     * @param modifiers The modifier key (SHIFT, CONTROL, ALT, META) state
     * @return true if this event is consumed by the view, false otherwise
     */
    public boolean startScrolling(int x, int y, int modifiers) {
	if (vsbSpace != 0 && 
	    vsb.startScrolling(x, y, modifiers)){
	    drawFocus();
	    return true;
	} else if (hsbSpace != 0&& hsb.startScrolling(x, y, modifiers)){
	    drawFocus();
	    return true;
	}
	return false;
    }

    /**
     * Stop scrolling the view.
     * @return true if this event is consumed by the view, false otherwise
     */
    public boolean stopScrolling() {
	if (vsbSpace != 0 && vsb.stopScrolling()){
	    drawFocus();
	    return true;
	} else if (hsbSpace != 0 && hsb.stopScrolling()){
	    drawFocus();
	    return true;
	}
	return false;
    }

    /**
     * An accessor method for the scrollbar dimensions associated with this view.
     * @return A Dimension object giving the vertical scrollbar width and the
     * horizontal scrollbar height
     */
    public Dimension getScrollbarDimensions() {
	return new Dimension(getVScrollbarWidth(), getHScrollbarHeight());
    }

    private int getHScrollbarHeight() {
	return (hsb != null) ? hsb.getSize().height : 0;
    }
    
    private int getVScrollbarWidth() {
	return (vsb != null) ? vsb.getSize().width : 0;
    }
}
