/*
 * @(#)ScrollableTextView.java	1.33 98/12/03
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

package sun.awt.touchable;

import java.awt.*;
import sun.awt.otk.*;
import java.awt.event.KeyEvent;
import java.awt.event.ActionEvent;
import java.awt.event.InputEvent;
import java.awt.event.AdjustmentEvent;
import java.awt.event.AdjustmentListener;
import sun.porting.toolkit.ToolkitEventHandler;

public class ScrollableTextView extends ScrollableTextAreaView {

    protected Rectangle dismissRect;                  // Dismiss button area
    protected boolean haveDismissButton;              // true if we have one
    // my state vars
    protected boolean dismissPressed;

    /**
     * Create a modal TextComponent view. This assumes that model is of
     * type TextComponentModel. A dismiss button is not created.
     * @param model Our text component model
     * @param toolkit Our toolkit reference
     * @param colors Our color policy
     */
    public ScrollableTextView(TextComponentModel model, ObjectToolkit toolkit, 
			      ContainerView parent,
			      ColorPolicy colors) {
	this(model, toolkit, parent, colors, NULL_BOUNDS, false);
    }

    /**
     * Create a modal TextComponent view. This assumes that model is of
     * type TextComponentModel. A dismiss button is not created.
     * @param model Our text component model
     * @param toolkit Our toolkit reference
     * @param colors Our color policy
     * @param bounds Bounds to use for the popup. If NULL_BOUNDS is specified
     * then use the values the dimensions returned by getPreferredSize() 
     * and position the popup centered horizontally and bottom-aligned 
     * veritcally. If zero width and height are specified, the (x, y) point
     * is used as a hint to place the popup, but we guarantee that it will fit
     * on the screen.
     */
    public ScrollableTextView(TextComponentModel model, ObjectToolkit toolkit,
			      ContainerView parent,
			      ColorPolicy colorPolicy, Rectangle bounds) {

        this(model, toolkit, parent, colorPolicy, bounds, false);
    }

    /**
     * Create a modal TextComponent view. This assumes that model is of
     * type TextComponentModel.
     * @param model Our text component model
     * @param toolkit Our toolkit reference
     * @param colors Our color policy
     * @param bounds Bounds to use for the popup. If NULL_BOUNDS is specified
     * then use the values the dimensions returned by getPreferredSize() 
     * and position the popup centered horizontally and bottom-aligned 
     * veritcally. If zero width and height are specified, the (x, y) point
     * is used as a hint to place the popup, but we guarantee that it will fit
     * on the screen.
     * @param createDismissButton If true a dismiss button will be created.
     */
    public ScrollableTextView(TextComponentModel model, ObjectToolkit toolkit,
			      ContainerView parent,
			      ColorPolicy colorPolicy, Rectangle bounds,
			      boolean createDismissButton) {

	super(model, toolkit, parent, colorPolicy, bounds);

	// we're not associated to ComponentModel so we set font manually
	super.updateDependent(model, model.getFont());

	// add our controllers
	addController(CvtVirtual.controller);
	addController(CutAndPaste.controller);
	addController(HorizPage.controller);
	addController(VertPage.controller);
	addController(HorizArrows.controller);
	addController(VertArrows.controller);
	addController(Backspace.controller);
	addController(TextPassThru.controller);
	addController(TextCatchall.controller);

	// set attributes from model
	editable = model.isEditable();
	
	// we're selected by virtue of being created
	selected = true;

	// create our text pane
	textpane = getTextPane();
	textpane.setEditable(editable);
	textpane.setFocus(selected);

	getContainerWindow().setBackgroundColor(colorPolicy.getEraseColor(null,
					     false, model.isEnabled(), false));

	// create the vertical scrollbar
	vsb = getVerticalScrollbar(vsb, model);

	this.haveDismissButton = createDismissButton;

	setBounds(bounds.x, bounds.y, bounds.width, bounds.height);
    }

    public Dimension getPreferredSize(int rows, int cols) {

	Dimension tp = new Dimension(0, 0);
        if (textpane != null) {
	    tp = textpane.getMinimumSize(DrawUtils.TEXTAREA_MIN_ROWS,
					 DrawUtils.TEXTAREA_MIN_COLS);
	}

        Dimension sb = getScrollbarDimensions(vsb);

	int w = DrawUtils.INTER_ELEM_PAD + tp.width + sb.width;
	int ctlHeight;
	if (haveDismissButton) {
	    ctlHeight = sb.height + dismissRect.height +
	                (DrawUtils.INTER_ELEM_PAD * 2);
	} else {
	    ctlHeight = sb.height;
	}

	return new Dimension(w, Math.max(tp.height, ctlHeight));
    }

    public void setBounds(int x, int y, int width, int height) {
        //System.err.println("+ScrollableTextView.setBounds: x=" + x + " y=" + y + " w=" + width + " h=" + height);
        
        super.setBounds(x, y, width, height);

        if (width > 0 || height > 0) {
	    
	    getComponentRects(width, height);

	    // set textRect
	    if (textpane != null) {
	        textpane.setBounds(textRect.x, textRect.y,
				   textRect.width, textRect.height);
	    }

	    // set the vertical scrollbar
	    if (vsb != null) {
	        vsb.setBounds(vScrollRect.x, vScrollRect.y,
			      vScrollRect.width, vScrollRect.height);
	    }
	}
	//System.err.println("-ScrollableTextView.setBounds: textR=" + textRect + " vScrollR=" + vScrollRect + " dismissR=" + dismissRect);
    }

    private void getComponentRects(int width, int height) {
	Dimension vScroll = getScrollbarDimensions(vsb);
	if (textRect == null) {
	    textRect = new Rectangle(0, 0,
				     width - DrawUtils.INTER_ELEM_PAD -
				         vScroll.width,
				     height);
	} else {
	    textRect.setSize(width - DrawUtils.INTER_ELEM_PAD -
				         vScroll.width,
				     height);
	}

	if (haveDismissButton) {
	    if (vScrollRect == null) {
	        vScrollRect = new Rectangle(textRect.x + textRect.width +
					    DrawUtils.INTER_ELEM_PAD,
					    textRect.y,
					    vScroll.width, vScroll.height);
	    } else {
	        vScrollRect.setLocation(textRect.x + textRect.width +
					DrawUtils.INTER_ELEM_PAD,
					textRect.y);
	    }
	} else {
	    if (vScrollRect == null) {
	        vScrollRect = new Rectangle(textRect.x + textRect.width +
					    DrawUtils.INTER_ELEM_PAD,
					    textRect.y +
					        ((textRect.height -
						  vScroll.height) / 2),
					    vScroll.width, vScroll.height);
	    } else {
	        vScrollRect.setLocation(textRect.x + textRect.width +
					DrawUtils.INTER_ELEM_PAD,
					textRect.y + ((textRect.height -
						       vScroll.height) / 2));
	    }
	}

	if (haveDismissButton) {
	    if (dismissRect == null) {
	        dismissRect = new Rectangle(textRect.x + textRect.width +
					    DrawUtils.INTER_ELEM_PAD,
					    textRect.y + textRect.height -
					    DrawUtils.CONTROL_BUTTON_H,
					    DrawUtils.CONTROL_BUTTON_W,
					    DrawUtils.CONTROL_BUTTON_H);
	    } else {
	        dismissRect.setLocation(textRect.x + textRect.width +
					DrawUtils.INTER_ELEM_PAD,
					textRect.y + textRect.height -
					DrawUtils.CONTROL_BUTTON_H);
	    }
	}
	//System.err.println("-ScrollableTextView.setBounds: textR=" + textRect + " vScrollR=" + vScrollRect + " dismissR=" + dismissRect);
    }

    public boolean handleInputEvent(View v,
				    int id,  long when, int modifiers,
				    int x, int y, int arg1, int arg2) {
        
        //System.err.println("ScrollableTextView.handleInputEvent(): v=" + v + " id=" + id + " x=" + x + " y=" + y + " arg1=" + arg1 + " arg2=" + arg2);
        boolean ret = false;

        switch(id) {

	case ToolkitEventHandler.POINTER_PRESSED:
	    if (haveDismissButton) {
	        if (dismissRect.contains(x, y)) {
		    if (arg1 > 0) {
		        dismissPressed = true;
			drawDismissButton(dismissPressed);
			ret = true;
		    }
		} else {
	            dismissPressed = false;
		}
	    }
	    break;

	case ToolkitEventHandler.POINTER_RELEASED:
	    if (haveDismissButton) {
	        if (dismissRect.contains(x, y)) {
		    if (dismissPressed) {
		        dismissPressed();
		    }
		    ret = true;
		}
	    }
	    break;

	case ToolkitEventHandler.POINTER_DRAGGED:
	    if (haveDismissButton) {
	        if (dismissPressed) {      // eat them all
		    if (!dismissRect.contains(x, y)) {
		        drawDismissButton(false);
		    } else {
		        drawDismissButton(true);
		    }
		    ret = true;
		}
	    }
	    break;

	default:
	    ret = false;
  	    break;
	}

	if (!ret) {
	    return super.handleInputEvent(v, id, when, modifiers,
					  x, y, arg1, arg2);
	}
	
	return ret;
    }

    protected void dismissPressed() {

        if (haveDismissButton) {
	    dismissPressed = false;
	    drawDismissButton(dismissPressed);
	}

	sendActionEvent();
	sendHideEvent();
    }


    public void sendActionEvent() {
	((TextComponentModel)getModel()).sendActionEvent();
    }

    public void sendHideEvent() {
	VirtualComponentMessage arg = null;
	if (getModel() instanceof TextComponentPeerModel) {
	    arg = getSelectionInfo();
	}
	((TextComponentModel)getModel()).sendHideEvent(arg);
    }


    protected VirtualComponentSelInfo getSelectionInfo() {
        return new VirtualComponentSelInfo(getSelectionStart(),
					   getSelectionEnd(),
					   getCaretPosition());
    }



    protected void paintView(Model m, Graphics g) {

        if (textpane != null) {
	    ((TextComponentModel) getModel()).getDataModel().setLineWidth(
				         textpane.getMaxPossibleStringWidth());
	}
        
        super.paintView(m, g);

	if (haveDismissButton) {
	    if (dismissRect != null) {
	        Rectangle oldClip = (Rectangle) g.getClipBounds();

		// draw the dismiss button
		g.setClip(oldClip);
		drawDismissButton(false);

		// restore graphics properties
		g.setClip(oldClip);
	    }
	}
    }

    /**
     * If we don't have a textpane yet, create one, and then return it.
     * @param tc Our text component
     * @return Data model for this component
     */
    public TextPane getTextPane() {

        if (textpane == null) {
	    TextComponentModel model = (TextComponentModel)getModel();
	    textpane = new TouchableMultiLinePane(
			    this, model,
			    getToolkit(), Palette.TEXTPOPUP_TEXTPANE_COLORS,
			    DrawUtils.TA_INSETS, 
			    DrawUtils.LEFT_MARGIN,
			    DrawUtils.RIGHT_MARGIN, 
			    true);
	}
	
	return textpane;
    }

    protected sun.porting.windowsystem.Window getGrabbingWindow() {
        return ((TouchableToolkit) getToolkit()).getGrabbingWindow();
    }

    protected void drawDismissButton(boolean pressed) {
        
        if (!haveDismissButton)
	    return;
      
        Graphics g = null;
	try {
	    g = getGraphics();
	    g.translate(dismissRect.x, dismissRect.y);

	    g.setColor(Palette.CTRLS_COLORS.getBackground(null, pressed,
							  true, false));
	    g.fillRect(0, 0, dismissRect.width, dismissRect.height);

	    int x = (dismissRect.width - DrawUtils.DIAMOND_SIZE)/2;
	    int y = (dismissRect.height - DrawUtils.DIAMOND_SIZE)/2;
	    g.translate(x, y);

	    g.setColor(Palette.CTRLS_COLORS.getForeground(null, pressed,
							  true, false));
	    int nVerts = Math.max(DrawUtils.diamondXpoints.length, 
				  DrawUtils.diamondYpoints.length);
	    g.fillPolygon(DrawUtils.diamondXpoints, 
			  DrawUtils.diamondYpoints, nVerts);
	    g.drawPolygon(DrawUtils.diamondXpoints, 
			  DrawUtils.diamondYpoints, nVerts);
	    Color outline2 = Palette.CTRLS_COLORS.getOutline2(pressed, true);
	    if (outline2 != null){
		g.setColor(outline2);
		g.drawLine(DrawUtils.diamondXpoints[nVerts - 2], 
			   DrawUtils.diamondYpoints[nVerts -2],
			   DrawUtils.diamondXpoints[nVerts - 1], 
			   DrawUtils.diamondYpoints[nVerts - 1]);
	    }

	    g.translate(-x-dismissRect.x, -y -dismissRect.y);
	} finally {
	    if (g != null) {
	        g.dispose();
	    }
	}
    }

    public void updateDependent(Subject m, Object arg) {

        if (!isVisible()) {
	    return;
	}

	if (arg instanceof AWTEvent) {
	    AWTEvent e = (AWTEvent) arg;

	    /* handle an event from AWT */
	    handleAWTEvent((Model)m, e);
	    if (e instanceof InputEvent) {
		((InputEvent)e).consume();
	    }
	    return;
	}

        super.updateDependent(m, arg);
    }

    public int lineNumberToY(int lineNum) {
	return (textRect.y + textpane.lineNumberToY(lineNum));
    }
}
