/*
 * @(#)ScrollbarView.java	1.17 98/11/30 SMI
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package sun.awt.touchable;

import java.awt.*;
import java.awt.event.AdjustmentEvent;
import sun.awt.otk.BasicView;
import sun.awt.otk.ColorPolicy;
import com.sun.util.*;

class ScrollbarView implements PTimerWentOffListener {
    protected ScrollButton buttons[] = null;
    protected ScrollButton currentButton = null;

    // if padding between buttons is not a constant 
    // then it should be set to 0;
    // that way getSize will return size without padding
    private int inter_button_pad;

    private BasicView view;
    private boolean enabled;
    protected Adjustable adj;
    protected Rectangle bounds = null;

    /**
     * The lock that controls scroll repeating
     * @see #startScrolling
     * @see ScrollRepeater
     */
    protected static final Object SCROLL_LOCK = new Object();

    /**
     * Invalid Scrollbar adjustment mode
     */
    protected static final int INVALID_MODE = -1;

    private PTimerSpec spec;
    private PTimer    timer;

    /**
     * Creates a ScrollbarView object with 2 buttons, using the givenView to paint into,
     * the supplied ColorPolicy to do its painting, the associated Adjustable object
     * for state data, and with the specified initial enabled state.
     * ScrollbarViews have no associated Model.
     * @param view the view to paint into.
     * @param colors the ColorPolicy to use.
     * @param adj the corresponding Adjustable object
     * @param enabled if true, the scrollbar is initially enabled; otherwise it is disabled.
     */
    ScrollbarView(BasicView view, ColorPolicy colors,
		  Adjustable adj, boolean enabled) {
	this(view, colors, adj, enabled, DrawUtils.INTER_ELEM_PAD);
    }

    /**
     * Creates a ScrollbarView object with 2 buttons, using the givenView to paint into,
     * the supplied ColorPolicy to do its painting, the associated Adjustable object
     * for state data, and with the specified initial enabled state, and amount of
     * padding between buttons.
     * ScrollbarViews have no associated Model.
     * @param view the view to paint into.
     * @param colors the ColorPolicy to use.
     * @param adj the corresponding Adjustable object
     * @param enabled if true, the scrollbar is initially enabled; otherwise it is disabled.
     * @param inter_button_pad padding between this ScrollbarView's buttons;
     * if padding between buttons is not a constant then it should be set to 0.
     */
    ScrollbarView(BasicView view, ColorPolicy colors, 
		  Adjustable adj, boolean enabled,
		  int inter_button_pad) {
	this(view, colors, adj, enabled, inter_button_pad,
	     AdjustmentEvent.UNIT_INCREMENT);
    }

    /**
     * Creates a ScrollbarView object with 2 buttons, using the givenView to paint into,
     * the supplied ColorPolicy to do its painting, the associated Adjustable object
     * for state data, and with the specified initial enabled state, amount of
     * padding between buttons, and button increment type.
     * ScrollbarViews have no associated Model.
     * @param view the view to paint into.
     * @param colors the ColorPolicy to use.
     * @param adj the corresponding Adjustable object
     * @param enabled if true, the scrollbar is initially enabled; otherwise it is disabled.
     * @param inter_button_pad padding between this ScrollbarView's buttons;
     * if padding between buttons is not a constant then it should be set to 0.
     * @param posIncr increment type for the ScrollbarView's buttons;
     * AdjustmentEvent.UNIT_INCREMENT or BLOCK_INCREMENT. 
     * The corresponding decrement type will also be used.
     * @see java.awt.event.AdjustmentEvent
     */
    ScrollbarView(BasicView view, ColorPolicy colors, 
		  Adjustable adj, boolean enabled,
		  int inter_button_pad, int posIncr) {

	int negIncr = (posIncr == AdjustmentEvent.UNIT_INCREMENT) ?
	    AdjustmentEvent.UNIT_DECREMENT : AdjustmentEvent.BLOCK_DECREMENT;

	Dimension minSize = ScrollButton.minimumSize();
	
	buttons = new ScrollButton[2];

	this.view = view;
	this.adj = adj;
	this.enabled = enabled;
	this.inter_button_pad = inter_button_pad;

	if (adj.getOrientation() == Adjustable.HORIZONTAL) {

	    buttons[0] = new ScrollButton(this, colors,
					  negIncr, false, 
					  0, 0, 
					 minSize.width, minSize.height,
					 DrawUtils.LEFT);
	    buttons[1] = new ScrollButton(this, colors,
					  posIncr, enabled,
					  minSize.width + inter_button_pad, 0,
					  minSize.width, minSize.height,
					  DrawUtils.RIGHT);
	} else {

	    buttons[0] = new ScrollButton(this, colors,
					  negIncr, false,
					  0, 0, 
					  minSize.width, minSize.height,
					  DrawUtils.UP);

	    buttons[1] = new ScrollButton(this, colors, 
					  posIncr, enabled,		  
					  0, minSize.height + inter_button_pad, 
					  minSize.width, minSize.height,
					  DrawUtils.DOWN);	    
	}

 	bounds = new Rectangle(getSize());
	updateEnabled();
    } 

    /**
     * Creates a ScrollbarView object with the given array of buttons,
     * using the givenView to paint into, the associated Adjustable object
     * for state data, and with the specified initial enabled state.
     * ScrollbarViews have no associated Model.
     * @param buttons the array of ScrollButtons to use.
     * @param view the view to paint into.
     * @param adj the corresponding Adjustable object
     * @param enabled if true, the scrollbar is initially enabled; otherwise it is disabled.
     */
    ScrollbarView(BasicView view, 
		  Adjustable adj, boolean enabled,
		  ScrollButton buttons[]) {
 	this.buttons = buttons;
	this.view = view;
	this.adj = adj;
 	this.enabled = enabled;
 	inter_button_pad = 0;
  	bounds = new Rectangle();
     } 

    /**
     * Paints the ScrollbarView.
     * @param g The Graphics object that should be used for painting
     */
    protected void paintScrollbar(Graphics g) {
	for (int i = 0; i < buttons.length; i++)
	    buttons[i].paintButton(g, false);
    }

    /**
     * Determines whether this ScrollbarView is enabled. An enabled scrollbar 
     * can respond to user input and generate events.
     *A scrollbar may be enabled or disabled by calling its setEnabled method.
     * @return true if the scrollbar is enabled; false otherwise
     * @see #setEnabled
     */
    protected boolean isEnabled() {
 	return (enabled);
    }

    /**
     * Enables or disables this ScrollbarView, depending on the value of the 
     * parameter enabled. An enabled scrolbar can respond to user 
     * input and generate events.
     * @param enabled if true, this scrollbar is enabled; otherwise it is disabled.
     * @see #isEnabled
     */
    protected void setEnabled(boolean enabled) {
	this.enabled = enabled;
	// TODO - need to set/paint buttons or something...
    }

    /**
     * Set the bounds of the ScrollbarView, relative to the view's parent; use the toolkit
     * default padding value for the amount of padding between buttons.
     * (DrawUtils.INTER_ELEM_PAD)
     * @param x The new x coordinate for the view's left edge
     * @param y The new y coordinate for the view's top edge
     * @param width The new width of the view
     * @param height The new height of the view
     */
    protected void setBounds(int x, int y, int width, int height) {
	setBounds(x, y, width, height, DrawUtils.INTER_ELEM_PAD);
    }

    /**
     * Set the bounds of the ScrollbarView, relative to the view's parent; use the specified
     * value for the amount of padding between buttons.
     * @param x The new x coordinate for the view's left edge
     * @param y The new y coordinate for the view's top edge
     * @param width The new width of the view
     * @param height The new height of the view
     * @param inter_button_pad padding between this ScrollbarView's buttons
     */
    protected void setBounds(int x, int y, int width, int height, 
			     int inter_button_pad) {
	Dimension minSize = ScrollButton.minimumSize();
	int x_offset = x;
	int y_offset = y;

	for(int i = 0; i < buttons.length; i++) {
	    buttons[i].setBounds(x_offset, y_offset, 
				 minSize.width, minSize.height);

	    if (adj.getOrientation() == Adjustable.VERTICAL)
		y_offset += buttons[i].getBounds().height + inter_button_pad;
	    else
		x_offset += buttons[i].getBounds().width + inter_button_pad;
	}

	if (adj.getOrientation() == Adjustable.VERTICAL)
	    bounds.setBounds(x, y, minSize.width, 
			     y_offset - y - inter_button_pad);
	else
	    bounds.setBounds(x, y, x_offset - x - inter_button_pad, 
			     minSize.height);
    }

    /**
     * Compute the minimum size for the scrollbar, using the amount of padding
     * between buttons originally specified to the constructor.
     * @return a Dimension object giving the scrollbar size
     */
    public Dimension getSize() {
	return (getSize(inter_button_pad));
    }

    /**
     * Compute the minimum size for the scrollbar, using the specified amount of
     * padding between buttons.
     * @param inter_button_pad padding between this ScrollbarView's buttons
     * @return a Dimension object giving the scrollbar size
     */
    public Dimension getSize(int inter_button_pad) {
	Dimension minSize = ScrollButton.minimumSize();
	if (adj.getOrientation() ==  Adjustable.VERTICAL) {
      	    return new Dimension(minSize.width, 
				 2*minSize.height + inter_button_pad);
	} else {
      	    return new Dimension(2*minSize.width + inter_button_pad, 
				 minSize.height);
	}
    }

    /**
     * Get the physical bounds (relative to parent) associated with this ScrollbarView.
     * @return A Rectangle describing the bounds of the view
     */
    public Rectangle getBounds() {
        return bounds;
    }

    private int scrollCount;

    /**
     * The PTimer listener, for implementing scrolling.
     * @param e the event specifying which timer and which timer specification
     * went off.
     * @see com.sun.util.PTimer
     * @see com.sun.util.PTimerWentOffListener
     */
    public void timerWentOff(PTimerWentOffEvent e) {
	/*
	 * THIS IS A COMPROMISE.   The basic issue is that performing
	 * the scroll itself can be very quick, e.g. on a list or menu,
	 * or it can take a lot of time and generate a lot of expose events,
	 * e.g. on a ScrollPane.  The desired behavior is for the button release
	 * to be detected immediately, but if a bunch of expose events were
	 * enqueued in front of it, a delay occurs.
	 *
	 * What we do is scroll the first time (i.e. at 1x delay), and then
	 * skip some cycles (until 4x delay).  When coupled with a check for
	 * (scrollCount == 0) in stopScrolling, this results in the behavior
	 * that a quick single click is treated properly, even when the scroll
	 * operation takes a long time; also, clicks in quick succession are
	 * handled adequately.  When scroll repeat eventually kicks in, we
	 * still display the problem that if the scroll operation generates
	 * lots of expose events, the "up" will be detected late, possibly
	 * resulting in an overscroll.  But unless we want to devise a way
	 * to get up events delivered before expose events (risky!!!) this
	 * is probably the best that we can do.
	 */
	if ((scrollCount == 0) || (scrollCount > 3)) {
	    scroll();
	}
	++scrollCount;
    }

    /**
     * Start scrolling, if the given event was on this scrollbar.
     * Starts a PTimer thread to control further scrolling if the button is held down.
     * @param x the x coordinate associated with the event
     * @param y the y coordinate associated with the event
     * @param modifiers The modifier key (SHIFT, CONTROL, ALT, META) state
     * @return true if this event is consumed by the scrollbar, false otherwise
     * @see #timerWentOff
     * @see com.sun.util.PTimer
     * @see com.sun.util.PTimerWentOffListener
     */
    public boolean startScrolling(int x, int y, int modifiers) {
	if (!isEnabled())
	    return false;

	if (currentButton == null && 
	    (modifiers & java.awt.event.InputEvent.BUTTON1_MASK) != 0) {

	    for (int i = 0; i < buttons.length; i++) {
		if (buttons[i].isEnabled() && buttons[i].inButton(x,y)) {

		    currentButton = buttons[i];
		    Graphics g = null;
		    try {
			g = view.getGraphics();
			currentButton.paintButton(g, true);
		    } finally {
			if (g != null)
			    g.dispose();
		    }

		    /*
		     * Do all scrolling on the timer thread, so that
		     * we get consistent behavior.
		     */
		    scrollCount = 0;
		    timer = PTimer.getTimer();
		    spec = new PTimerSpec();
		    spec.setDelayTime(200);
		    spec.setRepeat(true);
		    spec.addPTimerWentOffListener(this);
		    timer.schedule(spec);

		    return true;
		}
	    }
	}
	return false;
    }

    /**
     * Stop scrolling, reset scroll button appearance as appropriate. 
     * Stops the PTimer thread started by startScrolling.
     * @return true if this event is consumed by the scrollbar, false otherwise
     * @see #startScrolling
     */
    public boolean stopScrolling() {
	if (!isEnabled())
	    return false;

	if (currentButton != null) {
	    timer.deschedule(spec);
	    if (scrollCount == 0) {
		/* make sure we scroll at least once! */
		scroll();
	    }

	    Graphics g = null;
	    try {
		g = view.getGraphics();
		currentButton.paintButton(g, false);
	    } finally {
		if (g != null)
		    g.dispose();
	    }
	    currentButton = null;
	    return true;
	}
	return false;
    }

    /**
     * Returns the current button's scrolling mode, ie. the increment/decrement type.
     * @return the current button's scrolling mode
     * @see java.awt.event.AdjustmentEvent
     */
    protected int getScrollingMode() {
	return (currentButton != null) ? 
	    currentButton.getMode() : INVALID_MODE;
    }
    
    /**
     * Scroll one unit, as indicated by the button scrolling mode. Call setValue on
     * the corresponding Adjustable object to notify it of the change.
     */
    protected void scroll() {
	int oldValue = adj.getValue();
	int newValue;

        switch (getScrollingMode()) {
        case AdjustmentEvent.UNIT_DECREMENT:
	    newValue = oldValue - adj.getUnitIncrement();
	    break;

        case AdjustmentEvent.UNIT_INCREMENT:
            newValue = oldValue + adj.getUnitIncrement();
	    break;

        case AdjustmentEvent.BLOCK_DECREMENT:
	    newValue = oldValue - adj.getBlockIncrement();
            break;

        case AdjustmentEvent.BLOCK_INCREMENT:
	    newValue = oldValue + adj.getBlockIncrement();
            break;
	default:
	    return;
        }
	adj.setValue(newValue);

	if (adj.getValue() != oldValue) {
	    Sounds.play(Sounds.MOUSE_PRESS);
	}
    }

    /**
     * Updates the enabled/disabled state of the scrollbar as necessary, based on
     * the current minimum, maximum, and visible amount values of the associated 
     * Adjustable object.
     * @see java.awt.Adjustable
     */
    protected void updateEnabled() {
	updateEnabled(adj.getMinimum(), adj.getMaximum(), 
		      adj.getVisibleAmount());
    }

    /**
     * Updates the enabled/disabled state of the scrollbar as necessary, based on
     * the given minimum, maximum, and visible amount values.
     * @param min the minimum value of the scrollbar
     * @param max the maximum value of the scrollbar
     * @param vis the length of the proportional indicator of the scrollbar
     * @see java.awt.Adjustable
     */
    protected void updateEnabled(int min, int max, int vis) {
	// Check if any of the buttons changed their enabled state
	Graphics g = null;
	try {
	    for (int i = 0; i < buttons.length; i++) {
		boolean newEnabled = 
		    shouldBeEnabled(buttons[i], min, max, vis);
		if (buttons[i].isEnabled() != newEnabled) {	
		    if (g == null) {
			g = view.getGraphics();
		    }

		    buttons[i].setEnabled(newEnabled);
		    buttons[i].paintButton(g, false);
		}
	    }
	} finally {
	    if (g != null)
		g.dispose();
	}
    }

    private boolean shouldBeEnabled(ScrollButton button, 
				    int min, int max, int vis) {
	if (!isEnabled())
	    return false;

	int mode = button.getMode();
	int value = adj.getValue();

	// Already at minimum
	if (value == min &&
	    (mode == AdjustmentEvent.UNIT_DECREMENT ||
	     mode == AdjustmentEvent.BLOCK_DECREMENT)) {
	    return false;
	}
	
	// Already at maximum
	if ((value == (max - vis)) && 
	    (mode == AdjustmentEvent.UNIT_INCREMENT || 
	     mode == AdjustmentEvent.BLOCK_INCREMENT)) {
	    return false;
	}
	return true;
    }
}
