/*
 * @(#)TextAreaView.java	1.27 98/11/09
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

package sun.awt.touchable;

import java.awt.*;
import sun.awt.otk.*;
import java.awt.event.KeyEvent;
import java.awt.event.AdjustmentListener;
import java.awt.event.AdjustmentEvent;
import sun.porting.toolkit.ToolkitEventHandler;

class TextAreaView extends ScrollableTextAreaView {

    // cache scrollbar visibility as some looks may support only a few kinds.
    protected int sbVisibility;
    protected boolean wordwrap;

    public TextAreaView(TextAreaModel model, ObjectToolkit toolkit,
			ContainerView parent, ColorPolicy colors) {
	this(model, toolkit, parent, colors, NULL_BOUNDS);
    }

    public TextAreaView(TextAreaModel model, ObjectToolkit toolkit, 
			ContainerView parent, ColorPolicy colors,
			Rectangle bounds) {

	super(model, toolkit, parent, colors, bounds);

        // initialize scrollbarVisibility - don't set it directly so that
	// we can stub code for all cases in initializeScrollbars() but
	// over-ride behavior for our "look" here
        switch(model.getScrollbarVisibility()) {
        case TextArea.SCROLLBARS_NONE:
        case TextArea.SCROLLBARS_HORIZONTAL_ONLY:
	    sbVisibility = TextArea.SCROLLBARS_NONE;
	    break;

        case TextArea.SCROLLBARS_BOTH:
        case TextArea.SCROLLBARS_VERTICAL_ONLY:
        default:
	    sbVisibility = TextArea.SCROLLBARS_VERTICAL_ONLY;
	    break;
	}

	// now create the scrollbars
	initializeScrollbars(model);

	// set our bounds properly
	boundsChanged(model, bounds);
    }
    

    /**
     * This method computes the minimum size required for the given 
     * number of columns/rows of text
     * @param rows Number of rows
     * @param cols Number of cols
     */
    public synchronized Dimension getMinimumSize(int rows, int cols) {

	if ((rows < 0) || (cols < 0)) {
	    rows = DrawUtils.TEXTAREA_MIN_ROWS;
	    cols = DrawUtils.TEXTAREA_MIN_COLS;
	}
	
	// We always returnm inimumSize as if there was a scrollbar
	// present, otherwise for no scrollbar case the size of the 
	// textarea will not agree with the scrollbar case and cause
	// indeterminate layout behavior.
        Dimension tp = new Dimension(0, 0);
        if (textpane != null) {
	    tp = textpane.getMinimumSize(rows, cols);
	}
        Dimension sb = new Dimension(0, 0);
        if (vsb != null) {
	    sb = vsb.getSize();
	}
	int w = tp.width + DrawUtils.INTER_ELEM_PAD + sb.width;
	int h = Math.max(tp.height, sb.height);

	return new Dimension(w, h);
    }

    protected void boundsChanged(Model m, Rectangle r) {

        /*System.err.println("+boundsChanged(): m=" + m + " r=" + r);*/

        if (r != null) {
	    Dimension vScroll = getScrollbarDimensions(vsb);
	    Dimension hScroll = getScrollbarDimensions(hsb);
	
	    // set textRect 
	    Rectangle t = new Rectangle(0, 0, r.width -
					      DrawUtils.INTER_ELEM_PAD -
				              vScroll.width,
					r.height - hScroll.height);

	    if (textRect == null) {
	        textRect = t;
	    } else if (!textRect.equals(t)) {
	        textRect = t;
	    }

	    // set the textpane
	    if (textpane != null) {
	        textpane.setBounds(textRect.x, textRect.y,
				   textRect.width, textRect.height);
	    }

	    // set the vertical scrollbar
	    vScrollRect =
	        new Rectangle(textRect.x + textRect.width + 
			           DrawUtils.INTER_ELEM_PAD,
			      textRect.y +
			           (textRect.height - vScroll.height) / 2,
			      vScroll.width, vScroll.height);
	    if (vsb != null) {
	        vsb.setBounds(vScrollRect.x, vScrollRect.y,
			      vScrollRect.width, vScrollRect.height);
	    }

	    // set the horizontal scrollbar
	    //    - minumum stuff needed for TouchableLook
	    hScrollRect = new Rectangle(0, 0, hScroll.width, hScroll.height);
	    if (hsb != null) {
	        hsb.setBounds(hScrollRect.x, hScrollRect.y,
			      hScrollRect.width, hScrollRect.height);
	    }
	}
	
	repaintAll(m, r);

	/*System.err.println("-boundsChanged: textR=" + textRect + 
			   " vScrollR=" + vScrollRect + 
			   " hScrollR=" + hScrollRect);*/
    }

    /**
     * If we don't have a textpane yet, create one, and then return it.
     * @param tc Our text component
     * @return Data model for this component
     */
    public TextPane getTextPane() { 

        if (textpane == null) {
	    TextComponentModel model = (TextComponentModel)getModel();
	    textpane = new TouchableMultiLinePane(
				  this, model, getToolkit(),
				  Palette.TEXT_PANE_COLORS,
				  DrawUtils.TA_INSETS,
				  DrawUtils.LEFT_MARGIN,
				  DrawUtils.RIGHT_MARGIN,
				  true);
	}
	return textpane;
    }

    /**
     * Initialize scrollbars based on scrollbar visibility policy.
     */
    protected void initializeScrollbars(TextAreaModel model) {
        switch(sbVisibility) {
        case TextArea.SCROLLBARS_NONE:
            vsb = null;
	    hsb = null;
            wordwrap = true;
            break;

        case TextArea.SCROLLBARS_VERTICAL_ONLY:
            vsb = getVerticalScrollbar(vsb, model);
	    hsb = null;
            wordwrap = true;
            break;

        case TextArea.SCROLLBARS_HORIZONTAL_ONLY:
	    vsb = null;
            hsb = getHorizontalScrollbar(hsb, model);
            wordwrap = false;
            break;

        case TextArea.SCROLLBARS_BOTH:
        default:
	    vsb = getVerticalScrollbar(vsb, model);
            hsb = getHorizontalScrollbar(hsb, model);
            wordwrap = false;
            break;
        }

	if (model.getDataModel().getWordWrap() != wordwrap) {
	    model.getDataModel().setWordWrap(wordwrap);
	}
    }
}
