/*
 * @(#)TextPopupView.java	1.11 98/12/03 SMI
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

package sun.awt.touchable;

import java.awt.*;
import java.awt.event.*;
import sun.awt.otk.*;
import sun.porting.toolkit.ToolkitEventHandler;

class TextPopupView extends ModalContainerViewImpl
implements sun.awt.otk.VirtualKeyboardViewIntf{

    protected ScrollableTextView stv;

    /**
     * A minimal rectangle used to create objects with unspecified bounds.
     */
    protected static final Rectangle NULL_BOUNDS = new Rectangle(0, 0, 1, 1);

    TextPopupView(TextComponentModel textModel, ObjectToolkit toolkit, 
		  Rectangle bounds, Color bgColor){
	this(textModel, toolkit, bounds,
	     new Insets(DrawUtils.TEXT_POPUP_INSETS.top, 
			DrawUtils.TEXT_POPUP_INSETS.left,
			DrawUtils.TEXT_POPUP_INSETS.bottom,
			DrawUtils.TEXT_POPUP_INSETS.right),
	     bgColor,
	     (textModel instanceof TextComponentPeerModel));

	// keep TAB and Return from making it into the text buffer
	// This is safe because TextPopupView only comes up on TextFields
	stv.addController(TextFilter.controller);
	addKeyListener(stv);
	pack();
    }
    
    /**
     * That method should be called by sub-classes
     * since keyListener is not added and pack is not done
     */
    TextPopupView(TextComponentModel textModel, ObjectToolkit toolkit, 
		  Rectangle bounds, Insets insets, Color bgColor, 
		  boolean addDismissButton){

	super(toolkit, bounds, insets);

	setBackgroundColor(bgColor);

	stv = new ScrollableTextView(textModel, toolkit, this, 
				     Palette.DEF_COLORS,
				     NULL_BOUNDS, addDismissButton);
	add(stv);

	/* Register myself as a dependent on the text model.
	   For reasons of performance, there is a defaultView,
	   and then all additional views get added as dependents.
	   Hence this test.
	*/
	if (textModel.getDefaultView() == null) {
	    textModel.setDefaultView(stv);
	} else {
	    textModel.addDependent(stv);
	}
    }

    protected void pack(){
	Dimension prefSize = getPreferredSize(-1, -1);
	Dimension screen   = getToolkit().getScreenSize();
	Insets insets      = getInsets();
	Rectangle bounds   = getBounds();

	boolean centerPopup = bounds.equals(NULL_BOUNDS);

	// Make sure that popup is not less than preferred size
	bounds.width  = (bounds.width == 0 ? prefSize.width : 
			 Math.max(prefSize.width, bounds.width));
	bounds.height = (bounds.height == 0 ? prefSize.height : 
			 Math.max(prefSize.height, bounds.height));

	//Make sure that popup is less than the screen
	bounds.width  = Math.min(bounds.width, screen.width);
	bounds.height = Math.min(bounds.height, screen.height);

	// in case of a TextComponent align first line of Textpopup
	// with TextComponent first line
	if (stv.getModel() instanceof TextComponentPeerModel){
	    stv.setBounds(bounds.x, bounds.y, bounds.width, bounds.height);
	    int delta = stv.lineNumberToY(0);
	    if (bounds.y - delta >= 0) {
		bounds.y -= delta;
	    }
	} else if (centerPopup){
	    // if passed in bounds were empty - center popup on the screen
	    bounds.x = (screen.width - bounds.width)/2;
	    bounds.y = (screen.height - bounds.height)/2;
	}

	// make sure that popup does not go off the screen
	bounds.x = (bounds.x + bounds.width > screen.width) ? 
	            screen.width - bounds.width : bounds.x;
	bounds.y = (bounds.y + bounds.height > screen.height) ? 
	            screen.height - bounds.height : bounds.y;

	setBounds(Math.max(0, bounds.x), Math.max(0, bounds.y), 
		  bounds.width, bounds.height);
    }

    public void setBounds(int x, int y, int width, int height) {
	Insets insets = getInsets();
	doLayout(insets.left, insets.top, 
		 width - (insets.left + insets.right), 
		 height - (insets.top + insets.bottom));

	super.setBounds(x, y, width, height);
    }

    // Does layout before calling ContainerViewImpl.setBounds()
    protected void doLayout(int x, int y, int width, int height){
	stv.setBounds(x, y, width, height); 
    }

    protected void sendActionEvent() {
	stv.sendActionEvent();
    }

    protected void sendHideEvent() {
	stv.sendHideEvent();
    }

    /* The following methods implement the VirtualKeyboardIntf interface */
    public boolean dispatchEvent(AWTEvent e) {
	return false;
    }

    public void show(AWTEvent e) {
	if (!isVisible()) {

	    switch (e.getID()) {
	    case VirtualComponentEvent.VIRTUAL_COMPONENT_SHOW:
		VirtualComponentEvent evt = (VirtualComponentEvent)e;
	        setSelectionInfo((VirtualComponentSelInfo) evt.getMessage());
	        break;

	    default:
	        break;
	    }

	    setVisible(true);
	}
    }

    public void hide() {
	if (isVisible()) {
	    setVisible(false);
	}
    }

    public void dispose() {
    // remove ourselves as a dependent of the model - we're not visible
	hide();

	TextComponentModel textModel = (TextComponentModel)stv.getModel();    
	if (textModel.getDefaultView() == stv) {
	    textModel.setDefaultView(null);
	} else {
	    textModel.removeDependent(stv);
	}
    }

    boolean dismissPending = false;

    /**
     * We over-ride handleEvent to add functionality: Dismiss the popup
     * when mouse is clicked outside its bounds and when the return key is
     * pressed.
     * @param id The event type identifier
     * @param when A timestamp (in milliseconds) indicating when the event happened
     * @param modifiers The modifier key (SHIFT, CONTROL, ALT, META) state.
     * @param x The x coordinate associated with the event
     * @param y The y coordinate associated with the event
     * @param arg1 An optional argument whose meaning varies with the event type
     * @param arg2 An optional argument whose meaning varies with the event type
     * @see sun.porting.toolkit.ToolkitEventHandler
     */
    public void handleEvent(int id, long when, int modifiers,
		     int x, int y, int arg1, int arg2) {

	switch (id) {

	case ToolkitEventHandler.WINDOW_DAMAGED:
	    Graphics g = null;

	    try {
		g = getGraphics();
		g.clipRect(x, y, arg1, arg2);
		paintView(getModel(), g);
	    } finally {
		if (g != null) {
		    g.dispose();
		}
	    }
	    return;

	case ToolkitEventHandler.POINTER_PRESSED: {
	    Rectangle bounds = getBounds();
	    bounds.x = bounds.y = 0;
	    if (!bounds.contains(x, y)) {
	        dismissPending = true;
	    }
	    super.handleEvent(id, when, modifiers, x, y, arg1, arg2);
	    break;
	}

	case ToolkitEventHandler.POINTER_RELEASED: {
	    Rectangle bounds = getBounds();
	    bounds.x = bounds.y = 0;
	    if (dismissPending && !bounds.contains(x, y)) {
	        stv.dismissPressed();
	    } else {
	        super.handleEvent(id, when, modifiers, x, y, arg1, arg2);
	    }
	    dismissPending = false;
	    break;
	}

        case ToolkitEventHandler.KEY_PRESSED:
        case ToolkitEventHandler.KEY_RELEASED:
        case ToolkitEventHandler.KEY_TYPED:
	    if (stv.getModel() instanceof TextFieldModel) {
	        char d[] = {(char) arg2};

		switch(d[0]) {

		case '\n':
		case '\r':
		      // we act on the release deliberately, so that we end
		      // up handling all the events related to the Return press
		      if (id == ToolkitEventHandler.KEY_RELEASED)
		      	  stv.dismissPressed();
		    break;

		default:
		    super.handleEvent(id, when, modifiers, x, y, arg1, arg2);
		}
	    } else {
	        super.handleEvent(id, when, modifiers, x, y, arg1, arg2);
	    }
	    break;

        default:
	    super.handleEvent(id, when, modifiers, x, y, arg1, arg2);
	    break;
	}
    }

    protected void setSelectionInfo(VirtualComponentSelInfo selInfo) {
	if (selInfo != null) {
	    stv.setCaretPosition(selInfo.caret);
	    stv.select(selInfo.selStart, selInfo.selEnd);
	    stv.scrollTo(selInfo.selStart);
	}
    }


    protected void paintView(Model m, Graphics g) {
	if (!isVisible()) {
	    return;
	}

	// draw the borders
	Rectangle bounds = getBounds();
	Insets insets = getInsets();
	g.setColor(Palette.DIALOG_COLORS.getGroupingColor());

	g.fillRect(0, 0, 
		   Math.min(insets.left, DrawUtils.TEXT_POPUP_BORDER_THICK),
		   bounds.height);
	g.fillRect(0, 0, bounds.width, 
		   Math.min(insets.top, DrawUtils.TEXT_POPUP_BORDER_THICK));

	// right
	int borderThick = Math.min(insets.right, 
				   DrawUtils.TEXT_POPUP_BORDER_THICK);
	g.fillRect(bounds.width - borderThick, 0, 
		   borderThick, bounds.height);

	// bottom
	borderThick = Math.min(insets.bottom, 
			       DrawUtils.TEXT_POPUP_BORDER_THICK);
	g.fillRect(0, bounds.height - borderThick, bounds.width, borderThick);
    }
}
