/*
 * @(#)Window.java	1.51 99/09/27
 *
 * Copyright (c) 1996, 1997 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information"). You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

package sun.awt.aw;

import sun.porting.toolkit.SettableCursor;
import sun.porting.toolkit.ToolkitEventHandler;
import sun.porting.graphicssystem.CursorImage;
import sun.porting.graphicssystem.Drawable;
import sun.porting.graphicssystem.GeometryProvider;
import sun.porting.graphicssystem.GraphicsSystem;
import sun.porting.graphicssystem.Region;

import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;

/**
 * The window objects created by this window system.  Approximate memory
 * footprint for a single window object is < 100 bytes, but the size of the
 * region structures (which can vary over time) can push it over that if the
 * clip region for this window is particularly complex.
 */
public class Window implements sun.porting.windowsystem.Window, GeometryProvider {
    /**
     * @return A Graphics object suitable for drawing into this window.
     */
    public Graphics getGraphics() {
	return fbDrawable.getGraphics(this);
    }


    // these next few methods comprise the GeometryProvider interface.

    /**
     * Get the background color for the window. (I.e. the color which
     * will be used to erase it when damage occurs, etc.)
     * @return A <code>java.awt.Color</code> object describing the 
     * background color for the drawing surface.
     */
    public Color getBackgroundColor() {
	synchronized (this) {
	    return bgColor;
	}
    }

    /**
     * Get the lock object associated with the GeometryProvider.  All
     * operations must acquire this lock as the first thing they do,
     * and hold the lock until the draw has been completed.  (If the
     * object being drawn is a window, this can prevent the windows from
     * being moved during the draw; in any case it also prevents more than
     * one thread from drawing to the same object at the same time.)
     * @return The lock object that must be held before drawing on this window.
     */
    public Object getLock() {
	return winSys;
    }

    /**
     * Get the validation identifier.  This is the same number that is
     * returned by lock(); if it has not changed since the previous
     * acquisition of the lock, the clipping information can be assumed
     * not to have changed.
     * @return The change identifier.
     */
    public int getValidationID() {
	return validationCount;
    }

    // these 5 are for the DrawingSurfaceInfo interface 
    // (which GeometryProvider extends)

    /**
     * Lock the current drawing surface.
     * Lock the current state of the drawing surface in preparation
     * for drawing to it and return an integer "change identifier"
     * which can be used to determine when changes have occured to
     * the drawing surface specification and when to retrieve new
     * extent and physical handle data.  If the identifier has not
     * changed since the previous time lock and unlock were called
     * then the consumer of this information is free to reuse clipping
     * and handle information cached during the previous rendering
     * operation.  If the identifier changes, then all extent,
     * clipping, and handle information must be retrieved again.
     * The unlock() method must be called when rendering is complete
     * or the component may be permanently locked down on the screen
     * and other rendering threads may block indefinately.
     * @see getValidationID
     * @return The change identifier.
     */
    public int lock() {
	lock0(getLock());
	return validationCount;
    }

    /**
     * Unlock the drawing surface and allow other threads to
     * access or change its composition.
     * @see getValidationID
     */
    public void unlock() {
	unlock0(getLock());
    }

    /**
     * Get the bounding rectangle for this window.
     * Returns a rectangle describing the logical bounds of the given
     * screen object relative to the "drawing handle" information
     * returned by getSurface().  Note that there may be portions
     * of this bounding rectangle that are obscured.  This value
     * defines the location of the logical upper-left origin of
     * the component or screen object and its desired extents
     * (which may differ from the clipping extents due to the
     * presence of other overlapping screen objects).  The getClip()
     * method should be used to determine which portions of the
     * drawing surface can actually be drawn to.
     * @return A Rectangle object which describes the window's overall size
     * and position.  The position is in global coordinates.
     */
    public Rectangle getBounds() {
	return getBounds(true);
    }

    /**
     * Get the drawing surface corresponding to the window.
     * Returns an implementation-specific object which describes a
     * handle to the physical drawing surface on which this screen
     * object resides.  Depending on the type of the return value,
     * this object may describe a Windows hWnd handle, an X11
     * drawable handle, a Mac GrafPtr structure, or some other
     * platform-specific handle.
     * @return A handle to the drawing surface corresponding to the window.
     */
    public sun.awt.PhysicalDrawingSurface getSurface() {
	return fbDrawable;
    }

    /**
     * Get the Region that represents the window's current visible area.
     * Returns an explicit representation of the renderable portion
     * of the given screen object.  This clipping area will include
     * only those areas of the screen object that are not obscured
     * by other overlapping objects and which fall within the
     * boundaries of the physical drawing surface.  The area described
     * by the return value of this method can be directly used as a
     * graphics clip area to restrict rendering to the given screen
     * object.
     * @return The Region describing the window's current clip.
     */
    public Shape getClip() {
	synchronized (winSys) {
	    return canvasRegion;
	}
    }

    /**
     * Invalidate the clip by incrementing the validation count.
     */
    protected void invalidateClip() {
	++validationCount;
    }
    private int validationCount = 1;
    private native void lock0(Object lock);
    private native void unlock0(Object lock);

    /** 
     * Get the location and size of the window.  If global is true,
     * the position will be returned in root coordinates; otherwise
     * they will be relative to this window's parent.
     * @param global A flag indicating whether to return x and y in
     * global or parent-relative coordinates.
     * @return The position and size in pixels, as a Rectangle object.
     * @throws IllegalStateException if the window has been disposed.
     */
    public Rectangle getBounds(boolean global) throws IllegalStateException {
	if (isDisposed()) {
	    throw new IllegalStateException("Window " +winNum()+ " already disposed");
	}

	synchronized (this) {
	    if (global) {
		return new Rectangle(rootX, rootY, width, height);
	    } else {
		return new Rectangle(x, y, width, height);
	    }
	}
    }

    /**
     * Set the location and/or size of the window.  x and y are relative
     * to this window's parent.
     * @param x x coordinate of the upper left corner, relative to parent.
     * @param y y coordinate of the upper left corner, relative to parent.
     * @param w Width in pixels.
     * @param h Height in pixels.
     * @throws IllegalStateException if the window has been disposed.
     */
    public void setBounds(int x, int y, int w, int h) throws IllegalStateException {
	synchronized (winSys) {
	    if (isDisposed()) {
		throw new IllegalStateException("Window " +winNum()+ " already disposed");
	    }

	    reshape(x, y, w, h);
	}
    }

    /**
     * Move this window so that it is Nth in its list of siblings.
     * N = 0 moves the window to the front; N = -1 moves it to the back.
     * N = -2 moves it so it is above only the last window, etc.
     * @throws IllegalStateException if the window has been disposed.
     */
    public void setStackingOrder(int n) throws IllegalStateException {
	synchronized (winSys) {
	    if (isDisposed()) {
		throw new IllegalStateException("Window " +winNum()+ " already disposed");
	    }

	    parent.moveChildToPosition(n, this);
	}
    }

    /**
     * Set the background color for the window (i.e. the color which
     * will be used to erase it when damage occurs, etc.)
     * @param c The color to be used for the background.
     */
    public void setBackgroundColor(Color c) {
	synchronized (this) {
	    bgColor = c == null ? Color.white : c;
	}

	synchronized (winSys) {
	    /* now repaint the window */
	    if (canvasRegion != null) {
		handleExposure(canvasRegion);
	    }
	}
    }

    /**
     * Query the map state of the window
     * @return true if the window is mapped (visible) false otherwise.
     * @throws IllegalStateException if the window has been disposed.
     */
    public boolean isMapped() throws IllegalStateException {
	synchronized (winSys) {
	    if (isDisposed()) {
		throw new IllegalStateException("Window " +winNum()+ " already disposed");
	    }
	    return hierarchyMapped();
	}
    }

    /**
     * Make this window visible.
     * @throws IllegalStateException if the window has been disposed.
     */
    public void map() throws IllegalStateException {
	synchronized (winSys) {
	    if (isDisposed()) {
		throw new IllegalStateException("Window " +winNum()+ " already disposed");
	    }

	    mapLocked();
	}
    }

    /**
     * Make this window invisible.
     * If the window has the grab or the input focus, implicitly releases them.
     * @throws IllegalStateException if the window has been disposed.
     */
    public void unmap() throws IllegalStateException {
	synchronized (winSys) {
	    if (isDisposed()) {
		throw new IllegalStateException("Window " +winNum()+ " already disposed");
	    }

	    unmapLocked();
	}
    }

    /**
     * Dispose of any resources associated with the window, and mark it invalid.
     * If the window has the grab or the input focus, implicitly releases them.
     */
    public void dispose() {
	synchronized (winSys) {
	    if (!isDisposed()) {
		disposeLocked();
	    }
	}
    }


    /**
     * Determine whether this window has the input focus.
     * @return true if this window has focus, false otherwise.
     * @throws IllegalStateException if the window has been disposed.
     */
    public boolean hasInputFocus() throws IllegalStateException {
	synchronized (winSys) {
	    if (isDisposed()) {
		throw new IllegalStateException("Window " +winNum()+ " already disposed");
	    }

	    return winSys.getFocusWindow() == this;
	}
    }

    /**
     * Assign the keyboard input focus to this window.
     * @throws IllegalStateException if the window is disposed or not mapped.
     */
    public void acquireInputFocus() throws IllegalStateException {
	synchronized (winSys) {
	    if (isDisposed()) {
		throw new IllegalStateException("Window " +winNum()+ " already disposed");
	    }

	    winSys.setFocusWindow(this);
	}
    }


    /**
     * Find the ancestor of this window which is the top-level window.
     * This is normally the window parented by root.
     * @return This window's top-level ancestor.
     */
    protected Window findTopLevelWindow() {
	Window w = this;

	if (w.parent == null) {
	    return null;
	}

	while (w.parent != winSys.rootWindow) {
	    w = w.parent;
	}
	return w;
    }

    /**
     * Determine whether this window owns an input grab.
     * @return true if this window has the grab, false otherwise.
     * @throws IllegalStateException if the window has been disposed.  */
    public boolean hasGrab() throws IllegalStateException {
	synchronized (winSys) {
	    if (isDisposed()) {
		throw new IllegalStateException("Window " +winNum()+ " already disposed");
	    }

	    return winSys.getGrabbingWindow() == findTopLevelWindow();
	}
    }

    /**
     * Make this window grab all input. <b>Does not change keyboard focus.</b>
     * If this window is not a top-level window, the grab will be given to
     * its top-level ancestor instead.  The mode argument indicates what 
     * should be done with events that are not directed at the grab window.
     * @param mode how to handle events directed at other windows.
     * @throws IllegalStateException if the window is disposed or not mapped,
     * or if the grab is owned by someone else.
     * <strong>Important note: the actual window that owns the explicit grab 
     * must always be a top-level window.  It is the window system's 
     * responsibility to find the top-level ancestor when grabbing, and to 
     * ensure that the event target is owned by (not necessarily equal to) 
     * the grab owner.</strong>
     */
    public void grabInput(int mode) throws IllegalStateException {
	synchronized (winSys) {
	    if (isDisposed()) {
		throw new IllegalStateException("Window " +winNum()+ " already disposed");
	    }

	    winSys.setGrabbingWindow(findTopLevelWindow(), mode);
	}
    }

    /**
     * Make this window release a grab. <b>Does not change keyboard focus.</b>
     * @throws IllegalStateException if the window does not have the grab
     * or has been disposed.
     */
    public void releaseGrab() throws IllegalStateException {
	synchronized (winSys) {
	    if (isDisposed()) {
		throw new IllegalStateException("Window " + winNum() + 
						" already disposed");
	    }

   	    if (!winSys.isInGrabStack(findTopLevelWindow())) {
 		throw new IllegalStateException("Window " +winNum()+ 
						" Does not own grab");
 	    }

	    winSys.releaseGrabbingWindow(findTopLevelWindow());
	}
    }

    /**
     * Set the window's "user data" field to the given object.
     * This is a convenience for Toolkit implementors which allows
     * them to efficiently associate private data with each window.
     */
    public synchronized void setUserData(java.lang.Object o) {
	uData = o;
    }

    /**
     * Return the window's "user data" field.
     * This is a convenience for Toolkit implementors which allows
     * them to efficiently associate private data with each window.
     */
    public synchronized java.lang.Object getUserData() {
	return uData;
    }
    private java.lang.Object uData;


    // (private) implementation

    // position (relative to parent) and current size of window
    int x;
    int y;
    int width;
    int height;

    // Cursor image for this window.
    CursorImage cursorImage;

    // NOTE: if a small amount of information (mostly the clip regions) is
    // changed from a single region to an array, and some additional bookkeeping
    // is added, it may be possible for a Window to span more than one screen.
    // (which could be very useful for some devices); but the whole concept
    // of a root window then becomes a bit questionable--the parent of a
    // top-level window would need to be a "meta-root," and the windows
    // themselves would be clipped by "screen root" regions acting as viewports.

    // position (in global coordinates relative to root window)
    int rootX;
    int rootY;

    // the clipping region for this window (i.e. including children)
    // this is the area that is not clipped by the window's parent, or
    // obscured either by siblings or by ancestors' siblings.
    Region windowRegion;

    // the clipping region for primitives drawn into this window
    // This is windowRegion minus the areas covered by children.
    Region canvasRegion;

    // This window's parent
    Window parent;

    // A doubly linked list of this window's children
    Window child_head;
    Window child_tail;

    // for putting this window object in a doubly linked list with its siblings
    Window next;
    Window prev;

    int status;
    private void setStatus(int mask, boolean set) {
	if (set) {
	    status |= mask;
	} else {
	    status &= ~mask;
	}
    }

    private static final int DISPOSED = (1 << 31);
    final boolean isDisposed() {
	return (status & DISPOSED) != 0;
    }

    private static final int MAPPED = (1 << 30);
    private final boolean mapped() {
	return (status & MAPPED) != 0;
    }

    private static final int PARENTS_MAPPED = (1 << 29);
    private final boolean areParentsMapped() {
	return (status & PARENTS_MAPPED) != 0;
    }

    private static final int HIERARCHY_MAPPED_MASK = MAPPED | PARENTS_MAPPED;
    private final boolean hierarchyMapped() {
	return (status & HIERARCHY_MAPPED_MASK) == HIERARCHY_MAPPED_MASK;
    }

    private static final int VISIBLE = (1 << 28);
    private final boolean isVisible() {
	return (status & VISIBLE) != 0;
    }

    private static final int UNOBSCURED = (1 << 27);
    private final boolean isUnobscured() {
	return (status & UNOBSCURED) != 0;
    }

    private static final int VALID = (1 << 26);
    private final boolean isValid() {
	return (status & VALID) != 0;
    }

    private static final int WIN_NUM_MASK = 0xffffffff >>> 8;
    private final int winNum() {
	return status & WIN_NUM_MASK;
    }

    // This is used when calculating the closest ancestor of two
    // windows. It should be null in *all* instances except when
    // the window structure lock is held.
    Color bgColor = Color.white;

    static int winNumCounter = 0;

    /*
     * NB: This Window system implementation assumes only
     * one screen.  In order to support multiple screens, a
     * number of changes are required, probabli starting with 
     * a change of the following three variables to be instance
     * variables rather than class variables.
     */

    /**
     * The CursorImage to be used as a default for windows where
     * no cursor is specified.  Also used for the root window.
     */
    static protected CursorImage defaultCursorImage = null;
    static GraphicsSystem gfxSys = null;
    static Drawable       fbDrawable = null;

    WindowSystem          winSys;

    /**
     * Construct a window with the given parent, location and size.
     * If the parent is null, you are creating the root window.
     * @param ws The WindowSystem that this window belongs to.
     * @param parent The parent window for this window.
     * @param x The x coordinate for the window's upper-left corner.
     * Relative to parent.
     * @param y The y coordinate for the window's upper-left corner.
     * Relative to parent.
     * @param w The window's initial width.
     * @param h The window's initial height.
     */
    protected Window(WindowSystem ws, Window parent, 
		     int x, int y, int w, int h) {
	if (gfxSys == null) {
	    try {
		gfxSys = ws.getGS();
		fbDrawable = gfxSys.getScreen();
	    } catch (Throwable t) {
		throw new RuntimeException("Can't access graphics system");
	    }

	    String cursorClass = null;
	    try {
		cursorClass = 
		    System.getProperty("sun.awt.aw.DefaultCursor", 
				       "sun.awt.aw.NonExistentCursorClass");

		if (!cursorClass.equals("blank")) {
		    Class cls = Class.forName(cursorClass);
		    ImageProducer prod = (ImageProducer) cls.newInstance();
		    Point hotspot  = ((SettableCursor) prod).getHotSpot();
		    defaultCursorImage =
			ws.makeCursorImage(gfxSys.getImage(prod), 
					   hotspot.x, hotspot.y);
		}
	    } catch (Throwable t) {
		useDefaultCursor();
	    }
	}

	this.winSys = ws;
	this.parent = parent;
	this.x = x;
	this.y = y;
	this.width = Math.max(w, 1);
	this.height = Math.max(h, 1);
	status = VALID | (winNumCounter++ & WIN_NUM_MASK);

	if (parent == null) {
	    rootX            = x;
	    rootY            = y;
	    setStatus(PARENTS_MAPPED, true);
	    cursorImage      = defaultCursorImage;
	} else {
	    rootX            = x + parent.rootX;
	    rootY            = y + parent.rootY;
	    setStatus(PARENTS_MAPPED, parent.hierarchyMapped());
	    cursorImage      = parent.cursorImage;

	    parent.insertChildAtTop(this);
	}
    }

    /**
     * Change this window's position and or size.  Exposures may be generated
     * if this causes previously invisible window surface to become visible,
     * but any already-visible area should not have to be repainted.
     *
     * @param x The desired x coordinate for the window's upper-left corner.
     * Relative to parent.
     * @param x The desired y coordinate for the window's upper-left corner.
     * Relative to parent.
     * @param w The desired width.
     * @param h The desired height.
     */
    protected void reshape(int x, int y, int w, int h) {
	boolean size_changed = (this.width != w) || (this.height != h);

	int dx = x - this.x;
	int dy = y - this.y;
	int dw = w - this.width;
	int dh = h - this.height;

	// minimum width and height are 1, in order to avoid wierd edge cases
	if (w < 1) {
	    w = 1;
	}
	if (h < 1) {
	    h = 1;
	}

	if (this.x == x && this.y == y && this.width == w && this.height == h) {
	    return;
	}

	// check this up here in order to avoid repeated checks below
	if (!hierarchyMapped()) {
	    synchronized (this) {
		this.x = x;
		this.y = y;
		this.width = w;
		this.height = h;

		if ((dx | dy) != 0) {
		    changeChildParentPosition(dx, dy);
		}
	    }

	    if (size_changed) {
		winSys.windowEventOccurred(this, System.currentTimeMillis(),
					   dx, dy, dw, dh,
					   ToolkitEventHandler.WINDOW_RESHAPED);
	    }
	} else {
	    Region oldRegion = (windowRegion == null)
		? gfxSys.makeRegion(0, 0, 0, 0)
		: windowRegion;

	    boolean oldRegionEmpty = (windowRegion == null)
		|| windowRegion.isEmpty();

	    if ((!oldRegionEmpty) && ((dx | dy) != 0)) {
		Rectangle bbox = getBounds(true);
		int srcX = bbox.x - parent.rootX; // graphics context adds
		int srcY = bbox.y - parent.rootY; // parent's translation
		int srcW = Math.min(w,bbox.width);
		int srcH = Math.min(h,bbox.height);

		// Special code is required in order to quickly
		// blt this window in the context of its parent.
		// We calculate the clip region created by the
		// parent and the superior siblings of the window
		// to be moved; then we temporarily make that the 
		// parent's canvasRegion while we do the blt.
		Region reg;
		if (this == parent.child_head) {        // we're on top
		    reg = parent.windowRegion;
		} else if (this == parent.child_tail) {	// we're on bottom
		    reg = parent.canvasRegion.copy();
		    reg.union(this.windowRegion);
		} else {
		    reg = parent.windowRegion.copy();
		    Window sib = parent.child_head;
		    while (sib != null && sib != this) {
			if (sib.mapped()) {
			    reg.subtract(sib.windowRegion.getBounds());
			}
			sib = sib.next;
		    }
		}
		reg.defragment(true);

		/* Temporarily enlarging the clip requires remembering   */
		/* the canvasRegion field, so that we can reset and then */
		/* restore it.  We must call invalidateClip() each time. */
		Region tmp = parent.canvasRegion;
		parent.canvasRegion = reg;
		parent.invalidateClip();

		Graphics g = parent.getGraphics();
		g.copyArea(srcX, srcY, srcW, srcH, dx, dy);
		g.dispose();

		parent.canvasRegion = tmp;
		parent.invalidateClip();
	    }

	    // now do the actual work of moving the window, updating
	    // children, resetting all of the clip regions, etc.
	    synchronized (this) {
		this.x = x;
		this.y = y;
		this.width = w;
		this.height = h;

		if ((dx | dy) != 0) {
		    changeChildParentPosition(dx, dy);
		}
	    }


	    if (size_changed) {
		winSys.windowEventOccurred(this, System.currentTimeMillis(),
					   x, y, width, height,
					   ToolkitEventHandler.WINDOW_RESHAPED);
	    }

	    if (parent == null) {
		parentChanged(gfxSys.makeRegion(x, y, width, height), dx, dy);
	    } else {
		if (oldRegionEmpty
		    && ! parent.windowRegion.mayIntersect(rootX, rootY, width, height)) {
		    return;
		}

		parent.childChanged(this, dx, dy);
	    }

	    winSys.resetCursorWindow();
        }
    }

    /**
     * Indicate that this window is about to be moved to the front of all of 
     * its siblings.  
     * <strong>The supplied window must be a direct child of the one on 
     * which this method is invoked.</strong>
     * @param win The window that will be moved to the front.
     */
    protected void childToFront(Window win) {
        Region childRegion = gfxSys.makeRegion(win.rootX, win.rootY, win.width, win.height);
        childRegion.intersect(windowRegion);

        // update the regions of the new window and its children
        win.parentChanged(childRegion, 0, 0);

        // and handle any obscured areas of siblings below
        for (Window c = win.next; c != null; c = c.next) {
            if (c.mapped() && win.windowRegion.mayIntersect(c.windowRegion)) {
                c.obscured(win.rootX, win.rootY, win.width, win.height);
            }
        }
    }

    /**
     * Indicate that this window is about to be moved behind of all of 
     * its siblings.  <strong>This must be called before the stacking order
     * is actually changed, and the 
     * supplied window must be a direct child of the one on which this
     * method is invoked.</strong>
     * @param win The window that will be moved to the back.
     */
    protected void childToBack(Window win) {
        // update the regions of any window currently obscured by win
        if ((win.windowRegion == null) || win.windowRegion.isEmpty()) {
            return;
        }
        Region r = win.windowRegion.copy();
        for (Window c = win.next; c != null; c = c.next) {
            if (c.mapped() && r.mayIntersect(c.rootX, c.rootY, c.width, c.height)) {
                Region add = gfxSys.makeRegion(c.rootX, c.rootY, c.width, c.height);
                add.intersect(r);
                if (c.windowRegion != null) {
                    add.union(c.windowRegion);
                }
                c.parentChanged(add, 0, 0);
                r.subtract(c.rootX, c.rootY, c.width, c.height);
            }
        } 
        win.parentChanged(r, 0, 0);
    }

    /**
     * Indicate that the given window is to be unmapped.  Children should
     * already have had their parent mappings changed.  The supplied region
     * is the area that was occupied by the window before unmapping.
     * <strong>The supplied window must be a direct child of the one on 
     * which this method is invoked.</strong>
     * @param win The window that is being unmapped
     * @param r The region this window currently occupies
     */ 
    protected void childUnmapped(Window win, Region r) {
        // update the regions of any window currently obscured by win
        if ((r == null) || (r.isEmpty())) {
            return;
        }

        if (r.getBounds().equals(windowRegion.getBounds())) {
	    /* we're potentially uncovering the whole thing */
            recalculateRegions(0,0);
            return;
        }

        boolean noneMapped = true;
        for (Window c = win.next; c != null; c = c.next) {
            if (c.mapped()) {
                noneMapped = false;
                if (r.mayIntersect(c.rootX, c.rootY, c.width, c.height)) {
                    Region add = gfxSys.makeRegion(c.rootX, c.rootY, c.width, c.height);
                    add.intersect(r);
                    if (add.isEmpty()) {
                        continue;
                    }
                    if (c.windowRegion != null) {
                        add.union(c.windowRegion);
                    }

                    // possibly it would be more efficient
                    // to propagate the addition downward, rather
                    // than calling parentChanged??
                    c.parentChanged(add, 0, 0);
                    r.subtract(c.rootX, c.rootY, c.width, c.height);
                }
            }
        }

        // special check for the case where we are unmapping the only
        // mapped sibling.  (This is common with card layouts). In this
        // situation a copy is probably faster than a union and defragment.
        if (noneMapped) {
            for (Window c = win.prev; c != null; c = c.prev) {
                if (c.mapped()) {
                    noneMapped = false;
                    break;
                }
            }
        }

        if (noneMapped) {
            canvasRegion = windowRegion.copy();
	    invalidateClip();
        } else if (!r.isEmpty()) {
            canvasRegion.union(r);
            canvasRegion.defragment(false);
	    invalidateClip();
        }

        handleExposure(r);
    }

    /**
     * Indicate that the given window is to be mapped.  Children should
     * already have had their parent mappings changed.
     * <strong>The supplied window must be a direct child of the one on 
     * which this method is invoked.</strong>
     * @param win The window that is being mapped
     */ 
    protected void childMapped(Window win) {
        if (canvasRegion == null) {
            recalculateRegions(0, 0);
            return;
        } else {
            canvasRegion.subtract(win.rootX, win.rootY, win.width, win.height);
	    invalidateClip();
            setStatus(VISIBLE, !canvasRegion.isEmpty());
	    setStatus(UNOBSCURED, canvasRegion.contains(rootX, rootY, 
							width, height));
        }

        // compute windowRegion for the newly-mapped window by clipping
        // first to its parent region, then subtracting the bounding
        // boxes of any siblings which are above it
        Region childRegion = gfxSys.makeRegion(win.rootX, win.rootY, win.width, win.height);

        childRegion.intersect(windowRegion);
        for (Window c = win.prev; c != null; c = c.prev) {
            if (c.mapped()) {
                childRegion.subtract(c.rootX, c.rootY, c.width, c.height);
            }
        }

        // update the regions of the new window and its children
        win.parentChanged(childRegion, 0, 0);

        // and finally, handle any obscured areas of siblings below
        for (Window c = win.next; c != null; c = c.next) {
            if (c.mapped() && win.windowRegion.mayIntersect(c.windowRegion)) {
                c.obscured(win.rootX, win.rootY, win.width, win.height);
            }
        }
    }

    /**
     * Indicate that this window is being obscured in the given area.
     * @param rx x coord for the upper-left corner of the obscured rectangle.
     * @param ry y coord for the upper-left corner of the obscured rectangle.
     * @param rw The width of the obscured rectangle.
     * @param rh The height of the obscured rectangle.
     */
    protected void obscured(int rx, int ry, int rw, int rh) {
        // we're guaranteed that these exist if obscured() is called
        windowRegion.subtract(rx, ry, rw, rh);
        canvasRegion.subtract(rx, ry, rw, rh);
	invalidateClip();

        if (windowRegion.isEmpty()) {
            invisible(true);
        } else {
	    setStatus(UNOBSCURED, false);
	    setStatus(VISIBLE, !canvasRegion.isEmpty());

            for (Window c = child_head; c != null; c = c.next) {
                if (!c.mapped() || c.windowRegion.isEmpty()) {
                    continue;
                }

                if (c.windowRegion.mayIntersect(rx, ry, rw, rh)) {
                    c.obscured(rx, ry, rw, rh);
                }
            }
        }
    }

    /**
     * Indicate that this window is being made fully invisible.  This does
     * not mean it is not mapped; it is just completely covered by something.
     * If propagate is true, always call invisible recursively on any 
     * mapped children; otherwise propagation will only be done on currently
     * visible windows.
     * @param propagate If true, call invisible on all mapped children.
     */
    protected void invisible(boolean propagate) {
        boolean wasVisible = false;

        // we're not guaranteed that these exist, unfortunately.
        if ((windowRegion == null) || (canvasRegion == null)) {
            windowRegion = gfxSys.makeRegion(0,0,0,0);
            canvasRegion = gfxSys.makeRegion(0,0,0,0);
        } else {
            wasVisible = !windowRegion.isEmpty();
            windowRegion.setEmpty();
            canvasRegion.setEmpty();
        }
	invalidateClip();

	setStatus((VISIBLE | UNOBSCURED), false);

        if (!wasVisible && !propagate) {
            return;
        }

        for (Window c = child_head; c != null; c = c.next) {
            if (c.mapped()) {
                c.invisible(propagate);
            }
        }

	if (mapped()) {
	    winSys.windowEventOccurred(this, System.currentTimeMillis(),
				     x - rootX, y - rootY, width, height, 
				     ToolkitEventHandler.WINDOW_INVISIBLE);
	}
    }

    /**
     * Helper routine; dispose of the resources associated with a window.
     * The window system lock must be held.
     */
    protected void disposeLocked() {
        unmapLocked();

        // do these early in order to prevent event delivery to these windows
	setStatus(DISPOSED, true);

        if (parent != null && parent.child_head != null) {
            parent.deleteChild(this);

            // Dispose any "leftover" children
            for (Window c = child_head; c != null; c = c.next) {
                c.disposeLocked();
            }
        }
    }

    /**
     * Helper routine; map the window.  The window system lock must be held.
     */
    protected void mapLocked() {
        if ( ! mapped() ) {

	    setStatus(MAPPED, true);

            if (areParentsMapped()) {
                changeChildParentMappings(true);
                if (parent == null) {
                    parentChanged(gfxSys.makeRegion(x, y, width, height), 0, 0);
                } else {
                    parent.childMapped(this);
                }

		winSys.windowEventOccurred(this, System.currentTimeMillis(),
					 x - rootX, y - rootY, width, height,
					 ToolkitEventHandler.WINDOW_MAPPED);
            } else {
                windowRegion = gfxSys.makeRegion(0,0,0,0);
            }

	    winSys.resetCursorWindow();
	}
    }

    /**
     * Helper routine; unmap the window.  The window system lock must be held.
     */
    protected void unmapLocked() {
        if (mapped()) {
	    invalidateClip();	// prevent system from thinking that
				// window is still there!
	    setStatus(MAPPED | VISIBLE, false);

            if (areParentsMapped()) {
                // the following is needed for parent.childUnmapped()
                Region r = windowRegion;
                // reference to windowRegion is cleared by 
		// changeChildParentMappings, so retain it first!

                changeChildParentMappings(false);
                if (parent != null) {
                    parent.childUnmapped(this, r);
                }

		winSys.windowEventOccurred(this, System.currentTimeMillis(),
					 x - rootX, y - rootY, width, height,
					 ToolkitEventHandler.WINDOW_UNMAPPED);
            }

	    winSys.resetCursorWindow();
	}
    }

    // The following insert and delete operations are the result of
    // some performance improvements. Previously, the window lists
    // were implemented at SLists and SListIter enumerators. This was
    // deemed too inefficient in terms of memory allocations and CPU.
    // As a result, the Window class now uses its own private window
    // list implementation (explicit head/tail, next/prev pointers).

    // Utility routine to insert the given window into the head of
    // this window's children list.
    private final void insertChildAtTop(Window w) {

        if (child_head == null) {
            child_tail = w;
        } else {
            child_head.prev = w;
        }

        w.prev = null;
        w.next = child_head;
        child_head = w;
    }

    // Utility routine to insert the given window into the tail of
    // this window's children list.
    private final void insertChildAtBottom(Window w) {
	if (child_tail == null) {
	    child_head = w;
	} else {
	    child_tail.next = w;
	}

	w.next = null;
	w.prev = child_tail;
	child_tail = w;
    }

    // Delete the given child window from this window's child list.
    private final void deleteChild(Window w) {

        if (w.next == null) {
            child_tail = w.prev;
        } else {
            w.next.prev = w.prev;
        }

        if (w.prev == null) {
            child_head = w.next;
        } else {
            w.prev.next = w.next;
        }

        w.next = null;
        w.prev = null;
    }

    // NOTE: you must hold the window lock when calling this!
    private final void moveChildToPosition(int pos, Window child) {

	if (pos == 0) {
	    if (child_head != child) {
		if (child.hierarchyMapped()) {
		    childToFront(child);
		}

		deleteChild(child);
		insertChildAtTop(child);
	    }
	    return;
	} else if (pos == -1) {
	    if (child_tail != child) {
		if (child.hierarchyMapped()) {
		    childToBack(child);
		}

		deleteChild(child);
		insertChildAtBottom(child);
	    }
	    return;
	}

	if (child.hierarchyMapped()) {
	    childToBack(child);	// must restore regions of any siblings
	}
	deleteChild(child);	// guarantee that child is not in list

	Window p;

	if (pos > 0) {
	    p = child_head;
	    int index = 0;
	    while (p != child_tail) {
		++index;	// index of p.next
		p = p.next;
		if (index >= pos) {
		    break;	// p.next should end up pointing to child
		}
	    } 
	} else {
	    p = child_tail;
	    int index = 0;
	    while (p != null) {
		--index;	// index of p.next
		p = p.prev;
		if (index <= pos) {
		    break;	// p.next should end up pointing to child
		}
	    } 
	}

	// special cases, similar to above; this is faster than general case
	if (p == child_tail) {
	    insertChildAtBottom(child);
	    return;
	} else if (p == null) {
	    insertChildAtTop(child);

	    if (child.hierarchyMapped()) {
		childToFront(child);
	    }
	    return;
	}

	// We're guaranteed that we're not inserting at front, because
	// we checked to make sure that pos > 0; and we're not inserting
	// at end either, because p != child_tail.  So the pointer 
	// manipulation for the insertion is straightforward.
	child.next = p.next;
	child.next.prev = child;
	child.prev = p;
	p.next = child;

	if (child.hierarchyMapped()) {
	    recalculateRegions(0, 0);
	}
    }

    // should only be called on mapped windows. Child window w may
    // be just getting mapped or unmapped, however.
    private void childChanged(Window w, int dx, int dy) {
	invalidateClip();
        if (w.windowRegion == null) {
            w.windowRegion = gfxSys.makeRegion(0,0,0,0);
        }

        if (w.windowRegion.isEmpty()
            && (! windowRegion.mayIntersect(w.rootX, w.rootY, w.width, w.height)) ) {
            // it was empty, and it would stay empty
            return;
        }

        // Anything underneath w's old region could be affected
        // so we want to keep this around
        Region oldRegion = w.windowRegion.copy();

        // anything under w's new region could also be affected
        Region newRegion = w.mapped() ?
            gfxSys.makeRegion(w.rootX, w.rootY, w.width, w.height) : gfxSys.makeRegion(0,0,0,0);

        // this is a common enough case that it's worth checking for
        // so we do a separate mark step for all windows that may have
        // been affected by the move.
        boolean allWindowsAffected = true;
        for (Window c = w.next; c != null; c = c.next) {
            if (c.mapped()
                && (oldRegion.mayIntersect(c.rootX, c.rootY, c.width, c.height)
                    || newRegion.mayIntersect(c.rootX, c.rootY, c.width, c.height))) {
		c.setStatus(VALID, false);
            } else {
                allWindowsAffected = false;
		c.setStatus(VALID, true);
            }
        }

        // avoid doing the repeated unions if the end result is the
        // entire parent region
        if (allWindowsAffected && (w == child_head)) {
            canvasRegion = windowRegion.copy();
        } else {
            // add old region back into canvasRegion
            canvasRegion.union(oldRegion);

            // and also add back regions of all affected windows
            for (Window c = w.next; c != null; c = c.next) {
                if (!c.isValid()) {
                    canvasRegion.union(c.windowRegion);
                }
            }
        }

        // now recompute the clips for these windows only
        if (w.mapped()) {
            newRegion.intersect(canvasRegion);
            canvasRegion.subtract(newRegion.getBounds());
            w.parentChanged(newRegion, dx, dy);
        } else {
            w.windowRegion = null;
            w.canvasRegion = null;
        }

        for (Window c = w.next; c != null; c = c.next) {
            if (!c.isValid()) {
                Region nr = gfxSys.makeRegion(c.rootX, c.rootY, c.width, c.height);
                nr.intersect(canvasRegion);
                canvasRegion.subtract(nr.getBounds());
                c.parentChanged(nr, 0, 0);
		c.setStatus(VALID, true);
            }
        }

        canvasRegion.defragment(false);
        setStatus(UNOBSCURED,
		  canvasRegion.contains(rootX, rootY, this.width, this.height));
	setStatus(VISIBLE, !canvasRegion.isEmpty());

        if (isVisible()) {
            // Finally, repaint anything in oldRegion that was uncovered
            // by the change.
            if (w.mapped()) {
                oldRegion.subtract(w.rootX, w.rootY, w.width, w.height);
            }
            handleExposure(oldRegion);
        }
    }

    private void recalculateRegions(int dx, int dy) {
	invalidateClip();
        Region diffRegion =
            (windowRegion == null) ? gfxSys.makeRegion(0,0,0,0) : windowRegion.copy();

        for (Window child = child_head; child != null; child = child.next) {
            if (child.mapped()) {
                int cx = child.rootX;
                int cy = child.rootY;
                int cw = child.width;
                int ch = child.height;

                if (diffRegion.mayIntersect(cx, cy, cw, ch)) {
                    Region childRegion = gfxSys.makeRegion(cx, cy, cw, ch);

                    // calculate the real estate for the child and notify
                    childRegion.intersect(diffRegion);

                    child.parentChanged(childRegion, dx, dy);

                    // and subtract off this space from the diffRegion
                    diffRegion.subtract(cx, cy, cw, ch);
                } else {
                    child.invisible(true);
                }
            }
        }

        // any space remaining in diffRegion is our canvasRegion
        // (the clipping region for graphics primitives in our window)
        // If there's any of this that wasn't originally in the canvasRegion,
        // then we need to send some expose events.

        diffRegion.defragment(false);
        boolean wasUnobscured = isUnobscured();
	setStatus(UNOBSCURED, diffRegion.contains(rootX, rootY, width, height));
	setStatus(VISIBLE, !diffRegion.isEmpty());
        if (isVisible()) {
            if ((canvasRegion == null) || canvasRegion.isEmpty()) {
                canvasRegion = diffRegion;
                // must set before handleExposure
                handleExposure(diffRegion);
            } else {
		Rectangle bbox = canvasRegion.getBounds();
                if (isUnobscured() && wasUnobscured
                    && bbox.contains(rootX - dx, rootY - dy)
		    && bbox.contains(rootX - dx + width, rootY - dy + height)) {
                    // short cut
                } else {
                    if ((dx | dy) != 0) {
                        canvasRegion.translate(dx, dy);
                    }
                    Region exposureRegion = diffRegion.copy();
                    exposureRegion.subtract(canvasRegion);
                    canvasRegion = diffRegion;
                    // must set before handleExposure
                    handleExposure(exposureRegion);
                }
            }
        }

        canvasRegion = diffRegion;
    }

    private void translateAllRegions(int dx, int dy) {
        if ((windowRegion == null) || (canvasRegion == null)) {
            // this can happen if children are mapped but not
            // managed by the layout manager. It's a pathological
            // case, however.
            return;
        }
        windowRegion.translate(dx, dy);
        canvasRegion.translate(dx, dy);
	invalidateClip();
        for (Window c = child_head; c != null; c = c.next) {
            if (c.isMapped()) {
                c.translateAllRegions(dx, dy);
            }
        }
    }

    // The region available to our parent has changed, and therefore
    // the region available to us might have also changed.  Our new
    // window region is in newWindowRegion. This might be called on a
    // window that's just starting to be mapped, so we can't
    // count on windowRegion or canvasRegion.
    private void parentChanged(Region newWindowRegion, int dx, int dy) {
        // short cut if we're moving a window whose region hasn't
        // changed. Logic is convoluted in order to avoid doing a
        // region operation that isn't necessary.
        boolean OK = (windowRegion != null) && (canvasRegion != null);

        if (OK && ((dx | dy) != 0)) {
            windowRegion.translate(dx, dy);
        }

        if (OK) {
            newWindowRegion.defragment(false);
            OK = windowRegion.equals(newWindowRegion);
        } else {
            newWindowRegion.defragment(false);
        }

        if (OK) {
            if ((dx | dy) != 0) {
                canvasRegion.translate(dx, dy);
                for (Window c = child_head; c != null; c = c.next) {
                    if (c.mapped()) {
                        c.translateAllRegions(dx, dy);
                    }
                }
            }
	    invalidateClip();
            return;
        }

        windowRegion = newWindowRegion;
        if (canvasRegion == null) {
            // previous writing area was empty
            canvasRegion = gfxSys.makeRegion(0,0,0,0);
        }

        recalculateRegions(dx, dy);
    }

    // a parent has changed its position and therefore all the rootX
    // and rootY values need to be recalculated
    private void changeChildParentPosition(int dx, int dy) {

        if (parent != null) {
            rootX = x + parent.rootX;
            rootY = y + parent.rootY;
        } else {
            rootX = x;
            rootY = y;
        }

        for (Window w = child_head; w != null; w = w.next) {
            w.changeChildParentPosition(dx, dy);
        }
    }

    // a parent has changed its value of isMapped, and therefore
    // children need to be fixed up for areParentsMapped.
    // This does *not* send out exposure events or anything like that.
    private void changeChildParentMappings(boolean newValue) {
        if (!newValue) {
	    if (winSys.isInGrabStack(this)) {
		winSys.releaseGrabbingWindow(this);
	    }

	    if (this.hasInputFocus()) {
		winSys.setFocusWindow(null);
	    }

            windowRegion = null;
            canvasRegion = null;
	    setStatus(VISIBLE | UNOBSCURED, false);
	    invalidateClip();
        }

        for (Window w = child_head; w != null; w = w.next) {
            if (w.areParentsMapped() != newValue) {
		w.setStatus(PARENTS_MAPPED, newValue);
                if (w.mapped()) {
                    w.changeChildParentMappings(newValue);

		    int val = (newValue) ? ToolkitEventHandler.WINDOW_MAPPED
			                 : ToolkitEventHandler.WINDOW_UNMAPPED;

		    winSys.windowEventOccurred(w, System.currentTimeMillis(),
					       w.x - w.rootX, w.y - w.rootY,
					       width, height, val);
                }
            }
        }
    }

    private void handleExposure(Region exposureRegion) {
	// System.err.println("handleExposure " +exposureRegion);
        if (exposureRegion.isEmpty()) {
            return;
        }

	Rectangle exposeRectangle = exposureRegion.getBounds();
	Graphics g = getGraphics();
	g.translate(-rootX, -rootY);
	g.setClip(exposureRegion);
	g.clearRect(exposeRectangle.x, exposeRectangle.y, 
		    exposeRectangle.width, exposeRectangle.height);
	g.dispose();

	winSys.windowEventOccurred(this, System.currentTimeMillis(),
				 exposeRectangle.x - rootX,
				 exposeRectangle.y - rootY,
				 exposeRectangle.width,
				 exposeRectangle.height,
				 ToolkitEventHandler.WINDOW_DAMAGED);
    }

    private void handleExposure(Rectangle exposeRectangle) {
	// System.err.println("handleExposure " +exposeRectangle);
	Graphics g = getGraphics();
	g.clearRect(exposeRectangle.x - rootX, exposeRectangle.y - rootY, 
		    exposeRectangle.width, exposeRectangle.height);
	g.dispose();

	winSys.windowEventOccurred(this, System.currentTimeMillis(),
				 exposeRectangle.x - rootX,
				 exposeRectangle.y - rootY,
				 exposeRectangle.width,
				 exposeRectangle.height,
				 ToolkitEventHandler.WINDOW_DAMAGED);
    }

    /**
     * Test for whether a window contains the given point.
     * @param x The x coord of the point
     * @param y The y coord of the point
     * @return true if x,y is inside the window.
     */
    protected boolean contains(int x, int y) {
        return (windowRegion != null) && windowRegion.contains(x, y);
    }

    static final int ONSCREEN_MASK = HIERARCHY_MAPPED_MASK | DISPOSED;

    /**
     * Test for whether the window is currently on the screen (i.e. both
     * it and all its ancestors are mapped.
     */
    protected boolean onScreen() {
	return (status & ONSCREEN_MASK) == HIERARCHY_MAPPED_MASK;
    }

    /**
     * Set the cursor image associated with this window.
     * @param image The CursorImage to be used with this window.
     */
    public void setCursorImage(CursorImage image) {
	cursorImage = image;
    }

    /**
     * Get the cursor image associated with this window.
     * @return The CursorImage currently assigned to the window.
     */
    protected CursorImage getCursorImage() {
	return cursorImage;
    }

    /**
     * Create a String representation of this Window (for debugging).
     */
    public String toString() {
	return "[sun.awt.aw.Window #" +winNum()+ " at (" +rootX+ " " +rootY+ ")"
	    + " size " +width+ "x" +height+ "]";
    }

    /**
     * Repaint all windows in the given region.  This is normally used when 
     * the window system is running as an emulator, or when root depth has
     * been changed, e.g.
     * <strong>The window system lock must be held when this is called.</strong>
     * @param r The region to be refreshed
     * @param includeChildren repaint recursively
     */
    protected void refresh(sun.porting.graphicssystem.Region r,
			   boolean includeChildren) {

        if (canvasRegion != null) {
	    Region reg = (r == null ? canvasRegion : r);
            canvasRegion.defragment(true);
            handleExposure(reg);
        }

        if (includeChildren) {
            for (Window w = child_head; w != null; w = w.next) {
                if (w.mapped()) {
		    if (r == null) {
			w.refresh(null, true);
		    } else if (r.mayIntersect(w.windowRegion)) {
			Region reg = w.windowRegion.copy();
			reg.intersect(r);
			w.refresh(reg, true);
		    }
                }
            }
        }
    }
 
    /**
     * Use the system-specific default cursor.
     */
    protected void useDefaultCursor() {
	java.awt.Cursor c = 
	    new java.awt.Cursor(java.awt.Cursor.DEFAULT_CURSOR);
	defaultCursorImage = gfxSys.getCursorImage(c);
    }
}
