/*
 * @(#)WindowSystem.java	1.29 98/11/25
 *
 * Copyright (c) 1996, 1997 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information"). You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

package sun.awt.aw;

import sun.porting.windowsystem.EventHandler;
import sun.porting.windowsystem.WindowManagementFeedback;
import sun.porting.windowsystem.WindowFactory;
import sun.porting.windowsystem.Window;
import sun.porting.graphicssystem.GraphicsSystem;
import sun.porting.graphicssystem.Region;
import sun.porting.graphicssystem.CursorImage;

import sun.porting.toolkit.ToolkitEventHandler;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Point;

/*
 * Windows are created using a WindowFactory; this is done so that a
 * framebuffer-specific Window class can extend the general one below.
 * Setting the window factory must be done before the window system is
 * initialized.
 */
class WinFactory implements WindowFactory {
    WindowSystem winSys;
    
    public WinFactory(WindowSystem ws) {
	winSys = ws;
    }

    /**
     * @return a new window with the desired parent and bounds.
     */
    public synchronized Window makeWindow(Window parent, int x, int y,
					  int w, int h)
	throws IllegalArgumentException {

	if (winSys == null) {
	    throw new RuntimeException("no window system running??");
	}

	if (parent == null) {
	    throw new IllegalArgumentException("Parent may not be null");
	}

        return new sun.awt.aw.Window(winSys, (sun.awt.aw.Window) parent, 
				     x, y, w, h);
    }

    /**
     * @return a new top-level window with the desired type and bounds.
     */
    public synchronized Window makeTopLevelWindow(int type, int x, int y,
						  int w, int h)
	throws IllegalArgumentException {

	/* REMIND: type of window is ignored except for this validity check. */
	if (   (type != TOP_LEVEL_FRAME) 
	    && (type != TOP_LEVEL_DIALOG)
	    && (type != TOP_LEVEL_FRAMELESS)) {
	    throw new IllegalArgumentException("Not a valid type for a top-level window");
	}

	if (winSys == null) {
	    throw new RuntimeException("no window system running??");
	}


        return new sun.awt.aw.Window(winSys, 
				     (sun.awt.aw.Window)winSys.rootWindow, 
				     x, y, w, h);
    }
}

/**
 * The top-level class for controlling this window system.
 */
public class WindowSystem extends sun.porting.windowsystem.WindowSystem implements WindowManagementFeedback, EventHandler {

    // public interface for WindowSystem

    /**
     * Register a callback handler for receiving events.
     * @param h The callback handler, a ToolkitEventHandler object.
     * @throws IllegalStateException if there is already a handler
     * registered.
     */
    public void registerToolkitEventHandler(ToolkitEventHandler h) throws IllegalStateException {
	if (eventHandler != null) {
	    throw new IllegalStateException("ToolkitEventHandler already registered");
	}
	eventHandler = h;
        eventQueue = new WindowSystemEventQueue(eventHandler, this);
	eventQueue.start();

	// don't want to do this until we have an event handler registered!
	gfxSys.registerEventHandler(this);
	rootWindow.map();
    }

    /**
     * Get the root window for this windowsystem.
     * @return The root window.
     */
    public sun.porting.windowsystem.Window getRootWindow() {
	return rootWindow;
    }

    /**
     * Change the factory object used to create windows.  (The factory
     * mechanism is used in case a subclass of Window is needed.) 
     * If factory is null, use the window system's default factory.
     */
    public void setWindowFactory(WindowFactory factory) {
	synchronized (this) {
	    windowFactory = (factory == null) ? new WinFactory(this) : factory;
	}
    }

    /**
     * Create a new window as a child of the given parent window,
     * and having the specified dimensions.  If parent is null, create
     * a new top-level window.  Uses the current WindowFactory.
     * @param parent The parent window, or null for a top-level window
     * @param x the x coordinate for the window's upper left hand corner,
     * in the coordinate system of the parent window (global coordinates in
     * the case where parent is null.)
     * @param y the y coordinate for the window's upper left hand corner,
     * in the coordinate system of the parent window (global coordinates in
     * the case where parent is null.)
     * @param w the width of the window
     * @param h the height of the window
     * @return The new Window object.
     */
    public Window makeWindow(Window parent, int x, int y, int w, int h) {
	synchronized (this) {
	    return windowFactory.makeWindow(parent, x, y, w, h);
	}
    }

    /**
     * Create a new top-level window of the given type
     * and having the specified dimensions.  Uses the current 
     * <code>WindowFactory</code>.
     * @param windowType The type of the window
     * @param x the x coordinate for the window's upper left hand corner,
     * in the global coordinate system
     * @param y the y coordinate for the window's upper left hand corner,
     * in the global coordinate system
     * @param w the width of the window
     * @param h the height of the window
     * @return The new <code>Window</code> object.
     */
    public Window makeTopLevelWindow(int windowType,
				     int x, int y, int w, int h) {
	synchronized (this) {
	    return windowFactory.makeTopLevelWindow(windowType, x, y, w, h);
	}
    }

    /**
     * Get the window that has the keyboard input focus, or null if the
     * focus is not assigned to any window.
     * @return A Window object corresponding to the window that will receive
     * any keyboard events which occur.
     */
    public Window getFocusWindow() {
	synchronized (this) {
	    return focusOwner;
	}
    }

    /**
     * Get the window that has grabbed input, or null if none has.
     * @return A Window object corresponding to the window that has grabbed
     * input events.  Note that the grabbing window will not have the keyboard
     * focus unless it has been specifically assigned to it.
     */
    public Window getGrabbingWindow() {
	synchronized (this) {
 	    return grabStack.empty() ? null : 
 		((GrabData)grabStack.peek()).win;
	}
    }

    /**
     * Set the cursor image to match the one supplied.  The location
     * of the hotspot should be at (hotX, hotY) relative to the upper left
     * corner of the image.  Transparent pixels in the image are not painted.
     * @param image The contents of the cursor image
     * @param hotX  The x coordinate of the hotspot
     * @param hotY  The y coordinate of the hotspot
     */
    public void setCursorImage(CursorImage image) {
	synchronized (this) {
	    gfxSys.setCursorImage(image);
	}
    }


    // public interface for WindowManagementFeedback

    /**
     * Start window management feedback using a rectangle of the given
     * dimensions.  All coordinates are in global (framebuffer) space.
     * @param win The window to be reshaped.  (This allows the window system
     * to show the whole window during moves and/or reshapes).
     * @param x x coordinate of the feedback rectangle's upper-left corner.
     * @param y y coordinate of the feedback rectangle's upper-left corner.
     * @param w Width of the feedback rectangle.
     * @param h Height of the feedback rectangle.
     * @throws IllegalStateException if feedback is already in progress.
     * @throws IllegalArgumentException if win is not a top-level window.
     */
    public void startFeedback(Window win, int x, int y, int w, int h) 
	throws IllegalStateException, IllegalArgumentException {
	synchronized (this) {
	    moveFeedback(win, x, y, w, h, true);
	}
    }

    /**
     * Resize the feedback rectangle to the given dimensions.
     * @param win The window to be reshaped.  (This allows the window system
     * to show the whole window during moves and/or reshapes).
     * @param x x coordinate of the feedback rectangle's upper-left corner.
     * @param y y coordinate of the feedback rectangle's upper-left corner.
     * @param w Width of the feedback rectangle.
     * @param h Height of the feedback rectangle.
     * @throws IllegalStateException if no feedback is in progress.
     * @throws IllegalArgumentException if win is not the same window
     * passed to startFeedback.
     */
    public void moveFeedback(Window win, int x, int y, int w, int h)
	throws IllegalStateException, IllegalArgumentException {

	synchronized (this) {
	    // drawing before we update erases old rectangle
	    drawFeedback();
	    moveFeedback(win, x, y, w, h, true); // updates coordinates
	}
    }

    /**
     * End feedback; hide the feedback rectangle, for example.
     * When the window management operation in progress is about to be
     * completed, this should be called to clean up.
     * @param win The window to be reshaped.  (This allows the window system
     * to show the whole window during moves and/or reshapes).
     * @param x x coordinate of the feedback rectangle's upper-left corner.
     * @param y y coordinate of the feedback rectangle's upper-left corner.
     * @param w Width of the feedback rectangle.
     * @param h Height of the feedback rectangle.
     * @throws IllegalStateException if no feedback is in progress.
     * @throws IllegalArgumentException if win is not the same window
     * passed to startFeedback.
     */
    public void endFeedback(Window win, int x, int y, int w, int h) 
	throws IllegalStateException, IllegalArgumentException {

	synchronized (this) {
	    // drawing before we update erases old rectangle
	    drawFeedback();
	    moveFeedback(win, x, y, w, h, false); // don't update--just check args

	    clearFeedback();	// clear the variables
	}
    }


    //
    // (public) implementation of InputEventHandler functionality
    //

    /**
     * Called when an event on a pointing object (a mouse, track ball,
     * joystick etc) has occurred.
     * @param when Time at which the event occurred, in milliseconds since
     *        1/1/70 UTC.
     * @param dx Position change (x coordinate) since last pointer event
     * @param dy Position change (y coordinate) since last pointer event
     * @param ID Type identifier indicating the kind of event
     * @param number Used to distinguish among several buttons (etc) that 
     *        could have been pressed or released.
     * @see java.lang.System#currentTimeMillis
     */
    public void pointerEventOccurred(long when, int x, int y, int ID, int number) {
	// in order to avoid loading down the caller (who may be an interrupt
	// handler or other system-critical resource) we dump all events into
	// a queue and process them on a separate thread.
	eventQueue.addPointerEvent(when, x, y, ID, number);
    }

    /**
     * Called when an event on a key object (typewriter-style keyboard key,
     * button on a remote control, etc) has occurred.
     * @param when Time at which the event occurred, in milliseconds since
     *        1/1/70 UTC.
     * @param ID Type identifier indicating the kind of event
     * @param keycode A code identifying which key was pressed.
     * @param keychar Unicode character -- for TYPED events only.  (Events
     * with the TYPED id will only occur if the system is doing key-to-char
     * translation at a low level.)
     * @see java.awt.event.KeyEvent
     * @see java.lang.System#currentTimeMillis
     */
    public void keyboardEventOccurred(long when, int ID, int keycode, char keychar) {
	eventQueue.addKeyboardEvent(when, ID, keycode, keychar);
    }

    /**
     * Handle a graphicsChange event.  This event indicates either that
     * some windows got damaged (probably because the system is an emulator)
     * or some graphics parameter like the root window's size or depth was
     * modified.
     * @param when The timestamp for the event
     * @param r The region that needs to be repainted
     * @flags A set of flags indicating the nature of the change
     */
    public synchronized void graphicsChange(long when, Region r, int flags) {
	if ((flags & EventHandler.SIZE_CHANGE_FLAG) != 0) {
	    /* this can be implemented by unmapping and resizing */
	    /* the root window; then remapping it.  But I've chosen */
	    /* not to do so for now.  */

	    /* remapping the root window will cause a complete repaint, */
	    /* so we don't need to do another one! */
	    flags &= ~(EventHandler.GRAPHICS_INVALID_FLAG);

	    throw new RuntimeException("Changing root size is not implemented");
	}

	if ((flags & EventHandler.COLORMODEL_CHANGE_FLAG) != 0) {
	    /* This should probably not require anything of the window system */
	}

	if ((flags & EventHandler.GRAPHICS_INVALID_FLAG) != 0) {
	    ((sun.awt.aw.Window)rootWindow).refresh(r, true);
	}
    }

    /**
     * Handle a window event on the given window.  Just turns around and
     * posts it to the event queue.
     * @param win The window which experienced the event
     * @param timestamp The time when the event occurred
     * @param x The left edge of the rectangle affected
     * @param y The top edge of the rectangle affected
     * @param w The width of the area affected.
     * @param h The height of the area affected.
     * @param ID An identifier for the type of event which occurred.
     */
    protected void windowEventOccurred(Window win, long timestamp,
				       int x, int y, int w, int h, int ID) {
	if (win == rootWindow) {
	    return;
	}
	eventQueue.addWindowEvent(timestamp, x, y, w, h, 
				  ID, (sun.awt.aw.Window) win);
    }


    //
    // (private) Implementation of WindowSystem functionality
    //

    // The root window and a java.awt.Graphics object that draws
    // on the root window.  
    Window rootWindow = null;

    // Note: the system assumes that it has
    // exclusive use of the rootGraphics object. If you want to draw
    // on the root window, you must make a copy of the rootGraphics
    // object and reinitialize all fields of interest.
    Graphics rootGraphics = null;

    // The event handler supplied (i.e. registered) by the toolkit.
    ToolkitEventHandler eventHandler = null;

    // The window that currently owns the focus.  Often not null.
    Window focusOwner = null;

    // The stack of windows that have grabbed input events; top of the stack is the current grab owner.
    // Usually empty unless a modal dialog is visible.
    java.util.Stack grabStack = new java.util.Stack();
    boolean silentGrab = false;
    boolean discardEvents = false;
    boolean deliverAll = false;

    // data kept on grabStack
    class GrabData {
	sun.porting.windowsystem.Window win;
	int mode;
	GrabData(sun.porting.windowsystem.Window w, int m) {
	    win = w;
	    mode = m;
	}
	public boolean equals(Object gd) {
	    return win == ((GrabData)gd).win;
	}
    }

    /**
     * Set the focus window to the given window.  The window system lock
     * must be held before calling this.
     * @param w The window to receive text input focus.
     */
    protected void setFocusWindow(sun.porting.windowsystem.Window w) {
        if (focusOwner == w) {
	    return;
	}
	sun.awt.aw.Window win;

	if (focusOwner != null) {
	    win = (sun.awt.aw.Window) focusOwner;
	    windowEventOccurred(focusOwner, System.currentTimeMillis(), 
				win.x, win.y, win.width, win.height,
				ToolkitEventHandler.WINDOW_LOST_FOCUS);
	}

	focusOwner = w;
	if (w == null) {
	    return;
	}

	win = (sun.awt.aw.Window) w;
	windowEventOccurred(w, System.currentTimeMillis(), 
			    win.x, win.y, win.width, win.height,
			    ToolkitEventHandler.WINDOW_GOT_FOCUS);
    }

//     private void printGrabStack() {
// 	if (grabStack.empty())
// 	    System.out.println("GrabStack: [empty]");
// 	else {
// 	    System.out.println("GrabStack:");
// 	    java.util.Enumeration e = grabStack.elements();
// 	    while (e.hasMoreElements())
// 		System.out.println("   " + ((GrabData)e.nextElement()).win);
// 	}
//     }

    /***
     * Determine whether this window is in the stack of windows that have grabbed input.
     * (Note that the grab owner is the window at the top of the stack)
     * @param win The window whose grab status is to be checked.
     */
    protected boolean isInGrabStack(sun.porting.windowsystem.Window win) {
	synchronized (this) {
	    return grabStack.contains(new GrabData(win, 0));
	}
    }

    /**
     * Assign the input grab to the given window.  The window system lock
     * must be held before calling this.
     * @param w The window to receive the input grab
     * @param mode The mode of the grab (e.g. silent).
     */
    protected void setGrabbingWindow(sun.porting.windowsystem.Window w,
				     int mode) {
	eventQueue.redirectImplicitGrab(w);
	deliverAll    = (mode == w.DELIVER_ALL);
	discardEvents = (mode != w.DELIVER_ALL); 
	silentGrab    = (mode == w.DISCARD_SILENT);
	grabStack.push(new GrabData(w, mode));
    }

    /**
     * Release the grab held by the given window.  The window system lock
     * must be held before calling this.  If the given window is at the
     * top of the grab stack, the grab is passed along to its predecessor.
     * @param w The window releasing the input grab
     * @throws IllegalStateException if the grab stack is empty.
     */
    protected void releaseGrabbingWindow(sun.porting.windowsystem.Window w) {
	if (grabStack.isEmpty())
	    throw new IllegalStateException("Grab stack underflow?");

	GrabData topGrab = (GrabData)grabStack.peek();
	if (w == topGrab.win) {
	    grabStack.pop();
	    if (grabStack.isEmpty()) {
		eventQueue.releaseImplicitGrab(topGrab.win);
		deliverAll = discardEvents = silentGrab = false;
	    } else {
		topGrab = (GrabData)grabStack.peek();
		eventQueue.redirectImplicitGrab(topGrab.win);
		deliverAll    = (topGrab.mode == w.DELIVER_ALL);
		discardEvents = (topGrab.mode != w.DELIVER_ALL); 
		silentGrab    = (topGrab.mode == w.DISCARD_SILENT);
	    }
	} else {
	    // window releasing grab is not the top window, so we can't exactly "pop" it, but we do need
	    // to remove it from the grab stack - don't need to do anything else in this case
	    grabStack.removeElement(new GrabData(w, 0));
	}
    }

    /**
     * Indicates that the event queue discarded an event because of a grab.
     */
    protected void eventDiscarded() {
	if (!silentGrab) {
	    beep();
	}
    }

    // cursor position support
    Window cursorWindow = null;
    int cursorX = 0;
    int cursorY = 0;

    // window system must be locked during this call
    void handleCursorMovement(Point oldPos, Point newPos) {
	if (oldPos == null) {
	    cursorX = newPos.x;
	    cursorY = newPos.y;
	    checkCursorBounds();

	    findCursorWindow(null, cursorX, cursorY);
	} else {
	    if (      cursorWindow != null
		   && ((cursorX != oldPos.x) || (cursorY != oldPos.y))) {
		throw new RuntimeException("Cursor location out of sync");
	    }

	    cursorX = newPos.x;
	    cursorY = newPos.y;
	    checkCursorBounds();
	    cursorWindow = findCursorWindow(cursorWindow, cursorX, cursorY);
	}

	resetCursorWindow();
    }

    void handleCursorMovement(int x, int y) {
	cursorX = x;
	cursorY = y;

	checkCursorBounds();
	resetCursorWindow();
    }

    private final void checkCursorBounds() {
	if (cursorX < 0) cursorX = 0;
	if (cursorX >= width) cursorX = width - 1;
	if (cursorY < 0) cursorY = 0;
	if (cursorY >= height) cursorY = height - 1;
    }

    /* package-private */void resetCursorWindow() {
	Window oldCursorWindow = cursorWindow;
	cursorWindow = findCursorWindow(cursorWindow, cursorX, cursorY);

	if (cursorWindow != oldCursorWindow) {
	    gfxSys.setCursorImage(((sun.awt.aw.Window) cursorWindow).getCursorImage());
	}
    }

    // window system must be locked during this call
    /* package-private */
    sun.awt.aw.Window findCursorWindow(Window startingPoint, int x, int y) {
	sun.awt.aw.Window scanWindow = (sun.awt.aw.Window) startingPoint;
	long time = System.currentTimeMillis();

	// System.err.println("findCursorWindow(win, " +x+ "," +y+ ") starting at " +scanWindow);

	if ((scanWindow == null) || (scanWindow.isDisposed())) {
	    sun.awt.aw.Window w = locateWindow((sun.awt.aw.Window) rootWindow, x, y);

	    windowEventOccurred(w, time,
				     w.x + x - w.rootX, w.y + y - w.rootY,
				     w.width, w.height, 
				     ToolkitEventHandler.WINDOW_ENTERED);
	    return w;
	} else {
	    while ((scanWindow != rootWindow) && !scanWindow.contains(x, y)) {
		// System.err.println("findCursorWindow: up from " +scanWindow);
		// queue an exit/enter pair
		windowEventOccurred(scanWindow, time,
				    scanWindow.x + x - scanWindow.rootX, 
				    scanWindow.y + y - scanWindow.rootY,
				    scanWindow.width, scanWindow.height, 
				    ToolkitEventHandler.WINDOW_EXITED);
		scanWindow = scanWindow.parent;
		windowEventOccurred(scanWindow, time,
				    scanWindow.x + x - scanWindow.rootX, 
				    scanWindow.y + y - scanWindow.rootY,
				    scanWindow.width, scanWindow.height, 
				    ToolkitEventHandler.WINDOW_ENTERED);
	    }
	    
	    sun.awt.aw.Window prevWindow = scanWindow;
	    scanWindow = scanWindow.child_head;
	    while (scanWindow != null) {
		if (scanWindow.onScreen() && scanWindow.contains(x, y)) {
		    // System.err.println("findCursorWindow: down into " +scanWindow);
		    windowEventOccurred(prevWindow, time,
					prevWindow.x + x - prevWindow.rootX, 
					prevWindow.y + y - prevWindow.rootY,
					prevWindow.width, prevWindow.height, 
					ToolkitEventHandler.WINDOW_EXITED);
		    prevWindow = scanWindow;
		    windowEventOccurred(prevWindow, time,
					prevWindow.x + x - prevWindow.rootX, 
					prevWindow.y + y - prevWindow.rootY,
					prevWindow.width, prevWindow.height, 
					ToolkitEventHandler.WINDOW_ENTERED);

		    // now descend another level
		    scanWindow = scanWindow.child_head;
		} else {
		    scanWindow = scanWindow.next;
		}
	    }

	    return prevWindow;
	}
    }

    sun.awt.aw.Window locateWindow(sun.awt.aw.Window scanWindow, int x, int y) {
	sun.awt.aw.Window prevWindow = null;

	while (scanWindow != null) {
	    if (scanWindow.onScreen() && scanWindow.contains(x, y)) {
		prevWindow = scanWindow;
		scanWindow = scanWindow.child_head;
	    } else {
		scanWindow = scanWindow.next;
	    }
	}

	return (prevWindow == null) ? (sun.awt.aw.Window) rootWindow : prevWindow;
    }

    //
    // (private) Implementation of window management feedback functionality
    //

    private void moveFeedback(Window win, 
				int x, int y, int w, int h, boolean update) {
	if (fbWin != null) {
	    throw new IllegalStateException("Feedback already in progress");
	}
	if ((w <= 0) || (h <= 0)) {
	    throw new IllegalArgumentException("Width and height must be > 0");
	}
	if (win == null) {
	    throw new IllegalArgumentException("Window may not be null");
	}
	if (((sun.awt.aw.Window)win).parent != null) {
	    throw new IllegalArgumentException("not a top-level window");
	}

	if (update) {
	    fbX = x;
	    fbY = y;
	    fbW = w;
	    fbH = h;
	    fbWin = win;

	    drawFeedback();
	}
    }

    private void drawFeedback() {
	if (fbG == null) {
	    if (rootGraphics == null) {
		throw new IllegalStateException("Can't draw onto framebuffer");
	    }
	    if (rootGraphics == null) {
		rootGraphics = rootWindow.getGraphics();
	    }
	    fbG = rootGraphics.create();
	    fbG.setXORMode(Color.white);
	}

	fbG.drawRect(fbX, fbY, fbW, fbH);
    }

    private void clearFeedback() {
	fbX = fbY = fbW = fbH = -1; // for safety's sake....
	fbWin = null;
    }

    // the coordinates of the feedback rectangle
    private int fbX = -1;
    private int fbY = -1;
    private int fbW = -1;
    private int fbH = -1;

    // the current feedback window; null if no feedback in progress.
    private Window fbWin = null;

    // a Graphics object used only to draw feedback.
    private Graphics fbG = null;

    int width;
    int height;

    /**
     * Create a new WindowSystem instance.
     */
    public WindowSystem() {
	super();

	java.awt.Dimension d = gfxSys.getScreen().getSize();
	resize(d.width, d.height);
    }

    /**
     * Create a new WindowSystem instance, with a root of the given size.
     */
    public WindowSystem(int w, int h) {
	super();

	resize(w, h);
    }

    /**
     * Start this WindowSystem running, create a top-level window, etc.
     */
    public void start() {
	setWindowFactory(windowFactory);
	rootWindow   = 
	    windowFactory.makeTopLevelWindow(WindowFactory.TOP_LEVEL_FRAMELESS,
					     0, 0, width, height);

	rootWindow.setBackgroundColor(java.awt.Color.black);
    }

    /**
     * Change the size of the root window.
     */
    public synchronized void resize(int w, int h) {
	if ((w <= 0) || (h <= 0)) {
	    throw new RuntimeException("Width and Height must be positive");
	}

	if (rootWindow != null) {
	    rootWindow.setBounds(0, 0, w, h);
	}
	this.width  = w;
	this.height = h;
    }

    // expose gfxSys within this package.
    /* package-private */GraphicsSystem getGS() {
	return gfxSys;
    }

    private WindowSystemEventQueue eventQueue;

    /**
     * The window factory object that this WindowSystem is using
     * to create new windows.
     */
    protected WindowFactory windowFactory;
}
