/*
 * @(#)WindowSystemEventQueue.java	1.18 99/07/29 SMI
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package sun.awt.aw;

import sun.porting.toolkit.ToolkitEventHandler;
import sun.porting.windowsystem.EventHandler;

class WindowSystemEventQueue extends Thread {
    /**
     * Create an event queue instance which will take events from the
     * given window system, enqueue them, and deliver them to the toolkit
     * via the supplied handler.
     * @param handler The toolkit's event handler
     * @param ws The window system which is the source of the events.
     */
    public WindowSystemEventQueue(ToolkitEventHandler handler,
                                  WindowSystem ws) {
        super("aw Event Queue");
        setPriority(NORM_PRIORITY + 1); // if there are events, get them 
				        // delivered preferentially.
        eventHandler = handler;
        winSys       = ws;

        queue.start();
    }

    static final int POINTER_EVENT    = ToolkitEventHandler.POINTER_EVENT_START;
    static final int POINTER_MOVED    = POINTER_EVENT | EventHandler.NONE;
    static final int POINTER_PRESSED  = POINTER_EVENT | EventHandler.PRESSED;
    static final int POINTER_RELEASED = POINTER_EVENT | EventHandler.RELEASED;

    sun.porting.windowsystem.Window implicitGrab = null;

    /**
     * Redirect an implicit grab (i.e. a grab that was in response to a mouse
     * press) so that the rest of the events go to the supplied window.
     * An implicit grab means that the window which received the mouse down 
     * event is guaranteed to get the drag events and the up.
     * @param w The window to receive the rest of the mouse events.
     */
    protected synchronized void redirectImplicitGrab(sun.porting.windowsystem.Window w) {
        if (implicitGrab != null) {
            implicitGrab = w;
        }
    }

    /**
     * Release the implicit grab owned by the supplied window.  If that window
     * does not currently own an implicit grab, does nothing.
     * @param w The window which is asking to release the implicit grab.
     */
    protected synchronized void releaseImplicitGrab(sun.porting.windowsystem.Window w) {
	if (implicitGrab == w) {
	    implicitGrab = null;
	}
    }

    /**
     * Enqueue a pointer-related event.
     * @param when The timestamp for the event
     * @param x The x coordinate of the event
     * @param y The y coordinate of the event
     * @param ID The type of event
     * @param number Optional data (e.g. the number of the mouse button) 
     * associated with the event.
     */
    public synchronized void addPointerEvent(long when, int x, int y, 
                                             int ID, int number) {

        QueueRec queueRec = queue.getLastFullSlot();
        boolean add = false;

        // prevent overruns; if this queue can't keep up with mouse movements,
        // our best bet is to collapse them!
        if (   (queueRec == null) 
            || (queueRec.ID != POINTER_MOVED) || (ID != EventHandler.NONE)) {
            queueRec = queue.getNewEvent();
            add = true;
        }
        queueRec.x = x;
        queueRec.y = y;

        queueRec.timestamp = when;
        queueRec.ID        = ID | POINTER_EVENT; // cheap way to make ID unique
        queueRec.w         = number;

        if (add) {
            queue.addEvent(queueRec);
        }

        // System.err.println("addPointerEvent " +when+ " " +x+ " " +y+ " " +ID+ " " +number);
    }

    static final int KEY_EVENT = ToolkitEventHandler.KEY_EVENT_START;
    static final int KEY_PRESSED  = KEY_EVENT | EventHandler.PRESSED;
    static final int KEY_RELEASED = KEY_EVENT | EventHandler.RELEASED;
    static final int KEY_ACTION   = KEY_EVENT | EventHandler.ACTION;
    static final int KEY_TYPED    = KEY_EVENT | EventHandler.TYPED;


    /**
     * Enqueue a keyboard-related event.
     * @param when The timestamp for the event
     * @param ID The type of event
     * @param keycode The virtual key code for the event
     * @param keychar The character typed.
     */
    public synchronized void addKeyboardEvent(long when, int ID, 
                                              int keycode, char keychar) {

        QueueRec queueRec = queue.getNewEvent();

        queueRec.timestamp = when;
        queueRec.ID        = ID | KEY_EVENT; // cheap way to make ID unique
        queueRec.w         = keycode;
        queueRec.h         = keychar;

        queue.addEvent(queueRec);

        // System.err.println("addKeyboardEvent " +when+ " " +ID+ " " +keycode+ " " +keychar);
    }

    /**
     * Enqueue a window-related event.
     * @param when The timestamp for the event
     * @param x The x coordinate of the event
     * @param y The y coordinate of the event
     * @param w The width of the region affected by the event
     * @param w The height of the region affected by the event
     * @param ID The type of event
     * @param win The window which generated the event
     */
    public synchronized void addWindowEvent(long timestamp,
					    int x, int y, int w, int h, 
                                            int ID, Window win) {

        QueueRec queueRec = queue.getNewEvent();

        queueRec.timestamp = timestamp;
        queueRec.ID        = ID; // ID is already unique
        queueRec.x         = x;
        queueRec.y         = y;
        queueRec.w         = w;
        queueRec.h         = h;
        queueRec.win       = win;

        queue.addEvent(queueRec);

        // System.err.println("addWindowEvent " +timestamp+ " " +x+ " " + +y+ " " +w+ " " +h+ " " +ID+ " " +win);
        // Thread.dumpStack();
    }

    /**
     * Deliver events.  (This is the implementation of the Runnable interface,
     * required by the Thread.)
     */
    public void run() {
        QueueRec queueRec = null;
        // Preallocate exceptions because we don't want to be
        // allocating memory in the run thread!
        EventsGrabbedException notGrabOwner = 
            new EventsGrabbedException("The grab is owned by someone else.");
        UnknownEventTypeException unknownEventType =
            new UnknownEventTypeException("Unrecognized event type");


        while (true) {
            try {
                if (queueRec != null) {
                    queue.releaseEvent(queueRec); // we use the event from getNextEvent
                    queueRec = null;
                }

                queueRec = queue.getNextEvent();

                if (    (queueRec.ID > ToolkitEventHandler.WINDOW_EVENT_START)
                     && (queueRec.ID < ToolkitEventHandler.WINDOW_EVENT_END) ) {

		    if (queueRec.win == null) {
			continue;
		    }

		    if (!queueRec.win.isMapped()) {
		        int id = queueRec.ID;
		        if ((id != ToolkitEventHandler.WINDOW_UNMAPPED)
		         && (id != ToolkitEventHandler.WINDOW_LOST_FOCUS)) {
			    continue;
			}
		    }

		    eventHandler.windowEventOccurred(queueRec.win,
						     queueRec.timestamp,
						     queueRec.x, queueRec.y,
						     queueRec.w, queueRec.h,
						     queueRec.ID);
                } else if ((queueRec.ID & POINTER_EVENT) == POINTER_EVENT) {
                    synchronized (winSys) {
                        if ((queueRec.x | queueRec.y) != 0) {
                            winSys.handleCursorMovement(queueRec.x, queueRec.y);
                        }

                        sun.porting.windowsystem.Window dst = 
                            winSys.cursorWindow;

                        if ((queueRec.ID != POINTER_MOVED) || pointerIsDown) {
                            if (implicitGrab != null) {
 				if (implicitGrab.isMapped()){
 				    dst = implicitGrab;
 				} else {
 				    implicitGrab = null;
 				}
                            }

                            if (winSys.getGrabbingWindow() != null) {
                                boolean OK = (dst != null)
                                    && ((sun.awt.aw.Window)dst).hasGrab();

                                if (winSys.discardEvents && !OK) {
                                    throw notGrabOwner;
                                } else if (winSys.deliverAll) {
                                    dst = winSys.getGrabbingWindow();
                                }
                            }
                        }

                        queueRec.win = (sun.awt.aw.Window)dst;
                        queueRec.x   = winSys.cursorX - queueRec.win.rootX;
                        queueRec.y   = winSys.cursorY - queueRec.win.rootY;
                    } // window system lock is released before event is delivered

                    if ((queueRec.win == null) || !queueRec.win.isMapped()) {
                        continue;
                    }

                    switch (queueRec.ID) {
                    case POINTER_MOVED:
                        if (pointerIsDown) {
                            queueRec.ID  = ToolkitEventHandler.POINTER_DRAGGED;
                        } else {
                            queueRec.ID  = ToolkitEventHandler.POINTER_MOVED;
                        }
                        break;

                    case POINTER_PRESSED:
                        queueRec.ID  = ToolkitEventHandler.POINTER_PRESSED;
                        pointerIsDown = true;
                        implicitGrab = queueRec.win;
                        break;

                    case POINTER_RELEASED:
                        queueRec.ID  = ToolkitEventHandler.POINTER_RELEASED;
                        pointerIsDown = false;
                        implicitGrab = null;
                        break;

                    default:
                        throw unknownEventType;
                    }

                    eventHandler.pointerEventOccurred(queueRec.win,
                                                      queueRec.timestamp,
                                                      queueRec.x, queueRec.y,
                                                      queueRec.ID, queueRec.w);

                } else if ((queueRec.ID & KEY_EVENT) == KEY_EVENT) {
                    synchronized (winSys) {
                        queueRec.win = (sun.awt.aw.Window) winSys.focusOwner;

                        if (winSys.getGrabbingWindow() != null) {
                            boolean OK = (queueRec.win != null)
                                && ((sun.awt.aw.Window)queueRec.win).hasGrab();

                            if (winSys.discardEvents && !OK) {
                                throw notGrabOwner;
                            } else if (winSys.deliverAll) {
                                queueRec.win = (sun.awt.aw.Window)
				    winSys.getGrabbingWindow();
                            }
                        }

                        if ((queueRec.win == null) || !queueRec.win.isMapped()) {
                            continue;
                        }
                    } // release the window system lock right away

                    switch (queueRec.ID) {
                    case KEY_PRESSED:
                        lastKeyEventWasAction = false;
                        queueRec.ID = ToolkitEventHandler.KEY_PRESSED;
                        break;

                    case KEY_RELEASED:
                        if (lastKeyEventWasAction) {
                            queueRec.ID = ToolkitEventHandler.ACTION_RELEASED;
                        } else {
                            queueRec.ID = ToolkitEventHandler.KEY_RELEASED;
                        }
                        break;

                    case KEY_ACTION:
                        lastKeyEventWasAction = true;
                        queueRec.ID = ToolkitEventHandler.ACTION_PRESSED;
                        break;

                    case KEY_TYPED:
                        queueRec.ID = ToolkitEventHandler.KEY_TYPED;
                        break;

                    default:
                        throw unknownEventType;
                    }

                    eventHandler.keyboardEventOccurred(queueRec.win, 
                                                       queueRec.timestamp,
                                                       queueRec.ID,
                                                       queueRec.w, (char) queueRec.h);

                } else {
                    throw unknownEventType;
                }
            } catch (EventsGrabbedException e) {
                try {
                    // System.err.println(e);
                    winSys.eventDiscarded();
                } catch (Throwable t) {
                }
            } catch (UnknownEventTypeException e) {
                try {
                    // The event will be discarded.
                    // Perhaps this should beep, or something??
                } catch (Throwable t) {
                }
            } catch (Throwable ignored) {
                // this queue has got to keep running!
                try {
                    System.err.println("Exception occurred in event delivery thread:");
                    ignored.printStackTrace();
                } catch (Throwable t) {
                }
            }
        }
    }

    private ToolkitEventHandler eventHandler = null;
    private WindowSystem winSys = null;
    private boolean pointerIsDown = false;
    private boolean lastKeyEventWasAction = false;
    private EventQueueManager queue = new EventQueueManager(16, 4);
}

class UnknownEventTypeException extends IllegalArgumentException {
    public UnknownEventTypeException(String s) {
        super(s);
    }
}

class EventsGrabbedException extends IllegalStateException {
    public EventsGrabbedException(String s) {
        super(s);
    }
}

class EventQueueManager extends Thread {
    public EventQueueManager(int desiredSize, int lowWater) {
        this.desiredReserve = desiredSize;
        this.lowWaterMark   = lowWater;
    }

    QueueRec queue_head;
    QueueRec queue_tail;
    // Event queue management
    public synchronized void addEvent(QueueRec q) {
        if (queue_head == null) {
            queue_head = q;
        } else {

if (queue_tail.timestamp > q.timestamp) {
System.err.println("Events out of order: " +q);
System.err.println("was delivered after  " +queue_tail);
}
            queue_tail.next = q;
        }
        queue_tail = q;
        q.next = null;

        notifyAll();
    }

    public synchronized QueueRec getLastFullSlot() {
        return queue_tail;
    }

    public synchronized QueueRec getNextEvent() {
        while (queue_head == null) {
            try {
                wait();
            } catch (InterruptedException ignored) {
            }
        }

        QueueRec q = queue_head;
        queue_head = queue_head.next;
        return q;
    }

    // free list management
    private int desiredReserve  = 16;
    private int lowWaterMark    =  4;
    private int currentReserve  = 0;
    private QueueRec freeList   = new QueueRec();
    private Object freeListLock = new Object();

    public void run() {
        // whenever currentReserve drops below the low water mark,
        // this thread kicks in and adds new events to get us up
        // to the desired reserve.
        while (true) {
            try {
                synchronized (freeListLock) {
                    while (currentReserve > lowWaterMark) {
                        freeListLock.wait();
                    }
                }

                while (currentReserve <= desiredReserve) {
                    releaseEvent(new QueueRec());
                }
            } catch (Throwable t) {
                // mustn't exit
            }
        }
    }

    public QueueRec getNewEvent() {
        if (currentReserve != 0) {
            synchronized (freeListLock) {
                if (freeList != null) {
                    QueueRec q = freeList;
                    freeList = freeList.next;
                    --currentReserve;

                    freeListLock.notifyAll();
                    return q;
                }
            }
        }

        return new QueueRec();
    }

    public void releaseEvent(QueueRec q) {
        if (currentReserve > desiredReserve) {
            return;             // let the event get collected
        }

        synchronized (freeListLock) {
            q.next = freeList;
            freeList = q;
            ++currentReserve;
        }
    }
}

class QueueRec {
    public QueueRec() {
    }

    public long timestamp;
    public int ID;
    public int x;
    public int y;
    public int w;
    public int h;
    public Window win;

    public QueueRec next;
}
