/*
 * @(#)AdjustableObject.java	1.5 98/11/29 SMI
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package sun.awt.otk;

import java.awt.Adjustable;
import java.awt.AWTEventMulticaster;
import java.awt.event.AdjustmentEvent;
import java.awt.event.AdjustmentListener;

/**
 * An object which implements the Adjustable interface.  Useful for
 * views which want to interact with a scroll control that requires
 * an Adjustable.
 */
public class AdjustableObject implements Adjustable {
    private int orientation;
    private int minimum;
    private int maximum;
    private int value;
    private int visibleAmount;
    private int unitIncrement;
    private int blockIncrement;
    private AdjustmentListener adjustmentListener;

    /**
     * Creates an AdjustableObject that has orientation 
     * <code> orientation</code> and initial range.
     * If orientation is not <code>Adjustable.HORIZONTAL</code>, it defaults
     * to <code>Adjustable.VERTICAL</code>. 
     * Minimum is set to <code> min </code>, maximum is set to be greater than
     * minimum by at least 1. Value defaults to minimum. And 
     * length of the proportional indicator is set to <code> vis </code>, 
     * if it is not larger than minimum-maximum. 
     * Otherwise it defaults to minimum - maximum.
     * Unit increment is set to 1. 
     * While block increment is set to be visible amount - unitIncrement
     * but at least 1.
     * @param  orientation  orientation of AdjustableObject
     * @param  min          minimum value of AdjustableObject
     * @param  max          maximum value of AdjustableObject
     * @param  vis          visible amount of AdjustableObject
     */
    public AdjustableObject(int orientation, int min, int max, int vis){
	if (orientation == Adjustable.HORIZONTAL){
	    this.orientation = orientation;
	} else {
	    this.orientation = Adjustable.VERTICAL;
	}

	value = min;
	setUnitIncrement(1);
	setSpan(min, max, vis);  
    }

    /**
     * Retrieves the adjustable's orientation
     * @return the adjustable's orientation. This value could be either 
     * java.awt.VERTICAL or java.awt.HORIZONTAL.
     * @see java.awt.Adjustable#VERTICAL
     * @see java.awt.Adjustable#HORIZONTAL
     */
    public int getOrientation() {
	return orientation;
    }

    /**
     * Sets the minimum value of the adjustable object.
     * Maximum, length of the proportional indicator, and blockIncrement
     * will be adjusted if needed.
     * See constructor description for details.
     * @param    min  minimum value of adjustable object 
     */
    public synchronized void setMinimum(int min) {
	setSpan(min, maximum, visibleAmount);
    }

    /**
     * Retrieves the adjustable's minimum value.
     * @return the adjustable's minimu value that the adjustable can attain.
     *         See constructor description for details.
     * @see #setMinimum
     */
    public synchronized int getMinimum() {
	return minimum;
    }

    /**
     * Sets the maximum value of the adjustable object.
     * Length of the proportional indicator, and blockIncrement 
     * will be adjusted if needed.
     * See constructor description for details.
     * @param   max  maximum value of adjustable object
     */
    public synchronized void setMaximum(int max) {
	setSpan(minimum, max, visibleAmount);
    }

    /**
     * Retrieves the adjustable's maximum value.
     * @return the adjustable's maximum value that the adjustable can attain.
     *         See constructor description for details.
     * @see #setMaximum
     */
    public synchronized int getMaximum() {
	return maximum;
    }

    /**
     * Sets the unit value increment for the adjustable object.
     * It determines the amount to increase or decrease the adjustable 
     * value when an java.awt.event.AdjustmentEvent.UNIT_INCREMENT or 
     * java.awt.event.AdjustmentEvent.UNIT_DECREMENT is recieved
     * by adjustement listener.
     * @param u the unit increment
     */
    public synchronized void setUnitIncrement(int u) {
	unitIncrement = u;
    }

    /**
     * Retrieves the unit value increment for the adjustable object.
     * @return the unit increment which determines the amount to increase or
     *         decrease the adjustable value when an 
     *         java.awt.event.AdjustmentEvent.UNIT_INCREMENT or 
     *         java.awt.event.AdjustmentEvent.UNIT_DECREMENT is recieved
     *         by adjustement listener.
     * @see #setUnitIncrement
     */
    public synchronized int getUnitIncrement() {
	return unitIncrement;
    }

    /**
     * Sets the block value increment for the adjustable object.
     * It determines the amount to increase or decrease the adjustable 
     * value when an java.awt.event.AdjustmentEvent.BLOCK_INCREMENT or 
     * java.awt.event.AdjustmentEvent.BLOCK_DECREMENT is recieved
     * by adjustement listener.
     * @param b the block increment
     */
    public synchronized void setBlockIncrement(int b) {
	blockIncrement = b;
    }

    /**
     * Retrieves the block value increment for the adjustable object.
     * @return the block increment which determines the amount to increase or
     *         decrease the adjustable value when an 
     *         java.awt.event.AdjustmentEvent.BLOCK_INCREMENT or 
     *         java.awt.event.AdjustmentEvent.BLOCK_DECREMENT is recieved
     *         by adjustement listener.
     * @see #setBlockIncrement
     */
    public synchronized int getBlockIncrement() {
	return blockIncrement;
    }

    /**
     * Sets the length of the proportional indicator of the
     * adjustable object. See constructor descrition.
     * BlockIncrement will be adjusted if needed.
     * @param v the length of the indicator
     */
    public synchronized void setVisibleAmount(int v) {
	setSpan(minimum, maximum, v);
    }

    /**
     * Retrieves the length of the propertional indicator.
     * @returns current length of propertional indicator.
     * @see #setVisibleAmount
     */
    public synchronized int getVisibleAmount() {
	return visibleAmount;
    }

    /**
     * Sets the current value of the adjustable object. This
     * value must be within the range defined by the minimum and
     * (maximum - length of the proportional indicator) values for this object.
     * If the value was changed and adjustment listener was added, 
     * its adjustmentValueChanged method will be called with 
     * an AdjustmentEvent.
     * @param v the current value 
     * @see java.awt.event.AdjustmentEvent
     * @see java.awt.event.AdjustmentListener
     * @see #addAdjustmentListener
     */
    public synchronized void setValue(int v) {
	// bounds check
	v = Math.max(v, minimum);
	v = Math.min(v, maximum - visibleAmount);

	int delta = v - value;

        if (delta != 0) {
	    value = v;

	    // Synchronously notify the listeners so that they are 
	    // guaranteed to be up-to-date with the Adjustable before
	    // it is mutated again.
	    //
	    // Attempt to send a UNIT_* or BLOCK_* event if the delta
	    // matches, in order to give the client as much useful
	    // information as possible.
	    int adjustmentType;
	    if (delta == unitIncrement) {
		adjustmentType = AdjustmentEvent.UNIT_INCREMENT;
	    } else if (delta == -unitIncrement) {
		adjustmentType = AdjustmentEvent.UNIT_DECREMENT;
	    } else if (delta == blockIncrement) {
		adjustmentType = AdjustmentEvent.BLOCK_INCREMENT;
	    } else if (delta == -blockIncrement) {
		adjustmentType = AdjustmentEvent.BLOCK_DECREMENT;
	    } else {
		adjustmentType = AdjustmentEvent.TRACK;
	    }

	    AdjustmentEvent e =
		new AdjustmentEvent(this, 
				    AdjustmentEvent.ADJUSTMENT_VALUE_CHANGED,
				    adjustmentType, value);

	    adjustmentListener.adjustmentValueChanged(e);
	}
    }

    /**
     * Retrieves the current value of the adjustable object.
     * @return the current value of the adjustable object.
     * @see #setValue
     */
    public synchronized int getValue() {
	return value;
    }

    /**
     * Adds a listener for receiving adjustment events.
     * Adjustment events are fired whenever the value maintained in
     * adjustable object is changed.
     * @param   l  listener to receive adjustment events
     * @see java.awt.event.AdjustmentEvent
     * @see java.awt.event.AdjustmentListener
     * @see #removeAdjustmentListener
     * @see #setValue
     */
    public synchronized void addAdjustmentListener(AdjustmentListener l) {
	adjustmentListener = AWTEventMulticaster.add(adjustmentListener, l);
    }


    /**
     * Removes a listener from receiving adjustment events.
     * @param   l  listener to receive adjustment events
     * @see java.awt.event.AdjustmentEvent
     * @see java.awt.event.AdjustmentListener
     * @see #removeAdjustmentListener
     */
    public synchronized void removeAdjustmentListener(AdjustmentListener l){
	adjustmentListener = AWTEventMulticaster.remove(adjustmentListener, l);
    }

    private void setSpan(int min, int max, int visible) {
	// adjust the values to be reasonable
	minimum = min;
	maximum = Math.max(max, minimum + 1);
	visibleAmount  = Math.min(visible, maximum - minimum);
	visibleAmount  = Math.max(visibleAmount, 1);
        blockIncrement = Math.max(visibleAmount - unitIncrement, 1);
	setValue(value);
    }

}    
