/*
 * @(#)ArmableMouseController.java	1.7 98/11/25 SMI
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package sun.awt.otk;

import sun.porting.toolkit.ToolkitEventHandler;
import java.awt.AWTEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.Component;

/**
 * A controller that consumes mouse events and interprets them for the
 * benefit of View objects that implement Armable.   The consumeAll
 * flag is not really "state;" it allows for two variants of the
 * controller to both be instantiated as Singletons.
 */
public class ArmableMouseController extends InputController implements StatelessController {
    public static final ArmableMouseController consumingController = 
        new ArmableMouseController(true);

    public static final ArmableMouseController nonConsumingController = 
        new ArmableMouseController(false);

    public final boolean consumeAll;

    /**
     * Creates an instance of an ArmableMouseController.
     *
     * @param consumeAllMouseEvents determines whether the controller
     *                              consumes all mouse events or just 
     *                              the ones that it processes.
     */
    public ArmableMouseController(boolean consumeAllMouseEvents) {
	consumeAll = consumeAllMouseEvents;
    }

    /**
     * Processes mouse events on behalf of the Armable view to which it 
     * has been attached.
     * 
     * @param v The view that this controller is attached to
     * @param id The event type identifier
     * @param when A timestamp (in milliseconds) indicating when 
     *             the event happened.
     * @param modifiers The modifier key (SHIFT, CONTROL, ALT, META) state.
     * @param x The x coordinate associated with the event
     * @param y The y coordinate associated with the event
     * @param arg1 An optional argument whose meaning varies with the event type
     * @param arg2 An optional argument whose meaning varies with the event type
     *
     * @return A boolean indicating whether the event was handled. If the
     *         controller specifies that all mouse events are to be consumed,
     *         then a value of true is returned for all mouse events.
     */
    public boolean handleEvent(View v,
			       int id, long when, int modifiers,
			       int x, int y, int arg1, int arg2) {

	Armable a = (Armable)v;

	switch (id) {
	case ToolkitEventHandler.POINTER_PRESSED:
	    if (arg1 < 0) {
		return consumeAll; // click event.  Ignore it.
	    }

	    a.setArmed(true);
	    return true;

	case ToolkitEventHandler.WINDOW_EXITED:
	    if (a.isArmed()) {
		a.setArmed(false);
	    }
	    return true;

	case ToolkitEventHandler.WINDOW_ENTERED:
	    if (a.isRearmable()) {
		a.setArmed(true);
	    }
	    return true;

	case ToolkitEventHandler.POINTER_RELEASED:
	    a.clear(a.isArmed());
	    return true;

	case ToolkitEventHandler.POINTER_MOVED:
	case ToolkitEventHandler.POINTER_DRAGGED:
	    return consumeAll;

	default:
	    return false;
	}
    }
}




