/*
 * @(#)BasicView.java	1.41 99/09/27 SMI
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package sun.awt.otk;

import java.awt.*;
import java.awt.event.*;
import java.awt.peer.*;
import sun.porting.graphicssystem.Region;
import sun.porting.windowsystem.*;
import sun.porting.toolkit.*;

/**
 * An object which implements a View drawn in a real Window.
 * @author Richard Berlin
 * @version 1.36
 */
public abstract class BasicView implements View {
    private Model model;
    private ObjectToolkit tk;
    private Font font;
    private FontMetrics metrics;
    private ColorPolicy colors;

    /**
     * A minimal rectangle used to create BasicView objects with unspecified bounds.
     */
    protected static final Rectangle NULL_BOUNDS = new Rectangle(0, 0, 1, 1);

    /**
     * The actual window used to represent this View.
     */
    protected sun.porting.windowsystem.Window win;

    /**
     * Create a BasicView object as a view on the given Model, residing in the
     * specified toolkit and using the supplied ColorPolicy to do its painting.
     * The window used to represent this view will be a top-level window.
     *
     * @param model associated with this view.
     * @param toolkit to use.
     * @param colors the ColorPolicy to use.
     */
    public BasicView(Model model, ObjectToolkit toolkit, ColorPolicy colors) {
	this(model, toolkit, null, colors);
    }

    /**
     * Create a BasicView object as a view on the given Model, residing in the
     * specified toolkit and using the supplied ColorPolicy to do its painting.
     * Create the underlying window as a child of the given parent.
     *
     * @param model associated with this view.
     * @param toolkit to use.
     * @param parent to use when creating the underlying window.
     * @param colors the ColorPolicy to use.
     */
    public BasicView(Model model,
		     ObjectToolkit toolkit,
		     ContainerView parent,
		     ColorPolicy colors) {
	this(model, toolkit, parent, colors, NULL_BOUNDS);
    }

    /**
     * Create a BasicView object as a view on the given Model, residing in the
     * specified toolkit and using the supplied ColorPolicy to do its painting.
     * Create the underlying window as a child of the given parent, with initial
     * bounds as indicated.
     *
     * @param model associated with this view.
     * @param toolkit to use.
     * @param parent to use when creating the underlying window.
     * @param colors the ColorPolicy to use.
     * @param bounds to use when creating the underlying window.
     */
    public BasicView(Model model,
		     ObjectToolkit toolkit,
		     ContainerView parent,
		     ColorPolicy colors,
		     Rectangle bounds) {
	this(model, toolkit, parent, colors, bounds, 
	     WindowFactory.TOP_LEVEL_FRAMELESS);
    }


    /**
     * Create a BasicView object as a view on the given Model, residing in the
     * specified toolkit and using the supplied ColorPolicy to do its painting.
     * Create the underlying window as a child of the given parent, with initial
     * bounds as indicated and of the type specified.
     *
     * @param model associated with this view.
     * @param toolkit to use.
     * @param parent to use when creating the underlying window.
     * @param colors the ColorPolicy to use.
     * @param bounds to use when creating the underlying window.
     * @param windowType type of window.
     */
     public BasicView(Model model,
		     ObjectToolkit toolkit,
		     ContainerView parent,
		     ColorPolicy colors,
		     Rectangle bounds,
		     int windowType) {
	this.model  = model;
	this.tk     = toolkit;
	this.colors = colors;
	sun.porting.windowsystem.Window container =
	    (parent == null) ? null : parent.getContainerWindow();

	if (container == null) {
	    win = tk.winSys.makeTopLevelWindow(windowType, 
					       bounds.x, bounds.y, 
					       bounds.width, bounds.height);
	} else {
	    win = tk.winSys.makeWindow(container,
				       bounds.x, bounds.y, 
				       bounds.width, bounds.height);
	}

	win.setUserData(this);
    }

    /**
     * Return the window that should contain any children of this view.
     * (It is often the single window underlying a BasicView, but complex
     * containers made up of several window objects can override this method
     * to ensure that the correct window is being used as the container.
     * @return The Window object to use as the parent for this view's children.
     */
    protected sun.porting.windowsystem.Window getContainerWindow() {
	return win;
    }


    /**
     * The notification method for the Dependent interface.  Indicates that
     * an observable change has occurred on Subject m; arg is an optional 
     * argument which may help indicate what has changed.
     *
     * BasicView handles the following arg types.  Unrecognized types are ignored.
     * <pre>
     *     ViewEvent - a ViewEvent is an event that is posted by a View itself.
     *     When the model receives such an event, it is passed along to the Views.
     *     When received by BasicView, the handleViewEvent() method is called.
     *
     *     Font - The font associated with this object was changed.  The
     *     font and metrics fields of the BasicView will be set accordingly, and
     *     the fontChanged() method will be called.
     *
     *     null - do a full redisplay.  (Note: this is to be avoided!)
     * </pre>
     * @param m The Subject (model) whose state has changed
     * @param arg An argument indicating the substance of the change.  May be null.
     */
    public void updateDependent(Subject m, Object arg) {
	if (arg instanceof ViewEvent) {
	    /* we got back one of our own events  */
	    /* Note that ViewEvent is an AWTEvent */
	    handleViewEvent((Model)m, (ViewEvent)arg);
	    return;
	} else if (arg instanceof Font) {
	    /* Font changed */
	    font = (Font)arg;
	    metrics = tk.getFontMetrics(font);
	    fontChanged((Model)m, font, metrics);
	    return;
	} else if (arg != null) {
	    return;		// ignore unknown args
	}

	/* Full redisplay required. */

	// Dont bother doing the repaint if the component
	// is not visible.
	//
	if (!isVisible()) {
	    return;
	}

	Graphics  g = null;
	Rectangle r = null;
	try {
	    g = getGraphics();
	    r = getBounds();
	    g.clearRect(0, 0, r.width, r.height);
	    paintView((Model) m, g);
	} catch (Throwable t) {
	    /* Important for views to be bomb-proof! */

	    // System.err.println("Caught exception:  " +t);
	    // t.printStackTrace();
	} finally {
	    g.dispose();
	}
    }

    /**
     * The direct event delivery path to the View, always used for window-related
     * events and also used for input events if the View is not associated with an
     * awt.Component.  Event type identifiers are defined by 
     * sun.porting.toolkit.ToolkitEventHandler.
     *
     * All other types of events are passed along to the controllers.
     *
     * @param id The event type identifier
     * @param when A timestamp (in milliseconds) indicating when the event happened
     * @param modifiers The modifier key (SHIFT, CONTROL, ALT, META) state.
     * @param x The x coordinate associated with the event
     * @param y The y coordinate associated with the event
     * @param arg1 An optional argument whose meaning varies with the event type
     * @param arg2 An optional argument whose meaning varies with the event type
     * @see sun.porting.toolkit.ToolkitEventHandler
     */
    public void handleEvent(int id, long when, int modifiers,
			    int x, int y, int arg1, int arg2) {

	if (id == ToolkitEventHandler.WINDOW_DAMAGED) {
	    Graphics g = null;

	    try {
		g = getGraphics();
		g.clipRect(x, y, arg1, arg2);
		Model m = getModel();
		paintView(m, g);
	    } finally {
		if (g != null) {
		    g.dispose();
		}
	    }
	} else {
	    /* see if the controllers want it */
	    handleEvent(this, id, when, modifiers, x, y, arg1, arg2);
	}
    }

    /**
     * Called when a viewEvent was passed in to updateDependent()
     * @param m The model which generated the update
     * @param e The descriptive argument associated with the update
     */
    protected void handleViewEvent(Model m, ViewEvent e) {
	/* nothing to do in the default case */
    }

    /**
     * Called when null was passed in to updateDependent()
     * @param m The model which generated the update
     * @param g The Graphics object that should be used for painting
     */
    protected abstract void paintView(Model m, Graphics g);

    /**
     * Called when a font was passed in to updateDependent()
     * @param m The model which generated the update
     * @param f The new font associated with the view
     * @param fm The new font metrics associated with the view
     */
    protected void fontChanged(Model m, Font f, FontMetrics fm) {
	/* nothing to do in the default case */
    }

    /**
     * The Updateclient interface, used to support the ScreenUpdater.
     * @see sun.awt.ScreenUpdater
     * @see sun.awt.UpdateClient
     */
    public void updateClient(Object arg) {
	AWTEvent e = null;
	Rectangle r;

	synchronized (this) {
	if (arg instanceof Shape) {
	        r = ((Shape)arg).getBounds();
	    e = new ViewEvent(this, ViewEvent.UPDATE, arg);
	} else {
	        r = getBounds();
	    r.x = r.y = 0;
	    e = new ViewEvent(this, ViewEvent.UPDATE, r);
	}
	    clearUpdateRegion(r);
	}
		
	if (e != null) {
	    // System.err.println("posting " +e);
	    tk.postEvent(e);
	}
    }

    // helps with implementation of UpdateClient */
    private volatile Rectangle pending = null;

    /**
     * An implementation method that helps support updateClient.  Return a shape
     * which includes the given rectangle, possibly merged with any other update
     * that is still pending.
     * @param x The x coordinate for the rectangle's left edge
     * @param y The y coordinate for the rectangle's top edge
     * @param width The width of the rectangle
     * @param height The height of the rectangle
     */
    public synchronized Shape getScreenUpdateRegion(int x, int y, 
						    int width, int height) {
	if (pending == null) {
	    return pending = new Rectangle(x, y, width, height);
	} else {
	    int x1 = Math.min(x,          pending.x);
	    int y1 = Math.min(y,          pending.y);
            int x2 = Math.max(x + width,  pending.x + pending.width);
            int y2 = Math.max(y + height, pending.y + pending.height);
	    pending.setBounds(x1, y1, x2 - x1, y2 - y1);
	    return null;
	}
    }

    /**
     * Clear any pending updates that lie in the given rectangle.   (This is an
     * indication that the pending updates for this rectangle have been done.)
     * @param r The rectangle that has been updated.
     */
    public synchronized void clearUpdateRegion(Rectangle r) {
	if (pending == null) {
	    return;
	}

	if (   r.contains(pending.x,  pending.y)
	    && r.contains(pending.x + pending.width,
			  pending.y + pending.height)) {

	    /* if opposing corners are contained in r, we've just  */
	    /* painted an area that subsumes the pending update.   */
	    /* The effect of this assignment it to cause any other */
	    /* entries in the ScreenUpdater queue which reference  */
	    /* the same rectangle to become updates with no area.  */
	    pending.width = pending.height = 0;
	}

	/* since a paint event must have been delivered, start a new rect */
	pending = null;
    }

    /**
     * An accessor method for the toolkit with which this view is associated.
     * @return A toolkit object.
     */
    public final ObjectToolkit getToolkit() {
	return tk;
    }

    /**
     * Query the visible state of this view.
     * @return true if this view is supposed to be visible, false otherwise.
     * @see setVisible
     */
    public final boolean isVisible() {
 	return win.isMapped();
    }

    /**
     * Set the visible state of this view.
     * @param vis The desired visibility; true for visible, false for invisible.
     */
    public synchronized void setVisible(boolean vis) {
	if (vis) {
	    win.map();
	} else {
	    win.unmap();
	}
    }

    /**
     * Set the stacking order (Z coordinate) of this view with respect to its siblings.
     * A zero indicates the frontmost window, with positive numbers indicating how
     * many siblings should be ahead of this window.  Negative numbers indicate to
     * start at the back of the stack, with -1 indicating that this window should be
     * on the bottom, -2 meaning that 1 window should be behind it, etc.
     * @param order The desired position.
     */
    public void setStackOrder(int order) {
	win.setStackingOrder(order);
    }

    /**
     * Get the physical bounds (relative to parent) associated with the view. 
     * @return A Rectangle describing the bounds of the underlying window.
     */
    public Rectangle getBounds() {
	return win.getBounds(false);
    }

    /**
     * Get the physical bounds associated with the view, either in the
     * parent's coordinate system, or in the display's global coordinates.
     * @param global True if the bounds are desired in global coords, false otherwise.
     * @return A Rectangle describing the bounds of the underlying window.
     */
    public Rectangle getBounds(boolean global) {
	return win.getBounds(global);
    }

    /**
     * Set the bounds of the view, relative to the view's parent.
     * @param x The new x coordinate for the view's left edge
     * @param y The new y coordinate for the view's top edge
     * @param width The new width of the view
     * @param height The new height of the view
     */
    public void setBounds(int x, int y, int width, int height) {
	win.setBounds(x, y, width, height);
    }

    public Dimension getScrollbarDimensions() {
	return null;
    }

    /**
     * Get the preferred size for this view, on the assumption that the given
     * number of rows and columns must be displayed.  For a View that has no
     * row/column layout, this will return the same value as getPreferredSize().
     * @param rows The desired number of rows
     * @param cols The desired number of columns
     */
    public Dimension getPreferredSize(int rows, int cols) {
	Rectangle r = getBounds();
	return new Dimension(r.width, r.height);
    }

    /**
     * Get the minimum size for this view, on the assumption that the given
     * number of rows and columns must be displayed.  For a View that has no
     * row/column layout, this will return the same value as getMinimumSize().
     * @param rows The desired number of rows
     * @param cols The desired number of columns
     */
    public Dimension getMinimumSize(int rows, int cols) {
	return getPreferredSize(rows, cols);
    }

    /**
     * Get a Graphics object that has been configured to paint into this View.
     * @return A graphics object that can be used to paint into this View.
     */
    public Graphics getGraphics() {
	Graphics g = win.getGraphics();

	Rectangle r = getBounds();
	g.setClip(0, 0, r.width, r.height);

	Model m = getModel();
	Font f = getFont();
	g.setFont(f == null ? tk.getDefaultFont() : f);

	Color c = getColorPolicy().getForeground(m, false, m.isEnabled(), 
						 false);
	g.setColor(c);

	return g;
    }

    /**
     * An accessor method for the Font associated with this view.
     * @return A Font object.
     */
    public final Font getFont() {
	return font;
    }

    /**
     * An accessor method for the FontMetrics associated with this view.
     * @return A FontMetrics object.
     */
    public final FontMetrics getMetrics() {
	return metrics;
    }

    /**
     * An accessor method for the ColorPolicy associated with this view.
     * @return A ColorPolicy object.
     */
    public final ColorPolicy getColorPolicy() {
	return colors;
    }

    /**
     * An accessor method for the Model associated with this view.
     * @return A Model object.
     */
    public Model getModel() {
	return model;
    }

    private boolean focusTraversable = false;

    /**
     * turns on focusTraversal
     */
    public void acceptFocus(){
	focusTraversable = true;
    }

    /**
     * Query whether this View should accept the keyboard focus during traversal.
     * This query is usually passed along to the controllers.
     */
    public boolean isFocusTraversable() {
	return isVisible() && focusTraversable;
    }
 
    /**
     * Request that the keyboard focus be transferred to the window associated with
     * this View.
     */
    public void requestFocus() {
	if (isVisible()){ 
	    win.acquireInputFocus();
	}
    }

    /** 
     * Query whether this View has focus or not.
     * @return A boolean indicating focus status.
     */
    public boolean hasFocus(){
	return win.hasInputFocus();
    }

    private static final Controller noControllers[] = {};
    private Controller controllers[] = noControllers;
    private int numControllers = 0;

    /**
     * Handle an AWT event by delivering it to each of the controllers in turn
     * until one of them handles it.  Return true if the event was handled by one
     * of the controllers, false otherwise.
     * @param m The model which initially received the event
     * @param e The event itself
     * @return A boolean indicating whether the event was handled.
     */
    protected boolean handleAWTEvent(Model m, AWTEvent e) {
	synchronized(controllers) {
	    if (!m.isEnabled()) {
		return false;
	    }

	    // if a drag controller exists, dispatch to it first
	    if (dragController != null &&
		     dragController.handleEvent(this, e)) {
		return true;
	    }

	    for (int i = 0; i < numControllers; ++i) {
		if (!controllers[i].equals(dragController) &&
			controllers[i].handleEvent(this, e)) {
		    return true;
		}
	    }
	}

	return false;
    }

    /**
     * Handle an event by delivering it to each of the controllers in turn
     * until one of them handles it.  Return true if the event was handled by one
     * of the controllers, false otherwise.
     * @param v The View which initially received the event -- most likely "this."
     * @param id The event type identifier
     * @param when A timestamp (in milliseconds) indicating when the event happened
     * @param modifiers The modifier key (SHIFT, CONTROL, ALT, META) state.
     * @param x The x coordinate associated with the event
     * @param y The y coordinate associated with the event
     * @param arg1 An optional argument whose meaning varies with the event type
     * @param arg2 An optional argument whose meaning varies with the event type
     * @return A boolean indicating whether the event was handled.
     */
    protected boolean handleEvent(View v,
				  int id, long when, int modifiers,
				  int x, int y, 
				  int arg1, int arg2) {

	synchronized(controllers) {
	    if (!v.getModel().isEnabled()) {
		return false;
	    }

	    // if a drag controller exists, dispatch to it first
	    if (dragController != null &&
		dragController.handleEvent(this, id, when, modifiers,
					   x, y, arg1, arg2)) {
		return true;
	    }

	    for (int i = 0; i < numControllers; ++i) {

                if (!controllers[i].equals(dragController) &&
		    controllers[i].handleEvent(this, id, when, modifiers,
					       x, y, arg1, arg2)) {
		    return true;
		}
	    }
	}

	return false;
    }

    /**
     * Add a Controller which listens to events on this View.
     */
    public void addController(Controller c) {
	synchronized(controllers) {
	    Controller n[] = controllers;

	    if (numControllers >= n.length) {
		n = new Controller[n.length + 3];
		System.arraycopy(controllers, 0, n, 0, controllers.length);
	    }
	    n[numControllers++] = c;

	    controllers = n;	// IMPORTANT: must be last statement
	                        // in synchronized block, since it has
	                        // the effect of releasing the lock!
	}
    }

    /**
     * Remove a Controller which was previously added to this view.
     */
    public void removeController(Controller c) {
	synchronized(controllers) {
	    for (int i = 0; i < numControllers; ++i) {
		if (c == controllers[i]) {
		    System.arraycopy(controllers, i+1,
				     controllers, i,
				     numControllers - (i+1));

		    controllers[--numControllers] = null;

		    break;
		}
	    }

	    if (numControllers == 0) {
		controllers = noControllers; // IMPORTANT: must be last 
		                             // statement in the sync. block
	    }
	}
    }

    /**
     * Remove all Controllers which were previously added to this view.
     */
    public void removeControllers() {
	synchronized(controllers) {
	    dragController = null;
	    numControllers = 0;
	    controllers = noControllers; // IMPORTANT: must be last statement
				         // in the synchronized block
	}
    }

    private Controller dragController = null;

    /**
     * Designates an already added controller as the drag controller. This
     * controller is always the first one to receive events regardless of
     * the order in which the controllers were added. Note that setting
     * the drag controller DOES NOT add it to the controllers list and 
     * susequent sets will overwrite the previous value.
     * @param c Controller to set as the drag controller. Use null to reset.
     */
    public void setDragController(Controller c) {
        synchronized(controllers) {
	    dragController = c;
	}
    }

    /**
     * @return Returns the current drag controller.
     */
    public Controller getDragController() {
        synchronized(controllers) {
	    return dragController;
	}
    }
}
