/*
 * @(#)ChoiceModel.java	1.8 98/11/25 SMI
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package sun.awt.otk;

import java.awt.AWTEvent;
import java.awt.Choice;
import java.awt.Font;
import java.awt.event.ItemEvent;
import java.awt.peer.ChoicePeer;

public class ChoiceModel extends ComponentModel implements ChoicePeer, MenuItemOwner {
    private int selected_index;
    private OrderedList items;

    /**
     * Constructs a new ChoiceModel instance. Creates an ordered list to hold a copy of the 
     * items in the Choice component. Ensures the selected item will appear selected in the
     * view when the view is made visible.
     * @param target the AWT Choice component
     * @param tk the toolkit
     */
    public ChoiceModel(Choice target, ObjectToolkit tk) {
	this(target, tk, new OrderedListImpl(target.getItemCount()));
    }

    /**
     * Constructs a new ChoiceModel instance. Uses the supplied ordered list to hold a copy
     * of the items in the Choice component.  Ensures the selected item will appear selected
     * in all dependent views when they are made visible.
     * @param target the AWT Choice component
     * @param tk the toolkit
     * @param itemList the OrderedList to use
     */
    public ChoiceModel(Choice target, ObjectToolkit tk, OrderedList itemList) {
	super(target, tk);

	items = itemList;

	int count = target.getItemCount();
	for (int i = 0; i < count; ++i) {
	    add(target.getItem(i), i);
	}

	select(target.getSelectedIndex());
    }

    /**
     * Adds an item to the choice.
     * Notifies all dependent views to modify their appearance as necessary.
     * @param item the non-null item to be added
     * @param index the position in the choice list at which to place the item
     */
    public void add(String item, int index) {
	items.insertElementAt(new ChoiceMenuItem(item, this), index);
	notifyDependents(items);
    }

    /**
     * Removes an item from the choice. 
     * Notifies all dependent views to modify their appearance as necessary.
     * @param index the item's position in the choice
     */
    public void remove(int index) {
	ChoiceMenuItem item = (ChoiceMenuItem) items.elementAt(index);
	item.setOwner(null);

	items.removeElementAt(index);
	notifyDependents(items);
    }

    /**
     * Selects an item in the choice.
     * Notifies all dependent views to modify their appearance as necessary.
     * @param index the non-negative zero-based index of the item
     */
    public void select(int index) {
	selected_index = index;
	notifyDependents(new Integer(selected_index));
    }

    /**
     *  Notifies the ChoiceModel that an item has changed.
     * @param item the item that changed
     */
    public void itemChanged(MenuItemOperations item) {
	/* this only happens when we've initiated it */
    }

    /* for benefit of the view */

    /**
     * Sets the default view.
     * Notifies all dependent views to modify their appearance as necessary.
     * @param v the new value of the default view
     */
    public synchronized void setDefaultView(View v) {
	super.setDefaultView(v);

	notifyDependents(items);
	notifyDependents(new Integer(selected_index));
    }
    
    /**
     * @deprecated Replaced by add().
     * @see #add
     */
    public final void addItem(String item, int index) {
	add(item, index);
    }

    /**
     * Posts a MenuShowEvent to the EventQueue.
     * @param evt the MenuShowEvent
     * @see MenuShowEvent
     */
    public void postMenuShowEvent(AWTEvent evt) {
	((ObjectToolkit)getToolkit()).postEvent(evt);
    }
}

class ChoiceMenuItem extends LabelMenuItem {
    public ChoiceMenuItem(String label, ChoiceModel owner) {
	super(label, owner);
    }

    public void actionOccurred(int modifiers) {
	ChoiceModel cm   = (ChoiceModel)owner;
	ObjectToolkit tk = (ObjectToolkit)cm.getToolkit();
	Choice choice    = (Choice) cm.getTarget();
	String s         = choice.getSelectedItem();
	ItemEvent ev     = null;

	ev = new ItemEvent(choice, ItemEvent.ITEM_STATE_CHANGED,
			   getLabel(), ItemEvent.SELECTED);
	tk.postEvent(ev);

	/*
	 * Really gross to have to do it this way, but java.awt.Choice
	 * seems to depend on it!
	 */
	choice.select(getLabel());
    }
}
