/*
 * @(#)ColorPalette.java	1.7 98/11/30
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package sun.awt.otk;

import java.util.Hashtable;
import java.awt.Color;
import java.text.*;
import java.io.*;

import sun.io.FileIOFactory;

/**
 * A Hashtable which contains symbolic color names and the corresponding
 * color definitions.  It is used to externally represent sets of colors
 * which were chosen to work well together in a GUI.  The meanings of the
 * names are up to the GUI designer, so the palette definition file from
 * which the data is read should be closely tied to the look and feel.
 * <p>
 * This implementation of multiple color palettes uses FileReader, which is
 * an optional part of PJava.  The implementation was done so that multiple
 * palettes could be easily tested.  We do not expect most licensees to 
 * include this file in their shipping products; instead, they should hardwire
 * the desired palette or create their own mechanism for allowing users to
 * choose among palettes.
 * @see java.awt.Color
 */
public class ColorPalette extends Hashtable {

    /**
     * Create a new ColorPalette instance with the default palette.
     */
    public ColorPalette() {
	this(System.getProperty("sun.awt.palette"));
    }

    /**
     * Create a new ColorPalette instance with the named palette, using
     * an alternate database.
     * @param paletteName The name of the palette to use for color definitions.
     */
    public ColorPalette(String paletteName) {
	this(paletteName, System.getProperty("sun.awt.palette.definitions"));
    }


    /**
     * Create a new ColorPalette instance with the named palette.  Use an
     * alternate location for the palette database.
     * @param paletteName The name of the palette to use for color definitions.
     * @param databaseName The path to the palette database
     */
    public ColorPalette(String paletteName, String databaseName) {
	super(7);

	// System.err.println("Creating ColorPalette from " +databaseName);

	if (! parsePalette(paletteName, databaseName)) {
	    parsePalette(null, databaseName);	// match first palette
	}
    }

    private boolean parsePalette(String name, String file) {
	try {
	    InputStream is = FileIOFactory.getInputStream(file);
	    LineNumberReader r = new LineNumberReader(new InputStreamReader(is));
	    NumberFormat nf = NumberFormat.getInstance();
	    String s;

	    String target = "Palette: ";
	    if (name != null) target += name;

	    while ((s = r.readLine()) != null) {
		if (s.startsWith(target)) {
		    break;
		}
	    }

	    /* OK, we've either got the beginning of our palette, or null */
	    if (s == null) {
		return false;
	    }

	    while (((s = r.readLine()) != null) && !s.startsWith("END")) {
		ParsePosition pos = new ParsePosition(0);
		Number red = nf.parse(s, skipBlanks(s, pos));
		Number grn = nf.parse(s, skipBlanks(s, pos));
		Number blu = nf.parse(s, skipBlanks(s, pos));
		
		put(s.substring(pos.getIndex()).trim(),
		    new Color(red.intValue(), grn.intValue(), blu.intValue()));
	    }	    
	} catch (Throwable t) {
	    // t.printStackTrace();
	    return false;
	}

	return true;
    }

    private ParsePosition skipBlanks(String s, ParsePosition p) {
	try {
	    while (Character.isSpaceChar(s.charAt(p.getIndex()))) {
		p.setIndex(p.getIndex() + 1);
	    }
	} catch (ArrayIndexOutOfBoundsException oops) {
	}

	return p;
    }
}
