/*
 * @(#)ComponentModel.java	1.32 99/05/07 SMI
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package sun.awt.otk;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.awt.peer.ComponentPeer;
import java.util.Locale;
import com.sun.awt.*;

public class ComponentModel extends ObservableModel implements ComponentPeer {
    /**
     * The toolkit.
     */
    protected ObjectToolkit tk;

    private Component target;
    private boolean enabled;

    /**
     * Constructs a new ComponentModel instance.
     * @param target the AWT Component
     * @param tk the toolkit
     */
    public ComponentModel(Component target, ObjectToolkit tk) {
	this.tk      = tk;
	this.target  = target;
	this.enabled = target.isEnabled();
    }

    /* 
     * Return the component's z-order position relative to
     * other peer'd siblings (don't count lightweight siblings
     * or siblings who don't yet have valid peers).
     */
    private int getZOrderPosition() {
        Container p = target.getParent();
        if (p != null) {
            Component children[] = p.getComponents();

            int i;
            int index = 0;
            for (i = 0; i < children.length; i++) {
                if (children[i] == target) {
                    return index;
                } else {
		    Object cpeer = tk.targetToPeer(children[i]);
                    if ( (cpeer != null) && 
                        !(cpeer instanceof java.awt.peer.LightweightPeer)) {
                        index++;
                    }
                }
            }
        }
        return 0;
    }

    private boolean shouldBeVisible() {
	java.awt.Component target = this.target;
	java.awt.Container parent = tk.getNativeContainer(target);

	if (parent == null) {
	    return target.isVisible();
	}

	/* check all lightweight containers between us and real parent. */
	while (target != parent){
	    if (!target.isVisible()) {
		return false;
	    } else {
		target = target.getParent();
	    }
	}
	return true;
    }

    /**
     * Post an ActionEvent to the EventQueue.
     * @param eventId the type of event
     * @param command the command string for this action event
     * @see ActionEvent
     */
    public void postActionEvent(int eventId, String command) {

	ActionEvent ev = new ActionEvent(target, 
					 eventId,
					 command);
	ObjectToolkit.postEvent(ev);
    }

    
    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
    /*   These methods are for the benefit of Dependent Views  */
    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

    /**
     * Set the default view.
     * @param v the new value of the default view
     */
    public synchronized void setDefaultView(View v) {
	super.setDefaultView(v);

	if (v == null) {
	    return;
	}

	Color eraseColor = v.getColorPolicy().getEraseColor(this, false,
							    isEnabled(),
							    false);
	if (eraseColor != null) {
	    setBackground(eraseColor);
	}

	Font f = target.getFont();
	if (f != null) {
	    setFont(f);
	}

/*
	Cursor c = target.getCursor();
	if (c != null) {
	    setCursor(c);
	}
 */
    }

    /**
     * Notifies this ComponentModel that its view initialization is complete, so it 
     * is now "safe"  to make its view(s) visible. This is to prevent race conditions
     * whereby a view might become visible before it is correctly hooked up to
     * its model.
     */
    public void initComplete() {
	setStackOrder(getZOrderPosition());
	setVisible(shouldBeVisible());
    }

    /**
     * Get the corresponding AWT Component.
     * @return the AWT Component
     */
    protected final Component getTarget() {
	return target;
    }

    /**
     * Determines whether this component is visible. 
     * @return true if this component's default view is visible; false otherwise
     * @see #setVisible
     */
    public final boolean isVisible() {
	View dv = getDefaultView();
	return dv == null ? false : dv.isVisible();
    }

    /**
     * Determines whether this component is enabled. An enabled component 
     * can respond to user input and generate events. Components are 
     * enabled initially by default. A component may be enabled or disabled by 
     * calling its setEnabled method.
     * @return true if the component is enabled; false otherwise
     * @see #setEnabled
     */
    public final boolean isEnabled() {
	return enabled;
    }

    /**
     * Gets the font of this component.
     * @return This component's font. If a font has not been set 
     * for this component, the font of its parent is returned.
     * @see #setFont
     */
    public final Font getFont() {
	Component target = getTarget();
	return target.getFont();
    }

    // REMIND!!! We should come up with better names.
    // getReal.. returns a color set by a programmer using (target.set..)
    private native Color getRealForeground0(Component target);
    /**
     * Get this component's foreground color as set by the application.
     * @return the foreground color; null if none has been set
     * @see ColorPolicy
     */
    public final Color getRealForeground() {
	return getRealForeground0(target);
    }


    /**
     * Get the background color of this component's AWT Container.
     * @return the AWT Container's background color; null if not set or no parent container
     * @see ColorPolicy
     */
    public final Color getParentBackground() {
	Container parent = (Container)target.getParent();
	if (parent != null){
	    Color bg = parent.getBackground();
	    if (bg != null){
		return bg;
	    }
	}
	return null;
    }

    /**
     * Get the foreground color of this component's AWT Container.
     * @return the AWT Container's foreground color; null if not set or no parent container
     * @see ColorPolicy
     */
    public final Color getParentForeground() {
	if (target != null){
	    Container parent = (Container)target.getParent();
	    if (parent != null){
		Color fg = parent.getBackground();
		if (fg != null){
		    return fg;
		}
	    }
	}
	return null;
    }

    /**
     * Gets the foreground color of this component's AWT Component.
     * @return the AWT Component's foreground color. If it does 
     * not have a foreground color, the foreground color of its parent 
     * is returned.
     * @see ColorPolicy
     * @see #setForeground
     */
    public final Color getForeground() {
	return target.getForeground();
    }

    /**
     * Gets the background color of this component's AWT Component.
     * @return the AWT Component's background color. If it does 
     * not have a background color, the background color of its parent 
     * is returned.
     * @see ColorPolicy
     * @see setBackground
     */
    public final Color getBackground() {
	return target.getBackground();
    }

    private native Color getRealBackground0(Component target);
    /**
     * Get this component's background color as set by the application.
     * @return the background color; null if none has been set
     * @see ColorPolicy
     */
    public final Color getRealBackground() {
	return getRealBackground0(target);
    }

    private native Locale getRealLocale0(Component target);
    /**
     * Get this component's locale as set by the application.
     * @return the locale; null if none has been set
     * @see ColorPolicy
     */
    public final Locale getRealLocale() {
	return getRealLocale0(target);
    }

    /**
     * Gets the locale of this component's AWT Component.
     * @return the AWT Component's locale. If it does 
     * not have a locale, the locale of its parent is returned.
     */
    public final Locale getDefaultLocale() {
	return target.getLocale();
    }

    private void setStackOrder(int order) {
	View dv = getDefaultView();

	if (dv == null) {
	    throw new RuntimeException("Component has no visualization");
	}

	dv.setStackOrder(order);
    }

    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
    /* The rest of this implements the ComponentPeer interface */
    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

    /**
     * Shows or hides this component's default view, depending on the value of parameter b.
     * @param b if true, shows this component; otherwise, hides this component.
     * @see #isVisible
     */
     public void setVisible(boolean b) {
	View dv = getDefaultView();

	if (b && (dv == null)) {
	    throw new RuntimeException("Component has no visualization");
	}

	dv.setVisible(b);
	// We don't need to do a notify because the visiblity 
	// change should generate an event anyway??
	// notifyDependents();
    }

    /**
     * Enables or disables this component, depending on the value of the 
     * parameter b. An enabled component can respond to user 
     * input and generate events. Components are enabled initially by default.
     * <p>
     * Unlike setVisible, which benefits from a real window hierarchy,
     * the "nesting" quality of enabled components has to be enforced
     * by the toolkit; hence this method ensures that the enabled state
     * will not be set to true if a parent component is disabled. If this 
     * method determines that the requested state change is valid,
     * it calls doEnable to do the work.
     *
     * @param b if true, this component is enabled; otherwise this component is disabled.
     * @see #doEnable
     * @see #isEnabled
     */
    public final void setEnabled(boolean b) {
	Component p = target.getParent();
	while (b && (p != null)) {
	    b = p.isEnabled();
	    p = p.getParent();
	}

	if (enabled != b) {
	    doEnable(b);
	}
    }

    /** 
     * Sets the new enabled state, and notifies all dependent views to modify their appearance as necessary.
     * @param b if true, this component is enabled; otherwise this component is disabled.
     * @see #setEnabled
     */
    protected void doEnable(boolean b) {
	enabled = b;
	notifyDependents(); /* full repaint required */
    }

    /**
     * Paints this component's default view.
     * The explicit paint() call is part of the peer interface.
     * It is only used if someone makes an explicit call to
     * Container.paintComponents().
     * @param g The graphics context to use for painting
     */
    public void paint(Graphics g) {
	View dv = getDefaultView();
	if (dv != null) {
	    dv.updateDependent(this, g);
	}
    }

    /**
     * Prints this component.
     * The explicit print() call is part of the peer interface.
     * It is only used if someone makes an explicit call to
     * Container.printComponents().
     * @param g the graphics context to use for printing.
     */
    public void	print(Graphics g) {
	View dv = getDefaultView();
	if (dv != null) {
	    dv.updateDependent(this, g);
	}
    }

    /** 
     * Makes a request to repaint this component's default view.
     * @param tm maximum delay in milliseconds before the ScreenUpdater is called
     * @param x the x-coordinate of the rectangular area to repaint
     * @param y the y-coordinate of the rectangular area to repaint
     * @param width the width of the rectangular area to repaint
     * @param height the height of the rectangular area to repaint
     * @see sun.awt.ScreenUpdater
     */
    public void	repaint(long tm, int x, int y, int width, int height) {
	View dv = getDefaultView();
	
	if (dv == null) {
	    return;
	}

	/* This allows a view to coalesce updates, but doesn't require it. */
	Shape r = dv.getScreenUpdateRegion(x, y, width, height);
	if (r != null) {
	    sun.awt.ScreenUpdater.updater.notify(dv, tm, r);
	}
    }

    /**
     * Moves and resizes this component.
     * Notifies all dependent views to modify their appearance as necessary.
     * The new location of the top-left corner is specified by x and y, and the 
     * new size is specified by width and height.
     * @param x the new x-coordinate of the component in pixels
     * @param y the new y-coordinate of the component in pixels
     * @param width the new width of the component in pixels
     * @param height the new height of the component in pixels
     */
    public void	setBounds(int x, int y, int width, int height) {
	/* 
	 * This is meant to set the bounds on the default view.
	 * But other views may need to know that the bounds have
	 * been changed, so we send a notification with the new bounds.
	 */
	View dv = getDefaultView();
	if (dv == null) {
	    return;
	}

	dv.setBounds(x, y, width, height);
	notifyDependents(dv.getBounds(false));
    }

    /**
     * Posts a VIRTUAL_COMPONENT_SHOW VirtualComponentEvent to the EventQueue.
     * @param msg the message object for this VirtualComponentEvent
     * @see VirtualComponentEvent
     * @see VirtualComponentSelInfo
     */
    public void sendShowEvent(VirtualComponentMessage msg) {
	AWTEvent e = new VirtualComponentEvent(target,
		       VirtualComponentEvent.VIRTUAL_COMPONENT_SHOW, msg);
	ObjectToolkit.postEvent(e);
    }

    /**
     * Called from AWT when this component receives an event.
     * If e is a PaintEvent, sets the clip area to the damage rect and notifies the AWT Component
     * of the event.
     * Notifies all dependent views to modify their appearance as necessary.
     * @param e the non-null AWTEvent
     * @see java.awt.event.PaintEvent
     */
    public void handleEvent(AWTEvent e) {
	/* special case...event has gone to component, come back to us */
        //if (e instanceof KeyEvent) {
	//System.err.println("----------------------");
	  //System.err.println("ComponentModel.handleEvent()");
	  //System.err.println("\tthis = " + this);
	  //System.err.println("\tevent = " + e);
	//}
	Component target = this.target;
	if ((e instanceof PaintEvent) && (target != null)) {
	    Graphics g = null;
	    try {
		Rectangle r = ((PaintEvent)e).getUpdateRect();
		if ((r == null) || r.isEmpty()) {
		    return;
		}

		synchronized (target.getTreeLock()){
		    g = target.getGraphics();
		    if (g == null) {
			return;
		    }
		    g.clipRect(r.x, r.y, r.width, r.height);
		}

		View dv = getDefaultView();
		if (dv != null) {
		    dv.clearUpdateRegion(r);
		}

		if (e.getID() == PaintEvent.PAINT) {
		    target.paint(g);
		} else {
		    target.update(g);
		}
	    } finally {
		if (g != null) {
		    g.dispose();
		}
	    }
	    return;
	}

	/* 
	 * One might think that only the default view should handle
	 * events from AWT.  But it is worth letting all views know
	 * about these events, especially since there are cases where
	 * an alternate view might superimpose itself on the default
	 * view, and need to process events specially.
	 */

	notifyDependents(e);
    }

    /**
     * Gets the location of this component's default view.
     * @return an instance of Point representing 
     * the top-left corner of the default view's bounds in the 
     * coordinate space of the screen.
     */
    public Point getLocationOnScreen() {
	View dv = getDefaultView();

	if (dv == null) {
	    return null;
	}

	Rectangle r = dv.getBounds(true);
	return new Point(r.x, r.y);
    }

    /**
     * Calculates the preferred size of this component's default view.
     * @return a Dimension object indicating the default view's preferred size.
     * @see #getMinimumSize
     */
    public Dimension getPreferredSize() {
	View dv = getDefaultView();
	return dv == null ? null : dv.getPreferredSize(-1, -1);
    }

    /**
     * Gets the mininimum size of this component's default view.
     * @return a Dimension object indicating the default view's minimum size.
     * @see #getPreferredSize
     */
    public Dimension getMinimumSize() {
	View dv = getDefaultView();
	return dv == null ? null : dv.getMinimumSize(-1, -1);
    }

    /**
     * Gets the instance of ColorModel used to display 
     * the component on the output device.
     * @return the color model used by this component
     * @see java.awt.image.ColorModel
     * @see java.awt.Toolkit#getColorModel()
     */
    public ColorModel getColorModel() {
	return getToolkit().getColorModel();
    }

    /**
     * Gets the toolkit of this component.
     * @return the toolkit of this component
     * @see java.awt.Component#getToolkit
     * @see java.awt.Toolkit
     */
    public java.awt.Toolkit getToolkit() {
	return tk;
    }

    /**
     * Gets the graphics context for this component's default view. 
     * @return a graphics context for the default view, or null if there is not
     * yet a default view
     */
    public Graphics getGraphics() {
	View dv = getDefaultView();
	return dv == null ? null : dv.getGraphics();
    }

    /**
     * Gets the font metrics for the specified font.
     * @param font the font for which font metrics is to be obtained
     * @return The font metrics for font.
     * @see       java.awt.Component#getFont
     * @see       java.awt.Toolkit#getFontMetrics(java.awt.Font)
     */
    public FontMetrics getFontMetrics(Font font) {
	return getToolkit().getFontMetrics(font);
    }

    /**
     * Destroys this component. This method releases any resources associated
     * with this component.
     * <p>
     * This method should delete whatever is necessary, and should 
     * be careful to null out all references to other objects.
     */
    public synchronized void dispose() {
	tk.disposedPeer((Object)target, (Object)this);
	target = null;
	tk = null;

	super.dispose();
    }
    
    private Color lastFg = null;
    /**
     * Sets the foreground color of this component.
     * Notifies all dependent views to modify their appearance as necessary.
     * @param c the new foreground color, which can be null
     * @see #getForeground
     */
    public void	setForeground(Color c) {
	if (lastFg != c) {
	    lastFg = c;
	    notifyDependents();	/* full repaint required */
	}
    }

    /**
     * Sets the background color of this component.
     * Notifies all dependent views to modify their appearance as necessary.
     * @param c the new background color, which can be null
     * @see #getBackground
     */
    public void	setBackground(Color c) {
	notifyDependents(c);	// Set window b/g, then do full repaint.
    }

    /**
     * Sets the component's font.
     * @param f the font, which can be null
     * @see #getFont
     */
    public void setFont(Font f) {
	notifyDependents(f);	// font object allows specific detection
    }

    /**
     * Sets the cursor image to a predefined cursor.
     * Notifies all dependent views to modify their appearance as necessary.
     * @param cursor one of the constants defined by the Cursor class
     * @see java.awt.Cursor
     */
    public void	setCursor(Cursor cursor) {
	notifyDependents(cursor); // cursor object allows specific detection
    }

    /** 
     * Requests that this component's default view get the input focus. 
     * @see java.awt.event.FocusEvent
     * @see #isFocusTraversable
     */
    public void	requestFocus() {
	View dv = getDefaultView();

	if (dv != null) {
	    dv.requestFocus();
	}
    }

    /**
     * Determines whether this component's default view's focus can be traversed 
     * using the focus traversal mechanism.
     * @return true if the default view is focus-traverable, false otherwise.
     * @see sun.awt.FocusManagement
     * @see sun.awt.TabFocusManager
     */
    public boolean isFocusTraversable() {
	View dv = getDefaultView();
	
	return (dv == null) ? false : dv.isFocusTraversable();
    }

    /**
     * Returns true if all the drawing done during the update and paint methods
     * is automatically double buffered. 
     * The default value for the double buffering setting is platform-specific.
     * @return true if the default view is double buffered, false otherwise
     */
    public boolean isDoubleBuffered() {
	return false;
    }

    /**
     * Creates an image from the specified image producer.
     * @param producer  the non-null image producer
     * @return the image produced.    
     */
    public Image createImage(ImageProducer producer) {
	return tk.createImage(producer);
    }

    /**
     * Creates an off-screen image.
     * @param width the specified width
     * @param height the specified height
     * @return a new off-screen image
     */
    public Image createImage(int width, int height) {
	return tk.createImage(target, width, height);
    }

    /**
     * Triggers the loading of image data for an image.
     * @param img the non-null image to load
     * @param w -1 or the scaled width of the returned image
     * @param h -1 or the scaled height of the returned image
     * @param o the non-null image observer
     * @return true if all the image data for img is available; false otherwise
     */
    public boolean prepareImage(Image img, int w, int h, ImageObserver o) {
	return tk.prepareImage(img, w, h, o);
    }

    /**
     * Retrieves the construction status of an image.
     * @param img the non-null image to check
     * @param w if >= 0, specifies the width of the scaled version of the image to check
     * @param h if >= 0, specifies the height of the scaled version of the image to check
     * @param o if non-null, specifies the image observer to be notified whenever the status
     * changes.
     * @return the combination of status bits as defined by the ImageObserver interface
     */
    public int checkImage(Image img, int w, int h, ImageObserver o) {
	return tk.checkImage(img, w, h, o);
    }

    /**
     * @deprecated Replaced by getPreferredSize().
     * @see #getPreferredSize
     */
    public final Dimension preferredSize() {
	return getPreferredSize();
    }

    /**
     * @deprecated Replaced by getMinimumSize().
     * @see #getMinimumSize
     */
    public final Dimension minimumSize() {
	return getMinimumSize();
    }

    /**
     * @deprecated  Replaced by setVisible(boolean).
     * @see #setVisible
     */
    public final void show() {
	setVisible(true);
    }

    /**
     * @deprecated  Replaced by setVisible(boolean).
     * @see #setVisible
     */
    public final void hide() {
	setVisible(false);
    }

    /**
     * @deprecated  Replaced by setEnabled(boolean).
     * @see #setEnabled
     */
    public final void enable() {
	setEnabled(true);
    }

    /**
     * @deprecated  Replaced by setEnabled(boolean).
     * @see #setEnabled
     */
    public final void disable() {
	setEnabled(false);
    }

    /**
     * @deprecated  Replaced by setBounds(int, int, int, int).
     * @see #setBounds
     */
    public final void reshape(int x, int y, int width, int height) {
	setBounds(x, y, width, height);
    }


    private static native void initJNI();
    static {
	initJNI();
    }
}
