/*
 * @(#)ComponentView.java	1.31 99/09/29 SMI
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package sun.awt.otk;

import java.awt.*;
import java.awt.event.*;
import java.awt.peer.*;
import sun.porting.graphicssystem.Region;
import sun.porting.windowsystem.*;
import sun.porting.toolkit.*;

/**
 * An extension of BasicView which handles additional update types
 * and events that are common to representations of ComponentPeers.
 */
public abstract class ComponentView extends BasicView {
    /**
     * Create a ComponentView object as a view on the given Model, residing in the
     * specified toolkit and using the supplied ColorPolicy to do its painting.
     * The window used to represent this view will be a top-level window.
     *
     * @param model associated with this view.
     * @param toolkit to use.
     * @param colors the ColorPolicy to use.
     */
    public ComponentView(Model model, ObjectToolkit toolkit, 
			 ColorPolicy colors) {
	this(model, toolkit, null, colors);
    }

    /**
     * Create a ComponentView object as a view on the given Model, residing in the
     * specified toolkit and using the supplied ColorPolicy to do its painting.
     * Create the underlying window as a child of the given parent.
     *
     * @param model associated with this view.
     * @param toolkit to use.
     * @param parent to use when creating the underlying window.
     * @param colors the ColorPolicy to use.
     */
    public ComponentView(Model model,
			 ObjectToolkit toolkit,
			 ContainerView parent,
			 ColorPolicy colors) {
	this(model, toolkit, parent, colors, NULL_BOUNDS);
    }

    /**
     * Create a ComponentView object as a view on the given Model, residing in the
     * specified toolkit and using the supplied ColorPolicy to do its painting.
     * Create the underlying window as a child of the given parent, with initial
     * bounds as indicated.
     *
     * @param model associated with this view.
     * @param toolkit to use.
     * @param parent to use when creating the underlying window.
     * @param colors the ColorPolicy to use.
     * @param bounds to use when creating the underlying window.
     */
    public ComponentView(Model model,
			 ObjectToolkit toolkit,
			 ContainerView parent,
			 ColorPolicy colors,
			 Rectangle bounds) {
	super(model, toolkit, parent, colors, bounds);
    }


    /**
     * Create a ComponentView object as a view on the given Model, residing in the
     * specified toolkit and using the supplied ColorPolicy to do its painting.
     * Create the underlying window as a child of the given parent, with initial
     * bounds as indicated, and using the window type specified.  (This method
     * must be used to create Frame and Dialog windows.)
     *
     * @param model associated with this view.
     * @param toolkit to use.
     * @param parent to use when creating the underlying window.
     * @param colors the ColorPolicy to use.
     * @param bounds to use when creating the underlying window.
     */
    public ComponentView(Model model,
			 ObjectToolkit toolkit,
			 ContainerView parent,
			 ColorPolicy colors,
			 Rectangle bounds,
			 int windowType) {
	super(model, toolkit, parent, colors, bounds, windowType);
    }

    /**
     * The notification method for the Dependent interface.  Indicates that
     * an observable change has occurred on Subject m; arg is an optional 
     * argument which may help indicate what has changed.
     *
     * ComponentView handles the following arg types.  Unrecognized types are
     * passed to super.
     * <pre>
     *     ViewEvent is passed to super.  This must be done explicitly because
     *     a ViewEvent is a subclass of AWTEvent.
     *
     *     AWTEvent - handleAWTEvent is called.  In addition, if the event is
     *     a mouse press and the component can accept focus, requestFocus is called.
     *
     *     Rectangle - indicates that bounds were changed (via setBounds).  A
     *     call is made to boundsChanged() in case the View needs to be advised.
     *
     *     Cursor - The cursor image associated with the component was changed.
     *     The new cursor is associated with the underlying window and a call
     *     is made to cursorChanged().
     *
     *     Insets - The insets were changed.  A call is made to insetsChanged()
     *
     *     Color - The background color was set.  If it differs from the current
     *     background, the window background color is reset and a full repaint is
     *     done by calling super.updateDependent with a null argument.
     * </pre>
     * @param m The Subject (model) whose state has changed
     * @param arg An argument indicating the substance of the change.  May be null
     */
    public void updateDependent(Subject m, Object arg) {
	/* ComponentView handles all of the update types that can */
	/* be generated by a ComponentModel or ContainerModel     */

	if ((arg instanceof AWTEvent) && !(arg instanceof ViewEvent)) {
	    AWTEvent e = (AWTEvent) arg;

	    /* handle an event from AWT */
	    handleAWTEvent((Model)m, e);

	    /* Mouse Press always assigns input focus to components */
	    /* that want it.                                        */
	    if (isFocusTraversable() && e.getID()==MouseEvent.MOUSE_PRESSED){
		requestFocus();
	    }

	    return;
	} else if (arg instanceof Rectangle) {
	    /* Bounds changed (advisory--actual change was via setBounds) */
	    boundsChanged((Model)m, (Rectangle)arg);
	    return;
	} else if (arg instanceof Cursor) {
	    /* Cursor changed */
	    win.setCursorImage(getToolkit().getCursorImage((Cursor)arg));
	    cursorChanged((Model)m, (Cursor)arg);
	    return;
	} else if (arg instanceof Insets) {
	    /* Insets changed */
	    insetsChanged((Model)m, (Insets)arg);
	    return;
	} else if (arg instanceof Color) {
	    /* background changed */

	    // background color change can result in erase color change 
	    // (as a result of container background color change) and
	    // in background color change (through Component.setBackground);
	    // in both cases component has to be repainted
	    ComponentModel cm = (ComponentModel)m;
	    Color newErase    = getColorPolicy().getEraseColor(cm, false, 
							       cm.isEnabled(),
							       false);
	    boolean eraseChanged = !win.getBackgroundColor().equals(newErase);
	    if (eraseChanged) {
		win.setBackgroundColor((Color)arg);
	    }
	    super.updateDependent(m, null);
	    return;
	}

	super.updateDependent(m, arg);
    }

    /**
     * Default handling for a repaint on a Component.   Posts a PaintEvent to the target,
     * using the clip rectangle associated with g.
     * @param m The model which caused the repaint
     * @param g The graphics object originally allocated for the paint.  (In this case
     * it is unused.
     */
    protected void paintView(Model m, Graphics g) {
	try {
	    ComponentModel cm = (ComponentModel) m;
	    AWTEvent ev = new PaintEvent(cm.getTarget(), PaintEvent.PAINT,
					 g.getClipBounds());
	    getToolkit().postEvent(ev);
	} catch (IllegalArgumentException targetIsNull) {
	    // Target has gone.  Most likely unmapped, so we can safely ignore 
	    // this event.
	}
    }

    /**
     * The handleEvent method is overridden to translate most window-related
     * events to corresponding AWT events and post them to the target.  The
     * translation is as follows:
     * <pre>
     *     ToolkitEventHandler.WINDOW_ENTERED    ==>  MouseEvent.MOUSE_ENTERED
     *     ToolkitEventHandler.WINDOW_EXITED     ==>  MouseEvent.MOUSE_EXITED
     *     ToolkitEventHandler.WINDOW_RESHAPED   ==>  ComponentEvent.COMPONENT_MOVED
     *					       and/or ComponentEvent.COMPONENT_RESIZED
     *     ToolkitEventHandler.WINDOW_GOT_FOCUS  ==>  FocusEvent.FOCUS_GAINED
     *     ToolkitEventHandler.WINDOW_LOST_FOCUS ==>  FocusEvent.FOCUS_LOST
     *     ToolkitEventHandler.WINDOW_DAMAGED    ==>  PaintEvent.PAINT
     * </pre>
     *
     * @param id The event type identifier
     * @param when A timestamp (in milliseconds) indicating when the event happened
     * @param modifiers The modifier key (SHIFT, CONTROL, ALT, META) state.
     * @param x The x coordinate associated with the event
     * @param y The y coordinate associated with the event
     * @param arg1 An optional argument whose meaning varies with the event type
     * @param arg2 An optional argument whose meaning varies with the event type    
     */
    public void handleEvent(int id, long when, int modifiers,
			    int x, int y, int arg1, int arg2) {
	AWTEvent ev = null;

	try {
	    Component target = ((ComponentModel)getModel()).getTarget();

	    switch (id) {
	    case ToolkitEventHandler.WINDOW_ENTERED:
		/* REMIND: if grab architecture changes, so must this! */
		if ((getToolkit().getGrabbingWindow() == null) || win.hasGrab()) {
		    ev = new MouseEvent(target, MouseEvent.MOUSE_ENTERED,
					when, modifiers, x, y, 0, false);
		    break;
		}

	    case ToolkitEventHandler.WINDOW_EXITED:
		/* REMIND: if grab architecture changes, so must this! */
		if ((getToolkit().getGrabbingWindow() == null) || win.hasGrab()) {
		    ev = new MouseEvent(target, MouseEvent.MOUSE_EXITED,
					when, modifiers, x, y, 0, false);
		    break;
		}

	    case ToolkitEventHandler.WINDOW_RESHAPED:
		if (getModel() instanceof WindowModel) {
		    if ((x | y) != 0) {
			ev = new WindowEvent((java.awt.Window)target,
					     ComponentEvent.COMPONENT_MOVED);
		    }

		    if ((arg1 | arg2) != 0) {
			ev = new WindowEvent((java.awt.Window)target,
					     ComponentEvent.COMPONENT_RESIZED);
		    }
		}
		break;
       
	    case ToolkitEventHandler.WINDOW_GOT_FOCUS:
		ev = new FocusEvent(target, FocusEvent.FOCUS_GAINED);
		break;

	    case ToolkitEventHandler.WINDOW_LOST_FOCUS:
		ev = new FocusEvent(target, FocusEvent.FOCUS_LOST);
		break;

	    case ToolkitEventHandler.WINDOW_DAMAGED:
		ev = new PaintEvent(target, PaintEvent.PAINT,
				    new Rectangle(x, y, arg1, arg2));
		/* FALL THROUGH */
	    default:
		super.handleEvent(id, when, modifiers, x, y, arg1, arg2);
		break;
	    }
	} catch (IllegalArgumentException targetIsNull) {
	    super.handleEvent(id, when, modifiers, x, y, arg1, arg2);
	    return;
	}
 
	if (ev != null) {
	    // System.err.println("posting " +ev);
	    getToolkit().postEvent(ev);
	}
    }

    /**
     * Called when bounds have been changed on the component.
     * @param m The model that originated the change.
     * @param r Rectangle which describes the new bounds.
     */
    protected void boundsChanged(Model m, Rectangle r) {
	/* nothing to do in the default case */
    }

    /**
     * Called when cursor has been changed on the component.
     * @param m The model that originated the change.
     * @param c The new cursor for the component.
     */
    protected void cursorChanged(Model m, Cursor c) {
	/* nothing to do in the default case */
    }

    /**
     * Called when insets have been changed on the component.
     * @param m The model that originated the change.
     * @param i The new insets for the component.
     */
    protected void insetsChanged(Model m, Insets i) {
	/* nothing to do in the default case */
    }

    /**
     * This is an override of the method in BasicView; the
     * implementation of updateClient on a ComponentPeer posts an
     * UPDATE event to the target Component.
     */
    public void updateClient(Object arg) {
	AWTEvent ev = null;

	try {
	    Component target = ((ComponentModel)getModel()).getTarget();
	    Rectangle r;

	    synchronized (this) {
	        if (arg instanceof Region) {
		    r = ((Region)arg).getBounds();
		    ev = new RepaintEvent(target,
				          PaintEvent.UPDATE,
				          getToolkit().winSys, (Region)arg);
	        } else if (arg instanceof Shape) {
	            r = ((Shape)arg).getBounds();
		    ev = new PaintEvent(target, PaintEvent.UPDATE, r);
	        } else {
		    r = getBounds();
		    r.x = r.y = 0;
		    ev = new PaintEvent(target, PaintEvent.UPDATE, r);
		}

		clearUpdateRegion(r);
	    }
	} catch (IllegalArgumentException targetIsNull) {
	    /* this should not really happen.  But just in case... */
	    super.updateClient(arg);
	    return;
	}
		
	if (ev != null) {
	    // System.err.println("posting " +ev);
	    getToolkit().postEvent(ev);
	}
    }

    /**
     * This is an override of the method in BasicView; the implementation
     * of getPreferredSize for a component must consult the target by default.
     * @param rows The desired number of rows
     * @param cols The desired number of columns
     */
    public Dimension getPreferredSize(int rows, int cols) {
	try {
	    Component c = ((ComponentModel)getModel()).getTarget();
	    return c.getSize();
	} catch (NullPointerException targetIsNull) {
	    return new Dimension(1, 1);
	}
    }

    /**
     * This is an override of the method in BasicView; the implementation
     * of getMinimumSize for a component must consult the target by default.
     * @param rows The desired number of rows
     * @param cols The desired number of columns
     */
    public Dimension getMinimumSize(int rows, int cols) {
	return getPreferredSize(rows, cols);
    }

    /**
     * This is an override of the method in BasicView; it simply sets the
     * correct color based on the Component's specified foreground color.
     */
    public Graphics getGraphics() {
	Graphics g = super.getGraphics();
	Model m    = getModel();
	boolean e  = (m instanceof ComponentModel) ? 
	    ((ComponentModel)m).isEnabled() : true;

	g.setColor(getColorPolicy().getForeground(m, false, e, false));

	return g;
    }

    /** 
     * This is an override of the method in BasicView;
     * it checks if component is enabled before actually 
     * requesting focus
     */
    public void requestFocus(){
	if (((ComponentModel)getModel()).isEnabled()){ 
		super.requestFocus();
	}
    }

    /**
     * Query whether this View should accept the keyboard focus 
     * during traversal.
     */
    public boolean isFocusTraversable() {
	return (((ComponentModel)getModel()).isEnabled() && 
		super.isFocusTraversable());
    }
}
