/*
 * @(#)ContainerModel.java	1.5 98/11/25 SMI
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package sun.awt.otk;

import java.awt.*;
import java.awt.peer.PanelPeer;

public class ContainerModel extends ComponentModel implements PanelPeer {
    private Insets insets;

    /**
     * Constructs a new ContainerModel instance. Initializes the insets to zeroes.
     * @param target the AWT Container component
     * @param tk the toolkit
     */
    public ContainerModel(Container target, ObjectToolkit tk) {
	super(target, tk);
	insets = new Insets(0, 0, 0, 0);
    }

    /** 
     * Sets the new enabled state.
     * @param b if true, this component is enabled; otherwise this component is disabled,
     * and so are all of its children
     */
    protected void doEnable(boolean b) {
	Component children[] = ((Container)getTarget()).getComponents();
	for (int i = 0; i < children.length; ++i) {
	    ComponentModel m = (ComponentModel) tk.targetToPeer(children[i]);
	    m.doEnable(b && children[i].isEnabled());
	}

	super.doEnable(b);
    }

    /** 
     * Sets new Insets for this container.
     * Notifies all dependent views to modify their appearance as necessary.
     * @param v the view to set the insets for
     * @param top the inset's top value
     * @param left the inset's left value
     * @param bottom the inset's bottom value
     * @param right the inset's right value
     */
    public void setInsets(View v,
			  int top, int left, int bottom, int right) {

	synchronized (insets) {
	    insets.top = top;
	    insets.left = left;
	    insets.right = right;
	    insets.bottom = bottom;
	}
	notifyDependents(insets);
    }

    /**
     * Retrieves the container's insets.
     * @return a new non-null instance of the container's insets
     */
    public Insets getInsets() {
	synchronized (insets) {
	    return (Insets) insets.clone();
	}
    }

    /**
     * Sets the background color of this container, and its children.
     * Calls the color policy for each child component to determine
     * how the new background color will affect each child, if at all.
     * @param c the new background color, which can be null
     * @see ColorPolicy
     */
    public void	setBackground(Color c) {

	Container target = (Container)getTarget();
	Component[] comps = target.getComponents();
	int nComps = target.getComponentCount();

	for (int i=0; i< nComps; i++){
	    ComponentModel model = (ComponentModel)tk.targetToPeer(comps[i]);
	    if (model != null){
		View dv = model.getDefaultView();
		if (dv != null){
		    ColorPolicy cp  = dv.getColorPolicy();
		    Color erase     = cp.getEraseColor(model, false, 
						       model.isEnabled(), 
						       false);
		    // set new erase color if it is a component;
		    // or just do the repaint 
		    model.setBackground(erase);
		}
	    }
	}
	super.setBackground(c);	
    }

    /**
     * Signifies the start of validation of components in the container.
     */
    public void beginValidate() {
	// notifyDependents(/* What do we say for a validate? */);
    }

    /**
     * Signifies the end of validation of components in the container.
     */
    public void endValidate() {
	// notifyDependents(/* What do we say for a validate? */);
    }

    /**
     * @deprecated Replaced by getInsets().
     * @see #getInsets
     */
    public final Insets insets() {
	return getInsets();
    }
}
