/*
 * @(#)ContainerViewImpl.java	1.21 99/09/29 SMI
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package sun.awt.otk;
import java.awt.*;
import java.util.*;
import sun.porting.windowsystem.*;
import sun.porting.toolkit.*;

/**
 * This is a lightweight ContainerView.  By lightweight is
 * meant a view that has no model associated with it.
 * That is, one that has no associated java.awt target.
 *
 * 1.21, 09/29/99
 */
public abstract class ContainerViewImpl implements ContainerView {


    /** 
     * The views in this container.
     */
    private Vector children = new Vector(4);

    /**
     * see getContainerLock()
     */
    private Object LOCK = new Object();

    /**
     * The actual window used to represent this View.
     */
    private sun.porting.windowsystem.Window win;

    /**
     * Toolkit.
     */
    private ObjectToolkit tk;

    /**
     * Insets for the container.
     */
    private Insets insets;


    /**
     * Creates a ContainerViewImpl object residing in the specified toolkit
     * with initial bounds and insets as indicated.
     *
     * @param toolkit to use.
     * @param bounds to use when creating the underlying window.
     * @param insets to use when creating the underlying window.
     */
    public ContainerViewImpl(ObjectToolkit toolkit, 
			     Rectangle bounds, 
			     Insets insets) {
        this(toolkit, null, bounds, insets,
	     WindowFactory.TOP_LEVEL_FRAMELESS);
    }
    
    /**
     * Creates a ContainerViewImpl object residing in the specified toolkit
     * with initial bounds and insets as indicated and parented to the
     * ContainerView specified.
     *
     * @param toolkit to use.
     * @param parent for underlying window.
     * @param bounds to use for underlying window.
     * @param insets to use for underlying window.
     */
    public ContainerViewImpl(ObjectToolkit toolkit, 
			     ContainerView parent,
			     Rectangle bounds, 
			     Insets insets) {
        this(toolkit, parent, bounds, insets,
	     WindowFactory.TOP_LEVEL_FRAMELESS);
    }

    /**
     * Create a ContainerViewImpl object as a view residing in the
     * specified toolkit.  Create the underlying window as a 
     * child of the given parent, with initial bounds as indicated 
     * and of the type specified.
     *
     * @param toolkit to use.
     * @param parent to use for underlying window.
     * @param bounds to use for underlying window.
     * @param insets to use for underlying window.
     * @param windowType type of window.
     */
    public ContainerViewImpl(ObjectToolkit toolkit, 
			     ContainerView parent,
			     Rectangle bounds, 
			     Insets insets, 
			     int windowType) {
	tk = toolkit;

	sun.porting.windowsystem.Window container =
	    (parent == null) ? null : parent.getContainerWindow();

	if (container == null) {
	    win = tk.winSys.makeTopLevelWindow(windowType, 
					       bounds.x, bounds.y, 
					       bounds.width, bounds.height);
	} else {
	    win = tk.winSys.makeWindow(container,
				       bounds.x, bounds.y, 
				       bounds.width, bounds.height);
	}

	this.insets = (Insets)insets.clone();
	win.setUserData(this);
    }

    /**
     * Causes subcomponents of the containuer to be laid out at their
     * preferred size.
     */
    protected void pack() {
	// size this window
	Dimension preferredSize = getPreferredSize(-1, -1);
	Rectangle bounds = getBounds();
	setBounds(bounds.x, bounds.y, 
		  preferredSize.width, preferredSize.height);
    }

  
    /**
     * @return The Window object to use as the parent for this view's children.
     */
    public sun.porting.windowsystem.Window getContainerWindow() {
	return win;
    }


    /**
     * No model is associated with this view.  Hence this
     * always returns null.
     */
    public Model getModel() {
	return null;
    }

    /**
     * Used to synchronize access to the list of views that is
     * maintained by the class.
     */
    final Object getContainerLock() {
	return LOCK;
    }


    /**
     * Returns a handle to the toolkit.
     */
    public final ObjectToolkit getToolkit() {
	return tk;
    }
	
    /**
     * Adds a view to the Container.
     * v to add as a child
     */
    public void add(View v) {
	children.addElement(v);
    }

    /**
     * @return the number of views.
     */
    public int getViewCount() {
	return children.size();
    }

    /**
     * Returns the view at the "index"th location.  If there
     * is no view at that location, it throws an ArraIndexOutOfBoundsException.
     */
    public View getView(int index) {
	return (View) children.elementAt(index);
    }

    /**
     * Gets the list of view stored in the Container.
     */
    public Enumeration getViews() {
	return children.elements();
    }

    /**
     * Removes the specified view.
     */
    public void remove(View v) {
	children.removeElement(v);
    }

    /**
     * Removes the view at location - "index"
     */
    public void remove(int index) {
	children.removeElementAt(index);
    }

    /**
     * Removes all views.
     */
    public void removeAll() {
	children.removeAllElements();
    }
	
 
    private static final int MINIMUM_SIZE = 0;
    private static final int PREFERRED_SIZE = 1;

    private Rectangle getRect(View v, 
			      int type, 
			      int rows, 
			      int cols) {

	Rectangle viewRect = v.getBounds(false);
	Dimension size = ((type == MINIMUM_SIZE) ? 
			  v.getMinimumSize(rows, cols) : 
			  v.getPreferredSize(rows, cols));
	return new Rectangle(viewRect.x, viewRect.y, 
			     size.width, size.height);
    }



    private Dimension getSize(int type, int rows, int cols) {
	//System.err.println("getSize()");
	if (children.size() == 0) {
	    return new Dimension(0, 0);
	}
	
	Enumeration childViews = getViews();
	Rectangle containerRect = getRect((View) childViews.nextElement(), type, rows, cols);

	while (childViews.hasMoreElements()) {
	    Rectangle rect = getRect((View) childViews.nextElement(), type, rows, cols);
	    containerRect = containerRect.union(rect);
	}

	if (insets != null) {
	    containerRect.width += insets.left + insets.right;
	    containerRect.height += insets.top + insets.bottom;
	}
	return new Dimension(containerRect.width, containerRect.height);
    }

    /**
     * Returns the preferred size of this container.  This is 
     * computed by taking the union of the preferred sizes of
     * all the views contained within it.  And then adding
     * in the insets.
     * @param rows The desired number of rows
     * @param cols The desired number of columns
     */
    public Dimension getPreferredSize(int rows, int cols) {
	return getSize(PREFERRED_SIZE, rows, cols);
    }

    /**
     * Returns the minimum size of this container.  This is 
     * computed by taking the union of the minimum sizes of
     * all the views contained within it. And then adding
     * in the insets.
     * @param rows The desired number of rows
     * @param cols The desired number of columns
     */
    public Dimension getMinimumSize(int rows, int cols) {
	return getSize(MINIMUM_SIZE, rows, cols);
    }

    /**
     * Sets the background color of the container.
     */
    public void setBackgroundColor(Color color) {
	win.setBackgroundColor(color);
    }

    /**
     * Sets the bounds of the container.
     * @param x position
     * @param y position
     * @param width specification
     * @param height specification
     */
    public void setBounds(int x, int y, int width, int height) {
	win.setBounds(x, y, width, height);
    }

    /**
     * Draws focus to this container.
     */
    public void requestFocus() {
	win.acquireInputFocus();
    }

    /**
     * Returns the visible state of the container as a boolean.
     */
    public boolean isVisible() {
	return win.isMapped();
    }

    /**
     * Sets the visible state of the container.
     * It also iterates over its children and updates
     * their visible state.
     */
    public void  setVisible(boolean visible) {
	Enumeration childViews = getViews();

	while (childViews.hasMoreElements()) {
	    View v = (View) childViews.nextElement();
	    v.setVisible(visible);
	}
	if (visible) {
	    win.map();
	} else {
	    win.unmap();
	}
    }


    public void  setStackOrder(int order) {
	win.setStackingOrder(order);
    }

    /**
     * @return the bounds of the container.
     */
    public Rectangle getBounds() {
	return win.getBounds(false);
    }

    /**
     * @return the bounds of the container.
     * @param global True if the bounds are desired 
     *               in global coords, false otherwise.
     */
    public Rectangle getBounds(boolean global) {
	return win.getBounds(global);
    }

    /**
     * @return the graphics object.
     */
    public Graphics getGraphics() {
	Graphics g = win.getGraphics();
	Rectangle r = getBounds();
	g.setClip(0, 0, r.width, r.height);
	g.setFont(tk.getDefaultFont());
	g.setColor(tk.getDefaultForeground());
	return g;
    }
 
    /**
     * @return the insets of the container.
     */
    public Insets getInsets() {
	return new Insets(insets.top, insets.left, 
			  insets.bottom, insets.right);
    }
 
    /** 
     * Set new Insets
     * @param insets to use.
     */
    public void setInsets(Insets insets){
	this.insets = new Insets(insets.top, insets.left,
				 insets.bottom, insets.right);
    }

    /**
     * The Updateclient interface, used to support the ScreenUpdater.
     * This method does nothing by default.  It exists as part of 
     * implementing the view interface.  
     * @see sun.awt.ScreenUpdater
     * @see sun.awt.UpdateClient
     */
    public void updateClient(Object arg) {
	return;
    }


    /**
     * An implementation method that helps support updateClient.  Return a shape
     * which includes the given rectangle, possibly merged with any other update
     * that is still pending.
     * This method does nothing by default.  It exists as part of 
     * implementing the view interface.  
     * @param x The x coordinate for the rectangle's left edge
     * @param y The y coordinate for the rectangle's top edge
     * @param width The width of the rectangle
     * @param height The height of the rectangle
     */
    public Shape getScreenUpdateRegion(int x, int y, int width, int height) {
	return null;
    }

    /**
     * Clear any pending updates that lie in the given rectangle.   (This is an
     * indication that the pending updates for this rectangle have been done.)
     * This method does nothing by default.  It exists as part of 
     * implementing the view interface.      
     * @param r The rectangle that has been updated.
     */
    public void clearUpdateRegion(Rectangle r) {
	return;
    }

    /**
     * The notification method for the Dependent interface.  Indicates that
     * an observable change has occurred on Subject m; arg is an optional 
     * argument which may help indicate what has changed.
     * Given that this view is not associated with any model, this method
     * does nothing by default. It exists as part of implementing the view 
     * interface.      
     */
    public void updateDependent(Subject m, Object arg) {
	return;
    }
    
    /**
     * The direct event delivery path to the View, always used for window-related
     * events and also used for input events if the View is not associated with an
     * awt.Component.  Event type identifiers are defined by 
     * sun.porting.toolkit.ToolkitEventHandler.
     *
     * Does nothing by default.
     * 
     * @param id The event type identifier
     * @param when A timestamp (in milliseconds) indicating when the event happened
     * @param modifiers The modifier key (SHIFT, CONTROL, ALT, META) state.
     * @param x The x coordinate associated with the event
     * @param y The y coordinate associated with the event
     * @param arg1 An optional argument whose meaning varies with the event type
     * @param arg2 An optional argument whose meaning varies with the event type
     * @see sun.porting.toolkit.ToolkitEventHandler
     */   
    public void handleEvent(int id, long when, int modifiers,
		     int x, int y, int arg1, int arg2) {
	return;
    }

    /**
     * Query whether this View should accept the keyboard focus during traversal.
     * @return false always.
     *
     */   
    public boolean isFocusTraversable() {
	return false;
    }

    /**
     * Add a Controller which listens to events on this View.
     * No controller support for this view.  The method exists
     * as part of implementing the view interface.
     */
    public synchronized void addController(Controller c) {
	return;
    }

    /**
     * Remove a Controller which was previously added to this view.
     * No controller support for this view.  The method exists
     * as part of implementing the view interface.
     */
    public synchronized void removeController(Controller c) {
	return;
    }

    /**
     * Remove all Controllers which were previously added to this view.
     * No controller support for this view.  The method exists
     * as part of implementing the view interface.
     */
    public synchronized void removeControllers() {
	return;
    }

    /**
     * Designates an already added controller as the drag controller.
     * Since no controller support xists, this method does nothing.
     */
    public synchronized void setDragController(Controller c) {
        return;
    }

   /**
    * @return Returns the current drag controller.
    * Since no controller support exists, this method always returns
    * null..
    *
    */
    public synchronized Controller getDragController() {
        return null;
    }
   
    /**
     * NO scrollbars associated with the view, hence it
     * always returns null.
     */
    public Dimension getScrollbarDimensions() {
	return null;
    }
    
    /**
     * No ColorPolicy exists in the container view and hence
     * this method always returns null.
     */
    public ColorPolicy getColorPolicy() {
	return null;
    }
}
