/*
 * @(#)InputController.java	1.11 98/11/29 SMI
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package sun.awt.otk;

import java.awt.AWTEvent;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import sun.porting.toolkit.ToolkitEventHandler;

/**
 * Boilerplate method definition for re-evaluating AWT input events via the
 * "generic" handleEvent method.  This is much easier to maintain than
 * requiring each controller to write duplicate event handling routines.
 */
public abstract class InputController implements Controller {

    /**
     * Empty default constructor
     */
    public InputController() {
    }

    /** 
     * Implementation of this method is called after InputEvent was 
     * re-evaluated.
     * @param   v view where an event occurred
     * @param   id see "generic" handleEvent for description
     * @param   when see "generic" handleEvent for description
     * @param   modifiers see "generic" handleEvent for description
     * @param   x see "generic" handleEvent for description
     * @param   y see "generic" handleEvent for description
     * @param   arg1 see "generic" handleEvent for description
     * @param   arg2 see "generic" handleEvent for description
     * @return <code>true</code> if this event was handled and 
     *         <code>false</code> - otherwise
     * @see #handleEvent(View, AWTEvent)
     */
    abstract public boolean handleEvent(View v, int id, 
					long when, int modifiers,
					int x, int y, int arg1, int arg2);

    /**
     * This a "generic" handleEvent that does re-evaluation of an 
     * incoming AWT event and calles controller's handleEvent with 
     * correct values. Which are <br>
     * v - view where an event occurred
     * id - ToolkitEventHandler event id after evaluation of InputEvent
     * when - input event's timestamp
     * modifiers - state of input event's modifier keys
     * x - 0 if AWTEvent is a java.awt.KeyEvent;
     *            <ul>x-coordinate of the mouse event if it was 
     *            a java.awt.event.MouseEvent. </ul>
     * y - 0 if AWTEvent is a java.awt.KeyEvent;
     *            <ul>y-coordinate of the mouse event if it was  
     *            a java.awt.event.MouseEvent.</ul>
     * arg1 -  key code of a key event <br>
     *             <ul>if AWTEvent's id is java.awt.KeyEvent.KEY_PRESSED,
     *                 or java.awt.KeyEvent.KEY_RELEASED,
     *                 or java.awt.KeyEvent.KEY_TYPED; </ul> 
     *             <ul>0 if AWTEvent's id is
     *                 java.awt.event.MouseEvent.MOUSE_ENTERED,
     *                 or java.awt.event.MouseEvent.MOUSE_EXITED; </ul>
     *             <ul>number of clicks if AWTEvent's id is
     *                 java.awt.event.MouseEvent.MOUSE_PRESSED, 
     *                 or java.awt.event.MouseEvent.MOUSE_RELEASED, 
     *                 or java.awt.event.MouseEvent.MOUSE_MOVED,
     *                 or java.awt.event.MouseEvent.MOUSE_DRAGGED </ul> 
     * 
     *arg2  - key char of a KeyEvent<br>
     *              <ul>if AWTEvent's id is
     *                  java.awt.KeyEvent.KEY_PRESSED,
     *                  or java.awt.KeyEvent.KEY_RELEASED,
     *                  or java.awt.KeyEvent.KEY_TYPED; </ul>
     *               <ul>1 for being a popUpTrigger 
     *                  if AWTEvent's id is java.awt.event.MOUSE_PRESSED,
     *                  or java.awt.event.MOUSE_RELEASED,
     *                  or java.awt.event.MOUSE_MOVED, 
     *                  or java.awt.event.MOUSE_DRAGGED and 
     *                  0 - otherwise</ul> <br>
     *
     * @param  v view where an event occurred
     * @param  e AWT event that occurred in the view <code> v </code>
     * @return value returned from the controller's version of handleEvent
     *         if AWTEvent's id is one of the following: 
     *         java.awt.event.KeyEvent.KEY_PRESSED, 
     *         java.awt.event.KeyEvent.KEY_RELEASED, 
     *         java.awt.event.KeyEvent.KEY_TYPED, 
     *         java.awt.event.MouseEvent.MOUSE_ENTERED, 
     *         java.awt.event.MouseEvent.MOUSE_EXITED, 
     *         java.awt.event.MouseEvent.MOUSE_PRESSED, 
     *         java.awt.event.MouseEvent.MOUSE_RELEASED, 
     *         java.awt.event.MouseEvent.MOUSE_MOVED, 
     *         java.awt.event.MouseEvent.MOUSE_DRAGGED;
     *         otherwise - returns <code>false</code>
     * @see #handleEvent(View, int, long, int, int, int, int, int)
     */
    public boolean handleEvent(View v, AWTEvent e) {
	if (! (e instanceof InputEvent)) {
	    return false;
	}

	InputEvent ie = (InputEvent)e;
	long when = ie.getWhen();
	int  mod  = ie.getModifiers();

	if (e instanceof KeyEvent) {
	    KeyEvent ke = (KeyEvent) e;

	    int keycode = (int) ke.getKeyCode();
	    int keychar = (int) ke.getKeyChar();

	    switch (ke.getID()) {

	    case KeyEvent.KEY_PRESSED:
		return handleEvent(v, ToolkitEventHandler.KEY_PRESSED, 
				   when, mod, 0, 0, keycode, keychar);

	    case KeyEvent.KEY_RELEASED:
		return handleEvent(v, ToolkitEventHandler.KEY_RELEASED,
				   when, mod, 0, 0, keycode, keychar);

	    case KeyEvent.KEY_TYPED:
		return handleEvent(v, ToolkitEventHandler.KEY_TYPED,
				   when, mod, 0, 0, keycode, keychar);

	    default:
		/* implies some new event type was added? */
		// System.err.println("Unrecognized KeyEvent " +ke);
		return false;
	    }
	} else if (e instanceof MouseEvent) {
	    MouseEvent me = (MouseEvent) e;

	    int x      = me.getX();
	    int y      = me.getY();
	    int clicks = me.getClickCount();
	    int pop    = me.isPopupTrigger() ? 1 : 0;

	    switch (me.getID()) {
	    case MouseEvent.MOUSE_EXITED:
		return handleEvent(v, ToolkitEventHandler.WINDOW_EXITED, 
				   when, 0,
				   x, y, 0, 0);

	    case MouseEvent.MOUSE_ENTERED:
		return handleEvent(v, ToolkitEventHandler.WINDOW_ENTERED,
				   when, 0,
				   x, y, 0, 0);

	    case MouseEvent.MOUSE_PRESSED:
		return handleEvent(v, ToolkitEventHandler.POINTER_PRESSED,
				   when, mod,
				   x, y, clicks, pop);

	    case MouseEvent.MOUSE_RELEASED:
		return handleEvent(v, ToolkitEventHandler.POINTER_RELEASED,
				   when, mod,
				   x, y, clicks, pop);

	    case MouseEvent.MOUSE_MOVED:
		return handleEvent(v, ToolkitEventHandler.POINTER_MOVED,
				   when, mod,
				   x, y, clicks, pop);

	    case MouseEvent.MOUSE_DRAGGED:
		return handleEvent(v, ToolkitEventHandler.POINTER_DRAGGED,
				   when, mod,
				   x, y, clicks, pop);

	    case MouseEvent.MOUSE_CLICKED:
		if (clicks <= 0) {
		    System.err.println("Click event with count of zero?");
		    Thread.dumpStack();
		    clicks = 1;
		}

		/* 
		 * Toolkit defines no separate event type for a click event.
		 * Instead, we signal this by using a negative click count,
		 * which is safe because a multi-click event can't have a
		 * count of zero!
		 */
		return handleEvent(v, ToolkitEventHandler.POINTER_PRESSED,
				   when, mod,
				   x, y, -clicks, pop);

	    default:
		/* implies some new event type was added? */
		// System.err.println("Unrecognized MouseEvent " +me);
		return false;
	    }
	}

	return false;
    }
}
