/*
 * @(#)ItemSelectable.java	1.5 98/11/29 SMI
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package sun.awt.otk;

/**
 * The interface for controlling a view with several items that maintains 
 * an "armed" state. Only one item can be "armed" at a time. The "armed" 
 * state of an object usually indicates a pending action. Example: a desktop
 * implementation of a list GUI will arm a list item when a mouse press is 
 * received inside that item, indicating that an action will occur on this 
 * item. It will disarm the list item if the mouse is dragged outside this 
 * list item bounds without a mouse release, re-arm the item if the mouse
 * is dragged back inside, and trigger the pending action only if the mouse
 * release is received while the mouse is inside the bounds of the list item.
 */
public interface ItemSelectable{
    public final int NO_ACTION        = 0;
    public final int ACTION_PERFORMED = 1;
    public final int STATE_CHANGED    = 2;

    /**
     * Return the current arming state of the view.
     */
    boolean isArmed();

    /**
     * Determines if the current view has been armed and then disarmed.  
     * In other words, isRearmable() will return true only
     * if armItem(Object) with non-null object was called, 
     * and then armItem(null) was called,
     * leaving the view in a disarmed state.
     * To reset the state after an action has been triggered,
     * the controller should use clear() instead.
     *
     * @return <code>true</code> if the current view has been 
     *         armed and then disarmed; 
     *         otherwise returns - <code> false</code>
     */
    boolean isRearmable();


    /** 
     * Setd the arming state.  This is used for initial arming, for
     * disarming, and for rearming.
     * @param item if <code>item</code> is <code> null</code> view is 
     * disarmed; otherwise it is armed or re-armed.
     */
    void    armItem(Object item);

    /**
     * Clear the armed state; both isArmed() and isRearmable() will
     * return false after this call. 
     * If action is ACTION_PERFORMED or STATE_CHANGED, trigger the
     * pending action on the peer/model associated with this view.
     * @param action type of action to trigger
     */
    void    clear(int action);

    /**
     * Retrieves an item that corresponds to the specified coordinates.
     * @param x x-coordinate of a list item
     * @param y y-coordinate of a list item
     */
    Object getItem(int x, int y);

    /**
     * Retrieves currently armed item.
     * @return currently armed item; if no item is armed - 
     * <code> null </code> is returned.
     */
    Object getArmedItem();

    /**
     * Some of the actions have to be performed independently of 
     * view's armed state. 
     */
    //REMIND! This should not be part of this interface
    boolean performActionOnItem();
}
