/*
 * @(#)LabelMenuItem.java	1.7 98/11/29 SMI
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package sun.awt.otk;

import java.awt.*;

/**
 * It is an abstract base class for MenuItemModel, 
 * and ChoiceMenuItem (used in ChoiceModel). 
 * It is mostly a placeholder for states like font, enabled, and label
 * and a basic implementation for MenuItemPeer.
 */
abstract public class LabelMenuItem implements MenuItemOperations {
    protected String    label;
    protected MenuItemOwner owner = null;

    /**
     * Implementation of this method will generate an appropriate 
     * event for this menu item when it is called.
     * @see MenuItemModel#actionOccurred
     */
    abstract public synchronized void actionOccurred(int modifiers);

    /**
     * Creates new LabelMenuItem with a specified label.
     * Its menu owner will be set to <code>null</code>.
     * @param   label  menu item's label
     */
    public LabelMenuItem(String label) {
	this(label, null);
    }

    /**
     * Creates new LabelMenuItem with a specified label and owner.
     * @param   label  menu item's label
     * @param   owner  menu item's owner
     */
    public LabelMenuItem(String label, MenuItemOwner owner) {
	setLabel(label);
	setOwner(owner);
    }

    /**
     * This method should delete whatever is necessary, and should
     * be careful to null out all references to other objects.     
     */
    public synchronized void dispose() {
	owner = null;
    }

    /**
     * Sets the menu item's label.
     * @param    label the string speciying the menu item's label
     * @see java.awt.MenuItem#setLabel
     */
    public synchronized void setLabel(String label) {
	if (label == null) {
	    label = "";
	}

	if (!label.equals(this.label) ) {
	    this.label = label;

	    if (owner != null) {
		owner.itemChanged(this);
	    }
	}
    }

    /**
     * Enables or disables the menu item.
     * By default label items are always enabled.
     * Sub-classes can override that behavior.
     * @see java.awt.MenuItem#setEnabled
     * @see MenuItemModel#setEnabled
     * @see #isEnabled
     */
    public void setEnabled(boolean cond) {
	/* by default, label items are always enabled */
    }

    /**
     * Determines if this menu item is enabled or not.
     * Sub-classes can override that behavior.
     * @return  default label menu items are always enable,
     *          so <code>true</code> is always returned
     * @see #setEnabled
     * @see MenuItemModel#isEnabled
     */
    public synchronized boolean isEnabled() {
	return true;
    }

    /**
     * DEPRECATED:  Replaced by setEnabled(boolean).
     */
    public final void enable() {
	setEnabled(true);
    }

    /**
     * DEPRECATED:  Replaced by setEnabled(boolean).
     */
    public final void disable() {
	setEnabled(false);
    }

    /**
     * Sets menu item owner for that menu item.
     * @see MenuItemOwner
     */
    public synchronized final void setOwner(MenuItemOwner owner) {
	this.owner = owner;
    }

    /**
     * Determines if this menu item is a separator.
     * By default it is not a separator. So <code> false</code>
     * is always returned. 
     * @see MenuItemModel
     */
    public boolean isSeparator() {
	return false;
    }

    /**
     * Retrieves this menu items' label.
     * @return  this menu item's label
     * @see #setLabel
     */
    public synchronized final String getLabel() {
	return label;
    }

    /**
     * Retrieves this menu item's font
     * @return this menu item's font
     */
    public synchronized Font getFont() {
	Font f = null;

	if (owner != null) {
	    f = owner.getFont();
	}

	/* REMIND: we need a more flexible way of getting the default font; */
	/* if we can't inherit it from owner we should inherit it from the  */
	/* toolkit that is currently in use, for example.                   */

	return (f == null) ? new Font("Dialog", Font.PLAIN, 12) : f;
    }
}
