/*
 * @(#)LabelledComponentModel.java	1.10 98/11/29 SMI
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package sun.awt.otk;

import java.awt.*;
import java.awt.event.*;
import java.awt.peer.ButtonPeer;
import java.awt.peer.LabelPeer;

/**
 * This model suffices for Label and Button and can be extended for Checkbox. 
 */
public class LabelledComponentModel extends ComponentModel implements ButtonPeer, LabelPeer {

    private String text = "";
    private int    alignment;

    /**
     * Constructs LabelledComponentModel with specified (in the target)
     * text/label and alignment. It can be used as a Button peer, Label peer or
     * a super class of a Checkbox peer.
     * @param   target  Component that is expected to be java.awt.Label, or
     *                  java.awt.Button, or java.awt.Checkbox
     * @param   tk      current Toolkit
     */
    public LabelledComponentModel(Component target, ObjectToolkit tk) {
	super(target, tk);
    }

    /**
     * Sets and initializes the default view. Text is set 
     * for button, label, checkbox. Alignment is set for label.
     * @param   v   new default view
     */
    public synchronized void setDefaultView(View v) {
	super.setDefaultView(v);

	Component target = getTarget();

	if (target instanceof Button) {
	    setText(((Button)target).getLabel());
	} else if (target instanceof Label) {
	    setText(((Label)target).getText());
	    setAlignment(((Label)target).getAlignment());
	} else if (target instanceof Checkbox) {
	    setText(((Checkbox)target).getLabel());
	}

    }

    /**
     * Sets labelled component  (button, label, checkbox) peer's
     * text or label to a new value
     * @param   text   new text
     * @see #setLabel
     * @see java.awt.Label#setText
     */
    public void setText(String text) {
	if (text == null) {
	    text = "";
	}

	if (!text.equals(this.text)) {

	    this.text = text;

	    /* force recalculation of size  and full repaint*/
	    Font f = getTarget().getFont();
	    if (f != null) {
		setFont(f);
	    }
	}
    }

    /**
     * Sets the alignment for this label's peer to the specified one.
     * Possible values are: <code>java.awt.Label.CENTER</code>, 
     *                      <code>java.awt.Label.RIGHT</code>, 
     *                      <code>java.awt.Label.LEFT</code>)
     * @param  alignment  new label alignment 
     * @see java.awt.Label#setAlignment
     */
    public void setAlignment(int alignment) {
	if (this.alignment != alignment) {

	    this.alignment = alignment;
	    notifyDependents();	// re-alignment usually requires full repaint
	}
    }

    /**
     * Alias for ButtonPeer/CheckboxPeer interfaces 
     * @see #setText
     */
    public final void setLabel(String label) {
	setText(label);
    }

    /**
     * Retrieves the text of this labelled component (label, button, checkbox)
     * @return  the text of this labelled component
     * @see     #setText
     */
    public final String getLabel() {
	return text;
    }

    /** 
     * Retrieves the current alignment of this label. Possible values are
     * <code>java.awt.Label.LEFT</code>, 
     * <code>java.awt.Label.RIGHT</code>, and 
     * <code>java.awt.Label.CENTER</code>.
     * @see        #setAlignment
     */
    public final int getAlignment() {
	return alignment;
    }

    /**
     * Posts action event as a result of some feel specific action on
     * the view. (For example, a MOUSE_RELEASE inside the button).
     * Among labelled components only button is firing
     * action events.
     * @see java.awt.event.ActionEvent
     * @see java.awt.evetn.ActionListener
     */
    public void postActionPerformed() {
	Button target = (Button)getTarget();
	postActionEvent(ActionEvent.ACTION_PERFORMED, target.getActionCommand());
    }
}
