/*
 * @(#)ListModel.java	1.13 98/11/29 SMI
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

package sun.awt.otk;
import java.awt.*;
import java.awt.event.ItemEvent;

public class ListModel extends ComponentModel implements java.awt.peer.ListPeer{

    private OrderedList items;

    // Constant used to grow the items capacity and selected array
    private final static int   CAPACITY_INCR = 20;

    // used for single selection lists
    private int selectedItem;

    // index taht is specially painted to show that item is focus
    // it is NONE when no item is specially painted
    private static final int NONE           = -1; 
                                              
    private boolean        multipleSelections;

    /**
     * Constructs a new ListModel instance.
     * OrderedListImpl is created to store information about list items.
     * @param target the AWT Component
     * @param tk the toolkit
     */    
    public ListModel(List target, ObjectToolkit tk) {
	this(target, tk,
	     new OrderedListImpl(target.getItemCount(), CAPACITY_INCR));
    }

    /**
     * Constructs a new ListModel instance.
     * @param target the AWT Component
     * @param tk the toolkit
     * @param itemList OrderedList where ListModel will store data about
     *        each list item
     */
    public ListModel(List target, ObjectToolkit tk, OrderedList itemList){
	super(target, tk);

        int sel[] = target.getSelectedIndexes();

	/* set whether this list should allow multiple selections. */
        multipleSelections = target.isMultipleMode();

       	if (!multipleSelections && sel.length > 0){
	    selectedItem = sel[0];
	} else {
	    selectedItem = NONE;
	}

	items = itemList;

	/* add any items that were already inserted in the target.
	 * set corresponding selected element to true if 
	 * if the item is already selected in the target
	 */
	int nitems = target.getItemCount();

	for (int i = 0; i < nitems; i++) {
	    boolean selected = false;

	    // check if it is selected
	    for(int j=0; j < sel.length; j++){
		if (sel[j] == i){
		    selected = true;
		    break;
		}
	    }

	    items.addElement(new ListItem(target.getItem(i), selected));
	}
    }

    /**
     * Adds item to the list. 
     * Item is added to the OrderedList. Then dependents are notified that 
     * item was added by sending StateChangeNotification with type set 
     * to StateChangeNotification.ADD and with start being equal to 
     * item's index.
     * @param item   the item to be added
     * @param index  the position in the list at which to add the item
     * @see          StateChangeNotification
     * @see          java.awt.List#add
     */
    public synchronized void add(String item, int index) {

	// if the index of the item is < 0 or >= than items.size() 
	//then add the item to the end of the list
        if (index >= 0 && index < items.size()) {
	    items.insertElementAt(new ListItem(item, false), index);
        } else {
	    index = items.size();
	    items.addElement(new ListItem(item, false));
	}
	

	// update selectedItems if it is a single selection list
	if (!multipleSelections){
	    
	    if (selectedItem >= index){
		selectedItem ++;
	    }
	}

	notifyDependents(new StateChangeNotification(StateChangeNotification.ADD, index, index));
    }

    /**
     * Deletes a range of items from the list.
     * Items are deleted from to the OrderedList. 
     * Then dependents are notified that 
     * items were deleted by sending StateChangeNotification with type set 
     * to StateChangeNotification.DELETE and with correctly set 
     * start and end.
     * @param end    the 0-based index of the last item in the range
     * @param start  the 0-based index of the first item in the range
     * @see          StateChangeNotification
     * @see          java.awt.List#delItems
     */
    public void delItems(int start, int end){

	if (items.size() == 0){
	    return;
	}

	// if user passed in flipped args, reverse them
	if (start > end){
	    int tmp = start;
	    start = end;
	    end = tmp;
	}
	
	// make sure start and end exist
	start = Math.max(start, 0);
	end = Math.min(end, items.size()-1);

	// determine whether we're going to delete any visible elements
        // boolean repaintNeeded = isVisible(start, end);

        // delete items from start to end from 
	// items vector   
	int nDelItems = start - end +1;   
	for(int i = start; i <= end; i++){
	    items.removeElementAt(start); 
	}
	

	// update selectedItems if it is a single selection list
	if (!multipleSelections){
	    
	    if (selectedItem >= start && selectedItem <= end){
		selectedItem = NONE;
	    } else if (selectedItem > end){
		selectedItem -= nDelItems;
	    }
	}

	notifyDependents(new StateChangeNotification(StateChangeNotification.DELETE, start, end));
    }

    /**
     * Deselects an item in the list.
     * Then dependents are notified that 
     * item was deselected by sending StateChangeNotification with type set 
     * to StateChangeNotification.DESELECT and with start being equal to 
     * item's index.
     * @param index  the 0-based index of the item to deselect
     * @see          StateChangeNotification
     * @see          java.awt.List#deselect
     */
    public synchronized void deselect(int index){

	ListItem item = getListItem(index);
	if (item == null || !item.isSelected()){
	    return;
	}	
	
	// update selectedItem 
	if (!multipleSelections){
	    selectedItem = NONE;
	}
	
	// repaint deselected index 	    
	item.setSelected(false);

	notifyDependents(new StateChangeNotification(StateChangeNotification.DESELECT, index, index));
    }

    /**
     * This version of deselect is called in response to the user input.
     * After an item was deselected an ItemEvent is posted.
     * @param index the 0-based index of the item to deselect
     * @param postEvent <code>true </code> to post an ItemEvent;
     *                  <code>false</code> - otherwise.
     * @see #deselect(int)
     */
    public void deselect(int index, boolean postEvent) {
	deselect(index);

	// don't want to post event if deselect was called programmatically, 
	// as opposed to in response to user input
	if (postEvent) {
	    ItemEvent ev = new ItemEvent((List)getTarget(),
					 ItemEvent.ITEM_STATE_CHANGED,
					 new Integer(index),
					 ItemEvent.DESELECTED);
	    tk.postEvent(ev);
	}
    }



    /**
     * Calculates the minimum size of this list's default view with 
     * a specified number of rows.
     * @param  rows number of row that must be visible in the list
     * @return a Dimension object indicating the default list's minimum size.
     */
    public Dimension getMinimumSize(int rows){
	View lv = getDefaultView();
	return lv == null ? null : lv.getMinimumSize(rows, -1);
    }

    /**
     * Calculates the preferred  size of this list's default view with 
     * a specified number of rows.
     * @param  rows number of row that must be visible in the list
     * @return a Dimension object indicating the default list's 
     *         preferred size.
     */
    public Dimension getPreferredSize(int rows){
	View lv = getDefaultView();
	return (lv == null ? null : lv.getPreferredSize(rows, -1));
    }

    /**
     * Retrieves the indexes of the selected items.
     * @return    an array containing the list of indices
     * @see       java.awt.List#getSelectedIndexes
     * 
     */
    public synchronized int[] getSelectedIndexes(){
	// REMIND! This must be done more efficiently 
	int nSelected = 0;
	for(int i=0; i<items.size(); i++){
	    if (((ListItem)items.elementAt(i)).isSelected()){
		nSelected++;
	    }
	}

	int[] sel = new int[nSelected];
	for(int i=0, j=0; i<items.size(); i++){
	    if (((ListItem)items.elementAt(i)).isSelected()){
		sel[j++] = i; 
	    }
	}
	return sel;
    }

    /**
     * Notifies dependents with an Integer as an argument 
     * which is equal to the index of the item that has to be visible.
     * @param index    the 0-based index of the item to make visible
     * @see            java.awt.List#makeVisible
     */
    public void makeVisible(int index){
	notifyDependents(new Integer(index));
    }

    /**
     * Removes all items from the list.
     * Then dependents are notified that all items were removed 
     * by sending StateChangeNotification with type set 
     * to StateChangeNotification.CLEAR
     * @see          StateChangeNotification
     */
    public synchronized void removeAll() {

	items.removeAllElements();

	notifyDependents(new StateChangeNotification(StateChangeNotification.CLEAR, 0, 0));
    }

   /**
     * Selects an item in the list.
     * Then dependents are notified that 
     * item was deselected by sending StateChangeNotification with type set 
     * to StateChangeNotification.DESELECT and with start being equal to 
     * item's index.
     * @param index    the 0-based index of the item to select
     * @see            StateChangeNotification
     * @see            java.awt.List#select
     */
    public void select(int index){
	ListItem item = getListItem(index);
	if (item == null || item.isSelected()){
	    return;
	}      
	
	int oldSelectedItem = selectedItem;

	// update selectedItem
	if (!multipleSelections){
	    if (validItemIndex(oldSelectedItem)){
	  ((ListItem)items.elementAt(oldSelectedItem)).setSelected(false);
	    }
	    selectedItem = index;
	}
	
	// paint just selected item
	item.setSelected(true);

	notifyDependents(new StateChangeNotification(StateChangeNotification.SELECT, 
					   index, oldSelectedItem));
    }

    /**
     * This version of select is called in response to the user input.
     * After an item was selected an ItemEvent is posted.
     * @param index the 0-based index of the item to select
     * @param postEvent <code>true </code> to post an ItemEvent;
     *                  <code>false</code> - otherwise.
     * @see #select(int)
     */
    public void select(int index, boolean postEvent) {
	select(index);

	// don't want to post event if select was called programmatically, 
	// as opposed to in response to user input
	if (postEvent) {
	    ItemEvent ev = new ItemEvent((List)getTarget(),
					 ItemEvent.ITEM_STATE_CHANGED,
					 new Integer(index),
					 ItemEvent.SELECTED);
	    tk.postEvent(ev);
	}
    }

    /**
     * Sets selection mode. 
     * If <code> on</code> is <code> true</code>, this method sets this list
     * in multi-selection mode; otherwise, it sets this list in
     * single-selection mode.
     * Then dependents are notified that selection mode changed 
     * by sending StateChangeNotification with type set 
     * to StateChangeNotification.SELECT_MODE.
     * @param index    the 0-based index of the item to select
     * @see            StateChangeNotification
     * @param on specifies a new selection mode in the list
     */
    public synchronized void setMultipleMode(boolean on) {
	if (multipleSelections != on){

	    multipleSelections = on;
	    if (multipleSelections){
		selectedItem = NONE;
	    } else {
		int[] sel = getSelectedIndexes();
		if (sel.length > 0){
		    for(int i=1; i<sel.length; i++){
			deselect(i);
		    }
		    selectedItem = sel[0];
		} else {
		    selectedItem = NONE;
		}
	    }

	    notifyDependents(new StateChangeNotification(StateChangeNotification.SELECT_MODE));
	} 
    }


    /**
     * DEPRECATED: Replaced by add().
     */
    public synchronized void addItem(String item, int index) {
	add(item, index);
    }

    /**
     * DEPRECATED: Replaced by removeAll().
     */
    public synchronized void clear() {

	removeAll();
    }

    /**
     * DEPRECATED: Replaced by getMinimumSize(int rows).
     */
    public final Dimension minimumSize(int rows){
	return getMinimumSize(rows);
    }

    /**
     * DEPRECATED: Replaced by getPreferredSize().
     */
    public final Dimension preferredSize(int rows){
	return getPreferredSize(rows);
    }

    /**
     * DEPRECATED: Replaced by setMultipleMode().
     */
    public synchronized void setMultipleSelections(boolean on) {
	setMultipleMode(on);
    }

    private boolean validItemIndex(int index){
	return(index >= 0 && index < items.size());
    }

    // ********************************
    
    /**
     * Retrieves number of list item in the ordered list.
     */
    public int getItemsCount(){
	return items.size();
    }

    /**
     * Retrieves list item with a specified index.
     * @return list item with a specified index if this index is valid; 
     *         otherwise - returns null.
     * @see ListItem
     */
    public ListItem getListItem(int index){
	if (!validItemIndex(index)){
	    return null;
	}
	return (ListItem)items.elementAt(index);
    }
    
    /**
     * Retrieves current selection mode.
     * @return <code>true</code> if list is in multi-selection mode;
     *         otherwise - returns <code> false</code>
     */
    public boolean isMultipleMode() {
	return multipleSelections;
    }

    /** 
     * Retrieves the selected item if this list is in 
     * a single-selection mode.
     * @return ListModel.NONE - if nothing is selected or 
     *         list is in multi-selection mode;
     *         otherwise - index of a selected item
     */
    public int getSelectedItem(){
	if (multipleSelections){
	    return NONE;
	}
	
	return selectedItem;
    }

    /**
     * Retrieves index of a list item.
     * @param list list item which index is queried
     * @return list item's index if there is such list item in
     *         the ordered list and ListModel.NONE otherwise
     */
    public int getIndex(ListItem listItem){
	return (listItem == null ? NONE : 
		                   items.indexOf(listItem));
    }

 
    /** 
     * Retrieves number of rows that were specified to 
     * create this list.
     * @return number of rows 
     * @see java.awt.List#getRows
     */
    public final int getRows() {
	List target = (List)getTarget();
	return target.getRows();
    }
}


