/*
 * @(#)ListMouseController.java	1.8 98/11/29 SMI
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package sun.awt.otk;

import java.awt.event.ActionEvent;
import java.awt.event.ItemEvent;
import sun.porting.toolkit.ToolkitEventHandler;


/**
 * A controller that consumes mouse events and interprets them for the
 * benefit of View objects that implement ItemSelectable.   
 * The consumeAll flag is not really a "state;" 
 * it allows for two variants of the controller to both be instantiated 
 * as Singletons.
 */
public class ListMouseController extends InputController implements StatelessController {
    /**
     * Instance of ListMouseController that consumes mouse events.
     */
    public static final ListMouseController consumingController = 
        new ListMouseController(true);

    /**
     * Instance of ListMouseController that does not consumes mouse events.
     */
    public static final ListMouseController nonConsumingController = 
        new ListMouseController(false);

    /**
     * If it is <code> true</code> - mouse events are consumed,
     * otherwise - mouse events are not consumed.
     * It is set in the constructor of ListMouseController.
     */
    public final boolean consumeAll;

    /**
     * Constructs ListMouseController which will or will not consume
     * mouse events.
     * @param  consumeAllMouseEvents if it is <code> true</code> - 
     *                               mouse events are consumed, otherwise - 
     *                               mouse events are not consumed.
     */
    public ListMouseController(boolean consumeAllMouseEvents){
	consumeAll = consumeAllMouseEvents;
    }

    /** 
     * Implementation of this method is called after InputEvent was 
     * re-evaluated. It will arm an item that was pressed. Disarm an
     * armed item if mouse was dragged outside of that item and re-arm
     * it again it it was the mouse was dragged back into the item.
     * If an item is double-clicked performActionOnItem will be called.
     * @param   v view where an event occurred
     * @param   id see "generic" handleEvent for description
     * @param   when see "generic" handleEvent for description
     * @param   modifiers see "generic" handleEvent for description
     * @param   x see "generic" handleEvent for description
     * @param   y see "generic" handleEvent for description
     * @param   arg1 see "generic" handleEvent for description
     * @param   arg2 see "generic" handleEvent for description
     * @return <code>true</code> if this event was handled and 
     *         <code>false</code> - otherwise
     * @see InputController#handleEvent(View, AWTEvent)
     */
    public boolean handleEvent(View v,
			       int id, long when, int modifiers,
			       int x, int y, int clicks, int arg2) {
	ItemSelectable s = (ItemSelectable) v;

	switch (id) {
	case ToolkitEventHandler.POINTER_PRESSED:
	    {
               switch(clicks){
	       case 1:
			Object item = s.getItem(x, y);		

                        if (item == null){
		           return consumeAll;
		        }

		        s.armItem(item);
			
			return true;

	       case -2:
		   if (s.performActionOnItem()){
		       return true;

		   }
		   /* Fall through */
		default:
		    return consumeAll;
		}
	    }
	case ToolkitEventHandler.POINTER_RELEASED:
	    {

		if (clicks != 1 && clicks != 0){
		    return consumeAll;
		}
	      
		if (s.isArmed()){
		    s.clear(ItemSelectable.STATE_CHANGED);
		} else {	
		    s.clear(ItemSelectable.NO_ACTION);
		}

		return true;
	    }

	case ToolkitEventHandler.POINTER_MOVED:
	    return consumeAll;

	case ToolkitEventHandler.POINTER_DRAGGED:
	    {
		Object item = s.getItem(x, y);
		Object armedItem = s.getArmedItem();	

		if (item != armedItem && armedItem != null){
		    s.armItem(null);

		}

		if (s.isRearmable() && item != null){
		    s.armItem(item);
		}
	    return true;
	    }
	case ToolkitEventHandler.WINDOW_ENTERED:
	    {
		Object item = s.getItem(x, y);
		if (item == null){
		    return consumeAll;
		}
		
		if (s.isRearmable()){
		    s.armItem(item);
		}
	    }		
	case ToolkitEventHandler.WINDOW_EXITED:
		
	    if (s.isRearmable()){
		s.armItem(null);
	    }
	
	default:
	    return false;
	}
	
    }

}
