/*
 * @(#)MenuItemModel.java	1.10 98/11/25 SMI
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package sun.awt.otk;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.peer.MenuItemPeer;

/**
 * The model used for a simple menu item.
 * This model does not support views--that would be
 * very heavyweight, given JDK's restrictions on what little you can
 * do with a MenuItem!  (Set its label, set its font, enable/disable
 * it, add it to a menu.)  Instead, this model is merely a placeholder
 * for the appropriate data, and whenever a relevant change occurs, the
 * parent menu is notified.
 *
 * Most of the implementation is inherited from LabelMenuItem.
 */
public class MenuItemModel extends LabelMenuItem {
    private MenuItem  target  = null;
    private boolean   enabled;

    public MenuItemModel(MenuItem target) {
	super(target.getLabel());
	this.target = target;

	setEnabled(target.isEnabled());
    }

    /**
     * Set this menu item's enable state
     * @param b The enabled state: true for enabled, false for disabled.
     */
    public synchronized void setEnabled(boolean b) {
	if (enabled != b) {
	    enabled = b;
	    if (owner != null) {
		owner.itemChanged(this);
	    }
	}
    }

    /**
     * Get the enabled state for this item
     * @return The enabled state: true for enabled, false for disabled.
     */
    public synchronized boolean isEnabled() {
	return enabled;
    }

    /**
     * Destroy the peer and free its resources.
     */
    public synchronized void dispose() {
	target = null;
	super.dispose();
    }

    /**
     * Get the font used to render this item.
     * @return The font (a java.awt.Font object) used to render this item.
     */
    public synchronized final Font getFont() {
	Font f = target.getFont();

	return (f == null) ? super.getFont() : f;
    }

    /**
     * Called when an action is required on this menu item.  This happens,
     * for example, when the user clicks on the item.  A call to 
     * actionOccurred() usually results in posting an ACTION event.
     * @param modifiers The modifier state in effect when the action happened.
     */
    public synchronized void actionOccurred(int modifiers) {
	if (owner == null) {
	    return;
	}

	String cmd = target.getActionCommand();
	ActionEvent ev = new ActionEvent(target,
					 ActionEvent.ACTION_PERFORMED,
					 cmd == null ? target.getLabel() : cmd,
					 modifiers);
	((MenuModel)owner).getToolkit().postEvent(ev);
    }

    /**
     * A place-holder in menu item lists, indicating a separator.
     */
    public final static MenuItemOperations SEPARATOR = new MenuSeparator();

    /*
     * The class that implements the placeholder.  For the convenience
     * of the view, this implements both MenuSeparatorQuery and MenuItemPeer.
     */
    private static class MenuSeparator implements MenuItemOperations {
	public final boolean isSeparator() {
	    return true;
	}

	public final boolean isEnabled() {
	    return false;
	}

	public final String getLabel() {
	    return "-";
	}

	public final Font getFont() {
	    return null;
	}

	public final void setOwner(MenuItemOwner owner) { }
	public final void dispose() { }
	public final void setLabel(String label) { }
	public final void setEnabled(boolean b) { }
	public final void enable() { }
	public final void disable() { }
	public final void actionOccurred(int modifiers) { }
    }
}

