/*
 * @(#)MenuModel.java	1.13 98/11/25 SMI
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package sun.awt.otk;

import java.awt.*;
import java.awt.image.*;
import java.awt.peer.MenuPeer;
import java.util.Locale;

/**
 * A model that implements the peer interface for a full-fledged menu.
 * It allows menu items to be added and deleted, supports views, and functions
 * as a MenuItemOwner.
 */
public class MenuModel extends MenuComponentModel implements MenuPeer, MenuItemOperations, MenuItemOwner {
    /**
     * Create a new model, obtaining the list of items from the target.
     */
    public MenuModel(Menu target, ObjectToolkit tk) {
	this(target, tk, new OrderedListImpl(target.getItemCount()));
    }

    /** 
     * Create a new MenuModel using the specified list of items.
     */
    public MenuModel(Menu target, ObjectToolkit tk, OrderedList itemList) {
	super(target, tk);
	items = itemList;

	int count = target.getItemCount();
	for (int i = 0; i < count; ++i) {
	    addItem(target.getItem(i));
	}
    }

    /**
     * Dispose of all the resources associated with this peer.
     */
    public synchronized void dispose() {
        // this method should delete whatever is necessary, and should
        // be careful to null out all references to other objects.
	if (itemModel != null) {
	    itemModel.dispose();
	    itemModel = null;
	}

	items.removeAllElements();
	items = null;

	super.dispose();
    }

    /**
     * Append a separator (a horizontal bar) to the menu's peer representation.
     */
    public void addSeparator() {
	doAdd(MenuItemModel.SEPARATOR);
    }

    /**
     * Append the supplied item to the menu's peer representation.  If the
     * item's label is exactly equal to "-" append a separator instead.
     * @param item The item to add.
     */
    public void addItem(MenuItem item) {
	if (item.getLabel().equals("-")) {
	    // JDK1.1 creates a separator using a MenuItem with label "-". UGH.
	    addSeparator();
	} else {
	    item.addNotify();
	    doAdd((MenuItemOperations)getToolkit().targetToPeer(item));
	}
    }

    /**
     * Delete the menu item at the specified index.
     * @param index The index of the item to be deleted.
     */
    public void delItem(int index) {
	doRemove(index);
    }

    /**
     * Post the given menu show event via the AWT event queue.
     * @param evt The event to be posted.
     */
    public void postMenuShowEvent(AWTEvent evt) {
	getToolkit().postEvent(evt);
    }

    /* this is the MenuItemPeer interface */

    /**
     * Set the label (i.e. the title) associated with this menu.
     * @param label The desired label string.
     */
    public synchronized void setLabel(String label) {
	if (itemModel != null) {
	    itemModel.setLabel(label);
	}
    }

    /**
     * Set the enabled state on the menu.  This is normally useful only
     * when the menu is itself an item in another menu.
     * @param b The desired enabled state (true/false).
     */
    public synchronized void setEnabled(boolean b) {
	if (itemModel != null) {
	    itemModel.setEnabled(b);
	}
    }

    /**
     * DEPRECATED:  Replaced by setEnabled(boolean).
     */
    public final void enable() {
	setEnabled(true);
    }

    /**
     * DEPRECATED:  Replaced by setEnabled(boolean).
     */
    public final void disable() {
	setEnabled(false);
    }

    /* this is the rest of the MenuItemOperations interface */

    /**
     * Query whether this is a separator.
     * @return false always, because a Menu is not a separator.
     */
    public final boolean isSeparator() {
	return false;
    }

    /**
     * Set the owner.  This is normally of use when the menu
     * is functioning as a MenuItem or in an implementation of some
     * other component like a Choice.
     * @param owner The new owner.
     */
    public synchronized void setOwner(MenuItemOwner owner) {
	if ((itemModel == null) ^ (owner == null)) {
	    if (owner == null) {
		itemModel.dispose();
		itemModel = null;
		return;
	    }

	    itemModel = new MenuItemModel((MenuItem)getTarget());
	} 

	itemModel.setOwner(owner);
    }

    /**
     * Get the enabled state for this menu.  This is normally of
     * use only when the menu is functioning as a MenuItem.
     * @return true if this menu is enabled, false otherwise.
     */
    public synchronized boolean isEnabled() {
	if (itemModel == null) {
	    return true;
	} else {
	    return itemModel.isEnabled();
	}
    }

    /**
     * Get the label (title) associated with this menu.
     * @return the label (title) associated with this menu.
     */
    public synchronized String getLabel() {
	if (itemModel == null) {
	    return ((MenuItem)getTarget()).getLabel();
	} else {
	    return itemModel.getLabel();
	}
    }

    /**
     * Get the default font used to draw this menu.
     * @return the font used to draw this menu.
     */
    public synchronized Font getFont() {
	Font f = null;
	if (itemModel == null) {
	    f = getTarget().getFont();
	} else {
	    f = itemModel.getFont();
	}

	return (f == null) ? getToolkit().getDefaultFont() : f;
    }

    /**
     * Set the default view associated with this model.
     * @param newView The view that will be used to represent this component.
     */
    public synchronized void setDefaultView(View newView) {
	super.setDefaultView(newView);

	notifyDependents(items);
    }

    /**
     * Called when an action occurred.  For a MenuModel this won't normally
     * happen; the action is associated with a specific item.
     */
    public void actionOccurred(int modifiers) {
	/* 
	 * REMIND: must generate the appropriate event.  
	 * Probably none, since the item itself should get
	 * the actionOccurred() call instead.
	 */
    }

    /* this is used if the menu is added to another menu */
    private MenuItemModel itemModel;

    /**
     * The list of menu items. (Actually the item peers, i.e. MenuItemModel 
     * objects).
     */
    private OrderedList items;

    private void doAdd(MenuItemOperations item) {
	items.addElement(item);
	item.setOwner(this);
	notifyDependents(items);
    }    

    private void doRemove(int index) {
	MenuItemOperations item = (MenuItemOperations) items.elementAt(index);
	item.setOwner(null);

	items.removeElementAt(index);
	notifyDependents(items);
    }

    /**
     * A notification method used when a single item is changed.  Calls
     * notifyDependents with the single item in order to let the views know
     * about the change.
     */
    public void itemChanged(MenuItemOperations item) {
	notifyDependents(item);
    }
}
