/*
 * @(#)ModalContainerViewImpl.java	1.15 11/25/98
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

package sun.awt.otk;

import java.awt.*;
import java.awt.event.MouseEvent;
import sun.porting.windowsystem.WindowSystem;
import sun.porting.toolkit.ToolkitEventHandler;
import java.util.Enumeration;
import java.util.Vector;


/**
 * ModalContainerViewImpl extends ContainerViewImpl, to add in 
 * modal funcitonality which implies that all events get delivered
 * to this container.
 * It provides the ability for the views contained within it to
 * register interest for key events.  So when the following events,
 * ToolkitEventHandler.KEY_PRESSED, case ToolkitEventHandler.KEY_RELEASED,
 * and ToolkitEventHandler.KEY_TYPED are received, they are forwarded
 * to only those views that have registered themselves. All other 
 * events are delievered to those views that contain the x, y specified
 * in the event.
 *
 */
abstract public class ModalContainerViewImpl extends ContainerViewImpl {

    /**
     * This maintains a list of all the child views that
     * are interested in key events. 
     */
    Object keyListeners;


    /**
     * Creates a ModalContainerViewImpl object residing in the specified toolkit
     * with initial bounds and insets as indicated.
     *
     * @param toolkit to use.
     * @param bounds to use when creating the underlying window.
     * @param insets to use when creating the underlying window.
     */
    public ModalContainerViewImpl(ObjectToolkit toolkit, 
				  Rectangle bounds, 
				  Insets insets) {
        super(toolkit, bounds, insets);
    }

    /**
     * Creates a ModalContainerViewImpl object residing in the specified toolkit
     * with initial bounds and insets as indicated and parented to the
     * ContainerView specified.
     *
     * @param toolkit to use.
     * @param parent for underlying window.
     * @param bounds to use for underlying window.
     * @param insets to use for underlying window.
     */
    public ModalContainerViewImpl(ObjectToolkit toolkit, 
				  ContainerView parent,
				  Rectangle bounds, 
				  Insets insets) {
        super(toolkit, parent, bounds, insets);
    }

    /**
     * Create a ModalContainerViewImpl object as a view residing in the
     * specified toolkit.  Create the underlying window as a 
     * child of the given parent, with initial bounds as indicated 
     * and of the type specified.
     *
     * @param toolkit to use.
     * @param parent to use for underlying window.
     * @param bounds to use for underlying window.
     * @param insets to use for underlying window.
     * @param windowType type of window.
     */
    public ModalContainerViewImpl(ObjectToolkit toolkit, 
				  ContainerView parent,
				  Rectangle bounds, 
				  Insets insets, 
				  int windowType) {
	super(toolkit, parent,  bounds, insets, windowType);
    }


    /**
     * setVisiblle() is responsible for implementing the 
     * the modal behaviour. If setVisible() is called with 
     * show equal to true, then the underlying window does a grab all
     * so that all events will get delivered to this modal
     * container.  Conversely, if setVisible() is called with
     * a false value, the underlying window releases its grab.
     * Then its super class ContainerViewImpl's setVisible()
     * is called to map/unmap underlying window.
     *
     * @param show determines visible state.
     */
    public void setVisible(boolean show) {
	
	synchronized(this) {

	    boolean dialogVisible = isVisible();
	    
	    if ((show && dialogVisible) ||
		(!show && !dialogVisible)) {
		return;
	    }

	    sun.porting.windowsystem.Window win = getContainerWindow();
	    if (show) {
		win.grabInput(win.DELIVER_ALL);
	    } else { // hide
		win.releaseGrab();
	    }
	    super.setVisible(show);
 	}
    }


    /**
     * Adds a view that interested in being notified
     * when key events occur.
     * @param v view to add.
     */
    public void addKeyListener(View view) {
	if (keyListeners == null) {
	    keyListeners = view;
	} else {
	    Vector v = null;
	    if (! (keyListeners instanceof Vector)) {
		v = new Vector(2);
		v.addElement(keyListeners);
		keyListeners = v;
	    } else {
		v = (Vector)keyListeners;
	    }
	    v.addElement(view);
	}
    }

    /**
     * Remove a view that had previously registered
     * interest in being notified when key events occur.
     * 
     * @param v view to remove.
     */
    public void removeKeyListener(View view) {
	if (keyListeners ==  null) {
	    return;
	}
	if (keyListeners instanceof View) {
	    if (view == (View)keyListeners) {
		keyListeners = null;
	    }
	} else if (keyListeners instanceof Vector) {

	    Vector v = (Vector)keyListeners;
	    v.removeElement(view);

	    if (v.size() == 0) { 
		keyListeners = null;
	    }
	}
    }


    /*
     * Tracks the view that received the mouseDown.  Used to
     * ensure that if the mouse is released in another view,
     * then the view that had received the mouse down gets
     * notified, and it can register a release.
     */
    private View mouseDownRecipient = null;


    /**
     * When this method is invoked, this implies that this
     * modal container has done a grab and is therefore
     * receiving all events.  This method is responsible for
     * redirecting the events to the appropriate recipients.
     * <pre>
     * ToolkitEventHandler.WINDOW_DAMAGED:
     *        results in a repaint.
     * 
     * ToolkitEventHandler.KEY_PRESSED:
     * ToolkitEventHandler.KEY_RELEASED:
     * ToolkitEventHandler.KEY_TYPED:
     *        results in these events getting delivered to
     *	      all the views that have added themselves as
     *	      key listeners.
     *
     * ToolkitEventHandler.POINTER_RELEASED:
     * ToolkitEventHandler.POINTER_DRAGGED:
     *        delivered to the view that received the POINTER_PRESSED
     *        event.
     * 
     * All other events are delivered to the view that contains
     * the x, y specified in the event.
     * </pre>
     *
     * @param id The event type identifier
     * @param when A timestamp (in milliseconds) indicating when the event happened
     * @param modifiers The modifier key (SHIFT, CONTROL, ALT, META) state.
     * @param x The x coordinate associated with the event
     * @param y The y coordinate associated with the event
     * @param arg1 An optional argument whose meaning varies with the event type
     * @param arg2 An optional argument whose meaning varies with the event type
     * @see sun.porting.toolkit.ToolkitEventHandler
     */
    public void handleEvent(int id, long when, int modifiers,
		     int x, int y, int arg1, int arg2) {


	switch (id) {
	case ToolkitEventHandler.WINDOW_DAMAGED:
	    Graphics g = null;

	    try {
		g = getGraphics();
		g.clipRect(x, y, arg1, arg2);
		paintView(getModel(), g);
	    } finally {
		if (g != null) {
		    g.dispose();
		}
	    }
	    break;

        case ToolkitEventHandler.KEY_PRESSED:
        case ToolkitEventHandler.KEY_RELEASED:
        case ToolkitEventHandler.KEY_TYPED:
	{
	    if (keyListeners instanceof View) {
		((View)keyListeners).handleEvent(id, when, modifiers, x, y, arg1, arg2);
	    } else if (keyListeners instanceof Vector) {
		Vector v = (Vector)keyListeners;
		int cnt = v.size();
		while (--cnt >= 0) {
		    View view = (View)v.elementAt(cnt);
		    view.handleEvent(id, when, modifiers, x, y, arg1, arg2);
		}
	    }
	    break;
	}

	case ToolkitEventHandler.POINTER_RELEASED:
	case ToolkitEventHandler.POINTER_DRAGGED:
	    {
		if (mouseDownRecipient == null) {
		    break;
		}

		Rectangle bounds = mouseDownRecipient.getBounds(false);
		mouseDownRecipient.handleEvent(id, when, modifiers,
					       x - bounds.x, y - bounds.y,
					       arg1, arg2);
		if (id == ToolkitEventHandler.POINTER_RELEASED) {
		    mouseDownRecipient = null;
		}
	    }
	    break;

        default:
	    for (Enumeration childViews = getViews(); childViews.hasMoreElements();) {
		View v = (View) childViews.nextElement();
		Rectangle bounds = v.getBounds(false);
		if (bounds.contains(x, y)) {
		    if (id == ToolkitEventHandler.POINTER_PRESSED) {
			mouseDownRecipient = v;
		    }
		    v.handleEvent(id, when, modifiers, x - bounds.x, y - bounds.y, arg1, arg2);
		}
	    }
	}
    }

    /**
     * clears the contents of the window.
     * @param m The model which generated the update.  Always
     *          null since these are lightweight views.
     * @param g The Graphics object that should be used for painting
     */
    protected void paintView(Model m, Graphics g) {
	Rectangle myBounds = getBounds();
	g.clearRect(0, 0, myBounds.width, myBounds.height);
    }
}
