/*
 * @(#)MouseGobbler.java	1.11 98/11/30 SMI
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package sun.awt.otk;

import java.awt.AWTEvent;
import java.awt.event.MouseEvent;
import sun.porting.toolkit.ToolkitEventHandler;

/**
 * A helper for supporting both "drag" mode and "click" mode for
 * a menu.  This controller should be placed first in the chain, 
 * as it is designed to pre-empt events.
 *
 * Example: a typical desktop implementation of a menu GUI will behave
 * in two different ways depending on the user action that caused it to
 * be shown.
 *
 * <ul>
 * <li> if the mouse is released immediately after the initial click: 
 * <ul>
 *   <li> the initial up event is ignored
 *   <li> a mouse down which occurs outside the menu causes it to dismiss
 * </ul>
 * <li> if the mouse is held down and/or dragged after the initial click:
 * <ul>
 *   <li> a mouse up outside the menu causes it to dismiss
 * </ul>
 * </ul>
 *
 * In either case, the rest of the menu's behavior is identical.  In a
 * typical desktop the behavior resembles that of Armable, e.g. a mouse
 * up inside the menu triggers the corresponding action, and individual
 * elements are armed/disarmed based on the drag location.
 *
 * In order to implement this, the controller has to decide--based on the
 * speed of the release and whether a drag has been seen--whether it
 * needs to consume the first mouse release.  The gobbler only needs to
 * concern itself with whether to eat the initial mouse up; it is safe
 * for the mouse controller to treat any press/release that is outside
 * the menu as a dismissal event and to treat a press that is inside the
 * menu identically to a drag.  
 */
public class MouseGobbler extends InputController {
    /**
     * The timeout used to decide which mode the menu is being used in.
     * When a mouse press is seen
     * and the timeout elapses before a release, the menu is assumed to be
     * in click-drag-release mode.  If instead the initial release is seen 
     * quickly, another press and release will be required to cause an action.
     * A timeout < 0 means that dragging is disabled.
     */
    protected long timeout = 1000; // milliseconds

    /**
     * The time at which the menu was shown.  Event timestamps are compared
     * to this value to determine whether the timeout has elapsed.
     */
    protected long menuShownAt = 0;

    /**
     * Internal state variable describing the presumed operation mode
     * of the menu.  True means that drags are being
     * followed and the action will be issued on the release.  False means
     * that a mouse press is being watched for.
     */
    protected boolean dragMode = false;

    /**
     * Internal state variable, set to true when a mouse press was observed.
     */
    protected boolean sawPress = false;

    /**
     * Reset the state.  After this call completes, the controller will 
     * not be gobbling events; it will be waiting for a mouse press.
     * @param when The time at which the menu was shown.
     */
    protected synchronized void resetState(long when) {
	menuShownAt = when;
	dragMode    = false;
	sawPress    = false;
    }

    /**
     * Create a new MouseGobbler with the default timeout interval.
     */
    public MouseGobbler() {
    }

    /**
     * Create a new MouseGobbler with the specified timeout interval.
     * The timeout is used to decide whether the menu is being used in
     * click-drag-release mode, e.g. on a desktop.  
     * @see MouseGobbler#timeout
     * @param timeoutInterval the timeout used to decide which usage mode
     * the menu is in.
     */
    public MouseGobbler(long timeoutInterval) {
	this.timeout = timeoutInterval;
    }

    /**
     * Handle an AWT event that landed on the given view.
     * MenuShowEvent is directly handled here, causing the state to be
     * reset; other events are passed to super.
     * @param v The view which received the event
     * @param e The event that occurred
     */
    public synchronized boolean handleEvent(View v, AWTEvent e) {
	if (e instanceof MenuShowEvent) {
	    MenuShowEvent mse = (MenuShowEvent) e;

	    resetState(mse.getWhen());
	    return true;
	}

	return super.handleEvent(v, e);
    }

    /**
     * Handle an event which landed on the given view.
     * POINTER_PRESSED, POINTER_RELEASED and POINTER_DRAGGED
     * are handled here; other events are ignored.  The only
     * event consumed is POINTER_RELEASED, and then only if a press
     * was seen and the drag timeout has not expired.
     * @param v The View that received the event
     * @param id The type of the event
     * @param when The time (milliseconds since 1/1/70) when the event occurred
     * @param modifiers The modifier state at the time of the event
     * @param x The x coordinate associated with the event
     * @param y The y coordinate associated with the event
     * @param arg1 Optional data for events, ignored by this controller
     * @param arg2 Optional data for events, ignored by this controller
     */
    public synchronized boolean handleEvent(View v,
					    int id, long when, int modifiers,
					    int x, int y, int arg1, int arg2) {

	if (dragMode || sawPress) {
	    return false;
	}

	if ((timeout > 0) && ((when - menuShownAt) > timeout)) {
	    dragMode = true;
	    return false;
	}

	switch (id) {
	case ToolkitEventHandler.POINTER_RELEASED:
	    return true;	// dragMode and sawPress are known
				// to be false because they were
				// checked at the top of this method.

	case ToolkitEventHandler.POINTER_PRESSED:
	    sawPress = true;
	    return false;

	case ToolkitEventHandler.POINTER_DRAGGED:
	    /* timeout < 0 means always gobble the first      */
	    /* release event, so dragMode becomes irrelevant. */
	    if (timeout > 0) {
		dragMode = true;
	    }
	    return false;
	}


	return false;
    }

    /**
     * @return false always, because this controller does not affect
     * keyboard input in any way.
     */
    public boolean isFocusTraversable() {
	return false;
    }
}
