/*
 * @(#)MultiLineTextPane.java	1.40 98/12/18
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

package sun.awt.otk;

import java.awt.*;
import java.awt.event.*;
import sun.porting.toolkit.ToolkitEventHandler;

/**
 * This is an implementation of a multiline textpane. See 
 * <a href="TextComponentModel.html">TextComponentModel</a> for a high
 * level description of how textpanes are used. It is assumed that a
 * TextComponentView will provide a window to host this textpane.
 * <p>
 * A MultiLineTextPane also adds the concept
 * if partial scrolling, which provides a half line clipped text window
 * on top and bottom of the textpane as feedback of whether text exists
 * beyond what is visible in that direction.
 * @see sun.awt.otk.TextComponentModel
 * @see sun.awt.otk.TextComponentView
 */
public abstract class MultiLineTextPane
extends TextPaneImpl {

    /**
     * This tells us whether we're scrolling at line or halfline granularity.
     * If this is set to off then we revert to full line scrolling only.
     */
    protected boolean partialScrolling;

    /** Whether to show top half line or not */
    protected boolean showTopHalfLine;
    /** Whether to show bottom half line or not */
    protected boolean showBottomHalfLine;

    /** Bounds of top half line */
    protected Rectangle topHalfRect;
    /** Bounds of bottom half line */
    protected Rectangle bottomHalfRect;

    /**
     * Create a new MultiLineTextPane instance.
     * @param view Our view
     * @param model Our data model
     * @param tk Our toolkit
     * @param colors Our color policy
     * @param insets The insets used for this textpane. These define the
     * outer boundaries that we use for clipping before drawing text or
     * selections.
     * @param left Left padding used in addition to insets before drawing text
     * @param right Rigth pad used along with insets before drawing text
     */
    public MultiLineTextPane(BasicView view, TextComponentModel model,
			     ObjectToolkit tk, ColorPolicy colors,
			     Insets insets, int left, int right) {

        super(view, model, tk, colors, insets, left, right);
    }

    /**
     * Set the bounds of the textpane. Sets textRect, topHalfRect and
     * bottomHalfRect to the correct insets-adjusted value.
     * @param x New x-coordinate
     * @param y New y-coordinate
     * @param width New width
     * @param height New height
     */
    public void setBounds(int x, int y, int width, int height) {

        //System.err.println("+MultiLinePane.setBounds: x=" + x + " y=" + y + " w=" + width + " h=" + height + " insets=" + insets + " lPad=" + leftPad + " rPad=" + rightPad);

	super.setBounds(x, y, width, height);

	if (partialScrolling && width > 0 && height > 0) {
	    
	    Graphics g = view.getGraphics();
	    FontMetrics fm = g.getFontMetrics();
	    g.dispose();

	    // figure out how much dead space to leave on top based on the
	    // maximum number of "full" lines that will fit in given height
	    int lh = getLineHeight(fm);
	    int deadSpace = 0;
	    int nominalHeight = height - (insets.top + insets.bottom);
	    if (nominalHeight % lh != 0) {
	        deadSpace = (height -
			     ((nominalHeight / lh) * lh)
			    ) / 2;
	    }

	    int halfHt = getScrollIncrement(fm);
	    int w = width - insets.left - insets.right;
	    int s = x + insets.left;
	    topHalfRect = new Rectangle(s,
					y + Math.max(deadSpace, insets.top),
					w, halfHt);
	    bottomHalfRect = new Rectangle(s, y + height - halfHt -
					   Math.max(deadSpace, insets.bottom),
					   w, halfHt);
	    textRect = new Rectangle(s,
				     topHalfRect.y + topHalfRect.height,
				     w,
				     bottomHalfRect.y -
				         (topHalfRect.y + topHalfRect.height));
	} else {
	    topHalfRect = new Rectangle(0, 0, 0, 0);
	    bottomHalfRect = new Rectangle(0, 0, 0, 0);
	}

	// set the data model width
        data.setLineWidth(getMaxPossibleStringWidth());

	resizeScrollbars();

	//System.err.println("-MultiLinePane.setBounds: textRect=" + textRect + " topR=" + topHalfRect + " bottR=" + bottomHalfRect);
    }

    /**
     * Paint lines in the given range (inclusive).
     * @param g Current graphics context
     * @param s Starting line
     * @param e Ending line
     */
    public void paintLines(Graphics g, int s, int e) {

	int nrOfLines = data.getNumberOfLines();
        int li = linesInWindow();
        int lastVisLine = li + topLine - 1;
	int lastLine = nrOfLines - 1;

        s = Math.max(topLine, s);
	if (lastLine >= lastVisLine) {
	    e = Math.min(e, lastVisLine);
	} else {
	    e = Math.min(e, lastLine);
	}

	/*System.err.println("-----\npaintLines: s=" + s + " e=" + e +
			   " top=" + topLine + " last=" + lastLine + 
			   " lastVis=" + lastVisLine);*/
	FontMetrics fm = g.getFontMetrics();
	Font f = g.getFont();
	Color fg = getFgColor();
	Color bg = getBgColor();

	// paint lines with some text on them
        for (int i = s ; i <= e && i < nrOfLines; i++) {
	    if (isLineVisible(i)) {
	        paintLine(g, fm, f, fg, bg,
			  ln2y(fm, i), ln2baseline(fm, i),
			  selected && cursorOnThisLine(i, selStart, selEnd),
			  data.getLineStart(i),
			  data.getLineAsStringFromNumber(i));
	    }
        }

	if (lastLine < lastVisLine) {
	    while (lastLine++ < lastVisLine) {
	        //System.err.println(" eraseLine: " + lastLine);
	        paintLine(g, fm, f, fg, bg,
			  ln2y(fm, lastLine), ln2baseline(fm, lastLine),
			  false, 0, "");
	    }
	}
    }

    /**
     * Handle all mouse events for the textpane.
     * @param id Event type
     * @param when Time of event
     * @param modifiers Flags
     * @param x X coordinate
     * @param y Y coordinate
     * @param arg1 Click count
     * @param arg2 isPopupTrigger
     * @return True if the event was handled, false otherwise.
     */
    public boolean handleMouseEvent(int id,  long when, int modifiers,
				    int x, int y, int arg1, int arg2) {

        boolean handled = false;

        switch (id) {

        case ToolkitEventHandler.POINTER_PRESSED:

	    // We don't allow changing the insertion point when composing
	    if (composeMode) {
	        ((TextComponentView) view).playNotAllowedSound();
		handled = true;
		break;
	    }

	    if (textRect.contains(x, y)) {
	        // default case in super will do the right thing - let it go
		handled = false;
	    } else if (arg1 > 0 && topHalfRect.contains(x, y)) {
	        if (topLine > 0) {
		    int posn = xPos(x, topLine - 1);
		    positionCursor(posn);
		    scrollTo(selStart, true);
		}
		handled = true;
	    } else if (arg1 > 0 && bottomHalfRect.contains(x, y)) {
	        int vis = getLastVisibleLine();
	        if (vis < data.getNumberOfLines() - 1) {
		    int posn = xPos(x, vis + 1);
		    positionCursor(posn);
		    scrollTo(selStart, true);
		} 
		handled = true;
	    }
            break;

        case ToolkitEventHandler.POINTER_RELEASED:
	    // do nothing - left for future
            break;

        case ToolkitEventHandler.POINTER_DRAGGED:
	    if (textRect.contains(x, y)) {
	        int mpos2 = xy2pos(x, y);
		if (mpos < mpos2) {
		    select(mpos, mpos2);
		} else {
		    select(mpos2, mpos);
		}
		scrollTo(mpos2, true);
		handled = true;
	    }
            break;
	    
	default:
	    break;
        }

	if (!handled) {
	    handled = super.handleMouseEvent(id, when, modifiers,
					     x, y, arg1, arg2);
	}

	return handled;
    }
  
    /**
     * Draw the string to the screen.
     * @param g Graphics context
     * @param line line number to be rendered
     */
    protected void paintLine(Graphics g, int line) {

        if (partialScrolling) {
	    // Are we the top half line?
	    if (line == topLine - 1) {
	        paintTopHalfLine(line);
		return;
	    } else if (line == topLine + linesInWindow()) {
	        paintBottomHalfLine(line);
		return;
	    }
	}

        super.paintLine(g, line);
    }

    /**
     * Returns the current scroll increment.
     */
    protected int getScrollIncrement(FontMetrics fm) {
        return getLineHeight(fm);
    }

    /**
     * Paint lines that have changed in the given range. We make sure
     * that only lines in the visible range are painted.
     * @param start Start line number
     * @param end End line number
     */
    protected void paintChangedLines(int start, int end) {

	int nrOfLines = data.getNumberOfLines();
	int li = linesInWindow();
	    
	// make sure topLine is still valid
	if (topLine > nrOfLines - 1) {
	    // goto last screenfull
	    topLine = Math.max(0, nrOfLines - li);
	}

        int topPos = data.getLineStart(topLine);
	int bottomPos = data.getLineEnd(Math.min(topLine + li - 1,
						 nrOfLines - 1));

	if (selStart >= topPos && selStart <= bottomPos) {
	    // caret is in the visible area - do nothing

	    paintLines(start, end);
	    updateHalfLines();
	    resizeScrollbars();

	} else {                                  
	    // caret is not in the visible area - make edited line the bottom
	    paintLines(start, end);
	    int ln = Math.max(0, data.getLineNumber(selStart) - li + 1);
	    scrollAbsolute(data.getLineStart(Math.min(ln, nrOfLines - 1)), true);
	}
    }

    /**
     * Tell the peer to resize scrollbars.
     */
    protected void resizeScrollbars() {
        int nrOfLines = data.getNumberOfLines();
	int vis = Math.min(linesInWindow(), nrOfLines - topLine);
        ((TextPaneScroller) view).resizeScrollbars(
						   // vertical scrollbar values
						   topLine,
						   //data.getLineNumber(selStart),
						   vis, 0, nrOfLines,
						   // horiz. scrollbar values
						   0, 0, 0, 0);
    }

    /**
     * Scrolls vertically if necessary.
     * @param g Current graphics context
     * @param pos Index into the text
     */
    protected void handleVerticalScrolling(Graphics g, int pos) {

        int lnr = data.getLineNumber(pos);
        int li = linesInWindow();
	if (lnr >= li + topLine) {
	    if (selStart < selEnd) {       // we're in selection mode
	        scrollVertical(g, 1);
	    } else  {
	        //int lastLine = (topLine + l) - 1;
		scrollVertical(g, lnr - ((topLine + li) - 1));
	    }
	} else if (lnr < topLine) {
	    if (selStart < selEnd) {       // we're in selection mode
	        scrollVertical(g, -1);
	    } else {
	        scrollVertical(g, lnr - topLine);
	    }
	}

	updateHalfLines();
    }

    /**
     * Scroll the text "ml" times vertical.
     * @param g Current graphics context
     * @param ml Number of lines to move forward (positive value) or
     * backwards (negative values)
     */
    protected void scrollVertical(Graphics g, int ml) {

        //System.err.println("+scrollVertical: ml=" + ml);
        int nrOfLines = data.getNumberOfLines();
        int startLine, endLine;
        FontMetrics fm = g.getFontMetrics();
	int xmargin = textRect.x + leftPad;
	int w = getMaxPossibleStringWidth();

        if (ml == 0) {
            return;
        }

        if (topLine < -ml) {
            ml = -topLine;
        }

	int minTextH = getLineHeight(fm);
	int l = linesInWindow(fm);
        if (ml < 0) {
            topLine += ml;

            if (ml + l > 0) {
                int hg = ((ml + l) * minTextH);
                g.copyArea(xmargin, textRect.y, w, hg, 0, -ml * minTextH);
             }
            startLine = topLine;
            endLine = Math.min(nrOfLines - 1, topLine + -ml);

        } else {   // ml > 0
            topLine += ml;

            if (ml < l) {
                int y = textRect.y + (ml * minTextH);
                int hg = ((l - ml) * minTextH);
                g.copyArea(xmargin, y, w, hg, 0, -ml * minTextH);
            }
            startLine =  topLine + (l - ml);
            endLine = Math.min((topLine + l) - 1, nrOfLines - 1);
        }

        // after vert. scrolling movedRight should be recalculated
        int max = maxLength(topLine, topLine+linesInWindow(fm)-1);
        if (max - movedRight < w && movedRight != 0) {
            movedRight = Math.max(0, max - w);
            startLine = topLine;
            endLine = Math.min(topLine+linesInWindow(fm)-1, nrOfLines - 1);
        }
        paintLines(g, startLine, endLine);

        //System.err.println("-scrollVertical: ml=" + ml);
    }

    /**
     * Scroll the text to the position "pos" regardless of whether it is
     * already on the screen or not.
     * @param pos Index into the text
     * @param resizeScrollbars If true then scrollbars will be resized after
     * this operation.
     */
    protected void scrollAbsolute(int pos, boolean resizeScrollbars) {

        if (!data.isValidPosition(pos)) {
	    return;
	}

	Graphics g = null;
	try {

	    g = view.getGraphics();

	    int lnr = data.getLineNumber(pos);
	    if (lnr > topLine) {
		scrollVertical(g, lnr - topLine);
	    } else if (lnr < topLine) {
	        scrollVertical(g, lnr - topLine);
	    } else {
		scrollTo(selStart, true);
	    }

	    updateHalfLines();

	} finally {
	    if (g != null) {
	        g.dispose();
	    }
	}

	if (resizeScrollbars) {
	    resizeScrollbars();
	}
    }

    /**
     * Determine if the top and bottom half lines need to be repainted, and
     * then issue the appropriate paint calls.
     */
    protected void updateHalfLines() {

        int li = linesInWindow();
	int nrOfLines = data.getNumberOfLines();

        if (partialScrolling) {
	    if (topLine == 0) {
	        if (showTopHalfLine) {
		    paintTopHalfLine(-1);
		}
		showTopHalfLine = false;
	    } else {
	        showTopHalfLine = true;
	        paintTopHalfLine(topLine - 1);
	    }

	    if (nrOfLines > topLine + li) {
	        showBottomHalfLine = true;
	        paintBottomHalfLine(topLine + li);
	    } else {
	        if (showBottomHalfLine) {
		    paintBottomHalfLine(-1);
		}
		showBottomHalfLine = false;
	    }
	}
    }

    /**
     * Paint the given line in the top half line window.
     * @param line Line number of line to paint.
     */
    protected abstract void paintTopHalfLine(int line);

    /**
     * Paint the given line in the bottom half line window.
     * @param line Line number of line to paint.
     */
    protected abstract void paintBottomHalfLine(int line);
}

