/*
 * @(#)ObjectToolkit.java	1.49 99/07/28 SMI
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package sun.awt.otk;

import java.awt.*;
import java.awt.datatransfer.*;
import java.awt.event.*;
import java.awt.image.*;
import java.awt.peer.*;
import java.util.Properties;

import sun.awt.SunToolkit;
import sun.porting.toolkit.ToolkitEventHandler;
import sun.porting.windowsystem.WindowSystem;
import sun.porting.graphicssystem.CursorImage;

/**
 * ObjectToolkit is an abstract base class that helps the implementation
 * of personal java window toolkits with different look and feels.
 * This toolkit also implements the ToolkitEventHandler interface. It
 * is through this interface that events are delivered from the 
 * underlying window system to the toolkit.
 */
public abstract class ObjectToolkit extends SunToolkit implements ToolkitEventHandler {
    protected WindowSystem winSys;

    /* REMIND: make these configurable! */
    protected long MULTI_CLICK_THRESHOLD = 400; // milliseconds
    protected int  SMUDGE_FACTOR = 4;	        // pixels

    /**
     * Creates an ObjectToolkit, and registers itself with
     * the underlying window system as a ToolkitEventHandler.
     */
    protected ObjectToolkit() {
	winSys = WindowSystem.getDefaultWindowSystem();
	winSys.start();
	winSys.registerToolkitEventHandler(this);
    }

    private int clickCount = 0;
    private int modifierState = 0;
    private long lastPress = 0;
    private sun.porting.windowsystem.Window lastPressWin = null;
    private int lastPressX = 0;
    private int lastPressY = 0;

    /**
     * Returns the default font.
     */
    public abstract Font getDefaultFont();

    /**
     * Returns the default foreground color.
     */
    public abstract Color getDefaultForeground();

    /**
     * Returns the input method adapter to use. The
     * VirtualKeyboardInputMethod is used as the 
     * default input method adapter.  
     * @see sun.awt.otk.VirtualKeyboardInputMethod
     */
    public sun.awt.im.InputMethod getInputMethodAdapter() {
	return new VirtualKeyboardInputMethod();
    }

    private static String imStr = null;

    /**
     * Returns the input method adapter associated with the
     * sun.awt.im.InputMethod property.  If the property
     * value is default, then the SampleInputMethod in the
     * otk package is returned.
     */
    public sun.awt.im.InputMethod getRealInputMethodAdapter() {
        sun.awt.im.InputMethod im = null;
        try {
  	    if (imStr == null) {
	        imStr = System.getProperty("sun.awt.im.InputMethod");
	    }
	    if (imStr != null) {

	        if (imStr.equals("default")) {
		    imStr = "sun.awt.otk.SampleInputMethod";
		}

	        im = (sun.awt.im.InputMethod) Class.forName(imStr).newInstance();
	    }
	} catch (SecurityException e) {
	    System.err.println("Couldn't access system properties while creating InputMethod");
	} catch (ClassNotFoundException e) {
	    System.err.println("Class not found while creating InputMethod: " +
			       imStr);
	} catch (IllegalAccessException e) {
	    System.err.println("IllegalAccessException while creating InputMethod: " + imStr);
	} catch (InstantiationException e) {
	    System.err.println("InstantiationException while creating InputMethod: " + imStr);
	} catch (ClassCastException e) {
	    System.err.println("ClassCastException while creating InputMethod. It must implement sun.awt.im.InputMethod.");
	}

	return im;
    }

    /**
     * Returns whether a virtual keyboard is required or not.
     * This is determined by the value of the sun.awt.im.NoVirtualKeyboard
     * property.
     */
    public boolean isVirtualKeyboardRequired() {

	String nvk = System.getProperty("sun.awt.im.NoVirtualKeyboard");
	if ((nvk != null) && nvk.equalsIgnoreCase("true")) {
	    return false;
	}
	return true;

    }


    /**
     * Returns a VirtualKeyboardActivator object.
     * @see sun.awt.otk.VirtualKeyboardActivator
     */
    protected abstract VirtualKeyboardActivator getVirtualKeyboardActivator();

    /**
     * Returns a VirtualKeyboardViewIntf object.
     * @param inputMethod is the VirtualKeyboardInputMethod that controls the
     *                    activation / deactivation of the virtual keyboard.
     * @param textModel   is the text model that is associated with the component
     *                    that the virtual keyboard needs to be linked to.
     */
    protected abstract VirtualKeyboardViewIntf getVirtualKeyboard(VirtualKeyboardInputMethod inputMethod, 
								  TextComponentModel textModel);

    /**
     * Creates a VirtualKeybaordModel.
     */
    protected VirtualKeyboardModel getVirtualKeyboardModel(VirtualKeyboardInputMethod inputMethod) {
	return new VirtualKeyboardModel(inputMethod);
    }

    /**
     * This method returnds whether or not input methods should be enabled for
     * text components.
     * By default, this value is always true.
     */
    public boolean enableInputMethodsForTextComponent() {
        return true;
    }

    /**
     * Returns the window that is currently grabbing all events.
     */
    public sun.porting.windowsystem.Window getGrabbingWindow() {
	return winSys.getGrabbingWindow();
    }

    /**
     * Returns the window that currently has focus.
     * Returns null in the case where no owner can
     * be determined or an exception occurs
     */
    public Component getFocusOwner(){
	Component target = null;
	try {
	    sun.porting.windowsystem.Window focusWindow =
		winSys.getFocusWindow();
	    View v = (View) focusWindow.getUserData();
	    ComponentModel cm = (ComponentModel)v.getModel();
	    if (cm.isDefaultView(v)) {
		target = cm.getTarget();
	    }

	    /* target is now the peered component that has the focus, if */
	    /* any.  But if this is a container, the actual focus owner  */
	    /* could be a lightweight.  So we handle that possibility:   */
	    if (target instanceof Container) {
		while (! (target instanceof java.awt.Window)) {
		    target = target.getParent();
		}
		target = ((java.awt.Window)target).getFocusOwner();
	    }

	    return target;

	} catch (Throwable t) {
	    return null;
	}
    }


    /**
     * This method gets called in response to pointer/mouse events that occur in 
     * the underlying window system.  It translates a ToolkitEventHandler event
     * like POINTER_PRESSED/POINTER_RELEASED, to java.awt.MouseEvent and feeds
     * the event to the associated view via the handleEvent() method.  The 
     * view is associated with the window in which the event occured by storing
     * it as the window's user data.
     *
     * @see sun.porting.toolkit.ToolkitEventHandler
     */
    public void pointerEventOccurred(sun.porting.windowsystem.Window win, 
				     long when,
				     int x, int y, 
				     int ID, int number) {

	boolean clicked = (win == lastPressWin);
	int mask = 0;
	if (number == 1) {
	    mask = InputEvent.BUTTON1_MASK;
	} else if (number == 2) {
	    mask = InputEvent.BUTTON2_MASK;
	} else if (number == 3) {
	    mask = InputEvent.BUTTON3_MASK;
	}

	if (ID == POINTER_PRESSED) {
	    modifierState |= mask;

	    if ( ((when - lastPress) > MULTI_CLICK_THRESHOLD) || (!clicked)) {
		clickCount = 0;
	    }
	    ++clickCount;
	    lastPress = when;
	    lastPressWin = win;
	    lastPressX = x;
	    lastPressY = y;
	} else if (ID != POINTER_RELEASED) {
	    if (    (Math.abs(x - lastPressX) > SMUDGE_FACTOR)
	         || (Math.abs(y - lastPressY) > SMUDGE_FACTOR)  ) {
		clickCount = 0;
	    }
	}

	View tw = (View)win.getUserData();

	if (tw != null) {
	    try {
		ComponentModel cm = (ComponentModel)tw.getModel();

		if (cm == null || !cm.isDefaultView(tw)) {
		    throw new ClassCastException("not really a component");
		}

		if (!cm.isEnabled()) {
		    return;
		}

		Component target = cm.getTarget();
		MouseEvent ev = null;
		switch (ID) {
		case POINTER_PRESSED:
		    ev = new MouseEvent(target, MouseEvent.MOUSE_PRESSED,
					when, modifierState, x, y, 
					clickCount, false);
		    break;

		case POINTER_RELEASED:
		    ev = new MouseEvent(target, MouseEvent.MOUSE_RELEASED,
					when, modifierState, x, y, 
					clickCount, false);
		    if (clicked && (clickCount > 0)) {
			postEvent(ev);
			ev = new MouseEvent(target, MouseEvent.MOUSE_CLICKED,
					    when, modifierState, x, y, 
					    clickCount, false);
		    }

		    break;

		case POINTER_DRAGGED:
		    ev = new MouseEvent(target, MouseEvent.MOUSE_DRAGGED,
					when, modifierState, x, y, 
					clickCount, false);
		    break;

		case POINTER_MOVED:
		    ev = new MouseEvent(target, MouseEvent.MOUSE_MOVED,
					when, modifierState, x, y, 
					clickCount, false);
		    break;
		}

		if (ev != null) {
		    postEvent(ev);
		}
	    } catch (ClassCastException notAComponentModel) {
		tw.handleEvent(ID, when, modifierState, x, y, clickCount, 0);
		if (clicked && (clickCount > 0) && (ID == POINTER_PRESSED)) {
		    /*
		     * See MouseController.java; the way a MOUSE_CLICKED 
		     * event is signalled is by negating the click count.
		     * Doing this here allows a non-component to process
		     * double-click events correctly--and possibly 
		     * separately--from ordinary mouse press events.
		     */
		    tw.handleEvent(ID, when, modifierState, x, y, 
				   -clickCount, 0);
		}
	    }
	}

	/* must deliver event before clearing state */
	if (ID == POINTER_RELEASED) {
	    modifierState &= ~mask;
	}
    }


    /**
     * This method gets called in response to key events that occur in 
     * the underlying window system.  It translates a ToolkitEventHandler event
     * like KEY_PRESSED/KEY_RELEASED, to java.awt.KeyEvent and then feeds the
     * event to the appropriate view via the handleEvent() method. The appropriate 
     * view is stored in the window as part of its user data.
     *
     * @see sun.porting.toolkit.ToolkitEventHandler
     */
    public void keyboardEventOccurred(sun.porting.windowsystem.Window win,
				      long when,
				      int ID, int keycode, char keychar) {
	int mask = 0;
	if (keycode == KeyEvent.VK_META) {
	    mask = InputEvent.META_MASK;
	} else if (keycode == KeyEvent.VK_CONTROL) {
	    mask = InputEvent.CTRL_MASK;
	} else if (keycode == KeyEvent.VK_ALT) {
	    mask = InputEvent.ALT_MASK;
	} else if (keycode == KeyEvent.VK_SHIFT) {
	    mask = InputEvent.SHIFT_MASK;
	}

	if (ID == KEY_PRESSED) {
	    modifierState |= mask;
	} else if (ID == KEY_RELEASED) {
	    modifierState &= ~mask;
	}

	if (win == null) {	// no focus window
	    return;
	}

	View tw = (View)win.getUserData();
	if (tw != null) {
	    try {
		ComponentModel cm = (ComponentModel)tw.getModel();
		if (cm == null || !cm.isDefaultView(tw)) {
		    throw new ClassCastException("not really a component");
		}

		Component target = cm.getTarget();
		KeyEvent ev = null;

		switch (ID) {
		case KEY_PRESSED:
		    ev = new KeyEvent(target, KeyEvent.KEY_PRESSED,
				      when, modifierState,
				      keycode, keychar);
		    break;

		case KEY_RELEASED:
		    ev = new KeyEvent(target, KeyEvent.KEY_RELEASED,
				      when, modifierState,
				      keycode, keychar);
		    break;

		case KEY_TYPED:
		    /* According to KeyEvent.java, don't send typed */
		    /* event for action keys.                       */
		    if (isActionKey(keycode)) return;

		    ev = new KeyEvent(target, KeyEvent.KEY_TYPED,
				      when, modifierState,
				      keycode, keychar);
		    break;

		default:
		    System.err.println("keyboardEventOccurred: id " +ID+
				       " on " +tw);
		    System.err.println("   keycode = " +keycode+
				       ", keychar = "  +((int)keychar)); 
		    System.err.println();
		}

		if (ev != null) {
		    postEvent(ev);
		}
	    } catch (ClassCastException notAComponentModel) {
		tw.handleEvent(ID, when, modifierState, 0, 0, keycode, keychar);
	    }
	}
    }

    /**
     * This method gets called in response to window events that occur in 
     * the underlying window system.  It feeds the events to the associated
     * view via the handleEvent() method.  The appropriate view is stored 
     * in the window as part of its user data.
     *
     * @see sun.porting.toolkit.ToolkitEventHandler
     */
    public void windowEventOccurred(sun.porting.windowsystem.Window win,
				    long when,
				    int x, int y, int w, int h, 
				    int ID) {
	View tw = (View)win.getUserData();
	if (tw != null) {
	    tw.handleEvent(ID, when, modifierState, x, y, w, h);
	}
    }

    /**
     * Returns the cursor image to use given a cursor object.
     * @param Cursor to use.
     * @return CursorImage to use.
     */
    public CursorImage getCursorImage(Cursor c) {
	return winSys.getCursorImage(c);
    }

    /**
     * Get an implementation of a typeface with the given name and
     * style.  Style values are as in <code>java.awt.Font</code>.
     * @param name The name of the typeface
     * @param style The style (<code>PLAIN</code>, <code>BOLD</code>, <code>ITALIC</code>) of the font.
     * @return A <code>FontPeer</code> for the nearest matching font.
     * @see java.awt.Font
     */
    protected java.awt.peer.FontPeer getFontPeer(String name, int style) {
	return winSys.getFont(name, style);
    }

    /**
     * Return the resolution of the screen, in pixels per inch.
     */
    public int getScreenResolution() {
	return winSys.getScreenResolution();
    }

    /**
     * Get the color model of the screen.
     * @return The color model, as a java.awt.image.ColorModel object
     */
    public ColorModel getColorModel() {
	return winSys.getScreenColorModel();
    }

    /**
     * Get a valid <code>FontMetrics</code> object for the given font.
     * @param font The java font descriptor for the font
     * @return The corresponding <code>FontMetrics</code> object.
     * @see java.awt.Font
     */
    public FontMetrics getFontMetrics(Font font) {
	return winSys.getFontMetrics(font);
    }

    /**
     * Synchronizes the window system.  Which essentially implies
     * that iti synchornizes the graphics state.  Since for example,
     * some systems may do buffering of graphics events, this ensures
     * that the display is up-to-date.
     */
    public void sync() {
	winSys.sync();
    }

   /**
    * Indicates the construction status of a specified image that is
    * being prepared for display.
    * <p>
    * @param     image      the image for which to prepare a  
    *                           screen representation.
    * @param     width      the width of the desired screen 
    *                           representation, or <code>-1</code>.
    * @param     height     the height of the desired screen 
    *                           representation, or <code>-1</code>.
    * @param     observer   the <code>ImageObserver</code> 
    *                           object to be notified as the 
    *                           image is being prepared.
    * @return    <code>true</code> if the image has already been 
    *                 fully prepared; <code>false</code> otherwise.
    *
    * @see sun.porting.windowsystem.WindowSystem
    */
    public int checkImage(Image img, int w, int h, ImageObserver o) {
	return winSys.checkScrImage(img, w, h, o);
    }

    /**
     * Prepares an image for rendering. 
     * <p>
     * @param     image      the image for which to prepare a  
     *                           screen representation.
     * @param     width      the width of the desired screen 
     *                           representation, or <code>-1</code>.
     * @param     height     the height of the desired screen 
     *                           representation, or <code>-1</code>.
     * @param     observer   the <code>ImageObserver</code> 
     *                           object to be notified as the 
     *                           image is being prepared.
     * @return    <code>true</code> if the image has already been 
     *                 fully prepared; <code>false</code> otherwise.
     *
     * @see sun.porting.windowsystem.WindowSystem
     *
     */
    public boolean prepareImage(Image img, int w, int h, ImageObserver o) {
	return winSys.prepareScrImage(img, w, h, o);
    }

    /**
     * Get an implementation of an image that corresponds to the
     * given <code>ImageRepresentation</code> object.
     * @param image The <code>ImageRepresentation</code> object describing this image
     * @return An image, initialized from the given <code>ImageRepresentation</code>.
     *
     * @see sun.porting.windowsystem.WindowSystem
     */
    public Image createImage(ImageProducer producer) {
	return winSys.getImage(producer);
    }

    /**
     * Get an <code>Image</code> object for use as an offscreen
     * drawing area.  The object should have the specified size
     * and color model.
     * @param w The width of the offscreen image
     * @param h The height of the offscreen image
     * @return An offscreen image into which graphics can be drawn.
     */
    public Image createImage(Component target, int width, int height) {
	return (Image) winSys.makeDrawableImage(target, width, height);
    }

    /**
     * Returns a java.awt.PrintJob object.
     * Printing is no supported.  Hence the method
     * throws a RuntimeException.
     */
    public PrintJob getPrintJob(Frame frame, 
				String jobtitle, Properties props) {

	throw new UnsupportedOperationException("Printing is not supported");
    }

    /**
     * Emits an audio beep.
     */
    public void beep() {
	winSys.beep();
    }

    protected static Clipboard clipboard = null;

    /**
     * Returns a clipboard.  Constructs one if it
     * doesnt exist.
     */
    protected static Clipboard getClipboard() {
        if (clipboard == null) {
	    clipboard = new Clipboard("System");
	}
	return clipboard;
    }


    public Clipboard getSystemClipboard() {
        SecurityManager security = System.getSecurityManager();
        if (security != null) {
	  security.checkSystemClipboardAccess();
	}
	return getClipboard();
    }

    /**
     * Posts an AWTEvent in the system event queue.
     */
    protected static void postEvent(AWTEvent ev) {
	theEventQueue.postEvent(ev);
    }

    /**
     * Returns the dimensions of the screen.
     */
    public Dimension getScreenSize() {
	Rectangle r = winSys.getRootWindow().getBounds(true);
	return new Dimension(r.width, r.height);
    }

    /**
     * Returns the width of the screen.
     */
    protected int getScreenWidth() {
	Rectangle r = winSys.getRootWindow().getBounds(true);
	return r.width;
    }

    /**
     * Returns the height of the screen.
     */
    protected int getScreenHeight() {
	Rectangle r = winSys.getRootWindow().getBounds(true);
	return r.height;
    }

    /**
     * dispose the peer assoicated with the target
     */
    protected abstract void disposedPeer(Object target, Object peer);

    /**
     * Remove the assoicated between the target and the peer.
     */
    protected static void targetDisposedPeer(Object target, Object peer) {
	SunToolkit.targetDisposedPeer(target, peer);
    }

    /*
     * Fetch the peer associated with the given target (as specified
     * in the peer creation method).  This can be used to determine 
     * things like what the parent peer is.  If the target is null
     * or the target can't be found (either because the a peer was
     * never created for it or the peer was disposed), a null will
     * be returned.
     */
    protected static Object targetToPeer(Object target) {
	return SunToolkit.targetToPeer(target);
    }

    /**
     * Determine if given keyCode is an action key. NOTE: This code must
     * stay in sync with KeyEvent.isActionKey.
     * @param keyCode Key code to check
     * @see java.awt.event.KeyEvent
     */
    protected static boolean isActionKey(int keyCode) {
        switch (keyCode) {
	  case KeyEvent.VK_HOME:
          case KeyEvent.VK_END:
          case KeyEvent.VK_PAGE_UP:
          case KeyEvent.VK_PAGE_DOWN:
          case KeyEvent.VK_UP:
          case KeyEvent.VK_DOWN:
          case KeyEvent.VK_LEFT:
          case KeyEvent.VK_RIGHT:
          case KeyEvent.VK_F1:
          case KeyEvent.VK_F2:
          case KeyEvent.VK_F3:
          case KeyEvent.VK_F4:
          case KeyEvent.VK_F5:
          case KeyEvent.VK_F6:
          case KeyEvent.VK_F7:
          case KeyEvent.VK_F8:
          case KeyEvent.VK_F9:
          case KeyEvent.VK_F10:
          case KeyEvent.VK_F11:
          case KeyEvent.VK_F12:
          case KeyEvent.VK_PRINTSCREEN:
          case KeyEvent.VK_SCROLL_LOCK:
          case KeyEvent.VK_CAPS_LOCK:
          case KeyEvent.VK_NUM_LOCK:
          case KeyEvent.VK_PAUSE:
          case KeyEvent.VK_INSERT:
              return true;
        }
        return false;
    }

    /**
     * Determine if given modifier has the control bits turned on. NOTE: This
     * code must stay in sync with KeyEvent.isControlDown().
     * @param modifiers Modifier field to check
     * @see java.awt.event.InputEvent
     */
    protected static boolean isControlDown(int modifiers) {
        return (modifiers & java.awt.Event.CTRL_MASK) != 0;
    }

    /* model creation */
    protected ObservableModel makeObservableModel() {
	return new ObservableModel();
    }

    /**
     * CanvasModel implements the CanvasPeer.
     */
    class CanvasModel extends ComponentModel implements CanvasPeer {
	public CanvasModel(Component target, ObjectToolkit tk) {
	    super(target, tk);
	}
    }

    /**
     * Creates a CanvasModel.  This model implements java.awt.peer.CanvasPeer.
     * @param target to associate with the peer.
     * @return the peer created.  It is returned as a ComponentModel
     *         which is the parent class of a CanvasModel.
     */
    protected ComponentModel makeCanvasModel(Component target) {
	return new CanvasModel(target, this);
    }

    /**
     * Creates a ComponentModel.  This model implements java.awt.peer.ComponentPeer.
     * @param target to associate with the peer.
     * @return the peer created.
     */
    protected ComponentModel makeComponentModel(Component target) {
	return new ComponentModel(target, this);
    }


    /**
     * Creates a ContainerModel.  This model implements java.awt.peer.ContainerPeer.
     * @param target to associate with the peer.
     * @return the peer created.  It is returned as a ComponentModel
     *         which is the parent class of a ContainerModel.
     */
    protected ComponentModel makeContainerModel(Container target) {
	return new ContainerModel(target, this);
    }


    /**
     * Creates a WindowModel.  This model implements java.awt.peer.WindowPeer.
     * @param target to associate with the peer.
     * @return the peer created.
     */
    protected ComponentModel makeWindowModel(Window target) {
	return new WindowModel(target, this);
    }


    /**
     * Creates a DialogModel.  This model implements java.awt.peer.DialogPeer.
     * @param target to associate with the peer.
     * @return the peer created.
     */
    protected ComponentModel makeDialogModel(Dialog target) {
	return new DialogModel(target, this);
    }

    /**
     * Creates a FrameModel.  This model implements java.awt.peer.FramePeer.
     * @param target to associate with the peer.
     * @return the peer created.
     */
    protected ComponentModel makeFrameModel(Frame target) {
	return new FrameModel(target, this);
    }

    /**
     * Creates a LabelledComponentModel.  This model implements java.awt.peer.ButtonPeer
     * and LabelPeer.
     * @param target to associate with the peer.
     * @return the peer created.
     */
    protected ComponentModel makeLabelledComponentModel(Component target) {
	return new LabelledComponentModel(target, this);
    }

    /**
     * Creates a CheckboxModel.  This model implements java.awt.peer.CheckboxPeer.
     * @param target to associate with the peer.
     * @return the peer created.
     */
    protected ComponentModel makeCheckboxModel(Checkbox target) {
	return new CheckboxModel(target, this);
    }

    /**
     * Creates a PopupMeuModel.  This model implements java.awt.peer.PopupMenuPeer.
     * @param target to associate with the peer.
     * @return the peer created.
     */
    protected MenuModel makePopupMenuModel(PopupMenu target) {
	return new PopupMenuModel(target, this);
    }

    /**
     * Creates a MenuItemModel. 
     * @param target to associate with the model.
     * @return the model created.
     */
    protected MenuItemModel makeMenuItemModel(MenuItem target) {
	return new MenuItemModel(target);
    }

    /**
     * Creates a ChoiceModel.  This model implements java.awt.peer.ChoicePeer.
     * @param target to associate with the peer.
     * @return the peer created.
     */
    protected ChoiceModel makeChoiceModel(Choice target) {
	return new ChoiceModel(target, this);
    }

    /**
     * Creates a ScrollPaneModel.  This model implements java.awt.peer.ScrollPanePeer.
     * @param target to associate with the peer.
     * @return the peer created.
     */
    protected ComponentModel makeScrollPaneModel(ScrollPane target) {
	return new ScrollPaneModel(target, this);
    }

    /**
     * Creates a ListModel.  This model implements java.awt.peer.ListPeer.
     * @param target to associate with the peer.
     * @return the peer created.
     */
    protected ComponentModel makeListModel(List target) {
	return new ListModel(target, this);
    }

    /**
     * Creates a TextFieldModel.  This model implements java.awt.peer.TextFieldPeer.
     * @param target to associate with the peer.
     * @return the peer created.
     */
    protected TextFieldModel makeTextFieldModel(TextField target) {
        return new TextFieldModel(target, this);
    }

    /**
     * Creates a TextAreaModel.  This model implements java.awt.peer.TextAreaPeer.
     * @param target to associate with the peer.
     * @return the peer created.
     */
    protected TextAreaModel makeTextAreaModel(TextArea target) {
        return new TextAreaModel(target, this);
    }


    /**
     * Create a new window as a child of the given parent window,
     * and having default dimensions (0,0,1,1).
     */
    protected sun.porting.windowsystem.Window createWindow(ContainerView parent) {
	return winSys.makeWindow(parent.getContainerWindow(), 0, 0, 1, 1);
    }

    /* controller creation? */
}
