/*
 * @(#)ObservableModel.java	1.11 99/09/29 SMI
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package sun.awt.otk;

/**
 * A minimal Model which extends SubjectImpl.  It provides management
 * of dependent views and includes the concept of a "default view," which
 * is a crucial aspect of implementing an AWT toolkit: the view of the
 * Model that corresponds to the object managed by the parent's
 * LayoutManager--the default view--must be treated specially because it
 * must communicate its minimum and preferred sizes, inherit colorations
 * from parents, etc.
 */
public class ObservableModel extends SubjectImpl implements Model {
    private View defaultView;

    /**
     * Create a new ObservableModel.
     */
    public ObservableModel() {
	super();
    }

    /**
     * Get the default view.
     * @return A View, the one that has been set as the default view.
     */
    public final View getDefaultView() {
	return defaultView;
    }

    /**
     * Query whether a given view is the defaultView.  Equality is tested
     * using the equals() method, and equals() is NOT expected to be
     * commutative, i.e. v.equals(defaultView) and defaultView.equals(v)
     * could return different values because of an overridden equals() method.
     * @param v A view to be tested.
     * @return true if v is the default view, false otherwise.
     */
    public final synchronized boolean isDefaultView(View v) {
	return v.equals(defaultView) || defaultView.equals(v);
    }

    /**
     * Set the default view.  The default view is treated as a dependent, but it
     * is not added to the list of dependents.  This means that if the default view
     * is reset, it is as though the old default view has been removed.
     * @param v The new value of the default view.
     */
    public synchronized void setDefaultView(View newView) {
	defaultView = newView;
    }

    /**
     * Called when view initialization is complete to give the model the 
     * opportunity to set visibility.
     */
    public void initComplete() {
    }

    /**
     * Count the number of objects which depend on this Subject.
     * @return An integer, the number of dependents.
     */
    public int countDependents() {
	return super.countDependents() + ((defaultView != null) ? 1 : 0);
    }

    /**
     * Is the model interested in processing input events?
     * @return True if the model wants input, false otherwise.
     */
    public boolean isEnabled() {
	return true;
    }

    /**
     * Notify dependents that a change has occurred, supplying the provided
     * (optional) argument during the notification.
     *
     * This is explicitly defined to notify other dependents before
     * it notifies the default view.  This allows an alternate view
     * to process events (and mark them consumed) before the default 
     * view gets to see them.
     *
     * @param arg The argument to be passed during notification.
     */
    public void notifyDependents(Object arg) {
	View dv = null;

	synchronized (this) {
	    dv = defaultView;
	}

	super.notifyDependents(arg);

	if (consumed(arg)) {
	    return;
	}

	if (dv != null) {
	    dv.updateDependent(this, arg);
	}
    }

 
    /**
     * Dispose of this Model, i.e. break the connection to its Dependents.
     */
    public synchronized void dispose() {
	super.dispose();
	defaultView = null;
    }
}
