/*
 * @(#)RandomSelectionTextPane.java	1.6 98/11/30
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

package sun.awt.otk;

import java.awt.*;
import java.awt.event.*;
import sun.porting.toolkit.ToolkitEventHandler;

/**
 * This class implements the random selection mode. In this mode, if the 
 * user clicks on a line that is below the last line of text, new lines
 * are created to give the impression to the user of actually placi g the
 * cursor where they tapped (or at the beginning of the line they tapped on).
 * In reality, we just use the TextRandomData.insertSoftBreak() method to
 * create this illusion.
 * <p>
 * NOTE: This class assumes that the data buffer associated with it is of
 * type TextRandomData.
 * @see sun.awt.otk.TextComponentModel
 * @see sun.awt.otk.TextComponentView
 */
public abstract class RandomSelectionTextPane
extends MultiLineTextPane {

    /**
     * When true, a pointer pressed on textRect but off the text causes
     * implicit line return markers to be added to the text data.
     */
    protected static boolean randomSelectionMode;

    static {
	// set randomSelectionMode from system property
	if (Boolean.getBoolean("sun.awt.otk.noRandomSelectionMode")) {
	    randomSelectionMode = false;
	} else {
	    randomSelectionMode = true;
	}
    }

    /**
     * Create a new RandomSelectionTextPane instance.
     * @param view Our view
     * @param model Our data model
     * @param tk Our toolkit
     * @param colors Our color policy
     * @param insets The insets used for this textpane. These define the
     * outer boundaries that we use for clipping before drawing text or
     * selections.
     * @param left Left padding used in addition to insets before drawing text
     * @param right Rigth pad used along with insets before drawing text
     */
    public RandomSelectionTextPane(BasicView view, TextComponentModel model,
				   ObjectToolkit tk, ColorPolicy colors,
				   Insets insets, int left, int right) {

        super(view, model, tk, colors, insets, left, right);
    }

    /**
     * Handle all mouse events for the textpane.
     * @param id Event type
     * @param when Time of event
     * @param modifiers Flags
     * @param x X coordinate
     * @param y Y coordinate
     * @param arg1 Click count
     * @param arg2 isPopupTrigger
     * @return True if the event was handled, false otherwise.
     */
    public boolean handleMouseEvent(int id,  long when, int modifiers,
				    int x, int y, int arg1, int arg2) {

        boolean handled = false;

        switch (id) {

        case ToolkitEventHandler.POINTER_PRESSED:

	    // We don't allow changing the insertion point when composing
	    if (composeMode) {
	        ((TextComponentView) view).playNotAllowedSound();
		handled = true;
		break;
	    }

	    if (textRect.contains(x, y)) {

	        if (randomSelectionMode && arg1 > 0) {

		    int nrOfLines = data.getNumberOfLines();
		    int yLn = y2line(y);
		    int mpos = xy2pos(x, y);

		    // if we don't insert a break then let super handle event
		    if (yLn > nrOfLines - 1) {
		        int iLines = yLn - (nrOfLines - topLine - 1);
			if (insertSoftBreaks(iLines)) {
			    select(mpos + iLines, mpos + iLines);
			    handled = true;
			}
		    }
		} else {
		    // default case in super will do the right thing: let it go
		    handled = false;
		}
	    }
            break;
	    
	default:
	    break;
        }

	if (!handled) {
	    handled = super.handleMouseEvent(id, when, modifiers,
					     x, y, arg1, arg2);
	}

	return handled;
    }

    /**
     * This method either moves the cursor or the textpane down one line.
     * Over-ridden to allow insertion of a soft break.
     * @param absoluteMode When true, the whole textPane is scrolled down
     * by one line. Otherwise, only the cursor is moved down one line.
     * @return Returns true if the lineDown will cause the cursor to move 
     * outside the visible text. In this case, it is expected that the 
     * containing view for this textpane will issue a lineDown again with
     * absoluteMode set to true, which will then cause the appropriate
     * scroll to take place. If the cursor stays within the visible text,
     * then we return false. This return value is used by the container to
     * synchronize with any scrollbars that may exist.
     */
    public boolean lineDown(boolean absoluteMode) {
        if (randomSelectionMode) {
	    int lastLine = data.getNumberOfLines() - 1;
	    // Is the cursor on the last line?
	    if (cursorOnThisLine(lastLine, selStart, selEnd)) {
	        insertSoftBreaks(1);
	    } else if (selStart != selEnd) {
	        if (selStart >= data.getLineStart(lastLine) &&
		    selEnd <= data.getLineEnd(lastLine))
		    insertSoftBreaks(1);
	    }
	}

	return super.lineDown(absoluteMode);
    }

    /**
     * Given a y coordinate, figure out which line it would fall on within
     * the constraints of the visible lines.
     * @param y Y coordinate
     * @return Returns integer value of line that the hit occured on.
     * NOTE: This can be beyond the length of text, i.e. it can return a 
     * line number thatis past the last line of text. DON'T USE THIS 
     * FOR TEXT HIT TESTING.
     */
    protected int y2line(int y) {

        int li = linesInWindow();
        if (y < 0) {
	    return 0;
        } else if (y > (textRect.y + textRect.height)) {
	    return topLine + li - 1;
	} else {
	    int ln = topLine;
	    int minTextH = getLineHeight(view.getMetrics());
            int ys = textRect.y + minTextH;
            while ((y > ys) && (ln <= topLine + li - 1)) {
                ys += minTextH;
                ln++;
            }
	    return ln;
	}
    }

    /**
     * Append the given number of special random selection line break 
     * characters into the data buffer. These markers are handeled specially
     * by the TextDataRandomSelection class.
     * @param number The number of "soft" line breaks to insert
     */
    protected boolean insertSoftBreaks(int number) {

        if (number > 0) {

	    // Make sure that we have a TextRandomData object
	    try {
		int textLength = data.getLength();
		for (int i = 0; i < number; i++) {
		    ((TextRandomData) data).insertSoftBreak(textLength + i);
		}
		
		return true;
	    } catch (Exception e) {
	        throw new RuntimeException("Could not insert soft break.");
	    }
	}

	return false;
    }
}

