/*
 * @(#)SimpleStackActivationPolicy.java	1.11 99/07/29 SMI
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package sun.awt.otk;

import java.awt.Dialog;
import java.awt.Window;
import java.awt.AWTEvent;
import java.awt.event.WindowEvent;
import java.util.Vector;

/**
 * A simple activation policy whereby the window in front always
 * owns the input focus.
 */
public class SimpleStackActivationPolicy implements WindowActivationPolicy {
    Vector floaters = new Vector(1);

    /**
     * Constructs a new SimpleStackActivationPolicy object.
     */
    public SimpleStackActivationPolicy() {
    }

    /**
     * Adds a floating view to this activation policy object's list of windows that always
     * stay on top.
     * @param v the floating view to add; if it is already in the list, it is not added again
     */
    public void addFloatingView(View v) {
	if (!floaters.contains(v)) {
	    floaters.addElement(v);
	}
    }

    /**
     * Removes a floating view from this activation policy object's list of windows that
     * always stay on top.
     * @param v the floating view
     */
    public void removeFloatingView(View v) {
	floaters.removeElement(v);
    }

    /**
     * Returns the number of floating views in this activation policy object's list of
     * windows that always stay on top.
     * @return the number of floating views
     */
    public int countFloatingViews() {
	return floaters.size();
    }

    /**
     * Returns a copy of the list of floating views in this activation policy object's list of
     * windows that always stay on top.
     * @return an array containing references to the floating views; null if there are none
     */
    public View[] getFloatingViews() {
	if (floaters.size() == 0) return null;

	View v[] = new View[floaters.size()];
	floaters.copyInto(v);
	return v;
    }

    class WindowActivateListener extends java.awt.event.WindowAdapter {
	private WindowModel wm;
	public WindowActivateListener(WindowModel window) {
	    wm = window;
	}
	public void windowActivated(WindowEvent evt) {
 	    wm.requestFocus();
 	    evt.getWindow().removeWindowListener(this);
	};
    }

    /**
     * Notifies this window activation policy object that the visibility state of the given 
     * window has changed.  Pushes the given window onto the stack of currently 
     * visible windows if it has become visible; pops it off the stack if it has been hidden.
     * @param window the window whose visibility has changed
     * @param visible if true, this window has become visible; if false it has been hidden
     * @exception WindowActivationVetoException if window activation policy doesn't
     * allow this change
     */
    public void windowVisibilityChanged(WindowModel window, boolean visible)
	throws WindowActivationVetoException {

 	if (visible) {
	    // bootstrap problem with requesting focus before a
	    // window is mapped; use listener to wait until it is mapped
     	    ((Window)(window.getTarget())).addWindowListener(
     				    new WindowActivateListener(window));
	    pushActiveWindow(window);
	} else {
	    popActiveWindow(window);
	}
    }

    /**
     * Notifies this window activation policy object that the position of this window has
     * changed in the stack of currently visible windows. Moves the given window to
     * the appropriate position in the window stack and makes this window the active
     * window.
     * @param window the window whose stacking position has changed
     * @param stackOrder an integer specifying the new position of the window, where
     * a zero indicates the frontmost window, with positive numbers indicating how
     * many windows should be ahead of this window.  Negative numbers indicate to
     * start at the back of the stack, with -1 indicating that this window should be
     * on the bottom, -2 meaning that 1 window should be behind it, etc.
     * @exception WindowActivationVetoException if caller is attempting to hide a
     * floating view or a modal dialog
     */
    public void windowRestacked(WindowModel window, int stackOrder) 
	throws WindowActivationVetoException {
	
	if ((stackOrder > 0) && (stackOrder < countFloatingViews())) {
		throw new WindowActivationVetoException(window,	"Can't hide floating views");
	}

	stackOrder -= countFloatingViews();
	if (window instanceof DialogModel) {
	    Dialog d = (Dialog) window.getTarget();
	    if ((stackOrder != 0) && d.isModal()) {
		throw new WindowActivationVetoException(window,	"Modal dialogs can't be restacked");
	    }
	}

	if (stackOrder == 0) {
	    WindowModel oldWindow = null;
	    if (!activeWindowStack.isEmpty()) {
		oldWindow = (WindowModel) activeWindowStack.peek();
	    }

	    makeActive(window, oldWindow);
	}
    }


    private java.util.Stack activeWindowStack = new java.util.Stack();

    /**
     * Pushes the specified window onto the window stack and makes it the active 
     * window.
     * @param win the window to be pushed onto the stack
     */
    protected void pushActiveWindow(WindowModel win) {
	WindowModel oldWindow = null;

	if (!activeWindowStack.isEmpty()) {
	    oldWindow = (WindowModel) activeWindowStack.peek();
	}

	activeWindowStack.push(win);
	makeActive(win, oldWindow);
    }

    /**
     * "Pops" the specified window off the window stack. If the window is not on the
     * top of the stack, it is removed from its position within the stack. Makes the new
     * top window the active window.
     * @param win the window to be removed from the stack
     * @exception WindowActivationVetoException if window stack is empty
     */
    protected void popActiveWindow(WindowModel win) 
        throws WindowActivationVetoException {

	if (activeWindowStack.isEmpty()) {
	    throw new WindowActivationVetoException(win, 
						    "Window stack underflow?");
	}

	WindowModel oldWindow = (WindowModel) activeWindowStack.peek();

	if (oldWindow != win) {
	    /* window is not on the top of the stack, so "pop" is */
	    /* sort of a misnomer!                                */
	    activeWindowStack.removeElement(win);
	} else {
	    activeWindowStack.pop();
	    makeActive((activeWindowStack.isEmpty() ?
		       null : (WindowModel) activeWindowStack.peek()), 
		       oldWindow);
	}
    }

    /**
     * Makes the specified window the active window by posting a 
     * WINDOW_ACTIVATED WindowEvent to the EventQueue; also deactivates the
     * previously active window by posting a WINDOW_DEACTIVATED WindowEvent
     * to the EventQueue.
     * @param win the window to become the active window; may be null
     * @param old the window to be deactivated; may be null
     */
    protected void makeActive(WindowModel win, WindowModel old) {

	if (win == old) {
	    return;
	}

	if (old != null) {

	    ObjectToolkit tk = (ObjectToolkit) old.getToolkit();
	    AWTEvent ev = new WindowEvent((java.awt.Window) old.getTarget(), 
					  WindowEvent.WINDOW_DEACTIVATED);
	    tk.postEvent(ev);
	}

	if (win != null) {
	    win.requestFocus();

	    ObjectToolkit tk = (ObjectToolkit) win.getToolkit();
	    AWTEvent ev = new WindowEvent((java.awt.Window) win.getTarget(), 
					  WindowEvent.WINDOW_ACTIVATED);
 	    tk.postEvent(ev);
	}
    }
}
