/*
 * @(#)SingleLineTextPane.java	1.20 98/12/18
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

package sun.awt.otk;

import java.awt.*;
import java.awt.event.*;
import sun.porting.toolkit.ToolkitEventHandler;

/**
 * This is an implementation of a single line textpane used for TextField
 * peers. See 
 * <a href="TextComponentModel.html">TextComponentModel</a> for a high
 * level description of how textpanes are used. It is assumed that a
 * TextComponentView will provide a window to host this textpane.
 * @see sun.awt.otk.TextComponentModel
 * @see sun.awt.otk.TextComponentView
 */
public abstract class SingleLineTextPane
extends TextPaneImpl {

    /**
     * Create a new SingleLineTextPane instance.
     * @param view Our view
     * @param model Our data model
     * @param tk Our toolkit
     * @param colors Our color policy
     * @param insets The insets used for this textpane. These define the
     * outer boundaries that we use for clipping before drawing text or
     * selections.
     * @param left Left padding used in addition to insets before drawing text
     * @param right Rigth pad used along with insets before drawing text
     */
    public SingleLineTextPane(BasicView view, TextComponentModel model, 
			      ObjectToolkit tk, ColorPolicy colors,
			      Insets insets, int left, int right) {

        super(view, model, tk, colors, insets, left, right);
    }

    /**
     * Paint lines in the given range (inclusive).
     * @param g Current graphics context
     * @param s Starting line
     * @param e Ending line
     */
    public void paintLines(Graphics g, int s, int e) {
        paintLine(g, data.getTextString(), true);
    }

    /**
     * The text has been changed, scrollTo selection start.
     * @param start Begining line for painting
     * @param end Ending line for painting
     */
    protected void paintChangedLines(int start, int end) {
        paintLines(start, end);
        scrollTo(selStart, true);
    }

    /**
     * Set the bounds of the textpane. Sets textRect, topHalfRect and
     * bottomHalfRect to the correct insets-adjusted value.
     * @param x New x-coordinate
     * @param y New y-coordinate
     * @param width New width
     * @param height New height
     */
    public void setBounds(int x, int y, int width, int height) {

        super.setBounds(x, y, width, height); 

	// set the data model width
        data.setLineWidth(getMaxPossibleStringWidth());
    }

    /**
     * This method either moves the cursor or the textpane up one line.
     * Overriden to disable feature.
     */
    public void lineUp() {
    }

    /**
     * This method either moves the cursor or the textpane down one line.
     * Override to disable feature.
     */
    public void lineDown() {
    }

    /**
     * Move up one page.
     * Override to disable feature.
     */
    public void pageUp() {
    }

    /**
     * Move down one page.
     * Override to disable feature.
     */
    public void pageDown() {
    }

    /**
     * Get the number of lines in the textpane.
     */
    public int linesInWindow() {
        return 1;      // always only one line visible
    }

    /**
     * Get the number of lines in the textpane.
     */
    protected int linesInWindow(FontMetrics fm) {
        return 1;      // always only one line visible
    }

    /**
     * Scroll the text to the position "pos". If the pos is off-screen then
     * try to place as much as possible text on screen.
     * @param pos Index into the text
     * @param resizeScrollbars If true then scrollbars will be resized after
     * this operation.
     */
    protected void scrollTo(int pos, boolean resizeScrollbars) {

        if (!data.isValidPosition(pos)) {
	    return;
	}

	// try to make sure that we show all of the selection
	char [] text = data.getText();
        int xs = pos2x(selStart, text) - movedRight;
        int xe = pos2x(selEnd, text) - movedRight;
	int width = getMaxPossibleStringWidth();

	if (xs >= width) {               // start is to right of viewable text
	    pos = selEnd;
	} else if (xe <= 0) {   // end is to left of viewable text
	    pos = selStart;
	}

	super.scrollTo(pos, resizeScrollbars);
    }

    /**
     * Scroll the text to the position "pos" regardless of whether it is
     * already on the screen or not.
     * @param pos Index into the text
     * @param resizeScrollbars If true then scrollbars will be resized after
     * this operation.
     */
    protected void scrollAbsolute(int pos, boolean resizeScrollbars) {
        this.scrollTo(pos, resizeScrollbars);
    }

    /**
     * Determines if the textpane needs assistance with inline editing.
     * Currently, we only ask for assistance if the field is full.
     * @return Returns true if the field is not full, i.e. text is not over
     * flowing off the textpane, false otherwise.
     */
    public boolean supportsInlineEdit() {

        char [] text = data.getText();
	FontMetrics fm = view.getMetrics();

	if (text == null) {
	    return true;
	}
	if (fm == null) {
	    return false;
	}
	int w = getMaxPossibleStringWidth();
	int i = 0;

	while (i < text.length && w >= 0) {
	    w -= fm.charWidth(text[i++]);
	}

	return (w >= 0 || alwaysInlineEdit);
    }

    /**
     * return the x coordinate of the given character.
     * @param fm current FontMetrics
     * @param pos Position of character in buffer
     * @param buff Buffer containing the data string
     */
    protected int pos2x(FontMetrics fm, int pos, char [] buff) {
     
	//System.out.println("inside pos2x of SingleLineTextPane pos :"+pos+" buff is "+buff); 
        char echoChar = ((TextFieldModel)model).getEchoChar();
	if (echoChar == 0) {
	   //System.out.println("inside pos2x returning (normal) "+super.pos2x(fm, pos, buff)); 
	    return super.pos2x(fm, pos, buff);
	} else {
	    //System.out.println("inside pos2x returning (echoChar) "+fm.charWidth(echoChar) * pos); 
	    // have to take into account the space on the left of the field;
	    return textRect.x + leftPad + (fm.charWidth(echoChar) * pos);
	}
    }

    /**
     * return the position on the line "ln" given the old xPosition.
     * This is used for the UP and DOWN events.
     * @param xo position in text array on some line
     * @param ln linenumber
     * @return position in text array
     */
    protected int xPos(int xo, int ln) {
      
        char echoChar = ((TextFieldModel)model).getEchoChar();
	if (echoChar == 0) {
	    return super.xPos(xo, 0);
	} else {
	    FontMetrics fm = view.getMetrics();
	    int w = fm.charWidth(echoChar);
	    int len = data.getLength();
	    int div = xo / w;
	    return (div > len) ? len : div;
	}
    }

    /**
     * Tell the peer to resize scrollbars.
     */
    protected void resizeScrollbars() {
         // do nothing here
    }
}
