/*
 * @(#)SubjectImpl.java	1.6 98/11/25 SMI
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package sun.awt.otk;

import java.util.Vector;
import java.awt.event.InputEvent;

/**
 * An implementation class similar to java.util.Observable,
 * but designed to allocate objects only when absolutely necessary.
 *
 * In the otk Framework, most Subjects won't have multiple observers, and the
 * next most common case is two observers (with the second observer possibly
 * existing only some of the time).  Having more than two observers is a
 * very unusual circumstance in a toolkit.
 *
 * So we benefit from an implementation that differs from the one in
 * java.util: defaultView (the one required dependent) is stored
 * separately, since it must be distinguishable from other dependents.
 * And if there is only one other dependent, SubjectImpl stores it directly 
 * in its "dependents" field, avoiding both the creation of vector and
 * array objects, and an array copy which would be done on every
 * notify.
 */
public class SubjectImpl implements Subject {
    /**
     * This Subject's dependents.
     */
    protected Object dependents;
    private Object dTmp;

    /**
     * Create a new Subject with no Dependents.
     */
    public SubjectImpl() {
    }

    /**
     * Count the number of objects which depend on this Subject.
     *
     * NOTE: There is no point in synchronizing this.  If the caller wants to
     * guarantee that the value is still valid after the call completes,
     * the call and whatever follows must be in a synchronized block anyway.
     * @return An integer, the number of dependents.
     */
    public int countDependents() {
	Object d = dependents;

	if (d == null) {
	    return 0;
	} else if (d instanceof Vector) {
	    return ((Vector)d).size();
	} else {
	    return 1;
	}
    }

    /**
     * Add an object which is dependent on this Subject.
     * @param d The Dependent to be added.
     */
    public synchronized void addDependent(Dependent d) {
	if (dependents == null) {
	    dependents = d;
	} else {
	    Vector v = null;
	    if (! (dependents instanceof Vector)) {
		v = new Vector(2);
		v.addElement(dependents);
		dependents = v;
		dTmp = new Dependent[2];
	    } else {
		v = (Vector)dependents;
	    }

	    v.addElement(d);
	}
    }

    /**
     * Remove a Dependent from this Subject.  If the given object does not depend
     * on the subject, no action is taken.
     * @param d The dependent to remove.
     */
    public synchronized void removeDependent(Dependent d) {
	if (dependents == null) {
	    return;
	}

	if (dependents instanceof Vector) {
	    Vector v = (Vector)dependents;
	    v.removeElement(d);

	    if (v.size() == 0) {
		dependents = null;
		dTmp = null;
	    }
	} else {
	    dependents = null;
	}
    }

    /**
     * Remove all dependents from this Subject.
     */
    public synchronized void removeDependents() {
	if (dependents == null) {
	    return;
	}

	if (dependents instanceof Vector) {
	    ((Vector)dependents).removeAllElements();
	}

	dTmp = null;
	dependents = null;
    }

    /**
     * Notify all the dependents that a change has occurred on this Subject.
     * Equivalent to <code>notifyDependents(null);</code>
     */
    public void notifyDependents() {
	notifyDependents(null);
    }

    /**
     * Notify all dependents that a change has occurred on the Subject.  The
     * supplied argument is passed to each Dependent during the notification.
     * The argument is optional, i.e. its value may be null.
     * @param arg an optional argument to be passed to the Dependents.
     */
    public void notifyDependents(Object arg) {
	int cnt = 0;
	Dependent dt[] = null;

	synchronized (this) {
	    if (dependents == null) {
		return;
	    } else {
		if (dependents instanceof Vector) {
		    Vector v = (Vector)dependents;
		    cnt = v.size();
		    dt = (Dependent [])dTmp;
		    if (dt.length < cnt) {
			dTmp = new Dependent[cnt + 1];
			dt = (Dependent [])dTmp;
		    }
		    v.copyInto(dt);
		}
	    }
	}

	if (dt == null) {
	    ((Dependent)dependents).updateDependent(this, arg);
	} else {
	    while (--cnt >= 0) { 
		if (dt[cnt] != null) {
		    dt[cnt].updateDependent(this, arg);
		    dt[cnt] = null;
		    if (consumed(arg)) {
			break;
		    }
		}
	    }
	}
    }

    /**
     * Returns whether the given event has been consumed.
     * @param arg the event to test
     * @return true if arg is an InputEvent and has been consumed, false otherwise
     */
    protected boolean consumed(Object arg) {

	if (arg instanceof InputEvent) {
	    return ((InputEvent)arg).isConsumed();   
	}
	return false;
    }

    /**
     * Dispose of the Subject.  This includes removing all of its dependents.
     */
    public synchronized void dispose() {
	removeDependents();
    }
}
