/*
 * @(#)TextCharData.java	1.17 98/11/25
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

package sun.awt.otk;

import java.awt.FontMetrics;
import java.awt.Component;
import java.text.BreakIterator;

/**
 * This class implements a data buffer using an array of char's. Support for
 * line (see <a href="#breakChars">breakChars</a>) and word wrapping is also 
 * provided. When data is deleted, if there is less than 
 * <a href="#RELEASE_THRESHOLD">RELEASE_THRESHOLD</a> percent occupied space,
 * the buffer size is decreased.
 */
public class TextCharData extends TextData {

    /**
     * Data storage buffer
     */
    protected char data[];

    /**
     * Number of entries to grow the lines array when it runs out of space
     */
    protected static int TEXT_GROWTH_PAD = 32;

    /**
     * The following value controls the number of lines that are looked at
     * "before" the insertion/deletion point to properly handle wordwrap. This
     * value is used to synchornize the insertion/deletion and updateLinesArray
     * methods.
     */
    protected static final int wrapLineDecrement = 1;

    /**
     * The following values are used as constants for line wrapping. Although
     * we use a lineIterator, it does not handle these chars properly.
     */
    protected static final String breakChars = "\n\r";

    /**
     * Default character used for calculating width of a character when
     * tabbing. A tab is defined as tabChar * <a href="#tabWidth">tabWidth</a>.
     */
    protected static final char tabChar = 'O';

    /**
     * Default number of tabChar used when calculating widht of a tab.
     * A tab is defined as tabWidth * <a href="#tabChar">tabChar</a>.
     */
    protected static final int tabWidth = 8;

    /**
     * After deleting data, decrease the buffer size if we are less than
     * the specified percentage of previous length.
     */
    protected static double RELEASE_THRESHOLD = 0.8;


    /**
     * Create a TextData buffer that uses an array of char's to store data.
     * @param model The TextComponentModel that this data buffer is
     * associated with
     * @param text Initialization text for the data buffer
     * @param lineWidth Default line width. Use 0 if not interested in
     * word wrapping (see
     * <A href="#setWordWrap">setWordWrap()</A>
     * )
     * @param wrap Default value for the word wrap mode
     * @see sun.awt.otk.TextComponentModel
     */
    public TextCharData(TextComponentModel model, String text,
			int lineWidth, boolean wrap) {
        super(model, text, lineWidth, wrap);
    }

    /**
     * Set the current wordwrap flag.
     * @param wrap Turn on word wrap when true, off otherwise
     */
    public synchronized void setWordWrap(boolean wrap) {
        super.setWordWrap(wrap);
	nrOfLines = updateLinesArray(0, false);
	StateChangeNotification c = new StateChangeNotification(
				       StateChangeNotification.DATA_CHANGE, 0, nrOfLines - 1);
	notifyDependents((Object) c);
    }

    /**
     * Set the text in the data buffer to given text string.
     * @param text String to set the buffer value to
     */
    public synchronized void setText(String text) {
        if (setTextImpl(text)) {
	    StateChangeNotification c = new StateChangeNotification(
					   StateChangeNotification.DATA_CHANGE, -1, nrOfLines - 1);
	    sendNotifications(c);
	}
    }

    /**
     * Concrete setText implementation.
     * @param text String to set the buffer value to
     */
    protected synchronized boolean setTextImpl(String text) {
        // Always make sure that text is non-null
        if (text == null) {
            text = "";
        }

        textLength = text.length();
        data = new char[textLength + TEXT_GROWTH_PAD];
        int nr = Math.max(textLength / 20, 1);   // guess here: 20 chars/line
        lines = new int[nr];
        text.getChars(0, textLength, data, 0);
        nrOfLines = updateLinesArray(0, false);

	return true;
    }

    /**
     * Insert the given char array at given position.
     * @param text Character array to insert in data buffer
     * @param pos Index to insert the given string at
     */
    public synchronized void insert(char [] text, int pos) {
        StateChangeNotification c = insertImpl(text, pos);
	if (c != null) {
	    sendNotifications(c);
	}
    }

    /**
     * Concrete insert() implementation.
     * @param text Character array to insert in data buffer
     * @param pos Index to insert the given string at
     */
    protected synchronized StateChangeNotification insertImpl(char [] text,
							     int pos) {
        //System.err.println("TKCDM.insertImpl: pos=" + pos);

       

        if (isValidPosition(pos)) {

	    int l = text.length;
	    if (data.length <= textLength + l) {
	        char tmp[] = new char[textLength + l + TEXT_GROWTH_PAD];
		System.arraycopy(data, 0, tmp, 0, textLength);
		data = tmp;
	    }

	    if (textLength > pos) {
	        System.arraycopy(data, pos, data, pos + l, textLength - pos);
	    }
	
	    if (l > 1){
	        System.arraycopy(text, 0, data, pos, l);
	    } else if (l == 1){
	        data[pos] = text[0];
	    }
	    textLength += l;

	    int lnr = getLineNumber(pos);
            int ls = (lnr - wrapLineDecrement >= 0) 
	        ? lnr - wrapLineDecrement : lnr;

	    StateChangeNotification c = new StateChangeNotification(StateChangeNotification.DATA_CHANGE,
								 ls,
								 Math.min(ls + updateLinesArray(ls, true), nrOfLines - 1));
	    return c;

	} else {
	    return null;
	}
    }

    /**
     * Replace the data between the given indices with the iven text.
     * @param text String to replace with
     * @param start Starting index for replacement
     * @param end Ending index for replacement
     */
    public synchronized void replace(String text, int s, int e) {

        boolean textValueChanged = false;

	// do we need to flip s and e
	if (s > e) {
	    int temp = s;
	    s = e;
	    e = temp;
	}

	// truncate till end if we're out of range
	if (e > textLength) {
	    e = textLength;
	}

        if (isValidPosition(s) && isValidPosition(e)) {
            StateChangeNotification d = deleteImpl(s, e);
            StateChangeNotification i = insertImpl(text.toCharArray(), s);
	    if (d == null && i == null) {
		return;
	    }
	    if (d != null && i != null) {
		int ls = Math.min(d.getStartIndex(), i.getStartIndex());
		int le = Math.max(d.getEndIndex(), i.getEndIndex());
		StateChangeNotification t =
		    new StateChangeNotification(StateChangeNotification.DATA_CHANGE, ls, le);
		sendNotifications(t);
		return;
	    }
	    if (d != null) {
		sendNotifications(d);
	    } else {
		sendNotifications(i);
	    }
        }
    }

    /**
     * Delete the data between the given indices.
     * @param start Starting index for deletion
     * @param end Ending index for deletion
     */
    public synchronized void delete(int s, int e) {

        StateChangeNotification c = deleteImpl(s, e);
	if (c != null) {
	    sendNotifications(c);
	}
    }

    /**
     * Concrete delete() implementation.
     * @param start Starting index for deletion
     * @param end Ending index for deletion
     */
    protected synchronized StateChangeNotification deleteImpl(int s, int e) {

        //System.err.println("+TKCDM.deleteImpl: s=" + s + " e=" + e + " len=" + textLength + " buffLen=" + data.length);

        if (isValidPosition(s) && isValidPosition(e)) {

	    // do we need to flip s and e
	    if (s > e) {
	        int temp = e;
		s = e;
		e = temp;
	    }

	    // keep track of old buffer length
	    int tempLen = data.length;

	    // delete the data
	    System.arraycopy(data, e, data, s, textLength - e);
	    textLength -= (e - s);

	    // if we decreased buffer length by greater than release threshold
	    // then decrease size of buffer
	    if (((double) textLength / (double) tempLen) < RELEASE_THRESHOLD) {
	        char [] d = new char[textLength + TEXT_GROWTH_PAD];
		System.arraycopy(data, 0, d, 0, textLength);
		data = d;
	    }

	    //System.err.println("-TKCDM.deleteImpl: s=" + s + " e=" + e + " len=" + textLength + " buffLen=" + data.length);
	    int lnr = getLineNumber(s);
	    int ls = (lnr - wrapLineDecrement >= 0)
	        ? lnr - wrapLineDecrement : lnr;

	    StateChangeNotification c = new StateChangeNotification(StateChangeNotification.DATA_CHANGE,
							    ls, 
							    Math.min(ls + updateLinesArray(ls, true), nrOfLines - 1));
	    return c;

	} else {
	    return null;
	}
    }

    /**
     * Check if the given position is valid.
     * @param pos Position (index) of character in data buffer
     */
    public boolean isValidPosition(int p) {
        return (p >= 0 && p <= textLength);
    }

    /**
     * Return data as char array without removing linebreak characters at end.
     * @param pos Index in data buffer
     * @param length Number of characters to be returned 
     * @see getStrippedText
     */
    public char [] getNonStrippedText(int pos, int length) {
        if (pos + length <= textLength) {
	    char [] text = new char[length];
	    System.arraycopy(data, pos, text, 0, length);
	    return text;
	} else {
	    return null;
	}
    }

    /**
     * Return data as char array removing linebreak characters at end.
     * THIS CAN RETURN NULL!!!
     * @param pos Index in data buffer
     * @param length Number of characters to be returned 
     * @see getText
     */
    public char [] getStrippedText(int pos, int length) {

      //System.err.println("+getStrippedText(" + pos + ", " + length + "): buffLen=" + getLength() + " data.len=" + data.length);
        if (pos + length <= textLength && isValidPosition(pos) &&
	     isValidPosition(pos + length)) {

	    if (length > 1) {
	        if (data[pos + length] == '\r' &&
		    data[pos + length - 1] == '\n') {
		  
		    length--;
		}
	    }

	    char [] text = new char[length];
	    System.arraycopy(data, pos, text, 0, length);

	    //System.err.println(" getStrippedText(" + pos + ", " + length + "): >" + text + "<");
	    //printLines(null); 
	    return text;

	} else {
	    return null;
	}
    }

    /**
     * Calculates actual width in pixels that 
     * text will take on the screen from start to end positions  in text array;
     * and adds calculated length to passed in l 
     *  @param start first text position 
     *  @param end last text posion
     *  @param fm font metrics
     *  @param widths (0-255) character  widths in pixels
     *  @param l pixels are added to calculated length
     *  @return length in pixels
     */
    protected int charactersWidth(int start, int end, 
			   FontMetrics fm, int[] widths, int l) {

        for (int i = start; i <= end; i++) {
	    // Check for tabs
	    if (data[i] == '\t') {
	        // We found a tab, find out the length of this line
	        // up to this point
	        l = (((l/widths[tabChar])/tabWidth + 1) * 
		     tabWidth) * widths[tabChar];
	    } else {
	        if (data[i] > 255) {
		    l += fm.charWidth(data[i]);
		} else {
		    l += widths[data[i]];
		}
	    }
        }
        return l;
    }

    /**
     * Calculates actual width in pixels that 
     * text will take on the screen from start to end positions in data array;
     * and adds calculated length to passed in l. This method uses iterators.
     *  @param start first text position 
     *  @param end last text posion
     *  @param fm font metrics
     *  @param widths (0-255) character  widths in pixels
     *  @param l pixels are added to calculated length
     *  @return length in pixels
     */
    protected int charsWidth(int start, int end, 
			     FontMetrics fm, int[] widths, int l) {
      int cs = 0, ce = 0;

      if (start == end) {
	  l = charactersWidth(start, end, fm, widths, l);
      } else {
	  // Iterate using a char iterator
	  BreakIterator cIter = BreakIterator.getCharacterInstance();
	  cIter.setText(new String(data, start, end - start));
	  cs = cIter.first();
	
	  // search forward from first break to end of token
	  do {
	      ce = cIter.next();

	      if (ce - cs == 1) {
		  l = charactersWidth(start+cs, start+cs, fm, widths, l);
	      } else {
		  l += (ce == BreakIterator.DONE) 
		      ? fm.charWidth(data[start + cs])
		      : fm.charsWidth(data, start + cs, ce - cs);
	      }
      
              cs = ce;
	  } while (ce != BreakIterator.DONE);
      }
      //System.err.println("***** charsWidth: s:"+start+", e:"+end+", cs:"+cs+", ce:"+ce+", l:"+l);
      return l;
    }

    /**
     * Update lines array from the specified lineNumber. The basic algorithm 
     * here is as follows:
     *
     *    Iterate through all text starting at line nr using a lineIterator {
     *      if the last character of token is breakChar OR contains breakChar {
     *        Iterate through text using a character iterator
     *          handle hard breaks
     *          handle wordwrap
     *      } else {
     *          handle wordwrap
     *      }
     *    }
     *
     * The main attempt is to minimize the use of charIterators as much as
     * possible.
     * @param nr update lines array from this line number
     * @param returnChangedLines Return the number of changed lines if set to
     * true, otherwise return the new nrOfLines. This implementation will 
     * always return nrOfLines, i.e. returnChangedLines=false. The "true"
     * functionality is implemented in the sub-classes.
     * @return new number of  lines in lines array (note nrOfLines is not 
     * set yet)
     */
    public int updateLinesArray(int nr, boolean returnChangedLines) {

        if (this.lineWidth <= 0) {
	    return 0;
	}

        //printLines(null);
        Component target = (Component) model.getTarget();
        FontMetrics fm = target.getFontMetrics(target.getFont());
        int widths[] = fm.getWidths();
        int w = this.lineWidth;
        int lineWidth = 0;      // width of line being processed
	int wordWidth = 0;      // width of current "word"
	int origNrOfLines = nrOfLines;

	int linesIndex = nr;    // index into lines array

	// remember offset into text array where we start from - this keeps us
	// from keeping a local copy of the text buffer to represent what the
	// iterator has internally.
	int textOffset = lines[nr];

	// Initialize the BreakIterator
	BreakIterator iter = BreakIterator.getLineInstance();
	iter.setText(new String(data, textOffset, textLength - textOffset));
	int s = iter.first();
	//System.err.println("***** updateLinesArray: nr:"+nr+", textOffset:"+textOffset+", linesIndex:"+linesIndex+", s:"+s+", wordwarp:"+wordwrap+" textLength="+textLength);
	lines[linesIndex++] = textOffset + s;

	// Make sure that done only increments for "successive" iterations 
	// through this loop, i.e. make sure we kick only after 
	// wrapLineDecrement+1 consecutive iterations. oldDone ensures this.
	int done =0, oldDone = 0;
	int e;
	for (e = iter.next();
	     e != BreakIterator.DONE &&
	         e <= textLength &&
	         done <= wrapLineDecrement + 1;
	     s = e, e = iter.next()) {

	    //System.err.println("***** updateLinesArray: s:"+s+", e:"+e+", w:"+w+", linesIndex:"+linesIndex+", done:"+done+", str:"+new String(data, textOffset+s, e-s)+"<");

	    // If the last character of the token is any one of these chars
	    // then we need to scan it using a charIterator since we cannot
	    // be certain that there are no hard breaks in it. We only test
	    // the last character first because it is much cheaper and will
	    // satisfy a majority of the cases.
	    if (breakChars.indexOf(data[textOffset + e - 1]) != -1 ||
		 nextBreakChar(textOffset+s, textOffset+e, breakChars) != -1) {

	        // Iterate using a char iterator
	        BreakIterator cIter = BreakIterator.getCharacterInstance();
		cIter.setText(new String(data, textOffset + s, e - s));
		int cs = cIter.first();
		int off = textOffset + s;
		wordWidth = 0;
	
	        // search forward from first break to end of token
		int ce;
	        for (ce = cIter.next();
		     ce != BreakIterator.DONE & done <= wrapLineDecrement + 1;
		     cs = ce, ce = cIter.next()) {

		    //System.err.println("***** updateIndex(1): cs:"+cs+", ce:"+ce+", w:"+w+", lw:"+lineWidth+", ww:"+wordWidth+", done:"+done+", off:"+off+", char:"+data[off+cs]);

		    if (done > 0 && oldDone == done) {
		        done = oldDone = 0;
		    } else {
		        oldDone = done;
		    }

		    // make sure lines array is large enough
		    if (lines.length <= (linesIndex + 3)) {
		        growLinesArray(TEXT_GROWTH_PAD);
		    }

		    if (data[off + cs] == '\n') {
		        if (off + ce < textLength &&
			    data[off + ce] == '\r') {

			    if (lines[linesIndex] == off + ce + 1) {
			        done += (returnChangedLines) ? 1 : 0;
			    } else {
			        lines[linesIndex] = off + ce + 1;
			    }
			    //System.err.println("***** (1) break at " +(off+ce+1));
			    // eat next token
			    cs = ce;
			    ce = cIter.next();

			    // Line iterator does not handle CRLF properly
			    s = e;
			    e = iter.next();
			} else {
			    if (lines[linesIndex] == off + ce) {
			        done += (returnChangedLines) ? 1 : 0;
			    } else {
			        lines[linesIndex] = off + ce;
			    }
			    //System.err.println("***** (1) break at " +(off+ce));
			}
			linesIndex++;
			lineWidth = 0;     // reset lineWidth for next line
			wordWidth = 0;     // reset wordWidth for next token
		    } else if (data[off + cs] == '\r') {
		        // LFCR comes back as one char iterator!!!!
		        if (off + ce < textLength &&
			    data[off + cs + 1] == '\n') {

			    if (lines[linesIndex] == off + cs + 1) {
			        done += (returnChangedLines) ? 1 : 0;
			    } else {
			        lines[linesIndex++] = off + cs + 1;
			        lines[linesIndex] = off + ce;
			    }
			    // eat next token
			    cs = ce;
			    ce = cIter.next();
			} else {
			    if (lines[linesIndex] == off + ce) {
			        done += (returnChangedLines) ? 1 : 0;
			    } else {
			        lines[linesIndex] = off + ce;
			    }
			    //System.err.println("***** (1+) break at " +(off+ce));
			}
			linesIndex++;
			lineWidth = 0;     // reset lineWidth for next line
			wordWidth = 0;     // reset wordWidth for next token
		    } else if (w > 0 && wordwrap) {

		        wordWidth = charsWidth(cs, cs, fm, widths, wordWidth);
		        if (wordWidth >= w) {
			    //System.err.println("***** (1) breaking fat word at " + cs);
			    if (lines[linesIndex] == off + cs) {
			        done += (returnChangedLines) ? 1 : 0;
			    } else {
			        lines[linesIndex] = off + cs;
			    }
			    linesIndex++;
			    lineWidth = 0;
			    wordWidth = charsWidth(cs, cs, fm, widths, 0);
			}
			
			if (lineWidth + wordWidth >= w) {
			    //System.err.println("***** (1) breaking fat line at " + off);
			    if (lines[linesIndex] == off) {
			        done += (returnChangedLines) ? 1 : 0;
			    } else {
			        lines[linesIndex] = off;
			    }
			    linesIndex++;
			    lineWidth = 0;
			}
		    }
		}
	    }

	    // take care of line breaks due to TextArea width.
	    // updateIndex() can get called from setTextImpl() before the
	    // dimensions are set and w can be negative. Doing wordwrap in
	    // this case has no meaning - do nothing extra!
	    else if (w > 0 && wordwrap) {
	        wordWidth = charsWidth(textOffset + s,
				       textOffset + e - 1, fm, widths, 0);
		//System.err.println("***** updateLinesArray(2): s:"+s+", e:"+e+", w:"+w+", lw:"+lineWidth+", ww:"+wordWidth+", linesIndex:"+linesIndex);
		// fat word! lineWidth must be <= w
		if (wordWidth >= w) {
	        
		    // Iterate using a char iterator
		    BreakIterator cIter = BreakIterator.getCharacterInstance();
		    cIter.setText(new String(data, textOffset + s, e - s));
		    int cs = cIter.first();
	
		    // search forward
		    int ce, temp;
		    wordWidth = 0;
		    int wOff = textOffset + s;
		    for (ce = cIter.next();
			 ce != BreakIterator.DONE && done <= wrapLineDecrement + 1;
			 cs = ce, ce = cIter.next()) {

		        // make sure lines array is large enough
		        if (lines.length == linesIndex) {
			    growLinesArray(TEXT_GROWTH_PAD);
			}

			if (done > 0 && oldDone == done) {
			    done = oldDone = 0;
			} else {
			    oldDone = done;
			}

		        temp = s + textOffset + cs;
			//System.err.println("***** updateIndex(2): cs:"+cs+", ce:"+ce+", w:"+w+", lw:"+lineWidth+", ww:"+wordWidth+", done:"+done+", char:"+data[temp]);
			wordWidth =
			    charsWidth(temp, temp, fm, widths, wordWidth);
			if (wordWidth >= w) {
			  if (lineWidth != 0) {
			    //System.err.println("***** (2) breaking fat word at " + wOff);
			    if (lines[linesIndex] == wOff) {
			        done += (returnChangedLines) ? 1 : 0;
			    } else {
			        lines[linesIndex] = wOff;
			    }
			    linesIndex++;
			    lineWidth = 0;
			    wOff = textOffset + s + cs;

			  } else {

			    //System.err.println("***** (2) breaking fat line at " + temp);
			    if (lines[linesIndex] == temp) {
			        done += (returnChangedLines) ? 1 : 0;
			    } else {
			        lines[linesIndex] = temp;
			    }
			    linesIndex++;
			    lineWidth = 0;
			    wordWidth = charsWidth(temp, temp, fm, widths, 0);
			  }
			}
		    }

                } else if (lineWidth + wordWidth >= w) {

		    // make sure lines array is large enough
		    if (lines.length == linesIndex) {
		        growLinesArray(TEXT_GROWTH_PAD);
		    }

		    if (done > 0 && oldDone == done) {
		        done = oldDone = 0;
		    } else {
		        oldDone = done;
		    }

		    if (lines[linesIndex] == textOffset + s) {
		        done += (returnChangedLines) ? 1 : 0;
		    } else {
		        lines[linesIndex] = textOffset + s;
		    }
		    linesIndex++;
		    lineWidth = 0;     // reset lineWidth- current added below
		}
	    }
	    lineWidth += wordWidth;
	}

	/*
	System.err.println("***** updateLinesArray: nr:"+nr+", textOffset:"+textOffset+", linesIndex:"+linesIndex+", s:"+s+", e:"+e+", done:"+done+", nrOfLines:"+nrOfLines);
	if (done > wrapLineDecrement)
	    printLines(null);
	else
	    printLines(null, 0, linesIndex);
        */

	if (returnChangedLines) {
	    // set the nrOfLines only if it changed
	    if (e == BreakIterator.DONE || e >= textLength) {
	      //if (done <= wrapLineDecrement) {
	      nrOfLines = linesIndex;
	    }
	    if (nrOfLines - origNrOfLines >= 0)
	        return nrOfLines - nr;          // insertion
	    else
	        return origNrOfLines - nr;      // deletion
	} else {
	    // NOTE: WE DON'T SET nrOfLines HERE - THIS CAN BE DANGEROUS!!!
	    // Make sure you set nrOfLines to returned value
	    return linesIndex;
	}
    }

    /**
     * Given start and end index in data buffer, return index of first
     * line break character encountered.
     * @param s Start index
     * @param e End index
     * @param b String containing the line break characters
     * @return Returns the index of the first character in String b found
     * between start and end.
     */
    protected int nextBreakChar(int s, int e, String b) {
        // Don't need breakIterators as we're just doing char comparisons
        // against non-compositional characters.
        for (int i = s; i < e; i++)
	    if (b.indexOf(data[i]) != -1)
	        return i;

	return (-1);
    }
}
