/*
 * @(#)TextComponentModel.java	1.24 99/08/19
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

package sun.awt.otk;

import java.awt.*;
import java.awt.event.*;
import java.util.Vector;

import sun.porting.toolkit.ToolkitEventHandler;

/**
 * This is the super class for all text components supported by the PJava
 * toolkit. These include peered (TextArea and TextField) as well as non-
 * peered (virtual keyboard and text popup) components. The
 * TextComponentModel is responsible for defining the API for interacting
 * with the component. All text components
 * can be described using the following high-level architecture:
 * <p><li>1. Every text component has a TextComponentModel that defines the
 * interface for controling it. In the case of peered components, this
 * will correspond to atleast that components peer interface.</li>
 * <li>2. Associated with every TextComponentModel is atleast one 
 * TextComponentView. Each TextComponentView contains a
 * TextPane, which is a light-weight implementation of a text drawing
 * object and handles all drawing of text on this components behalf.</li>
 * <li>3. Every TextComponentView has atleast one TextComponentController
 * attached to it which is responsible for delivering events to the view
 * that are dispatched to it by the toolkit.</li>
 * <p>
 *     TextComponentModel                            TextComponentController
 *     ^     |             +-------------------+                 |    ^
 *     |     |    1..n     | TextComponentView |       1..n      |    |
 *     |     +-------------|                   |-----------------+    | Events
 *     |                   | +---------------+ |                      |
 *  Component/Peer APIs    | |   TextPane    | |                      |
 *                         | +---------------+ |                   Toolkit
 *                         +-------------------+            
 * <p>
 * TextComponentModel is also responsible for handling communication from the
 * Input Method if one is present. When designing text components, you need to
 * handle this protocol both for buffering and displaying text received from
 * the Input Method Framework (IMF).
 * <p>
 * The TextComponentModel stores its text data in a TextData buffer via
 * the TextDataInterface. The TextComponentView associated with this model
 * is registered as the Dependent of the TextData and receives paint requests
 * when the data buffer is updated.
 * <p>
 * @see sun.awt.otk.TextComponentPeerModel
 * @see sun.awt.otk.TextComponentView
 * @see sun.awt.otk.TextComponentController
 * @see sun.awt.otk.TextPane
 * @see sun.awt.otk.TextDataInterface
 * @see sun.awt.otk.Dependent
 */
public class TextComponentModel
    extends ComponentModel 
    implements InputMethodListener  {

    /** We cache the value of the peers editable field */
    protected boolean editable;

    /** Our data buffer */
    protected TextData data;

    /*
     * We maintain our own notion of which of our views is the top most
     * so we can accurately get/set information on it. We cannot rely on
     * the defaultView value as that is used to denote the View that is 
     * associated with the AWT side and is the one used for layout type
     * information. We implement a simple "last added view is on top" 
     * policy and enforce it by over-riding the add/removeDependent methods.
     * Note that no setTopView() method is provided.
     */
    private View myTopView;

    /**
     * Constructs a new TextComponentModel instance.
     * @param target the AWT Component
     * @param tk the toolkit
     */
    public TextComponentModel(Component target, ObjectToolkit tk) {
        super(target, tk);
	editable = true;
	data = getDataModel();
    }

    /**
     * Get the text from the data buffer.
     * @see java.awt.peer.TextComponentPeer
     */
    public synchronized String getText() {
        return data.getTextString();
    }

    /**
     * Set the text of the data buffer.
     * @text Text to use for setting the data buffer
     * @see java.awt.peer.TextComponentPeer
     */
    public synchronized void setText(String text) {
        data.setText(text);
    }

    /**
     * Get the start of the current selection.
     * @return Returns the index of the selection start
     * @see java.awt.peer.TextComponentPeer
     */
    public synchronized int getSelectionStart() {
	return 0;
    }

    /**
     * Get the end of the current selection.
     * @return Returns the index of the selection end
     * @see java.awt.peer.TextComponentPeer
     */
    public synchronized int getSelectionEnd() {
	return 0;
    }

    /**
     * Get the current caret position.
     * @return Returns the index of the caret position
     * @see java.awt.peer.TextComponentPeer
     */
    public synchronized int getCaretPosition() {
	return 0;
    }

    /**
     * Get the value of the edit mode.
     * @return Returns true when editable, false otherwise
     */
    public boolean isEditable() {
	return editable;
    }

    /**
     * Post an ACTION_PERFORMED event to the AWT component.
     */
    public void sendActionEvent() {
	if (editable) {
	    postActionEvent(ActionEvent.ACTION_PERFORMED,
			    getDataModel().getTextString());
	}
    }

    /**
     * Post a VIRTUAL_COMPONENT_SHOW event to the associated virtual component,
     * like virtual keyboard or text popup.
     * @param msg Message to be posted.
     * @see sun.awt.otk.VirtualComponentMessage 
     */
    public void sendShowEvent(VirtualComponentMessage msg) {
	AWTEvent e = new VirtualComponentEvent(getTarget(),
					       VirtualComponentEvent.VIRTUAL_COMPONENT_SHOW, msg);
	ObjectToolkit.postEvent(e);
    }

    /**
     * Post a VIRTUAL_COMPONENT_HIDE event to the associated virtual component,
     * like virtual keyboard or text popup.
     * @param msg Message to be posted.
     * @see sun.awt.otk.VirtualComponentMessage 
     */
    public void sendHideEvent(VirtualComponentMessage msg) {
	AWTEvent e = new VirtualComponentEvent(getTarget(),
					       VirtualComponentEvent.VIRTUAL_COMPONENT_HIDE, msg);
	ObjectToolkit.postEvent(e);
    }

    /**
     * Get the current TextData buffer. If one doesn't exist yet, create it.
     * @return Data model for this component
     */
    public TextData getDataModel() {
	if (data == null) {
	    data = new TextRandomData(this, "", 0, true);
	}
	return (TextData) data;
    }

    /**
     * Invoked when the text entered through an input method has changed.
     * This is the mechanism for the Input Method Framework (IMF) to
     * communicate to text components.
     * @param committedText IMF text that has already been committed
     * @param composedText IMF text that has not been committed yet
     * @param caret Location of caret in the IMF text.
     */
    public void inputMethodTextChanged(String committedText,
				       String composedText,
				       int caret) {
        ((TextComponentView) myTopView).inputMethodTextChanged(committedText,
				 composedText,
				 caret);
    }

    /**
     * Invoked when the caret within composed text has changed.
     * @param committedText IMF text that has already been committed
     * @param composedText IMF text that has not been committed yet
     * @param caret Location of caret in the IMF text.
     */
    public void caretPositionChanged(String committedText,
				     String composedText,
				     int caret) {

	
        ((TextComponentView) myTopView).inputMethodTextChanged(committedText,
							       composedText,
							       caret);
    }

    /**
     * Get the top most view.
     * @return Top most view associated with this TextModel that has the
     * user attention.
     */
    public final View getTopView() {
        return myTopView;
    }

    /**
     *
     * @param v A view to be tested.
     * @return true if v is the default view, false otherwise.
     */
    public final synchronized boolean isTopView(View v) {
	return v.equals(myTopView) || myTopView.equals(v);
    }

    /**
     * Add an object which is dependent on this Subject. We over-ride this
     * method to implement a simple "last dependent is top view" policy. We
     * basically set topView to the last added dependent.
     * @param d The Dependent to be added.
     */
    public synchronized void addDependent(Dependent d) {
	super.addDependent(d);
        myTopView = (View) d;
    }

    /**
     * Remove a Dependent from this Subject. We over-ride this method to
     * implement a simple "last dependent is top view" policy.
     * @param d The dependent to remove.
     */
    public synchronized void removeDependent(Dependent d) {

	super.removeDependent(d);

        if (dependents == null) {
	    myTopView = getDefaultView();
	} else if (dependents instanceof Vector) {
	    myTopView = (View) ((Vector) dependents).lastElement();
	} else {
	    myTopView = (View) dependents;
	}
    }

    /**
     * We over-ride this method to set the top view to be the default view
     * if there is now top view defined. This is the safe way to initialize
     * the top view value.
     * @param The new value of the default view.
     */
    public synchronized void setDefaultView(View newView) {
        super.setDefaultView(newView);
        myTopView = getDefaultView();
    }

    /**
     * Create a new StateChangeNotification and notify dependents that a
     * change has occured.
     * @param type Type of StateChangeNotification
     * @param arg1 First argument of change
     * @param arg2 Second argument of change
     * See StateChangeNotification for complete description of type and
     * arguments.
     * @see sun.awt.otk.StateChangeNotification
     */
    protected void notifyDependents(int type,
				    int arg1, int arg2) {
	StateChangeNotification c = new StateChangeNotification(type, arg1, arg2);
	super.notifyDependents(c);
    }

    /**
     * @return Returns true if this text component knows how to edit text
     * inline, i.e. without assistance. Toolkit may choose to provide support
     * for components that cannot handle inline editing by providing simple
     * text windows for this purpose. Returns false by default.
     */
    public boolean supportsInlineEdit() {
	return false;
    }

   /**
     * Get the echo character to use when in echo mode.
     * @see java.awt.peer.TextFieldPeer
     */
    public char getEchoChar() {
        return '*';
    }

    public boolean echoCharIsSet() {
        return false;
    }

}

