/*
 * @(#)TextData.java	1.17 99/10/25
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

package sun.awt.otk;

import java.awt.TextComponent;
import java.awt.event.TextEvent;
import sun.awt.otk.SubjectImpl;

/**
 * Default implementation of TextDataInterface. This is an abstract class
 * and does not actually provide any data store for the TextComponentModel
 * data. That is left for concrete implementations.
 * <p>
 * This class is implemented as a 
 * <a href="sun.awt.otk.Subject.html">Subject</a> 
 * with lightweight
 * <a href="sun.awt.otk.TextPane.html">TextPanes</a> 
 * registered as 
 * <a href="sun.awt.otk.Dependent.html">Dependents</a>. Changes to the 
 * data buffer are made and update notifications are directly sent to 
 * dependents, which then paint themselves accordingly.
 * @see sun.awt.otk.TextComponentModel
 * @see sun.awt.otk.Subject
 * @see sun.awt.otk.Dependent
 * @see sun.awt.otk.TextPane
 */
public abstract class TextData
extends SubjectImpl
implements TextDataInterface {

    /**
     * My text component
     */
    protected TextComponentModel model;

    /**
     * Array used for storing indices into the data buffer that signify the
     * start of every line
     */
    protected int lines[];

    /**
     * Number of lines in the data buffer
     */
    protected int nrOfLines;

    /**
     * Number of unicode characters in the data buffer
     */
    protected int textLength;

    /**
     * Width of a line - this really corresponds to the width of a line
     * in components displaying this data.
     */
    protected int lineWidth;

    /**
     * Wordwrap mode
     */
    protected boolean wordwrap;

    /**
     * Create a TextData buffer.
     * @param model The TextComponentModel that this data buffer is
     * associated with
     * @param text Initialization text for the data buffer
     * @param lineWidth Default line width. Use 0 if not interested in
     * word wrapping (see
     * <A href="#setWordWrap">setWordWrap()</A>
     * )
     * @param wrap Default value for the word wrap mode
     * @see sun.awt.otk.TextComponentModel
     */
    public TextData(TextComponentModel model, String text,
		    int lineWidth, boolean wrap) {

        this.model = model;
        this.lineWidth = lineWidth;
	wordwrap = wrap;
        setText(text);
    }

    /**
     * Get the length of the data buffer.
     * @return Returns the number of unicode characters that are contained 
     * in the data buffer.
     */
    public int getLength() {
        return textLength;
    }

    /**
     * Gets the number of lines in the buffer. The return value can be
     * dependent on the word wrapping and line breaking algortihms used by
     * the concrete classes.
     */
    public int getNumberOfLines() {
        return nrOfLines;
    }

    /**
     * Get the starting index for given line.
     * @param line Line number for the starting index
     */
    public int getLineStart(int line) {
      if (isValidLine(line)) {
          return lines[line];
      } else {
	  return 0;
      }
    }

    /**
     * Get the ending index for the given line.
     * @param line Line number for the ending index
     */
    public int getLineEnd(int line) {
      if (isValidLine(line)) {
          return (line == nrOfLines - 1) ? textLength : lines[line+1] - 1;
      } else {
	  return 0;
      }
    }

    protected int [] getLinesArray() {
        int ln[] = new int[lines.length];
	System.arraycopy(lines, 0, ln, 0, lines.length);
	return ln;
    }

    /**
     * make room for given number of additional lines in array.
     * @param num Number of additional lines to grow lines array by.
     */
    protected void growLinesArray(int num) {
        int tmp[] = new int[lines.length + num];
	System.arraycopy(lines, 0, tmp, 0, lines.length);
	lines = tmp;
    }

    /**
     *  Prints out lines array.
     *  @param initStr header to print
     */
    protected void printLines(String initStr) {
        printLines(initStr, 0, nrOfLines);
    }

    /**
     *  Prints out lines array.
     *  @param initStr header to print
     *  @param e elements of lines array to print
     */
    protected void printLines(String initStr, int e) {
        printLines(initStr, 0, e);
    }

    /**
     *  Prints out lines array.
     *  @param str header to print
     *  @param s start at element s
     *  @param e elements of lines array to print
     */
    protected void printLines(String initStr, int s, int e) {
        int numPerLine = 10;
	if (initStr == null) {
	    System.err.print("\nDumping lines: lines.length:" + lines.length + ", nrOfLines:" + nrOfLines + ", textLength:" + textLength);
	} else {
	    System.err.print(initStr);
	}
	for (int i = s; i < e; i++) {
	    if (i % numPerLine == 0) {
	        System.err.print("\nlines[" + i + "]:");
	    }
	    System.err.print("  " + lines[i]);
	}
	System.err.print("\n");
    }

    /**
     * Get the line number for the given position using binary search.
     * @param pos Position (index) of character in data buffer
     */
    public int getLineNumber(int pos) {
        int i = 0, j = nrOfLines - 1;
        while (i < j) {
            int m = i + ((j - i) >> 1);
            int s = lines[m];
            int e = lines[m + 1];
            if (pos < s) {
                j = m - 1;
            } else if (pos >= e) {
                i = m + 1;
            } else {
                return m;
            }
        }
        return i;
    }

    /**
     * Append the given string at end of data buffer.
     * @param text String to append to the data buffer
     */
    public synchronized void append(String text) {
        insert(text, textLength);
    }

    /**
     * Insert the given string at given position.
     * @param text String to insert in data buffer
     * @param pos Index to insert the given string at
     */
    public synchronized void insert(String text, int pos) {
        insert(text.toCharArray(), pos);
    }

    /**
     * Set the line width in the data buffer to correspond to the width of a
     * line in component displaying this data. We only need this information
     * since the data buffer contains the line break information which is
     * also set by any wordwrap algorithms. This information can be used 
     * if the concrete classes support wordwrap, etc.
     * @param width Width in pixels for a line
     * @see sun.awt.otk.TextDataInterface#setWordWrap
     * @see sun.awt.otk.TextDataInterface#getWordWrap
     */
    public synchronized void setLineWidth(int width) {

        if (width > 0) {

	    lineWidth = width;
	    nrOfLines = updateLinesArray(0, false);
	    // send a repaint request
            StateChangeNotification c = new StateChangeNotification(
					   StateChangeNotification.DATA_CHANGE,
					   -1, nrOfLines - 1);
            notifyDependents((Object) c);
	}
    }

    /**
     * Get the current lineWidth value.
     */
    public int getLineWidth() {
        return lineWidth;
    }

    /**
     * Set the current wordwrap flag.
     * @param wrap Turn on word wrap when true, off otherwise
     */
    public synchronized void setWordWrap(boolean wrap) {
        wordwrap = wrap;
    }

    /**
     * Get the current word wrap value.
     */
    public boolean getWordWrap() {
        return wordwrap;
    }

    /**
     * Get the text from the data buffer.
     * @return Returns a character array containing the text.
     */
    public char [] getText() {
        return getStrippedText();
    }

    /**
     * Get the stated number of characters from the data buffer starting 
     * at the given position.
     * @param pos Starting index of data that needs to be returned
     * @param length Number of characters that need to be returned
     * @return Returns a character array containing the text.
     */
    public char [] getText(int pos, int length) {
        return getStrippedText(pos, length);
    }

    /**
     * Get text from the data buffer and strip CR and LF from end of data.
     * @return Returns a character array containing the stripped text.
     */
    protected char [] getStrippedText() {
        return getStrippedText(0, textLength);
    }

    /**
     * Get text from the data buffer without stripping CR and LF from end
     * of data.
     * @return Returns a character array containing the non-stripped text.
     */
    protected char [] getNonStrippedText() {
        return getNonStrippedText(0, textLength);
    }

    /**
     * Get the text from the data buffer as a string.
     * @return Returns a string containing the text.
     */
    public String getTextString() {
        return new String(getText());
    }

    /**
     * Get a string containing the stated number of characters from the
     * data buffer starting at the given position.
     * @param pos Starting index of data that needs to be returned
     * @param length Number of characters that need to be returned
     * @return Returns a string containing the text.
     */
    public String getTextString(int pos, int length) {
        return new String(getText(pos, length));
    }

    /**
     * Get the data for the line containing the given position.
     * @param pos Index of a character in the line that needs to be returned
     * @return Returns a character array containing the text for the 
     * desired line.
     */
    public char [] getLineFromPosition(int pos) {
         if (isValidPosition(pos)) {
	     int ln = getLineNumber(pos);
	     int s = getLineStart(ln);
	     return getText(s, getLineEnd(ln) - s);
	 }  else {
	     return null;
	 }
    }

    /**
     * Get the string containing the data for the line containing the 
     * given position.
     * @param pos Index of a character in the line that needs to be returned
     * @return Returns a string containing the text for the 
     * desired line.
     */
    public String getLineAsStringFromPosition(int pos) {
        char [] text = getLineFromPosition(pos);
	if (text == null) {
	    return "";
	} else {
	    return new String(text);
	}	
    }

    /**
     * Get the data for the given line.
     * @param line Line number of the line that needs to be returned
     * @return Returns a character array containing the text for the 
     * desired line.
     */
    public char [] getLineFromNumber(int lineNumber) {
        if (isValidLine(lineNumber)) {
	    int s = getLineStart(lineNumber);
	    int e = getLineEnd(lineNumber);
	    return getText(s, (e == textLength) ? e - s : e - s + 1);
	} else {
	    return null;
	}
    }

    /**
     * Get the string containing the data for the given line.
     * @param line Line number of the line that needs to be returned
     * @return Returns a string containing the text for the 
     * desired line.
     */
    public String getLineAsStringFromNumber(int lineNumber) {
        char [] text = getLineFromNumber(lineNumber);
	if (text == null) {
	    return "";
	} else {
	    return new String(text);
	}	
    }

    /**
     * Check if the given line is valid.
     * @param line Line in data buffer
     */
    public boolean isValidLine(int line) {
        return (line >= 0 && line <= nrOfLines);
    }

    /**
     * Mechanism for notifying Dependents of state changes. Also posts
     * a TEXT_VALUE_CHANGED event to the component if data was changed.
     * @param c StateChangeNotification to be sent out to Dependents
     */
    protected void sendNotifications(StateChangeNotification c) {

        notifyDependents((Object) c);
	
	if (c.getType() == StateChangeNotification.DATA_CHANGE) {
	    ((ObjectToolkit) model.getToolkit()).postEvent(
				   new TextEvent(model.getTarget(),
					         TextEvent.TEXT_VALUE_CHANGED));
	}
    }

    /**
     * Update the lines array from the data buffer.
     * @param nr Line from which to start updating
     * @param returnChangedLines Return the number of changed lines if set to
     * true, otherwise return the new nrOfLines. This implementation will 
     * always return nrOfLines, i.e. returnChangedLines=false. The "true"
     * functionality is implemented in the sub-classes.
     * @return new number of  lines in lines array (note nrOfLines is not 
     * set yet)
     */
    protected abstract int updateLinesArray(int nr,
					    boolean returnChangedLines);

    /**
     * Get the stated number of characters from the data buffer starting 
     * at the given position, and strip CR and LF from end of data.
     * @return Returns a character array containing the stripped text.
     */
    protected abstract char [] getStrippedText(int pos, int length);

    /**
     * Get the stated number of characters from the data buffer starting 
     * at the given position, and doesn't strip CR and LF from end of data.
     * @return Returns a character array containing the non-stripped text.
     */
    protected abstract char [] getNonStrippedText(int pos, int length);
}
