/*
 * @(#)TextDataInterface.java	1.5 98/11/25
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

package sun.awt.otk;

/**
 * This interface defines the contract between text components and the data
 * buffer used to store its data. It exists to isolate the TextComponentModel
 * from its data and allows under the cover changes in the data storage 
 * implementations without modifying the TextComponentModel itself. A character
 * array based data storage class is provided as a reference (see 
 * TextCharData).
 * @see sun.awt.otk.TextCharData
 * @see sun.awt.otk.TextComponentModel
 */
public interface TextDataInterface {

    /**
     * Get the length of the data buffer.
     * @return Returns the number of unicode characters that are contained 
     * in the data buffer.
     */
    int getLength();

    /**
     * Gets the number of lines in the buffer. The return value can be
     * dependent on the word wrapping and line breaking algortihms used by
     * the concrete classes.
     */
    int getNumberOfLines();

    /**
     * Get the starting index for given line.
     * @param line Line number for the starting index
     */
    int getLineStart(int line);

    /**
     * Get the ending index for the given line.
     * @param line Line number for the ending index
     */
    int getLineEnd(int line);

    /**
     * Get the line number for the given position.
     * @param pos Position (index) of character in data buffer
     */
    int getLineNumber(int pos);

    /**
     * Check if the given position is valid.
     * @param pos Position (index) of character in data buffer
     */
    boolean isValidPosition(int pos);

    /**
     * Check if the given line is valid.
     * @param line Line in data buffer
     */
    boolean isValidLine(int line);

    /**
     * Set the text in the data buffer to given text string.
     * @param text String to set the buffer value to
     */
    void setText(String text);

    /**
     * Insert the given string at given position.
     * @param text String to insert in data buffer
     * @param pos Index to insert the given string at
     */
    void insert(String text, int pos);

    /**
     * Insert the given char array at given position.
     * @param text Character array to insert in data buffer
     * @param pos Index to insert the given string at
     */
    void insert(char [] text, int pos);

    /**
     * Append the given string at end of data buffer.
     * @param text String to append to the data buffer
     */
    void append(String text);

    /**
     * Delete the data between the given indices.
     * @param start Starting index for deletion
     * @param end Ending index for deletion
     */
    void delete(int start, int end);

    /**
     * Replace the data between the given indices with the iven text.
     * @param text String to replace with
     * @param start Starting index for replacement
     * @param end Ending index for replacement
     */
    void replace(String text, int start, int end);

    /**
     * Get the text from the data buffer.
     * @return Returns a character array containing the text.
     */
    char [] getText();

    /**
     * Get the stated number of characters from the data buffer starting 
     * at the given position.
     * @param pos Starting index of data that needs to be returned
     * @param length Number of characters that need to be returned
     * @return Returns a character array containing the text.
     */
    char [] getText(int pos, int length);

    /**
     * Get the data for the line containing the given position.
     * @param pos Index of a character in the line that needs to be returned
     * @return Returns a character array containing the text for the 
     * desired line.
     */
    char [] getLineFromPosition(int pos);

    /**
     * Get the data for the given line.
     * @param line Line number of the line that needs to be returned
     * @return Returns a character array containing the text for the 
     * desired line.
     */
    char [] getLineFromNumber(int lineNumber);

    /**
     * Get the text from the data buffer as a string.
     * @return Returns a string containing the text.
     */
    String getTextString();

    /**
     * Get a string containing the stated number of characters from the
     * data buffer starting at the given position.
     * @param pos Starting index of data that needs to be returned
     * @param length Number of characters that need to be returned
     * @return Returns a string containing the text.
     */
    String getTextString(int pos, int length);

    /**
     * Get the string containing the data for the line containing the 
     * given position.
     * @param pos Index of a character in the line that needs to be returned
     * @return Returns a string containing the text for the 
     * desired line.
     */
    String getLineAsStringFromPosition(int pos);

    /**
     * Get the string containing the data for the given line.
     * @param line Line number of the line that needs to be returned
     * @return Returns a string containing the text for the 
     * desired line.
     */
    String getLineAsStringFromNumber(int lineNumber);

    /**
     * Set the line width in the data buffer to correspond to the width of a
     * line in component displaying this data. We only need this information
     * since the data buffer contains the line break information which is
     * also set by any wordwrap algorithms. This information can be used 
     * if the concrete classes support wordwrap, etc.
     * @param width Width in pixels for a line
     * @see sun.awt.otk.TextDataInterface#setWordWrap
     * @see sun.awt.otk.TextDataInterface#getWordWrap
     */
    void setLineWidth(int width);

    /**
     * Get the current lineWidth value.
     */
    int getLineWidth();

    /**
     * Set the current wordwrap flag.
     * @param wrap Turn on word wrap when true, off otherwise
     */
    void setWordWrap(boolean wrap);

    /**
     * Get the current word wrap value.
     */
    boolean getWordWrap();
}
