/*
 * @(#)TextPane.java	1.14 98/12/08
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

package sun.awt.otk;

import java.awt.*;
import java.awt.event.*;

/**
 * This interface defines a light-weight implementation of a text drawing
 * component. It is called light-weight because it doesn't own it's own
 * window. It only knows how to draw its text inside a given rectangular
 * region. There is no concept of scrollbars or other visual artifacts, which
 * are assumed to be provided by a TextComponentView that hosts this TextPane.
 * This interface hierarchy was created so that text drawing code could be
 * reused in peered and non-peered components. It is intended that TextPane
 * code in this package handle only the management of the TextPane data
 * but leave all painting to concrete implementations that will be defined
 * in packages that are look specific. This way, users of TextPane don't have
 * to re-implement most of the text handling/editing functionality and
 * can concentrate on just over-riding/implementing the paint methods. 
 * <p>
 * See
 * <a href="sun.awt.otk.TextComponentModel.html">TextComponentModel</a> 
 * for a high level architectural description of how this components works.
 * <p>
 * Synchornization of scrollbars between the TextPane and its text view is
 * done vie a the
 * <a href="sun.awt.otk.TextPaneScroller.html">TextPaneScroller</a> 
 * interface.
 * @see sun.awt.otk.TextComponentModel
 * @see sun.awt.otk.TextComponentView
 * @see sun.awt.otk.TextPaneScroller
 */
public interface TextPane extends Dependent {

    /**
     * Go to the indicated line. Scrolling occurs if necessary.
     * @param line Line to go to
     */
    void gotoLine(int line);

    /**
     * Go to the indicated position. Scrolling occurs if necessary.
     * @param pos Index of character to go to
     */
    void gotoPosition(int pos);

    /**
     * This method either moves the cursor or the textpane up one line.
     * @param absoluteMode When true, the whole textPane is scrolled up
     * by one line. Otherwise, only the cursor is moved up one line.
     * @return Returns true if the lineUp will cause the cursor to move 
     * outside the visible text. In this case, it is expected that the 
     * containing view for this textpane will issue a lineUp again with
     * absoluteMode set to true, which will then cause the appropriate
     * scroll to take place. If the cursor stays within the visible text,
     * then we return false. This return value is used by the container to
     * synchronize with any scrollbars that may exist.
     */
    boolean lineUp(boolean absoluteMode);

    /**
     * This method either moves the cursor or the textpane down one line.
     * @param absoluteMode When true, the whole textPane is scrolled down
     * by one line. Otherwise, only the cursor is moved down one line.
     * @return Returns true if the lineDown will cause the cursor to move 
     * outside the visible text. In this case, it is expected that the 
     * containing view for this textpane will issue a lineDown again with
     * absoluteMode set to true, which will then cause the appropriate
     * scroll to take place. If the cursor stays within the visible text,
     * then we return false. This return value is used by the container to
     * synchronize with any scrollbars that may exist.
     */
    boolean lineDown(boolean absoluteMode);

    /**
     * Move up one page.
     */
    void pageUp();

    /**
     * Move down one page.
     */
    void pageDown();

    /**
     * Move to the very end of the data buffer.
     */
    void end();

    /**
     * Move to the end of the current line.
     */
    void endOfLine();

    /**
     * Move to the beginning of the data buffer.
     */
    void home();

    /**
     * Move to the start of the current line.
     */
    void homeOfLine();

    /**
     * Move to the left by one character.
     */
    void moveLeftChar();

    /**
     * Move to the right by one character.
     */
    void moveRightChar();

    /**
     * Select one white-space delimited word. White-space is implementation
     * defined.
     * @param pos Index to use as the basis for choosing the word to select
     */
    void selectWord(int pos);

    /**
     * Select one white-space delimited line. White-space is implementation
     * defined.
     * @param pos Index to use as the basis for choosing the line to select
     */
    void selectLine(int pos);

    /**
     * Select all the text.
     */
    void selectAll();

    /**
     * Select the text between the given range.
     * @param s Starting index into text
     * @param e Ending index into text
     * @see java.awt.peer.TextComponentPeer
     */
    void select(int selStart, int selEnd);

    /**
     * Callback used by our Subject to notify us of changes.
     * @param Subject Which subject the notification came from
     * @param arg Contents of the notification
     */
    void updateDependent(Subject subj, Object arg);

    /**
     * Set the editable flag.
     * @param editable Value to set the flag to
     * @see java.awt.peer.TextComponentPeer
     */
    void setEditable(boolean editable);

    /**
     * Get the start of the current selection.
     * @return Returns the index of the selection start
     * @see java.awt.peer.TextComponentPeer
     */
    int getSelectionStart();

    /**
     * Get the end of the current selection.
     * @return Returns the index of the selection end
     * @see java.awt.peer.TextComponentPeer
     */
    int getSelectionEnd();

    /**
     * Set the current caret position.
     * @param pos New position for the caret
     * @see java.awt.peer.TextComponentPeer
     */
    void setCaretPosition(int pos);

    /**
     * Get the current caret position.
     * @return Returns the index of the caret position
     * @see java.awt.peer.TextComponentPeer
     */
    int getCaretPosition();

    /**
     * Get the first visible line number.
     */
    int getFirstVisibleLine();

    /**
     * Get the last visible line number.
     */
    int getLastVisibleLine();

    /**
     * Check to see if line is currently visible.
     * @param lineNo Line number to check for visibility
     * @return Returns true if the line is currently visible, false otherwise
     */
    boolean isLineVisible(int lineNo);

    /**
     * Get the number of lines in the textpane.
     */
    int linesInWindow();

    /**
     * Set the bounds of the textpane.
     * @param x New x-coordinate
     * @param y New y-coordinate
     * @param width New width
     * @param height New height
     */
    void setBounds(int x, int y, int width, int height);

    /**
     * Get the current bounds of the textpane.
     */
    Rectangle getBounds();

    /**
     * Get the current size of the textpane.
     */
    Dimension getSize();

    /**
     * Get the minimum size of the textpane for the given rows and columns.
     * @param rows Number of rows to calculate minimum size
     * @param cols Number of columns to calculate minimum size
     */
    Dimension getMinimumSize(int rows, int cols);

    /**
     * Set the font to the given value.
     * @param f New font
     * @param fm New FontMetrics
     */
    void fontChanged(Font f, FontMetrics fm);

    /**
     * Get the current color policy.
     */
    ColorPolicy getColorPolicy();  

    /**
     * Repaint the textpane.
     */
    void paintComponent();

    /**
     * Repaint the textpane using the given Graphic.
     * @param g Graphics context to use for repainting
     */
    void paintComponent(Graphics g);

    /**
     * Paint the lines in the given range (inclusive). The textpane is free
     * to only paint lines that are visible.
     * @param s Index of first line
     * @param e Index of last line
     */
    void paintLines(int s, int e);

    /**
     * Handle mouse events.
     * @param   v view where an event occurred
     * @param   id see "generic" handleEvent for description
     * @param   when see "generic" handleEvent for description
     * @param   modifiers see "generic" handleEvent for description
     * @param   x see "generic" handleEvent for description
     * @param   y see "generic" handleEvent for description
     * @param   arg1 see "generic" handleEvent for description
     * @param   arg2 see "generic" handleEvent for description
     * @return <code>true</code> if this event was handled and 
     *         <code>false</code> - otherwise
     * @see sun.awt.otk.InputController#handleEvent(View, AWTEvent)
     */
    boolean handleMouseEvent(int id,  long when, int modifiers,
			     int x, int y, int arg1, int arg2);

    /**
     * Handle focus changes.
     * @param focusGained True when focus gained, false otherwise
     */
    void setFocus(boolean focusGained);

    /**
     * Get the maximum possible string width that would fit in this textpane.
     * This value can be used by word-wrapping algorithms to wrap text for
     * this textpane.
     * @return Returns the maximum width in pixels
     */
    int getMaxPossibleStringWidth();

    /**
     * @return Returns true if this text component knows how to edit text
     * inline, i.e. without assistance. Toolkit may choose to provide support
     * for components that cannot handle inline editing by providing simple
     * text windows for this purpose.
     * @see #setInlineEditMode()
     */
    boolean supportsInlineEdit();

    /**
     * Set the inline edit mode. When set to true, the component will edit
     * inline.
     * @param editMode New inline edit mode value
     */
    void setInlineEditMode(boolean editMode);

    /**
     * Get the y-coordinate to draw text on for the first line.
     */
    int lineNumberToY(int ln);

    /**
     * Copy the current selection to the clipboard.
     * @return Returns true if operation succeeded, false otherwise.
     */
    boolean copyToClipboard();

    /**
     * Cut the current selection to the clipboard.
     * @return Returns true if operation succeeded, false otherwise.
     */
    boolean cutToClipboard();

    /**
     * Paste the contents of the clipboard onto the current selection.
     * @return Returns true if operation succeeded, false otherwise.
     */
    boolean pasteFromClipboard();

    /**
     * Paste the contents of the string onto the current selection.
     * @param str String to paste
     * @return Returns true if operation succeeded, false otherwise.
     */
    boolean pasteOnSelection(String str);

    /**
     * Backspace, i.e. either delete selected text or delete backwards one
     * character.
     * @return Returns true if operation succeeded, false otherwise.
     */
    boolean backspace();

    /**
     * Invoked when the text entered through an input method has changed.
     * This is the mechanism for the Input Method Framework (IMF) to
     * communicate to text components.
     * @param committedText IMF text that has already been committed
     * @param composedText IMF text that has not been committed yet
     * @param caret Location of caret in the IMF text.
     */
    void inputMethodTextChanged(String committedText,
				String composedText,
				int caret);

    /**
     * Invoked when the caret within composed text has changed.
     * @param committedText IMF text that has already been committed
     * @param composedText IMF text that has not been committed yet
     * @param caret Location of caret in the IMF text.
     */
    void caretPositionChanged(String committedText,
			      String composedText,
			      int caret);

    /**
     * @return Returns true if currently in compose mode, false otherwise
     */
    boolean isComposing();
}
