/*
 * @(#)TextPaneImpl.java	1.53 99/08/19
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

package sun.awt.otk;

import java.awt.*;
import java.awt.event.*;
import java.awt.datatransfer.*;
import java.io.Reader;
import java.text.BreakIterator;
import sun.porting.toolkit.ToolkitEventHandler;

/**
 * Implementation of the TextPane interface.
 */
public abstract class TextPaneImpl
implements TextPane, ClipboardOwner {

    /**
     * When true, a pointer pressed on text results in the surrounding white
     * space delimited word to be selected. Once selected, another click
     * on the word will place the cursor at the nearest selection point.
     */
    protected static boolean wordSelection;

    static {
	// set wordSelection from system property
	if (Boolean.getBoolean("sun.awt.otk.textWordSelectionOff")) {
	    wordSelection = false;
	} else {
	    wordSelection = true;
	}
    }

    /** My x-coordinate */
    protected int x;
    /** My y-coordinate */
    protected int y;
    /** My width */
    protected int width;
    /** My height */
    protected int height;

    protected BasicView view;              // my view
    protected TextComponentModel model;    // my AWT target
    protected ObjectToolkit tk;            // my toolkit
  
    protected ColorPolicy colors;          // my color policy

    protected TextData data;               // my data buffer

    /**
     * Area that we don't draw text in. Generally this will
     * be used to set the clipping rect when drawing text and selecting.
     */
    protected Insets insets;

    /**
     * Padding from the left that is used for deteriming the start of text.
     * Left pad starts after the insets have been accounted for.
     */
    protected int leftPad;

    /**
     * Padding from the right that is used for deteriming the end of text.
     * Right pad starts after the insets have been accounted for.
     */
    protected int rightPad;

    /** Are we currently editable? */
    protected boolean editable;
    /** Are we currently selected, i.e. in focus? */
    protected boolean selected;

    /**
     * Start of currently selected text.
     */
    protected int selStart;

    /**
     * End of currently selected text.
     */
    protected int selEnd;

    /**
     * Location of caret.
     */
    protected int caretCol;

    /**
     * Index of top line that is visible, this is always maintained.
     */
    protected int topLine;

    /**
     * Number of pixels we've moved to the right as result of horizontal
     * scrolling.
     */
    protected int movedRight;

    /**
     * Bounds of the rectangle that the text is blatted into. This is the
     * actual text drawing region with the insets subtracted.
     */
    protected Rectangle textRect;

    /**
     * Default character used for calculating width of a character when
     * tabbing. A tab is defined as tabChar * <a href="#tabWidth">tabWidth</a>.
     * Override in concrete implementations.
     */
    protected char tabChar = 'n';

    /**
     * Default number of tabChar used when calculating widht of a tab.
     * A tab is defined as tabWidth * <a href="#tabChar">tabChar</a>.
     * Override in concrete implementations.
     */
    protected int tabWidth = 8;

    /**
     * Default character used for calculating width of a character.
     * Override in concrete implementations.
     */
    protected int defaultTextWidthChar = 'n';

    /**
     * The current mouse position. Cached to handle dragging of mouse.
     */
    protected int mpos;

    /**
     * Start of last selected word. Maintained only when in
     * <a href="#wordSelection">wordSelection</a> mode.
     */
    protected int wordStart;

    /**
     * End of last selected word. Maintained only when in
     * <a href="#wordSelection">wordSelection</a> mode.
     */
    protected int wordEnd;

    /**
     * Line number of last selected word. Maintained only when in
     * <a href="#wordSelection">wordSelection</a> mode.
     */
    protected int wordLine;

    /**
     * Whether inline editing is supported or not. Default value is true
     * since the peers know how to edit inline.
     */
    protected boolean alwaysInlineEdit = true;

    /**
     * Create a new TextPane instance.
     * A TextPane maintains an Insets that is used for clipping when
     * drawing text or selection rectangles. This is represented by the
     * inner rectangle below. There are two lines of text in the TextPane.
     * The left pad is the space between the first character that can
     * be rendered on a line (the "T" on the first line), note though that
     * any underlining can be drawn from the insets onward. The right pad
     * represents the analogous space on the right of the line, and text
     * cannot be drawn this many pixels to the right of the insets. All
     * values are in pixels.
     * <p>
     * +-----------------------------------------------------+
     * |+---------------------------------------------------+|
     * || This test string shows left and right padding, as ||
     * ||---------------------------------------------------||
     * || it is used in textpanes.                          ||
     * ||---------------------------------------------------||
     * ||                                                   ||
     * |+---------------------------------------------------+|
     * +-----------------------------------------------------+
     * <p>
     * For the partial scrolling case, i.e. half line scroll windows on top
     * and bottom, we maintain the insets.
     * <p>
     * @param view Our view
     * @param model Our data model
     * @param tk Our toolkit
     * @param insets The insets used for this textpane. These define the
     * outer boundaries that we use for clipping before drawing text or
     * selections.
     * @param left Left padding used in addition to insets before drawing text
     * @param right Rigth pad used along with insets before drawing text
     */
    public TextPaneImpl(BasicView view, TextComponentModel model,
                        ObjectToolkit tk, ColorPolicy colors,
			Insets insets,int left, int right) { 
        this.view = view;
	this.model = model;
	this.tk = tk;
	this.data = model.getDataModel();
	this.insets = new Insets(insets.top, insets.left, insets.bottom, insets.right);
	this.leftPad = left;
	this.rightPad = right;
	setCaretPosition(model.getCaretPosition());
	this.selStart = model.getSelectionStart();
	this.selEnd = model.getSelectionEnd();
	this.colors = colors;
	                        
	// initialize attributes that are common to both single and multiline
	topLine = 0;
	movedRight = 0;

	// Initialize all the bounds related variables
	setBounds(0, 0, 0, 0);
    }

    /**
     * Set the bounds of the textpane. Sets textRect to the
     * correct insets-adjusted value.
     * @param x New x-coordinate
     * @param y New y-coordinate
     * @param width New width
     * @param height New height
     */
    public void setBounds(int x, int y, int width, int height) {

        //System.err.println("+TextPaneImpl.setBounds: x=" + x + " y=" + y + " w=" + width + " h=" + height);

        this.x = x;
	this.y = y;
	this.width = width;
	this.height = height;

	// set the textRect
	textRect = new Rectangle(x + insets.left, y + insets.top,
				 width - insets.left - insets.right,
				 height - insets.top - insets.bottom);

	// topLine and movedRight should be recalculated
	int l = linesInWindow();
	int nrOfLines = data.getNumberOfLines();
	int e = Math.min(nrOfLines - 1, (topLine + l) - 1);
	if (nrOfLines - topLine < l && topLine != 0) {
	    topLine = Math.max(0, nrOfLines-l);
	    e = Math.max(0, Math.min(nrOfLines - 1, (topLine + l) - 1));
	}

	int w = getMaxPossibleStringWidth();
	int max = maxLength(topLine, e);
	if (max - movedRight < w && movedRight != 0) {
	    movedRight = Math.max(0, max - w);
	}

	//System.err.println("-TextPaneImpl.setBounds: textRect=" + textRect + " lW=" + l + " #lines=" + nrOfLines + " w=" + w + " max=" + max + " top=" + topLine + " e=" + e);
    }

    /**
     * Set the font to the given value. Causes a repaint.
     * @param f New font
     * @param fm New FontMetrics
     */
    public void fontChanged(Font f, FontMetrics fm) {
        // recalculate all sizing values
        Rectangle bounds = getBounds();
        setBounds(bounds.x, bounds.y, bounds.width, bounds.height);
    }

    /**
     * Get the current bounds of the textpane.
     */
    public Rectangle getBounds() {
        return new Rectangle(x, y, width, height);
    }

    /**
     * Get the current size of the textpane.
     */
    public Dimension getSize() {
        return new Dimension(width, height);
    }

    /**
     * Get the current color policy.
     */
    public ColorPolicy getColorPolicy(){
	return colors;
    }

    /**
     * Callback used by our data buffer (TextData) of changes via 
     * StateChangeNotification objects. We only repaint the changed lines
     * in the visible range.
     * @param Subject Which subject the notification came from
     * @param arg Contents of the notification
     * @see sun.awt.otk.TextData#sendNotifications
     */
    public void updateDependent(Subject subj, Object arg) {

        if (arg != null && arg instanceof StateChangeNotification) {
	    StateChangeNotification c = (StateChangeNotification) arg;
	    if (c.getType() == StateChangeNotification.DATA_CHANGE) {

	        if (c.getStartIndex() < 0) {                        // repaint requested
		    setCaretPosition(data.getLength());
		    paintComponent();
		    resizeScrollbars();
		} else {
		    paintChangedLines(c.getStartIndex(), c.getEndIndex());
		}
	    }
	} else {
	    Graphics g = null;
	    try {
	        g = view.getGraphics();
	        paintComponent(g);
	    } finally {
	      if (g != null) {
		  g.dispose();
	      }
	    }
	}
    }

    /**
     * Set the editable flag.
     * @param editable Value to set the flag to
     * @see java.awt.peer.TextComponentPeer
     */
    public synchronized void setEditable(boolean editable) {
        this.editable = editable;
    }

    /**
     * Notification mechanism of focus changes. We don't cause a repaint
     * as the view will force a paintComponent anyways.
     * WARNING: This may be a wrong assumption!!
     * @param focusGained True when focus gained.
     */
    public void setFocus(boolean focusGained) {

        selected = focusGained;

	if (!focusGained) {

	    // reset word selection on focus lost
	    wordStart = 0;
	    wordEnd = 0;

	    // reset compose mode on focuslost and delete composed text - we
	    // presume that inputMethod will send the text again if it wants
	    if (composeMode) {
	        data.delete(oldComposeStart, oldComposeEnd);
	        composeMode = false;
	    }
	}
    }

    /**
     * Get the start of the current selection.
     * @return Returns the index of the selection start
     * @see java.awt.peer.TextComponentPeer
     */
    public synchronized int getSelectionStart() {
        return selStart;
    }

    /**
     * Get the end of the current selection.
     * @return Returns the index of the selection end
     * @see java.awt.peer.TextComponentPeer
     */
    public synchronized int getSelectionEnd() {
        return selEnd;
    }

    /**
     * Set the current caret position. We don't differentiate between
     * caret position and selection start at this time.
     * @param pos New position for the caret
     * @see java.awt.peer.TextComponentPeer
     */
    public synchronized void setCaretPosition(int pos) {
	caretCol = pos - data.getLineStart(data.getLineNumber(pos));
	selStart = selEnd = pos;
    }

    /**
     * Get the current caret position. We don't differentiate between
     * selection start and caret position at this point.
     * @return Returns the index of the caret position
     * @see sun.awt.otk.setCaretPosition
     * @see java.awt.peer.TextComponentPeer
     */
    public synchronized int getCaretPosition() {
	return selStart;
    }

    /**
     * Get the last visible line number.
     */
    public synchronized int getFirstVisibleLine() {
        return topLine;
    }

    /**
     * Get the last visible line number.
     */
    public synchronized int getLastVisibleLine() {
        int li = linesInWindow();
        if (li == 1) {
	    return 0;
	} else {
	    return topLine + li - 1;
	}
    }

    /**
     * Get the number of lines in the textpane.
     */
    public synchronized boolean isLineVisible(int lineNo) {
	if (lineNo >= topLine && lineNo <= topLine + linesInWindow()) {
	    return true;
	} else {
	    return false;
	}
    }

    /**
     * Go to the indicated line. Scrolling occurs if necessary.
     * @param line Line to go to
     */
    public void gotoLine(int line) {
	gotoPosition(data.getLineStart(line));
    }

    /**
     * Go to the indicated position. Scrolling occurs if necessary.
     * @param pos Index of character to go to
     */
    public void gotoPosition(int pos) {
        if (data.isValidPosition(pos)) {
	    scrollAbsolute(pos, true);
	}
    }

    /**
     * Move to the very end of the data buffer.
     */
    public void end() {
        int textLength = data.getLength();
	select(textLength, textLength);
	scrollTo(selStart, true);
    }

    /**
     * Move to the end of the current line.
     */
    public void endOfLine() {
        int line = data.getLineNumber(selStart);
	int posright = 0;

	// if last line, then no need to call xPos
	if (line >= (data.getNumberOfLines() - 1)) {
	    posright = data.getLength();
	} else {
	    posright = xPos(0, line+1) - 1;
	}
	select(posright, posright);
	scrollTo(selStart, false);
    }

    /**
     * Move to the beginning of the data buffer.
     */
    public void home() {
        select(0, 0);
	scrollTo(selStart, true);
    }

    /**
     * Move to the start of the current line.
     */
    public void homeOfLine() {
        int posleft = xPos(0, data.getLineNumber(selStart));
	select(posleft, posleft);
	scrollTo(selStart, false);
    }


    /**
     * This method either moves the cursor or the textpane up one line.
     * @param absoluteMode When true, the whole textPane is scrolled up
     * by one line. Otherwise, only the cursor is moved up one line.
     * @param dir Direction to move in, +1 for down, -1 for up.
     * @return Returns true if the lineUp will cause the cursor to move 
     * outside the visible text. In this case, it is expected that the 
     * containing view for this textpane will issue a lineUp again with
     * absoluteMode set to true, which will then cause the appropriate
     * scroll to take place. If the cursor stays within the visible text,
     * then we return false. This return value is used by the container to
     * synchronize with any scrollbars that may exist.
     */
    private boolean moveLine(boolean absoluteMode, int dir) {
	int nrOfLines = data.getNumberOfLines();
	int newLine   = dir + data.getLineNumber(selStart);

	if ((newLine < 0) || (newLine >= nrOfLines)) {
	    return false;
	}

	int li = linesInWindow();
	int newTop  = topLine + ((dir < 0) ? -1 : li);

	int posn = Math.min(data.getLineStart(newLine) + caretCol,
			    data.getLineEnd(newLine));
	
	boolean ret = false;

	if (absoluteMode && ((newTop >= 0) && (newTop <= nrOfLines))) {
	    posn = Math.min(data.getLineStart(newTop) + caretCol,
			    data.getLineEnd(newTop));
	    scrollTo(selStart, false);
	} else if ((newLine >= topLine) && (newLine < topLine + li)) {
	    scrollTo(selStart, false);
	} else {
	    ret = true;
	}

	/* caretCol will be changed by select.  We need to preserve */
	/* the original value so that a subsequent moveLine goes to */
	/* the same target column that this call did.               */
	int targetCol = caretCol;
	positionCursor(posn);
	caretCol = targetCol;

	// the setValue on the scrollbar will generate lineDown
	return ret;
    }

    private void movePage(int dir) {
	int ln = data.getLineNumber(selStart);
	int newLine = ln + dir;
	int nrOfLines = data.getNumberOfLines();
	if ((newLine >= nrOfLines) || (newLine < 0)) {
	    return;
	}

	int page = linesInWindow();
	int showLn, newSelLn;

	if (dir > 0) {
	    /* page down */
	    showLn = Math.min(nrOfLines - 1, (topLine + page - 1) + page);
	    newSelLn = Math.min(nrOfLines - 1,
				(showLn - page + 1) + (ln - topLine));
	} else {
	    /* page up */
	    showLn = Math.max(0, topLine - page);
	    newSelLn = Math.max(0, showLn + (ln - topLine));
	}

	int lineStart = data.getLineStart(newSelLn);
	int newSelPos = Math.min(lineStart + caretCol, 
				 data.getLineEnd(newSelLn));

	lineStart = data.getLineStart(showLn);
	int showPos = Math.min(lineStart + caretCol, data.getLineEnd(showLn));

	/* caretCol will be changed by select.  We need to preserve */
	/* the original value so that a subsequent movePage goes to */
	/* the same target column that this call did.               */
	int targetCol = caretCol;
	positionCursor(newSelPos);
	caretCol = targetCol;

	scrollTo(showPos, true);
    }

    /**
     * This method either moves the cursor or the textpane up one line.
     * @param absoluteMode When true, the whole textPane is scrolled up
     * by one line. Otherwise, only the cursor is moved up one line.
     * @return Returns true if the lineUp will cause the cursor to move 
     * outside the visible text. In this case, it is expected that the 
     * containing view for this textpane will issue a lineUp again with
     * absoluteMode set to true, which will then cause the appropriate
     * scroll to take place. If the cursor stays within the visible text,
     * then we return false. This return value is used by the container to
     * synchronize with any scrollbars that may exist.
     */
    public boolean lineUp(boolean absoluteMode) {
	return moveLine(absoluteMode, -1);
    }

    /**
     * This method either moves the cursor or the textpane down one line.
     * @param absoluteMode When true, the whole textPane is scrolled down
     * by one line. Otherwise, only the cursor is moved down one line.
     * @return Returns true if the lineDown will cause the cursor to move 
     * outside the visible text. In this case, it is expected that the 
     * containing view for this textpane will issue a lineDown again with
     * absoluteMode set to true, which will then cause the appropriate
     * scroll to take place. If the cursor stays within the visible text,
     * then we return false. This return value is used by the container to
     * synchronize with any scrollbars that may exist.
     */
    public boolean lineDown(boolean absoluteMode) {
	return moveLine(absoluteMode, +1);
    }

    /**
     * Move up one page.
     */
    public void pageUp() {
	movePage(-1);
    }

    /**
     * Move down one page.
     */
    public void pageDown() {
	movePage(+1);
    }

    /**
     * Move to the right by one character.
     */
    public void moveRightChar() {
        handleRightArrowKey(selStart);
    }

    /**
     * Move to the left by one character.
     */
    public void moveLeftChar() {
        handleLeftArrowKey(selStart);
    }

    /**
     * Whether we support inline editing, i.e. handle text input without 
     * assistance from things like virtual keyboards, etc. Default 
     * implementation always tries to handle editing itself.
     * @return True if we can handle inline editing, false otherwise.
     */
    public boolean supportsInlineEdit() {
        return alwaysInlineEdit;
    }

    /**
     * Set the inline edit mode. When set to true, the component will edit
     * inline.
     * @param editMode New inline edit mode value
     */
    public void setInlineEditMode(boolean editMode) {
        alwaysInlineEdit = editMode;
    }

    /**
     * Paint lines that have changed in the given range. Default 
     * implementation doesn't check if lines are in visible range or not.
     * Over-ride for complex components.
     * @param start Start line number
     * @param end End line number
     */
    protected void paintChangedLines(int start, int end) {
        paintLines(start, end);
    }

    /**
     * Return the y co-ordinate of the given line number.
     * @param ln Line number to use for return value
     * @return Returns number of pixels from top of textpane
     */
    public int lineNumberToY(int ln) {
	return ln2y(view.getGraphics(), ln);
    }

    /**
     * Return the y co-ordinate of the given line number.
     * @param g Current graphics context
     * @param ln Line number to use for return value
     * @return Returns number of pixels from top of textpane
     */
    protected int ln2y(Graphics g, int ln) {
        FontMetrics fm = g.getFontMetrics();
	return ln2y(fm, ln);
    }

    /**
     * Return the y co-ordinate of the given line number.
     * @param fm Current FontMetrics
     * @param ln Line number to use for return value
     * @return Returns number of pixels from top of textpane
     */
    protected int ln2y(FontMetrics fm, int ln) {
        int y = textRect.y + ((ln - topLine) * getLineHeight(fm));
	return y;
    }

    /**
     * Return the y co-ordinate for the baseline for line ln.
     * @param g Current graphics context
     * @param ln Line number to use for return value
     * @return Returns number of pixels from top of textpane
     */
    protected int ln2baseline(Graphics g, int ln) {
        FontMetrics fm = g.getFontMetrics();
	return ln2baseline(fm, ln);
    }

    /**
     * Return the y position for the baseline for line ln.
     * @param fm Current FontMetrics
     * @param ln Line number to use for return value
     * @return Returns number of pixels from top of textpane
     */
    protected int ln2baseline(FontMetrics fm, int ln) {
	int baseline = getBaseline(fm);
	int minHeight = getLineHeight(fm);
	int y;
	if (ln - topLine <= 0) {
	    y = textRect.y + baseline;
	} else {
	    y = textRect.y + ((ln - topLine) * minHeight) + baseline;
	}
	/*System.err.println("ln2baseline(): ln=" + ln + ", top=" + topLine +
			   ", base=" + baseline + ", ymargin=" + ymargin +
			   ", y=" + y);*/

	return y;
    }

    /**
     * Calculate the minimum height of the text.
     * @param fm Current FontMetrics
     */
    protected int getLineHeight(FontMetrics fm) {
        return fm.getHeight();
    }

    /**
     * Get the baseline for text to be drawn on.
     * @param fm Current FontMetrics
     * @return Returns number of pixels from top of line, i.e. 0-based
     */
    protected int getBaseline(FontMetrics fm) {
	return ((fm.getHeight() +
		 fm.getMaxAscent() -
		 fm.getMaxDescent() + 1
		) / 2);
    }

    /**
     * Get the number of lines in the textpane.
     */
    public int linesInWindow() {
        return linesInWindow(view.getMetrics());
    }

    /**
     * Get the number of lines in the textpane.
     * @param fm Current FontMetrics
     */
    protected int linesInWindow(FontMetrics fm) {

	int l = 0;
	if (fm != null) {
	    int htotal = textRect.height;
	    int minHeight = getLineHeight(fm);

	    if (htotal <= 0) {
	        l = 0;
	    } else if (htotal < minHeight) {
	        l = 1;
	    } else {
	        l = htotal / minHeight;
	    }
	}

	return l;
    }

    /**
     * Return the x coordinate of the given character.
     * @param pos Position of character in string
     * @param buff Buffer containing the data string
     */
    protected int pos2x(int pos,  char [] buff) {
        return pos2x(view.getMetrics(), pos, buff);
    }

    /**
     * Return the x coordinate of the given character.
     * @param fm current FontMetrics
     * @param pos Position of character in string
     * @param str String containing the data string
     */
    protected int pos2x(FontMetrics fm, int pos, String str) {
        return pos2x(fm, pos, str.toCharArray());
    }

    /**
     * Return the x coordinate of the given character.
     * @param fm current FontMetrics
     * @param pos Position of character in buffer
     * @param buff Buffer containing the data string
     */
    protected int pos2x(FontMetrics fm, int pos, char [] buff) {

	//System.out.println("\nInside pos2x pos is "+pos+" buff is "+buff); 

        int x = textRect.x + leftPad;
        if (buff.length <= 0){
	   //System.out.println("pos2x returning (0 char) "+x); 
           return x;
	}

	if (((TextComponentModel)model).echoCharIsSet()){
            char echoChar = ((TextFieldModel)model).getEchoChar();
            x+=(fm.charWidth(echoChar) * pos);
	    //System.out.println("pos2x returning (echochar)"+x); 
	    return x;
        } else  {


            int widths[] = fm.getWidths();
            int tx = x;
            for (int i = 0; i < pos && i < buff.length; i++) {

                // search for tabs
                if (buff[i] == '\t') {
                   // keep track of the x location for use above
                   x = tx + ((((x/widths[tabChar])/tabWidth + 1) *
                           tabWidth) * widths[tabChar]);
                   } else {
                     // masking against 0xFF ensures that the index is
                     // between 0 and 255
                     if (buff[i] > 255) {
                        x += fm.charWidth(buff[i]);
                    } else {
                      x += widths[buff[i]];
                    }
                   }
           }
	   //System.out.println("pos2x returning (+ char) "+x); 
	   return x;
	}
    }

    /**
     * Return the position on the line "ln" given the old xPosition.
     * @param xo position in text array on some line
     * @param ln linenumber
     * @return Returns position in text array
     */
    protected int xPos(int xo, int ln) {
       if (((TextComponentModel)model).echoCharIsSet()){
            char echoChar = ((TextFieldModel)model).getEchoChar();
            FontMetrics fm = view.getMetrics();
            int w = fm.charWidth(echoChar);
            int len = data.getLength();
            int div = xo / w;
            return (div > len) ? len : div;
        } else {

           int xmargin = textRect.x + leftPad;
           FontMetrics fm = view.getMetrics();
           int x = xmargin, widths[] = fm.getWidths();
           char [] text;
           if (this instanceof SingleLineTextPane) {
               text = data.getText();
           } else {
            text = data.getLineFromNumber(ln);
           }

           // Loop through text until we arrive at the right posn.
           int posn;
           for (posn = 0; text != null && posn < text.length && x < xo; posn++
) {

               // Search for tabs
               if (text[posn] == '\t') {
                  x = xmargin + ((((x/widths[tabChar])/tabWidth + 1) *
                                tabWidth) * widths[tabChar]);
               } else {
                 // masking against 0xFF ensures that the index
                 // is between 0 and 255
                 if (text[posn] > 255) {
                    x += fm.charWidth(text[posn]);
                  } else {
                    x += widths[text[posn]];
                  }
              }
           }

           if (posn == text.length)
              return data.getLineEnd(ln);
           else
              return posn + data.getLineStart(ln);
      }

    }

    /*
     * Return the current line number from the x and y co-ordinates.
     * @param x X co-ordinate
     * @param y Y co-ordinate
     */
    protected int xy2pos(int x, int y) {

        if (this instanceof SingleLineTextPane) {
	    return xPos(x + movedRight, 0);
	}

        int ln;
        int endl = data.getLineNumber(selEnd);
        FontMetrics fm = view.getMetrics();
	int nrOfLines = data.getNumberOfLines();

        if (y < 0) {
            if (topLine - 1 >= 0) {
                ln = topLine - 1;
            } else {
                return 0;
            }
        } else if (y > (textRect.y + textRect.height)) {
            if (endl < nrOfLines - 1) {
                ln = endl + 1;
            } else {
                return data.getLength();
            }
        } else {
            ln = topLine;
	    int minTextH = getLineHeight(fm);
            int ys = textRect.y + minTextH;
            while ((y > ys) && (ln < nrOfLines - 1)) {
                ys += minTextH;
                ln++;
            }
            if (y > ys) {
                // this is the case where the y location is inside the
                // TextArea but not on the text - position caret to end
                // of text
                return data.getLength();
            }
        }

	return xPos(x + movedRight, ln);
    }

    /**
     * Repaint the textpane.
     */
    public void paintComponent() {
        Graphics g = null;
	try {
	    g = view.getGraphics();
	    paintComponent(g);
	} finally {
	    if (g != null) {
	        g.dispose();
	    }
	}
    }

    /**
     * Paint the lines in the given range (inclusive).
     * @param s Index of first line
     * @param e Index of last line
     */
    public void paintLines(int s, int e) {
        Graphics g = null;
	try {
	    g = view.getGraphics();
	    paintLines(g, s, e);
	} finally {
	    if (g != null) {
	        g.dispose();
	    }
	}
    }

    /**
     * Draw the string to the screen as the first line in textpane.
     * @param str String to be rendered
     */
    protected void paintLine(String str, boolean showCursor) {
        Graphics g = null;
	try {
	    g = view.getGraphics();
	    paintLine(g, str, showCursor);
	} finally {
	    if (g != null) {
	        g.dispose();
	    }
	}
    }

  
    /**
     * Draw the string to the screen.
     * @param g Graphics context
     * @param line line number to be rendered
     */
    protected void paintLine(Graphics g, int line) {
        FontMetrics fm = g.getFontMetrics();
	if (isLineVisible(line)) {
	    paintLine(g, fm, g.getFont(),
		      getFgColor(), getBgColor(),
		      ln2y(fm, line), ln2baseline(fm, line),
		      selected && cursorOnThisLine(line, selStart, selEnd), 
		      data.getLineStart(line),
		      data.getLineAsStringFromNumber(line));
	}
    }
  
    /**
     * Draw the string to the screen as the first line in textpane.
     * @param g Graphics context
     * @param str String to be rendered
     */
    protected void paintLine(Graphics g, String str, boolean showCursor) {
        FontMetrics fm = g.getFontMetrics();
	paintLine(g, fm, g.getFont(),
		  getFgColor(), getBgColor(),
		  ln2y(fm, 0), ln2baseline(fm, 0), showCursor, 0, str);
    }

    /**
     * Draw the line to the screen. This method draws the basic portions of
     * the line but uses the specialized paintString method to actually 
     * render the text string. This method MUST be implemented by the look
     * specific concrete instance of this textpane.
     * @param g Graphics context
     * @param fm FontMetrics for the text
     * @param f Font for the text
     * @param fg Foreground color
     * @param bg Background color
     * @param y Y co-ordinate of start of the selection rectangle
     * @param baseline Y co-ordinate for drawing the text
     * @param showCursor Whether to draw the cursor or not
     * @param startPos Index of first character of str in data buffer
     * @param str String to be rendered
     */
    protected abstract void paintLine(Graphics g, FontMetrics fm, Font f,
				      Color fg, Color bg, int y, int baseline,
				      boolean showCursor, int startPos,
				      String str);

    /**
     * Paint text in the given range as selected text.
     * @param g Current graphics context
     * @param start Starting index of selected text
     * @param end Ending index of selected text
     */
    void paintRange(Graphics g, int start, int end) {

        //System.err.println("paintRange: s=" + start + " e=" + end + " sS=" + selStart + " sE=" + selEnd);

        if (this instanceof SingleLineTextPane) {
	    paintLine(g, data.getTextString(), true);
	    return;
	}

        int lnrStart = data.getLineNumber(start);
	int lnrEnd = data.getLineNumber(end);

	if (lnrStart == lnrEnd) {
	    paintLine(g, lnrStart);
	} else {

	    while (lnrStart <= lnrEnd) {
	        paintLine(g, lnrStart++);
	    }
	}
    }

    /**
     * Select the text between the given range.
     * @param s Starting index into text
     * @param e Ending index into text
     * @see java.awt.peer.TextComponentPeer
     */
    public void select(int s, int e) {

        int textLength = data.getLength();
        /*System.err.println("+select(" + s + "," + e + "): sS=" + selStart +
			   " sE=" + selEnd + " textL=" + textLength);*/

        if ((s > textLength) || (s < 0) || (e > textLength) || (e < 0)) {
            return;
        }

        if ((this.selStart == s) && (this.selEnd == e)) {
            return;
        }

        int startOld = this.selStart;
        int endOld = this.selEnd;
	setCaretPosition(s);
        selEnd = e;
        Graphics g = null;
	try {
	    g = view.getGraphics();

	    if ((selStart >= endOld) || (selEnd <= startOld)) {
	        paintRange(g, startOld, endOld);
		paintRange(g, selStart, selEnd);
	    } else {
	        int start = Math.min(selStart, startOld);
		int end = Math.max(selStart, startOld);

		if (start != end) {
		    paintRange(g, start, end);
		}
		start = Math.min(selEnd, endOld);
		end = Math.max(selEnd, endOld);

		if (start != end) {
		    paintRange(g, start, end);
		}
	    }

	    //resizeScrollbars();

	} finally {
	  if (g != null) {
	      g.dispose();
	  }
	}

        /*System.err.println("-select(): sS=" + selStart + ", sE=" + selEnd);*/
    }

    /**
     * Scroll the text to the position "pos". If pos is already on 
     * the screen then do nothing.
     * @param pos Index into the text
     * @param resizeScrollbars If true then scrollbars will be resized after
     * this operation
     */
    protected void scrollTo(int pos, boolean resizeScrollbars) {

        if (!data.isValidPosition(pos)) {
	    return;
	}

	int lnr;
	char [] text;
	if (this instanceof SingleLineTextPane) {
	    lnr = 0;
	    text = data.getText();
	} else {
	    lnr = data.getLineNumber(pos);
	    text = data.getLineFromNumber(lnr);
	}

	int oldTop = topLine;
	Graphics g = null;
	try {
	    g = view.getGraphics();

	    // handle vertical scrolling if necessary
	    if (!(this instanceof SingleLineTextPane)) {
	        handleVerticalScrolling(g, pos);
	    }

	    int x = pos2x(g.getFontMetrics(), pos, text);
	    int w = getMaxPossibleStringWidth();
	    int xMin = movedRight + (textRect.x + leftPad);
	    int xMax = w + xMin;

	    /*System.err.println(" x=" + x + " w=" + w +
			       " movedRight=" + movedRight + " xMin=" + xMin +
			       " xMax=" + xMax);*/
	
	    // scroll horizontal
	    if (x > xMax) {
	        scrollHorizontal(g, x - xMax);
	    } else if (x < xMin) {
	        scrollHorizontal(g, x - xMin);
	    }

	    if (resizeScrollbars && topLine != oldTop) {
	        resizeScrollbars();
	    }

	} finally {
	    if (g != null) {
	       g.dispose();
	    }
	}
    }

    /**
     * Scroll horizontaly.
     * @param g Current graphics context
     * @param x Number of pixles to scroll - negative value for a left scroll,
     * positive for a right scroll
     */
    protected void scrollHorizontal(Graphics g, int x) {

        /*System.err.println("+scrollH(" + x + ")");*/

        int nrOfLines = data.getNumberOfLines();
	int xmargin = textRect.x + leftPad;
	int w = getMaxPossibleStringWidth();
        /*System.err.println(" scrollH(): #lines=" + nrOfLines + " xmar=" +
			   xmargin + " ymar=" + ymargin + " w=" + w +
			   " h=" + h);*/

	Graphics ng = null;
	try {
	    ng = g.create();
	    movedRight += x;

	    if (x < 0) {
	        ng.copyArea(xmargin, textRect.y,
			    w + x, textRect.height, -x, 0);
		// add 1 to width to accomodate cursor
		ng.clipRect(xmargin, textRect.y, (-x) + 1, textRect.height);
		int n = linesInWindow();
		int e = Math.min((topLine + n) - 1, nrOfLines - 1);
		paintLines(ng, topLine, e);
	    } else if (x > 0) {
	        ng.copyArea(xmargin + x, textRect.y,
			    w - x, textRect.height, -x, 0);
		// add 1 to width to accomodate cursor
		ng.clipRect((xmargin + w) - x, textRect.y, x + 1,
			    textRect.height);
		int n = linesInWindow();
		int e = Math.min((topLine + n) - 1, nrOfLines - 1);
		paintLines(ng, topLine, e);
	    }
	} finally {
	    if (ng != null) {
	        ng.dispose();
	    }
	}
        /*System.err.println("-scrollH()");*/
    }

    /**
     * Move left one character using iterators so internationalization is
     * supported.
     * @param p Index into the text
     */
    protected void handleLeftArrowKey(int p) {

        if (data.getNumberOfLines() > 0 && data.isValidPosition(p - 1)) { 
	    int ln = data.getLineNumber(p);
	    // keep track of previous line also
	    int begLn = Math.max(0, ln - 1);
	    BreakIterator cBoundary = BreakIterator.getCharacterInstance();
	    int start = data.getLineStart(begLn);
	    int end = data.getLineEnd(ln);
	    cBoundary.setText(new String(data.getText(start, end - start),
					 0, end - start));
	    cBoundary.following(p - start - 1);
	    int s = cBoundary.previous();
	    s = (s == BreakIterator.DONE) ? p : s + start;
	    select(s, s);
	    scrollTo(s, true);
	}
    }

    /**
     * Move right one character using iterators so internationalization is
     * supported.
     * @param p Index into the text
     */
    protected void handleRightArrowKey(int p) {

        int nrOfLines = data.getNumberOfLines();
        if (nrOfLines > 0 && data.isValidPosition(p + 1)) {
	    int ln = data.getLineNumber(p);
	    // keep track of next line also
	    int endLn = Math.min(data.getNumberOfLines() - 1, ln + 1);
	    BreakIterator cBoundary = BreakIterator.getCharacterInstance();
	    int start = data.getLineStart(ln);
	    int end = data.getLineEnd(endLn);
	    cBoundary.setText(new String(data.getText(start, end - start),
					 0, end - start));
	    int e = cBoundary.following(p - start);
	    e = (e == BreakIterator.DONE) ? end : e + start;
	    select(e, e);
	    scrollTo(e, true);
	}
    }

    /**
     * Position cursor. This method allows us to change the behavior when 
     * cursor is repositioned by letting us either call select or selectWord
     * depending on the value of 
     * <a href="#wordSelection">wordSelection</a> mode.
     * @param pos Index for placing the cursor
     */
    protected void positionCursor(int pos) {
        if (wordSelection) {
	    int len = data.getLength();
	    if ((pos >= wordStart && pos <= wordEnd) || pos == len) {

	        select(pos, pos);

		int ln = data.getLineNumber(pos);
		if (wordLine != ln || pos == len) {
		    wordStart = wordEnd = pos;
		    wordLine = ln;
		}
	    } else {
	        selectWord(pos);
	    }
	} else {
	    select(pos, pos);
	}
    }

    /**
     * Select one white-space delimited word. The word is selected based
     * on a word iterator.
     * @param pos Index to use as the basis for choosing the word to select
     * @see java.text.BreakIterator
     */
    public void selectWord(int pos) {

	int ln = data.getLineNumber(pos);
	int start = data.getLineStart(ln);
	int end = data.getLineEnd(ln);

	BreakIterator wordBoundary = BreakIterator.getWordInstance();
	// ignore double-click if not on text
	if (pos < end) {
	    wordBoundary.setText(new String(data.getLineFromNumber(ln),
					    0, end - start));
	    int e = wordBoundary.following(pos - start);
	    int s = wordBoundary.previous();
	    s = (s == BreakIterator.DONE) ? pos : s + start;
	    e = (e == BreakIterator.DONE) ? end : e + start;
	    select(s, e);
	} else {
	    select(pos, pos);
	}

	wordStart = selStart;
	wordEnd = selEnd;
	wordLine = ln;
    }

    /**
     * Select a line around the given position.
     * @param pos Index to use as the basis for choosing the line to select
     */
    public void selectLine(int pos) {

	int ln = data.getLineNumber(pos);
        if (this instanceof SingleLineTextPane) {
	    // select entire text area
	    select(0, data.getLength());
	} else {
	    // select current line
	    select(data.getLineStart(ln), data.getLineEnd(ln));
	}
    }

    /**
     * Select all text.
     */
    public void selectAll() {
        select(0, data.getLength());
    }

    /**
     * Implement handling of multiple mouse clicks. Double click selects
     * a word, triple click selects a line and four clicks select all text.
     * @param x X mouse location
     * @param y Y mouse location
     * @param count Mouse click count
     */
    protected void handleMultipleMouseClicks(int x, int y, int count) {

        int pos = xy2pos(x, y);

	switch (count) {

	case 2:     // select a localized "word"
	    selectWord(pos);
	    break;

	case 3:
	    selectLine(pos);
	    break;

	case 4:
	    selectAll();
	    break;

	default:
	    break;
	}
    }

    /**
     * Insert given text at specified position.
     * @param str String to insert
     * @param pos Index position for inserting text
     */
    protected void insertImpl(String str, int pos) {

        selStart += str.length();
        selEnd = selStart;

	data.insert(str, pos);
    }

    /**
     * Delete the text between indicated positions in the text buffer.
     * @param start Start index.
     * @param end End index.
     */
    protected void deleteImpl(int start, int end) {
	setCaretPosition(start);
	data.delete(start, end);
    }

    /**
     * Replace the text between the position "s" and "e" with "txt".
     * @param txt Text to replace range from s to e.
     * @param start Start index.
     * @param end End index.
     */
    protected void replaceRange(String txt, int s, int e) {

        selStart += txt.length();
        selEnd = selStart;

        data.replace(txt, s, e);
    }

    /**
     * Selection support
     */
    public void lostOwnership(Clipboard clipboard, Transferable contents) {
        // required method, but it needn't do anything in our implementation
    }

    /**
     * Set the contents of the system clipboard.
     * @param tk Current toolkit instance
     * @param contents New contents for the clipboard
     * @param owner Current clipboard owner
     */
    static void setClipboard(ObjectToolkit tk, 
			     String contents, ClipboardOwner owner) {
        Clipboard clipboard = tk.getClipboard();
        clipboard.setContents(new StringSelection(contents), owner);
    }

    /**
     * Get the contents of the system clipboard
     * @param tk Current toolkit instance
     * @param requestor Object requesting clipboard contents
     */
    static String getClipboard(ObjectToolkit tk, Object requestor) {
        Clipboard clipboard = tk.getClipboard();
        Transferable contents = clipboard.getContents(requestor);
        Reader stream = null;
        String value = null;
    
        try {
            if (contents.isDataFlavorSupported(DataFlavor.stringFlavor)) {
                value = (String) 
                    contents.getTransferData(DataFlavor.stringFlavor);
            } else {
                // This will throw UnsupportedFlavorException (caught
                // below) if the plainTextFlavor is not available.
                stream = (Reader) 
                    contents.getTransferData(DataFlavor.plainTextFlavor);
    
                // The plainText flavor returns a Reader object
                // so we must build a string from its contents.
                // Unfortunately, Reader can't tell us the number of
                // bytes that are available.  So the expedient method is to
                // read largish chunks of data and append them to a 
                // StringBuffer object.
                //
                // In normal usage we should rarely take this path;
                // to test it, the stringFlavor case had to be disabled.
                StringBuffer sbuf = new StringBuffer();
                char chars[] = new char[4096];
                for (;;) {
                    int numRead = stream.read(chars);
                    if (numRead > 0) {
                        sbuf.append(chars, 0, numRead);
                    } else if (numRead == -1) { // EOF signal
                        break;
                    } else {
                        System.err.println("Warning: " +numRead+ 
                               " chars read in getClipboard()");
    
                        break;  // possibly incorrect, but we want to avoid
                                // looping infinitely.
                    }
                }
                value = sbuf.toString();
            }
        } catch (java.io.IOException ignored) {
            // declared and thrown by read().
            value = null;
        } catch (java.awt.datatransfer.UnsupportedFlavorException ignored) {
            // declared and thrown by getTransferData()
            value = null;
        } finally {
            // clean up the reader stream.
            if (stream != null) {
                try { stream.close(); } catch (java.io.IOException e) { }
            }
        }
    
        return value;
    }

    /**
     * Copy the current selection to the clipboard.
     * @return Returns true if operation succeeded, false otherwise.
     */
    public boolean copyToClipboard() {

        if (selStart == selEnd) {
	    setClipboard(tk, "", this);
	} else {
	    setClipboard(tk,
			 data.getTextString(selStart, selEnd - selStart), 
			 this);
	}

	return true;
    }

    /**
     * Cut the current selection to the clipboard.
     * @return Returns true if operation succeeded, false otherwise.
     */
    public boolean cutToClipboard() {

        if (!editable) {
	    return false;
	}

	if (selStart == selEnd) {
	    setClipboard(tk, "", this);
	} else {
	    setClipboard(tk, data.getTextString(selStart, selEnd - selStart), 
			 this);
	    deleteImpl(selStart, selEnd);
	}

	return true;
    }

    /**
     * Paste the contents of the clipboard onto the current selection.
     * @return Returns true if operation succeeded, false otherwise.
     */
    public boolean pasteFromClipboard() {
	return pasteOnSelection(getClipboard(tk, this));
    }

    /**
     * Paste the contents of the string onto the current selection.
     * @param str String to paste
     * @return Returns true if operation succeeded, false otherwise.
     */
    public boolean pasteOnSelection(String str) {

        if (!editable) {
	    return false;
	}

	int pos = selStart + str.length();
	if (selStart == selEnd) {
	    insertImpl(str, selStart);
	} else {
	    replaceRange(str, selStart, selEnd);
	}

	return true;
    }

    /**
     * Backspace, i.e. either delete selected text or delete backwards one
     * character.
     * @return Returns true if operation succeeded, false otherwise.
     */
    public boolean backspace() {

        if (!editable) {
	    return false;
	}

	if (selStart != selEnd) {
	    deleteImpl(selStart, selEnd);
	} else if (selStart > 0) {
	    deleteImpl(selStart - 1, selStart);
	}

	return true;
    }

    /**
     * Handle all mouse events for the textpane.
     * @param id Event type
     * @param when Time of event
     * @param modifiers Flags
     * @param x X coordinate
     * @param y Y coordinate
     * @param arg1 Click count
     * @param arg2 isPopupTrigger
     * @return True if the event was handled, false otherwise.
     */
    public boolean handleMouseEvent(int id,  long when, int modifiers,
				    int x, int y, int arg1, int arg2) {

        boolean handled = false;

        switch (id) {

        case ToolkitEventHandler.POINTER_PRESSED:

	    // We don't allow changing the insertion point when composing
	    if (composeMode) {
	        ((TextComponentView) view).playNotAllowedSound();
		handled = true;
		break;
	    }
	  
	    // The toolkit uses POINTER_PRESSED with a negative click count
	    // to differentiate MOUSE_PRESSED to MOUSE_CLICKED.
	    if (arg1 < 0) {
	        handleMultipleMouseClicks(x, y, -arg1);
	    } else {
	        mpos = xy2pos(x, y);
		positionCursor(mpos);
	    }
	    handled = true;
            break;

        case ToolkitEventHandler.POINTER_RELEASED:
	    // do nothing - left for future
            break;

        case ToolkitEventHandler.POINTER_DRAGGED:
            int mpos2 = xy2pos(x, y);
            if (mpos < mpos2) {
                select(mpos, mpos2);
            } else {
                select(mpos2, mpos);
            }
            scrollTo(mpos2, true);
	    handled = true;
            break;
	    
	default:
	    break;
        }

	return handled;
    }

    /**
     * Get the maximum possible string width that would fit in this textpane.
     * This value can be used by word-wrapping algorithms to wrap text for
     * this textpane.
     * @return Returns the maximum width in pixels
     */
    public int getMaxPossibleStringWidth() {
        return textRect.width - (leftPad + rightPad);
    }

    // Indicates whether we're in IMF compose mode or not.
    protected boolean composeMode = false;
    protected int oldComposeStart;
    protected int oldComposeEnd;

    /**
     * Returns true if we're currently in compose mode.
     */
    public boolean isComposing() {
        return composeMode;
    }

    /**
     * Invoked when the text entered through an input method has changed.
     * This is the mechanism for the Input Method Framework (IMF) to
     * communicate to text components.
     * @param committedText IMF text that has already been committed
     * @param composedText IMF text that has not been committed yet
     * @param caret Location of caret in the IMF text.
     */
    public void inputMethodTextChanged(String committedText, String composedText, int caret) {

	if (composeMode) {
	    data.delete(oldComposeStart, oldComposeEnd);
	} else {
	    composeMode = true;
	}

	// get the caret location from our top view and insert text
	int commLen = 0;
	if (committedText != null) {
	    commLen = committedText.length();
	    data.insert(committedText, selStart);
	}
	if (composedText != null) {
	    int compLen = composedText.length();
	    oldComposeStart = selStart;
	    oldComposeEnd = selStart + compLen;
	    data.insert(composedText, selStart + commLen);
	    select(selStart + commLen, selStart + commLen + compLen);
	} else {
	    composeMode = false;
	    selStart += caret + 1;
	    select(selStart, selStart);
	}
    }

    /**
     * Invoked when the caret within composed text has changed. We currently
     * don't differentiate between selection start and caret position, so we
     * don't display the caret in compose mode, hence do nothing for now.
     * @param committedText IMF text that has already been committed
     * @param composedText IMF text that has not been committed yet
     * @param caret Location of caret in the IMF text.
     */
    public void caretPositionChanged(String committedText, String composedText, int caret) {
	//System.err.println("TextComponentModel.caretPositionChanged()");
	//tk.beep();
    }

    /**
     * Return the length of the widest line in given range.
     * @param startLine Starting line
     * @param endLine Ending line
     */
    protected int maxLength(int startLine, int endLine) {

        //System.err.println("+maxLength(" + startLine + ", " + endLine + ")");

        int max = 0;
        startLine = Math.max(0, startLine);
	int nrOfLines = data.getNumberOfLines();
        endLine = Math.min(nrOfLines - 1, endLine);

        if (this instanceof SingleLineTextPane) {
	    max = getTabbedLength(data.getText());
	} else {
	    for(int i = startLine ; i <= endLine ; i++) {
	        char [] text = data.getLineFromNumber(i);
		max = Math.max(max, getTabbedLength(text));
		//System.err.println(" i=" + i + " max=" + max + " text=" + text);
	    }
	}
	//System.err.println("-maxLength(" + startLine + ", " + endLine + ") = " + max);

        return max;
    }

    /**
     * Calculate tabified length of characters.
     * @param text Array of characters
     */
    protected int getTabbedLength(char [] text) {

	FontMetrics fm = view.getMetrics();
        if (text != null && fm != null) {

	    int l = 0;
	    int widths[] = fm.getWidths();
	    for (int i = 0 ; i < text.length ; i++) {
	        // Search for tabs
	        if (text[i] == '\t') {
		    // We found a tab, find out the length of this line
		    // up to this point
		    l = (((l/widths[tabChar])/tabWidth + 1) * 
			 tabWidth) * widths[tabChar];
		} else {
		    // masking against 0xFF ensures that the index is
		    // between 0 and 255
		    if (text[i] > 255) {
		        l += fm.charWidth(text[i]);
		    } else {
		        l += widths[text[i]];
		    }
		}
	    }

	    return l;
	}

	return 0;
    }

    /**
     * Return wether given character is consumed with KEY_PRESSED.
    protected static boolean checkConsumedChar(char ch) {
	if (Character.isISOControl(ch)) {
	    switch (ch) {
	      case 0x03:	// Ctrl-C    copy
	      case 0x16:	// Ctrl-V    paste
	      case 0x18:	// Ctrl-X    cut
	      case 0x08:	// Ascii BS  backspace
	      case 0x7F:	// Ascii DEL delete
		return true;
	    }
	}
	return false;
    }
     */

    /**
     * Remove CR, LF from end of given string.
     * @param str String to be stripped
     * @return Returns stripped string.
     */
    protected String strip(String str) {
        String t;
        if (str.endsWith("\n\r")) {
	    t =  new String(str.substring(0, str.length() - 2));
        } else if (str.endsWith("\n")) {
	    t =  new String(str.substring(0, str.length() - 1));
        } else if (str.endsWith("\r")) {
	    t =  new String(str.substring(0, str.length() - 1));
        } else {
	    t = str;
	}

	if (t.length() == 0) {
	    return "";
	} else {
	    return t;
	}
    }

    /**
     * Scroll the text to the position "pos" regardless of whether it is
     * already on the screen or not. Default implementation - override!
     * @param pos Index into the text
     * @param resizeScrollbars If true then scrollbars will be resized after
     * this operation.
     */
    protected void scrollAbsolute(int pos, boolean resizeScrollbars) {
        // do nothing for now.
    }
  
    /** 
     * Check to see if the cursor is currently on this line.
     * @param line Line to check
     * @param startSel Start of selection
     * @param endSel End of selection
     * @return Returns true when startSel == endSel and they are within the
     * bounds of the specified line.
     */
    protected boolean cursorOnThisLine(int line, int startSel, int endSel) {
	return (startSel == endSel &&
		startSel >= data.getLineStart(line) &&
		startSel <= data.getLineEnd(line));
    }

    /**
     * Return the foreground color for textpane.
     */
    protected abstract Color getFgColor();

    /**
     * Return the background color for textpane.
     */
    protected abstract Color getBgColor();

    /**
     * Paint lines in the given range (inclusive).
     * @param g Current graphics context
     * @param s Starting line
     * @param e Ending line
     */
    public abstract void paintLines(Graphics g, int s, int e);

    /**
     * Method for isntructing the component to completely paint itself.
     * @param g Current graphics context.
     */
    public abstract void paintComponent(Graphics g);

    /**
     * Paint the string. This method is provided so that each concrete 
     * implementation can render the string in it's own way, for ex. TextField
     * will have specila echoChar handling, or TextArea will have special
     * Tab handling, etc. The premise is that the above paintLine method
     * can draw the primitives of the line for textpanes, which will most
     * probably be the same and then call this specialized paintString
     * to render the text.
     * @param g Graphics context
     * @param fm Current fontMetrics
     * @param f Current font
     * @param x X coordinate for the string
     * @param y Y coordinate for the string
     * @param inStr The string to be rendered
     * @param selS Index of start of selection
     * @param selE Index of end of selection
     * @param fg Foreground color for string
     * @param bg Background color for string
     * @see sun.awt.tk.TKTextPaneImpl.paintLine
     */
    protected abstract void paintString(Graphics g, FontMetrics fm, Font f,
					int x, int y, String inStr,
					int selS, int selE,
					Color fg, Color bg);

    /**
     * Tell the peer to resize scrollbars.
     */
    protected abstract void resizeScrollbars();

    /**
     * Scrolls vertically if necessary. Not needed for single line panes.
     * @param g Current graphics context
     * @param pos Index into the text
     */
    protected void handleVerticalScrolling(Graphics g, int pos) {
        // Default case is do nothing
    }
}
