/*
 * @(#)TextRandomData.java	1.9 98/11/25
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

package sun.awt.otk;

import java.text.BreakIterator;
import java.awt.Component;
import java.awt.FontMetrics;

/**
 * This class adds support for the
 * random selection mode. This mode allows insertion of soft break characters
 * into the data buffer. These are used to simulate line breaks in
 * <a href="sun.awt.otk.TextComponentView.html">TextComponentViews</a>. For a
 * complete description of this mode, see 
 * <a href="sun.awt.otk.RandomSelectionTextPane.html">
 * RandomSelectionTextPane</a>.
 * Random selection mode is turned on based on the inverse of the value of  
 * sun.awt.otk.noRandomSelectionMode system property. When in random selection
 * mode we insert <a href="#newline">newline</a> for textareas and 
 * <a href="#space">space</a> for textfields as soft line breaks.
 */
public class TextRandomData extends TextCharData {

    /**
     * When true, a pointer pressed on textRect but off the text causes
     * implicit line return markers to be added to the text data.
     */
    protected static boolean randomSelectionMode;

    static {
	// set randomSelectionMode from system property
	if (Boolean.getBoolean("sun.awt.otk.noRandomSelectionMode")) {
	    randomSelectionMode = false;
	} else {
	    randomSelectionMode = true;
	}
    }

    /**
     * Array containing list of softbreaks
     */
    int [] softBreaks;

    /**
     * Number of softbreaks in the buffer
     */
    int numSoftBreaks;

    /**
     * Definition of a soft break that is inserted into textareas,
     * currently a backslash n.
     */
    static final char [] newline = { '\n' };

    /**
     * Definition of a soft break that is inserted into textfields,
     * currently a space.
     */
    static final char [] space = { ' ' };

    /**
     * Number of entries to grow the softBreaks list with when it overflows
     */
    static public final int SOFT_LIST_GROWTH = 3;

    /**
     * Create a TextData buffer that supports the random selection mode.
     * @param model The TextComponentModel that this data buffer is
     * associated with
     * @param text Initialization text for the data buffer
     * @param lineWidth Default line width. Use 0 if not interested in
     * word wrapping (see
     * <A href="#setWordWrap">setWordWrap()</A>
     * )
     * @param wrap Default value for the word wrap mode
     * @see sun.awt.otk.TextComponentModel
     */
    public TextRandomData(TextComponentModel model, String text,
			  int lineWidth, boolean wrap) {
        super(model, text, lineWidth, wrap);
	
	numSoftBreaks = 0;
    }

    /**
     * Adds a soft line break at the given location. In the case of textareas,
     * we actually insert a newline, and a space in the case of textfields.
     * @param pos Position for soft break
     */
    public synchronized void insertSoftBreak(int pos) {
        if (model instanceof TextAreaModel) {
	    insert(newline, pos);
	} else if (model instanceof TextFieldModel) {
	    int index = getLineNumber(pos);

	    // make sure list of softBreaks has room
	    if (numSoftBreaks <= 0) {

	        softBreaks = new int[SOFT_LIST_GROWTH];
	        
	    } else if (numSoftBreaks + 1 > softBreaks.length) {
	        int [] temp = new int[softBreaks.length + SOFT_LIST_GROWTH];
		System.arraycopy(softBreaks, 0, temp, 0, numSoftBreaks);
		softBreaks = temp;
	    }

	    softBreaks[numSoftBreaks++] = pos + 1;
	    
	    StateChangeNotification c = insertImpl(space, textLength, true);
	    if (c != null) {
	        sendNotifications(c);
	    }
	}
    }

    /**
     * Get the index of the first softbreak that is after the val, starting
     * with the given index.
     * @param val Value to compare against
     * @param start Start index for comparisons
     * @return The index of the first element that is larger than val
     */
    private int getFirstIndexLargerThan(int val, int start) {
        int i;
        for (i = start; i < numSoftBreaks; i++) {
	    if (softBreaks[i] > val)
	        break;
	}
	return i;
    }

    /**
     * Concrete insert() implementation. We over-ride this method here
     * to provide default support for the non-soft break insertImpl method.
     * @param text Character array to insert in data buffer
     * @param pos Index to insert the given string at
     */
    protected synchronized StateChangeNotification insertImpl(char [] text,
							     int pos) {

        if (!randomSelectionMode)
	    return super.insertImpl(text, pos);

        return this.insertImpl(text, pos, false);
    }

    /**
     * Concrete insert() implementation. New implementation provided to 
     * handle cases when we're inserting a soft break. The values of
     * soft breaks have to be incremented by the number of characters
     * inserted to keep in sync with the changed data buffer.
     * @param text Character array to insert in data buffer
     * @param pos Index to insert the given string at
     * @param sbreak When true, we know we're inserting a soft break
     */
    protected synchronized StateChangeNotification insertImpl(char [] text,
							      int pos,
							      boolean sbreak) {

	if (!sbreak && isValidPosition(pos)) {

	    // go to the appropriate entry in softBreaks
	    int start = getFirstIndexLargerThan(pos, 0);
	    int inserted = text.length;
	    for (int i = 0; i < numSoftBreaks - start; i++) {
	        softBreaks[start + i] += inserted;
	    }
	}

        return super.insertImpl(text, pos);
    }

    /**
     * Concrete delete() implementation. Over-ridden here to make sure that
     * the values in the soft breaks array are properly decremented by the
     * number of deleted characters.
     * @param s Start index for deletion
     * @param e End index for deletion
     */
    protected synchronized StateChangeNotification deleteImpl(int s, int e) {

        if (!randomSelectionMode)
	    return super.deleteImpl(s, e);
        
        if (isValidPosition(s) && isValidPosition(e)) {

	    // go to the appropriate entry in softBreaks
	    int start = getFirstIndexLargerThan(s, 0);
	    int end = getFirstIndexLargerThan(e, start);
	    int deleted = e - s;
	    for (int i = 0; i < numSoftBreaks - end; i++) {
	        softBreaks[start + i] = softBreaks[end + i] - deleted;
	    }

	    numSoftBreaks -= (end - start);
	}

        return super.deleteImpl(s, e);
    }

    /**
     * Update lines array from the specified lineNumber. The basic algorithm 
     * here is as follows:
     *
     *    Iterate through all text starting at line nr using a lineIterator {
     *      if the last character of token is breakChar OR contains breakChar {
     *        Iterate through text using a character iterator
     *          handle hard breaks
     *          handle wordwrap
     *      } else {
     *          handle wordwrap
     *      }
     *    }
     *
     * The main attempt is to minimize the use of charIterators as much as
     * possible.
     * @param nr update lines array from this line number
     * @param returnChangedLines Return the number of changed lines if set to
     * true, otherwise return the new nrOfLines. This implementation will 
     * always return nrOfLines, i.e. returnChangedLines=false. The "true"
     * functionality is implemented in the sub-classes.
     * @return new number of  lines in lines array (note nrOfLines is not 
     * set yet)
     */
    public int updateLinesArray(int nr, boolean returnChangedLines) {

        if (!randomSelectionMode)
	    return super.updateLinesArray(nr, returnChangedLines);

        if (this.lineWidth <= 0) {
	    return 0;
	}

        //printLines(null);
        Component target = (Component) model.getTarget();
        FontMetrics fm = target.getFontMetrics(target.getFont());
        int widths[] = fm.getWidths();
        int w = this.lineWidth;
        int lineWidth = 0;      // width of line being processed
	int wordWidth = 0;      // width of current "word"
	int origNrOfLines = nrOfLines;

	int linesIndex = nr;    // index into lines array

	// remember offset into text array where we start from - this keeps us
	// from keeping a local copy of the text buffer to represent what the
	// iterator has internally.
	int textOffset = lines[nr];

	// Initialize the BreakIterator
	BreakIterator iter = BreakIterator.getLineInstance();
	iter.setText(new String(data, textOffset, textLength - textOffset));
	int s = iter.first();
	//System.err.println("***** updateLinesArray: nr:"+nr+", textOffset:"+textOffset+", linesIndex:"+linesIndex+", s:"+s+", wordwarp:"+wordwrap+" textLength="+textLength);
	lines[linesIndex++] = textOffset + s;

	// go to the appropriate entry in softBreaks
	int softIndex = getFirstIndexLargerThan(textOffset, 0);

	// Make sure that done only increments for "successive" iterations 
	// through this loop, i.e. make sure we kick only after 
	// wrapLineDecrement+1 consecutive iterations. oldDone ensures this.
	int done =0, oldDone = 0;
	int e;
	for (e = iter.next();
	     e != BreakIterator.DONE &&
	         e <= textLength &&
	         done <= wrapLineDecrement + 1;
	     s = e, e = iter.next()) {

	    //System.err.println("***** updateLinesArray: s:"+s+", e:"+e+", w:"+w+", linesIndex:"+linesIndex+", done:"+done+", str:"+new String(data, textOffset+s, e-s)+"<");
	    // check for softbreaks
	    boolean softBreakInserted = false;
	    if (numSoftBreaks > 0 && softIndex < numSoftBreaks) {

	        //System.err.println("***** updateLinesArray: textOffset+e=" + (textOffset+e) + " softBreaks[softIndex]=" + softBreaks[softIndex] + " linesIndex=" + linesIndex + " softIndex=" + softIndex + " numSoftBreaks=" + numSoftBreaks);
	        if (textOffset + e >= softBreaks[softIndex]) {
		    // insert the softbreaks into the lines array
		    while (softIndex < numSoftBreaks &&
			   textOffset + e >= softBreaks[softIndex]) {

		        // make sure lines array is large enough
		        if (lines.length <= linesIndex + 3) {
			    growLinesArray(numSoftBreaks - softIndex +
					   TEXT_GROWTH_PAD);
			}
		        lines[linesIndex++] = softBreaks[softIndex++];
			softBreakInserted = true;
		    }
		}
	    }

	    if (softBreakInserted) {
	        lineWidth = 0;     // reset lineWidth for next line
		//wordWidth = charsWidth(s, e, fm, widths, 0);
		wordWidth = 0;
	    } else {

	    // If the last character of the token is any one of these chars
	    // then we need to scan it using a charIterator since we cannot
	    // be certain that there are no hard breaks in it. We only test
	    // the last character first because it is much cheaper and will
	    // satisfy a majority of the cases.
	    if (breakChars.indexOf(data[textOffset + e - 1]) != -1 ||
		 nextBreakChar(textOffset+s, textOffset+e, breakChars) != -1) {

	        // Iterate using a char iterator
	        BreakIterator cIter = BreakIterator.getCharacterInstance();
		cIter.setText(new String(data, textOffset + s, e - s));
		int cs = cIter.first();
		int off = textOffset + s;
		wordWidth = 0;
	
	        // search forward from first break to end of token
		int ce;
	        for (ce = cIter.next();
		     ce != BreakIterator.DONE & done <= wrapLineDecrement + 1;
		     cs = ce, ce = cIter.next()) {

		    //System.err.println("***** updateIndex(1): cs:"+cs+", ce:"+ce+", w:"+w+", lw:"+lineWidth+", ww:"+wordWidth+", done:"+done+", off:"+off+", char:"+data[off+cs]);

		    if (done > 0 && oldDone == done) {
		        done = oldDone = 0;
		    } else {
		        oldDone = done;
		    }

		    // make sure lines array is large enough
		    if (lines.length <= (linesIndex + 3)) {
		        growLinesArray(TEXT_GROWTH_PAD);
		    }

		    if (data[off + cs] == '\n') {
		        if (off + ce < textLength &&
			    data[off + ce] == '\r') {

			    if (lines[linesIndex] == off + ce + 1) {
			        done += (returnChangedLines) ? 1 : 0;
			    } else {
			        lines[linesIndex] = off + ce + 1;
			    }
			    //System.err.println("***** (1) break at " +(off+ce+1));
			    // eat next token
			    cs = ce;
			    ce = cIter.next();

			    // Line iterator does not handle CRLF properly
			    s = e;
			    e = iter.next();
			} else {
			    if (lines[linesIndex] == off + ce) {
			        done += (returnChangedLines) ? 1 : 0;
			    } else {
			        lines[linesIndex] = off + ce;
			    }
			    //System.err.println("***** (1) break at " +(off+ce));
			}
			linesIndex++;
			lineWidth = 0;     // reset lineWidth for next line
			wordWidth = 0;     // reset wordWidth for next token
		    } else if (data[off + cs] == '\r') {
		        // LFCR comes back as one char iterator!!!!
		        if (off + ce < textLength &&
			    data[off + cs + 1] == '\n') {

			    if (lines[linesIndex] == off + cs + 1) {
			        done += (returnChangedLines) ? 1 : 0;
			    } else {
			        lines[linesIndex++] = off + cs + 1;
			        lines[linesIndex] = off + ce;
			    }
			    // eat next token
			    cs = ce;
			    ce = cIter.next();
			} else {
			    if (lines[linesIndex] == off + ce) {
			        done += (returnChangedLines) ? 1 : 0;
			    } else {
			        lines[linesIndex] = off + ce;
			    }
			    //System.err.println("***** (1+) break at " +(off+ce));
			}
			linesIndex++;
			lineWidth = 0;     // reset lineWidth for next line
			wordWidth = 0;     // reset wordWidth for next token
		    } else if (w > 0 && wordwrap) {

		        wordWidth = charsWidth(cs, cs, fm, widths, wordWidth);
		        if (wordWidth >= w) {
			    //System.err.println("***** (1) breaking fat word at " + cs);
			    if (lines[linesIndex] == off + cs) {
			        done += (returnChangedLines) ? 1 : 0;
			    } else {
			        lines[linesIndex] = off + cs;
			    }
			    linesIndex++;
			    lineWidth = 0;
			    wordWidth = charsWidth(cs, cs, fm, widths, 0);
			}
			
			if (lineWidth + wordWidth >= w) {
			    //System.err.println("***** (1) breaking fat line at " + off);
			    if (lines[linesIndex] == off) {
			        done += (returnChangedLines) ? 1 : 0;
			    } else {
			        lines[linesIndex] = off;
			    }
			    linesIndex++;
			    lineWidth = 0;
			}
		    }
		}
	    }

	    // take care of line breaks due to TextArea width.
	    // updateIndex() can get called from setTextImpl() before the
	    // dimensions are set and w can be negative. Doing wordwrap in
	    // this case has no meaning - do nothing extra!
	    else if (w > 0 && wordwrap) {
	        wordWidth = charsWidth(textOffset + s,
				       textOffset + e - 1, fm, widths, 0);
		//System.err.println("***** updateLinesArray(2): s:"+s+", e:"+e+", w:"+w+", lw:"+lineWidth+", ww:"+wordWidth+", linesIndex:"+linesIndex);
		// fat word! lineWidth must be <= w
		if (wordWidth >= w) {
	        
		    // Iterate using a char iterator
		    BreakIterator cIter = BreakIterator.getCharacterInstance();
		    cIter.setText(new String(data, textOffset + s, e - s));
		    int cs = cIter.first();
	
		    // search forward
		    int ce, temp;
		    wordWidth = 0;
		    int wOff = textOffset + s;
		    for (ce = cIter.next();
			 ce != BreakIterator.DONE && done <= wrapLineDecrement + 1;
			 cs = ce, ce = cIter.next()) {

		        // make sure lines array is large enough
		        if (lines.length == linesIndex) {
			    growLinesArray(TEXT_GROWTH_PAD);
			}

			if (done > 0 && oldDone == done) {
			    done = oldDone = 0;
			} else {
			    oldDone = done;
			}

		        temp = s + textOffset + cs;
			//System.err.println("***** updateIndex(2): cs:"+cs+", ce:"+ce+", w:"+w+", lw:"+lineWidth+", ww:"+wordWidth+", done:"+done+", char:"+data[temp]);
			wordWidth =
			    charsWidth(temp, temp, fm, widths, wordWidth);
			if (wordWidth >= w) {
			  if (lineWidth != 0) {
			    //System.err.println("***** (2) breaking fat word at " + wOff);
			    if (lines[linesIndex] == wOff) {
			        done += (returnChangedLines) ? 1 : 0;
			    } else {
			        lines[linesIndex] = wOff;
			    }
			    linesIndex++;
			    lineWidth = 0;
			    wOff = textOffset + s + cs;

			  } else {

			    //System.err.println("***** (2) breaking fat line at " + temp);
			    if (lines[linesIndex] == temp) {
			        done += (returnChangedLines) ? 1 : 0;
			    } else {
			        lines[linesIndex] = temp;
			    }
			    linesIndex++;
			    lineWidth = 0;
			    wordWidth = charsWidth(temp, temp, fm, widths, 0);
			  }
			}
		    }

                } else if (lineWidth + wordWidth >= w) {

		    // make sure lines array is large enough
		    if (lines.length == linesIndex) {
		        growLinesArray(TEXT_GROWTH_PAD);
		    }

		    if (done > 0 && oldDone == done) {
		        done = oldDone = 0;
		    } else {
		        oldDone = done;
		    }

		    if (lines[linesIndex] == textOffset + s) {
		        done += (returnChangedLines) ? 1 : 0;
		    } else {
		        lines[linesIndex] = textOffset + s;
		    }
		    linesIndex++;
		    lineWidth = 0;     // reset lineWidth- current added below
		}
	    }
	    }
	    lineWidth += wordWidth;
	}

	/*
	System.err.println("***** updateLinesArray: nr:"+nr+", textOffset:"+textOffset+", linesIndex:"+linesIndex+", s:"+s+", e:"+e+", done:"+done+", nrOfLines:"+nrOfLines);
	if (done > wrapLineDecrement)
	    printLines(null);
	else
	    printLines(null, 0, linesIndex);
        */

	if (returnChangedLines) {
	    // set the nrOfLines only if it changed
	    if (e == BreakIterator.DONE || (textOffset + e) >= textLength) {
	      //if (done <= wrapLineDecrement) {
	      nrOfLines = linesIndex;
	    }
	    if (nrOfLines - origNrOfLines >= 0)
	        return nrOfLines - nr;          // insertion
	    else
	        return origNrOfLines - nr;      // deletion
	} else {
	    // NOTE: WE DON'T SET nrOfLines HERE - THIS CAN BE DANGEROUS!!!
	    // Make sure you set nrOfLines to returned value
	    return linesIndex;
	}
    }

    /**
     * Reset the soft breaks array. This method is called by clients 
     * when they need to remove all soft breaks from their data buffer.
     */
    public synchronized void resetSoftBreaks() {

        // reset the soft breaks
        if (numSoftBreaks > 0) {
	    softBreaks = null;
	    numSoftBreaks = 0;

	    nrOfLines = updateLinesArray(0, false);
	    StateChangeNotification c = new StateChangeNotification(
				       StateChangeNotification.DATA_CHANGE,
				       0, nrOfLines - 1);
	    notifyDependents((Object) c);
	}
    }

    /**
     *  Prints out lines array.
     *  @param str header to print
     *  @param s start at element s
     *  @param e elements of lines array to print
     */
    protected void printLines(String initStr, int s, int e) {
        super.printLines(initStr, s, e);

	System.err.print("Soft Breaks List:");
	for (int i = 0; i < numSoftBreaks; i++)
	    System.err.print("  " + softBreaks[i]);
	System.err.println("");
	//System.err.println(" Text=" + new String(data) + "<");
    }
}
