/*
 * @(#)View.java	1.18 99/09/29 SMI
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package sun.awt.otk;

import java.awt.*;
import sun.awt.UpdateClient;

public interface View extends Dependent, UpdateClient  {
    /**
     * The notification method for the Dependent interface.  Indicates that
     * an observable change has occurred on Subject m; arg is an optional 
     * argument which may help indicate what has changed.
     * @param m The Subject (model) whose state has changed
     * @param arg An argument indicating the substance of the change.  May be null.
     */
    void        updateDependent(Subject m, Object arg);

    /**
     * The UpdateClient interface, used to support the ScreenUpdater.
     * @param arg An argument indicating the substance of the change.  May be null.
     * @see sun.awt.ScreenUpdater
     * @see sun.awt.UpdateClient
     */
    void        updateClient(Object arg);
    
    /**
     * Helps with implementation of updateClient.
     * @param x The x coordinate for the rectangle's left edge
     * @param y The y coordinate for the rectangle's top edge
     * @param width The width of the rectangle
     * @param height The height of the rectangle
     * @return a shape which includes the given rectangle, possibly merged with any other update
     * that is still pending.
     */
    Shape       getScreenUpdateRegion(int x, int y, int width, int height);

    /**
     * Clear any pending updates that lie in the given rectangle.   (This is an
     * indication that the pending updates for this rectangle have been done.)
     * @param r The rectangle that has been updated.
     */
    void        clearUpdateRegion(Rectangle r);


    /**
     * Query whether this View should accept the keyboard focus during traversal.
     * This query is usually passed along to the controllers.
     * @return true if this View is focus-traverable, false otherwise.
     */
    boolean     isFocusTraversable();

    /**
     * Request that the keyboard focus be transferred to the window associated with this View.
     */
    void        requestFocus();

    /**
     * Query the visible state of this view.
     * @return true if this view is supposed to be visible, false otherwise.
     */
    boolean     isVisible();

    /**
     * Set the visible state of this view.
     * @param visible The desired visibility; true for visible, false for invisible.
     */
    void        setVisible(boolean visible);

    /**
     * Set the stacking order (Z coordinate) of this view with respect to its siblings.
     * A zero indicates the frontmost window, with positive numbers indicating how
     * many siblings should be ahead of this window.  Negative numbers indicate to
     * start at the back of the stack, with -1 indicating that this window should be
     * on the bottom, -2 meaning that 1 window should be behind it, etc.
     * @param order The desired position.
     */
    void        setStackOrder(int order);

    /**
     * Get the physical bounds associated with the view, either in the
     * parent's coordinate system, or in the display's global coordinates.
     * @param global True if the bounds are desired in global coords, false otherwise.
     * @return A Rectangle describing the bounds of the underlying window.
     */
    Rectangle   getBounds(boolean global);

    /**
     * Set the bounds of the view, relative to the view's parent.
     * @param x The new x coordinate for the view's left edge
     * @param y The new y coordinate for the view's top edge
     * @param width The new width of the view
     * @param height The new height of the view
     */
    void        setBounds(int x, int y, int width, int height);

    /**
     * Get the preferred size for this view, on the assumption that the given
     * number of rows and columns must be displayed.  For a View that has no
     * row/column layout, this will return the same value as getPreferredSize().
     * @param rows The desired number of rows
     * @param cols The desired number of columns
     */
    Dimension   getPreferredSize(int rows, int cols);

    /**
     * Get the minimum size for this view, on the assumption that the given
     * number of rows and columns must be displayed.  For a View that has no
     * row/column layout, this will return the same value as getMinimumSize().
     * @param rows The desired number of rows
     * @param cols The desired number of columns
     */
    Dimension   getMinimumSize(int rows, int cols);

    /**
     * Get a Graphics object that has been configured to paint into this View.
     * @return A graphics object that can be used to paint into this View.
     */
    Graphics    getGraphics();

    /**
     * An accessor method for the Model associated with this view.
     * @return A Model object.
     */
    Model       getModel();

    /**
     * An accessor method for the toolkit with which this view is associated.
     * @return A toolkit object.
     */
    ObjectToolkit getToolkit();

    /**
     * An accessor method for the ColorPolicy associated with this view.
     * @return A ColorPolicy object.
     */
    ColorPolicy getColorPolicy();

    /**
     * An accessor method for the scrollbar dimensions associated with this view.
     * @return A Dimension object giving the scrollbar size for this view; null if there is no associated scrollbar
     */
    Dimension   getScrollbarDimensions();

    /**
     * Add a Controller which listens to events on this View.
     * @param c the Controller to add
     */
    void        addController(Controller c);


    /**
     * Remove a Controller which was previously added to this view.
     * @param c the Controller to remove
     */
    void        removeController(Controller c);

    /**
     * Remove all Controllers which were previously added to this view.
     */
    void        removeControllers();

    /**
     * Designates an already added controller as the drag controller. This
     * controller is always the first one to receive events regardless of
     * the order in which the controllers were added. Note that setting
     * the drag controller DOES NOT add it to the controllers list and 
     * susequent sets will overwrite the previous value.
     * @param c Controller to set as the drag controller. Use null to reset.
     */
    void        setDragController(Controller c);

    /**
     * Returns the current drag controller.
     * @return the current drag controller
     */
    Controller  getDragController();

    /**
     * The direct event delivery path to the View, always used for window-related
     * events and also used for input events if the View is not associated with an
     * awt.Component.  Event type identifiers are defined by 
     * sun.porting.toolkit.ToolkitEventHandler.
     *
     * All other types of events are passed along to the controllers.
     *
     * @param id The event type identifier
     * @param when A timestamp (in milliseconds) indicating when the event happened
     * @param modifiers The modifier key (SHIFT, CONTROL, ALT, META) state.
     * @param x The x coordinate associated with the event
     * @param y The y coordinate associated with the event
     * @param arg1 An optional argument whose meaning varies with the event type
     * @param arg2 An optional argument whose meaning varies with the event type
     * @see sun.porting.toolkit.ToolkitEventHandler
     */
    void        handleEvent(int id, long when, int modifiers,
			    int x, int y, int arg1, int arg2);
}
