/*
 * @(#)VirtualKey.java	1.26 98/11/25 SMI
 * 
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 * 
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 * 
 * CopyrightVersion 1.1_beta
 *
 */

package sun.awt.otk;

import java.util.Locale;
import java.awt.*;
import java.awt.event.KeyEvent;
import sun.porting.toolkit.ToolkitEventHandler;

/**
 *
 * @version 1.26, 11/25/98
 * @author Sun Microsystems, Inc.
 */

public abstract class VirtualKey extends Object {
    private  VirtualKeySet vkSet;
    private  Rectangle myBounds;
    private  char myKeyChar;
    private  int myKeyCode;
    private  boolean pressed;
    private  boolean isEnabled;

    // NOTE: these modes overlap those from java.awt.Event
    //       1 << 0  - SHIFT_MASK
    //       1 << 1  - CTRL_MASK
    //       1 << 2  - META_MASK
    //       1 << 3  - ALT_MASK
    private  int allowableModes;

    /**
     * Constructs a <code>VirtualKey</code> object with the specified character,
     * key code, allowed modes, and bounds. 
     *
     * @param keyChar the character to be associated with the key
     * @param keyCode the key code
     * @param modes the allowable modes for the key - those for which the key
     * is allowed to be enabled 
     */
    public VirtualKey(char keyChar, int keyCode, int modes,
		      Rectangle bounds) {
	myKeyChar = keyChar;
	myKeyCode = keyCode;
	allowableModes = modes;
	myBounds = new Rectangle(bounds);
	isEnabled = true;
    }
 
    /**
     * Compares this <code>VirtualKey</code> object against a key code and
     * character for equality. 
     *
     * @param keyChar the character
     * @param keyCode the key code
     *
     * @returns true if the key is the same as that specified in the parameters
     */
    public boolean keyEquals(int keyCode, char keyChar) {
	keyChar = Character.toLowerCase(keyChar);

	if (keyChar != KeyEvent.CHAR_UNDEFINED) {
	    if (keyChar == myKeyChar) {
		return true;
	    }
	} else if (keyCode != KeyEvent.VK_UNDEFINED) {
	    if (keyCode == myKeyCode) {
		return true;
	    }
	}
	return false;
    }

    /**
     * Sets the key set for this key.
     * 
     * @param vkSet the key set to associate the key with
     *
     * @see #getVKSet()
     * @see sun.awt.otk.VirtualKeySet
     */
    public void setKeySet(VirtualKeySet vkSet) {
	this.vkSet = vkSet;
    }

    /**
     * Sets the key set for this key.
     * 
     * @param vkSet the key set to associate the key with
     *
     * @see #setKeySet
     * @see sun.awt.otk.VirtualKeySet
     */
    public final VirtualKeySet getVKSet() {
	return vkSet;
    }
   
    /**
     * Gets the key code for this key.
     *
     * @returns the key code for this key
     */
    public int getKeyCode() {
	return myKeyCode;
    }

    /**
     * Gets the character value for this key.
     *
     * @returns the character value for this key
     */
    public char getKeyChar(){
	return myKeyChar;
    }

    /**
     * Gets the bounds for this key.
     *
     * @returns the bounds for this key
     * @see #setBounds
     */
    public Rectangle getBounds() {
	return new Rectangle(myBounds);
    }

    /**
     * Gets the key code for this key.
     *
     * @param x the x coordinate
     * @param y the y coordinate
     * @param width the width of the key
     * @param height the height of the key
     * @see #getBounds
     */
    public void setBounds(int x, int y, int width, int height){
	myBounds.setBounds(x, y, width, height);
    }

    /**
     * Determines whether the point defined by the parameters is within the key.
     *
     * @param x the x coordinate
     * @param y the y coordinate
     * @returns true if the point is within the bounds of the key
     */
    public boolean contains(int x, int y){
	return myBounds.contains(x, y);
    }

    /**
     * Determines whether the key is pressed.
     *
     * @returns true if the key is pressed
     * @see #setPressed
     */
    public boolean isPressed(){
	return pressed;
    }

    /**
     * Sets the pressed state of the key.
     *
     * @param pressed the state to put the key into
     * @see #isPressed
     */
    public void setPressed(boolean pressed) {
	this.pressed = pressed;
    }

    /**
     * Resets the internal states of the key.
     */
    public void reset() {
	pressed = false;
	setEnabled(true);
    }

    /**
     * Notifies the key that conditions it may have used to set its internal state
     * have changed.
     *
     * @returns true if an update was required
     */
    public boolean update() {	
	boolean shouldBeEnabled = (vkSet.getModifiers() & ~allowableModes) == 0;
	if (isEnabled != shouldBeEnabled) {
	    isEnabled = shouldBeEnabled;
	    vkSet.updateDependent(this);
	    return true;
	}
	return false;
    }

    /**
     * Gets the enabled state of the key.
     *
     * @returns true if the key is enabled
     * @see #setEnabled
     */
    public boolean isEnabled(){

	// basic disabled
	if (!isEnabled) {
	    return false;
	}

	// is a mode set for which we aren't allowed to be enabled?
	if ((vkSet.getModifiers() & ~allowableModes) != 0) {
	    return false;
	}
	return true;
    }

    /**
     * Sets the enabled state of the key.
     *
     * @param enabled If true the key is enabled otherwise it is disabled.
     * @see #setEnabled
     */
    public void setEnabled(boolean enabled){
	isEnabled = enabled;
    }
    
    /**
     * Paints the key.  This method must be implemented in a concrete child class.
     *
     * @param g The graphics context to use for the drawing.
     */
    public abstract void paintKey(Graphics g);

    /**
     * Called when a pointer pressed event occurs for the key.  A concrete child
     * class must implement this method.
     */
    public abstract void pointerPressed();

    /**
     * Called when a pointer released event occurs for the key.  A concrete child
     * class must implement this method.
     */
    public abstract void pointerReleased();

    /**
     * Called when a key pressed event occurs for the key.  A concrete child
     * class must implement this method.
     */
    public abstract void keyPressed();

    /**
     * Called when a key released event occurs for the key.  A concrete child
     * class must implement this method.
     */
    public abstract void keyReleased();

    /**
     * Called when a key typed event occurs for the key.  A concrete child
     * class must implement this method.
     */
    public abstract void keyTyped();			    
}
