/*
 * @(#)VirtualKeySet.java	1.23 98/11/20 SMI
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

package sun.awt.otk;

import java.awt.*;
import java.awt.event.*;
import java.util.Locale;
import java.util.ResourceBundle;
import sun.porting.toolkit.ToolkitEventHandler;


abstract public class VirtualKeySet extends ContainerViewImpl {

    // Key Set Related Data Members
    public static final int NULL_MODE = 0;

    protected VirtualKey[] virtualKeys;
    protected VirtualKey lastKey;

    private int currentModifiers;
    private int lockedModifiers;
    private Locale locale;
    private String keySetResource;
    
    private VirtualKeyboardModel vkModel;
    

    public VirtualKeySet(VirtualKeyboardModel vkModel, 
			 ObjectToolkit toolkit, ContainerView parent,
			 Rectangle bounds, Insets insets,
			 String keySetResource) {
	this(vkModel, toolkit, parent, bounds, insets,
	     keySetResource, Locale.getDefault());
    }

    public VirtualKeySet(VirtualKeyboardModel vkModel,
			 ObjectToolkit toolkit, ContainerView parent,
			 Rectangle bounds, Insets insets,
			 String keySetResource, Locale locale) {

	super(toolkit, parent, bounds, insets);

	this.vkModel = vkModel;
	loadKeySet(keySetResource, locale);
	this.keySetResource = keySetResource;
	this.locale = locale;

	lastKey = virtualKeys[0];
    }

    public Model getModel(){
	return vkModel;
    }

    public Locale getLocale() {
	return locale;
    }

    public String getKeySetResource() {
	return keySetResource;
    }

    abstract public void paintKeys(Graphics g);

    protected VirtualKey getKey(int keyCode, char keyChar) {	
	for (int i = 0; i < virtualKeys.length; i++) {
	    VirtualKey vkey = virtualKeys[i];
	    if (vkey.keyEquals(keyCode, keyChar)) {
		return vkey;
	    }
	}	
	return null;
    }
 
    protected VirtualKey getKey(int x, int y) {
	for (int i = 0; i < virtualKeys.length; i++) {
	    VirtualKey vkey = virtualKeys[i];
	    if (vkey.contains(x, y)) {
	        return vkey;
	    }
	}
	return null;
    }

    protected VirtualKey getKey(int x, int y, int eventType) {
	VirtualKey vkey = null;
	if (eventType != ToolkitEventHandler.POINTER_RELEASED) {
	    if ((vkey = getKey(x, y)) != null) {
		lastKey = vkey;
	    }
	    return vkey;
	}
	return lastKey;
    }

    public Dimension getPreferredSize(int row, int cols) {
	Rectangle bounds = new Rectangle();

	if (virtualKeys.length != 0) {

	   bounds.setBounds(virtualKeys[0].getBounds());
	    for (int i = 1; i < virtualKeys.length; i++) {
	        bounds.add(virtualKeys[i].getBounds());
	    }
	}

	Insets insets = getInsets();
	bounds.width += insets.left + insets.right;
	bounds.height += insets.top + insets.bottom;
	return new Dimension(bounds.width, bounds.height);
    }

    public Graphics getGraphics() {
	Graphics g = getContainerWindow().getGraphics();

	Rectangle r = getBounds();
	g.setClip(0, 0, r.width, r.height);

	return g;
    }

    public int getModifiers() {
	return currentModifiers;  
    }

    public boolean getModifiers(int modifiers) {
	return (modifiers & currentModifiers) == modifiers;
    }

    public int getLockedModifiers() {
	return lockedModifiers;
    }

    public boolean getLockedModifiers(int modifiers) {
	return (modifiers & lockedModifiers) == modifiers;
    }

    public void setModifiers(int modifiers, boolean lock) {
	if (modifiers != currentModifiers) {
	    currentModifiers = modifiers;
	    if (lock) {
		lockedModifiers = modifiers;
	    } else {
		lockedModifiers = 0;
	    }
	    updateKeys();
	}
    }

    public boolean addModifier(int modifier, boolean lock) {
	if (lock && !getLockedModifiers(modifier)) {
	    lockedModifiers |= modifier;
	}

	if ((modifier & currentModifiers) == 0) {
	    currentModifiers |= modifier;
	    updateKeys();
	    return true;
	}
	return false;
    }

    public boolean clearModifier(int modifier, boolean lock) {
	if (((modifier & currentModifiers) != 0) &&
	    (lock || !getLockedModifiers(modifier))) {
	    currentModifiers &= ~modifier;
	    if (lock) {
		lockedModifiers &= ~modifier;
	    }
	    updateKeys();
	    return true;
	}
	return false;
    }

    public void clearModifiers() {
	currentModifiers = NULL_MODE;
	lockedModifiers = 0;
	updateKeys();
    }

    public void reset(int modifiers, boolean lock) {
	currentModifiers = modifiers;
	if (lock) {
	    lockedModifiers = modifiers;
	} else {
	    lockedModifiers = 0;
	}

	for (int i = 0; i < virtualKeys.length; i++) {
	    virtualKeys[i].reset();
	}
	updateDependent(null);
    }

    public void reset() {
	reset(NULL_MODE, false);
    }

    public void sendKeyEvent(VirtualKey vkey, int eventType) {
	((VirtualKeyboardModel) getModel()).sendKeyEvent(vkey, eventType);
    }

    public void updateKeys() {
	for (int i = 0; i < virtualKeys.length; i++) {
	    virtualKeys[i].update();
	}
    }

    protected void paintView(Model m, Graphics g) {
	if (!isVisible()) {
	    return;
	}
	
	Insets insets = getInsets();
	g.translate(insets.left, insets.top);
	paintKeys(g);
	g.translate(-insets.left, -insets.top);
    }

    protected void loadKeySet(String keySetResource, Locale locale) {

	ResourceBundle resource = ResourceBundle.getBundle(keySetResource,
							   locale);

	virtualKeys = (VirtualKey[]) resource.getObject("keys");
	for (int i = 0; i < virtualKeys.length; i++) {
	    virtualKeys[i].setKeySet(this);
	}
    }

    public void handleEvent(int id,  long when, int modifiers,
			       int x, int y, int arg1, int arg2) {

	VirtualKey vkey = null;

	if (id == ToolkitEventHandler.WINDOW_DAMAGED) {
	    Graphics g = null;

	    try {
		g = getGraphics();
		g.clipRect(x, y, arg1, arg2);
		paintView(getModel(), g);
	    } finally {
		if (g != null) {
		    g.dispose();
		}
	    }
	} else {
	    // Must translate positional events by our insets
	    Insets insets = getInsets();
	    x -= insets.left;
	    y -= insets.top;

	    switch (id) {
	    case ToolkitEventHandler.POINTER_PRESSED:
		if (arg1 > 0) {		
		    vkey = getKey(x, y, id);
		    if (vkey == null || !vkey.isEnabled()) {
			break;
		    }
		    pointerPressed(vkey);
		    }
		break;
	    case ToolkitEventHandler.POINTER_RELEASED:
		pointerReleased(lastKey);
		break;
	    case ToolkitEventHandler.KEY_PRESSED:
		vkey = getKey(arg1, (char) arg2);
		if (vkey == null || !vkey.isEnabled()) {
		    break;
		}
		keyPressed(vkey);
		break;
	    case ToolkitEventHandler.KEY_TYPED:
		vkey = getKey(arg1, (char) arg2);
		if (vkey == null || !vkey.isEnabled()) {
		    break;
		}
		keyTyped(vkey);
		break;
	    case ToolkitEventHandler.KEY_RELEASED:
		vkey = getKey(arg1, (char) arg2);
		if (vkey == null || !vkey.isEnabled()) {
		    break;
		}
		keyReleased(vkey);
		break;
	    }
	}
    }

    public void pointerPressed(VirtualKey vkey) {
	vkey.pointerPressed();
    }

    public void pointerReleased(VirtualKey vkey) {
	vkey.pointerReleased();
    }

    public void keyPressed(VirtualKey vkey) {
	vkey.keyPressed();
    }

    public void keyTyped(VirtualKey vkey) {
	vkey.keyTyped();
    }

    public void keyReleased(VirtualKey vkey) {
	vkey.keyReleased();	
    }

    public void updateDependent(Object obj) {
	Graphics g = null;

	try {
	    g = getGraphics();
	    if (obj == null) {
		Rectangle bounds = getBounds();
		g.clearRect(0, 0, bounds.width, bounds.height);
		paintView(((VirtualKeyboardModel) getModel()), g);
	    } else if (obj instanceof VirtualKey) {
		Insets insets = getInsets();
		g.translate(insets.left, insets.top);
		((VirtualKey) obj).paintKey(g);
		g.translate(-insets.left, -insets.top);
	    }
	} finally {
	    if (g != null) {
		g.dispose();
	    }
	}
    }
}
