/*
 * @(#)VirtualKeyboardInputMethod.java	1.36 98/12/17
 * 
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 * 
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 * 
 * CopyrightVersion 1.1_beta
 *
 */

package sun.awt.otk;

import java.util.Locale;
import java.awt.*;
import sun.awt.im.*;
import java.awt.event.*;
import java.awt.peer.*;


/**
 * VirtualKeyboardInputMethod class is an class which allows the chaining
 * of two input methods - one a real input method and the other a virtual
 * input device such as an on-screen keyboard.
 *
 * @see sun.awt.im.InputMethodAdapter
 *
 * @version 1.36 12/17/98
 */

public class VirtualKeyboardInputMethod 
    extends sun.awt.im.InputMethodAdapter {

    protected InputMethod inputMethod = null;
    protected boolean inputMethodCreationFailed = false;

    protected VirtualKeyboardViewIntf textPopup = null;
    protected boolean textPopupCreationFailed = false;
    protected VirtualKeyboardActivator keyboardActivator = null;

    private TextComponentModel textModel = null;

    /**
     * Constructs a new virtual keyboard input method.
     */
    protected VirtualKeyboardInputMethod() {
    }

    /**
     * Sets the input context which can dispatch input method
     * events to the client component.
     *
     * @param context the <code>InputMethodContext</code>
     */
    public void setInputContext(InputMethodContext context) {
	if (getRealInputMethod() != null) {
	    if (inputMethod instanceof sun.awt.otk.InputMethodAdapter) {
		((sun.awt.otk.InputMethodAdapter) inputMethod).setInputContext(context);
	    }
	}
    }



    /**
     * Activate input method. This is called by InputContext on a
     * FOCUS_GAINED event.
     */
    public void activate() {
	Component client = getClientComponent();

	if (getRealInputMethod() != null) {
	    if (inputMethod instanceof sun.awt.otk.InputMethodAdapter) {
                ((sun.awt.otk.InputMethodAdapter) inputMethod).setClientComponent(client);
		
	    }
	    inputMethod.activate();
	}

	/**
	 * If by some chance we get to calls to activate on the same
	 * target, we can ignore the second call.  If however the
	 * target is different, we need to make sure that we 
	 * deactivate the current popup before we create a new one.
	 */
	if (textPopup != null) {
	    if (textModel.getTarget() == client) {
		return;
	    }
	    deactivate();
	}
	    
	/**
	 * Create the virtual keyboard only when it is required.
	 */
	ObjectToolkit toolkit = (ObjectToolkit) Toolkit.getDefaultToolkit();
	boolean vkRequired = toolkit.isVirtualKeyboardRequired();
	
	

	if (vkRequired && !(client instanceof TextComponent)) {
	    if (keyboardActivator == null) {
		keyboardActivator = toolkit.getVirtualKeyboardActivator();
	    }
	    if (keyboardActivator != null) {
		keyboardActivator.setVisible(true);
	    }
	}
	if (vkRequired ||
	    (inputMethod != null && (!(client instanceof TextComponent)))) {
	    createTextPopup();
	}
    }

    /**
     * Deactivate input method.  This is called by InputContext on a
     * FOCUS_LOST event.
     */
    public void deactivate() {
	if (inputMethod != null) {
	    inputMethod.deactivate();
	}
	/**
	 * When the real input method is deactivated, completely
	 * dispose of the virtual keyboard.
	 */
	if (keyboardActivator != null) {
	    keyboardActivator.setVisible(false);
	}
	if (textPopup != null) {
	    hideTextPopup();
	    textPopup.dispose();
	    textModel = null;
	    textPopup = null;
	}
    }



    /**
     * Dispatch event to input method. InputContext dispatch event with this
     * method. Input method set consume flag if event is consumed in
     * input method.
     *
     * @param e event
     */
    public void dispatchEvent(AWTEvent e) {


	Component client = getClientComponent();
	if (client == null) {
	    return;
	}

	ObjectToolkit toolkit = (ObjectToolkit) Toolkit.getDefaultToolkit();

	
	 if (e instanceof BufferedKeyEvent) {
	    /**
	     * Buffereing is done in the case where:
	     * 1) target is not a TextComponent
	     * 2) a virtual keyboard is required OR a real input 
	     *    method is present.
	     *
	     * A BufferedKeyEvent is a tagging class to identify
	     * key events that have been buffered.
	     *
	     * A buffered key event needs to be passed on to AWT
	     * untouched.
	     */
	    return;
	 }

	 /* A VirtualComponentEvent.VIRTUAL_COMPONENT_SHOW event
	  * is currently fired by components when they decide
	  * that they need a virtual keyboard or text popup.
	  *
	  * A VirtualComponentEvent.VIRTUAL_COMPONENT_HIDDEN event
	  * gets generated when the virtual keybaord or text popup
	  * is deactivated.
	  *
	  */
	 if (e.getID() == VirtualComponentEvent.VIRTUAL_COMPONENT_SHOW) {
	     showTextPopup(e);
	     textModel.notifyDependents(e);
	     return;
	 } else if (e.getID() == VirtualComponentEvent.VIRTUAL_COMPONENT_HIDE) {
	     endComposition();
	     textModel.notifyDependents(e);
	     hideTextPopup();
	     return;
	 }

	 // let the input method have a look at it
	 if (inputMethod != null ) {
	     inputMethod.dispatchEvent(e);
	 }

	 /*
	  * No buffering for text components, 
	  * forward event back to AWT.
	  */
	 if (client instanceof TextComponent) {
	     return;
	 }

	 /*
	  * Now we decide whether we need to buffer
	  */
	 if (toolkit.isVirtualKeyboardRequired() || inputMethod != null) {
	     if (e instanceof VirtualKeyEvent) {
		 VirtualKeyEvent ve = (VirtualKeyEvent)e;
		 if (!ve.isConsumed()) {
		     // Merely update the model and consume the event.
		     //
		     textModel.notifyDependents(e);
		     ve.consume();
		     return;
		 }
	     }
	 }
    }

    /**
      * Returns the current client component of the input method.
      *
      * @returns client component
      */
    public Component getClientComponent() {
	return super.getClientComponent();
    }

    /**
      * This method posts the buffered data to client component.  The data
      * has been buffered when the client component is not an AWT text component.
      *
      * @param data the data buffer to post
      */
    public void postBufferedData(String data) {
	
	char[] array = data.toCharArray();
	for (int i = 0; i < array.length; i++) {
	    Component clientComponent = getClientComponent();
	    KeyEvent keyEvent = new BufferedKeyEvent(clientComponent,
						     KeyEvent.KEY_TYPED,
						     System.currentTimeMillis(),
						     0, KeyEvent.VK_UNDEFINED, 
						     array[i]);
	    clientComponent.dispatchEvent(keyEvent);
	}
    }

    /**
     * This method is called when the virtual keyboard needs
     * be shown.
     */
    protected void showTextPopup(AWTEvent e) {

	if (textPopup != null && textPopup.isVisible()) {
	    return;
	}
	Component clientComponent = (Component)e.getSource();

	// clientComponent gets set by the InputContext.
	// Because of package privacy issues, we cannot
	// set the clientComponent.  This check
	// garauntees that we wait for the clientComponent
	// to equal the event source, before we deliver
	// the event.
	if (clientComponent != getClientComponent()) {
	    return;
	}
	if (textPopup == null) {
	    createTextPopup();
	    if (textPopup == null) {
		return;
	    }
	}

	textPopup.show(e);
    }



    /**
     * This method is called when the virtual keyboard needs
     * be hidden.
     */
    protected void hideTextPopup() {
	if (textPopup != null && textPopup.isVisible()) {
	    textPopup.hide();

	    if (!(textModel instanceof TextComponentPeerModel)) {
 		postBufferedData(textModel.getText());
		textModel.setText("");
 	    }
	}
    }


    /**
     * Ends any input composition that may currently be going on in this
     * context. Depending on the platform and possibly user preferences,
     * this may commit or delete uncommitted text. Any changes to the text
     * are communicated to the active component using an input method event.
     *
     * <p>
     * A text editing component may call this in a variety of situations,
     * for example, when the user moves the insertion point within the text
     * (but outside the composed text), or when the component's text is
     * saved to a file or copied to the clipboard.
     *
     */
    public void endComposition() {
	if (inputMethod != null) { 
	    inputMethod.endComposition();
	}
    }

    /**
     * Disposes of the input method and releases the resources used by it.
     */
    public void dispose() {
	if (inputMethod != null) {
	    inputMethod.dispose();
	}

	/*
	 * dispose of the virtual keyboard if it was created.
	 */
	if (keyboardActivator != null) {
	    keyboardActivator.setVisible(false);
	}
	if (textPopup != null) {
	    hideTextPopup();
	    textPopup.dispose();
	    textModel = null;
	    textPopup = null;
	}
    }


    /**
     * An internal method used to get the real input method.  If it has been called
     * before then then the input method has been cached.  Otherwise an attempt is
     * made to create one.
     *
     * @returns the input method
     */
    protected InputMethod getRealInputMethod() {
 
        if (inputMethod != null) {
            return inputMethod;
        }
        
        if (inputMethodCreationFailed) {
            return null;
        }

        try {
            Toolkit toolkit = Toolkit.getDefaultToolkit();

            inputMethod = ((ObjectToolkit) toolkit).getRealInputMethodAdapter();
        } catch (Exception e) {
            // there are a number of bad things that can happen while creating
            // the input method. In any case, we just continue without an
            // input method.
            inputMethodCreationFailed = true;
        }
        
        if (inputMethod != null) {
            inputMethod.setInputContext((InputMethodContext) this.inputContext);
        }
        return inputMethod;
    }

    /**
     * An internal method used to create the text popup.
     */
    protected void createTextPopup() {

        if (textPopupCreationFailed) {
            return;
        }

	Component client = getClientComponent();

	if (client == null) {
	    return;
	}

	/**
	 * Establishing the textModel that is associated with the current
	 * client component
	 */
	ObjectToolkit toolkit = (ObjectToolkit) Toolkit.getDefaultToolkit();
	if (client instanceof TextComponent) {
	    textModel = (TextComponentModel)client.getPeer();
	} else if (inputMethod != null || toolkit.isVirtualKeyboardRequired()) {
	    textModel = new TextComponentModel(client, toolkit);
	    if (inputMethod instanceof sun.awt.otk.InputMethodAdapter) {
		((sun.awt.otk.InputMethodAdapter)inputMethod).setInputMethodListener((InputMethodListener)textModel);
	    }
	}

	try {
	
	    if (textModel.supportsInlineEdit() && !toolkit.isVirtualKeyboardRequired()) {
		return;
	    }
	    textPopup = toolkit.getVirtualKeyboard(this, textModel);

	} catch (Exception e) {
	    // there are a number of bad things that can happen while 
	    // creating the input method. In any case, we just continue 
	    // without an input method.
	    System.err.println("textPopup Creation Failed! \n - " + e);
	    e.printStackTrace();
	    textPopupCreationFailed = true;
	}
    }
}
