/*
 * @(#)VirtualKeyboardModel.java	1.19 98/11/30
 * 
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 * 
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 * 
 * CopyrightVersion 1.1_beta
 *
 */

package sun.awt.otk;

import java.lang.System;
import sun.awt.otk.VirtualKeyboardInputMethod;
import sun.awt.otk.VirtualKeyEvent;
import java.awt.event.KeyEvent;
import java.util.Hashtable;
import java.util.Locale;
import java.util.Enumeration;
import java.awt.*;

/**
  * This class is the container for the multiple key sets which can exist in a 
  * virtual keyboard system.
 */
public class VirtualKeyboardModel extends ObservableModel {


    protected VirtualKeyboardInputMethod vkim;
    protected Hashtable virtualKeySets;
    protected VirtualKeySet defaultKeySet;

    /**
      * Create a virtual keyboard model.
      *
      * @param inputMethod the virtual keyboard input method
      * @see sun.awt.otk.VirtualKeyboardInputMethod
      */
    public VirtualKeyboardModel(VirtualKeyboardInputMethod inputMethod) {
	vkim = inputMethod;
	virtualKeySets = new Hashtable();;
    }

    /**
      * Add a key set to the internal lists.
      *
      * @param vkSet the virtual key set to be added
      * @see sun.awt.otk.VirtualKeySet
      */
    public void addKeySet(VirtualKeySet vkSet) {
	Locale locale = vkSet.getLocale();
	String keySetResource = vkSet.getKeySetResource();
	String hashKey = new String(keySetResource + "_" + locale.toString());
	virtualKeySets.put(hashKey, vkSet);
	defaultKeySet = vkSet;
    }

    /**
      * Remove a key set to the internal lists.
      *
      * @param keySetResource the name of the virtual key set to be removed
      * @see sun.awt.otk.VirtualKeySet
      */
    public void removeKeySet(String keySetResource) {
	Locale locale = Locale.getDefault();
	removeKeySet(keySetResource, locale);
    }

    /**
      * Remove a key set to the internal lists.
      *
      * @param keySetResource the name of the virtual key set to be removed
      * @param locale the locale of the virtual key set to be removed
      * @see sun.awt.otk.VirtualKeySet
      */
    public void removeKeySet(String keySetResource, Locale locale) {
	String hashKey = new String(keySetResource + "_" + locale.toString());

	VirtualKeySet vkSet = (VirtualKeySet) virtualKeySets.remove(hashKey);
	if (vkSet == defaultKeySet) {  // must set the default to some value!
	    if (virtualKeySets.isEmpty()) {  // no other values -> make default null
		defaultKeySet = null;
	    } else {  // other values exist -> choose the first to be default
		Enumeration allHashKeys = virtualKeySets.elements();
		defaultKeySet = (VirtualKeySet) virtualKeySets.get(allHashKeys.nextElement());
	    }
	}
    }

    /**
      * Make a key set the default key set.  The key set must already have been
      * added to this virtual keyboard model.
      *
      * @param keySetResource the name of the virtual key set to be made default
      * @returns the key set that has been made the default one
      * @see sun.awt.otk.VirtualKeySet
      */
    public VirtualKeySet setDefaultKeySet(String keySetResource) {
	Locale locale = Locale.getDefault();
	return setDefaultKeySet(keySetResource, locale);
    }

    /**
      * Make a key set the default key set.  The key set must already have been
      * added to this virtual keyboard model.
      *
      * @param keySetResource the name of the virtual key set to be removed
      * @param locale the locale of the virtual key set to be removed
      * @returns the key set that has been made the default one
      * @see sun.awt.otk.VirtualKeySet
      */
    public VirtualKeySet setDefaultKeySet(String keySetResource, Locale locale) {
	boolean wasChanged = false;

	String hashKey = new String(keySetResource + "_" + locale.toString());
	VirtualKeySet vkSet = (VirtualKeySet) virtualKeySets.get(hashKey);
	if (vkSet != null) {
	    if (vkSet != defaultKeySet) {
		wasChanged = true;
	    }
	    defaultKeySet = vkSet;
	    if (wasChanged) {
		notifyDependents(vkSet);
	    }
	}
	return defaultKeySet;
    }

    /**
      * Get the current default key set.
      *
      * @returns the default key set
      * @see sun.awt.otk.VirtualKeySet
      */
    public VirtualKeySet getDefaultKeySet() {
	return defaultKeySet;
    }

    /**
      * Resets the default key set.
      *
      * @see sun.awt.otk.VirtualKeySet
      */
    public void reset() {
	defaultKeySet.reset();
    }
   
    /**
      * Sends a key event to the client component.
      *
      * @param vkey the virtual key that is requesting the event to be sent
      * @param eventType the type of key event that is being requested
      * @returns the key set that has been made the default one
      * @see sun.awt.otk.VirtualKey
      * @see java.awt.event.KeyEvent
      * @see sun.awt.otk.VirtualKeyEvent
      */
    public void sendKeyEvent(VirtualKey vkey, int eventType) {

	if (!vkey.isEnabled()) {
	    return;
	}

	int keyCode = vkey.getKeyCode();	
	char keyChar = vkey.getKeyChar();

	if (eventType == KeyEvent.KEY_TYPED) {
	    if (keyChar == KeyEvent.CHAR_UNDEFINED) {
		return;
	    }
	    keyCode = KeyEvent.VK_UNDEFINED;
	}

	long time = System.currentTimeMillis();
	
	Component clientComponent = vkim.getClientComponent();
	VirtualKeyEvent keyEvent = new VirtualKeyEvent(clientComponent, 
						       eventType, time, 
					          defaultKeySet.getModifiers(), 
						       keyCode, keyChar);
	clientComponent.dispatchEvent(keyEvent);
    }
}
