/*
 * @(#)WindowModel.java	1.14 99/07/29 SMI
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package sun.awt.otk;

import java.awt.Window;
import java.awt.peer.WindowPeer;

public class WindowModel extends ContainerModel implements WindowPeer {
    /* 
     * window activation events are sent whenever a new window becomes
     * the "active" one.  To control this, we use a separate object which
     * encapsulates and enforces the activation policy.
     */
    static private WindowActivationPolicy
        activationPolicy = new SimpleStackActivationPolicy();

    /**
     * Sets the window activation policy for all WindowModels. 
     * The default activation policy is SimpleStackActivationPolicy.
     * @param ac the window activation policy
     * @see SimpleStackActivationPolicy
     */
    static public void setActivationPolicy(WindowActivationPolicy ac) {
	if (activationPolicy != null) {
	    View v[] = activationPolicy.getFloatingViews();
	    if (v != null) {
		for (int i = 0; i < v.length; ++i) {
		    ac.addFloatingView(v[i]);
		}
	    }
	}

	activationPolicy = ac;
    }

    static /* package private */ WindowActivationPolicy getActivationPolicy() {
	return activationPolicy;
    }

    /**
     * Notifies the WindowActivationPolicy object that the visibility state of this window's default view
     * has changed.
     * @param b if true, this window's default view has become visible; if false it has been hidden
     * @exception WindowActivationVetoException if window activation policy doesn't
     * allow this change
     * @see WindowActivationPolicy
     */
    protected void windowVisibilityChanged(boolean b) 
        throws WindowActivationVetoException {
	activationPolicy.windowVisibilityChanged(this, b);
    }

    /**
     * Notifies the WindowActivationPolicy object that the position of this window has changed in the
     * stack of currently visible windows.
     * @param n an integer specifying the new position of the window, where
     * a zero indicates the frontmost window, with positive numbers indicating how
     * many windows should be ahead of this window.  Negative numbers indicate to
     * start at the back of the stack, with -1 indicating that this window should be
     * on the bottom, -2 meaning that 1 window should be behind it, etc.
     * @exception WindowActivationVetoException if window activation policy doesn't
     * allow this change
     * @see WindowActivationPolicy
     */
    protected void windowStackingChanged(int n)
        throws WindowActivationVetoException {
	activationPolicy.windowRestacked(this, n);
    }

    /**
     * Constructs a new WindowModel instance.
     * @param target the AWT Window component
     * @param tk the toolkit
     */
    public WindowModel(Window target, ObjectToolkit tk) {
	super(target, tk);
    }

    /**
     * Set the default view.
     * @param v the new value of the default view
     */
    public void setDefaultView(View v) {
	java.awt.Window target = (java.awt.Window)getTarget();

	if (target.getFont() == null) {
	    target.setFont(tk.getDefaultFont());
	}

	/* make sure background and foreground are set.  This turns out */
	/* to be essentially a noop if they're already set, because the */
	/* color policy will merely return the set value.               */
	ColorPolicy cp = v.getColorPolicy();
	target.setBackground(cp.getBackground(this, false, true, false));
	target.setForeground(cp.getForeground(this, false, true, false));

	super.setDefaultView(v);
    }

    /**
     * Shows or hides this window's default view, depending on the value of parameter b.
     * Notifies the WindowActivationPolicy object if the visibility state changed.
     * @param b if true, shows the window (if the window is already visible, brings it to the front);
     * otherwise, hides this component.
     * @see #windowVisibilityChanged
     * @see WindowActivationPolicy
     */
    public void setVisible(boolean b) {
	boolean wasVisible = isVisible();

	if (b && wasVisible) {
	    toFront();
	} else {
	    try {
		if (b != wasVisible) {
		    windowVisibilityChanged(b);
		}
		super.setVisible(b);

	    } catch (WindowActivationVetoException e) {
	    }
	}
    }

    /**
     * Moves the window's default view in front of all other windows.
     * Notifies the WindowActivationPolicy object if the window is currently visible.
     * @see #windowStackingChanged
     * @see WindowActivationPolicy
     */
    public synchronized void toFront() {
	try {
	    if (isVisible()) {
		windowStackingChanged(activationPolicy.countFloatingViews());
	    }

	    getDefaultView().setStackOrder(activationPolicy.countFloatingViews());
	} catch (WindowActivationVetoException e) {
	}
    }

    /**
     * Moves the window's default view behind all other windows.
     * Notifies the WindowActivationPolicy object if the window is currently visible.
     * @see #windowStackingChanged
     * @see WindowActivationPolicy
     */
    public synchronized void toBack() {
	try {
	    if (isVisible()) {
		windowStackingChanged(-1);
	    }

	    getDefaultView().setStackOrder(-1);
	} catch (WindowActivationVetoException e) {
	}
    }
}
