/*
 * @(#)resource.c	1.1 98/11/18
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*-
 *	Native methods needed to locate resource files in the system class loader
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "oobj.h"
#include "path.h"
#include "interpreter.h"
#include "monitor.h"
#include "sys_api.h"

#include "java_lang_ClassLoader.h"
#include "java_io_InputStream.h"

/*
 * Look for a resource file with a given name.
 */

static ClassClass *
classSunIoFileIOFactory()
{
    static ClassClass *cb = NULL;
    if (cb == NULL) {
	cb = FindStickySystemClass(EE(), "sun/io/FileIOFactory", TRUE);
    }
    return cb;
}

/*
 * Returns a Byte Array for a given member in a zip structure
 */

HArrayOfByte *
GetByteArrayInZIP(char *fn, zip_t *zip)
{
    struct stat st;
    unsigned char *data;
    HArrayOfByte *back;
    extern HArrayOfByte *MakeByteString(char *str, long len);

    LOADCLASS_LOCK();
    if (!zip_stat(zip, fn, &st)) {
      LOADCLASS_UNLOCK();
      return 0;
    }
    if ((data = sysMalloc(st.st_size)) == 0) {
      LOADCLASS_UNLOCK();
      return 0;
    }
    if (!zip_get(zip, fn, data, st.st_size)) {
      sysFree(data);
      LOADCLASS_UNLOCK();
      return 0;
    }
    back = MakeByteString((char *)data, st.st_size);
    sysFree(data);

    LOADCLASS_UNLOCK();
    return back;
}

/*
 * Does this zip structure have a member of the given name?
 */

bool_t
URLNameInZIP(char *fn, zip_t *zip)
{
    struct stat st;
    unsigned char *data;
    Hjava_io_InputStream *back;
    extern HArrayOfByte *MakeByteString(char *str, long len);

    LOADCLASS_LOCK();
    if (!zip_stat(zip, fn, &st)) {
      LOADCLASS_UNLOCK();
      return 0;
    }
    LOADCLASS_UNLOCK();
    return 1;
}

/*
 * Returns a name identifying the given system resource
 */

Hjava_lang_String *
GetURLName(char *name)
{
    int count = 0;
    cpe_t **cpp;
  
    /* FindClass before the upcall to Java code, because
       the Java code can raise an exception */
    ClassClass *classSecurityException
      = FindClass(EE(), JAVAPKG "SecurityException", TRUE);
    if (classSecurityException == (void *)0) {
        return 0;
    }

    for (cpp = sysGetClassPath(); cpp && *cpp != 0; cpp++, count++) {
	cpe_t *cpe = *cpp;
	char buf[MAXPATHLEN];
	if (cpe->type == CPE_DIR) {
	    if (jio_snprintf(buf, sizeof(buf), 
			     "%s%c%s", cpe->u.dir,
			     LOCAL_DIR_SEPARATOR, name) == -1) {
		return 0;
	    }
	    
	    /* Trying local file */
	    if (sysAccess(buf, 0) == 0) {
		HObject *myFile;
		Hjava_lang_String *myString;	
		char *npath = sysNativePath(buf);
		long exists = 1;
		    
		/* Now create a FileIO object
		   npath is non -null, so no exceptions in that call... */
		myString = makeJavaString(npath, strlen(npath));
		myFile = (HObject *)
		    execute_java_static_method(EE(),
					       classSunIoFileIOFactory(),
					       "newInstance",
					       "(Ljava/lang/String;)Lsun/io/FileIO;",
					       myString);

		/* Now check if the file exists.  This includes a security
		   manager check */
		exists = execute_java_dynamic_method(EE(),
						   myFile,
						   "exists",
						   "()Z");
		/* If there was a security exception, we clear it --
		   else the client can know that the file exists, and
		   they don't have read privileges. Null the
		   FileIO object pointed to by "myFile" -- it will
		   get GC-ed. */
		if (EE() && exceptionOccurred(EE())) {
		    ExecEnv *ee = EE();
		    JHandle *exc = ee->exception.exc;
		    if (is_instance_of(exc, classSecurityException, ee)) {
			exceptionClear(ee);
			myFile = (void *)0;
			return 0; /* cannot access! */
		    }
		}
		if (! exists) {
		    return 0;
		}

		if (jio_snprintf(buf, sizeof(buf),
				 "systemresource:/FILE%d/+/%s", count, name) == -1) {
		    return 0;
		}
		return makeJavaString(buf, strlen(buf));
	    }
	} else if (cpe->type == CPE_ZIP) {
	    if (URLNameInZIP(name, cpe->u.zip)) {
	      char url[MAXPATHLEN];
	      if (jio_snprintf(url, sizeof(url),
			       "systemresource:/ZIP%d/+/%s", count, name) == -1) {
		return 0;
	      }
		
	      return makeJavaString(url, strlen(url));
	    }
	}
    }

    return 0;
}

/*
 * REMIND: The following 4 routines share the same structure; they probably
 * should be organized to share more code.
 */

/*
 * Validate that a system resource is indeed correct
 */

long
resource_ValidateSystemResource(long iszip, char *base, char *member)
{
    cpe_t **cpp;

    /* FindClass before the upcall to Java code, because
       the Java code can raise an exception */
    ClassClass *classSecurityException
      = FindClass(EE(), JAVAPKG "SecurityException", TRUE);
    if (classSecurityException == (void *)0) {
        return 0;
    }
  
#if 0
    jio_fprintf(stderr, "resource_ValidateSystemResource\n");
    jio_fprintf(stderr, " iszip: %d\n", iszip);
    jio_fprintf(stderr, " base: %s\n", base);
    jio_fprintf(stderr, " member: %s\n", member);
#endif

    for (cpp = sysGetClassPath(); cpp && *cpp != 0; cpp++) {
	cpe_t *cpe = *cpp;
	char path[MAXPATHLEN];
	if (cpe->type == CPE_DIR) {
	    if (strcmp(cpe->u.dir, base) == 0) {
		if (iszip) {
		    return 0;
		}
		if (jio_snprintf(path, sizeof(path), 
				 "%s%c%s", cpe->u.dir,
				 LOCAL_DIR_SEPARATOR, member) == -1) {
		    return 0;
		}

		/* Trying local file */
		if (sysAccess(path, 0) == 0) {
		    HObject *myFile;
		    Hjava_lang_String *myString;
		    char *npath = sysNativePath(path);
		    long back = 1;
		    
		    /* Now create a FileIO object
		       npath is non -null, so no exceptions in that call... */
		    myString = makeJavaString(npath, strlen(npath));
		    myFile = (HObject *)
			execute_java_static_method(EE(),
						   classSunIoFileIOFactory(),
						   "newInstance",
						   "(Ljava/lang/String;)Lsun/io/FileIO;",
						   myString);

		    /* Now check if the file exists.  This includes a security
		       manager check */
		    back = execute_java_dynamic_method(EE(),
						       myFile,
						       "exists",
						       "()Z");
		    /* If there was a security exception, we clear it --
		       else the client can know that the file exists, and
		       they don't have read privileges. Null the
		       FileIO object pointed to by "myFile" -- it will
		       get GC-ed. */
		    if (EE() && exceptionOccurred(EE())) {
			ExecEnv *ee = EE();
			JHandle *exc = ee->exception.exc;
			if (is_instance_of(exc, classSecurityException, ee)) {
			    exceptionClear(ee);
			    myFile = (void *)0;
			    back = 0;
			}
			/* Anandp? do I have to do anything with the exception */
		    }
		    return back;
		}
		return 0;
	    }
	} else if (cpe->type == CPE_ZIP) {
	    if (strcmp(cpe->u.zip->fn, base) == 0) {
		
		if (iszip && URLNameInZIP(member, cpe->u.zip)) {
		    return 1;
		}
		return 0;
	    }
	}
    }
    return 0;
}


/*
 * Return a byte array with the contents of the given member of a ZIP file
 * REMIND - This could/should be done at the Java level...
 */

HArrayOfByte *
resource_GetZipByteArray(char *base, char *member)
{
  HArrayOfByte *back;
  cpe_t **cpp;
  
#if 0
  jio_fprintf(stderr, "resource_GetZipByteArray: %s\n", base);
  jio_fprintf(stderr, " base: %s\n", base);
  jio_fprintf(stderr, " member: %s\n", member);
#endif

  /* The getResource API disallows loading of .class files. See
     4062428.  Use case insensitive ASCII compare, so this works on
     Windows as well. */
  if (sysStricmp(member + strlen(member) - strlen(".class"), ".class") == 0) {
    SignalError(EE(), "java/lang/SecurityException", 
		"Can't read classfile as system resource");
    return 0;
  }

  for (cpp = sysGetClassPath(); cpp && *cpp != 0; cpp++) {
	cpe_t *cpe = *cpp;
	if (cpe->type == CPE_DIR) {
	    if (strcmp(cpe->u.dir, base) == 0) {
	        /* found the desired member, but we only handle ZIP, so
		 * return 0
		 */
	      return 0;
	    }
	} else if (cpe->type == CPE_ZIP) {
	    if (strcmp(cpe->u.zip->fn, base) == 0) {
	        /* found the desired member,, now check the member and
		 * return it
		 */

	      return (GetByteArrayInZIP(member, cpe->u.zip));
	    }
	}
  }

    return 0;
}

/*-
 * Convert a KEY value into the actual CLASSPATH entry it represents
 * By using opaque KEYS we are hiding information on files on the users machine
 *
 * GetURLName deals on keys also, the other methods deal on exposed values
 */

HString *
resource_GetEntryFromKey(long key) {
    int count = 0;
    cpe_t **cpp;

    for (cpp = sysGetClassPath(); cpp && *cpp != 0; cpp++, count++) {
	if (count == key) {
	    cpe_t *cpe = *cpp;

	    if (cpe->type == CPE_DIR) {
		return makeJavaString(cpe->u.dir, strlen(cpe->u.dir));
	    } else if (cpe->type == CPE_ZIP) {
		return makeJavaString(cpe->u.zip->fn, strlen(cpe->u.zip->fn));
	    }
	    return 0;
	}
    }
    return 0;
}


