/*
 * @(#)string.c	1.43 98/08/31
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "oobj.h"
#include "interpreter.h"
#include "javaString.h"
#include "utf.h"
#include "typecodes.h"
#include "path.h"
#include "sys_api.h"

enum {
    NO_ENCODING = 0,		/* Uninitialized */
    NO_FAST_ENCODING,		/* Platform encoding is not fast */
    FAST_8859_1,		/* ISO-8859-1 */
    FAST_CP1252			/* MS-DOS Cp1252 */
};

static int fastEncoding = NO_ENCODING;
static char *encoding = NULL;

/* enumeration of c1 row from Cp1252 */
static int cp1252c1chars[32] = {
    0x20AC,0xFFFD,0x201A,0x0192,0x201E,0x2026,0x2020,0x2021,
    0x02C6,0x2030,0x0160,0x2039,0x0152,0xFFFD,0xFFFD,0xFFFD,
    0xFFFD,0x2018,0x2019,0x201C,0x201D,0x2022,0x2013,0x2014,
    0x02Dc,0x2122,0x0161,0x203A,0x0153,0xFFFD,0xFFFD,0x0178
};

static unicode *cp1252ToUnicode(char *str, unicode *ustr, long len)
{
    unicode *dst = ustr;
    char *src = str, *end = str + len;
    while (src < end) {
        int n = *src++;
        if ((n >= 0x80) && (n <= 0x9f))
            *dst++ = cp1252c1chars[n-128];
        else
            *dst++ = n & 0xff;
    }
    return ustr;
}

static void unicodeToCp1252(unicode *str, char *ustr, long len)
{
    unicode *src = str, *end = str + len;
    char *dst = ustr;
    while (src < end) {
        unicode n = *src++;
        if (n < 256)
            *dst++ = n;
        else switch(n) {
            case 0x20AC: *dst++ = 0x80; break;
            case 0x201A: *dst++ = 0x82; break;
            case 0x0192: *dst++ = 0x83; break;
            case 0x201E: *dst++ = 0x84; break;
            case 0x2026: *dst++ = 0x85; break;
            case 0x2020: *dst++ = 0x86; break;
            case 0x2021: *dst++ = 0x87; break;
            case 0x02C6: *dst++ = 0x88; break;
            case 0x2030: *dst++ = 0x89; break;
            case 0x0160: *dst++ = 0x8A; break;
            case 0x2039: *dst++ = 0x8B; break;
            case 0x0152: *dst++ = 0x8C; break;
            case 0x2018: *dst++ = 0x91; break;
            case 0x2019: *dst++ = 0x92; break;
            case 0x201C: *dst++ = 0x93; break;
            case 0x201D: *dst++ = 0x94; break;
            case 0x2022: *dst++ = 0x95; break;
            case 0x2013: *dst++ = 0x96; break;
            case 0x2014: *dst++ = 0x97; break;
            case 0x02DC: *dst++ = 0x98; break;
            case 0x2122: *dst++ = 0x99; break;
            case 0x0161: *dst++ = 0x9A; break;
            case 0x203A: *dst++ = 0x9B; break;
            case 0x0153: *dst++ = 0x9C; break;
            case 0x0178: *dst++ = 0x9F; break;
            default:     *dst++ = '?';  break;
        }
    }
    *dst = '\0';
}

static Hjava_lang_String *makeEncodedJavaString(char *str, int len, int enc)
{
    HArrayOfChar *val;
    unicode *p;

    if (!str) len = 0;
    val = (HArrayOfChar *) ArrayAlloc(T_CHAR, len);
    if (val == 0) {
	SignalError(0, JAVAPKG "OutOfMemoryError", 0);
	return 0;
    }
    p = unhand(val)->body;
    if (str) {
        if (enc == FAST_8859_1)
            str2unicode(str, p, len);
        else if (enc == FAST_CP1252)
           cp1252ToUnicode(str, p, len);
    }
    KEEP_POINTER_ALIVE(p);
    return (Hjava_lang_String *)
      execute_java_constructor(EE(), 0, classJavaLangString, "([C)", val);
}

Hjava_lang_String *makeJavaString(char *str, int len)
{
    return makeEncodedJavaString(str, len, FAST_8859_1);
}

void javaStringPrint(Hjava_lang_String *s)
{
    if (s) {
	Classjava_lang_String *str = unhand(s);
	HArrayOfChar *hac = (HArrayOfChar *)str->value;
	unicode *body = unhand(hac)->body;
	jio_fprintus(stdout, body + str->offset, str->count);
	KEEP_POINTER_ALIVE(body);
    }
}

int javaStringLength(Hjava_lang_String *s)
{
    Classjava_lang_String *str;
    if (!s) {
	return 0;
    }
    str = unhand(s);
    return str->count;
}

unicode *javaString2unicode(Hjava_lang_String *s, unicode *buf, int len)
{
    if (s) {
	Classjava_lang_String *str = unhand(s);
	HArrayOfChar *hac = (HArrayOfChar *)str->value;
	unicode *body = unhand(hac)->body;
	memcpy(buf, body+str->offset, len*sizeof(unicode));
	KEEP_POINTER_ALIVE(body);
    } else {
	memset(buf, 0, len * sizeof(unicode));
    }
    return buf;
}

static char *javaString2EncodedCString(Hjava_lang_String *s, char *buf,
                                                 int buflen, int enc)
{
    Classjava_lang_String *str;
    HArrayOfChar *hac;

    if (s && (str = unhand(s)) && (hac = (HArrayOfChar *)str->value)) {
	unicode *src = unhand(hac)->body;
	int	len = javaStringLength(s);

	if (len >= buflen)
	    len = buflen - 1;
        if (enc == FAST_8859_1)
            unicode2str(src + str->offset, buf, len);
        else if (enc == FAST_CP1252)
	    unicodeToCp1252(src + str->offset, buf, len);
	KEEP_POINTER_ALIVE(src);
    } else if (buflen) {
	buf[0] = '\0';
    }
    return buf;
}

char *javaString2CString(Hjava_lang_String *s, char *buf, int buflen)
{
   return javaString2EncodedCString(s, buf, buflen, FAST_8859_1);
}

static char *makeEncodedCString(Hjava_lang_String *s, int enc)
{
    extern HArrayOfByte *MakeByteString(char *str, long len);
    int len = javaStringLength(s);
    char *buf;

    HArrayOfByte *hab = MakeByteString(NULL, len + 1);
    if (hab == 0) {
	SignalError(0, JAVAPKG "OutOfMemoryError", 0);
	return 0;
    }

    buf = unhand(hab)->body;
    return javaString2EncodedCString(s, buf, len + 1, enc);
}

char *makeCString(Hjava_lang_String *s)
{
    return makeEncodedCString(s, FAST_8859_1);
}

char *allocCString(Hjava_lang_String *s)
{
    int len = javaStringLength(s);
    char *buf = sysMalloc(len + 1);
    if (buf == 0) {
	SignalError(0, JAVAPKG "OutOfMemoryError", 0);
	return 0;
    }
    return javaString2CString(s, buf, len + 1);
}

#include "sun_io_ByteToCharConverter.h"
#include "sun_io_CharToByteConverter.h"

#define ENCODING_KEY     "file.encoding"
#define DEFAULT_ENCODING "8859_1"

static void check_encoding()
{
    if (fastEncoding == NO_ENCODING) {
	ExecEnv *ee = EE();
	Hjava_lang_String *enc;
	enc = (Hjava_lang_String *) execute_java_static_method
	    (ee, FindClass(ee, "java/lang/System", TRUE),
	     "getProperty",
	     "(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;",
	     makeJavaString(ENCODING_KEY, strlen(ENCODING_KEY)),
	     makeJavaString(DEFAULT_ENCODING, strlen(DEFAULT_ENCODING)));
	if (exceptionOccurred(ee)) {
	    exceptionClear(ee);
	    encoding = DEFAULT_ENCODING;
	}
	else
	    encoding = allocCString(enc);

        if (!strcmp(encoding, "8859_1"))
            fastEncoding = FAST_8859_1;
        else if (!strcmp(encoding, "Cp1252"))
            fastEncoding = FAST_CP1252;
        else
            fastEncoding = NO_FAST_ENCODING;
    }
} 

Hjava_lang_String *makeJavaStringFromPlatformCString(char *str, int len)
{
    Hsun_io_ByteToCharConverter *byteToChar;
    ExecEnv *ee = EE();
    Hjava_lang_String *ret;
    HArrayOfByte *hab;
    HArrayOfChar *hac;
    int length;
    extern HArrayOfByte *MakeByteString(char *str, long len);

    check_encoding();
    if (fastEncoding > NO_FAST_ENCODING)
	return makeEncodedJavaString(str, len, fastEncoding);

    byteToChar = (Hsun_io_ByteToCharConverter *)
	execute_java_static_method
	(ee, FindClass(ee, "sun/io/ByteToCharConverter", TRUE),
	 "getConverter",
	 "(Ljava/lang/String;)Lsun/io/ByteToCharConverter;", 
	 makeJavaString(encoding, strlen(encoding)));

    if (exceptionOccurred(ee)){
	exceptionClear(ee);
	return makeJavaString(str, len);
    }

    hab = MakeByteString(str, len);
    hac = (HArrayOfChar *)ArrayAlloc(T_CHAR, len);
    length = (int)execute_java_dynamic_method
	(ee, (HObject *)byteToChar,
	 "convert", "([BII[CII)I", hab, 0, len, hac, 0, len);

    if (exceptionOccurred(ee)){
	exceptionClear(ee);
	return makeJavaString(str, len);
    }

    ret = (Hjava_lang_String *)execute_java_constructor
	(ee, "java/lang/String", NULL, "([CII)", hac, 0, length);
    return ret;
}

char *makePlatformCString(Hjava_lang_String *s)
{
    Hsun_io_CharToByteConverter *charToByte;
    ExecEnv *ee = EE();
    int length;
    int strLen;
    HArrayOfByte *hab;
    char *string;

    check_encoding();
    if (fastEncoding > NO_FAST_ENCODING)
	return makeEncodedCString(s, fastEncoding);

    strLen = javaStringLength(s);
    charToByte = (Hsun_io_CharToByteConverter *)
	execute_java_static_method
	(ee, FindClass(ee, "sun/io/CharToByteConverter", TRUE),
	 "getConverter",
	 "(Ljava/lang/String;)Lsun/io/CharToByteConverter;", 
	 makeJavaString(encoding, strlen(encoding)));

    if (exceptionOccurred(ee)){
	exceptionClear(ee);
	return makeCString(s);
    }

    hab = (HArrayOfByte *)ArrayAlloc
	(T_BYTE, strLen * MB_CUR_MAX + 1);
    
    length = (int)execute_java_dynamic_method
	(ee, (HObject *)charToByte, "convert",
	 "([CII[BII)I", (HArrayOfChar *)(unhand(s)->value),
	 unhand(s)->offset, unhand(s)->offset + unhand(s)->count,
	 hab, 0, strLen * MB_CUR_MAX + 1);

    if (exceptionOccurred(ee)){
	exceptionClear(ee);
	return makeCString(s);
    }

    string = (char *)(unhand(hab)->body);
    string[length] = '\0';

    return string;
}

/*
 * Returns the number of bytes needed to hold a Java string in UTF
 * format, NOT including the terminating NULL.
 *
 * Returns -1 if something is wrong.
 */
int
javaStringUTFLength(HString *s)
{
    Classjava_lang_String *str;
    HArrayOfChar *hac;

    if (s && (str = unhand(s)) && (hac = (HArrayOfChar *)str->value)) {
	unicode *body = unhand(hac)->body;
	int unilength = str->count;
	int result = unicode2utfstrlen(body + str->offset, unilength) - 1;
	KEEP_POINTER_ALIVE(body);
	return result;
    } 
    return -1;
}

/*
 * Fill buf with unicode representation of hstr, upto buflen chars.
 * The UTF string will be NUL-terminated.
 *
 * If both buf and buflen are 0, malloc an appropriately sized
 * buffer for the result.
 *
 * Return result buffer.
 */
char *
javaString2UTF(HString *s, char *buf, int buflen)
{
    Classjava_lang_String *str;
    HArrayOfChar *hac;

    if (s && (str = unhand(s)) && (hac = (HArrayOfChar *)str->value)) {
	unicode *body = unhand(hac)->body;
	int unilength = str->count;
	char *result = unicode2utf(body + str->offset, unilength, buf, buflen);
	KEEP_POINTER_ALIVE(body);
	return result;
    } else if (buflen) {
	buf[0] = '\0';
    }
    return buf;
}

/*
 * Create a Java String whose characters are initialized from
 * the supplied NUL-terminated UTF string.
 */
HString *
makeJavaStringUTF(char *str)
{
    HArrayOfChar *val;
    unicode *unistr;
    int unilength = utfstrlen(str);

    val = (HArrayOfChar *) ArrayAlloc(T_CHAR, unilength);
    if (val == 0) {
        SignalError(0, JAVAPKG "OutOfMemoryError", 0);
        return 0;
    }
    unistr = unhand(val)->body;
    utf2unicode(str, unistr, unilength, &unilength);
    KEEP_POINTER_ALIVE(unistr);
    return (Hjava_lang_String *)
	execute_java_constructor(EE(), 0, classJavaLangString, "([C)", val);
}
