/*
 * @(#)plumbNum.c	1.6 98/07/01
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * plumbNum.c -- Source file for static methods comprising Java wrappers
 * for a subset of Colin Plumb's arbitrary precision positive integer
 * library.  These really are just wrappers; they don't add any functionality,
 * and they preserve all of Plumb's calls' quirks.  Look at Plumb's bn.doc
 * for descriptions of what each call does.
 */

#include "java_math_BigInteger.h"
#include "bn.h"
#include "prime.h"

typedef struct BigNum BigNum;

/* Forward refs for static functions */
static int byteArrayToBn(HArrayOfByte *byteArrayH, BigNum *bnP);
static HArrayOfByte *bnToByteArray(BigNum *bnP);
static int bnCall(int call);

/*
 * RAISE_MEMORY_EXC - Raise a Java OutOfMemoryError with the given text.
 */
#define RAISE_MEMORY_EXC(text)					\
    SignalError(EE(), "java/lang/OutOfMemoryError", text)


void java_math_BigInteger_plumbInit(struct Hjava_math_BigInteger * unused)
{
    bnInit();
}

HArrayOfByte *java_math_BigInteger_plumbAdd(
    struct Hjava_math_BigInteger *unused, HArrayOfByte *aH, HArrayOfByte *bH)
{
    BigNum	 a, b;
    HArrayOfByte *sumH = NULL;

    bnBegin(&a); bnBegin(&b);

    if (byteArrayToBn(aH, &a) && byteArrayToBn(bH, &b)
	&& bnCall(bnAdd(&a,&b)))
	sumH = bnToByteArray(&a);

    bnEnd(&a); bnEnd(&b);
    return sumH;
}

struct Hjava_math_BigInteger *java_math_BigInteger_plumbSubtract(
    struct Hjava_math_BigInteger *unused, HArrayOfByte *aH, HArrayOfByte *bH)
{
    BigNum	   	          a, b;
    int			          rc;
    long	   	          diffSignum;
    HArrayOfByte   	          *diffMagnitudeH;
    struct Hjava_math_BigInteger  *diffH = NULL;

    bnBegin(&a); bnBegin(&b);

    if (byteArrayToBn(aH, &a) && byteArrayToBn(bH, &b)) {
	bnCall(rc = bnSub(&a, &b));

	/* If subtraction succeeded, pack sign and magnitude into BigInteger */
	if (rc != -1 && ((diffMagnitudeH = bnToByteArray(&a)) != NULL)) {
	    diffSignum = (rc == 1 ? -1 :
			  (obj_length(diffMagnitudeH) == 0 ? 0 : 1));
	    diffH = (struct Hjava_math_BigInteger *) execute_java_constructor(
			EE(), "java/math/BigInteger", NULL,
			"([BI)", diffMagnitudeH, diffSignum);
	}
    }

    bnEnd(&a); bnEnd(&b);
    return diffH;
}

HArrayOfByte *java_math_BigInteger_plumbMultiply(
    struct Hjava_math_BigInteger *unused, HArrayOfByte *aH, HArrayOfByte *bH)
{
    BigNum	 a, b, dest;
    HArrayOfByte *productH = NULL;

    bnBegin(&a); bnBegin(&b); bnBegin(&dest);

    if (byteArrayToBn(aH, &a) && byteArrayToBn(bH, &b)
	&& bnCall(bnMul(&dest, &a, &b)))
	productH = bnToByteArray(&dest);

    bnEnd(&a); bnEnd(&b); bnEnd(&dest);
    return productH;
}

HArrayOfByte *java_math_BigInteger_plumbDivide(
    struct Hjava_math_BigInteger *unused, HArrayOfByte *aH, HArrayOfByte *bH)
{
    BigNum	 a, b, q, r;
    HArrayOfByte *quotientH = NULL;

    bnBegin(&a); bnBegin(&b); bnBegin(&q); bnBegin(&r);

    if (byteArrayToBn(aH, &a) && byteArrayToBn(bH, &b)
	&& bnCall(bnDivMod(&q, &r, &a, &b)))
	quotientH = bnToByteArray(&q);

    bnEnd(&a); bnEnd(&b); bnEnd(&q); bnEnd(&r);
    return quotientH;
}

HArrayOfByte *java_math_BigInteger_plumbRemainder(
    struct Hjava_math_BigInteger *unused, HArrayOfByte *aH, HArrayOfByte *bH)
{
    BigNum	 a, b, r;
    HArrayOfByte *remainderH = NULL;

    bnBegin(&a); bnBegin(&b); bnBegin(&r);

    if (byteArrayToBn(aH, &a) && byteArrayToBn(bH, &b)
	&& bnCall(bnMod(&r, &a, &b)))
	remainderH = bnToByteArray(&r);

    bnEnd(&a); bnEnd(&b); bnEnd(&r);
    return remainderH;
}

HArrayOfArray *java_math_BigInteger_plumbDivideAndRemainder(
    struct Hjava_math_BigInteger *unused, HArrayOfByte *aH, HArrayOfByte *bH)
{
    BigNum	   a, b, q, r;
    HArrayOfArray  * resultH = NULL;

    bnBegin(&a); bnBegin(&b); bnBegin(&q); bnBegin(&r);

    if (byteArrayToBn(aH, &a) && byteArrayToBn(bH, &b)
	&& bnCall(bnDivMod(&q, &r, &a, &b))) {
	/* Create a Byte array array to return results. */
	if ((resultH = (HArrayOfArray *) ArrayAlloc(T_CLASS, 2)) == NULL) {
	    RAISE_MEMORY_EXC("Allocating Java byte array array");
	} else {
	    unhand(resultH)->body[0] = (HObject *) bnToByteArray(&q);
	    unhand(resultH)->body[1] = (HObject *) bnToByteArray(&r);
	    unhand(resultH)->body[2] = (HObject *) FindClass(EE(), "[B", TRUE);
	}
    }

    bnEnd(&a); bnEnd(&b); bnEnd(&q); bnEnd(&r);
    return resultH;
}

HArrayOfByte *java_math_BigInteger_plumbGcd(
    struct Hjava_math_BigInteger *unused, HArrayOfByte *aH, HArrayOfByte *bH)
{
    BigNum	 a, b, dest;
    HArrayOfByte *gcdH = NULL;

    bnBegin(&a); bnBegin(&b); bnBegin(&dest);

    if (byteArrayToBn(aH, &a) && byteArrayToBn(bH, &b)
	&& bnCall(bnGcd(&dest, &a, &b)))
	gcdH = bnToByteArray(&dest);

    bnEnd(&a); bnEnd(&b); bnEnd(&dest);
    return gcdH;
}

HArrayOfByte *java_math_BigInteger_plumbModPow(
    struct Hjava_math_BigInteger *unused,
    HArrayOfByte *aH, HArrayOfByte *bH, HArrayOfByte *mH)
{
    BigNum	 a, b, m, dest;
    HArrayOfByte * resultH = NULL;

    bnBegin(&a); bnBegin(&b); bnBegin(&m); bnBegin(&dest);

    if (byteArrayToBn(aH, &a) && byteArrayToBn(bH, &b) && byteArrayToBn(mH, &m)
	&& bnCall(bnExpMod(&dest, &a, &b, &m)))
	resultH = bnToByteArray(&dest);

    bnEnd(&a); bnEnd(&b); bnEnd(&m); bnEnd(&dest);
    return resultH;
}

HArrayOfByte *java_math_BigInteger_plumbModInverse(
    struct Hjava_math_BigInteger *unused, HArrayOfByte *aH, HArrayOfByte *mH)
{
    BigNum	 a, m, dest;
    HArrayOfByte *modInvH = NULL;

    bnBegin(&a); bnBegin(&m); bnBegin(&dest);

    if (byteArrayToBn(aH, &a) && byteArrayToBn(mH, &m)
	&& bnCall(bnInv(&dest, &a, &m)))
	modInvH = bnToByteArray(&dest);

    bnEnd(&a); bnEnd(&m); bnEnd(&dest);
    return modInvH;
}

HArrayOfByte *java_math_BigInteger_plumbSquare(
    struct Hjava_math_BigInteger *unused, HArrayOfByte *aH)
{
    BigNum	 a, dest;
    HArrayOfByte *squareH = NULL;

    bnBegin(&a); bnBegin(&dest);

    if (byteArrayToBn(aH, &a) && bnCall(bnSquare(&dest, &a)))
	squareH = bnToByteArray(&dest);

    bnEnd(&a); bnEnd(&dest);
    return squareH;
}

HArrayOfByte *java_math_BigInteger_plumbGeneratePrime(
    struct Hjava_math_BigInteger *unused, HArrayOfByte *aH)
{
    BigNum	 a;
    HArrayOfByte *primeH = NULL;

    bnBegin(&a);

    if (byteArrayToBn(aH, &a) && bnCall(primeGen(&a, NULL, NULL, NULL, 0)))
	primeH = bnToByteArray(&a);

    bnEnd(&a);
    return primeH;
}


/*
 * byteArrayToBn - Converts a Java byte array into a bigNum.  Returns true
 * if the conversion succeeds, false if out of memory.  It is assumed that
 * the bigNum has already been "begun" (via bnBegin(bnP);).  If out of memory,
 * throws a Java exception.
 */
static int byteArrayToBn(HArrayOfByte *byteArrayH, BigNum *bnP)
{
    unsigned char *body = (unsigned char *) unhand(byteArrayH)->body;
    if (bnInsertBigBytes(bnP, body, 0, obj_length(byteArrayH)) == -1) {
        KEEP_POINTER_ALIVE(body);
	RAISE_MEMORY_EXC("Allocating C bignum");
	return FALSE;
    }

    return TRUE;
}

/*
 * bnToByteArray - Converts a bigNum into a Java byte array.  If out of memory,
 * returns NULL and throws a Java exception.
 */
static HArrayOfByte *bnToByteArray(BigNum *bnP)
{
    HArrayOfByte  *resultH;
    int		  len = (bnBits(bnP) + 7)/8;

    if ((resultH = (HArrayOfByte *) ArrayAlloc(T_BYTE, len)) == NULL)
	RAISE_MEMORY_EXC("Allocating Java byte array");
    else {
        unsigned char *body = (unsigned char *)unhand(resultH)->body;
	bnExtractBigBytes(bnP, body, 0, len);
	KEEP_POINTER_ALIVE(body);
    }
    return resultH;
}

/*
 * bnCall - Interpret the return code from an int-returning call in the bignum
 * package: if out of memory, throws a Java exception. Returns true if the
 * call succeeds.
 */

static int bnCall(int rc)
{
    int success = (rc != -1);

    if (!success)
	RAISE_MEMORY_EXC("Executing C bignum arithmetic operation");

    return success;
}
