/*
 * @(#)classinitialize.c	1.2 98/11/19
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*-
 *      Runtime routines to support classes.
 */

#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <fcntl.h>
#include <stdlib.h>

#include "timeval.h"
#include "oobj.h"
#include "tree.h"
#include "interpreter.h"
#include "bool.h"
#include "threads.h"
#include "monitor.h"
#include "exceptions.h"
#include "javaString.h"
#include "path.h"
#include "byteorder.h"
#include "utf.h"

sys_mon_t *_stringhash_lock;
sys_mon_t *_nametypehash_lock;

static bool_t Locked_ResolveClassConstantField(unsigned type,
					       ClassClass *class,
					       cp_item_type *constant_pool, 
					       unsigned index, 
					       struct execenv *ee);
static bool_t Locked_ResolveClassConstant(ClassClass *Class, cp_item_type *, 
					  unsigned, struct execenv *, unsigned);

#undef CONSTANT_POOL_TYPE_TABLE_PUT
#undef CONSTANT_POOL_TYPE_TABLE_SET_RESOLVED
void CONSTANT_POOL_TYPE_TABLE_PUT(unsigned char *cp,int i,int v) {
sysAssert(cp != 0);
sysAssert(i>=0);
sysAssert(i<=0xFFFF);
    CONSTANT_POOL_TYPE_TABLE_GET(cp,i) = (v);
}
void CONSTANT_POOL_TYPE_TABLE_SET_RESOLVED(unsigned char *cp,int i) {
sysAssert(cp != 0);
sysAssert(i>=0);
sysAssert(i<=0xFFFF);
    CONSTANT_POOL_TYPE_TABLE_GET(cp,i) |= CONSTANT_POOL_ENTRY_RESOLVED;
}

/* Perform any necessary initialization */

void InitializeClassConstantResolver() { 
    _stringhash_lock = (sys_mon_t *) sysMalloc(sysMonitorSizeof());
    memset((char*) _stringhash_lock, 0, sysMonitorSizeof());
    STRINGHASH_INIT();

    _nametypehash_lock = (sys_mon_t *) sysMalloc(sysMonitorSizeof());
    memset((char*) _nametypehash_lock, 0, sysMonitorSizeof());
    NAMETYPEHASH_INIT();
}

static void
InitOneClassIfNeeded(ClassClass* cb)
{
    if (CCIs(cb, Error) || !cbIsInROM(cb) || ROMIs(cb, Initialized)) {
	return;
    }
    monitorEnter(obj_monitor(cb));
    if (ROMFirstEncounter(cb)) {
        sysAssert(ROMIs(cbSuperclass(cb), Initialized) || 
		  ROMIs(cbSuperclass(cb), Initializing));
        ROMSet(cb, Initializing);
        RunStaticInitializers(cb);
        ROMSet(cb, Initialized);
    }
    monitorExit(obj_monitor(cb));
}

void
InitClassIfNeeded(ClassClass* cb)
{
    ClassClass *local[SUPER_INIT_CHAIN_SIZE];
    ClassClass **sclasses = local; /* Set the initial buffer */
    int size;
    int k;

    /*
     * A quick check to avoid extra work. Note that checking 
     * '!ROMFirstEncounter()' is fine even without having a lock.
     */
    if (CCIs(cb, Error) || !cbIsInROM(cb) || ROMIs(cb, Initialized)
	|| !ROMFirstEncounter(cb)) {
	return;
    }
    getSuperHier(cb, &sclasses, &size);
    if (size < 0) { /* Out of memory occurred */
	SignalError(EE(), JAVAPKG "OutOfMemoryError", 0);
	return;
    }
    /*
     * Now sclasses [0 ... size -1] has the class hierarchy. Initialize 
     * each of them. The higher indicates have superclasses.
     */
    for (k = size - 1; k >= 0; k--) {
	InitOneClassIfNeeded(sclasses[k]);
    }
    /*
     * Free sclasses, if getSuperclassHier() returned allocated buffer.
     */
    if (sclasses != local) {
	sysFree(sclasses);
    }
}

/*  
    Perform runtime-specific initialization on the class.  The
    definition of this routine if different depending on if we are
    compiling the interpreter or not.  Resolve the instance
    prototype, load class constants, and run the static
    initializers, if any.  
*/
/* No longer used.
char *
RuntimeInitClass(struct execenv *ee, ClassClass * cb)
{
    InitializeInvoker(cb);
    InitializeForCompiler(cb);
    if ((verifyclasses == VERIFY_ALL) ||
	  ((verifyclasses == VERIFY_REMOTE) && (cbLoader(cb) != NULL))) {
	if (!VerifyClass(cb))
	    return JAVAPKG "VerifyError";
    }
    RunStaticInitializers(cb);
    return (char *) 0;
}
*/

static bool_t
InitClassConstantClass(cp_item_type *constant_pool,
		       unsigned index, struct execenv *ee)
{
    ClassClass *class;
    unsigned char *type_table =
        constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
    if (type_table == NULL) {
        return TRUE;
    }
    sysAssert(CONSTANT_POOL_TYPE_TABLE_IS_RESOLVED(type_table, index));
    switch (CONSTANT_POOL_TYPE_TABLE_GET_TYPE(type_table, index)) {
    case CONSTANT_InterfaceMethodref: 
    case CONSTANT_Methodref: 
        class = methodclass(constant_pool[index].mb);
	break;
    case CONSTANT_Fieldref:
        class = fieldclass(constant_pool[index].fb);
	break;
    case CONSTANT_Class:
        class = constant_pool[index].clazz;
	break;
    default:
        class = 0;
    }
    if (class) {
        char *details = NULL;
        char *ename = ResolveClass(class, &details);
	if (ename) {
	    if (!exceptionOccurred(ee)) {
	        SignalError(ee, ename, details);
	    }
	    return FALSE;
	}
    }
    return TRUE;
}

/* ResolveClassStringConstant is called by the classloader to resolve the
 * constant pool entry, before assigning a value to a final [i.e. constant]
 * String.
 *
 * For the interpreter, we want this to be a String object with the
 * appropriate initial value. 
 */

bool_t
ResolveClassStringConstant(ClassClass *cb, unsigned index, struct execenv *ee)
{
    /* This function is called before the verifier has ever had a chance
     * to look at the class.  So we should be extra careful.
     */
    union cp_item_type *constant_pool = cbConstantPool(cb);
    unsigned char *type_table = 
	constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
    if (type_table == NULL) {
        sysAssert(cbIsInROM(cb));
        return TRUE;
    }
    if (type_table[index] == (CONSTANT_String | CONSTANT_POOL_ENTRY_RESOLVED))
	return TRUE; 
    else if (type_table[index] == CONSTANT_String) { 
	int string_index = constant_pool[index].i;
	if ((string_index < 1) || 
	    (string_index >= (int)cbConstantPoolCount(cb)) ||
	    (type_table[string_index] != 
	        (CONSTANT_Utf8 | CONSTANT_POOL_ENTRY_RESOLVED)))
	    return FALSE;
	return ResolveClassConstant(constant_pool, index, ee, 
				    1 << CONSTANT_String);
    } else { 
	return FALSE;
    }
}

void 
ResolveInit()
{
}

static bool_t 
ResolveClassConstantFromClass0(ClassClass *class, unsigned index, 
			       struct execenv *ee, unsigned mask)
{
    cp_item_type *constant_pool = cbConstantPool(class);
    unsigned char *type_table = 
	constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;

    if (type_table == NULL) { 
        sysAssert(cbIsInROM(class));
	return TRUE;
    }
    if (CONSTANT_POOL_TYPE_TABLE_IS_RESOLVED(type_table, index)) {
	return TRUE;
    } else { 
	bool_t ret;
	monitorEnter(obj_monitor(class));
	ret = Locked_ResolveClassConstant(class, constant_pool, 
					  index, ee, mask);
	monitorExit(obj_monitor(class));
	return ret;
    }
}    

bool_t
ResolveClassConstantFromClass2(ClassClass *class, unsigned index,
			       struct execenv *ee, unsigned mask,
			       bool_t init)
{
    if (!ResolveClassConstantFromClass0(class, index, ee, mask))
        return FALSE;
    if (init) {
        return InitClassConstantClass(cbConstantPool(class), index, ee);
    }
    return TRUE;
}

bool_t
ResolveClassConstantFromClass(ClassClass *class, unsigned index,
			      struct execenv *ee, unsigned mask)
{
    return ResolveClassConstantFromClass2(class, index, ee, mask, TRUE);
}

static bool_t
ResolveClassConstant0(cp_item_type *constant_pool, unsigned index, 
		      struct execenv *ee, unsigned mask)
{
    unsigned char *type_table = 
	constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
    if (type_table == NULL) { 
        return TRUE;
    } else if (CONSTANT_POOL_TYPE_TABLE_IS_RESOLVED(type_table, index)) {
	return TRUE;
    } else if (ee && ee->current_frame && ee->current_frame->current_method) { 
	ClassClass *class = 
	    methodclass(ee->current_frame->current_method);
	bool_t ret;
	monitorEnter(obj_monitor(class));
	ret = Locked_ResolveClassConstant(class, constant_pool, 
					  index, ee, mask);
	monitorExit(obj_monitor(class));
	return ret;
    } else { 
	/* A private, classless constant pool.  No need to lock anything. */
	return Locked_ResolveClassConstant(NULL, constant_pool, 
					   index, ee, mask);
    }
}

bool_t
ResolveClassConstant2(cp_item_type *cp, unsigned index,
		      struct execenv *ee, unsigned mask,
		      bool_t init)
{
    if (!ResolveClassConstant0(cp, index, ee, mask))
        return FALSE;
    if (init) {
        return InitClassConstantClass(cp, index, ee);
    }
    return TRUE;
}

bool_t
ResolveClassConstant(cp_item_type *cp, unsigned index,
		      struct execenv *ee, unsigned mask)
{
    /*
     * Extra stack check to shrink STACK_REDZONE
     */
    if (!sysThreadCheckStack(STACK_REDZONE)) {
	/* Got a stack overflow. Throw the preallocated error object. */
	exceptionThrow(ee ? ee : EE(), 
		       exceptionInternalObject(IEXC_StackOverflow));
	return FALSE;
    }
    return ResolveClassConstant2(cp, index, ee, mask, TRUE);
}

/*
 * This routine resolves class constants like field names,
 * methodnames, and class names, to their runtime-specific values.
 * Field names are resolved to offsets from the beginning of an
 * instance, methodnames are resolved to pointers to code, and class
 * names are resolved to pointers to the class.
 */
static bool_t
Locked_ResolveClassConstant(ClassClass *current_class, 
			    cp_item_type *constant_pool, 
			    unsigned index, 
			    struct execenv *ee, unsigned mask)
{
    unsigned char *type_table = 
	constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
    ClassClass *class = 0;

    if (type_table == NULL) { 
        sysAssert(cbIsInROM(current_class));
        return TRUE;
    } else if (type_table && !CONSTANT_POOL_TYPE_TABLE_IS_RESOLVED(type_table, index)) {
	unsigned type = CONSTANT_POOL_TYPE_TABLE_GET_TYPE(type_table, index);
	sysAssert((mask & (1 << type)) != 0);
	switch(type) {
	  case CONSTANT_Class: {
	      unsigned name_index = constant_pool[index].i;
	      char *name = constant_pool[name_index].cp;

	      class = FindClassFromClass(ee, name, FALSE, 
					 current_class);

	      if (CONSTANT_POOL_TYPE_TABLE_IS_RESOLVED(type_table, index)) {
		  return TRUE;
	      } else if (class == 0) {
		  /* Conformance: check for underlying errors */
		  if (ee == NULL)
		      ee = EE();
		  if (!exceptionOccurred(ee))
		      SignalError(ee, JAVAPKG "NoClassDefFoundError", name);
		  return FALSE;
	      } else if (VerifyClassAccess(current_class, class, TRUE)) {
		  constant_pool[index].clazz = class;
	      } else {
		  SignalError(ee, JAVAPKG "IllegalAccessError", name);
		  return FALSE;
	      }
	      break;
	  }

	  case CONSTANT_String:  {
	      unsigned value_index = constant_pool[index].i;
	      char *value = constant_pool[value_index].cp; /* string */
	      void **result;
	      STRINGHASH_LOCK();
	      if (Str2ID(&stringHash, value, &result, TRUE) == 0) {
		  STRINGHASH_UNLOCK();
		  SignalError(0, JAVAPKG "OutOfMemoryError", 0);
		  return FALSE;
	      } else if (*result == NULL) { 
		  /* We have never seen this string as a literal constant */
		  int length;
		  int i;
		  HObject *string;
		  HArrayOfChar *so;
		  unicode *body;
		  STRINGHASH_UNLOCK();	/* Don't allocate with this locked */
		  length = utfstrlen(value);
		  /* Allocate a byte array.  Copy bytes in. */
		  so = (HArrayOfChar *)ArrayAlloc(T_CHAR, length);
		  if (so == 0) {
		      SignalError(ee, JAVAPKG "OutOfMemoryError", 0);
		      return FALSE;
		  }
		  body = unhand(so)->body;
		  utf2unicode(value, body, length, &i);
		  KEEP_POINTER_ALIVE(body);
		  sysAssert(i == length);
		  /* 
		   * Call the privileged constructor that doesn't make a
		   * copy of the array
		   */
		  string = execute_java_constructor(PRIVILEGED_EE, 0, 
						    classJavaLangString, 
						    "(II[C)", 0, length, so);
		  /*
		   * Having build the string, see if we still need it.
		   * We can't just smash the old copy with the new one,
		   * because we may have handed out references to the old
		   * copy already, and subsequent ones must be the same.
		   */
		  STRINGHASH_LOCK();
		  if (*result == NULL) {    /* Nobody else has interned it */
		      *result = string;
		  }
		  STRINGHASH_UNLOCK();
		  if (ee == 0) {
		      ee = EE();
		  }
		  if (exceptionOccurred(ee)) { 
		      return FALSE;
		  }
	      } else {
	          STRINGHASH_UNLOCK();
	      }

	      constant_pool[index].cp = *result;
	      break;
 	  } 
	   
	  case CONSTANT_NameAndType: {
	      unsigned key = constant_pool[index].i;
	      unsigned name_index = key >> 16;
	      unsigned type_index = key & 0xFFFF;
	      char *name = constant_pool[name_index].cp;
	      char *type = constant_pool[type_index].cp;
	      constant_pool[index].i = NameAndTypeToHash(name, type);
	      break;
	  }

	  case CONSTANT_InterfaceMethodref: 
	  case CONSTANT_Fieldref:
	  case CONSTANT_Methodref: 
	      if (!Locked_ResolveClassConstantField(type, current_class, 
						    constant_pool, index, ee))
		  return FALSE;
	      break;
	}
	CONSTANT_POOL_TYPE_TABLE_SET_RESOLVED(type_table, index);
    }
    /* Netscape fix: class might be GC'ed before the constant pool 
       type is set to RESOLVED. */
    KEEP_POINTER_ALIVE(class);
    return TRUE;
} 

static bool_t
Locked_ResolveClassConstantField(unsigned type,
				 ClassClass *class, 
				 cp_item_type *constant_pool, 
				 unsigned index, 
				 struct execenv *ee)
{
    unsigned key = constant_pool[index].i; /* constant pool entry */
    unsigned class_index = key >> 16;      /* index of class */
    unsigned name_type_index = key & 0xFFFF; /* index of name/type index */
    unsigned ID;
    char buf[256];
    int len;
    ClassClass *cb, *ocb;
    char *name, *sig;

    if (!Locked_ResolveClassConstant(class, constant_pool, class_index, ee, 
				    1 << CONSTANT_Class))
        return FALSE;
    Locked_ResolveClassConstant(class, constant_pool, name_type_index, ee, 
				1 << CONSTANT_NameAndType);

    ID = constant_pool[name_type_index].i;
    ocb = constant_pool[class_index].clazz;
    {
        char *details = NULL;
	char *ename = LinkClass(ocb, &details);
	if (ename) {
	    if (!exceptionOccurred(ee)) {
	        SignalError(ee, ename, details);
	    }
	    return FALSE;
	}
    }
    switch (type) {
        case CONSTANT_Methodref: {
	    for (cb = ocb; ; cb = cbSuperclass(cb)) { 
		int n = cbMethodsCount(cb);
		struct methodblock *mb = cbMethods(cb);
		for (; --n >= 0; mb++) {
		    if (methodID(mb) == ID) {
			/* Check if erroneously "inheriting" a constructor */
			if (ocb != cb)
			    if (strcmp("<init>", methodname(mb)) == 0)
				goto fail;
			if (VerifyFieldAccess(class, cb, methodaccess(mb), TRUE)) {
			    constant_pool[index].mb = mb;
			    return TRUE;
			} else {
			    classname2string(cbName(ocb), buf, sizeof(buf));
			    len = strlen(buf);
			    NAMETYPEHASH_LOCK();
			    if (((name = ID2Str(nameTypeHash,
                                   (unsigned short)(ID >> 16), 0)) == 0) ||
			        ((sig = ID2Str(nameTypeHash,
                                   (unsigned short)(ID & 0xFFFF), 0)) == 0)) {
				NAMETYPEHASH_UNLOCK();
				SignalError(0, JAVAPKG "OutOfMemoryError", 0);
				return FALSE;
			    }
			    NAMETYPEHASH_UNLOCK();
			    /* On buffer overflow truncate string quietly */
			    (void) jio_snprintf(buf + len, sizeof(buf)-len,
				".%s%s", name, sig);
			    SignalError(ee, JAVAPKG "IllegalAccessError", buf);
			    return FALSE;
			}
		    }
		}
		if (cbSuperclass(cb) == NULL) /* We've hit top of chain */
		    break;
	    }

	  fail:
	    classname2string(cbName(ocb), buf, sizeof(buf));
	    len = strlen(buf);
	    NAMETYPEHASH_LOCK();
	    if (((name = ID2Str(nameTypeHash,
				(unsigned short)(ID >> 16), 0)) == 0) ||
	        ((sig = ID2Str(nameTypeHash,
				(unsigned short)(ID & 0xFFFF), 0)) == 0)) {
		NAMETYPEHASH_UNLOCK();
		SignalError(0, JAVAPKG "OutOfMemoryError", 0);
		return FALSE;
	    }
	    NAMETYPEHASH_UNLOCK();
	    /* On buffer overflow truncate string quietly */
	    (void) jio_snprintf(buf + len, sizeof(buf)-len,
		    ": method %s%s not found", name, sig);
	    SignalError(ee, JAVAPKG "NoSuchMethodError", buf);
	    return FALSE;
	}
	
        case CONSTANT_Fieldref: {
	    for (cb = ocb; ; cb = cbSuperclass(cb)) { 
		int n = cbFieldsCount(cb);
		struct fieldblock *fb = cbFields(cb);
		for (; --n >= 0; fb++) {
		    if (fieldID(fb) == ID) {
			if (VerifyFieldAccess(class, cb, fieldaccess(fb), TRUE)) {
			    constant_pool[index].fb = fb;
			    return TRUE;
			} else {
			    classname2string(cbName(ocb), buf, sizeof(buf));
			    len = strlen(buf);
			    NAMETYPEHASH_LOCK();
			    if ((name = ID2Str(nameTypeHash,
                                       (unsigned short)(ID >> 16), 0)) == 0) {
				NAMETYPEHASH_UNLOCK();
				SignalError(0, JAVAPKG "OutOfMemoryError", 0);
				return FALSE;
			    }
			    NAMETYPEHASH_UNLOCK();
			    /* Reporting error, so truncate string quietly */
			    (void) jio_snprintf(buf + len, sizeof(buf)-len, 
				".%s", name);
			    SignalError(ee, JAVAPKG "IllegalAccessError", buf);
			    return FALSE;
			}
		    }
		}
		if (cbSuperclass(cb) == NULL) 
		    break;
	    }
	    classname2string(cbName(ocb), buf, sizeof(buf));
	    len = strlen(buf);
	    NAMETYPEHASH_LOCK();
	    if ((name = ID2Str(nameTypeHash, 
				(unsigned short)(ID >> 16), 0)) == 0) {
		NAMETYPEHASH_UNLOCK();
		SignalError(0, JAVAPKG "OutOfMemoryError", 0);
		return FALSE;
	    }
	    NAMETYPEHASH_UNLOCK();
	    /* Reporting error, so truncate string quietly */
	    (void) jio_snprintf(buf + len, sizeof(buf)-len, 
				": field %s not found", name);
	    SignalError(ee, JAVAPKG "NoSuchFieldError", buf);
	    return FALSE;
	}

        case CONSTANT_InterfaceMethodref: {
	    /* The imethod table of an interface contains this interface
	     * in the first slot, and all of the superinterfaces in the
	     * remaining slots
	     */
	    struct imethodtable *itable = cbIntfMethodTable(ocb);
	    int icount = itable->icount;
	    int i;
	    sysAssert(icount > 0);
	    sysAssert(itable->itable[0].classdescriptor == ocb);
	    for (i = 0; i < icount; i++) { 
		ClassClass *cb = itable->itable[i].classdescriptor;
		int n = cbMethodsCount(cb);
		struct methodblock *mb = cbMethods(cb);
		for (; --n >= 0; mb++) {
		    if (methodID(mb) == ID) {
			constant_pool[index].mb = mb;
			return TRUE;
		    } 
		}
	    }
	    classname2string(cbName(ocb), buf, sizeof(buf));
	    len = strlen(buf);
	    NAMETYPEHASH_LOCK();
	    if (((name = ID2Str(nameTypeHash,
				(unsigned short)(ID >> 16), 0)) == 0) ||
	        ((sig = ID2Str(nameTypeHash,
				(unsigned short)(ID & 0xFFFF), 0)) == 0)) {
		NAMETYPEHASH_UNLOCK();
		SignalError(0, JAVAPKG "OutOfMemoryError", 0);
		return FALSE;
	    }
	    NAMETYPEHASH_UNLOCK();
	    /* On buffer overflow truncate string quietly */
	    (void) jio_snprintf(buf + len, sizeof(buf)-len,
		    ": method %s%s not found", name, sig);
	    SignalError(ee, JAVAPKG "NoSuchMethodError", buf);
	    return FALSE;
	}
	    
       default:
	   sysAssert(FALSE);
	   return FALSE;
    }
}


/* Verify that current_class can access new_class.  If the classloader_only
 * flag is set, we automatically allow any accesses in which current_class
 * doesn't have a classloader.
 */

bool_t
VerifyClassAccess(ClassClass *current_class, ClassClass *new_class, 
		  bool_t classloader_only) 
{
    if (current_class == NULL) {
	return TRUE;
    } else {
	return ((classloader_only && (cbLoader(current_class) == 0))
		|| (current_class == new_class)
		|| IsPublic(cbAccess(new_class))
		|| IsSameClassPackage(current_class, new_class));
    }
}    



/* Verify that current_class can a field of new_class, where that field's 
 * access bits are "access".  We assume that we've already verified that
 * class can access field_class.
 *
 * If the classloader_only flag is set, we automatically allow any accesses
 * in which current_class doesn't have a classloader.
 */


bool_t 
VerifyFieldAccess(ClassClass *current_class, ClassClass *field_class, 
		  int access, bool_t classloader_only)
{
    if (current_class == 0) {
	return TRUE;
    }
    if ((current_class == field_class) || IsPublic(access))
	return TRUE;
    else if (classloader_only && (cbLoader(current_class) == 0))
	return TRUE;
    if (IsProtected(access)) {
	/* See if current_class is a subclass of field_class */
	ClassClass *cb;
	for (  cb = cbSuperclass(current_class); ; cb = cbSuperclass(cb)) {
	    if (cb == field_class) 
		return TRUE;
	    if (cbSuperclass(cb) == 0) 
		break;
	}
    }
    return (!IsPrivate(access)
	    && IsSameClassPackage(current_class, field_class));
}
