/*
 * @(#)classloader.c	1.128 98/10/26
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*-
 *      Routines for loading and resolving class definitions.
 *      These routines should not depending up on the interpreter or
 *      garbage collector.
 *
 */

#include <string.h>
#include <sys/types.h>
#include <stddef.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <setjmp.h>

#include "oobj.h"
#include "path.h"
#include "interpreter.h"
#include "monitor.h"
#include "tree.h"
#include "bool.h"
#include "signature.h"
#include "exceptions.h"

#include "sys_api.h"
#include "byteorder.h"

#ifdef BREAKPTS
#include "breakpoints.h"
#endif

#ifdef HPROF
#include "vmprofiler.h"
#endif /* HPROF */
 
extern ClassClass *allocClassClass();

void
AddBinClass(ClassClass * cb)
{
    register int i = -1;
    BINCLASS_LOCK();
    if (cbLoader(cb) == 0) {
	for (i = nbinclasses; --i >= 0 && strcmp(cbName(binclasses[i]), cbName(cb)) != 0;);
    }
    if (i < 0) {
	if (nbinclasses >= sizebinclasses)
	    if (binclasses == 0)
		binclasses = (ClassClass **)
		    sysMalloc(sizeof(ClassClass *) * (sizebinclasses = 50));
	    else
		binclasses = (ClassClass **)
		    sysRealloc(binclasses, sizeof(ClassClass *)
			    * (sizebinclasses = nbinclasses * 2));
	if (binclasses == 0)
	    goto unlock;
	i = nbinclasses;
	nbinclasses++;
    }
    binclasses[i] = cb;

unlock:
    BINCLASS_UNLOCK();
#ifdef HPROF
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_CLASS_LOAD)) {
        jvmpi_load_class(cb);
    }
#endif /* HPROF */
#ifdef BREAKPTS
    if (debugging && cb) {
	notify_debugger_of_class_load(EE(), (JHandle *)cb);
    }
#endif
}

void
DelBinClass(ClassClass * cb)
{
    register int i;
    BINCLASS_LOCK();
    for (i = nbinclasses; --i >= 0; )
	if (binclasses[i] == cb) {
	    binclasses[i] = binclasses[--nbinclasses];
	    break;
	}
    BINCLASS_UNLOCK();
}

void
MakeClassSticky(ClassClass *cb)
{
    for(;;) {
        if (!cbIsInROM(cb)) {
	    /* monitorEnter(int_class_monitor(cb));*/
	    CCSet(cb, Sticky);
	    /* monitorExit(int_class_monitor(cb));*/
	}
	if (cbSuperclass(cb) == NULL) 
	    break;
	cb = cbSuperclass(cb);
	if (CCIs(cb, Sticky)) 
	    break;
    } 
}

/* 
 * When the interpreter askes for a file, in the end it's looking
 * for a classblock structure to be created.  The only way it can
 * get one of those is by loading a compiled class.
 *
 * OpenCode tries to open a .class file first.  -2 means it failed
 * because it couldn't open the file.
 *
 * If OpenCode returns a valid fd, that means it is a .class file.
 * If SkipCheckSource is false (meaning we're checking source
 * files), and the source file as specified in the compiled .class
 * file is more recent than the .class file, then the loaded class
 * is thrown away, and OpenCode is called again, only this time,
 * with arguments which will cause it to recompile the source.  In
 * the case of java, it forks javac to do this.  When this happens,
 * it has to reload the resulting .class file.
 *
 * Returns the class when it succeeds, 0 when it fails. 
*/


/* Load a .class file normally from the local disk */
static ClassClass *
LoadClassFromFile(char *fn, char *dir, char *class_name)
{
    extern int OpenCode(char *, char *, char *, struct stat*);
    struct stat st;
    ClassClass *cb = 0;
    int codefd = -1;
    int	retryCount = 0;
    unsigned char *external_class;
    char *detail;

    codefd = OpenCode(fn, NULL, dir, &st);

    /*
     * Extra stack check to shrink STACK_REDZONE
     */
    if (!sysThreadCheckStack(STACK_REDZONE)) {
	/* Got a stack overflow. Throw the preallocated error object. */
	exceptionThrow(EE(), exceptionInternalObject(IEXC_StackOverflow));
	return 0;
    }

again:
    if (codefd < 0)		/* open failed */
	return 0;

    /* Snarf the file into memory. */
    external_class = (unsigned char *)sysMalloc(st.st_size);
    if (external_class == 0)
        goto failed;
    if (sysRead(codefd, external_class, st.st_size) != st.st_size)
        goto failed;
    sysClose(codefd);
    codefd = -1;

    /* Create the internal class */
    cb = allocClassClass();
    if (cb == NULL || 
	!createInternalClass(external_class, external_class + st.st_size, 
			     cb, NULL, class_name, &detail)) {
	sysFree(external_class);
	goto failed;
    }
    sysFree(external_class);
    
    if (verbose)
	jio_fprintf(stderr, "[Loaded %s]\n", fn);
    return cb;
failed:
    if (codefd >= 0)
	sysClose(codefd);
    if (cb != 0)
        FreeClass(cb);
    return 0;
}

/* Load a .class file normally from a zip archive */
static ClassClass *
LoadClassFromZip(char *fn, zip_t *zip, char *class_name)
{
    struct stat st;
    ClassClass *cb;
    unsigned char *clazz;
    char *detail;

    /*
     * Extra stack check to shrink STACK_REDZONE
     */
    if (!sysThreadCheckStack(STACK_REDZONE)) {
	/* Got a stack overflow. Throw the preallocated error object. */
	exceptionThrow(EE(), exceptionInternalObject(IEXC_StackOverflow));
	return 0;
    }

    if (!zip_stat(zip, fn, &st)) {
	return 0;
    }
    if ((clazz = sysMalloc(st.st_size)) == 0) {
	return 0;
    }
    if (!zip_get(zip, fn, clazz, st.st_size)) {
	sysFree(clazz);
	return 0;
    }
    if ((cb = allocClassClass()) == 0) {
	sysFree(clazz);
	return 0;
    }
    if (!createInternalClass(clazz, clazz + st.st_size, cb, NULL, class_name,
			     &detail)) {
	sysFree(clazz);
	return 0;
    }
    sysFree(clazz);
    if (verbose) {
	jio_fprintf(stderr, "[Loaded %s from %s]\n", fn, zip->fn);
    }
    return cb;
}

/*
 * Find a class file that's somewhere local, and not from a classloader.  
 * We still need to search for it using the classpath.
 */

ClassClass *LoadClassLocally(char *name)
{
    ClassClass *cb = 0;
    cpe_t **cpp;

    if (name[0] == DIR_SEPARATOR || name[0] == SIGNATURE_ARRAY) 
	return 0;

    for (cpp = sysGetClassPath(); cpp && *cpp != 0; cpp++) {
	cpe_t *cpe = *cpp;
	char path[255];
	if (cpe->type == CPE_DIR) {
	    if (jio_snprintf(path, sizeof(path), 
			     "%s%c%s." JAVAOBJEXT, cpe->u.dir,
			     LOCAL_DIR_SEPARATOR, name) == -1) {
		return 0;
	    }
	    if (cb = LoadClassFromFile(sysNativePath(path), cpe->u.dir, name)) {
		return cb;
	    }
	} else if (cpe->type == CPE_ZIP) {
	    if (jio_snprintf(path, sizeof(path), "%s." JAVAOBJEXT, name) == -1){
		return 0;
	    }
	    if (cb = LoadClassFromZip(path, cpe->u.zip, name)) {
		return cb;
	    }
	}
    }
    return cb;
}

/*
 * A class loading context
 */
typedef struct _CICcontext {
    unsigned char *ptr;            /* The beginning of the .class data    */
    unsigned char *after_cp_ptr;   /* A pointer to right after the c.p.   */
    unsigned char *end_ptr;        /* The end of the .class data          */
    ClassClass *cb;                /* The class that is being loaded      */
    union cp_item_type *cp;        /* The constant pool of the class      */
    unsigned short minor_version;
    jmp_buf jump_buffer;           /* The context to 'rewind' to on error */
    char **detail;                 /* Space to fill in any error message  */

    /* 
     * We keep various different class loading memory buffers: One for
     * the constant pool entries and strings, one for the clinit
     * buffer, and one for all the rest of the data structures.  
     */
#define CLINIT_BUFF    0
#define MAIN_BUFF      1
#define NO_BUFF_TYPES  2
    int buffer_sizes[NO_BUFF_TYPES];
    unsigned char *buffers[NO_BUFF_TYPES];
    unsigned char *buffer_ptrs[NO_BUFF_TYPES];

} CICcontext;

#define ALIGN4(v) (((v) + 3) & ~3)

/*
 * The skip<n>bytes macros do truncation checks and skip over .class data.
 *
 * The _quick macros don't do the truncated class file checks.
 */
#define skipNbytes_quick(context, n)		\
    (context)->ptr += n

#define skipNbytes(context, n)			\
    if (context->end_ptr - context->ptr < n)	\
	JAVA_ERROR("Truncated class file");	\
    skipNbytes_quick(context, n)

#define skip1byte_quick(context)  skipNbytes_quick(context, 1)
#define skip2bytes_quick(context) skipNbytes_quick(context, 2)
#define skip4bytes_quick(context) skipNbytes_quick(context, 4)

#define skip1byte(context)        skipNbytes(context, 1)
#define skip2bytes(context)       skipNbytes(context, 2)
#define skip4bytes(context)       skipNbytes(context, 4)

#define skipAsciz_quick           skip2bytes_quick
#define skipAsciz                 skip2bytes

#define measureNBytes(context,size,buff_type)				\
{									\
    context->buffer_sizes[buff_type] += ALIGN4(size);			\
}

static void *allocNBytes(CICcontext *context, int n, int type)
{
    unsigned char **buffer_ptr = &context->buffer_ptrs[type];
    void *result = *buffer_ptr;
    *buffer_ptr += ALIGN4(n);
    sysAssert(*buffer_ptr <= context->buffers[type] +
	      context->buffer_sizes[type]);
    return result;
}
       
static char *getAsciz(CICcontext *);
static char *getAscizFromClass(CICcontext *, int i);

static unsigned long get1byte(CICcontext *);
static unsigned long get2bytes(CICcontext *);
static unsigned long get4bytes(CICcontext *);
static void getNbytes(CICcontext *, int count, char *buffer);

static void freeBuffers(CICcontext *);

static union cp_item_type *LoadConstantPool(CICcontext *);
static void ReadInCode(CICcontext *, struct methodblock *, int);
static void ReadExceptions(CICcontext *, struct methodblock *, int);

#ifdef CLASSDEBUGINFO
static void ReadLineTable(CICcontext *context, struct methodblock *mb, int);
static void ReadLocalVars(CICcontext *context, struct methodblock *mb, int);
#endif

#ifdef JCOV
static void ReadCovTable(CICcontext *, struct methodblock *mb, int);
#endif
static void InitializeStaticVar(char *fsig, struct fieldblock *fb, CICcontext *context);

static void
measureAndVerifyClass(CICcontext *context,
		      struct Hjava_lang_ClassLoader *loader);

static void
createInternalClass0(CICcontext *context, ClassClass *cb, 
		     struct Hjava_lang_ClassLoader *loader, char *name);

bool_t
createInternalClass1(unsigned char *ptr, unsigned char *end_ptr, 
		     ClassClass *cb, struct Hjava_lang_ClassLoader *loader, 
		     char *name, char **detail);

bool_t
createInternalClass(unsigned char *ptr, unsigned char *end_ptr, 
		    ClassClass *cb, struct Hjava_lang_ClassLoader *loader, 
		    char *name, char **detail)
{
    bool_t res;

#if !defined(NO_CLASSLOAD_HOOK) || defined(HPROF)
    unsigned char *saved_ptr = ptr;
#endif /* !NO_CLASSLOAD_HOOK || HPROF */
#ifndef NO_CLASSLOAD_HOOK
    static classload_hook hook_f = (classload_hook)(-1);

    if (hook_f == (classload_hook)(-1)) {
        char *hook_dll = getenv("_CLASSLOAD_HOOK");
	if (hook_dll) {
	    char buf[256];
	    sysBuildLibName(buf, sizeof(buf), "", hook_dll);
	    sysAddDLSegment(buf);
	    hook_f = (classload_hook)sysDynamicLink("ClassLoadHook");
	} else {
	    hook_f = NULL;
	}
    }

    if (hook_f != NULL && hook_f != (classload_hook)(-1)) {
        classload_event event;
	event.class_data = ptr;
	event.class_data_len = end_ptr - ptr;
	event.malloc_f = (void * (*)(int))sysMalloc;
        (*hook_f)(&event);
	if (event.new_class_data == NULL) {
	    return FALSE;
	}
	ptr = event.new_class_data;
	end_ptr = event.new_class_data + event.new_class_data_len;
    }

#endif /* NO_CLASSLOAD_HOOK */

#ifdef HPROF
 
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_CLASS_LOAD_HOOK)) {
        jvmpi_load_class_hook(&ptr, &end_ptr, sysMalloc);
        if (ptr == NULL) {
            ThrowOutOfMemoryError(0, 0);
            return FALSE;
        }
    }
#endif /* HPROF */


    res = createInternalClass1(ptr, end_ptr, cb, loader, name, detail);

#if !defined(NO_CLASSLOAD_HOOK) || defined(HPROF)
    if (ptr != saved_ptr) {
        sysFree(ptr);
    }
#endif /* !NO_CLASSLOAD_HOOK || HPROF */

    return res;
}

#define JAVA_ERROR(name) { *(context->detail) = name; \
                 	       longjmp(context->jump_buffer, 1); }

/* Create an internal class file from the indicated data.  It should be
 * in a buffer for which the first byte is at *ptr and the last byte is just
 * before *end_ptr.
 * The class's classloader is indicated by the classloader argument.
 *
 * We go through the buffer twice. In the first pass we determine the amount
 * of storage needed by the class. We then allocate a single chunk of memory,
 * free the temporary storage, and load from the buffer for the second time.
 *
 * Since all storage needed by the class initialization method <clinit>
 * can be freed after the class is loaded, we count the <clinit> space needs
 * separately and store the <clinit> method in a separate chunk of memory.  */

bool_t
createInternalClass1(unsigned char *ptr, unsigned char *end_ptr, 
		     ClassClass *cb, struct Hjava_lang_ClassLoader *loader, 
		     char *name, char **detail)
{
    CICcontext context_block;
    CICcontext *context = &context_block;
    struct Classjava_lang_Class *ucb = unhand(cb);
    int i;

    sysAssert(LOADCLASS_LOCKED());

    /*
     * Make sure all of the context is initialized.
     */
    memset(context, 0, sizeof(CICcontext));
    
    /* Set up the context */
    context->ptr = ptr;
    context->end_ptr = end_ptr;
    context->cb = cb;
    context->detail = detail;

#ifdef HPROF
    /* Make sure we don't dump a partial class */
    if (jvmpi_event_flags) {
        jvmpi_dump_context_lock();
    }
#endif /* HPROF */

    if (setjmp(context->jump_buffer)) {
	/* We've gotten an error of some sort */

	/* Free all unassigned memory associated with the current
	   loaded class. */
	freeBuffers(context);

#ifdef HPROF
        if (jvmpi_event_flags) {
            jvmpi_dump_context_unlock();
        }
#endif /* HPROF */
	return FALSE;
    }

    /* The first pass allows us to uncover any class format errors
       and find out the size of the various buffers needed. */
    measureAndVerifyClass(context, loader);

    /* rewind the raw class data */
    context->ptr = ptr;   
    
    /*
     * Allocate all the necessary buffers
     */
    for(i = 0; i < NO_BUFF_TYPES; i++) {
	int size_i = context->buffer_sizes[i];
	if (size_i > 0) {
	    unsigned char *space = sysCalloc(1, size_i);
	    if (space != NULL) {
		context->buffers[i] = space;
		context->buffer_ptrs[i] = space;
	    } else {
		JAVA_ERROR("out of memory");
	    }
	}
    }

    /*
     * Now create the internal class
     */
    createInternalClass0(context, cb, loader, name);

    /* Valid class - let's put it in the class table. */
#ifdef HPROF
    if (jvmpi_event_flags) {
	jvmpi_dump_context_unlock();
    }
#endif /* HPROF */
    AddBinClass(cb);
    return TRUE;
}

/*
 * Measure the size of a table that occurs in the .class file.
 *
 * rec_len - The size of each element of the table as it occurs in the .class
 * rec_len_alloc - The size of the internal representation of an element
 */
static int 
measureTable(CICcontext *context, int rec_len, int rec_len_alloc, int bufftype)
{
    int tab_len = get2bytes(context);

    if (tab_len > 0) {
	if (rec_len > 0) {
	    skipNbytes(context, tab_len * rec_len);
	}
	if (rec_len_alloc > 0) {
	    measureNBytes(context, tab_len * rec_len_alloc, bufftype);
	}
    }
    return tab_len;
}

static void
measureAndVerifyClass(CICcontext *context,
		      struct Hjava_lang_ClassLoader *loader)
{
    int i, j;
    union cp_item_type *constant_pool;
    int attribute_count;
    unsigned fields_count;
    unsigned methods_count;
    ClassClass *cb = context->cb;
    struct Classjava_lang_Class *ucb = unhand(cb);
    unsigned short minor_version;
    
    /*
     * Do the file type check now, so that we don't go too far if
     * we are not looking at a class file
     */
    if (get4bytes(context) != JAVA_CLASSFILE_MAGIC) 
	JAVA_ERROR("Bad magic number");

    /* Major version */
    skip2bytes(context);

    /* we are going to need this later */
    minor_version = get2bytes(context);

    context->cp = constant_pool = LoadConstantPool(context);

    /* 
     * Remember this place. We are going to jump to it during
     * createInternalClass0 
     */
    context->after_cp_ptr = context->ptr;

    /* access */
    skip2bytes(context);

    /* name */
    skipAsciz(context);

    /* super_name */
    skipAsciz(context);

    /* Measure implements table */
    measureTable(context, 2, sizeof(short), MAIN_BUFF);

    /* Measure field table */
    fields_count = measureTable(context, 0,
				sizeof(struct fieldblock), MAIN_BUFF);

    for (i = fields_count; --i >= 0; ) {
	char *field_name;
	char *field_sig;
	char isig;
	bool_t isStatic;
	int access;

	access = get2bytes(context) & ACC_WRITTEN_FLAGS;
	isStatic = ((access & ACC_STATIC) != 0) ? TRUE : FALSE;

	/* field name */
	skipAsciz(context);
	field_sig = getAsciz(context);
	attribute_count = get2bytes(context);
	/*
	 * Skip all field attributes
	 */
	for (j = 0; j < (int)attribute_count; j++) {
	    int length;

	    skipAsciz(context);
	    length = get4bytes(context);
	    /*
	     * Skip the field attribute completely; nothing to allocate here
	     */
	    skipNbytes(context, length);
	}
	if (isStatic) {
	    int static_field_mem_size;
	    isig = field_sig[0];
	    if ((isig == SIGNATURE_LONG) ||
		(isig == SIGNATURE_DOUBLE)) {
		static_field_mem_size = 2 * sizeof(long);
	    } else {
		static_field_mem_size = sizeof(long);
	    }
	    /*
	     * Make room for the static
	     */
	    measureNBytes(context, static_field_mem_size, MAIN_BUFF);
	}
    }

    /* The methods table */
    methods_count = measureTable(context, 0,
				 sizeof(struct methodblock), MAIN_BUFF);

    for (i = methods_count; --i >= 0;) {
	int method_access = get2bytes(context) & ACC_WRITTEN_FLAGS;
	char *method_name = getAsciz(context);
	char *method_sig = getAsciz(context);
	int allocType = MAIN_BUFF;

	if (method_name[0] == '<' &&
	    strcmp(method_name, "<clinit>") == 0 &&
	    strcmp(method_sig, "()V") == 0) {
	    allocType = CLINIT_BUFF;
	}

	attribute_count = get2bytes(context);
	for (j = 0; j < attribute_count; j++) {
#ifdef SUPPORTCOMPILER
	    extern char *CompiledCodeAttribute;
#endif
	    char *name = getAsciz(context);
	    int method_attribute_length = get4bytes(context);
	    if ((strcmp(name, "Code") == 0) 
		       && ((method_access & (ACC_NATIVE | ACC_ABSTRACT))==0)) {
		int code_attribute_count;
		int code_len;
		int k;

		if (minor_version <= 2) {
		    skip2bytes(context); /* 1-byte maxstack and nlocals */
		    code_len = get2bytes(context);
		} else {
		    skip4bytes(context); /* 2-byte maxstack and nlocals */
		    code_len = get4bytes(context);
		}
		measureNBytes(context, code_len, allocType);
		skipNbytes(context, code_len);
		measureTable(context,
			     2 + 2 + 2 + 2, /* exception table entry */
			     sizeof(struct CatchFrame),
			     allocType);
		code_attribute_count = get2bytes(context);
		for (k = 0; k < code_attribute_count; k++) {
		    int length;
#ifdef CLASSDEBUGINFO
		    char *code_attr_name = getAsciz(context);
		    length = get4bytes(context);

		    if (strcmp(code_attr_name,
			       "LineNumberTable") == 0) {
			/*
			 * Measure size of line number table
			 */
			measureTable(context, 4, sizeof(struct lineno),
				     allocType);
		    } else if (strcmp(code_attr_name,
				      "LocalVariableTable") == 0) {
			/*
			 * Measure size of local variable table
			 */
			measureTable(context, 10, sizeof(struct localvar),
				     allocType);
		    } else {
			skipNbytes(context, length);
		    }
#else
		    /*
		     * If not looking at class debug information, skip
		     * all code attributes
		     */
		    skipAsciz(context);
		    length = get4bytes(context);
		    skipNbytes(context, length);
#endif
		}
	    } else if (strcmp(name, "Exceptions") == 0) {
		measureTable(context, 2, sizeof(unsigned short), allocType);
	    } else {
		skipNbytes(context, method_attribute_length);
	    }
	}
	allocType = MAIN_BUFF; /* In case we started allocating for <clinit> */
    }

#ifndef JCOV
    /* See if there are class attributes, and skip them all */
    attribute_count = get2bytes(context); 
    for (j = 0; j < attribute_count; j++) {
	int length;

	skipAsciz(context); /* name */
	length = get4bytes(context);
	/* No allocations for class attributes */
	skipNbytes(context, length); 
    }
#else
#error "JCOV class loading not implemented"
#endif /* !JCOV */
}

static void
createInternalClass0(CICcontext *context, ClassClass *cb, 
		     struct Hjava_lang_ClassLoader *loader, char *name)
{
    int i, j;
    union cp_item_type *constant_pool;
    unsigned char *type_table;
    int attribute_count;
    unsigned fields_count;
    struct methodblock *mb;
    struct fieldblock *fb;
    struct Classjava_lang_Class *ucb = unhand(cb);
    unsigned short minor_version;
    unsigned short major_version;
    bool_t scannable = FALSE; /* By default the class does not have
				 scannable statics */
    
    /*
     * We already checked the magic number
     */
    skip4bytes_quick(context);

    context->minor_version = minor_version = get2bytes(context);
    major_version = get2bytes(context);

    /*
     * Signal that this class is a loaded class, and not in ROM
     */
    cbBodyRomIndex(ucb) = -1;

    cbBodyLoader(ucb) = loader;
    if (major_version != JAVA_VERSION) 
	JAVA_ERROR("Bad major version number");

    constant_pool = cbBodyConstantPool(ucb);

    /*
     * Skip the constant pool reading -- measureAndVerifyClass() did that
     */
    context->ptr = context->after_cp_ptr;

    type_table = constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;

    cbBodyAccess(ucb) = get2bytes(context) & ACC_WRITTEN_FLAGS;

    /* Get the name of the class */
    i = get2bytes(context);	/* index in constant pool of class */
    cbBodyName(ucb) = getAscizFromClass(context, i);
    if (name != NULL && strcmp(cbBodyName(ucb), name) != 0)
	JAVA_ERROR("Wrong name");
    constant_pool[i].clazz = cb;
    CONSTANT_POOL_TYPE_TABLE_SET_RESOLVED(type_table, i);

    if (loader) { 
	/* We don't trust a classloader to do the right thing. . . */
	ClassClass **pcb, **end_pcb;
	char *name = cbBodyName(ucb);
	if (name == NULL || !IsLegalClassname(name, FALSE)) {
	    JAVA_ERROR("Bad name");
	}
	BINCLASS_LOCK();
	for (pcb = binclasses, end_pcb = pcb + nbinclasses; 
	         pcb < end_pcb; pcb++) { 
	    ClassClass *cb = *pcb;
	    if ((cbLoader(cb) == loader) && (strcmp(name, cbName(cb)) == 0))
	        break;
	}
	BINCLASS_UNLOCK();
	if (pcb < end_pcb)
	    /* There's already a class with the same name and loader */
	    JAVA_ERROR("Duplicate name");
    }

    /* Get the super class name. */
    i = get2bytes(context);	/* index in constant pool of class */
    if (i > 0) {
        char *supername = getAscizFromClass(context, i);
	if (!IsLegalClassname(supername, FALSE)) { 
	    JAVA_ERROR("Bad superclass name");
	} else { 
	    /* Hack.  We rezero this slot when we initialize the class */
	    cbBodyMethodTable(ucb) = (void *)supername;
	}
    } 

    i = cbBodyImplementsCount(ucb) = get2bytes(context);
    if (i > 0) {
	unsigned j;
	cbBodyImplements(ucb) = allocNBytes(context, i * sizeof(short),
					    MAIN_BUFF);
	for (j = 0; j < i; j++) {
	    cbBodyImplements(ucb)[j] = get2bytes(context);
	}
    }

    fields_count = cbBodyFieldsCount(ucb) = get2bytes(context);
    if (fields_count > 0) 
        cbBodyFields(ucb) = (struct fieldblock *)
	    allocNBytes(context,
			cbBodyFieldsCount(ucb) * sizeof(struct fieldblock),
			MAIN_BUFF);
    for (i = fields_count, fb = cbBodyFields(ucb); --i >= 0; fb++) {
	char *field_name;
	char *field_sig;
	char isig;
	bool_t isStatic;

	fieldclass(fb) = cb;
	fieldaccess(fb) = get2bytes(context) & ACC_WRITTEN_FLAGS;
	isStatic = ((fieldaccess(fb) & ACC_STATIC) != 0) ? TRUE : FALSE;

	field_name = getAsciz(context);
	field_sig = getAsciz(context);
	fieldID(fb) = NameAndTypeToHash(field_name, field_sig);
	attribute_count = get2bytes(context);
	for (j = 0; j < (int)attribute_count; j++) {
	    char *name = getAsciz(context);
	    int length = get4bytes(context);
	    if (isStatic && strcmp(name, "ConstantValue") == 0) {
		if (length != 2) 
		    JAVA_ERROR("Wrong size for VALUE attribute");
		fieldaccess(fb) |= ACC_VALKNOWN;
		/* we'll change this below */
		fieldoffset(fb) = get2bytes(context); 
	    } else {
		getNbytes(context, length, NULL);
	    }
	}
	if (isStatic) {
	    InitializeStaticVar(field_sig, fb, context);
	}
	isig = field_sig[0];
	if ((isig == SIGNATURE_ARRAY) ||
	    (isig == SIGNATURE_CLASS)) {
	    fieldaccess(fb) |= ACC_REFERENCE;
	    /*
	     * If the class has scannable statics, mark it so
	     */
	    if (!scannable && isStatic) {
		CCSet(cb, Scannable);
		scannable = TRUE;
	    }
	}
	if ((isig == SIGNATURE_LONG) ||
	    (isig == SIGNATURE_DOUBLE)) {
	    fieldaccess(fb) |= ACC_DOUBLEWORD;
	}
    }

    if ((cbBodyMethodsCount(ucb) = get2bytes(context)) > 0) 
	cbBodyMethods(ucb) = (struct methodblock *)
	  allocNBytes(context,
		      cbBodyMethodsCount(ucb) * sizeof(struct methodblock),
		      MAIN_BUFF);
    for (i = cbMethodsCount(cb), mb = cbMethods(cb); --i >= 0; mb++) {
	char *method_name;
	char *method_sig;
	int allocType = MAIN_BUFF;

	methodclass(mb) = cb;
	methodaccess(mb) = get2bytes(context) & ACC_WRITTEN_FLAGS;
	method_name = getAsciz(context);
	method_sig = getAsciz(context);
	methodID(mb) = NameAndTypeToHash(method_name, method_sig);

	if (method_name[0] == '<' &&
	    strcmp(method_name, "<clinit>") == 0 &&
	    strcmp(method_sig, "()V") == 0) {
	    allocType = CLINIT_BUFF;
	}

	methodargssize(mb) = Signature2ArgsSize(method_sig) 
	                + ((methodaccess(mb) & ACC_STATIC) ? 0 : 1);
	if (methodargssize(mb) > 255)
	    JAVA_ERROR("Too many arguments");
	attribute_count = get2bytes(context);
	for (j = 0; j < attribute_count; j++) {
#ifdef SUPPORTCOMPILER
	    extern char *CompiledCodeAttribute;
#endif
	    char *name = getAsciz(context);
#ifdef SUPPORTCOMPILER
	    if (CompiledCodeAttribute != NULL 
		     && strcmp(name, CompiledCodeAttribute) == 0
		     && cbLoader(cb) == 0) {
		int attribute_length = get4bytes(context);
		ReadInCompiledCode(context, mb, attribute_length, 
				   get1byte, get2bytes, get4bytes, getNbytes);
	    } else
#endif
	    if ((strcmp(name, "Code") == 0) 
		       && ((methodaccess(mb) & (ACC_NATIVE | ACC_ABSTRACT))==0)) {
		ReadInCode(context, mb, allocType);
	    } else if (strcmp(name, "Exceptions") == 0) {
		ReadExceptions(context, mb, allocType);
	    } else {
		int length = get4bytes(context);
		skipNbytes_quick(context, length);
	    }
	}
	allocType = MAIN_BUFF;
    }

    /* See if there are class attributes */
    attribute_count = get2bytes(context); 
    for (j = 0; j < attribute_count; j++) {
	int length;
	
	/* 
	 * To pass the pJCK, we need to check if:
	 *  o SourceFile_attribute is 2
	 *  o sourcefile_index is a CONSTANT_Utf8_info
	 * even if we will discard the information.
	 */
	char *name = getAsciz(context);
	length = get4bytes(context);
	if (strcmp(name, "SourceFile") == 0) {
	    if (length != 2) {
		JAVA_ERROR("Wrong size for VALUE attribute");
	    }
#ifdef CLASSDEBUGINFO
	    cbBodySourceName(ucb) = getAsciz(context);
#else
	    /* 
	     * Use getAsciz() rather than skipAsciz_quick() 
	     * to check if the attribute_name_index item is valid 
	     * (this is required to pass pJCK).
	     */
	    getAsciz(context);
#endif
	} else {
	    skipNbytes_quick(context, length);
	}
    }
}

/* Create a fake array class that has the specified fields */

ClassClass *
createFakeArrayClass(char *name,           /* name */
		     int base_type,        /* base_type */
		     int depth,            /* array dimension */
		     ClassClass *inner_cb, /* base type if T_CLASS */
		     struct Hjava_lang_ClassLoader *loader)
{ 
    ClassClass *cb = allocClassClass();
    Classjava_lang_Class *ucb = unhand(cb);
    cp_item_type *constant_pool = 
	sysCalloc(CONSTANT_POOL_ARRAY_LENGTH, 
		  (sizeof(cp_item_type) + sizeof(unsigned char)));
    unsigned char *type_table = (unsigned char *)
        (constant_pool + CONSTANT_POOL_ARRAY_LENGTH);
    sysAssert(name[0] == SIGNATURE_ARRAY);
    cbBodyRomIndex(ucb) = -1;
    cbBodyName(ucb) = strDup(name);
    cbBodySuperclass(ucb) = classJavaLangObject;
    cbBodyConstantPool(ucb) = constant_pool;
    cbBodyConstantPoolCount(ucb) = CONSTANT_POOL_ARRAY_LENGTH;
    cbBodyLoader(ucb) = loader;

    constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type = type_table;
    constant_pool[CONSTANT_POOL_ARRAY_DEPTH_INDEX].i = depth;     
    constant_pool[CONSTANT_POOL_ARRAY_TYPE_INDEX].i = base_type;
    type_table[CONSTANT_POOL_ARRAY_DEPTH_INDEX] = 
	CONSTANT_Integer | CONSTANT_POOL_ENTRY_RESOLVED;
    type_table[CONSTANT_POOL_ARRAY_TYPE_INDEX] = 
	CONSTANT_Integer | CONSTANT_POOL_ENTRY_RESOLVED;
	        
    if (base_type == T_CLASS) {
	/* Initialize the appropriate fields of the constant pool */
	constant_pool[CONSTANT_POOL_ARRAY_CLASS_INDEX].clazz = inner_cb;
	type_table[CONSTANT_POOL_ARRAY_CLASS_INDEX] = 
	    CONSTANT_Class | CONSTANT_POOL_ENTRY_RESOLVED;
	/* The class is public iff its base class is public */
	cbBodyAccess(ucb) = ACC_FINAL | ACC_ABSTRACT | (cbAccess(inner_cb) & ACC_PUBLIC);
    } else {
	/* Set the class field to something innocuous */
	type_table[CONSTANT_POOL_ARRAY_CLASS_INDEX] = 
	    CONSTANT_Integer | CONSTANT_POOL_ENTRY_RESOLVED;
	cbBodyAccess(ucb) = ACC_FINAL | ACC_ABSTRACT | ACC_PUBLIC;
    }
    AddBinClass(cb);
    return cb;
}

/*
 * Create a classblock to represent a primitive type.
 * NOTE: We don't add it to the builtin class table, so this should be
 * called only once per primitive type.
 * See FindPrimitiveClass().
 */
ClassClass *
createPrimitiveClass(char *name, char sig, unsigned char typecode,
    unsigned char slotsize, unsigned char elementsize)
{
    ClassClass *cb = allocClassClass();
    Classjava_lang_Class *ucb = unhand(cb);

    cbBodyRomIndex(ucb) = -1;
    cbBodyName(ucb) = strDup(name);
    cbBodySuperclass(ucb) = classJavaLangObject;
    cbBodyConstantPool(ucb) = NULL;
    cbBodyConstantPoolCount(ucb) = 0;
    cbBodyLoader(ucb) = NULL;
    cbBodyAccess(ucb) = ACC_FINAL | ACC_ABSTRACT | ACC_PUBLIC;

    CCSet(cb, Primitive);
    cbTypeSig(cb) = sig;
    cbTypeCode(cb) = typecode;
    MakeClassSticky(cb);

#ifdef HPROF
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_CLASS_LOAD)) {
        jvmpi_load_class(cb);
    }
#endif /* HPROF */
    return cb;
}

/* Initialize the static variables in a class structore */

static void InitializeStaticVar(char *fsig, struct fieldblock *fb, CICcontext *context) {
    Java8 t1, t2;
    char isig = fsig[0];
    int index = fieldoffset(fb); /* set by "VALUE" attribute, if it's there */

    if (isig == SIGNATURE_DOUBLE || isig == SIGNATURE_LONG) { 
        twoword_static_address(fb) =
	    allocNBytes(context, 2 * sizeof(long), MAIN_BUFF);
    } else { 
        normal_static_address(fb) = allocNBytes(context, 1 * sizeof(long),
						MAIN_BUFF);
    }

    if (fieldaccess(fb) & ACC_VALKNOWN) { 
	ClassClass *cb = context->cb;
	union cp_item_type *constant_pool = cbConstantPool(cb);
	unsigned char *type_table = 
	      constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;

	if (index <= 0 || index >= cbConstantPoolCount(cb)) 
	    JAVA_ERROR("Bad initial value");

	switch (isig) { 
	    case SIGNATURE_DOUBLE: 
		if (type_table[index] != 
		    (CONSTANT_Double | CONSTANT_POOL_ENTRY_RESOLVED))
		    JAVA_ERROR("Bad index into constant pool");
		SET_DOUBLE(t1, twoword_static_address(fb),
			   GET_DOUBLE(t2, &constant_pool[index]));
	    break;

	    case SIGNATURE_LONG: 
		if (type_table[index] != 
		    (CONSTANT_Long | CONSTANT_POOL_ENTRY_RESOLVED))
		    JAVA_ERROR("Bad index into constant pool");
		SET_INT64(t1, twoword_static_address(fb),
			  GET_INT64(t2, &constant_pool[index]));
		break;

	    case SIGNATURE_CLASS:
		if (strcmp(fsig, "L" JAVAPKG "String;") != 0) {
		    JAVA_ERROR("Cannot set initial value for object");
		} else if (!ResolveClassStringConstant(cb, index, 0)) { 
		    JAVA_ERROR("Unable to resolve string");
		} else {
		    *(struct Hjava_lang_String **)normal_static_address(fb) = 
		        constant_pool[index].str;
		}
		break;

	    case SIGNATURE_BYTE: 
		if (type_table[index] != 
		    (CONSTANT_Integer | CONSTANT_POOL_ENTRY_RESOLVED))
		    JAVA_ERROR("Bad index into constant pool");
		*(int *)normal_static_address(fb) = 
		        (signed char)(constant_pool[index].i);
		break;

	    case SIGNATURE_CHAR:
		if (type_table[index] != 
		    (CONSTANT_Integer | CONSTANT_POOL_ENTRY_RESOLVED))
		    JAVA_ERROR("Bad index into constant pool");
		*(int *)normal_static_address(fb) = 
		        (unsigned short)(constant_pool[index].i);
		break;

	    case SIGNATURE_SHORT:
		if (type_table[index] != 
		    (CONSTANT_Integer | CONSTANT_POOL_ENTRY_RESOLVED))
		    JAVA_ERROR("Bad index into constant pool");
		*(int *)normal_static_address(fb) = 
		        (signed short)(constant_pool[index].i);
		break;
	    case SIGNATURE_BOOLEAN:
		if (type_table[index] != 
		    (CONSTANT_Integer | CONSTANT_POOL_ENTRY_RESOLVED))
		    JAVA_ERROR("Bad index into constant pool");
		*(int *)normal_static_address(fb) = 
		       (constant_pool[index].i != 0);
		break;

	    case SIGNATURE_INT:  
		if (type_table[index] != 
		    (CONSTANT_Integer | CONSTANT_POOL_ENTRY_RESOLVED))
		    JAVA_ERROR("Bad index into constant pool");
		*(int *)normal_static_address(fb) = constant_pool[index].i;
		break;
		
	    case SIGNATURE_FLOAT:
		if (type_table[index] != 
		        (CONSTANT_Float | CONSTANT_POOL_ENTRY_RESOLVED))
		    JAVA_ERROR("Bad index into constant pool");
		*(float *)normal_static_address(fb) = constant_pool[index].f;
		break;
			
	    default: 
		JAVA_ERROR("Unable to set initial value");
	}
    } else { 
	switch (isig) { 
	    default:
		*(int *)normal_static_address(fb) = 0;
		break;

	    case SIGNATURE_FLOAT:
		*(float *)normal_static_address(fb) = 0.0;
		break;

	    case SIGNATURE_LONG: 
		SET_INT64(t1, twoword_static_address(fb), ll_zero_const);
		break;

	    case SIGNATURE_DOUBLE: 
		SET_DOUBLE(t1, twoword_static_address(fb), 0.0);
		break;

	    case SIGNATURE_CLASS:
		*(JHandle **)normal_static_address(fb) = NULL;
		break;
	}
    }
}

static union cp_item_type *LoadConstantPool(CICcontext *context) 
{
    ClassClass *cb = context->cb;
    int nconstants = get2bytes(context);
    cp_item_type *constant_pool;
    unsigned char *type_table;
    int i;
    Java8 t1;
    int total_space;
    unsigned char *cp_space;
    unsigned char *cp_context;
    
    if (nconstants < CONSTANT_POOL_UNUSED_INDEX) {
	JAVA_ERROR("Illegal constant pool size");
    }
    
    total_space = nconstants * sizeof(cp_item_type);
    total_space += nconstants * sizeof(char);

    /* Remember the start of the constant pool in the .class file */
    cp_context = context->ptr;

    /* 
     * First we make a pass to see how much space we'd have to
     * allocate for the Utf8 strings.  
     */
    for (i = CONSTANT_POOL_UNUSED_INDEX; i < nconstants; i++) {
	int type = get1byte(context);
	switch (type) {
	  case CONSTANT_Utf8: {
	      int length = get2bytes(context);
	      total_space += length + 1;
	      skipNbytes(context, length);
	      break;
	  }
	
	  case CONSTANT_Class:
	  case CONSTANT_String:
	    skip2bytes(context);
	    break;

	  case CONSTANT_Fieldref:
	  case CONSTANT_Methodref:
	  case CONSTANT_InterfaceMethodref:
	  case CONSTANT_NameAndType:
	  case CONSTANT_Integer:
	  case CONSTANT_Float:
	    skip4bytes(context);
	    break;

	  case CONSTANT_Long:
	  case CONSTANT_Double:
	    skipNbytes(context, 8); /* high, low */
	    i++;
	    break;

	  default:
	    JAVA_ERROR("Illegal constant pool type");
	}
    }

    cp_space = sysCalloc(1, ALIGN4(total_space));
    if (cp_space == NULL) {
	JAVA_ERROR("No memory for constant pool");
    }

    context->cp = constant_pool = (union cp_item_type *)cp_space;
    cp_space += nconstants * sizeof(cp_item_type);

    type_table = cp_space;
    cp_space += nconstants * sizeof(char);

    /*
     * Now rewind and read the constant pool in
     */
    context->ptr = cp_context;

    for (i = CONSTANT_POOL_UNUSED_INDEX; i < nconstants; i++) {
	int type = get1byte(context);
	CONSTANT_POOL_TYPE_TABLE_PUT(type_table, i, type);
	switch (type) {
	  case CONSTANT_Utf8: {
	      int length = get2bytes(context);
	      char *result = cp_space;

	      cp_space += length + 1;
	      getNbytes(context, length, result);
	      result[length] = '\0';
	      constant_pool[i].cp = result;
	      CONSTANT_POOL_TYPE_TABLE_SET_RESOLVED(type_table, i);
	      break;
	  }
	
	  case CONSTANT_Class:
	  case CONSTANT_String:
	    constant_pool[i].i = get2bytes(context);
	    break;

	  case CONSTANT_Fieldref:
	  case CONSTANT_Methodref:
	  case CONSTANT_InterfaceMethodref:
	  case CONSTANT_NameAndType:
	    constant_pool[i].i = get4bytes(context);
	    break;
	    
	  case CONSTANT_Integer:
	  case CONSTANT_Float:
	    constant_pool[i].i = get4bytes(context);
	    CONSTANT_POOL_TYPE_TABLE_SET_RESOLVED(type_table, i);
	    break;

	  case CONSTANT_Long:
	  case CONSTANT_Double: {
	      unsigned high = get4bytes(context);
	      unsigned low = get4bytes(context);
	      int64_t value;
	      value = ll_add(ll_shl(uint2ll(high), 32), uint2ll(low));
	      SET_INT64(t1, &constant_pool[i], value);
	      CONSTANT_POOL_TYPE_TABLE_SET_RESOLVED(type_table, i);
	      i++;		/* increment i for loop, too */
	      /* Indicate that the next object in the constant pool cannot
	       * be accessed independently.    */
	      CONSTANT_POOL_TYPE_TABLE_PUT(type_table, i, 0);
	      CONSTANT_POOL_TYPE_TABLE_SET_RESOLVED(type_table, i);
	      break;
	  }

	  default:
	      JAVA_ERROR("Illegal constant pool type");
	}
    }
    /* It is important to only set these after everything is setup,
       so that the GC sees a consistent state.*/
    cbConstantPool(cb) = constant_pool;
    cbConstantPoolCount(cb) = nconstants;
    constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type = type_table;

    return constant_pool;
}



static void ReadInCode(CICcontext *context, struct methodblock *mb,
		       int allocType)
{
    int attribute_length = get4bytes(context);
    unsigned char *end_ptr = context->ptr + attribute_length;
    int attribute_count;
    int code_length;
    int i;

    if (context->minor_version <= 2) { 
	methodmaxstack(mb) = get1byte(context);
	methodnlocals(mb) = get1byte(context);
	code_length = methodcodelen(mb) = get2bytes(context);
    } else { 
	methodmaxstack(mb) = get2bytes(context);
	methodnlocals(mb) = get2bytes(context);
	code_length = methodcodelen(mb) = get4bytes(context);
    }
    if (code_length > 0xffff)
	JAVA_ERROR("Method larger than 64k");
    if (methodnlocals(mb) < methodargssize(mb))
        JAVA_ERROR("Arguments can't fit into locals");

    methodcode(mb) = allocNBytes(context, code_length, allocType);

    getNbytes(context, code_length, (char *)methodcode(mb));
    if ((methodexctablen(mb) = get2bytes(context)) > 0) {
	unsigned exception_table_size = methodexctablen(mb) 
	                                  * sizeof(struct CatchFrame);
	methodexctab(mb) = allocNBytes(context, exception_table_size,
				       allocType);
	for (i = 0; i < (int)methodexctablen(mb); i++) {
	    methodexctab(mb)[i].start_pc = get2bytes(context);
	    methodexctab(mb)[i].end_pc = get2bytes(context);		
	    methodexctab(mb)[i].handler_pc = get2bytes(context); 
	    methodexctab(mb)[i].catchType = get2bytes(context);
	}
    }
    attribute_count = get2bytes(context);
    for (i = 0; i < attribute_count; i++) {
	int length;
#ifdef CLASSDEBUGINFO
	char *code_attr_name = getAsciz(context);
	if (strcmp(code_attr_name, "LineNumberTable") == 0) {
	    ReadLineTable(context, mb, allocType);
	} else if (strcmp(code_attr_name, "LocalVariableTable") == 0) {
	    ReadLocalVars(context, mb, allocType);
	} else {
	    length = get4bytes(context);
	    skipNbytes_quick(context, length);
#ifdef JCOV
	} else if (strcmp(name, "CoverageTable") == 0 && cov_file) {
	    ReadCovTable(context, mb, allocType); 
#endif
	}
#else /* CLASSDEBUGINFO */
	/* 
	 * Use getAsciz() rather than skipAsciz_quick() 
	 * to check if the attribute_name_index item is valid 
	 * (this is required to pass pJCK).
	 */
	getAsciz(context);
	length = get4bytes(context);
	skipNbytes_quick(context, length);
#endif /* CLASSDEBUGINFO */
    }
    if (context->ptr != end_ptr) 
	JAVA_ERROR("Code segment was wrong length");
}


#ifdef CLASSDEBUGINFO
static void ReadLineTable(CICcontext *context, struct methodblock *mb,
			  int allocType)
{
    int attribute_length = get4bytes(context);
    unsigned char *end_ptr = context->ptr  + attribute_length;
    int i;
    if ((methodlnnotablen(mb) = get2bytes(context)) > 0) {
	struct lineno *ln = 
	    allocNBytes(context, methodlnnotablen(mb) * 
			sizeof(struct lineno), allocType);
	methodlnnotab(mb) = ln;
	for (i = methodlnnotablen(mb); --i >= 0; ln++) {
	    ln->pc = get2bytes(context);
	    ln->line_number = get2bytes(context);
	}
    }
    if (context->ptr != end_ptr)
	JAVA_ERROR("Line number table was wrong length?");
}	
	
static void ReadLocalVars(CICcontext *context, struct methodblock *mb,
			  int allocType)
{
    int attribute_length = get4bytes(context);
    unsigned char *end_ptr = context->ptr  + attribute_length;
    int i;
    if ((methodlocvartablen(mb) = get2bytes(context)) > 0) {
	struct localvar *lv = 
	  allocNBytes(context, methodlocvartablen(mb) * 
		      sizeof(struct localvar), allocType);
	methodlocvartab(mb) = lv;
	for (i = methodlocvartablen(mb); --i >= 0; lv++) {
	    lv->pc0 = get2bytes(context);
	    lv->length = get2bytes(context);
	    localvarname(context->cp, lv) = getAsciz(context);
	    localvarsig(context->cp, lv) = getAsciz(context);
	    lv->slot = get2bytes(context);
	}
    }
    if (context->ptr != end_ptr)
	JAVA_ERROR("Local variables table was wrong length?");
}	
#endif /* CLASSDEBUGINFO */

#ifdef JCOV
static void ReadCovTable(CICcontext *context, struct methodblock *mb,
			 int allocType)
{
    int attribute_length = get4bytes(context);
    unsigned char *end_ptr = context->ptr  + attribute_length;
    int old_clinit, i;

    old_clinit = context->in_clinit;		
    context->in_clinit = FALSE;		/* Since all storage needed by the class initialization method  
					 * <clinit> can be freed after the class is loaded
					 * we should not place Jcov data into this memory
				         */

    if ((methodcovtablen(mb) = get2bytes(context)) > 0) {
	int sz = methodcovtablen(mb) * sizeof(struct covtable)
                         + methodcodelen(mb) * sizeof(long);
	struct covtable *cv = allocNBytes(context, sz, allocType);
        long *hash;
	methodcovtab(mb) = cv;
        hash = (long*)(methodcovtab(mb) + methodcovtablen(mb));
	memset(hash, 0, methodcodelen(mb)*sizeof(long)); 
	for (i = 0; i < methodcovtablen(mb); cv++) {
	    cv->pc = get2bytes(context);
            *(hash+cv->pc) = i + 1;
	    cv->type = get2bytes(context);
	    cv->where = get4bytes(context);
	    cv->count = 0;
            i++;
	}
    }
    if (context->ptr != end_ptr)
	methodcovtab(mb) = NULL;	/* Don't use a wrong coverage table */

    context->in_clinit = old_clinit;
}	
#endif 

static void
ReadExceptions(CICcontext *context, struct methodblock *mb, int allocType)
{
    int attribute_length = get4bytes(context);
    unsigned char *end_ptr = context->ptr + attribute_length;
    unsigned short nexceptions = get2bytes(context);

    if ((methodnexceptions(mb) = nexceptions) > 0) {
	unsigned short *ep, *exceptions =
	    allocNBytes(context, nexceptions * sizeof (unsigned short),
			allocType);
	methodexceptions(mb) = ep = exceptions;
	while (nexceptions-- > 0) {
	    *ep++ = get2bytes(context);
	}
    }
    if (context->ptr != end_ptr)
	JAVA_ERROR("Exceptions attribute has wrong length");
}

unsigned Signature2ArgsSize(char *method_signature)
{
    char *p;
    int args_size = 0;
    for (p = method_signature; *p != SIGNATURE_ENDFUNC; p++) {
	switch (*p) {
	  case SIGNATURE_BOOLEAN:
	  case SIGNATURE_BYTE:
	  case SIGNATURE_CHAR:
	  case SIGNATURE_SHORT:
	  case SIGNATURE_INT:
	  case SIGNATURE_FLOAT:
	    args_size += 1;
	    break;
	  case SIGNATURE_CLASS:
	    args_size += 1;
	    while (*p != SIGNATURE_ENDCLASS) p++;
	    break;
	  case SIGNATURE_ARRAY:
	    args_size += 1;
	    while ((*p == SIGNATURE_ARRAY)) p++;
	    /* If an array of classes, skip over class name, too. */
	    if (*p == SIGNATURE_CLASS) { 
		while (*p != SIGNATURE_ENDCLASS) 
		  p++;
	    } 
	    break;
	  case SIGNATURE_DOUBLE:
	  case SIGNATURE_LONG:
	    args_size += 2;
	    break;
	  case SIGNATURE_FUNC:	/* ignore initial (, if given */
	    break;
	  default:
	    /* Indicate an error. */
	    return 0;
	}
    }
    return args_size;
}

void free_clinit_memory(struct methodblock *mb)
{
    /* This function is somewhat a hack. It fixes the problem in 1.1.3
     * and before. sysFree may be called on the wrong memory block if
     * the exception attribute comes before the code attribute.
     */
    /* If there is no exceptions attribute, or if both has already
     * been freed. */
    if (methodexceptions(mb) == NULL) {
        if (methodcode(mb)) {
	    sysFree(methodcode(mb));
	    methodcode(mb) = NULL;
	}
	return;
    }

    /* If both attributes exist, free the one at the lower address */
    if ((char *)methodcode(mb) < (char *)methodexceptions(mb))
        sysFree(methodcode(mb));
    else
        sysFree(methodexceptions(mb));

    methodcode(mb) = NULL;
    methodexceptions(mb) = NULL;
}

void FreeClass(ClassClass *cb) 
{
    int i;
    int classOffsetsToFree[] = {
	offsetOfCbField(implements),
	offsetOfCbField(fields),
	offsetOfCbField(methods)
    };
#define NO_CLASS_OFFSETS_TO_FREE \
    (sizeof(classOffsetsToFree) / sizeof(classOffsetsToFree[0]))

#ifdef SUPPORTCOMPILER
    CompilerFreeClass(cb);	/* indicate that class is free */
#endif

    /*
     * The class-loading phase mallocs at most three blocks of
     * contiguous malloc'ed memory for each loaded class. These are
     * the <clinit> memory, the constant pool memory, and memory for
     * all the rest. FreeClass() should be freeing all three of these.
     *
     * There is also the interface method table and the "v-table" allocated
     * during class resolution. Those are handled at the end.  
     *
     * That gives us a total of five things to do.
     */

    /*
     * 1. Find and release the memory for the <clinit> method.  
     */
    if (cbMethods(cb) != 0) {
	struct methodblock *mb = cbMethods(cb);
	struct methodblock *mbEnd = &cbMethods(cb)[cbMethodsCount(cb)];
	
	while ((mb < mbEnd) && (methodID(mb) != 0)) {
	    char *method_name = methodname(mb);
	    char *method_sig = methodsig(mb);
	    if (method_name[0] == '<' &&
		strcmp(method_name, "<clinit>") == 0 &&
		strcmp(method_sig, "()V") == 0 &&
		methodcodelen(mb) /* not external */ ) {
		free_clinit_memory(mb);
		break;
	    }
	    mb++;
	}
    }

    /*
     * 2. Get rid of the constant pool
     */
    if (cbConstantPool(cb) != 0) {
	sysFree(cbConstantPool(cb));
    }

    /*
     * 3. Get rid of the rest of the assigned memory. Find the first
     * assigned value, which holds the memory for all the rest.  
     */
    for (i = 0; i < NO_CLASS_OFFSETS_TO_FREE; i++) {
	void *field = cbFieldAtOffset(cb, classOffsetsToFree[i]);
	if (field != 0) {
	    sysFree(field);
	    break;
	}
    }

    /*
     * 4. Handle the methodtable
     */
    if (CCIs(cb, Initialized)) { 
        /* Before it is initialized, the method table slot can contain the
	 * super name.  It is only valid as a method table after initialization.
	 */
        sysFree(cbMethodTable(cb));
    }

    /*
     * 5. Handle the interface method table
     */

    /* Interface method tables can be shared between child and super classes
     */
    if (cbImplementsCount(cb) != 0 || cbIsInterface(cb)) {
        sysFree(cbIntfMethodTable(cb));
    }
}
	
static unsigned long get1byte(CICcontext *context)
{
    unsigned char *ptr = context->ptr;
    if (context->end_ptr - ptr < 1) {
	JAVA_ERROR("Truncated class file");
    } else {
	unsigned char *ptr = context->ptr;
	unsigned long value = ptr[0];
	(context->ptr) += 1;
	return value;
    }
}

static unsigned long get2bytes(CICcontext *context)
{
    unsigned char *ptr = context->ptr;
    if (context->end_ptr - ptr < 2) {
	JAVA_ERROR("Truncated class file");
    } else {
	unsigned long value = (ptr[0] << 8) + ptr[1];
	(context->ptr) += 2;
	return value;
    }
}


static unsigned long get4bytes(CICcontext *context)
{
    unsigned char *ptr = context->ptr;
    if (context->end_ptr - ptr < 4) {
	JAVA_ERROR("Truncated class file");
    } else {
	unsigned long value = (ptr[0] << 24) + (ptr[1] << 16) + 
	                             (ptr[2] << 8) + ptr[3];
	(context->ptr) += 4;
	return value;
    }
}

static void getNbytes(CICcontext *context, int count, char *buffer) 
{
    unsigned char *ptr = context->ptr;
    if (context->end_ptr - ptr < count)
	JAVA_ERROR("Truncated class file");
    if (buffer != NULL) 
	memcpy(buffer, ptr, count);
    (context->ptr) += count;
}


static char *getAsciz(CICcontext *context) 
{
    ClassClass *cb = context->cb;
    union cp_item_type *constant_pool = cbConstantPool(cb);
    int nconstants = cbConstantPoolCount(cb);
    unsigned char *type_table = 
	constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;

    int value = get2bytes(context);
    if ((value == 0) || (value >= nconstants) || 
	   type_table[value] != (CONSTANT_Utf8 | CONSTANT_POOL_ENTRY_RESOLVED))
	JAVA_ERROR("Illegal constant pool index");
    return constant_pool[value].cp;
}

static char *getAscizFromClass(CICcontext *context, int value) 
{
    ClassClass *cb = context->cb;
    union cp_item_type *constant_pool = cbConstantPool(cb);
    int nconstants = cbConstantPoolCount(cb);
    unsigned char *type_table = 
	constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
    if ((value <= 0) || (value >= nconstants) 
	|| (type_table[value] != CONSTANT_Class))
	JAVA_ERROR("Illegal constant pool index");
    value = constant_pool[value].i;
    if ((value <= 0) || (value >= nconstants) ||
	 (type_table[value] != (CONSTANT_Utf8 | CONSTANT_POOL_ENTRY_RESOLVED)))
	JAVA_ERROR("Illegal constant pool index");
    return constant_pool[value].cp;
}

/*
 * If a class loading buffer has not been assigned to any ClassClass
 * field yet, it needs to be freed here, as FreeClass() cannot find
 * its address.
 */
#define BUFF_UNASSIGNED(type)						  \
    (context->buffer_sizes[(type)] > 0 && (context->buffers[(type)] ==	  \
					   context->buffer_ptrs[(type)]))

#define FREE_BUFFER_IF_NEEDED(type)				\
     if (BUFF_UNASSIGNED(type)) {				\
	 sysFree(context->buffers[(type)]);			\
	 context->buffers[(type)] = 0; /* Just in case */	\
     }

static void freeBuffers(CICcontext * context)
{
    /*
     * Free all UNASSIGNED memory associated with a loaded class. All assigned
     * memory is reachable through the ClassClass structure, and will be freed
     * by FreeClass().
     */
    ClassClass *cb = context->cb;
    if (cb != NULL) {
	/* If the main buffer or the clinit buffer has not been
	   assigned, free it */
	FREE_BUFFER_IF_NEEDED(MAIN_BUFF);
	FREE_BUFFER_IF_NEEDED(CLINIT_BUFF);

	/* If constant pool has not been assigned, free it */
	if (cbConstantPool(cb) == 0) {
	    if (context->cp != NULL) {
		sysFree((unsigned char *)context->cp);
		context->cp = 0; /* Just in case */
	    }
	}
    }
}

/* In certain cases, we only need a class's name.  We don't want to resolve
 * the class reference if it isn't, already. 
 */

char *GetClassConstantClassName(cp_item_type *constant_pool, int index)
{
    unsigned char *type_table = 
        constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
    int type = (type_table == NULL) 
                    ? CONSTANT_Class | CONSTANT_POOL_ENTRY_RESOLVED
                    : type_table[index];

    switch(type) {
        case CONSTANT_Class | CONSTANT_POOL_ENTRY_RESOLVED: {
	    ClassClass *cb = constant_pool[index].clazz;
	    return cbName(cb);
	}

        case CONSTANT_Class: {
	    int name_index = constant_pool[index].i;
	    return constant_pool[name_index].cp;
	}
	    
	default:
	    return (char *)0;
    }
}


