/*
 * @(#)classresolver.c	1.116 98/11/18
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*-
 *      Routines for loading and resolving class definitions.
 *      These routines should not depending up on the interpreter or
 *      garbage collector.
 *
 */

#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <stdlib.h>

#include "timeval.h"
#include "oobj.h"
#include "interpreter.h"
#include "signature.h"
#include "tree.h"
#include "monitor.h"
#include "bool.h"
#include "path.h"
#include "sys_api.h"
#include "utf.h"
#include "javaString.h"
#include "exceptions.h"

#ifdef BREAKPTS
#include "breakpoints.h"
#endif

extern ClassClass *allocClassClass();

struct methodtable *javaLangClassMethods = 0;

static bool_t RunClinit(ClassClass * cb);

char *RuntimeInitClass(struct execenv *, ClassClass *);

static ClassClass *InitializeAndResolveClass(ClassClass *cb, bool_t resolve);
static char *Locked_InitializeClass(ClassClass * cb, char **detail);
static char *Locked_LinkClass(ClassClass * cb, char **detail);
static ClassClass *FindLoadedClass(char *name,
				   struct Hjava_lang_ClassLoader *loader);
static ClassClass *Locked_FindArrayClassFromClass(struct execenv *ee, 
						  char *name, 
						  ClassClass *from);
static void InitPrimitiveClasses();

/* An Explanation of Class-related Locks
 *
 * There are two global locks related to class loading:
 *
 * LOADCLASS_LOCK: ensures that only one thread is loading a class at
 * a given time. This eliminates the possibility of two threads loading
 * classes with the same name and same loader.
 *
 * BINCLASS_LOCK: ensures that only one thread is updating the global
 * class table (binclasses). This lock is also grabbed by the GC to ensure
 * that the GC always sees a valid global class table state.
 *
 * In addition, each class may have its own associated locks that are
 * created with monitorEnter(). ResolveClass, for example, need to
 * first grab this lock to ensure that the class is resolved only by
 * one thread, rather than simultanously by multiple threads.
 */

sys_mon_t *_loadclass_lock;
sys_mon_t *_binclass_lock;

static void
addslots(struct fieldblock ** fb, ClassClass * cb, int nslots)
{
    fb += nslots;
    while (cb) {
	long n = cbFieldsCount(cb);
	struct fieldblock *sfb = cbFields(cb);
	while (--n >= 0) {
	    *--fb = &sfb[n];
	}
	cb = cbSuperclass(cb);
    }
}

int
makeslottable(ClassClass * clb,
		     struct fieldblock ***slottable,
		     int *slottable_count)
{
    ClassClass *sclb;
    int         nslots = 0;

    sclb = clb;
    while (sclb) {
	nslots += cbFieldsCount(sclb);
	sclb = cbSuperclass(sclb);
    }
    *slottable_count = nslots;
    if (nslots == 0) {
	nslots++;
    }
    *slottable = (struct fieldblock **)
	sysMalloc(nslots * sizeof(struct fieldblock *));
    if (*slottable == 0) {
	return SYS_NOMEM;
    }
    addslots(*slottable, clb, *slottable_count);
    return SYS_OK;
}

struct fieldblock *
getslot(ClassClass * clb,
	       int slot,
	       int *slottable_count)
{
    ClassClass *sclb;
    int         nslots = 0;
    struct fieldblock *fb = NULL;

    sclb = clb;
    while (sclb) {
	nslots += cbFieldsCount(sclb);
	sclb = cbSuperclass(sclb);
    }
    *slottable_count = nslots;
    sclb = clb;
    while (sclb) {
	int slots = cbFieldsCount(sclb);
	nslots -= slots;
	if (slot >= nslots) {
	    fb = &cbFields(sclb)[slot - nslots];
	    break;
	}
	sclb = cbSuperclass(sclb);
    }
    return fb;
}

#if 0
static char *
copyclassname(char *src, char *dst)
{
    sysAssert(*src == SIGNATURE_CLASS);
    src++;
    while (*src && *src != SIGNATURE_ENDCLASS)
	*dst++ = *src++;
    *dst = 0;
    return src;
}
#endif

static char *
ResolveFields(ClassClass *cb, unsigned slot)
{
    struct fieldblock *fb;
    int thishash = 0;
    char *s;
    int c;
    int size;

    fb = cbFields(cb);
    for (size = 0; size < (int) cbFieldsCount(cb); size++, fb++) {
	int size = ((fieldaccess(fb) & ACC_DOUBLEWORD) ? 2 : 1);
	if (fieldaccess(fb) & ACC_STATIC) {
	    /* Do nothing.  Handled when the class is loaded. */
	} else {
	    fieldoffset(fb) = slot;
	    slot += size * sizeof(OBJECT);
	}
#ifdef UNUSED
	if ((fieldaccess(fb) & (ACC_STATIC | ACC_TRANSIENT)) == 0) {
	    for (s = fieldname(fb); (c = *s++) != 0;)
		thishash = thishash * 7 + c;
	    for (s = fieldsig(fb); (c = *s++) != 0;)
		thishash = thishash * 7 + c;
	}
#endif
    }
    if (slot > 65535) {
        return "java/lang/InternalError";
    }

    cbInstanceSize(cb) = slot;
#ifdef UNUSED
    cbThisHash(cb) = thishash;
    if (cbSuperclass(cb))
	cbTotalHash(cb) = thishash - cbTotalHash(unhand(cbSuperclass(cb)));
    else
        cbTotalHash(cb) = thishash;
    if (cbTotalHash(cb) < N_TYPECODES)
	cbTotalHash(cb) += N_TYPECODES;
#endif
    return NULL;
}


static char *
ResolveMethods(ClassClass *cb)
{
    struct methodblock *mb;
    int size;
    struct methodtable *new_table;
    struct methodblock **super_methods;
    unsigned mslot, super_methods_count;

    unsigned finalizerID = 0;	/* only used for cb == java.lang.Object */
    

    mb = cbMethods(cb);
    for (size = 0; size < (int) cbMethodsCount(cb); size++, mb++) {
	methodoffset(mb) = 0;
	methodinvokeridx(mb) = INVOKE_ILLEGAL;
    }

    if (cbIsInterface(cb)) { 
	/* We don't really need to built a method table for interfaces. */
	cbMethodTable(cb) = NULL;
	cbMethodTableSize(cb) = 0;
	mb = cbMethods(cb);
	/* Each method's offset is its index in the interface */
	for (size = 0; size < (int) cbMethodsCount(cb); size++, mb++) {
	    methodoffset(mb) = size;
	}
	return NULL;
    }

    if (cbSuperclass(cb) != NULL) { 
	ClassClass *super = cbSuperclass(cb);
	mslot = cbMethodTableSize(super);
	super_methods = methodtablemethods(cbMethodTable(super));
        super_methods_count = cbMethodTableSize(super);
    } else { 
        sysAssert(FINALIZER_MB_OFFSET == 1);
	mslot = 2;		/* mslot == 1 is reserved for finalizer! */
	super_methods = NULL;
	super_methods_count = 0;
	finalizerID = NameAndTypeToHash(FINALIZER_METHOD_NAME,
					FINALIZER_METHOD_SIGNATURE);
    }

    mb = cbMethods(cb);
    for (size = 0; size < (int) cbMethodsCount(cb); size++, mb++) {
	unsigned long ID = methodID(mb);
	struct methodblock **super_methods_p;
	int count;

	if ((methodaccess(mb) & ACC_STATIC) || (methodaccess(mb) & ACC_PRIVATE))
	    continue;
	if (strcmp(methodname(mb), "<init>") == 0)
	    continue;

	/* If this item has its own finalizer method, grab it */
	if (cb == classJavaLangObject && methodID(mb) == finalizerID) { 
	    methodoffset(mb) = FINALIZER_MB_OFFSET;
	    continue;
	} 

	for (super_methods_p = super_methods, count = super_methods_count;
	       count > 0;
	       super_methods_p++, count--) {
	    if ((*super_methods_p != NULL) && (methodID(*super_methods_p) == ID)) { 
	      /* Private methods are not inherited outside of the class. */
	        if (methodaccess(*super_methods_p) & ACC_PRIVATE)
		  continue;
	      /* Package-private methods are not inherited outside of the
		 package. */
	        if ((methodaccess(*super_methods_p) & ACC_PROTECTED) ||
		    (methodaccess(*super_methods_p) & ACC_PUBLIC) ||
		    IsSameClassPackage(methodclass(*super_methods_p), cb)) {
		    methodoffset(mb) = methodoffset(*super_methods_p);
		    break;
		} 
	      /*
		jio_fprintf(stderr, "!!!%s.%s\n", cbName(cb), methodname(mb));
		*/
	    }
	}

	if (methodoffset(mb) == 0) { 
	    methodoffset(mb) = mslot;
	    mslot++;
	}
    }

    if (mslot > 65535) {
        return "java/lang/InternalError";
    }

    /*
     *	This should be the only place that method tables are
     *	allocated.  We allocate more than we need here and mask the
     *	resulting pointer because the methodtable pointer field in
     *	object handles is overloaded and sometimes hold array types
     *	and array lengths.  We must ensure that method table pointers
     *	are allocated on an appropriate boundary so we don't lose any
     *	address bits when we mask off the type portion of the pointer.
     */
    new_table = (struct methodtable *)
         sysMalloc(sizeof(struct methodtable)
		          + (mslot - 1)* sizeof(struct methodblock *));
    if (new_table == NULL) {
	CCSet(cb, Error);
	return JAVAPKG "OutOfMemoryError";
    }
    memset((char *)methodtablemethods(new_table), 0, mslot * sizeof(struct methodblock *));
    if (super_methods) 
	memcpy((char *) methodtablemethods(new_table),
	       (char *) super_methods,
	       super_methods_count * sizeof(struct methodblock *));
    mb = cbMethods(cb);
    for (size = 0; size < (int) cbMethodsCount(cb); size++, mb++) {
	int offset = (int)methodoffset(mb);
	if (offset > 0) { 
	    sysAssert(offset < mslot);
	    mt_slot(new_table, offset) = mb;
	}
    }
    methodtablecb(new_table) = cb;
    cbMethodTable(cb) = new_table;
    cbMethodTableSize(cb) = mslot;
    if (methodtablemethods(new_table)[1] != 
	  methodtablemethods(cbMethodTable(classJavaLangObject))[1]) { 
        CCSet(cb, HasFinalizer);
    }
    return NULL;
}


static char *
ResolveInterfaces(ClassClass *cb, char **detail)
{
    const int ITEM_SIZE =  sizeof(cbIntfMethodTable(cb)->itable[0]);
    bool_t isInterface = cbIsInterface(cb);
    if (cbImplementsCount(cb) == 0 && !isInterface) {
	/* classes that don't implement their own interfaces can just inherit
	 * their parents imethodtable. */
	if (cb == classJavaLangObject) {
	    /* We can preinitialize the imethodtable of java.lang.Object */
	    static struct imethodtable t = { 0 };
	    cbIntfMethodTable(cb) = &t;
	} else { 
	    ClassClass *super = cbSuperclass(cb);
	    cbIntfMethodTable(cb) = cbIntfMethodTable(super);
	} 
	return NULL;
    } else { 
	cp_item_type *constant_pool = cbConstantPool(cb);
	ClassClass *super = cbSuperclass(cb);
	unsigned char *mallocSpace, *mallocSpaceEnd;
	ClassClass *icb;	/* temporary */
	struct imethodtable *super_itable = cbIntfMethodTable(super);
	struct imethodtable *this_itable = NULL;
	int super_itable_count = super_itable->icount;
	int i, j, k, icount, mcount;
	
	/* Resolve all the interfaces that this class implements, and 
	 * make sure that they all really are interfaces 
	 *
	 * icount will total all the interfaces that this class implements,
	 *    (include interfaces implemented by our superclass, and by 
	 *    interfaces that we implement.)
	 * mcount will total the total amount of space (in sizeof(long)) for 
	 *    which we'll have to allocate space for in the offsets table.
	 */
	icount = super_itable_count + (isInterface ? 1 : 0);
	mcount = 0;
	for (i = 0; i < (int)(cbImplementsCount(cb)); i++) {
	    int interface_index = cbImplements(cb)[i];
	    struct imethodtable *sub_itable;

	    icb = constant_pool[interface_index].clazz;
	    if (!cbIsInterface(icb)) {
	        *detail = "Implementing class";
		return JAVAPKG "IncompatibleClassChangeError";
	    }
	    sub_itable = cbIntfMethodTable(icb);
	    icount += sub_itable->icount;
	    if (!isInterface) 
		for (j = sub_itable->icount; --j >= 0; ) 
		    mcount += cbMethodsCount(sub_itable->itable[j].classdescriptor);
	}

	{ 
	    int this_itable_size = 
		offsetof(struct imethodtable, itable) + icount * ITEM_SIZE;
	    int offsets_size = mcount * sizeof(unsigned long);
	    mallocSpace = sysMalloc(this_itable_size + offsets_size);
	    if (mallocSpace == NULL) { 
		return JAVAPKG "OutOfMemoryError";
	    }
	    mallocSpaceEnd = mallocSpace + this_itable_size + offsets_size;
	    this_itable = (struct imethodtable *)mallocSpace;
	    mallocSpace += this_itable_size;
	    sysAssert(mallocSpace <= mallocSpaceEnd);
	}
	
	cbIntfMethodTable(cb) = this_itable;

	/* Start filling in the table. */
	icount = 0;
	if (isInterface) {
	    this_itable->itable[icount].classdescriptor = cb;
	    this_itable->itable[icount].offsets = NULL;	
	    icount++;
	}
	if (super_itable_count > 0) { 
	    /* We can copy our superclass's offset table.  The offsets
	       will stay the same.  */
	    memcpy(&this_itable->itable[icount], 
		   &super_itable->itable[0], 
		   super_itable_count * ITEM_SIZE);
	    icount += super_itable_count;
	}
	for (i = 0; i < (int)(cbImplementsCount(cb)); i++) {
	    /* Add the interfaces that we implement, either directly, or
	     * because those interfaces implement other interfaces.    */
	    int interface_index = cbImplements(cb)[i]; 
	    icb = constant_pool[interface_index].clazz;
	    memcpy(&this_itable->itable[icount], 
		   &cbIntfMethodTable(icb)->itable[0],
		   cbIntfMethodTable(icb)->icount * ITEM_SIZE);
	    icount += cbIntfMethodTable(icb)->icount;
	}
	sysAssert(!isInterface || super_itable_count == 0);
	/* Delete duplicates from the table.  This is very rare, so it can
	 * be quite inefficient.  */
	for (i = (isInterface ? 1 : super_itable_count); i < icount; i++) {
	    icb = this_itable->itable[i].classdescriptor;
	    for (j = 0; j < i; j++) { 
		if (icb == this_itable->itable[j].classdescriptor) { 
		    /* We have an overlap.  Item i is a duplicate.  Delete from
		     * the table */
		    for (k = i + 1; k < icount; k++) 
			this_itable->itable[k - 1] = this_itable->itable[k];
		    icount--;
		    i--; 	/* Reconsider this entry! */
		    break;
		}
	    }
	}
	this_itable->icount = icount;
	if (isInterface) { 
	    /* Nothing more to do for interfaces */
	    return NULL;
	}

	/* For each additional interface . . . */
	for (i = super_itable_count; i < icount; i++) { 
	    /* The table length is the number of interface methods */
	    ClassClass *intfi = this_itable->itable[i].classdescriptor;
	    int intfi_count = cbMethodsCount(intfi);
	    unsigned long *offsets = (unsigned long *)mallocSpace;
	    mallocSpace += intfi_count * sizeof(unsigned long);
	    sysAssert(mallocSpace <= mallocSpaceEnd);
	    this_itable->itable[i].offsets = offsets;
	    /* Look at each interface method */
	    for (j = 0; j < intfi_count; j++) { 
		struct methodblock *imb = cbMethods(intfi) + j;
		if ((methodaccess(imb) & ACC_STATIC) != 0) { 
		    sysAssert(!strcmp(methodname(imb), "<clinit>"));
		    offsets[j] = 0; 
		} else {
		    /* Find the class method that implements the interface 
		     * method */
		    unsigned ID = methodID(imb);
		    struct methodblock **mbt =
			methodtablemethods(cbMethodTable(cb));
		    for (k = cbMethodTableSize(cb) - 1; k >= 0; --k) { 
			struct methodblock *mb = mbt[k];
			if (mb != NULL 
			      && methodID(mb) == ID && IsPublic(methodaccess(mb))) {
			    offsets[j] = methodoffset(mb);
			    break;
			}
		    }
		    if (k == -1) { 
			*detail = "Unimplemented interface method";
			return JAVAPKG "IncompatibleClassChangeError";
		    }
		}
	    }
	}
	return NULL;
    }
}


char *
InitializeClass(ClassClass * cb, char **detail) 
{ 
    char *result;
    monitorEnter(obj_monitor(cb));
    result = Locked_InitializeClass(cb, detail);
    monitorExit(obj_monitor(cb));
    return result;
}


char *
ResolveClass(ClassClass *cb, char **detail) 
{
    char *result;
    if (CCIs(cb, Resolved))
        return 0;

    /*
     * Check for stack overflow which might be caused by possible 
     * recursive invocation of this function.
     */
    if (!sysThreadCheckStack(STACK_REDZONE)) {
	/* Caller will take care of throwing the error */
	return JAVAPKG "StackOverflowError";
    }

    result = LinkClass(cb, detail);
    if (result == 0) {
        if (!RunClinit(cb)) {
	    result = JAVAPKG "ExceptionInInitializerError";
	    *detail = cbName(cb);
	}
    }
    return result;
}

char *
LinkClass(ClassClass *cb, char **detail) 
{
    char *ret = NULL;
    ClassClass *scListLocalBuf[SUPER_INIT_CHAIN_SIZE];
    ClassClass **scList = scListLocalBuf; /* Set the initial buffer */
    int scSize;
    int k;

    /* A quick test to avoid extra work */
    if (CCIs(cb, Linked)) {
        return NULL;
    }

    getSuperHier(cb, &scList, &scSize);
    if (scSize < 0) { /* Out of memory occurred */
	return JAVAPKG "OutOfMemoryError";
    }
    /*
     * Now scList [0 ... scSize -1] has the class hierarchy. Link
     * each of them. The higher indicates have superclasses.
     */
    for (k = scSize - 1; k >= 0; k--) {
	ClassClass *sc = scList[k];

	if (CCIs(sc, Linked)) { /* A quick test */
	    continue;
	}

	/*
	 * Link the superclass, or set Error flag if error has occurred
	 * during linking its superclasses.
	 */
	monitorEnter(obj_monitor(sc));
	if (ret == NULL) {
	    ret = Locked_LinkClass(sc, detail);
	} else {
	    CCSet(sc, Error);
	}
	monitorExit(obj_monitor(sc));
    }
    /*
     * Free scList, if getSuperHier() returned allocated buffer.
     */
    if (scList != scListLocalBuf) {
	sysFree(scList);
    }
    return ret;
}

/*
 * Detect class circularities from InitializeClass()
 */

static void
pushSeen(ExecEnv *ee, struct seenclass *seen)
{
    struct seenclass *prev = ee->seenclasses.next;
    ee->seenclasses.next = seen;
    seen->next = prev;
}

static void
popSeen(ExecEnv *ee, struct seenclass *seen)
{
    if (seen != ee->seenclasses.next) /* paranoia */
	panic("popSeen: corrupt seen class stack");
    ee->seenclasses.next = ee->seenclasses.next->next;
}

static bool_t
checkSeen(ExecEnv *ee, ClassClass *cb)
{
    struct seenclass *seen = ee->seenclasses.next;
    while (seen) {
	if (cb == seen->cb)
	    return TRUE;
	seen = seen->next;
    }
    return FALSE;
}

static char *
Locked_InitializeClass(ClassClass * cb, char **detail)
{
    ClassClass *super = 0;
    char *ret = 0;
    bool_t noLoader;
    int i;

    if (CCIs(cb, Initialized))
	return NULL;

#ifndef TRIMMED
    if (verbose)
	jio_fprintf(stderr, "[Initializing %s]\n", cbName(cb));
#endif

    noLoader = (cbLoader(cb) == 0);
    if (cbFieldsCount(cb) > 2000) {
	return JAVAPKG "ClassFormatError";
    }
    if ((strcmp(cbName(cb), CLS_RESLV_INIT_CLASS) == 0) && noLoader) {
	/* Temporarily disable class circularity checks */
	ExecEnv *ee = EE();
	struct seenclass *save = ee->seenclasses.next;
	ee->seenclasses.next = NULL;
	/* Note: don't bother restoring on (fatal) error during bootstrap */

	classJavaLangClass = cb;
	MakeClassSticky(cb);

	{
	    static struct {
		char *name;
		ClassClass **cbp;
	    } clist[] = {
		{JAVAPKG "String", &classJavaLangString},
		{JAVAPKG "ThreadDeath", &classJavaLangThreadDeath},
		{JAVAPKG "Throwable", &classJavaLangThrowable},
		{JAVAPKG "Exception", &classJavaLangException},
		{JAVAPKG "Error", &classJavaLangError},
		{JAVAPKG "RuntimeException", &classJavaLangRuntimeException},
		{JAVAPKG "Cloneable", &interfaceJavaLangCloneable},
		{"java/io/Serializable", &interfaceJavaIoSerializable}
	    };
	    int i;

	    for (i = 0; i < sizeof clist / sizeof clist[0]; ++i) {
		char *cname = clist[i].name;
		ClassClass *cb = FindStickySystemClass(NULL, cname, TRUE);
		*clist[i].cbp = cb;
		if (cb == 0) {
		    *detail = cname;
		    return JAVAPKG "NoClassDefFoundError";
		}
	    }
	}

	/* Restore stack for class circularity checks */
	ee->seenclasses.next = save;

    } else if ((strcmp(cbName(cb), CLS_RESLV_INIT_OBJECT) == 0) && noLoader){
	classJavaLangObject = cb;
	MakeClassSticky(classJavaLangObject);
    }

    if (noLoader) { 
	char *name = cbName(cb);
	if (strcmp(name, CLS_RESLV_INIT_REF) == 0) {
	    CCSet(cb, SoftRef);
	    classSunMiscRef = cb;
	}
	if (strncmp(name, "java/", 5) || strncmp(name, "sun/", 4)) {
	    CCSet(cb, SysLock);
	}
    }
    if (cbSuperclass(cb) == NULL) {
        char *superName = (char *)cbMethodTable(cb); /* hack! */
	/* No, purge the name out of that slot, before it can do any damage. */
	cbMethodTable(cb) = 0;

	if (superName != NULL) {
	    /* Check for class definition circularities */
	    ExecEnv *ee = EE();
	    struct seenclass this[1];
	    if (checkSeen(ee, cb)) {
		*detail = cbName(cb);
		CCSet(cb, Error);
		return JAVAPKG "ClassCircularityError";
	    }
	    this->cb = cb;
	    this->next = NULL;
	    pushSeen(ee, this);
	    super = FindClassFromClass(ee, superName, FALSE, cb);
	    popSeen(ee, this);
	    if (super != NULL) { 
		sysAssert(CCIs(super, Initialized));
		/* Don't allow a class to have a superclass it can't access */
		if (!VerifyClassAccess(cb, super, FALSE)) { 
		    super = NULL;
		} 
	    }
	    if (super != NULL) {
		cbSuperclass(cb) = cbHandle(super);
		if (CCIs(super, SoftRef))
		    CCSet(cb, SoftRef);
	    } else {
		ret = JAVAPKG "NoClassDefFoundError";
		*detail = superName;
		cbSuperclass(cb) = NULL;
		CCSet(cb, Error);
	    } 
	} else if (cb == classJavaLangObject) {
	    cbSuperclass(cb) = 0;	    
	} else {
	    *detail = cbName(cb);
	    return JAVAPKG "ClassFormatException";
	}
    }

    for (i = 0; i < (int)(cbImplementsCount(cb)); i++) {
        /* Be very careful, since not verified yet. . . 
	 */
        int interface_index = cbImplements(cb)[i];
	cp_item_type *constant_pool = cbConstantPool(cb);
	unsigned char *type_table = 
	    constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
	ClassClass *icb;
	int nconstants = cbConstantPoolCount(cb);
	int iname_index;
	char *iname;

	if (interface_index <= 0 ||
	    interface_index >= nconstants ||
	    type_table[interface_index] != CONSTANT_Class ||
	    !(iname_index = constant_pool[interface_index].i) ||
	    iname_index <= 0 ||
	    iname_index >= nconstants ||
	    type_table[iname_index] != 
	    (CONSTANT_Utf8 | CONSTANT_POOL_ENTRY_RESOLVED)) {
		*detail = "Bad interface index";
		return JAVAPKG "ClassFormatError";
	}
		
	iname = constant_pool[iname_index].cp;
	if (iname == NULL || !IsLegalClassname(iname, FALSE)) {
	    *detail = "Bad interface name";
	    return JAVAPKG "ClassFormatError";
	}

	{
	    ExecEnv *ee = EE();
	    struct seenclass this[1];
	    if (checkSeen(ee, cb)) {
		*detail = cbName(cb);
		CCSet(cb, Error);
		return JAVAPKG "ClassCircularityError";
	    }
	    this->cb = cb;
	    this->next = NULL;
	    pushSeen(ee, this);
	    icb = FindClassFromClass(ee, iname, FALSE, cb);
	    popSeen(ee, this);
	    if (icb) {
	        constant_pool[interface_index].clazz = icb;
		type_table[interface_index] |= CONSTANT_POOL_ENTRY_RESOLVED;
	    } else {
	        *detail = iname;
		CCSet(cb, Error);
		return JAVAPKG "NoClassDefFoundError";
	    }
	}
    }

    CCSet(cb, Initialized);

    /* Make sure we know what classJavaLangClass is, and that it's method table
     * is filled in.
     */
    if (classJavaLangClass == 0) {
	classJavaLangClass = 
	      FindClassFromClass(0, CLS_RESLV_INIT_CLASS, TRUE, cb);
	if (classJavaLangClass == 0) {
	    return JAVAPKG "NoClassDefFoundError";
	}
    }
    
    /* The following may not do anything useful if classJavaLangClass hasn't
     * been completely resolved.  But we clean up in ResolveClass 
     */

    obj_methodtable(cbHandle(cb)) = javaLangClassMethods;

    return ret;
}

static bool_t initializePrimitives = TRUE;

static char *
Locked_LinkClass(ClassClass * cb, char **detail)
{
    ClassClass *super = NULL;
    unsigned slot = 0;
    char *ret = NULL;

    if (CCIs(cb, Error)) {
	*detail = cbName(cb);
	return JAVAPKG "NoClassDefFoundError";
    }
   
    sysAssert(CCIs(cb, Initialized));

    if (CCIs(cb, Linked)) {
	return NULL;
    }

    if (cbSuperclass(cb)) {
	/* If this object has a superclass. . . */
	super = cbSuperclass(cb);
	/* Superclasses have been resolved in advance by LinkClass() */
	sysAssert(CCIs(super, Linked));
	slot = cbInstanceSize(super);
    }
    sysAssert(!CCIs(cb, Error));
    sysAssert(!CCIs(cb, Linked));

#ifndef TRIMMED
    if (verbose)
        jio_fprintf(stderr, "[Resolving %s]\n", cbName(cb));
#endif
    cbInstanceSize(cb) = -1;
    if ((ret = ResolveFields(cb, slot))) { /* May return "InternalError" */
        *detail = cbName(cb);
	CCSet(cb, Error);
	return ret;
    }

    if ((ret = ResolveMethods(cb))) {
        /* May return "OutOfMemoryError" or "InternalError" */
	*detail = cbName(cb);
	CCSet(cb, Error);
	return ret;
    }

    if ((ret = ResolveInterfaces(cb, detail))) { /* All sorts of errors */
	CCSet(cb, Error);
	return ret;
    }

    InitializeInvoker(cb);

    if ((verifyclasses == VERIFY_ALL) ||
	  ((verifyclasses == VERIFY_REMOTE) && (cbLoader(cb) != NULL))) {
	/*
	 * Extra stack check for execution of InitPrimitiveClasses()
	 * to shrink STACK_REDZONE.
	 */
	if (!sysThreadCheckStack(STACK_REDZONE)) {
	    return JAVAPKG "StackOverflowError";
	}
	if (!VerifyClass(cb)) {
	    return JAVAPKG "VerifyError";
	}
    }

    CCSet(cb, Linked);

#ifdef BREAKPTS
    if (debugging) {
	notify_debugger_of_class_prepare(EE(), (JHandle *)cb);
    }
#endif

    /* We need this for bootstrapping.  We can't set the Handle's class block
     * pointer in Object or Class until Class has been resolved.
     */
    if (cb == classJavaLangClass) {
	int         j;
	ClassClass **pcb;

	/*
	 * Extra stack check for execution of InitPrimitiveClasses()
	 * to shrink STACK_REDZONE.
	 */
	if (!sysThreadCheckStack(STACK_REDZONE)) {
	    return JAVAPKG "StackOverflowError";
	}

	javaLangClassMethods = cbMethodTable(cb);

	BINCLASS_LOCK();
	for (j = nbinclasses, pcb = binclasses; --j >= 0; pcb++) {
	    obj_methodtable(cbHandle(*pcb)) = cbMethodTable(cb);
	}
	BINCLASS_UNLOCK();
	InitPrimitiveClasses();
    }
    return NULL;
}

static bool_t
RunClinit(ClassClass * cb)
{

    if (CCIs(cb, Resolved))
        return TRUE;

    if ((cbName(cb)[0] != SIGNATURE_ARRAY) && !cbIsPrimitive(cb)) {
	/* Don't need to initialize or verify array or primitive classes */
        return RunStaticInitializers(cb);
    } else if (cbName(cb)[0] == SIGNATURE_ARRAY) {
	ClassClass *inner_cb = 
	    cbConstantPool(cb)[CONSTANT_POOL_ARRAY_CLASS_INDEX].clazz;
	if (inner_cb) {
	    char *detail = 0;
	    char *ret = ResolveClass(inner_cb, &detail);
	    if (ret != NULL) {
	        if (!exceptionOccurred(EE())) { 
		    SignalError(0, ret, detail);
		}
		CCSet(cb, Error);
		return FALSE;
	    } else {
	        CCSet(cb, Resolved);
	    }
	} else {
	    CCSet(cb, Resolved);
	}
    } else {
        CCSet(cb, Resolved);
    }
    return TRUE;
}


/* Find an already loaded class with the given name.  If no such class exists
 * then return 0.
 */

#ifdef DEBUG
void
PrintLoadedClasses()
{
    int         j;
    ClassClass **pcb, *cb;
    BINCLASS_LOCK();
    pcb = binclasses;
    for (j = nbinclasses; --j >= 0; pcb++) {
	cb = *pcb;
	jio_fprintf(stdout, "class=%s, 0x%x\n", cbName(cb), cbLoader(cb));
    }
    BINCLASS_UNLOCK();
}
#endif

static ClassClass *
FindLoadedClass(char *name, struct Hjava_lang_ClassLoader *loader)
{
    int         j;
    ClassClass **pcb, *cb;

    BINCLASS_LOCK();
    pcb = binclasses;
    for (j = nbinclasses; --j >= 0; pcb++) {
	cb = *pcb;
	if ((cbLoader(cb) == loader) && (strcmp(name, cbName(cb)) == 0)) 
	    break;
    }
    BINCLASS_UNLOCK();
    if (j >= 0)
        return cb;
    return NULL;
}

/* We attempt to resolve it, also, if the "resolve" argument is true.
 * Otherwise, we only perform a minimal initialization on it, such that
 * it points to its superclass, which points to its superclass. . . .
 */

static ClassClass *
InitializeAndResolveClass(ClassClass *cb, bool_t resolve)
{
    char *err, *detail = NULL;
    if ((err = InitializeClass(cb, &detail))) {
        /* Check for underlying error already posted */
        ExecEnv *ee = EE();
	if (!exceptionOccurred(ee))
	    SignalError(0, err, detail);
	return 0;
    }
    if (resolve) {
        err = ResolveClass(cb, &detail);
	if (err) {
	    /* Check for underlying error already posted */ 
	    if (!exceptionOccurred(EE())) 
	        SignalError(0, err, detail);
	    return 0;
	}
    }
    INIT_CLASS_IF_NEEDED(cb);
    /* Check for error posted during INIT_CLASS_IF_NEEDED() */ 
    if (exceptionOccurred(EE()))
	return 0;

    return cb;
}

/* Find the class with the given name.  
 * If "resolve" is true, then we completely resolve the class.  Otherwise,
 * we only make sure that it points to its superclass, which points to its
 * superclass, . . . .
 */
ClassClass *(*class_loader_func)(HObject *loader, char *name, long resolve) = 0;

ClassClass *
FindClass(struct execenv *ee, char *name, bool_t resolve)
{
    /* skip pseudo frames */
    ClassClass *from = ee ? getCallerClass(ee, 0) : 0;
    return FindClassFromClass(ee, name, resolve, from);
}


/*
 * lock_classes and unlock_classes are exported by the JIT interface,
 * but no longer used anywhere else. It grabs both locks to ensure 
 * backward compatibility with 1.0.2.
 */
void
lock_classes()
{
    LOADCLASS_LOCK();
    BINCLASS_LOCK();
}

void
unlock_classes()
{
    BINCLASS_UNLOCK();
    LOADCLASS_UNLOCK();
}

/* DefineClass is the same as ClassLoader.DefineClass, except that
 * it does not initialize the class. Splitting this common code 
 * out of ClassLoader.DefineClass allows it to be shared by
 * RemoteClassLoader.DefineClass.
 */

Hjava_lang_Class *
DefineClass(struct Hjava_lang_ClassLoader *cl, 
	    struct Hjava_lang_String *name, 
	    HArrayOfByte *data,
	    long offset, long length) 
{
    ClassClass	    *cb;
    unsigned char   *body;
    int		    body_len;
    char	    *detail = 0;
    char	    *ename;
    char            *utf;

    if (data == 0) {
	SignalError(0, JAVAPKG "NullPointerException", 0);
	return 0;
    }
    cb = allocClassClass();
    if (cb == 0) {
        SignalError(0, JAVAPKG "OutOfMemoryError", 0);
        return 0;
    }
    cbLoader(cb) = cl;
    cbRomIndex(cb) = -1;
    body = (unsigned char *) unhand(data)->body;
    body_len = obj_length(data);
    if (offset < 0 || offset >= body_len || length < 0
		   || offset + length > body_len) {
	SignalError(0, JAVAPKG "ArrayIndexOutOfBoundsException", 0);
	return 0;
    }

    if (name != NULL) {
	Classjava_lang_String *str = unhand(name);
	unicode *str_body = unhand(str->value)->body;
	unicode *unistring = str_body + str->offset;
	int unilength = str->count;
	char *utfptr = unicode2utf(unistring, unilength, 0, 0);
	KEEP_POINTER_ALIVE(str_body);
	utf = utfptr;
	/* Convert any periods in the name to slashes */
	while (*utfptr != '\0') { 
	    unicode ch = next_utf2unicode(&utfptr);
	    if (ch == '.') 
		utfptr[-1] = '/';
	}
    } else { 
	utf = NULL;
    }

    /*
     * We need to lock the classes while creating the class and adding
     * the class to the list of classes
     */
    LOADCLASS_LOCK();
    if (!createInternalClass(body+offset, body+offset+length, cb, cl, utf,
                             &detail)) {
        SignalError(0, JAVAPKG "ClassFormatError", detail);
        LOADCLASS_UNLOCK();
        return 0;
    }
    LOADCLASS_UNLOCK();
    if (utf != NULL) {
        sysFree(utf);
    }

    KEEP_POINTER_ALIVE(body);

    return cb;
}

ClassClass *
FindClassFromClass(struct execenv *ee, char *name, bool_t resolve,
		   ClassClass *from)
{
    ClassClass *cb = 0;

    /*
     * Check for stack overflow which might be caused by possible 
     * recursive invocation of this function.
     */
    if (!sysThreadCheckStack(STACK_REDZONE)) {
	/* Got a stack overflow. Throw the preallocated error object. */
	exceptionThrow(ee ? ee : EE(), 
		       exceptionInternalObject(IEXC_StackOverflow));
	return 0;
    }

    if (name[0] == SIGNATURE_ARRAY) {
	cb = Locked_FindArrayClassFromClass(ee, name, from) ;
	if (cb && !exceptionOccurred(ee ? ee : EE()))
	    return InitializeAndResolveClass(cb, resolve);
    } else if (from != NULL && cbLoader(from)) {
	/* unlock the classes while the class is located by
	 * the class loader. The class loader should do this
	 * in a synchronized method
	 */
	return ClassLoaderFindClass(ee, cbLoader(from), name, resolve);
    } else {
        LOADCLASS_LOCK();
	cb = FindLoadedClass(name, 0);
	if (cb == 0) {
	    /* Check if there was an error; bail if so */
	    if (ee == NULL)
		ee = EE();
	    if (!exceptionOccurred(ee))
		cb = LoadClassLocally(name);
	}
	LOADCLASS_UNLOCK();
#ifdef REMOTECLASSLOADING
	if (cb == 0 && !exceptionOccurred(ee ? ee : EE()))
	    cb = FindRemoteClass(ee, name);
#endif
	if (cb && !exceptionOccurred(ee ? ee : EE()))
	    return InitializeAndResolveClass(cb, resolve);
    }
    return 0;
}

/*
 * FindStickySystemClass is a variant of FindClassFromClass that makes the 
 * class sticky (immune to class GC) if the class can be found.  It is only
 * used on system classes, i.e. classes with no class loader, so it's 
 * equivalent to FindClassFromClass with "from" argument 0.  It returns 0
 * if the class can't be found, and like FindClass assumes that the caller
 * will deal with that.  Note that until the class has been made sticky it
 * must be kept somewhere where GC will find it.
 */
ClassClass *
FindStickySystemClass(struct execenv *ee, char *name, bool_t resolve)
{
    ClassClass *cb;
    if ((cb = FindClassFromClass(ee, name, resolve, 0)) != NULL) {
	MakeClassSticky(cb);
    }
    return cb;
}

/* Find the array class with the specified name.  */
static ClassClass *
Locked_FindArrayClassFromClass(struct execenv *ee, char *name, 
			       ClassClass *from) { 
    struct Hjava_lang_ClassLoader *fromLoader = 
	        (from  == NULL) ? NULL : cbLoader(from);
    char *name_p;
    int depth, base_type;
    ClassClass *cb;
    struct Hjava_lang_ClassLoader *loader;
    ClassClass *inner_cb;

    sysAssert(name[0] == SIGNATURE_ARRAY);

    if (fromLoader == NULL) { 
	/* quick optimization if we know that we don't need a class loader */
	cb = FindLoadedClass(name, NULL);
	if (cb != NULL) 
	    return cb;
    }
	    


    /* Strip off all the initial SIGNATURE_ARRAY's to determine the depth,
     * and also what's left.  When we're done, name_p points at the first
     * non-character, and depth is the count of the array dpeth
     */
    for (name_p = name, depth = 0; *name_p == SIGNATURE_ARRAY; 
	 name_p++, depth++);

    /* Look at the character to determine what type of array we have. */
    switch(*name_p++) {
        case SIGNATURE_INT:    base_type = T_INT; break;
	case SIGNATURE_LONG:   base_type = T_LONG; break;
	case SIGNATURE_FLOAT:  base_type = T_FLOAT; break;
	case SIGNATURE_DOUBLE: base_type = T_DOUBLE; break;
	case SIGNATURE_BOOLEAN:base_type = T_BOOLEAN; break; 
	case SIGNATURE_BYTE:   base_type = T_BYTE; break;
	case SIGNATURE_CHAR:   base_type = T_CHAR; break;
	case SIGNATURE_SHORT:  base_type = T_SHORT; break;
	case SIGNATURE_CLASS:  base_type = T_CLASS; break;
        default:	       return NULL;	/* bad signature */
    }
    if (base_type == T_CLASS) { 
	char buffer_space[50];
	char *buffer = buffer_space;
	char *endChar = strchr(name_p, SIGNATURE_ENDCLASS);
	int nameLength = endChar - name_p;
	if (endChar == NULL || endChar[1] != '\0')
	    return NULL;
	if (nameLength >= sizeof(buffer_space)) /* need bigger buffer */
	    buffer = sysMalloc(nameLength + 1);
	memcpy(buffer, name_p, nameLength);
	buffer[nameLength] = '\0';
	inner_cb = FindClassFromClass(ee, buffer, FALSE, from);
	if (buffer != buffer_space) /* free buffer, if we malloc'ed it */
	    sysFree(buffer);
	if (inner_cb == NULL) 
	    return NULL;
	/* loader should either be fromLoader or NULL. */
	loader = cbLoader(inner_cb);
    } else { 
	if (*name_p != '\0')			/* bad signature */
	    return NULL;
	inner_cb = NULL;
	loader = NULL;
    }


    LOADCLASS_LOCK();
    cb = FindLoadedClass(name, loader);
    if (cb == NULL) { 
	/* Create the fake array class corresponding to this.
	 */
	cb = createFakeArrayClass(name, base_type, depth, inner_cb, loader);
    } 
    LOADCLASS_UNLOCK();
    return cb;
}

/*
 * Convert a name and type to a hash code
 *
 * We make copies of all strings that go into the hash table
 * in case the class that is the source of the strings being
 * inserted is eventually unloaded.
 */
unsigned NameAndTypeToHash(char *name, char *type)
{
    unsigned name_key;
    unsigned type_key;
    unsigned ret;

    NAMETYPEHASH_LOCK();
    if (((name_key = Str2ID(&nameTypeHash, name, 0, TRUE)) == 0) ||
	((type_key = Str2ID(&nameTypeHash, type, 0, TRUE)) == 0)) {
	SignalError(0, JAVAPKG "OutOfMemoryError", 0);
	ret = 0;
    } else {
	ret = (name_key << 16) + type_key;
    }
    NAMETYPEHASH_UNLOCK();

    return ret;
}

/*
 * Convert a NameAndType hash code to a name
 */
static char *
HashToKey(unsigned short nameOrSigID)
{
    char *key;

    NAMETYPEHASH_LOCK();
    key = ID2Str(nameTypeHash, nameOrSigID, 0);
    NAMETYPEHASH_UNLOCK();

    /*
     * 'key' will be 0 if ID2Str fails or if the entry is not there
     */
    return key;
}

/*
 * Convert a NameAndType hash code to a name
 */
char *HashToName(unsigned long ID)
{
    return HashToKey((unsigned short)(ID >> 16));
}

/*
 * Convert a NameAndType hash code to a name
 */
char *HashToSig(unsigned long ID)
{
    return HashToKey((unsigned short)(ID & 0xFFFF));
}

/*
 * Pseudo-classes to represent primitive Java types.
 */

static struct primtype {
    char		*name;
    char		typesig;
    unsigned char	typecode;
    unsigned char	slotsize;
    unsigned char	elementsize;
    ClassClass		**cellp;
} const primitive_types[] = {
    { "void",	SIGNATURE_VOID,	   T_VOID,	0, 0,	&class_void },
    { "boolean",SIGNATURE_BOOLEAN, T_BOOLEAN,	4, 1,	&class_boolean },
    { "byte",	SIGNATURE_BYTE,	   T_BYTE,	4, 1,	&class_byte },
    { "char",	SIGNATURE_CHAR,	   T_CHAR,	4, 2,	&class_char },
    { "short",	SIGNATURE_SHORT,   T_SHORT,	4, 2,	&class_short },
    { "int",	SIGNATURE_INT,	   T_INT,	4, 4,	&class_int },
    { "long",	SIGNATURE_LONG,	   T_LONG,	8, 8,	&class_long },
    { "float",	SIGNATURE_FLOAT,   T_FLOAT,	4, 4,	&class_float },
    { "double",	SIGNATURE_DOUBLE,  T_DOUBLE,	8, 8,	&class_double }
};

#define	PRIMITIVE_TYPE_COUNT	\
	(sizeof (primitive_types) / sizeof (primitive_types[0]))

#define	GET_PRIMITIVE_CLASS(nm) \
	(class_##nm ? class_##nm : FindPrimitiveClass(#nm))

ClassClass *
FindPrimitiveClass(char *name)
{
    int i;
    ClassClass *cb;
    
    for (i = 0; i < PRIMITIVE_TYPE_COUNT; i++) {
	if (strcmp(name, primitive_types[i].name) == 0) {
	    cb = *primitive_types[i].cellp;
	    if (cb == NULL) {
		char *err, *detail = NULL;
		cb = createPrimitiveClass(primitive_types[i].name,
					  primitive_types[i].typesig,
					  primitive_types[i].typecode,
					  primitive_types[i].slotsize,
					  primitive_types[i].elementsize);
		sysAssert(cb != NULL);
		if ((err = InitializeClass(cb, &detail)) != NULL)
		    return NULL; /* XXX */
		if ((err = ResolveClass(cb, &detail)) != NULL)
		    return NULL; /* XXX*/
		*primitive_types[i].cellp = cb;
	    }
	    return cb;
	}
    }
    return NULL;
}

static void
InitPrimitiveClasses()
{
    int i;
    initializePrimitives = FALSE;
    for (i = 0; i < PRIMITIVE_TYPE_COUNT; i++)
	FindPrimitiveClass(primitive_types[i].name);
}
