/*
 * @(#)classruntime.c	1.103 98/10/20
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*-
 *      Runtime routines to support classes.
 */

#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <stdlib.h>

#include "oobj.h"
#include "tree.h"
#include "interpreter.h"
#include "bool.h"
#include "timeval.h"
#include "threads.h"
#include "monitor.h"
#include "exceptions.h"
#include "javaString.h"
#include "path.h"

#include "sys_api.h"

#include "java_lang_Throwable.h"
#include "java_lang_Thread.h"

#ifdef BREAKPTS
#include "breakpoints.h"
#endif

#ifdef HPROF
#include "vmprofiler.h"
#endif /* HPROF */


#ifdef DEBUG
void
VerifyClasses()
{
    static      lastn;
    int         i;
    BINCLASS_LOCK();
    sysAssert(nbinclasses >= lastn);
    lastn = nbinclasses;
    sysAssert(nbinclasses <= sizebinclasses);
    sysAssert(binclasses != 0 || nbinclasses == 0);
    for (i = 0; i < nbinclasses; i++) {
	ClassClass *clb = binclasses[i];
	sysAssert(clb);
	sysAssert(((int) clb & 3) == 0);
	sysAssert(cbFieldsCount(clb) < 1000);
	sysAssert(cbMethodsCount(clb) < 1000);
	sysAssert(cbInstanceSize(clb) && cbInstanceSize(clb) < 4000);
	sysAssert(cbName(clb)[0]);
    }
    BINCLASS_UNLOCK();
}
#endif

const Invoker invokers[] = {
    {invokeJavaMethod},		        /* INVOKE_JAVA_METHOD */
    {invokeSynchronizedJavaMethod},     /* INVOKE_SYNCHRONIZED_JAVA_METHOD */
    {invokeNativeMethod}, 	        /* INVOKE_NATIVE_METHOD */
    {invokeSynchronizedNativeMethod},   /* INVOKE_SYNCHRONIZED_NATIVE_METHOD */
    {invokeJNINativeMethod}, 	        /* INVOKE_JNI_NATIVE_METHOD */
    {invokeJNISynchronizedNativeMethod},/* INVOKE_JNI_SYNCHRONIZED_NATIVE_.. */
    {invokeLazyNativeMethod}, 	        /* INVOKE_LAZY_NATIVE_METHOD */
    {invokeAbstractMethod}, 	        /* INVOKE_ABSTRACT_METHOD */
#ifdef SUPPORTCOMPILER
    {invokeCompiledMethod}, 	        /* INVOKE_COMPILED_METHOD */
#else
    {NULL},             	        /* INVOKE_COMPILED_METHOD */
#endif
    {NULL} 	       	        	/* INVOKE_ILLEGAL */
};

/*
 * These are all flavors of method invocation depending on the type
 * of method (defined by its access flags) being invoked.  There
 * are 4 possibilities: java method, synchronized java method,
 * native (C) method, synchronized native method.  Each methodblock
 * has a pointer to the appropriate one for its access, and each
 * one of these procedures optimizes its particular case.
 *
 * They don't all use all the parameters, which is unfortunate.
 */
bool_t
invokeJavaMethod(JHandle *o, struct methodblock *mb, int args_size, ExecEnv *ee)
{
    int        nlocals = methodnlocals(mb); /* maximum number of locals */
    JavaFrame  *old_frame = ee->current_frame; /* previous frame */
    stack_item *optop = old_frame->optop; /* Points to just before args */
    JavaStack  *stack = old_frame->javastack; /* Current stack pointer */
    JavaFrame  *frame;

#ifdef JDK12
    /* Reserve at least two words for the return value from callee. */
    if (nlocals < 2) {
	nlocals = 2;
    }
#endif

    frame = (JavaFrame *)(optop + nlocals); /* tentative  */

#ifdef TRACING
    if (trace_is_enabled) {
	char where_am_i[100];
	pc2string(methodcode(mb), mb, where_am_i, where_am_i + sizeof where_am_i);
	jio_fprintf(stdout, "Entering %s\n", where_am_i);
	fflush(stdout);
    }
#endif /* TRACING */

    if (frame->ostack + methodmaxstack(mb) >= stack->end_data) {
      /* Without tmp_XXX variables, the C compiler gets confused
       * about pointer aliasing, and generates much worse code.
       */
	JavaStack *tmp_stack = stack;
	JavaFrame *tmp_frame = frame;
	stack_item *tmp_optop = optop;
	if (!ExpandJavaStack(ee, &tmp_stack, &tmp_frame, &tmp_optop,
			     args_size, nlocals, methodmaxstack(mb)))
	    return FALSE;
	stack = tmp_stack;
	frame = tmp_frame;
	optop = tmp_optop;
    }

    frame->javastack = stack;
    frame->prev = old_frame;
    frame->vars = optop;
    frame->optop = frame->ostack;
    frame->current_method = mb; 
    frame->constant_pool = cbConstantPool(methodclass(mb));
    frame->returnpc = methodcode(mb);
    frame->monitor = 0;
#ifdef BREAKPTS
    frame->lastpc = 0;
#endif
#ifdef INCLUDE_PROF
    if (java_monitor) {
	frame->mon_starttime = now();
    }
#endif /* INCLUDE_PROF */
#ifdef HPROF
    if (jvmpi_event_flags & JVMPI_EVENT_METHOD_ENTRY_ON) {
        jvmpi_method_entry(ee, o);
    }
#endif /* HPROF */

    ee->current_frame = frame;

#ifdef BREAKPTS
    if (debugging) {
	notify_debugger_of_frame_push(ee);
    }
#endif /* BREAKPTS */

    return TRUE;
}


bool_t
invokeSynchronizedJavaMethod(JHandle *o, struct methodblock *mb, int args_size,
			    ExecEnv *ee)
{
    int        nlocals = methodnlocals(mb); /* maximum number of locals */
    JavaFrame  *old_frame = ee->current_frame; /* previous frame */
    stack_item *optop = old_frame->optop; /* Points to just before args */
    JavaStack  *stack = old_frame->javastack; /* Current stack pointer */
    JavaFrame  *frame;

#ifdef JDK12
    /* Reserve at least two words for the return value from callee. */
    if (nlocals < 2) {
	nlocals = 2;
    }
#endif

    frame = (JavaFrame *)(optop + nlocals); /* tentative  */

#ifdef TRACING
    if (trace_is_enabled) {
	char where_am_i[100];
	pc2string(methodcode(mb), mb, where_am_i, where_am_i + sizeof where_am_i);
	jio_fprintf(stdout, "Entering %s\n", where_am_i);
	fflush(stdout);
    }
#endif /* TRACING */

    if (frame->ostack + methodmaxstack(mb) >= stack->end_data) {
      /* Without tmp_XXX variables, the C compiler gets confused
       * about pointer aliasing, and generates much worse code.
       */
	JavaStack *tmp_stack = stack;
	JavaFrame *tmp_frame = frame;
	stack_item *tmp_optop = optop;
	if (!ExpandJavaStack(ee, &tmp_stack, &tmp_frame, &tmp_optop,
			     args_size, nlocals, methodmaxstack(mb)))
	    return FALSE;
	stack = tmp_stack;
	frame = tmp_frame;
	optop = tmp_optop;
    }

    frame->javastack = stack;
    frame->prev = old_frame;
    frame->vars = optop;
    frame->optop = frame->ostack;
    frame->current_method = mb; 
    frame->constant_pool = cbConstantPool(methodclass(mb));
    frame->returnpc = methodcode(mb);
    frame->monitor = o;
#ifdef BREAKPTS
    frame->lastpc = 0;
#endif
#ifdef INCLUDE_PROF
    if (java_monitor) {
	frame->mon_starttime = now();
    }
#endif /* INCLUDE_PROF */
    ee->current_frame = frame;
#ifdef HPROF
    if (jvmpi_event_flags & JVMPI_EVENT_METHOD_ENTRY_ON) {
        jvmpi_method_entry(ee, o);
    }
#endif /* HPROF */

    monitorEnter(obj_monitor(o));

#ifdef BREAKPTS
    if (debugging) {
	notify_debugger_of_frame_push(ee);
    }
#endif /* BREAKPTS */

    return TRUE;
}

bool_t
invokeNativeMethod(JHandle *o, struct methodblock *mb, int args_size, 
		   ExecEnv *ee)
{
  /* We need to set up a frame even for normal native methods so that
   * we can use JNI functions in normal native methods.
   */
    JavaFrame *old_frame = ee->current_frame;
    stack_item *optop = old_frame->optop;
    JavaStack  *stack = old_frame->javastack;
    JavaFrame  *frame = (JavaFrame *)(optop + args_size);

    if (frame->ostack + JNI_REF_INFO_SIZE + JNI_DEFAULT_LOCAL_CAPACITY >=
	stack->end_data) {
      /* Without tmp_XXX variables, the C compiler gets confused
       * about pointer aliasing, and generates much worse code.
       */
	JavaStack *tmp_stack = stack;
	JavaFrame *tmp_frame = frame;
	if (!ExpandJavaStackForJNI(ee, &tmp_stack, &tmp_frame,
			   JNI_REF_INFO_SIZE + JNI_DEFAULT_LOCAL_CAPACITY))
	    return FALSE;
	stack = tmp_stack;
	frame = tmp_frame;
    }

    frame->javastack = stack;
    frame->prev = old_frame;
    frame->optop = frame->ostack + JNI_REF_INFO_SIZE;
    frame->current_method = mb;
 
    JNI_REFS_FREELIST(frame) = NULL;
    JNI_N_REFS_IN_USE(frame) = 0;
    JNI_REFS_CAPACITY(frame) = JNI_DEFAULT_LOCAL_CAPACITY;
 
    ee->current_frame = frame;

#ifdef BREAKPTS
    if (debugging) {
	notify_debugger_of_frame_push(ee);
    }
#endif /* BREAKPTS */

#ifdef INCLUDE_PROF
    if (java_monitor) {
	long start_time, end_time;
	start_time = now();
	optop = (*(stack_item * (*) (stack_item*, ExecEnv*)) methodcode(mb)) (optop, ee);
	end_time = now();
	java_mon(old_frame->current_method, mb, end_time - start_time);
    } else
#endif /* INCLUDE_PROF */
    {
#ifdef HPROF
    if (jvmpi_event_flags & JVMPI_EVENT_METHOD_ENTRY_ON) {
        jvmpi_method_entry(ee, o);
    }
#endif /* HPROF */
	optop = (*(stack_item * (*) (stack_item*, ExecEnv*)) methodcode(mb)) (optop, ee);
#ifdef HPROF
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_METHOD_EXIT)) {
        jvmpi_method_exit(ee);
    }
#endif /* HPROF */

    }
    old_frame->optop = optop;
    ee->current_frame = old_frame;

    TRACE_METHOD(ee, mb, args_size, TRACE_METHOD_NATIVE_RETURN);
    return !(exceptionOccurred(ee));
}

bool_t
invokeSynchronizedNativeMethod(JHandle *o, struct methodblock *mb, int args_size,
			       ExecEnv *ee)
{
  /* We need to set up a frame even for normal native methods so that
   * we can use JNI functions in normal native methods.
   */
    JavaFrame *old_frame = ee->current_frame;
    stack_item *optop = old_frame->optop;
    JavaStack  *stack = old_frame->javastack;
    JavaFrame  *frame = (JavaFrame *)(optop + args_size);
    struct sys_mon * monitor;

    if (frame->ostack + JNI_REF_INFO_SIZE + JNI_DEFAULT_LOCAL_CAPACITY >=
	stack->end_data) {
      /* Without tmp_XXX variables, the C compiler gets confused
       * about pointer aliasing, and generates much worse code.
       */
	JavaStack *tmp_stack = stack;
	JavaFrame *tmp_frame = frame;
	if (!ExpandJavaStackForJNI(ee, &tmp_stack, &tmp_frame,
			   JNI_REF_INFO_SIZE + JNI_DEFAULT_LOCAL_CAPACITY))
	    return FALSE;
	stack = tmp_stack;
	frame = tmp_frame;
    }
 
    frame->javastack = stack;
    frame->prev = old_frame;
    frame->optop = frame->ostack + JNI_REF_INFO_SIZE;
    frame->current_method = mb;
 
    JNI_REFS_FREELIST(frame) = NULL;
    JNI_N_REFS_IN_USE(frame) = 0;
    JNI_REFS_CAPACITY(frame) = JNI_DEFAULT_LOCAL_CAPACITY;
 
    ee->current_frame = frame;

    monitorEnter(obj_monitor(o));

#ifdef BREAKPTS
    if (debugging) {
	notify_debugger_of_frame_push(ee);
    }
#endif /* BREAKPTS */

#ifdef INCLUDE_PROF
    if (java_monitor) {
	long start_time, end_time;
	start_time = now();
	optop = (*(stack_item * (*) (stack_item*, ExecEnv*)) methodcode(mb)) (optop, ee);
	end_time = now();
	java_mon(old_frame->current_method, mb, end_time - start_time);
    } else
#endif /* INCLUDE_PROF */
    {
#ifdef HPROF
        if (jvmpi_event_flags & JVMPI_EVENT_METHOD_ENTRY_ON) {
            jvmpi_method_entry(ee, o);
        }
#endif /* HPROF */
   	optop = (*(stack_item * (*) (stack_item*, ExecEnv*)) methodcode(mb)) (optop, ee);
        monitorExit(obj_monitor(o));
    }
#ifdef HPROF
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_METHOD_EXIT)) {
        jvmpi_method_exit(ee);
    }
#endif /* HPROF */

    old_frame->optop = optop;
    ee->current_frame = old_frame;

    TRACE_METHOD(ee, mb, args_size, TRACE_METHOD_NATIVE_RETURN);
    return !(exceptionOccurred(ee));
}

/* 
 * The invoke***JNINativeMethod functions are similar to
 * invoke***NativeMethod functions. The differences are here we call
 * sysInvokeNative to transfer the arguments onto the Java stack.
 * In addition, arguments that are Java objects or arrays are added
 * to the local registry.
 */

bool_t
invokeJNINativeMethod(JHandle *o, struct methodblock *mb, 
		      int args_size, ExecEnv *ee)
{
    JavaFrame  *old_frame = ee->current_frame; /* previous frame */
    stack_item *optop = old_frame->optop; /* Points to just before args */
    JavaStack  *stack = old_frame->javastack;
    JavaFrame  *frame = (JavaFrame *)(optop + args_size);
    ClassClass **isStatic = 0;
    JNIEnv *env = EE2JNIEnv(ee);

    if (frame->ostack + JNI_REF_INFO_SIZE + JNI_DEFAULT_LOCAL_CAPACITY >=
	stack->end_data) {
      /* Without tmp_XXX variables, the C compiler gets confused
       * about pointer aliasing, and generates much worse code.
       */
	JavaStack *tmp_stack = stack;
	JavaFrame *tmp_frame = frame;
	if (!ExpandJavaStackForJNI(ee, &tmp_stack, &tmp_frame,
			   JNI_REF_INFO_SIZE + JNI_DEFAULT_LOCAL_CAPACITY))
	    return FALSE;
	stack = tmp_stack;
	frame = tmp_frame;
    }

    frame->javastack = stack;
    frame->prev = old_frame;
    frame->optop = frame->ostack + JNI_REF_INFO_SIZE;
    frame->current_method = mb;

    JNI_REFS_FREELIST(frame) = NULL;
    JNI_N_REFS_IN_USE(frame) = 0;
    JNI_REFS_CAPACITY(frame) = JNI_DEFAULT_LOCAL_CAPACITY;


    ee->current_frame = frame;

    if (methodaccess(mb) & ACC_STATIC)
	isStatic = (ClassClass**)&(o);

#ifdef BREAKPTS
    if (debugging) {
	notify_debugger_of_frame_push(ee);
    }
#endif /* BREAKPTS */

#ifdef HPROF
    if (jvmpi_event_flags & JVMPI_EVENT_METHOD_ENTRY_ON) {
	jvmpi_method_entry(ee, o);
    }
#endif /* HPROF */


#ifdef INCLUDE_PROF
    if (java_monitor) {
	long start_time, end_time;
	start_time = now();
	optop = (stack_item *)sysInvokeNative((void *)env,
					      (void*)methodcode(mb),
					      (long*)optop,
					      methodsig(mb),
					      methodargssize(mb),
					      isStatic);
	end_time = now();
	java_mon(old_frame->current_method, mb, end_time - start_time);
	/* java_mon(methodclass(mb), mb, ee, 0, end_time - start_time); */
    } else
#endif /* INCLUDE_PROF */
    {
	optop = (stack_item *)sysInvokeNative((void *)env,
					      (void*)methodcode(mb),
					      (long*)optop,
					      methodsig(mb),
					      methodargssize(mb),
					      isStatic);
    }

#ifdef BREAKPTS
    if (debugging) {
	notify_debugger_of_frame_pop(ee);
    }
#endif /* BREAKPTS */
#ifdef HPROF
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_METHOD_EXIT)) {
	jvmpi_method_exit(ee);
    }
#endif /* HPROF */


    CHECK_CRITICAL_COUNT_ON_RETURN(ee);

    old_frame->optop = optop;
    ee->current_frame = old_frame;

    TRACE_METHOD(ee, mb, args_size, TRACE_METHOD_NATIVE_RETURN);
    return !(exceptionOccurred(ee));
}

bool_t
invokeJNISynchronizedNativeMethod(JHandle *o, struct methodblock *mb, 
				  int args_size, ExecEnv *ee)
{
    JavaFrame  *old_frame = ee->current_frame; /* previous frame */
    stack_item *optop = old_frame->optop; /* Points to just before args */
    JavaStack  *stack = old_frame->javastack;
    JavaFrame  *frame = (JavaFrame *)(optop + args_size);
    ClassClass **isStatic = 0;
    JNIEnv *env = EE2JNIEnv(ee);

    if (frame->ostack + JNI_REF_INFO_SIZE + JNI_DEFAULT_LOCAL_CAPACITY >=
	stack->end_data) {
      /* Without tmp_XXX variables, the C compiler gets confused
       * about pointer aliasing, and generates much worse code.
       */
	JavaStack *tmp_stack = stack;
	JavaFrame *tmp_frame = frame;
	if (!ExpandJavaStackForJNI(ee, &tmp_stack, &tmp_frame,
			   JNI_REF_INFO_SIZE + JNI_DEFAULT_LOCAL_CAPACITY))
	    return FALSE;
	stack = tmp_stack;
	frame = tmp_frame;
    }

    frame->javastack = stack;
    frame->prev = old_frame;
    frame->optop = frame->ostack + JNI_REF_INFO_SIZE;
    frame->current_method = mb;

    JNI_REFS_FREELIST(frame) = NULL;
    JNI_N_REFS_IN_USE(frame) = 0;
    JNI_REFS_CAPACITY(frame) = JNI_DEFAULT_LOCAL_CAPACITY;


    ee->current_frame = frame;

    if (methodaccess(mb) & ACC_STATIC)
	isStatic = (ClassClass**)&(o);

    monitorEnter(obj_monitor(o));

#ifdef BREAKPTS
    if (debugging) {
	notify_debugger_of_frame_push(ee);
    }
#endif /* BREAKPTS */

#ifdef HPROF
    if (jvmpi_event_flags & JVMPI_EVENT_METHOD_ENTRY_ON) {
	jvmpi_method_entry(ee, o);
    }
#endif /* HPROF */


#ifdef INCLUDE_PROF
    if (java_monitor) {
	long start_time, end_time;
	start_time = now();
	optop = (stack_item *)sysInvokeNative((void *)env,
					      (void*)methodcode(mb),
					      (long*)optop,
					      methodsig(mb),
					      methodargssize(mb),
					      isStatic);
	end_time = now();
	java_mon(old_frame->current_method, mb, end_time - start_time);
	/* java_mon(methodclass(mb), mb, ee, 0, end_time - start_time); */
    } else
#endif /* INCLUDE_PROF */
    {    
	optop = (stack_item *)sysInvokeNative((void *)env,
					      (void*)methodcode(mb),
					      (long*)optop,
					      methodsig(mb),
					      methodargssize(mb),
					      isStatic);
    }

#ifdef BREAKPTS
    if (debugging) {
	notify_debugger_of_frame_pop(ee);
    }
#endif /* BREAKPTS */
#ifdef HPROF
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_METHOD_EXIT)) {
	jvmpi_method_exit(ee);
    }
#endif /* HPROF */

 
    monitorExit(obj_monitor(o));

    CHECK_CRITICAL_COUNT_ON_RETURN(ee);

    old_frame->optop = optop;
    ee->current_frame = old_frame;

    TRACE_METHOD(ee, mb, args_size, TRACE_METHOD_NATIVE_RETURN);
    return !(exceptionOccurred(ee));
}

bool_t
invokeLazyNativeMethod(JHandle *o, struct methodblock *mb, int args_size, 
		       ExecEnv *ee)
{     
    monitorEnter(obj_monitor(cbHandle(methodclass(mb))));
    if (methodcode(mb) == 0) {
        if (dynoLink(mb)) {
	    methodinvokeridx(mb) = (methodaccess(mb) & ACC_SYNCHRONIZED) 
	      ? INVOKE_SYNCHRONIZED_NATIVE_METHOD : INVOKE_NATIVE_METHOD;

	} else if (dynoLinkJNI(mb)) {
	    methodinvokeridx(mb) = (methodaccess(mb) & ACC_SYNCHRONIZED) 
	      ? INVOKE_JNI_SYNCHRONIZED_NATIVE_METHOD : INVOKE_JNI_NATIVE_METHOD;
	}
    }
    monitorExit(obj_monitor(cbHandle(methodclass(mb))));

    if (methodcode(mb)) {
	return (methodinvoker(mb))(o, mb, args_size, ee);
    } else {
	if (!exceptionOccurred(ee)) {
	    /* 
	     * Do not overwrite OutOfMemoryError which could occur 
	     * during dynoLink***() 
	     */
	    SignalError(ee, JAVAPKG "UnsatisfiedLinkError", methodname(mb));
	}
	return FALSE;
    }
}

bool_t
invokeAbstractMethod(JHandle *o, struct methodblock *mb, int args_size, 
		     ExecEnv *ee)
{     
    SignalError(ee, JAVAPKG "AbstractMethodError", 
		methodname(mb));
    return FALSE;
}

bool_t
dynoLink(struct methodblock *mb)
{
    /* char sym[300]; */
#   define symSize 300
    char *sym;
    ClassClass *cb = methodclass(mb);
    extern int is_untrusted(ClassClass *);

    if (cbLoader(cb) && is_untrusted(cb)) {
	return FALSE;
    }

    /* Allocate buffer from heap to shrink stack requirement */
    if ((sym = sysMalloc(symSize * sizeof(char))) == NULL) {
	SignalError(EE(), JAVAPKG "OutOfMemoryError", 0);
	return FALSE;
    }

    mangleMethodName(mb, sym, symSize, MangleMethodName_JDK_1);
    methodcode(mb) = (unsigned char *)sysDynamicLink(sym);
    sysFree(sym);
    return (methodcode(mb) != 0);
}

bool_t
dynoLinkJNI(struct methodblock *mb)
{
    int i;
    /* char sym[1024]; */
#   define symSizeJNI 1024
    char *sym;
    ClassClass *cb = methodclass(mb);
    extern int is_untrusted(ClassClass *);

    if (cbLoader(cb) && is_untrusted(cb)) {
	return FALSE;
    }

    /* Allocate buffer from heap to shrink stack requirement */
    if ((sym = sysMalloc(symSizeJNI * sizeof(char))) == NULL) {
	SignalError(EE(), JAVAPKG "OutOfMemoryError", 0);
	return FALSE;
    }
   
    for (i=0; ;i++) {
        mangleMethodName(mb, sym, symSizeJNI, MangleMethodName_JNI_SHORT);
        if (!sysBuildFunName(sym, symSizeJNI, mb, i))
	    break;
        methodcode(mb) = (unsigned char *)sysDynamicLink(sym);
	if (methodcode(mb) != 0) {
	    sysFree(sym);
            return TRUE;
	}
    }

    for (i=0; ;i++) {
        mangleMethodName(mb, sym, symSizeJNI, MangleMethodName_JNI_LONG);
	if (!sysBuildFunName(sym, symSizeJNI, mb, i))
	    break;
        methodcode(mb) = (unsigned char *)sysDynamicLink(sym);
	if (methodcode(mb) != 0) {
	    sysFree(sym);
	    return TRUE;
	}
    }
    sysFree(sym);
    return FALSE;
}

/*  I put this into a separate function so that other parts of the
    interpreter can use it without knowing how an object is actually
    created. - csw
*/
HObject *
newobject(ClassClass *cb, unsigned char *pc, struct execenv *ee)
{
    HObject *ret;
    if ((ret = ObjAlloc(cb, 0)) == 0) {
	char buf[128];
	if (ee != NULL && pc != NULL) 
	    ee->current_frame->lastpc = pc;
	SignalError(ee, JAVAPKG "OutOfMemoryError", classname2string(cbName(cb), buf, sizeof(buf)));
	return 0;
    }
    return ret;
}

bool_t
RunStaticInitializers(ClassClass *cb)
{
    extern void free_clinit_memory(struct methodblock *mb);
    struct methodblock *mb;
    int size;
    char *s;
    struct methodblock *initializer = NULL;
    ExecEnv *ee = EE();

    static unsigned clinitID = 0;
    if (clinitID == 0)
	clinitID = NameAndTypeToHash("<clinit>", "()V");


    mb = cbMethods(cb);
    for (size = 0; size < (int) cbMethodsCount(cb); size++, mb++) {
	sysAssert(methodoffset(mb) < cbMethodTableSize(cb) || cbIsInterface(cb));
	s = methodname(mb);
	if (methodID(mb) == clinitID) {
	    initializer = mb;
	    break;				/* only one initializer */
	}
    }

    /* The exact initialization procedure as documented in 2.16.5 of 
       the VM Spec. */

    monitorEnter(obj_monitor(cb));                        /* Step 1 */
    {
        void *init_thread = cbInitThread(cb);

	while (init_thread && init_thread != (void *)ee) {/* Step 2 */
	    monitorWait(obj_monitor(cb), TIMEOUT_INFINITY);
	    init_thread = cbInitThread(cb);
	}
	if ((init_thread && init_thread == (void *)ee) || /* Step 3 */
	    ((!cbIsInROM(cb) && CCIs(cb, Resolved)) ||
	     (cbIsInROM(cb) && ROMIs(cb, Initialized)))) {
	    /* Step 4 */
	    monitorExit(obj_monitor(cb));
	    return TRUE;
	}
	if (CCIs(cb, Error)) {                            /* Step 5 */
	    SignalError(0, JAVAPKG "NoClassDefFoundError", cbName(cb));
	    monitorExit(obj_monitor(cb));
	    return FALSE;
	}
	cbInitThread(cb) = ee;                            /* Step 6 */
    }
    monitorExit(obj_monitor(cb));

    if (!cbIsInterface(cb)) {                             /* Step 7 */
        if (cbSuperclass(cb)) {
	    char *detail = 0;
	    char *exc = ResolveClass(cbSuperclass(cb), &detail);
	    if (exc) {
	        if (!exceptionOccurred(ee)) {
		    SignalError(ee, exc, detail);
		}
	    }
	}
    }

    if (ee && exceptionOccurred(ee)) {
        monitorEnter(obj_monitor(cb));
	cbInitThread(cb) = 0;
	if (!cbIsInROM(cb)) {
	    CCSet(cb,Error);
	}
	monitorNotifyAll(obj_monitor(cb));
	monitorExit(obj_monitor(cb));
	return FALSE;
    }

#ifndef TRIMMED
    if (verbose)
	jio_fprintf(stderr, "[Initializing %s]\n", cbName(cb));
#endif

                                                          /* Step 8 */
#ifdef SUPPORTCOMPILER
    InitializeForCompiler(cb);
#endif
    if (initializer != NULL) {
#ifndef TRIMMED
        if (verbose)
	    jio_fprintf(stderr, "[Running static initializer for %s]\n", 
		    cbName(cb));
#endif
	sysAssert(methodcode(initializer) != NULL);
	do_execute_java_method(ee, cb, 0, 0, initializer, TRUE);
	if (!cbIsInROM(cb)
#ifdef BREAKPTS
	    && !debugging
#endif
	) {
            /* Free the code for the initializer */
	    free_clinit_memory(initializer);
	    /* Clean out other pointers for GC */
	    methodcode(initializer) = NULL;
	    /* This isn't necessary.  But it makes sure mb->code 
	       isn't called */
	    methodinvokeridx(initializer) = INVOKE_ABSTRACT_METHOD;
        }
    }

    /* Now handle any exception */
    if (initializer == NULL || !ee || 
	!exceptionOccurred(ee)) {                         /* Step 9 */
        monitorEnter(obj_monitor(cb));
	cbInitThread(cb) = 0;
	if (!cbIsInROM(cb)) {
	    CCSet(cb,Resolved);
	}
	monitorNotifyAll(obj_monitor(cb));
	monitorExit(obj_monitor(cb));
	return TRUE;
    } else {                                              /* Step 10 */
        JHandle *oldError = ee->exception.exc;
	if (!is_instance_of(oldError, classJavaLangError, ee)) {
	    JHandle *newError;
	    exceptionClear(ee);
	    newError = execute_java_constructor(ee, 
					JAVAPKG "ExceptionInInitializerError",
						NULL, 
						"(L" JAVAPKG "Throwable;)",
						oldError);
	    if (!exceptionOccurred(ee)) { 
	        exceptionThrow(ee, newError);
	    } else { 
	        /* don't do anything. Some serious error must have occurred
		 * when constructing the ExceptionInInitializerError 
		 * object.  Let that error stand.
		 */
	    }
	}
	monitorEnter(obj_monitor(cb));                    /* Step 11 */
	cbInitThread(cb) = 0;
	if (!cbIsInROM(cb)) {
	    CCSet(cb,Error);
	}
	monitorNotifyAll(obj_monitor(cb));
	monitorExit(obj_monitor(cb));
	return FALSE;
    }
}

void
InitializeInvoker(ClassClass *cb)
{
    struct methodblock *mb;
    int size;

    mb = cbMethods(cb);
    for (size = 0; size < (int) cbMethodsCount(cb); size++, mb++) {
	unsigned int access = methodaccess(mb);
	sysAssert(methodoffset(mb) < cbMethodTableSize(cb) || cbIsInterface(cb));
#ifdef SUPPORTCOMPILER
	if (access & ACC_MACHINE_COMPILED) { 
	    methodinvokeridx(mb) = INVOKE_COMPILED_METHOD;
	} else 
#endif
	if ((access & (ACC_ABSTRACT | ACC_NATIVE)) == 0) {
	    /* Normal java method */
	    methodinvokeridx(mb) = (access & ACC_SYNCHRONIZED) 
		? INVOKE_SYNCHRONIZED_JAVA_METHOD
		: INVOKE_JAVA_METHOD;
	} else if (access & ACC_ABSTRACT) {
	    cbAccess(cb) |= ACC_ABSTRACT;
	    methodinvokeridx(mb) = INVOKE_ABSTRACT_METHOD;
	} else { 
	    methodinvokeridx(mb) = INVOKE_LAZY_NATIVE_METHOD;
	}
    }
}

void
SignalError(struct execenv * ee, char *ename, char *DetailMessage)
{
    ClassClass          *cb;
    Hjava_lang_Throwable *ret;
    Classjava_lang_Throwable *new_obj;
    JavaFrame            *frame;
#ifndef TRIMMED
    char where[100];
#endif

    if (ee == 0)
        ee = EE();

    /*
     * Check for stack overflow which might be caused by possible 
     * recursive invocation of this function.
     */
    if (!sysThreadCheckStack(STACK_REDZONE)) {
	/* Got a stack overflow. Throw the preallocated error object. */
	exceptionThrow(ee, exceptionInternalObject(IEXC_StackOverflow));
	return;
    }

#ifdef DEBUG
    if (exceptionOccurred(ee)) {
	/* Masking a pending error */
	ee = ee;
    }
#endif
    /*
     * Find the exception class (or the default).  If we can't find it
     * throw a preallocated object
     */
    if ((cb = FindClassFromClass(ee, ename, TRUE, NULL)) == 0 && 
	(cb = FindClassFromClass(ee, JAVAPKG "UnknownError", TRUE, NULL)) == 0) {
	/*
	 * If StackOveflowError occurred, do not try to call any stack
	 * consuming function. Instead just throw preallocated error object.
	 */
	if (strcmp(ename, JAVAPKG "StackOverflowError") == 0) {
	    exceptionThrow(ee, exceptionInternalObject(IEXC_StackOverflow));
	    return;
	}
#ifndef TRIMMED
	if (ee->current_frame) {
	    pc2string(ee->current_frame->lastpc, 
		      ee->current_frame->current_method, 
		      where, where + sizeof where);
	    jio_fprintf(stderr, "Class missing for error  \"%s\" at %s\n",
			ename, where);
	    if (DetailMessage) {
		jio_fprintf(stderr, 
			    "Detail message for error \"%s\" is \"%s\"\n", 
			    ename, DetailMessage);
	    }
	} else {
	    jio_fprintf(stderr, "Class missing for error  \"%s\"\n", ename);
	    if (DetailMessage) {
		jio_fprintf(stderr, 
			    "Detail message for error \"%s\" is \"%s\"\n", 
			    ename, DetailMessage);
	    }
	}
#endif
	exceptionThrow(ee, exceptionInternalObject(IEXC_NoClassDefinitionFound));
	return;
    }

    /*
     * Instantiate the object.  If we can't then throw a preallocated
     * object.
     */
    if ((ret = (Hjava_lang_Throwable *)ObjAlloc(cb, 0)) == NULL) { 
	exceptionThrow(ee, exceptionInternalObject(IEXC_OutOfMemory));
	return;
    }
    new_obj = unhand(ret);

    /* Backtrace through the interpreter frames. */
    frame = ee ? ee->current_frame : 0;
    if (frame == 0) {
#ifndef TRIMMED
	jio_fprintf(stderr, "Exception: %s (can't backtrace because of a missing "
		"context)\n", ename);
#endif
	/*	return;	*/
    }
#ifdef TRACING
    if (trace_is_enabled && frame) {
	jio_fprintf(stdout, "%6X %8X\tERROR %s\n", threadSelf(), frame->lastpc, ename);
	fflush(stdout);
    }
#endif /* TRACING */

    if (DetailMessage)
	new_obj->detailMessage = makeJavaStringUTF(DetailMessage);
    fillInStackTrace(ret, ee);

    /* This must be the last thing, since makeJavaString() may clear the
     * signal Error slot. 
     */
    exceptionThrow(ee, (HObject *)ret);
}


char       *
addhex(long n, char *buf, char *limit)
{
    int         i;
    for (i = 32; (i -= 4) >= 0 && buf < limit;)
	*buf++ = "0123456789ABCDEF"[(n >> i) & 0xF];
    return buf;
}

char       *
adddec(long n, char *buf, char *limit)
{
    char figs[10] = {0}; /* 32bit integer requires 10-digits at most */
    int  i;

    if (n < 0) { 
        n = -n; /* Critical for 0x80000000, converted into '-0' */ 
        if (buf < limit) 
            *buf++ = '-'; 
    } 
    for (i = 0; n > 0 && i < 10; i++) {
        figs[i] = n % 10;
        n /= 10;
    }
    /* figs[0] has been initialized to 0 */
    for (i = (i == 0) ? (0) : (i - 1); i >= 0 && buf < limit; i--) {
        *buf++ = figs[i] + '0'; 
    }
    return buf; 
}


int
pc2lineno(struct methodblock *mb, unsigned int pc_offset)
{
#ifdef CLASSDEBUGINFO
    int length = methodlnnotablen(mb);

    if (length > 0) {
	struct lineno *ln = &(methodlnnotab(mb)[length - 1]);
	int i;
	for (i = length; --i >= 0; ln--) {
	    if (pc_offset >= ln->pc) {
	        return ln->line_number;
	    }
	}
    }
    return -1;
#else
    return -1;
#endif /* CLASSDEBUGINFO */
}

static struct methodblock *pc2method(unsigned char *pc) {
    struct methodblock *mb;
    ClassClass *cb;
    int i, j;

    BINCLASS_LOCK();
    for (i = nbinclasses; --i >= 0; ) {
	cb = binclasses[i];
	for (mb = cbMethods(cb), j = cbMethodsCount(cb); --j >= 0; mb++) {
	    if (!(methodaccess(mb) & ACC_NATIVE)) {
		unsigned char *code_start;
#ifdef SUPPORTCOMPILER
		if (methodaccess(mb) & ACC_MACHINE_COMPILED) {
		    if (PCinCompiledCode(pc, mb))
		        goto unlock;
		} else
#endif
		{ 
		    code_start = (unsigned char *)methodcode(mb);
		    if (pc >= code_start && pc < code_start + methodcodelen(mb))
		        goto unlock;
		}
	    }
	}
    }
    mb = 0;

unlock:
    BINCLASS_UNLOCK();
    return mb;
}

char *
pc2string(unsigned char *pc, struct methodblock *mb, char *buf, char *limit)
{
    if (buf >= limit) {
	return buf;
    }
    --limit;	/* Save room for terminating '\0' */
    if (mb == 0) {
        mb = pc2method(pc);
    }
    if (mb != 0) {
	char cbuf[128];
	ClassClass *cb = methodclass(mb);
	buf = addstr(classname2string(cbName(cb),
		     cbuf, sizeof(cbuf)), buf, limit, 0);
	buf = addstr(".", buf, limit, 0);
	buf = addstr(methodname(mb), buf, limit, '(');
	if (cb != 0) {
#ifdef SUPPORTCOMPILER
	    if ((methodaccess(mb) & ACC_MACHINE_COMPILED) == 0)
#endif
	    {
#ifdef CLASSDEBUGINFO
		char        *fn;
		fn = cbSourceName(cb);
		if (fn) {
		    int         lineno;
		    /* Just want the short name */
		    fn = strrchr(fn, '/');
		    if (fn) {
			/* skip past last separator */
			fn += 1;
		    } else {
			fn = cbSourceName(cb);
		    }
		    buf = addstr("(", buf, limit, 0);
		    buf = addstr(fn, buf, limit, 0);
		    lineno = pc2lineno(mb, pc - methodcode(mb));
		    if (lineno >= 0) {
			buf = addstr(":", buf, limit, 0);
			buf = adddec(lineno, buf, limit);
		    }
		    buf = addstr(")", buf, limit, 0);
		} else {
		    buf = addstr("()", buf, limit, 0);
		}
#else
		buf = addstr("()", buf, limit, 0);
#endif /* CLASSDEBUGINFO */
	    }
#ifdef SUPPORTCOMPILER
	    else {
		buf = addstr("(Compiled Code)", buf, limit, 0);
	    }
#endif
	}
	*buf = 0;
    } else {
	buf[0] = '\0';
    }
    return buf;
}

static char *
pc2string12(unsigned char *pc, JavaFrame *frame, char *buf, char *limit)
{
    struct methodblock *mb = NULL;

    if (frame != NULL) {
	mb = frame->current_method;
    }
    return pc2string(pc, mb, buf, limit);
}

/*
   Make a Java unicode string out of a series of bytes that are
   already encoded as unicode.
*/
HArrayOfChar *
MakeUniString(unicode  *str, long len)
{
    unicode *new_str_body;
    JHandle *ret;
    unicode *tmp;

    sysAssert (str != 0);
    for (tmp = str; *tmp; tmp++)
	;
    if (len > tmp-str)
	len = (int) (tmp-str);
    ret = ArrayAlloc(T_CHAR, len);
    if (ret == 0)
	return 0;
    new_str_body = ((ArrayOfChar *)unhand(ret))->body;
    memmove(new_str_body, (char *) str, len * sizeof(unicode));
    KEEP_POINTER_ALIVE(new_str_body);
    return (HArrayOfChar *)ret;
}


/*
    Make a Java unicode string from a non-unicode C-style string.
*/
HArrayOfChar *
MakeString(const char *str, long len)
{
    HArrayOfChar  *ret;
    unicode * body;

    sysAssert(str != 0);
    ret = (HArrayOfChar *) ArrayAlloc(T_CHAR, len);
    if (ret == 0)
	return 0;
    body = unhand(ret)->body;
    str2unicode(str, body, len);
    KEEP_POINTER_ALIVE(body);
    return ret;
}

/*
    Make a Java array of bytes
*/
HArrayOfByte *
MakeByteString(char *str, long len)
{
    char *new_str_body;
    JHandle *ret;

    ret = ArrayAlloc(T_BYTE, len);
    if (ret == 0)
	return 0;
    new_str_body = ((ArrayOfByte *)unhand(ret))->body;
    if (str)
	memmove(new_str_body, (char *)str, len);
    else
	memset(new_str_body, 0, len);
    KEEP_POINTER_ALIVE(new_str_body);

    return (HArrayOfByte *)ret;
}

ExecEnv *
EE()
{
    ExecEnv *ee;
    TID self = threadSelf();

    if (self == 0) {
	ee = DefaultExecEnv;
    } else {
	ee = (ExecEnv *) THREAD(self)->eetop;
    }

    sysAssert(ee != 0);
    return ee;
}

void printStackTrace(ExecEnv *ee, int depth_limit)
{
    JavaFrame *frame, frame_buf;

    /* Print the stack back trace. 
     * This code is adapted from fillInStackTrace and printStackTrace.
     */
    frame = ee->current_frame;
    while (frame != NULL) {
    	if (frame->current_method != NULL) {
	    char where[MSG_BUF_LEN];
	    unsigned char *pc;
	    struct methodblock *mb = frame->current_method;

	    if (depth_limit-- <= 0)
	        break;

	    if (methodaccess(mb) & ACC_NATIVE) {
	        pc = methodcode(mb);
#ifdef SUPPORTCOMPILER
	    } else if (IS_JIT_FRAME(frame)) {
		pc = CompiledCodePC(frame, mb);
#endif
	    } else {
	        pc = frame->lastpc;
            }

	    strncpy(where, "\tat ", 4);
	    pc2string12(pc, frame, where + 4, where + sizeof(where));
	    jio_fprintf(stderr, "%s\n", where);
        }
	frame = FRAME_PREV(frame, &frame_buf);
    } 
}
