/*
 * @(#)debug.c	1.36 98/10/10
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Debugging help for users.
 */

#include <stdio.h>

#include "threads.h"
#include "debug.h"
#include "monitor.h"
#include "monitor_cache.h"
#include "exceptions.h"
#include "bool.h"
#include "interpreter.h"

#include "java_lang_Thread.h"

#ifdef LOGGING
int logging_level;
#endif

/*
 * Access to thread information.  Quietly truncate name on jio_snprintf()
 * buffer overflow.
 */
char *
threadName(HThread *tid)
{
    /* WARNING: There is no actual limit on the length of a thread name */
#define THREADNAMEMAX 80
    /* WARNING: Returning a name as a static buffer is not MT-safe! */
    static char	namebuf[THREADNAMEMAX];

    if (!tid) {
	(void) jio_snprintf(namebuf, sizeof(namebuf), "name unknown (nil)");
    } else {
        (void) jio_snprintf(namebuf, sizeof(namebuf), "%s (0x%8x)",
	        Object2CString((JHandle *) THREAD(tid)->name), tid);
    } 

    return (namebuf);
}

void
threadDumpInfo(TID tid, bool_t verbose)
{
    int priority;
    jio_fprintf(stderr, "    %s", Object2CString((JHandle *) THREAD(tid)->name));
    if (verbose == TRUE) {
	jio_fprintf(stderr, " (TID:0x%x", tid);
	if (THREAD(tid)->PrivateInfo) {
	    jio_fprintf(stderr, ", sys_thread_t:0x%x", SYSTHREAD(tid));	
            sysThreadDumpInfo(SYSTHREAD(tid));
            jio_fprintf(stderr, ")");
	    threadGetPriority(tid, &priority);
	    jio_fprintf(stderr, " prio=%d", priority);
	    if (unhand(tid)->eetop &&
		exceptionOccurred((struct execenv *)unhand(tid)->eetop)) {
	        char buf[256];
	        exception_t exc =
		    ((struct execenv *)unhand(tid)->eetop)->exception.exc;
	        jio_fprintf(stderr, ": pending=%s",
			classname2string(cbName(obj_classblock(exc)),
			buf, sizeof(buf)));
	    }
	    jio_fprintf(stderr, "%s\n",
		tid == threadSelf() ? " *current thread*" : "");
    	} else {
	    jio_fprintf(stderr, ") : <thread not active>\n");
	}
    } else {
	sys_thread_t *t = 0;
	if (THREAD(tid)->PrivateInfo) {
	    t = SYSTHREAD(tid);
	}
	jio_fprintf(stderr, " (0x%lx)\n", t);
    }

    return;
}

extern int
DumpThreadsHelper(sys_thread_t *t, void *arg)
{
    HThread *p = (HThread *) sysThreadGetBackPtr(t);
    struct Classjava_lang_Thread *tid = unhand(p);
    long n = 20;
    struct execenv *ee = (struct execenv *) tid->eetop;
    threadDumpInfo(p, TRUE);
    if (ee) {
	JavaFrame frame_buf;
	JavaFrame *frame = ee->current_frame;
	char pcb[256];

	/*
	 * Extra stack check to shrink STACK_REDZONE.
	 */
	if (!sysThreadCheckStack(STACK_REDZONE)) {
	    /*
	     * Not enough stack to process. Just return.
	     */
	    return SYS_NORESOURCE;
	}
	while (n > 0 && frame) {
            if (frame->current_method != 0) {
		pc2string(frame->lastpc, frame->current_method,
			  pcb, pcb + sizeof pcb);
		jio_fprintf(stderr, "\t%s\n", pcb);
		n--;
#ifdef SUPPORTCOMPILER
		frame = methodaccess(frame->current_method) & ACC_MACHINE_COMPILED ?
		    CompiledFramePrev(frame, &frame_buf) : frame->prev;
#else
		frame = frame->prev;
#endif
	    } else {
		frame = frame->prev;
	    }
	}
    }
    return SYS_OK;
}

void
DumpThreads()
{
    sys_thread_t * self;
    static int alreadyHere = 0;

    /* try to detect when we're just recursing and give up */
    if (alreadyHere) {
	panic( "Internal error dumping threads!");
    } else {
	alreadyHere++;
    }


    /* FIX THIS!!!!!!!
     * A hack for Win32. The Win32 ctrl-brk handler gets called in
     * a non-Java thread. Thus ee will be zero and we can't lock
     * the thread queue. 
     */
    self = sysThreadSelf();
    if (self) {
        QUEUE_LOCK();
    }

    jio_fprintf(stderr, "\nFull thread dump:\n");

    sysThreadEnumerateOver(DumpThreadsHelper, (void *) 0);

    if (self) {
        QUEUE_UNLOCK();
    }

    alreadyHere--;
}

/*
 * Access to monitor information
 */
static void
registeredDumpHelper(reg_mon_t *reg, void *ignored)
{
    jio_fprintf(stderr, "    %s: ", reg->name);
    sysMonitorDumpInfo(reg->mid);
}

/*
 * Clone of DumpMonitors() - should call the unlocked versions of the
 * enumerate routines.
 */
void
DumpMonitors_unlocked()
{
    /* Defined in gc.c as it needs to know about the heap. */
    extern void monitorDumpHelper(monitor_t *, void *);

    /*
     * Dump the contents of the monitor cache
     */
    jio_fprintf(stderr, "Monitor Cache Dump:\n");
    monitorEnumerate_unlocked(monitorDumpHelper, 0);

    /*
     * Also dump other well-known monitors: what these will be is
     * system-dependent and does not belong in share code.
     */
    jio_fprintf(stderr, "Registered Monitor Dump:\n");
    registeredEnumerate_unlocked(registeredDumpHelper, 0);
}

/*
 * If verbose == TRUE, we print information about unused cached monitors.
 * This is normally not done, but can be useful when debugging the cache
 * itself.
 */
void
DumpMonitors0(bool_t verbose)
{
    /* Defined in gc.c as it needs to know about the heap. */
    extern void monitorDumpHelper(monitor_t *, void *);

    /*
     * Dump the contents of the monitor cache
     */
    jio_fprintf(stderr, "Monitor Cache Dump:\n");
    monitorEnumerate(monitorDumpHelper, (void *) verbose);

    /*
     * Also dump other well-known monitors: what these will be is
     * system-dependent and does not belong in share code.
     */
    jio_fprintf(stderr, "Registered Monitor Dump:\n");
    registeredEnumerate(registeredDumpHelper, 0);
}

/*
 * Called from ^\, in which case we want to suppress information about
 * unused monitors in the cache.
 */
void
DumpMonitors()
{
    DumpMonitors0(FALSE);
}
