/*
 * @(#)finalize.c	1.45 98/10/10
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Support for finalization: the finalizer thread runs the methods
 * associated with objects that have been freed.
 */

#include "finalize.h"
#include "threads.h"
#include "monitor.h"
#include "oobj.h"
#include "log.h"
#include "sys_api.h"

sys_mon_t *_hasfinalq_lock;	     /* Lock for the HasFinalizerQ queue */
sys_mon_t *_finalmeq_lock;	     /* Lock for the FinalizeMeQ queue */

JHandle *HasFinalizerQ = NULL;       /* Live objects w/finalization method */
JHandle *FinalizeMeQ = NULL;         /* Garbage objects awaiting finalization */
JHandle *FinalizeMeQLast = NULL;     /* Last finalizer_t on FinalizeMeQ */

JHandle *BeingFinalized = NULL;	     /* The object currently being finalized; */
				     /* this has to be marked in GC so it's */
				     /* not thrown away if the finalizer */
				     /* thread is interrupted! */
static TID inBeingFinalized = NULL;  /* Who's got BeingFinalized */

bool_t finalize_on_exit = FALSE;     /* Control over finalization on exit */

extern int heap_memory_changes;

static void
execute_finalizer(JHandle *handle)
{
    struct methodblock *mb;
    struct execenv *ee = EE();

    sysAssert(unhand(handle) != 0);	/* Shouldn't have been reclaimed */
    sysAssert(BeingFinalized == handle);
    sysAssert(obj_isnt_array(handle));
    mb = mt_slot(obj_methodtable(handle), FINALIZER_MB_OFFSET);
    sysAssert(mb != NULL);
    do_execute_java_method(ee, handle, 0, 0, mb, FALSE);
    /*
     * The Java Language Spec says that exceptions not caught in a
     * finalize method are quietly ignored but terminate execution
     * of that object's finalization.
     */
    exceptionClear(ee);
}

/*
 * Explicitly, synchronously flush the FinalizeMeQ.  Normally this
 * is not necessary, as the finalizer thread will clean up.  However,
 * if the system is heavily loaded, there can be no guarantee that
 * the finalization thread will be run at all.  In this case, the
 * user can call Runtime.runFinalization() at reasonable intervals, or
 * when the allocation of a finalized resource fails.  Note that there
 * isn't a built-in equivalent to the garbage collector being run when
 * the heap fills up.  Finalized resources are arbitrary, so we can't
 * predict up front when to flush the queue.  However, for a given
 * resource users (or the system) can build equivalent functionality
 * into their allocators.
 *
 * runFinalization0 returns a boolean saying whether it made progress
 * or not: it might bail out without making progress if synchronous
 * finalization causes more synchronous finalization.  This boolean
 * is used by the memory planning code to decide whether it's worth
 * trying to GC.
 */

bool_t
runFinalization0(void)
{
    JHandle *handle;
    TID self = threadSelf();

    FINALMEQ_LOCK();

    /*
     * We might get called with FinalMeQ == NULL but with the last
     * finalized object still in BeingFinalized.  Hence do...while.
     */
    do {	
	/* If I am already finalizing, bail rather than self-deadlock */
	if (inBeingFinalized == self) {
	    /* Don't need to notify yourself */
	    FINALMEQ_UNLOCK();
	    return FALSE;
	}
	/* If someone else is finalizing, wait until they are done */
	while (BeingFinalized != NULL) {
	    /* ...that is, unless this thread has been interrupted. */
	    int ret = FINALMEQ_WAIT();	/* Don't clear interrupt state */
	    if (ret == SYS_INTRPT) {
		/* Finalizing thread will notify when done: we don't */
		FINALMEQ_UNLOCK();
		/* Only possible first time around, so FALSE is correct */
		return FALSE;
	    }
	}
	/* If someone else cleared the queue while we waited, return */
	if (FinalizeMeQ == NULL) {
	    break;
	}

	/* Else grab the next object to be finalized and finalize it */
	sysAssert(inBeingFinalized == NULL && BeingFinalized == NULL);
        handle = BeingFinalized = FinalizeMeQ;
        FinalizeMeQ = FINALIZER_T(handle)->next;
	if (FinalizeMeQ == NULL) {
	    FinalizeMeQLast = NULL;
	}
	inBeingFinalized = self;
        FINALMEQ_UNLOCK();
        execute_finalizer(handle);
        FINALMEQ_LOCK();
	inBeingFinalized = NULL;
	BeingFinalized = NULL;		/* Feel free to GC that object now! */
	heap_memory_changes++;

    } while (FinalizeMeQ != NULL) ;

    sysAssert(inBeingFinalized == NULL && BeingFinalized == NULL);

    /* Only do after finalizing everyone to avoid too much notification: */
    FINALMEQ_NOTIFY();		/* Notify one waiter on BeingFinalized */
    FINALMEQ_UNLOCK();
    return TRUE;
}

/*
 * This is what System.runFinalization actually calls: it will quietly
 * make no progress if called recursively.
 */
void
runFinalization(void)
{
    (void) runFinalization0();
}

/*
 * Finalize all outstanding objects.  There are two steps.  First,
 * synchronously garbage collect and finalize to take care of everything
 * that we can in the normal way.  Second, transfer all remaining objects
 * on the HasFinalizerQ to the FinalizeMeQ and sychronously finalize them.
 * Repeat step two until there are no objects to be finalized.
 * 
 * This function exists to support finalization upon exit.  It should
 * should be run only once one is committed to kill the runtime normally,
 * probably once all other user threads are dead.  We still synchronize
 * to avoid conflict with daemon threads, or finalize methods that may
 * create additional finalizable objects.  Exiting the runtime abnormally
 * needs no run finalizeOnExit().
 *
 * Like prepareFinalization() in gc.c, we put the cluster of finalizer_t's
 * moved to the FinalizeMeQ as the result of one GC at the end of the 
 * FinalizeMeQ; that is, the clusters are stored in FIFO order.  The
 * code that transfers finalizer_t's from one list to another does not
 * assume that it starts with an empty FinalizeMeQ, although in its
 * current use that is probably the case.  That should make this code
 * robust in case anyone else (e.g. a system thread) gets to run after
 * runFinalization and before grabbing HASFINALQ_LOCK!
 */
void
finalizeOnExit(void)
{
    /*
     * Extra stack check to shrink STACK_REDZONE
     */
    if (!sysThreadCheckStack(STACK_REDZONE)) {
	/*
	 * Got in a trouble. No way to continue the finalization.
	 */
	return;
    }

    gc(0, 0);			/* No point forcing compaction here */
    runFinalization();

    HASFINALQ_LOCK();
    while (HasFinalizerQ != NULL) {
	JHandle *handle;
	JHandle *next;
	JHandle **last;
	JHandle *newLast = NULL;
	JHandle *newFirst = NULL;

	FINALMEQ_LOCK();
	next = HasFinalizerQ;
	HasFinalizerQ = NULL;

	if (FinalizeMeQLast == NULL) {
	    sysAssert(FinalizeMeQ == NULL);
	    last = &FinalizeMeQLast;
	} else {
	    sysAssert(FinalizeMeQ != NULL);
	    last = &(FINALIZER_T(FinalizeMeQLast)->next);
	}

	/* Traverse HasFinalizerQ, flipping finalizers onto FinalizeMeQ */
	while ((handle = next) != NULL) {
	    /* The first new finalizer becomes the new FinalizeMeQLast */
	    if (newLast == NULL) {
		newLast = handle;
	    }
	    newFirst = handle;
	    next = FINALIZER_T(handle)->next;
	    FINALIZER_T(handle)->next = *last;
	    *last = handle;
	}
	if (newLast != NULL) {
	    FinalizeMeQLast = newLast;
	    if (FinalizeMeQ == NULL) {
		FinalizeMeQ = newFirst;
	    }
	}

	FINALMEQ_UNLOCK();
	HASFINALQ_UNLOCK();
	runFinalization();
	HASFINALQ_LOCK();
    }
    HASFINALQ_UNLOCK();
}

/*
 * Note that the finalizer thread runs at low priority, which means it
 * may *never* run.  Also note that it does not lock the scheduler: that
 * would be troublesome given that the finalization method is random
 * user Java code.  Rather, we just allow ourselves to be preempted as
 * the system chooses.  This shouldn't get us into trouble, as the
 * objects we are working with are known not to be referenced by any-
 * thing except the FinalizeMeQ.
 */
static void
finalizer_loop(TID tid)
{
    ExecEnv ee;
    JHandle *handle;
    TID self;

    InitializeExecEnv(&ee, tid);
    if (ee.initial_stack == 0) {
	out_of_memory();
    }
    threadInit(tid, (stackp_t) &tid);

    self = threadSelf();
    FINALMEQ_LOCK();
    while (1) {
	/*
	 * We can't get here from runFinalization of this same thread,
	 * so don't need to worry about self-deadlock.  We can get from
	 * here to runFinalization so we need to set inBeingFinalized.
	 */

	/* Someone may be synchronously finalizing (unlikely) */
	while (BeingFinalized != NULL || FinalizeMeQ == NULL) {
	    FINALMEQ_NOTIFY();	/* Someone else cleared, others waiting? */
	    FINALMEQ_WAIT();	/* Async finalizer shouldn't be interruptable */
	}

	Log(2, "Finalization thread is running\n");

	/*
	 * Drop and reacquire the FINALQ lock after every object so that
	 * a more important thread won't have to wait for it until we've
 	 * finalized all the outstanding objects before it gets to run.
	 */
	sysAssert(inBeingFinalized == NULL && BeingFinalized == NULL);
	handle = BeingFinalized = FinalizeMeQ;
	FinalizeMeQ = FINALIZER_T(handle)->next;  /* Unlink */
	if (FinalizeMeQ == NULL) {
	    FinalizeMeQLast = NULL; /* Do this before unlocking */
	}
	inBeingFinalized = self;
	FINALMEQ_UNLOCK();
	execute_finalizer(handle);
	FINALMEQ_LOCK();
	inBeingFinalized = NULL;
	BeingFinalized = NULL;	    /* Feel free to GC that object now! */
	heap_memory_changes++;
	FINALMEQ_NOTIFY();	    /* Notify one waiter on BeingFinalized */
    }

    /* If we ever got here: FINALMEQ_UNLOCK(); */
}

/*
 * Create the finalization thread.
 *
 * Interaction between finalization and GC: There can only be finalization
 * methods to run after GC has occurred.  Once there are finalization
 * methods to call, there is no expectation that a given run of the final-
 * izer thread will clear them all.  The finalizer should also be prepared
 * to be interrupted between finalizing any two objects, like async GC.
 *
 * An object cannot be considered free for compaction until its finalization
 * method has been run.  This means that either we finalize an object prior
 * to compaction in the GC run that freed the object, or the object must
 * only be reclaimed in the next GC run.  This happens automatically if we
 * trace the FinalizeMeQ, which after GC should retain the sole reference
 * to the object.
 */
void
InitializeFinalizerThread()
{
    /* 
     * The finalization thread can run arbitrary code, so is given
     * a normal-sized stack.  This may be overkill, but is safe.
     */
    (void) createSystemThread("Finalizer thread", MinimumPriority,
	ProcStackSize, finalizer_loop);
}

#ifdef DEBUG

/*
 * Debugging support: It is safe to call these things inside GC because
 * although the scheduler is locked, we always precede GC by grabbing
 * the finalization queue locks.
 */

void
DumpHasFinalizerQ()
{
    JHandle *handle;
    char *name;

    HASFINALQ_LOCK();
    jio_fprintf(stderr, "\nLiving objects with finalization methods:\n");
    handle = HasFinalizerQ;
    while (handle != NULL) {
	name = Object2CString(handle);
	jio_fprintf(stderr, "    Has finalizer: %s (handle 0x%lx)\n",
		name, handle);
	handle = FINALIZER_T(handle)->next;
    }
    HASFINALQ_UNLOCK();
}

void
DumpFinalizeMeQ()
{
    JHandle *handle;
    char *name;

    FINALMEQ_LOCK();
    jio_fprintf(stderr, "\nDead objects pending finalization:\n");
    handle = FinalizeMeQ;
    while (handle != NULL) {
	name = Object2CString(handle);
	jio_fprintf(stderr, "    To be finalized: %s (handle 0x%lx)\n",
		name, handle);
	handle = FINALIZER_T(handle)->next;
    }
    FINALMEQ_UNLOCK();
}

#endif
