/*
 * @(#)gc.c	1.185 98/12/07
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Java memory management and garbage collection
 *
 * Implementation notes:
 *
 * We really don't need to allocate mark bits for soft references.
 * It might be better to keep a small list of the soft references we
 * find while doing the mark phase.  This would save mark bits and
 * shouldn't cost too much in the way of cpu.
 *
 * I [Tim] tried to split this file up into two: heap.c, covering heap
 * initialization and allocation routines, and gc.c, covering GC and heap
 * expansion.  Doing so changed the basic heap pointers from statics,
 * and the result (for the SunPro compiler, at least), was a 10% slowdown
 * of the garbage collector.  So now we've got a big file again.
 *
 * There are 2 interesting flags for debug print: verbosegc is good for
 * lightweight garbage collecting info, and is turned on by the menu
 * item "Verbose GC". tracegc is good for heavy-duty print of lots of
 * info that will take a long time.
 *
 * TO DO:
 *    Think about the policy issues more - how much memory to pre-allocate,
 *    when do we gc vs. when do we expand the heap (manageAllocFailure),
 *    when do we shrink the heap (evacuationChunk).  Also default handle
 *    space vs. object space split.
 *    Need to make this tunable for each implementation of the VM, and should
 *    allow user controls.
 *
 */

#include "typedefs.h"
#include "profile.h"
#include "tree.h"
#include "oobj.h"
#include "interpreter.h"
#include "typecodes.h"
#include "monitor.h"
#include "threads.h"
#include "signature.h"
#include "monitor_cache.h"
#include "finalize.h"
#include "jni.h"

#include "sys_api.h"

#include "java_lang_Thread.h"
#include "gc.h"

#ifdef BREAKPTS
#include "breakpoints.h"
#endif

#undef max
#define max(a, b)	((a) < (b) ? (b) : (a))

#define IS_STATIC_AND_REFERENCE(fb) \
   ((fieldaccess(fb) & (ACC_REFERENCE | ACC_STATIC)) == (ACC_REFERENCE | ACC_STATIC))
#define IS_NONSTATIC_AND_REFERENCE(fb) \
   ((fieldaccess(fb) & (ACC_REFERENCE | ACC_STATIC)) == (ACC_REFERENCE))


float preallocFactor = 0.25;	/* Should be user-modifiable */


#ifndef PAGED_HEAPS /************ CONTIGUOUS HEAPS: ********************/

/*
 * Heap layout, pointers and pointer aliases set by SetLimits():
 *
 * |----------------------------mapped memory---------------------------|
 *	    |------------------committed memory----------------|
 *
 *  -------------------------------------------------------------------
 * |   ... <|  handles	 |	    objects		       |> ...  |
 *  -------------------------------------------------------------------
 *  ^	     ^		^ ^				      ^		^
 *  |	     |		| |				      |		|
 *  |	    hpool	| hpoollimit		      opoollimit	|
 * heapbase  |		| opool				      |	    heaptop
 *	    hpmin   hpmax opmin				   opmax
 *
 * Constraints:
 *    hpool/hpmin and hpoollimit are HANDLEGRAIN aligned
 *    opool/opmin must be OBJECTGRAIN aligned *plus* sizeof(hdr)
 *    the committed memory must be contained by the mapped memory
 */

/* The extent of the object pool (must be committed memory) */
static unsigned char *opool, *opoollimit, *opoolhand;
/* The extent of the handle pool (must be committed memory) */
static unsigned char *hpool, *hpoollimit;
/* The free handle list (replaces hpoolhand, which was a hint) */
static unsigned char *hpoolfreelist;
/* The extent of the mapped memory */
static unsigned char *heapbase, *heaptop;

static long markmax;

#define LIMIT hpoollimit

#else 		/************ PAGED HEAPS: ********************/

/*
 * Part of Java memory management and garbage collection.
 *
 * This supports a discontiguous gcable heap, which is useful for the
 * Mac OS, or other platforms without good memory mapping support.
 *
 * Reworked memory mgmt of gc to be more Mac friendly. Instead of allocating
 * all memory in a giant glob, and subdividing it into handle and object space,
 * we now allocate separate handle chunks and object chunks, and manage these
 * as a linked list of large chunks. The chunks still have to be locked down
 * in the Mac's system memory pool, but at least we allocate them more
 * rationally.
 * One more tidbit: due to requirement of 8B alignment, we alloc an extra
 * handle (or object word) so we can adjust startPtr to start at desired
 * alignment.
 */

/* doubly-linked list of handle chunks, in address order: */
ChunkBlkP firstHanBlkP = NULL;
ChunkBlkP lastHanBlkP = NULL;
ChunkBlkP curHanBlkP = NULL;
/* doubly-linked list of object chunks, in address order: */
ChunkBlkP firstObjBlkP = NULL;
ChunkBlkP lastObjBlkP = NULL;
ChunkBlkP curObjBlkP = NULL;

/* PAGE MAP:
 * Every page of memory from the lowest chunk to the highest chunk given to
 * Java by the OS has an entry in the "page_map".  The page map records the
 * type of the page (handle/object/free [encoded in the chunk_size field]),
 * the page offset to the beginning of the chunk, and a pointer to the markbits
 * for the page.  Since pages are aligned, it is simple to determinethe
 * PageMapEntry for any random pointer, and therefore "ValidHandle",
 * "isMarked", etc.
 */
typedef struct {
    short page_number;	      /* Which page is this in the chunk */
    short chunk_size;	      /* # of pages in this chunk, encodes chunk type */
    unsigned int*  mark_bits; /* ptr to the mark bits for this page */
} PageMapEntry;

PageMapEntry *page_map = NULL;	/* This can (and must) be re-allocated when
				 * lower pages are allocated
				 */
void* mem_base;		/* lowest page we have allocated */
void* mem_top;		/* ptr past highest page we have allocated */

#define LIMIT mem_top

static unsigned char *markbitsfreeptr;

/* number of entries in map: */
#define PageMapLength() (((char *)mem_top - (char *)mem_base) / PAGE_ALIGNMENT)

#define HANDLE_HEAP_TYPE    -1
#define FREE_HEAP_TYPE	     0
#define OBJECT_HEAP_TYPE     1

long gMaxMem = 0; /* DRW: Not used - maximum bytes of object memory to alloc */

#endif /************ END PAGED HEAPS ********************/

/* prototypes */
int manageAllocFailure(long n, int overflow_type, int overflow_act);
extern bool_t runFinalization0(void);
unsigned char *gc0(int, unsigned int);
void validatePool(long quick);

static int deferredHandleTableOverflow;

#define MAX_GC_DEPTH 100
/* The most number of variables that can be pushed onto our stack from a
 * recursive call is 5, plus we push a state value.  We want to allow at least
 * the same amount of recursion as the old code, at a cheaper cost 
 */
#define MC_STACK_SIZE (MAX_GC_DEPTH * 6)
static void markChildren(JHandle *h, JHandle *limit, int depth);
static void scanJavaFrame(JavaFrame *frame, stack_item *top_top_stack);

/*
 * Share buffers for mc_stack_global and soft ref "bag".
 *   static void* mc_stack_global[MC_STACK_SIZE];
 *   static JHandle *bag_ref[SOFTREFBAGSIZE];
 *   static int64_t bag_priority[SOFTREFBAGSIZE];
 */
static union {
    void *mc_stack[MC_STACK_SIZE];
    struct {
        JHandle *ref[SOFTREFBAGSIZE];
        int64_t priority[SOFTREFBAGSIZE];
    } bag_arrays;
} gcSharedBuffer;

#ifdef DEBUG
void verifyObjectCtrs(void);
void verifyHandleCtrs(void);
void DumpHeapInfo(void);
#ifdef PAGED_HEAPS
void checkChunk(ChunkBlkP);
void verifyPageMap(void);
#endif /* PAGED_HEAPS */
#endif /* DEBUG */

#ifdef HPROF
#include "byteorder.h"
#include "vmprofiler.h"
static void jvmpi_gc_dump_heap(void);
static bool_t jvmpi_in_dump_heap_flag = FALSE;
#endif /* HPROF */


/*
 * Lock against heap modification.
 */
sys_mon_t *_heap_lock;

#ifdef HPROF

#define GC_DISABLE_LOCK(self) sysMonitorEnter(_gc_disable_lock)
#define GC_DISABLE_UNLOCK(self) sysMonitorExit(_gc_disable_lock)
 
/*
 * Micro lock for disabling GC */
static sys_mon_t *_gc_disable_lock;
static int gc_disable_count = 0;

#endif /* HPROF */

extern sys_mon_t *_hasfinalq_lock;
extern sys_mon_t *_finalmeq_lock;

static unsigned int *markbits;  /* global markbits pool */
#define BITSPERMARK (BITSPERCHAR * sizeof(*markbits))	/* 32 */
static long FreeObjectCtr;	/* # bytes of objects free */
static long TotalObjectCtr;	/* # bytes of objects space allocated so far */
static long FreeHandleCtr;	/* # bytes of free handles (not # of handles) */
static long TotalHandleCtr;	/* # bytes of handles allocated so far */
static long marksize;           /* # bytes for global markbits pool */
static long FreeMemoryLowWaterMark;
static long nfreed = 0;		/* # of blocks freed in last gc */
static long bytesfreed = 0;	/* # of bytes freed in last gc */
static int markbitsInUse = FALSE; /* true during gc0_locked */
int tracegc = 0;

static bool_t allowClassGCNow;

/* List of unused classes found during GC */
static ClassClass *classFreeList = 0;
#ifdef ALLOC_BARRIERS

static long GCGenCtr;  /* generation count; add one each GC */
static long InGenCtr;
static bool_t AllocNoGC;

#define T_PRRED     T_MKTYPE(T_INTEGER, 0)
#define T_PRYELLOW  T_MKTYPE(T_INTEGER, 1)

#ifndef DEFAULT_RESERVE_RED
#define DEFAULT_RESERVE_RED    (32 * 1024)
#endif
#ifndef DEFAULT_RESERVE_YELLOW
#define DEFAULT_RESERVE_YELLOW (128 * 1024)
#endif

#ifndef _FIXED_BARRIER_SIZES_
long _barriersize_red = DEFAULT_RESERVE_RED;
long _barriersize_yellow = DEFAULT_RESERVE_YELLOW;
#define BARRIERSIZE_RED		_barriersize_red
#define BARRIERSIZE_YELLOW	_barriersize_yellow
#else
#define BARRIERSIZE_RED		DEFAULT_RESERVE_RED
#define BARRIERSIZE_YELLOW	DEFAULT_RESERVE_YELLOW
#endif

#define concat(x,y)	x ## y
#define xconcat(x,y)	concat(x,y)
#define quote(s)	#s
#define xquote(s)	quote(s)
#define eval(x)		x

#define RuntimeClassPrefix	sun_misc_VM
#define HRuntime xconcat(H,RuntimeClassPrefix)

#include xquote(RuntimeClassPrefix.h)

static void tryLowerAllocState();

static HRuntime *rtobj = NULL;

static HObject *PRRed    = NULL;
static HObject *PRYellow = NULL;

#define ALLOCSTATE_GREEN	(xconcat(RuntimeClassPrefix,_STATE_GREEN))
#define ALLOCSTATE_YELLOW	(xconcat(RuntimeClassPrefix,_STATE_YELLOW))
#define ALLOCSTATE_RED		(xconcat(RuntimeClassPrefix,_STATE_RED))

static int allocstate;

#define setAllocState(state) {allocstate = state;}

void
xconcat(RuntimeClassPrefix,_registerWithGC)(struct HRuntime *this)
{
    rtobj = this;
}

long
xconcat(RuntimeClassPrefix,_getState)(struct HRuntime *this)
{
#ifdef ALLOC_BARRIERS
    if (allocstate != ALLOCSTATE_GREEN) {
        gc(0, (unsigned)~0);  /* Force compaction by setting high target */
    }
    return allocstate;
#else
    /* This is a little strange; we always need sun_misc_VM_getState */
    return ALLOCSTATE_GREEN;
#endif /* ALLOC_BARRIERS */
}

static int
updateAllocState()		/* Assumes heap locked against GC */
{
    int notify = 0;

    if (rtobj != 0) {
	Classsun_misc_VM *rt = (Classsun_misc_VM *)unhand(rtobj);
	unsigned int mon = obj_monitor(rt->memoryAdviceLock);
	monitorEnter(mon);
	if (rt->memoryAdvice != allocstate) {
	    rt->memoryAdvice = allocstate;
	    monitorNotifyAll(mon);
	    notify = 1;
	}
	monitorExit(mon);
    }
    return notify;
}

static int as_old;
static int lock_count = 0;

static void
heap_lock()
{
    HEAP_LOCK();
    ++lock_count;
    if (lock_count == 1) {
	as_old = allocstate;
	InGenCtr = GCGenCtr;
    }
}

static void
heap_unlock(int trylower)
{
    int yield = 0;
    if (lock_count == 1) {
	yield = updateAllocState();
	trylower = trylower && (as_old >= allocstate);
	if (trylower && allocstate != ALLOCSTATE_GREEN &&
	    GCGenCtr != InGenCtr)
	{
	    tryLowerAllocState();
	}
    }
    --lock_count;
    HEAP_UNLOCK();
    if (yield) {
	threadYield();
    }
}

#ifdef HPROF

void jvmpi_heap_lock(ExecEnv* ee) {
    HEAP_LOCK();
}

void jvmpi_heap_unlock(ExecEnv* ee) {
    HEAP_UNLOCK();
}

#endif /* HPROF */

#undef HEAP_LOCK
#undef HEAP_UNLOCK
#define HEAP_LOCK()		heap_lock()
#define HEAP_UNLOCK()		heap_unlock(1)
#define HEAP_UNLOCK1(trylower)	heap_unlock(trylower)

#endif /* ALLOC_BARRIERS */

static void markPrimitiveClasses(void);

/* [Sheng] The "deferred handle" trick ensures that the recursive
 * GC scanning routines don't overflow the native stack.
 *
 * - Once the depth bound is exceeded, we do not scan the handle,
 *   but enter it into the "deferred handle" table.
 *
 * - If there is a handle table overflow, we iterate once more
 *   through all the handles in scanHandles() to reconcile the
 *   inconsistencies. Note that we are guaranteed that make some
 *   porgress in each pass, therefore the process always
 *   terminates.
 *
 * It is rare to run into deep markChildren recursions.
 * (markChildren is already rail recursive, and the limit argument
 * effectively cuts off deep recursions.) It is even rarer to
 * overflow the deferred handle table.
 *
 * I've run the system with a stack limit of 30 and handle table
 * size of 100. I haven't seen any cases of reiteration in
 * scanHandles().
 *
 * I've stress tested the system with a stack depth limit of 4 and
 * handle table size of 3, and haven't seen any significant overhead.
 * (Although reiteration did occur in scanHandles().)
 */


struct deferred_handle_t {
  JHandle *handle;
  struct deferred_handle_t *next;
};

#define DEFERRED_HANDLE_TABLE_SIZE 100

static struct deferred_handle_t
deferredHandleTable[DEFERRED_HANDLE_TABLE_SIZE];

static struct deferred_handle_t *free_deferred_handle_list;
static struct deferred_handle_t *used_deferred_handle_list;

static void initializeDeferredHandleTable()
{
    int i;
    used_deferred_handle_list = 0;
    free_deferred_handle_list = deferredHandleTable;
    for (i=0; i<DEFERRED_HANDLE_TABLE_SIZE - 1; i++)
        deferredHandleTable[i].next = deferredHandleTable + i + 1;
    deferredHandleTable[i].next = 0;
}

static void
putDeferredHandle(JHandle *h)
{
    struct deferred_handle_t *deferred_handle;
    if (free_deferred_handle_list == 0) {
        deferredHandleTableOverflow = 1;
        return;
    }
    deferred_handle = free_deferred_handle_list;
    free_deferred_handle_list = free_deferred_handle_list->next;

    deferred_handle->handle = h;
    deferred_handle->next = used_deferred_handle_list;
    used_deferred_handle_list = deferred_handle;
}

static JHandle *
getDeferredHandle()
{
    struct deferred_handle_t *deferred_handle;
    if (used_deferred_handle_list == 0)
        return 0;
    deferred_handle = used_deferred_handle_list;
    used_deferred_handle_list = used_deferred_handle_list->next;

    deferred_handle->next = free_deferred_handle_list;
    free_deferred_handle_list = deferred_handle;
    return deferred_handle->handle;
}

typedef struct JNIWeakRef {
    /* handle must be the first field. Code in jni.c assumes this! */ 
    JHandle *handle;
    struct JNIWeakRef *next;
    struct JNIWeakRef *prev;
    void *magic;
} JNIWeakRef;

/* A dummy cell to simplify pointer handling */
static JNIWeakRef weakRefDummy = {0,&weakRefDummy,&weakRefDummy,0};
/* The (doublely-linked) list of weak references created by JNI. */
static JNIWeakRef *weakRefList = &weakRefDummy;

JHandle **
newJNIWeakRef(ExecEnv *ee, JHandle *obj)
{
    JNIWeakRef * ref = (JNIWeakRef *)sysMalloc(sizeof(JNIWeakRef));
    if (ref == NULL)
        return NULL;
    ref->magic = (void *)ref;
    ref->handle = obj;
    HEAP_LOCK();
    { 
        JNIWeakRef *prev = weakRefList->prev;
	prev->next = ref;
	ref->prev = prev;
	weakRefList->prev = ref;
	ref->next = weakRefList;
    }
    HEAP_UNLOCK();
    return &ref->handle;
}

bool_t 
deleteJNIWeakRef(ExecEnv *ee, JHandle **handleP)
{
    JNIWeakRef * ref = (JNIWeakRef *)handleP;
    /* Sanity check that we are indeed deleting a weak ref. */
    if (ref->magic != (void *)ref)
        return FALSE;
    ref->magic = 0; /* Catch repeated deletes. */
    HEAP_LOCK();
    ref->next->prev = ref->prev;
    ref->prev->next = ref->next;
    HEAP_UNLOCK();
    sysFree(ref);
    return TRUE;
}

static void 
clearJNIWeakRefs()
{
    SetLimits();
    JNIWeakRef *ref = weakRefList->next;
    while (ref != weakRefList) {
        if (ref->handle && !IsMarked(ref->handle)) {
	    ref->handle = 0;
	}
	ref = ref->next;
    }
}

/*
 * Rough counter of heap modification events: a lack of known permu-
 * tations of the heap is used to suppress async GC.  While this isn't
 * a perfect indicator of good times to GC, it is unlikely that there
 * will be many good times to GC when we aren't turning over memory.
 * Current incrementers of heap_memory_changes include realObjAlloc()
 * and execute_finalizer().
 */
int heap_memory_changes = 0;

#ifdef DEBUG
long allocBytes = 0;   /* sum of bytes requested of realObjAlloc */
long allocObjects = 0; /* sum of objects allocated */
int  allocBogus = 0;   /* true if overflow occurs */
#endif /* DEBUG */

int64_t
TotalObjectMemory(void)
{
    return int2ll(TotalObjectCtr);
}

int64_t
FreeObjectMemory(void)
{
    return int2ll(FreeObjectCtr);
}

int64_t
TotalHandleMemory(void)
{
    return int2ll(TotalHandleCtr);
}

int64_t
FreeHandleMemory(void)
{
    return int2ll(FreeHandleCtr);
}

static void
clearMarkBits(void)
{

    sysAssert(!markbitsInUse);

#ifdef PAGED_HEAPS
    {   /* clear the pinned flag of each object chunk: */
        ChunkBlkP blkP = firstObjBlkP;
	do {
#ifdef LDEBUG
	    checkChunk(blkP);
#endif /* LDEBUG */
	    blkP->chunkFlags = blkP->chunkFlags & ~CHUNK_PINNED;
	    blkP = blkP->nextPtr;
	} while (blkP != firstObjBlkP);
    }
#endif

    clearLocalMarkBits();
    /* clear global markbits pool: */
    if (markbits) {
	memset((char *)markbits, 0, marksize);
    }
}

/* Search one chunk between p and limit for a free block of size "n".
 * If block found, sets block's header, decrements various counters,
 * and returns ptr.  Otherwise returns NULL.
 */
static void *
searchChunk(register unsigned char *p,
	    register unsigned char *limit,
	    register long n)
{
    /* note that p points to headers, so is not OBJECTGRAIN aligned */
    while (p < limit) {	    		/* search for a free block */
	register unsigned char *next;
	register hdr h = obj_geth(p);
	register hdr h2;		/*excess free space in this block */
	sysAssert(h_len(h) >= sizeof(hdr));
	next = p + h_len(h);
	sysAssert(next <= limit + sizeof(ClassObject));
	if (h_free(h)) {
	    while (h_free(h2 = obj_geth(next))) {
	      	/*coalesce next free run (why here?) */
		h_bumplen(h, h_len(h2));
#ifdef DEBUG
		/* old size might have been too small for a marker, so check */
		if (h_len(h) > OBJECTGRAIN) {
		    ((long *) p)[CHECK_WORD_INDEX] = 0x55555555;
		}
#endif
		next = p + h_len(h);
		obj_seth(p, h);
	    }
	    h2 = h_len(h) - n;
	    if (h2 >= 0) {  /* found block at least large enuf */
		if (h2 > 0) {
		    obj_setlf(p + n, h2, 1);
		    sysAssert(h2 >= OBJECTGRAIN);
#ifdef DEBUG
		    if (h2 > OBJECTGRAIN) {
			((long *) (p + n))[CHECK_WORD_INDEX] = 0x55555555;
		    }
#endif /* DEBUG */
		}
		obj_setlf(p, n, 0);
#ifdef DEBUG
		if (n > OBJECTGRAIN) {
		    ((long *) p)[CHECK_WORD_INDEX] = 0x55555555;
		}
#endif /* DEBUG */
		sysAssert((((long) p + sizeof(hdr)) & (OBJECTGRAIN - 1)) == 0);
		FreeObjectCtr -= n;
		sysAssert(FreeObjectCtr >= 0);
		return p;
	    }
	}
	p = next;
    } /* while (p < limit) */
    return NULL;
}

#ifdef PAGED_HEAPS /************ PAGED HEAPS: ********************/
/*
 * Allocate N pages, which are aligned and sized "alignment".
 * AllocChunk may use the first few bytes of the page for it's
 * own purposes (the CHUNK_BLK header), and the last chunk may be
 * slightly smaller than a page.
 *
 * Returns a ptr to the chunk block header, which is page aligned.
 * The usable space starts at "startPtr", and is aligned using inner_alignment.
 * The useful size (allocSize) is also aligned using inner_alignment.
 * The PAGE_ALIGNMENT must itself be inner_alignment aligned.
 *
 * Use FreeChunk to free.
 * This does NOT link the chunk into a free list.
 */
static ChunkBlkP
AllocChunk(size_t pages, size_t inner_alignment)
{
    void* chunkHandle;
    unsigned char* alignedPtr = NULL;
    ChunkBlkP chunk = NULL;
    size_t reqSize;

    reqSize = (PAGE_ALIGNMENT * pages) - OS_BLOCK_OVERHEAD;
    alignedPtr = sysAllocBlock(reqSize, &chunkHandle);

    if (alignedPtr != NULL) {
	size_t alignedHeaderSize = (sizeof(ChunkBlk) + inner_alignment - 1)
	                            & ~(inner_alignment - 1);

	chunk = (ChunkBlkP)alignedPtr;
	/* if PAGE_ALIGNMENT is inner aligned, then alignedPtr will be also.
	 * if OS_BLOCK_OVERHEAD is inner aligned,
	 * then reqSize will be aligned, and therefore allocSize and
	 * endPtr will be also.
	 */
	sysAssert((PAGE_ALIGNMENT % inner_alignment) == 0);
	sysAssert((OS_BLOCK_OVERHEAD % inner_alignment) == 0);
#ifdef DEBUG
	chunk->physEndPtr = alignedPtr + reqSize;
	/* save another (at least) 8 bytes for overwrite detection: */
	reqSize -= (inner_alignment > 8) ? inner_alignment : 8;
	/* use two words for overwrite protection - the first is more likely
	 * to stomped by the gc, the second by the user */
	*((int32_t *)(chunk->physEndPtr - 8)) = ALMOST_WORD;
	*((int32_t *)(chunk->physEndPtr - 4)) = ULTIMATE_WORD;
#endif /* DEBUG */

	chunk->chunkHandle = chunkHandle;
	chunk->nextPtr = NULL;				/* caller must link */
	chunk->prevPtr = NULL;
	chunk->allocSize = reqSize - alignedHeaderSize; /* now aligned */
	chunk->startPtr = alignedPtr + alignedHeaderSize; /* now aligned */
	chunk->freePtr = chunk->startPtr;
	chunk->freeCnt = chunk->allocSize;
	chunk->endPtr = chunk->startPtr + chunk->allocSize; /* now aligned */

#ifdef WASTED_SPACE_IN_LEADER
	initWastedSpaceInChunk(chunk);
#ifdef TRACEGC
	if (tracegc) {
	    jio_fprintf(stderr, "<GC: allocated chunk, size %d, wasted %d>\n",
		    chunk->allocSize, (int)chunk - (int)chunk->wasteStartPtr);
	}
#endif /* TRACEGC */
#endif /* END NOT WASTED_SPACE_IN_LEADER */

#ifdef DEBUG
	checkChunk(chunk);
#endif /* DEBUG */
    }

    return chunk;
}

static void
FreeChunk(ChunkBlkP chunk)
{
#ifdef DEBUG
   checkChunk(chunk);
#endif /* DEBUG */
   sysFreeBlock(chunk->chunkHandle);
}

static HObject *
realObjAlloc(struct methodtable *mptr, long n0)
{
    unsigned char *p;
    register long n;
    HObject *ret;
    int overflow_type = OVERFLOW_NONE;
    int on_overflow = OVERFLOW_ACT_GC;
    ChunkBlkP blkP = curObjBlkP;      /* for looping check */
#ifdef ALLOC_BARRIERS
    bool_t retryaftercallback = TRUE;
Retry:
#endif /* ALLOC_BARRIERS */

    n = (n0 + sizeof(hdr) + (OBJECTGRAIN - 1)) & ~(OBJECTGRAIN - 1);
    HEAP_LOCK();
    while (1) {
	/* make sure current object chunk has free space, else look in others.
	 * If none, alloc another object chunk. */
	while (curObjBlkP->freeCnt < n) {
	    if ((curObjBlkP = curObjBlkP->nextPtr) == blkP) {
	    	/* we looped thru all chunks, but no free space */
		overflow_type = OVERFLOW_OBJECTS;
		goto Overflow;
	    }
	}

	/* found a possible chunk */
	if ((p = searchChunk(curObjBlkP->freePtr, curObjBlkP->endPtr, n))
	     != NULL) {
	    goto GotIt;
	} else if ((curObjBlkP->freePtr != curObjBlkP->startPtr) &&
		    /* search other section of the chunk: */
		   ((p = searchChunk(curObjBlkP->startPtr,
		                     curObjBlkP->freePtr, n)) != NULL)) {
	    goto GotIt;
	} else {
	    if ((curObjBlkP = curObjBlkP->nextPtr) != blkP) {
		continue;   /* "while (1)" loop */
	    }
	}
#ifdef DEBUG
	if (verbosegc) {
	    jio_fprintf(stderr, "<GC: out of object space wanting %d bytes>\n", n);
	}
#endif /* DEBUG */
	overflow_type = OVERFLOW_OBJECTS;

Overflow:
#ifdef ALLOC_BARRIERS
        if (AllocNoGC) {
            AllocNoGC = FALSE; /* paranoia */
            HEAP_UNLOCK1(FALSE);
            return 0;
        }
#endif /* ALLOC_BARRIERS */
	/* Allocation failed for lack of space: garbage collect or expand */
	on_overflow = manageAllocFailure(n, overflow_type, on_overflow);
	if (on_overflow == OVERFLOW_ACT_FAIL) {
	    /* Tried everything, couldn't satisfy request, so fail */
	    ret = 0;
	    goto GetOut;
	} else {
	    /* back from GC - try alloc again.
	     * reset anchor point for freespace search */
	    blkP = curObjBlkP;
	}
    } /* while (1) */

GotIt:				/* p points to new object */
#ifdef LDEBUG
    checkChunk(curObjBlkP);
#endif
    curObjBlkP->freeCnt -= n;
    curObjBlkP->freePtr = (unsigned char *) ((long) p + n);
    sysAssert(curObjBlkP->freeCnt >= 0);
    /*
     * Reset overflow act for handles.  If there was also object overflow,
     * we will have escalated the overflow_act.  If we got as far as
     * expanding the object space, overflow_act == OVERFLOW_ACT_DESPERATE,
     * meaning that we'll fail the handle allocation if a handle isn't
     * immediately available.  If this is the case we will have already
     * tried GCing, finalizing and freeing Refs, so it only makes sense to
     * back off the overflow act to OVERFLOW_ACT_EXPAND.  Otherwise we
     * can leave the overflow act as it is.
     */
    if (overflow_type == OVERFLOW_OBJECTS &&
        on_overflow > OVERFLOW_ACT_EXPAND) {
        on_overflow = OVERFLOW_ACT_EXPAND;
    }
    ret = AllocHandle(mptr, (ClassObject *) (p + sizeof(hdr)));
    if (ret == 0) {
	obj_setfree(p);
	FreeObjectCtr += obj_len(p);
	curObjBlkP->freeCnt += obj_len(p);
	sysAssert(FreeObjectCtr <= TotalObjectCtr);
	/* on_overflow is OVERFLOW_ACT_GC */
	overflow_type = OVERFLOW_HANDLES;
	/* jump into the middle of the "while (1)" loop */
	goto Overflow;
    }

    memset(p + sizeof(hdr), 0, n0);

    heap_memory_changes++;
GetOut:
#ifdef HPROF
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_OBJ_ALLOC)) {
        jvmpi_alloc_object(EE(), ret);
    }
#endif /* HPROF */
#ifdef ALLOC_BARRIERS
    /* upwards transition on failing request; so remain in higher
       state for callback, then retry allocation, assuming
       callback will free some space */
    HEAP_UNLOCK1(!(ret == 0 && retryaftercallback));
    if ((ret == 0) && retryaftercallback) {
	/* upwards transition on failing request; so now retry
           allocation, assuming callback freed some space */
	retryaftercallback = FALSE;
	goto Retry;
    }
#else
    HEAP_UNLOCK();
#endif /* ALLOC_BARRIERS */

#ifdef DEBUG
    if (verbosegc) {
	long oldBytes = allocBytes;
	allocObjects++;
	allocBytes += n;
	if ((allocBytes < oldBytes)) {
	    allocBogus = 1;
	}
    }
#endif

    return ret;
}

#else  /************ CONTIGUOUS HEAPS: ********************/

static HObject *
realObjAlloc(struct methodtable *mptr, long n0)
{
    unsigned char *p;
    long n;
    HObject *ret;
    int overflow_type = OVERFLOW_NONE;
    int on_overflow = OVERFLOW_ACT_GC;
#ifdef ALLOC_BARRIERS
    bool_t retryaftercallback = TRUE;
Retry:
#endif /* ALLOC_BARRIERS */

    n = (n0 + sizeof(hdr) + (OBJECTGRAIN - 1)) & ~(OBJECTGRAIN - 1);
    HEAP_LOCK();
    while (1) {
        /* Found a possible chunk */
        if ((p = searchChunk(opoolhand, opoollimit, n)) != NULL) {
	    goto GotIt;
	} else if ((opoolhand > opool) &&
		   /* Search other section of the chunk */
		   ((p = searchChunk(opool, opoolhand, n)) != NULL)) {
	    goto GotIt;
	} else {   /* We failed to satisfy the allocation request */
	    /*
	     * It is possible that due to coalescing of free memory, and the
	     * fact that we only reset opoolhand if we actually find the space
	     * we are looking for, that after searching the section of the
	     * chunk before opoolhand and still failing to find space we
	     * could leave opoolhand pointing into the interior of a free
	     * block.  Normally we will just GC in this case, which resets
	     * opoolhand, but under some circumstances, like AllocNoGC being
	     * true, this might not happen.  So we conservatively reset
	     * opoolhand to the base of the object space just in case.  Finer
	     * tuning is possible but inconvenient.
	     */
	    opoolhand = opool;
	}

#ifdef DEBUG
	if (verbosegc) {
	    jio_fprintf(stderr, "<GC: out of object space wanting %d bytes>\n", n);
	}
#endif /* DEBUG */
	overflow_type = OVERFLOW_OBJECTS;

Overflow:
#ifdef ALLOC_BARRIERS
        if (AllocNoGC) {
            AllocNoGC = FALSE; /* paranoia */
            HEAP_UNLOCK1(FALSE);
            return 0;
        }
#endif /* ALLOC_BARRIERS */
	/* Allocation failed for lack of space: garbage collect or expand */
	on_overflow = manageAllocFailure(n, overflow_type, on_overflow);
	if (on_overflow == OVERFLOW_ACT_FAIL) {
	    /* Tried everything, couldn't satisfy request, so fail */
	    ret = 0;
	    goto GetOut;
	} /* opoolhand has been reset by GC */
    } /* end while (1). */

GotIt:				/* p points to new object */
    opoolhand = (unsigned char *) ((long) p + n);
    /*
     * Reset overflow act for handles.  If there was also object overflow,
     * we will have escalated the overflow_act.  If we got as far as
     * expanding the object space, overflow_act == OVERFLOW_ACT_DESPERATE,
     * meaning that we'll fail the handle allocation if a handle isn't
     * immediately available.  If this is the case we will have already
     * tried GCing, finalizing and freeing Refs, so it only makes sense
     * to back off the overflow act to OVERFLOW_ACT_EXPAND.  Otherwise
     * we can leave the overflow act as it is.
     */
    if (overflow_type == OVERFLOW_OBJECTS &&
        on_overflow > OVERFLOW_ACT_EXPAND) {
        on_overflow = OVERFLOW_ACT_EXPAND;
    }
    ret = AllocHandle(mptr, (ClassObject *) (p + sizeof(hdr)));
    if (ret == 0) {
        obj_setfree(p);
	FreeObjectCtr += obj_len(p);
	sysAssert(FreeObjectCtr <= TotalObjectCtr);
	/* on_overflow is OVERFLOW_ACT_GC */
	overflow_type = OVERFLOW_HANDLES;
	goto Overflow;
    }

    memset(p + sizeof(hdr), 0, n0);

    heap_memory_changes++;
GetOut:
#ifdef HPROF
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_OBJ_ALLOC)) {
        jvmpi_alloc_object(EE(), ret);
    }
#endif /* HPROF */
#ifdef ALLOC_BARRIERS
    /* upwards transition on failing request; so remain in higher
       state for callback, then retry allocation, assuming
       callback will free some space */
    HEAP_UNLOCK1(!(ret == 0 && retryaftercallback));
    if ((ret == 0) && retryaftercallback) {
	/* upwards transition on failing request; so now retry
           allocation, assuming callback freed some space */
	retryaftercallback = FALSE;
	goto Retry;
    }
#else
    HEAP_UNLOCK();
#endif /* ALLOC_BARRIERS */

#ifdef DEBUG
    if (verbosegc) {
	long oldBytes = allocBytes;
	allocObjects++;
	allocBytes += n;
	if ((allocBytes < oldBytes)) {
	    allocBogus = 1;
	}
    }
#endif

    return ret;
}

#endif  /************ END CONTIGUOUS HEAPS ********************/

/*
 * Free handles are linked together by the "methods" field.  There is a
 * seperate list for each handle chunk, pointed to by the chunk's "freePtr"
 * field.  The contiguous heap uses hpoolfreeptr as the head of the free list.
 */
HObject *
AllocHandle(struct methodtable *mptr, ClassObject *p)
{
    JHandle *hanP;
    SetLimits();

    if (p == 0) {
	return 0;
    }

#ifdef PAGED_HEAPS
    {
        ChunkBlkP blkP = curHanBlkP;
        /* make sure current handle chunk has free space, else look in others.
         * If none, fail. */
        while (curHanBlkP->freePtr == 0) {
  	    curHanBlkP = curHanBlkP->nextPtr;
	    if (curHanBlkP == blkP) {
	        goto NoLuck;
	    }
        }
    }
#endif /* PAGED_HEAPS */

    POP_FREE_HANDLE(hanP);
    if (hanP == NULL) {
        goto NoLuck;
    }

    /* p should always be valid, except that
     * Locked_InitializeClass calls us with a fake (non-java heap) object!
     */
    sysAssert(ValidObject(p) ||
	      (mptr == mkatype(T_BYTE, 1)));
    /* should assert heap is locked */
    sysAssert(unhand(hanP) == 0);		    /* make sure it's really free */
    sysAssert(ValidHandle(hanP));

#ifdef PAGED_HEAPS
    curHanBlkP->freeCnt -= sizeof(JHandle); /* update counters...*/
    sysAssert(curHanBlkP->freeCnt >= 0);
    /* these should run out at the same time: */
    sysAssert((curHanBlkP->freeCnt == 0) == (curHanBlkP->freePtr == 0));
#ifdef LDEBUG
    checkChunk(curHanBlkP);
#endif
#endif /* PAGED_HEAPS */
    obj_methodtable(hanP) = mptr;		    /* setup new handle... */
    unhand(hanP) = (ClassObject *) p;
    FreeHandleCtr -= sizeof(JHandle);
    sysAssert(FreeHandleCtr >= 0);
    return (HObject *) hanP;

NoLuck:
#ifdef DEBUG
    if (verbosegc) {
	jio_fprintf(stderr, "<GC: out of handle space>\n");
    }
#endif /* DEBUG */
    return 0;
}


/*
 * Calculate the size in bytes of an array of type t and length l.
 */
int
sizearray(int t, int l)
{
    int size = 0;

    switch(t){
    case T_CLASS:
        size = sizeof(OBJECT);
	break;
    default:
        size = T_ELEMENT_SIZE(t);
	break;
    }
    size *= l;
    return size;
}

ClassClass *
allocClassClass()
{
    return (ClassClass *)
	realObjAlloc(javaLangClassMethods, sizeof(Classjava_lang_Class));
}

/*
 * Allocate an array of type 't' with initial size 'l'.	 This routine
 * takes care of setting the method table pointer correctly.  It also
 * takes into consideration any special sizeing requirements, like
 * those needed for arrays of classes.
 */
HObject *
ArrayAlloc(int t, int l)
{
    sysAssert(t >= T_CLASS && t < T_MAXNUMERIC);
    /*
     * Check whether there is in principle enough memory to satisfy the
     * allocation.  We want to fail this early because the array size
     * calculation might otherwise overflow, which would pass a garbage
     * size to realObjAlloc().  Note that a T_CLASS array allocates
     * an extra OBJECT; hence the l - 1 which otherwise makes the test
     * one element conservative.  Note that l may be 0!
     *
     * This is even more bogus for PAGED_HEAPS, but we'll still try
     * to catch that overflow error.
     */
#ifdef PAGED_HEAPS
    if (l && ((l - 1) > (0x7FFFFFFF /
		    (t == T_CLASS ? sizeof(OBJECT) : T_ELEMENT_SIZE(t))))) {
 	return 0;
    }
#else
    if (l && ((l - 1) > ((heaptop - heapbase) /
		   (t == T_CLASS ? sizeof(OBJECT) : T_ELEMENT_SIZE(t))))) {
 	return 0;
    }
#endif

    if (t != T_CLASS) {
	int array_type = arrayinfo[t].array_type;
 	return realObjAlloc((struct methodtable *) mkatype(array_type, l),
	                    sizearray(t, l));
    } else {
	return realObjAlloc((struct methodtable *) mkatype(OBJ_FLAG_CLASS_ARRAY, l),
			    sizearray(t, l) + sizeof(OBJECT));
    }
}

/*
 * Allocate an object.	This routine is in flux.  For now, the second
 * parameter should always be zero and the first must alwaysd point to
 * a valid classblock.	This routine should not be used to allocate
 * arrays; use ArrayAlloc for that.
 */
HObject *
ObjAlloc(ClassClass *cb, long n0)
{
    HObject *handle;

#ifdef DEBUG
    if (n0 != 0 || cb == 0) {
	sysAbort();
    }
#endif /* DEBUG */

    /*
     * If the class has a finalization method, we try to allocate extra
     * space for a finalization_t structure which will be placed at the
     * end of the instance.
     */
    n0 = cbInstanceSize(cb) + (cbHasFinalizer(cb) ? sizeof(finalizer_t) : 0);
    handle = realObjAlloc(cbMethodTable(cb), n0);

    /*
     * If the class is a normal class and has a finalization method, flag
     * the instance and prepend it to the head of the HasFinalizerQ.  Note
     * that handle can be nil if we are out of memory!
     */
    if (handle && obj_isnt_array(handle) && cbHasFinalizer(cb)) {
	HASFINALQ_LOCK();
	FINALIZER_T(handle)->next = HasFinalizerQ;
	HasFinalizerQ = handle;
	HASFINALQ_UNLOCK();
    }

    return handle;
}

/*
 * Search the handle table for instances of class Ref, and clear the
 * object pointer in the lowest priority group.	 If async_call is true,
 * we are being called asynchronously, and should watch for pending
 * interrupts.
 *
 * Return the number of refs zeroed, and set the int argument pointer
 * to the number of refs found in the system.
 * NOTE: This assumes that we have GC'd first, and mark bits are still valid.
 */
static int
clearRefPointers(int async_call, int free_space_goal, int *totalrefs)
{
    int nsoftrefs = 0;
    /*
     * To shrink stack requirement for each thread, make the bag static.
     * Note no need to be reentrant. Malloc wouldn't be good for this 
     * case because of malloc failure.
     */
    /* Use shared buffer to save memory */
    /* JHandle *bag_ref[SOFTREFBAGSIZE]; */
    /* int64_t bag_priority[SOFTREFBAGSIZE]; */
    int64_t* const bag_priority = gcSharedBuffer.bag_arrays.priority;
    JHandle** const bag_ref = gcSharedBuffer.bag_arrays.ref;
    JHandle *hp;
    int ninbag = 0;
    int freed = 0;
    SetLimits();

    MAP_OVER_HANDLES_FROM_START(hp)
	if (ValidObject(unhand(hp))
	    && obj_isnt_array(hp)
	    && CCIs(obj_classblock(hp), SoftRef)
	    && ValidHandle(obj_getslot(hp, 0))
	    && IsMarked(obj_getslot(hp, 0)) == SoftMark) {
	    Java8 t;
	    int64_t prio = GET_INT64(t, &obj_getslot(hp, 1));
	    if (ninbag >= SOFTREFBAGSIZE
		&& ll_lt(prio, bag_priority[SOFTREFBAGSIZE - 1])) {
		ninbag--;
	    }
	    if (ninbag < SOFTREFBAGSIZE) {
		int i = ninbag;
		while (i > 0 && ll_lt(prio, bag_priority[i - 1])) {
		    bag_ref[i] = bag_ref[i - 1];
		    bag_priority[i] = bag_priority[i - 1];
		    i--;
		}
		bag_ref[i] = hp;
		SET_INT64(t, &bag_priority[i], prio);
		ninbag++;
	    }
	    nsoftrefs++;
	}
    END_MAP_OVER_HANDLES_FROM_START;

    if (async_call && sysInterruptsPending()) {
	return 0;
    }

    if (ninbag) {
	int i;
	int freemem;

	/* Don't free more than half the Refs when there are more than 10 */
	if ((i = (nsoftrefs + 1) >> 1) <= ninbag && ninbag > 10) {
	    ninbag = i;
	}

	freemem = FreeObjectCtr;
	/* Free the least recently used Refs first (see bug 1250395) */
	for (i = 0; i < ninbag; i++) {
#ifdef LDEBUG
	    sysAssert(ValidHandle(bag_ref[i]));
#endif /* LDEBUG */
	    /*
	     * Only free as many refs as needed to meet free_space_goal.
	     * Of course, this assumes that there are no non-ref references
	     * to the freed objects, which needs not be the case, so we
	     * may end up smashing refs that won't get us any space back.
	     * This all needs to be revisited.
	     */
	    hp = bag_ref[i];
	    freemem += obj_len((char *)
                unhand((JHandle*) obj_getslot(hp, 0)) - sizeof(hdr));
#ifdef TRACEGC
	    if (tracegc > 1) {
		jio_fprintf(stderr, "clear soft ref: 0x%x: %s\n",
			(unsigned int) hp, Object2CString(hp));
	    }
#endif /* TRACEGC */
	    obj_setslot(hp, 0, 0);	/* Clear the Ref */
	    freed++;
	    /* This is a weak check as it doesn't require contiguous memory */
	    if (freemem >= free_space_goal) {
		break;
	    }
	}
    }

    *totalrefs = nsoftrefs;
    return freed;		/* Return the number of refs freed */
}

/* Called for expandHandleSpace - zero obj fields, and link methods fields
 * into a free list. */
static void
linkNewHandles(JHandle *start, JHandle *limit, unsigned char *oldFreeList)
{
    register JHandle *hp = start;
    register JHandle *hanLimit = limit;

    while (1) {
	register JHandle *nexthp = hp + 1;
	unhand(hp) = 0;				/* set each handle free */
	if (nexthp == hanLimit) {
	   obj_methodtable(hp) = (struct methodtable *)oldFreeList;
	   break;
	} else {
	    /* link free handles via methods field */
	    obj_methodtable(hp) = (struct methodtable *)nexthp;
	    hp = nexthp;
	}
    } /* while */
}

#ifdef PAGED_HEAPS /************ PAGED HEAPS: ********************/

/* When we get a new chunk outside of the range mapped by the page_map,
 * then we need to allocate a new page map, copy the old one into the
 * right place, set globals, etc. Returns true or false.
 */
static int
reallocPageMap(ChunkBlkP blkP)
{
    if (!page_map) { /* initial case */
	size_t map_len;

	mem_base = blkP;
	mem_top = (void *)(((int)blkP->endPtr + PAGE_ALIGNMENT - 1)
			   & ~(PAGE_ALIGNMENT - 1));

	map_len = PageMapLength() * sizeof(PageMapEntry);
	page_map = sysMalloc(map_len);
	if (!page_map) {
	    goto FAIL;
	}

#ifdef TRACEGC
	if (tracegc) {
	    jio_fprintf(stderr, "<GC: Initialized page map to hold %d pages>\n",
		    PageMapLength());
	}
#endif

	/* clear new entries */
	memset(page_map, 0, map_len);
    } else {
	void* old_mem_base = mem_base;
	size_t old_map_len = PageMapLength();
	void* old_page_map = page_map;
	/* offset in pages of where old entries will start in new page map. */
	size_t page_offset;

	if ((void *)blkP < mem_base) {
	    mem_base = (unsigned char *)blkP;
	} else if ((void *)blkP >= mem_top) {
	    mem_top = (void *)(((int)blkP->endPtr + PAGE_ALIGNMENT - 1)
			       & ~(PAGE_ALIGNMENT - 1));
	}

	page_map = sysMalloc(PageMapLength() * sizeof(PageMapEntry));
	if (!page_map) {
	    goto FAIL;
	}

	/* note that old_mem_base must be >= mem_base */
	page_offset = ((int)old_mem_base - (int)mem_base) / PAGE_ALIGNMENT;

#ifdef TRACEGC
	if (tracegc) {
	    jio_fprintf(stderr,
	    "<GC: New page map starts %d pages earlier than the old one>\n",
		    page_offset);
	}
#endif

	/* clear all entries */
	memset(page_map, 0, PageMapLength() * sizeof(PageMapEntry));
	/* copy old entries */
	memmove((unsigned char *)page_map +
	         (page_offset * sizeof(PageMapEntry)),
		old_page_map,
		old_map_len * sizeof(PageMapEntry));
	sysFree(old_page_map);
    }
    return 1;
FAIL:
    if (verbosegc) {
	jio_fprintf(stderr, "<GC: Couldn't allocate page map>\n");
    }
    return 0;

}

/* Allocate mark bits for a page. Either allocate into
 * the chunk, or into our global markbits pool.
 */
static void*
allocMarkBits(ChunkBlkP blkP)
{
    void* result = NULL;

    result = allocMarkBitsLocally(blkP);
    if (result != NULL) {
#ifdef TRACEGC
	if (tracegc) {
	    jio_fprintf(stderr, "<GC: Page's mark bits allocated within chunk>\n");
	}
#endif
        return result;
    }

#ifdef LDEBUG
    sysAssert(page_map);
    sysAssert(!markbitsInUse);
#endif
    if ((markbitsfreeptr + MARK_BITS_SIZE) >
        ((unsigned char *)markbits + marksize)) {
	/* allocate new mark bits pool (this catches the intial case too ): */
	/* # of full pages used by chunk: */
	int numPages = (blkP->allocSize / PAGE_ALIGNMENT) + 1;
	void* oldPool = (char *)markbits;
	void* oldPoolEnd = markbitsfreeptr;
	int i;
	int limit = PageMapLength();

	/* mark bits only have to survive the extent of a gc, so toss old:*/
	if (markbits) {
	    sysFree((char *)markbits);
	}

	/* DRW REMIND: Look at this */
	/* # of new pages we want to handle (leave growing room) */
	numPages = max(numPages, MIN_OBJ_PAGES) * 2;
	marksize += (MARK_BITS_SIZE * numPages);
	markbits = (unsigned int*) sysMalloc(marksize);
	if (markbits == NULL) {
	    if (verbosegc) {
		jio_fprintf(stderr, "<GC: Couldn't allocate new mark bits pool>\n");
	    }
	    return 0;
	}
#ifdef TRACEGC
	if (tracegc) {
	    jio_fprintf(stderr, "<GC: Growing global mark bits pool to %d bytes>\n",
	            marksize);
	}
#endif
	markbitsfreeptr = (unsigned char *)markbits;

	/* Walk over page_map, and update the mark_bits fields
	 * if they used to come from the old pool. */
	if (oldPool) {
	    for (i = 0; i < limit; i++) {
		void* oldMarkBitsPtr = (void *) page_map[i].mark_bits;
		if ((oldMarkBitsPtr >= oldPool) &&
		    (oldMarkBitsPtr < oldPoolEnd)) {
		    page_map[i].mark_bits = (unsigned int*)markbitsfreeptr;
		    markbitsfreeptr += MARK_BITS_SIZE;
		}
	    }
	}
    }

#ifdef TRACEGC
    if (tracegc) {
	jio_fprintf(stderr, "<GC: Using global mark bits pool>\n");
    }
#endif
    result = markbitsfreeptr;
    markbitsfreeptr += MARK_BITS_SIZE;

    sysAssert(markbitsfreeptr <
              ((unsigned char *)markbits + marksize + MARK_BITS_SIZE));
    return result;
}

/* Set up entries in the page map for new chunk, and allocate mark bits.
 * Returns true or false. If heapType == FREE_HEAP_TYPE, it will clear
 * map entries for the chunk.
 */
static int
setPageMap(ChunkBlkP blkP, int heapType)
{
    int numPages = (blkP->allocSize >> PTR_2_PAGE_SHIFT) + 1;
    int firstMapIndex;
    int i;

    if (((void *)blkP < mem_base) || ((void *)blkP >= mem_top)) {
	if (!reallocPageMap(blkP)) {
	    return 0;
	}
    }

    firstMapIndex = ((int)blkP - (int)mem_base) >> PTR_2_PAGE_SHIFT;
    for (i = 0; i < numPages; i++) {
	int mapIndex = firstMapIndex + i;

	if (heapType == FREE_HEAP_TYPE) {
	    page_map[mapIndex].page_number = 0;
	    page_map[mapIndex].mark_bits = 0;
	    page_map[mapIndex].chunk_size = 0;
	} else {
	    void* markBits = allocMarkBits(blkP);

	    if (markBits == NULL) {
		return 0;
	    }

	    sysAssert(!page_map[mapIndex].mark_bits);
	    page_map[mapIndex].page_number = i;
	    page_map[mapIndex].mark_bits = markBits;
	    if (heapType == HANDLE_HEAP_TYPE) {
		page_map[mapIndex].chunk_size = -1;
	    } else {
		page_map[mapIndex].chunk_size = numPages;
	    }
	}
    }
    return 1;
}

/* insert new chunk into a list in order (currently address order).
 * May modify chunkPtr pointers.
 */
static void
insertChunk(ChunkBlkP* first, ChunkBlkP* last, ChunkBlkP chunk)
{
    if (!(*first)) {		/* first & last always defined together */
	(*first) = chunk;
	(*last) = chunk;
	chunk->nextPtr = chunk;	/* chain to itself */
	chunk->prevPtr = chunk;
    } else {
	ChunkBlkP chunksNextP = 0;
	/* when we find correct location, chunksNextP will will be set */
	sysAssert((*first)->prevPtr == (*last));
	sysAssert((*last)->nextPtr == (*first));
	for (chunksNextP = (*first);
	     /* - */;
	     chunksNextP = chunksNextP->nextPtr) {
	    if (chunksNextP > chunk) {		/* found link */
		if (chunksNextP == (*first)) {
		    (*first) = chunk;
		}
		break;
	    } else if (chunksNextP == (*last)) { /* end of list */
		chunksNextP = chunksNextP->nextPtr; /* should eql (*first) */
		(*last) = chunk;
		break;
	    }
	}
	chunk->nextPtr = chunksNextP;		/* setup chunk */
	chunk->prevPtr = chunksNextP->prevPtr;
	chunksNextP->prevPtr = chunk;		/* link in chunk */
	chunk->prevPtr->nextPtr = chunk;
    }
    sysAssert((*first)->prevPtr == (*last));
    sysAssert((*last)->nextPtr == (*first));
#ifdef DEBUG
    verifyPageMap(); /* callers add to page map first, so now in synch */
#endif
}

static void
removeChunk(ChunkBlkP* first, ChunkBlkP* last,
            ChunkBlkP* current, ChunkBlkP chunk)
{
#ifdef LDEBUG
    checkChunk(chunk);
    checkChunk((*first));
    checkChunk((*last));
    checkChunk((*current));
#endif

    /* if only one chunk, this doesn't change anything: */
    chunk->prevPtr->nextPtr = chunk->nextPtr;
    chunk->nextPtr->prevPtr = chunk->prevPtr;	/* "" */

    chunk->prevPtr = NULL;
    chunk->nextPtr = NULL;

    if ((*first) == chunk) {		/* update globals */
	(*first) = (*last)->nextPtr;
    }
    if ((*last) == chunk) {
	(*last) = (*first)->prevPtr;
    }
    if ((*current) == chunk) {
	(*current) = (*first);		/* which one should we pick??? */
    }

#ifdef DEBUG
    if (*first) {
	sysAssert((*first)->prevPtr == (*last));
    }
    if ((*last)) {
	sysAssert((*last)->nextPtr == (*first));
    }
#endif
    sysAssert(((*first) == NULL) == ((*last) == NULL));
    sysAssert(((*first) == NULL) == ((*current) == NULL));
}

static void
shrinkObjectSpace(ChunkBlkP chunk)
{
    if (verbosegc) {
	jio_fprintf(stderr, "<GC: Freeing object chunk 0x%lx>\n", chunk);
    }

#ifdef LDEBUG
    sysAssert(chunk->allocSize == chunk->freeCnt);
#endif /* LDEBUG */
    FreeObjectCtr -= chunk->allocSize;
    TotalObjectCtr -= chunk->allocSize;
    removeChunk(&firstObjBlkP, &lastObjBlkP, &curObjBlkP, chunk);
    setPageMap(chunk, FREE_HEAP_TYPE);
    FreeChunk(chunk);
}

/*
 * Expand the handle space.  The behavior of the handle space is simpler
 * than that of the object space: handles never move and the space can't
 * be fragmented, but that also means that contracting the handle space
 * is easy to block.
 *
 * Be sure to zero all the new handles: a free handle is recognized by
 * having its obj field zeroed.	 Remember that the base of the handle
 * space must be HANDLEGRAIN aligned within the committed memory.
 *
 * Expand the handle space by allocating several pages in one chunk.
 * Allocate markbits as needed for the chunk.  Sets curHanBlkP on exit.
 *
 */
int
expandHandleSpace(int needPages)
{
    ChunkBlkP blkP = NULL;
    int numPages;

    /* sysAssert(HEAP_LOCKED()); ...except during InitializeAlloc */

    for (numPages = max(needPages, MIN_HANDLE_PAGES);
	 numPages >= needPages;
	 numPages--) {
        if (((numPages * PAGE_ALIGNMENT) <=
	     (gMaxMem - (TotalObjectCtr + TotalHandleCtr))) &&
	    ((blkP = AllocChunk(numPages, HANDLEGRAIN)) !=
	     NULL)) {
	    break;
	}
    }

    if (blkP) {
	int szBlk = blkP->allocSize;

	if (!setPageMap(blkP, HANDLE_HEAP_TYPE)) {
	    return 0;
	}

	if (verbosegc) {
	    jio_fprintf(stderr,
	          "<GC: expanded handle space by 0x%lx bytes, chunk = 0x%lx>\n",
	           szBlk, blkP);
	}

	insertChunk(&firstHanBlkP, &lastHanBlkP, blkP);
	curHanBlkP = blkP;		      /* mark this block currrent */

	linkNewHandles((JHandle *)blkP->startPtr, (JHandle *)blkP->endPtr,
		       NULL);
	blkP->freePtr = blkP->startPtr;

	FreeHandleCtr += szBlk;
	TotalHandleCtr += szBlk;
	return 1;	    /* alloc worked */
    }
    if (verbosegc) {
        jio_fprintf(stderr, "<GC: Couldn't allocate new handle chunk>\n");
    }

    return 0;		/* alloc failed */
}

/*
 * Expand the object space.  To go on we need to at least cover the
 * currently failing allocation, meaning we need a contiguous chunk of
 * enough pages to hold something of its size.
 * However, we normally also want to leave a user-controllable amount
 * of preallocated space free *after* we have satisfied the current
 * allocation.	*** DRW: Not Yet Implemented. ****
 *
 * The last_free parameter passed in here is ignored in the segemented version,
 * because it's almost impossible to glue two chunks together in the kind of
 * machines this segmented code will run on.
 *
 * Allocate markbits as needed for the chunk.  Sets curObjBlkP on exit.
 */
int
expandObjectSpace(long sz, unsigned char *last_free)
{
	/* last_free is ignored */
	ChunkBlkP blkP = NULL;
	int numPages;
	int needPages;

#ifdef TRACEGC
	if (tracegc) {
	    jio_fprintf(stderr, "Going to expand becuase we need sz=0x%lx bytes\n",
		    sz);
	    if (firstObjBlkP) {
		DumpHeapInfo();
	    }
	}
#endif /* TRACEGC */

	/* sysAssert(HEAP_LOCKED()); ...except during InitializeAlloc */

	/*
	 * AllocChunk may not give us exactly nPages*alignemnt bytes,
	 * so ask for extra for now.
	 */
	needPages = (sz >> PTR_2_PAGE_SHIFT) + 2;
	for (numPages = max(needPages, MIN_OBJ_PAGES);
	     numPages >= needPages;
	     numPages--) {
	    if (((numPages * PAGE_ALIGNMENT) <=
		 (gMaxMem - (TotalObjectCtr + TotalHandleCtr))) &&
		(((blkP = AllocChunk(numPages, OBJECTGRAIN)) !=
		  NULL))) {
	        break;
	    }
	}

	if (blkP) {
	    int szBlk;
    	    /** Goofiness because we need the object ptr to be OBJECTGRAIN
    	     ** aligned, and we store a header before that, so reset
    	     ** allocation ptrs to make it so. */
    	    /* startPtr is pointer to the HEADER! */
    	    blkP->startPtr = blkP->startPtr + OBJECTGRAIN - sizeof(hdr);
    	    blkP->freePtr = blkP->startPtr;
    	    /* keep end aligned, and make room for an "out of bounds"
    	     * header, which also indicates end of chunk. */
    	    blkP->endPtr = blkP->endPtr - OBJECTGRAIN + sizeof(hdr)
    	                   - OBJECTGRAIN;
    	    blkP->allocSize = (long)(blkP->endPtr - blkP->startPtr);
    	    blkP->freeCnt = blkP->allocSize;

    	    szBlk = blkP->allocSize;

    	    if (!setPageMap(blkP, OBJECT_HEAP_TYPE)) {
    	        return 0;
	    }
    	    sysAssert(szBlk >= sz);
#ifdef DEBUG
    	    *(long *)(blkP->startPtr + sizeof(hdr) + sizeof(int*)) = 0x55555555;
#endif /* DEBUG */
#ifdef LDEBUG
    	    memset(blkP->startPtr, 0x55, szBlk);
#endif /* LDEBUG */
            if (verbosegc) {
    	        jio_fprintf(stderr,
		 "<GC: expanded object space by 0x%lx bytes, chunk = 0x%lx>\n",
	           szBlk, blkP);
	    }

    	    insertChunk(&firstObjBlkP, &lastObjBlkP, blkP);
    	    curObjBlkP = blkP;			    /* make block currrent */
    	    obj_setlf(blkP->startPtr, szBlk, 1);    /* mark chunk free */
    	    /* zero-length block used to terminate some searches
    	     * in realObjAlloc */
    	    obj_setlf(blkP->endPtr, 0, 0);
#ifdef DEBUG
    	    checkChunk(blkP);
#endif /* DEBUG */
    	    FreeObjectCtr += szBlk;
    	    TotalObjectCtr += szBlk;
    	    return 1;				/* alloc worked */
	}
	if (verbosegc) {
	    jio_fprintf(stderr, "<GC: Couldn't allocate new object chunk>\n");
	}

	return 0;				/* alloc failed */
}

#else  /************ CONTIGUOUS HEAPS: ********************/

/*
 * Expand the mark bits array.	We don't care whether the expansion that
 * led to this is of the handle or object space, just that the contiguous
 * memory that the mark bits must span has expanded by the given amount.
 *
 * There is no slack in expanding the mark bits: we have already alloca-
 * ted the additional memory that the increment represents, so we have
 * got to be able to allocate mark bits to cover it.  We don't need to
 * initialize the new mark bits: they're zeroed on the start of each GC.
 *
 * Note: Given that sysCommitMem() should be assumed to actually allocate
 * swap space, it is arguable that we should put off committing memory
 * for the mark bits until GC actually needs it.  But GC is pretty much
 * inevitable in Java, so it's not clear there's any real benefit.
 */
static int
expandMarkBits(long incr)
{
    static long extra = 0;
    long markincr;
    unsigned int *incrbits;

    sysAssert(!markbitsInUse);

    markincr =
	((incr * 2 + OBJECTGRAIN*BITSPERMARK - 1)/(OBJECTGRAIN*BITSPERMARK)) *
	      sizeof(*markbits);
    if (markincr > extra) {
	markincr -= extra;	/* Use leftover from last expansion */
    } else {
	extra -= markincr;	/* Completely covered: no need to expand */
	marksize += markincr;
	return 1;
    }

    if (marksize + extra + markincr > markmax) {
        if (verbosegc) {
	    jio_fprintf(stderr, "<GC: tried to expand mark bits over limit>\n");
	}
	return 0;
    } else {
	long memincr;
        incrbits = sysCommitMem((char *)markbits+marksize+extra,
				markincr, &memincr);
	sysAssert(memincr >= markincr);
	if (!incrbits) {
	    if (verbosegc) {
	        jio_fprintf(stderr, "<GC: expansion of mark bits failed>\n");
	    }
	    return 0;
	}
	sysAssert(incrbits == (unsigned int *)((char *)markbits+marksize+extra));
	/* Avoid setting any globals until all error checking is done */
	marksize += extra + markincr;
	extra = memincr - markincr;	/* Leftover to use next expansion */
	return 1;
    }
}

/*
 * Expand the handle space.  The behavior of the handle space is simpler
 * than that of the object space: handles never move and the space can't
 * be fragmented, but that also means that contracting the handle space
 * is easy to block.
 *
 * If we were quicker to allocate new handles than GC to reclaim handles,
 * initially expanding by a fixed amount would be quick and easy.  But
 * given the attempt to stay small as much as possible, we are instead
 * using a strategy similar to the object store's, where we GC first,
 * and if we have to expand, we preallocate new space that scales with
 * the amount of existing space, subject to a minimum.
 *
 * Be sure to zero all the new handles: a free handle is recognized by
 * having its obj field zeroed.	 Remember that the base of the handle
 * space must be HANDLEGRAIN aligned within the committed memory.
 *
 * If we are going to seriously try to contract the handle space we
 * should instead allocate handles from higher addresses toward lower
 * addresses.
 */

#define MIN_HANDLE_EXPANSION 4*1024

int
expandHandleSpace(int numPages)
{   /* ignore numPages */
    long incr;
    long preallocIncr;
    unsigned char * oldhpool = hpool;

    /*
     * Expand by the maximum of a constant and the amount we need to
     * make FreeObjectCtr-n >= TotalObjectCtr*preallocFactor.  The
     * constant 1 rounds up after the division.
     */
    preallocIncr = 1 +
	((TotalHandleCtr*preallocFactor - (FreeHandleCtr-sizeof(JHandle))) /
		(1 - preallocFactor));
    incr = MIN_HANDLE_EXPANSION;
    incr = max(incr, preallocIncr);

    /* Always scale incr up to a multiple of HANDLEGRAIN */
    incr = (incr + HANDLEGRAIN - 1) & ~(HANDLEGRAIN - 1);

    if (hpool-incr < heapbase) { /* Future: We might get by with less */
        if (verbosegc) {
	    jio_fprintf(stderr, "<GC: tried to expand handle space over limit>\n");
	}
	return 0;
    } else {
        hpool -= incr;
	hpool = sysCommitMem(hpool, incr, &incr);
	if (!hpool) {
	    if (verbosegc) {
	        jio_fprintf(stderr, "<GC: expansion of handle space failed>\n");
	    }
	    return 0;
	}
	/* Need to be HANDLEGRAIN aligned within committed memory */
	hpool = (unsigned char *)
	  (((int) hpool + HANDLEGRAIN-1) & ~(HANDLEGRAIN-1));

	linkNewHandles((JHandle*)hpool, (JHandle*)oldhpool, hpoolfreelist);
	hpoolfreelist = hpool;
	FreeHandleCtr  += incr;
	TotalHandleCtr += incr;
	if (verbosegc) {
	  jio_fprintf(stderr,
		  "<GC: expanded handle space by %d to %d bytes, %d%% free>\n",
		  incr, TotalHandleCtr, 100 * FreeHandleCtr / TotalHandleCtr);
	}
	return expandMarkBits(incr);
    }
}

/*
 * Expand the object space.  To go on we need to at least cover the
 * currently failing allocation, meaning we need a contiguous chunk of
 * its size minus the amount of space in the trailing free chunk, if any.
 * However, we normally also want to leave a user-controllable amount
 * of preallocated space free *after* we have satisfied the current
 * allocation.	This preallocated space is calculated as a percentage
 * of the total free space, so given possible fragmentation, that free
 * space does not guarantee large contiguous allocations.  But given
 * that most allocations are not that large, preallocation can generally
 * reduce the number of interruptions due to memory management operations.
 * The user controls the rate of preallocation by varying the value of
 * the preallocFactor variable.
 *
 * [REMIND: Currently preallocFactor is global.	 That is probably right
 * for a single application, but could it lead to applets fighting over
 * its value???]
 *
 * The last_free parameter passed in here is the last free chunk
 * of memory as determined by the most recent heap compaction in
 * *this* round of memory management.  If the last chunk isn't known,
 * last_free is 0.  There is no guarantee that the last chunk abuts
 * the end of the object space, but if it does, any new memory allocated
 * here is coalesced with the last chunk.
 *
 * Because we are expanding at the top of the space, we leave opoolhand
 * where it is.
 */
int
expandObjectSpace(long n, unsigned char *last_free)
{
    long incr;
    long preallocIncr;
    unsigned char *incrptr;

    /*
     * Expand by the maximum of the minimum we need to satisfy the current
     * allocation (taking into account the last_free chunk) and the amount
     * we need such that FreeObjectCtr-n >= TotalObjectCtr*preallocFactor.
     * The constant 1 rounds up after the division.
     */
    preallocIncr = 1 +
	((TotalObjectCtr*preallocFactor - (FreeObjectCtr-n)) /
		(1 - preallocFactor));
    if (last_free) {
        if (n > obj_len(last_free)) {
	    /* May be expanding to satisfy an allocation, not preallocate */
	    incr = n - obj_len(last_free);
	} else {
	    /* We must be expanding to preallocate */
	    incr = 0;
	}
    } else {
        incr = n;
    }

    incr = max(incr, preallocIncr);
    /* Always scale incr up to a multiple of OBJECTGRAIN */
    incr = (incr + OBJECTGRAIN - 1) & ~(OBJECTGRAIN - 1);

    /* Expansion request should succeed when we don't need to expand.
     * If we don't return here, sysCommitMem will return 0 and cause
     * the memory expansion request to fail.
     */
    if (incr == 0)
        return 1;

    if (opoollimit + incr > heaptop) {
        if (verbosegc) {
	    jio_fprintf(stderr, "<GC: tried to expand object space over limit>\n");
	}
	return 0;
    } else {
        /* We could go on with anything down to n - obj_len(last_free) */
       incrptr = sysCommitMem(opoollimit + sizeof(hdr), incr, &incr);
       if (!incrptr) {
	   if (verbosegc) {
	       jio_fprintf(stderr, "<GC: expansion of object space failed>\n");
	   }
	   return 0;
       }
       sysAssert(incrptr == opoollimit + sizeof(hdr));
       if (last_free && last_free + obj_len(last_free) == opoollimit) {
  	  /*
	   * The last free chunk abuts the end of the object space, so
	   * we can coalesce it with the new memory.
	   */
	   obj_bumplen(last_free, incr);
#ifdef DEBUG
	   ((long *) last_free)[1] = 0x55555555;
#endif /* DEBUG */
       } else {
	  /*
	   * The last free chunk either doesn't abut the end of the
	   * object space or we don't know where it is, so we don't
	   * try to coalesce the new space.
	   */
	   obj_setlf(opoollimit, incr, 1);
#ifdef DEBUG
	   ((long *) opoollimit)[1] = 0x55555555;
#endif /* DEBUG */
       }
       opoollimit = opoollimit + incr;
       obj_setlf(opoollimit, 0, 0);
       FreeObjectCtr  += incr;
       TotalObjectCtr += incr;
       if (verbosegc) {
	 jio_fprintf(stderr,
		 "<GC: expanded object space by %d to %d bytes, %d%% free>\n",
		 incr, TotalObjectCtr, 100 * FreeObjectCtr / TotalObjectCtr);
       }
       return expandMarkBits(incr);
    }
}

#endif  /************ END CONTIGUOUS HEAPS ********************/

/*
 * Explicitly free an object and zero out a pointer to it
 */
static void
objFree_locked(HObject ** o)
{
    register unsigned char *p;

    /* Paranoia */
    if ((o == 0) || (*o == 0) || (unhand(*o) == 0)) {
        return;
    }

    p = ((unsigned char *) unhand(*o)) - sizeof(hdr);

    if (!obj_free(p)) {
	FreeObjectCtr += obj_len(p);
	FreeHandleCtr += sizeof(JHandle);
	obj_setfree(p);
	unhand(*o) = 0;
	*o = 0;
    }
}

/*
 * Deal with an out-of-memory situation caused by an overflow within
 * the Java heap, either of the handle area or the object area.	 This
 * thing is called with the heap lock, so doesn't have to worry about
 * someone else coming along and allocating or causing another GC.
 */
int
manageAllocFailure(long n, int overflow_type, int overflow_act) {
    int freeObject;
    int freeHandle;
    int nsoftrefs = 0;
    int ninbag = 0;
    unsigned char *last_free = 0;
    int minFree;

    sysAssert(HEAP_LOCKED());

    while (1) {
	if (verbosegc) {
	    jio_fprintf(stderr,
	"<GC: managing allocation failure. need %d bytes, type=%d, action=%d>\n",
		    n, overflow_type, overflow_act);
	}

	/*
	 * Because of fragmentation, it isn't good enough only to know
	 * how much free space is available.  Before we go try to
	 * allocate again, we want to know whether the last memory
	 * management operation we did made new progress.  So each
	 * time around the while() we reset the progress indicators.
	 */
	if (overflow_type == OVERFLOW_OBJECTS) {
	    freeObject = FreeObjectCtr;
	} else {
	    freeHandle = FreeHandleCtr;
	}

	switch (overflow_act) {

	case OVERFLOW_ACT_GC:
	    /*
	     * Try normal garbage collection; otherwise try finalizing.
	     */
	    last_free = gc0(0, n);
	    /*
	     * I'm not sure it's really worth finalizing and freeing
	     * refs for a handle overflow, as those are not good sources
	     * of many handles.  The alternative is going straight to
	     * OVERFLOW_ACT_EXPAND if overflow_type == OVERFLOW_HANDLES.
	     */
	    overflow_act = OVERFLOW_ACT_FINALIZE;
	    break;

	case OVERFLOW_ACT_FINALIZE:
	    /*
	     * Try finalizing then GCing again; next free refs
	     *
	     * If there are objects on the FinalizeMeQ they are known to
	     * be garbage that can't be freed until they are finalized.
	     * Synchronously finalize them, then garbage collect again.
	     */
	    overflow_act = OVERFLOW_ACT_REFS;
	    if (FinalizeMeQ != NULL) {
		bool_t ret = FALSE;
	        if (verbosegc) {
		    int count = 0;
		    JHandle *handle = FinalizeMeQ;
		    while (handle != NULL) {
		        count++;
			handle = FINALIZER_T(handle)->next;
		    }
		    jio_fprintf(stderr,
			    "<GC: synchronously running %d finalizers>\n",
			    count);
		}
		HEAP_UNLOCK();
		ret = runFinalization0();
		HEAP_LOCK();
		if (ret == TRUE) {	/* Finalization made progress */
		    last_free = gc0(0, n);
		}
		break;
	    } else {
	      continue; /* Don't bother checking for progress */
	    }

	case OVERFLOW_ACT_REFS:
	    /*
	     * If there are any refs defined, free some and GC again;
	     * keep trying this so long as there remain refs to free,
	     * and otherwise expand.
	     */
	    if ((ninbag = clearRefPointers(0, TotalObjectCtr, &nsoftrefs))
		!= 0) {
	        if (verbosegc) {
		    jio_fprintf(stderr, "<GC: zeroed %d of %d soft refs>\n",
			    ninbag, nsoftrefs);
		}
		last_free = gc0(0, n);
		overflow_act = OVERFLOW_ACT_REFS;
		break;
	    } else {
	        overflow_act = OVERFLOW_ACT_EXPAND;
		continue;	    /* Don't bother checking for progress */
	    }

	case OVERFLOW_ACT_EXPAND:
	    /*
	     * Expand the overflowing area: Successful expansion should
	     * guarantee that the allocation will succeed, and should not
	     * over(pre)allocate.
	     */
	    overflow_act = OVERFLOW_ACT_DESPERATE;
	    if (overflow_type == OVERFLOW_HANDLES) {
	        if (expandHandleSpace(0)) {
		    /* We got *some* memory */
		    return overflow_act;
		}
	    } else if (overflow_type == OVERFLOW_OBJECTS) {
	        if (expandObjectSpace(n, last_free)) {
		    /* We got *some* memory */
		    return overflow_act;
		}
	    }
	    break;

	case OVERFLOW_ACT_DESPERATE:
#ifdef ALLOC_BARRIERS
	    if (allocstate == ALLOCSTATE_GREEN) {
		if (BARRIERSIZE_YELLOW != 0) {
	    	    objFree_locked(&PRYellow); 	/* Free the yellow reserve */
		    setAllocState(ALLOCSTATE_YELLOW);
		    if (verbosegc) {
			jio_fprintf(stderr, "<GC: allocstate up to Yellow because of allocation request of %d bytes>\n", n);
		    }
		    return OVERFLOW_ACT_GC;  /* retry, GC next if still fails */
		} else if (BARRIERSIZE_RED != 0) {
	    	    objFree_locked(&PRRed); 	/* Free the red reserve */
		    setAllocState(ALLOCSTATE_RED);
		    if (verbosegc) {
			jio_fprintf(stderr, "<GC: allocstate up to Red because of allocation request of %d bytes>\n", n);
		    }
		    return OVERFLOW_ACT_GC;  /* retry, GC next if still fails */
		}   /* No reserves, fall through to out of heap handling */
	    } else if (allocstate == ALLOCSTATE_YELLOW) {
		if (BARRIERSIZE_RED != 0) {
	            objFree_locked(&PRRed); 	/* Free the red reserve */
		    setAllocState(ALLOCSTATE_RED);
		    if (verbosegc) {
		        jio_fprintf(stderr, "<GC: allocstate up to Red>\n");
		    }
		    overflow_act = OVERFLOW_ACT_GC;
		    continue;	/* Back to GC, don't check for progress */
		}   /* No reserves, fall through to out of heap handling */
	    } else {
	        sysAssert(allocstate == ALLOCSTATE_RED);
		/* Already red, fall through to out of heap handling */
	    }
#endif /* ALLOC_BARRIERS */
	    if (verbosegc) {
	        jio_fprintf(stderr, "<GC: totally out of heap space>\n");
	    }
	    /* Failed to make any progress: really out of heap */
	    return OVERFLOW_ACT_FAIL;

	default:
	    return OVERFLOW_ACT_FAIL;
	}

	/*
	 * Test whether the last memory management operation made
	 * sufficient progress to try allocating again.	 Otherwise,
	 * go directly to trying the next operation.  Subtlety: because
	 * of fragmentation, apparently adequate free space is not always
	 * enough, so we also require evidence of recent progress.  Even
	 * so, there's still no guarantee allocation will go through.
	 * Note that we subtract off the current allocation amount: the
	 * preallocated memory is the amount after this allocation.
	 *
	 * DRW: The old pre-alloc factor required too much free mem at large
	 * values. PAGE_ALIGNMENT is probably too small, but it's better
	 * than nothing. *** REMIND: Try other policies!
	 */
#ifdef PAGED_HEAPS
	minFree = PAGE_ALIGNMENT;
#else /* not PAGED_HEAPS: */
	if (overflow_type == OVERFLOW_OBJECTS) {
	    minFree = TotalObjectCtr*preallocFactor;
	} else {
	    minFree = TotalHandleCtr*preallocFactor;
	}
#endif /* not PAGED_HEAPS */

	if (overflow_type == OVERFLOW_OBJECTS) {
	    if (FreeObjectCtr > freeObject &&
		FreeObjectCtr-n >= minFree) {
	        return overflow_act;	/* Try to allocate again */
	    }	/* else go back and switch on new overflow_act */
	} else {
	    if (FreeHandleCtr > freeHandle &&
		FreeHandleCtr-sizeof(JHandle) >= minFree) {
	        return overflow_act;    /* Try to allocate again */
	    }	/* else go back and switch on new overflow_act */
	}
    }
}

#ifdef ALLOC_BARRIERS
/*
 * Check to see if we can lower the allocstate back to green (or yellow).
 * This should be called with the heap locked, so we don't get duplicate
 * allocations of the panic reserves (and so the AllocNoGC flag won't
 * be corrupted by other threads).
 *
 * Note that if we were permitted to transition to a yellow or red state
 * then we must have nonzero BARRIERSIZE_YELLOW or BARRIERSIZE_RED,
 * respectively.
 */
static void
tryLowerAllocState()
{
    if (allocstate == ALLOCSTATE_RED) {
	AllocNoGC = TRUE;
	PRRed = ArrayAlloc(T_PRRED,BARRIERSIZE_RED);
	AllocNoGC = FALSE;
	if (PRRed == NULL) {
	    return;
	}
	if (BARRIERSIZE_YELLOW != 0) {
	    setAllocState(ALLOCSTATE_YELLOW);
	    if (verbosegc) {
	        jio_fprintf(stderr, "<GC: allocstate down to Yellow>\n");
	    }
	} else {	/* Transition straight to green */
	    setAllocState(ALLOCSTATE_YELLOW);
	    if (verbosegc) {
		jio_fprintf(stderr, "<GC: allocstate down to Green>\n");
	    }
	}
    }
    if (allocstate == ALLOCSTATE_YELLOW) {
	AllocNoGC = TRUE;
	PRYellow = ArrayAlloc(T_PRYELLOW,BARRIERSIZE_YELLOW>>1);
	AllocNoGC = FALSE;
	if (PRYellow == NULL) {
	    return;
	}
	setAllocState(ALLOCSTATE_GREEN);
	if (verbosegc) {
	    jio_fprintf(stderr, "<GC: allocstate down to Green>\n");
	}
    }
}
#endif /* ALLOC_BARRIERS */

/* Mark the handles enclosed by the object "h" refers to, and recursively
 * call markChildren on those handles.
 *
 * The "limit" is used to prevent double-marking.  Since handles are scanned
 * in address order, we only want to recursively mark children of handles that
 * the handle scan has passed over (ie handles < limit), because scanHandles()
 * will get to the later handles in a while anyway.
 */


#define MC_STACK
/* #define MC_COVERAGE */

#ifdef MC_STACK

/* Use shared buffer to save memory */
/* static void*     mc_stack_global[MC_STACK_SIZE]; */
static void** const mc_stack_global = gcSharedBuffer.mc_stack;
#  ifdef MC_COVERAGE
static int push_states[8];
static int pop_states[8];
static int deferral_states, undeferral_states;
#  endif
#endif

static void
markChildren(register JHandle * h, register JHandle * limit, int depth)
{

/* MARKIT: The basic recursive marker.  Mark the handle "han" 
 * with the mark "pMark",   
 * if necessary
 *    pMark  - mark to assign to han
 *    han    - valid handle to be marked with pMark, if it has a weaker mark
 *    label  - a unique label
 *    count  - number of pairs in the following list
 *    saves  - (type1 var1 . . . . typen varn).   Variables whose value
 *             is to be saved/restored in case of a recursive call.
 *    cleanup- Additional cleanup to be done on return from recursive call.
 *
 * All live variables must either be be one of the var-i, or must
 * have their value re-created in the cleanup from one of the saved 
 * variables
 */

#define MARKIT(pMark, han, label, num, saves, cleanup) \
    if (IsMarked(han) < pMark) {	    	       \
        MarkPtr(han, pMark);			       \
	if (han < limit) {			       \
	    RECURSE(han, label, num, saves, cleanup);  \
	}					       \
    }

/* Similar to MARKIT.  But we use the needsmark variable to give us tail
 * recursion.  We recurse on "needsmark" instead of on "han" if it is 
 * non-NULL, and save "han" in the variable needsmark when we're done.
 *
 * The value of "han" is live after the recursive call.  It must be
 * saved.
 */

#define MARKITsub(pMark, han, label, num, saves, cleanup) \
    MARKITsubx(pMark, han, IsMarked(han), label, num, saves, cleanup)

/* Identical to MARKITsub, except that we already now the value of
 * isMarked(han)
 */

#define MARKITsubx(pMark, han, hanMark, label, num, saves, cleanup) \
    if (hanMark < pMark) {	         	       \
        MarkPtr(han, pMark);			       \
	if (han < limit) {			       \
            if (needsmark) {                           \
                JHandle *_t_ = needsmark;              \
		needsmark = NULL;                      \
	        RECURSE(_t_, label, num, saves, cleanup); \
	    }                                          \
	    needsmark = han;                           \
	}					       \
    }

/* Used by MARKIT to specify no cleanup (this makes compiler happy) */
#define NOP ((void)0)

#ifdef MC_STACK
#define RECURSE(han, label, count, saves, cleanup)  \
     {  void **new = mc_stack_ptr + count + 1;      \
        if (new >= mc_stack_end) {                  \
            COVERAGE_DEFERRAL();                    \
            putDeferredHandle(han);		    \
        } else {                                    \
            PUSH_VALUES ## count ## saves;          \
	    COVERAGE_ON(label);                     \
            PUSH_STATE(label);                      \
            h = han;                                \
            goto ScanNext;                          \
         STATE(label):                              \
            POP_VALUES ## count ## saves;           \
            cleanup  ;                              \
	    COVERAGE_OFF(label);                    \
         }                                          \
     }
#else
#define RECURSE(han, label, count, saves, cleanup)  \
    if (depth >= MAX_GC_DEPTH) {                    \
       putDeferredHandle(han);                      \
    } else {                                        \
       markChildren((JHandle *)han, limit, depth+1);\
    }
#endif

/* Macros for pushing values onto the stack */

#define PUSH_VALUES5(type1, var1, type2, var2, type3, var3,              \
		     type4, var4, type5, var5)                           \
    PUSH_VALUES4(type1, var1, type2, var2, type3, var3, type4, var4);    \
    *mc_stack_ptr++ = (void *)var5;

#define PUSH_VALUES4(type1, var1, type2, var2, type3, var3, type4, var4) \
    PUSH_VALUES3(type1, var1, type2, var2, type3, var3);                 \
    *mc_stack_ptr++ = (void *)var4;

#define PUSH_VALUES3(type1, var1, type2, var2, type3, var3)              \
    PUSH_VALUES2(type1, var1, type2, var2);                              \
    *mc_stack_ptr++ = (void *)var3;

#define PUSH_VALUES2(type1, var1, type2, var2)                           \
    PUSH_VALUES1(type1, var1);                                           \
    *mc_stack_ptr++ = (void *)var2;

#define PUSH_VALUES1(type1, var1)                                        \
    *mc_stack_ptr++ = (void *)var1;

#define PUSH_VALUES0()

/* Macros for popping values off the stack */

#define POP_VALUES5(type1, var1, type2, var2, type3, var3,           \
		    type4, var4, type5, var5)                        \
    var5 = (type5)(*--mc_stack_ptr);                                 \
    POP_VALUES4(type1, var1, type2, var2, type3, var3, type4, var4);

#define POP_VALUES4(type1, var1, type2, var2, type3, var3, type4, var4)  \
    var4 = (type4)(*--mc_stack_ptr);                                 \
    POP_VALUES3(type1, var1, type2, var2, type3, var3);

#define POP_VALUES3(type1, var1, type2, var2, type3, var3)           \
    var3 = (type3)(*--mc_stack_ptr);                                 \
    POP_VALUES2(type1, var1, type2, var2);

#define POP_VALUES2(type1, var1, type2, var2)                        \
    var2 = (type2)(*--mc_stack_ptr);                                 \
    POP_VALUES1(type1, var1);

#define POP_VALUES1(type1, var1)                                     \
    var1 = (type1)(*--mc_stack_ptr);

#define POP_VALUES0()

#define STATE(label) State_ ## label

#ifdef MC_COVERAGE
#  define COVERAGE_ON(i)        push_states[i]++;
#  define COVERAGE_OFF(i)       pop_states[i]++;
#  define COVERAGE_DEFERRAL()   deferral_states++;
#  define COVERAGE_UNDEFERRAL() undeferral_states++;
#else 
#  define COVERAGE_ON(i)  
#  define COVERAGE_OFF(i) 
#  define COVERAGE_DEFERRAL() 
#  define COVERAGE_UNDEFERRAL() 
#endif

/* Macros for returning to a particular place in the code, upon returning
 * from recursion.  gcc is nice and let's us save the address of a label
 * as a void*.  In standard C, we have to save an int, and then use a switch
 * statement on the value of that int 
 */

#define MARK_OBJECT_CLASS  0
#define MARK_FIELDS        1
#define MARK_CONSTANT_POOL 2
#define MARK_CLASS_STATICS 3
#define MARK_CLASS_LOADER  4
#define MARK_CLASS_SIGNERS 5
#define MARK_SUPER         6
#define MARK_ARRAY         7


#ifdef __GNUC__
#  define PUSH_STATE(label) *mc_stack_ptr++ = &&STATE(label)
#  define POP_STATE() \
      if (mc_stack_ptr > mc_stack) goto **--mc_stack_ptr;
#else
#  define PUSH_STATE(label) *mc_stack_ptr++ = (void *)label
#  define POP1_STATE(label) case label: goto STATE(label);
#  define POP_STATE()                     \
       if (mc_stack_ptr > mc_stack) {     \
  	  switch((int)*--mc_stack_ptr) {  \
	      POP1_STATE(MARK_OBJECT_CLASS);    \
              POP1_STATE(MARK_FIELDS);          \
              POP1_STATE(MARK_CONSTANT_POOL);   \
              POP1_STATE(MARK_CLASS_STATICS);   \
              POP1_STATE(MARK_CLASS_LOADER);    \
              POP1_STATE(MARK_CLASS_SIGNERS);   \
              POP1_STATE(MARK_SUPER);           \
              POP1_STATE(MARK_ARRAY);           \
	      default:							 \
	         panic("Bad state in markChildren recursion stack\n");	 \
	  }								 \
       }
#endif


    register ClassClass *cb;
    register JHandle *sub, *needsmark = NULL;

    int parentMark;
    SetLimits();

    extern ClassClass* classJavaLangClass;
    register ClassClass *classJavaLangClass_local = classJavaLangClass;


#ifdef MC_STACK
    register void **mc_stack = mc_stack_global;
    register void **mc_stack_ptr = mc_stack;
    void          **mc_stack_end = &mc_stack[MC_STACK_SIZE];
#endif

ScanNext:
    parentMark = IsMarked(h);
    sysAssert(parentMark != NoMark);

#ifdef TRACEGC
    if (tracegc > 1) {
	jio_fprintf(stderr, "<%X/%X>\n", h, unhand(h));
	jio_fprintf(stderr, " <  %s\n", Object2CString(h));
    }
#endif /* TRACEGC */
    if (obj_isnt_array(h)) {
        cb = obj_classblock(h);

	if (!cbIsInROM(cb)) {			
	    MARKIT(parentMark, (JHandle *)cb, MARK_OBJECT_CLASS,
		   2, (JHandle *, h, int, parentMark),
		   (cb = obj_classblock(h)));
	}
	if (cb != classJavaLangClass_local) {
	    register ClassObject *p = unhand(h);
	    do {
		register long n = cbFieldsCount(cb);
		register struct fieldblock *fb = cbFields(cb);
		/* sun.misc.Ref has only one non-static reference field.
		 * Anything in that field gets marked as soft. 
		 */
		bool_t isSoftRef = (cb == classSunMiscRef);
		for ( ; --n >= 0; fb++) {
		    if (IS_NONSTATIC_AND_REFERENCE(fb)) {
			int offset = fieldoffset(fb);
			sub = *(JHandle **) ((char *) p + offset);
			if (ValidHandle(sub)) {
			    int subMark = IsMarked(sub);
			    if (subMark != HardMark) {
				int newMark = isSoftRef ? SoftMark : parentMark;
				/* Strengthen the mark as necessary */
				MARKITsubx(newMark, sub, subMark, 
					   MARK_FIELDS, 
					   4,
					   (ClassObject *, p,
					    int, parentMark,
					    int, n,
					    struct fieldblock *, fb),
					   (cb = fieldclass(fb),
					    offset = fieldoffset(fb), 
					    sub = *(JHandle **) ((char *) p + offset)));
			    }
			}
		    }
		}
		cb = cbSuperclass(cb);
	    } while (cb != NULL);
	} else if (!cbIsInROM((ClassClass *)h)) {
	    ClassClass *cb = (ClassClass *)h;
	    int parentMark = IsMarked(cbHandle(cb));
	    int i, j;
	    struct fieldblock *fb;

	    /* We must be extra careful in scanning the internals of a class
	     * structure, because this routine may be called when a class
	     * is only partially loaded (in createInternalClass).
	     */
	    if (cbConstantPool(cb) &&
		cbConstantPool(cb)[CONSTANT_POOL_TYPE_TABLE_INDEX].type) {
		union cp_item_type *constant_pool = cbConstantPool(cb);
		union cp_item_type *cpp = 
		    &constant_pool[CONSTANT_POOL_UNUSED_INDEX];
		union cp_item_type *end_cpp = 
		    &constant_pool[cbConstantPoolCount(cb)];
		unsigned char *type_tab =
		    constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type;
		unsigned char *this_type = 
		    &type_tab[CONSTANT_POOL_UNUSED_INDEX];
		for ( ; cpp < end_cpp; cpp++, this_type++) {
		    switch (*this_type) {
		    case CONSTANT_Class|CONSTANT_POOL_ENTRY_RESOLVED: {
			/* Only mark if the entry is a resolved Class */
			ClassClass *sub_cb = (*cpp).clazz;
			/* Kona: if (ValidHorO(sub_cb) && sub_cb != cb) { */
			if (sub_cb && sub_cb != cb && !cbIsInROM(sub_cb)) {
#ifdef TRACEGC
			    if (tracegc > 1 && !IsMarked(sub_cb)) {
				if (sub_cb->obj && sub_cb->methods) {
				    jio_fprintf(stderr, " ms %s\n",
						Object2CString(sub_cb));
				} else {
				    jio_fprintf(stderr, " ms %8X\n", sub_cb);
				}
			    }
#endif /* TRACEGC */
			    MARKIT(parentMark,(JHandle *)sub_cb,
				   MARK_CONSTANT_POOL,
				   5,
				   (ClassClass *, cb,
				    int, parentMark,
				    union cp_item_type *, cpp,
				    union cp_item_type *, end_cpp,
				    unsigned char *, this_type), NOP);
			}
			break;
		    }
		    }
		} /* loop over constant pool*/
	    }

	    /* Scan class definitions looking for statics */
	    if (CCIs(cb, Scannable) &&
		cbFields(cb) && (cbFieldsCount(cb) > 0)) {
		for (i = cbFieldsCount(cb), fb = cbFields(cb); --i >= 0; fb++) {
		    if (IS_STATIC_AND_REFERENCE(fb)) {
			JHandle *sub = *(JHandle **)normal_static_address(fb);
			if (ValidHandle(sub)) {
			    MARKIT(parentMark, sub, MARK_CLASS_STATICS,
				   3,
				   (int, parentMark,
				    int, i,
				    struct fieldblock *, fb),
				   (cb = fieldclass(fb)));
#ifdef TRACEMARK
			    if (tracegc > 1) {
				jio_fprintf(stderr, " mc1 %s\n",
					    Object2CString(sub));
			    }
#endif /* TRACEMARK */
			}
		    }
		}
	    }
	    if (ValidHandle(cbLoader(cb)))
		MARKIT(parentMark, (JHandle *)cbLoader(cb),
		       MARK_CLASS_LOADER,
		       2, (ClassClass *, cb, int, parentMark), NOP);
	    if (ValidHandle(cbSigners(cb)))
		MARKIT(parentMark, (JHandle *)(cbSigners(cb)),
		       MARK_CLASS_SIGNERS,
		       2, (ClassClass *, cb, int, parentMark), NOP);
	    cb = cbSuperclass(cb);
	    if (cb && (ValidHandle(cb)))
		MARKIT(parentMark, (JHandle *)cb, MARK_SUPER, 0, (), NOP);
	}
    } else if (obj_array_flags(h) == OBJ_FLAG_CLASS_ARRAY) { 
	register long n = obj_length(h);
	HObject **body = unhand((HArrayOfObject *)h)->body;
	for (; n >= 0; n--) {
	    sub = body[n];
	    if (ValidHandle(sub)) {
		MARKITsub(parentMark, sub, MARK_ARRAY,
			  3, (HObject **, body, int, parentMark, int, n),
			  (sub = body[n]));
	    }
	}
    }
    if (needsmark != NULL) { 
        h = needsmark;
	needsmark = NULL;
	goto ScanNext;          /* tail recursion */
    }

    POP_STATE();

    if ((depth == 0) && ((h = getDeferredHandle()) != NULL)) { 
         COVERAGE_UNDEFERRAL();
         goto ScanNext;
    } 
#ifdef TRACEGC
    if (tracegc > 1) {
	jio_fprintf(stderr, " >  %s\n", Object2CString(h));
    }
#endif /* TRACEGC */
}


static int
scanThread(sys_thread_t *t, void *arg)
{
    JHandle *me = (JHandle *) arg;
    JHandle *p = (JHandle *) sysThreadGetBackPtr(t);
    unsigned char **ssc, **limit;
    stackp_t base;
    ExecEnv *ee = (ExecEnv *)THREAD(p)->eetop;
    SetLimits();

#ifdef DEBUG
    if (!ValidHandle(p)) {
	sysAbort();
    }
#endif /* DEBUG */
#ifdef TRACEMARK
    if (tracegc) {
	jio_fprintf(stderr, "scanThread, t:%x, arg:%x, p:%x\n", t, arg, p);
    }
    if (tracegc > 1 && !IsMarked(p)) {
	if (((JHandle *)p)->obj && ((JHandle *)p)->methods) {
	    jio_fprintf(stderr, " ms1 %s\n", Object2CString((JHandle *)p));
	} else {
	    jio_fprintf(stderr, " ms1 %8X\n", p);
	}
    }
#endif /* TRACEMARK */

    /* Mark thread object */
    MarkPtr(p, HardMark);

    if (SYSTHREAD(p) == 0 || ee == NULL || !ee->has_been_born) {
	/* The thread must be unborn or exiting... */
	goto ScanJavaStack;
    }

    base = threadStackBase((TID)p);
    ssc	 = threadStackPointer((TID)p);

    if (p == me) {
	ssc = (unsigned char **) &me;
    }

    if (ssc == 0 || base == 0 || (ssc == base)) {
	/*
	 * If the stack does not have a top of stack pointer or a base
	 * pointer then it hasn't run yet and we don't need to scan
	 * its stack.  When exactly each of these data becomes available
	 * may be system-dependent, but we need both to bother scanning.
	 */
	goto ScanJavaStack;
   }

    /* Align stack top, important on Windows 95. */
    if ((long)ssc % sizeof(void *)) {
        ssc = (unsigned char **)((long)(ssc) & ~(sizeof(void *) - 1));
    }

    limit = (unsigned char **) base;

#ifdef TRACEMARK
    if (tracegc) {
	jio_fprintf(stderr, "%8X [%10X,%10X] %d\n", p, ssc, limit, limit - ssc);
    }
#endif /* TRACEMARK */
    sysAssert(ssc != limit);

    while (ssc < limit) {
	register unsigned char *ptr = *ssc;
	if (ValidHorO(ptr)) {
	    /* If ptr is (apparently) an object pointer, it is pinned */
#ifdef TRACEMARK
	    if (tracegc > 1 && !IsMarked(ptr) && ValidHandle(ptr)) {
		if (((JHandle *)ptr)->obj && ((JHandle *)ptr)->methods) {
		    jio_fprintf(stderr, " ms2 %s\n",
			    Object2CString((JHandle *) ptr));
		} else {
		    jio_fprintf(stderr, " ms2 0x%8X\n", ptr);
		}
	    }
#endif /* TRACEMARK */
	    MarkPtr(ptr, HardMark);
	}
	ssc++;
    }

    /*
     * Whether or not we scan the thread stack, we decide independently
     * whether to scan the Java stack.  Doing so should be more robust
     * in the face of partially-initialized or partially-zeroed threads
     * during thread creation or exit, or changes to any of that code.
     */
ScanJavaStack:
    {
	JavaFrame *frame;
	register unsigned char *ptr;

	if (ee != 0) {
	    /*
	     * Because of the Invocation API, the EE may not be on the C
	     * stack anymore.
	     */
	    if (ValidHandle(ee->exception.exc))
	        MarkPtr(ee->exception.exc, HardMark);
	}

	if ((ee != 0) && ((frame = ee->current_frame) != 0)) {
	    struct methodblock *current_method = frame->current_method;
	    struct methodblock *prev_current_method;
	    stack_item *top_top_stack = (!IS_NATIVE_MB(current_method))
		  ? &frame->ostack[methodmaxstack(frame->current_method)]
		  : frame->optop;

	    for(;;) {
		struct methodblock *mb = frame->current_method;

		if (mb == 0 || !IS_JIT_FRAME(frame))
                    scanJavaFrame(frame, top_top_stack);

		prev_current_method = current_method;
		frame = frame->prev;
		if (!frame) break;
		current_method = frame->current_method;
		/*
		 * If the previous frame was a transition frame from C back
		 * to Java (indicated by prev_current_method == NULL), then
		 * this new frame might not have set its optop.  We must be
		 * conservative.   Otherwise, we can use the optop value.
		 *
		 * Also permit two consecutive frames with NULL current
		 * methods, in support of JITs.  See bug 4022856.
		 */
		top_top_stack =
		    (prev_current_method == NULL &&
			!IS_NATIVE_MB(current_method))
			    ? &frame->ostack[methodmaxstack(current_method)]
			    : frame->optop;
	    }
	} else {
#ifdef TRACEMARK
	    if (tracegc) {
		jio_fprintf(stderr, "ee=%x\n", ee);
		if (ee) {
		    jio_fprintf(stderr, "frame=%x\n", ee->current_frame);
		}
	    }
#endif /* TRACEMARK */
	}

    }

    return SYS_OK;
}

static void
scanThreads(void)
{
    JHandle *self;
    int err;

    self = (JHandle *) sysThreadGetBackPtr(sysThreadSelf());
    err = sysThreadEnumerateOver(scanThread, (void *) self);
    /* REMIND - should do something with the return value. */
}

static void
scanJavaFrame(JavaFrame *frame, stack_item *top_top_stack)
{
    stack_item *ssc, *limit;
    JHandle *ptr;
    JavaStack *javastack;
    struct methodblock *mb = frame->current_method;
    SetLimits();

#ifdef TRACEMARK
    if (tracegc) {
        jio_fprintf(stderr, " scanning frame %8X\n", frame);
    }
#endif /* TRACEMARK */
 
    if (mb) {
	ClassClass *mbClass = methodclass(mb);
	if (!cbIsInROM(mbClass)) {
	    MarkPtr(cbHandle(mbClass), HardMark);
	}
    }

    limit = top_top_stack;
    javastack = frame->javastack;
 
    /* Scan the operand stack. */
    /*CONSTCOND*/
    while (1) {
        int is_first_chunk = IN_JAVASTACK((stack_item *)frame, javastack);
        for (ssc = is_first_chunk ? frame->ostack : javastack->data;
             ssc < limit; ssc++) {
            ptr = ssc->h;
            if (ValidHandle(ptr)) {     /* Never an object pointer */
#ifdef TRACEMARK
                if (tracegc > 1 && !IsMarked(ptr)) {
                    jio_fprintf(stderr, " ms3 %8X\n", ptr);
                }
#endif /* TRACEMARK */
                MarkPtr(ptr, HardMark);
            }
        }
        if (is_first_chunk)
            break;
        javastack = javastack->prev;
        limit = javastack->end_data;
    }
 
    /* Nothing more to do for pseudo frames. */
    if (mb == 0 || IS_JIT_FRAME(frame))
        return;
 
    if (methodaccess(mb) & ACC_NATIVE) {
        /* For native frames, we scan the arguments stored at the top
           of the previous frame. */
        JavaFrame *prev_frame = frame->prev;
        if (prev_frame == 0)
            return;
        ssc = prev_frame->optop;
        limit = ssc + methodargssize(mb);
    } else {
        /* Scan local variables in Java frame */
        ssc = frame->vars;
        if (ssc == 0)
            return;
        limit = (stack_item *)frame;
    }
 
    for (; ssc < limit; ssc++) {
        ptr = ssc->h;
        if (ValidHandle(ptr)) {  /* Never an object pointer */
#ifdef TRACEMARK
            if (tracegc > 1 && !IsMarked(ptr)) {
                jio_fprintf(stderr, " ms4 %8X\n", ptr);
            }
#endif /* TRACEMARK */
            MarkPtr(ptr, HardMark);
        }
    }
}

static void
setClassGCMode()
{
    extern int no_verifiers;     /* Updated by verify_class_codes    */

    /* We are single threaded here, so no need to lock following check */
    if (classgc && no_verifiers == 0) {
        allowClassGCNow = TRUE;
    } else {
        allowClassGCNow = FALSE;
    }
}

static void markInternedString(char *utf, void *ptr);

/* Scan all the strings that are in the intern table */
static void scanInternedStrings(void)
{
    sysAssert(STRINGHASH_LOCKED());
    Str2IDCallback(&stringHash, markInternedString);
}

/* Mark one string that's in the intern table.  This is called back */
static void markInternedString(char *utf, void *ptr) {
    SetLimits();
    if (ValidHandle(ptr)) {
        MarkPtr(ptr, HardMark);
    }
}

/* Mark the global references created by the Native Method interface */
static void scanJNIGlobalRefs(void)
{
    scanJavaFrame(globalRefFrame, globalRefFrame->optop);
}

/*
 * Scan the heap and chase pointers.  Scanning from pointers into the
 * middle of objects is normally disabled, apparently because it was
 * slow.  This should be revisited, especially if it reduces pinning.
 */
static void
scanHeap(void)
{
#ifdef CHECK_INTERIOR_POINTERS
    SetLimits();
    unsigned char *o, *oo, *next;
    int i, size;

#ifdef TRACEGC
    if (tracegc) {
        jio_fprintf(stderr, "\nRecursive mark scan:\n");
    }
#endif /* TRACEGC */

    /*
     * Search the entire memory pool looking for references into the
     * the middle of objects.  If any are found, mark the object they
     * are part of.  Pointers into the middle of objects can happen
     * frequently, especially in the case of arrays.
     *
     * This piece of code is tricky because it needs to be fast.
     * It is basically walking through most of the mark bits looking
     * for set bits.  It does it in 32-bit increments for speed.
     */
    o = oo = opmin;
    next = o + obj_len(o);
    size = ((int) (opoollimit - opmin) / (OBJECTGRAIN * BITSPERMARK) + 1) * 2;
    for (i=0; i < size; i++) {
	sysAssert(((unsigned int)(oo - opmin) >> 7) == i);
	if (markbits[i]) {
	    unsigned int bits = markbits[i];
	    unsigned int shifts = OBJECTGRAIN * BITSPERMARK / 2;
	    while (bits && oo < opmax) {
	        if (bits & 0x3) {
		    /* Mark the object, not the header */
		    MarkPtr((o + sizeof(hdr)), HardMark);
		}
		bits >>= 2;
		shifts -= OBJECTGRAIN;
		oo += OBJECTGRAIN;
		if (oo == next) {
		  o = next;
		  next = o + obj_len(o);
		}
	    }
	    oo += shifts;
	} else {
	    oo += OBJECTGRAIN * BITSPERMARK / 2;
	}

	if (oo >= opmax) {
	    break;
	}

	while (oo >= next) {
	  o = next;
	  next = o + obj_len(o);
	}
    }
#endif /* CHECK_INTERIOR_POINTERS */
}


/*
 * Walk the entire handle pool, and do a recursive search for
 * references.
 */
static void
scanHandles(void)
{
    SetLimits();
    JHandle *hp;


    deferredHandleTableOverflow = 1;

    while (deferredHandleTableOverflow) {
        deferredHandleTableOverflow = 0;
        MAP_OVER_HANDLES_FROM_START(hp)
	    if (unhand(hp)) {
	        if (ValidObject(unhand(hp)) && IsMarked(unhand(hp))) {
#ifdef PAGED_HEAPS
		    ChunkBlkP ObjChunk = ChunkBase(unhand(hp));
		    /* note that this chunk is pinned */
		    ObjChunk->chunkFlags |= CHUNK_PINNED;
#endif /* PAGED_HEAPS */
		    MarkPtr((JHandle *) hp, HardMark);
		}
		if (IsMarked(hp)) {
		    markChildren(hp, hp + 1, 0);
		}
	    }
	END_MAP_OVER_HANDLES_FROM_START;
    }
}


static void
markROMClassStatics() {
    extern JHandle* StaticRefStorage[];
    extern uint32_t StaticRefStorageCount;
    JHandle **ptr;
    int i;
    
    SetLimits();
    for (ptr = StaticRefStorage, i = StaticRefStorageCount; --i >= 0; ptr++) { 
        JHandle *sub = *ptr;
	if (ValidHandle(sub)) 
	    MarkPtr(sub, HardMark);
    }
}


/* Mark all classes that shouldn't be unloaded. */
static void
markStickyClasses()
{
    ClassClass **pcb =    &binclasses[numRomClasses];
    ClassClass **pcbEnd = &binclasses[nbinclasses];
    SetLimits();

    /* We purposely avoid marking the ROM'ed classes.  They're automatically
     * sticky, and can't be marked anyway.
     */
    for ( ; pcb < pcbEnd; pcb++) {
        ClassClass *cb = *pcb;
	if (!allowClassGCNow || CCIs(cb, Sticky))
	    MarkPtr(cbHandle(cb), HardMark);
    }
}


/*
 * Mark handles to the 9 primitive classes.  Currently, these are not
 * in binclasses, and they have no internal object pointers.
 */
static void
markPrimitiveClasses()
{
    static ClassClass **primitive_classes[] = {
	&class_void, &class_boolean, &class_byte, &class_char, &class_short,
	&class_int, &class_long, &class_float, &class_double
    };
    ClassClass ***cbpp;
    int count;
    /* If ROMizing, then primitive classes go in rom, and no need to mark */
    if (numRomClasses == 0) { 
        SetLimits();
	count = sizeof(primitive_classes)/sizeof(primitive_classes[0]);
	for (cbpp = primitive_classes; --count >= 0; cbpp++) { 
	    ClassClass *cb = **cbpp;
	    if (ValidHandle(cb)) {
	        MarkPtr(cb, HardMark);
		sysAssert(ValidObject(unhand(cb)));
	    }
	}
    } 
}

/*
 * Take care of finalization
 *
 * At this point we have marked all reachable objects.	Anything
 * not marked and having a finalization method will be queued for
 * finalization.  However, it is not the case that everything
 * else that's unmarked can just be freed.  An object to be
 * finalized expects everything it references to be retained by
 * GC, whether or not those things are themselves finalized.  So
 * we make a pass over the HasFinalizerQ that moves freed objects
 * needing finalization to the FinalizeMeQ.  We then make a pass
 * over the FinalizeMeQ recursively marking everything on it.  It
 * is important that there is one pass over each queue: doing it
 * in a single pass over the HasFinalizerQ causes different
 * behavior as finalizable objects referenced by earlier finali-
 * zable objects will be marked and not be themselves finalized.
 * We could potentially allow interruption between the two passes.
 *
 * After this pass we are in a consistent state (only objects that
 * are unreachable have been put on the FinalizeMeQ), so we can
 * allow async GC to be interrupted.  Note that this means that
 * an object may be finalized earlier than its non-finalized peers
 * are freed.  Note that the FinalizeMeQ scan also marks objects that
 * were on the FinalizeMeQ when we started, which we needed to
 * do anyhow.  Removing an object from the HasFinalizerQ ensures
 * that we will never try to finalize it again.
 */
static void
prepareFinalization(void)
{
    SetLimits();
    JHandle *handle;

    deferredHandleTableOverflow = 0;

    if (HasFinalizerQ != NULL) {
        JHandle **prev = &HasFinalizerQ;
	/* last is a pointer to a pointer to original FinalizeMeQLast */
	JHandle **last;
	/* newLast becomes the new FinalizeMeQLast */
	JHandle *newLast = NULL;
	/* newFirst becomes the new FinalizeMeQ if latter was NULL */
	JHandle *newFirst = NULL;

	if (FinalizeMeQLast == NULL) {
	    sysAssert(FinalizeMeQ == NULL);
	    last = &FinalizeMeQLast;
	} else {
	    sysAssert(FinalizeMeQ != NULL);
	    last = &(FINALIZER_T(FinalizeMeQLast)->next);
	}

        /* Pass over HasFinalizerQ: */
        while ((handle = *prev) != NULL) {
	    sysAssert(ValidHandle(handle));
	    if (unhand(handle) && !IsMarked(handle) && ValidObject(unhand(handle))) {
		/* Important: Put in FinalizeMeQ in reverse order */
		/* Put the first one found in FinalizeMeQLast... */
		*prev = FINALIZER_T(handle)->next; /* Unlink */
		/* The first new JHandle becomes the new FinalizeMeQLast */
		if (newLast == NULL) {
		    newLast = handle;
		}
		newFirst = handle;
		FINALIZER_T(handle)->next = *last;
		*last = handle;
	    } else { 	/* still in use, so leave alone */
		prev = &(FINALIZER_T(handle)->next);
	    }
	}
	if (newLast != NULL) {	/* If we had a new one, update globals */
	    FinalizeMeQLast = newLast;
	    if (FinalizeMeQ == NULL) {
		FinalizeMeQ = newFirst;
	    }
	}
    }

    /* Trace the FinalizeMeQ: */
    handle = FinalizeMeQ;
    while (handle != NULL) {
	sysAssert(ValidHandle(handle));
	/* Mark the handle itself */
	MarkPtr(handle, HardMark);
	/* Recursively mark all objects referencable via the handle */
	markChildren(handle, (JHandle *) LIMIT, 0);
	handle = FINALIZER_T(handle)->next;
    }

    /*
     * Trace the BeingFinalized object (see comments in finalize.c).
     * It is essentially guaranteed that this object will have been
     * traced from the thread stack of the finalizing thread, but
     * we want to be clear that BeingFinalized has to be traced, and
     * we need BeingFinalized for synchronization purposes.
     */
    handle = BeingFinalized;
    if (handle != NULL) {
	sysAssert(ValidHandle(handle));
	if (IsMarked(handle) != HardMark) {
	    MarkPtr(handle, HardMark);
	    markChildren(handle, (JHandle *) LIMIT, 0);
	}
    }
    if (deferredHandleTableOverflow) {
	scanHandles();
    }
}

static int freeHandle(JHandle *hp)
{
	    int freed;
#ifdef PAGED_HEAPS
	    ChunkBlkP MOH_chunk = ChunkBase(hp);
#endif /* PAGED_HEAPS */
#ifdef TRACEFREE
	    if (tracegc) {
		jio_fprintf(stderr, " fr %s (%8X/%8X)\n",
			Object2CString(hp), hp, unhand(hp));
	    }
#endif /* TRACEFREE */
	    freed = obj_len((char *) unhand(hp) - sizeof(hdr));
#ifdef DEBUG
	    memset((char *) unhand(hp), 0x55, freed - sizeof(hdr));
#endif /* DEBUG */
	    FreeObjectCtr += freed;
	    sysAssert(FreeObjectCtr <= TotalObjectCtr);

#ifdef PAGED_HEAPS
	    {
		ChunkBlkP objP = ChunkBase(unhand(hp));
#ifdef LDEBUG
		checkChunk(objP);
#endif /* LDEBUG */
		objP->freeCnt += freed;
	    }
#endif /* PAGED_HEAPS */

	    nfreed++;
	    bytesfreed += freed;

#ifdef HPROF
            /* we should do the jvmpi notification before freeing h->obj
             * as we pass the actual obj pointer to the profiling agent. */
            if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_OBJ_FREE)) {
                jvmpi_free_object(hp);
            }
#endif /* HPROF */

	    obj_setfree((char *) unhand(hp) - sizeof(hdr));
	    unhand(hp) = 0;

	    PUSH_FREE_HANDLE(hp);
#ifdef PAGED_HEAPS
	    MOH_chunk->freeCnt += sizeof(JHandle);
#endif /* PAGED_HEAPS */

	    FreeHandleCtr += sizeof(JHandle);
	    sysAssert(FreeHandleCtr <= TotalHandleCtr);
	    return freed;
}

/*
 * Scan for remaining unmarked objects and free them.  All objects
 * that are to be finalized have been marked.
 */
static int
freeSweep(unsigned int free_space_goal)
{
    SetLimits();
    JHandle *hp;
    int do_compact_heap = 1;

#ifdef TRACEGC
    if (tracegc) {
	jio_fprintf(stderr, "freeSweep, free_space_goal:0x%x\n", free_space_goal);
    }
#endif

    MAP_OVER_HANDLES_FROM_START(hp)
	if (unhand(hp) && ValidObject(unhand(hp)) && !IsMarked(hp)) {
	    if (freeHandle(hp) >= free_space_goal) {
		do_compact_heap = 0;
	    }
	}
    END_MAP_OVER_HANDLES_FROM_START;

/* Reset allocation hands, so that freed memory gets used */
#ifdef PAGED_HEAPS
    {
	ChunkBlkP chunk = firstObjBlkP;
	do {
	    chunk->freePtr = chunk->startPtr; /* reset object allocation ptr */
	    chunk = chunk->nextPtr;	      /* walk thru object chunks */
	} while (chunk != firstObjBlkP);
    }

    /* Prefer low memory over high memory. */
    curHanBlkP = firstHanBlkP;
    curObjBlkP = firstObjBlkP;
#else /* not PAGED_HEAPS: */
    opoolhand = opmin;
#endif /* PAGED_HEAPS */

    return do_compact_heap;
}

static void
unloadUnusedClasses(void)
{
    ClassClass **pcb;
    int i;
    SetLimits();

    pcb = binclasses + nbinclasses - 1;
    for (i = nbinclasses; --i >= 0; pcb--) {
	ClassClass *cb = *pcb;
	if (cbIsInROM(cb) || IsMarked(cbHandle(cb))) {
  	    continue;
	}
	/* Mark the class block so that it won't be freed by freeSweep. */
	MarkPtr(cbHandle(cb), HardMark);
	/*
	 * Remove the class and add it to list of classes to be freed.
	 * It is illegal to call free() during single-threading mode,
	 * so classes are freed immediately after resuming multi-
	 * threading mode.
	 */
	cbSigners(cb) = (void *) classFreeList;
	classFreeList = cb;

	binclasses[i] = binclasses[--nbinclasses];
    }
}

static void
freeClasses()
{
    int classesFreed = 0;

    sysAssert(HEAP_LOCKED());

#ifdef HPROF
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_CLASS_UNLOAD)) {
        ClassClass *cb = classFreeList;
        while (cb) {
            jvmpi_free_class(cb);
#ifdef JDK12
	    cb = (ClassClass *)cbFinalizer(cb);
#else
	    cb = (ClassClass *)cbSigners(cb);
#endif
        }
    }
#endif /* HPROF */

#ifdef BREAKPTS
    if (debugging) {
	ClassClass *cb = classFreeList;
	while (cb) {
	    notify_debugger_of_class_unload(EE(), (JHandle *)cb);
#ifdef JDK12
	    cb = (ClassClass *)cbFinalizer(cb);
#else
	    cb = (ClassClass *)cbSigners(cb);
#endif
	}
    }
#endif

    while (classFreeList != 0) {
	ClassClass *cb = classFreeList;
	classFreeList = (ClassClass *) cbSigners(cb);

	if (verbosegc) {
	    char cbuf[128];
	    jio_fprintf(stderr, "<GC: freeing class %s>\n",
            	    classname2string(cbName(cb), cbuf, sizeof(cbuf)));
	}

	FreeClass(cb);
	freeHandle((JHandle *)cbHandle(cb));
	classesFreed++;
    }

    if (verbosegc && classesFreed > 0) {
	jio_fprintf(stderr, "<GC: unloaded and freed %d class%s>\n",
		classesFreed, classesFreed == 1 ? "" : "es");
    }
}

/*
 * COMPACT THE HEAP:
 * Concept: Assuming optimal packing, there are only two places that
 * holes, or bubbles in the heap can occur.  These are immediately before
 * each pinned range of memory, and at the tail of the heap (the last
 * range is always free).
 *
 * The goal of these schemes is to leave as many large free blocks available
 * as is possible.
 */

/*
 * Put a pointer to each object's handle in the first word of the object,
 * saving the first word of the object in the object pointer of the
 * handle. This way we can compact the heap looking only at the heap,
 * then go back and fix up the handles.	 (we need to scan all objects in
 * heap order, not handle order to effectively compact everything).
 */
#define obj_swapped(p)	    (obj_geth(p) & 2)
#define obj_setswapped(p)   (obj_geth(p) |= 2)
#define obj_clearswapped(p) (obj_geth(p) &= ~2)

#define handle_swapped(hp)	(IsMarked(hp) & 0x1)
#define handle_setswapped(hp)	(MarkPtr(hp, 0x1))
#define handle_clearswapped(hp) (ClearMarkPtr(hp, 0x1))

/*
 * This is the reverse of the above swap.  It is called once
 * on each allocated block and fixes the handle/pointer swap.
 */
#define REVERSE_SWAP(p) { \
    JHandle *hp, *nexthp; \
    sysAssert(obj_swapped(p)); \
    obj_clearswapped(p); \
    hp = *(JHandle **) (p + sizeof(hdr)); \
    while (handle_swapped(hp)) { \
		nexthp = (JHandle *) (unhand(hp)); \
		unhand(hp) = (OBJECT *) (p + sizeof(hdr)); \
		hp = nexthp; \
    } \
    *(OBJECT *) (p + sizeof(hdr)) = (OBJECT) unhand(hp); \
    unhand(hp) = (OBJECT *) (p + sizeof(hdr)); \
    handle_setswapped(hp); \
}

#ifdef PAGED_HEAPS /************ PAGED HEAPS: ********************/

/* Picks a chunk to free:
 * To do this right, we'd like to know how the OS allocates chunks for us,
 * but I haven't noticed any real order.  For now, only try to free the last
 * unpinned chunk that is small enough to make things worthwhile. Note that we
 * need to have already processed the free lists of the possible destiniation
 * chunks.  If we chose to evacuate the first chunk, we wouldn't have any free
 * lists to copy things to.  Should change compact heap to process all
 * destination chunks before the evacuation chunks.
 */
static ChunkBlkP
evacuationChunk()
{
    ChunkBlkP chunk = firstObjBlkP;
    ChunkBlkP highestChunk = NULL;	    /* highest unpinned chunk */

    /* Only try to evacuate if there might be enough room in the other chunks,
     * "*2" to make sure it we won't immediatly ask for the chunk back again. */
    do {
	if (((chunk->chunkFlags & CHUNK_PINNED) == 0) &&
	    (FreeObjectCtr > (chunk->allocSize * 2)))
	    highestChunk = chunk;
	chunk = chunk->nextPtr;
    } while (chunk != firstObjBlkP);

    return highestChunk;
}

/*
 * This algorithm walks a chunk from beginning to end, performing
 * several operations:
 *  0. Pick a destination chunk for the chunk being compacted.	Usually we
 *     compact into the same chunk, but we can choose some chunks as candidates
 *     for evacuation if there are no pinned objects in the chunk.  Currently
 *     evacuates the last unpinned chunk, and only when async_call is true.
 *  1. Coalesce adjacent free blocks.
 *  2. Record free blocks in linked lists based off of object chunk's "freePtr".
 *     This list may be handy globally, but currently only created during
 *     compactHeap(). The list is sorted in address order. Be careful not to
 *     swap free objects, because swapping and free links use the same word
 *     in an object! Also note that the link is stored where we used to store
 *     a debugging check word (0x55555555), so the check word is now stored
 *     in the next word if the object is big enough.
 *  3. Copy unpinned, allocated blocks down into the lowest free
 *     block available.
 *  4. Free object chunk if no objects remain.
 *
 */

#define free_list_next(p)   \
        (*((unsigned char **)((unsigned char *)(p) + sizeof(hdr))))

/* Compact all data in this chunk.  If we want to evacuate a chunk, pass in
 *  a non-NULL value for destChunk.
 */
static int
compactChunk(int async_call, ChunkBlkP chunk, ChunkBlkP destChunk)
{
    unsigned char *p, *next, *prev;
    unsigned long plen;
    int blocks_moved = 0;
    /* use last_free as a pointer to the tail of the free list */
    unsigned char *last_free = 0;
    /* flags when we are evacuating to another chunk. */
    ChunkBlkP evacChunk = NULL;
#ifdef HPROF
    bool_t jvmpi_obj_move_on = JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_OBJ_MOVE);
#endif

#define RemoveLink(linkp, prevp, nextp) {	\
    if (prevp != NULL) {			\
	free_list_next(prevp) = (nextp);	\
    } else {					\
	destChunk->freePtr = (nextp);		\
    }						\
    if (last_free == (linkp)) {			\
	last_free = (nextp) ? (nextp) : (prevp);\
    }						\
}

#ifdef LDEBUG
    checkChunk(chunk);
    if (destChunk) {
	checkChunk(destChunk);
    }
#endif

    /* note that p points to headers, not aligned objects */
    p = (unsigned char *) chunk->startPtr;
    /* from now the freePtr field is a linked free list, so initialize list: */
    chunk->freePtr = NULL;
    prev = 0;
    if (destChunk) {
	evacChunk = chunk;
    } else {
	destChunk = chunk;		    /* copy into itself */
    }

#ifdef TRACEGC
    if (tracegc && evacChunk) {
	jio_fprintf(stderr, "Going to evacuate chunk 0x%lx to chunk 0x%lx\n",
		evacChunk, destChunk);
    }
#endif /* TRACEGC */

    /*
     *	- look at every object in this chunk
     */
    while (p < (unsigned char *) chunk->endPtr) {
	plen = obj_len(p);
	next = p + plen;

#ifdef LDEBUG
	sysAssert(ValidObject(p + sizeof(hdr)));
	sysAssert(ValidObject(next + sizeof(hdr)));
	sysAssert(plen > 0);
#endif

	if (async_call && sysInterruptsPending()) {
	    /*
	     * Cycle through the remaining blocks as quickly as possible,
	     * reversing all of the handle pointers
	     */
#ifdef TRACEGC
	    if (tracegc) {
		jio_fprintf(stderr,
			"compactHeap bailing due to interrupted async gc\n");
	    }
#endif /* TRACEGC */
	    do {
		while (p < chunk->endPtr) {
		    if (!obj_free(p) && obj_swapped(p)) {
			REVERSE_SWAP(p);
		    }
		    p += obj_len(p);
		 }
		 chunk = chunk->nextPtr;
	     } while (firstObjBlkP != chunk);
	     return -1;
	 }

	/* There are three cases here (in order):
	 * p is free, p is used and not pinned, and p is used and pinned.
	 */
	if (obj_free(p)) {
	    int old_len = plen;
	    /* 1. Coalesce adjacent free blocks */

	    while (next && obj_free(next)) {
#ifdef LDEBUG
		sysAssert(obj_len(next) > 0);
#endif

		plen += obj_len(next);
		next += obj_len(next);
		blocks_moved++;
	    }
	    obj_setlf(p, plen, 1);
#ifdef DEBUG
	    if ((old_len <= OBJECTGRAIN) && (plen > OBJECTGRAIN)) {
		((long *) p)[CHECK_WORD_INDEX] = 0x55555555;
	    }
#endif /* DEBUG */

#ifdef TRACECOMPACT
	    if (tracegc) {
		if (plen > old_len) {
		    jio_fprintf(stderr,
		       "coaleced 0x%lx bytes into 0x%lx free bytes at 0x%lx\n",
			    (plen - old_len), old_len, p);
		} else {
		    jio_fprintf(stderr, "found	0x%lx free bytes at 0x%lx\n",
			    plen, p);
		}
	    }
#endif /* TRACECOMPACT */

	    /* 2. Record p in the free list in address order.  Since we are
	     * walking the heap in address order, simply place p at the end
	     * of the list.
	     */
	    if (last_free == NULL) {
		chunk->freePtr = p;
	    } else {
		free_list_next(last_free) = p;
	    }
	    free_list_next(p) = NULL;
	    last_free = p;
	    prev = p;
	} else if (!IsMarked(p + sizeof(hdr)) && !obj_pinned(p)) {
	    unsigned char *freep;		/* loop variable */
	    unsigned char *foundp = NULL;
	    unsigned char *free_before_foundp;	/* foundp's previous pointer */

	    /* 3. Copy p down into the lowest available free block */

	    /*
	     * Find lowest block that is big enough to hold the block.
	     * Free list is in address order, so just return the first one.
	     * Ignore 'prev' as we will handle it with a simpler case.
	     * If we are evacuationg this chunk, first search all other chunks
	     * for free space.	Failing that, compact into itself.
	     */
	    if (evacChunk) { /* first look other chunks for free space */
		ChunkBlkP someDestChunk = destChunk;
		do {
		    free_before_foundp = NULL;
		    if ((someDestChunk != chunk) &&
			(someDestChunk->freeCnt >= plen))
			for (freep = someDestChunk->freePtr;
			     freep != NULL;
			     freep = free_list_next(freep)) {
			    if (obj_len(freep) >= plen) {
			    	/* test was GT, but that didn't seem right */
				foundp = freep;
				break;
			    }
			    free_before_foundp = freep;
			}
		    if (!foundp) {
			someDestChunk = someDestChunk->nextPtr;
#ifdef TRACEGC
			if (tracegc) {
			    jio_fprintf(stderr,
			    "Try to evacuate rest of chunk 0x%lx to 0x%lx\n",
				    chunk, someDestChunk);
			}
#endif /* TRACEGC */
		    }
		} while (!foundp && (someDestChunk != destChunk)
		         && (someDestChunk < chunk));

		if (foundp) {
		    destChunk = someDestChunk;
#ifdef LDEBUG
		    checkChunk(destChunk);
		    /* only chunks lower in memory have their free lists set. */
		    sysAssert(someDestChunk < chunk);
		    sysAssert(ValidObject(foundp + sizeof(hdr)));
#endif
		} else {
    		    /* once we fail to evacaute a block, give up. */
		    evacChunk = NULL;
		    destChunk = chunk;
		}
	    }

	    if (!foundp) {
		free_before_foundp = NULL;
		for (freep = chunk->freePtr;
		     freep != NULL;
		     freep = free_list_next(freep)) {
		    if (obj_len(freep) >= plen) {
		        /* test was GT, but that didn't seem right */
			foundp = freep;
			break;
		    }
		    free_before_foundp = freep;
		}
	    }

	    /* did we find a block? */
	    if (foundp && foundp != prev) {
		unsigned char *oldp = p;
		/* remember this stuff because memmove will bash it: */
		unsigned long lowest_len = obj_len(foundp);
		unsigned char *next_freep = free_list_next(foundp);
		/*
		 * If a slot is found, move the block, and mark the old
		 * slot as free.  If the free space was bigger than we needed,
		 * break off a new chunk.
		 */
#ifdef TRACECOMPACT
		if (tracegc) {
		    jio_fprintf(stderr, "move1 0x%lx bytes from 0x%lx to 0x%lx\n",
		            plen, p, foundp);
		}
#endif /* TRACECOMPACT */
		sysAssert(obj_free(foundp));
		memmove(foundp, p, plen);
#ifdef HPROF
		if (jvmpi_obj_move_on) {
                    jvmpi_move_object(p + sizeof(hdr),
                                      foundp + sizeof(hdr));
                }
#endif /* HPROF */
		blocks_moved++;
		obj_setfree(p);
		sysAssert(lowest_len >= plen);
#ifdef DEBUG
		if (plen > OBJECTGRAIN) {
		    ((long *) p)[CHECK_WORD_INDEX] = 0x55555555;
		}
#endif /* DEBUG */
		p = foundp;

		/* it was bigger than we needed, break off a piece. */
		if (lowest_len > plen) {
		    unsigned char *fp = foundp;
		    fp += plen;
		    obj_setlf(fp, (lowest_len - plen), 1);
#ifdef DEBUG
#ifdef LDEBUG
		    if (!((lowest_len - plen) & (OBJECTGRAIN - 1)) == 0) {
		        /* size still aligned? */
			sysAssert(((lowest_len - plen)
                                   & (OBJECTGRAIN - 1)) == 0);
		    }
		    sysAssert(ValidObject(fp + sizeof(hdr)));
#endif /* LDEBUG */
		    if ((lowest_len - plen) > OBJECTGRAIN) {
			((long *) fp)[CHECK_WORD_INDEX] = 0x55555555;
		    }
#endif /* DEBUG */
		    /* link fp to the next free block, and replace foundp
		     * in the free list with fp */
		    free_list_next(fp) = next_freep;
		    RemoveLink(foundp, free_before_foundp, fp);
		} else {
		    /* remove foundp from the free list */
		    RemoveLink(foundp, free_before_foundp, next_freep);
		}

		if (evacChunk) {
		    destChunk->freeCnt -= plen;
		    evacChunk->freeCnt += plen;
		}

		if (prev && obj_free(prev)) {
		    /* back up by two, to coalece adjacent free blocks */
		    next = prev;
		    prev = 0;
		} else {
		    /* look at this block again as a free block */
		    next = oldp;
		}
	    } else {
		/*
		 * no free slot found or it is the previous slot.  See if prev
		 * is free, and if so, just slide this range back.  Note that
		 * the previous slot does not need to bigger than current
		 * object, because we are just sliding.
		 */
		if (prev && obj_free(prev)) {
		    unsigned char *next_freep = free_list_next(prev);
#ifdef TRACECOMPACT
		    if (tracegc) {
			jio_fprintf(stderr, "move2 0x%lx bytes from 0x%lx to 0x%lx\n",
			        plen, p, prev);
		    }
#endif /* TRACECOMPACT */
		    memmove(prev, p, plen);
#ifdef HPROF
		    if (jvmpi_obj_move_on) {
			int len = plen;
			int dist = p - prev;
			unsigned char * lp = prev;
			while (len > 0) {
			    int l = obj_len(lp);
			    sysAssert(l > 0);
			    sysAssert(obj_swapped(lp));
			    sysAssert(!obj_free(lp));
			    jvmpi_move_object(lp + sizeof(hdr) + dist,
					      lp + sizeof(hdr));
			    lp += l;
			    len -= l;
			}
		    }
#endif /* HPROF */
		    blocks_moved++;
		    next = prev + plen;
		    obj_setlf(next, (p - prev), 1);
		    sysAssert(ValidObject(prev + sizeof(hdr)));
		    sysAssert(next_freep == NULL);
		    sysAssert(last_free == prev);
#ifdef DEBUG
		    if ((p - prev) > OBJECTGRAIN) {
			((long *) next)[CHECK_WORD_INDEX] = 0x55555555;
		    }
#endif /* DEBUG */
		    p = prev;
		    /* remove prev from the free list.	prev must be the last
		     * thing on the list */
		    if (foundp != prev) {
		    	/* we need to find the free block before prev */
			free_before_foundp = NULL;
			for (freep = chunk->freePtr;
			     freep != prev;
			     freep = free_list_next(freep)) {
			    free_before_foundp = freep;
			}
		    }
		    if (free_before_foundp) {
			free_list_next(free_before_foundp) = NULL;
		    } else {
			chunk->freePtr = next_freep;
		    }
		    last_free = free_before_foundp; /* DRW double check this! */
		} else {
		    prev = p;
		}
	    }
	} else {
	    prev = p;
	}

	/* at this point, p needs to == the place the object was moved to,
	 * so we can unswap it */
	if (next > p && !obj_free(p)) {
	    if (obj_swapped(p)) {
		REVERSE_SWAP(p);
	    }
	}
	p = next;
    }
#ifdef TRACEGC
    if (tracegc) {
	unsigned char* freep;
	for (freep = chunk->freePtr;
	     freep != NULL;
	     freep = free_list_next(freep)) {
	    jio_fprintf(stderr, "free block 0x%lx  holds 0x%lx bytes\n",
	            freep, obj_len(freep));
	}
    }
#endif /* TRACEGC */
    return blocks_moved;
}

static int
compactHeap(int async_call, unsigned char **last_freep,
            unsigned int free_space_goal)
{
    register JHandle *hp;
    ChunkBlkP chunk;
    ChunkBlkP evacChunk = NULL;
    ChunkBlkP destChunk = NULL;
    int blocks_moved = 0;
    unsigned char *last_free = 0;
#ifdef DEBUG
    long start_time = now();
#endif
#ifdef HPROF
    bool_t jvmpi_obj_move_on = JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_OBJ_MOVE);
#endif


    /*
     * Put a pointer to each object's handle in the first word of the object,
     * saving the first word of the object in the object pointer of the
     * handle. This way we can compact the heap looking only at the heap,
     * then go back and fix up the handles.  (we need to scan all objects in
     * heap order, not handle order to effectively compact everything).
     * if multiple handles point to one object, handles are linked together.
     * DRW: To Do - Shouldn't swap pinned objects!
     * Sheng: pinned objects are no longer swapped.
     */

    MAP_OVER_HANDLES_FROM_START(hp)
	if (ValidObject(unhand(hp))) {
	    OBJECT T = *unhand(hp);
	    unsigned char *p = (unsigned char *)(unhand(hp)) - sizeof(hdr);
	    if (!IsMarked(p + sizeof(hdr)) && !obj_pinned(p)) {
	        *unhand(hp) = (OBJECT) hp;
		unhand(hp) = (OBJECT *) T;
		if (obj_swapped(p)) {
		    handle_setswapped(hp);
		} else {
		    obj_setswapped(p);
		    handle_clearswapped(hp);
		}
	    }
	}
    END_MAP_OVER_HANDLES_FROM_START;

    /* There are many possible preconditions to freeing a chunk, but this is
     * the most basic: Don't evacute if gc-ing due to memory request, or if
     * there's only one chunk.
     */
    if ((firstObjBlkP != lastObjBlkP) &&
	(async_call || (free_space_goal == 0xFFFFFFFF))) {
	evacChunk = evacuationChunk();
	if (evacChunk) {
	    /* becuase of the way we calc the free lists, the dest needs to be
	     * before the evacChunk, so we can't evac the first chunk. */
	    if (evacChunk != firstObjBlkP) {
		destChunk = firstObjBlkP; /* not great, but good guess */
	    } else {
		evacChunk = NULL;
	    }
#ifdef TRACEGC
	    if (tracegc) {
		jio_fprintf(stderr, "Attempting to free block 0x%lx\n", evacChunk);
	    }
#endif /* TRACEGC */
	}
    }

    /*
     * The main loop - look at every object in every chunk
     */
    chunk = firstObjBlkP;		 /* iterates over all object chunks */
    do {
	int result = compactChunk(async_call, chunk,
	                          (chunk == evacChunk) ? destChunk : NULL);
	if (result >= 0) {
	    blocks_moved += result;
	} else {	/* bail out due to async gc getting interupted */
	    *last_freep = 0;
	    break;	/* still need to do some clean up */
	}
	chunk = chunk->nextPtr;			/* advance to next chunk */
    } while (chunk != firstObjBlkP);

    if (evacChunk) {
	if (evacChunk->freeCnt == evacChunk->allocSize) {
	    shrinkObjectSpace(evacChunk);
	} else {
#ifdef TRACEGC
	    if (tracegc) {
		jio_fprintf(stderr,
"Tried to free block 0x%lx, but it still is using 0x%lx bytes \n",
			evacChunk, evacChunk->allocSize - evacChunk->freeCnt);
	    }
#endif /* TRACEGC */
	}
    }

    /* turn free lists back into free hints the way searchChunk expects */
    chunk = firstObjBlkP;
    do {
	if (!chunk->freePtr) {	/* non-NULL free lists make good hints, */
	    chunk->freePtr = chunk->startPtr;	/* but NULL hints are bad. */
	}
	chunk = chunk->nextPtr;
    } while (chunk != firstObjBlkP);

    sysAssert(last_free == 0 || obj_free(last_free));
    *last_freep = last_free;

#ifdef DEBUG
    if (verbosegc) {
	long total_time = now() - start_time;
	jio_fprintf(stderr, "<GC: compactHeap took %d ms, blocks_moved=0x%lx>\n",
		total_time, blocks_moved);
    }
#endif

    return async_call ? blocks_moved : 0;
}

#else  /************ CONTIGUOUS HEAPS: ********************/

/*
 * Concept: Assuming optimal packing, there are only two places that
 * holes, or bubbles in the heap can occur.  These are immediately before
 * each pinned range of memory, and at the tail of the heap (the last
 * range is always free).
 *
 * This algorithm walks the heap from beginning to end, performing
 * several operations:
 *  1. Coalesce adjacent free blocks.
 *  2. Record bubbles in the free_blocks[] array.  This is an array
 *     containing the biggest free blocks found.
 *  3. Copy unpinned, allocated blocks down into the lowest free
 *     block available.
 *
 * The goal of this scheme is to leave as many large free blocks available
 * as is possible.
 */
static int
compactHeap(int async_call, unsigned char **last_freep,
            unsigned int free_space_goal)
{
    SetLimits();
    register JHandle *hp;
    unsigned char *p, *next, *prev;
#define NUMFREEBLOCKS 16
    unsigned char *free_blocks[NUMFREEBLOCKS];
    unsigned int plen;
    int blocks_moved = 0;
    unsigned char *last_free = 0;
#ifdef DEBUG
    long start_time = now();
#endif
#ifdef HPROF
    bool_t jvmpi_obj_move_on = JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_OBJ_MOVE);
#endif

    p = opmin;
    prev = 0;
    memset((char *)free_blocks, 0, sizeof(free_blocks));

#ifdef HPROF
    MarkPtr(unhand(threadSelf()), HardMark);
	/* hprof gets very upset if EE() can't find the current thread,
	   so we make sure that it's marked.  */
#endif
    MAP_OVER_HANDLES_FROM_START(hp)
    if (ValidObject(unhand(hp))) {
	OBJECT T = *unhand(hp);
	unsigned char *p = (unsigned char *)(unhand(hp)) - sizeof(hdr);
	/* Pinned objects are not swapped. */
	if (!IsMarked(p + sizeof(hdr)) && !obj_pinned(p)) {
	    *unhand(hp) = (OBJECT) hp;
	    unhand(hp) = (OBJECT *) T;
	    if (obj_swapped(p)) {
	        handle_setswapped(hp);
	    } else {
	        obj_setswapped(p);
		handle_clearswapped(hp);
	    }
	}
    }
    END_MAP_OVER_HANDLES_FROM_START;

    /*
     * The main loop - look at every object in the pool
     */
    while (p < opmax) {
        plen = obj_len(p);
	next = p + plen;

	if (async_call && sysInterruptsPending()) {
	    blocks_moved = 0;
	    /*
	     * Cycle through the remaining blocks as quickly as possible,
	     * reversing all of the handle pointers
	     */
	    while (p < opmax) {
	        if (!obj_free(p) && obj_swapped(p)) {
		    REVERSE_SWAP(p);
		}
		p += obj_len(p);
	    }
	    last_free = 0;
	    break;
	}

	if (obj_free(p)) {
	    unsigned int i, smallest = 0, len = (unsigned) ~0;

	    /* 1. Coalesce adjacent free blocks */

	    while (next && obj_free(next)) {
	        plen += obj_len(next);
		sysAssert(obj_len(next) != 0);
		next += obj_len(next);
		blocks_moved++;
	    }
	    obj_setlf(p, plen, 1);

	    /*
	     * Record the most recently seen free block.  When compaction
	     * finishes this will be the highest free block.  It is very
	     * likely that it will span to the end of the object space,
	     * giving us the option of freeing some of it.
	     */
	    last_free = p;

	    /* 2. Record the fact that p is a free block */

	    /*
	     * Search for the smallest sized free range, and if it
	     * is smaller than the size of p, replace it with p.
	     * There may be empty slots in the buffer. Use them
	     * if possible.
	     */
	    for (i = 0; i < NUMFREEBLOCKS; i++) {
	        sysAssert(free_blocks[i] == 0 || obj_free(free_blocks[i]));
		if (free_blocks[i] == 0 || free_blocks[i] == p) {
		    len = 0;
		    smallest = i;
		    break;
		} else if (obj_len(free_blocks[i]) < len) {
		    smallest = i;
		    len = obj_len(free_blocks[i]);
		}
	    }

	    if (len < obj_len(p)) {
	        free_blocks[smallest] = p;
	    }

	    prev = p;

	} else if (!IsMarked(p + sizeof(hdr)) && !obj_pinned(p)) {
	    int i, lowest, prev_block = -1;
	    unsigned int len, lowest_len;
	    unsigned char *lowest_addr;

	    /* 3. Copy p down into the lowest available free block */

	    /*
	     * Find lowest block that is big enough to hold the block.
	     * Ignore 'prev' as we will handle it with a simpler case.
	     */
	    lowest = -1;
	    lowest_len = 0;
	    lowest_addr = opmax;
	    for (i = 0; i < NUMFREEBLOCKS; i++) {
	        if (!free_blocks[i]) {
		    continue;
		}
		if (free_blocks[i] == prev) {
		    prev_block = i;
		}
		len = obj_len(free_blocks[i]);
		if (len > plen && lowest_addr > free_blocks[i]) {
		    lowest = i;
		    lowest_addr = free_blocks[i];
		    lowest_len = len;
		}
	    }

	    /* did we find a slot? */
	    if (prev && lowest >= 0 && lowest != prev_block) {
	        unsigned char *oldp = p;
		/*
		 * If a slot is found, move the block, and mark the old
		 * slot as free.  If the free space was bigger than we needed,
		 * break off a new chunk.
		 */
		sysAssert(obj_free(free_blocks[lowest]));
		memmove(free_blocks[lowest], p, plen);
#ifdef HPROF
                if (jvmpi_obj_move_on) {
                    jvmpi_move_object(p + sizeof(hdr),
                                      free_blocks[lowest] + sizeof(hdr));
                }
#endif /* HPROF */
		blocks_moved++;

		obj_setfree(p);
#ifdef DEBUG
		((long *) p)[1] = 0x55555555;
#endif /* DEBUG */
		p = free_blocks[lowest];

		/* it was bigger than we needed, break off a piece. */
		if (lowest_len > plen) {
		    unsigned char *fp = free_blocks[lowest];
		    fp += plen;
		    obj_setlf(fp, (lowest_len - plen), 1);
#ifdef DEBUG
		    ((long *) fp)[1] = 0x55555555;
#endif /* DEBUG */
		    if (prev <= free_blocks[lowest]) {
		        free_blocks[lowest] = 0;
		    } else {
		        free_blocks[lowest] = fp;
		    }
		} else {
		    free_blocks[lowest] = 0;
		}

		if (obj_free(prev)) {
		    next = prev;
		    prev = 0;
		} else {
		   next = oldp;
		}
	    } else {
	        /*
		 * no free slot found or it is the previous slot.  See if prev
		 * is free, and if so, just slide this range back.
		 */
	        if (prev && obj_free(prev)) {
		    memmove(prev, p, plen);
#ifdef HPROF
		    if (jvmpi_obj_move_on) {
			int len = plen;
			int dist = p - prev;
			unsigned char * lp = prev;
			while (len > 0) {
			    int l = obj_len(lp);
			    sysAssert(l > 0);
			    sysAssert(obj_swapped(lp));
			    sysAssert(!obj_free(lp));
			    jvmpi_move_object(lp + sizeof(hdr) + dist,
					      lp + sizeof(hdr));
			    lp += l;
			    len -= l;
			}
		    }
#endif /* HPROF */
		    blocks_moved++;
		    next = prev + plen;
		    obj_setlf(next, (p - prev), 1);
#ifdef DEBUG
		    ((long *) next)[1] = 0x55555555;
#endif /* DEBUG */
		    p = prev;
		    if (prev_block >= 0) {
		        free_blocks[prev_block] = 0;
		    }

		} else {
		    prev = p;
		}
	    }
	} else {
	    prev = p;
	}

	if (next > p && !obj_free(p) && obj_swapped(p)) {
	    REVERSE_SWAP(p);
	}
	p = next;
    }

#ifdef DEBUG
    if (verbosegc) {
	long total_time = now() - start_time;
	jio_fprintf(stderr, "<GC: compactHeap took %d ms, blocks_moved=0x%lx>\n",
		total_time, blocks_moved);
    }
#endif

    sysAssert(last_free == 0 || obj_free(last_free));
    *last_freep = last_free;

    return async_call ? blocks_moved : 0;
}

#endif  /************ END CONTIGUOUS HEAPS ********************/

/*
 * Garbage collect memory.  If 'async_call' is true, this function is being
 * called from the asych GC thread, and should immediately exit whenever
 * sysInterruptsPending() returns true.  The 'free_space_goal' is
 * the amount of memory in a single block which the caller is interested
 * in having free.  If such a block does not exist, compact the heap.
 *
 * gc0_locked() should only be called when holding the heap lock and the
 * finalization queue locks.  gc() or realObjAlloc() acquires the heap
 * lock, then calls gc0() which grabs the finalization locks and calls
 * gc_locked(), which does system-specific tasks then calls gc0_locked().
 */

static long init_time;
static long scan_time;
static long sweep_time;
static long compact_time;
static long total_time;

static int gc_finished; /* whether GC finished. */

unsigned char *
gc0_locked(int async_call, unsigned int free_space_goal)
{
    /* SPARC only: "ta ST_FLUSH_WINDOWS" done by "gc" stub in machgc.s */
    int do_compact_heap;
    long start_time = now();

    unsigned char *last_free = 0;
#ifdef ALLOC_BARRIERS
    SetLimits();
#endif /* ALLOC_BARRIERS */

    init_time = scan_time = sweep_time = compact_time = 0;
    gc_finished = 0;

    sysAssert(HEAP_LOCKED());
    sysAssert(HASFINALQ_LOCKED());
    sysAssert(FINALMEQ_LOCKED());
    sysAssert(QUEUE_LOCKED());
    sysAssert(STRINGHASH_LOCKED());
    sysAssert(BINCLASS_LOCKED());
    sysAssert(GLOBALREF_LOCKED());

    nfreed = 0;
    bytesfreed = 0; /* used by realObjAlloc to decide gc vs. allocate more */

    setClassGCMode();
    clearMarkBits();
    if (async_call && sysInterruptsPending()) {
	goto unlock;
    }

#ifdef ALLOC_BARRIERS
    if (ValidHandle(PRRed)) {MarkPtr(PRRed, HardMark);}
    if (ValidHandle(PRYellow)) {MarkPtr(PRYellow, HardMark);}
#endif /* ALLOC_BARRIERS */

    markbitsInUse = TRUE;
    MARK(GCTHREAD);
    scanThreads();
    if (async_call && sysInterruptsPending()) {
        goto unlock;
    }

    scanInternedStrings();
    if (async_call && sysInterruptsPending()) {
	goto unlock;
    }

    MARK(GCJNIGLOBALREFS);
    scanJNIGlobalRefs();
    if (async_call && sysInterruptsPending()) {
	goto unlock;
    }

    MARK(GCHEAP);
    scanHeap();
    init_time = now() - start_time;
    if (async_call && sysInterruptsPending()) {
	goto unlock;
    }

    MARK(GCMARKSTICKYCLASSES);
    markStickyClasses();
    markPrimitiveClasses();
    if (async_call && sysInterruptsPending()) {
	goto unlock;
    }


    MARK(GCROMCLASSES);
    if (numRomClasses > 0)
        markROMClassStatics();

    MARK(GCHANDLES);
    scanHandles();

    scan_time = now() - start_time - init_time;

#ifdef HPROF
    if (jvmpi_event_flags && jvmpi_in_dump_heap_flag) {
        jvmpi_gc_dump_heap();
    }
#endif /* HPROF */

    if (async_call && sysInterruptsPending()) {
	goto unlock;
    }

    clearJNIWeakRefs();
    if (async_call && sysInterruptsPending()) {
        goto unlock;
    }

    MARK(GCFINAL);
    prepareFinalization();
    if (async_call && sysInterruptsPending()) {
        goto unlock;
    }

    MARK(GCCLASSES)
    if (classgc) {
	if (allowClassGCNow) {
	    unloadUnusedClasses();
	}
    }

/* We do not do a interrupt-pending check here:

    if (async_call && sysInterruptsPending()) {
	goto unlock;
    }

   Since from now on unused classes will ultimately be freed in freeClasses,
   we must make sure all unused objects are freed as well. The system must
   never enter a situation where normal objects point to freed objects.
 */

    MARK(GCSWEEP);
    do_compact_heap = freeSweep(free_space_goal);
    sweep_time = now() - start_time - init_time - scan_time;
    if (async_call && sysInterruptsPending()) {
	goto unlock;
    }

    MARK(COMPACT);
#ifdef PDEBUG    /* always compact...*/
    do_compact_heap = 1;
#endif

#ifdef HPROF
    /* don't compact the heap during a dump, as we don't want the profiler
     * agent to get object_move events for objects already recorded in the
     * dump */
    if (jvmpi_event_flags && jvmpi_in_dump_heap_flag) {
        do_compact_heap = 0;
    }
#endif

    if (do_compact_heap) {
	/* The heap is compacted in adress order, so there shouldn't be
	 * any benefit to doing it multiple times */
	compactHeap(async_call, &last_free, free_space_goal);
    }
    compact_time = now() - start_time - init_time - scan_time - sweep_time;
    gc_finished = 1;

  unlock:
    total_time = now() - start_time;

#ifdef INCLUDE_PROF
    if (java_monitor && !async_call) {
	ExecEnv *ee = EE();
	if (ee && ee->current_frame && ee->current_frame->current_method) {
	    java_mon(ee->current_frame->current_method,
		    (struct methodblock *)-1, now() - start_time);
	}
    }
#endif /* INCLUDE_PROF */
    markbitsInUse = FALSE;

#ifdef ALLOC_BARRIERS
    /* Only take GC's that do something into account */
    if (nfreed > 0) {
	GCGenCtr++;
    }
#endif /* ALLOC_BARRIERS */
    return last_free;

}

static
void printGCMessages(int async_call)
{
    extern int no_verifiers;
    char async_msg[30];

    if (async_call) {
        jio_snprintf(async_msg, sizeof(async_msg), "(async%s)",
		     gc_finished ? "" : ", interrupted");
    } else
        async_msg[0] = 0;

    if (!allowClassGCNow) {
        jio_fprintf(stderr,
		    "<GC%s: not GC'ing classes: %d verifier%s running>\n",
		    async_msg, no_verifiers, (no_verifiers > 1) ? "s" : "");
    }
    jio_fprintf
        (stderr,
	 "<GC%s: freed %d objects, %d bytes in %d ms, %d%% free (%ld/%ld)>\n",
	 async_msg, nfreed, bytesfreed, total_time,
	 (int)(100 * (double)FreeObjectCtr / (double)TotalObjectCtr),
	 FreeObjectCtr, TotalObjectCtr);
    jio_fprintf(stderr,
		"  <GC: init&scan: %d ms, scan handles: %d ms,"
		" sweep: %d ms, compact: %d ms>\n",
		init_time, scan_time, sweep_time, compact_time);
}

/*
 * Locking and unlocking special system monitors for GC
 */

void
LOCK_GC_LOCKS()
{
    sysAssert(HEAP_LOCKED());

    /*
     * To ensure consistency of locked data structures, we should not
     * be holding any of the following locks when GC occurs.  Special
     * dispensation may be given to locks that protect code that is
     * proven safe at points of possible recursive entry.
     */
    sysAssert(!HASFINALQ_LOCKED());
    sysAssert(!FINALMEQ_LOCKED());
    sysAssert(!QUEUE_LOCKED());
    sysAssert(!STRINGHASH_LOCKED());
    sysAssert(!BINCLASS_LOCKED());
    sysAssert(!GLOBALREF_LOCKED());

    HASFINALQ_LOCK();
    FINALMEQ_LOCK();
    QUEUE_LOCK();
    STRINGHASH_LOCK();
    BINCLASS_LOCK();
    GLOBALREF_LOCK();

#ifdef HPROF
    GC_DISABLE_LOCK(EE());
    while (gc_disable_count) {
        if (SYS_INTRPT == sysMonitorWait(_gc_disable_lock,
                                         SYS_TIMEOUT_INFINITY, TRUE)) {
	    SignalError((struct execenv *) THREAD(threadSelf())->eetop,
			JAVAPKG "InterruptedException",
			"operation interrupted");
	    return;
        }
    }

#endif /* HPROF */
}

void
UNLOCK_GC_LOCKS()
{
    sysAssert(HEAP_LOCKED());

    /* Make sure we are holding the things we want to unlock */
    sysAssert(HASFINALQ_LOCKED());
    sysAssert(FINALMEQ_LOCKED());
    sysAssert(QUEUE_LOCKED());
    sysAssert(STRINGHASH_LOCKED());
    sysAssert(BINCLASS_LOCKED());
    sysAssert(GLOBALREF_LOCKED());

#ifdef HPROF
    GC_DISABLE_UNLOCK(EE());
#endif /* HPROF */

    /* Remember that unlocking order must be the reverse of lock order! */
    GLOBALREF_UNLOCK();
    BINCLASS_UNLOCK();
    STRINGHASH_UNLOCK();
    QUEUE_UNLOCK();
    FINALMEQ_UNLOCK();
    HASFINALQ_UNLOCK();
}

/*
 * Interface to synchronous garbage collection from other memory manage-
 * ment code.  This is called call when memory allocation fails.  We
 * expect to have the heap lock when it is called.
 */

unsigned char *gc_locked(int, unsigned int);	/* Call holding locks! */

unsigned char *
gc0(int async_call, unsigned int free_space_goal)
{
    unsigned char *last_free = 0;

#ifdef TRACEGC
    if (tracegc) {
	jio_fprintf(stderr, "*** gc ***	 goal=%ld\n", free_space_goal);
	DumpHeapInfo();
    }
#ifdef LDEBUG
    verifyHandleCtrs();
    verifyObjectCtrs();
    validatePool(0);
#endif /* LDEBUG */
#endif /* TRACEGC */

    sysAssert(HEAP_LOCKED());
    LOCK_GC_LOCKS();

#ifdef HPROF
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_GC_START)) {
        jvmpi_gc_start();
    }
#endif /* HPROF */

    if (sysThreadSingle() == SYS_OK) {
	last_free = gc_locked(async_call, free_space_goal);
    }
    sysThreadMulti();
    freeClasses();
    /* safe to print in multi-threaded mode */
    if (verbosegc)
	printGCMessages(FALSE);

    /* Notify the finalizer thread if objects pending finalization */
    if (FinalizeMeQ != NULL) {
	FINALMEQ_NOTIFY();
    }

    heap_memory_changes = 0;

#ifdef HPROF
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_GC_FINISH)) {
        jvmpi_gc_finish(TotalHandleCtr - FreeHandleCtr,
                        TotalObjectCtr - FreeObjectCtr,
                        TotalObjectCtr);
    }
#endif /* HPROF */

    UNLOCK_GC_LOCKS();

#ifdef TRACEGC
    if (tracegc) {
	jio_fprintf(stderr, "** gc done **\n");
	DumpHeapInfo();
    }
#ifdef LDEBUG
    verifyHandleCtrs();
    verifyObjectCtrs();
    validatePool(0);
#endif /* LDEBUG */
#endif /* TRACEGC */

    return last_free;
}

/*
 * User interface to synchronous garbage collection.  This is called
 * by an explicit call to GC.  Asynchronous GC has its own entry point.
 * At the call we do not already have the heap lock, so grab it.
 */
void
gc(int async_call, unsigned int free_space_goal)
{
    HEAP_LOCK();
    (void) gc0(async_call, free_space_goal);
    HEAP_UNLOCK();
}

/*
 * Called whenever the system thinks it is idle, and might want to
 * do a GC.  This might also want to grow the heap spaces to maintain
 * the desired free space.
 */
void
asyncGC()
{
    int freemem = FreeObjectCtr;
    int totalrefs, freedrefs;
    int print_freedrefs = 0;

#ifdef DEBUG
    if (verbosegc && allocObjects) {
       if (allocBogus)
	   jio_fprintf(stderr,
"<GC(async): Start - Avg. object size for last %d objects is unknown>\n",
		   allocObjects);
       else
	   jio_fprintf(stderr,
"<GC(async): Start - Avg. object size for last %d objects was %d bytes>\n",
		   allocObjects, allocBytes / allocObjects);
    }
    allocBytes = 0;
    allocBogus = 0;
    allocObjects = 0;
#endif

#ifdef TRACEGC
    if (tracegc) {
	jio_fprintf(stderr, "** asyncGC\n");
	DumpHeapInfo();
    }
#ifdef LDEBUG
    verifyHandleCtrs();
    verifyObjectCtrs();
    validatePool(0);
#endif
#endif

    HEAP_LOCK();
    LOCK_GC_LOCKS();

    if (sysThreadSingle() != SYS_OK) {
	goto unlock;
    }

    /*
     * We currently don't try to expand asynchronously, so ignore the
     * return value (last_free) of gc_locked().
     */
    (void) gc_locked(1, ~0);
    if (sysInterruptsPending()) {
        goto unlock;
    }

    /* Don't want to zero this unless gc actually completed... */
    heap_memory_changes = 0;

    /*
     * If we didn't free anything, and we hit the low water mark,
     * try clearing some soft refs.
     */
    if ((freemem == FreeObjectCtr) && (freemem < FreeMemoryLowWaterMark)) {
	/* If we can't clear any, just return */
	if (!(freedrefs = clearRefPointers(1, FreeMemoryLowWaterMark,
	                                   &totalrefs))) {
	    goto unlock;
	}
	print_freedrefs = 1;
	if (sysInterruptsPending()) {
	    goto unlock;
	}
	(void) gc_locked(1, ~0);
    }

unlock:
    sysThreadMulti();
    /* safe to print in multi-threaded mode */
    if (verbosegc && print_freedrefs) {
        jio_fprintf(stderr,
	            "<GC(async): Asynchronously zeroed %d of %d soft refs>\n",
		    freedrefs, totalrefs);
    }
    freeClasses();
    if (verbosegc)
	printGCMessages(TRUE);

    /* Notify the finalizer thread if objects pending finalization */
    if (FinalizeMeQ != NULL) {
	FINALMEQ_NOTIFY();
    }

    UNLOCK_GC_LOCKS();
    HEAP_UNLOCK();

#ifdef TRACEGC
    if (tracegc) {
	DumpHeapInfo();
    }
#ifdef LDEBUG
    verifyHandleCtrs();
    verifyObjectCtrs();
    validatePool(0);
#endif
#endif

    return;
}

/*
 * Initialize the Java heap.  Argument max_request is the maximum heap
 * memory we are prepared to devote to the heap, and min_request is the
 * minimum memory and amount allocated on startup.
 */

#ifdef PAGED_HEAPS /************ PAGED HEAPS: ********************/

void
InitializeAlloc0(long max_request, long min_request)
{
    /* set up global mark bits pool */
    markbits = NULL;
    markbitsfreeptr = NULL;
    marksize = 0;

    /* setup page map so it will get allocated */
    page_map = NULL;
    /* not really needed, but could flush out problems: */
    mem_base = (void *)0xffffffff;
    mem_top = NULL;

    /* setup linked lists */
    /* doubly-linked list of handle chunks, in address order */
    firstHanBlkP = NULL;
    lastHanBlkP = NULL;
    curHanBlkP = NULL;
    /* doubly-linked list of object chunks, in address order */
    firstObjBlkP = NULL;
    lastObjBlkP = NULL;
    curObjBlkP = NULL;

    FreeObjectCtr = 0;
    TotalObjectCtr = FreeObjectCtr;
    FreeHandleCtr = 0;
    TotalHandleCtr = FreeHandleCtr;

    gMaxMem = max_request;		/* save max memory to be used */

#ifdef LDEBUG
    min_request = min_request / 4;
#endif

    /* initial handle chunk - should be 20-25% object total space  */
    if (! expandHandleSpace((((int)(min_request * 0.20)) >>
			    PTR_2_PAGE_SHIFT) + 1)) {
	out_of_memory();
    }
    if (! expandObjectSpace(((int)(min_request * 0.80)), NULL)) {
	out_of_memory();
    }
    initializeDeferredHandleTable();
}

#else  /************ CONTIGUOUS HEAPS: ********************/

/*
 * Initialize structures supporting GC: allocate and zero the mark bits.
 * [Should we be allocating the mark bits at GC time instead?]
 */
static void
InitializeGC(int max, int min)
{
#ifdef ALLOC_BARRIERS
    GCGenCtr = 0;
    AllocNoGC = FALSE;
#endif /* ALLOC_BARRIERS */
#ifdef HPROF
    jvmpi_new_arena();
#endif /* HPROF */

    /*
     * Map the mark bits array, whose size is defined by the maximum
     * (mapped) memory, and commit to enough memory to map the committed
     * heap.
     */
    markmax = ((max/(OBJECTGRAIN*BITSPERMARK) + 1) * 2) * sizeof(*markbits);
    markbits = (unsigned int *) sysMapMem(markmax, &markmax);
    if (!markbits) {
        out_of_memory();	/* Can't start up */
    }

    /* Could postpone commit until GC; see comment at expandMarkBits() */
    expandMarkBits(min);
}

#define BACKOFF_FACTOR 0.75

void
InitializeAlloc0(long max_request, long min_request)
{
    long max, min;

    /*
     * Allocate the initial heap memory.  In the abstract we first map
     * the maximum we are interested in, then commit to the minimum/initial
     * amount.	Note that as actually implemented on some platforms,
     * commitment might be a no-op, min might equal max etc.
     *
     * Because we don't really need more than min_request to start up,
     * we can back off on the heap mapping if it initially fails.  There
     * is no guarantee that we will be allowed to expand to the max anyhow.
     * Related spaces (e.g. mark bits) scale off what we actually get.
     * We should back off in fairly large chunks so if we are successful
     * we'll be able to map the rest of what we'll need.
     */
    sysAssert(hpool == 0);
    heapbase = (unsigned char *) sysMapMem(max_request, &max);
    while (!heapbase) {
        max_request -= (max_request * BACKOFF_FACTOR);
	if (max_request < min_request) {
	    out_of_memory();    /* Can't start up */
	}
	heapbase = (unsigned char *) sysMapMem(max_request, &max);
    }
    heaptop = heapbase + max;

    /*
     * We want same partition proportions (20%/80%) of the mapped
     * memory as the committed memory.	The handle pool base must be
     * HANDLEGRAIN aligned within the committed memory.
     */
    hpool = heapbase + (int) ((heaptop - heapbase) * 0.20);
    hpool = hpool - (int) (min_request * 0.20);
    hpool = sysCommitMem(hpool, min_request, &min);
    if (!hpool) {
        out_of_memory();	/* Can't start up */
    }

    /* The previous code could have resulted in hpoollimit
     * past heaptop.  This makes sure that doesn't happen.
     */
    hpoollimit = hpool + min;
    hpoollimit = (unsigned char*)((int)hpoollimit & ~(HANDLEGRAIN - 1));
    hpool = (unsigned char *)
	(((int) hpool + HANDLEGRAIN-1) & ~(HANDLEGRAIN-1));

    /*
     * Allocate the initial GC mark bits array
     */
    InitializeGC(max, min);

    /*
     * Leave the bottom 20% of the pool for handles, the rest for objects.
     */
    opoollimit = hpoollimit - sizeof(hdr);
    hpoollimit = hpool +
        ((int) ((hpoollimit - hpool) * 0.20) & ~(HANDLEGRAIN - 1));

    /*
     * The object pool base must be sizeof(hdr) past OBJECTGRAIN alignment
     */
    opool = hpoollimit;
    while (((int) opool + sizeof(hdr)) & (OBJECTGRAIN - 1)) {
        opool += sizeof(hdr);
    }

    /*
     * Make sure that all the handles are zero.	 We recognize a free handle
     * by the fact that its ->obj field is 0.
     */
    linkNewHandles((JHandle*)hpool, (JHandle*)hpoollimit, 0);
    hpoolfreelist = hpool;

    obj_setlf(opool, opoollimit - opool, 1);
#ifdef DEBUG
    ((long *) opool)[1] = 0x55555555;
#endif /* DEBUG */
    obj_setlf(opoollimit, 0, 0);
    FreeObjectCtr = opoollimit - opool;
    TotalObjectCtr = FreeObjectCtr;
    FreeHandleCtr = hpoollimit - hpool;
    TotalHandleCtr = FreeHandleCtr;

    opoolhand = opool;
    initializeDeferredHandleTable();
}

#endif  /************ END CONTIGUOUS HEAPS ********************/

/*
 * Initialize the Java heap.  Argument max_request is the maximum heap
 * memory we are prepared to devote to the heap, and min_request is the
 * minimum memory and amount allocated on startup.
 *
 * If preallocating the entire heap (#ifdef USE_MALLOC), then we always
 * allocate max_request on startup.
 *
 * The heap and heap reserve sizes, whether defaults or as overridden on
 * the command line, must satisfy the constraint
 *
 *    initial heap + yellow reserve + red reserve <= maximum heap
 *
 * else the runtime will not start up.  In this case InitializeAlloc
 * returns FALSE to its caller, so the caller can do/say something
 * appropriate in the context or on the platform.
 */

bool_t
InitializeAlloc(long max_request, long min_request)
{
    sysAssert(OBJECTGRAIN >= sizeof(hdr));

#ifdef USE_MALLOC
    /* If we are just malloc'ing, allocate max amount up front */
    min_request = max_request;
#endif /* USE_MALLOC */

    if (min_request > max_request) {
	/*
	 * At this point we return to the platform-specific startup code
	 * so we can give a sensible diagnostic about incompatible startup
	 * and maximum heap sizes.  Subsequent failures are true out-of-
	 * memory conditions that the runtime can't survive.
	 */
	return FALSE;
    }

#ifdef ALLOC_BARRIERS
    GCGenCtr = 0;
    AllocNoGC = FALSE;
#endif /* ALLOC_BARRIERS */

    InitializeAlloc0(max_request, min_request); /* max or min may be changed */

    /*
     * The free memory low water mark is used when deciding whether to
     * free refs.
     */
    FreeMemoryLowWaterMark = (long) (0.2 * FreeObjectCtr) & ~(OBJECTGRAIN - 1);

    /*
     * Allocate memory locks early to avoid order conflicts on early GC.
     * See bug 1259363, for instance.
     */
    _finalmeq_lock = (sys_mon_t *) sysMalloc(sysMonitorSizeof());
    memset(_finalmeq_lock, 0, sysMonitorSizeof());
    FINALMEQ_LOCK_INIT();
    _hasfinalq_lock = (sys_mon_t *) sysMalloc(sysMonitorSizeof());
    memset(_hasfinalq_lock, 0, sysMonitorSizeof());
    HASFINALQ_LOCK_INIT();
    _heap_lock = (sys_mon_t *) sysMalloc(sysMonitorSizeof());
    memset((char *) _heap_lock, 0, sysMonitorSizeof());
    HEAP_LOCK_INIT();
#ifdef HPROF
    _gc_disable_lock = sysMalloc(sysMonitorSizeof());
    memset((char *) _gc_disable_lock, 0, sysMonitorSizeof());
    sysMonitorInit(_gc_disable_lock);
    jvmpi_new_arena();
#endif /* HPROF */

#ifdef TRACEGC
    verifyHandleCtrs();
    verifyObjectCtrs();
    validatePool(0);
#endif

#ifdef ALLOC_BARRIERS
    setAllocState(ALLOCSTATE_GREEN);

    /*
     * Allocate the red and yellow panic reserves
     */
    if (BARRIERSIZE_RED != 0) {
	AllocNoGC = TRUE;	/* Too early to GC; will crash */
	PRRed = ArrayAlloc(T_PRRED, BARRIERSIZE_RED);
	AllocNoGC = FALSE;
	if (PRRed == NULL) {
	    out_of_memory();
	}
    }
    if (BARRIERSIZE_YELLOW != 0) {
	AllocNoGC = TRUE;	/* Too early to GC; will crash */
	PRYellow = ArrayAlloc(T_PRYELLOW, BARRIERSIZE_YELLOW>>1);
	AllocNoGC = FALSE;
	if (PRYellow == NULL) {
	    out_of_memory();
	}
    }
#endif /* ALLOC_BARRIERS */

    return TRUE;
}

#ifdef HPROF

void EnableGC(ExecEnv *ee)
{
    GC_DISABLE_LOCK(EE2SysThread(ee));
    if (--gc_disable_count == 0) {
        sysMonitorNotifyAll(_gc_disable_lock);
    }
    GC_DISABLE_UNLOCK(EE2SysThread(ee));
}
 
void DisableGC(ExecEnv *ee)
{
    GC_DISABLE_LOCK(EE2SysThread(ee));
    gc_disable_count++;
    GC_DISABLE_UNLOCK(EE2SysThread(ee));
}
 

#endif /* HPROF */

bool_t isHandle(p)
{
    SetLimits();
    return ValidHandle(p);
}

bool_t isObject(p)
{
    SetLimits();
    return ValidObject(p);
}

void pin_object(void *obj)
{
    obj_pin((char *)obj - sizeof(hdr));
}

void unpin_object(void *obj)
{
    obj_unpin((char *)obj - sizeof(hdr));
}

int pinned_object(void *obj)
{
    return obj_pinned((char *)obj - sizeof(hdr));
}

void lock_code();
void unlock_code();
void lock_verifier();
void unlock_verifier();

void lock_for_debugger()
{
    NAMETYPEHASH_LOCK();
    LOADCLASS_LOCK();
    HEAP_LOCK();
    LOCK_GC_LOCKS();
    REGISTRY_LOCK();
    lock_verifier();
    lock_code(); /* from interpreter.c */
    CACHE_LOCK();
}

void unlock_for_debugger()
{
    CACHE_UNLOCK();
    unlock_code();
    unlock_verifier();
    REGISTRY_UNLOCK();
    UNLOCK_GC_LOCKS();
    HEAP_UNLOCK();
    LOADCLASS_UNLOCK();
    NAMETYPEHASH_UNLOCK();
}

#ifdef HPROF

/* Get the type and size information of an object */
ClassClass * jvmpi_get_object_info(JHandle *h, int *is_array, int *size)
{
    SetLimits();

    *size = obj_len((char *)unhand(h) - sizeof(hdr)) - sizeof(hdr);
    if (!obj_is_array(h)) {
	*is_array = 0;
	if (obj_methodtable(h) == NULL) {
	    /* OK, if javaLangClassMethods is still 0 */
	    return NULL;
	}
	return obj_classblock(h);
    } else {
        switch(obj_array_type(h)) {
	case T_BOOLEAN:      
	case T_BYTE:      
	case T_SHORT:      
	case T_CHAR:      
	case T_INT:      
	case T_LONG:      
	case T_FLOAT:      
	case T_DOUBLE:
	    *is_array = obj_array_type(h);
	    return NULL;

	case T_CLASS: {
	    ArrayOfObject *array = (ArrayOfObject *)unhand(h);
	    int length = obj_length(h);
	    ClassClass *array_cb = (ClassClass *) array->body[length];
	    *is_array = T_CLASS;
	    return array_cb;
	}
	default:
	    return JVMPI_INVALID_CLASS; /* RED and ORANGE reserve,
					 * cache blocks. */
	}
    }
    return 0;
}

JHandle *jvmpi_obj_reverse_map(void *obj)
{
    SetLimits();
    JHandle *hp;

    MAP_OVER_HANDLES_FROM_START(hp)
        if ((void *) unhand(hp) == obj) {
	    return hp;
	}
    END_MAP_OVER_HANDLES_FROM_START;
    return NULL;
}

/* Set by jvmpi routines when we want a heap dump. */
void jvmpi_in_dump_heap(bool_t flag) 
{
    jvmpi_in_dump_heap_flag = flag;
}

unsigned long jvmpi_get_heap_size(void)
{
    return TotalObjectCtr;
}

/*
 * The following routines are cloned from the appropriate scan* 
 * GC funcitons. REMIND: keep them up to date!
 */

static void 
jvmpi_scan_java_frame(JavaFrame *frame,
		      stack_item *top_top_stack,
		      int jni_info,
		      ExecEnv *ee,
		      int dist_from_top)
{
    stack_item *ssc, *limit;
    JHandle *ptr;
    JavaStack *javastack;
    struct methodblock *mb = frame->current_method;
    bool_t is_native = mb ? methodaccess(mb) & ACC_NATIVE : TRUE;
    int root_info = IS_JIT_FRAME(frame) ? GC_ROOT_NATIVE_STACK :
		    (is_native ? jni_info : GC_ROOT_JAVA_FRAME);
    SetLimits();

    if (mb) {
	ClassClass *mbClass = methodclass(mb);
	jvmpi_dump_add_root((JHandle *) mbClass, root_info, (void*) dist_from_top, ee);
    }

    limit = top_top_stack;
    javastack = frame->javastack;
 
    /* Scan the operand stack. */
    /*CONSTCOND*/
    while (1) {
        int is_first_chunk = IN_JAVASTACK((stack_item *)frame, javastack);
        for (ssc = is_first_chunk ? frame->ostack : javastack->data;
             ssc < limit; ssc++) {
            ptr = ssc->h;
            if (ValidHandle(ptr)) {     /* Never an object pointer */
		if (jni_info == GC_ROOT_JNI_GLOBAL) {
		    jvmpi_dump_add_root(ptr,
					GC_ROOT_JNI_GLOBAL,
					&(ssc->h),
					0);
		} else {
		    jvmpi_dump_add_root(ptr, root_info, (void *)dist_from_top, ee);
		}
            }
        }
        if (is_first_chunk)
            break;
        javastack = javastack->prev;
        limit = javastack->end_data;
    }
 
    /* Nothing more to do for pseudo frames. */
    if (mb == 0 || IS_JIT_FRAME(frame))
        return;
 
    if (methodaccess(mb) & ACC_NATIVE) {
        /* For native frames, we scan the arguments stored at the top
           of the previous frame. */
        JavaFrame *prev_frame = frame->prev;
        if (prev_frame == 0)
            return;
        ssc = prev_frame->optop;
        limit = ssc + methodargssize(mb);
    } else {
        /* Scan local variables in Java frame */
        ssc = frame->vars;
        if (ssc == 0)
            return;
        limit = (stack_item *)frame;
    }
 
    for (; ssc < limit; ssc++) {
        ptr = ssc->h;
        if (ValidHandle(ptr)) {  /* Never an object pointer */
	    jvmpi_dump_add_root(ptr, root_info, (void *) dist_from_top, ee);
        }
    }
}

static JHandle *
jvmpi_find_handle_from_object(void *obj)
{
    JHandle *hp;
    SetLimits();

    MAP_OVER_HANDLES_FROM_START(hp)
	if (unhand(hp) == obj) {
	    return hp;
	}
    END_MAP_OVER_HANDLES_FROM_START;
    return NULL;
}

static int 
jvmpi_scan_thread_traces(sys_thread_t *t, void *arg)
{
    ExecEnv *ee = SysThread2EE(t);

    if (ee == NULL) {
	return SYS_OK;
	/* This is true if the finalizer thread hasn't started yet */
    }

    if (ee->initial_stack == NULL) {
        /* EE already destroyed. Thread is about to exit */
        return SYS_OK;
    }

    if (ee->current_frame) {
        jvmpi_dump_add_trace(ee);
    }
    return SYS_OK;
}

    /* This is a clone fo scanThread */
static int 
jvmpi_scan_thread_roots(sys_thread_t *t, void *arg)
{
    ExecEnv *ee = SysThread2EE(t);
    JHandle *me = (JHandle *) arg;
    JHandle *p = (JHandle *) sysThreadGetBackPtr(t);
    unsigned char **ssc, **limit;
    stackp_t base;
    SetLimits();

    if (SYSTHREAD(p) == 0 || ee == NULL || !ee->has_been_born) {
	/* The thread must be unborn or exiting... */
	return SYS_OK;
    }

    base = threadStackBase((TID)p);
    ssc	 = threadStackPointer((TID)p);

    if (p == me) {
	ssc = (unsigned char **) &me;
    }

    if (ssc == 0 || base == 0 || (ssc == base)) {
	/*
	 * If the stack does not have a top of stack pointer or a base
	 * pointer then it hasn't run yet and we don't need to scan
	 * its stack.  When exactly each of these data becomes available
	 * may be system-dependent, but we need both to bother scanning.
	 */
	goto ScanJavaStack;
   }

    /* Align stack top, important on Windows 95. */
    if ((long)ssc % sizeof(void *)) {
        ssc = (unsigned char **)((long)(ssc) & ~(sizeof(void *) - 1));
    }

    limit = (unsigned char **) base;

    sysAssert(ssc != limit);

    while (ssc < limit) {
	register unsigned char *ptr = *ssc;
	if (ValidHandle(ptr)) {
	    jvmpi_dump_add_root((JHandle *)ptr,
				GC_ROOT_NATIVE_STACK,
				NULL,
				ee);
	} else if (ValidObject(ptr)) {
	    JHandle *h = jvmpi_find_handle_from_object(ptr);
	    if (h) {
		jvmpi_dump_add_root(h,
				    GC_ROOT_NATIVE_STACK,
				    NULL,
				    ee);
	    }
	}
	ssc++;
    }

    /*
     * Whether or not we scan the thread stack, we decide independently
     * whether to scan the Java stack.  Doing so should be more robust
     * in the face of partially-initialized or partially-zeroed threads
     * during thread creation or exit, or changes to any of that code.
     */
ScanJavaStack:
    {
	ExecEnv *ee = (ExecEnv *)THREAD(p)->eetop;
	JavaFrame *frame;
	register unsigned char *ptr;

	if (ee != 0) {
	    /*
	     * Because of the Invocation API, the EE may not be on the C
	     * stack anymore.
	     */
	    if (ValidHandle(ee->exception.exc)) {
		jvmpi_dump_add_root(ee->exception.exc,
				    GC_ROOT_THREAD_BLOCK,
				    NULL,
				    ee);
	    }
	}

	if ((ee != 0) && ((frame = ee->current_frame) != 0)) {
	    struct methodblock *current_method = frame->current_method;
	    struct methodblock *prev_current_method;
	    stack_item *top_top_stack = (!IS_NATIVE_MB(current_method))
		  ? &frame->ostack[methodmaxstack(frame->current_method)]
		  : frame->optop;
	    int dist_from_top = 0;

	    for(;;) {
		struct methodblock *mb = frame->current_method;

		if (mb == 0 || !IS_JIT_FRAME(frame)) {
                    jvmpi_scan_java_frame(frame, top_top_stack,
					  GC_ROOT_JNI_LOCAL, ee, dist_from_top);
		}
		{
		    JavaFrame *frame_ptr = frame;
		    JavaFrame frame_buf;
		    frame = frame->prev;
		    while (frame_ptr != frame) {
			if (frame_ptr->current_method) {
			    dist_from_top++;
			}
			frame_ptr = FRAME_PREV(frame_ptr, &frame_buf);
		    }
		}
		prev_current_method = current_method;
		if (!frame) break;
		current_method = frame->current_method;
		/*
		 * If the previous frame was a transition frame from C back
		 * to Java (indicated by prev_current_method == NULL), then
		 * this new frame might not have set its optop.  We must be
		 * conservative.   Otherwise, we can use the optop value.
		 *
		 * Also permit two consecutive frames with NULL current
		 * methods, in support of JITs.  See bug 4022856.
		 */
		top_top_stack =
		    (prev_current_method == NULL &&
			!IS_NATIVE_MB(current_method))
			    ? &frame->ostack[methodmaxstack(current_method)]
			    : frame->optop;
	    }
	}
    }

    return SYS_OK;
}

static void
jvmpi_scan_threads(void)
{
    JHandle *self;
    int err;

    self = (JHandle *) sysThreadGetBackPtr(sysThreadSelf());
    err = sysThreadEnumerateOver(jvmpi_scan_thread_traces, (void *) self);
    err = sysThreadEnumerateOver(jvmpi_scan_thread_roots, (void *) self);
    /* REMIND - should do something with the return value. */
}



static void
jvmpi_scan_globalrefs(void)
{
    jvmpi_scan_java_frame(globalRefFrame, globalRefFrame->optop,
			  GC_ROOT_JNI_GLOBAL, 0, 0);
}

static void
jvmpi_scan_primitive_classes(void)
{
    static ClassClass **primitive_classes[] = {
	&class_void, &class_boolean, &class_byte, &class_char, &class_short,
	&class_int, &class_long, &class_float, &class_double
    };
    ClassClass ***cbpp;
    int count;
    SetLimits();
    count = sizeof(primitive_classes)/sizeof(primitive_classes[0]);
    for (cbpp = primitive_classes; --count >= 0; cbpp++) { 
	ClassClass *cb = **cbpp;
	if (ValidHandle(cb)) {
	    jvmpi_dump_add_root((JHandle *) cb, GC_ROOT_STICKY_CLASS, 0, 0);
	    sysAssert(ValidObject(unhand(cb)));
	}
    } 
}

static void jvmpi_scan_sticky_classes(void)
{
    ClassClass **pcb =    binclasses;
    int i;
    SetLimits();

    for (i = nbinclasses; --i > 0; pcb++) {
        ClassClass *cb = *pcb;
	if (!allowClassGCNow || CCIs(cb, Sticky))
	    jvmpi_dump_add_root((JHandle *) cb, GC_ROOT_STICKY_CLASS, 0, 0);
    }
}


static void jvmpi_gc_dump_heap(void)
{
    JHandle *hp;
    SetLimits();

    jvmpi_scan_threads();
    jvmpi_scan_globalrefs();
    jvmpi_scan_sticky_classes();

    MAP_OVER_HANDLES_FROM_START(hp)
	if (unhand(hp)) {
	    if (ValidObject(unhand(hp)) && IsMarked(hp)) {
		jvmpi_dump_object(hp);
	    }
	}
    END_MAP_OVER_HANDLES_FROM_START;
}

#endif /* HPROF */
/*
 * Garbage collector debugging and diagnostics
 */

/*
 * Support for DumpMonitors(): we want to print whose monitor this is
 * if it is associated with an object.	To check whether that's so we
 * need to see whether the handle is in the handle pool, and the limits
 * of the handle pool aren't visible out of gc.c.
 *
 * The arg that is passed in here is a verbose flag that controls
 * whether information is to be printed about unused monitors.  Monitors
 * that are passed in here are always in the monitor cache.
 */
void
monitorDumpHelper(monitor_t *mid, void *arg)
{
    unsigned int key = mid->key;
    bool_t verbose = (bool_t)(int)arg;
    SetLimits();

    if (verbose != FALSE || mid->use_count != 0) {
	if (ValidHandle(key)) {
	    jio_fprintf(stderr, "    %s: ", Object2CString((JHandle *) key));
	} else {
	    jio_fprintf(stderr, "    <unknown key> (0x%lx): ", key);
	}
	sysMonitorDumpInfo((sys_mon_t *)&mid->mid);
    }
}

/*
 * Profiling support
 */
#define MAX_CLASS_HASH 1023

const struct arrayinfo arrayinfo[] = {
    { 0,		      0,                "N/A",        0},
    { 0,		      0,                 0,           0},
    { OBJ_FLAG_CLASS_ARRAY,   SIGNATURE_CLASS,   "class",     sizeof(OBJECT)},
    { 0,		      0,                 0,           0},
    { OBJ_FLAG_BOOLEAN_ARRAY, SIGNATURE_BOOLEAN, "bool",      sizeof(char)},
    { OBJ_FLAG_CHAR_ARRAY,    SIGNATURE_CHAR,	 "char",      sizeof(unicode)},
    { OBJ_FLAG_FLOAT_ARRAY,   SIGNATURE_FLOAT,   "float",     sizeof(float)},
    { OBJ_FLAG_DOUBLE_ARRAY,  SIGNATURE_DOUBLE,  "double",    sizeof(double)},
    { OBJ_FLAG_BYTE_ARRAY,    SIGNATURE_BYTE,    "byte",      sizeof(char)},
    { OBJ_FLAG_SHORT_ARRAY,   SIGNATURE_SHORT,   "short",     sizeof(short)},
    { OBJ_FLAG_INT_ARRAY,     SIGNATURE_INT,     "int",       sizeof(int)},
    { OBJ_FLAG_LONG_ARRAY,    SIGNATURE_LONG,    "long",      sizeof(int64_t)},
};

#define ARRAYTYPES	(sizeof(arrayinfo) / sizeof(arrayinfo[0]))

#define	T_BAD	T_XXUNUSEDXX1

/* This table should be 2 ^ ARRAY_FLAG_BITS long */

const unsigned char array_type_to_type[1 << ARRAY_FLAG_BITS] = { 
    T_NORMAL_OBJECT, T_CLASS,  T_NORMAL_OBJECT, T_BYTE,    
    T_NORMAL_OBJECT, T_SHORT,  T_NORMAL_OBJECT, T_INT, 
    T_NORMAL_OBJECT, T_LONG,   T_NORMAL_OBJECT, T_FLOAT,
    T_NORMAL_OBJECT, T_DOUBLE, T_NORMAL_OBJECT, T_BOOLEAN, 
    T_NORMAL_OBJECT, T_CHAR,   T_NORMAL_OBJECT, T_BAD,
    T_NORMAL_OBJECT, T_BAD,    T_NORMAL_OBJECT, T_BAD,
    T_NORMAL_OBJECT, T_BAD,    T_NORMAL_OBJECT, T_BAD,
    T_NORMAL_OBJECT, T_BAD,    T_NORMAL_OBJECT, T_BAD
};


/* print debug info for handles & classes */
void
profHandles(FILE *fp) {
    SetLimits();
    register JHandle *hp;
    ClassClass *cb;
    struct hash_entry {
        ClassClass *cb;
        int count;
        int arraycount;
        int arraylength;
    } *tab, *p;
    int arraycount[ARRAYTYPES];
    int arraylength[ARRAYTYPES];
    int handles_used = 0;
    int i;

    tab = (struct hash_entry *) sysMalloc(sizeof(struct hash_entry) *
          (MAX_CLASS_HASH+1));
    memset((char *) tab, 0, sizeof(struct hash_entry) * (MAX_CLASS_HASH+1));
    memset((char *) arraycount, 0, sizeof(arraycount));
    memset((char *) arraylength, 0, sizeof(arraylength));

    MAP_OVER_HANDLES_FROM_START(hp)
	if ((unhand(hp) == 0) || (obj_free((unsigned char *) unhand(hp)))) {
	    continue;
	}
        if (obj_isnt_array(hp)) { 
	    cb = obj_classblock(hp);
	    p = &tab[(((long) cb) >> 2) % MAX_CLASS_HASH];
	    /* p is desired hash ptr - handle collisions by linear scan down */
	    while ((p->cb != 0) && (p->cb != cb)) {
	        if (--p == tab) {
		    p = &tab[MAX_CLASS_HASH]; /* restart from end of table */
	   /* ??? this loops forever if more than MAX_CLASS_HASH classes... */
		}
	    }
	    if (p->cb == 0) {
	        p->cb = cb;
	    }
	    p->count++;		  /* count refs to this class */
	} else if (obj_array_flags(hp) == OBJ_FLAG_CLASS_ARRAY) { 
	    cb = ((ClassClass **) unhand((HArrayOfObject*)hp)->body)
		                                [obj_length(hp)];
	    p = &tab[(((long) cb) >> 2) % MAX_CLASS_HASH];
	    while ((p->cb != 0) && (p->cb != cb)) {
		if (--p == tab) {
		    p = &tab[MAX_CLASS_HASH];
		    /* restart from end of hash table */
		    /* ??? this loops forever if more than MAX_CLASS_HASH classes...*/
		}
	    }
	    if (p->cb == 0) {
		p->cb = cb;
	    }
	    p->arraycount++;
	    p->arraylength += obj_length(hp);
	} else { 
	    int type = obj_array_type(hp);
	    int array_length = obj_length(hp);
	    arraycount[type]++;
	    arraylength[type] += array_length;
	}
	handles_used++;
    END_MAP_OVER_HANDLES_FROM_START;

    jio_fprintf(fp,
            "handles_used: %d, handles_free: %d, heap-used: %d, heap-free: %d\n",
	    handles_used, TotalHandleCtr / sizeof(JHandle),
	    TotalObjectCtr - FreeObjectCtr, FreeObjectCtr);

    jio_fprintf(fp, "sig  count  bytes  indx\n");
    for (i = 0 ; i < ARRAYTYPES ; i++) {
       	if (arraycount[i]) {
       	    jio_fprintf(fp, "[%c   %5d	%5d  %4d\n",
      		    arrayinfo[i].sig, arraycount[i],
		    arraylength[i] * arrayinfo[i].factor, i);
	}
    }
    p = &tab[MAX_CLASS_HASH];
    i = MAX_CLASS_HASH;
    while (p != tab) {
	if (p->cb != 0) {
       	    jio_fprintf(fp, "*** tab[%d] p=%x cb=%x cnt=%d ac=%d al=%d\n",
       		    i, p, p->cb, p->count, p->arraycount,
       		    p->arraylength);
       	    /*JVE*/
       	    if (p->count > 0) {
       	       	jio_fprintf(fp, "	L%s; %d %d\n",
       	       		cbName(p->cb), p->count,
	       			p->count * cbInstanceSize(p->cb));
	    }
       	    if (p->arraycount > 0) {
       	      	jio_fprintf(fp, "  [L%s; %d %d\n",
	      	      	cbName(p->cb), p->arraycount,
       			p->arraylength * sizeof(OBJECT));
	    }
	}
       	p--;
       	i--;
    }
    sysFree(tab);
}

void
prof_heap(FILE *fp)
{

    gc(0, ~0);		/* Force compaction */
    profHandles(fp);
}

/*
 * Debugging facilities and useful stuff to call from dbx...
 */

#ifdef DEBUG

#ifdef PAGED_HEAPS /************ PAGED HEAPS: ********************/

void
checkChunk(ChunkBlkP chunk)
{
    sysAssert(chunk);
    sysAssert((long)(chunk->endPtr - chunk->startPtr) == chunk->allocSize);
    /* freePtr is either a head of a free list, or it's a hint, and
     * must be in range: */
    sysAssert((chunk->freePtr == NULL) ||
	      ((chunk->freePtr >= chunk->startPtr) &&
	       (chunk->freePtr <= chunk->endPtr)));
    sysAssert((chunk->freeCnt >= 0) &&
	      (chunk->freeCnt <= chunk->allocSize));
    sysAssert(chunk->endPtr <= chunk->physEndPtr);
    sysAssert((*(int32_t *)(chunk->physEndPtr - 8)) == ALMOST_WORD);
    sysAssert((*(int32_t *)(chunk->physEndPtr - 4)) == ULTIMATE_WORD);
    sysCheckWastedSpace(chunk);
}

void
verifyPageMap(void)
{
    ChunkBlkP blkP;
    int handlePagesByList = 0;
    int objectPagesByList = 0;
    int handlePagesByMap = 0;
    int objectPagesByMap = 0;
    int i;
    int limit = PageMapLength();

    /* may be called before heaps have been initialized, so be careful */
    sysAssert(page_map);
    blkP = firstHanBlkP;
    if (blkP) {
	do {
	    checkChunk(blkP);
	    handlePagesByList += (blkP->allocSize >> PTR_2_PAGE_SHIFT) + 1;
	    blkP = blkP->nextPtr;
	} while (blkP != firstHanBlkP);
    }

    blkP = firstObjBlkP;
    if (blkP) {
	do {
	    checkChunk(blkP);
	    objectPagesByList += (blkP->allocSize >> PTR_2_PAGE_SHIFT) + 1;
	    blkP = blkP->nextPtr;
	} while (blkP != firstObjBlkP);
    }

    for (i = 0; i < limit; i++) {
	if (page_map[i].chunk_size > 0) {
	    objectPagesByMap++;
	    sysAssert(page_map[i].mark_bits != NULL);
	} else if (page_map[i].chunk_size < 0) {
	    handlePagesByMap++;
	    sysAssert(page_map[i].mark_bits != NULL);
	} else {
	    sysAssert(page_map[i].mark_bits == NULL);
	}
    }

    sysAssert(handlePagesByList == handlePagesByMap);
    sysAssert(objectPagesByList == objectPagesByMap);
}

void
DumpHeapInfo()
{
    ChunkBlkP hanP = firstHanBlkP;
    ChunkBlkP blkP = firstObjBlkP;

    jio_fprintf(stderr, "DumpHeapInfo\n");

    do {
	jio_fprintf(stderr,
	 "handle chunk 0x%lx; freeCnt=0x%lx, freePtr=0x%lx, endPtr=0x%lx\n",
		hanP, hanP->freeCnt, hanP->freePtr, hanP->endPtr);
	checkChunk(hanP);
	hanP = hanP->nextPtr;
    } while (hanP != firstHanBlkP);

    do {
	jio_fprintf(stderr,
"object chunk 0x%lx; freeCnt=0x%lx, freePtr=0x%lx, endPtr=0x%lx, pinned=%d\n",
	        blkP, blkP->freeCnt, blkP->freePtr, blkP->endPtr,
		blkP->chunkFlags & CHUNK_PINNED);
	checkChunk(blkP);
	blkP = blkP->nextPtr;
    } while (blkP != firstObjBlkP);

    jio_fprintf(stderr, "\n");
}

#else  /************ CONTIGUOUS HEAPS: ********************/

void
DumpHeapInfo()
{
    HEAP_LOCK();

    jio_fprintf(stderr, "Heap layout info:\n");
    jio_fprintf(stderr, "  Mapped memory:\t[0x%x,0x%x), size: 0x%x\n",
	    heapbase, heaptop, heaptop-heapbase);
    jio_fprintf(stderr, "  Committed memory:\t[0x%x,0x%x), size: 0x%x\n",
	    hpool, opoollimit, opoollimit-hpool);
    jio_fprintf(stderr, "  Handle space: \t[0x%x,0x%x], size: 0x%x\n",
	    hpool, hpoollimit, opool-hpool);
    jio_fprintf(stderr, "  Object space: \t[0x%x,0x%x], size: 0x%x\n",
	    opool, opoollimit, opoollimit-opool+sizeof(hdr));

    HEAP_UNLOCK();
}

#endif  /************ END CONTIGUOUS HEAPS ********************/

int
isMarked(char *p)
{
    SetLimits();
    return (IsMarked(p) ? 1 : 0);
}

int
isHorO(p)
{
    SetLimits();
    return (ValidHorO(p) ? 1 : 0);
}

int GCValidHandle(JHandle *h) {
    SetLimits();

    return ValidHandle(h) && unhand(h) != 0;
}

void
printPool(char *s)
{
    SetLimits();
    register unsigned char *p;
    register col = 0;

    if (s)
	jio_fprintf(stderr, "\nprintPool: %s\n", s);

    MAP_OVER_OBJECTS_FROM_START(p)
	if (++col > 7) {
	    jio_fprintf(stderr, "\n");
	    col = 1;
	}
	if (obj_len(p) < 4
#ifdef PAGED_HEAPS
	    || obj_len(p) > (MOO_chunk->allocSize)
#endif
	    ) {
	    jio_fprintf(stderr, "Bogus len 0x%lx in blk 0x%lx\n", obj_len(p), p);
	    break;
	}
	jio_fprintf(stderr, "%8X%c%c%c\t", obj_len(p),
		p == MOO_start ? '*' : ' ',
		obj_free(p) ? 'F' : ' ',
		IsMarked(p + sizeof(hdr)) ? 'P' : ' ');
    END_MAP_OVER_OBJECTS_FROM_START;
}

void
printHandles(char *s)
{
    SetLimits();
    JHandle *hp;

    if (s)
	jio_fprintf(stderr, "\nprintHandles: %s", s);

    MAP_OVER_HANDLES_FROM_START(hp)
	if (unhand(hp)) {
	  jio_fprintf(stderr, "han=%X %c  obj=%X %c\n",
		  hp,
		  ValidHandle(hp) ? (IsMarked(hp) ? 'M' : ' ') : '*',
		  unhand(hp),
		  ValidObject(unhand(hp)) ? (IsMarked(unhand(hp)) ? 'P' : ' ') : '*');
	}
    END_MAP_OVER_HANDLES_FROM_START;
    jio_fprintf(stderr, "\n");
}

void
verifyObjectCtrs(void)
{
    SetLimits();
    register unsigned char *p;
    int free = 0;
    int inuse = 0;
    int total = 0;
    int chunkFree = 0;

    MAP_OVER_OBJECTS_FROM_START(p)
	if (obj_free(p)) {
	    chunkFree += obj_len(p);
	} else {
	    inuse += obj_len(p);
	}
	total += obj_len(p);
	if ((p + obj_len(p)) >= MOO_limit) {
#ifdef PAGED_HEAPS
	    sysAssert(chunkFree == MOO_chunk->freeCnt);
#endif /* PAGED_HEAPS */
	    free += chunkFree;
	    chunkFree = 0;
	}
    END_MAP_OVER_OBJECTS_FROM_START;

    sysAssert(free == FreeObjectCtr);
    sysAssert(inuse == TotalObjectCtr - FreeObjectCtr);
    sysAssert(total == TotalObjectCtr);

/*  For more verbose output:

    jio_fprintf(stderr, "verifyObjectCtrs() detected:\n");
    jio_fprintf(stderr, "    %d bytes free versus %d bytes accumulated\n",
	    free, FreeObjectCtr);
    jio_fprintf(stderr, "    %d bytes in use versus %d bytes accumulated\n",
	    inuse, TotalObjectCtr - FreeObjectCtr);
    jio_fprintf(stderr, "    %d bytes total versus %d bytes accumulated\n",
	    total, TotalObjectCtr);
*/
}

void
verifyHandleCtrs(void)
{
    SetLimits();
    register JHandle *hp;
    int free = 0;
    int inuse = 0;
    int total = 0;
    int chunkFree = 0;

    MAP_OVER_HANDLES_FROM_START(hp);
	if (unhand(hp) == 0) {
	    chunkFree += sizeof(JHandle);
	} else {
	    inuse += sizeof(JHandle);
	}
	total += sizeof(JHandle);
	if ((hp + 1) >= MOH_limit) {
#ifdef PAGED_HEAPS
	    sysAssert(chunkFree == MOH_chunk->freeCnt);
#endif /* PAGED_HEAPS */
	    free += chunkFree;
	    chunkFree = 0;
	}
    END_MAP_OVER_HANDLES_FROM_START;

    sysAssert(free == FreeHandleCtr);
    sysAssert(inuse == TotalHandleCtr - FreeHandleCtr);
    sysAssert(total == TotalHandleCtr);

/*  For more verbose output:

    jio_fprintf(stderr, "verifyHandleCtrs() detected:\n");
    jio_fprintf(stderr, "    %d bytes free versus %d bytes accumulated\n",
	    free, FreeHandleCtr);
    jio_fprintf(stderr, "    %d bytes in use versus %d bytes accumulated\n",
	    inuse, TotalHandleCtr - FreeHandleCtr);
    jio_fprintf(stderr, "    %d bytes total versus %d bytes accumulated\n",
	    total, TotalHandleCtr);
*/
}

void
validatePool(long quick)
{
    SetLimits();
    register unsigned char *p;
    register JHandle *hp;
    int err = 0;

#ifdef PAGED_HEAPs
    verifyPageMap();
#endif /* PAGED_HEAPs */

    MAP_OVER_OBJECTS_FROM_START(p)
	sysAssert(obj_len(p) >= OBJECTGRAIN);  /* 4 didn't seem right */
	sysAssert((p + obj_len(p)) <= MOO_limit);
	sysAssert(ValidObject(p + sizeof(hdr)));
#ifdef PAGED_HEAPS
	sysAssert(p >= MOO_chunk->startPtr);
#endif /* PAGED_HEAPs */
	if (obj_free(p)) {
	    if (obj_len(p) > OBJECTGRAIN) {
	        sysAssert(((long *) p)[CHECK_WORD_INDEX] == 0x55555555);
	     }
	} else if (!quick) {
	    MAP_OVER_HANDLES_FROM_START(hp)
		if (unhand(hp) == (ClassObject *) (p + sizeof(hdr))) {
		    goto HasHandle;
		}
	    END_MAP_OVER_HANDLES_FROM_START;
	    jio_fprintf(stderr, "@%X -- allocated object with no handle\n", p);
	    err++;
	}
HasHandle:
    err = err;	/* the label needs something to attach itself to... */
    END_MAP_OVER_OBJECTS_FROM_START;


    MAP_OVER_HANDLES_FROM_START(hp)
	if (unhand(hp)) {
	    if (ValidObject(unhand(hp)) &&
	        obj_free((char *) unhand(hp) - sizeof(hdr))) {
		jio_fprintf(stderr,
		       "@%X/%X -- Handle points to free block\n", hp, unhand(hp));
		err++;
	    } else if (obj_isnt_array(hp)) {
		if (obj_methodtable(hp) == 0) {
		    jio_fprintf(stderr,
				"1 @%X/%X -- Invalid method table\n", hp, obj_methodtable(hp));
		    err++;
		} else if (obj_methodtable(hp) != 
			   cbMethodTable(obj_classblock(hp))) {
		    jio_fprintf(stderr,
				"2 @%X/%X -- Invalid method table\n", hp, obj_methodtable(hp));
		    err++;
		}
	    } else { 
		if (obj_array_flags(hp) > OBJ_FLAG_MAX_ARRAY) {
		    jio_fprintf(stderr,
				"3 @%X/%X -- Invalid method table\n", hp, obj_methodtable(hp));
		    err++;
		}
	    }
	}
    END_MAP_OVER_HANDLES_FROM_START;

    if (tracegc)
	jio_fprintf(stderr, err == 0
	        ? "OK memory pool\n"
	        : "***%d errors in the memory pool***\n", err);
}

#endif /* DEBUG */
