/*
 * @(#)interpreter.c	1.243 98/10/20
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Interpret a Java object file
 */

#include <sys/types.h>
#include <string.h>
#include <stdarg.h>

#include "bool.h"
#include "opcodes.h"
#include "oobj.h"
#include "interpreter.h"
#include "signature.h"
#include "typecodes.h"
#include "tree.h"
#include "threads.h"
#include "monitor.h"
#include "exceptions.h"
#include "javaString.h"
#include "path.h"
#include "byteorder.h"
#include "jmath.h"
#include "sys_api.h"

#include "java_lang_Thread.h"
#include "java_lang_Throwable.h"

#ifdef BREAKPTS
#include "breakpoints.h"
#endif

#ifdef HPROF
#include "vmprofiler.h"
#endif /* HPROF */

int min_javastack_chunk_size = 500; /* Shrunk from 1000 to suit pJava apps */

#ifdef TRACING
int trace_is_enabled;
int tracem_is_enabled;
#  define TRACE(a) (trace_is_enabled ? jio_fprintf(stdout, "%6X %8X", (int)threadSelf(), (int)pc), \
	jio_printf a, fflush(stdout), 0 : 0)
#  define TRACEIF(o, a) (opcode == o ? TRACE(a) : 0)

#else
#  define TRACE(a)
#  define TRACEIF(a, b)
#endif

#ifdef STATISTICS
extern int instruction_count[256][256];
#endif

ExecEnv *DefaultExecEnv;

/* Some JIT compilers don't want to use the versions of
 * opc_invokevirtual_quick and opc_getfield_quick that lose information
 */
bool_t UseLosslessQuickOpcodes;	

/*
 * Lock around rewriting method code for quick_ instruction optimization
 */
static sys_mon_t *_code_lock;
#define CODE_LOCK_INIT()    monitorRegister(_code_lock, "Code rewrite lock")
#define CODE_LOCK()	    sysMonitorEnter(_code_lock)
#define CODE_UNLOCK()	    sysMonitorExit(_code_lock)

void lock_code()
{
    CODE_LOCK();
}

void unlock_code()
{
    CODE_UNLOCK();
}

JNIEXPORT sys_mon_t *get_code_lock()
{
    return _code_lock;
}

/*
 * Java stack management
 */

/*
 * Lock around code freeing deleted Java stack segments
 */
static sys_mon_t *_ostack_lock;
#define OSTACK_LOCK_INIT()  monitorRegister(_ostack_lock, "Java stack lock")
#define OSTACK_LOCK()	    sysMonitorEnter(_ostack_lock)
#define OSTACK_UNLOCK()	    sysMonitorExit(_ostack_lock)

static void FreeJavaStackMemory(JavaStack *);

static JavaStack *
InitializeJavaStack(ExecEnv *ee, JavaStack *previous_stack,
		    JavaStack *stack, size_t size)
{
    stack->execenv = ee;
    stack->prev = previous_stack;
    stack->next = 0;
    if (previous_stack) {
	sysAssert(previous_stack->next == 0);
        previous_stack->next = stack;
	stack->stack_so_far = previous_stack->stack_so_far 
	          + size * sizeof(stack->data[0]);
    } else {
	stack->stack_so_far = size * sizeof(stack->data[0]);
    }
    stack->end_data = &stack->data[size];
    return stack;
}

/*
 * Create the *first* Java stack segment for a given thread.  As this
 * is done by the new thread as it is starting up, it should not allocate
 * memory -- the new thread isn't ready to handle out-of-memory.  The one
 * exception is on startup of the priomordial thread, when there is no
 * Thread object yet.  In that case we allocate the segment, but the
 * calling code is prepared to handle allocation failure.  This function
 * is not exported by the JIT interface.
 */
static JavaStack *
CreateFirstJavaStack(ExecEnv *ee)
{
    JavaStack *stack;

    if (ee->thread == 0) {	/* Starting the primordial thread */
	return CreateNewJavaStack(ee, 0, 0);
    } else {
	stack = (JavaStack *)THREAD(ee->thread)->initial_stack_memory;
	sysAssert(stack != NULL);
	return InitializeJavaStack(ee, 0, stack,
	    sizeof stack->data / sizeof stack->data[0]);
    }
}

/*
 * Create a *subsequent* Java stack segment for a given thread.  This
 * version should be backward compatible with the JDK 1.0.2 version,
 * which is surfaced in the JIT API.  It should only be called from
 * a thread that has already started up, so should be free to allocate
 * memory (and possibly fail at doing so).
 */
JavaStack *
CreateNewJavaStack(ExecEnv *ee, JavaStack *previous_stack, size_t size)
{
    JavaStack *stack;

    if (size < min_javastack_chunk_size)
        size = min_javastack_chunk_size;

    stack = sysMalloc(sizeof(JavaStack) + size * sizeof(stack_item)
	- sizeof stack->data);
    if (stack == 0) {
	return 0;
    }

    return InitializeJavaStack(ee, previous_stack, stack, size);
}

/*
 * Initialize the ExecEnv of a new thread.  This function is called by
 * the new thread on its startup.  The initial Java stack segment will
 * have been allocated by the parent thread, and if that wasn't possible
 * the parent thread should have handled the error.
 *
 * All ExecEnv fields not explicitly set here are initially zeroed.
 */
void
InitializeExecEnv(ExecEnv *ee, HThread *thread) {
    memset((char *) ee, 0, sizeof(ExecEnv));	/* All fields initially 0 */
    if (DefaultExecEnv == 0 && thread == 0) {
	DefaultExecEnv = ee;
    }
    ee->thread = thread;
    /* The first segment of the Java stack is preallocated by parent thread */
    ee->initial_stack = CreateFirstJavaStack(ee);
    if (thread) { 	/* We don't have thread at initialization */
        THREAD(thread)->eetop = (long)ee;
    }
    exceptionClear(ee);		/* Don't want to assume this means 0'ed... */

    /* Stuff for the JNI: */
    InitializeJNIRootFrame(ee);
}

#ifdef DEBUG
static int java_stack_chunks_freed = 0;
#endif

/*
 * The list on which we put freed thread stacks pending reclamation.
 */
typedef struct free_javastack_t {
    struct free_javastack_t *next;
} free_javastack_t;

static free_javastack_t *javastackFreeList = (free_javastack_t *) 0;

/*
 * I (Tim) don't think that it is necessary any more to lazily reclaim
 * Java stack segments.  If segments were eagerly freed, we might be able
 * to do without locking.  Alternatively, freed segments could be kept
 * in a free list and used as a cache for subsequent Java stack creation
 * and expansion, but we'd need to retain OSTACK_LOCK.
 */
void DeleteExecEnv(ExecEnv *ee) {
    JavaStack *stack;

    /*
     * release the java stack.  We want to make sure that the ee doesn't
     * contain a pointer to the stack while we are freeing it, so we clear
     * it first and put in a store barrier for machines with weak memory
     * ordering.
     */  
    stack = ee->initial_stack;
    ee->initial_stack = 0;
#ifdef JDK12
    sysStoreBarrier();
#endif
    FreeJavaStackMemory(stack);
 
    /* JNI */
    DeleteJNIRootFrame(ee);
 
#ifdef JDK12
    memset(ee, 0, offsetof(ExecEnv, sys_thr));
#endif
}

static void FreeJavaStackMemory(JavaStack *this_stack) {
    while (this_stack) {
        JavaStack *next_stack = this_stack->next;
	sysFree(this_stack);
	this_stack = next_stack;
    }
}

int ExpandJavaStackForJNI(ExecEnv *ee,
			  JavaStack **stackP, 
			  JavaFrame **frameP,
			  int capacity)
{
    return ExpandJavaStack(ee, stackP, frameP, 0, 0, 0, capacity);
}

int ExpandJavaStack(ExecEnv *ee,
		    JavaStack **stackP, 
		    JavaFrame **frameP, 
		    stack_item **optopP,
		    int args_size,
		    int nlocals, 
		    int maxstack)
{
    JavaStack *stack = *stackP;
    JavaFrame *frame = *frameP;
    size_t required_size =
        frame->ostack - (stack_item *)frame + nlocals + maxstack;

    if (stack->next != 0) {
        if (required_size > stack->next->end_data - stack->next->data) {
	    JavaStack *this_stack = stack->next;
	    stack->next = 0;
	    FreeJavaStackMemory(this_stack);
	} 
    }

    if (stack->next != 0) 
        stack = stack->next;
    else {
        if (stack->stack_so_far +
	    JAVASTACK_CHUNK_SIZE(stack)*sizeof(stack->data[0]) > 
	    JavaStackSize) {
	    ThrowStackOverflowError(0, 0);
	    return 0;
	}
	stack = CreateNewJavaStack(ee, stack, required_size);
	if (stack == 0) {
	    ThrowOutOfMemoryError(0, 0);
	    return 0;
	} 
    }	
    frame = (JavaFrame *)(stack->data + nlocals);
    if (optopP) {
        stack_item *copied_optop;
	stack_item *optop = *optopP;
      	for (copied_optop = stack->data; --args_size >= 0 ; ) {
	    copied_optop[args_size] = optop[args_size];
	}
	*optopP = copied_optop;
    }
    *frameP = frame;
    *stackP = stack;
    return 1;
}

bool_t ExecuteJava(unsigned char  *, ExecEnv *ee);


/* Call a Java method from C.  The java method may have any number of arguments
 * and may return a single 4 byte value which may be an integer, a pointer to
 * an object handle, whatever.
 */
long
execute_java_dynamic_method(ExecEnv *ee, HObject *obj, 
			   char *method_name, char *signature, ...)
{
    va_list args;
    long result;
    va_start(args, signature);
        result = do_execute_java_method_vararg(ee, obj, method_name, signature, 
					      0, FALSE, args, 0, FALSE);
    va_end(args);
    return result;
}


/* Call a static Java method from C.  The java method may have any number of
 * arguments and may return a single 4-byte value.
 */
long
execute_java_static_method(ExecEnv *ee, ClassClass *cb, 
			  char *method_name, char *signature, ...)
{
    va_list args;
    long result;
    va_start(args, signature);
        result = do_execute_java_method_vararg(ee, cb, method_name, signature,
					      0, TRUE, args, 0, FALSE);
    va_end(args);
    return result;
}

/* Create a new Java object from C.  For example:
 *
 *       obj = execute_java_constructor(0, "java/lang/Float", 0, "(F)", f);
 *
 * EE should be either 
 *     .) the current execution environment
 *     .) 0, in which case the current execution environment is calculated
 *     .) PRIVILEGED_EE, meaning explicitly allow this call to access 
 *        constructors for which the current EE doesn't have privileges.
 *        [i.e. you can access private constructors, etc.]
 *
 * Either classname or cb gives the class of the object to be constructed.  
 * Pass the name of the class in classname, or a pointer to the class object
 * in cb.  Whichever one you give, the other should be NULL.
 *
 * signature is the signature of the arguments to the constructor.  It's 
 * first and last character should be "(" and ")" respectively.  Following
 * are the arguments to the function.
 */

static HObject *
execute_java_constructor_vararg(struct execenv *ee,
				  char *classname,
				  ClassClass *cb,
				  char *signature,
				  int skip,
				  va_list args)
{
#ifdef TRACING
    int pc = 0;			/* for TRACE */
#endif

    HObject *obj;
    char real_signature[256];
    unsigned initID;
    int len;
    ClassClass *from;
    struct methodblock *mb;
    bool_t security = TRUE;
    
    if (ee == PRIVILEGED_EE) {
	ee = EE();
	security = FALSE;
    } else if (ee == 0) {
        ee = EE();
    }

    if (ee)
	/* skip pseudo frames */
	from = getCallerClass(ee, skip);
    else
	from = 0;

    /* Find the class.  Make sure we can allocate an object of that class */
    if (cb == NULL) {
	cb = FindClass(0, classname, TRUE);
	if (cb == NULL) {
	    if (!ee || !exceptionOccurred(ee)) {
		SignalError(0, JAVAPKG "NoClassDefFoundError", classname);
            }
	    return 0;
	}
    } else { 
	classname = cbName(cb);
    }
    if (cbAccess(cb) & (ACC_INTERFACE | ACC_ABSTRACT)) { 
	SignalError(0, JAVAPKG "InstantiationException", cbName(cb));
	return 0;
    }
    if (security && !VerifyClassAccess(from, cb, FALSE)) {
	SignalError(0, JAVAPKG "IllegalAccessException", cbName(cb));
	return 0;
    }
    
    /* Find the constructor */
    if (jio_snprintf(real_signature, sizeof(real_signature), "%sV", signature) == -1) {
	SignalError(0, JAVAPKG "InternalError", "signature overflow");
	return 0;
    }
    initID = NameAndTypeToHash("<init>", real_signature);
    for (len = cbMethodsCount(cb), mb = cbMethods(cb); --len >= 0; mb++) 
	if (methodID(mb) == initID) 
	    break;
    if (len < 0) { 
	SignalError(0, JAVAPKG "NoSuchMethodError", 0);
	return 0;
    }
    if (security && !VerifyFieldAccess(from, methodclass(mb), 
					methodaccess(mb), FALSE)) {
	SignalError(0, JAVAPKG "IllegalAccessException", 0);
	return 0;
    } 

    /* Allocate the object and call the constructor */
    if ((obj = newobject(cb, 0, ee)) == 0) { 
	SignalError(0, JAVAPKG "OutOfMemoryError", 0);
	return 0;
    }
    TRACE(("execute_java_constructor new %s => %s\n", 
	   cbName(cb), Object2CString(obj)));
    do_execute_java_method_vararg(ee, obj, 0, 0, mb, FALSE, args, 0, FALSE);
    return obj;
}

HObject *
execute_java_constructor(struct execenv *ee,
			 char *classname,
			 ClassClass *cb,
			 char *signature, ...)
{
    va_list args;
    HObject *result;
    
    va_start(args, signature);
    result =
	execute_java_constructor_vararg(ee, classname, cb, signature, 0, args);
    va_end(args);
    return result;
}

HObject *
execute_java_constructor_from(struct execenv *ee,
			      char *classname,
			      ClassClass *cb,
			      char *signature,
			      int skip, ...)
{
    va_list args;
    HObject *result;
    
    va_start(args, skip);
    result =
	execute_java_constructor_vararg(ee, classname, cb, signature, 
					skip, args);
    va_end(args);
    return result;
}


long
do_execute_java_method(ExecEnv *ee, void *obj, 
		      char *method_name, char *signature, 
		      struct methodblock *mb,
		      bool_t isStaticCall, ...) {
    va_list args;
    long result;
    va_start(args, isStaticCall);
        result = do_execute_java_method_vararg(ee, obj, 
					  method_name, signature, 
					  mb, isStaticCall, args, 0, FALSE);
    va_end(args);
    return result;
}


long
do_execute_java_method_vararg(ExecEnv *ee, void *obj, 
			      char *method_name, char *method_signature, 
			      struct methodblock *mb,
			      bool_t isStaticCall, va_list args, 
			      long *otherBits, bool_t shortFloats) {

    JavaFrame *frame;
    bool_t isInitializer;
    JNIEnv *env;

    int retval = 0;
    char *p;
    Java8 tdub;

    if (ee == 0)
        ee = EE();
    env = EE2JNIEnv(ee);

    if (otherBits) {
	*otherBits = 0;
    }

    if (mb) {
        method_name = methodname(mb);
	method_signature = methodsig(mb);
    } else {
	NAMETYPEHASH_LOCK();
	if ((Str2ID(&nameTypeHash, method_signature, 0, TRUE) == 0) ||
	    (Str2ID(&nameTypeHash, method_name, 0, TRUE) == 0)) {
	    NAMETYPEHASH_UNLOCK();
	    SignalError(0, JAVAPKG "OutOfMemoryError", 0);
	    return 0;
	}
	NAMETYPEHASH_UNLOCK();
    }
    isInitializer = (method_name[0] == '<'  &&  /* fast test */
		     !isStaticCall && !strcmp(method_name, "<init>"));

    {
	int stack_size = mb ? methodargssize(mb) :
	    (Signature2ArgsSize(method_signature) + 1);
	if (stack_size < 2) {
	    stack_size = 2;
	}
	if ((*env)->PushLocalFrame(env, stack_size) < 0) {
	    return 0;
	}
    }

    frame = ee->current_frame;
    frame->monitor = NULL;
#ifdef BREAKPTS
    frame->lastpc = 0;
#endif
    
    /* Push the target object, if not a static call */
    if (!isStaticCall) 
        (frame->optop++)->p = obj;

    for (p = method_signature + 1; *p != SIGNATURE_ENDFUNC; p++) {
	switch (*p) {
	  case SIGNATURE_BOOLEAN:
	  case SIGNATURE_SHORT:
	  case SIGNATURE_BYTE:
	  case SIGNATURE_CHAR:
	  case SIGNATURE_INT:
	    (frame->optop++)->i = va_arg(args, long);
	    break;
	  case SIGNATURE_FLOAT:
	      if (shortFloats) 
		  (frame->optop++)->i = va_arg(args, long);
	      else 
		  (frame->optop++)->f = va_arg(args, double);
	      break;
	  case SIGNATURE_CLASS:
	    (frame->optop++)->h = va_arg(args, HObject *);
	    while (*p != SIGNATURE_ENDCLASS) p++;
	    break;
	  case SIGNATURE_ARRAY:
	    while ((*p == SIGNATURE_ARRAY)) p++;
	    if (*p == SIGNATURE_CLASS) 
	        { while (*p != SIGNATURE_ENDCLASS) p++; } 
	    (frame->optop++)->p = va_arg(args, void *);
	    break;
	  case SIGNATURE_LONG:
	    SET_INT64(tdub, frame->optop, va_arg(args, int64_t));
	    frame->optop += 2;
	    break;
	  case SIGNATURE_DOUBLE:
	    SET_DOUBLE(tdub, frame->optop, va_arg(args, double));
	    frame->optop += 2;
	    break;
	  default:
	    jio_fprintf(stderr, "Invalid method signature '%s'\n", method_name);
	    goto done;
	}
    }
    {
	unsigned char pc[4];
	cp_item_type  constant_pool[6];
	unsigned char cpt[6];
	
	pc[0] =  isStaticCall ? (mb ? opc_invokestatic_quick : opc_invokestatic)
	       : (mb != NULL) ? opc_invokenonvirtual_quick
	       : isInitializer ? opc_invokespecial : opc_invokevirtual;
	pc[1] = 0; pc[2] = 1;	/* constant pool entry #1 */
	pc[3] = opc_return;

	if (mb) {
	    constant_pool[1].mb = mb;
	    cpt[1] = CONSTANT_POOL_ENTRY_RESOLVED | CONSTANT_Methodref;
	} else {
	    ClassClass *cb = isStaticCall ? (ClassClass *)obj 
		                      : obj_array_classblock((JHandle *)obj);
	    /* Type table */
	    constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].type = cpt;
	    cpt[0] = CONSTANT_POOL_ENTRY_RESOLVED;
	    /* Method entry */
	    constant_pool[1].i = (2 << 16) + 3; /* Class, NameAndType */
	    cpt[1] = CONSTANT_Methodref;
	    /* Class entry */
	    constant_pool[2].clazz = cb; 
	    cpt[2] = CONSTANT_POOL_ENTRY_RESOLVED | CONSTANT_Class;
	    /* Name and type entry */
	    constant_pool[3].i = (4 << 16) + 5;	/* Name, Type */
	    cpt[3] = CONSTANT_NameAndType;
	    /* Method name and type  */
	    constant_pool[4].cp = method_name;
	    constant_pool[5].cp = method_signature;
	    cpt[4] = cpt[5] = CONSTANT_POOL_ENTRY_RESOLVED | CONSTANT_Utf8;
	}

	frame->constant_pool = constant_pool;

	/* Run the byte codes outside the lock and catch any exceptions. */
	exceptionClear(ee);
	ee->current_frame = frame;
	if (ExecuteJava(pc, ee) && (p[1] != SIGNATURE_VOID)) {
	    retval = frame->optop[-1].i;
	    if ((p[1] == SIGNATURE_DOUBLE || p[1] == SIGNATURE_LONG) 
		  && (otherBits != 0))
		*otherBits = frame->optop[-2].i;
	}
    }

 done:
    /* Our caller can look at ee->exceptionKind and ee->exception. */
    (*env)->PopLocalFrame(env, 0);
    return retval;
}

char *
Object2CString(JHandle *oh)
{                               /* ARB to C string */
    ClassObject *o;
    ClassClass *cb;
    /* Using a static buffer is not MT-safe! */
    static char buf[100];
    int maxlen = sizeof(buf) - 10;
    bool_t isHandle(void *p);
    bool_t isObject(void *p);

    if (oh == 0)
	return "NULL";
    if (!isHandle(oh)) { 
	(void) jio_snprintf(buf, sizeof(buf), "BOGUS-HANDLE[%x]", (int)oh);
	return buf;
    }
    o = unhand(oh);
    if (!isObject(o)) { 
	/* Silently truncate on unlikely overflow */
	(void) jio_snprintf(buf, sizeof(buf), "BOGUS-PTR[%d]", (int)o);
    } else if (obj_isnt_array(oh)) { 
        char cbuf[128];
	cb = obj_classblock(oh);
	/* Quietly truncate on buffer overflow */
	(void) jio_snprintf(buf, sizeof(buf), "%s@%X%c%X",
			    classname2string(cbName(cb), cbuf, sizeof(cbuf)),
			    (int)oh, DIR_SEPARATOR, (int)o);
    } else { 
        switch (obj_array_flags(oh)) {
	case OBJ_FLAG_CHAR_ARRAY: {
	    HArrayOfChar *str = (HArrayOfChar *) oh;
	    int len = obj_length(str);
	    unicode *s = unhand(str)->body;
	    unicode * body = s;
	    char *p = buf;
	    
	    len = (len > maxlen ? maxlen : len);

	    *p++ = '"';
	    while (--len >= 0)
		*p++ = *s++;
	    *p++ = '"';
	    *p++ = '\0';
	    KEEP_POINTER_ALIVE(body);
	    break;
	    }
	case OBJ_FLAG_CLASS_ARRAY:
#define BODYOF(h)   unhand(h)->body
	    {
		char cbuf[128];

		cb = ((ClassClass **) BODYOF((HArrayOfObject*)oh))[obj_length(oh)];
		/* Quietly truncate on buffer overflow */
		(void) jio_snprintf(buf, sizeof(buf), "%s[%d]",
			cb ? classname2string(cbName(cb), cbuf, sizeof(cbuf)) : "*MissingClassName*",
			obj_length(oh));
	    }
	    break;
	default:
	    /* Quietly truncate on buffer overflow */
	    (void) jio_snprintf(buf, sizeof(buf), "%s[%d]", 
		    arrayinfo[obj_array_type(oh)].name, 
				obj_length(oh));
	}
    }
    return buf;
}

bool_t
is_instance_of(JHandle * h, ClassClass *dcb, ExecEnv *ee)
{
    if (h == NULL) {		/* null can be cast to anything */
	return TRUE;
    } else if (obj_isnt_array(h)) {
	/* h is an object.  Get its class and to the work */
	return is_subclass_of(obj_classblock(h), dcb, ee);
    } else {
	/* h is an array.  dcb must either be classJavaLangObject, or 
	   interfaceJavaLangCloneable, or an array class */
	return cbName(dcb)[0] == SIGNATURE_ARRAY
	           ? array_is_instance_of_array_type(h, dcb, ee)
		   : (dcb == classJavaLangObject 
		          || dcb == interfaceJavaLangCloneable
		          || dcb == interfaceJavaIoSerializable);
    }
}

bool_t
is_subclass_of(ClassClass *cb, ClassClass *dcb, ExecEnv *ee)
{
    if (cbAccess(dcb) & ACC_INTERFACE) {
        for (; ; cb = cbSuperclass(cb)) {
            if (ImplementsInterface(cb, dcb, ee))
                return TRUE;
            if (cbSuperclass(cb) == 0)
                return FALSE;
        }
    } else {
        for (; ; cb = cbSuperclass(cb)) {
            if (cb == dcb)
                return TRUE;
            if (cbSuperclass(cb) == 0)
                return FALSE;
        }
    }    
}

/* Return TRUE if cb implements the interface icb */

bool_t 
ImplementsInterface(ClassClass *cb, ClassClass *icb, ExecEnv *ee)
{
    /* This has become a lot faster, with the new interface scheme. . .*/
    int i;
    struct imethodtable *imtable = cbIntfMethodTable(cb);
    for (i = imtable->icount - 1; i >= 0; i--)
	if (imtable->itable[i].classdescriptor == icb)
	    return TRUE;
    return FALSE;
}

bool_t array_is_instance_of_array_type(JHandle * h, ClassClass *cb, ExecEnv *ee)
{
    int class_depth = cbConstantPool(cb)[CONSTANT_POOL_ARRAY_DEPTH_INDEX].i;
    int class_type = cbConstantPool(cb)[CONSTANT_POOL_ARRAY_TYPE_INDEX].i;
    int item_depth, item_type;
    ClassClass *item_class, *class_class;

    class_class = (class_type == T_CLASS) 
	             ? cbConstantPool(cb)[CONSTANT_POOL_ARRAY_CLASS_INDEX].clazz 
	             : NULL;
    
    if (obj_array_flags(h) == OBJ_FLAG_CLASS_ARRAY) {
	/* We have an array of classes.  Find out the types of the elements. */
	ArrayOfObject *array = (ArrayOfObject *)unhand(h);
	int length = obj_length(h);
	ClassClass *array_cb = (ClassClass *) array->body[length];
	if (cbName(array_cb)[0] == SIGNATURE_ARRAY) {
	    /* an Array of arrays */
	    union cp_item_type *array_cp = cbConstantPool(array_cb);
	    item_depth = array_cp[CONSTANT_POOL_ARRAY_DEPTH_INDEX].i + 1;
	    item_type =  array_cp[CONSTANT_POOL_ARRAY_TYPE_INDEX].i;
	    item_class = array_cp[CONSTANT_POOL_ARRAY_CLASS_INDEX].clazz;
	} else {
	    /* An array of clases */
	    item_depth = 1;
	    item_type = T_CLASS;
	    item_class = array_cb;
	}
    } else {
	item_depth = 1;
	item_type = obj_array_type(h);
	item_class = 0;		/* immaterial */
    }
    if (item_depth > class_depth) { 
	return (class_class == classJavaLangObject) || 
	       (class_class == interfaceJavaLangCloneable) ||
	       (class_class == interfaceJavaIoSerializable);
    } else if (item_depth == class_depth) { 
	return    (class_type == item_type) 
	       && (   (class_type != T_CLASS) 
		   || (class_class == item_class)
		   || is_subclass_of(item_class, class_class, ee));
    } else { 
	return FALSE;
    }
}

/* An exception has been thrown.  Find the pc at which this exception
 * is handled.  Alternatively, rethrow the error if no appropriate pc
 * is found.
 */

unsigned char *
ProcedureFindThrowTag(ExecEnv *ee, JavaFrame *frame, 
		      JHandle *object, unsigned char *pc)
{
    if (frame->current_method == 0) {
	return NULL;
    } else {
	/* Search current class for catch frame, based on pc value,
	   matching specified type. */
	struct methodblock *mb = frame->current_method;
	ClassClass *methodClass = methodclass(mb);
	ClassClass *objectClass = obj_array_classblock(object);
	struct CatchFrame *cf = methodexctab(mb);
	struct CatchFrame *cfEnd = cf + methodexctablen(mb);
	int pcTarget = pc - methodcode(mb); /* offset from start of code */
	cp_item_type *constant_pool = cbConstantPool(methodclass(mb));
	for (; cf < cfEnd; cf++) { 
	    if (cf->start_pc <= pcTarget && pcTarget < cf->end_pc) { 
		/* Is the class name a superclass of object's class. */
		if (cf->catchType == 0) {
		    /* Special hack put in for finally.  Always succeed. */
		    goto found;
		} else { 
		    /* The usual case */
		    int catchType = cf->catchType;
		    char *catchName = 
			GetClassConstantClassName(constant_pool, catchType);
		    ClassClass *catchClass = NULL;
		    ClassClass *cb;
		    for (cb = objectClass;; cb = cbSuperclass(cb)) {
			if (strcmp(cbName(cb), catchName) == 0) {
			    if (cbLoader(cb) == cbLoader(methodClass)) 
				goto found;
			    if (catchClass == NULL) { 
				/* I haven't yet turned found the actual
				 * class corresponding to catchType, but it's
				 * got the right name.  Find the actual class */
				if (!ResolveClassConstantFromClass(
				       methodClass, catchType, ee, 
				       1 << CONSTANT_Class)) 
				    return NULL;
				catchClass = constant_pool[catchType].clazz;
			    }
			    if (cb == catchClass) 
				goto found;
			}
			if (cbSuperclass(cb) == 0) {
			    /* Not found.  Try next catch frame. */
			    break;
			}
		    }
		}
	    }
	}
	/* not found in this execution environment */
	return NULL;
      found:
	return methodcode(mb) + cf->handler_pc;
    }
}


/* Called by opc_multianewarray.  Create a filled in multidimensional array
 *    dimensions:  The number of dimensions to fill in
 *    array_cb:    The type of the array
 *    sizes:       An array of the dimension sizes
 *
 * Return NULL if we've run out of space.  Otherwise, return the 
 * allocated array. 
 */

HObject *
MultiArrayAlloc(int dimensions, ClassClass *array_cb, stack_item *lengths)
{
    /* The environment */
    struct execenv *ee = EE();
    /* Information about array_cb */
    cp_item_type *array_cp = cbConstantPool(array_cb);
    int array_depth = array_cp[CONSTANT_POOL_ARRAY_DEPTH_INDEX].i;
    char *array_classname =  cbName(array_cb);

    int depth, index, this_length, child_length;
    HArrayOfObject *this_link, *next_link, *child_link;
    ArrayOfObject  fake_array;
    HArrayOfObject fake_handle;

    /* Protect ourselves, in case someone redefines ArrayOfObject not to
     * have at least one element in the body. */
    sysAssert(sizeof(fake_array) >= 
              offsetof(ArrayOfObject, body[0]) + sizeof(fake_array.body[0]));

    sysAssert(dimensions <= array_depth);

    /* Create a fake array handle and body (of length = 1) to hold
     * the root of the array we're creating. 
     * Create a linked list containing this single array (see loop invariant
     *     below).
     * This lets us simplify the algorithm by treating depth == 0 exactly like
     *     depth > 0.
     */
    unhand(&fake_handle) = &fake_array;
    this_link = &fake_handle;
    fake_array.body[0] = NULL;	/* terminate the linked list */
    this_length = 1;		/* number of elements in pseudo array */

    for (depth = 0; ; 
	 this_length = child_length, this_link = child_link, depth++) { 
        /* LOOP_INVARIANT
	     We have a completely bushy tree to depth "depth".
	     Each of the "leaf" arrays,  i.e. those at depth "depth",
	        has length "this_length", which is guaranteed to be > 0.
	     The leaf arrays are linked into a linked list.
	        This linked list starts at "this_link" and continues through
		each array's element[0] field.
	   TNAIRAVNI_POOL
	   We now fill in the next level: 
	     (1) Completely fill in each of the "this_length" elements of each
	         of the arrays in the linked list with a new array of
		 length "lengths[depth].i" and the appropriate type.
	     (2) If this isn't the last iteration through the loop, link the
	         newly created arrays into a new linked list called
		 "child_link".  This will become "this_link" on the next 
                 iteration through the loop.
	*/
	bool_t more_loops;	          /* last time through outer loop? */
	int child_base_type;	          /* argument to ArrayAlloc()  */
	ClassClass *child_clazz_type;     /* element type, if T_CLASS */

	child_length = lengths[depth].i;   /* length of newly created arrays */
	child_link = NULL;	           /* linked list of new arrays */
	
	/* Determine child_clazz_type and child_base_type for the arrays we're
	 * about to create. */
	if (depth + 1 == array_depth) {
	    /* Not an array of arrays.  Must be last iteration of loop */
	    sysAssert(depth == dimensions - 1);
	    child_base_type = array_cp[CONSTANT_POOL_ARRAY_TYPE_INDEX].i;
	    child_clazz_type = (child_base_type == T_CLASS) 
	        /* an array of classes */
	        ? array_cp[CONSTANT_POOL_ARRAY_CLASS_INDEX].clazz 
   	        /* an array of primitive types */
	        : NULL;
	    more_loops = FALSE;	/* must be last iteration through loop */
	} else { 
	    /* An array of arrays. */
	    child_base_type = T_CLASS;
	    child_clazz_type = FindClassFromClass(ee, array_classname+depth+1, 
						 FALSE, array_cb);
	    /* This array class must exist.  If we get NULL, then we've
	     * run out of memory, or something bizarre. 
	     */
	    if (child_clazz_type == NULL) 
	        return NULL;
	    /* Is this the last iteration through loop? */
	    more_loops = ((depth < dimensions - 1) && child_length > 0);
	}

	/* For each array on the linked list. . . . */
	for ( ; this_link != NULL; this_link = next_link) {
	    ArrayOfObject *uthis_link = unhand(this_link);
	    /* Grab the next link before we bash it */
	    next_link = (HArrayOfObject *)uthis_link->body[0];
	    /* Fill in each element of each array on the linked list */
	    for (index = this_length - 1; index >= 0; index--) { 
	        HObject *new_array = ArrayAlloc(child_base_type, child_length);
		if (new_array == NULL) return NULL;
		if (child_clazz_type != NULL) { 
		    ArrayOfObject *unew = unhand((HArrayOfObject *)new_array);
		    unew->body[child_length] = (HObject *)child_clazz_type;
		    if (more_loops) { 
			/* Link the newly created arrays for the next
			 * iteration through the outer loop.  */
		        unew->body[0] = (HObject *)child_link;
			child_link = (HArrayOfObject *)new_array;
		    }
		}
		uthis_link->body[index] = new_array;
	    } 
	}
	sysAssert(!more_loops == (child_link == NULL));
	if (!more_loops) 
	    return fake_array.body[0];
    }
}



/*
    This routine is used to get the current time in milliseconds for
    the profiling routines.
*/
long now()
{
    return(sysGetMilliTicks());
}


#define JAVA_ERROR( name, data ) \
	if (1) { ee->current_frame->lastpc = pc; \
		 SignalError( ee, JAVAPKG name, data ); \
		 return -1; } else
/* 
 * Should  a call to the specified methodblock be turned into
 * an invokesuper_quick instead of a invokenonvirtual_quick?
 *
 * The four conditions that have to be satisfied:
 *    The method isn't private
 *    The method isn't an <init> method
 *    The ACC_SUPER flag is set in the current class
 *    The method's class is a superclass (and not equal) to the current class.
 */ 
bool_t
isSpecialSuperCall(ClassClass *current_class, struct methodblock *mb) {
    /* Remove ifndef when code for opc_invokesuper_quick in the asm FI works. */
    if (((cbAccess(current_class) & ACC_SUPER) != 0)
	  && (!IsPrivate(methodaccess(mb)))
	  && (strcmp(methodname(mb), "<init>") != 0)
	  && (current_class != classJavaLangObject)) {
	ClassClass *methodClass = methodclass(mb);
	ClassClass *cb = cbSuperclass(current_class);
	for (; ; cb = cbSuperclass(cb)) {
	    if (cb == methodClass) 
		return TRUE;
	    if (cbSuperclass(cb) == 0)
		return FALSE;
	}
    }
    return FALSE;
}


int
quickFieldAccess(int opcode, unsigned char * pc, 
		 struct fieldblock *fb, ExecEnv *ee ){
    int access = fieldaccess(fb);
    int result = 0;
    TRACE(("\t%s %s.%s\n", opnames[opcode], 
	   cbName(fieldclass(fb)), fieldname(fb)));
    if (fieldaccess(fb) & ACC_STATIC) {
	char buf[256];
	int len;
	classname2string(cbName(fieldclass(fb)), 
			 buf, sizeof(buf));
	len = strlen(buf);
	/* If buffer overflow, quietly truncate */
	(void) jio_snprintf(buf + len, sizeof(buf)-len,
		": field %s did not used to be static",
		fieldname(fb));
	JAVA_ERROR("IncompatibleClassChangeError", buf);
    } else if ((fieldaccess(fb) & ACC_FINAL) 
	       && (opcode == opc_putfield) 
	       && ( (ee->current_frame->current_method == NULL)
		   || (fieldclass(fb) != 
		        methodclass(ee->current_frame->current_method)))) { 
	char buf[256];
	int len;
	classname2string(cbName(fieldclass(fb)), buf, sizeof(buf));
	len = strlen(buf);
	/* If buffer overflow, quietly truncate */
	(void) jio_snprintf(buf + len, sizeof(buf)-len, 
			    ": field %s is final", fieldname(fb));
	JAVA_ERROR("IllegalAccessError", buf);
    } else {
	unsigned int offset = fieldoffset(fb) / sizeof (OBJECT);
	CODE_LOCK();
	    if (pc[0] != opcode && pc[0] != opc_breakpoint) { 
		result = 1;
	    } if (offset <= 0xFF && !UseLosslessQuickOpcodes) { 
		int delta = ((access & ACC_DOUBLEWORD) != 0)
		        ? (opc_getfield2_quick - opc_getfield)
		        : (opc_getfield_quick - opc_getfield);
		pc[1] = offset;
		pc[0] = opcode + delta;
	    } else { 
		int delta = (opc_getfield_quick_w - opc_getfield);
		pc[0] = opcode + delta;
	    }
	CODE_UNLOCK();
    }
    return result;
}

int quickStaticAccess(int opcode, unsigned char * pc, 
		      struct fieldblock *fb, ExecEnv *ee ){
    TRACE(("\t%s %s.%s\n", opnames[opcode], 
	   cbName(fieldclass(fb)), fieldname(fb)));
    if ((fieldaccess(fb) & ACC_STATIC) == 0) {
	char buf[256];
	int len;
	classname2string(cbName(fieldclass(fb)), 
			 buf, sizeof(buf));
	len = strlen(buf);
	/* If buffer overflow, quietly truncate */
	(void) jio_snprintf(buf + len, sizeof(buf)-len,
		": field %s used to be static",
		fieldname(fb));
	JAVA_ERROR("IncompatibleClassChangeError", buf);
    } else if ((fieldaccess(fb) & ACC_FINAL) 
	       && (opcode == opc_putstatic) 
	       && ( (ee->current_frame->current_method == NULL)
		   || (fieldclass(fb) != 
		        methodclass(ee->current_frame->current_method)))) { 
	char buf[256];
	int len;
	classname2string(cbName(fieldclass(fb)), buf, sizeof(buf));
	len = strlen(buf);
	/* If buffer overflow, quietly truncate */
	(void) jio_snprintf(buf + len, sizeof(buf)-len, 
			    ": field %s is final", fieldname(fb));
	JAVA_ERROR("IllegalAccessError", buf);
    } else {
	char *signature = fieldsig(fb);
	int delta = ((fieldaccess(fb) & ACC_DOUBLEWORD) != 0)
		   ? (opc_getstatic2_quick - opc_getstatic)
		   : (opc_getstatic_quick - opc_getstatic);
	pc[0] = opcode + delta;
    }
    return 0;
}

int quickInvocation(int opcode, unsigned char * pc, 
		    struct methodblock *mb, ExecEnv *ee ) {
    int result = 0;
    TRACE(("\t%s %s.%s%s\n", opnames[opcode], 
	   cbName(methodclass(mb)), methodname(mb), methodsig(mb)));
    if ((opcode == opc_invokestatic) == ((methodaccess(mb) & ACC_STATIC) == 0)) { 
	char buf[256];
	int len;
	classname2string(cbName(methodclass(mb)), buf, sizeof(buf));
	len = strlen(buf);
	/* If buffer overflow, quietly truncate */
	(void )jio_snprintf(buf + len, sizeof(buf)-len, 
			    ": method %s%s %sused to be static",
			    methodname(mb), methodsig(mb),
			    (opcode == opc_invokestatic ? "" : "did not "));
	JAVA_ERROR("IncompatibleClassChangeError", buf);
    }

    CODE_LOCK();
        if (pc[0] != opcode && pc[0] != opc_breakpoint) {
	    result = 1;
	} else if (opcode == opc_invokestatic) { 
	    pc[0] = opc_invokestatic_quick;
	} else if (opcode == opc_invokevirtual) {
	    unsigned int offset = methodoffset(mb);
	    if (methodaccess(mb) & ACC_PRIVATE) {
	        pc[0] = opc_invokenonvirtual_quick;
	    } else if (offset <= 0xFF && !UseLosslessQuickOpcodes) {
		pc[1] = methodoffset(mb); 
		pc[2] = methodargssize(mb); 
		pc[0] = methodclass(mb) == classJavaLangObject 
		              ? opc_invokevirtualobject_quick
		              : opc_invokevirtual_quick;
	    } else { 
		pc[0] = opc_invokevirtual_quick_w;
	    }
	} else if (opcode == opc_invokespecial) {
	    JavaFrame  *frame = ee->current_frame;
	    ClassClass *current_class = frame->current_method ? 
		methodclass(frame->current_method) : NULL;
	    struct methodblock *new_mb = mb;
	    unsigned int offset = methodoffset(mb); 
	    if ((current_class != NULL) && 
		     isSpecialSuperCall(current_class, mb)) {
		ClassClass *superClass = cbSuperclass(current_class);
		new_mb = mt_slot(cbMethodTable(superClass), offset);
	    }
	    if (mb == new_mb) { 
		pc[0] = opc_invokenonvirtual_quick;
	    } else { 
		pc[1] = offset >> 8;
		pc[2] = offset & 0xFF;
		pc[0] = opc_invokesuper_quick;
		mb = new_mb;
	    } 
	} 
    CODE_UNLOCK();
    return result;
}


/* Called just after a method has been called from a particular pc for the
 * first time.  See if we can eliminate the method call and inline instead.
 */
void FixupQuickInvocation(unsigned char *pc, 
			  struct methodblock *currentMethod,
			  struct methodblock *targetMethod)
{ 
    unsigned char rewrite[3];
    if (MethodCallInline(pc, currentMethod, targetMethod, rewrite)) {
	CODE_LOCK(); 
	    memcpy(pc, rewrite, 3); 
	CODE_UNLOCK();
    }
}    

bool_t
invokeInterfaceError(ExecEnv *ee, unsigned char *pc, 
			  ClassClass *cb, ClassClass *intf)
{ 
    char buf[256];
    char *bufptr = buf, *endbuf = buf + 256;

    jio_snprintf(bufptr, endbuf - bufptr, "class ");
       bufptr += strlen(bufptr);

    classname2string(cbName(cb), bufptr, endbuf - bufptr);
       bufptr += strlen(bufptr);

    jio_snprintf(bufptr, endbuf - bufptr, " does not implement interface ");
       bufptr += strlen(bufptr);

    classname2string(cbName(intf), bufptr, endbuf - bufptr);
    JAVA_ERROR("IncompatibleClassChangeError", buf);
}


/*
 * Note: Call InitializeInterpreter() BEFORE InitializeClassThread() to
 * ensure that the Java stack lock is initialized before the idle thread
 * has a chance to run!
 */
void
InitializeInterpreter()
{
    _code_lock=(sys_mon_t *) sysMalloc(sysMonitorSizeof());
    memset((char*) _code_lock, 0, sysMonitorSizeof());
    CODE_LOCK_INIT();
    _ostack_lock=(sys_mon_t *) sysMalloc(sysMonitorSizeof());
    memset((char*) _ostack_lock, 0, sysMonitorSizeof());
    OSTACK_LOCK_INIT();

    _loadclass_lock=(sys_mon_t *) sysMalloc(sysMonitorSizeof());
    memset((char*) _loadclass_lock, 0, sysMonitorSizeof());
    LOADCLASS_LOCK_INIT();
    _binclass_lock=(sys_mon_t *) sysMalloc(sysMonitorSizeof());
    memset((char*) _binclass_lock, 0, sysMonitorSizeof());
    BINCLASS_LOCK_INIT();

    InitializeJNI();
    InitializeClassConstantResolver();
#ifdef STATISTICS
    memset(instruction_count, 0, sizeof(instruction_count));
#endif
#if defined(IAL) && defined(SELFTEST)
	ExecuteJavaSelfTest();
#endif

}

#ifdef DEBUG
void PrintObject(JHandle *oh) {
    if (oh == 0)
        jio_fprintf(stdout, "NULL\n");
    else {
	ClassObject *o = unhand(oh);
	if (-1000 < (long) o && (long) o < 1000)
	    jio_fprintf(stdout, "BOGUS-PTR[%d]\n", (int)o);
	else if (obj_is_array(oh))
	    jio_fprintf(stdout, "%s\n", Object2CString(oh));
	else {
	    ExecEnv *current_ee = threadSelf() 
	                            ? (ExecEnv *)(unhand(threadSelf())->eetop)
				    : 0;
	    ExecEnv lee;
	    ClassClass *system_class;
	    ExecEnv *ee = current_ee ? current_ee 
	                         : (InitializeExecEnv(&lee, 0), &lee);
	    if (ee->initial_stack == 0) {
		jio_fprintf(stderr, "Out of memory\n");
	    }
	    system_class = FindClassFromClass(&lee, "java/lang/System", TRUE, NULL);
	    if (system_class == 0) {
		jio_fprintf(stderr, "Unable to find System class\n");
	    } else {
		JHandle *out = *(JHandle **)getclassvariable(system_class, "out");
#ifdef TRACING
		int oldtrace = trace_is_enabled;   /* we don't want to see this */
		trace_is_enabled = 0;
#endif /* TRACING */
		execute_java_dynamic_method(ee, out, 
					   "println", "(Ljava/lang/Object;)V",
					   oh);
#ifdef TRACING
		trace_is_enabled = oldtrace;
#endif /* TRACING */
	    }
	    if (current_ee == 0)
	        DeleteExecEnv(&lee);
	}
    }
}    
#endif

void *
ResolveClassConstantFromPC(unsigned char *pc, unsigned char opcode, 
			       cp_item_type *constant_pool, 
			       struct execenv *ee, 
			       unsigned mask) { 
    unsigned char opcode2;
    unsigned index;
    CODE_LOCK();
	opcode2 = pc[0];
        index = GET_INDEX(pc + 1);
    CODE_UNLOCK();
#ifdef BREAKPTS
    if (opcode2 == opc_breakpoint) {	/* Find the real opcode */
	opcode2 = get_breakpoint_opcode(ee, pc, FALSE);
    }
#endif /* BREAKPTS */
    if (opcode2 != opcode) { 
	return NULL;
    } else { 
	ResolveClassConstant(constant_pool, index, ee, mask);
	return (exceptionOccurred(ee)) ? NULL : constant_pool[index].p;
    } 
}
    


#ifdef STATISTICS
void
DumpStatistics ()
{
    int totals[256], total = 0;
    int percent;
    int i, j;
    for (i = 0; i < 256; i++) {
	int sum = 0;
	for (j = 0; j < 256; j++) sum += instruction_count[i][j];
	totals[i] = sum;
	total += sum;
    }
    for (i = 0; i < 256; i++) { 
	if (totals[i])
	    jio_fprintf(sdtout, "%23.-23s %8d\n", opnames[i], totals[i]);
    }
    jio_fprintf(stdout, "%23.-23s %8d\n", "TOTAL", total);
    percent = total/100;
    for (i = 0; i < 256; i++) {
	for (j = 0; j < 256; j++) { 
	    if (instruction_count[i][j] >= percent) { 
		jio_fprintf(stdout, "%23.-23s => %23.-23s %8d %4.2f (%4.2f)\n", 
		       opnames[i], opnames[j], instruction_count[i][j], 
		       instruction_count[i][j] * 100.0 / total, 
		       instruction_count[i][j] * 100.0 / totals[i]);
	    }
	}
    }

}
#endif

#ifdef TRACING
void
trace_method(struct execenv* ee, struct methodblock* mb,
	     int args_size, int type)
{
    HArrayOfChar* threadName = threadSelf() ? getThreadName() : NULL;
    JavaFrame *frame;
    int depth;
    int i;
    for (frame = ee->current_frame, depth = 0;
	   frame != NULL;
	   frame = frame->prev)
	if (frame->current_method != 0) depth++;
    if (type == TRACE_METHOD_RETURN)
	depth--;
    jio_fprintf(stdout, "# ");
    if (threadName != NULL) {
	int threadNameLen = obj_length(threadName);
	unicode *threadNameStr = unhand(threadName)->body;
	jio_fprintus(stdout, threadNameStr, threadNameLen);
    } else {
	jio_fprintf(stdout, "???");
    }
    jio_fprintf(stdout, " [%2d] ", depth);
    for (i = 0; i < depth; i++)
	jio_fprintf(stdout, "| ");
    jio_fprintf(stdout, "%c %s.%s%s ",
		((type == TRACE_METHOD_ENTER) ? '>' : '<'),
		cbName(methodclass(mb)), methodname(mb), methodsig(mb));
    if (type == TRACE_METHOD_ENTER)
	jio_fprintf(stdout, "(%d) entered\n", args_size);
    else if (exceptionOccurred(ee))
	jio_fprintf(stdout, "throwing %s\n", cbName(obj_classblock(ee->exception.exc)));
    else
	jio_fprintf(stdout, "returning\n");
    fflush(stdout);
}
#endif


/* Step back n frames, skip and pseudo frames in between.
 * This function is used in Class.forName, Class.newInstance, Method.Invoke, 
 * AccessController.doPrivileged.
 */

struct methodblock *reflect_invoke_mb = 0;

JavaFrame *
getCallerFrame(JavaFrame * frame, int n, JavaFrame * frame_buf)
{
    struct methodblock *invoke_mb = reflect_invoke_mb;

    while (n-- > 0) {
        /* skip pseudo frames */
        while (frame && (frame->current_method == 0 ||
			 frame->current_method == invoke_mb)) {
	    frame = frame->prev;
	}

	if (frame == 0) {
	    return 0;
	}

        /* skip a real frame */
	frame = FRAME_PREV(frame, frame_buf);
    }

    /* skip pseudo frames */
    while (frame && (frame->current_method == 0 ||
		     frame->current_method == invoke_mb)) {
        frame = frame->prev;
    }

    return frame;
}

ClassClass *
getCallerClass(ExecEnv * ee, int skip)
{
    JavaFrame frame_buf;
    JavaFrame * frame = getCallerFrame(ee->current_frame, skip, &frame_buf);
    if (frame) {
	return methodclass(frame->current_method);
    } else {
	return NULL;
    }
}
