/*
 * @(#)monitor.c	1.47 98/11/13
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * System-independent, reentrant monitors.
 */

#include "monitor.h"
#include "monitor_cache.h"
#include "threads.h"
#include "log.h"
#include "sys_api.h"

#include "java_lang_Thread.h"

/*-
 * Look up a monitor in the monitor cache and enter it using
 * monitorEnter().  If there is no monitor cached for the given key
 * create one on the fly.
 */
void
monitorEnter(unsigned int key)
{
    monitor_t *mon;

    mon = createMonitor(key); 		/* Increments mon->use_count */
    sysAssert(mon != MID_NULL);
    sysMonitorEnter(sysmon(mon));
}

void
monitorExit(unsigned int key)
{
    monitor_t *mon;

    CACHE_LOCK();
    mon = lookupMonitor(key);
    if (mon == MID_NULL || sysMonitorExitLocked(sysmon(mon)) != SYS_OK) {
	CACHE_UNLOCK();		/* Cache lock not guaranteed reentrant */
	SignalError((struct execenv *) THREAD(threadSelf())->eetop,
		    JAVAPKG "IllegalMonitorStateException",
		    "current thread not owner");
	return;
    }

    /* REMIND: is this needed any more or can it be done in monitor GC? */
    mon->use_count--;  /* Have to decrement before monitorDestroy */
    CACHE_UNLOCK();
}

/*-
 * Wake up the first thread waiting on the condition variable
 * associated with this monitor.
 */
void
monitorNotify(unsigned int key)
{
    monitor_t *mon;

    CACHE_LOCK();
    mon = lookupMonitor(key);
    CACHE_UNLOCK();

    if (mon == MID_NULL || sysMonitorNotify(sysmon(mon)) != SYS_OK) {
	SignalError((struct execenv *) THREAD(threadSelf())->eetop,
		    JAVAPKG "IllegalMonitorStateException",
		    "current thread not owner");
	return;
    }
}

/*-
 * Wake up *all* the threads waiting on the condition variable
 * associated with this monitor.
 */
void
monitorNotifyAll(unsigned int key)
{
    monitor_t *mon;

    CACHE_LOCK();
    mon = lookupMonitor(key);
    CACHE_UNLOCK();

    if (mon == MID_NULL || sysMonitorNotifyAll(sysmon(mon)) != SYS_OK) {
	SignalError((struct execenv *) THREAD(threadSelf())->eetop,
		    JAVAPKG "IllegalMonitorStateException",
		    "current thread not owner");
	return;
    }
}

void
monitorWait(unsigned int key, int millis)
{
    monitor_t *mon;
    int ret;
    struct execenv *ee = EE();

    CACHE_LOCK();
    mon = lookupMonitor(key);
    CACHE_UNLOCK();

    if (mon == MID_NULL) {
	ret = SYS_ERR;
    } else {
	ret = sysMonitorWait(sysmon(mon), millis, TRUE);
	if (ret == SYS_TIMEOUT) {
	    ret = SYS_OK;
	}
    }

    /* An asynchronous exception might have been thrown while we
     * wait on the monitor.
     * REMIND: 
     * This still does not fix the inherent race condition
     * where an async exception can be thrown after this check
     * is done. However, this race condition exists everywhere
     * else in the VM and should be fixed thoroughly in a future
     * release. 
     */
    if (exceptionOccurred(ee))
        return;

    if (ret == SYS_INTRPT) {
	SignalError(ee,
		    JAVAPKG "InterruptedException",
		    "operation interrupted");
	return;
    }

    if (ret != SYS_OK) {
	SignalError(ee,
		    JAVAPKG "IllegalMonitorStateException",
		    "current thread not owner");
	return;
    }
}

/*-
 * Registration of monitors not allocated from the monitor cache.
 *
 * All monitors except those in the monitor cache should be registered
 * via monitorRegister().
 *
 * One tricky point is that access to the monitor registry is
 * protected by a registered monitor.  That may require special
 * treatment when doing some operations, e.g. when a thread is dying
 * it leaves all held monitors, but uses the registry monitor to do
 * so.  The responsibility is currently put on the helper function
 * passed to registeredEnumerate() to take care of this special case.
 * registeredEnumerate() is intended to be the only access to the
 * monitor registry.
 *
 * Note also that monitorRegister() is used on the registry lock, and
 * also uses the lock.  This works because we don't actually need the
 * monitor to be registered to use it -- we only need it to be
 * initialized.
 */

sys_mon_t *_registry_lock;
reg_mon_t *MonitorRegistry;  /* Root of registered monitors */

void
monitorRegistryInit()
{
    _registry_lock = (sys_mon_t *) sysMalloc(sysMonitorSizeof());
    memset((char*)_registry_lock, 0, sysMonitorSizeof());
    REGISTRY_LOCK_INIT();
}

void
monitorRegister(sys_mon_t *mid, char *name)
{
    reg_mon_t *reg;

    sysMonitorInit(mid);
    reg = (reg_mon_t *) sysMalloc(sizeof(reg_mon_t));
    if (!reg) {
	SignalError((struct execenv *) THREAD(threadSelf())->eetop,
		    JAVAPKG "OutOfMemoryError", 0);
	return;
    }
    reg->mid = mid;
    if (!(reg->name = strDup(name))) {
	SignalError((struct execenv *) THREAD(threadSelf())->eetop,
		    JAVAPKG "OutOfMemoryError", 0);
	return;
    }
    REGISTRY_LOCK();
    reg->next = MonitorRegistry;
    MonitorRegistry = reg;
    REGISTRY_UNLOCK();
}

/*-
 * If monitors are dynamically allocated and deallocated (none currently
 * are), it will be necessary that a registered monitor be unregistered
 * prior to deallocation to avoid dangling pointers in the monitor 
 * registry.
 */
void
monitorUnregister(sys_mon_t *mid)
{
    reg_mon_t **prevp = &MonitorRegistry;
    reg_mon_t *reg;

    REGISTRY_LOCK();
    reg = *prevp;
    while (reg && reg->mid != mid) {
	prevp = &reg->next;
	reg = reg->next;
    }

    sysAssert(reg);	/* Else bad call or inconsistent state */

    *prevp = reg->next;
    sysFree(reg->name);
    sysFree(reg);
    REGISTRY_UNLOCK();
}

/*-
 * Call a function for each registered monitor.  Cf. monitorEnumerate().
 *
 * It is intended that this is the only access to the monitor registry.
 * The monitors it contains, of course, are as accessible as desired.
 * The unlocked version should be called with the registry lock held.
 * The unlocked version is used for debug purposes when the lock is alredy
 * held - see debug.c, and solaris/native_threads/interrupt_md.c
 */

void
registeredEnumerate_unlocked(void (*fcn)(reg_mon_t *, void *), void *cookie)
{
    reg_mon_t *reg;

    /*
     * Commented out because the Win32 control-break handler isn't called
     * from a Java thread:
     *
     * sysAssert(REGISTRY_LOCKED());
     */
    reg = MonitorRegistry;
    while (reg) {
	reg_mon_t *next = reg->next;
	fcn(reg, cookie);
	reg = next;
    }
}

void
registeredEnumerate(void (*fcn)(reg_mon_t *, void *), void *cookie)
{
    REGISTRY_LOCK();
    registeredEnumerate_unlocked(fcn, cookie);
    REGISTRY_UNLOCK();
}
