/*
 * @(#)monitor_cache.c	1.57 98/08/31
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Code to manage the monitor cache.
 *
 * The monitor cache is a hash table which contains monitors.  The
 * hash function hashes an arbitrary value (typically the address of
 * an object) to a single monitor structure.  The monitor cache and
 * its hash table grow as necessary, and are only limited by available
 * memory.  The hash table is grown when the total number of monitors
 * is HASHTABLE_MULTIPLE times the number of hash table buckets, i.e.
 * when the average hash chain length with all monitors in use would
 * be HASHTABLE_MULTIPLE.
 *
 * Once the cache has expanded, it is never contracted, although doing
 * so would be possible.
 */

#include "threads.h"
#include "monitor.h"
#include "monitor_cache.h"
#include "log.h"

#include "sys_api.h"
#include "monitor_md.h"

/*
 * The following defines should preserve for initialization to work
 *	INIT_MONITOR_COUNT <= EXPANSION_QUANTUM * INIT_MONITOR_COUNT
 */
#define INIT_HASHTABLE_BUCKETS	32 /* Must be a power of two! */
#define INIT_MONITOR_COUNT	32 /* Initial number of monitors */
#define EXPANSION_QUANTUM	16 /* Number of monitors to expand by */
#define HASHTABLE_MULTIPLE	2  /* Length of avg hash chain before expand */
#define HASH_MASK		(monHashTableBuckets - 1)
#define HASH(k)			(((k >> 2) ^ (k >> 10)) & HASH_MASK)

/*
 * When the number of free monitors falls below FREE_MONITOR_THRESHOLD we
 * expand.  This is needed because some platforms need monitors in the
 * expansion code (e.g. Green threads on Solaris does a malloc which
 * may use at least two monitors).  Because the monitor cache is in
 * shared code, this number has to be conservative.
 */
#define FREE_MONITOR_THRESHOLD 5

/*
 * The hash table itself and its size.  We can't dynamically allocate the
 * initial hash table because we would need monitors to do so.  If the
 * table is eventually expanded, we just lose the initial memory.
 */
static monitor_t *initMonHashTable[INIT_HASHTABLE_BUCKETS] = {0};
monitor_t **monHashTable = &initMonHashTable[0];
int monHashTableBuckets = INIT_HASHTABLE_BUCKETS;

/*
 * The total number of monitors in the monitor cache
 */
int monCount = 0;

/*
 * The monitor cache free list and count of free monitors.  We use
 * the count to determine when we have fewer than FREE_MONITOR_THRESHOLD
 * monitors free, in which case we expand the cache.  We can't wait
 * for the free list to be empty before we expand, as some platforms
 * need monitors to expand (e.g. malloc on Solaris may use as many
 * as two monitors).  We need to overestimate here because this code
 * and FREE_MONITOR_THRESHOLD are shared across all platforms. 
 */
static monitor_t *monFreeList = 0;
static int monFreeCount = 0;

/*
 * This is a simple last used monitor cache, based on the assumption
 * that a single monitor is used repeatedly before another monitor is
 * used.
 */
static monitor_t *lastMonUsed = 0;

#ifdef DEBUG
static int monLookups = 0;
static int monQuickFind = 0;
static int monHashFind = 0;
static int monCreate = 0;
#endif

/*
 * Must be called for each cached monitor prior to use.
 */
void 
monitorInit(monitor_t *mon) 
{
    sysAssert(mon != MID_NULL);
    sysMonitorInit(sysmon(mon));
}

/*
 * Must be called for each cached monitor once it will never again be
 * used.  Note that cached monitors are never monitorDestroy'd until
 * sysExit, and doing so would be disasterous on some platforms.  DO NOT
 * CONFUSE THIS WITH monitorExit!  This cleans up resources on runtime
 * exit and is the converse of monitorInit.
 */
void 
monitorDestroy(monitor_t *mon) 
{
    sysAssert(mon != MID_NULL);
    sysMonitorDestroy(sysmon(mon));
}

/*
 * Scan for unused cached monitors and free them to the cache.  Return
 * the number freed.
 *
 * All monitors in the hash table should be considered, as they are
 * nominally in use.  A monitor is actually unused if its entry_count
 * is 0 and nobody is on any of its wait queues.  When these conditions
 * are true when a thread leaves a monitor (see sysMonitorExit())
 * the monitor's flags and monitor_owner field are also reset, but those
 * are not the determiners.
 *
 * We don't reclaim memory or other resources for monitors here.  The
 * monitor is simply reinitialized and put into a free list for reuse.
 * As we never allocate more monitors than we actually need (modulo some
 * padding), eagerly freeing monitors causes unnecessary overhead.  If we
 * grew that big once we are likely to do so again.
 */
static int
monitorCacheScavenge()
{
    int i;
    int count = 0;
    monitor_t **monp;
#ifdef DEBUG
    long start_time = now();
    long end_time;
    int monCountLocal;
    int monFreeCountLocal;
#endif

    for (i = monHashTableBuckets; --i >= 0; ) {
	for (monp = &monHashTable[i]; *monp != 0; ) {
	    monitor_t *mon = *monp;
	    if (mon->use_count == 0) {
		sysAssert(sysMonitorOwner(sysmon(mon)) == NULL);
		/* Unlink the monitor from the hash chain */
		*monp = mon->next;
		/*
		 * Reinitialize monitor and put it on the free list.
		 * sys_mon_t reinitialization is done on sysMonitorExit.
		 */
		mon->key = 0;
		mon->next = monFreeList;
#ifdef DEBUG
		mon->flags = 0;		/* Only checked by assertions */
#endif
		monFreeList = mon;
		monFreeCount++;
		count++;
	    } else {			/* In use; skip over this one */
		monp = &((*monp)->next);
	    }
	}
    }

#ifdef DEBUG
    monCountLocal = monCount;
    monFreeCountLocal = monFreeCount;
#endif /* DEBUG */

    /*
     * The current value of lastMonUsed must be garbage because it caches a
     * monitor that's in the cache, and we wouldn't be collecting if we were
     * wanting a monitor that's already in the cache.  Clearing it is not an
     * option, though: it might end up a dangling reference!
     */
    lastMonUsed = 0;

#ifdef DEBUG
    end_time = now();
    if (verbosegc) {
	jio_fprintf(stderr,
    "<Monitor cache GC freed %d of %d monitors in %ld ms (%d total free)>\n",
		count, monCountLocal, end_time-start_time, monFreeCountLocal);
    }
#endif

    return count;
}

/*
 * Add count monitors to the front of the monitor free list.  This
 * doesn't assume that the monFreeList is empty, and in fact it
 * probably isn't (see comment below when monitorCacheExpand() is
 * called from createMonitor()).  This is called with the monitor
 * cache *unlocked*.
 *
 * There is a crucial and nonobvious connection between the delayed
 * increment of monFreeCount in monitorCacheExpand() and the threshold
 * test in createMonitor().  Once we have started to expand, we want
 * to to get other threads wanting monitors into the createMonitor()
 * code that causes them to wait until expansion finishes.  This
 * avoids nasty problems like grabbing monitors in the unprotected
 * region of createMonitor() between adding new monitors to the
 * cache and expanding the hash table.  By postponing resetting
 * monFreeCount until after all expansion is done, we guarantee that
 * all threads creating monitors will be caught by the single, fast
 * threshold test.  Once inside, further tests decide what to do
 * with them, but the crucial normal case (no expansion) remains fast.
 */
void
monitorCacheExpand(int count)
{
    int nbytes = count * (sizeof(monitor_t) + sysMonitorSizeof());
    monitor_t *mon;
    monitor_t *start;
    int i;

    if ((mon = (monitor_t *) sysMalloc(nbytes)) == 0) {
	out_of_memory();	/* We currently can't survive this */
    }

    Log2(2, "Expanding monitor cache by %d monitors to %d\n",
	 count, monCount+count);

    start = mon;
    memset((char *) mon, 0, nbytes);
    CACHE_LOCK();
    monCount += count;
    /* monFreeCount += count; Postponed!  See comment above */
    i = count;
    while (--i > 0) {
	monitor_t *next = (monitor_t *)
	    ((char *)mon + (sizeof(monitor_t) + sysMonitorSizeof()));
	monitorInit(mon);
	mon->next = next;
	mon = next;
    }
    monitorInit(mon);
    mon->next = monFreeList;
    monFreeList = start;

    /*
     * If after expansion the monCount is greater than or equal to 
     * HASHTABLE_MULTIPLE times the hash table size, expand the hash table
     * as well.  That is to say that we expand when the average hash chain
     * with all monitors in use is HASHTABLE_MULTIPLE long.
     */
    if (monCount < HASHTABLE_MULTIPLE * monHashTableBuckets) {
	monFreeCount += count; 		/* Increment when done expanding */
	CACHE_UNLOCK();
    } else {
	/* Expand the hash table: size must be a multiple of two */
	int hash, i;
	int oldMonHashTableBuckets = monHashTableBuckets;
	int newMonHashTableBuckets = monHashTableBuckets;
	monitor_t **oldMonHashTable = monHashTable;
	monitor_t **newMonHashTable;
	monitor_t **hashp;

	do {
	    newMonHashTableBuckets *= 2;
	} while (monCount >= HASHTABLE_MULTIPLE * newMonHashTableBuckets);

	Log2(2, "Expanding monitor cache hash table from %d to %d buckets\n",
	     monHashTableBuckets, newMonHashTableBuckets);
	nbytes = newMonHashTableBuckets * sizeof(monitor_t *);
	/*
	 * Remember that malloc may use a monitor, so needs to have
	 * a consistent monHashTable and monHashTableBuckets.  After
	 * we reset monHashTable and monHashTableBuckets we need to
	 * be sure not to enter any cache monitors before we've
	 * rehashed everything.
	 */
        CACHE_UNLOCK();
	if ((newMonHashTable = (monitor_t **) sysMalloc(nbytes)) == 0) {
	    /*
	     * We didn't get what we had hoped for, but we should be able
	     * to continue using the old table albeit with slower cache 
	     * access.  We'll try to expand the hash table again next time
	     * we allocate another block of monitors, if we haven't died
	     * by running out of malloc space yet.
	     */
	    CACHE_LOCK();
	    monFreeCount += count;	/* Increment when done expanding */
	    CACHE_UNLOCK();
	    return;
	}
	CACHE_LOCK();
	monHashTable = newMonHashTable;
	monHashTableBuckets = newMonHashTableBuckets;
	memset((char *) monHashTable, 0, nbytes);
	for (i = 0; i < oldMonHashTableBuckets; i++) {
	    mon = oldMonHashTable[i];
	    while (mon != 0) {
		monitor_t *next = mon->next;
		hash = HASH(mon->key);	/* must have new size set here */
		hashp = &monHashTable[hash];
		mon->next = *hashp;
		*hashp = mon;
		mon = next;
	    }
	}
	monFreeCount += count; 		/* Increment when done expanding */
	CACHE_UNLOCK();
        if (oldMonHashTable != &initMonHashTable[0]) {
            sysFree(oldMonHashTable);
        }
    }
}

/*
 * Expand the monitor cache if someone else isn't already doing so.
 * Return TRUE if the caller should immediately try to reallocate the
 * monitor, and FALSE otherwise.  The latter case can occur when
 * recursively expanding the cache, in which case we want to keep
 * working out of the FREE_MONITOR_THRESHOLD buffer of monitors.
 */
static sys_thread_t *expandingFreeList = 0;

bool_t
monitorCacheMaybeExpand()
{
    bool_t tryAgain = FALSE;

    sysAssert(CACHE_LOCKED());
    CACHE_UNLOCK();

    /*
     * During bootstrap we might not have a queue_lock yet.  In this
     * case we can't do any scavenging so we just need to expand the
     * cache.  This shouldn't happen in the normal case but it depends
     * on what the starting size of the monitor_cache is.
     */
    if (_queue_lock == NULL) {
	monitorCacheExpand(EXPANSION_QUANTUM);
	CACHE_LOCK();
	return TRUE;
    }

    /* The lock usage here is a little bizarre because on Green
     * Threads the CACHE_LOCK is the scheduler lock.  You can't
     * acquire new locks when holding the scheduler lock and you can't
     * wait on it.  This means that we can't use the simple approach
     * of locking the CACHE_LOCK and simply doing whatever we need to
     * do.  Instead we're going to use QUEUE_LOCK for that.  We need
     * to hold the QUEUE_LOCK for the call to monitorCacheScavenge
     * since it calls sysThreadEnumerateOver which requires this lock
     * to be held.  We still need to lock CACHE_LOCK to protect the
     * monitor cache itself.  Other threads that want to expand the
     * monitor cache end up waiting on the QUEUE_LOCK.  Once an
     * expansion happens, the waiters are woken up with a notifyAll on
     * the QUEUE_LOCK.  This is less than ideal but is better than the
     * alternatives available right now.
     */

    QUEUE_LOCK();

    CACHE_LOCK();
    /* Redundant threshold test in case this thread blocked and some else GC */
    if (!expandingFreeList && monFreeCount < FREE_MONITOR_THRESHOLD) {
	monitorCacheScavenge();
	if (monFreeCount < FREE_MONITOR_THRESHOLD) {
	    CACHE_UNLOCK();
	    expandingFreeList = sysThreadSelf();
	    QUEUE_UNLOCK();

	    /*
	     * We can't hold CACHE_LOCK when calling this because of Green
	     * threads.  monitorCacheExpand ends up calling malloc which
	     * will drop back into the monitor code to lock the heap.
	     */ 
	    monitorCacheExpand(EXPANSION_QUANTUM);


	    /*
	     * We're done so notify any threads which are waiting for
	     * us to finish expanding the monitor cache.
	     */
	    QUEUE_LOCK();
	    expandingFreeList = 0;
	    sysMonitorNotifyAll(_queue_lock);
	    QUEUE_UNLOCK();
	    tryAgain = TRUE;
	} else {
	    /*
	     * We've scavenged enough monitors to keep on going so
	     * return, telling our caller to try again.
	     */
	    CACHE_UNLOCK();
	    QUEUE_UNLOCK();
	    tryAgain = TRUE;
	}
    } else if (expandingFreeList != sysThreadSelf()) {
	CACHE_UNLOCK();
	/* Somebody is already expanding the free list, so wait */
	while (expandingFreeList) {
	    QUEUE_WAIT();
	}
	/*
	 * Someone might have tried to create a monitor but was
	 * forced to wait because someone else was expanding the
	 * cache.  During that time the expander might have created
	 * a monitor for the waiter's key, so when the waiter wakes
	 * up he needs to start the allocation from scratch.
	 */
	QUEUE_UNLOCK();
	tryAgain =  TRUE;
    } else {
	CACHE_UNLOCK();
	QUEUE_UNLOCK();
	/*
         * Else *I* am expanding the list, so let me continue using
         * monitors out of the FREE_MONITOR_THRESHOLD buffer.
         */
    }

    CACHE_LOCK();
    return tryAgain;
}


/* Initialize the monitor cache
 * 
 * Called during bootstrapping, so should be no chance of interruption.
 *
 * We need a sys_-level monitor initialization step.  We don't need
 * a sys_-level monitor destruction, as we never really destroy a
 * cache monitor -- they are just left in the table to be used again.
 */

bool_t  monitorsInitialized = FALSE;

void
monitorCacheInit()
{
    CACHE_LOCK_INIT();		/* The monitor cache lock is special */

    monitorCacheExpand(INIT_MONITOR_COUNT);
    monitorsInitialized = TRUE;
}

/*
 * Destroy the current monitor cache.  This is only done at sysExit, and
 * must not be done earlier; Java code cannot be run safely after the
 * monitor cache has been destroyed.  Doing this explicitly makes gives
 * tools like Purify a clearer view of what memory and system resources
 * are being leaked as opposed to being implicitly cleaned up on exit.
 * 
 * We currently do NOT free memory for the platform-independent portion
 * of the cache on sysExit.
 */
void
monitorCacheDestroy()
{
    int i;
    monitor_t *mon;

    CACHE_LOCK();	/* Just in case... */

    monitorsInitialized = FALSE;
    for (i = 0; i < monHashTableBuckets; i++) {
	mon = monHashTable[i];
	if (mon != 0) {
	    while (mon != 0) {
		monitorDestroy(mon);
	        mon = mon->next;
	    }
	}
    }

    CACHE_UNLOCK();
}

/*
 * Look up a monitor in the hash table.  Return MID_NULL if it is not
 * there.  Assumes we have control of the monitor cache lock.
 */
monitor_t *
lookupMonitor(unsigned int key)
{
    monitor_t *mon;
    int	hash;

    sysAssert(CACHE_LOCKED());
    
#ifdef DEBUG
    monLookups += 1;
#endif

    if ((mon = lastMonUsed) != 0 && mon->key == key) {
	sysAssert((mon->flags & MON_IN_USE) != 0);
#ifdef DEBUG
	monQuickFind += 1;
#endif
	return mon;
    }

    hash = HASH(key);
    mon = monHashTable[hash];
    while (mon != 0) {
	/* All monitors in hash table are in use */
	sysAssert((mon->flags & MON_IN_USE) != 0);
	if (mon->key == key) {
#ifdef DEBUG
	    monHashFind += 1;
#endif
	    return (lastMonUsed = mon);
	}
	mon = mon->next;
    }

    return MID_NULL;
}

/*
 * Create a monitor for the specified key.  If it already exists,
 * just return it.
 */
monitor_t *
createMonitor(unsigned int key)
{
    int	hash;
    monitor_t **hashp, *mon;
    monitor_t *ret;

    CACHE_LOCK();

#ifdef DEBUG
    monLookups += 1;
#endif

RetryCreate:
    if ((mon = lastMonUsed) != 0 && mon->key == key) {
	sysAssert((mon->flags & MON_IN_USE) != 0);
#ifdef DEBUG
	monQuickFind += 1;
#endif
	ret = mon;
	goto unlock;
    }

    /* First look for the monitor, in case it's in use */
    hash = HASH(key);
    mon = monHashTable[hash];
    while (mon != 0) {
	/* All monitors in hash table are in use */
	sysAssert((mon->flags & MON_IN_USE) != 0);
	if (mon->key == key) {
#ifdef DEBUG
	    monHashFind += 1;
#endif
	    ret = (lastMonUsed = mon);
	    goto unlock;
	}
	mon = mon->next;
    }

    /*
     * If we get here, a monitor for key has not been allocated.  Now
     * we grab one from the free list, and add it to the hash table.  If
     * handing over the next monitor pushes us below FREE_MONITOR_THRESHOLD
     * free monitors, expand the monitor cache *first*.  On systems that
     * don't use native threads, like on Solaris with Green threads, 
     * sysMalloc() is probably made thread safe using a monitor from the
     * monitor cache, and in order to expand the free list, we need to
     * call sysMalloc().  So we make sure we have some small number of
     * monitors available when we expand.  We also have to play some
     * tricks to avoid deadlock between the malloc and monitor cache
     * monitors.
     */
    if (monFreeCount < FREE_MONITOR_THRESHOLD) {
        if (monitorCacheMaybeExpand() == TRUE) {
    	    goto RetryCreate;
        }   /* Otherwise, allocate out of FREE_MONITOR_THRESHOLD */
    }

    sysAssert(monFreeList != 0);
    /* We don't want to recreate anything that was created by expansion: */
    sysAssert(lookupMonitor(key) == MID_NULL);

#ifdef DEBUG
    monCreate += 1;
#endif
    mon = monFreeList;
    monFreeList = mon->next;
    monFreeCount--;
    sysAssert((mon->flags & MON_IN_USE) == 0);

    mon->key = key;
#ifdef DEBUG
    mon->flags |= MON_IN_USE;
#endif
    hashp = &monHashTable[hash];
    mon->next = *hashp;
    *hashp = mon;
    /* REMIND: Why don't we put the new monitor in lastMonUsed??? */
    ret = mon;

unlock:
    ret->use_count++;
    CACHE_UNLOCK();
    return ret;
}

/*
 * monitorEnumerate_unlocked() should be called with the cache lock held.
 * It is used for debug purposes.
 */
void
monitorEnumerate_unlocked(void (*fcn)(monitor_t *, void *), void *cookie)
{
    monitor_t **hashtable = monHashTable;
    monitor_t *mon, *next;
    int	i;

    sysAssert(CACHE_LOCKED());
    for (i = monHashTableBuckets; --i >= 0; ) {
	mon = hashtable[i];
	while (mon != 0) {
	    next = mon->next;
	    fcn(mon, cookie);
	    mon = next;
	}
    }
}

void
monitorEnumerate(void (*fcn)(monitor_t *, void *), void *cookie)
{
    CACHE_LOCK();
    monitorEnumerate_unlocked(fcn, cookie);
    CACHE_UNLOCK();
}

#ifdef DEBUG

void DumpMonitorCacheInfo()
{
    int i;
    monitor_t *mon;

    jio_fprintf(stderr, "Monitor cache info:\n");
    jio_fprintf(stderr, "  Initial monitor count: %d\n", INIT_MONITOR_COUNT);    
    jio_fprintf(stderr, "  Initial hash table buckets: %d\n",
	    INIT_HASHTABLE_BUCKETS);  
    jio_fprintf(stderr, "  Hash table expansion multiple: %d\n", HASHTABLE_MULTIPLE);
    jio_fprintf(stderr, "  Minimum number of free monitors before expansion: %d\n",
	    FREE_MONITOR_THRESHOLD);    
    jio_fprintf(stderr, "  Number of new monitors added per expansion: %d\n",
	    EXPANSION_QUANTUM);    
    jio_fprintf(stderr, "  Current total number of monitors: %d\n", monCount);
    jio_fprintf(stderr, "  Current number of free monitors: %d\n", monFreeCount);
    jio_fprintf(stderr, "  Current number of hash table buckets (power of 2!): %d\n",
	    monHashTableBuckets);

    jio_fprintf(stderr, "  Current hash table bucket dump:\n");
    for (i = 0; i < monHashTableBuckets; i++) {
	mon = monHashTable[i];
	if (mon != 0) {
	    jio_fprintf(stderr, "    Bucket %d: ", i);
	    while (mon != 0) {
	        jio_fprintf(stderr, "0x%lx ", mon);
	        mon = mon->next;
	    }
	    jio_fprintf(stderr, "\n");
	}
    }
}

#endif
