/*
 * @(#)threads.c	1.66 98/11/25
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include <sys/types.h>
#include <string.h>

#include "exceptions.h"
#include "threads.h"
#include "log.h"
#include "timeval.h"
#include "monitor.h"
#include "monitor_cache.h"
#include "sys_api.h"

#include "java_lang_Thread.h"
#include "java_lang_Throwable.h"

#ifdef HPROF
#include "vmprofiler.h"
#endif /* HPROF */

#ifdef BREAKPTS
#include "breakpoints.h"
#endif

/*
 * A count of all active threads and of those all user threads.  These
 * counts are maintained by code below the HPI.
 */
int ActiveThreadCount = 0;		/* All threads */
int UserThreadCount   = 0;		/* User threads */

sys_mon_t *_queue_lock;	/* Protects thread queue, thread count */

/*
 * The following are defined as macros in threads.h:
 *
 * stackp_t threadStackBase(TID tid)
 * void * threadStackPointer(TID tid)
 * void threadYield(void)
 * int threadSetPriority(TID tid, int priority)
 * int threadGetPriority(TID tid, int *priority)
 */

void
threadInit(TID tid, stackp_t sb)
{
    ExecEnv *ee = EE();
    sysThreadInit(SYSTHREAD(tid), sb);
#ifdef HPROF
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_THREAD_START)) {
        jvmpi_thread_start(ee->thread);
    }
#endif /* HPROF */
 
#ifdef BREAKPTS
    if (debugging) {
        notify_debugger_of_thread_start(ee, ee->thread);
    }
#endif /* BREAKPTS */

    ee->has_been_born = 1;
}

void
threadSuspend(TID p)
{
    sys_thread_t *t = SYSTHREAD(p);

    if (t != NULL) {
        (void) sysThreadSuspend(t);
    }
}

void
threadResume(TID p)
{
    sys_thread_t *t = SYSTHREAD(p);

    if (t != NULL) {
        (void) sysThreadResume(t);
    }
}

void
threadExit(void) 
{
#ifdef HPROF
    if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_THREAD_END)) {
	ExecEnv *ee = EE();
        jvmpi_thread_end(ee->thread);
    }
#endif /* HPROF */

#ifdef BREAKPTS
    if (debugging) {
	ExecEnv *ee = EE();
        notify_debugger_of_thread_end(ee, ee->thread);
    }
#endif /* BREAKPTS */
    sysThreadExit();
}

/*
 * Initialize the threads package.  The caller will return, executing
 * in the context of the first thread.
 */
void
threadBootstrap(TID tid)
{
    sys_thread_t *t;

    sysAssert(SYSTHREAD(tid) == 0);

    /* Initialize the queue lock monitor */
    _queue_lock = (sys_mon_t *)sysMalloc(sysMonitorSizeof());
    memset((char *)_queue_lock, 0, sysMonitorSizeof());
    QUEUE_LOCK_INIT();

    /* Bootstrap the current process into the main, primordial thread */

    sysThreadBootstrap(&t, (void *) tid);

    THREAD(tid)->PrivateInfo = (long) t;

    THREAD(tid)->priority = NormalPriority;
    threadSetPriority(tid, NormalPriority);

    /* Create system dependent threads */
    sysThreadInitializeSystemThreads();

    /* Preallocate the internal exception objects */
    exceptionInit();
}

/*
 * threadCreate() creates a new thread, returning 0 if things went OK
 * and 1 if not.  A return value of 1 will currently be interpreted as
 * an out of memory situation although it actually covers additional 
 * resource-related problems.  See the comment below.
 */
int
threadCreate(TID tid, unsigned int flags, size_t stack_size, void (*func)(TID))
{
    int err;
    sys_thread_t *t;

    sysAssert(SYSTHREAD(tid) == 0);

    /* Parent thread preallocates first Java stack segment */
    THREAD(tid)->initial_stack_memory = (long) sysMalloc(sizeof (JavaStack)); 
    if (THREAD(tid)->initial_stack_memory == 0) {
	return 1;
    }

    /*
     * sysThreadCreate() returns SYS_OK if things are OK, and otherwise
     * may currently return either SYS_NOMEM, or SYS_NORESOURCE on some
     * platforms.  It's not clear what SYS_NORESOURCE really means, so
     * threadCreate effectively lumps this with SYS_NOMEM.  This could
     * be improved in the future.  We check below that an error return
     * really is one of those, so we don't lump more than we intended.
     */
    err = sysThreadCreate(stack_size, flags,
			  (void (*)(void *))func, &t, (void *) tid);
    if (err != SYS_OK) {
	sysAssert(err == SYS_NOMEM || err == SYS_NORESOURCE);
        return 1;		/* SYS_NOMEM or SYS_NORESOURCE */
    }

    THREAD(tid)->PrivateInfo = (long) t;
    return 0;
}

    //
    //  Returns 0 on error
    // 
TID createSystemThread(char *name, int priority,
                       long stack_size, void (*f)(TID))
{
    TID tid =
        (TID)execute_java_constructor(NULL, NULL, classJavaLangThread,
            "(Ljava/lang/ThreadGroup;Ljava/lang/String;)",
            systemgroup, makeJavaString(name, strlen(name)));
 
    if (tid == 0 || threadCreate(tid, THR_SYSTEM, stack_size, f) != 0) {
        ThrowOutOfMemoryError(0, 0);
        return 0;
    } else {
        /* Set fields others will expect... */
        THREAD(tid)->priority = priority;
        THREAD(tid)->daemon = 1;
        threadSetPriority(tid, priority);
        threadResume(tid);
	return tid;
    }
}

TID
threadSelf(void)				/* Common: should be a macro? */
{
    sys_thread_t *t = sysThreadSelf();
    if (t) {
	return (TID) sysThreadGetBackPtr(t);
    } else {
	return 0;
    }
}

/*
 * This structure is used during the enumeration of threads.  It
 * holds the state of the enumeration while walking through the
 * threads list.
 */
typedef struct {
    TID *array;
    int index;
    int maxindex;
} tid_array_t;

static int
threadEnumeratorHelper(sys_thread_t *t, void *arg)
{
    tid_array_t *tarr = (tid_array_t *) arg;
    TID tid;

    tid = (TID) sysThreadGetBackPtr(t);
    if (tarr->array != NULL) {
	if (tarr->index >= tarr->maxindex) {
	    return SYS_ERR;
	}
	tarr->array[tarr->index] = tid;
    }
    tarr->index++;
    return SYS_OK;
}

/*
 * Enumerate all threads in the system
 */
int
threadEnumerate(TID *tarr, int count)
{
    tid_array_t helper_data;
    int i;

    helper_data.array = tarr;
    helper_data.maxindex = count;
    helper_data.index = 0;

    /* Currently ignoring error return */
    (void) sysThreadEnumerateOver(threadEnumeratorHelper, (void *) &helper_data);
    for (i = helper_data.index; i < count; i++) {
	tarr[i] = (TID) 0;
    }

    return helper_data.index;
}

/*
 * Count all threads in the system.
 */
int
threadCount()
{
    tid_array_t helper_data;

    helper_data.array = 0;
    helper_data.maxindex = 0;
    helper_data.index = 0;

    sysThreadEnumerateOver(threadEnumeratorHelper, (void *) &helper_data);
    return helper_data.index;
}

void
threadSleep(int millis)
{
    /* Need a unique key per thread */
    int sleep_key = (int) SYSTHREAD(threadSelf());

    if (millis == 0) {
	threadYield();
    } else {
	Log2(4, "threadSleep: TID %s millis %d\n",
	     threadName(threadSelf()), millis);
	monitorEnter((uint_t) sleep_key);
	monitorWait((uint_t) sleep_key, millis);
	monitorExit((uint_t) sleep_key);
    }
}

int
threadPostException(TID tid, void *vexc)
{ 
    JHandle *exc = (JHandle*) vexc;
    struct execenv *ee;  


    ee = (ExecEnv *) THREAD(tid)->eetop;
    if (ee != (ExecEnv *) 0) {   
	/*
	 * Remote exceptions don't currently fill in the backtrace; see
	 * bug 4010164.  Make sure we don't throw an old backtrace.
	 */
	if (is_instance_of(exc, classJavaLangThrowable, ee)) {
	    Classjava_lang_Throwable *throwable =
		unhand((Hjava_lang_Throwable *)exc);
	    throwable->backtrace = NULL;
	}
	sysThreadPostException(SYSTHREAD(tid), exc);
	return TRUE;
    }
    return FALSE;
} 
