/*
 * @(#)util.c	1.56 98/11/18
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*-
	utility routines needed by both the compiler and the interpreter.
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stddef.h>

#include "oobj.h"
#include "interpreter.h"
#include "monitor.h"
#include "utf.h"
#include "sys_api.h"
#include "path.h"       /* DIR_SEPARATOR */

extern struct StrIDhash *nameTypeHash;
extern struct StrIDhash *stringHash;

/* declared in interpreter.h */
PanicHook panic_hook = NULL;
OutOfMemoryHook out_of_memory_hook = NULL;

unicode *
str2unicode(const char *str, unicode *ustr, long len)
{
    unicode *dst = ustr;

    memset((char *) dst, 0, len * sizeof(*dst));
    while (*str && --len >= 0)
	*ustr++ = 0xff & *str++;
    return dst;
}

void
unicode2str(const unicode *src, char *dst, long len)
{
    while (--len >= 0)
	*dst++ = *src++;
    *dst = '\0';
}


int
jio_printf (const char *format, ...)
{
    int len;

    va_list args;
    va_start(args, format);
    len = jio_vfprintf(stdout, format, args);
    va_end(args);

    return len;
}

int
jio_fprintf (FILE *handle, const char *format, ...)
{
    int len;

    va_list args;
    va_start(args, format);
    len = jio_vfprintf(handle, format, args);
    va_end(args);

    return len;
}

int
jio_vfprintf (FILE *handle, const char *format, va_list args)
{
    extern jint (JNICALL *vfprintf_hook)(FILE *fp, const char *format, 
					 va_list args);
    /* vfprintf_hook may not have been installed in javah and javap */
    if (vfprintf_hook)
	return vfprintf_hook(handle, format, args);
    else
	return vfprintf(handle, format, args);
}

/*
 * Print a Java-style unicode string
 */
void
jio_fprintus(FILE *handle, unicode *str, long len)
{
#define UNISTRLEN 100
    char obuf[UNISTRLEN+1];
    register int p;
    if (str == 0 || len <= 0)
	return;
    while (len > 0) {
	unicode c;
	for (p = 0; p < sizeof obuf - 10 && --len >= 0;)
	    if ((c = *str++) >= 256) {
		/*
		 * Make unicode characters readable.  Should not be
		 * possible to overflow, but truncate silently if so.
		 */
		(void) jio_snprintf(obuf + p, UNISTRLEN+1 - p, "\\<%X>", c);
		p += strlen(obuf + p);
	    } else
		obuf[p++] = c;
	obuf[p] = 0;
	if (p)
	    jio_fprintf(handle, "%s", obuf);
    }
}

#undef  HTSIZE			/* Avoid conflict on PC */
#define HTSIZE 2003		/* Default size -- must be prime */
#define HTOVERFLOWPOINT(h)	(h->size * 4 / 5)

/*
 * We store most of the result of the hash in hash_bits to quickly
 * reject impossible matches because strcmp() is fairly expensive,
 * especially when many strings will have common prefixes.
 */
typedef struct {
    char *hash;			/* the string for this entry */
    unsigned long info;		/* high bit set if malloc'ed
				 * low 31 bits are the low 31 bits of hash */
#define MALLOCED_FLAG (1u << 31)
#define HASHBIT_MASK (MALLOCED_FLAG - 1)
} StrIDhashSlot;

typedef void (*hash_fn)(const char *s, unsigned *h1, unsigned *h2);

typedef struct StrIDhash {
    int size;			/* number of entries */
    hash_fn hash;		/* hash function for this table */
    union {
       struct StrIDhash *next;	/* for normal buckets */
       struct StrIDhash **nextp; /* for locked buckets */
    } u;
    short used;			/* number of full entries.  -1 means "locked" */
    short baseid;		/* ID for item in slot[0] */
    void **params;		/* param table, if needed */
    StrIDhashSlot slot[1];	/* expanded to be number of slots */
} StrIDhash;

#define StrID_ISLOCKED(h) (h->used == -1)


/* Hash a string into a primary and secondary hash value */
void
default_hash(const char *s, unsigned *h1, unsigned *h2)
{
    int i;
    unsigned raw_hash;
    for (raw_hash = 0; (i = *(const unsigned char *)s) != '\0'; ) { 
        if (i < 0x80) { 
	    s++;
	} else { 
	    /* We want "s" to stay in a register, if possible.  So we have
	     * to fool the C compiler into believing we're not taking its 
	     * address */
	    const char *t = s;
	    i = next_utf2unicode((char **)&t);
	    s = t;
	}
	raw_hash = raw_hash * 37 + i;
    }
    *h1 = raw_hash;
    *h2 = (raw_hash & 7) + 1; /* between 1 and 8 */
}

/* Create a hash table of the specified size */
static StrIDhash *
createHash(int entries)
{
    StrIDhash *h;
    int size = offsetof(struct StrIDhash, slot) +  (entries * sizeof(h->slot));
    h = (StrIDhash *)sysCalloc(1, size); 
    if (h != NULL) { 
	h->size = entries;
	h->hash = default_hash;	/* only custom tables get something else */
	h->u.next = NULL;
    }
    return h;
}

/*
 * Given a string, return a unique 16 bit id number.  
 * If param isn't null, we also set up an array of void * for holding info
 *     about each object.  The address of this void * is stored into param
 * If CopyNeeded is true, then the name argument "s" must be dup'ed, since
 * the current item is allocated on the stack.
 *
 * Note about returning 0 in the case of out of memory errors: 0 is *not*
 * a valid ID!  The out of memory error should be thrown by the calling code.
 */
unsigned short
Str2ID(StrIDhash **hash_ptr, register char *s, void ***param,
	      int CopyNeeded)
{
    /*
     * The database is a hash table.  When the hash table overflows, a new
     * hashtable is created chained onto the previous one.  This is done so
     * that we can use the hashtable slot index as the ID, without worrying
     * about having the IDs change when the hashtable grows.
     */
    StrIDhash *h = *hash_ptr;
    StrIDhash **nextp;
    long i;
    unsigned primary_hash;
    unsigned secondary_hash;
    unsigned hash_bits;
    hash_fn current_hash_func = NULL;

#ifdef DEBUG
    /*
     * Ensure locking of well-known hash tables.  Str2ID is currently used
     * used on the stringHash table, the nameTypeHash table, and on local
     * hash tables used in the verifier.
     */ 
    if (hash_ptr == &stringHash) {
	sysAssert(STRINGHASH_LOCKED());
    } else if (hash_ptr == &nameTypeHash) {
	sysAssert(NAMETYPEHASH_LOCKED());
    }   /* Should be one of the verifier's private hash tables */
#endif /* DEBUG */

    if (h == NULL) 
	goto not_found;

    /* Create the hash values */
    current_hash_func = h->hash;
    current_hash_func(s, &primary_hash, &secondary_hash);
    hash_bits = primary_hash & HASHBIT_MASK;

    for (;;) {
	char *s2;
	int bucketSize = h->size;
	
	/* See if the new hash table has a different hash function */
	if (h->hash != current_hash_func) {
	    current_hash_func = h->hash;
	    current_hash_func(s, &primary_hash, &secondary_hash);
	    hash_bits = primary_hash & HASHBIT_MASK;
	}
	i = primary_hash % bucketSize;
	
	while ((s2 = h->slot[i].hash) != NULL) {
	    if ((h->slot[i].info & HASHBIT_MASK)== hash_bits
		     && strcmp(s2, s) == 0)
		goto found_it;
	    if ((i -= secondary_hash) < 0)
		i += bucketSize;
	}
	/* Not found in this table.  Try the next table. */
	nextp = StrID_ISLOCKED(h) ? h->u.nextp : &h->u.next;
	if (*nextp == NULL) 
	    break;
	h = *nextp;
    }
    
not_found:
    /* Either the hash table is empty, or the item isn't yet found. */
    if (h == NULL || (h->used >= HTOVERFLOWPOINT(h)) || StrID_ISLOCKED(h)) {
	/* Need to create a new bucket */
	StrIDhash *next;
        if (h && h->baseid > 32000 && h != stringHash) {
	    panic("16-bit string hash table overflow");
	}
	next = createHash(HTSIZE);
	if (next == NULL) {
	    /* Calling code should signal OutOfMemoryError */
	    return 0;
	}
	if (h == NULL) { 
	    /* Create a new table */
	    *hash_ptr = h = next; 
	    h->baseid = 1;  /* guarantee that no ID is 0 */
	} else { 
	    next->baseid = h->baseid + h->size;
	    *nextp = next;
	    h = next;
	}
	if (h->hash != current_hash_func) {
	    current_hash_func = h->hash;
	    current_hash_func(s, &primary_hash, &secondary_hash);
	    hash_bits = primary_hash & HASHBIT_MASK;
	}
	i = primary_hash % h->size;
    }
    if (CopyNeeded) {
	char *d = strDup(s);
	if (d == NULL) {
	    /* Calling code should signal OutOfMemoryError */
            return 0;
	}
	s = d;
    } 
    h->slot[i].hash = s;
    h->slot[i].info = hash_bits | (CopyNeeded ? MALLOCED_FLAG : 0);
    h->used++;

found_it:
    /* We have found or created slot "i" in hash bucket "h" */
    if (param != NULL) { 
	if (h->params == NULL) { 
	    h->params = sysCalloc(h->size, sizeof(void *));
	    if (h->params == NULL) {
		/* Calling code should signal OutOfMemoryError */
		return 0;
	    }
	}
	*param = &(h->params[i]);
    }
    return h->baseid + i;
}

/* Free an StrIDhash table and all the entries in it */
void Str2IDFree(StrIDhash **hash_ptr)
{
    StrIDhash *hash = *hash_ptr;

#ifdef DEBUG
    /*
     * Ensure locking of well-known hash tables.  ID2Str is not currently
     * used except on local hash tables in the verifier.
     */ 
    if (hash_ptr == &stringHash) {
	sysAssert(STRINGHASH_LOCKED());
    } else if (hash_ptr == &nameTypeHash) {
	sysAssert(NAMETYPEHASH_LOCKED());
    }   /* Should be one of the verifier's private hash tables */
#endif /* DEBUG */

    while (hash != NULL) {
        /* We shouldn't be freeing ROM classes.  So no need to check 
	 * if StrID_ISLOCKED(h). */
	StrIDhash *next = hash->u.next;
	StrIDhashSlot *ptr, *endPtr;
	for (ptr = &hash->slot[0], endPtr = ptr + hash->size; 
	          ptr < endPtr; ptr++) { 
	    if (ptr->info & MALLOCED_FLAG)
	        sysFree(ptr->hash);
	}
	if (hash->params != NULL) 
	    sysFree(hash->params);
	sysFree(hash);
	hash = next;
    }
    *hash_ptr = 0;
}

/*
 * Call the callback function on every entry in the table.  This
 * should only be invoked holding the string hash table lock.
 */
void Str2IDCallback(StrIDhash **hash_ptr, void (*callback)(char *, void *))
{
    StrIDhash *hash, *next;
    int i;

#ifdef DEBUG
    /*
     * Ensure locking of well-known hash tables.  Str2IDCallback is currently
     * only used on the stringHash table.
     */ 
    if (hash_ptr == &stringHash) {
	sysAssert(STRINGHASH_LOCKED());
    } else if (hash_ptr == &nameTypeHash) {
	sysAssert(NAMETYPEHASH_LOCKED());
    }   /* Should be one of the verifier's private hash tables */
#endif /* DEBUG */

    /* Currently only used on the string intern hash table */
    sysAssert(STRINGHASH_LOCKED());

    hash = *hash_ptr;
    while (hash) {
	void **params = hash->params;
	next = StrID_ISLOCKED(hash) ? *(hash->u.nextp) : hash->u.next;
	for (i = 0; i < hash->size; i++) {
	    if (hash->slot[i].hash != 0) { 
		callback(hash->slot[i].hash, params ? params[i] : NULL);
	    }
	}
	hash = next;
    }
}

/*
 * Returns NULL in the case of an error.
 */
char *
ID2Str(StrIDhash *h, unsigned short ID, void ***param)
{
    int entry;

#ifdef DEBUG
    /*
     * Ensure locking of well-known hash tables.  ID2Str is currently used
     * on the nameTypeHash table and on local hash tables in the verifier.
     */
    if (h == stringHash) {
	sysAssert(STRINGHASH_LOCKED());
    } else if (h == nameTypeHash) {
	sysAssert(NAMETYPEHASH_LOCKED());
    }   /* Should be one of the verifier's private hash tables */
#endif /* DEBUG */

    while ((long)(ID - h->baseid) >= h->size) {
        h = (StrID_ISLOCKED(h)) ? *(h->u.nextp) : h->u.next;
    }
    entry = ID - h->baseid;
    if (param != NULL) { 
	if (h->params == NULL) {
	    h->params = (void **)sysCalloc(h->size, sizeof(*param));
	    if (h->params == NULL) {
		/* Calling code should signal OutOfMemoryError */
		return NULL;
	    }
	}
	*param = &h->params[entry];
    } 
    return h->slot[entry].hash;
}

char *
addstr(char *s, char *buf, char *limit, char term)
{
    char c;
    while ((c = *s) && c != term && buf < limit) {
	*buf++ = c;
	s++;
    }
    return buf;
}

char *
unicode2rd(unicode *s, long len)
{       /* unicode string to readable C string */
#define CSTRLEN 40
    static char buf[CSTRLEN+1];
    char *dp = buf;
    int c;
    if (s == 0)
	return "NULL";
    *dp++ = '"';
    while (--len>=0 && (c = *s++) != 0 && dp < buf + sizeof buf - 10)
	if (040 <= c && c < 0177)
	    *dp++ = c;
	else
	    switch (c) {
	    case '\n':
		*dp++ = '\\';
		*dp++ = 'n';
		break;
	    case '\t':
		*dp++ = '\\';
		*dp++ = 't';
		break;
	    case '\r':
		*dp++ = '\\';
		*dp++ = 'r';
		break;
	    case '\b':
		*dp++ = '\\';
		*dp++ = 'b';
		break;
	    case '\f':
		*dp++ = '\\';
		*dp++ = 'f';
		break;
	    default:
		/* Should not be possible to overflow, truncate if so */
		(void) jio_snprintf(dp, CSTRLEN+1 - (dp - buf), "\\%X", c);
		dp += strlen(dp);
		break;
	    }
    *dp++ = '"';
    if (len >= 0 && c != 0)
	*dp++ = '.', *dp++ = '.', *dp++ = '.';
    *dp++ = 0;
    return buf;
}

char *
int642CString(int64_t number, char *buf, int buflen)
{
    ll2str(number, buf, buf + buflen);
    return buf;
}

/*
 * WARNING: out_of_memory() aborts the runtime!  It should not be used
 * except in the case of out of memory situations that are clearly not
 * survivable, like running out of memory before the runtime is initialized.
 * If the runtime has finished initializing and you run out of memory, you
 * should throw an OutOfMemoryError instead.
 */
void
out_of_memory()
{
    if (out_of_memory_hook) {
	(*out_of_memory_hook) ();
    } else {
	jio_fprintf(stderr, "**Out of memory, exiting**\n");
    	sysExit(1);
    }
}

void
panic(const char *format, ...)
{
    va_list ap;
    char buf[256];

    va_start(ap, format);
    
    /* If buffer overflow, quietly truncate */
    (void) jio_vsnprintf(buf, sizeof(buf), format, ap);

    if (panic_hook) {
    	(*panic_hook) (buf);
    } else {
    	jio_fprintf(stdout, "\n*** panic: %s\n", buf);
    }

    sysAbort();
}

char *
classname2string(char *src, char *dst, int size) {
    char *buf = dst;
    for (; (--size > 0) && (*src != '\0') ; src++, dst++) {
	if (*src == '/') {
	    *dst = '.';
	}  else {
	    *dst = *src;
	}
    }
    *dst = '\0';
    return buf;
}

/*
 * Warning: the buffer overflow checking here is pretty bogus although
 * should prevent overruns.
 */
void 
mangleMethodName(struct methodblock *mb, char *buffer, int buflen, 
		 int mangleType) 
{ 
    ClassClass *cb = methodclass(mb);

    char *bufptr = buffer;
    char *bufend = buffer + buflen; 

    /* Copy over Java_ */
    (void) jio_snprintf(bufptr, bufend - bufptr, "Java_");
    bufptr += strlen(bufptr);

    if (mangleType == MangleMethodName_JDK_1) {
        /* JDK 1.0.2 convention */
        /* Mangle the class name */
        bufptr += mangleUTFString(cbName(cb), bufptr, bufend - bufptr, 
				  MangleUTF_Class);
        if (bufend - bufptr > 1) 
	    *bufptr++ = '_';
        bufptr += mangleUTFString(methodname(mb), bufptr, bufend - bufptr, 
				  MangleUTF_Field);
	(void) jio_snprintf(bufptr, bufend - bufptr, "%s", "_stub");
	return;
    }

    /* New JNI convention */ 
    bufptr += mangleUTFString(cbName(cb), bufptr, bufend - bufptr, 
			      MangleUTF_JNI);
    if (bufend - bufptr > 1) 
        *bufptr++ = '_';
    bufptr += mangleUTFString(methodname(mb), bufptr, bufend - bufptr, 
			      MangleUTF_JNI);

    /* JNI long name: includes a mangled signature. */
    if (mangleType == MangleMethodName_JNI_LONG) {
        char sig[1024];
	int i;
        if (bufend - bufptr > 2) {
	    *bufptr++ = '_';
	    *bufptr++ = '_';
        }
	for (i = 0; i<1023; i++)
	    if ((sig[i] = methodsig(mb)[i+1]) == ')')
	       break;
        sig[i] = 0;
        bufptr += mangleUTFString(sig, bufptr, bufend - bufptr, 
				  MangleUTF_JNI);
    }
}
    
int
mangleUnicodeChar(int ch, char *bufptr, char *bufend) {
    /* DANGER.  
     *   DO NOT USE jio_snprintf directly with format "%.5x"
     *   It doesn't understand it, and you will lose horribly.
     * REGNAD
     */
    /* So, we use "%05x" which works as expected */
    return jio_snprintf(bufptr, bufend - bufptr, "_%05x", ch);
}

int
mangleUTFString(char *name, char *buffer, int buflen, int type) { 
    char *ptr = name;
    char *bufptr = buffer;
    char *bufend = buffer + buflen - 1;	/* save space for NULL */
    int ch;
    while (((ch = next_utf2unicode(&ptr)) != '\0') && (bufend - bufptr > 0)) {
	if (ch <= 0x7f && isalnum(ch)) {
	    *bufptr++ = ch;
	} else if (ch == '/' && type == MangleUTF_Class) { 
	    *bufptr++ = '_';
	} else if (ch == '_' && type == MangleUTF_FieldStub) { 
	    *bufptr++ = '_';
        } else if (type == MangleUTF_JNI) {
	    char *esc = 0;
	    if (ch == '_')
	        esc = "_1";
	    else if (ch == '/')
	        esc = "_";
	    else if (ch == ';')
	        esc = "_2";
	    else if (ch == '[')
	        esc = "_3";
	    if (esc) {
	        bufptr += jio_snprintf(bufptr, bufend - bufptr, esc);
	    } else {
	        bufptr += mangleUnicodeChar(ch, bufptr, bufend);
	    }
        } else if (type == MangleUTF_Signature) {
	    if (isprint(ch)) {
	        *bufptr++ = ch;
	    } else {
	        bufptr += mangleUnicodeChar(ch, bufptr, bufend);
	    }
 	} else {
	    bufptr += mangleUnicodeChar(ch, bufptr, bufend);
	}
    }
    *bufptr = '\0';
    return bufptr - buffer;
}


/*
 * jio_snprintf(): sprintf with buffer limit.
 */

#include <stdarg.h>
#include "oobj.h"
#include "interpreter.h"

typedef struct InstanceData {
    char *buffer;
    char *end;
} InstanceData;

#define ERROR_RETVAL -1
#undef  SUCCESS
#define SUCCESS 0
#undef  CheckRet
#define CheckRet(x) { if ((x) == ERROR_RETVAL) return ERROR_RETVAL; }

static int 
put_char(InstanceData *this, int c)
{
    if (iscntrl(0xff & c) && c != '\n' && c != '\t') {
        c = '@' + (c & 0x1F);
        if (this->buffer >= this->end) {
            return ERROR_RETVAL;
        }
        *this->buffer++ = '^';
    }
    if (this->buffer >= this->end) {
        return ERROR_RETVAL;
    }
    *this->buffer++ = c;
    return SUCCESS;
}

static int
format_string(InstanceData *this, char *str, int left_justify, int min_width,
              int precision)
{
    int pad_length;
    char *p;

    if (str == 0) {
	return ERROR_RETVAL;
    }

    if ((int)strlen(str) < precision) {
        pad_length = min_width - strlen(str);
    } else {
        pad_length = min_width - precision;
    }
    if (pad_length < 0)
        pad_length = 0;
    if (left_justify) {
        while (pad_length > 0) {
            CheckRet(put_char(this, ' '));
            --pad_length;
        }
    }

    for (p = str; *p != '\0' && --precision >= 0; p++) {
        CheckRet(put_char(this, *p));
    }

    if (!left_justify) {
        while (pad_length > 0) {
            CheckRet(put_char(this, ' '));
            --pad_length;
        }
    }
    return SUCCESS;
}

#define MAX_DIGITS 32

static int
format_number(InstanceData *this, long value, int format_type, int left_justify,
              int min_width, int precision, bool_t zero_pad)
{
    int sign_value = 0;
    unsigned long uvalue;
    char convert[MAX_DIGITS+1];
    int place = 0;
    int pad_length = 0;
    static char digits[] = "0123456789abcdef";
    int base = 0;
    bool_t caps = FALSE;
    bool_t add_sign = FALSE;

    switch (format_type) {
      case 'o': case 'O':
          base = 8;
          break;
      case 'd': case 'D':
          add_sign = TRUE; /* fall through */
      case 'u': case 'U':
          base = 10;
          break;
      case 'X':
          caps = TRUE; /* fall through */
      case 'x':
          base = 16;
          break;
    }
    sysAssert(base > 0 && base <= 16);

    uvalue = value;
    if (add_sign) {
        if (value < 0) {
            sign_value = '-';
            uvalue = -value;
        }
    }

    do {
        convert[place] = digits[uvalue % (unsigned)base];
        if (caps) {
            convert[place] = toupper(convert[place]);
        }
        place++;
        uvalue = (uvalue / (unsigned)base);
        if (place > MAX_DIGITS) {
            return ERROR_RETVAL;
        }
    } while(uvalue);
    convert[place] = 0;

    pad_length = min_width - place;
    if (pad_length < 0) {
        pad_length = 0;
    }
    if (left_justify) {
        if (zero_pad && pad_length > 0) {
            if (sign_value) {
                CheckRet(put_char(this, sign_value));
                --pad_length;
                sign_value = 0;
            }
            while (pad_length > 0) {
                CheckRet(put_char(this, '0'));
                --pad_length;
            }
        } else {
            while (pad_length > 0) {
                CheckRet(put_char(this, ' '));
                --pad_length;
            }
        }
    }
    if (sign_value) {
        CheckRet(put_char(this, sign_value));
    }

    while (place > 0 && --precision >= 0) {
        CheckRet(put_char(this, convert[--place]));
    }

    if (!left_justify) {
        while (pad_length > 0) {
            CheckRet(put_char(this, ' '));
            --pad_length;
        }
    }
    return SUCCESS;
}

int
jio_vsnprintf(char *str, size_t count, const char *fmt, va_list args)
{
    char *strvalue;
    long value;
    InstanceData this;
    bool_t left_justify, zero_pad, long_flag, add_sign, fPrecision;
    int min_width, precision, ch;

    if (str == NULL) {
	return ERROR_RETVAL;
    }
    str[0] = '\0';

    this.buffer = str;
    this.end = str + count - 1;
    *this.end = '\0';		/* ensure null-termination in case of failure */

    while ((ch = *fmt++) != 0) {
        if (ch == '%') {
            zero_pad = long_flag = add_sign = fPrecision = FALSE;
            left_justify = TRUE;
            min_width = 0;
            precision = this.end - this.buffer;
        next_char:
            ch = *fmt++;
            switch (ch) {
              case 0:
                  return ERROR_RETVAL;
              case '-':
                  left_justify = FALSE;
                  goto next_char;
              case '0': /* set zero padding if min_width not set */
                  if (min_width == 0)
                      zero_pad = TRUE;
              case '1': case '2': case '3':
              case '4': case '5': case '6':
              case '7': case '8': case '9':
                  if (fPrecision == TRUE) {
                      precision = precision * 10 + (ch - '0');
                  } else {
                      min_width = min_width * 10 + (ch - '0');
                  }
                  goto next_char;
              case '.':
                  fPrecision = TRUE;
                  precision = 0;
                  goto next_char;
              case 'l':
                  long_flag = TRUE;
                  goto next_char;
              case 's':
                  strvalue = va_arg(args, char *);
                  CheckRet(format_string(&this, strvalue, left_justify,
                                         min_width, precision));
                  break;
              case 'c':
                  ch = va_arg(args, int);
                  CheckRet(put_char(&this, ch));
                  break;
              case '%':
                  CheckRet(put_char(&this, '%'));
                  break;
              case 'd': case 'D':
              case 'u': case 'U':
              case 'o': case 'O':
              case 'x': case 'X':
                  value = long_flag ? va_arg(args, long) : va_arg(args, int);
                  CheckRet(format_number(&this, value, ch, left_justify,
                                         min_width, precision, zero_pad));
                  break;
              default:
                  return ERROR_RETVAL;
            }
        } else {
            CheckRet(put_char(&this, ch));
        }
    }
    *this.buffer = '\0';
    return strlen(str);
}

int 
jio_snprintf(char *str, size_t count, const char *fmt, ...)
{
    va_list args;
    int len;

    va_start(args, fmt);
    len = jio_vsnprintf(str, count, fmt, args);
    va_end(args);
    return len;
}

/* Return true if the two classes are in the same class package */

bool_t
IsSameClassPackage(ClassClass *class1, ClassClass *class2) 
{
    if (cbLoader(class1) != cbLoader(class2)) 
	return FALSE;
    else {
	char *name1 = cbName(class1);
	char *name2 = cbName(class2);
	char *last_slash1 = strrchr(name1, DIR_SEPARATOR);
	char *last_slash2 = strrchr(name2, DIR_SEPARATOR);
	if ((last_slash1 == NULL) || (last_slash2 == NULL)) {
	    /* One of the two doesn't have a package.  Only return true
	     * if the other one also doesn't have a package. */
	    return (last_slash1 == last_slash2); 
	} else {
	    int length1, length2;
	    if (*name1 == SIGNATURE_ARRAY) { 
		do name1++; while (*name1 == SIGNATURE_ARRAY);
		if (*name1 != SIGNATURE_CLASS) { 
		    /* Something is terribly wrong.  Shouldn't be here */
		    return FALSE;
		}
		name1++;
	    }
	    if (*name2 == SIGNATURE_ARRAY) { 
		do name2++; while (*name2 == SIGNATURE_ARRAY);
		if (*name2 != SIGNATURE_CLASS) { 
		    /* Something is terribly wrong.  Shouldn't be here */
		    return FALSE;
		}
		name2++;
	    }
	    length1 = last_slash1 - name1;
	    length2 = last_slash2 - name2;
	    return ((length1 == length2) 
		    && (strncmp(name1, name2, length1) == 0));
	}
    }
}

/*
 * Given a class or interface, put the transitive closure of its superclasses
 * and/or superinterfaces in (*plist)[1 ... *pidx -1].
 * The higher indexes have superinterfaces and (*plist)[0] has the given 
 * class. 
 * For efficiency reason, it takes an initial buffer. Caller have to 
 * provide it (the size must be SUPER_INIT_CHAIN_SIZE) and assign its 
 * pointer to *plist.
 * If the buffer gets too small, a new buffer is allocated and its pointer
 * is assigned to *plist. Note that you need to take care to free it in 
 * your caller. ex. if (returnedList != localList) sysFree(returnedList);
 * '*psize' will be set to '-1' if error occurred.
 */
static void
getSuperHierImpl(ClassClass *cb, ClassClass ***plist, int *psize,
		 bool_t superclasses, bool_t superinterfaces)
{
    ClassClass **local = *plist; /* Keep pointer to the initial buffer */
    ClassClass **list = *plist;
    int iSize = SUPER_INIT_CHAIN_SIZE; /* Initial buffer size */
    int iTop;
    int iEnd;
    int iNext;
    int i;

    sysAssert(cb != NULL);
    sysAssert(SUPER_INIT_CHAIN_SIZE > 1);
    sysAssert(superclasses || superinterfaces);
    
    /* Put the given class at list[0] to make the process easier */
    list[0] = cb;
    iTop = 0;
    iEnd = 1;
    iNext = 1;

    do {
	/*
	 * list: |WXYZ|ABCD|...........|
	 *        ^    ^    ^           ^
	 *        0  iTop  iEnd,iNext  iSize
	 *
	 * 0 to iTop-1: Interfaces whose superinterfaces have already been
	 *              searched.
	 * iTop to iEnd-1: Superinterfaces whose superinterfaces are to be 
	 *              searched during the next iteration.
	 */
	for (i = iTop; i < iEnd; i++) {
	    ClassClass *icb = list[i];
	    cp_item_type *constantPool = cbConstantPool(icb);
	    int j;
	    int icount = superinterfaces ? (int)cbImplementsCount(icb) : 0;
	    ClassClass *super = superclasses ? cbSuperclass(icb) : NULL;
	    int ncount = icount;
	    int nSize = iSize;

	    if (super != NULL) {
		++ncount;
	    }

	    /* If exceed the buffer size... */
	    while (iNext + ncount >= nSize) {
		/* Double the buffer size */
		nSize = nSize << 1;
	    }

	    /* We need to grow... */
	    if (nSize > iSize) {
		/* Grow the buffer and copy what we already filled */
		int k;
		ClassClass **newList = 
		    (ClassClass **)sysMalloc(sizeof(ClassClass *) * nSize);
		if (newList == NULL) {
		    /* Out of memory. Set '*psize' -1 and return */
		    if (list != local) {
			sysFree(list);
		    }
		    /* Reminder: *plist has been kept as the given value */
		    sysAssert(*plist == local);
		    *psize = -1; /* Error occurred */
		    return;
		}
		for (k = 0; k < iSize; k++) {
		    newList[k] = list[k];
		}
		/* Free the old buffer if it was malloced */
		if (list != local) {
		    sysFree(list);
		}
		list = newList;
		iSize = nSize;
	    }

	    if (super != NULL) {
		list[iNext++] = super;
	    }
	    for (j = 0; j < icount; j++) {
		int interfaceIndex = cbImplements(icb)[j];
		list[iNext++] = constantPool[interfaceIndex].clazz;
	    }

	}
	/*
	 * iEnd to iNext-1: Superinterfaces for "ABCD" which are added
	 *                  to the list during the last iteration.
	 */
	iTop = iEnd;
	iEnd = iNext;
	sysAssert(iNext < 1000000); /* Check for broken (recursive) hier */
	/* iNext == iEnd if no more superinterface */
    } while (iEnd > iTop);

    *plist = list;
    *psize = iEnd;
}

/*
 * Given a class or interface, put the transitive closure of its superclasses
 * and superinterfaces in (*plist)[1 ... *pidx -1].
 */
void
getSuperHier(ClassClass *cb, ClassClass ***plist, int *psize)
{
    getSuperHierImpl(cb, plist, psize, TRUE, TRUE);
}

char *
strDup(const char *s)
{
    int len = strlen(s);
    char *d = sysMalloc(len + 1);
    if (d != NULL) {
	strcpy(d, s);
    }
    return d;
}
