/*
 * @(#)JarVerifierStream.java	1.10 99/04/01
 * 
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/* 
 * This file is: JarVerifierStream.java.optional.
 * This is the correct file to resync with jdk.
 */

package sun.tools.jar;

import java.io.*;
import java.util.*;
import java.util.zip.*;
import java.math.BigInteger;
import java.security.*;

import sun.security.provider.*;
import sun.security.pkcs.*;
import sun.net.www.MessageHeader;
import sun.misc.BASE64Decoder;

/**
 *
 * @version 	1.10 99/11/17
 * @author	Benjamin Renaud
 * @author	Roland Schemers
 */
public class JarVerifierStream extends ZipInputStream {

    /* Are we debugging ? */
    private static final boolean debug = false;

    /* The current JAR entry */
    private ZipEntry currentEntry;

    /* a table mapping signature files to blocks parsed */
    private Hashtable signatures;

    /* a table mapping names to identities for entries that have 
       had their actual hashes verified */
    private Hashtable verifiedIdentities;

    /* a table mapping names to identities for entries that have
       passed the .SF/.DSA -> MANIFEST check */
    private Hashtable sigFileIdentities;

    /* Are we parsing a block? */
    private boolean parsingBlock = false;

    /* Are we done parsing META-INF entries? */
    private boolean parsingMeta = true;

    /* Are we parsing the manifest? */
    private boolean parsingManifest = false;

    /* Has this entry been processed yet? */
    private boolean processed = false;

    /* The manifest file */
    private Manifest manifest;

    /* The output stream to use when keeping track of files we are interested
       in */
    private ByteArrayOutputStream baos;

    /* The JarEntryVerifier for the current entry */
    JarEntryVerifier jev;

    /* Scopes in which we look for identities. */
    private IdentityScope scope;

    /**
     */
    public JarVerifierStream(InputStream is) 
	throws IOException, JarException { 
	super(is);
	jev = new JarEntryVerifier();
	signatures = new Hashtable();
	sigFileIdentities = new Hashtable();
	verifiedIdentities = new Hashtable();
	baos = new ByteArrayOutputStream();
	scope = IdentityScope.getSystemScope();

    }

    /**
     * This method scans to see which entry we're parsing and
     * keeps various state information depending on what type of
     * file is being parsed. Files it treats specially are: <ul>
     *
     * <li>Manifest files. At any point, this stream can be queried 
     * for a manifet. If it is present, a Manifest object will be
     * returned.
     *
     * <li>Block Signature file. Like with the manifest, the stream
     * can be queried at any time for all block parsed thus far.
     *
     * </ul>
     */
    public synchronized ZipEntry getNextEntry() throws IOException {

	/* Get the next entry */
	if (currentEntry != null) {
	    closeEntry();
	}

	currentEntry = super.getNextEntry();

	if (currentEntry == null)
	    return null;

	processed = false;
	String name = currentEntry.getName();
	String uname = name.toUpperCase();

	/*
	 * Assumptions:
	 * 1. The manifest should be the first entry in the META-INF directory.
	 * 2. The .SF/.DSA files follow the manifest, before any normal entries
	 * 3. Any of the following will throw a SecurityException:
	 *    a. unable to parse/verify signature on a .DSA file
	 *    b. digest mismatch between a manifest section and
	 *       the SF section.
	 *    c. digest mismatch between the actual jar entry and the manifest
	 */

	if (parsingMeta && (manifest == null) && 
	    Manifest.isManifestName(uname)) {
		parsingManifest = true;
		baos.reset();
		jev.setEntry(null, null);
		return currentEntry;
	}

	if (parsingMeta && 
           (uname.startsWith("META-INF/") ||
            uname.startsWith("/META-INF/") ||
            uname.startsWith("META-INF\\") ||   // for javakey bug 4087128
            uname.startsWith("\\META-INF\\") ||
            uname.startsWith("/META-INF\\") ||
            uname.startsWith("\\META-INF/"))) {

	    if (uname.endsWith(".DSA") || uname.endsWith(".RSA")) {
		/* We parse only DSA or RSA PKCS7 blocks. */
		parsingBlock = true;
		baos.reset();
		jev.setEntry(null, null);
	    }
	    return currentEntry;
	}

	if (parsingMeta)
	    parsingMeta = false;

	if (currentEntry.isDirectory()) {
	    jev.setEntry(null, null);
	    return currentEntry;
	}

	// be liberal in what you accept. Try without the "./" if it
	// has it, or without the "./" if it doesn't.
	if (name.startsWith("./"))
	    name = name.substring(2);

	// only set the jev object for entries that have a signature
	if (sigFileIdentities.get(name) != null) {
	    jev.setEntry(manifest, name);
	    return currentEntry;
	} 

	// don't compute the digest for this entry
	jev.setEntry(null, null);

	return currentEntry;
    }

    /**
     * read a single byte. 
     */
    public int read() throws IOException {
	int b = super.read();

	if (b != -1) {
	    if (parsingBlock || parsingManifest) {
		baos.write(b);
	    } else {
		jev.update((byte)b);
	    }
	} else {
	    processEntry();
	}
	return b;
    }

    /**
     * read an array of bytes.
     */

    public int read(byte[] b, int off, int len) throws IOException {
	int n = super.read(b, off, len);
	if (n != -1) {
	    if (parsingBlock || parsingManifest) {
		baos.write(b, off, n);
	    } else {
		jev.update(b, off, n);
	    }
	} else {
	    processEntry();
	}
	return n;
    }

    /**
     * called when we reach the end of entry in one of the read() methods.
     */
    private void processEntry() 
	throws IOException 
    {
	if (processed) return;

	processed = true;

	if (parsingManifest) {
	    parsingManifest = false;
	    manifest = new Manifest(baos.toByteArray());
	} else if (parsingBlock) {
	    parsingBlock = false;
	    try {
		PKCS7 block = new PKCS7(baos.toByteArray());
		byte[] bytes = block.getContentInfo().getData();
		SignatureFile sf = new SignatureFile(bytes);
		/* this hash could probably go away...*/
		signatures.put(sf, block);
		processSignature(sf, block);
	    } catch (sun.security.pkcs.ParsingException pe) {
		// ignore and treat as unsigned
	    } catch (IOException ioe) {
		// ignore and treat as unsigned
	    } catch (SignatureException se) {
		// ignore and treat as unsigned
	    } catch (NoSuchAlgorithmException nsae) {
		// ignore and treat as unsigned
	    }
	} else {
	    jev.verify(verifiedIdentities, sigFileIdentities);
	}
    }

    /* Verifies the signature blocks, get a the signature file,
       verifies the manifest entries included in it, and maps the
       names to the identities. */

    private void processSignature(SignatureFile sf, PKCS7 block) 
	throws JarException, SignatureException, NoSuchAlgorithmException 
    {
	MessageDigest sha, md5;
	BASE64Decoder decoder;
	SignerInfo[] infos;
	Vector identities = null;

	if (manifest == null) return;

	infos = block.verify();
	    
	if (infos == null)
	    return;
	decoder = new BASE64Decoder();
	sha = MessageDigest.getInstance("SHA");
	md5 = MessageDigest.getInstance("MD5");
	identities = getIdentities(infos, block);

	Enumeration entries = sf.entries();

	while(entries.hasMoreElements()) {

	    String name;	   
	    MessageHeader mh = (MessageHeader)entries.nextElement();
	    if ((name = mh.findValue("Name")) == null) {
                continue;
            }

	    MessageHeader mmh = manifest.getEntry(name);

	    if (mmh != null) {
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		PrintStream ps = new PrintStream(baos);
		mmh.print(ps);
		byte[] data = baos.toByteArray();
		// try SHA and MD5. If it didn't have either,
		// treat this code as unsigned. 
		// XXX: should we throw an exception instead?
		boolean shaOK = 
		    verifySigFileEntry(mh, "SHA-Digest", 
				       data, sha, decoder, name);
		boolean md5OK = 
		    verifySigFileEntry(mh, "MD5-Digest", 
				       data, md5, decoder, name);
		   
		if (shaOK || md5OK) addIdentities(name, identities);
	    }
	}
    }

    /**
     * given the .SF digest header, and the data from the
     * section in the manifest, see if the hashes match.
     * if not, throw a SecurityException.
     *
     * @return true if we found the digest header
     * @exception SecurityException if the hash was not equal
     */

    private boolean verifySigFileEntry(MessageHeader mh,
				       String header,
				       byte data[],
				       MessageDigest md,
				       BASE64Decoder decoder,
				       String name)
	throws JarException
    {
	String base64Hash = mh.findValue(header);
	if (base64Hash == null) 
	    return false;
	byte[] sfHash;
	try {
	    sfHash = decoder.decodeBuffer(base64Hash);
	} catch (IOException ioe) {
	    throw new SecurityException("unable to decode base64 digest");
	}
	md.reset();
	byte [] ourHash = md.digest(data);
	if (debug) {
	    System.out.println("Signature Block File: "+ name + " digest=" +
			       md.getAlgorithm());
	    System.out.println("  sigfile  " + toHex(sfHash));
	    System.out.println("  computed " + toHex(ourHash));
	    System.out.println();
	}
	if (!MessageDigest.isEqual(ourHash, sfHash)) {
	    throw new SecurityException("invalid " + md.getAlgorithm() + 
				   " signature file digest for " +
				   name);
	}
	return true;
    }

    /**
     * Given the PKCS7 blocks and SignerInfo[], create a Vector
     * of Identity objects. We do this only *once* for a given
     * .DSA file.
     */
    private Vector getIdentities(SignerInfo infos[], PKCS7 block) {

	Vector identities = null;

	if (scope == null)
	    return null;

	for (int i = 0; i < infos.length; i++) {
	    SignerInfo info = infos[i];

	    Certificate cert = info.getCertificate(block);
	    PublicKey key = cert.getPublicKey();

	    Identity identity = scope.getIdentity(key);

	    if (identity == null) {
		try {
		    String name = cert.getPrincipal().getName();
		    identity = new SystemIdentity(name, scope);
		    identity.setPublicKey(key);
		    identity.addCertificate(cert);
		} catch (KeyManagementException e) {
		}
	    }
	    if (identity != null) {
		if (debug) {
		    System.out.println("Signature Block Identity: "+identity);
		}
		if (identities == null)
		    identities = new Vector();
		identities.addElement(identity);
	    }
	}
	return identities;
    }

    /**
     * Add the new Identity objects to the sigFileIdentities hash for
     * the given entry.
     */

    private void addIdentities(String name, Vector newIds) {

	if (newIds == null) return;

	if (name.startsWith("./"))
	    name = name.substring(2);

	Vector identities = (Vector)sigFileIdentities.get(name);

	if (identities == null) {
	    identities = new Vector();
	    sigFileIdentities.put(name, identities);
	}

	Enumeration e = newIds.elements();
	while (e.hasMoreElements()) {
	    Object id = e.nextElement();
	    identities.addElement(id);
	}

    }

    /**
     * Returns a hashtable mapping filenames to vectors of identities.
     */
    public Hashtable getVerifiedSignatures() {
	/* we may want to return a copy of this at some point. 
	   For now we simply trust the caller */
	return verifiedIdentities;
    }

    /**
     * Returns an enumeration of PKCS7 blocks.
     */
    public Enumeration getBlocks() {
	return signatures.elements();
    }

    /**
     * This method used to be called by various versions of
     * AppletResourceLoader, even though it didn't do anything with 
     * the result. We leave it in and return null for backwards compatability.
     */
    public Hashtable getNameToHash() {
	return null;
    }

    public Manifest getManifest() {
	return manifest;
    }

    /**
     * return an array of Identity objects for the given file in the jar.
     */
    public Object[] getIdentities(String name)
    {
	Vector ids = (Vector)verifiedIdentities.get(name);
	if (ids == null) 
	    return null;

	// copy from the vector into an array
	Identity[] identities = new Identity[ids.size()];
	ids.copyInto(identities);
	return (Object[])identities;
    }

    // for the toHex function
    private static final char[] hexc =
            {'0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f'};
    /**
     * convert a byte array to a hex string for debugging purposes
     * @param the binary data to be converted to a hex string
     * @return an ASCII hex string
     */

    public static String toHex(byte[] data) {

	StringBuffer sb = new StringBuffer(data.length*2);

	for (int i=0; i<data.length; i++) {
	    sb.append(hexc[(data[i] >>4) & 0x0f]);
	    sb.append(hexc[data[i] & 0x0f]);
	}
	return sb.toString();
    }

}

/**
 * This class is used to verify each entry in a jar file with its
 * manifest value. We re-use it for each entry for peformance reasons...
 */

class JarEntryVerifier {

    private static final boolean debug = false;

    /** the supported digest header values */
    private String supportedDigestHeader[] = { "MD5-Digest", "SHA-Digest"};

    /** the supported digest values for getInstance */
    private String supportedDigestAlg[] = { "MD5", "SHA"};

    /** the digest objects for all the supported digests */
    private java.security.MessageDigest supportedDigest[] = { null, null };

    /** the digests that are actually in use for the current entry */
    private java.security.MessageDigest digest[] = { null, null };

    /** the hash values in the Manifest */
    private byte manifestHash[][] = { null, null};
    private BASE64Decoder decoder;
    private String name;

    /**
     * Create a new Jarverifier object. This object is re-used
     * for all the entries in the jar.
     */
    JarEntryVerifier()
	throws JarException
    {
	try {
	    for (int i=0; i < supportedDigest.length; i++) {
		supportedDigest[i] = 
		    MessageDigest.getInstance(supportedDigestAlg[i]);
	    }
	} catch (NoSuchAlgorithmException nsae) {
	    throw new 
                  SecurityException("unable to create MessageDigest objects");
	}
	decoder = new BASE64Decoder();
    }

    /**
     * Find the hashes in the
     * manifest for this entry, save them, and set the MessageDigest
     * objects to calculate the hashes on the fly. If man or name is
     * null it signifies that update/verify should ignore this entry.
     */
    void setEntry(Manifest man, String name)
	throws IOException
    {
	digest[0] = null;
	digest[1] = null;

	if (man == null || name == null) {
	    return;
	}

	/* get the headers from the manifest for this entry */
	/* if there aren't any, we can't verify any digests for this entry */
	MessageHeader mh = man.getEntry(name);
	if (mh == null) {
	    mh = man.getEntry("./"+name);
	    if (mh == null)
		return;
	}

	this.name = name;

	for (int i=0; i < supportedDigest.length; i++) {
	    String base64 = mh.findValue(supportedDigestHeader[i]);
	    if (base64 != null) {
		digest[i] = supportedDigest[i];
		digest[i].reset();
		manifestHash[i] = decoder.decodeBuffer(base64);
	    } else {
		digest[i] = null;
		manifestHash[i] = null;
	    }
	}
    }

    /**
     * update the digests for the digests we are interested in 
     */
    void update(byte buffer) {
	for (int i=0; i < digest.length; i++) {
	    if (digest[i] != null)
		digest[i].update(buffer);
	}
    }

    /**
     * update the digests for the digests we are interested in 
     */
    void update(byte buffer[], int off, int len) {
	for (int i=0; i < digest.length; i++) {
	    if (digest[i] != null)
		digest[i].update(buffer, off, len);
	}
    }

    /**
     * go through all the digests, calculating the final digest
     * and comparing it to the one in the manifest. If this is
     * the first time we have verified this object, remove its
     * identities from sigFileIdentities and place in verifiedIdentities.
     *
     *
     */
    void verify(Hashtable verifiedIdentities, 
		Hashtable sigFileIdentities) 
	throws JarException 
    {
	boolean done = false;
	for (int i=0; i < digest.length; i++) {
	    if (digest[i] != null) {
		byte [] theHash = digest[i].digest();

		if (debug) {
		    System.out.println("Manifest Entry: " +
				    name + " digest=" + supportedDigestAlg[i]);
		    System.out.println("  manifest " +
			       JarVerifierStream.toHex(manifestHash[i]));
		    System.out.println("  computed " +
			       JarVerifierStream.toHex(theHash));
		    System.out.println();
		}

		if (!MessageDigest.isEqual(theHash, manifestHash[i]))
		    throw new SecurityException(supportedDigestAlg[i] +
						" digest error for "+name);
		done = true;
	    }
	}

	if (done) {
	    // take it out of sigFileIdentities and put it in verifiedIds...
	    Object obj = sigFileIdentities.remove(name);
	    if (obj != null) verifiedIdentities.put(name, obj);
	}
    }
}
