/*
 * Copyright (c) 1996-1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * @(#)romfile.c	1.5 99/01/06
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include <bool.h>
#include <debug.h>
#include <sys_api.h>	/* sysAssert */

#include <limits.h>
#include "romfile.h"

struct npfd {
    const unsigned char *data;
    int length;
    int offset;
};

static struct npfd npfd[MAX_FILES];

/*
 * We keep a set of internal file descriptors to describe IO to in-memory
 * file images.                                                 KGH Oct 95
 */
 
typedef const struct _JavaFileInfo {
    const unsigned char* name;
    const unsigned char* data;
    unsigned length;
} JavaFileInfo;

extern JavaFileInfo javaFileNames[];

static JavaFileInfo *
find_JavaFile(const char *fname)
{
    int i;
    JavaFileInfo *ptr = javaFileNames;

    for (; ptr->name; ptr++) {
	if (strcmp(fname, ptr->name) == 0) {
            return ptr;
        }
    }
    return (0);
}

const unsigned char *
get_romfile_data(const char *fname) {
    JavaFileInfo *info = find_JavaFile(fname);
    if (info == NULL) {
	return NULL;
    } else {
	return info->data;
    }
}

static int
is_JavaDirectory(const char *fname)
{
    int i;
    int len = strlen(fname);
    JavaFileInfo *ptr = javaFileNames;

    /* 
     * If the name already ends with a '/', check for existence of
     * directory instead
     */
    if (fname[len - 1] == '/') {
	len--;
    }
    for (; ptr->name; ptr++) {
        if (!strncmp(fname, ptr->name, len)) {
	    if (strlen(ptr->name) > len && ptr->name[len] == '/') {
		return 1;
	    }
        }
    }
    return 0;
}

long
rom_read(int fd, void *buf, long nbytes)
{
    struct npfd *n = &npfd[fd];
    int max = n->length - n->offset;
    if (max < nbytes) {
	nbytes = max;
    }
    memcpy(buf, n->data + n->offset, nbytes);
    n->offset += nbytes;

    return (nbytes);
}

long
rom_write(int fd, const void *buf, long nbytes)
{
    panic("rom_write");
    return -1;
}

int
rom_stat(unsigned fsid, const char *fname, struct stat *sbuf)
{
    int ret;
    JavaFileInfo *j = find_JavaFile(fname);
 
    if (j != 0) {
        sbuf->st_mode = 0444 | S_IFREG;
        sbuf->st_nlink = 1;
        sbuf->st_size = j->length;
        sbuf->st_dev = 0;
        sbuf->st_ino = (ino_t)j->data;
        sbuf->st_mtime = 0;
        sbuf->st_atime = 0;
	sbuf->st_ctime = 0;
        return (0);
    }

    /* wasn't a file.  Maybe it's a directory? */
    if (is_JavaDirectory(fname)) {
        sbuf->st_mode = 0444 | S_IFDIR;
        sbuf->st_nlink = 1;
        sbuf->st_size = 0;
        sbuf->st_dev = 0;
        sbuf->st_ino = 0;
        sbuf->st_mtime = 0;
        sbuf->st_atime = 0;
	sbuf->st_ctime = 0;
        return (0);
    }

    return -1;
}

int rom_fstat(int fd, struct stat *sbuf)
{
    struct npfd *n = &npfd[fd];
    sbuf->st_mode = 0444 | S_IFREG;
    sbuf->st_nlink = 1;
    sbuf->st_size = n->length;
    sbuf->st_dev = 0;
    sbuf->st_ino = (ino_t)n->data;
    sbuf->st_mtime = 0;
    sbuf->st_atime = 0;
    sbuf->st_ctime = 0;
    return (0);
}

off_t rom_lseek(int fd, off_t offset, int type)
{
    struct npfd *n = &npfd[fd];

    switch (type) {

    case SEEK_SET:
	n->offset = offset;
	break;
    case SEEK_CUR:
	n->offset += offset;
	break;
    case SEEK_END:
	n->offset = n->length + offset;
	break;
    }

    return (0);
}

long rom_avail(int fd)
{
    struct npfd *n = &npfd[fd];
    return n->length - n->offset;
}

int rom_close(int fd)
{
    struct npfd *n = &npfd[fd];

    sysAssert((unsigned)fd < MAX_FILES);

    n->length = 0;
    n->offset = 0;
    n->data = 0;
    return 0;
}

static iovector rom_vector = {
    rom_read,
    rom_write,
    rom_fstat,
    rom_lseek,
    rom_avail,
    rom_close,
    NULL,		// readdir not supported
    NULL		// closedir not supported
};

iovector *
rom_open(unsigned fsid, const char *fname, int flags, int fd)
{
    if (flags & O_CREAT) {
	errno = EACCES;
	return NULL;
    } else {
	JavaFileInfo *jv = find_JavaFile(fname);
	if (jv == 0) {
	    errno = ENOENT;
	    return NULL;
	}
	npfd[fd].data = jv->data;
	npfd[fd].length = jv->length;
	npfd[fd].offset = 0;
	return &rom_vector;
    }
}

int
rom_access(unsigned fsid, const char *path, int mode)
{
    struct stat buf;
    int status = rom_stat(fsid, path, &buf);
    if (status != -1) {
	if ((mode & W_OK) == W_OK) {
	    return -1;
	}
    }
    return status;
}

int
rom_mkdir(unsigned fsid, const char *path, int mode)
{
    return -1;
}

int rom_unlink(unsigned fsid, const char *path)
{
    return -1;
}

int
rom_rmdir(unsigned fsid, const char *path)
{
    return -1;
}

int
rom_rename(unsigned fsid, const char *src, const char *dst)
{
    return -1;
}

pathops rom_ops = {
    rom_open,
    rom_access,
    rom_mkdir,
    rom_stat,
    rom_unlink,
    rom_rename,
    rom_rmdir,
    NULL	// opendir not supported
};

