/*
 * @(#)AudioPlayer.java	1.32 98/11/02
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.audio;

import java.util.Vector;
import java.util.Enumeration;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

/**
 * This class provides an interface to play audio streams.
 *
 * To play an audio stream use:
 * <pre>
 *	AudioPlayer.player.start(audiostream);
 * </pre>
 * To stop playing an audio stream use:
 * <pre>
 *	AudioPlayer.player.stop(audiostream);
 * </pre>
 * To play an audio stream from a URL use:
 * <pre>
 *	AudioStream audiostream = new AudioStream(url.openStream());
 *	AudioPlayer.player.start(audiostream);
 * </pre>
 * To play a continuous sound you first have to
 * create an AudioData instance and use it to construct a
 * ContinuousAudioDataStream.
 * For example:
 * <pre>
 *	AudoData data = new AudioStream(url.openStream()).getData();
 *	ContinuousAudioDataStream audiostream = new ContinuousAudioDataStream(data);
 *	AudioPlayer.player.stop(audiostream);
 * </pre>
 *
 * @see AudioData
 * @see AudioDataStream
 * @see AudioDevice
 * @see AudioStream
 * @author Arthur van Hoff, Thomas Ball
 * @version 	1.32, 11/02/98
 */
public
class AudioPlayer extends Thread {

    private AudioDevice devAudio;
    
    /**
     * The default audio player. This audio player is initialized
     * automatically.
     */
    public static final AudioPlayer player = new AudioPlayer();

    private static ThreadGroup getAudioThreadGroup() {
	ThreadGroup g = currentThread().getThreadGroup();
	while ((g.getParent() != null) && (g.getParent().getParent() != null)) {
	    g = g.getParent();
	}
	return g;
    }

    /**
     * Construct an AudioPlayer.
     */
    private AudioPlayer() {
	super(getAudioThreadGroup(), "Audio Player");
	devAudio = AudioDevice.device;
	setPriority(MAX_PRIORITY);
	setDaemon(true);
	start();
    }

    /**
     * Start playing a stream. The stream will continue to play
     * until the stream runs out of data, or it is stopped.
     * @see AudioPlayer#stop
     */
    public synchronized void start(InputStream in) {
	devAudio.openChannel(in);
	notify();
    }

    /**
     * Stop playing a stream. The stream will stop playing,
     * nothing happens if the stream wasn't playing in the
     * first place.
     * @see AudioPlayer#start
     */
    public synchronized void stop(InputStream in) {
	devAudio.closeChannel(in);
    }

    /**
     * Main mixing loop. This is called automatically when the AudioPlayer
     * is created.
     */
    public void run() {
	devAudio.open();
	devAudio.play();
	devAudio.close();
	System.out.println("audio player exit");
    }
}
