package sun.awt;

import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.TextArea;
import java.awt.TextField;
import java.awt.event.KeyEvent;

/**
 * This is an example of an alternative focus manager.
 * It transfers the focus in two dimensions.  It was originally
 * intended to operate using the up/down/left/right arrow keys--hence
 * the name--but because Motif gobbles the arrows, it is currently coded 
 * to use the vi keys (HJKL) in conjunction with the modifier pair 
 * CTRL+SHIFT instead.   It should be trivial to restore it to
 * arrow keys by changing the constants at the top of the class, but
 * if the Text peers use the arrow keys for traversal within the
 * component you may want to do some additional hacking to allow
 * these keys to get through.  (Unfortunately, then the question
 * arises of how you *do* transfer the focus out of a text component?)
 */
public synchronized class ArrowKeyFocusManager extends TabFocusManager {
    static final int MASK  = KeyEvent.CTRL_MASK | KeyEvent.SHIFT_MASK;

    // vi keys
    static final int LEFT  = KeyEvent.VK_H;
    static final int DOWN  = KeyEvent.VK_J;
    static final int UP    = KeyEvent.VK_K;
    static final int RIGHT = KeyEvent.VK_L;

    public ArrowKeyFocusManager(Container cont) {
	super(cont);
    }

    public boolean handleKeyEvent(KeyEvent e) {
	if ((focusRoot == null) || (focusOwner == null)) {
	    return false;
	}

	switch (e.getKeyCode()) {
	case LEFT:
	case RIGHT:
	    /*
	     * If using arrow keys, might want to let them go through
	     * to the text component rather than transferring focus.
	     */
	    // if (e.getSource() instanceof TextField) {
	    //    return false;
	    // }

	    /* FALL THROUGH */
	case DOWN:
	case UP:
	    /*
	     * If using arrow keys, might want to let them go through
	     * to the text component rather than transferring focus.
	     */
	    // if (e.getSource() instanceof TextArea) {
	    //    return false;
	    // }

	    if (e.getModifiers() != MASK) {
		return false;
	    }
	    break;

	default:
	    return false;
        }

        int id = e.getID();
	if (id == KeyEvent.KEY_RELEASED || id == KeyEvent.KEY_TYPED) {
	    return true;	// consume the matching typed and up
	}

	/* now do the real work */
        transferFocus(focusOwner, e.getKeyCode());
	return true;
    }

    /**
     * Search for the nearest component in the correct direction.
     * It does this in a single pass through the list of components,
     * by keeping track of as many as three different
     * components that were found to be in the correct direction
     * relative to base.  If direction is UP, for example, prev
     * will be the nearest component whose lower-right corner is above
     * and left of base's upper-left corner; even will be the nearest
     * component whose bottom edge is directly above the top edge of base, 
     * and next will be the nearest component whose lower-left corner is 
     * above and right of base's top right corner.
     * 
     * At the end of the search, a component which is even with base is
     * preferred; if there isn't one, then the nearer of prev and next
     * is used.
     */
    protected Component advance(Component base, Container ignore,
				Container p, int direction) {
	Point     baseLoc  = base.getLocationOnScreen();
	Dimension baseSize = base.getSize();

	Component prev = null;
	Component even = null;
	Component next = null;

	/* These are not really used as dimensions, just distance pairs. */
	Dimension pDist = new Dimension(0x7fffffff, 0x7fffffff);
	int       eDist = 0x7fffffff;
	Dimension nDist = new Dimension(0x7fffffff, 0x7fffffff);

	/* distance adjustments allow us to compare the correct corners */
	/* depending on the relative orientation of the two components  */
	int prevAdj = 0;	// distance adjustment for prev components
	int nextAdj = 0;	// distance adjustment for next components

	if ((direction == LEFT) || (direction == RIGHT)) {
	    nextAdj = baseSize.height;
	} else {
	    nextAdj = baseSize.width;
	}

	for (int i = 0; i < p.getComponentCount(); i++) {
	    Component c = p.getComponent(i);
	    if ((c == base) || (c == ignore)) {
		continue;	// don't want to advance to self
				// or an already-searched container
	    }

	    if (! (c.isVisible() && c.isEnabled())) {
		continue;	// ignore invisible or disabled components
	    }

	    if (! ((c instanceof Container) || c.isFocusTraversable())) {
		continue;	// skip components that won't take focus
				// and can't contain anything that would
				// take focus.
	    }

	    int offAxis   = 0;	// distance which determines prev/even/next
	    int onAxis    = 0;	// distance in the direction we wish to move

	    Point pt      = c.getLocationOnScreen();
	    int dx        = pt.x - baseLoc.x;
	    int dy        = pt.y - baseLoc.y;
	    Dimension dim = c.getSize();

	    /* 
	     * The off/on axis distances and the adjustment for prev/next
	     * are the only values that differ depending on the direction
	     * we wish to move.
	     */
	    switch (direction) {
	    case UP:
	    case DOWN:
		offAxis = dx;
		prevAdj = dim.width;

		if (direction == UP) {
		    onAxis = -(dy + dim.height);
		} else {
		    onAxis = dy - baseSize.height;
		}
		break;

	    case LEFT:
	    case RIGHT:
		offAxis = dy;
		prevAdj = dim.height;

		if (direction == LEFT) {
		    onAxis = -(dx + dim.width);
		} else {
		    onAxis = dx - baseSize.width;
		}
		break;
	    }

	    /* now process the component. */

	    if (onAxis < 0) {
		continue;	// c is on the wrong side of base
	    }

	    if (offAxis < 0) {	// prev or even
		/* adjust distance to take correct corners into account */
		offAxis = (-offAxis) - prevAdj;

		if (   ((offAxis >= 0) && (offAxis < pDist.width))
		    || ((offAxis == pDist.width) && (onAxis < pDist.height))) {
		    /* adjusted distance > 0  =>  prev component */
		    prev = c;
		    pDist.width  = offAxis;
		    pDist.height = onAxis;
		}
	    } else if (offAxis > 0) {	// next or even
		/* adjust distance to take correct corners into account */
		offAxis = offAxis - nextAdj;

		if (   ((offAxis >= 0) && (offAxis < nDist.width))
		    || ((offAxis == nDist.width) && (onAxis < nDist.height))) {
		    /* adjusted distance > 0  =>  next component */
		    next = c;
		    nDist.width  = offAxis;
		    nDist.height = onAxis;
		}
	    } else {
		/* offAxis == 0  =>  an even component.  Don't adjust. */
	    }

	    if (offAxis <= 0) {
		// adjusted offAxis distance <= 0 is an even component
		if (onAxis < eDist) {
		    even = c;
		    eDist = onAxis;
		}
	    } 
	}

	if (even != null) {
	    return even;
	} else if (nDist.height < pDist.height) {
	    return next;
	} else {
	    return prev;
	} 
    }

    /*
     * If there is a component in the given direction which can
     * take the focus, assign it to that component.  This potentially
     * requires searching upward and downward through containers.
     */
    public void transferFocus(Component base, int direction) {
	if (base == null) {
	    return;
	}

        synchronized (focusRoot.getTreeLock()) {
            Container target = base.getParent();
	    Container last   = null;
	    Point baseLoc = base.getLocationOnScreen();

	    /* this loop searches up through containers */
            while (target != null) {
		Component p = target;
		
		/* this loop searches down through containers */
		while (p instanceof Container) {
		    p = advance(base, last, (Container)p, direction);

		    /* try assigning focus */
		    if ((p != null) && assignFocus(p)) {
			return;
		    }
		}

		if (target == focusRoot) {
		    return;
		} else {
		    last   = target;
		    target = target.getParent();
		}
	    }
        }
    }
}
