/*
 * @(#)InputContext.java	1.6 98/10/29
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.im;

import java.awt.Component;
import java.util.Locale;
import java.awt.AWTEvent;
import java.awt.Toolkit;
import java.awt.event.FocusEvent;
import java.awt.event.ComponentEvent;
import sun.awt.SunToolkit;

/**
 * An InputContext object manages the communication between text editing
 * components and input methods. It dispatches events between them, and
 * forwards requests for information from the input method to the text
 * editing component.
 *
 * <p>
 * By default, one InputContext instance is created per Window instance,
 * and this input context is shared by all components within the window's
 * container hierarchy. However, this means that only one text input
 * operation is possible at any one time within a window, and that the
 * text needs to be committed when moving the focus from one text component
 * to another. If this is not desired, text components can create their
 * own input context instances.
 *
 * @see java.awt.Component#getInputContext
 * @version 1.6, 10/29/98
 * @author Sun Microsystems, Inc.
 */

public class InputContext {

    private InputMethod inputMethod;
    private boolean inputMethodCreationFailed;

    private Component currentClientComponent;

    /**
     * Constructs an InputContext.
     */
    protected InputContext() {
    }

    /**
     * Returns a new InputContext instance.
     */
    public static InputContext getInstance() {
	return new InputMethodContext();
    }

    /**
     * Dispatches an event to the active input method. Called by AWT.
     *
     * @param event The event
     */
    public synchronized void dispatchEvent(AWTEvent event) {

        InputMethod inputMethod = getInputMethod();
        int id = event.getID();
        switch (id) {
	
	case ComponentEvent.COMPONENT_HIDDEN:
		dispose();
		currentClientComponent = null;
		break;
	case FocusEvent.FOCUS_GAINED:
	    activate((Component)event.getSource());
	    if(inputMethod != null)
		inputMethod.dispatchEvent(event);
	    break;

	case FocusEvent.FOCUS_LOST:
	    if(inputMethod != null)
		inputMethod.dispatchEvent(event);
	    deactivate((Component)event.getSource());
	    break;

	default:
	    if(inputMethod != null)
		inputMethod.dispatchEvent(event);
        }
    }

    /**
     * Activates input context for the given component. If the input
     * context previously was active for a different component and there
     * is uncommitted text for that component, the text is committed.
     * Called by AWT when the keyboard focus moves to the component.
     *
     * @param client The client component into which text is entered.
     */
    synchronized void activate(Component client) {
	currentClientComponent = client;
	if (inputMethod != null) {
	    if (inputMethod instanceof InputMethodAdapter) {
                ((InputMethodAdapter) inputMethod).setClientComponent(client);
            }
            inputMethod.activate();
        }
    }
    
    /**
     * Deactivates input context for the given component. Called by AWT
     * when the keyboard focus moves away from the component.
     *
     * @param target The component into which text was entered.
     */
    synchronized void deactivate(Component client) {
        if (inputMethod != null) {
            inputMethod.deactivate();
        }
	currentClientComponent = null;
    }

    /**
     * Returns the client component.
     */
    Component getClientComponent() {
        Component client = currentClientComponent;
        return client;
    }

    /**
     * Disposes of the input context and release the resources used by it.
     * Called by AWT.
     */
    public void dispose() {
        if (inputMethod != null) {
            inputMethod.dispose();
            inputMethod = null;
        }
    }

    InputMethod getInputMethod() {
 
        if (inputMethod != null) {
            return inputMethod;
        }
        
        if (inputMethodCreationFailed) {
            return null;
        }

        try {
            Toolkit toolkit = Toolkit.getDefaultToolkit();

            inputMethod = ((SunToolkit) toolkit).getInputMethodAdapter();
        } catch (Exception e) {
            // there are a number of bad things that can happen while creating
            // the input method. In any case, we just continue without an
            // input method.
            inputMethodCreationFailed = true;
        }
        
        if (inputMethod != null) {
            inputMethod.setInputContext((InputMethodContext) this);
        }
        return inputMethod;
    }

}
