/*
 * @(#)HttpClient.java	1.80 98/10/29
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.net.www.http;

import java.io.*;
import java.net.*;
import java.util.*;
import sun.net.NetworkClient;
import sun.net.ProgressEntry;
import sun.net.ProgressData;
import sun.net.www.MessageHeader;
import sun.net.www.HeaderParser;
import sun.net.www.MeteredStream;
import sun.misc.Regexp;
import sun.misc.RegexpPool;


/**
 * @version 1.80, 10/29/98
 * @author Herb Jellinek
 * @author Dave Brown
 */
public class HttpClient extends NetworkClient {

    // Http requests we send
    MessageHeader requests;
    // if we've had one io error
    boolean failedOnce = false;

    // Keep-alive stream
    KeepAliveStream kas;
    
    /** regexp pool of hosts for which we should connect directly, not Proxy 
     *  these are intialized from a property. 
     */
    private static RegexpPool dontProxy = null;

    /** Default port number for http daemons. REMIND: make these private */
    static final int	httpPortNumber = 80;

    /** return default port number (subclasses may override) */
    protected int getDefaultPort () { return httpPortNumber; }

    public static String proxyHost = null;
    public static int proxyPort = 80;
    public static boolean httpKeepAliveSet = true;

    /* instance-specific proxy fields override the static fields if set. 
     * Used by FTP.  These are set to the true proxy host/port if
     * usingProxy is true.
     */
    private String instProxy = null;
    private int instProxyPort = -1;

    /* All proxying (generic as well as instance-specific) may be
     * disabled through use of this flag
     */
    protected boolean proxyDisabled;

    // are we using proxy in this instance?
    public boolean usingProxy = false;
    // target host, port for the URL
    private String host;
    private int port;
    
    /* where we cache currently open, persistent connections */
    protected static KeepAliveCache kac = new KeepAliveCache();
    
    boolean keepingAlive = false;     // this is a keep-alive connection
    long keepAliveExpirationTime = 0; // end of life for this keep-alive 
    int keepAliveTimeout = 0;         // seconds in keep-alive response

    /** Url being fetched. */
    protected URL	url;

    /* set properties on startup */
    static {
	resetProperties();
    }
    
    
    /**
     * re-initialize the dontProxy list, proxyHost/proxyPort, and
     * maxConnections from system properties. 
     */

    public static synchronized void resetProperties() {
	proxyHost = System.getProperty("http.proxyHost");
        proxyPort = Integer.getInteger("http.proxyPort", 80).intValue();
	if (proxyHost == null) {
	    /* maintain compatibility with 1.0.2, before
	     * the properties namespace was so polluted
	     * and these properties were called, respectively,
	     * "proxyHost" "proxyPort".  Hopefully there won't
	     * be any conflicts.
	     */
	    proxyHost = System.getProperty("proxyHost");
	    proxyPort = Integer.getInteger("proxyPort", 80).intValue();
	}
	if (proxyHost != null && proxyHost.length() == 0) {
	    /* System.getProp() will give us an empty String, ""
	     * for a defined but "empty" property.
	     */
	    proxyHost=null;
	}
	int maxConnections = 
	    Integer.getInteger("http.maxConnections", 2).intValue();

	if (maxConnections <= 0) { // pilot error, Houston..
	    maxConnections = 2;
	}
	kac.setMaxConnections(maxConnections);
	
	String keepAlive = System.getProperty("http.keepAlive");
	if (keepAlive != null) {
	    httpKeepAliveSet = Boolean.valueOf(keepAlive).booleanValue();
	}

	dontProxy = new RegexpPool();
	String rawList = System.getProperty("http.nonProxyHosts");
	if(rawList != null) {
	    java.util.StringTokenizer st
		= new java.util.StringTokenizer(rawList, "|", false);
	    try {
		while (st.hasMoreTokens()) {
		    dontProxy.add(st.nextToken().toLowerCase(),
			    new Boolean(true));
		}
	    } catch (Exception e) {
		e.printStackTrace();
	    }
	}
    }

    /* This package-only CTOR should only be used for FTP piggy-backed on HTTP 
     * HTTP URL's that use this won't take advantage of keep-alive.
     * Additionally, this constructor may be used as a last resort when the
     * first HttpClient gotten through New() failed (probably b/c of a 
     * Keep-Alive mismatch).
     *
     * XXX That documentation is wrong ... it's not package-private any more
     */
    public HttpClient(URL url, String proxy, int proxyPort)
    throws IOException {
	this (url, proxy, proxyPort, false);
    }

    /*
     * This constructor gives "ultimate" flexibility, including the ability
     * to bypass implicit proxying.  Sometimes we need to be using tunneling
     * (transport or network level) instead of proxying (application level),
     * for example when we don't want the application level data to become
     * visible to third parties.
     *
     * @param url		the URL to which we're connecting
     * @param proxy		proxy to use for this URL (e.g. forwarding)
     * @param proxyPort		proxy port to use for this URL
     * @param proxyDisabled	true to disable default proxying
     */
    private HttpClient (URL url, String proxy, int proxyPort, boolean proxyDisabled) 
	throws IOException {

	this.proxyDisabled = proxyDisabled;
	if (!proxyDisabled) {
	    this.instProxy = proxy;
	    this.instProxyPort = (proxyPort < 0)
		? getDefaultPort ()
		: proxyPort;
	}

	/* try to set host to "%d.%d.%d.%d" string if
	 * visible - Sprint bug - brown */
	try {
	    InetAddress addr = InetAddress.getByName(url.getHost());
	    this.host = addr.getHostAddress();
	} catch (UnknownHostException ignored) {
	    this.host = url.getHost();
	}
	this.url = url;
	port = url.getPort();
	if (port == -1) {
	    port = getDefaultPort ();
	}
	openServer();
    }

    protected HttpClient (URL url, boolean proxyDisabled) throws IOException {
	this(url, null, -1, proxyDisabled);
    }

    private HttpClient (URL url) throws IOException {
	this(url, null, -1, false);
    }

    /* This method is used to get an HttpClient to the specifed URL.  
     * If there's currently an active HttpClient to that server/port, 
     * you'll get that one.
     */
    public static HttpClient New(URL url) throws IOException {
	/* see if one's already around */
	HttpClient ret = kac.findAliveHttpClient(url);
	if (ret == null) {
	    ret = new HttpClient (url);  // CTOR called openServer()
	} else {
	    ret.url = url;
	}
	// don't know if we're keeping alive until we parse the headers
	// for now, keepingAlive is false
	return ret;
    }

    /* return it to the cache as still usable, if:
     * 1) It's keeping alive, AND
     * 2) It hasn't had a error (PrintStream.checkError())
     *
     * If this client is not keepingAlive, it should have been
     * removed from the cache in the parseHeaders() method.
     */
    public static void finished (HttpClient done) {
	if (done.keepingAlive && !(done.serverOutput.checkError())) {
	    /* This connection is keepingAlive && still valid.
	     * Return it to the cache.
	     */
	    kac.addHttpClient(done);
	} else {
	    done.closeServer();
	}
    }	    

    /* We're very particular here about what our InputStream to the server
     * looks like for reasons that are apparent if you can decipher the
     * method parseHTTP().  That's why this method is overidden from the
     * superclass.
     */
    public void openServer(String server, int port) throws IOException {
	serverSocket = doConnect (server, port);
	serverOutput = new PrintStream(
	    new BufferedOutputStream(serverSocket.getOutputStream()));
	serverSocket.setTcpNoDelay(true);
    }

    /*
     */
    private synchronized void openServer() throws IOException {

	SecurityManager security = System.getSecurityManager();
	if (security != null) {
	    security.checkConnect(host, port);
	}

	if (keepingAlive) { // already opened
	    return;
	}

	/* Try to open connections in this order, return on the 
	 * first one that's successful:
	 * 1) if (instProxy != null)
	 *        connect to instProxy
	 *        _raise_exception_  if failure
	 *        usingProxy=true
	 * 2) if (proxyDisabled || host is on dontProxy list)
	 *        connect directly (locally);
	 *        usingProxy = false;
	 * 3) if (proxyHost != null)
	 *        connect to proxyHost;
	 *        usingProxy=true;
	 * 4) connect locally
	 *    usingProxy = false;
	 */

	// see if the host is on dontProxy list

	if (instProxy != null) {
	    openServer(instProxy, instProxyPort);
	    usingProxy = true;
	    return;
	}

	if (proxyDisabled
		|| dontProxy.match(url.getHost().toLowerCase()) != null
		|| dontProxy.match(host) != null) {
	    openServer(host, port);
	    usingProxy = false;
	    return;
	}
	    

	if (url.getProtocol().equals("http")) {
	    if (proxyHost != null) {
		try {
		    openServer(proxyHost, proxyPort);
		    instProxy = proxyHost;
		    instProxyPort = proxyPort;
		    usingProxy = true;
		    return;
		} catch (IOException e) {
		    // continue in this case - try locally
		}
	    }
	    // try locally - let Exception get raised
	    openServer(host, port);
	    return;

	} else {
	    /* we're opening some other kind of url, most likely an
	     * ftp url. In this case:
	     * 1) try instProxy (== FTP proxy)
	     * 2) try HttpProxy
	     * 3) directly
	     */
	    if (instProxy != null) {
		try {
		    super.openServer(instProxy, instProxyPort);
		    usingProxy = true;
		} catch (IOException keepGoing) {}

	    } else if (proxyHost != null) {
		try {
		    super.openServer(proxyHost, proxyPort);
		    instProxy = proxyHost;
		    instProxyPort = proxyPort;
		    usingProxy = true;
		} catch (IOException keepGoing) {}
	    } else {
		super.openServer(host, port);
	    }
	}
    }


    /** Parse the first line of the HTTP request.  It usually looks
	something like: "HTTP/1.0 <number> comment\r\n". */

    public String getURLFile() {
	if (usingProxy) {
	    String result = url.getProtocol() + "://" + url.getHost();
	    if (url.getPort() != -1) {
		result += ":" + url.getPort();
	    }
	    return result + url.getFile();
	} else {
	    return url.getFile();
	}
    }

    public void writeRequests(MessageHeader head) {
	requests = head;
	requests.print(serverOutput);
	serverOutput.flush();
    }

    public boolean parseHTTP(MessageHeader responses, ProgressEntry pe)
    throws IOException {
	/* If "HTTP/*" is found in the beginning, return true.  Let
         * HttpURLConnection parse the mime header itself.
	 *
	 * If this isn't valid HTTP, then we don't try to parse a header
	 * out of the beginning of the response into the responses,
	 * and instead just queue up the output stream to it's very beginning.
	 * This seems most reasonable, and is what the NN browser does.
	 */

	keepAliveTimeout = 0;         
	boolean ret = false;
	byte[] b = new byte[7];

	try {
	    serverInput = new PushbackInputStream(
		    serverSocket.getInputStream(), 7);
	    int nread = 0;
	    while (nread < 7) {
		int r = serverInput.read(b, nread, 7 - nread);
		if (r < 0) {
		    break;
		} 
		nread += r;
	    }
	    String keep=null;
	    ret = b[0] == 'H' && b[1] == 'T'
		    && b[2] == 'T' && b[3] == 'P' && b[4] == '/' &&
		b[5] == '1' && b[6] == '.';
	    ((PushbackInputStream) serverInput).unread(b);
	    if (ret) { // is valid HTTP - respose started w/ "HTTP/1."
		responses.parseHeader(serverInput);
		/* decide if we're keeping alive: 
		 * This is a bit tricky.  There's a spec, but most current
		 * servers (10/1/96) that support this differ in dialects.
		 * If the server/client misunderstand each other, the
		 * protocol should fall back onto HTTP/1.0, no keep-alive.
		 */
		if (usingProxy) {
		    keep = responses.findValue("Proxy-Connection");
		} else {
		    keep = responses.findValue("Connection");
		}
		if (keep != null && keep.toLowerCase().equals("keep-alive")) {
		    /* some servers, notably Apache1.1, send something like:
		     * "Keep-Alive: timeout=15, max=1" which we should respect.
		     */
		    HeaderParser p = new HeaderParser(
			    responses.findValue("Keep-Alive"));
		    keepAliveTimeout = p.findInt("timeout", 15);
		    keepAliveExpirationTime = (new Date().getTime()) 
					      + (keepAliveTimeout * 1000);
		}
	    } else if (nread != 7) {
		if (!failedOnce && requests != null) {
		    failedOnce = true;
		    closeServer();
		    openServer();
		    writeRequests(requests);
		    return parseHTTP(responses, pe);
		} 
		throw new SocketException("Unexpected end of file from server");
	    } else { 
		// we can't vouche for what this is....
		responses.set("Content-type", "unknown/unknown");
	    }
	} catch (IOException e) {
	    closeServer();
	    if (!failedOnce && requests != null 
		    && !(e instanceof InterruptedIOException)) {
		// try once more
		failedOnce = true;
		openServer();
		writeRequests(requests);
		return parseHTTP(responses, pe);
	    } else {
		throw e;
	    }
	}
	
	int cl = -1;
	try {
	    cl = Integer.parseInt(responses.findValue("content-length"));
	} catch (Exception e) {}

	if (keepAliveTimeout > 0 && cl > 0) {
	    keepingAlive = true;
	} else if (keepingAlive) {
	    /* Previously we were keeping alive, and now we're not.  Remove
	     * this from the cache (but only here, once) - otherwise we get
	     * multiple removes and the cache count gets messed up.
	     */
	    keepingAlive=false;
	}

	/* finally wrap a KeepAlive/MeteredStream around it if appropriate */
	if (cl > 0) {
	    pe.setType(url.getFile(), responses.findValue("content-type"));
	    pe.update(0, cl);
	    if (keepingAlive) {
		kas = new KeepAliveStream(
		    serverSocket.getInputStream(), pe, this);
		serverInput = new BufferedInputStream(kas);
		failedOnce = false;
	    } else {
		serverInput = new BufferedInputStream(
                     new MeteredStream(serverSocket.getInputStream(), pe));
	    }
	} else {
	    serverInput = new BufferedInputStream(
		serverSocket.getInputStream());
	    ProgressData.pdata.unregister(pe);
	}
	return ret;
    }

    public synchronized InputStream getInputStream() {
	return serverInput;
    }

    public OutputStream getOutputStream() {
	return serverOutput;
    }

    public String toString() {
	return getClass().getName()+"("+url+")";
    }

    public final boolean isKeepingAlive() {
	return httpKeepAliveSet && keepingAlive;
    }

    protected void finalize() throws Throwable {
	// This should do nothing.  The stream finalizer will 
	// close the fd.
    }
    
    /* Use only on connections in error. */
    public void closeServer() {
	try {
	    keepingAlive = false;
	    serverSocket.close();
	} catch (Exception e) {}
    }

    /**
     * @return the proxy host being used for this client, or null
     *		if we're not going through a proxy
     */
    public String getProxyHostUsed() {
	if (!usingProxy) {
	    return null;
	} else {
	    return instProxy;
	}
    }

    /**
     * @return the proxy port being used for this client.  Meaningless
     *		if getProxyHostUsed() gives null.
     */
    public int getProxyPortUsed() {
	return instProxyPort;
    }

    public long getExpirationTime() {
	return keepAliveExpirationTime;
    }

    public URL getURL() {
	return url;
    }
}
