/*
 * @(#)HttpURLConnection.java	1.51 98/11/02
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.net.www.protocol.http;

import java.net.URL;
import java.net.ProtocolException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.io.ByteArrayOutputStream;
import java.io.*;
import sun.net.*;
import sun.net.www.*;
import sun.net.www.http.HttpClient;

/**
 * A class to represent an HTTP connection to a remote object.
 *
 * @author  James Gosling
 * @author  Herb Jellinek
 */

public class HttpURLConnection extends java.net.HttpURLConnection {
    
    static final String version = System.getProperty("java.version");
    public static final String userAgent =
        System.getProperty("http.agent", "Java"+version);
    static final String httpVersion = "HTTP/1.0";
    static final String acceptString =
        "text/html, image/gif, image/jpeg, *; q=.2, */*; q=.2";

    /* max # of allowed re-directs, by HTTP spec */
    static final int maxRedirects = 5;

    protected HttpClient http;
    protected Handler handler;

    /* output stream to server */
    protected PrintStream ps = null;

    /* We only have a single static authenticator for now.
     */
    private static HttpAuthenticator defaultAuth;
    
    /* all the headers we send */
    private MessageHeader requests;

    /* Progress entry */
    protected ProgressEntry pe;

    /* all the response headers we get back */
    private MessageHeader responses;
    /* the stream _from_ the server */
    private InputStream inputStream = null;
    /* post stream _to_ the server, if any */
    private ByteArrayOutputStream poster = null;

    /* Indicates if the std. request headers have been set in requests. */
    private boolean setRequests=false;

    /* Indicates whether a request has already failed or not */
    private boolean failedOnce=false;

    private boolean shouldFollowRedirects = true;

    /* adds the standard key/val pairs to reqests if necessary & write to
     * given PrintStream
     */
    private void writeRequests() throws IOException {

	/* print all message headers in the MessageHeader 
	 * onto the wire - all the ones we've set and any
	 * others that have been set
	 */
	if (!setRequests) {

	    /* We're very particular about the order in which we
	     * set the request headers here.  The order should not
	     * matter, but some careless CGI programs have been
	     * written to expect a very particular order of the
	     * standard headers.  To name names, the order in which
	     * Navigator3.0 sends them.  In particular, we make *sure*
	     * to send Content-type: <> and Content-length:<> second
	     * to last and last, respectively, in the case of a POST
	     * request.
	     */
	    requests.prepend(method + " " + http.getURLFile()+" "  + 
			     httpVersion, null);
	    requests.setIfNotSet("User-Agent", userAgent);
	    int port = url.getPort();
	    String host = url.getHost();
	    if (port != -1 && port != 80) {
		host += ":" + String.valueOf(port);
	    }
	    requests.setIfNotSet("Host", host);
	    requests.setIfNotSet("Accept", acceptString);

	    /* Here, we have to work around an egregious bug in the Netscape
	     * proxy (at least in version 2.0p1 and 2.0p1+ by my reckoning):
	     * 1) If we're doing a POST
	     * 2) And we send "Proxy-Connection: keep-alive"
	     * 3) And the server the proxy connects to DOES NOT reply w/ 
	     *    "Connection: keep-alive"
	     * The proxy will *still* send back with the responses the headers
	     * "Connection: keep-alive" and "Proxy-connection: keep-alive",
	     * no if's, and's, or but's.
	     * Which deceives us, the client, into thinking we have a keep-alive
	     * connection with the ultimate server when in fact we do not.
	     * The solution, unfortunately, is to never attempt keep-alive
	     * when talking to a proxy && (POST || PUT).  I don't want to get 
	     * into the game of deciphering which proxy we're talking to. -brown
	     */
	     
	    // Try keep-alive only on first attempt
	    if (!failedOnce && http.httpKeepAliveSet) {
		if (http.usingProxy && !(method.equals("POST") 
					 || method.equals("PUT"))) {
		    requests.set("Proxy-Connection", "keep-alive");
		} else if (!(http.usingProxy)) {
		    requests.set("Connection", "keep-alive");
		}
	    }
	    // send any pre-emptive authentication
	    if (http.usingProxy) {
		AuthenticationInfo pauth
		    = AuthenticationInfo.getProxyAuth(http.getProxyHostUsed(),
						      http.getProxyPortUsed());
		if (pauth != null && pauth.supportsPreemptiveAuthorization()) {
		    // Sets "Proxy-authorization"
		    requests.setIfNotSet(pauth.getHeaderName(),
					 pauth.getHeaderValue());
		}
	    }
	    AuthenticationInfo sauth = AuthenticationInfo.getServerAuth(url);
	    if (sauth != null && sauth.supportsPreemptiveAuthorization()) {
		// Sets "Authorization"
		requests.setIfNotSet(sauth.getHeaderName(),
				     sauth.getHeaderValue());
	    }

	    if (poster != null) {
		/* add Content-length & POST/PUT data */
		synchronized (poster) {
		    requests.setIfNotSet("Content-type", "application/x-www-form-urlencoded");
		    requests.set("Content-length", String.valueOf(poster.size()));
		}
	    }
	    setRequests=true;
	}
	http.writeRequests(requests);
	if (poster != null) {
	    poster.writeTo(ps);
	    ps.flush();
	}

	if (ps.checkError()) {
	    disconnect();
	    if (failedOnce) {
		throw new IOException("Error writing to server");
	    } else { // try once more
		failedOnce=true;
		http = getNewClient (url);
		ps = (PrintStream) http.getOutputStream();
		connected=true;
		responses = new MessageHeader();
		requests = new MessageHeader();
		setRequests=false;
		writeRequests();
	    }
	}
    }


    /**
     * Create a new HttpClient object, bypassing the cache of
     * HTTP client objects/connections.
     *
     * @param url	the URL being accessed
     */
    protected HttpClient getNewClient (URL url)
    throws IOException {
	return new HttpClient (url, null, -1);
    }


    /**
     * Create a new HttpClient object, set up so that it uses
     * per-instance proxying to the given HTTP proxy.  This
     * bypasses the cache of HTTP client objects/connections.
     *
     * @param url	the URL being accessed
     * @param proxyHost	the proxy host to use
     * @param proxyPort	the proxy port to use
     */
    protected HttpClient
    getProxiedClient (URL url, String proxyHost, int proxyPort)
    throws IOException {
	return new HttpClient (url, proxyHost, proxyPort);
    }


    protected HttpURLConnection(URL u, Handler handler)
    throws IOException {
	super(u);
	requests = new MessageHeader();
	responses = new MessageHeader();
	this.handler = handler;
    }
    
    /** this constructor is used by other protocol handlers such as ftp
        that want to use http to fetch urls on their behalf. */
    public HttpURLConnection(URL u, String host, int port) {
	super(u);
	requests = new MessageHeader();
	responses = new MessageHeader();
    	handler = new Handler(host, port);
    }

    public static void setDefaultAuthenticator(HttpAuthenticator a) {
	defaultAuth = a;
    }

    /**
     * opens a stream allowing redirects only to the same host.
     */
    public static InputStream openConnectionCheckRedirects(java.net.URLConnection c)
	throws IOException
    {
        boolean redir;
        int redirects = 0;
        InputStream in = null;

        do {
            if (c instanceof HttpURLConnection) {
                ((HttpURLConnection) c).setInstanceFollowRedirects(false);
            }
 
            // We want to open the input stream before
            // getting headers, because getHeaderField()
            // et al swallow IOExceptions.
            in = c.getInputStream();
            redir = false;
 
            if (c instanceof HttpURLConnection) {
                HttpURLConnection http = (HttpURLConnection) c;
                int stat = http.getResponseCode();
                if (stat >= 300 && stat <= 305 &&
                        stat != HttpURLConnection.HTTP_NOT_MODIFIED) {
                    URL base = http.getURL();
                    String loc = http.getHeaderField("Location");
                    URL target = null;
                    if (loc != null) {
                        target = new URL(base, loc);
                    }
                    http.disconnect();
                    if (target == null
                        || !base.getProtocol().equals(target.getProtocol())
                        || base.getPort() != target.getPort()
                        || !hostsEqual(base, target)
                        || redirects >= 5)
                    {
                        throw new SecurityException("illegal URL redirect");
		    }
                    redir = true;
                    c = target.openConnection();
                    redirects++;
                }
            }
        } while (redir);
        return in;
    }


    //
    // Same as java.net.URL.hostsEqual
    //
    private static boolean hostsEqual(URL u1, URL u2) {
	String h1 = u1.getHost();
	String h2 = u2.getHost();

	if (h1 == null) {
	    return h2 == null;
	} else if (h2 == null) {
	    return false;
	} else if (h1.equalsIgnoreCase(h2)) {
	    return true;
	}
        // Have to resolve addresses before comparing, otherwise
        // names like tachyon and tachyon.eng would compare different
        try {
            InetAddress a1 = InetAddress.getByName(h1);
            InetAddress a2 = InetAddress.getByName(h2);
            return a1.equals(a2);
        } catch(UnknownHostException e) {
        } catch(SecurityException e) {
        }
        return false;
    }

    // overridden in HTTPS subclass

    public void connect() throws IOException {
	if (connected) {
	    return;
	}
	if ("http".equals(url.getProtocol()) && !failedOnce) {
	    http = HttpClient.New(url);
	} else {
	    // make sure to construct new connection if first attempt failed
	    http = new HttpClient(url, handler.proxy, handler.proxyPort);
	}
	ps = (PrintStream)http.getOutputStream();

	// constructor to HTTP client calls openserver
	connected = true;
    }

    /*
     * Allowable input/output sequences:
     * [interpreted as POST/PUT]
     * - get output, [write output,] get input, [read input]
     * - get output, [write output]
     * [interpreted as GET]
     * - get input, [read input]
     * Disallowed:
     * - get input, [read input,] get output, [write output]
     */

    public synchronized OutputStream getOutputStream() throws IOException {

	try {
	    if (!doOutput) {
		throw new ProtocolException("cannot write to a URLConnection"
			       + " if doOutput=false - call setDoOutput(true)");
	    }
	    
	    if (method.equals("GET")) {
		setRequestMethod("POST"); // Backward compatibility
	    }
	    if (!"POST".equals(method) && !"PUT".equals(method) && 
		"http".equals(url.getProtocol())) {
		throw new ProtocolException("HTTP method " + method + " doesn't support output");
	    }

	    // if there's already an input stream open, throw an exception
	    if (inputStream != null) {
		throw new ProtocolException("Cannot write output after reading input.");
	    }

	    connect();

	    /* REMIND: This exists to fix the HttpsURLConnection subclass.
	     * Hotjava needs to run on JDK1.1FCS.  Do proper fix in subclass
	     * for 1.2 and remove this.
	     */
	    ps = (PrintStream)http.getOutputStream();

	    if (poster == null)
		poster = new ByteArrayOutputStream();
	    return poster;
	} catch (RuntimeException e) {
	    disconnect();
	    throw e;
	} catch (IOException e) {
	    disconnect();
	    throw e;
	}
    }

    public synchronized InputStream getInputStream() throws IOException {

	if (!doInput) {
	    throw new ProtocolException("Cannot read from URLConnection"
		   + " if doInput=false (call setDoInput(true))");
	}
	if (inputStream != null) {
	    return inputStream;
	}

	int redirects = 0;
	AuthenticationInfo serverAuthentication = null;
	AuthenticationInfo proxyAuthentication = null;
	try {
	    do {

		pe = new ProgressEntry(url.getFile(), null);
		ProgressData.pdata.register(pe);
		connect();

		/* REMIND: This exists to fix the HttpsURLConnection subclass.
		 * Hotjava needs to run on JDK1.1FCS.  Do proper fix in subclass
		 * for 1.2 and remove this.
		 */
		ps = (PrintStream)http.getOutputStream();

		writeRequests();
		http.parseHTTP(responses, pe);
		inputStream = http.getInputStream();

		int respCode = getResponseCode();
		if (respCode == HTTP_PROXY_AUTH) {
		    if (proxyAuthentication != null) {
			proxyAuthentication.removeFromCache();
		    }
		    proxyAuthentication = getHttpProxyAuthentication();
		    if (proxyAuthentication != null) {
			disconnect();
			redirects++;
			continue;
		    }
		}
		if (respCode == HTTP_UNAUTHORIZED) {
		    if (serverAuthentication != null) {
			/* previous authorization failed - uncache */
			serverAuthentication.removeFromCache();
		    } 
		    serverAuthentication = getServerAuthentication();
		    if (serverAuthentication != null) {
			disconnect();
			redirects++; // don't let things loop ad nauseum
			continue;
		    }
		}

		if (respCode == HTTP_OK || (respCode >= 300 && respCode <= 305))
		{
		    if (proxyAuthentication != null) {
			// cache auth info on success
			proxyAuthentication.addToCache();
		    }
		    if (serverAuthentication != null) {
			// cache auth info on success
			serverAuthentication.addToCache();
		    }
		}

		if (followRedirect()) {
		    /* if we should follow a redirect, then the followRedirects()
		     * method will disconnect() and re-connect us to the new
		     * location
		     */
		    redirects++;
		    if (!url.getProtocol().startsWith("http")) {
			return url.openStream();
		    }
		    continue;
		}
		if (method.equals("HEAD") || method.equals("TRACE")) {
		    disconnect();
		    return inputStream = new EmptyInputStream();
		}
		

		String fname = url.getFile();
		if (respCode >= 400 && fname != null && !fname.endsWith(".html") &&
		       !fname.endsWith(".htm") && !fname.endsWith("/") && 
		       !fname.endsWith(".txt")) {
		    /* for unfound html files, the server's response
		     * can sometimes contain useful information, like a
		     * help page.  In all other cases, e.g., ".class" or
		     * ".gif" or ".jpg" or ".ser", it is more telling
		     * to raise FileNotFoundException(), which for example
		     * keeps the AppletClassLoader from parsing an HTML page
		     * into a Class.
		     */
		    throw new java.io.IOException("File Not Found: " + url.toString());
		}

		return inputStream;
	    } while (redirects < maxRedirects);

	    // too many redirects if we got here
	    throw new ProtocolException("Server redirected too many times ("
					    + redirects + ")");
	} catch (RuntimeException e) {
	    disconnect();
	    throw e;
	} catch (IOException e) {
	    disconnect();
	    throw e;
	}
    }


    /**
     * Gets the authentication for an HTTP proxy, and applies it to
     * the connection.
     */
    private AuthenticationInfo getHttpProxyAuthentication() {
	/* get authorization from authenticator */
	AuthenticationInfo ret = null;
	String raw = getHeaderField("Proxy-authenticate");
	String host = http.getProxyHostUsed();
	int port = http.getProxyPortUsed();
	if (host != null && raw != null) {
	    HeaderParser p = new HeaderParser(raw);
	    String realm = p.findValue("realm");
	    String scheme = p.findKey(0);
	    ret = AuthenticationInfo.getProxyAuth(host, port, realm);

	    if (ret == null && defaultAuth != null
		&& defaultAuth.schemeSupported(scheme)) {
		try {
		    URL u = new URL("http", host, port, "/");
		    String a = defaultAuth.authString(u, scheme, realm);
		    if (a != null) {
			ret = new BasicAuthentication (true, host, port, realm, a);
			// not in cache by default - cache on success
		    }
		} catch (java.net.MalformedURLException ignored) {
		}
	    }
	    if (ret != null) {
		if (!ret.setHeaders(this, p)) {
		    ret = null;
		}
	    }
	}
	return ret;
    }

    /**
     * Gets the authentication for an HTTP server, and applies it to
     * the connection.
     */
    private AuthenticationInfo getServerAuthentication() {
	/* get authorization from authenticator */
	AuthenticationInfo ret = null;
	String raw = getHeaderField("WWW-Authenticate");
	if (raw != null) {
	    HeaderParser p = new HeaderParser(raw);
	    String realm = p.findValue("realm");
	    String scheme = p.findKey(0);
	    ret = AuthenticationInfo.getServerAuth(url, realm);

	    // For backwards compatibility, we also try defaultAuth
	    // REMIND:  Get rid of this for JDK2.0.

	    if (ret == null && defaultAuth != null
		&& defaultAuth.schemeSupported(scheme)) {
		String a = defaultAuth.authString(url, scheme, realm);
		if (a != null) {
		    ret = new BasicAuthentication (false, url, realm, a); 
		    // not in cache by default - cache on success
		}
	    }

	    if (ret != null) {
		if (!ret.setHeaders(this, p)) {
		    ret = null;
		}
	    }
	}
	return ret;
    }

    /* Tells us whether to follow a redirect.  If so, it
     * closes the connection (break any keep-alive) and
     * resets the url, re-connects, and resets the request
     * property.
     */
    private boolean followRedirect() throws IOException {
	if (!shouldFollowRedirects) {
	    return false;
	}

	if (!getFollowRedirects()) {
	    return false;
	}

	int stat = getResponseCode();
	if (stat < 300 || stat > 305 || stat == HTTP_NOT_MODIFIED) {
	    return false;
	}
	String loc = getHeaderField("Location");
	if (loc == null) { 
	    /* this should be present - if not, we have no choice
	     * but to go forward w/ the response we got
	     */
	    return false;
	}
	disconnect();
	// clear out old response headers!!!!
	responses = new MessageHeader();
	if (stat == HTTP_USE_PROXY) {
	    /* This means we must re-request the resource through the
	     * proxy denoted in the "Location:" field of the response.
	     * Judging by the spec, the string in the Location header
	     * _should_ denote a URL - let's hope for "http://my.proxy.org"
	     * Make a new HttpClient to the proxy, using HttpClient's
	     * Instance-specific proxy fields, but note we're still fetching
	     * the same URL.
	     */
	    URL urlp = new URL(loc);
	    http = getProxiedClient (url, urlp.getHost(), urlp.getPort());
	    requests.set(0, method + " " + http.getURLFile()+" "  + 
			     httpVersion, null);
	    connected = true;
	} else {
	    // maintain previous headers, just change the name
	    // of the file we're getting
	    url = new URL(url, loc);

	    if (!url.getProtocol().startsWith("http")) {
		/* we're still following a redirect, but we shouldn't
		 * pursue it via the HTTP protocol.  So just blow out
		 * of here and return true.
		 */
		return true;
	    }

	    if (method.equals("POST") && !Boolean.getBoolean("http.strictPostRedirect")) {
		/* The HTTP/1.1 spec says that a redirect from a POST 
		 * *should not* be immediately turned into a GET, and
		 * that some HTTP/1.0 clients incorrectly did this.
		 * Correct behavior redirects a POST to another POST.
		 * Unfortunately, since most browsers have this incorrect
		 * behavior, the web works this way now.  Typical usage
		 * seems to be:
		 *   POST a login code or passwd to a web page.
		 *   after validation, the server redirects to another
		 *     (welcome) page
		 *   The second request is (erroneously) expected to be GET
		 * 
		 * We will do the incorrect thing (POST-->GET) by default.
		 * We will provide the capability to do the "right" thing
		 * (POST-->POST) by a system property, "http.strictPostRedirect=true"
		 */

		requests = new MessageHeader();
		setRequests = false;
		setRequestMethod("GET");
		poster = null;
		connect();
	    } else {
		connect();
		requests.set(0, method + " " + http.getURLFile()+" "  + 
			     httpVersion, null);
		requests.set("Host", url.getHost() + ((url.getPort() == -1 || url.getPort() == 80) ?
			                             "": ":"+String.valueOf(url.getPort())));
	    }
	}
	return true;
    }
	
    /**
     * Disconnect from the server.
     */
    public void disconnect() {

	responseCode = -1;
	if (pe != null) {
	    ProgressData.pdata.unregister(pe);
	    pe = null;
	}
	if (http != null) {
	    http.closeServer();
	    //	    poster = null;
	    http = null;
	    connected = false;
	}
    }

    public boolean usingProxy() {
	if (http != null) {
	    return http.usingProxy;
	}
	return false;
    }

    /**
     * Gets a header field by name. Returns null if not known.
     * @param name the name of the header field
     */
    public String getHeaderField(String name) {
	try {
	    getInputStream();
	} catch (IOException e) {}
	return responses.findValue(name);
    }

    /**
     * Gets a header field by index. Returns null if not known.
     * @param n the index of the header field
     */
    public String getHeaderField(int n) {
	try {
	    getInputStream();
	} catch (IOException e) {}
	return responses.getValue(n);
    }

    /**
     * Gets a header field by index. Returns null if not known.
     * @param n the index of the header field
     */
    public String getHeaderFieldKey(int n) {
	try {
	    getInputStream();
	} catch (IOException e) {}
	return responses.getKey(n);
    }

    public void setRequestProperty(String key, String value) {
        if (connected)
            throw new IllegalAccessError("Already connected");
	/* There should be only one header per key, but special-case
	 * HTTP cookies.
	 */
	if ("set-cookie".equalsIgnoreCase(key)) {
	    requests.add(key, value);
	} else {
	    requests.set(key, value);
	}
    }

    //
    // Set a property for authentication.  This can safely disregard
    // the connected test.
    //
    void setAuthenticationProperty(String key, String value) {
        requests.set(key, value);
    }
 
    public String getRequestProperty (String key) {
	return requests.findValue(key);
    }

    public void finalize() {
	/* if pe isn't null, we unregister it, just in case the applet
           is misbehaved and didn't call close on its streams. */
	if (pe != null) {
	    ProgressData.pdata.unregister(pe);
	    pe = null;
	}
	// The stream finalizer will close the fd
    }

    // This is a temporary workaround to provide a per-instance mechanism
    // for controlling the handling of redirections from the protocol
    // handler
    public void setInstanceFollowRedirects(boolean shouldFollow) {
	shouldFollowRedirects = shouldFollow;
    }
    String getMethod() {
	return method;
    }
}

/** An input stream that just returns EOF.  This is for
 * HTTP URLConnections that are KeepAlive && use the
 * HEAD method - i.e., stream not dead, but nothing to be read.
 */

class EmptyInputStream extends InputStream {

    public int available() {
	return 0;
    }

    public int read() {
	return -1;
    }
}
