/*
 * @(#)SystemResourceManager.java	1.20 99/07/12
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.net.www.protocol.systemresource;

import java.net.*;
import java.util.*;
import java.io.*;
import java.applet.*;
import java.awt.*;
import java.awt.image.*;

import sun.applet.*;
import sun.awt.image.*;

import sun.io.FileIO;
import sun.io.FileIOFactory;

/**
 * Manage system resources, including caching images and audioclips.
 * This class is an auxiliary class to SystemResourceConnection.
 *
 * NOTE: Changes to this class need to be synchronized with 
 *       src/share/optional/noappletsupport/sun/net/www/protocol/
 *       systemresource/SystemResourceManager.java
 *
 * @see		SystemResourceConnection
 * @version 	1.20, 07/12/99
 * @author	Eduardo Pelegri-Llopart
 */

public class SystemResourceManager {
    private static boolean debug = false;

    /**
     * The URL for this system resource
     */
    private URL url;
    private String base;
    private String member;
    private boolean isZip;
    private String compoundName;

    /**
     * Construct a SystemResourceManager for this URL.
     * May raise exceptions with malformed URLs and with security access
     */

    SystemResourceManager(URL url) 
	    throws MalformedURLException {
	this.url = url;
	ParseSystemURL parse = new ParseSystemURL(url);
	String key = parse.getBase(); // the key
	try {
	    base = getEntryFromKey(Integer.parseInt(key));
	} catch (Exception ex) {
	    base = null;
	} 
	member = parse.getMember();
	isZip = parse.isZip();
	if (base == null) {
	    throw new MalformedURLException(url+
					    " is not a valid system resource URL");
	    // used to be: compoundName = member;
	} else {
	    compoundName = base + FileIO.separator + member;
	}
	if (!parse.isValid()) {
	    throw new MalformedURLException(url+
					    " is not a valid system resource URL");
	    }
	if (!validateSystemResource(isZip, base, member)) {
	    throw new SecurityException(url +
					" refers to a non system resource");
	}

    }

    /**
     * Acquire a local resource as an Object
     */

    synchronized Object getLocalResource() {
	if (isZip) {
	    return getFromZip();
	} else {
	    return getFromFile();
	}
    }

    /**
     * Get a local resource as an InputStream
     * We DO NOT go through the ClassLoader methods, as we don't want
     * to expose this access path
     */

    InputStream getLocalResourceStream() {
	debug("SystemResourceManager.getLocalResourceStream::");
	if (isZip) {
	    debug("  isZip; base: "+base+" member: "+member);
	    byte[] data = getZipResourceByteArray(base, member);
	    return new ByteArrayInputStream(data);
	} else {
	    try {
		return new BufferedInputStream(FileIOFactory.getInputStream(compoundName));
	    } catch (Exception ex) {
		debug("Could not locate a resource base: "+base
		      +" member: "+member);
		return null;
	    }
	}
    }

    /**
     * Private methods
     */

    private Object getFromZip() {
	Object back;
	String type;		// the type of the stream

	byte[] buf = getZipResourceByteArray(base, member);
	InputStream stream = new ByteArrayInputStream(buf);
	try {
	    type = URLConnection.guessContentTypeFromStream(stream);
	} catch (IOException ex) {
	    type = null;
	}

	if (type != null) {
	    if (type.startsWith("image")) {
		// REMIND - the image code should get it from the Image cache
		return new ByteArrayImageSource(buf);
	    } else if (type.startsWith("audio")) {
		// REMIND - check what the souce code does
		return new AppletAudioClip(buf);
		}
	}
	return stream;		// return as an InputStream
    }
    
    private Object getFromFile() {
	Object back;
	String type;		// the type of the stream

        if (compoundName.regionMatches(true, compoundName.length()-6,
				       ".class", 0, 6)) {
	     throw new java.lang.SecurityException("Can't read classfile as system resource");
	}

	InputStream stream;
	try {
	    stream = new BufferedInputStream(FileIOFactory.getInputStream(compoundName));
	} catch (Exception ex) {
	    debug("Could not locate a resource base: "+base
		  +" member: "+member);
	    return null;
	}

	try {
	    type = URLConnection.guessContentTypeFromStream(stream);
	} catch (IOException ex) {
	    debug("Exception while guessing Content Type; e: "+ex);
	    type = null;
	}

	if (type != null) {
	    if (type.startsWith("image")) {
		debug("getFromFile:: base: "+base+" member: "+member);
		try {
		  stream.close();
		} catch (IOException ioe) {
		}
		return new FileImageSource(compoundName);
	    } else if (type.startsWith("audio")) {
		byte[] buf = byteArrayFromStream(stream);
		try {
		  stream.close();
		} catch (IOException ioe) {
		}
		// REMIND - There should be a way of creating an AppletAudioClip
		// from a generic stream.
		return new AppletAudioClip(buf);
	    }
	}
	return stream;		// return as an InputStream
    }

    /*
     * Auxiliary: Get a byte array with the contents of an input stream.
     */

    private static byte[] byteArrayFromStream(InputStream s) {
	int len;
	try {
	    len = s.available();
	    byte[] buf = new byte[len];
	    int count = 0;
	    while (count < len) {
		int n = s.read(buf, count, len-count);
		if (n < 0) {
		    return null;	// error reading data
		}
		count += n;
		// Thread.currentThread().yield(); // probably not needed here
	    }
	    return buf;
	} catch (Exception e) {
	    debug("Unexpected internal exception... "+e);
	    throw new Error ("Unexpected internal exception");
	}
    }

    static void debug(String s) {
	if (debug)
	    System.out.println("SystemResourceManager:: "+s);
    }

    /**
     * Convert a KEY value into the actual CLASSPATH entry it represents
     * By using opaque KEYS we are hiding information on files on the users machine
     */

    private native static String getEntryFromKey(int key);
	

    /**
     * Access the byte array.
     * REMIND - this can/should be replaced by Java code
     */

    private native static byte[] getZipResourceByteArray(String base,
						 String member);

    /**
     * Validate that the URL is indeed to a system resource (and not to some
     * file in the local file system)
     * This needs to be done in native code.
     */ 

    private native static boolean validateSystemResource(boolean isZip,
						 String base,
						 String member);

    /**
     * Load the library containing the native methods
     */

    static {
	System.loadLibrary("net");
    }
}
