/*
 * @(#)RemoteClassLoader.java	1.1 98/02/11 SMI
 *
 * Copyright (c) 1996 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

/* Code to support the loading of classes from a well known server. This
 * version loads classes from zip files. The appropriate zip file is
 * located by using the remote-classes.properties file.
 */

package sun.pjava.lang;

import java.io.*;
import java.util.*;
import java.util.zip.*;
import java.net.*;

import sun.io.FileIO;
import sun.io.FileIOFactory;

public final class RemoteClassLoader
{
    // string containing the URL directory where the implementation
    // of the remote classes are located.
    private static String codebase;

    // Properites database mapping classnames to the remote file
    // that each remote classes' implemention is contained in.
    private static Properties remoteClasses;

    // Database of classes that have already been remotely loaded
    // and defined. This ensures that we don't load and define
    // the same class twice.
    private static Hashtable definedClasses = new Hashtable();

    /* Don't let anyone instantiate this class */
    private RemoteClassLoader() {
    }

    private static void initRemoteClassLoader() {
	readProperties();
	init();  // let native code do some initialization
    }

    private native static void init();

    /* findRemoteClass is called by the vm when it has failed to locate a
     * class using normal methods (CLASSPATH and AppletClassLoader).
     *
     * findRemoteClass checks to see if the class is registered in the
     * remote-classes.properties file. If it is then it creates an
     * InputStream to the remote file that contains the class. This
     * InputStream is handed off to processRemoteFile to do whatever
     * is necessary to register all class in the file.
     *
     * The Class object for the requested class is returned.
     */
    public synchronized static Class findRemoteClass(String classname) {
	if (remoteClasses == null)
	    return null;  // we don't know about any remote classes

	Class result = null;

	try {
	    // If we've already defined this class, then just return it
	    result = (Class)definedClasses.get(classname);
	    if (result != null) {
		//System.out.println(classname + " already loaded");
		return result;
	    }

	    // Get a url input stream for the file that the remote class is in
	    String remoteFileName = remoteClasses.getProperty(classname);
	    if (remoteFileName == null) {
		//System.out.println(classname + " is not a remote class");
		return null;
	    }
	    String remoteFileURLString = codebase + remoteFileName;
	    URL remoteFileURL = new URL(remoteFileURLString);
	    InputStream remoteFileURLStream = remoteFileURL.openStream();

	    // Process all the classes that are in the remote file
	    result = processRemoteFile(remoteFileURLStream, classname);
	}
	catch (Throwable e) {
	    System.out.println("findRemoteClass: Exception thrown!");
	    //e.printStackTrace();
	    result = null;
	}

	return result;
    }

    /* processRemoteFile is responsible for extracting each class out of the
     * InputStream and calling defineClass on each class. It returns the
     * Class object for the classname parameter passed in.
     *
     * This implementation of processRemoteFile handles an InputStream for
     * a zip file. Other implementation are also possible, including supporting
     * MClass files or files on a PC-Card. If a URL can point to the file
     * then the only code that would need to change is processRemoteFile.
     */
    private static Class processRemoteFile(InputStream remoteStream,
					   String classname) {
	Class result = null;

	try {
	    // get a ZipInputStream for the remote file
	    ZipInputStream zin = new ZipInputStream(remoteStream);

	    // register all the classes in the zip file
	    ZipEntry entry;
	    while ((entry = zin.getNextEntry()) != null) {
		Class cb = null;

		//System.out.println(entry.getName() + " "  + entry.getSize());
		if (entry.isDirectory())
		    continue;

		// convert a/b/myclass.class to a/b/myclass
		int namelen = entry.getName().length() - 6; // less .class length
		String name = entry.getName().substring(0, namelen);

		// make sure we don't define any classes twice
		if (definedClasses.get(name) != null)
		    continue;
		
		// read the class in and define it
		byte classdata[] = readFile(zin, (int)entry.getSize());
		try {
		    cb = defineClass(name, classdata, 0, (int)entry.getSize());
		} catch (Throwable ex) {
			//System.err.println("Bad Class not Loaded: " + name);
			//ex.printStackTrace();
		}
		if (cb != null) {
		    definedClasses.put(name, cb);
		    // if this is the classes requrested then this is the one
		    // that we want to return when we're all done.
		    if (name.compareTo(classname) == 0)
			result = cb;
		}

		zin.closeEntry();
	    }
	    zin.close();
	}
	catch (Throwable e) {
	    //e.printStackTrace();
	    result = null;
	}

	//System.out.println("result is " + result);
	return result;
     }

    /* defineClass is exactly the same as ClassLoader.defineClass(), except
     * it doesn't initialize the class. This happens automatically later.
     */
    private native static Class defineClass(String name, byte[] data,
				     int offset, int length);

    /* readProperites reads in the remote-classes.properties database and
     * sets up the codebase field.
     */
    private static  void readProperties() {
	String jhome = System.getProperty("java.home");
	if (jhome != null) {
	    try {
		FileIO f = FileIOFactory.newInstance(jhome + FileIO.separator +
				  "lib" + FileIO.separator + 
				  "remote-classes.properties");
		if (!f.canRead())
		    return;
		InputStream in =
		    new BufferedInputStream(FileIOFactory.getInputStream(f.getPath()));
		remoteClasses = new Properties();
		remoteClasses.load(in);
		in.close();
		codebase = remoteClasses.getProperty("codebase");
		if (codebase == null) {
		    System.out.println("RemoteClassLoader: can't find codebase property");
		    remoteClasses = null;
		}
	    }
	    catch (IOException e) {
		e.printStackTrace();
		return;
	    }
	}
    }

    /* Reads the entire InputStream into a byte array and returns the
     * byte array. Note that if the InputStream is a zip file then the
     * current entry is read, not the entire zip file.
     */
    static byte[] readFile(InputStream in, int len) throws IOException {

        byte buf[] = new byte[len];
        int off = 0;
        
        while (len > 0) {
            int n = in.read(buf, off, len);
            if (n == -1) {
                //System.err.println("EOF while reading file");
                throw new EOFException();
            }
            off += n;
            len -= n;
        }

        return buf;
    }

}
