/*
 * @(#)RegistryImpl.java	1.8 98/07/01
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.rmi.registry;

import java.util.Enumeration;
import java.util.Hashtable;
import java.net.InetAddress;
import java.rmi.*;
import java.rmi.server.ObjID;
import java.rmi.server.RemoteServer;
import java.rmi.server.ServerNotActiveException;
import java.rmi.registry.Registry;
import sun.rmi.server.UnicastServerRef;
import sun.rmi.transport.LiveRef;
import sun.rmi.transport.ObjectTable;
import sun.rmi.transport.Target;

/**
 * A "registry" exists on every node that allows RMI connections to
 * servers on that node.  The registry on a particular node contains a
 * transient database that maps names to remote objects.  When the
 * node boots, the registry database is empty.  The names stored in the
 * registry are pure and are not parsed.  A service storing itself in
 * the registry may want to prefix its name of the service by a package
 * name (although not required), to reduce name collisions in the
 * registry.
 *
 * The LocateRegistry class is used to obtain registry for different hosts.
 *
 * @see java.rmi.registry.LocateRegistry
 */
public class RegistryImpl extends java.rmi.server.RemoteServer
	implements Registry
{
    private Hashtable bindings = new Hashtable(101);
    private static RegistryImpl registry;
    private static InetAddress localhost = null;
    private static ObjID id = new ObjID(ObjID.REGISTRY_ID);

    /**
     * Construct a new RegistryImpl on the default port.
     */
    public RegistryImpl() throws RemoteException {
	setup(new LiveRef());
    }

    /**
     * Construct a new RegistryImpl on the specified port.
     */
    public RegistryImpl(int port)
	throws RemoteException
    {
	LiveRef lref = new LiveRef(id, port);

	setup(lref);
    }
    
    /*
     * Create the UnicastServerRef from the liveRef and export the object
     */
    private void setup(LiveRef lref)
	throws RemoteException
    {
	/* Server ref must be created and assigned before remote object 
	 * 'this' can be exported.
	 */
	UnicastServerRef uref = new UnicastServerRef(lref);
	ref = uref;
	uref.exportObject(this, null);
	Target t = ObjectTable.getTarget(this);
	t.setPermanent(true);
    }


    /**
     * Returns the remote object for specified name in the registry.
     * @exception RemoteException If remote operation failed.
     * @exception NotBound If name is not currently bound.
     */
    public Remote lookup(String name)
	throws RemoteException, NotBoundException
    {
	synchronized (bindings) {
	    Remote obj = (Remote)bindings.get(name);
	    if (obj == null)
		throw new NotBoundException(name);
	    return obj;
	}
    }
    

    /**
     * Binds the name to the specified remote object.
     * @exception RemoteException If remote operation failed.
     * @exception AlreadyBoundException If name is already bound.
     */
    public void bind(String name, Remote obj)
	throws RemoteException, AlreadyBoundException, AccessException
    {
	checkAccess("bind");
	synchronized (bindings) {
	    Remote curr = (Remote)bindings.get(name);
	    if (curr != null)
		throw new AlreadyBoundException(name);
	    bindings.put(name, obj);
	}
    }
    
    
    /**
     * Unbind the name.
     * @exception RemoteException If remote operation failed.
     * @exception NotBound If name is not currently bound.
     */
    public void unbind(String name)
	throws RemoteException, NotBoundException, AccessException
    {
	checkAccess("unbind");
	synchronized (bindings) {
	    Remote obj = (Remote)bindings.get(name);
	    if (obj == null)
		throw new NotBoundException(name);
	    bindings.remove(name);
	}
    }
    
    /** 
     * Rebind the name to a new object, replaces any existing binding.
     * @exception RemoteException If remote operation failed.
     */
    public void rebind(String name, Remote obj)
	throws RemoteException, AccessException
    {
	checkAccess("rebind");
	bindings.put(name, obj);
    }

    /**
     * Returns an enumeration of the names in the registry.
     * @exception RemoteException If remote operation failed.
     */
    public String[] list()
	throws RemoteException
    {
	String[] names;
	synchronized (bindings) {
	    int i = bindings.size();
	    names = new String[i];
	    Enumeration enum = bindings.keys();
	    while ((--i) >= 0)
		names[i] = (String)enum.nextElement();
	}
	return names;
    }

    /*
     * Check that the Client is on the same host as this server.
     */
    void checkAccess(String op) throws AccessException {
	try {
	    if (localhost == null)
		localhost = InetAddress.getLocalHost();

	    String clientname = getClientHost();
	    InetAddress host = InetAddress.getByName(clientname);

	    if (!host.equals(localhost) &&
		!clientname.equals(localhost.getHostName()))
		throw new AccessException("Registry." + op + " " +
					  host + " != " + localhost);
	} catch (ServerNotActiveException ex) {
	    // Ignore, called from local host, allow modification
	} catch (java.net.UnknownHostException ex) {
	    // Don't know our own name, can't compare, disallow
	    throw new AccessException("Registry." + op);
	}
    }

    public static ObjID getID() {
	return id;
    }
    
    /**
     * Main program to start a registry. <br>
     * The port number can be specified on the command line.
     */
    public static void main(String args[])
    {
	// Create and install the security manager
	System.setSecurityManager(new RMISecurityManager());

	try {
	    int port = Registry.REGISTRY_PORT;
	    if (args.length >= 1) {
		port = Integer.parseInt(args[0]);
	    }
	    registry = new RegistryImpl(port);
	    // prevent registry from exiting
	    while (true) {
		try {
		    // The following timeout is used because a bug in the
		    // native C code for Thread.sleep() cause it to return
		    // immediately for any higher value.
		    Thread.sleep(Integer.MAX_VALUE - 1);
		} catch (InterruptedException e) {
		}
	    }
	} catch (NumberFormatException e) {
	    System.out.println("Port is not a number.");
	} catch (Exception e) {
	    System.out.println("RegistryImpl.main: an exception occurred: " +
			       e.getMessage());
	    e.printStackTrace();
	}
	System.exit(1);
    }
}
