/*
 * @(#)MarshalInputStream.java	1.8 98/07/01
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.rmi.server;

import java.io.*;
import java.net.*;
import sun.rmi.transport.Utils;

/**
 * The MarshalInputStream is extended from ObjectInputStream to extract the 
 * location of stub classes to dynamically load and to create stubs for remote 
 * objects. The location is used only if the stubs are not available locally.
 * <p>
 * A new MarshalInputStream should be created to deserialize remote objects or 
 * graphs containing remote objects. Objects are created from the stream using 
 * the ObjectInputStream.readObject method. MarshalInputStream uses information 
 * embedded in the stream to locate and load stub classes, if those stubs are 
 * not available locally.
 */

public class MarshalInputStream extends ObjectInputStream
{
    
    /**
     * Create a new MarshalInputStream object.
     */
    public MarshalInputStream(InputStream in)
	throws IOException, StreamCorruptedException
    {	      
	super(in);
    }


    /**
     * resolveClass is extended to acquire (if present) the location
     * from which to load the specified class.
     * It will find, load, and return the class.
     */
    protected Class resolveClass(ObjectStreamClass classDesc)
	throws IOException, ClassNotFoundException
    {
	// read annotation describing from where to load class
	Object obj = readObject();

	/* Always try loading class using default mechanism first (using
	 * first class loader on the stack, if any) so that we don't reload
	 * the same class with the RMIClassLoader, which would not be
	 * compatible with the other as far as the runtime is concerned.
	 */
	ClassNotFoundException defaultException;
	try {
	    return super.resolveClass(classDesc);
	} catch (ClassNotFoundException e) {
	    defaultException = e;
	} // should we ignore and proceed on IOException too??

	String classname = classDesc.getName();
	RMIClassLoader loader = null;
	String location = null;

	/* If loading only from the codebase is not set and there
	 * is a location to load the class from, use its stub loader,
	 * otherwise use the default local loader.
	 */
	try {
	    if (!codebaseonly && obj != null && (obj instanceof String)) {
		location = (String)obj;
		URL url = new URL(location);
		loader = RMIClassLoader.getClassLoader(url);
	    } else {
		loader = RMIClassLoader.getLocalLoader();
	    }
	} catch (MalformedURLException ex) {
	    throw new IOException("Malformed URL " + location);
	}

	if (loader != null)
	    return loader.loadClass(classname);
	else
	    throw defaultException;
    }


    /* Cache the value of the java.rmi.server.useCodebaseOnly property.
     * If true enable loading only from the locally defined stub base.
     */
    private static boolean codebaseonly =
    	Utils.getBoolean("java.rmi.server.useCodebaseOnly");
}
