/*
 * @(#)MarshalOutputStream.java	1.10 98/07/01
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.rmi.server;

import java.io.*;
import java.rmi.Remote;
import sun.rmi.transport.Utils;

/**
 * A MarshalOutputStream extends ObjectOutputStream to add functions
 * specific to marshaling of remote object references. If it is
 * necessary to serialize remote objects or objects that contain
 * references to remote objects a MarshalOutputStream must be used
 * instead of ObjectOutputStream. <p>
 *
 * A new MarshalOutputStream is constructed to serialize remote
 * objects or graphs containing remote objects. Objects are written to
 * the stream using the ObjectOutputStream.writeObject method. <p>
 *
 * MarshalOutputStream maps remote objects to the corresponding remote
 * stub and embeds the location from which to load the stub
 * classes. The location may be ignored by the client but is supplied.
 */
public class MarshalOutputStream extends ObjectOutputStream
{
    private static String codebase;

    /** Create a marshaling stream to handle RMI marshaling
     */
    public MarshalOutputStream(OutputStream out) throws IOException {
	super(out);
	enableReplaceObject(true);
    }

    /** replaceObject is extended to check for instances of Remote
     * that need to be serialized as proxy objects.  RemoteProxy.getProxy
     * is called to check for and find the stub.
     */
    protected Object replaceObject(Object obj) throws IOException {
	if (obj instanceof Remote) {
	    obj = RemoteProxy.getProxy((Remote)obj);
	}
	return obj;
    }


    /**
     * annotateClass is extended to serialize a location from which
     * to load the the specified class.
     */
    protected void annotateClass(Class cl) throws IOException {

	/* If the class is in a "java.*" package, use null.
	 * If the class was loaded by an RMIClassLoader, use its codebase URL.
	 * If the class was loaded by another ClassLoader, use the security
	 * context, if it is a URL.  If it isn't, or it has no class loader,
	 * then default to the local codebase property, if set.
	 */
	String location = null;
	String classname = cl.getName();
	if (!classname.startsWith("java.")) {
	    ClassLoader loader = cl.getClassLoader();
	    if (loader != null) {
		if (loader instanceof RMIClassLoader) {
		    RMIClassLoader stubloader = (RMIClassLoader)loader;
		    location = stubloader.getCodeBase().toExternalForm();
		} else {
		    /* Guess/Assume that its an applet class loader
		     * and use the current security context for the location
		     * if its a URL.
		     */
		    SecurityManager sm = System.getSecurityManager();
		    Object sc;
		    if (sm != null &&
			(sc = sm.getSecurityContext()) != null &&
			(sc instanceof java.net.URL)) {
			location = ((java.net.URL)sc).toExternalForm();
		    }
		}
	    }

	    if (location == null) {
		// default to using codebase
		location = codebase;
	    }
	}
	
	// write the specified location (may be null).
	writeObject(location);
    }

    static 
    {
	// initialize the codebase to propagate with classes
	codebase = Utils.getProperty("java.rmi.server.codebase");
	if (codebase != null) {
	    try {
		// make sure that codebase is a proper URL
		new java.net.URL(codebase);
	    } catch (java.net.MalformedURLException e) {
		// do not propagate a malformed URL
		codebase = null;
	    }
	}
    }
}
