/*
 * @(#)UnicastRef.java	1.9 98/07/14
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.rmi.server;

import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.net.InetAddress;
import java.rmi.MarshalException;
import java.rmi.RemoteException;
import java.rmi.Remote;
import java.rmi.server.LogStream;
import java.rmi.server.Operation;
import java.rmi.server.RemoteCall;
import java.rmi.server.RemoteObject;
import java.rmi.server.RemoteRef;
import sun.rmi.transport.Channel;
import sun.rmi.transport.Connection;
import sun.rmi.transport.LiveRef;
import sun.rmi.transport.Transport;
import sun.rmi.transport.StreamRemoteCall;

public class UnicastRef implements RemoteRef, java.io.Serializable {

    protected LiveRef ref;

    /**
     * Create a new (empty) Unicast remote reference.
     */
    public UnicastRef()
    {}
    
    /** 
     * Create a new Unicast RemoteRef.
     */
    public UnicastRef(LiveRef liveRef) {
	ref = liveRef;
    }

    /**
     * Create an appropriate call object for a new call on this object.
     * Passing operation array and index, allows the stubs generator to
     * assign the operation indexes and interpret them. The RemoteRef
     * may need the operation to encode in for the call.
     */
    public RemoteCall newCall(RemoteObject obj, Operation[] ops, int opnum,
			      long hash)
	throws RemoteException
    {
	/* create remote connection and call object*/
	if (RemoteProxy.logLevel >= LogStream.BRIEF)
	    LogStream.log("transport").println(
		"UnicastRef.newCall: get connection");

	Connection conn = ref.getChannel().newConnection();
	try {
	    if (RemoteProxy.logLevel >= LogStream.VERBOSE)
		LogStream.log("transport").println(
		    "UnicastRef.newCall: create call");

	    RemoteCall call =
		new StreamRemoteCall(conn, ref.getObjID(), opnum, hash);
	    try {
		marshalCustomCallData(call.getOutputStream());
	    } catch (IOException e) {
		throw new MarshalException("error marshaling " +
					   "custom call data");
	    }
	    return call;
	} catch (RemoteException e) {
	    ref.getChannel().free(conn, false);
	    throw e;
	}
    }
  
    protected void marshalCustomCallData(ObjectOutput out)
	throws IOException
    {}
    
    /**
     * Invoke makes the remote call present in the RemoteCall object.
     * 
     * Invoke will raise any "user" exceptions which
     * should pass through and not be caught by the stub.  If any
     * exception is raised during the remote invocation, invoke should
     * take care of cleaning up the connection before raising the
     * "user" or remote exception.
     */
    public void invoke(RemoteCall call) throws Exception
    {
	try {
	    if (RemoteProxy.logLevel >= LogStream.VERBOSE)
		LogStream.log("transport").println(
		    "UnicastRef.invoke: execute call");

	    call.executeCall();

	} catch (RemoteException e) {
	    /* should cleanup connection for some remote exceptions...
	     * until they are separated appropriately (by class),
	     * assume the connection is not reusable. */
	    if (RemoteProxy.logLevel >= LogStream.BRIEF)
		LogStream.log("transport").println(
		    "UnicastRef.invoke: exception occurred " +
		    e.getMessage());

	    Connection conn = ((StreamRemoteCall)call).getConnection();
	    ref.getChannel().free(conn, false);
	    /* reraise remote exception */
	    throw e;
	
	} catch (Exception e) {
	    /* assume that these other exceptions leave the connection in
	     * a reusable state. */
	    if (RemoteProxy.logLevel >= LogStream.BRIEF)
		LogStream.log("transport").println(
		    "UnicastRef.invoke: exception occurred " +
		    e.getMessage());

	    Connection conn = ((StreamRemoteCall)call).getConnection();
	    ref.getChannel().free(conn, true);
	    /* reraise user (and unknown) exceptions. */
	    throw e;
	    
	}
    }

    /**
     * Done should only be called if the invoke returns successfully
     * (non-exceptionally) to the stub. It allows the remote reference to
     * clean up (or reuse) the connection.
     */
    public void done(RemoteCall call) throws RemoteException {
	
	/* Done only uses the connection inside the call to obtain the
	 * channel the connection uses.  Once all information is read
	 * from the connection, the connection may be freed.  
	 */
	if (RemoteProxy.logLevel >= LogStream.BRIEF)
	    LogStream.log("transport").println
		("UnicastRef.done: free connection (reuse = "
		 + " true)");
	
	/* Free the call connection early. */
	Connection conn = ((StreamRemoteCall) call).getConnection();
	ref.getChannel().free(conn, true);
	
	try {
	    call.done();
	} catch (IOException e) {
	    /* WARNING: If the conn has been reused early, then it is
	     * too late to recover from thrown IOExceptions caught
	     * here. This code is relying on StreamRemoteCall.done()
	     * not actually throwing IOExceptions.  
	     */
	}
    }

    /**
     * Returns the class of the ref type to be serialized
     */
    public String getRefClass(ObjectOutput out)
    {
	return "UnicastRef";
    }

    /**
     * Write out external representation for remote ref.
     */
    public void writeExternal(ObjectOutput out) throws IOException 
    {
	ref.writeExternal(out);
    }

    /**
     * Read in external representation for remote ref.
     * @exception ClassNotFoundException If the class for an object
     * being restored cannot be found.
     */
    public void readExternal(ObjectInput in)
	throws IOException, ClassNotFoundException
    {
	ref = new LiveRef();
	ref.readExternal(in);
    }
    
    //----------------------------------------------------------------------;
    /**
     * Method from object, forward from RemoteObject
     */
    public String remoteToString() {
	return "RemoteStub [ref: " + ref + "]";
    }

    /**
     * default implementation of hashCode for remote objects
     */
    public int remoteHashCode() {
	return ref.hashCode();
    }

    /** default implementation of equals for remote objects
     */
    public boolean remoteEquals(RemoteRef sub) {
	if (sub instanceof UnicastRef)
	    return ref.remoteEquals(((UnicastRef)sub).ref);
	return false;
    }
}
