/*
 * @(#)UnicastServerRef.java	1.12 98/07/14
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.rmi.server;

import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.net.InetAddress;
import java.rmi.*;
import java.rmi.server.ExportException;
import java.rmi.server.Operation;
import java.rmi.server.RemoteCall;
import java.rmi.server.RemoteRef;
import java.rmi.server.RemoteStub;
import java.rmi.server.RemoteServer;
import java.rmi.server.ServerNotActiveException;
import java.rmi.server.ServerRef;
import java.rmi.server.Skeleton;
import sun.rmi.transport.LiveRef;
import sun.rmi.transport.Target;
import sun.rmi.transport.Channel;
import sun.rmi.transport.tcp.TCPChannel;
import sun.rmi.transport.tcp.TCPEndpoint;
import sun.rmi.transport.tcp.TCPTransport;

/**
 * Server-side ref for a remote impl.
 *
 * @author Ann Wollrath
 * @author Roger Riggs
 */
public class UnicastServerRef extends UnicastRef
	implements ServerRef, Dispatcher
{
    transient private Skeleton skel;
        
    /**
     * Create a new (empty) Unicast server remote reference.
     */
    public UnicastServerRef() 
    {}

    /**
     * Construct a Unicast server remote reference for a specified
     * liveRef.
     */
    public UnicastServerRef(LiveRef ref)
    {
	super(ref);
    }

    /**
     * Export this object, create the skeleton and stubs for this
     * dispatcher.  Create a stub based on the type of the impl,
     * initialize it with the appropriate remote reference. Create the
     * target defined by the impl, dispatcher (this) and stub.
     * Export that target via the Ref. The portData is the port number.
     * If portData is null, then use current LiveRef 'ref' if it is
     * non-null; otherwise assume port 0.
     */ 
    public RemoteStub exportObject(Remote impl, Object portData)
	throws RemoteException
    {
	int port = ((portData != null) ?
		    ((Integer)portData).intValue() :
		    0);

	if (ref == null) {
	    ref = new LiveRef(port);
	}

	RemoteStub stub = setSkeleton(impl);
	Target target = new Target(impl, (Dispatcher)this, stub);
	ref.exportObject(target);
	return stub;
    }

    /**
     * Return the hostname of the current client.  When called from a
     * thread actively handling a remote method invocation the
     * hostname of the client is returned.
     * @exception ServerNotActiveException If called outside of servicing
     * a remote method invocation.
     */
    public String getClientHost() throws ServerNotActiveException
    {
	Channel ch = TCPTransport.getCurrentChannel();
	if (ch instanceof TCPChannel) {
	    TCPEndpoint ep = (TCPEndpoint)((TCPChannel)ch).getEndpoint();
	    try {
		return ep.getHost();
	    } catch (Exception ex) {
	    }
	}
	throw new ServerNotActiveException("not in a remote call");
    }

    /**
     * setSkeleton discovers and sets the appropriate skeleton for the impl
     * and returns the appropriate stub for the impl.
     */
    public RemoteStub setSkeleton(Remote impl) throws RemoteException
    {
	skel = RemoteProxy.getSkeleton(impl);
	return RemoteProxy.getStub(impl, getClientRef());
    }
    
    /**
     * Call to dispatch to the remote object (on the server side).
     * The up-call to the server and the marshaling of return result
     * (or exception) should be handled before returning from this
     * method.
     * @param obj the target remote object for the call
     * @param call the "remote call" from which operation and
     * method arguments can be obtained.
     * @exception IOException If unable to marshal return result or
     * release input or output streams
     */
    public void dispatch(Remote obj, RemoteCall call)
	throws IOException
    {
	int op;			// operation number
	long hash;		// hash for matching stub with skeleton
	
	try {
	    // read remote call header
	    ObjectInput in;
	    try {
		in = call.getInputStream();
		op = in.readInt();
		hash = in.readLong();
	    } catch (Exception readEx) {
		throw new UnmarshalException("Error unmarshaling call header",
					     readEx);
	    }

	    // if calls are being logged, write out object id and operation
	    java.io.PrintStream callLog = RemoteServer.getLog();
	    if (callLog != null) {
		Operation[] ops = skel.getOperations();
		String opname = ((op >= 0 && op < ops.length) ?
				 ops[op].toString() : "<<method unknown>>");
		callLog.println(//"UnicastServerRef.dispatch(): " +
		    "[" + getClientHost() + ": " +
		    obj.getClass().getName() +
		    ref.getObjID().toString() + ": " +
		    opname + "]");
	    }
	    
	    unmarshalCustomCallData(in);
	    // dispatch to skeleton for remote object
	    skel.dispatch(obj, call, op, hash);
	    
	} catch (Throwable e) {
	    // if calls are being logged, write out exception
	    java.io.PrintStream callLog = RemoteServer.getLog();
	    if (callLog != null) {
		synchronized (callLog) {
		// include client host name if possible
		    try {
			callLog.print(
			    "[" + getClientHost() + "] ");
		    } catch (ServerNotActiveException ex) {
		    }
		    callLog.print("exception: ");
		    e.printStackTrace(callLog);
		}
	    }
	    
	    ObjectOutput out = call.getResultStream(false);
	    if (e instanceof RuntimeException) {
		e = new ServerRuntimeException("Server RuntimeException",
					       (Exception)e);
	    } else if (e instanceof Error ) {
		e = new ServerError("Server Error", (Error)e);
	    } else if (e instanceof RemoteException) {
		e = new ServerException("Server RemoteException",
					(Exception)e);
	    }
	    out.writeObject(e);
	} finally {
	    call.releaseInputStream(); // in case skeleton doesn't
	    call.releaseOutputStream();
	}
    }
    
    protected void unmarshalCustomCallData(ObjectInput in)
	throws IOException, ClassNotFoundException
    {}
    
    /**
     * Returns the class of the ref type to be serialized
     */
    public String getRefClass(ObjectOutput out)
    {
	if (out instanceof MarshalOutputStream) {
	    return "UnicastRef";
	} else {
	    return "UnicastServerRef";
	}
    }
    
    /**
     * Return the client remote reference for this remoteRef.
     * In the case of a client RemoteRef "this" is the answer.
     * For  a server remote reference, a client side one will have to
     * found or created.
     */
    protected RemoteRef getClientRef() {
	return new UnicastRef(ref);
    }

    /**
     * Write out external representation for remote ref.
     */
    public void writeExternal(ObjectOutput out) throws IOException 
    {
	if (out instanceof MarshalOutputStream) {
	    throw new Error("UnicastServerRef.writeExternal: internal error");
	}
    }

    /**
     * Read in external representation for remote ref.
     * @exception ClassNotFoundException If the class for an object
     * being restored cannot be found.
     */
    public void readExternal(ObjectInput in)
	throws IOException, ClassNotFoundException
    {
	// object is re-exported elsewhere (e.g., by UnicastRemoteObject)
	ref = null;
	skel = null;
    }
    
}
