/*
 * @(#)DGCClient.java	1.9 98/07/01
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.rmi.transport;

import java.util.*;
import java.rmi.*;
import java.rmi.dgc.DGC;
import java.rmi.dgc.Lease;
import java.rmi.dgc.VMID;
import java.rmi.server.ObjID;
import java.rmi.server.LogStream;

/**
 * The DGCClient embodies most of the client-side of distributed garbage
 * collection.  The countTable maps live references to a reference count.
 * The reference count is the number of local references exist for a
 * given remote object.
 * 
 * When a LiveRef is unmarshaled, it is entered into the countTable via a
 * call to "referenced".  This action increments the reference count of the
 * LiveRef, and if the reference is not already in the countTable, a
 * "dirty" call to the distributed collector at the server's
 * site when the reference is first added to the table.
 *
 * When a LiveRef is garbage collected locally, a call to
 * "unreferenced" must be made. This action will decrement the
 * reference count of the ref.  When the ref count becomes zero, a
 * "clean" call is made to the distributed collector at server's site.
 *
 * The client must ping the server site periodically to "renew" the
 * leases on remote references so that the client remains in a
 * server's remembered set. If a lease is not renewed before the lease
 * expires, the server can assume that the client is no longer
 * accessible and remove this client from its remembered set.
 *
 * @java.rmi.dgc.DGC, sun.rmi.transport.DGCImpl
 *
 * @author Ann Wollrath
 */
final class DGCClient implements Runnable {

    /** Lease request time for DGC; defaults to 10 minutes */
    private final static long leaseValue =
	getLong("java.rmi.dgc.leaseValue", 600000);

    private static Object lock   = new Object();

    // this should be a *private* method since it is privileged
    private static long getLong(String name, long def) {
        return Utils.getLong(name, def).longValue();
    }
        
    /** DGC table mapping LiveRef to CountTableEntry */
    private static Hashtable countTable = new Hashtable();
    /** Table mapping endpoints to list of refs containing that endpoint */
    private static Hashtable leaseTable = new Hashtable(); 
    /** Hashtable mapping endpoints to clean requests */
    private static Hashtable cleanTable = new Hashtable();
    /** List of references that need their CountTableEntry decremented */
    private static Vector unrefList = new Vector();

   /** 
    * Indicates whether leaseTable has changed so we can know whether
    * to look search for the next set of ref's to renew
    **/
    private static boolean leaseTableModified = false;


    /** This client's VMID */
    private static VMID vmid;

    static {
	if (VMID.isUnique())
	    vmid = new VMID();
	else
	    vmid = null;
    }

    /** next sequence number to use to order dirty and clean calls */
    private static long nextSequenceNum = Long.MIN_VALUE;

    /** 
     * Thread for sending asynchronous clean calls.  Please note:
     * This thread executes the DGCClient.run() method.
     **/
    private static Thread cleaner = null;
    /** Thread for sending asynchronous lease renewals */
    private static Thread leasee = null;

    private DGCClient() 
    {}

    /**
     * Increment reference counts associated with LiveRef objects.  Make dirty call
     * if first reference to ref.  Important!! This method assumes that all the
     * the LiveRef's in the array parameter have the same Endpoint.  This is not an
     * unreasonable assumption b/c they are passed this way from 
     * sun.rmi.transport.ConnectionInputStream.registerRefs()
     */
    static void referenced(LiveRef [] refArray) throws RemoteException {
	/* here we lock the whole table.  This only allows one dirty
	 * call at a time, but the synchronization is *really* tricky
	 * if we lock on an entry after getting it from the table
	 * (since the entry can be removed in the interim).
	 */

	synchronized (lock) {
            /**
             * get endpoint from any of the the LiveRef objects in refArray
             * and use it to figure out whether endpoint is in leaseTable.
             * if not, add a LeaseTableEntry to leaseTable.  Then start up
             * lease renewal and cleaner threads if they are null
             **/

            Endpoint refEp = refArray[0].getEndpoint();
	    if (DGCImpl.logLevel >= LogStream.VERBOSE)
		LogStream.log("dgc").println("DGCClient.referenced:endpoint =" + 
                                              refEp);

	    LeaseTableEntry leaseEntry = (LeaseTableEntry) leaseTable.get(refEp);
	    if (leaseEntry == null) {
                // add entry for the remote server's endpoint
		leaseEntry = new LeaseTableEntry(refEp);
		leaseTable.put(refEp, leaseEntry);
                    
	        if (DGCImpl.logLevel >= LogStream.VERBOSE)
		    LogStream.log("dgc").println("DGCClient.referenced: " +
		        "added Endpoint =" + refEp + "to the leaseTable" + 
			"LeaseTableEntry =" +leaseEntry);

                // If lease renewal thread isn't running, start it up
 	        // since leaseTable is no longer empty.
		if (leasee == null) {
		    leasee = RMIThread.newThread(new LeaseRenewer(),
						     "LeaseRenewer", true);
		    leasee.start();
		}

		// start up cleaner here too since it must run
		// finalization periodically
		if (cleaner == null) {
		    cleaner = RMIThread.newThread(new DGCClient(),
						      "Cleaner", true);
	            cleaner.start();
		}
	    }

	    if (DGCImpl.logLevel >= LogStream.VERBOSE)
		LogStream.log("dgc").println("DGCClient.referenced: " +
					     "leaseTableEntry" + leaseEntry);


            // Make a vector of Refs for which dirty calls
            // have not yet been made (some of the LiveRefs have already
            // had their objID's added to the leaseTable).
	    Vector dirtyRefList = new Vector();	    
	    ObjID  refId;
	    for (int i = 0; i < refArray.length; i++) {        
	        CountTableEntry countEntry = (CountTableEntry) countTable.get(refArray[i]);
	        if (countEntry == null) {
    	  	    // if entry doesn't exist; add entry to countTable
		    LiveRef newRef = (LiveRef) refArray[i].copy();
		    refId = newRef.getObjID();
		
		    countEntry = new CountTableEntry();
		    countTable.put(newRef, countEntry);

	            if (DGCImpl.logLevel >= LogStream.VERBOSE)
		        LogStream.log("dgc").println("DGCClient.referenced: " +
		            " Added LiveRef = " + refArray[i] + 
                            "ObjID = " + refId + " to the countTable. " + 
			    "CountTableEntry = " + countEntry);

                } else {
		    refId = refArray[i].getObjID();
                }           

		// If the ref's ObjID isn't in the leaseTable, then
                // add it and put it in list of ObjId's to be sent
 	 	// in next batch of dirty calls
		if (!leaseEntry.objIdList.contains(refId)) {
 		    leaseEntry.objIdList.addElement(refId);
		    dirtyRefList.addElement(refArray[i]);

		    if (DGCImpl.logLevel >= LogStream.VERBOSE)
		        LogStream.log("dgc").println("DGCClient.referenced: " +
                            "added ref = " + refArray[i] + 
                            " to the dirtyRefList" +
			    " and added ObjID = " + refId + " to objIdList");
                }
                
    	       	refArray[i].counted();
	    	countEntry.refCount++;

	        if (DGCImpl.logLevel >= LogStream.VERBOSE)
		    LogStream.log("dgc").println("DGCClient.referenced: " +
					         "leaseTableEntry" + leaseEntry);

            }

	    // if no dirty calls are required, return
            if(dirtyRefList.isEmpty()) {
	        if (DGCImpl.logLevel >= LogStream.VERBOSE)
		    LogStream.log("dgc").println("DGCClient.referenced: " +
		        "No dirty calls necessary -- returning.");
		return;
	    }

            // Copy Vector of ObjID's into an array of ObjID's and
            // send batched dirty call
	    ObjID [] objIdArray = getObjIdArray(dirtyRefList);

	    DGC dgc = null;
	    try {
	        dgc = LocateDGC.getDGC(refEp);
            } catch (RemoteException re) {
		//###what do I do here?
		//continue;  //##for now
	    }

	    try {
 		if (DGCImpl.logLevel >= LogStream.VERBOSE)
		    LogStream.log("dgc").println(
	               "DGCClient.referenced: making dirty call...");
			
		Lease lease      = dgc.dirty(objIdArray, getNextSequenceNum(),
				             new Lease(vmid, leaseValue));
		long currentTime = System.currentTimeMillis();
		vmid             = lease.getVMID();

	        if (DGCImpl.logLevel >= LogStream.VERBOSE)
		    LogStream.log("dgc").println("DGCClient.referenced: " +
		        "New lease -- vmid = " + lease.getVMID() + 
                        "duration = " + lease.getValue());
               
		setUpLeasing(leaseEntry, currentTime, lease);

	    } catch (RemoteException e) {
		// dirty call failed; schedule clean call
		if (DGCImpl.logLevel >= LogStream.BRIEF)
		    LogStream.log("dgc").println(
			"DGCClient.referenced: dirty call failed");
		scheduleClean(objIdArray, refEp, true);
                cleanUpCountTable(dirtyRefList);
		// ###do I want to do the below -- I think yes
		//if(leaseEntry.renewTime == 0) 
		//  leaseTable.remove(refEp);
		throw e;
            }
	}
    }


    /**
     * When this method is finished, the renewal time for a given endpoint
     * is calculated and the leasing thread is notified if the
     * renewal time needs to be recalculated.
     **/
    private static void setUpLeasing(LeaseTableEntry entry, 
                                     long currentTime, Lease lease) 
    {
        /* length of lease in entry -- undefined when entry.renewTime == 0 */
	long oldLength;   
	/* length of lease parameter */
        long newLength;   
        
        if (entry.renewTime == 0) {
	    oldLength = 0; 
	} else {
	    oldLength = entry.lease.getValue(); 
	}

	/* calc renewTime and then set lease for leaseTableEntry */
        newLength      = lease.getValue();
        long renewTime = currentTime + getRenewTime(newLength);
	entry.lease    = lease;

	if (DGCImpl.logLevel >= LogStream.VERBOSE)
	    LogStream.log("dgc").println("DGCClient.setUpLeasing: " +
	        "old lease length = " + oldLength + 
                " new lease length = " + newLength + 
                " current time = " + currentTime );

	/*
	 * if the entry.lease is shorter than it was, or if it's 
	 * a brand new LeaseTableEntry, notify the leasee thread.  This
	 * assumes that new leaseTableEntries and only new
	 * leaseTableEntries have renewTime == 0
	 */

	if ( (newLength < oldLength) || (entry.renewTime == 0) ) {
	    entry.renewTime = renewTime;
	    leaseTableModified = true;
	    lock.notifyAll();
	
	    if (DGCImpl.logLevel >= LogStream.VERBOSE) {
	        LogStream.log("dgc").println("DGCClient.setUpLeasing: " +
	            "notifyAll called on leasing thread.");
            }
	} else {
	    entry.renewTime = renewTime;
	}
    }


    /* 
     * This method takes the length of the new lease
     * and returns the amount of time the leasee should wait
     * before starting the renewal process.
     */
    private static long getRenewTime(long leaseLength) {        
        return (long) ( 0.5 * (double) leaseLength);
    }


    /*
     * Wrapper for real referenced method.  This provides backward
     * compatibility.  It _is_ called from a LiveRef constructor.  It
     * takes the LiveRef object passed as a parameter and passes it to
     * the real referenced method as an array of size one.
     */
    static void referenced(LiveRef ref) throws RemoteException 
    { 
	LiveRef [] refArray = new LiveRef[1];
	refArray[0]         = ref;
        referenced(refArray);
    }

   /*
    * Utility method.  This method removes all the refs passed in it's 
    * Vector parameter from the countTable
    * Please note: This method accesses the cleanTable without synchronizing 
    * on it so it should be called from a method within which countTable
    * is synchronized.
    *
    * After a call to schedule clean is made, the "cleaned" refs need to be 
    * removed from the countTable so this method should most likely be called
    * following, or within, that method.
    *
    * Please note:  This method is not usually called when only one ref is 
    * being removed from the countTable.  In that case, it is simply removed
    * via the class Hashtable remove method.
    */

    private static void cleanUpCountTable(Vector refList) {
	for (int i = 0; i < refList.size(); i++)
	    countTable.remove((LiveRef) refList.elementAt(i));
    }


    /*
     * Saves ref into the unrefList, and notifies the cleaner thread
     * that there are refs in the refTable.  This method is called by
     * LiveRef.finalize()
     * Please note:  Because this method is called from the LiveRef
     * finalizer (which can be called when memory needs to be freed
     * up) memory should not be allocated within blocks synchronized
     * on unrefList!!  The possible result is deadlock.  
     */
    static void unreferenced(LiveRef ref) {
	synchronized (unrefList) {
	    unrefList.addElement(ref);
	    unrefList.notifyAll();
	}
    }


    /**
     * Return next value in a monotonically increasing sequence.  This
     * function will return a greater value than the previous time
     * it was called (for 2^64 calls).
     */
    private synchronized static long getNextSequenceNum()
    {
	return nextSequenceNum++;
    }


    /**
     * Handle reference count decrements.  Each time unreferenced is
     * called on a LiveRef, the refCount of the ref needs to be 
     * decrement, and if the refCount reaches zero, the ref needs
     * to be removed from the countTable and the leaseTable
     */
    private static void handleDecrements() 
    {

        // I chose the length of this array rather arbitrarily.
        final int  ARRAY_SIZE = 128;
	LiveRef    ref        = null;
 	LiveRef [] refArray   = new LiveRef[ARRAY_SIZE];
	int        length     = 0;


	/* 
	 * This block copies up to ARRAY_SIZE elements of the  
	 * unreflist into the array refArray and then removes
	 * those elements from the unreflist.
	 * This may seem to be an overly complicated way to do
	 * this; however, it must be done this way to avoid
	 * allocating memory within the block synchronized on
	 * unreflist.
	 */
	synchronized (unrefList) {
	    length = unrefList.size();
	    if (length == 0) {
	        return;
	    } else if (length <= ARRAY_SIZE) {
	        unrefList.copyInto(refArray);
		unrefList.removeAllElements();
	    } else {
	        length = ARRAY_SIZE;
		int i = 0;
		for (i = 0; i < ARRAY_SIZE; i++) 
		    refArray[i] = (LiveRef) unrefList.elementAt(i);
		for (i = 0; i < ARRAY_SIZE; i++) 
		    unrefList.removeElementAt(0);
	    }
	}

	// decrement ref count for each of the refs in refArray. 
	// Remove them from the countTable and leaseTable, and
	// schedule a clean call if the refCount reaches zero
	synchronized (lock) {
	    for (int i = 0; i < length; i++) { 
	        ref = refArray[i];
		CountTableEntry countEntry = 
		    (CountTableEntry)countTable.get(ref);
		if (countEntry != null) {
		    Endpoint        refEp      = ref.getEndpoint();
		    ObjID           refId      = ref.getObjID();
		    --countEntry.refCount;
		    if (countEntry.refCount == 0) {
		      
			if (DGCImpl.logLevel >= LogStream.VERBOSE)
			    LogStream.log("dgc").println(
			        "DGCClient.handleDecrements: " +
                                "ObjID = " + ref.getObjID() + 
				" refCount == 0"); 
   
		        // Remove ObjID from leaseTable and if 
			// objIdTable is empty, remove entire
			// leaseTableEntry 
			LeaseTableEntry leaseEntry = 
			    (LeaseTableEntry) leaseTable.get(refEp);
			if (leaseEntry != null) {
			    leaseEntry.objIdList.removeElement(refId);
			    if (leaseEntry.objIdList.isEmpty()) {
			        leaseTable.remove(refEp);
				leaseTableModified = true;

				if (DGCImpl.logLevel >= LogStream.VERBOSE)
				    LogStream.log("dgc").println(
				        "DGCClient.handleDecrement: " +
					"Removing ep = " + refEp +
					" from the leaseTable.");
			    }
			}

			// reference count reached zero; make clean call
			// and take refCount out of countTable
			scheduleClean(ref, false);
			countTable.remove(ref);
		    }
		}    
	    }
	}
    }


    /**
     * Internal routine to schedule clean call.  
     **/
    private static void 
        scheduleClean(ObjID [] objIdArray, Endpoint ep, boolean strong) 
    {
        long  seqNum = getNextSequenceNum();

        if(objIdArray == null)
	    return;

	CleanRequest request = (CleanRequest) cleanTable.get(ep);
        if (request == null) {
	    request = new CleanRequest(objIdArray, ep, seqNum, strong);
	    cleanTable.put(ep, request);        
        } else {
            request.objIdArray = mergeObjIdArrays(request.objIdArray, objIdArray); 
	    request.strong     = strong;
	    /* What should be done about the sequence number? */
	}
    }

    /**
     *  This method merges two arrays of type ObjID.  
     **/

    private static ObjID [] 
        mergeObjIdArrays(ObjID [] array1, ObjID [] array2) 
    {
	ObjID [] newObjIdArray;

        if ((array1.length == 0) && (array2.length == 0)) {
	    newObjIdArray = null;
        } else if (array1.length == 0) {
	    newObjIdArray = array2;
	} else if (array2.length == 0) {
	    newObjIdArray = array1;
	} else {
            int newLength = array1.length + array2.length; 
	    newObjIdArray = new ObjID[newLength];
            System.arraycopy(array1,0, newObjIdArray, 0, array1.length);
            System.arraycopy(array2, 0, newObjIdArray, 
                             array1.length, array2.length);
	}
	return newObjIdArray;
    } 

    
    /* Wrapper to preserve logic of handleDecrements() */
    private static void scheduleClean(LiveRef ref, boolean strong) {
        Endpoint refEp      = ref.getEndpoint();
        ObjID id            = ref.getObjID();
        ObjID [] objIdArray = new ObjID[1];
	objIdArray[0]       = id;
	scheduleClean(objIdArray, refEp, strong);
    }

    /**
     * Send out pending clean calls.  
     */
    private static void sendCleans() {
        // Keep a list of entries to remove from cleanTable
        // so they can be removed _after_ the cleanTable enum 
        // is being iterated through 
	Vector      entriesToRemove = new Vector();
	Enumeration enum            = cleanTable.elements();
	try {
            // make a batched clean call for each entry in
	    // cleanTable.
	    while (enum.hasMoreElements()) {

		CleanRequest request = (CleanRequest)enum.nextElement();
                Endpoint     ep      = request.ep; 
		DGC          dgc     = null;

		// get server's remote collector
		try {
		    dgc = LocateDGC.getDGC(ep);
		} catch (RemoteException re) {
		    // ###what do I do here?
		    continue;
		}

		// make clean call
		try {
		    if (vmid != null) {

		        if (DGCImpl.logLevel >= LogStream.VERBOSE) {
			    LogStream.log("dgc").println(         
			         "DGCClient.sendCleans: making " +
                                 request.objIdArray.length + 
                                 " clean call(s) for Endpoint = " + ep);
                            for(int i = 0; i < request.objIdArray.length; i++)
			        LogStream.log("dgc").println("\tObjID = " + 
                                    request.objIdArray[i]);
			}

			dgc.clean(request.objIdArray, request.sequenceNum, 
                                  vmid, request.strong);
                        entriesToRemove.addElement(ep); 
		    }
		} catch (ConnectException ce) {
		    // target VM went away, forget clean request
		    //cleanTable.remove(ep);
		    entriesToRemove.addElement(ep);
		} catch (RemoteException re) {
		    // possible transient failure, retain clean request
		    if (DGCImpl.logLevel >= LogStream.VERBOSE)
			LogStream.log("dgc").println(
			    "DGCClient.sendCleans: clean call failed");
		}
	    }
	} catch (NoSuchElementException e) {
          //###should something be done here
        }
	
	// Now remove entries that had successful clean call returns
	enum = entriesToRemove.elements();
	while (enum.hasMoreElements()) {
	    Endpoint ep = (Endpoint)enum.nextElement();
	    cleanTable.remove(ep);
	}
    }  

    /**
     * Loop that handles reference count decrements and send of
     * pending clean calls.
     */
    public void run() 
    {
	while (true) {
	    // wait until either notify or timeout occurs
	    long interval = getLong("sun.rmi.transport.cleanInterval",180000);

            System.runFinalization();
	    
	    synchronized (unrefList) {
		if (unrefList.isEmpty()) {
		    try {
			unrefList.wait(interval);
		    }
		    catch (InterruptedException e) {
		    }
		}
	    }

	    handleDecrements();
	    sendCleans();
	}
    }
	    


    /**
     * This method manages the thread responsible lease renewal.  It
     * it schedules which leaseTableEntry should be renewed next, and
     * causes the thread to wait until it is time for lease renewal.
     **/
    public static boolean doRenewal() {
        LeaseTableEntry nextRenewal  = null;
	boolean         keepRenewing = true;
	long            currentTime, waitInterval, wakeUpTime;
	synchronized(lock) {
	    // loop while renewLeases() returns true.
	    do { 

                nextRenewal = getNextRenewal();
		if (nextRenewal == null) {
		    leasee = null;
		    return false;
		}

     	        if (DGCImpl.logLevel >= LogStream.BRIEF)
	           LogStream.log("dgc").println("DGCClient.doRenewal: " +
                       "next lease to renew has Endpoint = " +
			nextRenewal.ep + "renewTime = " +
			nextRenewal.renewTime);
	        
                // wait until leasing thread is awakened.  if the
		// leaseTable was modified, re-calculate nextRenewal
		// and waitTime, otherwise, if it's time, renewleases().   
		while (true)
                {
		    currentTime  = System.currentTimeMillis();
   		    waitInterval = nextRenewal.renewTime -  currentTime;
		    wakeUpTime   = currentTime + waitInterval;

		    if (waitInterval > 0) {
     	                if (DGCImpl.logLevel >= LogStream.BRIEF)
	                    LogStream.log("dgc").println(
                                "DGCClient.doRenewal:  " +  
		                "Waiting for " + waitInterval + 
                                "milliseconds");
		
		        try {
		            lock.wait(waitInterval);		    
		        } catch (InterruptedException e) {
		          //###should I do something here?
		        }
		    }		
		
	            currentTime = System.currentTimeMillis();
		    if (leaseTableModified) {
			nextRenewal = getNextRenewal();
			leaseTableModified = false;
			if (nextRenewal == null) {
			    leasee = null;
			    return false;
			}

     	                if (DGCImpl.logLevel >= LogStream.BRIEF)
	                    LogStream.log("dgc").println(
                                "DGCClient.doRenewal: " +
                                "next lease to renew has Endpoint = " +
			        nextRenewal.ep + "renewTime = " +
			        nextRenewal.renewTime);

		    } else if (currentTime >= wakeUpTime) {
	                break;
		    }
		}

	        if(nextRenewal != null)
                    keepRenewing = renewLeases(nextRenewal); 
	    } while(keepRenewing);

	    return keepRenewing;
	}
    }



    /**
     * This method returns the LeaseTableEntry to be 
     * renewed next.  This method assumes that leaseTable
     * is not empty.  
     */
    private static LeaseTableEntry getNextRenewal() {
	Enumeration enum      = leaseTable.elements();            
	long soonest          = Long.MAX_VALUE;
	LeaseTableEntry entry = null;

	while (enum.hasMoreElements()) {
	    LeaseTableEntry nextEntry = (LeaseTableEntry) enum.nextElement();
	    if (nextEntry.renewTime < soonest) {
		soonest = nextEntry.renewTime;
		entry   = nextEntry;
	    }
	}    

	return entry;
    }


    /**
     * Renew leases before they expire.
     *
     * @return if true, there are leases outstanding; otherwise leases
     * no longer need to be renewed.  Right now, this method only runs on
     * LeaseTableEntry at a time (the one with the soonest renewal time)
     */
    static boolean renewLeases(LeaseTableEntry leaseEntry) 
    {
        Endpoint ep         = leaseEntry.ep;
        DGC      dgc        = null;    
	ObjID[]  objIdArray = null;
        int      numObjIds  = 0;
       	try {
	    try {
	        dgc = LocateDGC.getDGC(ep);
            } catch (RemoteException re) {
	      //###what shoudl I do here?
	    }
	    // Copy ObjID vector into an array for dirty call
	    if (leaseEntry == null) {
	        String s = Thread.currentThread().getName();
	    }

            //&&&I don't think I need this synchronization, but what
	    //synchronication do I need?
	    synchronized (leaseEntry.objIdList) {
	        numObjIds  = leaseEntry.objIdList.size();
	        objIdArray = new ObjID[numObjIds];
                leaseEntry.objIdList.copyInto(objIdArray);
	    }

	    if (DGCImpl.logLevel >= LogStream.BRIEF)
		LogStream.log("dgc").println(
		    "DGCClient.renewLeases: Renewing " + numObjIds +
		    " leases for Endpoint = " + ep);

	    Lease lease      = dgc.dirty(objIdArray, getNextSequenceNum(),
				         new Lease(vmid, leaseValue));
	    vmid             = lease.getVMID();
	    long currentTime = System.currentTimeMillis();

	    if (DGCImpl.logLevel >= LogStream.VERBOSE) {
		LogStream.log("dgc").println("DGCClient.renew:eases: " +
		    "New lease -- vmid = " + lease.getVMID() + 
                    "duration = " + lease.getValue());
		
		LogStream.log("dgc").println("DGCClient.renewLeases: " +
		    "*Renewed the leases for the following ObjID's*");
                for (int i = 0; i < numObjIds; i++)
		    LogStream.log("dgc").println("DGCClient.renewLeases: " + 
		        "ObjID #" + i + " = " + objIdArray[i]);
            }
            setUpLeasing(leaseEntry, currentTime, lease);
	} catch (RemoteException e) {
	    // ###nyi: what do I do here? for now just forget about
	    // ###keeping track of these refs...
	    leaseTable.remove(ep);
            // ### shouldn't I also clean up countTable??
	}

        // If leaseTable is empty, turn off lease thread.
	if (leaseTable.isEmpty()) {
	    leasee = null;
	    return false;
	} else {
	    return true;
	}
    }
	
    /**
     * Utility method.  Keeps ugliness out of code. If this is used
     * elsewhere, it might need to be modified to check for empty
     * refList
     **/
    private static ObjID [] getObjIdArray(Vector refList) {
       ObjID [] objIdList = null;

       objIdList = new ObjID[refList.size()]; 
       for(int i = 0; i < refList.size(); i++)
	   objIdList[i] = ((LiveRef) refList.elementAt(i)).getObjID();

       return objIdList;
    }


    /**
     * The DGCClient.countTable is made up of object of this type.  Member
     * refCount refers to the number of references the client has to a given
     * LiveRef
     **/
    private static final class CountTableEntry {
        int refCount = 0;
    }


    /**
     * This Class holds the information for one entry into
     * DGCClient.cleanTable. There is one CleanRequest object per
     * endpoint. 
     **/
  
    private static final class CleanRequest {
    
        /**
         * This array points to an array of ObjID's that are ready to
	 * be cleaned. This is structure passed as a parameter in the
	 * clean call. 
         **/
        ObjID [] objIdArray;
        Endpoint ep;
        long sequenceNum;
        boolean strong;
    
        //###I'm not sure this is necessary
        public CleanRequest(LiveRef ref, long sequenceNum, boolean strong) {
            this.ep          = ref.getEndpoint();
            this.strong      = strong;
            this.sequenceNum = sequenceNum;
            ObjID [] array   = new ObjID[1];
            array[0]         = ref.getObjID();
 	    this.objIdArray  = array;
        }

        public CleanRequest(ObjID [] objIdArray, Endpoint ep, 
                            long sequenceNum, boolean strong) 
        {
	    this.sequenceNum = sequenceNum;
	    this.objIdArray  = objIdArray;
	    this.strong      = strong;
	    this.ep 	     = ep;
        }

        public boolean equals(Object obj) {
            if ((obj != null) && (obj instanceof CleanRequest)) {
                CleanRequest req = (CleanRequest)obj;
                if (req.objIdArray.length != objIdArray.length) {
		    return false;
	        } else if  ( ! ( (sequenceNum == req.sequenceNum) && 
                                 (strong == req.strong)           && 
                                 (req.ep.equals(ep)) ) ) 
                {
	            return false;
                }
	        for (int i = 0; i < objIdArray.length; i++) {  
                    if (!objIdArray[i].equals(req.objIdArray[i]))
	                return false;                        
                }
                return true;
	    }
	    return false;
        }
    } /* end of class CleanRequest */

  
    /**
     * Elements of the DGCClient.leaseTable are of 
     * this type and are keyed on endpoint.  
     **/
    private static final class LeaseTableEntry {

        /** Holds renewal time for endpoints */
        public long renewTime = 0;
    
        /** 
         * Endpoint for  all the LiveRef objects in the 
         * Vector refList
         **/
        public Endpoint ep = null;

        /**
         *  Vector holding list of objID's for endpoint 
         *  Please note:  endpoint is also key into leaseTable.
         **/
        public Vector objIdList = new Vector();

        /** 
         * Lease for the remote refs with ObjID's held in
         * the objIdList.  Please note that when there is only 
         * one Lease for all remote refs with the same endpoint
         **/
        public Lease lease = null;
  
        public LeaseTableEntry(Endpoint ep) {
	    this.ep = ep;
        }
    } /* end of class LeaseTableEntry */

    /**
     * This class implements the lease renewal thread
     **/
    private static class LeaseRenewer implements Runnable {
        public void run() {
            boolean keepRenewing;

	    if (DGCImpl.logLevel >= LogStream.BRIEF)
	        LogStream.log("dgc").println("LeaseRenewer.run: " +
                    "Lease renewing thread starting up.");

	    do {
                keepRenewing = doRenewal();
            } while (keepRenewing);
        }
    } /* end of class LeaseRenewer */

} //end of class DGCClient.java

    
