/*
 * @(#)LiveRef.java	1.12 98/07/01
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.rmi.transport;

import java.io.*;
import java.rmi.*;
import java.rmi.server.ObjID;
import sun.rmi.transport.tcp.TCPEndpoint;

public class LiveRef implements java.io.Externalizable {
    /** wire representation for the object*/
    private transient Endpoint ep;
    private transient ObjID id;

    /** connection service for the object */
    private transient Channel ch;

    /** flag to indicate whether this ref was reference counted */
    private transient boolean counted = false;

    /** flag to indicate whether this ref specifies a local server or
     * is a ref for a remote object (surrogate)
     */
    private transient boolean isLocal = false;

    /**
     * Construct a "well-known" live reference to a remote object.
     */
    private LiveRef(ObjID objID,
		    Endpoint endpoint,
		    Channel channel,
		    boolean isLocalServer) 
    {
	ep = endpoint;
	id = objID;
	ch = channel;
	isLocal = isLocalServer;
    }

    /**
     * Returns a shallow copy of the ref.
     */
    LiveRef copy() {
	return new LiveRef(id, ep, ch, isLocal);
    }
    
    /**
     * Construct a "well-known" live reference to a remote object
     * @param isLocalServer If true, indicates this ref specifies a local
     * server in this address space; if false, the ref is for a remote
     * object (hence a surrogate or proxy) in another address space.
     */
    public LiveRef(ObjID objID, Endpoint endpoint, boolean isLocalServer) 
    {
	this(objID, endpoint, null, isLocalServer);
    }

    /**
     * Construct a new (empty) live reference.
     */
    public LiveRef() {}

    /**
     * Construct a new live reference for a server object in the local
     * address space.
     */
    public LiveRef(int port)
    {
	this((new ObjID()), port);
    }

    /**
     * Return the port number associated with this ref.
     */
    public int getPort() 
    {
	return ((TCPEndpoint)ep).getPort();
    }
    
    /**
     * Construct a new live reference for a "well-known" server object
     * in the local address space.
     */
    public LiveRef(ObjID objID, int port) 
    {
	this(objID, TCPEndpoint.getLocalEndpoint(port), true);
    }

    /**
     * Export the object to accept incoming calls.
     */
    public void exportObject(Target target)
	throws RemoteException
    {
	ep.exportObject(id, target);
    }
    
    public Channel getChannel() throws RemoteException {
	if (ch == null) {
	    ch = ep.getChannel();
	}
	
	return ch;
    }

    public ObjID getObjID() 
    {
	return id;
    }

    Endpoint getEndpoint()
    {
	return ep;
    }

    void counted() 
    {
	counted = true;
    }

    public String toString() 
    {
	String type;
	
	if (isLocal)
	    type = "local";
	else
	    type = "remote";
	return "[endpoint:" + ep + "(" + type + ")," +
	    "objID:" + id + "]";
    }

    public int hashCode() 
    {
	return id.hashCode();
    }

    public boolean equals(Object obj) 
    {
	if (obj instanceof LiveRef) {
	    LiveRef ref = (LiveRef)obj;
	    return (ep.equals(ref.ep) && id.equals(ref.id) &&
		    isLocal == ref.isLocal);
	} else {
	    return false;
	}
    }
    
    public boolean remoteEquals(Object obj) 
    {
	if (obj instanceof LiveRef) {
	    LiveRef ref = (LiveRef)obj;
	    return (ep.equals(ref.ep) && id.equals(ref.id));
	} else {
	    return false;
	}
    }
    
    public void writeExternal(ObjectOutput out) throws IOException
    {
	boolean isResultStream = false;

	if (out instanceof ConnectionOutputStream) {
	    ConnectionOutputStream stream = (ConnectionOutputStream)out;
	    if (isLocal) {
		// make sure impl is pinned before sending out reference
		Target target = ObjectTable.getTarget(id);
		if (target != null)
		    target.pinImpl();
		else
		    throw new IOException("attempt to marshal unexported " +
					  "object reference: " + this);
	    }
	    isResultStream = stream.isResultStream();
	    if (isResultStream) {
		// save liveRef until an ack is received.
		stream.saveRef(this);
	    }
	}
	// All together now write out the endpoint, id, and flag
	
	((TCPEndpoint) ep).write(out);
	id.write(out);
	out.writeBoolean(isResultStream);
    }
    
    public void readExternal(ObjectInput in)
	throws IOException, ClassNotFoundException
    {
	// Now read in the endpoint, id, and result flag
	ep =  TCPEndpoint.read(in);
	id = ObjID.read(in);
	boolean isResultStream = in.readBoolean();

	ch = null;
	isLocal = false;
	counted = false;
	if (in instanceof ConnectionInputStream) {
	    ConnectionInputStream stream = (ConnectionInputStream)in;
	    // save ref to send "dirty" call after all args/returns
	    // have been unmarshaled.
	    stream.saveRef(this);
	    if (isResultStream) {
		// set flag in stream indicating that remote objects were
		// unmarshaled.  A DGC ack should be sent by the transport.
		stream.setAckNeeded();
	    }
	} else {
	    DGCClient.referenced(this);
	}
    }

    public void finalize() throws Throwable
    {
	if (isLocal == false) {
	    if (!id.equals(DGCImpl.getID()) && counted) {
		DGCClient.unreferenced(this);
	    }
	}
    }
}
