/*
 * @(#)ObjectTable.java	1.8 98/07/01
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.rmi.transport;

import java.rmi.Remote;
import java.rmi.dgc.VMID;
import java.rmi.server.ExportException;
import java.rmi.server.LogStream;
import java.rmi.server.ObjID;
import java.util.*;


/**
 * Object table shared by all implementors of the Transport interface.
 * This table maps object ids to remote object targets in this address
 * space.
 *
 * @author Ann Wollrath
 */
public final class ObjectTable {
    /** Table that maps ObjID to Target */
    private static Hashtable objTable = new Hashtable();
    /** Table that maps ImplKey to Target */
    private static Hashtable implTable = new Hashtable();
    /** WeakRef wrapper object (re)used for lookup operations */
    private static WeakRef lookupKey = new WeakRef();
    /** The reaper gets rid of table entries of impls that are no
     * longer referenced. */
    private static Reaper reaper = null;
    /** Thread to keep process alive while there are exported objects */
    private static Thread keepAlive = null;
    /** constructor is private to avoid instantiation */
    private ObjectTable()
    {}
    
    /**
     * Returns the target associated with the object id.
     */
    static Target getTarget(ObjID id) {
	// this is already synchronized on objTable
	return (Target)objTable.get(id);
    }

    /**
     * Returns the target associated with the remote object
     */
    public static Target getTarget(Remote impl) {
	synchronized (implTable) {
	    try {
		lookupKey.setThing(impl);
		return (Target)implTable.get(lookupKey);
	    } finally {
		lookupKey.setThing(null);
	    }
	}
    }

    /**
     * Adds the target to both the object and impl tables.
     */
    static synchronized void putTarget(ObjID id, Target target)
	throws java.rmi.server.ExportException
    {
	if (DGCImpl.logLevel >= LogStream.VERBOSE)
	    LogStream.log("dgc").println(
		"ObjectTable.putTarget: add object " + id +
		", target = " + target);
	
	WeakRef weakImpl = target.getWeakImpl();

	if (objTable.containsKey(id) || implTable.containsKey(weakImpl))
	    throw new ExportException("Object ID already in use");
	objTable.put(id, target);
	implTable.put(weakImpl, target);
	if (reaper == null) {
	    reaper = new Reaper();
	    Thread t = RMIThread.newThread(reaper, "Reaper", true);
	    t.start();
	}
	if (keepAlive == null && !target.isPermanent()) {
	    keepAlive = RMIThread.newThread(new KeepAlive(), "KeepAlive",
					    false);
	    keepAlive.start();
	}
    }
    
    static synchronized void referenced(ObjID id, long sequenceNum, VMID vmid)
    {
	Target target = getTarget(id);
	// add client to dirty set.
	if (target != null)
	    target.referenced(sequenceNum, vmid);
    }

    static synchronized void unreferenced(ObjID id, long sequenceNum, VMID vmid,
					  boolean strong)
    {
	Target target = getTarget(id);
	// remove  client to dirty set.
	if (target != null)
	    target.unreferenced(sequenceNum, vmid, strong);
    }
    
    /**
     * Scan the object table looking for targets that are unpinned
     * and the weak reference is null, having been cleared by the
     * Java VM garbage collector.
     * If so the target should be removed from both the impl and
     * object tables.
     */
    static void reap() {
	Enumeration enum = (Enumeration)objTable.keys();
	while (enum.hasMoreElements()) {
	    ObjID id = (ObjID)enum.nextElement();
	    Target t = getTarget(id);
	
	    if (t == null)
		continue;

	    // Get the weak reference and check it
	    WeakRef weakimpl = t.getWeakImpl();
	    if (weakimpl.get() == null) {
		if (DGCImpl.logLevel >= LogStream.VERBOSE)
		    LogStream.log("dgc").println(
			"ObjectTable.reap: remove object " + id);
		    
		if (!t.isEmpty()) {
		    throw new Error("Reaping an object with references");
		}

		// impl has been garbage collected; remove entry
		implTable.remove(weakimpl);
		// also remove object table entry
		objTable.remove(id);
	    }
	}
	checkKeepAlive();
    }

    /**
     * Check to see whether the keepAlive should continue to execute.
     * If the object table is populated with only permanent objects,
     * then the keepAlive thread should be stopped.
     */
    private static synchronized void checkKeepAlive() 
    {
	if (keepAlive != null) {
	    Enumeration enum = (Enumeration)objTable.elements();
	    while (enum.hasMoreElements()) {
		Target t = (Target)enum.nextElement();
		if (!t.isPermanent())
		    return;
	    }
	    keepAlive.stop();
	    keepAlive = null;
	}
    }
}

/**
 * Nondaemon thread that keeps process alive
 */
class KeepAlive implements Runnable {
    public void run() 
    {
	while (true) {
	    try {
		// The following argument is used because a bug in the
		// native C code for Thread.sleep() causes it to return
		// immediately for any higher value.
		Thread.sleep(Integer.MAX_VALUE - 1);
	    } catch (InterruptedException e) {
	    }
	}
    }
}
    
/**
 * The reaper runs in a thread that calls the object table's reap
 * routine every so often. Defaults to reaping every 3 seconds.
 */
class Reaper implements Runnable {
    public void run() {
	while (true) {
	    try {
		int interval = Utils.getLong("sun.rmi.transport.reapInterval",
					     3000).intValue();
		Thread.sleep(interval);
	    } catch (InterruptedException e) {
	    }
	    ObjectTable.reap();
	}
    }
}
