/*
 * @(#)Target.java	1.8 98/07/01
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.rmi.transport;

import java.util.*;
import java.rmi.Remote;
import java.rmi.dgc.VMID;
import java.rmi.server.LogStream;
import java.rmi.server.RemoteServer;
import java.rmi.server.RemoteStub;
import java.rmi.server.Unreferenced;
import sun.rmi.server.Dispatcher;

/**
 * A target contains information pertaining to a remote object that
 * resides in this address space.  Targets are located via the
 * ObjectTable.
 */
public final class Target {
    /** flag indicating whether target is subject to collection */
    private boolean permanent = false;
    /** weak reference to remote object implementation */
    private WeakRef weakImpl;
    /** dispatcher for remote object */
    private Dispatcher disp;
    /** stub for remote object */
    private RemoteStub stub;
    /** set of clients that hold references to this target */
    private Vector refSet = new Vector();
    /** table that maps client endpoints to sequence numbers */
    private Hashtable sequenceTable = new Hashtable(5);
    
    /**
     * Construct a Target for the remote object "impl".
     */
    public Target(Remote impl, Dispatcher disp, RemoteStub stub)
    {
	this.weakImpl = new WeakRef(impl);
	this.disp = disp;
	this.stub = stub;
    }

    /** Get the stub (proxy) object for this target
     */
    public RemoteStub getStub() {
	return stub;
    }

    /**
     * Get the weak reference for the Impl of this target.
     */
    WeakRef getWeakImpl() {
	return weakImpl;
    }

    /**
     * Returns the dispatcher for this remote object target.
     */
    Dispatcher getDispatcher()
    {
	return disp;
    }

    /**
     * Get the impl for this target.
     * Note: this may return null if the impl has been garbage collected.
     * (currently, there is no need to make this method public)
     */
    synchronized Remote getImpl() {
	return (Remote)weakImpl.get();
    }

    /**
     * Returns true if the target is permanent.
     */
    boolean isPermanent() 
    {
	return permanent;
    }

    /**
     * Pin impl in target. Pin the WeakRef object so it holds a strong
     * reference to the object to it will not be garbage collected locally.
     * This way there is a single object responsible for the weak ref 
     * mechanism.
     */
    synchronized void pinImpl() {
	weakImpl.pin();
    }

    /**
     * Unpin impl in target.  Weaken the reference to impl so that it
     * can be garbage collected locally. But only if there the refSet
     * is empty.  All of the weak/strong handling is in WeakRef
     */
    synchronized void unpinImpl() {
	/* only unpin if:
	 * a) impl is not permanent, and
	 * b) impl is not already unpinned, and
	 * c) there are no external references (outside this
	 *    address space) for the impl
	 */
	if (!permanent && refSet.isEmpty()) {
	    weakImpl.unpin();
	}
    }

    /**
     * If "on" is true, then the impl is pinned permanently (the impl
     * will not be collected via distributed and/or local GC).  If
     * "on" is false, than the impl is subject to collection. Permanent
     * objects do not keep a server from exiting.
     */
    public synchronized void setPermanent(boolean on) {
	permanent = on;
	if (permanent) {
	    pinImpl();
	} else {
	    /* try to unpin; will only succeed if refSet is empty */
	    unpinImpl();
	}
    }


    /**
     * Add an endpoint to the remembered set.  Also adds a notifier
     * to call back if the address space associated with the endpoint
     * dies.
     */
    synchronized void referenced(long sequenceNum, VMID vmid)
    {
	// check sequence number for vmid
	SequenceEntry entry = (SequenceEntry) sequenceTable.get(vmid);
	if (entry == null) {
	    sequenceTable.put(vmid, new SequenceEntry(sequenceNum));
	} else if (entry.sequenceNum < sequenceNum) {
	    entry.update(sequenceNum);
	} else  {
	    // late dirty call; ignore.
	    return;
	}

	if (DGCImpl.logLevel >= LogStream.VERBOSE)
	    LogStream.log("dgc").println(
		"Target.referenced: add to dirty set: " + vmid);

	// the impl should already be pinned at this point.
	if (refSet.contains(vmid))
	    return;
	else
	    refSet.addElement(vmid);

	DGCImpl.getDGCImpl().registerTarget(vmid, this);
    }

    /**
     * Remove endpoint from remembered set.  If set becomes empty,
     * remove server from Transport's object table.
     */
    synchronized void unreferenced(long sequenceNum, VMID vmid, boolean strong)
    {
	// check sequence number for vmid
	SequenceEntry entry = (SequenceEntry) sequenceTable.get(vmid);
	if (entry == null || entry.sequenceNum > sequenceNum) {
	    // late clean call; ignore
	    return;
	} else if (strong) {
	    // strong clean call; retain sequenceNum
	    entry.retain(sequenceNum);
	} else if (entry.keep == false) {
	    // get rid of sequence number
	    sequenceTable.remove(vmid);
	}

	if (DGCImpl.logLevel >= LogStream.VERBOSE)
	    LogStream.log("dgc").println(
		"Target.unreferenced: remove from dirty set: " + vmid);

	refSetRemove(vmid);
    }

    /**
     * Remove endpoint from the reference set.
     */
    synchronized private void refSetRemove(VMID vmid) {
	// remove notification request
	DGCImpl.getDGCImpl().unregisterTarget(vmid, this);

	if (refSet.removeElement(vmid) && refSet.isEmpty()) {
	    // reference set is empty, so server can be garbage collected.
	    // remove object from table.
	    if (DGCImpl.logLevel >= LogStream.VERBOSE)
		LogStream.log("dgc").println(
		    "Target.unreferenced: " +
		    "reference set is empty: target = " + this);

	    unpinImpl();
	    
	    // invoke unreferenced callback (if supported) on the server
	    // so that the server can do some cleanup if neccessary.
	    Remote obj = getImpl();
	    if (obj instanceof Unreferenced) {
		// spawn thread to do callback.
		new UnreferencedObj((Unreferenced)obj);
	    }
	}
    }
    
    /**
     * Returns true if remembered set is empty; otherwise returns
     * false
     */
    boolean isEmpty() 
    {
	return refSet.isEmpty();
    }

    /**
     * This method is called if the address space associated with the
     * vmid dies.  In that case, the vmid should be removed
     * from the reference set.
     */
    synchronized public void vmidDead(VMID vmid) 
    {
	if (DGCImpl.logLevel >= LogStream.BRIEF)
	    LogStream.log("dgc").println(
		"Target.notify: removing endpoint " +
		vmid + " from reference set");

	sequenceTable.remove(vmid);
	refSetRemove(vmid);
    }
}

class SequenceEntry {
    long sequenceNum;
    boolean keep;

    SequenceEntry(long sequenceNum) {
	this.sequenceNum = sequenceNum;
	keep = false;
    }

    void retain(long sequenceNum) {
	this.sequenceNum = sequenceNum;
	keep = true;
    }

    void update(long sequenceNum) 
    {
	this.sequenceNum = sequenceNum;
    }
}

/**
 * This class is used to spawn a thread to asychronously call the
 * "unreferenced" method when the reference set for an impl becomes
 * empty.
 */
class UnreferencedObj implements Runnable {
    Unreferenced obj;
    private static int num = 0;
    
    UnreferencedObj(Unreferenced obj) {
	this.obj = obj;
	(RMIThread.newThread(this, "Unreferenced-" + ++num, true)).start();
    }

    public void run()
    {
	obj.unreferenced();
    }
}


