/*
 * @(#)Transport.java	1.9 98/07/01
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.rmi.transport;

import java.io.IOException;
import java.io.ObjectOutput;
import java.util.Hashtable;
import java.rmi.MarshalException;
import java.rmi.NoSuchObjectException;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.server.LogStream;
import java.rmi.server.ObjID;
import java.rmi.server.RemoteCall;
import java.rmi.server.RemoteServer;
import java.rmi.server.ServerNotActiveException;
import sun.rmi.server.Dispatcher;

/**
 * Transport abstraction for enabling communication between different
 * VMs.
 *
 * @author Ann Wollrath
 */
public abstract class Transport {

    /** "transport" package log level */
    static int logLevel = LogStream.parseLevel(
	Utils.getProperty("sun.rmi.transport.logLevel"));
    
    /**
     * Returns a <I>Channel</I> that generates connections to the
     * endpoint <I>ep</I>. A Channel is an object that creates and
     * manages connections of a particular type to some particular
     * address space.
     * @param ep the endpoint to which connections will be generated.
     * @return the channel or null if the transport cannot
     * generate connections to this endpoint
     */
    public abstract Channel getChannel(Endpoint ep);

    /**
     * Returns an endpoint for this address space.
     */
    public abstract Endpoint thisEndpoint();

    /**
     * Export the object so that it can accept incoming calls.
     */
    public void exportObject(ObjID id, Target target)
	throws RemoteException
    {
	ObjectTable.putTarget(id, target);
    }
    
    /**
     * Service an incoming remote call. When a message arrives on the
     * connection indicating the beginning of a remote call, the
     * threads are required to call the <I>serviceCall</I> method of
     * their transport.  The default implementation of this method
     * locates and calls the dispatcher object.  Ordinarily a
     * transport implementation will not need to override this method.
     * At the entry to <I>tr.serviceCall(conn)</I>, the connection's
     * input stream is positioned at the start of the incoming
     * message.  The <I>serviceCall</I> method processes the incoming
     * remote invocation and sends the result on the connection's
     * output stream.  If it returns "true", then the remote
     * invocation was processed without error and the transport can
     * cache the connection.  If it returns "false", a protocol error
     * occurred during the call, and the transport should destroy the
     * connection.
     */
    public boolean serviceCall(RemoteCall call) {
	Dispatcher disp;
	Remote impl;

	try {
	    /* read object id */
	    ObjID id;

	    if (Transport.logLevel >= LogStream.VERBOSE)
		LogStream.log("transport").println(
		    "Transport.serviceCall: waiting to read object id");
	    try {
		id = ObjID.read(call.getInputStream());
	    } catch (java.io.IOException e) {
		throw new MarshalException("unable to read objID", e);
	    }
	    
	    /* get the remote object */
	    if (Transport.logLevel >= LogStream.BRIEF)
		LogStream.log("transport").println(
		    "Transport.serviceCall: get target " + id);
	    Target target = ObjectTable.getTarget(id);
	    if (target == null || (impl = target.getImpl()) == null) 
		throw new NoSuchObjectException("no such object in table");
	    disp = target.getDispatcher();

	}
	catch (RemoteException e) {
	    // if calls are being logged, write out exception
	    java.io.PrintStream callLog = RemoteServer.getLog();
	    if (callLog != null) {
		synchronized (callLog) {
		    // include client host name if possible
		    try {
			callLog.print("[" + RemoteServer.getClientHost() +
				      "] ");
		    } catch (ServerNotActiveException ex) {
		    }
		    callLog.print("exception: ");
		    e.printStackTrace(callLog);
		}
	    }

	    /* we haven't started marshalling stuff yet, return
	     * the exception to the client.
	     */
	    try {
		if (Transport.logLevel >= LogStream.BRIEF) {
		    LogStream ls = LogStream.log("transport");
		    ls.println("Transport.serviceCall: exception: ");
		    e.printStackTrace(ls);
		}

		ObjectOutput out = call.getResultStream(false);
		out.writeObject(e);
		call.releaseOutputStream();
		
	    } catch (IOException ie) {
		return false;
	    }
	    return true;
	}

	try {
	    // call the dispatcher
	    if (Transport.logLevel >= LogStream.VERBOSE)
		LogStream.log("transport").println(
		    "Transport.serviceCall: call dispatcher");

	    disp.dispatch(impl, call);
	} catch (IOException e) {
	    return false;
	}
	return true;
    }
}
