/*
 * @(#)WeakRef.java	1.7 98/07/01
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.rmi.transport;

import java.rmi.server.LogStream;

/**
 * A WeakRef is a weak reference (in the Java sense) to an Object.
 * The referenced can be made stronger using the pin method and weak
 * again with the unpin method. Instances of this class may be used as
 * keys or values in a hash table.
 */
public class WeakRef extends sun.misc.Ref {

    /** value of the referent's "identity" hash code */
    private int hashValue;

    /** hard reference to the reference, for when this WeakRef is "pinned" */
    private Object hardref = null;

    /**
     * Construct a weak reference to the specified object.
     */
    public WeakRef(Object obj) {
	/*
	 * Cache referent's "identity" hash code (so that we still have it
	 * after it gets cleared).  We cannot use the value from the object's
	 * hashCode() method, since if the object is of a remote class not
	 * extended from RemoteObject and it is trying to implement hashCode()
	 * and equals() so that it can be compared to stub objects, its own
	 * hash code could not have been initialized yet (see bugid 4102938).
	 */
	setHashValue(obj);
	setThing(obj);
    }

    /**
     * Construct a weak reference object with no referent set.
     */
    public WeakRef() {
	setHashValue(null);
    }
    
    /**
     * Pin, turns this weak ref into a strong ref by keeping another reference.
     * Only if the hardref is null should it be set to the current value.
     */
    public synchronized void pin() {
	if (hardref == null) {
	    if (DGCImpl.logLevel >= LogStream.VERBOSE)
		LogStream.log("dgc").println(
		    "WeakRef.pin: ref = " + this);

	    hardref = check();
	}
    }

    /**
     * Unpin, turns a strong ref into a weak reference.
     * If the softref has been null'ed, restore its value from hardref.
     */
    public synchronized void unpin() {
	if (check() == null && hardref != null) {
	    if (DGCImpl.logLevel >= LogStream.VERBOSE)
		LogStream.log("dgc").println(
		    "WeakRef.unpin: ref = " + this);

	    setThing(hardref);
	}
	hardref = null;
    }

    /*
     * reconstitute restores the value of the soft ref to the hard ref
     * in the case where the gc has cleared it.
     */
    public Object reconstitute() {
	return hardref;
    }

    /**
     * setThing must update the saved hashvalue.
     * Don't use setThing when these are inserted in a hashtable
     * because the hashvalue is not allowed to change.
     */
    public synchronized void setThing(Object thing) {
	setHashValue(thing);
	super.setThing(thing);

	// If this ref is pinned, pin the new object
	if (hardref != null)
	    hardref = thing;
    }

    /**
     * Set value of this WeakRef's hash code based on the "identity"
     * hash code of the given object.  See comment in constructor for
     * more explanation.
     */
    private void setHashValue(Object obj) {
	if (obj != null) {
	    hashValue = System.identityHashCode(obj);
	} else {
	    hashValue = 0;
	}
    }

    /**
     * Always return the "identity" hash code for the original referent.
     */
    public int hashCode() 
    {
	return hashValue;
    }
	    
    /**
     * Return true if "obj" is this identical WeakRef object, or, if the
     * contained reference has not been cleared, if "obj" is another WeakRef
     * object with the identical non-null referent.  Otherwise, return false.
     */
    public boolean equals(Object obj) {
	if (obj instanceof WeakRef) {
	    if (obj == this)
		return true;

	    Object referent = get();
	    return (referent != null) && (referent == ((WeakRef) obj).get());
	} else {
	    return false;
	}
    }
}

