/*
 * @(#)RMIMasterSocketFactory.java	1.10 98/07/01
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.rmi.transport.proxy;

import java.io.*;
import java.net.Socket;
import java.net.ServerSocket;
import java.net.NoRouteToHostException;
import java.net.UnknownHostException;
import java.util.Hashtable;
import java.util.Vector;
import java.rmi.server.LogStream;
import java.rmi.server.RMISocketFactory;
import sun.rmi.transport.RMIThread;
import sun.rmi.transport.Utils;

/**
 * RMIMasterSocketFactory attempts to create a socket connection to the
 * specified host using successively less efficient mechanisms
 * until one succeeds.  If the host is successfully connected to,
 * the factory for the successful mechanism is stored in an internal
 * hash table keyed by the host name, so that future attempts to
 * connect to the same host will automatically use the same
 * mechanism.
 */
public class RMIMasterSocketFactory extends RMISocketFactory {

    /** "proxy" package log level */
    static int logLevel = LogStream.parseLevel(
	Utils.getProperty("sun.rmi.transport.proxy.logLevel"));

    /** timeout for attemping direct socket connections */
    private static long connectTimeout =
	Utils.getLong("sun.rmi.transport.proxy.connectTimeout",
		      15000).longValue();	// default: 15 seconds

    /** table of hosts successfully connected to and the factory used */
    Hashtable successTable = new Hashtable();

    /** maximum number of hosts to remember successful connection to */
    private static final int MaxRememberedHosts = 64;

    /** list of the hosts in successTable in initial connection order */
    private Vector hostList = new Vector(MaxRememberedHosts);

    /** default factory to initally use for direct socket connection */
    protected RMISocketFactory initialFactory = new RMIDirectSocketFactory();

    /** ordered list of factories to try as alternate connection
      * mechanisms if a direct socket connections fails */
    protected Vector altFactoryList;

    /**
     * Create a RMIMasterSocketFactory object.  Establish order of
     * connection mechanisms to attempt on createSocket, if a direct
     * socket connection fails.
     */
    public RMIMasterSocketFactory()
    {
	altFactoryList = new Vector(2);
	boolean setFactories = false;
	
	try {
	    String proxyHost = System.getProperty("http.proxyHost");
	    if (proxyHost == null)
		proxyHost = System.getProperty("proxyHost");

	    if (!Boolean.getBoolean("java.rmi.server.disableHttp") &&
		(proxyHost != null && proxyHost.length() > 0)) {
		setFactories = true;
	    }
	}
	catch (Exception e) {
	    // unable to obtain the properties, so assume default behavior.
	    setFactories = true;
	}

	if (setFactories) {
	    altFactoryList.addElement(new RMIHttpToPortSocketFactory());
	    altFactoryList.addElement(new RMIHttpToCGISocketFactory());
	}
    }

    /**
     * Create a new client socket.  If we remember connecting to this host
     * successfully before, then use the same factory again.  Otherwise,
     * try using a direct socket connection and then the alternate factories
     * in the order specified in altFactoryList.
     */
    public Socket createSocket(String host, int port)
	throws IOException
    {
	if (RMIMasterSocketFactory.logLevel >= LogStream.BRIEF)
	    LogStream.log("transport").println(
		"RMIMasterSocketFactory.createSocket: " +
		"host: " + host + ", port: " + port);

	RMISocketFactory factory;

	// If we remember successfully connecting to this host before,
	// use the same factory.
	factory = (RMISocketFactory) successTable.get(host);
	if (factory != null) {
	    if (RMIMasterSocketFactory.logLevel >= LogStream.BRIEF)
		LogStream.log("transport").println(
		    "RMIMasterSocketFactory.createSocket: " +
		    "previously successful factory found: " + factory);
	    return factory.createSocket(host, port);
	}

	// Next, try a direct socket connection.  Open socket in another
	// thread and only wait for specified timeout, in case the socket
	// would otherwise spend minutes trying an unreachable host.
	Socket initialSocket = null;
	Socket fallbackSocket = null;
	AsyncConnector connector =
	    new AsyncConnector(this, initialFactory, host, port);
	IOException initialFailure = null;

	try {
	    synchronized (connector) {
		Thread connectorThread =
		    RMIThread.newThread(connector, "AsyncConnector", true);
		connectorThread.start();
		try {
		    connector.wait(connectTimeout);
		} catch (InterruptedException e) {
		}
		initialSocket = checkConnector(connector);
	    }

	    // assume no route to host (for now) if no connection yet
	    if (initialSocket == null)
		throw new NoRouteToHostException(
		    "connect timed out: " + host);

	    if (RMIMasterSocketFactory.logLevel >= LogStream.BRIEF)
		LogStream.log("transport").println(
		    "RMIMasterSocketFactory.createSocket: " +
		    "direct socket connection successful");

	    return initialSocket;

	} catch (UnknownHostException e) {
	    initialFailure = e;
	} catch (NoRouteToHostException e) {
	    initialFailure = e;
	} finally {
	    if (initialFailure != null) {

		if (RMIMasterSocketFactory.logLevel >= LogStream.BRIEF)
		    LogStream.log("transport").println(
			"RMIMasterSocketFactory.createSocket: " +
			"direct socket connection failed: " +
			initialFailure);

		// Finally, try any alternate connection mechanisms.
		for (int i = 0; i < altFactoryList.size(); ++ i) {
		    factory = (RMISocketFactory) altFactoryList.elementAt(i);
		    try {
			if (RMIMasterSocketFactory.logLevel >= LogStream.BRIEF)
			    LogStream.log("transport").println(
				"RMIMasterSocketFactory.createSocket: " +
				"trying with factory: " + factory);

			// For HTTP connections, the output (POST request) must
			// be sent before we verify a successful connection.
			// So, sacrifice a socket for the sake of testing...
			// The following sequence should verify a successful
			// HTTP connection if no IOException is thrown.
			Socket testSocket = factory.createSocket(host, port);
			InputStream in = testSocket.getInputStream();
			int b = in.read(); // probably -1 for EOF...
			testSocket.close();
		    } catch (IOException ex) {
			if (RMIMasterSocketFactory.logLevel >= LogStream.BRIEF)
			    LogStream.log("transport").println(
				"RMIMasterSocketFactory.createSocket: " +
				"factory failed: " + ex);

			continue;
		    }
		    if (RMIMasterSocketFactory.logLevel >= LogStream.BRIEF)
			LogStream.log("transport").println(
			    "RMIMasterSocketFactory.createSocket: " +
			    "factory succeeded");

		    // factory succeeded, open new socket for caller's use
		    try {
			fallbackSocket = factory.createSocket(host, port);
		    } catch (IOException ex) {	// if it fails 2nd time,
		    }				// just give up
		    break;
		}
	    }
	}

	synchronized (successTable) {
	    try {
		// check once again to see if direct connection succeeded
		synchronized (connector) {
		    initialSocket = checkConnector(connector);
		}
		if (initialSocket != null) {
		    // if we had made another one as well, clean it up...
		    if (fallbackSocket != null)
			fallbackSocket.close();
		    return initialSocket;
		}
		// if connector ever does get socket, it won't be used
		connector.notUsed();
	    } catch (UnknownHostException e) {
		initialFailure = e;
	    } catch (NoRouteToHostException e) {
		initialFailure = e;
	    }
	    // if we had found an alternate mechanism, go and use it
	    if (fallbackSocket != null) {
		// remember this successful host/factory pair
		rememberFactory(host, factory);
		return fallbackSocket;
	    }
	    throw initialFailure;
	}
    }

    /**
     * Remember a successful factory for connecting to host.
     * Currently, excess hosts are removed from the remembered list
     * using a Least Recently Created strategy.
     */
    void rememberFactory(String host, RMISocketFactory factory)
    {
	synchronized (successTable) {
	    while (hostList.size() >= MaxRememberedHosts) {
		successTable.remove(hostList.elementAt(0));
		hostList.removeElementAt(0);
	    }
	    hostList.addElement(host);
	    successTable.put(host, factory);
	}
    }

    /**
     * Check if an AsyncConnector succeeded.  If not, return socket
     * given to fall back to.
     */
    Socket checkConnector(AsyncConnector connector)
	throws IOException
    {
	IOException e = connector.getException();
	if (e != null) {
	    e.fillInStackTrace();
	    throw e;
	}
	return connector.getSocket();
    }

    /**
     * Create a new server socket.
     */
    public ServerSocket createServerSocket(int port) throws IOException
    {
	return new HttpAwareServerSocket(port);
    }
}

/**
 * AsyncConnector is used by RMIMasterSocketFactory to attempt socket
 * connections on a separate thread.  This allows RMIMasterSocketFactory
 * to control how long it will wait for the connection to succeed.
 */
class AsyncConnector implements Runnable {

    /** the RMIMasterSocketFactory that created this */
    protected RMIMasterSocketFactory master;

    /** what factory to use to attempt connection */
    protected RMISocketFactory factory;

    /** the host to connect to */
    protected String host;

    /** the port to connect to */
    protected int port;

    /** IOException that occurred during connection, if any */
    protected IOException exception = null;

    /** the connected socket, if successful */
    protected Socket socket = null;

    /** socket should be closed after created, if ever */
    boolean cleanUp = false;

    /**
     * Create a new asynchronous connector object.
     */
    AsyncConnector(RMIMasterSocketFactory master, RMISocketFactory factory,
                   String host, int port)
    {
	this.master  = master;
	this.factory = factory;
	this.host    = host;
	this.port    = port;

	SecurityManager security = System.getSecurityManager();
	if (security != null) {
	    security.checkConnect(host, port);
	}
    }

    /**
     * Attempt socket connection in separate thread.  If successful,
     * notify master waiting, 
     */
    public void run()
    {
	try {
	    Socket temp = factory.createSocket(host, port);
	    synchronized (this) {
		socket = temp;
		notify();
	    }
	    master.rememberFactory(host, factory);
	    synchronized (this) {
		if (cleanUp)
		    try {
			socket.close();
		    } catch (IOException e) {
		    }
	    }
	} catch (IOException e) {
	    synchronized (this) {
		exception = e;
		notify();
	    }
	}
    }

    /**
     * Get exception that occurred during connection attempt, if any.
     */
    synchronized IOException getException()
    {
	return exception;
    }

    /**
     * Get successful socket, if any.
     */
    synchronized Socket getSocket()
    {
	return socket;
    }

    /**
     * Note that this connector's socket, if ever successfully created,
     * will not be used, so it should be cleaned up quickly
     */
    synchronized void notUsed()
    {
	if (socket != null)
	    try {
		socket.close();
	    } catch (IOException e) {
	    }
	cleanUp = true;
    }
}
