/*
 * @(#)TCPChannel.java	1.16 98/07/01
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.rmi.transport.tcp;

import java.io.*;
import java.net.Socket;
import java.util.*;
import java.rmi.*;
import java.rmi.server.ExportException;
import java.rmi.server.LogStream;
import java.rmi.server.RMISocketFactory;
import sun.rmi.transport.Channel;
import sun.rmi.transport.Connection;
import sun.rmi.transport.Endpoint;
import sun.rmi.transport.Notifiable;
import sun.rmi.transport.RMIThread;
import sun.rmi.transport.Transport;
import sun.rmi.transport.TransportConstants;
import sun.rmi.transport.Utils;

class InEntry {

    InputStream in = null;
    boolean available = false;

    InEntry()
    {}
    
    InEntry(InputStream in) 
    {
	this.in = in;
    }
}

/**
 * TCPChannel is the socket-based implementation of the RMI Channel
 * abstraction.
 *
 * @author Ann Wollrath
 */
public class TCPChannel implements Channel, Runnable {
    /** endpoint for this channel */
    private TCPEndpoint ep;
    /** transport for this channel */
    private TCPTransport tr;
    /** objects to notify if the channel becomes dead */
    private Vector notifyList = new Vector();
    /** list of cached connections */
    private Vector freeList = new Vector();
    /** frees cached connections that have expired */
    private Reaper reaper = null;
    
    /** using multiplexer (for bi-directional applet communication */
    private boolean usingMultiplexer = false;
    /** connection multiplexer, if used */
    private ConnectionMultiplexer multiplexer = null;
    /** connection acceptor (should be in TCPTransport) */
    private ConnectionAcceptor acceptor;
    
    private static long timeout =
    	Utils.getLong("sun.rmi.transport.connectionTimeout",
		      15000).longValue();

    /**
     * Create channel for endpoint.
     */
    TCPChannel(TCPTransport tr, TCPEndpoint ep) {
	this.tr = tr;
	this.ep = ep;
    }

    /**
     * Return the endpoint for this channel.
     */
    public Endpoint getEndpoint() {
	return ep;
    }

    /**
     * Generates a new connection to the endpoint of the address space
     * for which this is a channel.
     */
    public Connection newConnection() throws RemoteException
    {
	TCPConnection conn;
	boolean isNewConnection = false;


	// try to get a free connection 
	synchronized (freeList) {
	    if (freeList.size() > 0) {
		SecurityManager security = System.getSecurityManager();
		if (security != null) {
		    security.checkConnect(ep.getHost(), ep.getPort());
		}

		conn = (TCPConnection)freeList.lastElement();
		freeList.removeElementAt(freeList.size()-1);

		if (TCPTransport.logLevel >= LogStream.BRIEF)
		    LogStream.log("tcp").println(
			"TCPChannel.newConnection: reuse connection");

		return conn;
	    }
	}

	// none free, so create a new connection
	if (TCPTransport.logLevel >= LogStream.BRIEF)
	    LogStream.log("tcp").println(
		"TCPChannel.newConnection: create connection");

	if (!usingMultiplexer) {
	    Socket sock = openSocket();
	    conn = new TCPConnection(this, sock);

	    try {
		DataOutputStream out =
		    new DataOutputStream(conn.getOutputStream());

		// choose protocol (single op if not reusable socket)
		if (!conn.isReusable()) {
		    out.writeByte(TransportConstants.SingleOpProtocol);
		} else {
		    out.writeByte(TransportConstants.StreamProtocol);
		    out.flush();
		    DataInputStream in =
			new DataInputStream(conn.getInputStream());
		    byte ack = in.readByte();
		    if (ack != TransportConstants.ProtocolAck)
			throw new MarshalException(
			"Transport protocol not supported by server");

		    TCPEndpoint endpoint;
		    endpoint = TCPEndpoint.read(in);
		    if (TCPTransport.logLevel >= LogStream.VERBOSE)
			LogStream.log("tcp").println(
			    "TCPChannel.newConnection: " +
			    "server suggested endpoint " + endpoint);

		    // set local host name, if unknown
		    TCPEndpoint.setLocalHost(endpoint.getHost());
		    // do NOT set the default port, because we don't
		    // know if we can't listen YET...

		    // write out default endpoint to match protocol
		    // (but it serves no purpose)
		    endpoint = TCPEndpoint.getLocalEndpoint(0);
		    endpoint.write(out);
		    if (TCPTransport.logLevel >= LogStream.VERBOSE)
			LogStream.log("tcp").println(
			    "TCPChannel.newConnection: " +
			    "using endpoint " + endpoint);
		    out.flush();
		}
	    } catch (IOException e) {
		if (e instanceof RemoteException)
		    throw (RemoteException) e;
		else
		    throw new MarshalException(
			"Error marshaling transport header", e);
	    }
	} else {
	    try {
		conn = multiplexer.openConnection();
	    } catch (IOException e) {
		synchronized (this) {
		    usingMultiplexer = false;
		    multiplexer = null;
		}
		throw new ConnectIOException(
		    "Error creating multiplexed connection", e);
	    }
	}
	return conn;
    }

    /**
     * Free the connection generated by this channel.
     * @param conn The connection
     * @param reuse If true, the connection is in a state in which it
     *        can be reused for another method call.
     */
    public void free(Connection conn, boolean reuse)
    {
	if (conn == null) return;

	if (reuse && conn.isReusable()) {
	    if (TCPTransport.logLevel >= LogStream.BRIEF)
		LogStream.log("tcp").println(
		    "TCPChannel.free: reuse connection");

	    /*
	     * Cache connection; if reaper for expired connection
	     * doesn't exist, then create one.
	     */
	    synchronized (freeList) {
		freeList.addElement(conn);
		if (reaper == null) {
		    if (TCPTransport.logLevel >= LogStream.BRIEF)
			LogStream.log("tcp").println(
		    		"TCPChannel.free: create reaper");
		    reaper = new Reaper();
		    Thread t = RMIThread.newThread(reaper,
						   "ConnectionExpiration-" +
						   ep.getHost() + ":" +
						   ep.getPort(), true);
		    t.start();
		}
	    }
	    ((TCPConnection)conn).setExpiration(System.currentTimeMillis() +
						timeout);
	} else { 
	    if (TCPTransport.logLevel >= LogStream.BRIEF)
		LogStream.log("tcp").println(
		    "TCPChannel.free: close connection");

	    try {
		conn.close();
	    } catch (java.io.IOException e) {
		// eat exception
	    }
	}
    }

    /**
     * Open socket connection to endpoint and send transport header.
     */
    private Socket openSocket() throws RemoteException
    {
	if (TCPTransport.logLevel >= LogStream.VERBOSE)
	    LogStream.log("tcp").println(
		"TCPChannel.openSocket: opening socket to " +
		ep.getHost() + ":" + ep.getPort());

	Socket sock;

	RMISocketFactory factory = RMISocketFactory.getSocketFactory();
	if (factory == null)
	    factory = TCPTransport.defaultSocketFactory;
	try {
	    sock = factory.createSocket(ep.getHost(), ep.getPort());
	} catch (java.net.UnknownHostException e) {
	    throw new UnknownHostException("Unknown host: " + ep, e);
	} catch (java.net.ConnectException e) {
	    throw new ConnectException("Connection refused to host: " + ep, e);
	} catch (IOException e) {
	    throw new ConnectIOException("Error creating connection to: " + ep,
					 e);
	}

	// set socket to disable Nagle's algorithm (always send immediately)
	try {
	    sock.setTcpNoDelay(true);
	} catch (Exception e) {
	    // if we fail to set this, ignore and proceed anyway
	}

	try {
	    InputStream sockIn = sock.getInputStream();
	    OutputStream sockOut = sock.getOutputStream();

	    // write out transport header
	    DataOutputStream out =
		new DataOutputStream(sockOut);
	    out.writeInt(TransportConstants.Magic);
	    out.writeShort(TransportConstants.Version);
	} catch (IOException e) {
	    throw new MarshalException("Error marshaling transport header", e);
	}
	return sock;
    }

    /**
     * Open multiplexed connection to server so that it can use it
     * to open connections back to here.
     */
    private ConnectionMultiplexer openMultiplexer()
	throws RemoteException
    {
	Socket sock = openSocket();
	TCPConnection conn = new TCPConnection(this, sock);

	ConnectionMultiplexer mult;	
	try {
	    DataOutputStream out =
		new DataOutputStream(sock.getOutputStream());

	    // choose multiplexing protocol (fail if not reusable)
	    if (!conn.isReusable()) {
		throw new ConnectIOException(
		    "Cannot open multiplexed connection to " + ep);
	    } else {
		out.writeByte(TransportConstants.MultiplexProtocol);
		out.flush();
		DataInputStream in =
		    new DataInputStream(sock.getInputStream());
		byte ack = in.readByte();
		if (ack != TransportConstants.ProtocolAck) {
		    throw new MarshalException(
			"Transport protocol not supported by receiver");
		}

		TCPEndpoint endpoint;
		endpoint = TCPEndpoint.read(in);
		if (TCPTransport.logLevel >= LogStream.VERBOSE)
		    LogStream.log("tcp").println(
			"TCPTransport(" + endpoint.getPort() + ").run: " +
			"server suggested endpoint " + endpoint);

		// set default endpoint info, if not already
		TCPEndpoint.setLocalHost(endpoint.getHost());
		TCPEndpoint.setDefaultPort(endpoint.getPort());

		endpoint = TCPEndpoint.getLocalEndpoint(0);
		endpoint.write(out);
		if (TCPTransport.logLevel >= LogStream.VERBOSE)
		    LogStream.log("tcp").println(
			"TCPTransport(" + endpoint.getPort() + ").run: " +
			"using endpoint " + endpoint);
		out.flush();

		mult = new ConnectionMultiplexer(this, sock.getInputStream(),
		    sock.getOutputStream(), true);
	    }
	} catch (IOException e) {
	    if (e instanceof RemoteException)
		throw (RemoteException) e;
	    else
		throw new MarshalException(
		    "Error marshaling transport header", e);
	}
	return mult;
    }

    /**
     * Execute thread that deals with multiplexed connections.
     */
    public void run()
    {
	try {
	    multiplexer.run();
	} catch (IOException e) {
	    if (TCPTransport.logLevel >= LogStream.VERBOSE) {
		LogStream ls = LogStream.log("tcp");
		synchronized (ls) {
		    ls.print("exception occurred in multiplexer: ");
		    e.printStackTrace(ls);
		}
	    }
	    multiplexer = null;

	    // try once to reopen multiplex connection??
	    try {
		haveMultiplexer();
	    } catch (RemoteException ex) {
	    }
	}
    }

    /**
     * Make sure that this endpoint has a multiplexed connection to its
     * endpoint to allow back communication.
     */
    void haveMultiplexer() throws RemoteException
    {
	if (multiplexer == null) {
	    if (TCPTransport.logLevel >= LogStream.VERBOSE)
		LogStream.log("tcp").println(
		    "TCPChannel.haveMultiplexer(): " +
		    "attempting to open multiplex connection");

	    multiplexer = openMultiplexer();
	    Thread t = RMIThread.newThread(this,
		"Multiplexer-" + ep.getHost() + ":" + ep.getPort(), true);
	    t.start();
	}
    }

    /**
     * Use given connection multiplexer object to obtain new connections
     * through this channel.
     */
    synchronized void useMultiplexer(ConnectionMultiplexer newMultiplexer)
    {
	// for now, always just use the last one given
	multiplexer = newMultiplexer;

	usingMultiplexer = true;
    }

    /**
     * Accept a connection provided over a multiplexed channel.
     */
    void acceptMultiplexConnection(Connection conn)
    {
	if (acceptor == null) {
	    acceptor = new ConnectionAcceptor(tr);
	    acceptor.startNewAcceptor();
	}
	acceptor.accept(conn);
    }


    /**
     * Remove each cached connection whose time out has expired.
     */
    private boolean freeCachedConnections() 
    {
	long time = System.currentTimeMillis();
	synchronized (freeList) {
	    Enumeration enum = freeList.elements();
	    while (enum.hasMoreElements()) {
		TCPConnection conn = (TCPConnection)enum.nextElement();
		if (conn.expired(time)) {
		    if (TCPTransport.logLevel >= LogStream.BRIEF)
			LogStream.log("tcp").println(
 				"TCPChannel.freeCachedConnections: " +
				"connection timeout expired");
		    try {
			conn.close();
		    } catch (java.io.IOException e) {
			// eat exception
		    }
		    freeList.removeElement(conn);
		}
	    }
	    
	    if (freeList.isEmpty()) {
		reaper = null;
		return false;
	    } else {
		return true;
	    }
	}
    }

    /**
     * Reaps expired cached connections; thread exits when freelist
     * is empty.
     */
    private class Reaper implements Runnable {
	public void run()
	{
	    do {
		try {
		    Thread.sleep(timeout);
		} catch (InterruptedException e) {
		}
		if (TCPTransport.logLevel >= LogStream.VERBOSE)
		    LogStream.log("tcp").println(
		      "TCPChannel.Reaper: " +
		      "wake up");
	    } while (freeCachedConnections());
		if (TCPTransport.logLevel >= LogStream.VERBOSE)
		    LogStream.log("tcp").println(
		      "TCPChannel.Reaper: " +
		      "exit");
	}
    }
    
    /**
     * Informs this channel that its endpoint is "dead". The
     * transport is responsible for monitoring the liveness of
     * channels (e.g., by pinging the endpoint of the channel) and
     * informing channels when they are "dead".  A channel is "dead"
     * when the program (address space) to which it is connected
     * exits, crashes, or cannot be reached for some period of time.
     * @param state Either Channel.Dead or Channel.Failed.
     */
    public void dead(int state) 
    {
	if (state == Channel.Dead) {
	    // get rid of all connections
	    synchronized (freeList) {
		Enumeration enum = freeList.elements();
		while (enum.hasMoreElements()) {
		    Connection conn = (Connection)enum.nextElement();
		    try {
			conn.close();
		    } catch (java.io.IOException e) {
			// eat exception
		    }
		}
		freeList.removeAllElements();
	    }
	}

	// create new notify list to allow new registrations during
	// notification...
	Vector oldNotifyList;
	
	synchronized (notifyList) {
	    oldNotifyList = notifyList;
	    notifyList = new Vector();
	}
	
	// inform those objects wanting notification...
	Enumeration enum = oldNotifyList.elements();
	while (enum.hasMoreElements()) {
	    Notifiable obj = (Notifiable)enum.nextElement();
	    obj.notify(ep, state);
	}
	oldNotifyList = null;
    }

    /**
     * Register interest in receiving a callback when this channel
     * becomes inaccessible.
     */
    public void addNotifiable(Endpoint ep, Notifiable obj) 
    {
	synchronized (notifyList) {
	    if (!notifyList.contains(obj))
		notifyList.addElement(obj);
	}
    }

    /**
     * Remove notification request.
     */
    public void removeNotifiable(Endpoint ep, Notifiable obj)
    {
	notifyList.removeElement(obj);
    }
}

/**
 * ConnectionAcceptor manages accepting new connections and giving them
 * to TCPTransport's message handler on new threads.
 *
 * Since this object only needs to know which transport to give new
 * connections to, it doesn't need to be per-channel as currently
 * implemented.
 */
class ConnectionAcceptor implements Runnable {

    /** transport that will handle message on accepted connections */
    private TCPTransport transport;

    /** queue of connections to be accepted */
    private Vector queue = new Vector(4);

    /** thread ID counter */
    private static int threadNum = 0;

    /**
     * Create a new ConnectionAcceptor that will give connections
     * to the specified transport on a new thread.
     */
    public ConnectionAcceptor(TCPTransport transport)
    {
	this.transport = transport;
    }

    /**
     * Start a new thread to accept connections.
     */
    public void startNewAcceptor()
    {
	Thread t = RMIThread.newThread(this, 
				       "Multiplex Accept-" + ++ threadNum,
				       true);
	t.start();
    }

    /**
     * Add connection to queue of connections to be accepted.
     */
    public void accept(Connection conn)
    {
	synchronized (queue) {
	    queue.addElement(conn);
	    queue.notify();
	}
    }

    /**
     * Give transport next accepted conection, when available.
     */
    public void run()
    {
	Connection conn;

	synchronized (queue) {
	    while (queue.size() == 0) {
		try {
		    queue.wait();
		} catch (InterruptedException e) {
		}
	    }
	    startNewAcceptor();
	    conn = (Connection) queue.elementAt(0);
	    queue.removeElementAt(0);
	}

	transport.handleMessages(conn, true);
    }
}
