/*
 * @(#)TCPConnection.java	1.11 98/07/01
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.rmi.transport.tcp;

import java.io.*;
import java.net.InetAddress;
import java.net.Socket;
import java.rmi.*;
import java.rmi.server.LogStream;
import java.rmi.server.RMISocketFactory;
import sun.rmi.transport.*;
import sun.rmi.transport.proxy.*;

public class TCPConnection implements Connection {

    private Socket socket;
    private Channel channel;
    private InputStream in = null;
    private OutputStream out = null;
    private long expiration = Long.MAX_VALUE;

    /**
     * Constructor used by subclass when underlying input and output streams
     * are already available.
     */
    public TCPConnection(TCPChannel ch, InputStream in, OutputStream out)
    {
	socket   = null;
	channel  = ch;
	this.in  = in;
	this.out = out;
    }

    /**
     * Constructor used for initiating a new output connection
     * (call, ping, etc).
     */
    public TCPConnection(Channel ch, TCPEndpoint ep) throws RemoteException
    {	
	channel = ch;
	RMISocketFactory factory = RMISocketFactory.getSocketFactory();
	if (factory == null)
	    factory = TCPTransport.defaultSocketFactory;
	try {
	    socket = factory.createSocket(ep.getHost(), ep.getPort());
	} catch (java.net.UnknownHostException e) {
	    throw new UnknownHostException("Unknown host: " + ep, e);
	} catch (java.net.ConnectException e) {
	    throw new ConnectException("Connection refused to host: " + ep, e);
	} catch (IOException e) {
	    throw new ConnectIOException("Error creating connection to: " + ep,
					 e);
	}

	// set socket to disable Nagle's algorithm (always send immediately)
	try {
	    socket.setTcpNoDelay(true);
	} catch (Exception e) {
	    // if we fail to set this, ignore and proceed anyway
	}
    }

    /**
     * Constructor used for creating a connection to accept call
     * (an input connection).
     */
    public TCPConnection(TCPChannel ch, Socket s) 
    {
	socket = s;
	channel = ch;
    }
    
    /**
     * Gets the output stream for this connection
     */
    public OutputStream getOutputStream() throws IOException
    {
	if (out == null)
	    out = new BufferedOutputStream(socket.getOutputStream());
	return out;
    }
    
    /**
     * Release the output stream for this connection.
     */
    public void releaseOutputStream() throws IOException
    {
	if (out != null)
	    out.flush();
    }

    /**
     * Gets the input stream for this connection.
     */
    public InputStream getInputStream() throws IOException
    {
	if (in == null)
	    in = new BufferedInputStream(socket.getInputStream());
	return in;
    }
    

    /**
     * Release the input stream for this connection.
     */
    public void releaseInputStream()
    {
    }

    /**
     * Determine if this connection can be used for multiple operations.
     * If the socket implements RMISocketInfo, then we can query it about
     * this; otherwise, assume that it does provide a full-duplex
     * persistent connection like java.net.Socket.
     */
    public boolean isReusable()
    {
	if ((socket != null) && (socket instanceof RMISocketInfo))
	    return ((RMISocketInfo) socket).isReusable();
	else
	    return true;
    }

    /**
     * Set the expiration time of this connection.
     * @param time The time at which the time out expires.
     */
    void setExpiration(long time)
    {
	expiration = time;
    }

    /**
     * Returns true if the timeout has expired on this connection;
     * otherwise returns false.
     * @param time The current time.
     */
    boolean expired(long time) 
    {
	return expiration <= time;
    }

    /**
     * Close the connection.
     */
    public void close() throws IOException
    {
	if (TCPTransport.logLevel >= LogStream.BRIEF)
	    LogStream.log("tcp").println(
		"TCPConnection.close: close connection");

	if (socket != null)
	    socket.close();
	else {
	    in.close();
	    out.close();
	}
    }

    /**
     * Returns the channel for this connection.
     */
    public Channel getChannel() 
    {
	return channel;
    }
} 
