/*
 * @(#)TCPEndpoint.java	1.12 98/07/07
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.rmi.transport.tcp;

import java.io.*;
import java.net.InetAddress;
import java.util.*;
import java.rmi.RemoteException;
import java.rmi.RMISecurityException;
import java.rmi.server.LogStream;
import java.rmi.server.ObjID;
import java.rmi.server.RemoteServer;
import sun.rmi.transport.Channel;
import sun.rmi.transport.Endpoint;
import sun.rmi.transport.RMIThread;
import sun.rmi.transport.Target;
import sun.rmi.transport.Transport;
import sun.rmi.transport.Utils;

/**
 * TCPEndpoint represents some communication endpoint for an address
 * space (VM).
 *
 * @author Ann Wollrath
 */
public class TCPEndpoint implements Endpoint, java.io.Serializable {
    /** IP address or host name */
    private String host;
    /** port number */
    private int port;

    /** maps port numbers to transports */
    private static Hashtable transportTable = new Hashtable(2);

    /** the local host name */
    private static String localHost;
    /** true if real local host name is known yet */
    private static boolean localHostKnown;

    /**
     * Find host name of local machine.  Property "java.rmi.server.hostname"
     * is used if set, so server administrator can compensate for the possible
     * inablility to get fully qualified host name from VM.
     */
    static {
	localHostKnown = true;
	localHost = Utils.getProperty("java.rmi.server.hostname");
	// could try querying CGI program here?
	if (localHost == null) {
    	    try {
		InetAddress localAddr;

		// debug option to simulate restricted applet behavior
		// property "sun.rmi.transport.tcp.simulateApplet":
		//	>= 1	can only get "localhost" for host name
		//	>= 2	listen on socket throws security exception
		if (Utils.getLong(
		    "sun.rmi.transport.tcp.simulateApplet", 0).intValue() >= 1) {
		    localAddr = InetAddress.getByName("");
		} else {
		    localAddr = InetAddress.getLocalHost();
		}
		byte[] raw = localAddr.getAddress();
		if ((raw[0] == 127) &&
		    (raw[1] ==   0) &&
		    (raw[2] ==   0) &&
		    (raw[3] ==   1)) {
		    localHostKnown = false;
		}

		/* if the user wishes to use a fully qualified domain
		 * name then attempt to find one.  
		 */
		if (Utils.getBoolean("java.rmi.server.useLocalHostName")) {
		    localHost = FQDN.attemptFQDN(localAddr);
		} else {
		    /* default to using ip addresses, names will 
		     * work across seperate domains. 
		     */
		    localHost = localAddr.getHostAddress();
		}
	    } catch (Exception e) {
		localHostKnown = false;
		localHost = null;
	    }
	}

	if (TCPTransport.logLevel >= LogStream.BRIEF)
	    LogStream.log("tcp").println(
		"TCPEndpoint.<clinit>: " +
		"localHostKnown = " + localHostKnown +
		", localHost = " + localHost);
    }

    /** default endpoint for this address space */
    private static TCPEndpoint defaultEndpoint =
	new TCPEndpoint(localHost, 0);

    /**
     * Create an endpoint for a specified host and port.
     */
    public TCPEndpoint(String host, int port) 
    {
	if (host == null)
	    host = "";
	this.host = host;
	this.port = port;
    }

    /**
     * Get an endpoint for the local address space on specified port.
     * If port number is 0, it returns shared default endpoint object
     * whose host name and port may or may not have been determined.
     */
    public static TCPEndpoint getLocalEndpoint(int port)
    {
	if (port == 0)
	    return defaultEndpoint;
	else
	    return new TCPEndpoint(localHost, port);
    }

    /**
     * Set the local host name, if currently unknown.
     */
    static void setLocalHost(String host)
    {
	if (!localHostKnown) {
	    defaultEndpoint.host = host;
	    localHost = host;
	    localHostKnown = true;

	    if (TCPTransport.logLevel >= LogStream.BRIEF)
		LogStream.log("tcp").println(
		    "TCPEndpoint.setLocalHost: set local host to " + host);
	}
    }

    /**
     * Set the port of the (shared) default endpoint object.
     * When first created, it contains port 0 because the transport
     * hasn't tried to listen to get assigned a port, or if listening
     * failed, a port hasn't been assigned from the server.
     */
    static void setDefaultPort(int port)
    {
	// do nothing if it has already been set
	if (defaultEndpoint.port != 0)
	    return;

	synchronized (transportTable) {
	    Integer oldPort = new Integer(defaultEndpoint.port);
	    Transport tr = (Transport) transportTable.get(oldPort);
	    if (tr != null) {
		transportTable.remove(oldPort);
		transportTable.put(new Integer(port), tr);
	    }
	    defaultEndpoint.port = port;

	    if (TCPTransport.logLevel >= LogStream.BRIEF)
		LogStream.log("tcp").println(
		    "TCPEndpoint.setDefaultPort: set default port to " + port);
	}
    }

    /**
     * Returns transport for remote endpoint.
     */
    public Transport getTransport() 
    {
	return getTransport(false);
    }
    
    /**
     * Returns the transport for this endpoint. 
     * @param local If true, indicates endpoint is for local server,
     * otherwise the endpoint is for a remote server and the default
     * transport (at port 0) should be used.
     */
    private Transport getTransport(boolean local)
    {
	Transport tr = null;

	synchronized (transportTable) {
	    if (local) {
		// set to default port (may still be zero, though)
//		if (port == 0)
//		    port = TCPTransport.DefaultPort;
		Integer portNum = new Integer(port);
		tr = (Transport)transportTable.get(portNum);
		if (tr == null) {
		    /* create transport */
		    tr = new TCPTransport(port);
		    transportTable.put(portNum, tr);
		}
	    } else {
		Integer defaultPort = new Integer(TCPTransport.DefaultPort);
		tr = (Transport)transportTable.get(defaultPort);
		if (tr == null)
		    tr = defaultEndpoint.getTransport(true);
	    }
	}
	
	return tr;
    }

    /**
     * Export the object to accept incoming calls.
     */
    public void exportObject(ObjID id, Target target)
	throws RemoteException
    {
	Transport tr =  getTransport(true);
	tr.exportObject(id, target);
    }

    /**
     * Returns a channel for this (remote) endpoint.
     */
    public Channel getChannel() 
    {
	Transport tr = getTransport(false);
	return tr.getChannel(this);
    }

    /**
     * Returns address for endpoint
     */
    public String getHost() {
	return host;
    }

    /**
     * Returns the port for this endpoint.  Note that the port may be
     * zero which indicates the default port for the VM for which this
     * is an endpoint.
     */
    public int getPort() {
	return port;
    }

    /**
     * Return string representation for endpoint.
     */
    public String toString() {
	return "[" + host + ":" + port + "]";
    }

    public int hashCode() 
    {
	return port;
    }

    public boolean equals(Object obj) 
    {
	if ((obj != null) && (obj instanceof TCPEndpoint)) {
	    TCPEndpoint ep = (TCPEndpoint)obj;
	    return (port == ep.port) && host.equals(ep.host);
	} else {
	    return false;
	}
    }

    /**
     * Write endpoint to output stream.
     */
    public void write(DataOutput out) throws java.io.IOException
    {
	out.writeUTF(host);
	out.writeInt(port);
    }
    
    /**
     * Get the endpoint from the input stream.
     * @param in the input stream
     * @exception IOException If id could not be read (due to stream failure)
     */
    public static TCPEndpoint read(DataInput in) throws java.io.IOException
    {
	String host = in.readUTF();
	int port = in.readInt();
	return new TCPEndpoint(host, port);
    }

    /**
     * The class FQDN encapsulates a routine that makes a best effort
     * attempt to retrieve the fully qualified domain name of the local
     * host.
     *
     * @author 	Laird Dornin
     */
    private static class FQDN implements Runnable {
	
	/**
	 * strings in which we can store discovered fqdn
	 */
	private String reverseLookup;

	private String hostAddress;

	private FQDN(String hostAddress) {
	    this.hostAddress = hostAddress;
	}

	/**
	 * Do our best to obtain a fully qualified hostname for the local
	 * host.  Perform the following steps to get a localhostname:
	 * 
	 * 1. InetAddress.getLocalHost().getHostName() - if contains
	 *    '.' use as FQDN
	 * 2. if no '.' query name service for FQDN in a thread
	 *    Note: We query the name service for an FQDN by creating 
	 *    an InetAddress via a stringified copy of the local ip 
	 *    address; this creates an InetAddress with a null hostname.
	 *    Asking for the hostname of this InetAddress causes a name
	 *    service lookup.
	 *
	 * 3. if name service takes too long to return, use ip address
	 * 4. if name service returns but response contains no '.' 
	 *    default to ipaddress.
	 */
	static String attemptFQDN(InetAddress localAddr) 
	    throws java.net.UnknownHostException {
	    
	    String hostName = localAddr.getHostName();
	    
	    if (hostName.indexOf('.') < 0 ) {
		
		String hostAddress = localAddr.getHostAddress();
		FQDN f = new FQDN(hostAddress);
		
		int nameServiceTimeOut = Utils.getLong
		    ("sun.rmi.transport.tcp.localHostNameTimeOut", 
		     10000).intValue();

		try {
		    synchronized(f) {
			f.getFQDN();

			/* wait to obtain an FQDN */
			f.wait(nameServiceTimeOut);
		    }
		} catch (InterruptedException e) {
		    /* propagate the exception to the caller */
		    Thread.currentThread().interrupt();
		}
		hostName = f.getHost();

		if ((hostName == null) || (hostName.equals("")) 
		    || (hostName.indexOf('.') < 0 )) {
		
		    hostName = hostAddress;
		}
	    }
	    return hostName;
	}
	
	/** 
	 * Method that that will start a thread to wait to retrieve a
	 * fully qualified domain name from a name service.  The spawned
	 * thread may never return but we have marked it as a daemon so the vm
	 * will terminate appropriately.  
	 */
	private void getFQDN() {
	    
	    Thread fqdnThread;

	    /* FQDN finder will run in RMI threadgroup. */
	    fqdnThread = RMIThread.newThread(this, "FQDN Finder", true);
	    fqdnThread.start();
	}
	
	private synchronized String getHost() {
	    return reverseLookup;
	}

	/**
	 * thread to query a name service for the fqdn of this host.
	 */
	public void run()  {

	    String name = null;

	    try {
		name = InetAddress.getByName(hostAddress).getHostName();
	    } catch (java.net.UnknownHostException e) {
	    } finally {
		synchronized(this) {
		    reverseLookup = name;
		    this.notify();
		}
	    }
	}
    }
}
