/*
 * @(#)TCPTransport.java	1.21 98/07/14
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.rmi.transport.tcp;

import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.io.*;
import java.util.Enumeration;
import java.util.Hashtable;
import java.rmi.RemoteException;
import java.rmi.server.ExportException;
import java.rmi.server.LogStream;
import java.rmi.server.ObjID;
import java.rmi.server.RMIFailureHandler;
import java.rmi.server.RMISocketFactory;
import java.rmi.server.RemoteCall;
import java.rmi.server.SocketSecurityException;
import java.rmi.server.UID;
import sun.rmi.transport.Channel;
import sun.rmi.transport.Connection;
import sun.rmi.transport.DGCAckHandler;
import sun.rmi.transport.Endpoint;
import sun.rmi.transport.RMIThread;
import sun.rmi.transport.StreamRemoteCall;
import sun.rmi.transport.Target;
import sun.rmi.transport.Transport;
import sun.rmi.transport.TransportConstants;
import sun.rmi.transport.Utils;

/**
 * TCPTransport is the socket-based implementation of the RMI Transport
 * abstraction.
 *
 * @author Ann Wollrath
 */
public class TCPTransport extends Transport implements Runnable {

    /** "tcp" package log level */
    static int logLevel = LogStream.parseLevel(
	Utils.getProperty("sun.rmi.transport.tcp.logLevel"));

    /** default port for TCPTransport; will be set when transport listens
     * on anonymous port. */
    static int DefaultPort = 0;

    /** flag marking if listen failed due to security exception */
    private boolean listenNotAllowed = false;
    /** flag marking if export is impossible (behind firewall) */
    private boolean exportNotAllowed = false;

    /** to tag threads */
    private static int threadNum = 0;
    /** table mapping threads to connections */
    private static Hashtable clientConnections = new Hashtable(37);

    /** endpoint for this transport */
    private TCPEndpoint endpoint;
    /** server socket for this transport */
    private ServerSocket server = null;
    /** table mapping endpoints to channels */
    private Hashtable channelTable = new Hashtable();

    static RMISocketFactory defaultSocketFactory =
	new sun.rmi.transport.proxy.RMIMasterSocketFactory();

    /**
     * Create a transport for this host on an anonymous port.
     */
    TCPTransport() {
	this(0);
    }

    TCPTransport(int port)  {
	if (TCPTransport.logLevel >= LogStream.BRIEF)
	    LogStream.log("tcp").println(
		"TCPTransport.<init>: Version = " +
		TransportConstants.Version + ", port = " + port);

	endpoint = TCPEndpoint.getLocalEndpoint(port);
    }
    
    /**
     * Returns a <I>Channel</I> that generates connections to the
     * endpoint <I>ep</I>. A Channel is an object that creates and
     * manages connections of a particular type to some particular
     * address space.
     * @param ep the endpoint to which connections will be generated.
     * @return the channel or null if the transport cannot
     * generate connections to this endpoint
     */
    public Channel getChannel(Endpoint ep)
    {
	Channel ch = null;

	if (ep instanceof TCPEndpoint) {
	    synchronized (channelTable) {
		ch = (Channel)channelTable.get(ep);
		if (ch == null) {
		    ch = new TCPChannel(this, (TCPEndpoint)ep);

		    // If we've determined that this transport cannot listen
		    // for new connections (and we need to), then all
		    // outgoing channels need to open a multiplexed
		    // connection.
		    if (listenNotAllowed && !exportNotAllowed) {
			try {
			    ((TCPChannel) ch).haveMultiplexer();
			} catch (RemoteException e) {
			    if (TCPTransport.logLevel >= LogStream.BRIEF) {
				LogStream ls = LogStream.log("tcp");
				synchronized (ls) {
				ls.print(
				    "TCPTransport(" + endpoint.getPort() +
				    ").getChannel: " +
				    "failed to have multiplexed connection " +
				    "on channel, exception: ");
				e.printStackTrace(ls);
				}
			    }

			    exportNotAllowed = true;

			    // Here we should remove all objects that have
			    // so far tried to export on this transport from
			    // the object table??
			}
		    }

		    channelTable.put(ep, ch);
		}
	    }
	}
	return ch;
    }

    /**
     * Returns an endpoint for this address space.
     */
    public Endpoint thisEndpoint() 
    {
	return endpoint;
    }

    /**
     * Export the object so that it can accept incoming calls.
     */
    public void exportObject(ObjID id, Target target)
	throws RemoteException
    {
	if (exportNotAllowed)
	    throw new ExportException(
		"failed to create multiplex protocol connection");

	super.exportObject(id, target);

	// don't try to listen if we know we can't on default port
	if (listenNotAllowed &&
	    endpoint.equals(TCPEndpoint.getLocalEndpoint(0)))
	    return;

	// try to listen
	try {
	    listen();
	} catch (SocketSecurityException e) {
	    if (TCPTransport.logLevel >= LogStream.BRIEF)
		LogStream.log("tcp").println(
		    "TCPTransport(" + endpoint.getPort() +
		    ").exportObject: " +
		    "listen failed with exception " + e.toString());

	    // fail if specific port was requested
	    if (endpoint.getPort() != 0)
		throw e;

	    // Since we are not allowed to listen for new connections, to
	    // be able to export objects, channels must open multiplexed
	    // connections to allow back communication.
	    synchronized (channelTable) {
		listenNotAllowed = true;
		Enumeration enum = channelTable.elements();
		while (enum.hasMoreElements()) {
		    TCPChannel ch = (TCPChannel) enum.nextElement();
		    try {
			ch.haveMultiplexer();
		    } catch (RemoteException ex) {
			if (TCPTransport.logLevel >= LogStream.BRIEF) {
			    LogStream ls = LogStream.log("tcp");
			    synchronized (ls) {
			    ls.print(
				"TCPTransport(" + endpoint.getPort() +
				").exportObject: " +
				"failed to have multiplexed connection " +
				"on channel, exception: ");
			    e.printStackTrace(ls);
			    }
			}
				    
			exportNotAllowed = true;
			throw new ExportException(
			    "failed to create multiplex protocol connection");
		    }
		}
		System.out.println("\n*** RMI unable to listen on socket: using multiplexed connections instead ***\n");
		
	    }
	}
    }
    
    /**
     * Listen on transport's endpoint.
     */
    private synchronized void listen() throws RemoteException
    {
	if (server == null) {
	    int port = endpoint.getPort();

	    if (TCPTransport.logLevel >= LogStream.BRIEF)
		LogStream.log("tcp").println(
		    "TCPTransport(" + endpoint.getPort() + ").listen: " +
		    "create socket, port = " +  port);

	    try {
		RMISocketFactory factory =
		    RMISocketFactory.getSocketFactory();

		// debug option to simulate restricted applet behavior
		// property "sun.rmi.transport.tcp.simulateApplet":
		//	>= 1	can only get "localhost" for host name
		//	>= 2	listen on socket throws security exception
		if (Utils.getLong(
		    "sun.rmi.transport.tcp.simulateApplet", 0).intValue() >= 2)
		    throw new sun.applet.AppletSecurityException(
			"socket.listen");

		server = ((factory != null) ? factory :
			  defaultSocketFactory).createServerSocket(port);
		
		if (DefaultPort == 0 && port == 0) {
		    DefaultPort = server.getLocalPort();

		    if (TCPTransport.logLevel >= LogStream.BRIEF)
			LogStream.log("tcp").println(
			    "TCPTransport(" + endpoint.getPort() + ").listen: " +
			    "default port set to " + DefaultPort);

		    // If our endpoint was at port 0, this call will fix
		    // its port number and fix the endpoint->transport table.
		    // Otherwise, this will define the default port for
		    // future anonymous exports.
		    TCPEndpoint.setDefaultPort(DefaultPort);
		}

		newListener();
	    } catch (java.lang.SecurityException e) {
		throw new
		    SocketSecurityException("Listen disallowed on port: " +
					    port, e); 
	    } catch (java.net.BindException e) {
		throw new ExportException("Port already in use: " + port, e);
	    } catch (IOException e) {
		throw new ExportException("Listen failed on port: " + port, e);
	    }
	}
    }

    /**
     * Private method to create a new thread to accept incoming calls
     * for this transport.
     */
    private void newListener() 
    {
	Thread t = RMIThread.newThread(this, "TCP Accept-" + ++threadNum,
				       true);
	t.start();
    }

    /**
     * Private method to retry creation of a server socket.
     *
     * @return if true, retry was successful
     */
    synchronized private boolean retryServerSocket() 
    {
	int port = endpoint.getPort();
	
	while (true) {
	    try {
		RMISocketFactory factory =
		    RMISocketFactory.getSocketFactory();
		server = ((factory != null) ? factory :
		          defaultSocketFactory).createServerSocket(port);
		return true;
	    } catch (IOException ex) {
		/*
		 * Retry server socket if no failure handler
		 * is installed.  If one is installed, leave it
		 * up to the failure handler to decide.
		 */
		RMIFailureHandler fh =
		    RMISocketFactory.getFailureHandler();
		if (fh != null) {
		    if (fh.failure(ex) == false) {
			server = null;
			return false;
		    }
		} else {
		    /*
		     * Default behavior if no failure handler is
		     * installed: sleep for a bit in the hope that
		     * some resources will be freed so the next
		     * socket creation will succeed.
		     */
		    try {
			Thread.sleep(10000);
		    } catch (InterruptedException e) {
			// ignore
		    }
		}
		
	    }
	}
    }

    /**
     * Accept and service server-side input from client.
     */
    public void run() 
    {
	Socket socket;

	if (TCPTransport.logLevel >= LogStream.BRIEF)
	    LogStream.log("tcp").println(
		"TCPTransport.run: listening on port " + endpoint.getPort());

	// accept connection
	while(true) {
	    ServerSocket myServer = server;
	    if (myServer == null)
		// the server socket will be null if another thread
		// attempted to retry creating the server socket
		// and failed.
		return;

	    Throwable acceptFailure = null;
	    try {
		socket = myServer.accept();
		break;

	    } catch (IOException e) {
		acceptFailure = e;
	    } catch (RuntimeException e) {
		acceptFailure = e;
	    } catch (Error e) {
		acceptFailure = e;
	    } finally {
		if (acceptFailure != null) {
		/*
		 * First, if *any* exception or error occurred trying to
		 * do an accept() (such a NoClassDefFoundError because there
		 * were no available file descriptors, and the VM could not
		 * load the java.net.SocketException class), then we must
		 * make sure that the server socket object is closed, or else
		 * clients connecting to it will be left hanging forever.
		 *
		 * (Normally, ServerSocket.accept() automatically closes
		 * the server socket on when an IOException occurs, but that
		 * doesn't cover the trickier cases like running out of file
		 * descriptors.)
		 */
		    try {
			server.close();
		    } catch (IOException e) {
			// ignore problems closing server
		    }
		/*
		 * Next, we want to execute the fallback mechanism for
		 * recreating server socket on failure by calling the
		 * retryServerSocket(), but we shouldn't do that for
		 * really severe errors.  So, the decision to retry is
		 * made if the original failure was just an IOException,
		 * or if we got a NoClassDefFoundError, which should
		 * handle the difficult case of running out of file
		 * descriptors described above.
		 */
		    if ((acceptFailure instanceof IOException) ||
			(acceptFailure instanceof NoClassDefFoundError)) {
			/* transport unable to accept connection.
			 * when socket fails, it closes the connection so a new
			 * server socket must be created here... */
			if (retryServerSocket() == false)
			    return;
		    } else {
		/*
		 * Otherwise, the error is probably too severe to handle here,
		 * so throw it up.  The preceding logic requires that the
		 * "acceptFailure" must be a RuntimeException or an Error,
		 * and hence unchecked, but the compiler can't tell that,
		 * so we have to convince it of that explicitly...
		 */
			if (acceptFailure instanceof RuntimeException)
			    throw (RuntimeException) acceptFailure;
			else // acceptFailure must be instance of Error...
			    throw (Error) acceptFailure;
		    }
		}
	    }
	}

	// set socket to disable Nagle's algorithm (always send immediately)
	try {
	    socket.setTcpNoDelay(true);
	} catch (Exception e) {
	    // if we fail to set this, ignore and proceed anyway
	}

	// create another thread to accept on this port 
	newListener();
	
	try {
	    InputStream sockIn = socket.getInputStream();
	    DataInputStream in = new DataInputStream(sockIn);

	    OutputStream sockOut = socket.getOutputStream();
	    DataOutputStream out = new DataOutputStream(sockOut);

	    // read and verify transport header
	    int  magic = in.readInt();
	    short version = in.readShort();
	    if (magic != TransportConstants.Magic ||
		version != TransportConstants.Version) {
		// protocol mismatch detected...
		// just close socket: this would recurse if we marshal an
		// exception to the client and the protocol at other end
		// doesn't match.
		closeSocket(socket);
		return;
	    }

	    // find client host name (or "" if from behind firewall)
	    InetAddress addr = socket.getInetAddress();
	    String host = (addr != null) ? addr.getHostAddress() : "";

	    if (TCPTransport.logLevel >= LogStream.BRIEF)
		LogStream.log("tcp").println(
		    "TCPTransport.run: accepted socket from [" +
		    host + ":" + socket.getPort() + "]");

	    TCPEndpoint ep;
	    TCPChannel ch;
	    TCPConnection conn;

	    // send ack (or nack) for protocol
	    byte protocol = in.readByte();
	    switch (protocol) {
	    case TransportConstants.SingleOpProtocol:
		// no ack for protocol

		// create dummy channel for receiving messages
		ep = new TCPEndpoint(host, socket.getLocalPort());
		ch = new TCPChannel(this, ep);
		conn = new TCPConnection(ch, socket);

		// read input messages
		handleMessages(conn, false);
		break;

	    case TransportConstants.StreamProtocol:
		// send ack
		out.writeByte(TransportConstants.ProtocolAck);
		out.flush();

		// suggest endpoint (in case client doesn't know host name)
		ep = new TCPEndpoint(host, socket.getPort());
		if (TCPTransport.logLevel >= LogStream.VERBOSE)
		    LogStream.log("tcp").println(
			"TCPTransport(" + endpoint.getPort() + ").run: " +
			"suggesting endpoint " + ep);
		ep.write(out);
		out.flush();

		// read and discard (possibly bogus) endpoint
		ep = TCPEndpoint.read(in);
		if (TCPTransport.logLevel >= LogStream.VERBOSE)
		    LogStream.log("tcp").println(
			"TCPTransport(" + endpoint.getPort() + ").run: " +
			"using endpoint " + ep);

		// create dummy channel for receiving messages
		ep = new TCPEndpoint(host, socket.getLocalPort());
		ch = new TCPChannel(this, ep);
		conn = new TCPConnection(ch, socket);

		// read input messages
		handleMessages(conn, true);
		break;

	    case TransportConstants.MultiplexProtocol:
		if (TCPTransport.logLevel >= LogStream.VERBOSE)
		    LogStream.log("tcp").println(
			"TCPTransport(" + endpoint.getPort() + ").run: " +
			"accepting multiplex protocol");

		// send ack
		out.writeByte(TransportConstants.ProtocolAck);

		// suggest endpoint (in case client doesn't already have one)
		ep = new TCPEndpoint(host, socket.getPort());
		if (TCPTransport.logLevel >= LogStream.VERBOSE)
		    LogStream.log("tcp").println(
			"TCPTransport(" + endpoint.getPort() + ").run: " +
			"suggesting endpoint " + ep);
		ep.write(out);
		out.flush();

		// read endpoint client has decided to use
		ep = TCPEndpoint.read(in);
		if (TCPTransport.logLevel >= LogStream.VERBOSE)
		    LogStream.log("tcp").println(
			"TCPTransport(" + endpoint.getPort() + ").run: " +
			"using endpoint " + ep);

		ConnectionMultiplexer multiplexer;
		synchronized (channelTable) {
		    // create or find channel for this endpoint
		    ch = (TCPChannel) getChannel(ep);
		    multiplexer = new ConnectionMultiplexer(ch,
			sockIn, sockOut, false);
		    ch.useMultiplexer(multiplexer);
		}
		multiplexer.run();
		break;

	    default:
		// protocol not understood, send nack and close socket
		out.writeByte(TransportConstants.ProtocolNack);
		out.flush();
		break;
	    }

	} catch (IOException e) {
	    // socket in unknown state: destroy socket
	    if (TCPTransport.logLevel >= LogStream.BRIEF) {
		LogStream ls = LogStream.log("tcp");
		ls.println("TCPTransport.run: terminated with exception: ");
		e.printStackTrace(ls);
	    }
	} finally {
	    closeSocket(socket);
	}
    }

    /** close socket and eat exception */
    private static void closeSocket(Socket sock) 
    {
	try {
	    sock.close();
	} catch (IOException ex) {
	    // eat exception
	}
    }
    
    /**
     * handleMessages decodes transport operations and handles messages
     * appropriately.  If an exception occurs during message handling,
     * the socket is closed.
     */
    void handleMessages(Connection conn, boolean persistent)
    {
	// put the connection in the current connections by thread
	Thread currthread = Thread.currentThread();
	clientConnections.put(currthread, conn);

	try {
	    do {
		DataInputStream in =
		    new DataInputStream(conn.getInputStream());
		byte op = in.readByte();// transport op

		if (TCPTransport.logLevel >= LogStream.BRIEF)
		    LogStream.log("tcp").println(
			"TCPTransport(" + endpoint.getPort() +
			").handleMessages: op = " + op);
		    
		switch (op) {
		case TransportConstants.Call:
		    // service incoming RMI call
		    RemoteCall call = new StreamRemoteCall(conn);
		    if (serviceCall(call) == false)
			return;
		    break;

		case TransportConstants.Ping:
		    // send ack for ping
		    DataOutputStream out =
			new DataOutputStream(conn.getOutputStream());
		    out.writeByte(TransportConstants.PingAck);
		    conn.releaseOutputStream();
		    break;

		case TransportConstants.DGCAck:
		    DGCAckHandler.received(UID.read(in));
		    break;
		    
		default:
		    throw new IOException("unknown transport op " + op);
		}
	    } while (persistent);
	    
	} catch (IOException e) {
	    // exception during processing causes connection to close (below)
	    if (TCPTransport.logLevel >= LogStream.BRIEF) {
		LogStream ls = LogStream.log("tcp");
		ls.println("TCPTransport(" + endpoint.getPort() +
		    ").handleMessages: exception: ");
		e.printStackTrace(ls);
	    }
	} finally {
	    // clean out the current connection from table
	    clientConnections.remove(currthread);
	    try {
		conn.close();
	    } catch (IOException ex) {
		// eat exception
	    }
	}
    }
    
    /**
     * Return the Channel for the Connection of the current thread, if
     * one is active.
     */
    /*
     * This could return the Connection but that would
     * give the caller a chance to muck with the connection and all
     * that is needed is the channel so the host:port of the client
     * can be known.
     * This is used by RemoteServer:getClientHost and probably
     * should be redesigned.
     */
    public static Channel getCurrentChannel() {
	Thread currthread = Thread.currentThread();

	Object o = clientConnections.get(currthread);
	if (o != null && o instanceof Connection) {
	    Channel ch = ((Connection)o).getChannel();
	    return ch;
	}
	return null;
    }
}
