/*
 * @(#)PKCS10.java	1.19 99/03/08
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


package sun.security.pkcs;

import java.io.ByteArrayOutputStream;
import java.io.PrintStream;
import java.io.IOException;

import java.security.NoSuchAlgorithmException;
import java.security.InvalidKeyException;
import java.security.Signature;
import java.security.SignatureException;

import sun.misc.BASE64Encoder;

import sun.security.util.*;	// DER
import sun.security.x509.*;



/**
 * PKCS #10 certificate requests are created and sent to Certificate
 * Authorities, which then create X.509 certificates and return them to
 * the entity which created the certificate request.  These cert requests
 * basically consist of the subject's X.500 name and public key, signed
 * using the corresponding private key.
 *
 * <P>PKCS #10 supports an optional set of attributes.  At this time,
 * such attributes are not supported by this class.
 *
 * @version 1.19
 * @author David Brownell
 */
public
class PKCS10
{
    /**
     * Constructs an unsigned PKCS #10 certificate request.  Before this
     * request may be used, it must be encoded and signed.  Then it
     * must be retrieved in some conventional format (e.g. string).
     * 
     * @param publicKey the public key that should be placed
     *		into the certificate generated by the CA.
     */
    public PKCS10 (X509Key publicKey)
    {
	subjectPublicKeyInfo = publicKey;
    }


    /**
     * Parses an encoded, signed PKCS #10 certificate request, verifying
     * the request's signature as it does so.  This constructor would
     * typically be used by a Certificate Authority, from which a new
     * certificate would then be constructed.
     *
     * @param data the DER-encoded PKCS #10 request.
     * @except IOException for low level errors reading the data
     * @except SignatureException when the signature is invalid
     * @except NoSuchAlgorithmException when the signature
     *	algorithm is not supported in this environment
     */
    public PKCS10 (byte data [])
    throws IOException, SignatureException, NoSuchAlgorithmException
    {
	DerInputStream	in;
	DerValue	seq [];
	AlgorithmId	id;
	byte		sigData [];
	Signature	sig;

	certificateRequest = data;

	//
	// Outer sequence:  request, signature algorithm, signature.
	// Parse, and prepare to verify later.
	//
	in = new DerInputStream (data);
	seq = in.getSequence (3);

	if (seq.length != 3)
	    throw new IllegalArgumentException ("not a PKCS #10 request");
	
	data = seq [0].toByteArray ();		// reusing this variable
	id = AlgorithmId.parse (seq [1]);
	sigData = seq [2].getBitString ();

	//
	// Inner sequence:  version, name, key, attributes
	//
	BigInt		serial;
	DerValue	val;

	serial = seq [0].data.getInteger ();
	if (serial.toInt () != 0)
	    throw new IllegalArgumentException ("not PKCS #10 v1");

	subject = new X500Name (seq [0].data);
	subjectPublicKeyInfo = X509Key.parse (seq [0].data.getDerValue ());

	// Cope with a somewhat common illegal PKCS #10 format
	if (seq [0].data.available () != 0)
	    attributeSet = seq [0].data.getDerValue ();
	else
	    attributeSet = null;
	
	if (seq [0].data.available () != 0)
	    throw new IllegalArgumentException ("illegal PKCS #10 data");

	//
	// OK, we parsed it all ... validate the signature using the
	// key and signature algorithm we found.
	//

	try {
	    sig = Signature.getInstance(id.getName ());
	    sig.initVerify (subjectPublicKeyInfo);
	    sig.update (data);
	    if (!sig.verify (sigData))
		throw new SignatureException ("Invalid PKCS #10 signature");
	} catch (InvalidKeyException e) {
	    throw new SignatureException ("invalid key");
	}
    }

    // XXX public void addAttribute (AVA attribute)

    /**
     * Create the signed certificate request.  This will later be
     * retrieved in either string or binary format.
     *
     * @param requester identifies the signer (by X.500 name)
     *		and provides the private key used to sign.
     */
    public void encodeAndSign (X500Signer requester)
    throws IOException, SignatureException
    {
	DerOutputStream		out, scratch;
	byte 			certificateRequestInfo [];
	byte			sig [];

	if (certificateRequest != null)
	    throw new SignatureException ("request is already signed");

	subject = requester.getSigner ();

	/*
	 * Encode empty set of AVA ... 0xA0 is context-specific
	 * tag zero, in "constructed" form, which (here) indicates
	 * "set of AVA".
	 *
	 * XXX we'll need support for these attributes before long
	 */
	scratch = new DerOutputStream ();	// empty set
	out = new DerOutputStream ();
	out.write ((byte) 0x0A0, scratch);
	attributeSet = new DerValue (out.toByteArray ());

	/*
	 * Emit cert request info, wrap in a sequence for signing
	 */
	scratch = new DerOutputStream ();
	scratch.putInteger (new BigInt (0));		// version zero
	subject.emit (scratch);				// X.500 name
	subjectPublicKeyInfo.emit (scratch);		// public key
	// scratch.putDerValue (attributeSet);		// attributes
	attributeSet.emit (scratch);

	out = new DerOutputStream ();
	out.write (DerValue.tag_Sequence, scratch);	// wrap it!
	certificateRequestInfo = out.toByteArray ();
	scratch = out;

	/*
	 * Sign it ...
	 */
	requester.update (certificateRequestInfo, 0,
		certificateRequestInfo.length);
	sig = requester.sign ();

	/*
	 * Build guts of SIGNED macro
	 */
	requester.getAlgorithmId ().emit (scratch);	// sig algorithm
	scratch.putBitString (sig);			// sig

	/*
	 * Wrap those guts in a sequence
	 */
	out = new DerOutputStream ();
	out.write (DerValue.tag_Sequence, scratch);
	certificateRequest = out.toByteArray ();
    }


    /**
     * Returns the subject's name.
     */
    public X500Name getSubjectName ()
	{ return subject; }


    /**
     * Returns the subject's public key.
     */
    public X509Key getSubjectPublicKeyInfo ()
	{ return subjectPublicKeyInfo; }


    /**
     * Returns the encoded and signed certificate request as a
     * DER-encoded byte array.
     *
     * @return the certificate request, or null if encodeAndSign()
     *		has not yet been called.
     */
    public byte [] toByteArray ()
    {
	return certificateRequest;
    }


    /**
     * Prints an E-Mailable version of the certificate request on the print
     * stream passed.  The format is a common base64 encoded one, supported
     * by most Certificate Authorities because Netscape web servers have
     * used this for some time.  Some certificate authorities expect some
     * more information, in particular contact information for the web
     * server administrator.
     *
     * @param out the print stream where the certificate request
     *	will be printed.
     * @except IOException when an output operation failed
     * @except SignatureException when the certificate request was
     *	not yet signed.
     */
    public void print (PrintStream out)
    throws IOException, SignatureException
    {
	if (certificateRequest == null)
	    throw new SignatureException ("Cert request was not signed");
	
	BASE64Encoder		encoder = new BASE64Encoder ();

	out.println ("-----BEGIN NEW CERTIFICATE REQUEST-----");
	encoder.encodeBuffer (certificateRequest, out);
	out.println ("-----END NEW CERTIFICATE REQUEST-----");
    }

    /**
     * Provides a short description of this request.
     */
    public String toString ()
    {
	return "[PKCS #10 certificate request, "
	    + subjectPublicKeyInfo.getAlgorithmId ().getName ()
	    + " public key, for <" + subject + ">]";
    }

    private X500Name	subject;
    private X509Key	subjectPublicKeyInfo;

    private DerValue	attributeSet;	// not yet exposed

    private byte	certificateRequest [];	// signed
}
