/*
 * @(#)PKCS7.java	1.21 99/03/08
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.security.pkcs;

import java.io.*;
import java.util.*;
import java.security.*;

import sun.security.util.*;
import sun.security.x509.*;

/**
 * PKCS7 as defined in RSA Laboratories PKCS7 Technical Note. Profile
 * supported: <tt>data</tt> and <tt>signedData</tt> ContentInfo
 * types. For signedData, <tt>crls</tt>, <tt>attributes</tt> and
 * PKCS#6 Extended Certificates are not supported.
 *
 * @version 1.21 99/11/17
 * @author Benjamin Renaud 
 */
public class PKCS7 {

    /* Are we debugging? */
   // private boolean debug = false;
    private boolean debug = true;

    private ObjectIdentifier contentType;

    // the ASN.1 members for a signedData (and other) contentTypes
    private BigInt version;
    private AlgorithmId[] digestAlgorithmIds;
    private ContentInfo contentInfo;
    private X509Cert[] certificates;
    private SignerInfo[] signerInfos;

    public PKCS7(InputStream in) throws ParsingException, IOException {
	DataInputStream dis = new DataInputStream(in);
	byte[] data = new byte[dis.available()];
	dis.readFully(data);
	
	parse(new DerInputStream(data));
    }
    
    public PKCS7(DerInputStream derin) throws ParsingException {
	parse(derin);
    }

    public PKCS7(byte[] bytes) throws ParsingException {
	DerInputStream derin = new DerInputStream(bytes);
	parse(derin);
    }

    private void parse(DerInputStream derin) throws ParsingException {
	try {
	    ContentInfo contentInfo = new ContentInfo(derin);
	    contentType = contentInfo.contentType;
	
	    if (contentType.equals(ContentInfo.SIGNED_DATA_OID)) {
		parseSignedData(contentInfo.getContent());
	    } else {
		throw new ParsingException("content type " + contentType +
					   "not supported.");
	    }
	} catch (IOException e) {
	    debug(e);
	    throw new ParsingException("IOException.");	    
	}
    }

    public PKCS7(AlgorithmId[] digestAlgorithmIds,
		 ContentInfo   contentInfo,
		 X509Cert[]    certificates,
		 SignerInfo[]  signerInfos) {

	version = new BigInt(1);
	this.digestAlgorithmIds = digestAlgorithmIds;
	this.contentInfo = contentInfo;
	this.certificates = certificates;
	this.signerInfos = signerInfos;
    }

    private void parseSignedData(DerValue val)
	throws ParsingException, IOException {
	    
	DerInputStream dis = val.toDerInputStream();
	
	// Version
	version = dis.getInteger();

	// digestAlgorithmIds

	DerValue[] digestAlgorithmIdVals = dis.getSet(1);
	int len = digestAlgorithmIdVals.length;

	digestAlgorithmIds = new AlgorithmId[len];
	try {
	    for (int i = 0; i < len; i++) {
		DerValue oid = digestAlgorithmIdVals[i];
		digestAlgorithmIds[i] = AlgorithmId.parse(oid);
	    }

	} catch (IOException e) {
	    debug(e);
	    throw new ParsingException("error parsing digest AlgorithmId IDs");
	}
	// contentInfo
	contentInfo = new ContentInfo(dis);

	// certificates
	DerValue[] certificateVals = dis.getSet(2);
	
	len = certificateVals.length;
	certificates = new X509Cert[len];
	
	for (int i = 0; i < len; i++) {
	    X509Cert cert = new X509Cert(certificateVals[i]);
	    certificates[i] = cert;
	}

	// crls are ignored.
	dis.getSet(0);

	// signerInfos
	DerValue[] signerInfoVals = dis.getSet(1);
	
	len = signerInfoVals.length;
	signerInfos = new SignerInfo[len];

	for (int i = 0; i < len; i++) {
	    DerInputStream in = signerInfoVals[i].toDerInputStream();
	    signerInfos[i] = new SignerInfo(in);
	}
    }


    public void encodeSignedData(OutputStream out) throws IOException {
	DerOutputStream derout = new DerOutputStream();
	encodeSignedData(derout);
	out.write(derout.toByteArray());
    }

    public void encodeSignedData(DerOutputStream out) 
    throws IOException {

	DerOutputStream signedDataSeq = new DerOutputStream();

	// version
	signedDataSeq.putInteger(version);
	
	// digestAlgorithmIds
	DerOutputStream digestAlgSet = new DerOutputStream();
	for (int i = 0; i < digestAlgorithmIds.length; i++) {
	    digestAlgorithmIds[i].encode(digestAlgSet);
	}

	signedDataSeq.write(DerValue.tag_Set, digestAlgSet);

	// contentInfo
	contentInfo.encode(signedDataSeq);
	
	// certificates
	DerOutputStream certSet = new DerOutputStream();
	for (int i = 0; i < certificates.length; i++) {
	    certificates[i].encode(certSet);
	}
	signedDataSeq.write(DerValue.tag_Set, certSet);

	// crls -- null sequence
	DerOutputStream nullStream = new DerOutputStream();
	signedDataSeq.write(DerValue.tag_Set, nullStream);

	// signerInfos
	DerOutputStream signerInfoSet = new DerOutputStream();
	for (int i = 0; i < signerInfos.length; i++) {
	    signerInfos[i].encode(signerInfoSet);
	}
	signedDataSeq.write(DerValue.tag_Set, signerInfoSet);

	// making it a signed data block
	DerValue signedData = new DerValue(DerValue.tag_Sequence, 
					   signedDataSeq.toByteArray());

	// making it a content info sequence
	ContentInfo block = new ContentInfo(ContentInfo.SIGNED_DATA_OID,
					    signedData);

	// writing out the contentInfo sequence.
	block.encode(out);

    }

    /**
     * This verifies a given SignerInfo.
     */
    public SignerInfo verify(SignerInfo info, byte[] bytes) 
    throws NoSuchAlgorithmException, SignatureException {
	return info.verify(this, bytes);
    }

    /** 
     * Returns all signerInfos which self-verify
     */
    public SignerInfo[] verify(byte[] bytes)
    throws NoSuchAlgorithmException, SignatureException {

	Vector intResult = new Vector();

	for (int i = 0; i < signerInfos.length; i++) {
	    SignerInfo signerInfo = verify(signerInfos[i], bytes);
	    if (signerInfo != null) {
		intResult.addElement(signerInfo);
	    } else {
	    }

	}

	if (intResult.size() != 0) {

	    SignerInfo[] result = new SignerInfo[intResult.size()];
	    intResult.copyInto(result);
	    return result;
	}
	return null;
    }

    public SignerInfo[] verify() 
    throws NoSuchAlgorithmException, SignatureException {
	return verify(null);
    }
    
    public  BigInt getVersion() {
	return version;
    }

    public AlgorithmId[] getDigestAlgorithmIds() {
	return  digestAlgorithmIds;
    }

    public ContentInfo getContentInfo() {
	return contentInfo;
    }

    public X509Cert[] getCertificates() {
	return certificates;
    }

    public SignerInfo[] getSignerInfos() {
	return signerInfos;
    }

    public X509Cert getCertificate(BigInt serial, X500Name name) {
	for (int i = 0; i < certificates.length; i++) {
	    X509Cert cert = certificates[i];
	    X500Name thisName = cert.getSubjectName();
	    BigInt thisSerial = cert.getSerialNumber();
	    if (serial.equals(thisSerial) && name.equals(thisName)) {
		return cert;
	    }
	}
	return null;
    }

    public String toString() {
	String out = "";

	out += "PKCS7 :: version: " + version + "\n";
	out += "PKCS7 :: digest AlgorithmIds: \n";
	for (int i = 0; i < digestAlgorithmIds.length; i++) {
	    out += "\t" + digestAlgorithmIds[i] + "\n";
	}
	out += contentInfo + "\n";
	out += "PKCS7 :: certificates: \n";
	for (int i = 0; i < certificates.length; i++) {
	    out += "\t" + i + ".   " + certificates[i] + "\n";
	}
	out += "PKCS7 :: signer infos: \n";
	for (int i = 0; i < signerInfos.length; i++) {
	    out += ("\t" + i + ".  " + signerInfos[i] + "\n");
	}
	return out;
    }

    private void debug(Throwable t) {
	if (debug) {
	    t.printStackTrace();
	}
    }
}


	
