/*
 * @(#)PKCS8Key.java	1.18 99/03/08
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.security.pkcs;

import java.io.*;
import java.math.*;
import java.security.Key;
import java.security.Provider;
import java.security.PrivateKey;
import java.security.Security;
import java.security.InvalidKeyException;
import java.security.ProviderException;
import java.util.Properties;

import sun.misc.HexDumpEncoder;
import sun.security.x509.*;
import sun.security.util.*;

// this will go away.
import sun.security.provider.Sun;

/**
 * Holds a PKCS#8 key, for example a private key
 *
 * @version 1.18, 99/11/17
 * @author Dave Brownell
 * @author Benjamin Renaud
 */
public
class PKCS8Key implements PrivateKey {

    /* The algorithm information (name, parameters, etc). */
    protected AlgorithmId algid;

    /* The key bytes, without the algorithm information */
    protected byte[] key;

    /* The encoded for the key. */
    private byte[] encodedKey;

    /* The version for this key */
    public static final BigInteger version = BigInteger.valueOf(0);
    
    /**
     * Default constructor.  The key constructed must have its key
     * and algorithm initialized before it may be used, for example
     * by using <code>decode</code>.
     */
    protected PKCS8Key () { }

    /*
     * Build and initialize as a "default" key.  All PKCS#8 key
     * data is stored and transmitted losslessly, but no knowledge
     * about this particular algorithm is available.
     */
    private PKCS8Key (AlgorithmId algid, byte key [])
    throws InvalidKeyException {
	this.algid = algid;
	this.key = key;
	encode();
    }


    /**
     * Construct PKCS#8 subject public key from a DER value.  If
     * the runtime environment is configured with a specific class for
     * this kind of key, a subclass is returned.  Otherwise, a generic
     * PKCS8Key object is returned.
     * 
     * <P>This mechanism gurantees that keys (and algorithms) may be
     * freely manipulated and transferred, without risk of losing
     * information.  Also, when a key (or algorithm) needs some special
     * handling, that specific need can be accomodated.
     *
     * @param in the DER-encoded SubjectPublicKeyInfo value
     * @exception IOException on data format errors
     */
    public static PKCS8Key parse (DerValue in) throws IOException
    {
	AlgorithmId	algorithm;
	PKCS8Key		subjectKey;
	
	if (in.tag != DerValue.tag_Sequence)
	    throw new IOException ("corrupt private key");

	BigInteger parsedVersion = in.data.getInteger().toBigInteger();
	if (!version.equals(parsedVersion)) {
	    throw new IOException("version mismatch: (supported: " + 
				  version + ", parsed: " + 
				  parsedVersion);
	}

	algorithm = AlgorithmId.parse (in.data.getDerValue ());

	try {
	    subjectKey = buildPKCS8Key (algorithm, in.data.getOctetString ());

	} catch (InvalidKeyException e) {
	    throw new IOException("corrupt private key");
	}

	if (in.data.available () != 0)
	    throw new IOException ("excess private key");
	return subjectKey;
    }


    /**
     * Parse the key bits.  This may be redefined by subclasses to take
     * advantage of structure within the key.  For example, RSA public
     * keys encapsulate two unsigned integers (modulus and exponent) as
     * DER values within the <code>key</code> bits; Diffie-Hellman and
     * DSS/DSA keys encapsulate a single unsigned integer.
     *
     * <P>This function is called when creating PKCS#8 SubjectPublicKeyInfo
     * values using the PKCS8Key member functions, such as <code>parse</code>
     * and <code>decode</code>.
     */
    protected void parseKeyBits () throws IOException, InvalidKeyException {
	encode();
    }

    /* Temporary workaround for X509Key */
    static String getSunProperty(String prop) {
	Provider sun = Security.getProvider("SUN");
	return sun.getProperty(prop);
    }
    /* End workaround */
    
    /*
     * Factory interface, building the kind of key associated with this
     * specific algorithm ID or else returning this generic base class.
     * See the description above.
     */
    static PKCS8Key buildPKCS8Key (AlgorithmId algid, byte key [])
    throws IOException, InvalidKeyException
    {
	Properties	props;
	String		keytype;
	String		classname;
	
	/* This is temporarily mapped back to SUN till we
	   decide on a proper interface for java.* stuff. */
	classname = getSunProperty("PrivateKey.PKCS#8." + algid.getName());
	
	if (classname == null) {
	    classname = "sun.security.x509.PKCS8Key";
	}

	try {
	    Class	keyClass = Class.forName(classname);
	    Object	inst;
	    PKCS8Key	result;

	    inst = keyClass.newInstance();
	    if (inst instanceof PKCS8Key) {
		result = (PKCS8Key) inst;
		result.algid = algid;
		result.key = key;
		result.parseKeyBits();
		return result;
	    }

	    System.err.println ("Misconfiguration:  faulty key config, " +
				classname);
	    return new PKCS8Key (algid, key);

	} catch (ClassNotFoundException e) {
	    System.err.println ("Misconfiguration:  unknown key class, " + 
			      classname);
	    return new PKCS8Key (algid, key);

	} catch (IllegalAccessException e) {
	    // this should not happen.
	    e.printStackTrace ();
	    throw new IOException (classname + " [internal error]");

	} catch (InstantiationException e) {
	    System.err.println ("Misconfiguration:  faulty key class, " + 
			      classname);
	    return new PKCS8Key (algid, key);
	}
    }


    /**
     * Returns the algorithm to be used with this key.
     */
    public String getAlgorithm() { 
	return algid.getName();
    }

    /**
     * Returns the algorithm ID to be used with this key.
     */
    public AlgorithmId	getAlgorithmId () { return algid; }


    /**
     * PKCS#8 sequence on the DER output stream.
     */
    public final void emit (DerOutputStream out) throws IOException
    {
	DerOutputStream		tmp = new DerOutputStream ();
	tmp.putInteger(new BigInt(version.toByteArray()));
	algid.emit (tmp);
	tmp.putOctetString (key);
	out.write (DerValue.tag_Sequence, tmp);
    }

    /**
     * Returns the DER-encoded form of the key as a byte array.
     */
    public synchronized byte[] getEncoded() {
	if (encodedKey == null) {
	    try {
		encode();
	    } catch (InvalidKeyException e) {
	    }
	}				   
	return encodedKey;
    }


    /**
     * Returns the format for this key: "PKCS#8"
     */
    public String getFormat() {
	return "PKCS#8";
    }

    /**
     * Returns the DER-encoded form of the key as a byte array.
     */
    public byte[] encode() throws InvalidKeyException {
	if (encodedKey == null) {
	    try {
		DerOutputStream	out;
		
		out = new DerOutputStream ();
		emit (out);
		encodedKey = out.toByteArray();

	    } catch (IOException e) {
		throw new InvalidKeyException ("IOException : " + 
					       e.getMessage());
	    }
	} 
	return encodedKey;
    }

    /*
     * Returns a printable representation of the key
     */
    public String toString ()
    {
	HexDumpEncoder	encoder = new HexDumpEncoder ();

	return "algorithm = " + algid.toString ()
	    + ", unparsed keybits = \n" + encoder.encodeBuffer (key);
    }

    /** 
     * Initialize an PKCS8Key object from an input stream.  The data
     * on that input stream must be encoded using DER, obeying the
     * PKCS#8 format: a sequence consisting of a version, an algorithm
     * ID and a bit string which holds the key.  (That bit string is
     * often used to encapsulate another DER encoded sequence.)
     *
     * <P>Subclasses should not normally redefine this method; they should
     * instead provide a <code>parseKeyBits</code> method to parse any
     * fields inside the <code>key</code> member.
     *
     * @param in an input stream with a DER-encoded PKCS#8
     * SubjectPublicKeyInfo value 
     */
    public void decode(InputStream in)
    throws InvalidKeyException
    {

	DerValue	val;

	try {
	    val = new DerValue (in);
	    if (val.tag != DerValue.tag_Sequence)
		throw new InvalidKeyException ("invalid key format");


	    BigInteger version = val.data.getInteger().toBigInteger();
	    if (!version.equals(this.version)) {
		throw new IOException("version mismatch: (supported: " + 
				      this.version + ", parsed: " + 
				      version);
	    }
	    algid = AlgorithmId.parse (val.data.getDerValue ());
	    key = val.data.getOctetString ();
	    parseKeyBits ();
	    if (val.data.available () != 0)
		throw new InvalidKeyException ("excess key data");

	} catch (IOException e) {
	    // e.printStackTrace ();
	    throw new InvalidKeyException();
	}
    }

    public void decode(byte[] encodedKey) throws InvalidKeyException {
	decode(new ByteArrayInputStream(encodedKey));
    }

    /**
     * Serialization write ... PKCS#8 keys serialize as
     * themselves, and they're parsed when they get read back.
     */
    private synchronized void
    writeObject (java.io.ObjectOutputStream stream)
    throws IOException { 
	stream.write(getEncoded()); 
    }


    /**
     * Serialization read ... PKCS#8 keys serialize as
     * themselves, and they're parsed when they get read back.
     */
    private synchronized void
    readObject (ObjectInputStream stream) 
    throws IOException {

	try {
	    decode(stream); 

	} catch (InvalidKeyException e) {
	    e.printStackTrace();
	    throw new IOException("deserialized key is invalid: " + 
				  e.getMessage());
	}
    }

}
