/*
 * @(#)SignerInfo.java	1.16 99/03/08
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.security.pkcs;

import java.io.*;
import java.security.*;

import sun.security.util.*;
import sun.security.x509.*;

/**
 * A SignerInfo, as defined in PKCS#7's signedData type.
 *
 * @version 1.16 99/11/17
 * @author Benjamin Renaud 
 */
public class SignerInfo {

    BigInt version;
    X500Name issuerName;
    BigInt certificateSerialNumber;
    AlgorithmId digestAlgorithmId;
    AlgorithmId digestEncryptionAlgorithmId;
    byte[] encryptedDigest;

    public SignerInfo(X500Name 	issuerName, 
		      BigInt serial,
		      AlgorithmId digestAlgorithmId,
		      AlgorithmId digestEncryptionAlgorithmId,
		      byte[] encryptedDigest) {
	this.version 			= new BigInt(1);
	this.issuerName 			= issuerName;
	this.certificateSerialNumber 	= serial;
	this.digestAlgorithmId 		= digestAlgorithmId;
	this.digestEncryptionAlgorithmId 	= digestEncryptionAlgorithmId;
	this.encryptedDigest 		= encryptedDigest;
    }

    public SignerInfo(DerInputStream derin) 
    throws IOException, ParsingException {

	// version
	version = derin.getInteger();	
	
	// issuerAndSerialNumber
	DerValue[] issuerAndSerialNumber = derin.getSequence(2);
	byte[] issuerBytes = issuerAndSerialNumber[0].toByteArray();
	issuerName = new X500Name(new DerValue(DerValue.tag_Sequence, 
					       issuerBytes));
	certificateSerialNumber = issuerAndSerialNumber[1].getInteger();

	// digestAlgorithmId
	DerValue tmp = derin.getDerValue();

	digestAlgorithmId = AlgorithmId.parse(tmp);

	    // we punt on set of auth attributes
	derin.getSet(0);
	
	// digestEncryptionAlgorithmId - little RSA naming scheme - 
	// signature == encryption...
	tmp = derin.getDerValue();

	digestEncryptionAlgorithmId = AlgorithmId.parse(tmp);

	// encryptedDigest
	encryptedDigest = derin.getOctetString();

	// we punt on set of unauth attributes
	derin.getSet(0);
	
	// all done
	if (derin.available() != 0) {
	    throw new ParsingException("extra data at the end");
	}
    }

    public void encode(DerOutputStream out) throws IOException {
	
	DerOutputStream seq = new DerOutputStream();
	seq.putInteger(version);

	DerOutputStream issuerAndSerialNumber = new DerOutputStream();
	issuerName.emit(issuerAndSerialNumber);
	issuerAndSerialNumber.putInteger(certificateSerialNumber);
	seq.write(DerValue.tag_Sequence, issuerAndSerialNumber);

	digestAlgorithmId.encode(seq);
	
	DerOutputStream nullStream = new DerOutputStream();
	// auth attributes
	seq.write(DerValue.tag_Set, nullStream);
	    
	digestEncryptionAlgorithmId.encode(seq);
	
	seq.putOctetString(encryptedDigest);
	
	// unauth attributes
	seq.write(DerValue.tag_Set, nullStream);	
	
	out.write(DerValue.tag_Sequence, seq);
    }

    public X509Cert getCertificate(PKCS7 block) {
	return block.getCertificate(certificateSerialNumber, issuerName);
    }

    /* Returns null if verify fails, and the key that succeeded if
       verify succeed. */
    SignerInfo verify(PKCS7 block, byte[] bytes) 
    throws NoSuchAlgorithmException, SignatureException {

	IdentityScope scope = IdentityScope.getSystemScope();

	try {
	    if (bytes == null) {
		ContentInfo content = block.getContentInfo();
		bytes = content.getContentBytes();
	    }

	    String algname = getDigestEncryptionAlgorithmId().getName();
	    Signature sig = Signature.getInstance(algname);
	    X509Cert cert = getCertificate(block);
						 
	    if (cert == null) {
		throw new SignatureException("No cert for " + issuerName);
	    }

	    PublicKey key = cert.getPublicKey();
	    sig.initVerify(key);
	    
	    sig.update(bytes);

	    if (sig.verify(encryptedDigest)) {
		return this;
	    }

	} catch (IOException e) {
	    throw new SignatureException("IO error verifying signature:\n" +
					 e.getMessage());

	} catch (InvalidKeyException e) {
	    throw new SignatureException("InvalidKey: " + e.getMessage());

	}
	return null;
    }
	
    /* Verify the content of the pkcs7 block. */
    SignerInfo verify(PKCS7 block)
    throws NoSuchAlgorithmException, SignatureException {
	return verify(block, null);
    }
	

    public BigInt getVersion() {
	    return version;
    }

    public X500Name getIssuerName() {
	return issuerName;
    }

    public BigInt getCertificateSerialNumber() {
	return certificateSerialNumber;
    }

    public AlgorithmId getDigestAlgorithmId() {
	return digestAlgorithmId;
    }

    public AlgorithmId getDigestEncryptionAlgorithmId() {
	return digestEncryptionAlgorithmId;
    }

    public byte[] getEncryptedDigest() {
	return encryptedDigest;
    }

    public String toString() {
	String out = "";

	out += "Signer Info for (issuer): " + issuerName + "\n";
	out += "\tversion: " + version + "\n";
	out += "\tcertificateSerialNumber: " + certificateSerialNumber + 
	    "\n";
	out += "\tdigestAlgorithmId: " + digestAlgorithmId + "\n";
	out += "\tdigestEncryptionAlgorithmId: " + digestEncryptionAlgorithmId +
	    "\n";
	out += "\tencryptedDigest: " + encryptedDigest + "\n";
	
	return out;
    }

}


	
	
