/*
 * @(#)DSAPrivateKey.java	1.40 99/03/08
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.security.provider;

import java.util.*;
import java.io.*;
import java.math.BigInteger;
import java.security.*;
import java.security.interfaces.DSAKey;
import java.security.interfaces.DSAParams;

import sun.security.x509.*;
import sun.security.util.*;
import sun.security.pkcs.*;

/**
 * A PKCS#8 private key for the Digital Signature Algorithm.
 *
 * @see DSAPublicKey
 * @see AlgIdDSA
 * @see DSA
 * 
 * @author Benjamin Renaud 
 */
public final class DSAPrivateKey extends PKCS8Key 
implements java.security.interfaces.DSAPrivateKey, Serializable {

    /* the private key */
    private BigInteger x;

    /**
     * This constructor is for use only with PKCS8Key factory
     * functionality and should not be called directly.
     */
    public DSAPrivateKey() {
    }

    /**
     * A constructor which takes a PKCS#8/DER encoded byte array.
     *
     * @param encodedKey an PKCS#8/DER encoded byte array.
     * 
     * @exception InvalidKeyException when the encoded key is invalid
     * or when it is improperly encoded.
     * 
     * @see sun.security.pkcs.PKCS8Key#decode 
     */
    public DSAPrivateKey(byte[] encodedKey) throws InvalidKeyException {
	decode(encodedKey);
    }

    /**
     * Make a DSA private key out of three parameters and a private key.
     */
    public DSAPrivateKey(BigInteger x, BigInteger p, 
			 BigInteger q, BigInteger g) 
    throws InvalidKeyException {
	super();
	algid = new AlgIdDSA(p, q, g);

	try {
	    key = new DerValue(DerValue.tag_Integer, 
			       x.toByteArray()).toByteArray();
	    encode();
	} catch (IOException e) {
	  throw new InvalidKeyException ("counld not DER encode y: " +
					 e.getMessage());
	}

	this.x = x;
    }

    /**
     * Get the DSA key's public algorithm parameters.
     */
    public DSAParams getParams() {
	try {
	    if (algid instanceof DSAParams) {
		return (DSAParams)algid;
		
	    } else {
		byte[] encoded = algid.encode();
		algid = new AlgIdDSA(encoded);
		return (DSAParams)algid;
	    }

	} catch (IOException e) {
	    throw new ProviderException("unable to parse algorithm params.");
	}
    }
    
    /**
     * Get X.
     */
    public BigInteger getX() {
	return x;
    }

    /**
     * Hook provided by PKCS8 base class to allow conversion of
     * key and other bits to a convenient format.
     */
    protected void parseKeyBits() 
    throws InvalidKeyException {
	DerInputStream	in = new DerInputStream (key);

	try {
	    x = in.getInteger().toBigInteger();

	} catch (IOException e) {
	    throw new InvalidKeyException(e.getMessage());
	}
    }

    byte[] getEncodedKey() throws InvalidKeyException {
	return getEncoded();
    }


    public String toString() {

	return "Sun DSA Private Key \nparameters:" + algid + "\nx: " +
	    x.toString(16) + "\n";
    }

}
