/*
 * @(#)DSAPublicKey.java	1.44 99/03/08
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.security.provider;

import java.util.*;
import java.io.*;
import java.math.BigInteger;
import java.security.InvalidKeyException;
import java.security.interfaces.DSAParams;

import sun.security.x509.X509Key;
import sun.security.x509.AlgIdDSA;
import sun.security.util.BigInt;
import sun.security.util.DerValue;
import sun.security.util.DerInputStream;
import sun.security.util.DerOutputStream;

/**
 * An X.509 public key for the Digital Signature Algorithm..
 *
 * @see DSAPrivateKey
 * @see AlgIdDSA
 * @see DSA
 * 
 * @author Benjamin Renaud 
 */
public final class DSAPublicKey extends X509Key 
implements java.security.interfaces.DSAPublicKey, Serializable {

    /* Are we debugging */
    private static boolean debug = false;

    /* the public key */
    private BigInteger y;

    /**
     * This constructor is for use only with X509Key factory
     * functionality and should not be called directly.
     */
    public DSAPublicKey() {
    }

    /**
     * A constructor which takes an X.509/DER encoded DSA public key.
     *
     * @param encodedKey an X.509/DER encoded DSA public key..
     * 
     * @see #decode */
    public DSAPublicKey(byte[] encodedKey) throws InvalidKeyException {
	decode(encodedKey);
    }

    /**
     * Make a DSA public key out of three parameters and a public key.
     */
    public DSAPublicKey(BigInteger y, 
			BigInteger p, BigInteger q, BigInteger g) 
    throws InvalidKeyException {
	super();
	this.y = y;
	algid = new AlgIdDSA(p, q, g);

	try {
	    key = new DerValue(DerValue.tag_Integer, 
			       y.toByteArray()).toByteArray();
	    
	} catch (IOException e) {
	  throw new InvalidKeyException ("could not DER encode y: " +
					 e.getMessage());
	}
    }

    /**
     * Get the y value of the key. This is the raw public value, y, without
     * the parameters.
     *
     * @see getParams
     */
    public BigInteger getY() {
	return y;
    }

    /**
     * Return the DSA parameters associated with this key.
     */
    public DSAParams getParams() {

	return (DSAParams)algid;
    }

    protected void parseKeyBits() throws InvalidKeyException {

	try {
	    DerInputStream in = new DerInputStream (key);
	    y = in.getInteger().toBigInteger();
	    
	} catch (IOException e) {
	    debug(e);
	    throw new InvalidKeyException("Invalid key: y value\n" +
					  e.getMessage());
	}
    }

    public String toString() {

	return "Sun DSA Public Key \nparameters:" + algid + "\ny: " +
	    y.toString(16) + "\n";
    }

    private static void debug(Throwable t) {
	if (debug) {
	    t.printStackTrace();
	}
    }
}
