/*
 * @(#)Main.java	1.96 98/11/10
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
 
package sun.security.provider;

import java.io.*;
import java.util.*;
import java.util.zip.*;
import java.math.BigInteger;
import java.security.*;
import java.security.interfaces.*;

import sun.net.www.MessageHeader;
import sun.security.x509.*;
import sun.security.pkcs.*;
import sun.security.util.*;
import sun.tools.jar.Manifest;
import sun.tools.jar.SignatureFile;

import sun.io.FileIO;
import sun.io.FileIOFactory;

/**
 * <p>The javakey utility. It is the SUN provider tool for managing
 * signers, identities, their keys and their certificates.
 *
 * @version 1.96 11/10/98
 * @author Benjamin Renaud 
 */

public class Main {

    /* Is debugging on */
    private boolean debug = false;

    /* Is this being used as a program, or as a programatically
       controlled instance. Used to error handling: calling exit or
       not. This class does yet behave properly when quitOnError is
       false. */
    private boolean quitOnError = false;

    String program;
    String usage;

    PrintStream out;
    PrintStream err;
    InputStream in;


    /* The system scope */
    protected IdentityScope scope;

    /**
     * Constructor.
     */
    public Main(PrintStream out, PrintStream err, InputStream in, 
		   String program, boolean quitOnError) {
	this.out = out;
	this.err = err;
	this.program = program;
	this.quitOnError = quitOnError;
    }

    /**
     * Set javakey's scope
     */
    protected void setScope(IdentityScope scope) {
	this.scope = scope;
    }

    /**
     * List all identities in the database. The syntax is as follows:
     *
     * -l lists all identities in short notation
     * -ld lists all identities in detailed notation
     * -li <id> lists a given identity in detailed notation
     */
    void listCmd(String[] args) {
	
	String action = args[0];
	boolean detailed = false;
	
	/* Are we listing a single identity? */
	if (action.equals("li")) {
	    if (args.length != 2) {
		listUsage();
		return;

	    } else {
		Identity identity = scope.getIdentity(args[1]);
		if (identity == null) {
		    error("No such identity: " + args[1] + " in scope " +
			  scope);

		} else {
		    list(identity);
		}
	    }

	} else if (action.equals("l") || 
		   (detailed = action.equals("ld"))) {
	    if (args.length != 1) {
		listUsage();
	    }
	    list(detailed);
	    
	} else {
	    listUsage();
	}
    }

    void listUsage() {
	String  usage = "";
	usage += "legal options for list:\n";
	usage += "\tno argument \tlist all identities in succint form.\n";
	usage += "\ti <identity>\tlist an identity in detail.\n";
	usage += "\td\t\tdetail: list all identities in detail.";
	
	out.println(usage);
	error("illegal arguments to list");
    }
    
    /**
     * Prints a list of all identity profiles in the present scope to
     * standard out.
     *
     * @param detailed whether or not to provider a details about each
     * identity 
     */
    public boolean list(boolean detailed) {

	out.println();
	out.println("Scope: " + scope);
	Enumeration e = scope.identities();
	while(e.hasMoreElements()) {
	    Identity identity = (Identity)e.nextElement();
	    out.println();
	    out.println(identity.toString(detailed));
	}
	return true;
    }

    /**
     * Prints a detailed provider of a given Identity to standard out.
     */
    public void list(Identity identity) {
	out.println("Identity: " + identity.getName());
	out.println(identity.toString(true));
    }

    /* Create an identity or a signer, optionally specifying 
       trusted or not.  */

    void createCmd(String[] args) {
	boolean success = false;
	
	String action = args[0];
	boolean trust = false;
	int argLen = args.length;

	boolean createIdentity = action.equalsIgnoreCase("c");
	boolean createSigner = action.equalsIgnoreCase("cs");
	boolean validArgs = argLen == 2 || argLen == 3;
	
	if (argLen == 3) {
	    trust = new Boolean(args[2]).booleanValue();
	}

	if (validArgs && createIdentity) {
	    success = createIdentity(args[1], trust);

	} else if (validArgs && createSigner) {
	    success = createSigner(args[1], trust);

	} else {
	    String  usage = "";
	    usage += "legal options for create:\n";
	    usage += "\tno arguments\t<name> {trusted}\tcreate " + 
		"a new identity.\n";
	    usage += "\ts\t\t<name> {trusted}\tcreate a new signer.\n";

	    out.println(usage);
	    error("illegal arguments to create");
	}
	if (success) {
	    String msg = "Created identity " + scope.getIdentity(args[1]);
	    save(scope, msg);
	}
    }
    
    public boolean createIdentity(String name, boolean trust) {
	try {
	    SystemIdentity id = new SystemIdentity(name, scope);
	    scope.addIdentity(id);
	    id.setTrusted(trust);

 	} catch (KeyManagementException e) {
 	    return error("An error occured trying to add " + name + " to " +
			 scope, e);

	} catch (InvalidParameterException e) {
 	    return error("Invalid parameter trying to add " + name + " to " + 
			 scope, e);
	}
	return true;
    }

    public boolean createSigner(String name, boolean trust) {
	try {

	    if (scope.getIdentity(name) != null) {
		return error("Name conflict. Please remove the identity " + 
			     "named " + name + 
			     " before adding a new identity with that name.");
	    }

	    SystemSigner id = new SystemSigner(name, scope);
	    scope.addIdentity(id);
	    id.setTrusted(trust);

 	} catch (KeyManagementException e) {
 	    return error("An error occured trying to add " + name + " to " + 
		  scope, e);
	}
	return true;
    }
    
    /* Remove an identity or a signer from the database */
    protected void removeCmd(String args[]) {
	String name = "";
	boolean success = false;

	if (args.length == 2) {

	    name = args[1];
	    if (scope.getIdentity(name) == null) {
		success = error("No one named " + name + " in the system.");
	    } else {
		success = remove(name);
	    }

	} else {
	    String  usage = "";
	    usage += "legal options for remove:\n";
	    usage += "\tno arguments\t<name>\tthe name of the identity " + 
		"to remove\n";

	    out.println(usage);
	    success = error("illegal arguments to remove");
	}
	if (success) {
	    save(scope, "Removed " + name + ".");
	}
    }

    /**
     * Removes the identity named name from the current scope. name
     * must be in the scope.
     */
    public boolean remove(String name) {
    	try {
	    Identity identity = scope.getIdentity(name);
	    if (identity == null) {
		return error(name + " not found in database.");
	    }
	    scope.removeIdentity(identity);
	} catch (KeyManagementException e) {
	    return internalError("remove " + name);
	}
	return true;
    }

    protected void exportCmd(String[] args) {
	boolean success = false;
	String msg = "";
	
	String action = args[0];
	int argLen = args.length;
	
	if (argLen < 3 || argLen > 4) {
	    exportUsageError();

	} else {
	    String name = args[1];

	    if (action.equals("ek")) {
		FileIO publicFile = FileIOFactory.newInstance(args[2]);
		FileIO privateFile = null;
		if (args.length == 4) {
		    privateFile = FileIOFactory.newInstance(args[3]);
		    
		}
		success = exportKeys(name, publicFile, privateFile);
		msg = "Public key exported to " + publicFile + ".";
		if (privateFile != null) {
		    msg += "\nPrivate key exported to " + privateFile + ".";
		}
	    } else if (action.equals("ec")) {
		if (argLen != 4) {
		    exportUsageError();
		}
		int certnum = Integer.parseInt(args[2]);
		FileIO certFile = FileIOFactory.newInstance(args[3]);
		success = exportCert(name, certnum, certFile);
		msg = "Certificate " + certnum + " exported to " + 
		    certFile + ".";
	    } else {
		exportUsageError();
	    }
	    if (success) {
		save(scope, msg);
	    }
	}
    }

    boolean exportKeys(String name, FileIO pub, FileIO priv) {
	try {
	    OutputStream fos;
	    BufferedOutputStream bos;

	    Identity id = scope.getIdentity(name);
	    if (id == null) {
		return error("There are no identities named " + name +
			     " in the system.");
	    }
	    if (id instanceof Identity) {
		PublicKey key = id.getPublicKey();
		if (key == null) {
		    return error("identity " + name + 
				 " does not have a public key.");
		} 
		fos = pub.getOutputStream();
		bos = new BufferedOutputStream(fos);
		bos.write(key.getEncoded());
		bos.flush();
		fos.close();
	    } 
	    if (id instanceof Signer && priv != null) {
		Signer signer = (Signer)id;
		fos = priv.getOutputStream();
		bos = new BufferedOutputStream(fos);
		PrivateKey key = signer.getPrivateKey();
		if (key == null) {
		    return error("signer " + name + 
				 " does not have a private key.");
		}
		bos.write(key.getEncoded());
		bos.flush();
		fos.close();
	    }
	} catch (IOException e) {
	    error("Error: io error", e);
	}
	return true;
    }

    boolean exportCert(String name, int certnum, FileIO certFile) {
	try {
	    OutputStream fos;
	    BufferedOutputStream bos;

	    Identity id = scope.getIdentity(name);
	    if (id == null) {
		return error("There are no identities named " + name + 
			     " in the system.");
	    }
	    Certificate[] certs = id.certificates();
	    if (certs.length < certnum) {
		return error("identity " + name + " does not have a cert " + 
			     certnum + ".");
	    }
	    Certificate cert = certs[certnum - 1];
	    fos = certFile.getOutputStream();
	    bos = new BufferedOutputStream(fos);
	    cert.encode(bos);
	    bos.flush();
	    fos.close();
	    
	} catch (IOException e) {
	    return error("Error: io error", e);
	} catch (KeyException e) {
	    return error("Key exception: invalid key", e);
	}
	return true;
    }

    private boolean exportUsageError() {
	String  usage = "";
	usage += "legal options for export (<id> is the identity " +
	    "for which the import is done.)\n";
	usage += "\tk\t<id> <pub> {priv}\t\texport a public key,\n\t\t " +
	    "(and optionally a private key) to a file.\n";
	usage += "\tc\t<id> <certnum> <certFile>\texport a cert " +
	    "to a  files.\n";
	out.println(usage);
	return error("illegal arguments to export.");
    }

    /**
     * Import a key or a cert. Syntax is: <identity/signer>
     * [pubkey,certificate] {privkey} 
     */
    protected void importCmd(String[] args) {
	boolean success = false;
	String msg = "";

	String action = args[0];
	int argLen = args.length;
	
	if (argLen < 2 || argLen > 4) {
	    importUsageError();

	} else {
	    String name = args[1];
	    String file = "";
	    
	    if (argLen > 2) {
		file = args[2];
	    }
	    /* Import an X509 certificate */
	    if (action.equals("ic")) {
		success = importCertificate(name, file);
		if (success)
		    msg = "Imported certificate from " + file + " for " +
			name + ".";

	    /* Import a public key */
	    } else if (action.equals("ik")) {
		success = importPublicKey(name, file);
		if (success)
		    msg = "Set public key from " + file + " for " + name + ".";

	    /* Import a key pair */
	    } else if (action.equals("ikp") && argLen == 4) {
		success = setKeyPair(name, file, args[3]);
		if (success) {
		    msg = "Set public key from " + file + " for " + name + ".";
		    msg += "\nSet private key from " + args[3] + 
			" for " + name + ".";
		}

	    /* Import some textual information */
	    } else if (action.equals("ii") && argLen == 2) {
		success = setInfo(name);
		if (success) msg = "Set information for " + name;
		
	    } else {
		success = importUsageError();
	    }
	    if (success) {
		save(scope, msg);
	    }
	}
    }

    private boolean importUsageError() {
	String  usage = "";
	usage += "legal options for import (<id> is the identity for " + 
	    "which the import is done.)\n";
	usage += "\tk\t<id> <pub>\t\timport a public key from a file.\n";
	usage += "\tkp\t<id> <pub> <priv>\timport a key pair " + 
	    "from two files.\n";
	usage += "\tc\t<id> <cert>\t\timport a certificate from a file.\n";
	out.println(usage);
	return error("illegal arguments to import.");
	}

    private boolean setInfo(String name) {

	Identity id = (Identity)scope.getIdentity(name);
	if (id == null) {
	    return error("There are no identities named " + id + 
			 " in the system.");
	}
	out.println("Please enter the info for this identity.");
	out.println("[End with ctrl-D on Unix and ctrl-Z on Win32]");
	DataInputStream dis = new DataInputStream(System.in);
	String info = "";
	String line = "";
	try {
	    while((line = dis.readLine()) != null) {
		info += "\n" + line;
	    }
	} catch (IOException e) {
	    debug(e);
	    return error("IOException importing information.");
	}
	id.setInfo(info);
	return true;
    }

    public boolean importCertificate(String name, String certfile) {
	Identity id = null;

	try {
	    if (certfile == null || certfile.length() == 0)
		return error("Certificate file missing");
	    FileIO file = FileIOFactory.newInstance(certfile);
	    if (!file.exists()) {
		return error("File " + certfile + "  not found.");
	    }
	    
	    DataInputStream dis = new DataInputStream(file.getInputStream());

	    byte[] data = new byte[dis.available()];
	    dis.readFully(data);
	    X509Cert cert = new X509Cert(data);
	    
	    id = scope.getIdentity(name);
	    if (id == null) {
		return error(name + " not found in the system.");
	    }
	    importCertificate(id, cert);

	} catch (IOException e) {
	    return error("Error: io error", e);

	} catch (KeyManagementException e) {
	    return error("key management exception: " + e.getMessage(), e);

	} catch (CertException e) {
	    return error("invalid cert: " + certfile, e);
	}
	return true;
    }

    boolean importCertificate(Identity identity, Certificate cert) 
    throws KeyManagementException {

	PublicKey certKey = cert.getPublicKey();
	PublicKey identityKey = identity.getPublicKey();

	if (identityKey == null) {
	    identity.setPublicKey(certKey);

	} else {
	    if (! identityKey.equals(certKey)) {
		return error("Certificate key does not match identity key.");
	    }
	}
	identity.addCertificate(cert);
	return true;
    }

    boolean importPublicKey(String name, String keyfile) {
	try {
	    Identity identity = scope.getIdentity(name);
	    if (identity == null) {
		return error(name + " not found in database.");
	    }

	    if (keyfile == null || keyfile.length() == 0)
		return error("Public-key file missing");
	    FileIO file = FileIOFactory.newInstance(keyfile);
	    if (!file.exists()) {
		return error("File " + keyfile + " not found.");
	    }

	    DataInputStream dis = new DataInputStream(file.getInputStream());

	    byte[] data = new byte[dis.available()];
	    dis.readFully(data);
	    DerValue keyData = new DerValue(data);
	    X509Key key = X509Key.parse(keyData);
	    
	    PublicKey identityKey = identity.getPublicKey();
	    if (identityKey != null && !key.equals(identityKey)) {
		message(identity + " already has a public key." + 
			"\nClearing any exisiting certificates...\n");
		Certificate[] certs = identity.certificates();
		for (int i = 0; i < certs.length; i++) {
		    identity.removeCertificate(certs[i]);
		}
	    }
	    identity.setPublicKey(key);

	} catch (IOException e) {
	    return error("Error: io error", e);
	} catch (KeyException e) {
	    return error("Key exception: invalid key", e);
	}
	return true;
    }
    
    boolean setKeyPair(String name, String pubfile, String privfile) {
	try {
	    if (pubfile == null || pubfile.length() == 0)
		return error("Public-key file missing");
	    FileIO pub = FileIOFactory.newInstance(pubfile);

	    if (privfile == null || privfile.length() == 0)
		return error("Private-key file missing");
	    FileIO priv = FileIOFactory.newInstance(privfile);

	    if (!pub.exists()) { 
		return error("File " + pub + " not found.");
	    }
	    if (!priv.exists()) { 
		return error("File " + priv + " not found.");
	    }

	    InputStream fis = pub.getInputStream();
	    DataInputStream dis = new DataInputStream(fis);
	    byte[] data = new byte[dis.available()];
	    dis.readFully(data);
	    fis.close();

	    DerValue keyData = new DerValue(data);
	    X509Key pubkey = X509Key.parse(keyData);

	    fis = priv.getInputStream();
	    dis = new DataInputStream(fis);
	    data = new byte[dis.available()];
	    dis.readFully(data);
	    fis.close();
	    keyData = new DerValue(data);

	    PKCS8Key privkey = PKCS8Key.parse(keyData);
	    KeyPair keyPair = new KeyPair(pubkey, privkey);
	    
	    return setKeyPair(name, keyPair);
	} catch (IOException e) {
	}
	return true;
    }

    boolean setKeyPair(String name, KeyPair keyPair) {	

	try {
	    Signer signer = (Signer)scope.getIdentity(name);
	    if (signer == null) {
		return error("Signer " + name + " not found in system.");
	    }
	    signer.setKeyPair(keyPair);

	} catch (KeyException e) {
	    return error("Key exception: invalid key", e);

	} catch (InvalidParameterException e) {
	    return error("Internal error!", e);	    
	}
	return true;
    }


    private void save(IdentityScope scope, String msg) {
	if (scope instanceof IdentityDatabase) {
	    try { 
		IdentityDatabase db = (IdentityDatabase)scope;
		db.save();
	    } catch (IOException e) {
		error("IOException trying to save the database.", e);
	    }
	} else {
	    out.println("Cannot save the database: not a SUN database.");
	}
	exit(msg);
    }

    private void save(IdentityScope scope) {
	save(scope, "Operation succesful.");
    }


    /* Generate key pairs, certs and signatures. Syntax is:
     *
     * key pairs:
     * gk alg modsize {<pub>} {<priv>}
     *
     * certs
     * gc <certdir>
     * 
     * sig
     * gs <sigdir> <file>
     */
    private void generateCmd(String[] args) {
	boolean success = false;
	String msg = "";

	String action = args[0];
	int argLen = args.length;

	Properties directive = new Properties();
	FileIO directiveFile = null;

	if (argLen == 2 || argLen == 3) {
	    /* We load the directive file. */
	    directiveFile = FileIOFactory.newInstance(args[1]);
	    try {
		InputStream is = directiveFile.getInputStream();
		directive.load(is);
	    } catch (IOException e) {
		error("IOException loading directive file " + directiveFile);
	    } 
	}
	
	if (action.equals("gk") || action.equals("g")) {
	    if (argLen == 2) {
		success = generateKeyPair(directive);
	    } else if (argLen < 7 && argLen > 3) {
		directive.put("signer.name", args[1]);
		directive.put("algorithm", args[2]);
		directive.put("strength", args[3]);
		if (argLen > 4) {
		    directive.put("pub", args[4]);
		} 
		if (argLen > 5) {
		    directive.put("priv", args[5]);
		}
		success = generateKeyPair(directive);
	    }
	    msg = "Generated " + directive.get("algorithm") + " keys for " + 
		directive.get("signer.name") + " (strength: " + 
		directive.get("strength") + ").";
	    String pub = directive.getProperty("pub");
	    String priv = directive.getProperty("priv");
	    if (pub != null) {
		msg += "\nSaved public key to " + pub + ".";
	    }
	    if (priv != null) {
		msg += "\nSaved private key to " + priv + ".";
	    }		    
	} else if (action.equals("gc")) {
	    if (argLen != 2) {
		usageError("generate certificate only takes one " + 
			   "argument, the directive file.");
	    } else {
		success = generateCertificate(directive);
		msg = "Generated certificate from directive file " + 
		    directiveFile + ".";
	    }

	} else if (action.equals("gs")) {
	    try {
		if (args.length != 3) {
		    usageError("wrong number of arguments. " +
			       "gs takes 2 arguments: " + 
			       "the directive and jar file.");
		} else {
		    FileIO jarFile = FileIOFactory.newInstance(args[2]);
		    if (!jarFile.exists()) {
			error("JAR file " + jarFile + " not found.");
		    }
		    success = signJar(directive, jarFile);
		    msg = "Signed JAR file " + jarFile + 
			" using directive file " + directiveFile + ".";
		}
	    } catch (Exception e) {
		error("error generating signature: " + e, e);
	    }
	} else {
	    generateUsageError();
	}
	if (success) {
	    save(scope, msg);
	}
    }

    private boolean generateUsageError() {
	String  usage = "";
	usage += "legal options for generate:\n";
	usage += "\tk\t<id> <alg> <mod>\t\tgenerate a key pair.\n";
	usage += "\tc\t<certdir> \tgenerate a certificate.\n";
	usage += "\ts\t<sigdir> <file>\t\tgenerate a signature.\n";
	out.println(usage);
	return error("illegal arguments to generate.");
    }

    boolean generateCertificate(Properties directive) {

	int certnum = -1;

	BigInt serial = null;
	Date start = null;
	Date end = null;
	X509Cert issuerCert = null;

	PrivateKey privateKey = null;
	X509Key publicKey = null;

	String signingAlgorithm = "DSA";

	try {
		
	    /* We get all the properties we need. */
	    String issuerName = assertProp("issuer.name", directive);
	    String subjectName = assertProp("subject.name", directive);

	    /* If we're not self signed, we need a CA cert */
	    String issuerCertNum = null;
	    boolean selfSigned = (issuerName.equals(subjectName));
	    if (!selfSigned) {
		issuerCertNum = assertProp("issuer.cert", directive);
	    }
	    
	    String subjectCN = assertProp("subject.real.name", directive);
	    String subjectOU = assertProp("subject.org.unit", directive);
	    String subjectO = assertProp("subject.org", directive);
	    String subjectC = assertProp("subject.country", directive);

	    String startDate = assertProp("start.date", directive);
	    String endDate = assertProp("end.date", directive);
	    String serialNum = assertProp("serial.number", directive);

	    signingAlgorithm = directive.getProperty("signature.algorithm");

	    /* The subject distinguised name */
	    X500Name dnSubject = new X500Name(subjectCN, subjectOU,
					      subjectO, subjectC);
	    
	    /* Try parsing the date  and the cert num */
	 
	    try {
		start = new Date(startDate);
		end = new Date(endDate);
		Date now = new Date();

		if (now.after(end)) {
		    out.println("warning: certificate's expiration date is earlier " +
				"than the current date.");
		}
		if (end.before(start)) {
		    out.println("warning: certificate's expiration date is earlier " +
				"than the start date.");
		}   
		if (!selfSigned) {
		    certnum = Integer.parseInt(issuerCertNum) - 1;
		}
		
		BigInteger num = new BigInteger(serialNum);
		serial = new BigInt(num.toByteArray());

	    } catch (NumberFormatException e) {
		return error("Error parsing issuer.cert number.");
	    }
	    catch (IllegalArgumentException e) {
		return error("Date formatted incorrectly.");

	    } catch (Exception e) {
		debug(e);
		return error("error signing: " + e.getMessage());
	    }

	    /* Try to get the signer, and check that he's got all he
               needs. */
	    Signer signer = (Signer)scope.getIdentity(issuerName);
		
	    if (signer == null) {
		return error("no signer " + issuerName + 
			     " found in the database.");
	    }
	    /* no matter what, we need a private key */
	    privateKey = signer.getPrivateKey();

	    if (privateKey == null) {
		return error(signer + " does not have a private key.");
	    }

	    if (signingAlgorithm == null) {
		signingAlgorithm = "DSA";
	    }

	    if (!selfSigned) {
		Certificate[] certs = signer.certificates();
		if (certnum < certs.length) {
		    Certificate tmpCert = certs[certnum];

		    if (tmpCert == null ||
			(! (tmpCert instanceof X509Cert))) {
			return error("invalid cert for signer");
		    } else {
			issuerCert = (X509Cert)tmpCert;
		    }
		} else {
		    return error("No such cert: " + (certnum + 1) + 
				 " for issuer " + signer + ".");
		}
	    }
	    try {
		if (privateKey == null) {
		    return error("Signer " + signer + " has no signing key.");
		}
		/* The victim needs a public key */
		Identity identity = scope.getIdentity(subjectName);

		if (signer == null) {
		    return error("No such guarantor: " + issuerName);

		} else if (identity == null) {
		    return error("No such identity: " + subjectName);
		}

		PublicKey pubKey = identity.getPublicKey();

		if (pubKey == null) {
		    return error(identity + " does not have a public key.");
		}

		if (! (pubKey.getFormat().equals("X.509"))) {
		    return error("Public key format is not X.509");
		}
		/* If it's not an X.509 key instance, build one */
		try {
		    if (pubKey instanceof X509Key) {
			publicKey = (X509Key)pubKey;

		    } else {
			DerValue der = new DerValue(pubKey.getEncoded());
			publicKey = X509Key.parse(der);
		    }
	    
		} catch (IOException e) {
		    return error("Bad X509 key: " + e.getMessage(), e);
		}

		/* The basic cert */
		X509Cert cert = new X509Cert(dnSubject, publicKey, 
					     start, end);

		if (selfSigned) {
		    issuerCert = cert;
		}
		
		AlgorithmId algid = AlgorithmId.get(signingAlgorithm);

		X500Signer x500Signer = issuerCert.getSigner(algid, 
							     privateKey);
				    
		byte[] certBytes = cert.encodeAndSign(serial, x500Signer);
		X509Cert signedCert = new X509Cert(certBytes);

		identity.addCertificate(signedCert);

		// save it to file if spec'ed
		String out = directive.getProperty("out.file");
		if (out != null) {
		    FileIO outfile = FileIOFactory.newInstance(out);
		    signedCert.encode(outfile.getOutputStream());
		}
	    
	    } catch (IOException e) {
		return error("IO exception during cert signing", e);

	    } catch (SignatureException e) {
		return error("A signature exception occured", e);

	    } catch (KeyException e) {
		return error("A key exception occured", e);

	    } catch (NoSuchAlgorithmException e) {
		return error("algorithm " + signingAlgorithm + " not available", e);
	    }
	} catch (IOException e) {
	    e.printStackTrace();
	    return false;
	}
	return true;
    }

    boolean generateKeyPair(Properties directive) {
	String name = assertProp("signer.name", directive);
	String strength = assertProp("strength", directive);
	String algorithm = assertProp("algorithm", directive);
	
	String provider = directive.getProperty("provider");
	String pub = directive.getProperty("pub");
	String priv = directive.getProperty("priv");
	
	if (provider != null)
	    provider = provider.trim();
	if (pub != null)
	    pub = pub.trim();
	if (priv != null)
	    priv = priv.trim();

	Identity identity = scope.getIdentity(name);
	if (identity == null) {
	    error(name + " not found in system.");
	} else if (! (identity instanceof Signer)) {
	    error("Identity " + identity + " not a signer.");
	} else {
	    Signer signer = (Signer)identity;
	    int str = Integer.parseInt(strength);

	    byte[] seed = machineSeed(64);

	    KeyPairGenerator keyGen = null;
	    KeyPair keyPair = null;

	    try {
		keyGen = KeyPairGenerator.getInstance(algorithm, provider);
		keyGen.initialize(str, new SecureRandom());
		keyPair = keyGen.generateKeyPair();
		
	    } catch (NoSuchAlgorithmException e) {
		return error("Algorithm " + algorithm + " not available.");	

	    } catch (Exception e) {
		return error("Invalid algorithm params for key generation.", e);
	    }
	    try {
		signer.setKeyPair(keyPair);
	    } catch (InvalidParameterException e) {
		err.println("Internal error!");
		e.printStackTrace(err);	
	    } catch (KeyException e) {
		err.println("Internal error!");
		e.printStackTrace(err);
	    }	    
	    if ((pub != null) && (pub.length() != 0)) {
		FileIO pubFile = FileIOFactory.newInstance(pub);
		if (!saveKey(keyPair.getPublic(), pubFile)) {
		    error("error saving public key to file " + pubFile);
		}
	    }
	    if ((priv != null) && (priv.length() != 0)) {
		FileIO privFile = FileIOFactory.newInstance(priv);
		if (!saveKey(keyPair.getPrivate(), privFile)) {
		    error("error saving private key to file " + privFile);
		}
	    }
	    return true;
	}
	return false;
    }

    private boolean saveKey(Key key, FileIO file) {

	try {
	    OutputStream fos = file.getOutputStream();
	    fos.write(key.getEncoded());
	    fos.close();
	} catch (IOException e) {
	    return error("IOException saving keys.", e);	

	} catch (ClassCastException e) {
	    return error("Not a SUN key- cannot save", e);
	}
	return true;
    }

    boolean signJar(Properties directive, FileIO jarFile) throws Exception {

	/* Parsing and checking directive file entries. */
	String sfNameStr = assertProp("signature.file", directive);

	if (sfNameStr == null) {
	    return false;
	} else {
	    sfNameStr = sfNameStr.toUpperCase();
	}

	/* output file name. See if they asked for one, and make sure
           it's not the same as the jar file itself. */
	String jarName = jarFile.getName();
	String signedJarName = directive.getProperty("out.file");
	if (signedJarName != null)
	    signedJarName = signedJarName.trim();

	if (signedJarName == null ||
	    signedJarName.equals(jarName)) {
	    signedJarName = jarName + ".sig";
	}
	FileIO signedJarFile = FileIOFactory.newInstance(signedJarName);

	/* Done parsing arguments */
	Manifest manifest;

	ZipInputStream zis = new ZipInputStream(jarFile.getInputStream());
	    
	PrintStream ps = new PrintStream(signedJarFile.getOutputStream());
	ZipOutputStream zos = new ZipOutputStream(ps);
	    
	ZipEntry entry = zis.getNextEntry();	    

	/* First guess at what they might be - we don't xclude RSA ones. */

	String sfFilename = 
	    ("META-INF/" + sfNameStr + ".SF").toUpperCase();
	String bkFilename = 
	    ("META-INF/" + sfNameStr + ".DSA").toUpperCase();

	while(entry != null) {
	    if (Manifest.isManifestName(entry.getName())) {
		byte[] data = getBytesToEOF(zis);
		manifest = new Manifest(data);

		// re-output the manifest
		writeEntry(new ByteArrayInputStream(data), zos, entry);

		// if sf is null, an error has occured.
		SignatureFile sf = computeSignatures(manifest, directive);
		if (sf == null) {
		    return false;
		}

		PKCS7 bk = sf.getBlock();
		
		sfFilename = sf.getName();
		bkFilename = sf.getBlockName();

		ZipEntry sfFile = new ZipEntry(sf.getName());
		ZipEntry bkFile = new ZipEntry(sf.getBlockName());
		
		Date date = new Date();
		sfFile.setTime(date.getTime());
		bkFile.setTime(date.getTime());
		
		zos.putNextEntry(sfFile);
		sf.stream(zos);
		out.println("Creating entry: " + sfFile.getName());

		zos.putNextEntry(bkFile);
		bk.encodeSignedData(zos);
		out.println("Creating entry: " + bkFile.getName());
		
	    } else if (entry.getName().equalsIgnoreCase(sfFilename) || 
		       entry.getName().equalsIgnoreCase(bkFilename)) {
		out.println("(Replaced existing entry: " + entry.getName() +
			    ")");
	    } else {
		writeEntry(zis, zos, entry);
	    }
	    entry = zis.getNextEntry();
	}
	/* This will close fos as well */
	zos.close();
	
	return true;
    }

    private byte[] getBytesToEOF(InputStream is) throws IOException {

	/* Note that we can't buffer here, because InflaterInputStream gets
	   upset when we do a read beyond a compression boundary. */
	ByteArrayOutputStream out = new ByteArrayOutputStream();

	byte[] buf = new byte[1024];
	int len;
	while ((len = is.read(buf, 0, buf.length)) != -1) {
	    out.write(buf, 0, len);
	}
	return out.toByteArray();
    }
    
		    
    private void writeEntry(InputStream is, ZipOutputStream os, ZipEntry e) 
    throws IOException {
	out.println("Adding entry: " + e.getName());
	byte[] data = getBytesToEOF(is);
	/*
	byte[] data = new byte[(int)e.getSize()];
	is.read(data);
	*/

	os.putNextEntry(e);
	if (data.length > 0) {
	    os.write(data);
	}
    }
		    
    SignatureFile computeSignatures(Manifest manifest, Properties directive) 
    throws IOException, InvalidKeyException, SignatureException,
	   NoSuchAlgorithmException {

	String sfNameStr = assertProp("signature.file", directive);
	String signerStr = assertProp("signer", directive);
	String certNumberStr = assertProp("cert", directive);
	String chainNumberStr = assertProp("chain", directive);

	Identity id  = scope.getIdentity(signerStr);
	Signer signer = null;

	if (id == null) {
	    error(signerStr +  " not found in database.");	    
	    return null;
	}
	if (id instanceof Signer) {
	    signer = (Signer)id;
	} else {
	    error(id + " is not a Signer.");
	    return null;
	}
	PrivateKey privateKey = signer.getPrivateKey();

	if (privateKey == null) {
	    error(signer + " does not have a public key.");	    
	    return null;
	}
	int chain = 0;

	if (chainNumberStr != null) {
	    chain = Integer.parseInt(chainNumberStr);
	}
	if (certNumberStr == null) {
	    error("illegal directive file");
	    return null;    
	}
	if (sfNameStr.length() > 8) {
	    error("signature file specified longer than 8 characters.");
	    return null;
	}
	int certNumber = Integer.parseInt(certNumberStr);
	
	Certificate[] certs = signer.certificates();
	
	if (certNumber > certs.length || certNumber < 1) {
	    error("Invalid cert number (" + certNumber + 
		  "). Valid numbers are 1 through " + certs.length + ".");
	    return null;
	}
	Certificate cert = certs[certNumber - 1];

	if (!(cert instanceof X509Cert)) {
	    error("only X509 certificate format supported " + 
			 "for signing JAR");
	    return null;
	}

	SignatureFile sf = new SignatureFile(manifest, sfNameStr);

	// compute PKCS7 block and save the .DSA file.
	PKCS7 pkcs7 = computePKCS7(sf, privateKey,
				   (X509Cert)cert);

	sf.setBlock(pkcs7);
	return sf;
    }

    /**
     * This computes the signature file and returns the PKCS7
     * block for it.
     */
    private PKCS7 computePKCS7(SignatureFile sf, PrivateKey privateKey, 
			       X509Cert cert) 
    throws IOException, InvalidKeyException, SignatureException,
    NoSuchAlgorithmException {
	X500Name name = cert.getSubjectName();
	BigInt serial = cert.getSerialNumber();
	String keyAlgorithm = privateKey.getAlgorithm();

	String digestAlgorithm = "SHA";
	String signatureAlgorithm = "SHA/DSA";

	if (keyAlgorithm.equals("DSA")) {
	} else if (keyAlgorithm.equals("RSA")) {
	    signatureAlgorithm = "MD5/RSA";
	    digestAlgorithm = "MD5";
	} else {
	    error("private key is not a DSA or RSA key.");
	}

	AlgorithmId digestAlg = AlgorithmId.get(digestAlgorithm);
	AlgorithmId sigAlg = AlgorithmId.get(signatureAlgorithm);

	Signature sig = Signature.getInstance(signatureAlgorithm);
	sig.initSign(privateKey);

	ByteArrayOutputStream baos = new ByteArrayOutputStream();
	sf.stream(baos);

	byte[] bytes = baos.toByteArray();
	ContentInfo contentInfo = new ContentInfo(bytes);
	
	sig.update(bytes);
	byte[] signature = sig.sign();

	SignerInfo signerInfo = new SignerInfo(name, serial, 
					       digestAlg, sigAlg,
					       signature);
	    
	AlgorithmId[] algs = {digestAlg};
	SignerInfo[] infos = {signerInfo};
	X509Cert[] certs = {cert};
	    
	return new PKCS7(algs, contentInfo, certs, infos);
    }

    void displayCmd(String[] args) {

	if ((! args[0].equals("dc")) || args.length != 2) {
	    String  usage = "";
	    usage += "legal options to display certificate:\n";
	    usage += "\tc\t<cert>\tdisplay a certificate.\n";
	    out.println(usage);
	    error("illegal arguments to display.");

	} else {
	    displayCertificate(args[1]);
	}
    }
	

    /**
     * display a cert.
     */
    private void displayCertificate(String filename) {
	FileIO file = FileIOFactory.newInstance(filename);
	try { 
	    if (!file.exists()) {
		error("File " + file + " not found.");

	    } else {
		DataInputStream dis = new DataInputStream(file.getInputStream());
		byte[] data = new byte[dis.available()];
		dis.readFully(data);

		X509Cert cert = new X509Cert(data);
		out.println(cert);
	    }
	} catch (IOException e) {
	    error("IO error", e);
	}
    }

    private void trustCmd(String args[]) {
	boolean success = false;

	if (!(args.length == 3)) {
	    usageError("Invalid number of arguments for set trust.");
	} else {
	    String name = args[1];
	    Boolean trust = new Boolean(args[2]);
	    Identity id = scope.getIdentity(name);

	    if (id == null) {
		error(name + " not found in the database.");
	    } else {
		if (setTrusted(id, trust.booleanValue())) {
		    save(scope, name + " is trusted: " + trust);
		}
	    }
	}
    }

    private boolean setTrusted(Identity id, boolean trusted) {
	if (id instanceof SystemIdentity) {
	    SystemIdentity sid = (SystemIdentity)id;
	    sid.setTrusted(trusted);

	} else if (id instanceof SystemSigner) {
	    SystemSigner signer = (SystemSigner)id;
	    signer.setTrusted(trusted);

	} else {
	    return error(id + " is not a SUN identity.");
	}
	return true;
    }

    void run(String[] args) throws KeyManagementException {
	if (args.length < 1) {
	    usageError("for more information, see documentation.");
	    return;
	}
	/* Accept the -arg syntax, but don't require it. */
	String arg = args[0];
	if (arg.startsWith("-")) {
	    arg = arg.substring(1);
	}
	args[0] = arg;

	switch (arg.charAt(0)) {

	case 'l':
	    listCmd(args);
	    break;
	case 'c':
	    createCmd(args);
	    break;
	case 'r':
	    removeCmd(args);
	    break;
	case 'i':
	   importCmd(args);
	    break;
	case 'g':
	    generateCmd(args);
	    break;
	case 'd':
	    displayCmd(args);
	    break;
	case 't':
	    trustCmd(args);
	    break;
	case 'v':
	    //	    verifyCmd(args);
	    break;
	case 'e':
	    exportCmd(args);
	    break;
	default:
	    usage();
	    break;
	}
    }

    public static void main(String args[]) throws KeyManagementException {

	Main main = new Main(System.out, System.err, System.in, "javakey",
			     true);
	main.setScope(IdentityScope.getSystemScope());
	if (main.scope == null) {
	    main.error("could not initialize scope");
	}

	main.run(args);
    }

    byte[] machineSeed(int numBytes) {
	return SecureRandom.getSeed(numBytes);
    }

    /* Should we use user seed anyway? Make that an option. -br */
    byte[] userSeed(int numBytes) {
	out.println("Please enter some random data (" + numBytes + 
		    " key strokes) and hit return.");
	InputStream in = System.in;
	
	byte[] result = new byte[numBytes];
	int count = 0;
	String seed = "";

	try {
	    while(count < numBytes) {
		result[count++] = (byte)in.read();
	    }

	} catch (IOException e) {
	    err.println("IOException");
	    debug(e);
	}
	return result;
    }

    void usage() {
	if (usage == null) {
	    usage = "";
	    usage += "javakey\n";
	    usage += "\tl\tlist of the identities in the database.\n";
	    usage += "\tc\tcreate an new identity.\n";
	    usage += "\tr\tremove an identity from the database.\n";
	    usage += "\ti\timport a public key, a key pair, etc.\n";
	    usage += "\tg\tgenerate a key pair, a certificate, etc.\n";
	    usage += "\td\tdisplay a certficate.\n";
	}
	out.println(usage);
    }

    boolean usageError(String msg) {
	usage();
	return error(msg);
    }

    boolean error(String msg) {
	err.println(msg);
	if (quitOnError) {
	    System.exit(1);
	}
	return false;
    }

    boolean internalError(String msg) {
	err.println("Internal error! " + msg);
	err.println("please submit a bug report via " +
		    System.getProperty("java.vendor.url.bug"));
	if (quitOnError) {
	    System.exit(1);
	}
	return false;
    }

    boolean error(String msg, Throwable t) {
	debug(t);
	return error(msg + "[" + t.getMessage() +"]");
    }

    void debug(Throwable t) {
	if (debug) {
	    t.printStackTrace(err);
	}
    }

    void debug(String s, Throwable t) {
	if (debug) {
	    t.printStackTrace(err);
	    System.err.println(s);
	}
    }

    void debug(String s) {
	if (debug) {
	    System.err.println(s);
	}
    }

    void message(String msg) {
	out.print(msg);
    }

    void exit(String msg) {
	out.println(msg);
	if (quitOnError) {
	    System.exit(0);
	}
    }

    String assertProp(String key, Properties props) {
	
	String data = props.getProperty(key);
	if (data != null)
	    data = data.trim();

	if ((data == null) || (data.length() == 0))
	    error(key + " must be specified in the directive file.");

	return data;
    }

}    

