/*
 * @(#)BigInt.java	1.21 98/07/01
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.security.util;

import java.math.BigInteger;


/**
 * A low-overhead arbitrary-precision <em>unsigned</em> integer.
 * This is intended for use with ASN.1 parsing, and printing of
 * such parsed values.  Convert to "BigInteger" if you need to do
 * arbitrary precision arithmetic, rather than just represent
 * the number as a wrapped array of bytes.
 *
 * <P><em><b>NOTE:</b>  This class may eventually disappear, to
 * be supplanted by big-endian byte arrays which hold both signed
 * and unsigned arbitrary-precision integers.
 *
 * @version 1.21
 * @author David Brownell
 */
public final
class BigInt
{
    // Big endian -- MSB first.
    private byte		places [];

    /**
     * Constructs a "Big" integer from a set of (big-endian) bytes.
     * Leading zeroes should be stripped off.
     *
     * @param data a sequence of bytes, most significant bytes/digits
     *		first.  CONSUMED.
     */
    public BigInt (byte data [])
	{ places = data; }


    /**
     * Constructs a "Big" integer from a "BigInteger", which must be
     * positive (or zero) in value.
     */
    public BigInt (BigInteger i)
    {
	byte	temp [] = i.toByteArray ();

	if ((temp [0] & 0x80) != 0)
	    throw new IllegalArgumentException ("negative BigInteger");

	// XXX we assume exactly _one_ sign byte is used...

	if (temp [0] != 0)
	    places = temp;
	else {
	    places = new byte [temp.length - 1];
	    for (int j = 1; j < temp.length; j++)
		places [j - 1] = temp [j];
	}
    }


    /**
     * Constructs a "Big" integer from a normal Java integer.
     *
     * @param i the "small" integer
     */
    public BigInt (int i)
    {
	if (i < (1 << 8)) {
	    places = new byte [1];
	    places [0] = (byte) i;
	} else if (i < (1 << 16)) {
	    places = new byte [2];
	    places [0] = (byte) i;
	    places [1] = (byte) (i >> 8);
	} else if (i < (1 << 24)) {
	    places = new byte [3];
	    places [0] = (byte) i;
	    places [1] = (byte) (i >> 8);
	    places [2] = (byte) (i >> 16);
	} else {
	    places = new byte [4];
	    places [0] = (byte) i;
	    places [1] = (byte) (i >> 8);
	    places [2] = (byte) (i >> 16);
	    places [3] = (byte) (i >> 24);
	}
    }

    /**
     * Converts the "big" integer to an integer value.
     *
     * @excpet NumberFormatException if 32 bits is insufficient.
     */
    public int	toInt ()
    {
	int i;
	int retval;

	if (places.length > 4)
	    throw new NumberFormatException ("BigInt.toLong, too big");
	for (i = 0, retval = 0; i < places.length; i++) {
	    retval <<= 8;
	    retval += 0x0ff & (int) places [i];
	}
	return retval;
    }

    /**
     * Returns a hexadecimal printed representation.  The value is
     * formatted to fit on lines of at least 75 characters, with
     * embedded newlines.  Words are separated for readability,
     * with eight words (32 bytes) per line.
     */
    public String toString ()
	{ return hexify (); }

    /**
     * Returns a BigInteger value which supports many arithmetic
     * operations.
     */
    public BigInteger toBigInteger ()
	{ return new BigInteger(1, places); }

    /**
     * Returns the data as a byte array.  The most significant bit
     * of the array is bit zero.
     */
    public byte [] toByteArray () { return places; }


    private static final String digits = "0123456789abcdef";

    private String hexify ()
    {
	StringBuffer	buf = new StringBuffer (places.length * 2);
	int		i;

	buf.append ("    ");	// four spaces
	for (i = 0; i < places.length; i++) {
	    buf.append (digits.charAt ((places [i] >> 4) & 0x0f));
	    buf.append (digits.charAt (places [i] & 0x0f));
	    if (((i + 1) % 32) == 0) {
		if ((i +  1) != places.length)
		    buf.append ("\n    ");	// line after four words
	    } else if (((i + 1) % 4) == 0)
		buf.append (' ');		// space between words
	}
	return buf.toString ();
    }


    /**
     * Returns true iff the parameter is a numerically equivalent
     * BigInt.
     *
     * @param other the object being compared with this one.
     */
    public boolean equals (Object other)
    {
	if (other instanceof BigInt)
	    return equals ((BigInt) other);
	return false;
    }

    /**
     * Returns true iff the parameter is numerically equivalent.
     *
     * @param other the object being compared with this one.
     */
    public boolean equals (BigInt other)
    {
	// XXX use straight byte comparison of 'places'
	BigInteger num1 = new BigInteger(1, this.toByteArray ());
	BigInteger num2 = new BigInteger(1, other.toByteArray ());
	return num1.equals (num2);
    }
}
