/*
 * @(#)DerInputBuffer.java	1.9 98/07/01
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.security.util ;

import java.io.IOException;
import java.io.ByteArrayInputStream;


/**
 * DER input buffer ... this is the main abstraction in the DER library
 * which actively works with the "untyped byte stream" abstraction.  It
 * does so with impunity, since it's not intended to be exposed to the
 * anyone who could violate the "typed value stream" DER model and hence
 * corrupt the input stream of DER values.
 *
 * @version 1.9
 * @author David Brownell
 */
class DerInputBuffer
    extends ByteArrayInputStream
    implements Cloneable
{
    DerInputBuffer (byte buf [])
	{ super (buf); }

    DerInputBuffer (byte buf [], int offset, int len)
	{ super (buf); }

    DerInputBuffer dup ()
    {
	try {
	    DerInputBuffer retval = (DerInputBuffer) clone ();

	    retval.mark (Integer.MAX_VALUE);
	    return retval;

	} catch (CloneNotSupportedException e) {
	    throw new IllegalArgumentException (e.toString ());
	}
    }

    int peek () throws IOException
    {
	if (pos >= count)
	    throw new IOException ("out of data");
	else
	    return buf [pos]; 
    }

    public boolean equals (Object other)
    {
	if (other instanceof DerInputBuffer)
	    return equals ((DerInputBuffer) other);
	else
	    return false;
    }

    boolean equals (DerInputBuffer other)
    {
	int	max = this.available ();
	int	i;

	if (this == other)
	    return true;
	if (other.available () != this.available ())
	    return false;
	for (i = 0; i < max; i++) {
	    if (this.buf [this.pos + i] != other.buf [other.pos + i]) {
		return false;
	    }
	}
	return true;
    }

    void truncate (int len) throws IOException
    {
	if (len > available ())
	    throw new IOException ("insufficient data");
	count = pos + len;
    }

    /**
     * Returns the unsigned integer which takes up the specified number
     * of bytes in this buffer.
     */
    BigInt getUnsigned  (int len) throws IOException
    {
	if (len > available ())
	    throw new IOException ("short read, getInteger");

	/*
	 * A prepended zero is used to ensure that the integer is
	 * interpreted as unsigned even when the high order bit is
	 * zero.  We don't support signed BigInts.
	 *
	 * Fix this here ... BigInts aren't expected to have these,
	 * and stuff like signing (sigsize = f(modulus)) misbehaves.
	 */
	if (buf [pos] == 0) {
	    len--;
	    skip (1);
	}

	/*
	 * Consume the rest of the buffer, returning its value as
	 * an unsigned integer.
	 */
	byte bytes [] = new byte [len];

	System.arraycopy (buf, pos, bytes, 0, len);
	skip (len);
        return new BigInt (bytes);
    }

    /**
     * Returns the bit string which takes up the rest of this buffer.
     */
    byte [] getBitString ()
    {
	/*
	 * This byte affects alignment and padding; we don't currently
	 * support non-aligned bitstrings.
	 */
	if (pos >= count || buf [pos] != 0)
	    return null;
	
	/*
	 * Just copy the data into an aligned, padded octet buffer,
	 * and consume the rest of the buffer.
	 */
	int	len = available ();
	byte	retval [] = new byte [len - 1];

	System.arraycopy (buf, pos + 1, retval, 0, len - 1);
	pos = count;
	return retval;
    }
}
