/*
 * @(#)DerOutputStream.java	1.21 98/07/01
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.security.util;

import java.io.FilterOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.OutputStream;
import java.io.IOException;
import java.util.Date;
import java.util.Vector;
import java.util.TimeZone;
import java.text.SimpleDateFormat;

/**
 * Output stream marshaling DER-encoded data.  This is eventually provided
 * in the form of a byte array; there is no advance limit on the size of
 * that byte array.
 *
 * <P>At this time, this class supports only a subset of the types of
 * DER data encodings which are defined.  That subset is sufficient for
 * generating most X.509 certificates.
 *
 * <P><em><b>NOTE:</b> Do not rely on the fact that this class currently
 * exposes an untyped data stream model, in addition to the DER model
 * of a stream of tagged data items.  That functionality will be removed,
 * so that it is no longer possible to create illegal encodings.</em>
 *
 * @version 1.21
 * @author David Brownell
 */
public
class DerOutputStream extends ByteArrayOutputStream
{
    /**
     * Construct an DER output stream.
     *
     * @param size how large a buffer to preallocate.
     */
    public DerOutputStream (int size)
    {
	super (size);
    }

    /**
     * Construct an DER output stream.
     */
    public DerOutputStream ()
    {
    }


    /**
     * Writes tagged, pre-marshaled data.  This calcuates and encodes
     * the length, so that the output data is the standard triple of
     * { tag, length, data } used by all DER values.
     *
     * @param tag the DER value tag for the data, such as
     *		<em>DerValue.tag_Sequence</em>
     * @param buf buffered data, which must be DER-encoded 
     */
    public void write (byte tag, byte buf []) throws IOException
    {
	write (tag);
	putLength (buf.length);
	write (buf, 0, buf.length);
    }


    /**
     * Writes tagged data using buffer-to-buffer copy.  As above,
     * this writes a standard DER record.  This is often used when
     * efficiently encapsulating values in sequences.
     *
     * @param tag the DER value tag for the data, such as
     *		<em>DerValue.tag_Sequence</em>
     * @param out buffered data
     */
    public void write (byte tag, DerOutputStream out) throws IOException
    {
	write (tag);
	putLength (out.count);
	write (out.buf, 0, out.count);
    }


    /**
     * Marshals pre-encoded DER value onto the output stream.
     */
    public void putDerValue (DerValue val) throws IOException
    {
	// XXX eventually this will get a "clean" implementation
	val.emit (this);
    }


    /*
     * PRIMITIVES -- these are "universal" ASN.1 simple types.
     *
     * 	INTEGER, BIT STRING, OCTET STRING, NULL
     *	OBJECT IDENTIFIER, SEQUENCE (OF), SET (OF)
     *	PrintableString, T61String, IA5String, UTCTime
     */

    /**
     * Marshals a DER unsigned integer on the output stream.
     */
    public void putInteger (BigInt i) throws IOException
    {
	byte	buf [] = i.toByteArray ();

	write (DerValue.tag_Integer);
	if ((buf [0] & 0x080) != 0) {
	    /*
	     * prepend zero so it's not read as a negative number
	     */
	    putLength (buf.length + 1);
	    write (0);
	} else
	    putLength (buf.length);
	write (buf, 0, buf.length);
    }

    /**
     * Marshals a DER bit string on the output stream.  Note that
     * only octet aligned bit strings are supported.
     *
     * @param bits the bit string, MSB first
     */
    public void putBitString (byte bits []) throws IOException
    {
	write (DerValue.tag_BitString);
	putLength (bits.length + 1);
	write (0);		// all of last octet is used
	write (bits);
    }


    /**
     * DER-encodes an ASN.1 OCTET STRING value on the output stream.
     *
     * @param octets the octet string
     */
    public void putOctetString (byte octets []) throws IOException
    {
	write (DerValue.tag_OctetString, octets);
    }

    /**
     * Marshals a DER "null" value on the output stream.  These are
     * often used to indicate optional values which have been omitted.
     */
    public void putNull () throws IOException
    {
	write (DerValue.tag_Null);
	putLength (0);
    }

    /**
     * Marshals an object identifier (OID) on the output stream.
     * Corresponds to the ASN.1 "OBJECT IDENTIFIER" construct.
     */
    public void putOID (ObjectIdentifier oid) throws IOException
    {
	oid.emit (this);
    }


    /**
     * Marshals a sequence on the output stream.  This supports both
     * the ASN.1 "SEQUENCE" (zero to N values) and "SEQUENCE OF"
     * (one to N values) constructs.
     */
    public void putSequence (DerValue seq []) throws IOException
    {
	DerOutputStream	bytes = new DerOutputStream ();
	int i;

	for (i = 0; i < seq.length; i++)
	    seq [i].emit (bytes);

	write (DerValue.tag_Sequence, bytes);
    }


    /**
     * Marshals a set on the output stream.  These are semantically
     * unordered (unlike sequences).  This supports both the ASN.1
     * "SET" (zero to N members) and "SET OF" (one to N members)
     * constructs.
     */
    public void putSet (DerValue set []) throws IOException
    {
	DerOutputStream	bytes = new DerOutputStream ();
	int i;

	for (i = 0; i < set.length; i++)
	    set [i].emit (bytes);

	write (DerValue.tag_Set, bytes);
    }

    /**
     * XXX what character set is this?
     */
    public void putPrintableString (String s) throws IOException
    {
	int i;

	write (DerValue.tag_PrintableString);
	putLength (s.length ());
	for (i = 0; i < s.length (); i++)
	    write ((byte) s.charAt (i));
    }

//    /*
//     * T61 is an 8 bit extension to ASCII, escapes e.g. to Japanese
//     */
//    void putT61String (String s) throws IOException
//    {
//	// XXX IMPLEMENT ME
//
//	throw new IOException ("DerOutputStream.putT61String() NYI");
//    }

    /**
     * Marshals a string which is consists of IA5 (ASCII) characters
     */
    public void putIA5String (String s) throws IOException
    {
	int i;

	write (DerValue.tag_IA5String);
	putLength (s.length ());
	for (i = 0; i < s.length (); i++)
	    write ((byte) s.charAt (i));
    }

    /**
     * Marshals a DER UTC time/date value.
     *
     * <P>YYMMDDhhmmss{Z|+hhmm|-hhmm} ... emits only using Zulu time
     * and with seconds (even if seconds=0) as per IETF-PKIX partI.
     */
    public void putUTCTime (Date d) throws IOException {
	/*
	 * Format the date.
	 */
	TimeZone tz = TimeZone.getTimeZone("GMT");
        String pattern = "yyMMddHHmmss'Z'";
        SimpleDateFormat sdf = new SimpleDateFormat(pattern);
        sdf.setTimeZone(tz);
        byte[] utc = (sdf.format(d)).getBytes();
	/*
	 * Write the formatted date.
	 */
	write(DerValue.tag_UtcTime);
	putLength(utc.length);
	write(utc);
    }


    /*
     * Utilities ... 
     */

    void putLength (int len) throws IOException
    {
	if (len < 128) {
	    write ((byte)len);

	} else if (len < (1 << 8)) {
	    write ((byte)0x081);
	    write ((byte)len);

	} else if (len < (1 << 16)) {
	    write ((byte)0x082);
	    write ((byte) (len >> 8));
	    write ((byte) len);

	} else if (len < (1 << 24)) {
	    write ((byte)0x083);
	    write ((byte) (len >> 16));
	    write ((byte) (len >> 8));
	    write ((byte) len);

	} else {
	    write ((byte)0x084);
	    write ((byte) (len >> 24));
	    write ((byte) (len >> 16));
	    write ((byte) (len >> 8));
	    write ((byte) len);
	}
    }
}
