/*
 * @(#)AlgorithmId.java	1.61 99/03/08
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.security.x509;

import java.io.*;
import java.security.*;
import java.util.Properties;

import sun.security.util.*;


/**
 * This class identifies algorithms, such as cryptographic transforms, each
 * of which may be associated with parameters.  Instances of this base class
 * are used when this runtime environment has no special knowledge of the
 * algorithm type, and may also be used in other cases.  Equivalence is
 * defined according to OID and (where relevant) parameters.
 *
 * <P>Subclasses may be used, for example when when the algorithm ID has
 * associated parameters which some code (e.g. code using public keys) needs
 * to have parsed.  Two examples of such algorithms are Diffie-Hellman key
 * exchange, and the Digital Signature Standard Algorithm (DSS/DSA).
 *
 * <P>The OID constants defined in this class correspond to some widely
 * used algorithms, for which conventional string names have been defined.
 * This class is not a general repository for OIDs, or for such string names.
 * Note that the mappings between algorithm IDs and algorithm names is
 * not one-to-one.
 *
 * @version 1.61
 * @author David Brownell
 */
public
class AlgorithmId implements Serializable
{
    /* Are we debugging? */
    private static boolean debug = false;

    /**
     * The object identitifer being used for this algorithm.
     */
    private ObjectIdentifier	algid;

    /**
     * Parameters for this algorithm.  These are stored in unparsed
     * DER-encoded form; subclasses can be made to automaticaly parse
     * them so there is fast access to these parameters.
     */
    protected DerValue		params;


    /**
     * Returns one of the algorithm IDs most commonly associated
     * with this algorithm name.
     *
     * @param algname the name being used
     *
     * @deprecated use the short get form of this method.
     */
    public static AlgorithmId getAlgorithmId (String algname) 
    throws NoSuchAlgorithmException 
    {
	return get(algname);
    }


    /**
     * Returns one of the algorithm IDs most commonly associated
     * with this algorithm name.
     *
     * @param algname the name being used
     */
    public static AlgorithmId get(String algname) 
    throws NoSuchAlgorithmException
    {
	ObjectIdentifier oid = algOID (algname);

	if (oid == null)
	    throw new NoSuchAlgorithmException
		("unrecognized algorithm name: " + algname);

	return new AlgorithmId(oid);
    }

    /**
     * Parse (unmarshal) an ID from a DER sequence input value.  This form
     * parsing might be used when expanding a value which has already been
     * partially unmarshaled as a set or sequence member.
     *
     * @param val the input value, which contains the algid and, if
     *		there are any parameters, those parameters.
     * @return an ID for the algorithm.  If the system is configured
     * 		appropriately, this may be an instance of a class
     *		with some kind of special support for this algorithm.
     *		In that case, you may "narrow" the type of the ID.
     */
    public static AlgorithmId
    parse (DerValue val)
    throws IOException
    {
	if (val.tag != DerValue.tag_Sequence)
	    throw new IOException ("algid parse error, not a sequence");
	
	/*
	 * Get the algorithm ID and any parameters.
	 */
	ObjectIdentifier	algid;
	DerValue		params;
	DerInputStream		in = val.toDerInputStream ();

	algid = in.getOID();
	if (in.available () == 0)
	    params = null;
	else {
	    params = in.getDerValue ();
	    if (params.tag == DerValue.tag_Null)
		params = null;
	}

	/*
	 * Figure out what class (if any) knows about this oid's
	 * parameters.  Make one, and give it the data to decode.
	 */
	AlgorithmId		alg;

	alg = buildAlgorithmId (algid, params);
	alg.decodeParams ();
	return alg;
    }


    /*
     * Return an AlgorithmId object which understands how to parse this
     * object's parameters, if one is known.  Otherwise, return an
     * instance of this base class -- it holds any parameters that exist,
     * in a generic manner.
     *
     * When the algorithm configuration has errors, this does its best
     * to report the error and continue processing.
     */
    private static AlgorithmId
    buildAlgorithmId (ObjectIdentifier algid, DerValue params)
    throws IOException
    {
	/*
	 * For algorithm ids, we rely on package-specific properties;
	 * the keys are OIDs in generic RFC 1779 syntax, values are
	 * the names of classes.
	 */
	String classname = Security.getAlgorithmProperty (algid.toString(),
							 "Class");

	if (debug)
	    System.out.println ("dynamic class name for [" + algid.toString()
		+ "|Class]: " + classname);

	if (classname == null || classname.equals (""))
	    return new AlgorithmId (algid, params);

	try {
	    Class	algClass = Class.forName (classname);
	    Object	inst;
	    AlgorithmId	id;

	    inst = algClass.newInstance ();
	    if (!(inst instanceof AlgorithmId)) {
		System.err.println (
			"Misconfiguration:  faulty algorithm config, " 
				    + classname);
		return new AlgorithmId (algid, params);
	    }
	    id = (AlgorithmId) inst;
	    id.algid = algid;
	    id.params = params;
	    id.decodeParams();
	    return id;

	} catch (ClassNotFoundException e) {
	    System.err.println ("Misconfiguration:  unknown algorithm class, "
				+ classname);
	    return new AlgorithmId (algid, params);

	} catch (IllegalAccessException e) {
	    // this should not happen.
	    throw new IOException (classname + " [internal error]");

	} catch (InstantiationException e) {
	    System.err.println ("Misconfiguration:  faulty algorithm class, " 
				+ classname);
	    return new AlgorithmId (algid, params);
	}
    }


    /**
     * Constructs a parameterless algorithm ID.
     *
     * @param oid the identifier for the algorithm
     */
    public AlgorithmId (ObjectIdentifier oid)
    {
	algid = oid;
    }


    private AlgorithmId (ObjectIdentifier oid, DerValue params)
    throws IOException
    {
	this.algid = oid;
	this.params = params;
	decodeParams ();
    }


    /**
     * Constructs an algorithm ID which will be initialized
     * separately, for example by deserialization.
     */
    // XXX deprecated for general use
    public AlgorithmId () { }


    /**
     * When redefined by subclassers, this function allows algorithm
     * parameters to be automatically parsed.  For example, this could
     * cause encoded DSA algorithm parameters P, Q, and G to be stored
     * so that both private (X) and public (Y) keys can share them.
     */
    protected void decodeParams () throws IOException
    {
	// nothing by default
    }

    /**
     * Marshal a DER-encoded "AlgorithmID" sequence on the DER stream.
     */
    public final void emit (DerOutputStream out)
    throws IOException
    {
	DerOutputStream	bytes = new DerOutputStream ();

	bytes.putOID (algid);
	if (params == null)
	    bytes.putNull ();
	else
	    bytes.putDerValue (params);
	out.write (DerValue.tag_Sequence, bytes);
    }


    /**
     * Emit DER-encoded X.509 "AlgorithmId" sequence on the output stream.
     *
     * @deprecated use the emit method instead
     */
    public final void encode (OutputStream out) throws IOException
    {
	DerOutputStream	bytes = new DerOutputStream ();

	emit (bytes);
	out.write (bytes.toByteArray());
    }

    /**
     * Returns the DER-encoded X.509 AlgorithmId as a byte array.
     *
     * @deprecated use the emit method instead
     */
    public final byte [] encode () throws IOException
    {
	DerOutputStream out = new DerOutputStream ();

	emit (out);
	return out.toByteArray();
    }

    /*
     * Translates from some common algorithm names to the
     * OID with which they're usually associated ... this mapping
     * is the reverse of the one below, except in those cases
     * where synonyms are supported or where a given algorithm
     * is commonly associated with multiple OIDs.
     */
    private static ObjectIdentifier algOID (String name)
    {

	// Digesting algorithms

	if (name.equals ("MD5"))
	    return AlgorithmId.MD5_oid;
	if (name.equals ("MD2"))
	    return AlgorithmId.MD2_oid;
	if (name.equals ("SHA") || name.equals ("SHA1") || name.equals("SHA-1"))
	    return AlgorithmId.SHA_oid;

	// Various public key algorithms

	if (name.equals ("RSA"))
	    return AlgorithmId.RSA_oid;
	if (name.equals ("Diffie-Hellman") || name.equals("DH"))
	     return AlgorithmId.DH_oid;
	if (name.equals ("DSA"))
	    return AlgorithmId.DSA_oid;

	// Common signature types

	if (name.equals ("MD5withRSA") || name.equals("MD5/RSA"))
	    return AlgorithmId.md5WithRSAEncryption_oid;
	if (name.equals ("MD2withRSA") || name.equals("MD2/RSA"))
	    return AlgorithmId.md2WithRSAEncryption_oid;
	if (name.equals ("SHAwithDSA") || name.equals("SHA/DSA"))
	    return AlgorithmId.shaWithDSA_oid;

	return null;
    }

    /*
     * For the inevitable cases where key or signature types are not
     * configured in an environment which encounters such keys or
     * signatures, we still attempt to provide user-friendly names
     * for some of the most common algorithms.  Subclasses can of
     * course override getName().
     *
     * Wherever possible, the names are those defined by the IETF.
     * Such names are noted below.
     */
    private String algName ()
    {
	// Common message digest algorithms

	if (algid.equals (AlgorithmId.MD5_oid))
	    return "MD5";	// RFC 1423
	if (algid.equals (AlgorithmId.MD2_oid))
	    return "MD2";	// RFC 1423
	if (algid.equals (AlgorithmId.SHA_oid))
	    return "SHA";

	// Common key types

	if (algid.equals (AlgorithmId.RSAEncryption_oid)
		|| algid.equals (AlgorithmId.RSA_oid))
	    return "RSA";
	if (algid.equals (AlgorithmId.DH_oid))
	    return "Diffie-Hellman";
	if (algid.equals (AlgorithmId.DSA_oid))
	    return "DSA";

	// Common signature types

	if (algid.equals (AlgorithmId.md5WithRSAEncryption_oid))
	    return "MD5withRSA";
	if (algid.equals (AlgorithmId.md2WithRSAEncryption_oid))
	    return "MD2withRSA";
	if (algid.equals (AlgorithmId.shaWithDSA_oid))
	    return "SHA1withDSA";

	// default returns a dot-notation ID

	return "OID." + algid.toString ();
    }

    /**
     * Returns the ISO OID for this algorithm.  This is usually converted
     * to a string and used as part of an algorithm name, for example
     * "OID.1.3.14.3.2.13" style notation.  Use the <code>getName</code>
     * call when you do not need to ensure cross-system portability
     * of algorithm names, or need a user friendly name.
     */
    final public ObjectIdentifier	getOID ()
	{ return algid; }


    /**
     * Returns a name for the algorithm which may be more intelligible
     * to humans than the algorithm's OID, but which won't necessarily
     * be comprehensible on other systems.  For example, this might
     * return a name such as "MD5withRSA" for a signature algorithm on
     * some systems.  It also returns names like "OID.1.2.3.4", when
     * no particular name for the algorithm is known.
     */
    public String getName()
	{ return algName (); }

    /**
     * Returns a string describing the algorithm and its parameters.
     */
    public String toString ()
    {
	return "[" + algName () + paramsToString () + "]";
    }


    /**
     * Provides a human-readable description of the algorithm parameters.
     * This may be redefined by subclasses which parse those parameters.
     */
    protected String paramsToString ()
    {
	if (params == null)
	    return "";
	else
	    return ", params unparsed";
    }


    /**
     * Returns true iff the argument indicates the same algorithm
     * with the same parameters.
     */
    public boolean equals (AlgorithmId other)
    {
        if (!algid.equals (other.algid))
            return false;
	else if (params == null && other.params == null)
	    return true;
	else if (params == null)
	    return false;
	else
	    return params.equals (other.params);
    }


    /**
     * Compares this AlgorithmID to another.  If algorithm parameters are
     * available, they are compared.  Otherwise, just the object IDs
     * for the algorithm are compared.
     *
     * @param other preferably an AlgorithmId, else an ObjectIdentifier
     */
    public boolean equals (Object other)
    {
	if (other instanceof AlgorithmId)
	    return equals ((AlgorithmId) other);
	else if (other instanceof ObjectIdentifier)
	    return equals ((ObjectIdentifier) other);
	else
	    return false;
    }


    /**
     * Compares two algorithm IDs for equality.  Returns true iff
     * they are the same algorithm, ignoring algorithm parameters.
     */
    public final boolean equals (ObjectIdentifier id)
    {
	return algid.equals (id);
    }


    /*****************************************************************/

    /*
     * HASHING ALGORITHMS
     */
    private static final int MD2_data [] = 
	{ 1, 2, 840, 113549, 2, 2 };
    private static final int MD5_data [] = 
	{ 1, 2, 840, 113549, 2, 5 };
    private static final int SHA1_data [] =
	{ 1, 3, 14, 3, 2, 26 };

    /**
     * Algorithm ID for the MD2 Message Digest Algorthm, from RFC 1319.
     */
    public static final ObjectIdentifier
	MD2_oid = new ObjectIdentifier (MD2_data);

    /**
     * Algorithm ID for the MD5 Message Digest Algorthm, from RFC 1321.
     */
    public static final ObjectIdentifier
	MD5_oid = new ObjectIdentifier (MD5_data);

    /**
     * Algorithm ID for the SHA1 Message Digest Algorithm, from FIPS 180-1.
     * This is sometimes called "SHA", though that is often confusing since
     * many people refer to FIPS 180 (which has an error) as defining SHA.
     */
    public static final ObjectIdentifier 
	SHA_oid = new ObjectIdentifier (SHA1_data);


    /*
     * COMMON PUBLIC KEY TYPES
     */
    private static final int DH_data [] = 
	{ 1, 2, 840, 113549, 1, 3, 1 };
    private static final int dsa_data [] = 
	{ 1, 3, 14, 3, 2, 12 };
    private static final int RSA_data [] = 
	{ 1, 2, 5, 8, 1, 1 };
    private static final int RSAEncryption_data [] = 
	{ 1, 2, 840, 113549, 1, 1, 1 };

    /**
     * Algorithm ID for Diffie Hellman Key agreement, from PKCS #3.
     * Parameters include public values P and G, and may optionally specify
     * the length of the private key X.  Alternatively, algorithm parameters
     * may be derived from another source such as a Certificate Authority's
     * certificate.
     */
    public static final ObjectIdentifier
	DH_oid = new ObjectIdentifier (DH_data);

    /**
     * Algorithm ID for the Digital Signing Algorithm (DSA), from the
     * NIST Object Implementors' Workshop.
     * Parameters
     * may include public values P, Q, and G; or these may be derived from
     * another source such as a Certificate Authority's certificate.
     */
    public static final ObjectIdentifier
	DSA_oid = new ObjectIdentifier (dsa_data);

    /**
     * Algorithm ID for RSA keys used for any purpose, as defined in X.509.
     * The algorithm parameter is a single value, the number of bits in the
     * public modulus.
     */
    public static final ObjectIdentifier
	RSA_oid = new ObjectIdentifier (RSA_data);

    /**
     * Algorithm ID for RSA keys used with RSA encryption, as defined
     * in PKCS #1.  There are no parameters associated with this algorithm.
     */
    public static final ObjectIdentifier
	RSAEncryption_oid = new ObjectIdentifier (RSAEncryption_data);



    /*
     * COMMON SIGNATURE ALGORITHMS
     */
    private static final int md2WithRSAEncryption_data [] = 
	{ 1, 2, 840, 113549, 1, 1, 2 };
    private static final int md5WithRSAEncryption_data [] = 
	{ 1, 2, 840, 113549, 1, 1, 4 };
    private static final int shaWithDSA_data [] = 
	{ 1, 3, 14, 3, 2, 13 };

    /**
     * Identifies a signing algorithm where an MD2 digest is encrypted
     * using an RSA private key; defined in PKCS #1.  Use of this
     * signing algorithm is discouraged due to MD2 vulnerabilities.
     */
    public static final ObjectIdentifier
	md2WithRSAEncryption_oid = new
		ObjectIdentifier (md2WithRSAEncryption_data);

    /**
     * Identifies a signing algorithm where an MD5 digest is
     * encrypted using an RSA private key; defined in PKCS #1.
     */
    public static final ObjectIdentifier
	md5WithRSAEncryption_oid = new
		ObjectIdentifier (md5WithRSAEncryption_data);

    /**
     * Identifies the FIPS 186 "Digital Signature Standard" (DSS), where a
     * SHA1 digest is signed using the Digital Signing Algorithm (DSA).
     */
    public static final ObjectIdentifier
	shaWithDSA_oid = new ObjectIdentifier (shaWithDSA_data);
}
