/*
 * @(#)CertAndKeyGen.java	1.34 99/03/08
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.security.x509;

import java.io.*;
import java.security.*;
import java.util.Date;

import sun.security.util.BigInt;
import sun.security.pkcs.PKCS10;


/**
 * Generate a pair of keys, and provide access to them.  This class is
 * provided primarily for ease of use.
 *
 * <P>This provides some simple certificate management functionality.
 * Specifically, it allows you to create self-signed X.509 certificates
 * as well as PKCS 10 based certificate signing requests.
 *
 * <P>Keys for some public key signature algorithms have algorithm
 * parameters, such as DSS/DSA.  Some sites' Certificate Authorities
 * adopt fixed algorithm parameters, which speeds up some operations
 * including key generation and signing.  <em>At this time, this interface
 * does not provide a way to provide such algorithm parameters, e.g.
 * by providing the CA certificate which includes those parameters.</em>
 *
 * <P>Also, note that at this time only signature-capable keys may be
 * acquired through this interface.  Diffie-Hellman keys, used for secure
 * key exchange, may be supported later.
 *
 * @see X509Cert
 * @see PKCS10
 *
 * @version 1.34
 * @author David Brownell
 */
public final
class CertAndKeyGen
{
    /**
     * Creates a CertAndKeyGen object for a particular key type
     * and signature algorithm.
     *
     * @param keyType type of key, e.g. "RSA", "DSA"
     * @param sigAlg name of the signature algorithm, e.g. "MD5WithRSA",
     *		"MD2WithRSA", "SHAwithDSA".
     */
    public CertAndKeyGen (String keyType, String sigAlg)
    throws NoSuchAlgorithmException
    {
	keyGen = KeyPairGenerator.getInstance(keyType);
	this.sigAlg = sigAlg;
    }

    /**
     * Sets the source of random numbers used when generating keys.
     * If you do not provide one, a system default facility is used.
     * You may wish to provide your own source of random numbers
     * to get a reproducible sequence of keys and signatures, or
     * because you may be able to take advantage of strong sources
     * of randomness/entropy in your environment.
     */
    public void		setRandom (SecureRandom generator)
    {
	prng = generator;
    }

    // want "public void generate (X509Cert)" ... inherit DSA/D-H param

    /**
     * Generates a random public/private key pair, with a given key
     * size.  Different algorithms provide different degrees of security
     * for the same key size, because of the "work factor" involved in
     * brute force attacks.  As computers become faster, it becomes
     * easier to perform such attacks.  Small keys are to be avoided.
     *
     * <P>Note that not all values of "keyBits" are valid for all
     * algorithms, and not all public key algorithms are currently
     * supported for use in X.509 certificates.  If the algorithm
     * you specified does not produce X.509 compatible keys, an
     * invalid key exception is thrown.
     *
     * @param keyBits the number of bits in the keys.
     * @except InvalidKeyException if the environment does not
     *	provide X.509 public keys for this signature algorithm.
     */
    public void generate (int keyBits)
    throws InvalidKeyException
    {
	KeyPair	pair;

	try {
	    if (prng == null)
		prng = new SecureRandom ();
	    keyGen.initialize (keyBits, prng);
	    pair = keyGen.generateKeyPair ();

	} catch (Exception e) {
	    throw new IllegalArgumentException (e.getMessage ());
	}
	
	PublicKey publicKey = pair.getPublic();

	if (publicKey instanceof X509Key) {
	    this.publicKey = (X509Key) publicKey;
	    
	} else { 
	    throw new InvalidKeyException ("public key " + publicKey + 
					  " not an X509Key.");
	}
	privateKey = pair.getPrivate ();
    }


    /**
     * Returns the public key of the generated key pair.
     */
    public X509Key getPublicKey ()
    {
	return publicKey; 
    }


    /**
     * Returns the private key of the generated key pair.
     *
     * <P><STRONG><em>Be extremely careful when handling private keys.
     * When private keys are not kept secret, they lose their ability
     * to securely authenticate specific entities ... that is a huge
     * security risk!</em></STRONG>
     */
    public PrivateKey	getPrivateKey ()
    {
	return privateKey;
    }


    /**
     * Returns a self-signed X.509v1 certificate for the public key.
     * The certificate is immediately valid.
     *
     * <P>Such certificates normally are used to identify a "Certificate
     * Authority" (CA).  Accordingly, they will not always be accepted by
     * other parties.  However, such certificates are also useful when
     * you are bootstrapping your security infrastructure, or deploying
     * system prototypes.
     *
     * @param myname X.500 name of the subject (who is also the issuer)
     * @param validity how long the certificate should be valid, in seconds
     */
    public X509Cert		getSelfCert (X500Name myname, long validity)
    throws InvalidKeyException, SignatureException, NoSuchAlgorithmException
    {
	X500Signer	issuer;
	X509Cert	cert;
	Date		firstDate, lastDate;
	
	try {
	    issuer = getSigner (myname);

	    firstDate = new Date ();
	    lastDate = new Date ();
	    lastDate.setTime (lastDate.getTime () + validity * 1000);

	    cert = new X509Cert (myname, publicKey, firstDate, lastDate);
	    cert.encodeAndSign (
		    new BigInt ((int) (firstDate.getTime () / 1000)),
		    issuer);

	    return cert;

	} catch (IOException e) {
	    throw new CertException (CertException.err_CONSTRUCTION,
		    " getSelfCert");
	}
    }


    /**
     * Returns a PKCS #10 certificate request.  The caller uses either 
     * <code>PKCS10.print</code> or <code>PKCS10.toByteArray</code>
     * operations on the result, to get the request in an appropriate
     * transmission format.
     *
     * <P>PKCS #10 certificate requests are sent, along with some proof
     * of identity, to Certificate Authorities (CAs) which then issue
     * X.509 public key certificates.
     *
     * @param myname X.500 name of the subject
     */
    public PKCS10 getCertRequest (X500Name myname)
    throws InvalidKeyException, SignatureException
    {
	PKCS10	req = new PKCS10 (publicKey);

	try {
	    req.encodeAndSign (getSigner (myname));

	} catch (IOException e) {
	    throw new SignatureException (sigAlg + " IOException");

	} catch (NoSuchAlgorithmException e) {
	    // "can't happen"
	    throw new SignatureException (sigAlg + " unavailable?");
	}	    
	return req;
    }

    private X500Signer getSigner (X500Name me)
    throws InvalidKeyException, NoSuchAlgorithmException
    {
	Signature	signature = Signature.getInstance(sigAlg);

	// XXX should have a way to pass prng to the signature
	// algorithm ... appropriate for DSS/DSA, not RSA

	signature.initSign (privateKey);
	return new X500Signer (signature, me);
    }

    private SecureRandom	prng;
    private String		sigAlg;
    private KeyPairGenerator	keyGen;
    private X509Key		publicKey;
    private PrivateKey		privateKey;
}
