/*
 * @(#)CertException.java	1.17 98/10/29
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.security.x509;


/**
 * CertException indicates one of a variety of certificate problems.
 *
 * <P><em>In the future this may be replaced by a family of related
 * exception classes, eliminating the use of the code indicating
 * the specific fault.</em>
 *
 * @version 1.17
 * @author David Brownell
 */
public
class CertException extends SecurityException {

    // Zero is reserved.

    /** Indicates that the signature in the certificate is not valid. */
    public static final int verf_INVALID_SIG = 1;

    /** Indicates that the certificate was revoked, and so is invalid. */
    public static final int verf_INVALID_REVOKED = 2;

    /** Indicates that the certificate is not yet valid. */
    public static final int verf_INVALID_NOTBEFORE = 3;

    /** Indicates that the certificate has expired and so is not valid. */
    public static final int verf_INVALID_EXPIRED = 4;	// Cert expired

    /** Indicates that a certificate authority in the certification
     * chain is not trusted. */
    public static final int verf_CA_UNTRUSTED = 5;

    /** Indicates that the certification chain is too long. */
    public static final int verf_CHAIN_LENGTH = 6;

    /** Indicates an error parsing the ASN.1/DER encoding of the certificate. */
    public static final int verf_PARSE_ERROR = 7;

    /** Indicates an error constructing a certificate or cert chain. */
    public static final int err_CONSTRUCTION = 8;

    /** Indicates an problem with the public key */
    public static final int err_INVALID_PUBLIC_KEY = 9;

    private int		verfCode;
    private String	moreData;


    /**
     * Constructs a cert exception using an error code (<CODE>verf_*</CODE>)
     * and a string describing the context of the error.
     */
    public CertException (int code, String moredata)
    {
	verfCode = code;
	moreData = moredata;
    }


    /**
     * Constructs a cert exception using just an error code, without a
     * string describing the context.
     */
    public CertException (int code)
    {
	verfCode = code;
    }


    /**
     * Returns the error code with which the exception was created.
     */
    public int getVerfCode () { return verfCode; }


    /**
     * Returns a string describing the context in which the exception
     * was reported.
     */
    public String getMoreData () { return moreData; }


    /**
     * Return a string corresponding to the error code used to create
     * this exception.
     */
    public String getVerfDescription ()
    {
	switch (verfCode) {
	  case verf_INVALID_SIG:
	    return "The signature in the certificate is not valid.";
	  case verf_INVALID_REVOKED:
	    return "The certificate was revoked.";
	  case verf_INVALID_NOTBEFORE:
	    return "The certificate is not yet valid.";
	  case verf_INVALID_EXPIRED:
	    return "The certificate expired.";
	  case verf_CA_UNTRUSTED:
	    return "The Authority which issued the is not trusted.";
	  case verf_CHAIN_LENGTH:
	    return "The certificate path to a trusted authority is too long.";
	  case verf_PARSE_ERROR:
	    return "The certificate could not be parsed.";
	  case err_CONSTRUCTION:
	    return "There was an error when constructing the certificate.";
	  case err_INVALID_PUBLIC_KEY:
	    return "The public key was not in the correct format.";

	  default:
	    return "Unknown code:  " + verfCode;
	}
    }


    /**
     * Returns a string describing the certificate exception.
     */
    public String toString ()
    {
	return "[Certificate Exception:  "
		+ getVerfDescription ()
		+ ( (moreData != null)
		    ? ( "\n  (" + moreData + ")" )
		    : "" )
		+ "]";
    }
}
