/*
 * @(#)X509Key.java	1.62 99/04/01
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.security.x509;

import java.io.*;
import java.security.Key;
import java.security.Provider;
import java.security.PublicKey;
import java.security.Security;
import java.security.InvalidKeyException;
import java.security.ProviderException;
import java.security.MessageDigest;
import java.util.Properties;

import sun.misc.HexDumpEncoder;
import sun.security.util.*;

// this will go away.
import sun.security.provider.Sun;

/**
 * Holds an X.509 key, for example a public key found in an X.509
 * certificate.  Includes a description of the algorithm to be used
 * with the key; these keys normally are used as
 * "SubjectPublicKeyInfo".
 *
 * <P>While this class can represent any kind of X.509 key, it may be
 * desirable to provide subclasses which understand how to parse keying
 * data.   For example, RSA public keys have two members, one for the
 * public modulus and one for the prime exponent.  If such a class is
 * provided, it is used when parsing X.509 keys.  If one is not provided,
 * the key still parses correctly.
 *
 * @version 1.62, 99/11/17
 * @author David Brownell 
 */
public
class X509Key implements PublicKey {

    /* The algorithm information (name, parameters, etc). */
    protected AlgorithmId algid;

    /* The key bytes, without the algorithm information */
    protected byte[] key;

    /* The encoded for the key. */
    private byte[] encodedKey;

    /**
     * Default constructor.  The key constructed must have its key
     * and algorithm initialized before it may be used, for example
     * by using <code>decode</code>.
     */
    protected X509Key () { }

    /*
     * Build and initialize as a "default" key.  All X.509 key
     * data is stored and transmitted losslessly, but no knowledge
     * about this particular algorithm is available.
     */
    private X509Key (AlgorithmId algid, byte key [])
    throws InvalidKeyException {
	this.algid = algid;
	this.key = key;
	encode();
    }


    /**
     * Construct X.509 subject public key from a DER value.  If
     * the runtime environment is configured with a specific class for
     * this kind of key, a subclass is returned.  Otherwise, a generic
     * X509Key object is returned.
     * 
     * <P>This mechanism gurantees that keys (and algorithms) may be
     * freely manipulated and transferred, without risk of losing
     * information.  Also, when a key (or algorithm) needs some special
     * handling, that specific need can be accomodated.
     *
     * @param in the DER-encoded SubjectPublicKeyInfo value
     * @exception IOException on data format errors
     */
    public static X509Key parse (DerValue in) throws IOException
    {
	AlgorithmId	algorithm;
	X509Key		subjectKey;

	if (in.tag != DerValue.tag_Sequence)
	    throw new IOException("corrupt subject key");

	algorithm = AlgorithmId.parse (in.data.getDerValue ());

	try {
	    subjectKey = buildX509Key (algorithm, in.data.getBitString ());

	    /* REMIND Remove this soon.
	    subjectKey.initialize(subjectKey.encode(), "X.509", 
				  algorithm.getName());
	     */

	} catch (InvalidKeyException e) {
	    throw new IOException("subject key, " + e.toString());
	}

	if (in.data.available () != 0)
	    throw new IOException("excess subject key");
	return subjectKey;
    }


    /**
     * Parse the key bits.  This may be redefined by subclasses to take
     * advantage of structure within the key.  For example, RSA public
     * keys encapsulate two unsigned integers (modulus and exponent) as
     * DER values within the <code>key</code> bits; Diffie-Hellman and
     * DSS/DSA keys encapsulate a single unsigned integer.
     *
     * <P>This function is called when creating X.509 SubjectPublicKeyInfo
     * values using the X509Key member functions, such as <code>parse</code>
     * and <code>decode</code>.
     */
    protected void parseKeyBits () throws IOException, InvalidKeyException {
	encode();
    }

    /* Temporary workaround for X509Key */
    static String getSunProperty(String prop) {
	Provider sun = Security.getProvider("SUN");
	return sun.getProperty(prop);
    }
    /* End workaround */
    
    /*
     * Factory interface, building the kind of key associated with this
     * specific algorithm ID or else returning this generic base class.
     * See the description above.
     */
    static X509Key buildX509Key (AlgorithmId algid, byte key [])
    throws IOException, InvalidKeyException
    {
	Properties	props;
	String		keytype;
	String		classname;
	
	/* This is temporarily mapped back to SUN till we
	   decide on a proper interface for java.* stuff. */
	classname = getSunProperty("PublicKey.X.509." + algid.getName());
	
	if (classname == null) {
	    classname = "sun.security.x509.X509Key";
	}

	try {
	    Class	keyClass = Class.forName(classname);
	    Object	inst;
	    X509Key	result;

	    inst = keyClass.newInstance();
	    if (inst instanceof X509Key) {
		result = (X509Key) inst;
		result.algid = algid;
		result.key = key;
		result.parseKeyBits();
		return result;
	    }

	    System.err.println ("Misconfiguration:  faulty key config, " +
				classname);
	    return new X509Key (algid, key);

	} catch (ClassNotFoundException e) {
	    System.err.println ("Misconfiguration:  unknown key class, " + 
			      classname);
	    return new X509Key (algid, key);

	} catch (IllegalAccessException e) {
	    // this should not happen.
	    e.printStackTrace ();
	    throw new IOException (classname + " [internal error]");

	} catch (InstantiationException e) {
	    System.err.println ("Misconfiguration:  faulty key class, " + 
			      classname);
	    return new X509Key (algid, key);
	}
    }


    /**
     * Returns the algorithm to be used with this key.
     */
    public String getAlgorithm() { 
	return algid.getName();
    }

    /**
     * Returns the algorithm ID to be used with this key.
     */
    public AlgorithmId	getAlgorithmId () { return algid; }


    /**
     * Emit SubjectPublicKeyInfo sequence on the DER output stream.
     */
    public final void emit (DerOutputStream out) throws IOException
    {
	DerOutputStream		tmp = new DerOutputStream ();

	algid.emit (tmp);
	tmp.putBitString (key);
	out.write (DerValue.tag_Sequence, tmp);
    }

    /**
     * Returns the DER-encoded form of the key as a byte array.
     */
    public synchronized byte[] getEncoded() {
	if (encodedKey == null) {
	    try {
		encode();
	    } catch (InvalidKeyException e) {
	    }
	}				   
	return encodedKey;
    }


    /**
     * Returns the format for this key: "X.509"
     */
    public String getFormat() {
	return "X.509";
    }

    /**
     * Returns the DER-encoded form of the key as a byte array.
     */
    public byte[] encode() throws InvalidKeyException {
	if (encodedKey == null) {
	    try {
		DerOutputStream	out;
		
		out = new DerOutputStream ();
		emit (out);
		encodedKey = out.toByteArray();

	    } catch (IOException e) {
		throw new InvalidKeyException ("IOException : " + 
					       e.getMessage());
	    }
	} 
	return encodedKey;
    }

    /*
     * Returns a printable representation of the key
     */
    public String toString ()
    {
	HexDumpEncoder	encoder = new HexDumpEncoder ();

	return "algorithm = " + algid.toString ()
	    + ", unparsed keybits = \n" + encoder.encodeBuffer (key);
    }

    /** 
     * Initialize an X509Key object from an input stream.  The data on that
     * input stream must be encoded using DER, obeying the X.509
     * <code>SubjectPublicKeyInfo</code> format.  That is, the data is a
     * sequence consisting of an algorithm ID and a bit string which holds
     * the key.  (That bit string is often used to encapsulate another DER
     * encoded sequence.)
     *
     * <P>Subclasses should not normally redefine this method; they should
     * instead provide a <code>parseKeyBits</code> method to parse any
     * fields inside the <code>key</code> member.
     *
     * <P>The exception to this rule is that since private keys need not
     * be encoded using the X.509 <code>SubjectPublicKeyInfo</code> format,
     * private keys may override this method, <code>encode</code>, and
     * of course <code>getFormat</code>.
     *
     * @param in an input stream with a DER-encoded X.509
     *		SubjectPublicKeyInfo value
     */
    public void decode(InputStream in)
    throws InvalidKeyException
    {
	DerValue	val;

	try {
	    val = new DerValue (in);
	    if (val.tag != DerValue.tag_Sequence)
		throw new InvalidKeyException ("invalid key format");

	    algid = AlgorithmId.parse (val.data.getDerValue ());
	    key = val.data.getBitString ();
	    parseKeyBits ();
	    if (val.data.available () != 0)
		throw new InvalidKeyException ("excess key data");

	} catch (IOException e) {
	    // e.printStackTrace ();
	    throw new InvalidKeyException(e.toString());
	}
    }

    public void decode(byte[] encodedKey) throws InvalidKeyException {
	decode(new ByteArrayInputStream(encodedKey));
    }

    /**
     * Serialization write ... X.509 keys serialize as
     * themselves, and they're parsed when they get read back.
     */
    private synchronized void
    writeObject (java.io.ObjectOutputStream stream)
    throws IOException { 
	stream.write(getEncoded()); 
    }


    /**
     * Serialization read ... X.509 keys serialize as
     * themselves, and they're parsed when they get read back.
     */
    private synchronized void
    readObject (ObjectInputStream stream) 
    throws IOException {

	try {
	    decode(stream); 

	} catch (InvalidKeyException e) {
	    e.printStackTrace();
	    throw new IOException("deserialized key is invalid: " + 
				  e.getMessage());
	}
    }

    public boolean equals(Object object) {
	if (this == object) {
	    return true;
	}

	if (object instanceof Key) {
	    Key key = (Key)object;

	    byte[] b1 = getEncoded();
	    byte[] b2 = key.getEncoded();

	    return MessageDigest.isEqual(b1, b2);
	}

	return false;
    }
	    

}
