/*
 * @(#)palette4.c	1.2 98/05/13 SMI
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
/* Iterative color palette generation */
/* Richard Berlin, June 24 1996 */
/* rearchitected for speed improvements by Jim Graham August 1996 */
/* changed to CCIR Rec 709 on 13 Nov 1997 */
/* output format changed for pjava on 13 Jan 1998 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#define TRUE 1
#define FALSE 0

/* This information is specific to the monitor */
/* the default values are for CCIR Rec 709 */
static float monitor_gamma[3] = {1.0/.45, 1.0/.45, 1.0/.45}; /* r,g,b */
static float redxy[2]   = {0.640,  0.330};
static float greenxy[2] = {0.300,  0.600};
static float bluexy[2]  = {0.150,  0.060};
static float whitexy[2] = {0.3127, 0.3291};
static float whiteY     = 1.0;
/* end of information specific to the monitor */

/* these are derived from the monitor information above */
/* see recalculate_constants() for the formulae */
static float mat[3][3] = {
    {0.4121, 0.2125, 0.0193},
    {0.3577, 0.7154, 0.1192},
    {0.1804, 0.0721, 0.9499}
};
static float uwht = 4*0.3127/(-2*0.3127 + 12*0.3291 + 3);
static float vwht = 9*0.3291/(-2*0.3127 + 12*0.3291 + 3);

static float Rmat[3][256];
static float Gmat[3][256];
static float Bmat[3][256];
static float Ltab[256], Utab[256], Vtab[256];

typedef struct {
    unsigned char red;
    unsigned char green;
    unsigned char blue;
    unsigned char bestidx;
    float L, U, V;
    float dist;
    float dE;
    float dL;
} CmapEntry;

int gray_start, gray_stop;

/* this is a multiplier--it should not be zero */
#define LIGHTNESS_PENALTY 10
#define WEIGHT 250.0
#ifdef WEIGHT
#define WEIGHT_DIST(d,l)   (WEIGHT*(d)/(WEIGHT+(l)))
#define UNWEIGHT_DIST(d,l) ((WEIGHT+(l))*(d)/WEIGHT)
#else
#define WEIGHT_DIST(d,l) (d)
#define UNWEIGHT_DIST(d,l) (d)
#endif

#define xR redxy[0]
#define yR redxy[1]
#define xG greenxy[0]
#define yG greenxy[1]
#define xB bluexy[0]
#define yB bluexy[1]
#define xW whitexy[0]
#define yW whitexy[1]
static void
recalculate_constants() {
    float D  = xR*(yG - yB) + xG*(yB - yR) + xB*(yR - yG);
    float Cr = (whiteY*(xW*(yG - yB) - yW*(xG - xB) + xG*yB - xB*yG)) / (yW*D);
    float Cg = (whiteY*(xW*(yB - yR) - yW*(xB - xR) - xR*yB + xB*yR)) / (yW*D);
    float Cb = (whiteY*(xW*(yR - yG) - yW*(xR - xG) + xR*yG - xG*yR)) / (yW*D);

    mat[0][0] = xR*Cr;
    mat[0][1] = yR*Cr;
    mat[0][2] = (1 - xR - yR)*Cr;
    mat[1][0] = xG*Cg;
    mat[1][1] = yG*Cg;
    mat[1][2] = (1 - xG - yG)*Cg;
    mat[2][0] = xB*Cb;
    mat[2][1] = yB*Cb;
    mat[2][2] = (1 - xB - yB)*Cb;

    uwht = 4*whitexy[0]/(-2*whitexy[0] + 12*whitexy[1] + 3);
    vwht = 9*whitexy[1]/(-2*whitexy[0] + 12*whitexy[1] + 3);
}
#undef yW
#undef xW
#undef yB
#undef xB
#undef yG
#undef xG
#undef yR
#undef xR

static int
init_matrices()
{
    int i, j;

    for (i = 0; i < 256; ++i)
    {
	float iG = pow(i/255.0, monitor_gamma[0]);
	Rmat[0][i] = mat[0][0] * iG;
	Rmat[1][i] = mat[0][1] * iG;
	Rmat[2][i] = mat[0][2] * iG;

	iG = pow(i/255.0, monitor_gamma[1]);
	Gmat[0][i] = mat[1][0] * iG;
	Gmat[1][i] = mat[1][1] * iG;
	Gmat[2][i] = mat[1][2] * iG;

	iG = pow(i/255.0, monitor_gamma[2]);
	Bmat[0][i] = mat[2][0] * iG;
	Bmat[1][i] = mat[2][1] * iG;
	Bmat[2][i] = mat[2][2] * iG;
    }
}

static void
LUV_convert(int red, int grn, int blu, float *L, float *u, float *v)
{
    float X = Rmat[0][red] + Gmat[0][grn] + Bmat[0][blu];
    float Y = Rmat[1][red] + Gmat[1][grn] + Bmat[1][blu];
    float Z = Rmat[2][red] + Gmat[2][grn] + Bmat[2][blu];

    float sum  = X+Y+Z;
    float x    = X/sum;
    float y    = Y/sum;
    float dnm  = -2*x + 12*y + 3;
    float uprm = 4*x/dnm;
    float vprm = 9*y/dnm;
    float ytmp = pow(Y/whiteY, 1.0/3.0);

    if (ytmp < .206893) {
	*L = 903.3*Y/whiteY;
    } else {
	*L = 116*(ytmp) - 16;
    }
    *u = 13*(*L)*(uprm-uwht);
    *v = 13*(*L)*(vprm-vwht);
}

int total;
unsigned char cmap_r[256], cmap_g[256], cmap_b[256];

#define DIST_THRESHOLD 7
int no_close_color(float l, float u, float v, int c_tot, int exact) {
    int i;
    for (i = 0; i < c_tot; ++i) {
	float t, dist = 0;
	t = Ltab[i] - l; dist += t*t*LIGHTNESS_PENALTY;
	t = Utab[i] - u; dist += t*t;
	t = Vtab[i] - v; dist += t*t;

	if (dist < (exact ? 0.001 : DIST_THRESHOLD))
	    return 0;
    }

    return 1;
}

int add_color(int r, int g, int b, int f) {
    cmap_r[total] = r;
    cmap_g[total] = g;
    cmap_b[total] = b;
    LUV_convert(cmap_r[total],cmap_g[total],cmap_b[total],
		Ltab + total, Utab + total, Vtab + total);
    if (f) fprintf(stderr, "forced add of %d %d %d, total = %d...",
		   r, g, b, total);
    if (no_close_color(Ltab[total], Utab[total], Vtab[total], total-1, f)) {
	++total;
	if (f) fprintf(stderr, "added\n");
	return 1;
    } else {
	if (f) fprintf(stderr, "already there?\n");
	return 0;
    }
}

static int red_idx, green_idx, blue_idx, yellow_idx, magenta_idx, cyan_idx;
static int white_idx;

static void init_primaries() {
    add_color(0x00, 0x00, 0x00, TRUE); /* black */
    add_color(0x55, 0x55, 0x55, TRUE); /* dark gray */
    add_color(0xaa, 0xaa, 0xaa, TRUE); /* light gray */
 /* add_color(0xff, 0xff, 0xff, TRUE); white -- belongs in last slot */

    red_idx = total;
    add_color(0xff, 0x00, 0x00, TRUE); /* red */
    add_color(0xaa, 0x00, 0x00, TRUE); /* dark red */
    yellow_idx = total;
    add_color(0xff, 0xff, 0x00, TRUE); /* yellow */
    add_color(0xaa, 0x55, 0x00, TRUE); /* dark yellow */

    green_idx = total;
    add_color(0x00, 0xff, 0x00, TRUE); /* green */
    add_color(0x00, 0xaa, 0x00, TRUE); /* dark green */

    cyan_idx  = total;
    add_color(0x00, 0xff, 0xff, TRUE); /* cyan */
    add_color(0x00, 0x55, 0xaa, TRUE); /* dark cyan */

    blue_idx = total;
    add_color(0x00, 0x00, 0xff, TRUE); /* blue */
    add_color(0x00, 0x00, 0xaa, TRUE); /* dark blue */

    magenta_idx = total;
    add_color(0xff, 0x00, 0xff, TRUE); /* magenta */
    add_color(0x55, 0x00, 0xaa, TRUE); /* dark magenta */

    white_idx = total;
    add_color(0xff, 0xff, 0xff, TRUE); /* white */
}

int find_nearest(CmapEntry *pCmap)
{
    int red = pCmap->red;
    int grn = pCmap->green;
    int blu = pCmap->blue;
    float L = pCmap->L;
    float dist;
    int i, ix;

    if ((red == grn) && (grn == blu)) {

	ix = pCmap->bestidx;
	dist = pCmap->dist;

	/* special case for black */
	if ((L*L) < dist) {
	    ix = 0;
	    dist = (L*L);
	}

	for (i = gray_start; i <= gray_stop; ++i) {
	    float dL = Ltab[i] - L; dL *= dL;
	    if (i == gray_stop) {
		dL = Ltab[15] - L; dL *= dL;
	    }

	    if (dL <= dist) {
		ix = i;
		dist = dL;
	    }

	    if ((dL > dist) && (dist < 2)) break; /* we've past the best one */
	}

	pCmap->dist = dist;
	pCmap->dL = dist;
	pCmap->dE = WEIGHT_DIST(dist*LIGHTNESS_PENALTY,L);
	pCmap->bestidx = ix;
    } else {
	float U = pCmap->U;
	float V = pCmap->V;
	ix = pCmap->bestidx;
	dist = pCmap->dist;

	for (i = 0; i < total; ++i)
	{
	    float dL, dU, dV;

	    /* avoid letting irreversable impurities creep in! */
	    if (
		   ((pCmap->red == 255) && (cmap_r[i] != 255))
		|| ((pCmap->red == 0) && (cmap_r[i] != 0))
		|| ((pCmap->green == 255) && (cmap_g[i] != 255))
		|| ((pCmap->green == 0) && (cmap_g[i] != 0))
		|| ((pCmap->blue == 255) && (cmap_b[i] != 255))
		|| ((pCmap->blue == 0) && (cmap_b[i] != 0))
	    ) {
		continue;
	    }

	    dL = Ltab[i] - L; dL *= (dL*LIGHTNESS_PENALTY);
	    dU = Utab[i] - U; dU *= dU;
	    dV = Vtab[i] - V; dV *= dV;

	    if ((dL + dU + dV) < dist)
	    {
		dist = dL + dU + dV;
		/* *delta = (dL/4) + dU + dV; */
		/* *delta = dist */
		/* *delta = dL + 100*(dU+dV)/(100+L); */
		pCmap->dist = dist;
		pCmap->dE = WEIGHT_DIST(dist, L);
		pCmap->dL = dL/LIGHTNESS_PENALTY;
		pCmap->bestidx = i;

		ix = i;
		if (dist < 5) break;
	    }
	}
    }

    return ix;
}


static void
dump_palette() {
    float dE;
    int r,g,b,i,cnt;
    float max_dL, max_dE, t, ave_dL;
    double ave_dE;
    float L, U, V, dL;

    fprintf(stderr, "Final palette contains:\n");
    for (i = 0; i < 16; ++i) {
	fprintf(stderr, "#%02x%02x%02x\n",
		cmap_r[i],cmap_g[i],cmap_b[i]);
    }
    fflush(stderr);

    printf("static XColor color4_cells[16] = {\n");
    for (i = 0; i < 16; ++i) {
	printf("    {%3d, 0x%02x%02x, 0x%02x%02x, 0x%02x%02x, ", i, 
	       cmap_r[i],cmap_r[i], 
	       cmap_g[i],cmap_g[i],
	       cmap_b[i],cmap_b[i]);

	printf("DoRed | DoGreen | DoBlue, 0},\n");
    }
    printf("};\n\n");

    printf("#define find_index(r,g,b) index_table[((r>>4)<<8)|((g>>4)<<4)|(b>>4)]\n\n");
    printf("static unsigned char index_table[0x1000] = {\n    ");

    max_dL = 0;
    max_dE = 0;
    ave_dL = 0;
    ave_dE = 0;
    cnt = 0;

    for (r = 0; r < 16; ++r)
    {
	for (g = 0; g < 16; ++g)
	{
	    for (b = 0; b < 16; ++b)
	    {
		CmapEntry cmap_entry;
		cmap_entry.red = (int)(floor(r*255.0/15.0));
		cmap_entry.green = (int)(floor(g*255.0/15.0));
		cmap_entry.blue = (int)(floor(b*255.0/15.0));
		LUV_convert(cmap_entry.red, cmap_entry.green, cmap_entry.blue,
			    &cmap_entry.L, &cmap_entry.U, &cmap_entry.V);

		if (r == g && g == b) {
		    cmap_entry.bestidx = white_idx;
		    cmap_entry.dist = ((100.0 - cmap_entry.L)
				       * (100.0 - cmap_entry.L));
		} else {
		    /* this is gross, but if we don't do it,
		       some colors tend to desaturate rather badly. */
		    if (r == 15) {
			if (g == 15) {
			    if (b == 15) {
				cmap_entry.bestidx = white_idx;
			    } else {
				cmap_entry.bestidx = yellow_idx;
			    }
			} else if (b == 15) {
			    cmap_entry.bestidx = magenta_idx;
			} else {
			    cmap_entry.bestidx = red_idx;
			}
		    } else if (g == 15) {
			if (b == 15) {
			    cmap_entry.bestidx = cyan_idx;
			} else {
			    cmap_entry.bestidx = green_idx;
			}
		    } else if (b == 15) {
			cmap_entry.bestidx = blue_idx;
		    } else {
			cmap_entry.bestidx = white_idx;
		    }

		    /* this is the distance for white */
		    cmap_entry.dist = (8*(100-cmap_entry.L)*(100-cmap_entry.L)
				       + cmap_entry.U*cmap_entry.U
				       + cmap_entry.V*cmap_entry.V);
		}

		i = find_nearest(&cmap_entry);
		if ((b&15)==0) printf("\n    ");
		printf("%3d,", i);
		dL = cmap_entry.dL;
		dE = cmap_entry.dE;

		if (dL > max_dL) max_dL = dL;
		t = UNWEIGHT_DIST(dE,dL) - dL*(LIGHTNESS_PENALTY-1);
		if (t > max_dE) max_dE = t;

		ave_dL += (dL > 0) ? sqrt(dL) : 0;
		ave_dE += (t > 0) ? sqrt(t) : 0;
		++cnt;
	    }
	}
    }
    printf("\n};\n\n");
    fflush(stdout);

    fprintf(stderr, "Worst case error dL = %5.3f, dE = %5.3f\n", 
	    sqrt(max_dL), sqrt(max_dE));
    fprintf(stderr, "Average error dL = %5.3f, dE = %5.3f\n", 
	    ave_dL / cnt,  ave_dE / cnt);
}

main() {
    int p;

    recalculate_constants();

    init_matrices();
    init_primaries();

    gray_start = 0; gray_stop = 2;
    fprintf(stderr, "gray_stop = %d\n", gray_stop);

    dump_palette();
    return 0;
}


