/*
 * @(#)palette8.c	1.10 98/11/18
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

/* Iterative color palette generation by Richard Berlin, June 24 1996
 * Rearchitected for speed improvements by Jim Graham, August 1996
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#define TRUE 1
#define FALSE 0

/*
 * This generates a 256-color palette, and an associated 32x32x32
 * lookup table, using a color distribution algorithm based on LUV
 * or LAB rather than RGB.
 *
 * Best results for a given device will be obtained by adjusting the
 * values just below.  The default values given are for a monitor
 * which complies with CCIR Rec 709.  This should be reasonably useful
 * for CRT displays but could be quite far off on another display
 * technology, e.g. flat-panel LCD.
 */

/* ** *** BEGIN INFORMATION SPECIFIC TO THE MONITOR *** ** */
/* ** ***                                           *** ** */
static double monitor_gamma[3] = {
    1.0/.45,			/* gamma of red   channel */
    1.0/.45,			/* gamma of green channel */
    1.0/.45			/* gamma of blue  channel */
};
static double redxy[2]   = {0.6400, 0.3300}; /* x,y coords of red primary */
static double greenxy[2] = {0.3000, 0.6000}; /* x,y coords of green primary */
static double bluexy[2]  = {0.1500, 0.0600}; /* x,y coords of blue primary */
static double whitexy[2] = {0.3127, 0.3291}; /* x,y of white point */
static double whiteY     = 1.0;	             /* nominal Y of white */
/* ** ***                                           *** ** */
/* ** *** END INFORMATION SPECIFIC TO THE MONITOR *** ** */

/* We restrict all colors input to the algorithm to be a multiple of */
/* STEP_SIZE.  17 is a good choice; it corresponds to 4 bits of resolution */
/* in each channel (i.e. a 4096-color cube). */
#define STEP_SIZE 17

static double mat[3][3];	/* RGB to XYZ conversion matrix */
static double uwht, vwht;	/* u,v of white */
static double Rmat[3][256];	/* Pre-multiplied vectors for */
static double Gmat[3][256];	/* conversion of RGB to XYZ */
static double Bmat[3][256];

/* LUV corresponding to palette being constructed. */
static double Ltab[256], Utab[256], Vtab[256];

typedef struct {
    unsigned char red;
    unsigned char green;
    unsigned char blue;
    unsigned char bestidx;
    double L, U, V;
    double dist;
    double dE;
    double dL;
} CmapEntry;

static int virt_cmap_step_size;
static int virt_cmap_checked;
static int num_virt_cmap_entries;
static CmapEntry *virt_cmap;

int gray_start, gray_stop;

/* this is a multiplier--it should not be zero */
#define LIGHTNESS_PENALTY 10
#define WEIGHT 250.0
#ifdef WEIGHT
#define WEIGHT_DIST(d,l)   (WEIGHT*(d)/(WEIGHT+(l)))
#define UNWEIGHT_DIST(d,l) ((WEIGHT+(l))*(d)/WEIGHT)
#else
#define WEIGHT_DIST(d,l) (d)
#define UNWEIGHT_DIST(d,l) (d)
#endif

#define xR redxy[0]
#define yR redxy[1]
#define xG greenxy[0]
#define yG greenxy[1]
#define xB bluexy[0]
#define yB bluexy[1]
#define xW whitexy[0]
#define yW whitexy[1]
static void
recalculate_constants() {
    double D  = xR*(yG - yB) + xG*(yB - yR) + xB*(yR - yG);
    double Cr = (whiteY*(xW*(yG - yB) - yW*(xG - xB) + xG*yB - xB*yG)) / (yW*D);
    double Cg = (whiteY*(xW*(yB - yR) - yW*(xB - xR) - xR*yB + xB*yR)) / (yW*D);
    double Cb = (whiteY*(xW*(yR - yG) - yW*(xR - xG) + xR*yG - xG*yR)) / (yW*D);

    mat[0][0] = xR*Cr;
    mat[0][1] = yR*Cr;
    mat[0][2] = (1 - xR - yR)*Cr;
    mat[1][0] = xG*Cg;
    mat[1][1] = yG*Cg;
    mat[1][2] = (1 - xG - yG)*Cg;
    mat[2][0] = xB*Cb;
    mat[2][1] = yB*Cb;
    mat[2][2] = (1 - xB - yB)*Cb;

    uwht = 4*whitexy[0]/(-2*whitexy[0] + 12*whitexy[1] + 3);
    vwht = 9*whitexy[1]/(-2*whitexy[0] + 12*whitexy[1] + 3);
}
#undef yW
#undef xW
#undef yB
#undef xB
#undef yG
#undef xG
#undef yR
#undef xR

static void
init_matrices()
{
    int i;

    for (i = 0; i < 256; ++i)
    {
	double iG = pow(i/255.0, monitor_gamma[0]);
	Rmat[0][i] = mat[0][0] * iG;
	Rmat[1][i] = mat[0][1] * iG;
	Rmat[2][i] = mat[0][2] * iG;

	iG = pow(i/255.0, monitor_gamma[1]);
	Gmat[0][i] = mat[1][0] * iG;
	Gmat[1][i] = mat[1][1] * iG;
	Gmat[2][i] = mat[1][2] * iG;

	iG = pow(i/255.0, monitor_gamma[2]);
	Bmat[0][i] = mat[2][0] * iG;
	Bmat[1][i] = mat[2][1] * iG;
	Bmat[2][i] = mat[2][2] * iG;
    }
}

static void
LUV_convert(int red, int grn, int blu, double *L, double *u, double *v)
{
    double X = Rmat[0][red] + Gmat[0][grn] + Bmat[0][blu];
    double Y = Rmat[1][red] + Gmat[1][grn] + Bmat[1][blu];
    double Z = Rmat[2][red] + Gmat[2][grn] + Bmat[2][blu];

    double sum  = X+Y+Z;
    double x    = (sum == 0) ? 0 : (X/sum);
    double y    = (sum == 0) ? 0 : (Y/sum);
    double dnm  = -2*x + 12*y + 3;
    double uprm = 4*x/dnm;
    double vprm = 9*y/dnm;
    double ytmp = pow(Y/whiteY, 1.0/3.0);

    if (ytmp < .206893) {
	*L = 903.3*Y/whiteY;
    } else {
	*L = 116*(ytmp) - 16;
    }
    *u = 13*(*L)*(uprm-uwht);
    *v = 13*(*L)*(vprm-vwht);
}

int total;
unsigned char cmap_r[256], cmap_g[256], cmap_b[256];

static int DIST_THRESHOLD = 10;
int no_close_color(double l, double u, double v, int c_tot, int exact) {
    int i;
    for (i = 0; i < c_tot; ++i) {
	double t, dist = 0;
	t = Ltab[i] - l; dist += t*t*LIGHTNESS_PENALTY;
	t = Utab[i] - u; dist += t*t;
	t = Vtab[i] - v; dist += t*t;

	if (dist < (exact ? 0.001 : DIST_THRESHOLD))
	    return 0;
    }

    return 1;
}

static int red_idx, green_idx, blue_idx, yellow_idx, magenta_idx, cyan_idx;
static int white_idx;
int add_color(int r, int g, int b, int f) {

    if (r == 255) {
	if ((g == 255) && (b == 255)) {
	    white_idx = total;
	} else if ((g | b) == 0) {
	    red_idx = total;
	} else if ((g == 255) && (b == 0)) {
	    yellow_idx = total;
	} else if ((g == 0) && (b == 255)) {
	    magenta_idx = total;
	}
    } else if (g == 255) {
	if ((r | b) == 0) {
	    green_idx = total;
	} else if ((r == 0) && (b == 255)) {
	    cyan_idx = total;
	}
    } else if (b == 255) {
	if ((r | g) == 0) {
	    blue_idx = total;
	}
    }

    cmap_r[total] = r;
    cmap_g[total] = g;
    cmap_b[total] = b;
    LUV_convert(cmap_r[total],cmap_g[total],cmap_b[total],
		Ltab + total, Utab + total, Vtab + total);
    if (f) fprintf(stderr, "forced add of %d %d %d, total = %d...",
		   r, g, b, total);
    if (no_close_color(Ltab[total], Utab[total], Vtab[total], total-1, f)) {
	++total;
	if (f) fprintf(stderr, "added\n");
	return 1;
    } else {
	if (f) fprintf(stderr, "already there?\n");
	return 0;
    }
}

static void init_primaries() {
    int r, g, b;

    for (r = 0; r < 256; r += 85) {
	for (g = 0; g < 256; g += 85) {
	    for (b = 0; b < 256; b += 85) {
		if ((r == g) && (g == b)) continue; /* black or white */
		add_color(r, g, b, TRUE);
	    }
	}
    }
}

static void init_pastels() {
    int i;
    /* very light colors */
    for (i = 6; i > 0; --i)
	add_color((i&4) ? 0xff : 0xee,
		  (i&2) ? 0xff : 0xee,
		  (i&1) ? 0xff : 0xee, TRUE);
}

static void init_grays() {
    int i;
    for (i = STEP_SIZE; i < 255; i += STEP_SIZE)
	add_color(i, i, i, TRUE);
}

static void init_touchable_palette() {
    add_color(0x33, 0x00, 0x99, TRUE);
    add_color(0xaa, 0xaa, 0xee, TRUE);
    add_color(0xee, 0x88, 0x66, TRUE);
    add_color(0xff, 0xbb, 0xaa, TRUE);
    add_color(0xff, 0xcc, 0xbb, TRUE);
    add_color(0xff, 0xdd, 0xdd, TRUE);
    add_color(0xff, 0xee, 0xee, TRUE);
}

static void init_colors_from_file(FILE *f)
{
    char buf[1024];
    while (!feof(f)) {
	int r, g, b;
	char *p = fgets(buf, sizeof(buf), f);
	if (sscanf(buf, "%d%d%d", &r, &g, &b) == 3) {
	    /* white was moved to 255 and hence won't be found */
	    /* by the duplicate detection.  Prevent that bug here. */
	    if ((r&g&b) != 255) {
		add_color(r, g, b, TRUE);
	    }
	}
    }
}

static void
init_virt_cmap(int step_size)
{
    int r, g, b;
    int steps;
    CmapEntry *pCmap;

    if (virt_cmap) {
	if (virt_cmap_step_size == step_size) {
	    return;
	}
	free(virt_cmap);
	virt_cmap = 0;
    }

    steps = 255 / step_size + 1;
    num_virt_cmap_entries = steps * steps * steps;
    virt_cmap = malloc(sizeof(CmapEntry) * num_virt_cmap_entries);
    pCmap = virt_cmap;
    for (r = 255; r >= 0; r -= step_size) {
	for (g = 255; g >= 0; g -= step_size) {
	    for (b = 255; b >= 0; b -= step_size) {
		if (pCmap >= virt_cmap + num_virt_cmap_entries) {
		    fprintf(stderr, "OUT OF pCmap CONVERSION SPACE!\n");
		    exit(1);
		}
		pCmap->red = r;
		pCmap->green = g;
		pCmap->blue = b;
		LUV_convert(r,g,b, &pCmap->L, &pCmap->U, &pCmap->V);
		pCmap->bestidx = 255;
		if (r == g && g == b) {
		    pCmap->dist = (100.0 - pCmap->L) * (100.0 - pCmap->L);
		} else {
		    pCmap->dist = (8*(100-pCmap->L)*(100-pCmap->L)
				   + pCmap->U*pCmap->U + pCmap->V*pCmap->V);
		}
		pCmap++;
	    }
	}
    }
    if (pCmap < virt_cmap + num_virt_cmap_entries) {
	fprintf(stderr, "Didn't fill pCmap conversion table!\n");
	exit(1);
    }
    virt_cmap_step_size = step_size;
    virt_cmap_checked = 0;
}

int find_nearest(CmapEntry *pCmap)
{
    int red = pCmap->red;
    int grn = pCmap->green;
    int blu = pCmap->blue;
    double L = pCmap->L;
    double dist;
    int i, ix;

    if ((red == grn) && (grn == blu)) {

	ix = pCmap->bestidx;
	dist = pCmap->dist;

	/* special case for black */
	if ((L*L) < dist) {
	    ix = 0;
	    dist = (L*L);
	}

	for (i = gray_start; i <= gray_stop; ++i) {
	    double dL = Ltab[i] - L; dL *= dL;
	    if (i == gray_stop) {
		dL = Ltab[255] - L; dL *= dL;
	    }

	    if (dL <= dist) {
		ix = i;
		dist = dL;
	    }

	    if ((dL > dist) && (dist < 2)) break; /* we've past the best one */
	}

	pCmap->dist = dist;
	pCmap->dL = dist;
	pCmap->dE = WEIGHT_DIST(dist*LIGHTNESS_PENALTY,L);
	pCmap->bestidx = ix;
    } else {
	double U = pCmap->U;
	double V = pCmap->V;
	ix = pCmap->bestidx;
	dist = pCmap->dist;

	for (i = virt_cmap_checked; i < total; ++i)
	{
	    double dL, dU, dV;

	    /* avoid letting irreversable impurities creep in! */
	    if (
		   ((pCmap->red == 255) && (cmap_r[i] != 255))
		|| ((pCmap->red == 0) && (cmap_r[i] != 0))
		|| ((pCmap->green == 255) && (cmap_g[i] != 255))
		|| ((pCmap->green == 0) && (cmap_g[i] != 0))
		|| ((pCmap->blue == 255) && (cmap_b[i] != 255))
		|| ((pCmap->blue == 0) && (cmap_b[i] != 0))
	    ) {
		continue;
	    }

	    dL = Ltab[i] - L; dL *= (dL*LIGHTNESS_PENALTY);
	    dU = Utab[i] - U; dU *= dU;
	    dV = Vtab[i] - V; dV *= dV;

	    if ((dL + dU + dV) < dist)
	    {
		dist = dL + dU + dV;
		/* *delta = (dL/4) + dU + dV; */
		/* *delta = dist */
		/* *delta = dL + 100*(dU+dV)/(100+L); */
		pCmap->dist = dist;
		pCmap->dE = WEIGHT_DIST(dist, L);
		pCmap->dL = dL/LIGHTNESS_PENALTY;
		pCmap->bestidx = i;

		ix = i;
		if (dist < 5) break;
	    }
	}
    }

    return ix;
}


#define MAX_OFFENDERS 32
CmapEntry *offenders[MAX_OFFENDERS];
int num_offenders;

static void
insert_in_list(CmapEntry *pCmap)
{
    int i, j;
    double dE = pCmap->dE;

    for (i = num_offenders; i > 0; --i) {
	if (dE < offenders[i-1]->dE) break;
    }

    if (i < MAX_OFFENDERS) {
	if (num_offenders < MAX_OFFENDERS) ++num_offenders;

	for (j = num_offenders-1; j > i; --j) {
	    offenders[j] = offenders[j-1];
	}
	offenders[i] = pCmap;
    }
}

static void
handle_biggest_offenders(int step_size) {
    int i, j;
    double max_dL = 0;
    double max_dE = 0;
    double t;
    CmapEntry *pCmap;

    num_offenders = 0;
    init_virt_cmap(step_size);

    pCmap = virt_cmap;
    for (i = 0; i < num_virt_cmap_entries; i++) {
	double dL, dE;
	find_nearest(pCmap);
	dL = pCmap->dL;
	dE = pCmap->dE;
	if (dL > max_dL) max_dL = dL;
	t = UNWEIGHT_DIST(dE,dL) - dL*(LIGHTNESS_PENALTY-1);
	if (t > max_dE) max_dE = t;
	insert_in_list(pCmap);
	pCmap++;
    }

    virt_cmap_checked = total;

    fprintf(stderr, "Maximum error in L* = %5.2f\n", sqrt(max_dL));
    fprintf(stderr, "Maximum delta E = %5.2f\n", sqrt(max_dE));
    /* total < 255, not 256, because the last element contains white */
    for (i = 0; (total < 255) && (i < num_offenders); ++i) {
	double threshold;
	int ix, curr_total;
	double t;
	double deltaE;
	CmapEntry *offender = offenders[i];

	if (!offender || offender->dE <= 2.0) continue;

	threshold = .5 * UNWEIGHT_DIST(offender->dE, offender->L);
	ix = offender->bestidx;
	deltaE = 0.0;

	t = offender->V - Vtab[ix]; t *= t; deltaE += t;
	t = offender->U - Utab[ix]; t *= t; deltaE += t;
	t = offender->L - Ltab[ix]; t *= t; deltaE += t;

	deltaE = sqrt(deltaE);

	fprintf(stderr, "adding #%02x%02x%02x (key %5.1f thrsh %5.1f err %5.1f dL %5.1f)...", 
		offender->red, offender->green, offender->blue, 
		offender->dE, threshold, deltaE, sqrt(t));

	if (add_color(offender->red, offender->green, offender->blue, FALSE)) {
	    fprintf(stderr, "done\n");
	} else
	    fprintf(stderr, "nearby color already there\n");

	curr_total = total;
	for (j = i+1; j < num_offenders; ++j) {
	    double d, t;

	    if (!offenders[j]) {
		continue;
	    }

	    d = 0;
	    t = offender->L - offenders[j]->L; t *= 4*t; d += t;
	    t = offender->U - offenders[j]->U; t *= t; d += t;
	    t = offender->V - offenders[j]->V; t *= t; d += t;

	    if (d < threshold) {
		offenders[j] = 0;
#if 0
		fprintf(stderr,
			"    Discarding nearby color #%02x%02x%02x (%f)\n",
			offenders[j].r, offenders[j].g, offenders[j].b, d);
#endif
	    }
	}
	if ((total == curr_total) && (DIST_THRESHOLD > 1)) {
	    --i;		/* try this slot again */
	    --DIST_THRESHOLD;
	}
    }
}

static void
dump_palette() {
    int r, g, b, i, cnt;
    double max_dL, max_dE, t, ave_dL, ave_dE, dL, dE;

    fprintf(stderr, "Final palette contains:\n");
    for (i = 0; i < 256; ++i) {
	fprintf(stderr, "#%02x%02x%02x\n",
		cmap_r[i],cmap_g[i],cmap_b[i]);
    }
    fflush(stderr);

#ifdef unix
    printf("static XColor color8_cells[256] = {\n");
    for (i = 0; i < 256; ++i) {
	printf("    {%3d, 0x%02x%02x, 0x%02x%02x, 0x%02x%02x, ", i, 
	       cmap_r[i],cmap_r[i], 
	       cmap_g[i],cmap_g[i],
	       cmap_b[i],cmap_b[i]);

	printf("DoRed | DoGreen | DoBlue, 0},\n");
    }
    printf("};\n\n");

    printf("#define find_index(r,g,b) index_table[((r>>3)<<10)|((g>>3)<<5)|(b>>3)]\n\n");
    printf("unsigned char index_table[0x8000] = {");
#else
    printf("unsigned char awt_reds[256] = {");
    for (i = 0; i < 256; ++i) {
	if ((i&15)==0) printf("\n    ");
	printf("%3d,", cmap_r[i]);
    }
    printf("\n};\n");
    printf("unsigned char awt_greens[256] = {");
    for (i = 0; i < 256; ++i) {
	if ((i&15)==0) printf("\n    ");
	printf("%3d,", cmap_g[i]);
    }
    printf("\n};\n");
    printf("unsigned char awt_blues[256] = {");
    for (i = 0; i < 256; ++i) {
	if ((i&15)==0) printf("\n    ");
	printf("%3d,", cmap_b[i]);
    }
    printf("\n};\n");
    printf("unsigned char AwtImage::cube[LOOKUPSIZE * LOOKUPSIZE * LOOKUPSIZE] = {");
#endif

    max_dL = 0;
    max_dE = 0;
    ave_dL = 0;
    ave_dE = 0;
    cnt = 0;
    virt_cmap_checked = 0;

    for (r = 0; r < 32; ++r)
    {
	for (g = 0; g < 32; ++g)
	{
	    for (b = 0; b < 32; ++b)
	    {
		CmapEntry cmap_entry;
		cmap_entry.red = (int)(floor(r*255.0/31.0));
		cmap_entry.green = (int)(floor(g*255.0/31.0));
		cmap_entry.blue = (int)(floor(b*255.0/31.0));
		LUV_convert(cmap_entry.red, cmap_entry.green, cmap_entry.blue,
			    &cmap_entry.L, &cmap_entry.U, &cmap_entry.V);
		if (r == g && g == b) {
		    cmap_entry.bestidx = white_idx;
		    cmap_entry.dist = ((100.0 - cmap_entry.L)
				       * (100.0 - cmap_entry.L));
		} else {
		    /* this is gross, but if we don't do it,
		       some colors tend to desaturate rather badly. */
		    if (r == 15) {
			if (g == 15) {
			    if (b == 15) {
				cmap_entry.bestidx = white_idx;
			    } else {
				cmap_entry.bestidx = yellow_idx;
			    }
			} else if (b == 15) {
			    cmap_entry.bestidx = magenta_idx;
			} else {
			    cmap_entry.bestidx = red_idx;
			}
		    } else if (g == 15) {
			if (b == 15) {
			    cmap_entry.bestidx = cyan_idx;
			} else {
			    cmap_entry.bestidx = green_idx;
			}
		    } else if (b == 15) {
			cmap_entry.bestidx = blue_idx;
		    } else {
			cmap_entry.bestidx = white_idx;
		    }

		    /* this is the distance for white, plus a little */
		    cmap_entry.dist = (8*(100-cmap_entry.L)*(100-cmap_entry.L)
				       + cmap_entry.U*cmap_entry.U
				       + cmap_entry.V*cmap_entry.V)
			+ 1;
		}

		i = find_nearest(&cmap_entry);
		if ((b&15)==0) printf("\n    ");
		printf("%3d,", i);
		dL = cmap_entry.dL;
		dE = cmap_entry.dE;

		if (dL > max_dL) max_dL = dL;
		t = UNWEIGHT_DIST(dE,dL) - dL*(LIGHTNESS_PENALTY-1);
		if (t > max_dE) max_dE = t;

		ave_dL += (dL > 0) ? sqrt(dL) : 0;
		ave_dE += (t > 0) ? sqrt(t) : 0;
		++cnt;
	    }
	}
    }
    printf("\n};\n\n");
    fflush(stdout);

    fprintf(stderr, "Worst case error dL = %5.3f, dE = %5.3f\n", 
	    sqrt(max_dL), sqrt(max_dE));
    fprintf(stderr, "Average error dL = %5.3f, dE = %5.3f\n", 
	    ave_dL / cnt,  ave_dE / cnt);
}

main(int argc, char *argv[]) {
    int p;

    recalculate_constants();

    init_matrices();

    /* we have to ensure that element 0 is black and 255 is white */
    total = 255;
    add_color(255,255,255, TRUE);
    total = 0;
    add_color(0, 0, 0, TRUE);

    /* do grays next; otherwise find_nearest may break! */
    gray_start = total;  init_grays(); gray_stop = total;
    fprintf(stderr, "gray_stop = %d\n", gray_stop);

    init_primaries();
    init_pastels();

    if (argc < 2) {
	init_touchable_palette();
    } else {
	for (++argv; --argc > 0; ++argv) {
	    FILE *f = fopen(*argv, "r");
	    if (f != NULL) {
		fprintf(stderr, "Reading colors from file %s\n", *argv);
		init_colors_from_file(f);
		fclose(f);
	    }
	}
    }

    p = 0;
    while (total < 255) {
	fprintf(stderr, "pass %d, total colors in palette = %d\n", ++p, total);
	handle_biggest_offenders(STEP_SIZE);
    }

    dump_palette();
    return 0;
}


