/*
 * @(#)FontMetricsImpl.java	1.1 98/07/27
 * 
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 * 
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

package sun.awt.gfW;

import java.awt.*;
import java.util.Hashtable;
import sun.awt.FontDescriptor;
import sun.awt.PlatformFont;
import sun.awt.CharsetString;
import java.io.CharConversionException;


/** 
 * A font metrics object for a WServer font.
 * 
 * @version 1.6 24 Nov 1995
 * @author Jim Graham
 */
class FontMetricsImpl extends FontMetrics {
    /**
     * The widths of the first 256 characters.
     */
    int widths[];

    /** 
     * The standard ascent of the font.  This is the logical height
     * above the baseline for the Alphanumeric characters and should
     * be used for determining line spacing.  Note, however, that some
     * characters in the font may extend above this height.
     */
    int ascent;

    /** 
     * The standard descent of the font.  This is the logical height
     * below the baseline for the Alphanumeric characters and should
     * be used for determining line spacing.  Note, however, that some
     * characters in the font may extend below this height.
     */
    int descent;

    /** 
     * The standard leading for the font.  This is the logical amount
     * of space to be reserved between the descent of one line of text
     * and the ascent of the next line.  The height metric is calculated
     * to include this extra space.
     */
    int leading;

    /** 
     * The standard height of a line of text in this font.  This is
     * the distance between the baseline of adjacent lines of text.
     * It is the sum of the ascent+descent+leading.  There is no
     * guarantee that lines of text spaced at this distance will be
     * disjoint; such lines may overlap if some characters overshoot
     * the standard ascent and descent metrics.
     */
    int height;

    /** 
     * The maximum ascent for all characters in this font.  No character
     * will extend further above the baseline than this metric.
     */
    int maxAscent;

    /** 
     * The maximum descent for all characters in this font.  No character
     * will descend further below the baseline than this metric.
     */
    int maxDescent;

    /** 
     * The maximum possible height of a line of text in this font.
     * Adjacent lines of text spaced this distance apart will be
     * guaranteed not to overlap.  Note, however, that many paragraphs
     * that contain ordinary alphanumeric text may look too widely
     * spaced if this metric is used to determine line spacing.  The
     * height field should be preferred unless the text in a given
     * line contains particularly tall characters.
     */
    int maxHeight;

    /** 
     * The maximum advance width of any character in this font. 
     */
    int maxAdvance;

    /**
     * Calculate the metrics from the given WServer and font.
     */
    public FontMetricsImpl(Font font) {
	super(font);
	init();
    }

    /**
     * Get leading
     */
    public int getLeading() {
	return leading;
    }

    /**
     * Get ascent.
     */
    public int getAscent() {
	return ascent;
    }

    /**
     * Get descent
     */
    public int getDescent() {
	return descent;
    }

    /**
     * Get height
     */
    public int getHeight() {
	return height;
    }

    /**
     * Get maxAscent
     */
    public int getMaxAscent() {
	return maxAscent;
    }

    /**
     * Get maxDescent
     */
    public int getMaxDescent() {
	return maxDescent;
    }

    /**
     * Get maxAdvance
     */
    public int getMaxAdvance() {
	return maxAdvance;
    }

    /** 
     * Return the width of the specified string in this Font. 
     */
    public int stringWidth(String string) {
	return charsWidth(string.toCharArray(), 0, string.length());
    }

    /**
     * Return the width of the specified char[] in this Font.
     */
    public int charsWidth(char chars[], int offset, int length) {
	Font font = getFont();
	PlatformFont pf = ((PlatformFont) font.getPeer());
	if (pf.mightHaveMultiFontMetrics()) {
	    return getMFStringWidth(chars, offset, length, font, pf);
	} else {
	    if (widths != null) {
		int w = 0;
		for (int i = offset; i < offset + length; i++) {
		    int ch = chars[i];
		    if (ch < 0 || ch >= widths.length) {
			w += maxAdvance;
		    } else {
			w += widths[ch];
		    }
		}
		return w;
	    } else {
		return maxAdvance * length;
	    }
	}
    }

    /*
     * get multi font string width with a multiple native font
     */
    private int getMFStringWidth(char string[], int offset, int length, 
				 Font font, PlatformFont pf) {
        if (length == 0) {
            return 0;
        }

	sun.awt.CharsetString[] css = 
            pf.makeMultiCharsetString(string, offset, length);

	byte[] csb = null;
        int baLen = 0;
	int w = 0;
	for (int i = 0; i < css.length; i++) {
	    CharsetString cs = css[i];
            if (needsConversion(font, cs.fontDescriptor)) {
                try {
                    int l = cs.fontDescriptor.fontCharset.getMaxBytesPerChar()
                        * cs.length;
                    if (l > baLen) {
                        baLen = l;
                        csb = new byte[baLen];
                    }
                    int csbLen = cs.fontDescriptor.fontCharset.convert(
				  cs.charsetChars, 
                                  cs.offset, cs.offset + cs.length,
				  csb, 0, csb.length);
		
                    w += getMFCharSegmentWidth(font, cs.fontDescriptor, true,
                                               null, 0, 0, csb, csbLen);
                } catch (CharConversionException ignored) {
                }
            } else {
                w += getMFCharSegmentWidth(font, cs.fontDescriptor, false,
                                           cs.charsetChars, cs.offset, 
                                           cs.length, null, 0);
            }
	}
	return w;
    }

    static native boolean needsConversion(Font font, FontDescriptor des);

    // Get the width of the char segment, which is in the native charset
    // of font.
    private native int getMFCharSegmentWidth(
        Font font, FontDescriptor des, boolean converted, 
        char[] chars, int offset, int charLength,
        byte[] bytes, int byteLength);

    /**
     * Return the width of the specified byte[] in this Font. 
     */
    public native int bytesWidth(byte data[], int offset, int len);

    /**
     * Get the widths of the first 256 characters in the font.
     */
    public int[] getWidths() {
	return widths;
    }

    native void init();

    static Hashtable table = new Hashtable();
    
    static FontMetrics getFontMetrics(Font font) {
	FontMetrics fm = (FontMetrics)table.get(font);
	if (fm == null) {
	    table.put(font, fm = new FontMetricsImpl(font));
	}
	return fm;
    }

    private native static void initJNI();
    static { initJNI(); }
}
