/*
 * @(#)GraphicsImpl.java	1.5 98/11/25
 * 
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 * 
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

package sun.awt.gfW;                  

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Image;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.image.ImageObserver;
import java.awt.image.ImageProducer;
import java.io.CharConversionException;
import java.util.Enumeration;
import sun.awt.CharsetString;
import sun.awt.FontDescriptor;
import sun.awt.PlatformFont;
import sun.awt.image.ImageRepresentation;
import sun.porting.graphicssystem.Drawable;
import sun.porting.graphicssystem.GeometryProvider;
import sun.porting.graphicssystem.Region;


/**
 * An implementation of Graphics which maps AWT drawing primitives to equivalent
 * Win32 API calls.
 *
 * @version 1.5, 11/25/98
 * @author Martin Robins
 */
class GraphicsImpl extends Graphics
{
   int                pData;
   Drawable           drawable;
   GeometryProvider   geometry;
   GraphicsSystem     graphicsSystem;
   Color              foreground;
   Font               font;
   Point              ptTrans;
   boolean            chgTrans;
   boolean            chgClipGraphics;
   int                lastGeometryID;
   Region             rgnClipGraphics;


   /**
    * Constructs a new GraphicsImpl object for drawing into a specified Drawable
    * object and clipped according to a specified GeometryProvider which 
    * typically represents the visible area of a window within the WindowSystem.
    * 
    * @param  d a Drawable object representing the drawing surface (context).
    * @param  p a GeometryProvider object maintining clipping bounds.
    * @param  g the GraphicsSystem object.
    */
   public GraphicsImpl( Drawable d, GeometryProvider p, GraphicsSystem g )
   {
      drawable        = d;
      geometry        = p;
      graphicsSystem  = g;
      nInit( drawable );
      ptTrans         = new Point( 0, 0 );
      chgTrans        = false;
      chgClipGraphics = false;
      lastGeometryID  = 0;
      setColor( Color.black );

      validateGeometry();
   }


   /**
    * Constructs a GraphicsImpl object based on an existing object.
    * 
    * @param  g the GraphicsImpl object to copy.
    */
   protected GraphicsImpl( GraphicsImpl g ) 
   {
      drawable        = g.drawable;
      geometry        = g.geometry;
      graphicsSystem  = g.graphicsSystem;
      font            = g.font;
      nInit( drawable );
      ptTrans         = new Point( g.ptTrans.x, g.ptTrans.y );
      chgTrans        = true;
      chgClipGraphics = true;
      lastGeometryID  = 0;
      if (g.rgnClipGraphics != null) 
      {
         rgnClipGraphics = g.rgnClipGraphics.copy();
      }

      setColor( g.foreground );

      validateGeometry();
   }


   /** 
    * Clears the specified rectangle by filling it with the background
    * color of the current drawing surface. This operation does not 
    * use the current paint mode. 
    * <p>
    * Beginning with Java&nbsp;1.1, the background color 
    * of offscreen images may be system dependent. Applications should 
    * use <code>setColor</code> followed by <code>fillRect</code> to 
    * ensure that an offscreen image is cleared to a specific color. 
    * @param       x the <i>x</i> coordinate of the rectangle to clear.
    * @param       y the <i>y</i> coordinate of the rectangle to clear.
    * @param       width the width of the rectangle to clear.
    * @param       height the height of the rectangle to clear.
    * @see         java.awt.Graphics#fillRect(int, int, int, int)
    * @see         java.awt.Graphics#drawRect
    * @see         java.awt.Graphics#setColor(java.awt.Color)
    * @see         java.awt.Graphics#setPaintMode
    * @see         java.awt.Graphics#setXORMode(java.awt.Color)
    * @since       JDK1.0
    */
   public void clearRect(int x, int y, int width, int height) 
   {
      Color   background;


      synchronized (geometry.getLock()) 
      {
         validateGeometry();
         background = geometry.getBackgroundColor();
         nSetColor( background.getRGB() );
         nFillRect( x, y, width, height );
         nSetColor( foreground.getRGB() );
      }
   }


   /** 
    * Intersects the current clip with the specified rectangle.
    * The resulting clipping area is the intersection of the current
    * clipping area and the specified rectangle.
    * This method can only be used to make the current clip smaller.
    * To set the current clip larger, use any of the setClip methods.
    * Rendering operations have no effect outside of the clipping area.
    * @param x the x coordinate of the rectangle to intersect the clip with
    * @param y the y coordinate of the rectangle to intersect the clip with
    * @param width the width of the rectangle to intersect the clip with
    * @param height the height of the rectangle to intersect the clip with
    * @see #setClip(int, int, int, int)
    * @see #setClip(Shape)
    */
   public void clipRect(int x, int y, int width, int height)
   {
      if (rgnClipGraphics != null)
      {
         rgnClipGraphics.intersect( x + ptTrans.x
                                  , y + ptTrans.y
                                  , width
                                  , height
                                  );
      }
      else
      {
         rgnClipGraphics = graphicsSystem.makeRegion( x + ptTrans.x
                                                    , y + ptTrans.y
                                                    , width        
                                                    , height       
                                                    );             
      }

      chgClipGraphics = true;
   }
   

   /**
    * Copies an area of the component by a distance specified by 
    * <code>dx</code> and <code>dy</code>. From the point specified
    * by <code>x</code> and <code>y</code>, this method
    * copies downwards and to the right.  To copy an area of the 
    * component to the left or upwards, specify a negative value for 
    * <code>dx</code> or <code>dy</code>.
    * If a portion of the source rectangle lies outside the bounds 
    * of the component, or is obscured by another window or component, 
    * <code>copyArea</code> will be unable to copy the associated
    * pixels. The area that is omitted can be refreshed by calling 
    * the component's <code>paint</code> method.
    * @param       x the <i>x</i> coordinate of the source rectangle.
    * @param       y the <i>y</i> coordinate of the source rectangle.
    * @param       width the width of the source rectangle.
    * @param       height the height of the source rectangle.
    * @param       dx the horizontal distance to copy the pixels.
    * @param       dy the vertical distance to copy the pixels.
    * @since       JDK1.0
    */
   public void copyArea(int x, int y, int width, int height, int dx, int dy)
   {
      if ((width <= 0) || (height <= 0) || ((dx == 0) && (dy == 0))) 
      {
         return;
      }

      synchronized (geometry.getLock()) 
      {
         validateGeometry();

         Rectangle   rcBounds = geometry.getBounds();
         Region      rgnClip  = getClipRegion();
  
         if (rgnClip == null) 
         {
            return;
         }
        
         /* Calculate the destination region allowing for clipping */
         Region   rgnDst = graphicsSystem.makeRegion( x + rcBounds.x + dx
                                                    , y + rcBounds.y + dy
                                                    , width
                                                    , height
                                                    );
         rgnDst.intersect( rgnClip );
        
         /* Calculate the source region allowing for clipping */
         Region   rgnSrc = rgnDst.copy();
         rgnSrc.translate( -dx, -dy );
         rgnSrc.intersect( rgnClip );
        
         /* Copy the clipped area */
         Rectangle   rcSrc = rgnSrc.getBounds();
         nCopyArea( rcSrc.x - rcBounds.x
                  , rcSrc.y - rcBounds.y
                  , rcSrc.width
                  , rcSrc.height
                  , dx
                  , dy
                  );
        
         /* compute an exposure region (consisting of source areas that */
         /* could not be copied to the destination).  Pass the region   */
         /* off to the window system to generate the repaint events.    */
         rgnSrc.translate( dx, dy );
         rgnDst.subtract( rgnSrc );
         
         /* remainder in dst is the area that must be exposed */
         graphicsSystem.sendGraphicsChange( rgnDst );
      }
   }


   /**
    * Creates a new <code>Graphics</code> object that is 
    * a copy of this <code>Graphics</code> object.
    * @return     a new graphics context that is a copy of 
    *             this graphics context.
    * @since      JDK1.0
    */
   public Graphics create()
   { 
      return (new GraphicsImpl( this ));
   }


   /**
    * Draws the outline of a circular or elliptical arc 
    * covering the specified rectangle.
    * <p>
    * The resulting arc begins at <code>startAngle</code> and extends  
    * for <code>arcAngle</code> degrees, using the current color.
    * Angles are interpreted such that 0&nbsp;degrees 
    * is at the 3&nbsp;o'clock position. 
    * A positive value indicates a counter-clockwise rotation
    * while a negative value indicates a clockwise rotation.
    * <p>
    * The center of the arc is the center of the rectangle whose winBounds 
    * is (<i>x</i>,&nbsp;<i>y</i>) and whose size is specified by the 
    * <code>width</code> and <code>height</code> arguments. 
    * <p>
    * The resulting arc covers an area 
    * <code>width&nbsp;+&nbsp;1</code> pixels wide
    * by <code>height&nbsp;+&nbsp;1</code> pixels tall.
    * @param        x the <i>x</i> coordinate of the 
    *                    upper-left corner of the arc to be drawn.
    * @param        y the <i>y</i>  coordinate of the 
    *                    upper-left corner of the arc to be drawn.
    * @param        width the width of the arc to be drawn.
    * @param        height the height of the arc to be drawn.
    * @param        startAngle the beginning angle.
    * @param        arcAngle the angular extent of the arc, 
    *                    relative to the start angle.
    * @see         java.awt.Graphics#fillArc
    * @since       JDK1.0
    */
   public void drawArc(int x, int y, int width, int height, int startAngle, int arcAngle) 
   {
      synchronized (geometry.getLock()) 
      {
         validateGeometry();
         nDrawArc( x, y, width, height, startAngle, arcAngle );
      }
   }


   /** 
    * Draws the text given by the specified byte array, using this 
    * graphics context's current font and color. The baseline of the 
    * first character is at position (<i>x</i>,&nbsp;<i>y</i>) in this 
    * graphics context's coordinate system.
    * @param data the data to be drawn
    * @param offset the start offset in the data
    * @param length the number of bytes that are drawn
    * @param x the <i>x</i> coordinate of the baseline of the text
    * @param y the <i>y</i> coordinate of the baseline of the text
    * @see         java.awt.Graphics#drawChars
    * @see         java.awt.Graphics#drawString
    * @since       JDK1.0
    */
   public void drawBytes(byte data[], int offset, int length, int x, int y)
   {
      synchronized (geometry.getLock()) 
      {
         validateGeometry();
         nDrawBytes( data, offset, length, x, y );
      }
   }


   /**
    * Draws an image at x,y in nonblocking mode with a callback object.
    */
   public boolean drawImage(Image img, int x, int y, ImageObserver observer) 
   {
      ImageRepresentation   ir;


      ir = imageCheck( img, observer );

      if (ir != null) 
      {
         validateGeometry();
         return (ir.drawImage( this, x, y, null, observer ));
      }
      else
      {
         return (false);
      }
   }


   /**
    * Draws an image scaled to x,y,w,h in nonblocking mode with a
    * callback object.
    */
   public boolean drawImage(Image img, int x, int y, int width, int height, ImageObserver observer)
   {
      ImageRepresentation   ir;


      ir = imageCheck( img, observer );

      if (ir != null) 
      {
         validateGeometry();
         return (ir.drawScaledImage( this, x, y, width, height, null, observer ));
      }
      else
      {
         return (false);
      }
   }


   /**
    * Draws an image at x,y in nonblocking mode with a solid background
    * color and a callback object.
    */
   public boolean drawImage(Image img, int x, int y, Color bgcolor, ImageObserver observer) 
   { 
      ImageRepresentation   ir;


      ir = imageCheck( img, observer );

      if (ir != null) 
      {
         validateGeometry();
         return (ir.drawImage( this, x, y, bgcolor, observer ));
      }
      else
      {
         return (false);
      }
   }


   /**
    * Draws an image scaled to x,y,w,h in nonblocking mode with a
    * solid background color and a callback object.
    */
   public boolean drawImage(Image img, int x, int y, int width, int height, Color bgcolor, ImageObserver observer) 
   { 
      ImageRepresentation   ir;


      ir = imageCheck( img, observer );

      if (ir != null) 
      {
         validateGeometry();
         return (ir.drawScaledImage( this, x, y, width, height, bgcolor, observer ));
      }
      else
      {
         return (false);
      }
   }


   /**
    * Draws a subrectangle of an image scaled to a destination rectangle
    * in nonblocking mode with a callback object.
    */
   public boolean drawImage(Image img, int dx1, int dy1, int dx2, int dy2, int sx1, int sy1, int sx2, int sy2, ImageObserver observer) 
   {
      ImageRepresentation   ir;


      if (dx1 == dx2 || dy1 == dy2)
      {
          return true;
      }
      
      ir = imageCheck( img, observer );

      if (ir != null) 
      {
         validateGeometry();
         return (ir.drawStretchImage( this, dx1, dy1, dx2, dy2, sx1, sy1, sx2, sy2, null, observer ));
      }
      else
      {
         return (false);
      }
   }


   /**
    * Draws a subrectangle of an image scaled to a destination rectangle in
    * nonblocking mode with a solid background color and a callback object.
    */
   public boolean drawImage(Image img, int dx1, int dy1, int dx2, int dy2, int sx1, int sy1, int sx2, int sy2, Color bgcolor, ImageObserver observer)
   { 
      ImageRepresentation   ir;


      if (dx1 == dx2 || dy1 == dy2)
      {
          return true;
      }
      
      ir = imageCheck( img, observer );

      if (ir != null) 
      {
         validateGeometry();
         return (ir.drawStretchImage( this, dx1, dy1, dx2, dy2, sx1, sy1, sx2, sy2, bgcolor, observer ));
      }
      else
      {
         return (false);
      }
   }


   /** 
    * Draws a line, using the current color, between the points 
    * <code>(x1,&nbsp;y1)</code> and <code>(x2,&nbsp;y2)</code> 
    * in this graphics context's coordinate system. 
    * @param   x1  the first point's <i>x</i> coordinate.
    * @param   y1  the first point's <i>y</i> coordinate.
    * @param   x2  the second point's <i>x</i> coordinate.
    * @param   y2  the second point's <i>y</i> coordinate.
    * @since   JDK1.0
    */
   public void drawLine(int x1, int y1, int x2, int y2)
   {
      synchronized (geometry.getLock()) 
      {
         validateGeometry();
         nDrawLine( x1, y1, x2, y2 );
      }
   }


   /** 
    * Draws the outline of an oval.
    * The result is a circle or ellipse that fits within the 
    * rectangle specified by the <code>x</code>, <code>y</code>, 
    * <code>width</code>, and <code>height</code> arguments. 
    * <p> 
    * The oval covers an area that is 
    * <code>width&nbsp;+&nbsp;1</code> pixels wide 
    * and <code>height&nbsp;+&nbsp;1<code> pixels tall. 
    * @param       x the <i>x</i> coordinate of the upper left 
    *                     corner of the oval to be drawn.
    * @param       y the <i>y</i> coordinate of the upper left 
    *                     corner of the oval to be drawn.
    * @param       width the width of the oval to be drawn.
    * @param       height the height of the oval to be drawn.
    * @see         java.awt.Graphics#fillOval
    * @since       JDK1.0
    */
   public void drawOval(int x, int y, int width, int height) 
   {
      synchronized (geometry.getLock()) 
      {
         validateGeometry();
         nDrawOval( x, y, width, height );
      }
   }


   /** 
    * Draws a sequence of connected lines defined by 
    * arrays of <i>x</i> and <i>y</i> coordinates. 
    * Each pair of (<i>x</i>,&nbsp;<i>y</i>) coordinates defines a point.
    * The figure is not closed if the first point 
    * differs from the last point.
    * @param       xPoints an array of <i>x</i> points
    * @param       yPoints an array of <i>y</i> points
    * @param       nPoints the total number of points
    * @see         java.awt.Graphics#drawPolygon(int[], int[], int)
    * @since       JDK1.1
    */
   public void drawPolyline(int xPoints[], int yPoints[], int nPoints)
   {
      synchronized (geometry.getLock()) 
      {
         validateGeometry();
         nDrawPolyline( xPoints, yPoints, nPoints );
      }
   }


   /** 
    * Draws a closed polygon defined by 
    * arrays of <i>x</i> and <i>y</i> coordinates. 
    * Each pair of (<i>x</i>,&nbsp;<i>y</i>) coordinates defines a point.
    * <p>
    * This method draws the polygon defined by <code>nPoint</code> line 
    * segments, where the first <code>nPoint&nbsp;-&nbsp;1</code> 
    * line segments are line segments from 
    * <code>(xPoints[i&nbsp;-&nbsp;1],&nbsp;yPoints[i&nbsp;-&nbsp;1])</code> 
    * to <code>(xPoints[i],&nbsp;yPoints[i])</code>, for 
    * 1&nbsp;&le;&nbsp;<i>i</i>&nbsp;&le;&nbsp;<code>nPoints</code>.  
    * The figure is automatically closed by drawing a line connecting
    * the final point to the first point, if those points are different.
    * @param        xPoints   a an array of <code>x</code> coordinates.
    * @param        yPoints   a an array of <code>y</code> coordinates.
    * @param        nPoints   a the total number of points.
    * @see          java.awt.Graphics#fillPolygon
    * @see          java.awt.Graphics#drawPolyline
    * @since        JDK1.0
    */
   public void drawPolygon(int xPoints[], int yPoints[], int nPoints)
   {
      synchronized (geometry.getLock()) 
      {
         validateGeometry();
         nDrawPolygon( xPoints, yPoints, nPoints );
      }
   }


   /** 
    * Draws the outline of the specified rectangle. 
    * The left and right edges of the rectangle are at 
    * <code>x</code> and <code>x&nbsp;+&nbsp;width</code>. 
    * The top and bottom edges are at 
    * <code>y</code> and <code>y&nbsp;+&nbsp;height</code>. 
    * The rectangle is drawn using the graphics context's current color.
    * @param         x   the <i>x</i> coordinate 
    *                         of the rectangle to be drawn.
    * @param         y   the <i>y</i> coordinate 
    *                         of the rectangle to be drawn.
    * @param         width   the width of the rectangle to be drawn.
    * @param         height   the height of the rectangle to be drawn.
    * @see          java.awt.Graphics#fillRect
    * @see          java.awt.Graphics#clearRect
    * @since        JDK1.0
    */
   public void drawRect(int x, int y, int width, int height) 
   {
      synchronized (geometry.getLock()) 
      {
         validateGeometry();
         nDrawRect( x, y, width, height );
      }
   }


   /** 
    * Draws an outlined round-cornered rectangle using this graphics 
    * context's current color. The left and right edges of the rectangle 
    * are at <code>x</code> and <code>x&nbsp;+&nbsp;width</code>, 
    * respectively. The top and bottom edges of the rectangle are at 
    * <code>y</code> and <code>y&nbsp;+&nbsp;height</code>. 
    * @param      x the <i>x</i> coordinate of the rectangle to be drawn.
    * @param      y the <i>y</i> coordinate of the rectangle to be drawn.
    * @param      width the width of the rectangle to be drawn.
    * @param      height the height of the rectangle to be drawn.
    * @param      arcWidth the horizontal diameter of the arc 
    *                    at the four corners.
    * @param      arcHeight the vertical diameter of the arc 
    *                    at the four corners.
    * @see        java.awt.Graphics#fillRoundRect
    * @since      JDK1.0
    */
   public void drawRoundRect(int x, int y, int width, int height, int arcWidth, int arcHeight) 
   {
      synchronized (geometry.getLock()) 
      {
         validateGeometry();
         nDrawRoundRect( x, y, width, height, arcWidth, arcHeight );
      }
   }


   /** 
    * Draws the text given by the specified string, using this 
    * graphics context's current font and color. The baseline of the 
    * first character is at position (<i>x</i>,&nbsp;<i>y</i>) in this 
    * graphics context's coordinate system. 
    * @param       str      the string to be drawn.
    * @param       x        the <i>x</i> coordinate.
    * @param       y        the <i>y</i> coordinate.
    * @see         java.awt.Graphics#drawBytes
    * @see         java.awt.Graphics#drawChars
    * @since       JDK1.0
    */
   public void drawString(String string, int x, int y) 
   {
      drawChars( string.toCharArray(), 0, string.length(), x, y );
   }


   /** 
    * Draws the text given by the specified character array, using this 
    * graphics context's current font and color. The baseline of the 
    * first character is at position (<i>x</i>,&nbsp;<i>y</i>) in this 
    * graphics context's coordinate system. 
    * @param data the array of characters to be drawn
    * @param offset the start offset in the data
    * @param length the number of characters to be drawn
    * @param x the <i>x</i> coordinate of the baseline of the text
    * @param y the <i>y</i> coordinate of the baseline of the text
    * @see         java.awt.Graphics#drawBytes
    * @see         java.awt.Graphics#drawString
    * @since       JDK1.0
    */
   public void drawChars(char chars[], int offset, int length, int x, int y) 
   {
      PlatformFont pf = ((PlatformFont)font.getPeer());


      synchronized (geometry.getLock()) 
      {
         validateGeometry();
         if (pf.mightHaveMultiFontMetrics()) 
         {
            drawMFChars( chars, offset, length, x, y, pf );
         }
         else
         {
            nDrawSFChars( chars, offset, length, x, y );
         }
      }
   }


   /**
    * Disposes of this graphics context and releases 
    * any system resources that it is using. 
    * A <code>Graphics</code> object cannot be used after 
    * <code>dispose</code>has been called.
    * <p>
    * When a Java program runs, a large number of <code>Graphics</code>
    * objects can be created within a short time frame.
    * Although the finalization process of the garbage collector 
    * also disposes of the same system resources, it is preferable 
    * to manually free the associated resources by calling this
    * method rather than to rely on a finalization process which 
    * may not run to completion for a long period of time.
    * <p>
    * Graphics objects which are provided as arguments to the 
    * <code>paint</code> and <code>update</code> methods 
    * of components are automatically released by the system when 
    * those methods return. For efficiency, programmers should
    * call <code>dispose</code> when finished using
    * a <code>Graphics</code> object only if it was created 
    * directly from a component or another <code>Graphics</code> object.
    * @see         java.awt.Graphics#finalize
    * @see         java.awt.Component#paint
    * @see         java.awt.Component#update
    * @see         java.awt.Component#getGraphics
    * @see         java.awt.Graphics#create
    * @since       JDK1.0
    */
   public synchronized void dispose()
   {
      nDispose();
   }
   

   /** 
    * Fills an oval bounded by the specified rectangle with the
    * current color.
    * @param       x the <i>x</i> coordinate of the upper left corner 
    *                     of the oval to be filled.
    * @param       y the <i>y</i> coordinate of the upper left corner 
    *                     of the oval to be filled.
    * @param       width the width of the oval to be filled.
    * @param       height the height of the oval to be filled.
    * @see         java.awt.Graphics#drawOval
    * @since       JDK1.0
    */
   public void fillArc(int x, int y, int width, int height, int startAngle, int arcAngle) 
   {
      synchronized (geometry.getLock()) 
      {
         validateGeometry();
         nFillArc( x, y, width, height, startAngle, arcAngle );
      }
   }


   /** 
    * Fills an oval bounded by the specified rectangle with the
    * current color.
    * @param       x the <i>x</i> coordinate of the upper left corner 
    *                     of the oval to be filled.
    * @param       y the <i>y</i> coordinate of the upper left corner 
    *                     of the oval to be filled.
    * @param       width the width of the oval to be filled.
    * @param       height the height of the oval to be filled.
    * @see         java.awt.Graphics#drawOval
    * @since       JDK1.0
    */
   public void fillOval(int x, int y, int width, int height)
   {
      synchronized (geometry.getLock()) 
      {
         validateGeometry();
         nFillOval( x, y, width, height );
      }
   }


   /** 
    * Fills a closed polygon defined by 
    * arrays of <i>x</i> and <i>y</i> coordinates. 
    * <p>
    * This method draws the polygon defined by <code>nPoint</code> line 
    * segments, where the first <code>nPoint&nbsp;-&nbsp;1</code> 
    * line segments are line segments from 
    * <code>(xPoints[i&nbsp;-&nbsp;1],&nbsp;yPoints[i&nbsp;-&nbsp;1])</code> 
    * to <code>(xPoints[i],&nbsp;yPoints[i])</code>, for 
    * 1&nbsp;&le;&nbsp;<i>i</i>&nbsp;&le;&nbsp;<code>nPoints</code>.  
    * The figure is automatically closed by drawing a line connecting
    * the final point to the first point, if those points are different.
    * <p>
    * The area inside the polygon is defined using an 
    * even-odd fill rule, also known as the alternating rule.
    * @param        xPoints   a an array of <code>x</code> coordinates.
    * @param        yPoints   a an array of <code>y</code> coordinates.
    * @param        nPoints   a the total number of points.
    * @see          java.awt.Graphics#drawPolygon(int[], int[], int)
    * @since        JDK1.0
    */
   public void fillPolygon(int xPoints[], int yPoints[], int nPoints) 
   {
      synchronized (geometry.getLock()) 
      {
         validateGeometry();
         nFillPolygon( xPoints, yPoints, nPoints );
      }
   }


   /** 
    * Fills the specified rectangle. 
    * The left and right edges of the rectangle are at 
    * <code>x</code> and <code>x&nbsp;+&nbsp;width&nbsp;-&nbsp;1</code>. 
    * The top and bottom edges are at 
    * <code>y</code> and <code>y&nbsp;+&nbsp;height&nbsp;-&nbsp;1</code>. 
    * The resulting rectangle covers an area 
    * <code>width</code> pixels wide by 
    * <code>height</code> pixels tall.
    * The rectangle is filled using the graphics context's current color. 
    * @param         x   the <i>x</i> coordinate 
    *                         of the rectangle to be filled.
    * @param         y   the <i>y</i> coordinate 
    *                         of the rectangle to be filled.
    * @param         width   the width of the rectangle to be filled.
    * @param         height   the height of the rectangle to be filled.
    * @see           java.awt.Graphics#fillRect
    * @see           java.awt.Graphics#clearRect
    * @since         JDK1.0
    */
   public void fillRect( int x, int y, int width, int height )
   {
      synchronized (geometry.getLock()) 
      {
         validateGeometry();
         nFillRect( x, y, width, height );
      }
   }


   /** 
    * Fills the specified rounded corner rectangle with the current color.
    * The left and right edges of the rectangle 
    * are at <code>x</code> and <code>x&nbsp;+&nbsp;width&nbsp;-&nbsp;1</code>, 
    * respectively. The top and bottom edges of the rectangle are at 
    * <code>y</code> and <code>y&nbsp;+&nbsp;height&nbsp;-&nbsp;1</code>. 
    * @param       x the <i>x</i> coordinate of the rectangle to be filled.
    * @param       y the <i>y</i> coordinate of the rectangle to be filled.
    * @param       width the width of the rectangle to be filled.
    * @param       height the height of the rectangle to be filled.
    * @param       arcWidth the horizontal diameter 
    *                     of the arc at the four corners.
    * @param       arcHeight the vertical diameter 
    *                     of the arc at the four corners.
    * @see         java.awt.Graphics#drawRoundRect
    * @since       JDK1.0
    */
   public void fillRoundRect(int x, int y, int width, int height, int arcWidth, int arcHeight) 
   {
      synchronized (geometry.getLock()) 
      {
         validateGeometry();
         nFillRoundRect( x, y, width, height, arcWidth, arcHeight );
      }
   }


   /**
    * Gets the current clipping area.
    * @return      a <code>Shape</code> object representing the 
    *                      current clipping area.
    * @see         java.awt.Graphics#getClipBounds
    * @see         java.awt.Graphics#clipRect
    * @see         java.awt.Graphics#setClip(int, int, int, int)
    * @see         java.awt.Graphics#setClip(Shape)
    * @since       JDK1.1
    */
   public Shape getClip()
   {
      return (getClipBounds()); 
   }


   /**
    * Returns the bounding rectangle of the current clipping area.
    * The coordinates in the rectangle are relative to the coordinate
    * system origin of this graphics context.
    * @return      the bounding rectangle of the current clipping area.
    * @see         java.awt.Graphics#getClip
    * @see         java.awt.Graphics#clipRect
    * @see         java.awt.Graphics#setClip(int, int, int, int)
    * @see         java.awt.Graphics#setClip(Shape)
    * @since       JDK1.1
    */
   public synchronized java.awt.Rectangle getClipBounds() 
   {
      Rectangle   rcClip;
      Point       ptOrigin;


      /*
       * Map clip bounds relative to origin
       */
      rcClip   = getClipRegion().getBounds();
      ptOrigin = getOrigin();
      rcClip.translate( -ptOrigin.x, -ptOrigin.y );

      return (rcClip);
   }


   /**
    * Gets this graphics context's current color.
    * @return    this graphics context's current color.
    * @see       java.awt.Color
    * @see       java.awt.Graphics#setColor
    * @since     JDK1.0
    */
   public Color getColor() 
   {
      return (foreground); 
   }


   /**
    * Gets the current font.
    * @return    this graphics context's current font.
    * @see       java.awt.Font
    * @see       java.awt.Graphics#setFont
    * @since     JDK1.0
    */
   public Font getFont() 
   {
      return (font); 
   }


   /**
    * Gets the font metrics of the current font.
    * @return    the font metrics of this graphics 
    *                    context's current font.
    * @see       java.awt.Graphics#getFont
    * @see       java.awt.FontMetrics
    * @see       java.awt.Graphics#getFontMetrics(Font)
    * @since     JDK1.0
    */
   public java.awt.FontMetrics getFontMetrics()
   {
      return (getFontMetrics( font ));
   }


   /**
    * Gets the font metrics for the specified font.
    * @return    the font metrics for the specified font.
    * @param     f the specified font
    * @see       java.awt.Graphics#getFont
    * @see       java.awt.FontMetrics
    * @see       java.awt.Graphics#getFontMetrics()
    * @since     JDK1.0
    */
   public FontMetrics getFontMetrics(Font f) 
   {
      if (f == null)
      {
         return (null);
      }

      return (graphicsSystem.getFontMetrics( f )); 
   }


   /**
    * Sets the current clip to the rectangle specified by the given
    * coordinates.
    * Rendering operations have no effect outside of the clipping area.
    * @param       x the <i>x</i> coordinate of the new clip rectangle.
    * @param       y the <i>y</i> coordinate of the new clip rectangle.
    * @param       width the width of the new clip rectangle.
    * @param       height the height of the new clip rectangle.
    * @see         java.awt.Graphics#clipRect
    * @see         java.awt.Graphics#setClip(Shape)
    * @since       JDK1.1
    */
   public void setClip(int x, int y, int width, int height)
   {
      rgnClipGraphics = graphicsSystem.makeRegion( x + ptTrans.x
                                                 , y + ptTrans.y
                                                 , width        
                                                 , height       
                                                 );             

      chgClipGraphics = true;
   }


   /**
    * Sets the current clipping area to an arbitrary clip shape.
    * Not all objects which implement the <code>Shape</code> 
    * interface can be used to set the clip.  The only 
    * <code>Shape</code> objects which are guaranteed to be 
    * supported are <code>Shape</code> objects which are
    * obtained via the <code>getClip</code> method and via 
    * <code>Rectangle</code> objects.
    * @see         java.awt.Graphics#getClip()
    * @see         java.awt.Graphics#clipRect
    * @see         java.awt.Graphics#setClip(int, int, int, int)
    * @since       JDK1.1
    */
   public void setClip(Shape clip) 
   {
      if (clip != null)
      {
         if (clip instanceof Region) 
         {
            rgnClipGraphics = ((Region)clip).copy();
         } 
         else 
         {
            Rectangle rcClip = clip.getBounds();

            rgnClipGraphics = graphicsSystem.makeRegion( rcClip.x
                                                       , rcClip.y
                                                       , rcClip.width
                                                       , rcClip.height
                                                       );
         }

         rgnClipGraphics.translate( ptTrans.x, ptTrans.y );
      }
      else
      {
         rgnClipGraphics = null;
      }

      chgClipGraphics = true;
   }
   

   /**
    * Sets this graphics context's current color to the specified 
    * color. All subsequent graphics operations using this graphics 
    * context use this specified color. 
    * @param     c   the new rendering color.
    * @see       java.awt.Color
    * @see       java.awt.Graphics#getColor
    * @since     JDK1.0
    */
   public void setColor(Color c) 
   {
      if (c == null)
      {
         c = Color.black;
      }

      foreground = c;
      nSetColor( c.getRGB() );
   }


   /**
    * Sets this graphics context's font to the specified font. 
    * All subsequent text operations using this graphics context 
    * use this font. 
    * @param  font   the font.
    * @see     java.awt.Graphics#getFont
    * @see     java.awt.Graphics#drawChars(java.lang.String, int, int)
    * @see     java.awt.Graphics#drawString(byte[], int, int, int, int)
    * @see     java.awt.Graphics#drawBytes(char[], int, int, int, int)
    * @since   JDK1.0
   */
   public void setFont(Font font) 
   {
      if ((font != this.font) && (font != null))
      {
         this.font = font;
         nSetFont( font );
      }
   }
   

   /**
    * Sets the paint mode of this graphics context to overwrite the 
    * destination with this graphics context's current color. 
    * This sets the logical pixel operation function to the paint or
    * overwrite mode.  All subsequent rendering operations will
    * overwrite the destination with the current color. 
    * @since   JDK1.0
    */
   public void setPaintMode() 
   {
      nSetPaintMode();
   }
   

   /**
    * Sets the paint mode of this graphics context to alternate between 
    * this graphics context's current color and the new specified color. 
    * This specifies that logical pixel operations are performed in the 
    * XOR mode, which alternates pixels between the current color and 
    * a specified XOR color. 
    * <p>
    * When drawing operations are performed, pixels which are the 
    * current color are changed to the specified color, and vice versa. 
    * <p>
    * Pixels that are of colors other than those two colors are changed 
    * in an unpredictable but reversible manner; if the same figure is 
    * drawn twice, then all pixels are restored to their original values. 
    * @param     c1 the XOR alternation color
    * @since     JDK1.0
    */
   public void setXORMode(Color xorColor)
   {
      nSetXORMode( xorColor.getRGB() );
   }
   

   /**
    * Translates the origin of the graphics context to the point
    * (<i>x</i>,&nbsp;<i>y</i>) in the current coordinate system. 
    * Modifies this graphics context so that its new origin corresponds 
    * to the point (<i>x</i>,&nbsp;<i>y</i>) in this graphics context's 
    * original coordinate system.  All coordinates used in subsequent 
    * rendering operations on this graphics context will be relative 
    * to this new origin.
    * @param  x   the <i>x</i> coordinate.
    * @param  y   the <i>y</i> coordinate.
    * @since   JDK1.0
    */
   public void translate(int x, int y)
   {
      ptTrans.translate( x, y );
      chgTrans = true;
   }


   /*
    * Validates that the geometry information held by the native object is
    * correct.  If native object is out of date, it is updated.
    */
   private void validateGeometry()
   {
      int           newGeometryID;
      Point         ptOrigin;
      Region        rgnClip;
      Enumeration   enumClipRects;
      Rectangle     rcClip;


      /*
       * We'll need to know if geometry has been modified
       */
      newGeometryID = geometry.getValidationID();

      /*
       * Could origin have changed?
       */
      if ((newGeometryID != lastGeometryID) || chgTrans)
      {
         /*
          * Update native origin
          */
         ptOrigin = getOrigin();
         nSetOrigin( ptOrigin.x, ptOrigin.y );
      }

      /*
       * Could clip region have changed?
       */
      if ((newGeometryID != lastGeometryID) || chgClipGraphics)
      {
         rgnClip = getClipRegion();

         /*
          * Valid clip region? 
          */
         if (rgnClip != null)
         {
            /*
             * Update native clip region
             */
            enumClipRects = rgnClip.getRectangles();
            rcClip        = (Rectangle)(enumClipRects.nextElement());

            nSetClipRect( rcClip.x, rcClip.y, rcClip.width, rcClip.height );
                        
            rcClip = (Rectangle)enumClipRects.nextElement();
  
            while (rcClip != null)
            {
               nUnionClipRect( rcClip.x, rcClip.y, rcClip.width, rcClip.height );
               rcClip = (Rectangle)enumClipRects.nextElement();
            }
         }
      }     


      /*
       * Geometry is now valid
       */
      lastGeometryID  = newGeometryID;
      chgClipGraphics = false;
      chgTrans        = false;
   }


   /**
    * Returns the true origin of this context, relative to the screen.
    * 
    * @return a Point object holding the origin.
    */
   private Point getOrigin()
   {
      Point       ptOrigin;
      Rectangle   rcBounds;


      /*
       * Offset origin from window bounds by current translation
       */
      rcBounds = geometry.getBounds();
      ptOrigin = new Point( rcBounds.x + ptTrans.x, rcBounds.y + ptTrans.y );

      return (ptOrigin);
   }


   /**
    * Returns the complete clip region of this context which is the composite 
    * of the GeometryProvider's clip region and the application specified 
    * clip region.
    * 
    * @return a Region object representing the composite region.
    */
   private Region getClipRegion()
   {
      Region        rgnClipComposite;
      Region        rgnClipGeometry;
      Rectangle     rcBounds;
      Enumeration   enumClipComposite;


      /* 
       * Need geometry info
       */
      rcBounds        = geometry.getBounds();
      rgnClipGeometry = (Region)(geometry.getClip());

      /*
       * Do we have a graphics clip region?
       */
      if (rgnClipGraphics != null)
      {
         /*
          * Graphics clip region is relative to origin
          */
         rgnClipComposite = rgnClipGraphics.copy();
         rgnClipComposite.translate( rcBounds.x, rcBounds.y );

         /*
          * Geometry clip region is not
          */
         rgnClipComposite.intersect( rgnClipGeometry );
      }
      else
      {
         /*
          * Just use geometry clip region 
          */
         rgnClipComposite = rgnClipGeometry;
      }
  
      return (rgnClipComposite);
   }

   /**
    * Validates and returns the ImageRepresentation object associated with an 
    * image.
    * 
    * @param img      the Image object to check.
    * @param observer the ImageObserver object to update on error.
    * 
    * @return an ImageRepresentation object encapsulating the native image.
    */
   private final ImageRepresentation imageCheck(Image img, ImageObserver observer) 
   {
      if (((sun.awt.image.Image)img).hasError())
      {
         if (observer != null) 
         {
            observer.imageUpdate( img
                                , ImageObserver.ERROR | ImageObserver.ABORT
                                , -1
                                , -1
                                , -1
                                , -1
                                );
         }

         return (null);
      }

      if (img instanceof ImageImpl) 
      {
          return  (((ImageImpl)img).getImageRep());
      } 
      else 
      {
          return  (((DrawableImageImpl)img).getImageRep());
      }
   }


   /**
    * Draws an array of characters which may be represented using multiple 
    * fonts.
    * 
    * @param chars  the array of characters to draw from.
    * @param offset offset of first character in array to draw.
    * @param length number of characters to draw.
    * @param x      the <i>x</i> coordinate of the baseline of the text
    * @param y      the <i>y</i> coordinate of the baseline of the text
    * @param pf     the PlatformFont object encapsulating the native font(s).
    */
   private void drawMFChars(char chars[], int offset, int length, int x, int y, PlatformFont pf) 
   {
      if (length == 0) 
      {
          return;
      }

      sun.awt.CharsetString[] css = pf.makeMultiCharsetString(chars, offset, length);

      byte[] csb = null;
      int baLen = 0;
      int w = 0;
      for (int i = 0; i < css.length; i++) 
      {
         CharsetString cs = css[i];

         if (FontMetricsImpl.needsConversion(font, cs.fontDescriptor)) 
         {
            try
            {
               int l = cs.fontDescriptor.fontCharset.getMaxBytesPerChar() 
                     * cs.length;
               if (l > baLen) 
               {
                  baLen = l;
                  csb = new byte[baLen];
               }

               int csbLen = 
               cs.fontDescriptor.fontCharset.convert( cs.charsetChars
                                                    , cs.offset
                                                    , cs.offset + cs.length
                                                    , csb
                                                    , 0
                                                    , csb.length
                                                    );

               w += nDrawMFCharsConvertedSegment( font
                                                , cs.fontDescriptor
                                                , csb
                                                , csbLen
                                                , x + w
                                                , y
                                                );
             }         
             catch (CharConversionException ignored) {}
         } 
         else
         {
            w += nDrawMFCharsSegment( font
                                    , cs.fontDescriptor
                                    , cs.charsetChars
                                    , cs.offset
                                    , cs.length
                                    , x + w
                                    , y
                                    );
         }          
      }
   }


   private native void nCopyArea( int x, int y, int width, int height, int dx, int dy );
   private native void nDispose();
   private native void nDrawArc( int x, int y, int width, int height, int startAngle, int arcAngle );
   private native void nDrawBytes( byte data[], int offset, int length, int x, int y );
   private native void nDrawLine( int x1, int y1, int x2, int y2 );
   private native int  nDrawMFCharsSegment( Font font, FontDescriptor des, char[] chars, int offset, int length,int x, int y );
   private native int  nDrawMFCharsConvertedSegment( Font font, FontDescriptor des, byte[] bytes, int length, int x, int y );
   private native void nDrawOval( int x, int y, int width, int height ); 
   private native void nDrawPolygon( int[] xPoints, int[] yPoints, int nPoints );
   private native void nDrawPolyline( int[] xPoints, int[] yPoints, int nPoints );
   private native void nDrawRect( int x, int y, int width, int height ); 
   private native void nDrawRoundRect( int x, int y, int width, int height, int arcWidth, int arcHeight ); 
   private native void nDrawSFChars( char[] chars, int offset, int length, int x, int y );
   private native void nFillArc( int x, int y, int width, int height, int startAngle, int arcAngle );
   private native void nFillOval( int x, int y, int width, int height ); 
   private native void nFillPolygon( int[] xPoints, int[] yPoints, int nPoints );
   private native void nFillRect( int x, int y, int width, int height ); 
   private native void nFillRoundRect( int x, int y, int width, int height, int arcWidth, int arcHeight ); 
   private native void nInit( Drawable d );
   private native void nSetClipRect( int x, int y, int width, int height );
   private native void nSetColor( int color );
   private native void nSetFont( Font font );
   private native void nSetOrigin( int x, int y );
   private native void nSetPaintMode();
   private native void nSetXORMode( int color );
   private native void nUnionClipRect( int x, int y, int width, int height );

   private native static void initJNI();
   static { initJNI(); }
} 
