/*
 * @(#)GraphicsSystem.java	1.11 99/07/29
 * 
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 * 
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */

package sun.awt.gfW;

import java.awt.Component;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.FontMetrics;
import java.awt.Font;
import java.awt.Image;
import java.awt.Point;
import java.awt.image.ColorModel;
import java.awt.image.ImageObserver;
import java.awt.image.ImageProducer;
import java.awt.peer.FontPeer;
import sun.porting.graphicssystem.CursorImage;
import sun.porting.graphicssystem.Drawable;
import sun.porting.graphicssystem.Region;
import sun.porting.windowsystem.EventHandler;
import sun.porting.utils.RegionImpl;
       

//
// Class: GraphicsSystem
//
public class GraphicsSystem extends sun.porting.graphicssystem.GraphicsSystem implements Runnable
{
   // Static Variables
   private static final int NONE     = EventHandler.NONE;
   private static final int PRESSED  = EventHandler.PRESSED;
   private static final int RELEASED = EventHandler.RELEASED;
   private static final int ACTION   = EventHandler.ACTION;
   private static final int TYPED    = EventHandler.TYPED;


   // Member Variables
   private int            screenWidth;
   private int            screenHeight;
   private EventHandler   eventHandler = null;
   private Drawable       defaultScreen = null;
   private ColorModel     colorModel;
   private int            cursorHidden = 0;
   private int            cursorX = 0;
   private int            cursorY = 0;


   // Static Initializers
   static 
   {
      System.loadLibrary( "pswsgs" );
      initJNI(); 
   }


   // Constructors
   public GraphicsSystem()
   {
      this( getDesiredWidth()
          , getDesiredHeight()
          , getDesiredDepth()
          , getPixelType().startsWith("color")
          );
   }

   public GraphicsSystem( int       width
                        , int       height
                        , int       depth
                        , boolean   useColors
                        )
   {
      Thread   eventThread;


      screenWidth  = width;
      screenHeight = height;

      /* 
       * This should never be referenced, because init
       * should reset colorModel to the correct value.
       */
      colorModel = java.awt.image.ColorModel.getRGBdefault();

      /*
       * Create event processing thread in synchronized block 
       * to allow us to wait for the thread to initialize 
       */
      synchronized (this)
      {
         /*
          * Create a new thread to process this object, which in turn will
          * start the event processor.
          */
         eventThread = new Thread( this );
         eventThread.start();
   
         /*
          * Wait for event processor to initialize 
          */
         try 
         {
            wait();
         }
         catch (InterruptedException ignored) {}
      }

      defaultScreen = new ScreenImpl( width, height, colorModel, this );
   }


   // Public Java Methods
   public void run()
   {
      eventLoop();
   }

   public void registerEventHandler( EventHandler h )
   {
      synchronized (this)    
      {
         if (eventHandler == null)
         {
            eventHandler = h;
         }
         else
         {
            return;
         }
      }
   }

   public FontPeer getFont( String name, int style )
   {
      return (new FontImpl( name, style ));
   }

   public FontMetrics getFontMetrics( Font key )
   {
      return (FontMetricsImpl.getFontMetrics( key ));
   }

   public java.awt.Image getImage( ImageProducer producer )
   {
      return (new ImageImpl( producer ));
   }

   public Image makeDrawableImage( Component c, int w, int h ) 
   {
      return (new DrawableImageImpl( c, w, h, this ));
   }

   public Drawable getScreen()
   {
      return (defaultScreen);
   }

   public Drawable[] getScreens() 
   {
      Drawable   d[] = new Drawable[1];


      d[0] = defaultScreen;

      return (d);
   }

   public Region makeRegion( int x, int y, int w, int h )
   {                               
      return (new RegionImpl( x, y, w, h ));
   }

   public void setCursorVisibility( boolean visible )
   {
      synchronized (this)
      {
         // Showing or hiding?
         if (visible)
         {  
            // Not already visible?
            if (cursorHidden > 0)
            {
               // One step closer to visibility
               cursorHidden--;

               // Just made visible?
               if (cursorHidden == 0)
               {
                  showCursor( true );
               }
            }
         }
         else
         {
            // One step further from visibility
            cursorHidden++;

            // Just hidden?
            if (cursorHidden == 1)
            {
               showCursor( false );
            }
         }
      }
   }

   public boolean isCursorVisible()
   {
      synchronized (this)
      {
         return (cursorHidden == 0);
      }
   }

   public Point getCursorLocation() 
   {
      synchronized (this)
      {
         Point   pt = new Point();

         getCursorPos( pt );

         return (pt);
      }
   }

   public CursorImage makeCursorImage( java.awt.Image img, int hotX, int hotY )
   {
       return (new CursorImageImpl( img, hotX, hotY ));
   }

   public CursorImage getCursorImage(Cursor c)
   {
      return (new CursorImageImpl( c ));
   }

   public void setCursorImage( CursorImage image ) 
   {
      synchronized (this) 
      {
         setCursor( image );
      }
   }

   public Dimension getMaximumCursorSize() 
   {
      return (getBestCursorSize( screenWidth, screenHeight ));
   }

   public Dimension getBestCursorSize( int width, int height )
   {
      Dimension d = new Dimension();

      getCursorSize(d);

      return (d);
   }

   public int getMaximumCursorColors()
   {
      return (2);
   }

   public boolean prepareScrImage( Image img, int w, int h, ImageObserver o )
   {
      ImageImpl ximg = (ImageImpl)img;


      if (w == 0 || h == 0) 
      {
         return true;
      }

      if (ximg.hasError()) 
      {
         if (o != null) 
         {
            o.imageUpdate( img
                         , java.awt.image.ImageObserver.ERROR
                         | java.awt.image.ImageObserver.ABORT
                         , -1
                         , -1
                         , -1
                         , -1
                         );
         }

         return false;
      }

      sun.awt.image.ImageRepresentation ir = ximg.getImageRep();
      return ir.prepare(o);
   }

   public int checkScrImage( Image img, int w, int h, ImageObserver o )
   {
      ImageImpl ximg = (ImageImpl)img;
      int       repbits;


      if (w == 0 || h == 0) 
      {
         repbits = java.awt.image.ImageObserver.ALLBITS;
      }
      else
      {
         repbits = ximg.getImageRep().check( o );
      }
      
      return (ximg.check( o ) | repbits);
   }

   // Protected Java Methods
   protected void sendGraphicsChange( Region rgnInvalid ) 
   {
      if ((eventHandler != null) && (!rgnInvalid.isEmpty()))
      {
         eventHandler.graphicsChange( System.currentTimeMillis()
                                    , rgnInvalid
                                    , EventHandler.GRAPHICS_INVALID_FLAG
                                    );
      }
   }

   // Private Java Methods
   private static int getDesiredDepth()
   {
      try 
      {
         String s = getPixelType();
         int    n = Integer.parseInt(s.substring(s.indexOf(':') + 1));
         return n;
      }
      catch (NumberFormatException e) 
      {
         return 8;
      }
   }

   private static String getPixelType() 
   {
      return System.getProperty("sun.awt.gfW.pixelType", "color:8");
   }

   private static int getDesiredWidth() 
   {
      int w = 640;


      try 
      {
         String s = System.getProperty("sun.graphicssystem.width", "640");
         int    n = Integer.parseInt(s);

         if ((n > 0) && (n < 65536)) 
         {
    	    w = n;
         }
      } catch (Throwable ignored) {}

      return w;
   }

   private static int getDesiredHeight() 
   {
      int h = 480;


      try
      {
          String s = System.getProperty("sun.graphicssystem.height", "480");
          int    n = Integer.parseInt(s);

          if ((n > 0) && (n < 65536)) 
          {
             h = n;
          }
      } catch (Throwable ignored) {}

      return h;
   }

   private void handleExposeEvent( int x, int y, int w, int h )
   {
      sendGraphicsChange( makeRegion( x, y, w, h ) );
   }

   private void handleMouseInput( int id, int dx, int dy, int button )
   {
      if (eventHandler != null)
      {
         eventHandler.pointerEventOccurred( System.currentTimeMillis()
                                          , dx
                                          , dy
                                          , id
                                          , button
                                          );
      }
   }

   private void handleKeyInput( int id, int keycode, int keychar )
   {
      if (eventHandler != null)
      {
         eventHandler.keyboardEventOccurred( System.currentTimeMillis()
                                           , id
                                           , keycode
                                           , (char)keychar
                                           );
      }
   }

   // Public Native Methods
   public native int  getScreenResolution();
   public native void sync();
   public native void beep();

   // Private Native Methods
   private native void eventLoop();
   private native void getCursorSize( Dimension d );
   private native void setCursor( CursorImage image );
   private native void showCursor( boolean show );
   private native void getCursorPos( Point pt );
   private native static void initJNI();
}
