/*
 * @(#)condvar_md.c	1.18 98/09/02
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Win32 implementation of condition variables
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <windows.h>
#include <mmsystem.h>

#include "condvar_md.h"

/*
 * Create and initialize condition variable.
 */
bool_t
condvarInit(condvar_t *condvar)
{
    HANDLE e = CreateEvent(NULL, TRUE, FALSE, NULL);
    if (e == 0) {
        return FALSE;
    }
    condvar->event = e;
    return TRUE;
}

/*
 * Delete the condition variable by closing the event object handle.
 */
void
condvarDestroy(condvar_t *condvar)
{
    CloseHandle(condvar->event);
}

/*
 * Block on a condition variable until signaled.
 *
 * REMIND: We could see which event the wait wakes up on and, if it
 * is the interrupt event, we could pass back a return value that
 * would serve instead of the (possibly more expensive) interrupted
 * check.
 */

#define CONDVAR_EVENT   0
#define INTERRUPT_EVENT 1

void
condvarWait(sys_thread_t *tid, condvar_t *condvar, mutex_t *mutex)
{
    unsigned int c = condvar->counter;
    HANDLE events[2];
    int retry = 0;

    events[CONDVAR_EVENT] = condvar->event;
    events[INTERRUPT_EVENT] = tid->interrupt_event;
    condvar->waiters++;
    /*
     * Loop until condition variable is signaled. The event object is
     * set whenever the condition variable is signaled, and tickets will
     * reflect the number of threads which have been notified. The counter
     * field is used to make sure we don't respond to notifications that
     * have occurred *before* we started waiting, and is incremented each
     * time the condition variable is signaled.
     */
    while (1) {
        int which;
	LeaveCriticalSection(mutex);
	/* If this is a retry, let other low-priority threads have a 
	 * chance to run. 
	 * Make sure that we sleep outside of the critical section.
	 */
	if (retry)
	    Sleep(1);
	else
	    retry = 1;

	which = WaitForMultipleObjects(2, events, FALSE, INFINITE);
	EnterCriticalSection(mutex);
	if (which == INTERRUPT_EVENT) {
	    /* The interrupted flag is set by sysThreadInterrupt */
	    --condvar->waiters;
	    return;
	}
	if (condvar->tickets != 0 && condvar->counter != c)
	    break;
    }
    --condvar->waiters;
    /*
     * If this was the last thread to be notified, then we need to reset
     * the event object.
     */
    if (--condvar->tickets == 0) {
	ResetEvent(condvar->event);
    }
}

/*
 * Block on a condition variable until signaled or timeout has expired.
 */
void
condvarTimedWait(sys_thread_t *tid, condvar_t *condvar, mutex_t *mutex,
		 unsigned long millis)
{
    unsigned int c = condvar->counter;
    unsigned long t0, t1, elapsed;
    HANDLE events[2]; 
    int retry = 0;

    events[CONDVAR_EVENT] = condvar->event; 
    events[INTERRUPT_EVENT] = tid->interrupt_event; 
    condvar->waiters++;
    t0 = GetTickCount();
    while (TRUE) {
	int which;
	LeaveCriticalSection(mutex);
	/* Let other lower-priority threads have a chance to run. */
	if (retry)
	    Sleep(1);
	else
	    retry = 1;

	which = WaitForMultipleObjects(2, events, FALSE, millis);
	EnterCriticalSection(mutex);
	if (which == INTERRUPT_EVENT) {
	    /* The interrupted flag is set by sysThreadInterrupt */
	    --condvar->waiters;
	    return;
	}
	if (condvar->tickets > 0 && condvar->counter != c) {
	    --condvar->waiters;
	    if (--condvar->tickets == 0) {
		ResetEvent(condvar->event);
	    }
	    return;
	}
	t1 = GetTickCount();
	/* Need to check for wraparound of elapsed time */
	elapsed = t1 >= t0 ? t1 - t0 : ULONG_MAX - t1 + t0;
	if (elapsed >= millis) { /* timed out */
	    --condvar->waiters;
	    return;
	}
	millis -= elapsed;
	t0 = t1;
    }
}

/*
 * Notify single thread waiting on a condition variable.
 */
bool_t
condvarSignal(condvar_t *condvar)
{
    if (condvar->waiters > condvar->tickets) {
	if (!SetEvent(condvar->event)) {
	    return FALSE;
	}
	condvar->tickets++;
	condvar->counter++;
    }
    return TRUE;
}

/*
 * Notify all threads waiting on a condition variable.
 */
bool_t
condvarBroadcast(condvar_t *condvar)
{
    if (condvar->waiters > 0) {
	if (!SetEvent(condvar->event)) {
	    return FALSE;
	}
	condvar->tickets = condvar->waiters;
	condvar->counter++;
    }
    return TRUE;
}
