/*
 * @(#)dirent.c	1.9 98/09/02
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Posix-compatible directory access routines
 */

#include <windows.h>
#include <errno.h>

#include "sys_api.h"
#include "dirent.h"


/* Caller must have already run dirname through sysNativePath, which removes
   duplicate slashes and converts all instances of '/' into '\\'. */

DIR *
opendir(const char *dirname)
{
    DIR* dirp;
    DWORD fattr;
    int dirnamelen = strlen(dirname);

    /* Must fail if dirname == "" */
    if (dirnamelen == 0) {
	errno = ENOENT;
	return NULL;
    }
    if ((dirp = (DIR*)sysMalloc(sizeof(DIR))) == NULL) {
	errno = ENOMEM;
	return NULL;
    }
    /*
     * GetFileAttributes() works fine for "\\".
     * FindFirstFile() does not work for "\\", but we're going to append 
     * "*.*" before using it, which makes FindFirstFile() working happily.
     */
    dirp->path = (char *)sysMalloc(dirnamelen + 5 /* for "\\*.*\0" */);
    if (dirp->path == NULL) {
	sysFree(dirp);
	errno = ENOMEM;
	return NULL;
    }
    strcpy(dirp->path, dirname);

    fattr = GetFileAttributes(dirp->path);
    if (fattr == 0xffffffff) { /* error */
	sysFree(dirp->path);
	sysFree(dirp);
	errno = ENOENT;
	return NULL;
    } else if (fattr & FILE_ATTRIBUTE_DIRECTORY == 0) {
	sysFree(dirp->path);
	sysFree(dirp);
	errno = ENOTDIR;
	return NULL;
    }

    /* Append "*.*", or possibly "\\*.*", to path */
    sysAssert(dirnamelen >= 1);
#ifndef _WIN32_WCE
    /* No '\\' needed for cases like "X:" or "XXX\\" */
    if ((dirnamelen == 2 && dirp->path[1] == ':')
	|| dirp->path[dirnamelen - 1] == '\\')
#else
    /* No drive letter for WinCE */
    if (dirp->path[dirnamelen - 1] == '\\')
#endif
    {
	strcat(dirp->path, "*.*");
    } else {
	strcat(dirp->path, "\\*.*");
    }
    dirp->handle = FindFirstFile(dirp->path, &dirp->find_data);
    if (dirp->handle == INVALID_HANDLE_VALUE) {
        if (GetLastError() != ERROR_FILE_NOT_FOUND) {
	    sysFree(dirp->path);
	    sysFree(dirp);
	    errno = EACCES;
	    return NULL;
	}
    }
    return dirp;
}

struct dirent *
readdir(DIR *dirp)
{
    if (dirp->handle == INVALID_HANDLE_VALUE) {
	return 0;
    }

    strcpy(dirp->dirent.d_name, dirp->find_data.cFileName);

    if (!FindNextFile(dirp->handle, &dirp->find_data)) {
	if (GetLastError() == ERROR_INVALID_HANDLE) {
	    errno = EBADF;
	    return 0;
	}
	FindClose(dirp->handle);
	dirp->handle = INVALID_HANDLE_VALUE;
    }

    return &dirp->dirent;
}

int
closedir(DIR *dirp)
{
    if (dirp->handle != INVALID_HANDLE_VALUE) {
	if (!FindClose(dirp->handle)) {
	    errno = EBADF;
	    return -1;
	}
	dirp->handle = INVALID_HANDLE_VALUE;
    }
    sysFree(dirp->path);
    sysFree(dirp);
    return 0;
}

void
rewinddir(DIR *dirp)
{
    if (dirp->handle != INVALID_HANDLE_VALUE) {
	FindClose(dirp->handle);
    }
    dirp->handle = FindFirstFile(dirp->path, &dirp->find_data);
}
