/*
 * @(#)io_md.c	1.22 98/09/02
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include <io.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifdef BREAKPTS
#include "oobj.h"
#include "interpreter.h"
#include "monitor.h"
#endif

#include "path_md.h"
#include "io_md.h"
#include "sys_api.h"
#include "dirent.h"

#define IO_EXCEPTION "java/io/IOException"

int
sysOpen(const char *path, int oflag, int mode)
{
    char pathbuf[MAX_PATH];
    int rc;

    if (strlen(path) > MAX_PATH - 1) {
	errno = ENAMETOOLONG;
	return -1;
    }

#ifdef BREAKPTS
    /*
     * Under Win32, the debugger often deadlocks when a thread is suspended in an open()
     * call. This code prevents such deadlocks.
     */
    if (debugging) {
        BINCLASS_LOCK();
    }
#endif
    rc = open(sysNativePath(strcpy(pathbuf, path)), oflag | O_BINARY, mode);

#ifdef BREAKPTS
    if (debugging) {
        BINCLASS_UNLOCK();
    }
#endif
    return rc;
}


int
sysAccess(const char *path, int mode)
{
    char pathbuf[MAX_PATH];

    if (strlen(path) > MAX_PATH - 1) {
	errno = ENAMETOOLONG;
	return -1;
    }
    return access(sysNativePath(strcpy(pathbuf, path)), mode);
}

int
sysStat(const char *path, struct stat *sbuf)
{
    char pathbuf[MAX_PATH];

    if (strlen(path) > MAX_PATH - 1) {
	errno = ENAMETOOLONG;
	return -1;
    }
    return stat(sysNativePath(strcpy(pathbuf, path)), sbuf);
}

int
sysUnlink(const char *path)
{
    char pathbuf[MAX_PATH];

    if (strlen(path) > MAX_PATH - 1) {
	errno = ENAMETOOLONG;
	return -1;
    }
    sysNativePath(strcpy(pathbuf, path));
    SetFileAttributes(pathbuf, 0);
    if (DeleteFile(pathbuf) == FALSE) {
	return -1;
    }
    return 0;
}

int
sysMkdir(const char *path, int mode)
{
    char pathbuf[MAX_PATH];

    if (strlen(path) > MAX_PATH - 1) {
	errno = ENAMETOOLONG;
	return -1;
    }
    return mkdir(sysNativePath(strcpy(pathbuf, path)));
}

int
sysRmdir(const char* path)
{
    char pathbuf[MAX_PATH];

    if (strlen(path) > MAX_PATH - 1) {
	errno = ENAMETOOLONG;
	return -1;
    }
    return _rmdir(sysNativePath(strcpy(pathbuf, path)));
}

int
sysRename(const char* srcName, const char* dstName)
{
    char from[MAX_PATH];
    char to[MAX_PATH];

    if (strlen(srcName) > MAX_PATH - 1 || strlen(dstName) > MAX_PATH - 1) {
	errno = ENAMETOOLONG;
	return -1;
    }
    // This would be on one line but it seems to get
    // compiled incorrectly that way
    strcpy(from, srcName);
    strcpy(to, dstName);
    sysNativePath(from);
    sysNativePath(to);
    return rename(from, to);
}

DIR *
sysOpenDir(const char *path)
{
    char pathbuf[MAX_PATH];

    if (strlen(path) > MAX_PATH - 1) {
       errno = ENAMETOOLONG;
       return 0;
    }
    return (DIR*) opendir(sysNativePath(strcpy(pathbuf, path)));
}


static int nonSeekAvailable(int, long *);
static int stdinAvailable(int, long *);

int
sysAvailable(int fd, long *pbytes)
{
    long cur, end;
    struct stat stbuf;

    if (fstat(fd, &stbuf) >= 0) {
	int mode = stbuf.st_mode;
#ifndef _WIN32_WCE
	/* Neither pipe nor console I/O supported by WinCE */
	if (S_ISCHR(mode) || S_ISFIFO(mode)) {
	    return (fd == 0) ? stdinAvailable(fd, pbytes) :
		nonSeekAvailable(fd, pbytes);
	}
#endif /* !_WIN32_WCE */
	if ((cur = lseek(fd, 0L, SEEK_CUR)) == -1) {
	    return FALSE;
	} else if ((end = lseek(fd, 0L, SEEK_END)) == -1) {
	    return FALSE;
	} else if (lseek(fd, cur, SEEK_SET) == -1) {
	    return FALSE;
	}
	*pbytes = end - cur;
	return TRUE;
    } else {
	return FALSE;
    }
}

#ifndef _WIN32_WCE

static int nonSeekAvailable(int fd, long *pbytes) { 
    /* This is used for available on non-seekable devices 
     * (like both named and anonymous pipes, such as pipes
     *  connected to an exec'd process).
     * Standard Input is a special case.
     *
     */
    HANDLE han;

    if ((han = (HANDLE) _get_osfhandle(fd)) == (HANDLE)(-1)) {
	return FALSE;
    }

    if (! PeekNamedPipe(han, NULL, 0, NULL, pbytes, NULL)) {
	/* PeekNamedPipe fails when at EOF.  In that case we
	 * simply make *pbytes = 0 which is consistent with the
	 * behavior we get on Solaris when an fd is at EOF.
	 * The only alternative is to raise and Exception, 
	 * which isn't really warranted.
	 */
	if (GetLastError() != ERROR_BROKEN_PIPE) {
	    return FALSE;
	}
	*pbytes = 0;
    }
    return TRUE;
}

static int stdinAvailable(int fd, long *pbytes) {
    /* Standard input is a special case on win32.
     * What's implemented here is not strictly correct:
     * Stdin is line-buffered, yet WaitForSingleObject()
     * will notify us that something is available on
     * standard in when e.g., a user types something but
     * hasn't entered a carriage return.  So 
     * System.in.available() returns non-zero even though
     * System.in.read() will block.
     * We also can't tell how much is available - we have
     * to just return 1 or 0.
     *
     * PeekConsoleInput() looked somewhat promising, but
     * had serious problems (might be worth investigating
     * in the future)
     * 
     * If anyone has a better idea have at it!
     */

    HANDLE han;

    if ((han = GetStdHandle(STD_INPUT_HANDLE)) == 
	INVALID_HANDLE_VALUE) {
	return FALSE;
    }

    switch (WaitForSingleObject(han, 0)) {
    case WAIT_ABANDONED: /* this is an unsignaled state */
    case WAIT_TIMEOUT:
	*pbytes = 0;
	return TRUE;
    case WAIT_OBJECT_0:
	*pbytes = 1;
	return TRUE;
    default:
	return FALSE;
    }
}
     
#endif /* !_WIN32_WCE */
	
