/*
 * @(#)linker_md.c	1.37 98/11/13
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Maintains a list of currently loaded DLLs (Dynamic Link Libraries)
 * and their associated handles. Library names are case-insensitive.
 */

#include <windows.h>
#include <stdio.h>
#include <string.h>

#include "sys_api.h"
#include "path_md.h"
#include "tree.h"
#include "oobj.h"

typedef struct {
    const char *fn;
    HANDLE h;
} dllentry_t;

static dllentry_t * dllarray;
static int	    dllarraysize;
static int	    dllcount;
static char *	    libdir;

#define BUILTINLIB "-*-BUILTIN-*-"

/*
 * Check whether the dynamic library is built-in
 */
static bool_t isBuiltin(char *fname)
{
    static const char * const builtins[] = {
#ifdef STATICLIBRARIES
	"awt", "net", "zip", "agent", "math", "pswsgs", "hprof", "javafile",
#endif
	"winawt", "mmedia", "dst", "jpeg", "text", "sysresource"
    };
#define NO_BUILTINS sizeof(builtins) / sizeof(char *)
    int i;
    for (i = 0; i < NO_BUILTINS; i++) {
	if (strcmp(fname, builtins[i]) == 0) {
	    return TRUE;
	}
    }
    return FALSE;
}

/*
 * create a string for the JNI native function name by adding the
 * appropriate decorations.
 *
 * On Win32, "__stdcall" functions are exported differently, depending
 * on the compiler. In MSVC 4.0, they are decorated with a "_" in the 
 * beginning, and @nnn in the end, where nnn is the number of bytes in 
 * the arguments (in decimal). Borland C++ exports undecorated names.
 *
 * sysBuildFunName handles different encodings depending on the value 
 * of encodingIndex. It returns 0 when handed an out-of-range
 * encodingIndex.
 */
int
sysBuildFunName(char *name, int nameMax, void *mb_, int encodingIndex)
{
  if (encodingIndex == 0) {
    /* For Microsoft MSVC 4.0 */
    char suffix[6];    /* This is enough since Java never has more than 
			   256 words of arguments. */
    int real_args_size;
    int nameLen;
    int i;

    struct methodblock *mb = (struct methodblock*)mb_;
    if (methodaccess(mb) & ACC_STATIC)
        real_args_size = methodargssize(mb) * 4 + 8;
    else
        real_args_size = methodargssize(mb) * 4 + 4;
    jio_snprintf(suffix, sizeof(suffix), "@%d", real_args_size);
    
    nameLen = strlen(name);
    if (nameLen >= nameMax - 7)
        return 1;
    for(i = nameLen; i > 0; i--)
        name[i] = name[i-1];
    name[0] = '_';
    
    jio_snprintf(name + nameLen + 1, 6, "%s", suffix);
    return 1;
  } else if (encodingIndex == 1)
    /* For Borland, etc. */
    return 1;
  else
    return 0;
}

/*
 * Build a machine dependent library name out of a path and file name.
 */
void
sysBuildLibName(char *holder, int holderlen, char *pname, char *fname)
{
    const int pnamelen = strlen(pname);

    if (isBuiltin(fname)) {
	(void) jio_snprintf(holder, holderlen, "%s", BUILTINLIB);
	return;
    }

    /* Quietly truncates on buffer overflow. Should be an error. */
    if (pnamelen == 0) {
	jio_snprintf(holder, holderlen, "%s.dll", fname);
    } else {
	const char c = pname[pnamelen-1];
	if (c == ':' || c == '\\') {
	    jio_snprintf(holder, holderlen, "%s%s.dll", pname, fname);
	} else {
	    jio_snprintf(holder, holderlen, "%s\\%s.dll", pname, fname);
	}
    }
}

/*
 * Initialize the linker but return an empty DLL search path so that we
 * can handle the looking up of native library DLLs here instead of in
 * Java.
 */
char *
sysInitializeLinker(void)
{
    extern HINSTANCE hJavaInst;
    char buf[MAX_PATH+1];
    int ret;

#ifndef STATICLIBRARIES
    if (libdir == 0) {
	/* Get location of <java-home>\bin directory */
	ret = GetModuleFileName(hJavaInst, buf, sizeof(buf));
	sysAssert(ret > 0);
	*strrchr(buf, '\\') = '\0';
	libdir = strdup(buf);
	if (libdir == 0) {
	    out_of_memory();
	}
	/* Make sure that runtime DLL is included first */
	ret = sysAddDLSegment("javai.dll");
	if (ret == -1) {
	    out_of_memory();
	} else if (ret == 0) {
	    jio_fprintf(stderr, "*** Linker initialization failed ***\n");
	    sysExit(1);
	}
    }
#endif
    return "";
}

/* 
 * Add a shared library to the list. DLL names are case-insensitive.
 * Return 1 if things went OK, 0 if you couldn't read the shared library,
 * and -1 if you ran out of memory.
 */
int
sysAddDLSegment(char *fname)
{
    HANDLE h;
    int i;
    char buf[MAX_PATH+1];

    if (libdir == 0) {
	sysInitializeLinker();
    }

    if (strcmp(fname, BUILTINLIB) == 0) {
	return 1;
    }

#ifdef DEBUG
    {
	/*
	 * Before change, assumes a ".dll" is already present.
	 * Therefore we can assume a dot will always be present
	 * in anything passed by Java itself. The suffix length can
	 * vary, though.
	 */
	char *dot = strrchr(fname,'.');
	if ( dot )
	{
		char tmp[MAX_PATH+1];
		strcpy(tmp, fname);
		strcpy(tmp + (dot-fname), "_g.");
		strcpy(tmp + (dot-fname) + 3, dot + 1 );
		// strcpy(tmp + strlen(tmp) - 4, "_g.dll");
		fname = strdup(tmp);
	}
	else
		fname = strdup(fname);
    }
#else
    fname = strdup(fname);
#endif

    if (fname == 0) {
	return -1;			/* Out of memory! */
    }

    for (i = 0; i < dllcount; i++) {
        if (stricmp(fname, dllarray[i].fn) == 0) {
	    return 1;    // dll is already loaded
	}
    }

    /* First, try the <java-home>\bin directory */
    jio_snprintf(buf, sizeof(buf), "%s\\%s", libdir, fname);
    h = LoadLibrary(buf);

    /* If not found there, then try the standard DLL search path */
    if (h == 0) {
	h = LoadLibrary(fname);
	if (h == 0) {
	    return 0;
	}
    }

    if (dllcount >= dllarraysize) {
	dllarraysize += 16;
	dllarray = realloc(dllarray, sizeof(dllentry_t) * dllarraysize);
	if (dllarray == 0) {
	    return -1;			/* Out of memory! */
	}
    }

    dllarray[dllcount].fn = fname;
    dllarray[dllcount++].h = h;

    return 1;
}

/*
 * Look up symbol name in currently loaded DLLs and return it's handle.
 */
long
sysDynamicLink(char *sym)
{
    int i;

    if (libdir == 0) {
	sysInitializeLinker();
    }

    for (i = 0; i < dllcount; i++) {
	FARPROC *f = (FARPROC *)GetProcAddress(dllarray[i].h, sym);
	if (f != NULL) {
	    return (long)f;
	}
    }

    return 0;
}
