/*
 * @(#)monitor_md.c	1.35 98/09/02
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Win32 implementation of Java monitors
 */

#include <windows.h>

#include "sys_api.h"
#include "threads.h"
#include "threads_md.h"
#include "monitor.h"
#include "monitor_md.h"

/*
 * The high-speed lock controlling cache access
 */
cache_lock_t _moncache_lock;

/*
 * Create and initialize monitor. This can be called before threads have
 * been initialized.
 */
int
sysMonitorInit(sys_mon_t *mid)
{
    sysAssert(mid != SYS_MID_NULL);
    mid->monitor_owner = SYS_THREAD_NULL;
    mid->entry_count = 0;
    mutexInit(&mid->mutex);
    if (!condvarInit(&mid->condvar)) {
	mutexDestroy(&mid->mutex);
	return SYS_ERR;
    }
    return SYS_OK;
}

/*
 * Free any system-dependent resources held by monitors.  On Win32 this
 * means releasing the critical section (mutex) and condition variable
 * that are part of each monitor.
 */
int
sysMonitorDestroy(sys_mon_t *mid)
{
    sysAssert(mid != SYS_MID_NULL);

    mutexDestroy(&mid->mutex);
    condvarDestroy(&mid->condvar);

    return SYS_OK;
}

/*
 * Take ownership of monitor. This can be called before threads have
 * been initialized, in which case we do nothing since locked monitors
 * are not yet needed.
 */
int
sysMonitorEnter(sys_mon_t *mid)
{
    sys_thread_t *self = sysThreadSelf();

    sysAssert(mid != SYS_MID_NULL);
    if (self == 0) {
	return ThreadsInitialized ? SYS_ERR : SYS_OK;
    }
    sysAssert(ThreadsInitialized);

    if (mid->monitor_owner == self) {
	/* Fast path when this thread already owns the monitor. */
	sysAssert(mid->entry_count > 0);
    } else {
 	self->state = MONITOR_WAIT;
	mutexLock(&mid->mutex);
	self->state = RUNNABLE;
	sysAssert(mid->monitor_owner == SYS_THREAD_NULL);
	mid->monitor_owner = self;
    }
    mid->entry_count++;
    return SYS_OK;
}

/*
 * Return TRUE if this thread currently owns the monitor. This can be
 * called before threads have been initialized, in which case we always
 * return TRUE.
 */
bool_t
sysMonitorEntered(sys_mon_t *mid)
{
    sys_thread_t *self = sysThreadSelf();
    sysAssert(mid != SYS_MID_NULL);
    return !ThreadsInitialized || self != 0 && mid->monitor_owner == self;
}

/*
 * Release ownership of monitor. This can be called before threads have
 * been initialized, in which case we do nothing as locked monitors are
 * not yet needed.
 */
int
sysMonitorExit(sys_mon_t *mid)
{
    sys_thread_t *self = sysThreadSelf();

    sysAssert(mid != SYS_MID_NULL);
    if (self == 0) {
	return ThreadsInitialized ? SYS_ERR : SYS_OK;
    }
    sysAssert(ThreadsInitialized);
    if (mid->monitor_owner != self) {
        return SYS_ERR;
    }
    sysAssert(mid->entry_count > 0);
    if (--mid->entry_count == 0) {
        mid->monitor_owner = 0;
	mutexUnlock(&mid->mutex);
    }
    return SYS_OK;
}

/*
 * Notify single thread waiting on condition variable.
 */
int
sysMonitorNotify(sys_mon_t *mid)
{
    sysAssert(mid != SYS_MID_NULL);
    if (!sysMonitorEntered(mid)) {
	return SYS_ERR;
    }
    return condvarSignal(&mid->condvar) ? SYS_OK : SYS_ERR;
}

/*
 * Notify all threads waiting on condition variable.
 */
int
sysMonitorNotifyAll(sys_mon_t *mid)
{
    sysAssert(mid != SYS_MID_NULL);
    if (!sysMonitorEntered(mid)) {
	return SYS_ERR;
    }
    return condvarBroadcast(&mid->condvar) ? SYS_OK : SYS_ERR;
}

/*
 * Atomically drop mutex and wait for notification.
 */
int
sysMonitorWait(sys_mon_t *mid, int millis, bool_t clear)
{
    sys_thread_t *self = sysThreadSelf();
    int entry_count;

    sysAssert(mid != SYS_MID_NULL);
    entry_count = mid->entry_count;
    if (self == 0) {
        return SYS_ERR;
    }
    sysAssert(ThreadsInitialized);
    if (mid->monitor_owner != self) {
        return SYS_ERR;
    }
    if (sysThreadIsInterrupted(self, clear)) {
	return SYS_INTRPT;
    }

    mid->monitor_owner = SYS_THREAD_NULL;
    mid->entry_count = 0;
    self->state = CONDVAR_WAIT;
    if (millis == SYS_TIMEOUT_INFINITY) {
        condvarWait(self, &mid->condvar, &mid->mutex);
    } else {
        condvarTimedWait(self, &mid->condvar, &mid->mutex, millis);
    }
    self->state = RUNNABLE;
    mid->entry_count = entry_count;
    mid->monitor_owner = self;

    /* Did we get interrupted in mid-wait? */
    if (sysThreadIsInterrupted(self, clear)) {
	return SYS_INTRPT;
    } else {
	return SYS_OK;
    }
}

/*
 * Print monitor information.
 */
void
sysMonitorDumpInfo(sys_mon_t *mid)
{
    sysAssert(mid != SYS_MID_NULL);
    if (mid->monitor_owner != SYS_THREAD_NULL) {
        TID tid = mid->monitor_owner->cookie;
	jio_fprintf(stderr, "owner %s (0x%lx, %ld %s)\n",
		    Object2CString((JHandle *)THREAD(tid)->name),
		    mid->monitor_owner, mid->entry_count,
		    mid->entry_count == 1 ? "entry" : "entries");
    } else {
	jio_fprintf(stderr, "<unowned>\n");
    }
    if (mid->condvar.waiters > 0) {
	jio_fprintf(stderr, "        Waiters: %d\n", mid->condvar.waiters);
    }
    if (mid->condvar.tickets > 0) {
	jio_fprintf(stderr, "        Tickets: %d\n", mid->condvar.tickets);
    }
}

/*
 * Return size of system-dependent monitor structure.
 */
int
sysMonitorSizeof(void)
{
    return sizeof(struct sys_mon);
}
