/*
 * @(#)audio.c	1.1 98/11/18
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include <io.h> 
#include <fcntl.h> 
#include <sys/types.h> 
#include <sys/stat.h> 
#include <stdio.h> 
#include <windows.h> 
#ifdef WIN32_LEAN_AND_MEAN
#include <mmsystem.h>
#endif

#include "oobj.h"
#include "interpreter.h"

#include "sunaudio_common.h"

HWAVEOUT waveHandle;		/* Handle of wave output device */
struct buf *free_list;		/* List of available buffers */
CRITICAL_SECTION free_list_lock;
int free_buffers;		/* Number of buffers in free list */
int total_buffers;		/* Total number of buffers allocated */
HANDLE free_buffer_event;	/* Automatic-reset event triggered whenever
				   a buffer is freed by the callback routine */

/* ulaw8tol16 - Convert 8-bit ulaw to 16-bit linear */

__inline int
ulaw8tol16(int datum)
{
    static const int exp_lut[8] =
	{ 0, 132, 396, 924, 1980, 4092, 8316, 16764 };
    int sign, exponent, mantissa;

    datum = ~datum;
    sign = datum >> 7;
    exponent = (datum >> 4) & 0x07;
    mantissa = datum & 0x0f;
    datum = exp_lut[exponent] + (mantissa << (exponent + 3));
    return (sign < 0 ? -datum : datum);
}

void
printError(char *f, int r)
{
    TCHAR msg[80];

    jio_fprintf(stderr, "Error: %s: ", f);
    if (waveOutGetErrorText(r, msg, sizeof(msg)/sizeof(TCHAR)) 
	== MMSYSERR_NOERROR) {
#ifdef UNICODE
	jio_fprintus(stderr, msg, wcslen(msg));
	jio_fprintf(stderr, "\n");
#else
	jio_fprintf(stderr, "%s\n", msg);
#endif
    } else {
	jio_fprintf(stderr, "Unknown error\n");
    }
}
	
void
audioWrite(char *data, long dataLen)
{
    struct buf *bp;
    int len, i, r;
 
    while (dataLen > 0) {
	if (free_list == NULL && total_buffers < MAXBUF) {
	    /* Expand available buffer space */
	    bp = (struct buf *)malloc(sizeof(struct buf));
	    total_buffers++;
	} else {
	    if (free_list == NULL) {
		/* We must wait for a free buffer */
		while (free_list == NULL) {
		    WaitForSingleObject(free_buffer_event, INFINITE);
		}
	    }
	    EnterCriticalSection(&free_list_lock);
	    bp = free_list;
	    free_list = free_list->next;
	    --free_buffers;
	    LeaveCriticalSection(&free_list_lock);
	    r = waveOutUnprepareHeader(waveHandle, &bp->hdr, sizeof(WAVEHDR));
	    if (r != 0) {
		printError("waveOutUnprepareHeader", r);
		return;
	    }
	}
	len = min(dataLen, DATALEN);
	bp->hdr.lpData = (char *)bp->data;
	bp->hdr.dwBufferLength = len * sizeof(short);
	bp->hdr.dwBytesRecorded = len * sizeof(short);
	bp->hdr.dwUser = (DWORD)bp;
	bp->hdr.dwFlags = 0;
	bp->hdr.dwLoops = 0;
	r = waveOutPrepareHeader(waveHandle, &bp->hdr, sizeof(WAVEHDR));
	if (r != 0) {
	    printError("waveOutPrepareHeader", r);
	    return;
	}
	for (i = 0; i < len; i++) {
	    bp->data[i] = ulaw8tol16(data[i]);
	}
	r = waveOutWrite(waveHandle, &bp->hdr, sizeof(WAVEHDR));
	if (r != 0) {
	    printError("waveOutWrite", r);
	    return;
	}
	data += len;
	dataLen -= len;
    }
}

void CALLBACK 
waveCallBack(HWAVE waveHandle, UINT msg, DWORD inst, DWORD p1, DWORD p2)
{
    if (msg == MM_WOM_DONE) {
	struct buf *bp = (struct buf *)p1;
	EnterCriticalSection(&free_list_lock);
	bp->next = free_list;
	free_list = bp;
	free_buffers++;
	SetEvent(free_buffer_event);
	LeaveCriticalSection(&free_list_lock);
    }
}
