/*
 * @(#)SharedDC.cpp	1.5 99/07/29
 *
 * Copyright (c) 1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * Implementation of SharedDC class for managing concurrent access to a DC.  
 * See include file for details.
 *
 */

/*
 * Includes
 */
#include <windows.h>
#include <stdio.h>

#include "SharedDC.h"


/*
 * Class:       SharedDC
 *
 * Function:    CONSTRUCTOR
 *
 * Description: Constructs a SharedDC object to manage access to a Windows HDC.
 *
 * Parameters:  hdc - DC handle to manage.
 *
 * Returns:     void
 */
SharedDC::SharedDC( HDC hdc )
{
   /*
    * Initialize
    */
   m_hdc  = hdc;
   m_hmtx = CreateMutex( NULL, 0, NULL );
   memset( &m_dcbOriginal, 0, sizeof (m_dcbOriginal) );
   memset( &m_dcbCurrent, 0, sizeof (m_dcbCurrent) );

   /* 
    * Make sure that text drawing calls don't erase the background first
    */
   SetBkMode( hdc, TRANSPARENT );
}


/*
 * Class:       SharedDC
 *
 * Function:    DESTRUCTOR
 *
 * Description: Frees resources including the Windows DC being managed.  No 
 *              special allowances are made for objects selected into the DC.
 *              It is the responsibility of users of this object to call 
 *              Deselect when finished using it.
 *
 * Parameters:  void
 *
 * Returns:     void
 */
SharedDC::~SharedDC( void )
{
   /*
    * Done with semaphore
    */
   CloseHandle( m_hmtx );

   /*
    * Done with DC
    */
   DeleteDC( m_hdc );
}


/*
 * Class:       SharedDC
 *
 * Function:    Lock
 *
 * Description: Locks the DC for exclusive access by one thread.  Once the lock
 *              is aquired, the DC attributes are set according to the DCBUNDLE
 *              passed in.  If a new object is selected into the DC as a result,
 *              the existing object is NOT deleted.  This is the responsibility
 *              of the thread that originally passed them in with a previous
 *              call to this function.
 *
 * Parameters:  pdcb - Ptr to bundle attributes to set.  Can be NULL.
 *
 * Returns:     hdc  - Handle of Windows DC.
 */
HDC SharedDC::Lock( PDCBUNDLE pdcb )
{
   HBITMAP   hbmOld;
   HPEN      hpenOld;
   HBRUSH    hbrOld;
   HFONT     hfontOld;


   /*
    * Get exclusive access to this DC.
    */
   WaitForSingleObject( m_hmtx, INFINITE );

   /*
    * Set up our bundle
    *
    * NOTE 1: If no bundle is provided, leave the DC alone.
    *         This saves time if the DC is only being used as
    *         a source, eg. for BitBlt.
    *
    * NOTE 2: We only set an attribute in the DC if it is 
    *         not the same as the last time we set it.
    *
    * NOTE 3: When an object is selected into the DC, the old
    *         object is not deleted.  It is the resposibility
    *         of the caller to delete objects when finished
    *         with them.
    */
   if (pdcb != NULL)
   {
      if ((pdcb->hbm != 0) && (m_dcbCurrent.hbm != pdcb->hbm))
      {
         m_dcbCurrent.hbm = pdcb->hbm;
         hbmOld = (HBITMAP)SelectObject( m_hdc, m_dcbCurrent.hbm );

         if (m_dcbOriginal.hbm == 0)
         {
            m_dcbOriginal.hbm = hbmOld;
         }
      }

      if ((pdcb->hpen != 0) && (m_dcbCurrent.hpen != pdcb->hpen))
      {
         m_dcbCurrent.hpen = pdcb->hpen;
         hpenOld = (HPEN)SelectObject( m_hdc, m_dcbCurrent.hpen );

         if (m_dcbOriginal.hpen == 0)
         {
            m_dcbOriginal.hpen = hpenOld;
         }
      }

      if ((pdcb->hbr != 0) && (m_dcbCurrent.hbr != pdcb->hbr))
      {
         m_dcbCurrent.hbr = pdcb->hbr;
         hbrOld = (HBRUSH)SelectObject( m_hdc, m_dcbCurrent.hbr );

         if (m_dcbOriginal.hbr == 0)
         {
            m_dcbOriginal.hbr = hbrOld;
         }
      }

      if ((pdcb->hfont != 0) && (m_dcbCurrent.hfont != pdcb->hfont))
      {
         m_dcbCurrent.hfont = pdcb->hfont;
         hfontOld = (HFONT)SelectObject( m_hdc, m_dcbCurrent.hfont );

         if (m_dcbOriginal.hfont == 0)
         {
            m_dcbOriginal.hfont = hfontOld;
         }
      }

      if (m_dcbCurrent.clrPaint != pdcb->clrPaint)
      {
          m_dcbCurrent.clrPaint = pdcb->clrPaint;
          SetTextColor( m_hdc, m_dcbCurrent.clrPaint );
      }

      if (m_dcbCurrent.iROP2 != pdcb->iROP2)
      {
         m_dcbCurrent.iROP2 = pdcb->iROP2;
         SetROP2( m_hdc, m_dcbCurrent.iROP2 );
      }

      if ((pdcb->hrgnClip != 0) && (m_dcbCurrent.hrgnClip != pdcb->hrgnClip))
      {
         m_dcbCurrent.hrgnClip = pdcb->hrgnClip;
         SelectClipRgn( m_hdc, m_dcbCurrent.hrgnClip );
      }
   }

   return (m_hdc);
}


/*
 * Class:       SharedDC
 *
 * Function:    Unlock
 *
 * Description: Unlocks the DC when drawing operations have been completed.
 *              Any objects selected into the DC using SharedDC::Lock, could
 *              potentially remain selected for a long time.  To force
 *              deselection, a call to SharedDC::Lock or SharedDC::Deselect
 *              should be made.
 *
 * Parameters:  void
 *
 * Returns:     void
 */
void SharedDC::Unlock( void )
{
   ReleaseMutex( m_hmtx );
}


/*
 * Class:       SharedDC
 *
 * Function:    Deselect
 *
 * Description: Deselects objects from a DC.  This call should be made prior to
 *              deleting any graphics objects to ensure that they are no longer
 *              selected into the DC.
 *
 * Parameters:  pdcb - Ptr to bundle attributes to be deselected.
 *
 * Returns:     void
 */
void SharedDC::Deselect( PDCBUNDLE pdcb )
{
   /*
    * Get exclusive access to this DC.
    */
   WaitForSingleObject( m_hmtx, INFINITE );

   /*
    * Deselect any objects currently selected into this
    * DC by restoring to their original values.
    */
   if (pdcb->hbm == m_dcbCurrent.hbm)
   {
      SelectObject( m_hdc, m_dcbOriginal.hbm );
      m_dcbCurrent.hbm = m_dcbOriginal.hbm;
   }

   if (pdcb->hpen == m_dcbCurrent.hpen)
   {
      SelectObject( m_hdc, m_dcbOriginal.hpen );
      m_dcbCurrent.hpen = m_dcbOriginal.hpen;
   }

   if (pdcb->hbr == m_dcbCurrent.hbr)
   {
      SelectObject( m_hdc, m_dcbOriginal.hbr );
      m_dcbCurrent.hbr = m_dcbOriginal.hbr;
   }

   if (pdcb->hfont == m_dcbCurrent.hfont)
   {
      SelectObject( m_hdc, m_dcbOriginal.hfont );
      m_dcbCurrent.hfont = 0;
   }

   if (pdcb->hrgnClip == m_dcbCurrent.hrgnClip)
   {
      SelectClipRgn( m_hdc, 0 );
      m_dcbCurrent.hrgnClip = 0;
   }

   /*
    * Done with DC
    */
   ReleaseMutex( m_hmtx );
}
