/*
 * @(#)WGraphics.java	1.27 98/08/19
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.awt.windows;

import java.awt.*;
import java.io.*;
import java.util.*;
import java.awt.image.ImageObserver;
import sun.awt.image.OffScreenImageSource;
import sun.awt.image.ImageRepresentation;
import sun.awt.FontDescriptor;
import sun.awt.PlatformFont;
import sun.awt.CharsetString;
import java.io.CharConversionException;
import sun.io.CharToByteConverter;

/**
 * WGraphics is an object that encapsulates a graphics context for a
 * particular canvas. 
 * 
 * @version 1.18 18 Mar 1996
 */

public class WGraphics extends Graphics {
    int		pData;
    Color	foreground;
    Font	font;
    int		originX;
    int		originY;

    private native void createFromComponent(WComponentPeer w);
    private native void createFromGraphics(WGraphics g);
    private native void createFromPrintJob(WPrintJob pj);
    private native void createFromHDC(int hdc);
    private native void imageCreate(ImageRepresentation ir);
    private native void pSetFont(Font f);
    private native void pSetForeground(int value);

    private sun.awt.image.Image		image;

    WGraphics(WPrintJob pj) {
	createFromPrintJob(pj);
    }
    WGraphics(WGraphics g) {
	createFromGraphics(g);
    }
    WGraphics(WComponentPeer comp) {
	createFromComponent(comp);
    }
    WGraphics(int hdc) {
        createFromHDC(hdc);
    }

    public WGraphics(WImage image) {
	OffScreenImageSource osis = (OffScreenImageSource)image.getSource();
	imageCreate(osis.getImageRep());
	this.image = image;
    }

    /**
     * Create a new WGraphics Object based on this one.
     */
    public Graphics create() {
	WGraphics g = new WGraphics(this);
	g.foreground = foreground;
	g.font = font;
	g.originX = originX;
	g.originY = originY;
	g.image = image;
	return g;
    }

    /**
     * Translate
     */
    public void translate(int x, int y) {
	originX += x;
	originY += y;
    }

    private boolean disposed = false;

    /*
     * Subclasses should override disposeImpl() instead of dispose(). Client
     * code should always invoke dispose(), never disposeImpl().
     *
     * A WGraphics cannot be used after being disposed.
     */
    protected native void disposeImpl();
    public final void dispose() {
        boolean call_disposeImpl = false;

        if (!disposed) {
            synchronized (this) {
                if (!disposed) {
                    disposed = call_disposeImpl = true;
                }
            }
        }

        if (call_disposeImpl) {
            disposeImpl();
        }
    }
    public void finalize() {
        dispose();
	super.finalize();
    }

    public void setFont(Font font) {
	if ((font != null) && (this.font != font)) {
	    this.font = font;
	    pSetFont(font);
	}
    }
    public Font getFont() {
	return font;
    }

    /**
     * Gets font metrics for the given font.
     */
    public FontMetrics getFontMetrics(Font font) {
	return WFontMetrics.getFontMetrics(font);
    }

    /**
     * Sets the foreground color.
     */
    public void setColor(Color c) {
	if ((c != null) && (c != foreground)) {
	    foreground = c;
	    pSetForeground(c.getRGB());
	}
    }
    public Color getColor() {
	return foreground;
    }

    /**
     * Sets the paint mode to overwrite the destination with the
     * current color. This is the default paint mode.
     */
    public native void setPaintMode();

    /**
     * Sets the paint mode to alternate between the current color
     * and the given color.
     */
    public native void setXORMode(Color c1);

    /** 
     * Gets the current clipping area 
     */
    public native Rectangle getClipBounds();

    private native void changeClip(int X, int Y, int W, int H, boolean set);
    private native void removeClip();

    /** Crops the clipping rectangle for this WGraphics context. */
    public void clipRect(int X, int Y, int W, int H) {
	changeClip(X, Y, W, H, false);
    }

    /** Sets the clipping rectangle for this WGraphics context. */
    public void setClip(int X, int Y, int W, int H) {
	changeClip(X, Y, W, H, true);
    }

    /** Returns a Shape object representing the clip. */
    public Shape getClip() {
	return getClipBounds();
    }

    /** Sets the clip to a Shape (only Rectangle allowed). */
    public void setClip(Shape clip) {
	if (clip == null) {
	    removeClip();
	} else if (clip instanceof Rectangle) {
	    Rectangle r = (Rectangle) clip;
	    changeClip(r.x, r.y, r.width, r.height, true);
	} else {
	    throw new IllegalArgumentException("setClip(Shape) only supports Rectangle objects");
	}
    }

    /** Clears the rectangle indicated by x,y,w,h. */
    public native void clearRect(int x, int y, int w, int h);

    /** Fills the given rectangle with the foreground color. */
    public native void fillRect(int X, int Y, int W, int H);

    /** Draws the given rectangle. */
    public native void drawRect(int X, int Y, int W, int H);


    /** Draws the given string. */
    public void drawString(String string, int x, int y) {
        drawChars(string.toCharArray(), 0, string.length(), x, y);
    }

    /** Draws the given character array. */
    public void drawChars(char chars[], int offset, int length, int x, int y) {
	PlatformFont pf = ((PlatformFont) font.getPeer());
	if (pf.mightHaveMultiFontMetrics()) {
	    drawMFChars(chars, offset, length, x, y, font, pf);
	} else {
            drawSFChars(chars, offset, length, x, y);
	}
    }

    private native void drawSFChars(char[] chars, int offset, int length, 
                                    int x, int y);

    private void drawMFChars(char chars[], int offset, int length, 
                             int x, int y, Font font, PlatformFont pf) {
        if (length == 0) {
            return;
        }

	sun.awt.CharsetString[] css = 
            pf.makeMultiCharsetString(chars, offset, length);

        byte[] csb = null;
	int baLen = 0;
	int w = 0;
	for (int i = 0; i < css.length; i++) {
	    CharsetString cs = css[i];

            if (WFontMetrics.needsConversion(font, cs.fontDescriptor)) {
                try {
                    int l = cs.fontDescriptor.fontCharset.getMaxBytesPerChar()
                        * cs.length;
                    if (l > baLen) {
                        baLen = l;
                        csb = new byte[baLen];
                    } 
                // use a new instance of the fontCharset (bug 4078870)
				CharToByteConverter cv = (CharToByteConverter)
					cs.fontDescriptor.fontCharset.getClass().newInstance();
                int csbLen = cv.convert(cs.charsetChars, cs.offset,
										cs.offset + cs.length, csb, 0, 
										csb.length);
                    w += drawMFCharsConvertedSegment(font, cs.fontDescriptor, 
                                                     csb, csbLen, x+w, y);
                } catch (InstantiationException ignored) {
				} catch (IllegalAccessException ignored) {
					// CharToByte converters are public and can be instantiated
					// so no exceptions should be thrown
                } catch (CharConversionException ignored) {
                }
	    } else {
                    w += drawMFCharsSegment(font, cs.fontDescriptor, 
                                            cs.charsetChars, cs.offset,
                                            cs.length, x+w, y);
            }
	}
    }

    private native int drawMFCharsSegment(Font font, FontDescriptor des, 
                                          char[] chars, int offset, int charLength,
                                          int x, int y);
    private native int drawMFCharsConvertedSegment(Font font, FontDescriptor des, 
                                                   byte[] bytes, int length, 
                                                   int x, int y);

    /** Draws the given byte array. */
    public native void drawBytes(byte data[], int offset, int length, 
                                 int x, int y);

    /** Draws the given line. */
    public native void drawLine(int x1, int y1, int x2, int y2);

    /**
     * Draws an image at x,y in nonblocking mode with a callback object.
     */
    public boolean drawImage(Image img, int x, int y, ImageObserver observer) {
	WImage wImg = (WImage) img;
	if (wImg.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img,
				     ImageObserver.ERROR|ImageObserver.ABORT,
				     -1, -1, -1, -1);
	    }
	    return false;
	}
	ImageRepresentation ir = wImg.getImageRep();
	return ir.drawImage(this, x, y, null, observer);
    }

    /**
     * Draws an image scaled to x,y,w,h in nonblocking mode with a
     * callback object.
     */
    public boolean drawImage(Image img, int x, int y, int width, int height,
			     ImageObserver observer) {
	if (width == 0 || height == 0) {
	    return true;
	}
	WImage wImg = (WImage) img;
	if (wImg.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img,
				     ImageObserver.ERROR|ImageObserver.ABORT,
				     -1, -1, -1, -1);
	    }
	    return false;
	}
	ImageRepresentation ir = wImg.getImageRep();
	return ir.drawScaledImage(this, x, y, width, height, null, observer);
    }

    /**
     * Draws an image at x,y in nonblocking mode with a solid background
     * color and a callback object.
     */
    public boolean drawImage(Image img, int x, int y, Color bg,
			     ImageObserver observer) {
	WImage wImg = (WImage) img;
	if (wImg.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img,
				     ImageObserver.ERROR|ImageObserver.ABORT,
				     -1, -1, -1, -1);
	    }
	    return false;
	}
	ImageRepresentation ir = wImg.getImageRep();
	return ir.drawImage(this, x, y, bg, observer);
    }

    /**
     * Draws an image scaled to x,y,w,h in nonblocking mode with a
     * solid background color and a callback object.
     */
    public boolean drawImage(Image img, int x, int y, int width, int height,
			     Color bg, ImageObserver observer) {
	if (width == 0 || height == 0) {
	    return true;
	}
	WImage wImg = (WImage) img;
	if (wImg.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img,
				     ImageObserver.ERROR|ImageObserver.ABORT,
				     -1, -1, -1, -1);
	    }
	    return false;
	}
	ImageRepresentation ir = wImg.getImageRep();
	return ir.drawScaledImage(this, x, y, width, height, bg, observer);
    }

    /**
     * Draws a subrectangle of an image scaled to a destination rectangle
     * in nonblocking mode with a callback object.
     */
    public boolean drawImage(Image img,
			     int dx1, int dy1, int dx2, int dy2,
			     int sx1, int sy1, int sx2, int sy2,
			     ImageObserver observer) {
	if (dx1 == dx2 || dy1 == dy2) {
	    return true;
	}
	WImage wImg = (WImage) img;
	if (wImg.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img,
				     ImageObserver.ERROR|ImageObserver.ABORT,
				     -1, -1, -1, -1);
	    }
	    return false;
	}
	ImageRepresentation ir = wImg.getImageRep();
	return ir.drawStretchImage(this,
				   dx1, dy1, dx2, dy2,
				   sx1, sy1, sx2, sy2,
				   null, observer);
    }

    /**
     * Draws a subrectangle of an image scaled to a destination rectangle in
     * nonblocking mode with a solid background color and a callback object.
     */
    public boolean drawImage(Image img,
			     int dx1, int dy1, int dx2, int dy2,
			     int sx1, int sy1, int sx2, int sy2,
			     Color bgcolor, ImageObserver observer) {
	if (dx1 == dx2 || dy1 == dy2) {
	    return true;
	}
	WImage wImg = (WImage) img;
	if (wImg.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img,
				     ImageObserver.ERROR|ImageObserver.ABORT,
				     -1, -1, -1, -1);
	    }
	    return false;
	}
	ImageRepresentation ir = wImg.getImageRep();
	return ir.drawStretchImage(this,
				   dx1, dy1, dx2, dy2,
				   sx1, sy1, sx2, sy2,
				   bgcolor, observer);
    }

    /**
     * Copies an area of the canvas that this graphics context paints to.
     * @param X the x-coordinate of the source.
     * @param Y the y-coordinate of the source.
     * @param W the width.
     * @param H the height.
     * @param dx the x-coordinate of the destination.
     * @param dy the y-coordinate of the destination.
     */
    public native void copyArea(int X, int Y, int W, int H, int dx, int dy);


    /** Draws a rounded rectangle. */
    public native void drawRoundRect(int x, int y, int w, int h,
				     int arcWidth, int arcHeight);

    /** Draws a filled rounded rectangle. */
    public native void fillRoundRect(int x, int y, int w, int h,
				     int arcWidth, int arcHeight);

    /** Draws a polygon defined by an array of x points and y points */
    public native void drawPolygon(int xPoints[], int yPoints[], int nPoints);

    /** Draws a polygon defined by an array of x points and y points */
    public native void drawPolyline(int xPoints[], int yPoints[], int nPoints);

    /** Fills a polygon with the current fill mask */
    public native void fillPolygon(int xPoints[], int yPoints[], int nPoints);

    /** Draws an oval to fit in the given rectangle */
    public native void drawOval(int x, int y, int w, int h);

    /** Fills an oval to fit in the given rectangle */
    public native void fillOval(int x, int y, int w, int h);

    /**
     * Draws an arc bounded by the given rectangle from startAngle to
     * endAngle. 0 degrees is a vertical line straight up from the
     * center of the rectangle. Positive angles indicate clockwise
     * rotations, negative angle are counter-clockwise.
     */
    public native void drawArc(int x, int y, int w, int h,
			       int startAngle,
			       int endAngle);

    /** fills an arc. arguments are the same as drawArc. */
    public native void fillArc(int x, int y, int w, int h,
			       int startAngle,
			       int endAngle);

    public String toString() {	
	return getClass().getName() + "[" + originX + "," + originY + "]";
    }

    // Print a peer using this graphics object.
    native void print(WComponentPeer peer);

    /* Outline the given region. */
    //public native void drawRegion(Region r);

    /* Fill the given region. */
    //public native void fillRegion(Region r);
}
