/*
 * @(#)awt_Component.cpp	1.172 98/09/02
 *
 * Copyright (c) 1996 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

#include "awt_Component.h"
#include "awt_Toolkit.h"
#include "awt_Window.h"
#include "awt_Frame.h"
#include "awt_MenuItem.h"
#include "awt_Graphics.h"
#ifdef DEBUG
#include "awt_Unicode.h"
#endif

#include <java_awt_Toolkit.h>
#include <java_awt_FontMetrics.h>
#include <java_awt_Color.h>
#include <java_awt_Cursor.h>
#include <java_awt_Event.h>
#include <java_awt_event_KeyEvent.h>
#include <java_awt_Insets.h>
#include <java_awt_Graphics.h>
#include <sun_awt_windows_WPanelPeer.h>
#include <java_awt_event_InputEvent.h>

// Begin -- Win32 SDK include files
#include <tchar.h>
// End -- Win32 SDK include files


extern "C" {
  #include <exceptions.h>   // we need this for java exception logging
};

const char* szAwtComponentClassName = "SunAwtComponent";

// Bug #4039858 (Selecting menu item causes bogus mouse click event)
BOOL g_bMenuLoop = FALSE;       
static HWND g_hwndDown = NULL;

// initialize static variable
UINT AwtComponent::m_CodePage = GetACP();

//////////////////////////////////////////////////////////////////////////
// AwtComponent class methods

AwtComponent::AwtComponent() : m_lock("AwtComponent")
{
    m_callbacksEnabled = FALSE;
    m_hwnd = NULL;

    m_colorForeground = 0;
    m_colorBackground = 0;
    m_backgroundColorSet = FALSE;
    m_penForeground = NULL;
    m_brushBackground = NULL;
    m_sharedDC = NULL;
    m_DefWindowProc = NULL;
    m_nextControlID = 1;
    m_hCursor = NULL;
    m_childList = NULL;
    m_myControlID = 0;
    m_dragged = FALSE;
    m_hdwp = NULL;
    m_validationNestCount = 0;

}

AwtComponent::~AwtComponent()
{
    // all component deletion should be done on the toolkit thread
    // to avoid race conditions (see bug 4051487)
    ASSERT(::GetCurrentThreadId() == AwtToolkit::MainThread());

    CriticalSection::Lock l(GetLock());

    if (m_hdwp != NULL) {
    // end any deferred window positioning, regardless
    // of m_validationNestCount
	::EndDeferWindowPos(m_hdwp);
    }
    
    // Stop message filtering.
    UnsubclassWindow();

    // Disconnect all links.
    UnlinkObjects();

    if (m_childList != NULL)
	delete m_childList;

    if (m_myControlID != 0) {
	AwtComponent* parent = GetParent();
	if (parent != NULL)
	    parent->RemoveChild(m_myControlID);
    }

    // Waiting until this base destructor assumes that we don't care about 
    // getting destroy related messages in derived classes, as this object 
    // is already torn down.
    DestroyHWnd();

    // Release any allocated resources.
    if (m_penForeground != NULL) {
        m_penForeground->Release();
        m_penForeground = NULL;
    }
    if (m_brushBackground != NULL) {
        m_brushBackground->Release();
        m_brushBackground = NULL;
    }
#if 0
    // DestroyCursor is only valid on cursors created with CreateCursor...
    if (m_hCursor != NULL) {
        ::DestroyCursor(m_hCursor);
        m_hCursor = NULL;
    }
#endif
}

///////////////////////////////////////////////////////////////////////////
// AwtComponent window class routines

ATOM thisPropertyAtom;

void AwtComponent::Init()
{
    char buf[40];
    VERIFY(sprintf(buf, "%s-0x%x", szAwtComponentClassName, 
                   ::GetCurrentThreadId()) < 40);
    thisPropertyAtom = ::GlobalAddAtom(buf);
    ASSERT(thisPropertyAtom != 0);
}

void AwtComponent::Term()
{
    VERIFY(::GlobalDeleteAtom(thisPropertyAtom) == NULL);
}

//
// Single window proc for all the components. Delegates real work to the component's
// WindowProc() member function.
//
LRESULT CALLBACK AwtComponent::WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    ASSERT(::IsWindow(hWnd));

    AwtComponent* self = GetComponent(hWnd);

    if (!self || self->GetHWnd() != hWnd) {
        return ::DefWindowProc(hWnd, message, wParam, lParam);
    }

    return self->WindowProc(message, wParam, lParam);
}

///////////////////////////////////////////////////////////////////////////
// AwtComponent instance routines

////////////////////////////////////////
// Window class registration routines

// Register window class; returns whether registry actually happened.
void AwtComponent::RegisterClass()
{
    const char* className = GetClassName();
    WNDCLASS  wc;

    if (!::GetClassInfo(::GetModuleHandle(NULL), className, &wc)) {
        wc.style         = CS_OWNDC;
	wc.lpfnWndProc   = (WNDPROC)::DefWindowProc;
	wc.cbClsExtra    = 0;
	wc.cbWndExtra    = 0;
	wc.hInstance     = ::GetModuleHandle(NULL);
	wc.hIcon         = AwtToolkit::GetInstance().GetAwtIcon();
	wc.hCursor       = NULL;
	wc.hbrBackground = NULL;
	wc.lpszMenuName  = NULL;
	wc.lpszClassName = className;

        ATOM ret = ::RegisterClass(&wc);
        ASSERT(ret != 0);
    }
}

void AwtComponent::UnregisterClass()
{
    ::UnregisterClass(GetClassName(), ::GetModuleHandle(NULL));
}

//
// Create the HWND for this component, & link up everyone
//
void 
AwtComponent::CreateHWnd(const char* title, long windowStyle, long windowExStyle,
                         int x, int y, int w, int h,
                         HWND hWndParent, HMENU hMenu, 
                         COLORREF colorForeground, COLORREF colorBackground,
                         Hsun_awt_windows_WComponentPeer* peer)
{
    CriticalSection::Lock l(GetLock());
    //The window class of multifont label must be "BUTTON" because "STATIC" class
    //can't get WM_DRAWITEM message, and m_peerObject member is referred in the
    //GetClassName method of AwtLabel class.
    //So m_peerObject member must be set here.
    m_peerObject = (Hsun_awt_windows_WObjectPeer*) peer; // C++ -> JavaPeer

    RegisterClass();

    Hjava_awt_Component* target = (Hjava_awt_Component*)(unhand(peer)->target);
    if (unhand(target)->visible) {
        windowStyle |= WS_VISIBLE;
    } else {
        windowStyle &= ~WS_VISIBLE;
    }

    HWND hwnd = ::CreateWindowEx(windowExStyle,
                               GetClassName(),
                               title,
                               windowStyle,
                               x, y, w, h,
                               hWndParent,
                               hMenu,
                               ::GetModuleHandle(NULL),
                               NULL);
    ASSERT(hwnd != NULL);
    m_hwnd = hwnd;

    LinkObjects(peer);

    // Subclass the window now so that we can snoop on its messages
    SubclassWindow();

    // Set default colors.
    m_colorForeground = colorForeground;
    m_colorBackground = colorBackground;

    // Only set background color if the color is actually set on the target -- this
    // avoids inheriting a parent's color unnecessarily, and has to be done here
    // because there isn't an API to get the real background color from outside the
    // AWT package.
    if (unhand(target)->background) {
        execute_java_dynamic_method(EE(), (HObject*)peer, 
                                    "setBackground", "(Ljava/awt/Color;)V", 
                                    unhand(target)->background);
    }
}

//
// Destroy this window's HWND
//
void AwtComponent::DestroyHWnd() {
    if (m_hwnd != NULL) {
        AwtToolkit::DestroyComponent(this);
        m_hwnd = NULL;
    }
}

//
// Install our window proc as the proc for our HWND, and save off the 
// previous proc as the default
//
void AwtComponent::SubclassWindow()
{
    m_DefWindowProc = (WNDPROC)::SetWindowLong(GetHWnd(), GWL_WNDPROC, (long)WndProc);
}

//
// Reinstall the original window proc as the proc for our HWND
//
void AwtComponent::UnsubclassWindow()
{
    if (m_DefWindowProc) {
        ::SetWindowLong(GetHWnd(), GWL_WNDPROC, (long)m_DefWindowProc);
        m_DefWindowProc = NULL;
    }
}

/////////////////////////////////////
// (static method)
// Determines the top-level ancestor for a given window. If the given
// window is a top-level window, return itself.
//
// 'Top-level' includes dialogs as well.
//
HWND AwtComponent::GetTopLevelParentForWindow(HWND hwndDescendant) {
    if (hwndDescendant == NULL) {
	return NULL;
    }

    ASSERT(IsWindow(hwndDescendant));
    HWND hwnd = hwndDescendant;
    for(;;) {
	LONG style = ::GetWindowLong(hwnd, GWL_STYLE);
	if ( (style & WS_CHILD) == 0 ) {
	    // found a non-child window so terminate
	    break;
	}
	hwnd = ::GetParent(hwnd);
    }

    ASSERT( !(::GetWindowLong(hwnd, GWL_STYLE) & WS_CHILD) );
    return hwnd;
}

////////////////////////////////////////
// 

void AwtComponent::SetColor(COLORREF c)
{
    if (m_colorForeground == c) {
        return;
    }
    m_colorForeground = c;
    if (m_penForeground != NULL) {
        m_penForeground->Release();
        m_penForeground = NULL;
    }
    VERIFY(::InvalidateRect(GetHWnd(), NULL, FALSE));
}

void AwtComponent::SetBackgroundColor(COLORREF c)
{
    if (m_colorBackground == c) {
        return;
    }
    m_colorBackground = c;
    m_backgroundColorSet = TRUE;
    if (m_brushBackground != NULL) {
        m_brushBackground->Release();
        m_brushBackground = NULL;
    }
    VERIFY(::InvalidateRect(GetHWnd(), NULL, TRUE));
}

HPEN AwtComponent::GetForegroundPen()
{
    if (m_penForeground == NULL) {
        m_penForeground = AwtPen::Get(m_colorForeground);
    }
    return m_penForeground->GetHandle();
}

COLORREF AwtComponent::GetBackgroundColor()
{
    if (m_backgroundColorSet == FALSE) {
        AwtComponent* c = this;
        while ((c = c->GetParent()) != NULL) {
            if (c->IsBackgroundColorSet()) {
                return c->GetBackgroundColor();
            }
        }
    }
    return m_colorBackground;
}

HBRUSH AwtComponent::GetBackgroundBrush()
{
    if (m_backgroundColorSet == FALSE) {
        if (m_brushBackground != NULL) {
            m_brushBackground->Release();
            m_brushBackground = NULL;
        }
          AwtComponent* c = this;
          while ((c = c->GetParent()) != NULL) {
              if (c->IsBackgroundColorSet()) {
                  m_brushBackground = 
                      AwtBrush::Get(c->GetBackgroundColor());
                  break;
              }
          }
    }
    if (m_brushBackground == NULL) {
        m_brushBackground = AwtBrush::Get(m_colorBackground);
    }
    return m_brushBackground->GetHandle();
}

void AwtComponent::SetFont(AwtFont* font)
{
    ASSERT(font != NULL);
    if (font->GetAscent() < 0) {
	AwtFont::SetupAscent(font);
    }
    SendMessage(WM_SETFONT, (WPARAM)font->GetHFont(), MAKELPARAM(FALSE, 0));
    VERIFY(::InvalidateRect(GetHWnd(), NULL, TRUE));
}

AwtSharedDC* AwtComponent::GetSharedDC()
{
    if (m_sharedDC == NULL) {
        HDC hWndDC = ::GetDC(GetHWnd());
        if (hWndDC != 0) {
            ::SetBkMode(hWndDC, TRANSPARENT);
            ::SelectObject(hWndDC, ::GetStockObject(NULL_BRUSH));
            ::SelectObject(hWndDC, ::GetStockObject(NULL_PEN));
            ::GdiFlush();
            m_sharedDC = new AwtSharedDC(this, hWndDC);
        }
    }
    return m_sharedDC;
}

AwtComponent* AwtComponent::GetParent() 
{
    HWND hwnd = ::GetParent(GetHWnd());
    if (hwnd == NULL) {
        return NULL;
    }
    return GetComponent(hwnd);
}

AwtWindow* AwtComponent::GetContainer() 
{
    AwtComponent* comp = this;
    while (comp != NULL) {
        if (comp->IsContainer()) {
            return (AwtWindow*)comp;
        }
        comp = comp->GetParent();
    }
    return NULL;
}

////////////////////////////////////////
// 

void AwtComponent::Show()
{
    VERIFY(::PostMessage(GetHWnd(), WM_AWT_COMPONENT_SHOW, SW_SHOWNA, 0));
}

void AwtComponent::Hide()
{
    VERIFY(::PostMessage(GetHWnd(), WM_AWT_COMPONENT_SHOW, SW_HIDE, 0));
}

void AwtComponent::Reshape(int x, int y, int w, int h)
{
    AwtWindow* container = GetContainer();
    AwtComponent* parent = GetParent();
    if (container != NULL && container == parent) {
        container->SubtractInsetPoint(x, y);
    }
    long flags = SWP_NOACTIVATE | SWP_NOZORDER | SWP_NOCOPYBITS;

    if (parent && strcmp(parent->GetClassName(), "SunAwtScrollPane") == 0) {
        if (x > 0) {
            x = 0;
        }
        if (y > 0) {
            y = 0;
        }

	RECT childRect;
	::GetClientRect(GetHWnd(), &childRect);
 
	/* if Reshape() is called and it is not as a result of a move, 
	   suppress repaints.  Given that scrollpane always adjusts
	   the dimensions of its child to be atleast as large as the
	   veiwport,  we can safely assume that these resizes
	   will be outside the viewrect. */

	if ((childRect.right != w) || (childRect.bottom != h)) {
	    flags |= SWP_NOREDRAW;
	}
    }
    if (m_hdwp != NULL) {
	m_hdwp = ::DeferWindowPos(m_hdwp, GetHWnd(), 0, x, y, w, h, flags);
	ASSERT(m_hdwp != NULL);
    } else {
	VERIFY(::SetWindowPos(GetHWnd(), 0, x, y, w, h, flags));
    }
}

void AwtComponent::SetScrollValues(UINT bar, int min, int value, int max)
{
    int minTmp, maxTmp;

    ::GetScrollRange(GetHWnd(), bar, &minTmp, &maxTmp);
    if (min == INT_MAX) {
        min = minTmp;
    }
    if (value == INT_MAX) {
        value = ::GetScrollPos(GetHWnd(), bar);
    }
    if (max == INT_MAX) {
        max = maxTmp;
    }
    if (min == max) {
        max++;
    }
    ::SetScrollRange(GetHWnd(), bar, min, max, FALSE);
    ::SetScrollPos(GetHWnd(), bar, value, TRUE);
}

////////////////////////////////////////
// 

//
// Oportunity to process and/or eat a message before it is dispatched
//
MsgRouting AwtComponent::PreProcessMsg(MSG& msg)
{
    return mrPassAlong;
}

static UINT lastMessage = WM_NULL;

//
// Dispatch messages for this window class--general component
//
LRESULT AwtComponent::WindowProc(UINT message, WPARAM wParam, LPARAM lParam)
{
    LRESULT retValue = 0;
    MsgRouting mr = mrDoDefault;

    lastMessage = message;

    switch (message) {
      case WM_CREATE:     mr = WmCreate(); break;
      case WM_CLOSE:      mr = WmClose(); break;
      case WM_DESTROY:    mr = WmDestroy(); break;

      case WM_ERASEBKGND: 
          mr = WmEraseBkgnd((HDC)wParam, *(BOOL*)&retValue); break;
      case WM_PAINT:      mr = WmPaint((HDC)wParam); break;
      case WM_MOVE:       mr = WmMove(LOWORD(lParam), HIWORD(lParam)); break;
      case WM_SIZE:       
      {
	  RECT r;
	  // fix 4128317 : use GetClientRect for full 32-bit int precision and
	  // to avoid negative client area dimensions overflowing 16-bit params - robi
	  ::GetClientRect( GetHWnd(), &r );
          mr = WmSize(wParam, r.right - r.left, r.bottom - r.top);
	  //mr = WmSize(wParam, LOWORD(lParam), HIWORD(lParam));
	  break;
      }
      case WM_SHOWWINDOW:
          mr = WmShowWindow(wParam, lParam); break;
      case WM_SYSCOMMAND: 
          mr = WmSysCommand(wParam & 0xFFF0, LOWORD(lParam), HIWORD(lParam)); break;
      case WM_EXITSIZEMOVE:
          mr = WmExitSizeMove(); break;

      // Bug #4039858 (Selecting menu item causes bogus mouse click event)
      case WM_ENTERMENULOOP:              
	  mr = WmEnterMenuLoop((BOOL)wParam);
          g_bMenuLoop = TRUE;
          break;     
      case WM_EXITMENULOOP:
	  mr = WmExitMenuLoop((BOOL)wParam);
          g_bMenuLoop = FALSE;
          break;     

      case WM_SETFOCUS:   mr = WmSetFocus((HWND)wParam); break;
      case WM_KILLFOCUS:  mr = WmKillFocus((HWND)wParam); break;
      case WM_ACTIVATE:   
          mr = WmActivate(LOWORD(wParam), (BOOL)HIWORD(wParam)); break;

#if defined(WIN32)
      case WM_CTLCOLORMSGBOX:
      case WM_CTLCOLOREDIT:
      case WM_CTLCOLORLISTBOX:
      case WM_CTLCOLORBTN:
      case WM_CTLCOLORDLG:
      case WM_CTLCOLORSCROLLBAR:
      case WM_CTLCOLORSTATIC:
          mr = WmCtlColor((HDC)wParam, (HWND)lParam, 
                          message-WM_CTLCOLORMSGBOX+CTLCOLOR_MSGBOX, 
                          *(HBRUSH*)&retValue); 
          break;
      case WM_HSCROLL:    
          mr = WmHScroll(LOWORD(wParam), HIWORD(wParam), (HWND)lParam); 
          break;
      case WM_VSCROLL:    
          mr = WmVScroll(LOWORD(wParam), HIWORD(wParam), (HWND)lParam); 
          break;
#else
      case WM_CTLCOLOR:   
          mr = WmCtlColor((HDC)wParam, (HWND)LOWORD(lParam), HIWORD(lParam),
                          *(HBRUSH*)&retValue); 
          break;
      case WM_HSCROLL:    
          mr = WmHScroll(wParam, LOWORD(lParam), (HWND)HIWORD(lParam)); 
          break;
      case WM_VSCROLL:
          mr = WmVScroll(wParam, LOWORD(lParam), (HWND)HIWORD(lParam)); 
          break;
#endif
      case WM_SETCURSOR:  
          mr = WmSetCursor((HWND)wParam, LOWORD(lParam), HIWORD(lParam), 
                           *(BOOL*)&retValue); 
          break;
      case WM_AWT_MOUSEENTER: 
          mr = WmMouseEnter(wParam, LO_INT(lParam), HI_INT(lParam)); break;
      case WM_LBUTTONDOWN:
      case WM_LBUTTONDBLCLK:
          mr = WmMouseDown(wParam, LO_INT(lParam), HI_INT(lParam), LEFT_BUTTON); break;
      case WM_MBUTTONDOWN:
      case WM_MBUTTONDBLCLK:
          mr = WmMouseDown(wParam, LO_INT(lParam), HI_INT(lParam), MIDDLE_BUTTON); break;
      case WM_RBUTTONDOWN:
      case WM_RBUTTONDBLCLK:
          mr = WmMouseDown(wParam, LO_INT(lParam), HI_INT(lParam), RIGHT_BUTTON); break;
      case WM_LBUTTONUP:
          mr = WmMouseUp(wParam, LO_INT(lParam), HI_INT(lParam), LEFT_BUTTON); break;
      case WM_RBUTTONUP:
          mr = WmMouseUp(wParam, LO_INT(lParam), HI_INT(lParam), RIGHT_BUTTON);
          break;
      case WM_MBUTTONUP:      
          mr = WmMouseUp(wParam, LO_INT(lParam), HI_INT(lParam), MIDDLE_BUTTON);
          break;
      case WM_MOUSEMOVE:
          mr = WmMouseMove(wParam, LO_INT(lParam), HI_INT(lParam)); 
          break;
      case WM_AWT_MOUSEEXIT:
          mr = WmMouseExit(wParam, LO_INT(lParam), HI_INT(lParam)); 
          break;

      case WM_KEYDOWN:
          mr = WmKeyDown(wParam, LOWORD(lParam), HIWORD(lParam), FALSE); 
          break;
      case WM_KEYUP:
          mr = WmKeyUp(wParam, LOWORD(lParam), HIWORD(lParam), FALSE); 
          break;
      case WM_CHAR:
          mr = WmChar(wParam, LOWORD(lParam), HIWORD(lParam), FALSE); 
          break;

      case WM_IME_CHAR:
          mr = WmIMEChar(wParam, LOWORD(lParam), HIWORD(lParam), FALSE); 
          break;

      case WM_INPUTLANGCHANGE:
      	  mr = WmInputLangChange(wParam, lParam);
	  // should return non-zero if we process this message
	  retValue = 1;
	  break;

      case WM_AWT_FORWARD_CHAR:
      	  mr = WmForwardChar(wParam, lParam);
	  break;

      case WM_AWT_FORWARD_BYTE:
          mr = HandleEvent((MSG*)lParam);
          break;

      case WM_SYSKEYDOWN:
          mr = WmKeyDown(wParam, LOWORD(lParam), HIWORD(lParam), TRUE); 
          break;
      case WM_SYSKEYUP:
          mr = WmKeyUp(wParam, LOWORD(lParam), HIWORD(lParam), TRUE); 
          break;
      case WM_SYSCHAR:
          mr = WmChar(wParam, LOWORD(lParam), HIWORD(lParam), TRUE); 
          break;

      case WM_TIMER:
          mr = WmTimer(wParam);
          break;

      case WM_COMMAND:    
#if defined(WIN32)
          mr = WmCommand(LOWORD(wParam), (HWND)lParam, HIWORD(wParam));
#else
          mr = WmCommand(LOWORD(wParam), (HWND)LOWORD(lParam), HIWORD(lParam));
#endif
          break;

      case WM_COMPAREITEM:
          mr = WmCompareItem(wParam, *(COMPAREITEMSTRUCT*)lParam, retValue);
          break;
      case WM_DELETEITEM: 
          mr = WmDeleteItem(wParam, *(DELETEITEMSTRUCT*)lParam);
          break;
      case WM_DRAWITEM:
          mr = WmDrawItem(wParam, *(DRAWITEMSTRUCT*)lParam);
          break;
      case WM_MEASUREITEM:
          mr = WmMeasureItem(wParam, *(MEASUREITEMSTRUCT*)lParam);
          break;

      case WM_AWT_HANDLE_EVENT:
          mr = HandleEvent((MSG*)lParam);
          break;

      case WM_AWT_PRINT_COMPONENT:
	  mr = PrintComponent((AwtGraphics*)wParam);
	  break;
      case WM_PRINT:
	  mr = WmPrint((HDC)wParam, lParam);
	  break;
       case WM_PRINTCLIENT:
	  mr = WmPrintClient((HDC)wParam, lParam);
	  break;

      case WM_NCCALCSIZE:
	  mr = WmNcCalcSize((BOOL)wParam, (LPNCCALCSIZE_PARAMS)lParam, 
			    *(UINT*)&retValue);
	  break;
      case WM_NCPAINT:
	  mr = WmNcPaint((HRGN)wParam);
	  break;
      case WM_NCHITTEST:
	  mr = WmNcHitTest(LOWORD(lParam), HIWORD(lParam), *(UINT*)&retValue);
	  break;

      case WM_AWT_RESHAPE_COMPONENT: {
          RECT* r = (RECT*)lParam;
          Reshape(r->left, r->top, r->right - r->left, r->bottom - r->top);
          delete r;
          mr = mrConsume;
          break;
      }

      case WM_AWT_BEGIN_VALIDATE:
          BeginValidate();
          mr = mrConsume;
          break;
      case WM_AWT_END_VALIDATE:
          EndValidate();
          mr = mrConsume;
          break;
			  
      case WM_QUERYNEWPALETTE:
	  mr = WmQueryNewPalette(*(UINT*)&retValue);
          break;
      case WM_PALETTECHANGED:
          mr = WmPaletteChanged((HWND)wParam);
          break;
      case WM_STYLECHANGED:
	  mr = WmStyleChanged(wParam, (LPSTYLESTRUCT)lParam);
	  break;
      case WM_SETTINGCHANGE:
	  mr = WmSettingChange(wParam, (LPCTSTR)lParam);
	  break;
      // These messages are used to route Win32 calls to the creating thread,
      // since these calls fail unless executed there.
      case WM_AWT_COMPONENT_SHOW:
          ::ShowWindow(GetHWnd(), wParam);
          mr = mrConsume;
          break;
      case WM_AWT_COMPONENT_SETFOCUS:
          if (wParam != NULL) {
              if ( ((HWND) wParam) == ::GetActiveWindow()) {
                  ::SetFocus(GetHWnd());
              } else {
                  // Don't call SetFocus if our top-level window is not
                  // active since SetFocus would activate it.
                  ::SendMessage((HWND) lParam, WM_KILLFOCUS, (WPARAM)GetHWnd(), 0);
                  ::SendMessage(GetHWnd(), WM_SETFOCUS, lParam, 0);
              }
          } else {
              ::SetFocus(GetHWnd());
          }
          mr = mrConsume;
          break;

      case WM_AWT_SET_SCROLL_INFO:
	  SCROLLINFO *si = (SCROLLINFO *) lParam;
	  ::SetScrollInfo(GetHWnd(), (int) wParam, si, TRUE);
	  delete si;
	  mr = mrConsume;
	  break;
    }

    // If not a specific Consume, it was a specific DoDefault, or a 
    // PassAlong (since the default is the next in chain), then call the 
    // default proc
    //
    if (mr != mrConsume) {
        retValue = DefWindowProc(message, wParam, lParam);
    }
    return retValue;
}

//
// Call this instance's default window proc, or if none set, call the stock 
// Window's one.
//
LRESULT AwtComponent::DefWindowProc(UINT msg, WPARAM wParam, LPARAM lParam)
{
    return m_DefWindowProc
        ? ::CallWindowProc((FARPROC)m_DefWindowProc, GetHWnd(), msg, wParam, lParam)
        : ::DefWindowProc(GetHWnd(), msg, wParam, lParam);
}

// This message should only be received when a window is destroy by
// Windows, and not Java.  Window termination has been reworked so
// this method should never be called during termination.
MsgRouting AwtComponent::WmDestroy()
{
    delete this;
    return mrConsume;
}

MsgRouting AwtComponent::WmMove(int x, int y)
{
    return mrDoDefault;
}

MsgRouting AwtComponent::WmSize(int type, int w, int h)
{
    return mrDoDefault;
}
  
MsgRouting AwtComponent::WmSysCommand(UINT uCmdType, UINT xPos, UINT yPos)
{
    return mrDoDefault;
}

MsgRouting AwtComponent::WmExitSizeMove()
{
    return mrDoDefault;
}

MsgRouting AwtComponent::WmEnterMenuLoop(BOOL isTrackPopupMenu)
{
    return mrDoDefault;
}

MsgRouting AwtComponent::WmExitMenuLoop(BOOL isTrackPopupMenu)
{
    return mrDoDefault;
}

MsgRouting AwtComponent::WmShowWindow(int show, long status)
{
    return mrDoDefault;
}

MsgRouting AwtComponent::WmSetFocus(HWND hWndLost)
{
    SendFocusEvent(java_awt_event_FocusEvent_FOCUS_GAINED);
    return mrDoDefault;
}

MsgRouting AwtComponent::WmKillFocus(HWND hWndGotFocus)
{
    HWND newTop = NULL;
    HWND oldTop = NULL;

    // Determine new top-level window with component gaining the focus
    newTop = AwtComponent::GetTopLevelParentForWindow(hWndGotFocus);

    // Determine old top-level window with component losing the focus
    oldTop = AwtComponent::GetTopLevelParentForWindow(GetHWnd());

    // FOCUS_LOST is temporary if we are switching to another top-level window
    BOOL isTemporary = (newTop != oldTop);

    // Post the FOCUS_LOST event. It will be redispatched to lightweight
    // components at the java.awt level
    SendFocusEvent(java_awt_event_FocusEvent_FOCUS_LOST, isTemporary);

    return mrDoDefault;
}

MsgRouting AwtComponent::WmCtlColor(HDC hDC, HWND hCtrl, UINT ctlColor, HBRUSH& retBrush)
{
    AwtComponent* child = AwtComponent::GetComponent(hCtrl);
    if (child) {
        ::SetBkColor(hDC, child->GetBackgroundColor());
        ::SetTextColor(hDC, child->GetColor());
        retBrush = child->GetBackgroundBrush();
        return mrConsume;
    }
    return mrDoDefault;
/*
    switch (ctlColor) {
        case CTLCOLOR_MSGBOX:
        case CTLCOLOR_EDIT:
        case CTLCOLOR_LISTBOX:
        case CTLCOLOR_BTN:
        case CTLCOLOR_DLG:
        case CTLCOLOR_SCROLLBAR:
        case CTLCOLOR_STATIC:
    }
*/
}

MsgRouting AwtComponent::WmHScroll(UINT scrollCode, UINT pos, HWND hScrollbar) {
    if (hScrollbar && hScrollbar != GetHWnd()) {  // the last test should never happen
        AwtComponent* sb = GetComponent(hScrollbar);
        if (sb) {
            sb->WmHScroll(scrollCode, pos, hScrollbar);
        }
    }
    return mrDoDefault;
}

MsgRouting AwtComponent::WmVScroll(UINT scrollCode, UINT pos, HWND hScrollbar) {
    if (hScrollbar && hScrollbar != GetHWnd()) {  // the last test should never happen
        AwtComponent* sb = GetComponent(hScrollbar);
        if (sb) {
            sb->WmVScroll(scrollCode, pos, hScrollbar);
        }
    }
    return mrDoDefault;
}

#define FT2INT64(ft) \
	((__int64)(ft).dwHighDateTime << 32 | (__int64)(ft).dwLowDateTime)

__int64 nowMillis()
{
    SYSTEMTIME st, st0;
    FILETIME   ft, ft0;

    (void)memset(&st0, 0, sizeof st0);
    st0.wYear  = 1970;
    st0.wMonth = 1;
    st0.wDay   = 1;
    SystemTimeToFileTime(&st0, &ft0);

    GetSystemTime(&st);
    SystemTimeToFileTime(&st, &ft);

    return (FT2INT64(ft) - FT2INT64(ft0)) / 10000;
    return 0;
}

MsgRouting AwtComponent::WmPaint(HDC)
{
    // Get the rectangle that covers all update regions, if any exist.
    RECT r;
    if (::GetUpdateRect(GetHWnd(), &r, FALSE)) {
        if ((r.right-r.left) > 0 && (r.bottom-r.top) > 0 &&
            m_peerObject != NULL && m_callbacksEnabled) {
            // Always call handlePaint, because the underlying control
            // will have painted itself (the "background") before any
            // paint method is called.
            DoCallback("handlePaint", "(IIII)V", 
                       r.left, r.top, r.right-r.left, r.bottom-r.top);
        }
    }
    return mrDoDefault;
}

MsgRouting AwtComponent::WmSetCursor(HWND hWnd, UINT hitTest, UINT message, 
                                     BOOL& retVal)
{
    if (hitTest == HTCLIENT) {
        ::SetCursor(GetCursor(TRUE));
        retVal = TRUE;
        return mrConsume;
    } else {
        return mrDoDefault;
    }
}

MsgRouting AwtComponent::WmMouseEnter(UINT flags, int x, int y)
{
    SendMouseEvent(java_awt_event_MouseEvent_MOUSE_ENTERED, 
                   nowMillis(), x, y, 0, 0, 0);
    return mrConsume;	// Don't pass our synthetic event on!
}

static MSG* CreateMessage(UINT message, WPARAM wParam, LPARAM lParam)
{
    MSG* msg = new MSG();
    msg->message = message;
    msg->wParam = wParam;
    msg->lParam = lParam;
    return msg;
}

// Double-click variables.
static ULONG multiClickTime = ::GetDoubleClickTime();
static int multiClickMaxX = ::GetSystemMetrics(SM_CXDOUBLECLK);
static int multiClickMaxY = ::GetSystemMetrics(SM_CYDOUBLECLK);
static AwtComponent* lastClickWnd = NULL;
static ULONG lastTime = 0;
static int lastClickX = 0;
static int lastClickY = 0;
static int clickCount = 0;

MsgRouting AwtComponent::WmMouseDown(UINT flags, int x, int y, int button)
{
    __int64 tm = nowMillis();

    ULONG now = (ULONG)tm;
    if (lastClickWnd == this &&
        (now - lastTime) <= multiClickTime && 
        abs(x - lastClickX) <= multiClickMaxX &&
        abs(y - lastClickY) <= multiClickMaxY) {
        clickCount++;
    }
    else {
        clickCount = 1;
    }
    lastClickWnd = this;
    lastTime = now;
    lastClickX = x;
    lastClickY = y;
    m_dragged = FALSE;

    MSG* msg = CreateMessage(lastMessage, flags, MAKELPARAM(x, y));
    SendMouseEvent(java_awt_event_MouseEvent_MOUSE_PRESSED, tm, x, y, 
                   GetJavaModifiers(flags, button), clickCount, 0, msg);

    return mrConsume;
}

MsgRouting AwtComponent::WmMouseUp(UINT flags, int x, int y, int button)
{
    MSG* msg = CreateMessage(lastMessage, flags, MAKELPARAM(x, y));
    SendMouseEvent(java_awt_event_MouseEvent_MOUSE_RELEASED, nowMillis(), 
                   x, y, GetJavaModifiers(flags, button), clickCount, 
                   (button == RIGHT_BUTTON? 1 : 0), msg);

    // If no movement, then report a click following the button release
    //
    if (!m_dragged) {
        SendMouseEvent(java_awt_event_MouseEvent_MOUSE_CLICKED, 
                       nowMillis(), x, y, 
                       GetJavaModifiers(flags, button), clickCount, 0);
    }
    m_dragged = FALSE;

    return mrConsume;
}

MsgRouting AwtComponent::WmMouseMove(UINT flags, int x, int y)
{
    static AwtComponent* lastComp = NULL;
    static int lastX = 0;
    static int lastY = 0;

    // Only report mouse move and drag events if a move or drag actually
    // happened -- Windows sends a WM_MOUSEMOVE in case the app wants to
    // modify the cursor.
    if (lastComp != this || x != lastX || y != lastY) {
        lastComp = this;
        lastX = x;
        lastY = y;

        long id;
        if ( flags & ALL_MK_BUTTONS ) {
            m_dragged = TRUE;
            id = java_awt_event_MouseEvent_MOUSE_DRAGGED;
        } 
        else 
            id = java_awt_event_MouseEvent_MOUSE_MOVED;

        MSG* msg = CreateMessage(lastMessage, flags, MAKELPARAM(x, y));
        SendMouseEvent(id, nowMillis(), x, y, 
                       GetJavaModifiers(flags, 0), 0, 0, msg);
    } else {
        ::SetCursor(GetCursor(TRUE));
    }
    return mrConsume;
}

MsgRouting AwtComponent::WmMouseExit(UINT flags, int x, int y)
{
    SendMouseEvent(java_awt_event_MouseEvent_MOUSE_EXITED, 
                   nowMillis(), x, y, 0, 0, 0);
    return mrConsume;	// Don't pass our synthetic event on!
}

long AwtComponent::GetJavaModifiers(UINT flags, int mouseButton)
{
    int modifiers = 0;

    if ((1<<24) & flags) {
        modifiers |= java_awt_event_InputEvent_META_MASK;
    }	
    if (HIBYTE(::GetAsyncKeyState(VK_CONTROL)) || 
        mouseButton == MIDDLE_BUTTON) {
        modifiers |= java_awt_event_InputEvent_CTRL_MASK;
    }
    if (HIBYTE(::GetAsyncKeyState(VK_SHIFT))) {
        modifiers |= java_awt_event_InputEvent_SHIFT_MASK;
    }
    if (HIBYTE(::GetAsyncKeyState(VK_MBUTTON)) || 
        HIBYTE(::GetAsyncKeyState(VK_MENU))) {
        modifiers |= java_awt_event_InputEvent_ALT_MASK;
    }

    // set "right mouse" mask depending on whether system is set to be
    // left or right-handed.
    if (::GetSystemMetrics(SM_SWAPBUTTON)) {
	if (HIBYTE(::GetAsyncKeyState(VK_LBUTTON))) {
	    modifiers |= java_awt_event_InputEvent_META_MASK;
	} 

    } else {
	if (HIBYTE(::GetAsyncKeyState(VK_RBUTTON))) {
	    modifiers |= java_awt_event_InputEvent_META_MASK;
	}
    }

    // Mouse buttons are already set correctly for left/right handedness
    if (mouseButton == LEFT_BUTTON) {
        modifiers |= java_awt_event_InputEvent_BUTTON1_MASK;
    }
    if (mouseButton == MIDDLE_BUTTON) {
        modifiers |= java_awt_event_InputEvent_ALT_MASK;
    }
    if (mouseButton == RIGHT_BUTTON) {
        modifiers |= java_awt_event_InputEvent_META_MASK;
    }
    return modifiers;
}

UINT AwtComponent::JavaKeyToWindowsKey(UINT wkey)
{
    switch (wkey) {
      case java_awt_event_KeyEvent_VK_DELETE: return VK_DELETE;
      case java_awt_event_KeyEvent_VK_ENTER: return VK_RETURN;
      case java_awt_event_KeyEvent_VK_BACK_SPACE: return VK_BACK;
      case java_awt_event_KeyEvent_VK_TAB: return VK_TAB;
      case java_awt_event_KeyEvent_VK_PRINTSCREEN: return VK_SNAPSHOT;
      case java_awt_event_KeyEvent_VK_INSERT: return VK_INSERT;
      case java_awt_event_KeyEvent_VK_HELP: return VK_HELP;
    }
    return wkey;
}

UINT AwtComponent::WindowsKeyToJavaKey(UINT wkey)
{
    switch (wkey) {
      case VK_DELETE:    return java_awt_event_KeyEvent_VK_DELETE;
      case VK_RETURN:    return java_awt_event_KeyEvent_VK_ENTER;
      case VK_BACK:      return java_awt_event_KeyEvent_VK_BACK_SPACE;
      case VK_TAB:       return java_awt_event_KeyEvent_VK_TAB;
      case VK_SNAPSHOT:  return java_awt_event_KeyEvent_VK_PRINTSCREEN;
      case VK_INSERT:    return java_awt_event_KeyEvent_VK_INSERT;
      case VK_HELP:      return java_awt_event_KeyEvent_VK_HELP;
    }
    return wkey;
}

UINT AwtComponent::WindowsKeyToJavaChar(UINT wkey, UINT modifiers)
{
    BYTE    keyboardState[AwtToolkit::KB_STATE_SIZE];
    WCHAR   unicodeChar[2];
    char    mbChar[4];
    UINT    scancode;
    int     conversionResult;

    scancode = ::MapVirtualKey(wkey, 0);
    AwtToolkit::GetAsyncKeyboardState(keyboardState);
    
    // apply modifiers if necessary
    if (modifiers)
    {
        if (modifiers & java_awt_Event_SHIFT_MASK)
        {
            keyboardState[VK_SHIFT] |= 0x80;
        }

        if (modifiers & java_awt_Event_CTRL_MASK)
        {
	    //fix for bug #4098210
	    keyboardState[VK_CONTROL] &= ~0x80;
	    if (wkey >= java_awt_event_KeyEvent_VK_A &&
		wkey <= java_awt_event_KeyEvent_VK_Z)
	    {
		keyboardState[VK_CONTROL] |= 0x80;
	    }
	    if ((((wkey & 0x7F) == java_awt_event_KeyEvent_VK_OPEN_BRACKET ) ||
		 ((wkey & 0x7F) == java_awt_event_KeyEvent_VK_BACK_SLASH   ) ||
		 ((wkey & 0x7F) == java_awt_event_KeyEvent_VK_CLOSE_BRACKET)) && 
		 ( (modifiers & java_awt_Event_SHIFT_MASK) == 0 ) )
	    {
		keyboardState[VK_CONTROL] |= 0x80;
	    }
	    if ( ( wkey == (UINT)::VkKeyScan('-')        ) && 
		 ( modifiers & java_awt_Event_SHIFT_MASK ) )
	    {
		keyboardState[VK_CONTROL] |= 0x80;
	    }
        }
        if (modifiers & java_awt_Event_ALT_MASK)
        {
 	    //fix for bug #4098210
 	    if (keyboardState[VK_CONTROL] & 0x80)
 		keyboardState[VK_MENU] &= ~0x80;
	    else
	        keyboardState[VK_MENU] |= 0x80;
        }
    }
    

    // instead of creating our own conversion tables, I'll let Win32
    // convert the character for me.
    conversionResult = ToAscii(wkey, scancode, keyboardState, (LPWORD)mbChar, 0);

    if (conversionResult == 0) 
    {
        // No translation available -- try known conversions or else punt.
        if (wkey == java_awt_event_KeyEvent_VK_DELETE ||
            wkey == java_awt_event_KeyEvent_VK_ENTER) 
        {
	    return wkey;
        }
	if (wkey >= VK_NUMPAD0 && wkey <= VK_NUMPAD9) 
        {
            return wkey - VK_NUMPAD0 + '0';
        }
	return 0;
    }
    // the caller expects a Unicode character.
    VERIFY(::MultiByteToWideChar(GetCodePage(), MB_PRECOMPOSED, mbChar, 1, unicodeChar, 1));
    return unicodeChar[0];
}

MsgRouting AwtComponent::WmKeyDown(UINT wkey, UINT repCnt, UINT flags, BOOL system)
{
  //VK_PROCESSKEY is not a valid Java key and has already been consumed.
    if (wkey == VK_PROCESSKEY) return mrDoDefault;

    UINT modifiers = GetJavaModifiers(flags, 0);
    UINT jkey = WindowsKeyToJavaKey(wkey);
    UINT character = WindowsKeyToJavaChar(jkey, modifiers);

    MSG* msg = CreateMessage((system ? WM_SYSKEYDOWN : WM_KEYDOWN), wkey, 
                             MAKELPARAM(repCnt, flags));
    SendKeyEvent(java_awt_event_KeyEvent_KEY_PRESSED, nowMillis(), jkey, 
                 character, modifiers, msg);
    return mrConsume;
}

MsgRouting AwtComponent::WmKeyUp(UINT wkey, UINT repCnt, UINT flags, BOOL system)
{
    if (wkey == VK_PROCESSKEY) return mrDoDefault;

    UINT modifiers = GetJavaModifiers(flags, 0);
    UINT jkey = WindowsKeyToJavaKey(wkey);
    UINT character = WindowsKeyToJavaChar(jkey, modifiers);

    MSG* msg = CreateMessage((system ? WM_SYSKEYUP : WM_KEYUP), wkey, 
                             MAKELPARAM(repCnt, flags));
    SendKeyEvent(java_awt_event_KeyEvent_KEY_RELEASED, nowMillis(), jkey,
                 character, modifiers, msg);
    return mrConsume;
}

static BOOL IsDBCSEditClass(AwtComponent *comp)
{
    const char* className = comp->GetClassName();
    if  ( stricmp(className,"Edit")==0 || stricmp(className,"ListBox")==0
        || stricmp(className,"ComboBox")==0 )
      return TRUE;
    else
      return FALSE;
}


MsgRouting AwtComponent::WmInputLangChange(UINT charset, UINT hKeyboardLayout)
{
    // Normally we would be able to use charset and TranslateCharSetInfo
    // to get a code page that should be associated with this keyboard
    // layout change. However, there seems to be an NT 4.0 bug associated
    // with the WM_INPUTLANGCHANGE message, which makes the charset parameter
    // unreliable, especially on Asian systems. Our workaround uses the 
    // keyboard layout handle instead.

    TCHAR strCodePage[MAX_ACP_STR_LEN];
    LANGID idLang;
    LCID idLocale;

    // a LANGID is stored in the bottom word of hKeyboardLayout
    idLang = LOWORD(hKeyboardLayout);
    // use the LANGID to create a LCID
    idLocale = MAKELCID(idLang, SORT_DEFAULT);
    // get the ANSI code page associated with this locale
    if (GetLocaleInfo(idLocale, LOCALE_IDEFAULTANSICODEPAGE, strCodePage, sizeof(strCodePage)) > 0 )
    {
        m_CodePage = _ttoi(strCodePage);
    }
    return mrConsume;
}

MsgRouting AwtComponent::WmIMEChar(UINT character, UINT repCnt, UINT flags, BOOL system)
{

    // We will simply create Java events here.
    TCHAR mbChar[3] = { _T('\0'), _T('\0'), _T('\0')};
    WCHAR unicodeChar;

    mbChar[0] = HIBYTE(character)?HIBYTE(character):LOBYTE(character);
    mbChar[1] = HIBYTE(character)?LOBYTE(character):_T('\0');

    ::MultiByteToWideChar(GetCodePage(), 0, mbChar, -1, &unicodeChar, 1);
    MSG* pMsg = CreateMessage(WM_IME_CHAR, character, MAKELPARAM(repCnt, flags));
    SendKeyEvent(java_awt_event_KeyEvent_KEY_TYPED, nowMillis(), 
                 java_awt_event_KeyEvent_CHAR_UNDEFINED, unicodeChar, 
                 GetJavaModifiers(flags, 0), pMsg);
    return mrConsume;
}

MsgRouting AwtComponent::WmChar(UINT character, UINT repCnt, UINT flags, BOOL system)
{

    // Will only get WmChar messages with DBCS if we create them for an Edit class 
    // in the WmForwardChar method. These synthesized DBCS chars are ok to pass on 
    // directly to the default window procedure. They've already been filtered through 
    // the Java key event queue. We will never get the trail byte since the edit class 
    // will PeekMessage(&msg, hwnd, WM_CHAR, WM_CHAR, PM_REMOVE).
    // I would like to be able to pass this character off via WM_AWT_FORWARD_BYTE, but 
    // the Edit classes don't seem to like that.
    // Begin pollution
    if (IsDBCSLeadByteEx(GetCodePage(), BYTE(character)))
    {  
        return mrDoDefault;
    }
    // End pollution
  
    // We will simply create Java events here.
    WCHAR unicodeChar = L'\0';
    UINT message = 0;

    message = system?WM_SYSCHAR:WM_CHAR;

    //fix for bug #4098210
    if (character == java_awt_event_KeyEvent_VK_DELETE) {
	if ((::GetKeyState(character) & 0x80) == 0)
	    unicodeChar = java_awt_event_KeyEvent_VK_BACK_SPACE;
    }
    else
        ::MultiByteToWideChar(GetCodePage(), 0, (TCHAR*)&character, 1, &unicodeChar, 1);

    MSG* pMsg = CreateMessage(message, character, MAKELPARAM(repCnt, flags));
    SendKeyEvent(java_awt_event_KeyEvent_KEY_TYPED, nowMillis(), 
                 java_awt_event_KeyEvent_CHAR_UNDEFINED, unicodeChar, 
                 GetJavaModifiers(flags, 0), pMsg);
    return mrConsume;
}

MsgRouting AwtComponent::WmForwardChar(UINT character, UINT lParam)
{
    // This message is sent from the Java key event handler.
    TCHAR mbChar[2];
    MSG* pMsg;
    mbChar[0] = HIBYTE(character)?HIBYTE(character):LOBYTE(character);
    mbChar[1] = HIBYTE(character)?LOBYTE(character):_T('\0');
    

    if (IsDBCSEditClass(this) && mbChar[1])
    {
        // The first WM_CHAR message will get handled by the WmChar, but
        // the second WM_CHAR message will get picked off by the Edit class.
        // WmChar will never see it.
        // If an Edit class gets a lead byte, it immediately calls PeekMessage
        // and pulls the trail byte out of the message queue.
        ::PostMessage(GetHWnd(), WM_CHAR, mbChar[0] & 0x00ff, lParam);
        ::PostMessage(GetHWnd(), WM_CHAR, mbChar[1] & 0x00ff, lParam);
    }
    else
    {
        pMsg = CreateMessage(WM_CHAR, mbChar[0] & 0x00ff, lParam);
        ::PostMessage(GetHWnd(), WM_AWT_FORWARD_BYTE, 0, (LPARAM)pMsg);
        if (mbChar[1])
        {
            pMsg = CreateMessage(WM_CHAR, mbChar[1] & 0x00ff, lParam);
            ::PostMessage(GetHWnd(), WM_AWT_FORWARD_BYTE, 0, (LPARAM)pMsg);
        }
    }
    return mrConsume;
}

MsgRouting AwtComponent::WmCommand(UINT id, HWND hWndChild, UINT notifyCode)
{
    // Menu/Accelerator
    if (hWndChild == 0) {
        AwtObject* obj = AwtToolkit::GetInstance().LookupCmdID(id);
        ASSERT(obj);
	ASSERT(((AwtMenuItem*)obj)->GetID() == id);
        obj->DoCommand();
        return mrConsume;
    }
    // Child id notification
    //
    else {
        AwtComponent* child = AwtComponent::GetComponent(hWndChild);
        if (child) {
            child->WmNotify(notifyCode);
        }
    }
    return mrDoDefault;
}

MsgRouting AwtComponent::WmNotify(UINT notifyCode)
{
    return mrDoDefault;
}

MsgRouting AwtComponent::WmCompareItem(UINT ctrlId, 
                                       COMPAREITEMSTRUCT far& compareInfo, 
                                       long& result)
{
    AwtComponent* child = AwtComponent::GetComponent(compareInfo.hwndItem);
    if (child == this) {
        //DoCallback("handleItemDelete",
    }
    else if (child) {
        return child->WmCompareItem(ctrlId, compareInfo, result);
    }
    return mrConsume;
}

MsgRouting AwtComponent::WmDeleteItem(UINT ctrlId, DELETEITEMSTRUCT far& deleteInfo)
{
    CriticalSection::Lock l(GetLock());
    /* Workaround for NT 4.0 bug -- if SetWindowPos is called on a AwtList
     * window, a WM_DELETEITEM message is sent to its parent with a window
     * handle of one of the list's child windows.  The property lookup
     * succeeds, but the HWNDs don't match. */
    if (deleteInfo.hwndItem == NULL) {
        return mrConsume;
    }
    AwtComponent* child = (AwtComponent*)::GetProp(deleteInfo.hwndItem, 
                                                   ThisProperty());
    if (child && child->GetHWnd() != deleteInfo.hwndItem) {
        return mrConsume;
    }

    if (child == this) {
        //DoCallback("handleItemDelete",
    }
    else if (child) {
        return child->WmDeleteItem(ctrlId, deleteInfo);
    }
    return mrConsume;
}
                            
MsgRouting AwtComponent::WmDrawItem(UINT ctrlId, DRAWITEMSTRUCT far& drawInfo)
{
    if (drawInfo.CtlType == ODT_MENU) {
        if (drawInfo.itemData != 0) {
            AwtMenu* menu = (AwtMenu*)(drawInfo.itemData);
	    AWT_TRACE(("AwtComponent::WmDrawItem %x\n", menu));
	    ASSERT( !IsBadReadPtr(menu, sizeof(*menu)) );
	    ASSERT( menu->GetHMenu() != NULL);
            menu->DrawItem(drawInfo);
        }
    } else {
        AwtComponent* child = AwtComponent::GetComponent(drawInfo.hwndItem);
        if (child == this) {
          //DoCallback("handleItemDelete",
        }
        else if (child != NULL) {
            return child->WmDrawItem(ctrlId, drawInfo);
        }
    }
    return mrConsume;
}
        
MsgRouting AwtComponent::WmMeasureItem(UINT ctrlId, MEASUREITEMSTRUCT far& measureInfo)
{
    if (measureInfo.CtlType == ODT_MENU) {
        if (measureInfo.itemData != 0) {
            AwtMenu* menu = (AwtMenu*)(measureInfo.itemData);
            HDC hDC = ::GetDC(GetHWnd());
            menu->MeasureItem(hDC, measureInfo);
            ::ReleaseDC(GetHWnd(), hDC);
        }
    } else {
        HWND  hChild = ::GetDlgItem(GetHWnd(), measureInfo.CtlID);
        AwtComponent* child = AwtComponent::GetComponent(hChild);
	// If the parent cannot find the child's instance from its handle,
	// maybe the child is in its creation.  So the child must be searched 
	// from the list linked before the child's creation.
        if (child == NULL)
            child = SearchChild((UINT)ctrlId);
        
        if (child == this) {
            //DoCallback("handleItemDelete", 
        }
        else if (child)
            return child->WmMeasureItem(ctrlId, measureInfo);
    }
    return mrConsume;
}


//for WmDrawItem method of Label, Button and Checkbox
void AwtComponent::DrawWindowText(HDC hDC, Hjava_awt_Font *hJavaFont, 
			Hjava_lang_String *hJavaString, int x, int y)
{
    int nOldBkMode = ::SetBkMode(hDC,TRANSPARENT);
    ASSERT(nOldBkMode != 0);
    AwtFont::drawMFString(hDC, hJavaFont, hJavaString, x, y);
    VERIFY(::SetBkMode(hDC,nOldBkMode));
}

// Draw text in gray (the color being set to COLOR_GRAYTEXT) when 
// the component is disabled.
// Used only for label, checkbox and button in OWNER_DRAW.
// It draws the text in emboss.
void AwtComponent::DrawGrayText(HDC hDC, Hjava_awt_Font *hJavaFont, 
			Hjava_lang_String *hJavaString, int x, int y)
{
    ::SetTextColor(hDC, ::GetSysColor(COLOR_BTNHILIGHT));
    AwtComponent::DrawWindowText(hDC, hJavaFont, hJavaString, x+1, y+1);
    ::SetTextColor(hDC, ::GetSysColor(COLOR_BTNSHADOW));
    AwtComponent::DrawWindowText(hDC, hJavaFont, hJavaString, x, y);
}

//for WmMeasureItem method of List and Choice
Hjava_lang_String* AwtComponent::GetItemString(Hjava_awt_Component *target, long index)
{
    Hjava_lang_String* str = (Hjava_lang_String*) execute_java_dynamic_method(EE(),
	(HObject *)target, "getItem",	"(I)Ljava/lang/String;", index);
    // ignore exceptions, which may occur due to race-conditions between
    // when a item is added and displayed.
    return str;
}

//for WmMeasureItem method of List and Choice
void AwtComponent::MeasureListItem(MEASUREITEMSTRUCT far& measureInfo)
{
    Hjava_awt_Dimension* dimension = PreferredItemSize();
    measureInfo.itemWidth = unhand(dimension)->width;
    measureInfo.itemHeight = unhand(dimension)->height;
}

//for WmDrawItem method of List and Choice
void AwtComponent::DrawListItem(DRAWITEMSTRUCT far& drawInfo)
{
    Hsun_awt_windows_WComponentPeer * self = GetPeer();
    Hjava_awt_Component* target = (Hjava_awt_Component*)unhand(self)->target;

    HDC hDC = drawInfo.hDC;
    RECT rect = drawInfo.rcItem;
    
    BOOL bEnabled = isEnabled();

    DWORD crBack, crText;
    if (drawInfo.itemState & ODS_SELECTED) {
	// Set background and text colors for selected item
	crBack = ::GetSysColor (COLOR_HIGHLIGHT);
	crText = ::GetSysColor (COLOR_HIGHLIGHTTEXT);
    } else {
	// Set background and text colors for unselected item
        crBack = GetBackgroundColor();
        crText = bEnabled ? GetColor() : ::GetSysColor(COLOR_GRAYTEXT);
    }

    // Fill item rectangle with background color
    AwtBrush* brBack = AwtBrush::Get(crBack);
    ASSERT(brBack->GetHandle());
    VERIFY(::FillRect(hDC, &rect, brBack->GetHandle()));
    brBack->Release();
    
    // Set current background and text colors
    ::SetBkColor (hDC, crBack);
    ::SetTextColor (hDC, crText);
    
    //draw string (with left mergin of 1 point)
    if ((int) (drawInfo.itemID) >= 0) {
	    Hjava_awt_Font *hJavaFont = GET_FONT(self);
	    Hjava_lang_String *hJavaString = GetItemString(target,
							   drawInfo.itemID);
	    SIZE size = AwtFont::getMFStringSize(hDC,hJavaFont,hJavaString);
	    AwtFont::drawMFString(hDC, hJavaFont, hJavaString,
				  rect.left+1,
				  (rect.top + rect.bottom-size.cy) / 2);
    }
    if ((drawInfo.itemState & ODS_FOCUS)  &&
        (drawInfo.itemAction & (ODA_FOCUS | ODA_DRAWENTIRE))) {
        VERIFY(::DrawFocusRect(hDC, &rect));
    }
}

//for MeasureListItem method and WmDrawItem method of Checkbox
long AwtComponent::GetFontHeight()
{
    Hsun_awt_windows_WComponentPeer * self = GetPeer();
    Hjava_awt_Component* target = (Hjava_awt_Component*)unhand(self)->target;

    Hjava_awt_Font* hJavaFont = GET_FONT(self);
    Hjava_awt_Toolkit* toolkit = 
        (Hjava_awt_Toolkit*) execute_java_dynamic_method(
            EE(), (HObject *)(target), "getToolkit", "()Ljava/awt/Toolkit;");
    ASSERT(!exceptionOccurred(EE()));
    Hjava_awt_FontMetrics* fontMetrics = 
        (Hjava_awt_FontMetrics*)execute_java_dynamic_method(
            EE(), (HObject *)(toolkit), "getFontMetrics", 
            "(Ljava/awt/Font;)Ljava/awt/FontMetrics;", hJavaFont);
    ASSERT(!exceptionOccurred(EE()));
    long height = execute_java_dynamic_method(EE(),
	(HObject *)(fontMetrics), "getHeight",	"()I");
    ASSERT(!exceptionOccurred(EE()));

    return height;
}

// Forward the print request to the Graphics object.  It's necessary
// to bounce the print request from WGraphics_print to here and back
// to avoid race conditions when sending print messages, by running on 
// the creating thread.
MsgRouting AwtComponent::PrintComponent(AwtGraphics* g)
{
    g->PrintComponent(this);
    return mrConsume;
}

// If you override WmPrint, make sure to save a copy of the DC on the GDI
// stack to be restored in WmPrintClient. Windows mangles the DC in
// ::DefWindowProc.
MsgRouting AwtComponent::WmPrint(HDC hDC, UINT flags)
{
    // Save two copies of the DC: one for WmPrintClient and one for
    // us to use after DefWindowProc returns
    ::SaveDC(hDC);
    ::SaveDC(hDC);

    DefWindowProc(WM_PRINT, (WPARAM) hDC, (LPARAM) flags);

    ::RestoreDC(hDC, -1);

    // Special case for components with a sunken border. Windows does not
    // print the border correctly, so we have to do it ourselves.
    if (IS_WIN4X && (::GetWindowLong(GetHWnd(), GWL_EXSTYLE) &
		     WS_EX_CLIENTEDGE)) {
        RECT r;
        ::GetClipBox(hDC, &r);
	::DrawEdge(hDC, &r, EDGE_SUNKEN, BF_RECT);
    }

    return mrConsume;
}

// If you override WmPrintClient, maker sure to obtain a valid copy of
// the DC from the GDI stack. The copy of the DC should have been placed
// there by WmPrint. Windows mangles the DC in ::DefWindowProc.
MsgRouting AwtComponent::WmPrintClient(HDC hDC, UINT)
{
    // obtain valid DC from GDI stack
    ::RestoreDC(hDC, -1);

    return mrDoDefault;
}

MsgRouting AwtComponent::WmNcCalcSize(BOOL fCalcValidRects, 
				      LPNCCALCSIZE_PARAMS lpncsp, UINT& retVal)
{
    return mrDoDefault;
}

MsgRouting AwtComponent::WmNcPaint(HRGN hrgn)
{
    return mrDoDefault;
}

MsgRouting AwtComponent::WmNcHitTest(UINT x, UINT y, UINT& retVal)
{
    return mrDoDefault;
}

MsgRouting AwtComponent::WmQueryNewPalette(UINT& retVal)
{
    if (AwtImage::m_hPal != NULL) {
        GetSharedDC()->Lock();
        HDC hDC = GetSharedDC()->GrabDC(NULL);
        ASSERT(hDC);
		
		#if defined(NETSCAPE)
		 	BOOL bBackgroundPalette = AwtToolkit::m_bInsideNavigator;
		#else
			BOOL bBackgroundPalette = FALSE;
		#endif // defined (NETSCAPE)

		ASSERT(::SelectPalette(hDC, AwtImage::m_hPal, bBackgroundPalette) != NULL);
		//Netscape : We dont want to invalidate here - RealizePalette will
		//automatically post a WM_PALETTECHANGED message, where we can
		//invalidate if necessary.
		
        ::RealizePalette(hDC);
        GetSharedDC()->Unlock();
        retVal = TRUE;
    } else {
        retVal = FALSE;
    }

    return mrDoDefault;
}

MsgRouting AwtComponent::WmPaletteChanged(HWND hwndPalChg)
{
//Netscape : We need to re-realize our palette here (unless we're the one
//that was realizing it in the first place).  That will let us match the
//remaining colors in the system palette as best we can.  We always invalidate
//because the palette will have changed when we recieve this message.

    if (hwndPalChg == GetHWnd()) {
        GetSharedDC()->Lock();
		GetSharedDC()->GrabDC(NULL);
		HDC hDC = GetSharedDC()->GetDC();
		ASSERT(hDC);
		ASSERT(::RealizePalette(hDC) != GDI_ERROR);
		GetSharedDC()->Unlock();
	}
	Invalidate(NULL);
    return mrDoDefault;
}

MsgRouting AwtComponent::WmStyleChanged(WORD wStyleType, LPSTYLESTRUCT lpss)
{
    ASSERT(!IsBadReadPtr(lpss, sizeof(STYLESTRUCT)));
    return mrDoDefault;
}

MsgRouting AwtComponent::WmSettingChange(WORD wFlag, LPCTSTR pszSection)
{
    ASSERT(!IsBadStringPtr(pszSection, 20));
    AWT_TRACE(("WM_SETTINGCHANGE: wFlag=%d pszSection=%s\n", (int)wFlag, pszSection));
    return mrDoDefault;
}

AwtComponent* AwtComponent::SearchChild(UINT id) {
    ChildListItem* child;
    for (child = m_childList; child != NULL;child = child->m_next) {
	if (child->m_ID == id)
	    return child->m_Component;
    }
    //ASSERT(FALSE); // This should not be happend if all children are recorded
    return NULL;	// To avoid compiler error.
}

void AwtComponent::RemoveChild(UINT id) {
    ChildListItem* child = m_childList;
    ChildListItem* lastChild = NULL;
    while (child != NULL) {
	if (child->m_ID == id) {
	    if (lastChild == NULL) {
		m_childList = child->m_next;
	    } else {
		lastChild->m_next = child->m_next;
	    }
	    child->m_next = NULL;
	    ASSERT(child != NULL);
	    delete child;
	    return;
	}
	lastChild = child;
	child = child->m_next;
    }
}

void AwtComponent::SendKeyEvent(long id, int64_t when, long raw, long cooked,
                                long modifiers, MSG* msg)
{
    CriticalSection::Lock l(GetLock());
    if (GetPeer() == NULL) {
        // event received during termination.
        return;
    }

    static ClassClass* classEvent = NULL;
    if (classEvent == NULL) {
        char* clsName = "java/awt/event/KeyEvent";
        classEvent = FindStickySystemClass(EE(), clsName, TRUE);
        if (classEvent == NULL) {
            SignalError(0, JAVAPKG "ClassNotFoundException", clsName);
            return;
        }
        ASSERT(classEvent);
    }

    Hjava_awt_event_KeyEvent* hEvent = (Hjava_awt_event_KeyEvent*)
        execute_java_constructor(EE(), NULL, classEvent,
                                 "(Ljava/awt/Component;IJIIC)", GetTarget(), 
                                 id, when, modifiers, raw, cooked);
    ASSERT(!exceptionOccurred(EE()));
    ASSERT(hEvent != NULL);
    unhand(hEvent)->data = (long)msg;
    SendEvent((Hjava_awt_AWTEvent*)hEvent);
}

void AwtComponent::SendMouseEvent(long id, int64_t when, long x, long y, 
                                  long modifiers, long clickCount, 
                                  long popupTrigger,
                                  MSG* msg)
{
    CriticalSection::Lock l(GetLock());
    if (GetPeer() == NULL) {
        // event received during termination.
        return;
    }

    static ClassClass* classEvent = NULL;
    if (classEvent == NULL) {
        char* clsName = "java/awt/event/MouseEvent";
        classEvent = FindStickySystemClass(EE(), clsName, TRUE);
        if (classEvent == NULL) {
            SignalError(0, JAVAPKG "ClassNotFoundException", clsName);
            return;
        }
        ASSERT(classEvent);
    }

    RECT insets;
    GetInsets(&insets);

    Hjava_awt_event_MouseEvent* hEvent = (Hjava_awt_event_MouseEvent*)
        execute_java_constructor(EE(), NULL, classEvent,
                                 "(Ljava/awt/Component;IJIIIIZ)", GetTarget(), 
                                 id, when, modifiers, x+insets.left, y+insets.top, 
				 clickCount, popupTrigger);
    ASSERT(!exceptionOccurred(EE()));
    ASSERT(hEvent != NULL);
    unhand(hEvent)->data = (long)msg;
    SendEvent((Hjava_awt_AWTEvent*)hEvent);
}

void AwtComponent::SendFocusEvent(long id, BOOL temporary)
{
    CriticalSection::Lock l(GetLock());
    if (GetPeer() == NULL) {
        // event received during termination.
        return;
    }

    static ClassClass* classEvent = NULL;
    if (classEvent == NULL) {
        char* clsName = "java/awt/event/FocusEvent";
        classEvent = FindStickySystemClass(EE(), clsName, TRUE);
        if (classEvent == NULL) {
            SignalError(0, JAVAPKG "ClassNotFoundException", clsName);
            return;
        }
    }

    Hjava_awt_event_FocusEvent* hEvent = (Hjava_awt_event_FocusEvent*)
        execute_java_constructor(EE(), NULL, classEvent,
                                 "(Ljava/awt/Component;IZ)", 
                                 GetTarget(), id, (long)temporary);
    ASSERT(!exceptionOccurred(EE()));
    ASSERT(hEvent != NULL);
    SendEvent((Hjava_awt_AWTEvent*)hEvent);
}

// Forward a filtered event directly to the subclassed window.
// This method is needed so that DefWindowProc is invoked on the 
// component's owning thread.
MsgRouting AwtComponent::HandleEvent(MSG* msg)
{
    DefWindowProc(msg->message, msg->wParam, msg->lParam);
    delete msg;
    return mrConsume;
}

void AwtComponent::SynthesizeKeyMessage(Hjava_awt_event_KeyEvent* hEvent)
{
    UINT message;
    switch (unhand(hEvent)->id) {
      case java_awt_event_KeyEvent_KEY_PRESSED:
	  message = WM_KEYDOWN; 
	  break;
      case java_awt_event_KeyEvent_KEY_RELEASED:
	  message = WM_KEYUP; 
	  break;
      case java_awt_event_KeyEvent_KEY_TYPED:
	  message = WM_CHAR; 
	  break;
      default:
	  return;
    }

    // KeyEvent.modifiers aren't supported -- the Java app must send separate
    // KEY_PRESSED and KEY_RELEASED events for the modifier virtual keys.

    if (unhand(hEvent)->id == java_awt_event_KeyEvent_KEY_TYPED) {
	// Convert unicode character to DBCS chars, and send each DBCS char as
	// a separate message.
	char mbs[2];
	UINT forwardChar;
	
	int retValue = ::WideCharToMultiByte(GetCodePage(), 0, 
					    (WCHAR*)&(unhand(hEvent)->keyChar),
					    1, mbs, 2, NULL, NULL);
	if (retValue == 2)
        {
            // if DBCS, then we'll package wParam up like wParam in a WM_IME_CHAR
            // message
            forwardChar = ((mbs[0] & 0x00FF) << 8) | (mbs[1] & 0x00FF);
        }
	else
        {
            forwardChar = mbs[0] & 0x00ff;
	}

        ::PostMessage(GetHWnd(), WM_AWT_FORWARD_CHAR, forwardChar, 0);

    } else {
        UINT key = AwtComponent::JavaKeyToWindowsKey(unhand(hEvent)->keyCode);
        MSG* msg = CreateMessage(message, key, 0);
	::PostMessage(GetHWnd(), WM_AWT_HANDLE_EVENT, 0, (LPARAM)msg);
    }
}

void AwtComponent::SynthesizeMouseMessage(Hjava_awt_event_MouseEvent* hEvent)
{
    enum Buttons { left, middle, right } button;
    if ((unhand(hEvent)->modifiers & java_awt_event_InputEvent_META_MASK) > 0) {
	button = right;
    } else if ((unhand(hEvent)->modifiers & java_awt_event_InputEvent_ALT_MASK) > 0) {
	button = middle;
    } else {
	button = left;
    }

    UINT message;
    switch (unhand(hEvent)->id) {
      case java_awt_event_MouseEvent_MOUSE_PRESSED: {
	  switch (button) {
            case left:   message = WM_LBUTTONDOWN; break;
            case middle: message = WM_MBUTTONDOWN; break;
            case right:  message = WM_RBUTTONDOWN; break;
	  }
	  break;
      }
      case java_awt_event_MouseEvent_MOUSE_RELEASED: {
	  switch (button) {
            case left:   message = WM_LBUTTONUP; break;
            case middle: message = WM_MBUTTONUP; break;
            case right:  message = WM_RBUTTONUP; break;
	  }
	  break;
      }
      case java_awt_event_MouseEvent_MOUSE_MOVED:
      // MOUSE_DRAGGED events must first have sent a MOUSE_PRESSED event.
      case java_awt_event_MouseEvent_MOUSE_DRAGGED:
	  message = WM_MOUSEMOVE;
	  break;
      default:
	  return;
    }

    MSG* msg = CreateMessage(message, 0, 
                             MAKELPARAM(unhand(hEvent)->x, unhand(hEvent)->y));
    ::PostMessage(GetHWnd(), WM_AWT_HANDLE_EVENT, 0, (LPARAM)msg);
}

void AwtComponent::Invalidate(RECT* r)
{
    ::InvalidateRect(GetHWnd(), r, FALSE);
}

void AwtComponent::BeginValidate()
{
    ASSERT(m_validationNestCount >= 0 &&
    	   m_validationNestCount < 1000); // sanity check

    if (m_validationNestCount == 0) {
    // begin deferred window positioning if we're not inside
    // another Begin/EndValidate pair
	ASSERT(m_hdwp == NULL);
	m_hdwp = ::BeginDeferWindowPos(32);
    }

    m_validationNestCount++;
}

void AwtComponent::EndValidate()
{
    ASSERT(m_validationNestCount > 0 &&
    	   m_validationNestCount < 1000); // sanity check
    ASSERT(m_hdwp != NULL);

    m_validationNestCount--;
    if (m_validationNestCount == 0) {
    // if this call to EndValidate is not nested inside another
    // Begin/EndValidate pair, end deferred window positioning
        ::EndDeferWindowPos(m_hdwp);
        m_hdwp = NULL;
    }
}

////////////////////////////////////////
// HWND, AwtComponent and Java Peer interaction

// Link the C++, Java peer, and HWNDs together.
//
void AwtComponent::LinkObjects(Hsun_awt_windows_WComponentPeer* hPeer)
{
    CriticalSection::Lock l(GetLock());

    // Any caller should hold the objects CRITICALSECTION.
    // Bind all three objects together thru this C++ object, two-way to each:
    //     JavaPeer <-> C++ <-> HWND
    //
    m_peerObject = (Hsun_awt_windows_WObjectPeer*) hPeer;  // C++ -> JavaPeer
    SET_PDATA(hPeer, this);                                // JavaPeer -> C++
    VERIFY(::SetProp(GetHWnd(), ThisProperty(), (HANDLE)this)); // HWND -> C++
}

void AwtComponent::UnlinkObjects()
{
    CriticalSection::Lock l(GetLock());

    // Any caller should hold the objects CRITICALSECTION.
    // Cleanup above binding
    //
    if (m_hwnd) {
        VERIFY(::RemoveProp(m_hwnd, ThisProperty()) != NULL);
    }
        // Note: order of operations should be the reverse of the link operation.
    if (m_peerObject) {
        unhand(m_peerObject)->pData = NULL;
        m_peerObject = NULL;
    }

    // Also cleanup any after-the-fact binding
    if (m_sharedDC) {
        AwtSharedDC* sdc = m_sharedDC;
        m_sharedDC = NULL;
        sdc->Detach(NULL);
    }
}

//////////////////////////////////////////////////////////////////////////////
// ComponentPeer native methods

void 
sun_awt_windows_WComponentPeer_show(Hsun_awt_windows_WComponentPeer* self)
{
    CHECK_PEER(self);
    AwtComponent* p = PDATA(AwtComponent, self);
    CriticalSection::Lock l(p->GetLock());
    p->Show();
}

void 
sun_awt_windows_WComponentPeer_hide(Hsun_awt_windows_WComponentPeer* self)
{
    CHECK_PEER(self);
    AwtComponent* p = PDATA(AwtComponent, self);
    CriticalSection::Lock l(p->GetLock());
    p->Hide();
}

void 
sun_awt_windows_WComponentPeer_enable(Hsun_awt_windows_WComponentPeer* self)
{
    CHECK_PEER(self);
    AwtComponent* p = (AwtComponent*)unhand(self)->pData;
    ::EnableWindow(p->GetHWnd(), TRUE);
    ::InvalidateRect(p->GetHWnd(), NULL, TRUE); // Bug #4038881 Labels don't enable and disable properly
    CriticalSection::Lock l(p->GetLock());
    p->VerifyState();
}

void 
sun_awt_windows_WComponentPeer_disable(Hsun_awt_windows_WComponentPeer* self)
{
    CHECK_PEER(self);
    AwtComponent* p = (AwtComponent*)unhand(self)->pData;
    ::EnableWindow(p->GetHWnd(), FALSE);
    ::InvalidateRect(p->GetHWnd(), NULL, TRUE); // Bug #4038881 Labels don't enable and disable properly
    CriticalSection::Lock l(p->GetLock());
    p->VerifyState();
}

void 
sun_awt_windows_WComponentPeer_reshape(Hsun_awt_windows_WComponentPeer* self,
                                       long x,
                                       long y,
                                       long w,
                                       long h)
{
    CHECK_PEER(self);
    AwtComponent* p = PDATA(AwtComponent, self);
    RECT* r = new RECT;
    ::SetRect(r, x, y, x+w, y+h);
    ::PostMessage(p->GetHWnd(), WM_AWT_RESHAPE_COMPONENT, 0, (LPARAM)r);
}

void
sun_awt_windows_WComponentPeer_nativeHandleEvent(Hsun_awt_windows_WComponentPeer* self,
                                              Hjava_awt_AWTEvent* hEvent)
{
    CHECK_NULL(hEvent, "null AWTEvent");
    if (unhand(hEvent)->data != 0) 
    {
        static BOOL keyDownConsumed = FALSE;
	static BOOL bCharChanged = FALSE;
	static WCHAR modifiedChar;	
	WCHAR unicodeChar;
        MSG* pMsg = (MSG*)unhand(hEvent)->data;
        unhand(hEvent)->data = 0;

        // Remember if a KEY_PRESSED event is consumed, as an old model 
        // program won't consume a subsequent KEY_TYPED event.
        if (unhand(hEvent)->consumed) 
        {
            keyDownConsumed = 
                (unhand(hEvent)->id == java_awt_event_KeyEvent_KEY_PRESSED);
            delete pMsg;
            return;
        }

        // Consume a KEY_TYPED event if a KEY_PRESSED had been, to support 
        // the old model.
        if (unhand(hEvent)->id == java_awt_event_KeyEvent_KEY_TYPED && 
          keyDownConsumed) 
        {
            delete pMsg;
            keyDownConsumed = FALSE;
            return;
        }

        // Modify any event parameters, if necessary.
        if (self && unhand(self)->pData && 
            unhand(hEvent)->id >= java_awt_event_KeyEvent_KEY_FIRST &&
            unhand(hEvent)->id <= java_awt_event_KeyEvent_KEY_LAST) 
        {

            AwtComponent* p = (AwtComponent*)unhand(self)->pData;
            CriticalSection::Lock l(p->GetLock());

            Hjava_awt_event_KeyEvent* keyEvent = 
                (Hjava_awt_event_KeyEvent*)hEvent;

            switch (unhand(hEvent)->id) 
            {
              case java_awt_event_KeyEvent_KEY_PRESSED:
	      {
                  UINT key = p->JavaKeyToWindowsKey(unhand(keyEvent)->keyCode);
	          if (!key) 
	          {
	              key = unhand(keyEvent)->keyCode;
	          }
	          UINT modifiers = unhand(keyEvent)->modifiers;
	          modifiedChar = p->WindowsKeyToJavaChar(key, modifiers);
	          bCharChanged = (unhand(keyEvent)->keyChar != modifiedChar);
	      }

                  break;
              case java_awt_event_KeyEvent_KEY_RELEASED:
                  keyDownConsumed = FALSE;
		  bCharChanged = FALSE;
		  break;
              case java_awt_event_KeyEvent_KEY_TYPED: 
              {
		  if (bCharChanged) 
		  {
		      unicodeChar = modifiedChar;
		  } 
		  else
		  {
		      unicodeChar = (WCHAR)unhand(keyEvent)->keyChar;
		  }
		  bCharChanged = FALSE;

                  TCHAR mbChar[2] = {'\0','\0',};
                  int len;
                  len = ::WideCharToMultiByte(p->GetCodePage(), 0, &unicodeChar, 1, mbChar, 2, NULL, NULL);
                  // after converting back to ANSI, send the character back to the native
                  // window for processing.
                  if (len>1)
                  {
                      // if DBCS, then we'll package wParam up like wParam in a WM_IME_CHAR
                      // message
                      pMsg->wParam = ((mbChar[0] & 0x00FF) << 8) | (mbChar[1] & 0x00FF);
                  }
		          else
                  {
                      pMsg->wParam = mbChar[0] & 0x00ff;
                  }
                  // the WM_AWT_FORWARD_CHAR handler will send this character to the
                  // default window proc
                  ::PostMessage(p->GetHWnd(), WM_AWT_FORWARD_CHAR, pMsg->wParam, pMsg->lParam);
                  delete pMsg;
                  return;
              }
                  break;
              default:
                  break;
            }
        }

        // Post the message directly to the subclassed component.
        // KEY_TYPED events get posted above.
        if (self && unhand(self)->pData) 
	    {
            AwtComponent* p = (AwtComponent*)unhand(self)->pData;
            ::PostMessage(p->GetHWnd(), WM_AWT_HANDLE_EVENT, 0, (LPARAM)pMsg);
        }
        return;
    } 
    
    // Forward any valid synthesized events.  Currently only mouse and key events
    // are supported.
    if (self == NULL || unhand(self)->pData == NULL) 
    {
	    return;
    }
    
    AwtComponent* p = (AwtComponent*)unhand(self)->pData;
    if (unhand(hEvent)->id >= java_awt_event_KeyEvent_KEY_FIRST &&
        unhand(hEvent)->id <= java_awt_event_KeyEvent_KEY_LAST) 
    {
	    p->SynthesizeKeyMessage((Hjava_awt_event_KeyEvent*)hEvent);
    } else if (unhand(hEvent)->id >= java_awt_event_MouseEvent_MOUSE_FIRST &&
        unhand(hEvent)->id <= java_awt_event_MouseEvent_MOUSE_LAST) 
    {
	    p->SynthesizeMouseMessage((Hjava_awt_event_MouseEvent*)hEvent);
    }
}

void 
sun_awt_windows_WComponentPeer__dispose(Hsun_awt_windows_WComponentPeer* self)
{
    ASSERT(self != NULL && unhand(self) != 0);
    AwtComponent* p = (AwtComponent*)unhand(self)->pData;
    if (p == NULL) {
        return;
    }

    ::SendMessage(AwtToolkit::GetInstance().GetHWnd(), WM_AWT_DISPOSE,
		  (WPARAM)self, 0);
}

void 
sun_awt_windows_WComponentPeer__setForeground(Hsun_awt_windows_WComponentPeer* self,
                                              long rgb)
{
    CHECK_PEER(self);
    AwtComponent* p = PDATA(AwtComponent, self);
    p->SetColor(PALETTERGB((rgb>>16)&0xff, 
		           (rgb>>8)&0xff, 
		           (rgb)&0xff));
    p->VerifyState();
}

void 
sun_awt_windows_WComponentPeer__setBackground(Hsun_awt_windows_WComponentPeer* self,
				 	     long rgb)
{
    CHECK_PEER(self);
    AwtComponent* p = PDATA(AwtComponent, self);
    p->SetBackgroundColor(PALETTERGB((rgb>>16)&0xff, 
		                     (rgb>>8)&0xff, 
		                     (rgb)&0xff));
    p->VerifyState();
}

Hjava_awt_Point*
sun_awt_windows_WComponentPeer_getLocationOnScreen(Hsun_awt_windows_WComponentPeer* self)
{
    CHECK_PEER_RETURN(self);
    AwtComponent* component = PDATA(AwtComponent, self);

    RECT rect;
    VERIFY(::GetWindowRect(component->GetHWnd(), &rect));

    Hjava_awt_Point *point =
        (Hjava_awt_Point*)execute_java_constructor(
            EE(), "java/awt/Point", 0, "(II)", rect.left, rect.top);

    if (point == NULL || exceptionOccurred(EE())) {
	SignalError(
            0, JAVAPKG "NullPointerException: point constructor failed", 0);
	return NULL;
    }

    return point;
}

void 
sun_awt_windows_WComponentPeer_setFont(Hsun_awt_windows_WComponentPeer* self,
				       Hjava_awt_Font* hFont)
{
    CHECK_PEER(self);
    ASSERT(hFont != NULL);

    Classjava_awt_Font* jFont = (Classjava_awt_Font*)unhand(hFont);
    AwtFont* font = (AwtFont*)jFont->pData;
    if (font == NULL) {
	//arguments of AwtFont::Create are changed for multifont component
	font = AwtFont::Create(hFont);
    }
    SET_PDATA(hFont, font);

    AwtComponent* p = PDATA(AwtComponent, self);
    p->SetFont(font);
}

void 
sun_awt_windows_WComponentPeer_requestFocus(Hsun_awt_windows_WComponentPeer* self)
{
    CHECK_PEER(self);
    AwtComponent* p = PDATA(AwtComponent, self);
    VERIFY(::PostMessage(p->GetHWnd(), WM_AWT_COMPONENT_SETFOCUS, 0, 0));
}

void
sun_awt_windows_WComponentPeer_start(Hsun_awt_windows_WComponentPeer* self)
{
    CHECK_PEER(self);
    AwtComponent* p = PDATA(AwtComponent, self);
    Hjava_awt_Component* target = (Hjava_awt_Component*)p->GetTarget();

    // Disable window if specified -- windows are enabled by default.
    if (!unhand(target)->enabled) {
	::EnableWindow(p->GetHWnd(), FALSE);
    }

    // The peer is now ready for callbacks, since this is the last 
    // initialization call
    p->EnableCallbacks(TRUE);
}

void
sun_awt_windows_WComponentPeer_setCursor(Hsun_awt_windows_WComponentPeer* self,
                                         Hjava_awt_Cursor* hJavaCursor)
{
    CHECK_PEER(self);
    CHECK_NULL(hJavaCursor, "null cursor");
    AwtComponent* comp = PDATA(AwtComponent, self);
    CriticalSection::Lock l(comp->GetLock());
    LPCTSTR cursor;

    switch (unhand(hJavaCursor)->type) {
      case java_awt_Cursor_DEFAULT_CURSOR:
      default:
          cursor = IDC_ARROW;
          break;
      case java_awt_Cursor_CROSSHAIR_CURSOR:
          cursor = IDC_CROSS;
          break;
      case java_awt_Cursor_TEXT_CURSOR:
          cursor = IDC_IBEAM;
          break;
      case java_awt_Cursor_WAIT_CURSOR:
          cursor = IDC_WAIT;
          break;
      case java_awt_Cursor_NE_RESIZE_CURSOR:
      case java_awt_Cursor_SW_RESIZE_CURSOR:
          cursor = IDC_SIZENESW;
          break;
      case java_awt_Cursor_SE_RESIZE_CURSOR:
      case java_awt_Cursor_NW_RESIZE_CURSOR:
          cursor = IDC_SIZENWSE;
          break;
      case java_awt_Cursor_N_RESIZE_CURSOR:
      case java_awt_Cursor_S_RESIZE_CURSOR:
          cursor = IDC_SIZENS;
          break;
      case java_awt_Cursor_W_RESIZE_CURSOR:
      case java_awt_Cursor_E_RESIZE_CURSOR:
          cursor = IDC_SIZEWE;
          break;
      case java_awt_Cursor_HAND_CURSOR:
          cursor = "HAND_CURSOR";
          break;
      case java_awt_Cursor_MOVE_CURSOR:
          cursor = IDC_SIZEALL;
          break;
    }

    HCURSOR hCursor;

#if 0
    hCursor = comp->GetCursor(FALSE);
    // DestroyCursor is only valid on cursors created with CreateCursor...
    if (hCursor != NULL && unhand(hCursor)->type == <custom type when it exists>) {
        VERIFY(::DestroyCursor(hCursor));
    }
#endif

    hCursor = ::LoadCursor(NULL, cursor);
    if (hCursor == NULL) {
        // Not a system cursor, check for resource.
        hCursor = ::LoadCursor(AwtToolkit::GetInstance().GetModuleHandle(), 
                               cursor);
    }
    if (hCursor == NULL) {
        // Not found anywhere, use default.
        hCursor = ::LoadCursor(NULL, IDC_ARROW);
        ASSERT(hCursor != NULL);
    }
    comp->SetCursor(hCursor);

    // Force the cursor to change if mouse is over it.
    POINT pt;
    ::GetCursorPos(&pt);
    // removed fix for 4040388, which caused 4117036 and
    // used Oracle's suggested fix - robi.khan@eng
    ::SetCursorPos(pt.x, pt.y);
}

void
sun_awt_windows_WComponentPeer__beginValidate(Hsun_awt_windows_WComponentPeer* self)
{
    CHECK_OBJ(self);
    if (unhand(self)->pData) {
        AwtComponent* p = PDATA(AwtComponent, self);
        ::PostMessage(p->GetHWnd(), WM_AWT_BEGIN_VALIDATE, 0, 0);
    }
}

void
sun_awt_windows_WComponentPeer_endValidate(Hsun_awt_windows_WComponentPeer* self)
{
    CHECK_OBJ(self);
    if (unhand(self)->pData) {
        AwtComponent* p = PDATA(AwtComponent, self);
        ::PostMessage(p->GetHWnd(), WM_AWT_END_VALIDATE, 0, 0);
    }
}

void
sun_awt_windows_WComponentPeer_setZOrderPosition(Hsun_awt_windows_WComponentPeer *self,
Hsun_awt_windows_WComponentPeer *aboveComp)
{
    CHECK_PEER(self);
    if (unhand(self)->pData) {
        AwtComponent *p = PDATA(AwtComponent, self);
	HWND hWndSelf = p->GetHWnd();
	HWND hWndAbove = NULL;
	HWND hWndInsertAfter = NULL;
	LONG selfStyle = GetWindowLong(hWndSelf, GWL_STYLE);

	if (!(selfStyle & WS_CHILD)) {
	// frames/windows/dialogs do not follow component front-to-back z-ordering rules
	    return;
	}

	// aboveComp refers to the component that should appear on top
	// of this one, so we must insert ourselves below it
        if (aboveComp != NULL && unhand(aboveComp)->pData) {
            AwtComponent* aboveW = PDATA(AwtComponent, aboveComp);
            hWndAbove = aboveW->GetHWnd();
        }
	
	// fixed 4059430 (re-opened) -- previous 1.1.7 fix was wrong, restored
	// 1.1.6 code and just added a couple of checks
	if (hWndAbove != NULL) {
	// need to insert in the middle of the z-order
	    ASSERT(IsWindow(hWndAbove));
	    hWndInsertAfter = hWndAbove;
	} else if (hWndAbove == NULL) {
	// insert at top of z-order when aboveComp is null
	    hWndInsertAfter = HWND_TOP;
	}

	// move the window into the correct position in the z-order
	::SetWindowPos(hWndSelf, hWndInsertAfter, 
		       0, 0, 0, 0, SWP_NOACTIVATE |SWP_NOMOVE | SWP_NOSIZE);
    }
}


/////////////////////////////////////////////////////////////////////////////
// Diagnostic routines
#ifdef DEBUG

void AwtComponent::VerifyState()
{
    if (AwtToolkit::GetInstance().VerifyComponents() == FALSE) {
        return;
    }

    if (m_callbacksEnabled == FALSE) {
        // Component is not fully setup yet.
        return;
    }

    // Get target bounds.
    Hjava_awt_Component* hTarget = (Hjava_awt_Component*)GetTarget();
    Classjava_awt_Component* target = unhand(hTarget);
    long x = target->x;
    long y = target->y;
    long width = target->width;
    long height = target->height;

    // Convert target origin to absolute coordinates
    Hjava_awt_Component* hComp = hTarget;
    while (TRUE) {
        Hjava_awt_Component* hParent = 
            (Hjava_awt_Component*)unhand(hComp)->parent;
        if (hParent == NULL) {
            break;
        }
        x += unhand(hParent)->x;
        y += unhand(hParent)->y;

        // If this component has insets, factor them in, but ignore
        // top-level windows.
        if (unhand(hParent)->parent != NULL) {
            Hsun_awt_windows_WPanelPeer* peer = (Hsun_awt_windows_WPanelPeer*)
		GetPeerForTarget((HObject*)hParent);
            if (peer != NULL && IsInstanceOf((HObject*)peer, 
                                             "sun/awt/windows/WPanelPeer")) {
                Hjava_awt_Insets* insets = (Hjava_awt_Insets*)
                    execute_java_dynamic_method(EE(), (HObject*)peer, "insets", 
                                                "()Ljava/awt/Insets;");
                x += unhand(insets)->left;
                y += unhand(insets)->top;
            }
        }

        hComp = hParent;
    }

    // Test whether component's bounds match the native window's
    RECT rect;
    VERIFY(::GetWindowRect(GetHWnd(), &rect));
#if 0
    ASSERT( (x == rect.left) && 
            (y == rect.top) &&
            (width == (rect.right-rect.left)) && 
            (height == (rect.bottom-rect.top)) );
#else
    BOOL fSizeValid = ( (x == rect.left) && 
            (y == rect.top) &&
            (width == (rect.right-rect.left)) && 
            (height == (rect.bottom-rect.top)) );
#endif

    // See if visible state matches
    long targetVisible = execute_java_dynamic_method(EE(), GetTarget(), 
                                                     "isShowing", "()Z");
    ASSERT(!exceptionOccurred(EE()));
    BOOL wndVisible = ::IsWindowVisible(GetHWnd());
#if 0
    ASSERT( (targetVisible && wndVisible) ||
            (!targetVisible && !wndVisible) );
#else
    BOOL fVisibleValid = ( (targetVisible && wndVisible) ||
            (!targetVisible && !wndVisible) );
#endif

    // Check enabled state
    BOOL wndEnabled = ::IsWindowEnabled(GetHWnd());
#if 0
    ASSERT( (target->enabled && wndEnabled) ||
            (!target->enabled && !wndEnabled) );
#else
    BOOL fEnabledValid = ( (target->enabled && wndEnabled) ||
            (!target->enabled && !wndEnabled) );

    if (!fSizeValid || !fVisibleValid || !fEnabledValid) {
        printf("AwtComponent::ValidateState() failed:\n");
        Hjava_lang_String* targetStr =  (Hjava_lang_String*)
            execute_java_dynamic_method(EE(), GetTarget(), "getName", 
                                        "()Ljava/lang/String;");
        ASSERT(!exceptionOccurred(EE()));
        printf("\t%S\n", GET_WSTRING(targetStr));
        printf("\twas:       [%d,%d,%dx%d]\n", x, y, width, height);
        if (!fSizeValid) {
            printf("\tshould be: [%d,%d,%dx%d]\n", rect.left, rect.top,
                   rect.right-rect.left, rect.bottom-rect.top);
        }
        if (!fVisibleValid) {
            printf("\tshould be: %s\n", 
                   (targetVisible) ? "visible" : "hidden");
        }
        if (!fEnabledValid) {
            printf("\tshould be: %s\n", 
                   (target->enabled) ? "enabled" : "disabled");
        }
    }
#endif
}
#endif //DEBUG
