/*
 * @(#)awt_Dialog.cpp	1.39 98/10/05
 *
 * Copyright (c) 1996 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

#include "awt_Toolkit.h"
#include "awt_Dialog.h"

#if defined(DEBUG)
// counts how many nested modal dialogs are open, a sanity
// check to ensure the somewhat complicated disable/enable
// code is working properly
int AwtModalityNestCounter = 0;
#endif
// property name tagging windows disabled by modality
static const char * ModalDisableProp = "SunAwtModalDisableProp";

///////////////////////////////////////////////////////////////////////////
// AwtDialog class methods

AwtDialog::AwtDialog() {
    m_modalWnd = NULL;
}

AwtDialog::~AwtDialog() {
    if (m_modalWnd != NULL) {
	WmEndModal();
    }
}

const char* AwtDialog::GetClassName() {
  return "SunAwtDialog";
}

// Override RegisterClass so dialogs do not have icon by default
void AwtDialog::RegisterClass()
{
    const char* className = GetClassName();
    WNDCLASS  wc;

    if (!::GetClassInfo(::GetModuleHandle(NULL), className, &wc)) {
        wc.style         = CS_OWNDC | CS_SAVEBITS; 
	wc.lpfnWndProc   = (WNDPROC)::DefWindowProc;
	wc.cbClsExtra    = 0;
	wc.cbWndExtra    = DLGWINDOWEXTRA;
	wc.hInstance     = ::GetModuleHandle(NULL);
	wc.hIcon         = NULL; // Dialogs do not have an icon by default
	wc.hCursor       = NULL;
	wc.hbrBackground = NULL;
	wc.lpszMenuName  = NULL;
	wc.lpszClassName = className;

        ATOM ret = ::RegisterClass(&wc);
        ASSERT(ret != 0);
    }
}

// Create a new AwtDialog object and window.  
AwtDialog* AwtDialog::Create(Hsun_awt_windows_WDialogPeer* self,
	                   Hsun_awt_windows_WComponentPeer* hParent)
{
    Hjava_awt_Dialog* target = (Hjava_awt_Dialog*)unhand(self)->target;
    CHECK_NULL_RETURN(target, "null target");

    HWND hwndParent = NULL;
    if (hParent != NULL) {
	CHECK_PEER_RETURN(hParent);
	AwtFrame* parent = PDATA(AwtFrame, hParent);
	hwndParent = parent->GetHWnd();
    }

    AwtDialog* dialog = new AwtDialog();
    CHECK_NULL_RETURN(dialog, "AwtDialog() failed");
	
    //Netscape : Bug # 79863 : Make sure the size of the dialog is at 
    //least the minimim tracking size because making a dialog of
    //zero size renders it invisible.

    if ((int)unhand(target)->width == 0) {
	    unhand(target)->width = ::GetSystemMetrics(SM_CXMINTRACK);
    }
    if ((int)unhand(target)->height == 0) {
	    unhand(target)->height = ::GetSystemMetrics(SM_CYMINTRACK) + ::GetSystemMetrics(SM_CYSIZEFRAME);
    }	

    int colorId = IS_WIN4X ? COLOR_3DFACE : COLOR_WINDOW;
    DWORD style = WS_POPUP | WS_CAPTION | WS_SYSMENU | WS_CLIPCHILDREN;
    style &= ~(WS_MINIMIZEBOX|WS_MAXIMIZEBOX);
    //DWORD exStyle = IS_WIN4X ? WS_EX_WINDOWEDGE | WS_EX_DLGMODALFRAME: 0;
    DWORD exStyle = IS_WIN4X ? WS_EX_WINDOWEDGE : 0; // WS_EX_DLGMODALFRAME gets added in SetResizable now...
    
    dialog->CreateHWnd("",
                       style, exStyle,
                       (int)unhand(target)->x,
                       (int)unhand(target)->y,
                       (int)unhand(target)->width,
                       (int)unhand(target)->height,
                       hwndParent,
                       NULL,
                       ::GetSysColor(COLOR_WINDOWTEXT),
                       ::GetSysColor(colorId),
                       (Hsun_awt_windows_WComponentPeer*)self
                      );

    //Netscape: Force the WM_NCALCSIZE to be sent to the window which 
    //in turn makes the window compute its insets.  This has to happen
    //now so that subsequent calls to getInsets() will return proper
    //values.
    dialog->RecalcNonClient();
    dialog->UpdateDialogIcon();
    dialog->UpdateSystemMenu();
 
    if (unhand(target)->foreground == NULL) {
	execute_java_dynamic_method(EE(), (HObject*)self, "setDefaultColor", 
			            "(Z)V", (IS_WIN4X != 0));
    }

    return dialog;
}

//
// Disables top-level windows for a modal dialog
//	hWndDlg		Modal dialog 
//
void AwtDialog::ModalDisable(HWND hWndDlg)
{
#if defined(DEBUG)
    ASSERT(AwtModalityNestCounter >= 0 && AwtModalityNestCounter <= 1000); // sanity check
    AwtModalityNestCounter++;
#endif
    AwtDialog::ResetDisabledLevel(hWndDlg);
    ::EnumThreadWindows(AwtToolkit::MainThread(),
		    (WNDENUMPROC)DisableTopLevelsCallback,(LPARAM)hWndDlg) ;
}

// 
// Re-enables top-level windows for a modal dialog
//	hWndDlg		Modal dialog
//
void AwtDialog::ModalEnable(HWND hWndDlg)
{
#if defined(DEBUG)
    ASSERT(AwtModalityNestCounter > 0 && AwtModalityNestCounter <= 1000); // sanity check
    AwtModalityNestCounter--;
#endif
    int disabledLevel = AwtDialog::GetDisabledLevel(hWndDlg);
    ::EnumThreadWindows(AwtToolkit::MainThread(),
		    (WNDENUMPROC)EnableTopLevelsCallback, (LPARAM)disabledLevel);
    AwtDialog::ResetDisabledLevel(hWndDlg);
}

//
// Brings a window to foreground when modal dialog goes away
// (the Win32 Z-order is very screwy if another app is active
//  when the modal dialog goes away, so we explicitly activate 
//  the first enabled, visible, window we find)
//
//	hWndParent	Parent window of the dialog
//
void AwtDialog::ModalNextWindowToFront(HWND hWndParent)
{
    if ( ::IsWindowEnabled(hWndParent) && ::IsWindowVisible(hWndParent) ) {
    // always give parent first shot at coming to the front
	::SetForegroundWindow(hWndParent);
    } else {
    // otherwise, activate the first enabled, visible window we find
	::EnumThreadWindows(AwtToolkit::MainThread(),
		    (WNDENUMPROC)NextWindowToFrontCallback, 0L );
    }
}

//
// Disables top-level windows, incrementing a counter on each disabled
// window to indicate how many nested modal dialogs need the window disabled
//
BOOL CALLBACK AwtDialog::DisableTopLevelsCallback(HWND hWnd, LPARAM lParam)
{
    HWND	hWndDlg = (HWND)lParam;

    if (hWnd == hWndDlg || hWnd == AwtToolkit::GetInstance().GetHWnd()) {
	return TRUE; // ignore myself and toolkit when disabling
    }

    // the drop-down list portion of combo-boxes are actually top-level
    // windows with WS_CHILD, so ignore these guys...
    if ( (::GetWindowLong(hWnd, GWL_STYLE) & WS_CHILD) != 0) {
	return TRUE;
    }

    ASSERT(hWndDlg == NULL || IsWindow(hWndDlg));

    AwtDialog::IncrementDisabledLevel(hWnd, 1);
    return TRUE;
}

//
// Enables top-level window if no modal dialogs need it disabled
//
BOOL CALLBACK AwtDialog::EnableTopLevelsCallback(HWND hWnd, LPARAM lParam)
{
    HWND	hWndDlg = (HWND)lParam;
    int		disabledLevel = AwtDialog::GetDisabledLevel(hWnd);
    int		dlgDisabledLevel = (int)lParam;

    if (disabledLevel == 0 || disabledLevel  < dlgDisabledLevel) {
    // skip enabled windows or windows disabled after the 
    // given modal dialog became modal
	return TRUE;
    }

    AwtDialog::IncrementDisabledLevel(hWnd, -1);
    return TRUE;
}

//
// Brings the first enabled, visible window to the foreground
//
BOOL CALLBACK AwtDialog::NextWindowToFrontCallback(HWND hWnd, LPARAM lParam)
{
    if ( ::IsWindowVisible(hWnd) && ::IsWindowEnabled(hWnd) ) {
	::SetForegroundWindow(hWnd);
	return FALSE; // stop callbacks
    }
    return TRUE;
}

int AwtDialog::GetDisabledLevel(HWND hWnd) {
    int	disabledLevel = (int)::GetProp(hWnd, ModalDisableProp);
    ASSERT(disabledLevel >= 0 && disabledLevel <= 1000);
    return disabledLevel;
}

#if 0 // defined(DEBUG)
// debugging-only function; useful for tracking down disable/enable problems
static void DbgShowDisabledLevel(HWND hWnd, int disabledLevel)
{    
    // put disabled counter in title bar text
    char	szDisabled[256];
    wsprintf(szDisabled, "disabledLevel = %d", disabledLevel);
    ::SetWindowText(hWnd, szDisabled);
}
    #define DBG_SHOW_DISABLED_LEVEL( _win, _dl ) \
	    DbgShowDisabledLevel( (_win), (_dl) )
#else    
    #define DBG_SHOW_DISABLED_LEVEL( _win, _dl )
#endif

void AwtDialog::IncrementDisabledLevel(HWND hWnd, int increment) {
    int	disabledLevel = GetDisabledLevel(hWnd);
    ASSERT( increment == -1 || increment == 1); // only allow increment/decrement
    disabledLevel += increment;
    if (disabledLevel == 0 ) {
	// don't need the property any more...
	ResetDisabledLevel(hWnd);
    } else {
	::SetProp(hWnd, ModalDisableProp, (HANDLE)disabledLevel);
    }
    ::EnableWindow(hWnd, disabledLevel == 0);
    DBG_SHOW_DISABLED_LEVEL(hWnd, GetDisabledLevel(hWnd));
}

void AwtDialog::ResetDisabledLevel(HWND hWnd) {
    ::RemoveProp(hWnd, ModalDisableProp);
    DBG_SHOW_DISABLED_LEVEL(hWnd, GetDisabledLevel(hWnd));
}

MsgRouting AwtDialog::WmShowModal()
{
    ASSERT(::GetCurrentThreadId() == AwtToolkit::MainThread());

    // disable top-level windows
    AwtDialog::ModalDisable( GetHWnd() );
    // enable myself (could have been explicitly disabled via setEnabled)
    ::EnableWindow(GetHWnd(), TRUE); // is this the right behaviour? (rnk 6/4/98)
    // show window and set focus to it
    SendMessage(WM_AWT_COMPONENT_SHOW, SW_SHOW);
    // normally done by DefWindowProc(WM_ACTIVATE),
    // but we override that message so we do it explicitly
    ::SetFocus(GetHWnd());

    m_modalWnd = GetHWnd();

    return mrConsume;
}

MsgRouting AwtDialog::WmEndModal()
{
    ASSERT( ::GetCurrentThreadId() == AwtToolkit::MainThread() );
    ASSERT( ::IsWindowVisible( GetHWnd() ) );
    ASSERT( ::IsWindow(m_modalWnd) );

    // re-enable top-level windows
    AwtDialog::ModalEnable( GetHWnd() );
    
    HWND	hWndParent = ::GetParent( GetHWnd() );
    BOOL	isEnabled = ::IsWindowEnabled(hWndParent);

    // Need to temporarily enable the dialog's owner window
    // or Windows could activate another application. This
    // happens when another app was activated previous to
    // a nested dialog being activated/closed (in which case the
    // owner wouldn't be re-enabled in ModalEnable).
    ::EnableWindow(hWndParent, TRUE);
    // hide the dialog
    SendMessage(WM_AWT_COMPONENT_SHOW, SW_HIDE);
    // restore the owner's true enabled state
    ::EnableWindow(hWndParent, isEnabled);

    // bring the next window in the stack to the front
    AwtDialog::ModalNextWindowToFront( hWndParent );
    m_modalWnd = NULL;

    return mrConsume;
}

void AwtDialog::SetResizable(BOOL isResizable)
{
    // call superclass
    AwtFrame::SetResizable(isResizable);

    long    style = GetStyle();
    long    xstyle = GetStyleEx();
    if (isResizable) {
    // remove modal frame
	xstyle &= ~WS_EX_DLGMODALFRAME;
    } else {
    // add modal frame
	xstyle |= WS_EX_DLGMODALFRAME;
    }
    // dialogs are never minimizable/maximizable, so remove those bits
    style &= ~(WS_MINIMIZEBOX|WS_MAXIMIZEBOX);
    SetStyle(style);
    SetStyleEx(xstyle);
    RedrawNonClient();
}

// Adjust system menu so that:
//  Non-resizable dialogs only have Move and Close items
//  Resizable dialogs only have Move/Close/Size
// Normally, Win32 dialog system menu handling is done via
// CreateDialog/DefDlgProc, but our dialogs are using DefWindowProc
// so we handle the system menu ourselves
void AwtDialog::UpdateSystemMenu()
{
    HWND    hWndSelf = GetHWnd();
    BOOL    isResizable = ((GetStyle() & WS_THICKFRAME) != 0);

    // first of all, restore the default menu
    ::GetSystemMenu(hWndSelf, TRUE);
    // now get a working copy of the menu
    HMENU   hMenuSys = GetSystemMenu(hWndSelf, FALSE);
    // remove inapplicable sizing commands
    ::DeleteMenu(hMenuSys, SC_MINIMIZE, MF_BYCOMMAND);
    ::DeleteMenu(hMenuSys, SC_RESTORE, MF_BYCOMMAND);
    ::DeleteMenu(hMenuSys, SC_MAXIMIZE, MF_BYCOMMAND);
    if (!isResizable) {
	::DeleteMenu(hMenuSys, SC_SIZE, MF_BYCOMMAND);
    }
    // remove separator if only 3 items left (Move, Separator, and Close)
    if (::GetMenuItemCount(hMenuSys) == 3) {
	MENUITEMINFO	mi;
	mi.cbSize = sizeof(MENUITEMINFO);
	mi.fMask = MIIM_TYPE;
	::GetMenuItemInfo(hMenuSys, 1, TRUE, &mi);
	if (mi.fType & MFT_SEPARATOR) {
	    ::DeleteMenu(hMenuSys, 1, MF_BYPOSITION);
	}
    }
}

// Call this when dialog isResizable property changes, to 
// hide or show the small icon in the upper left corner
void AwtDialog::UpdateDialogIcon()
{
    BOOL    isResizable = ((GetStyle() & WS_THICKFRAME) != 0);
    HWND    hWndSelf = GetHWnd();
    HWND    hWndOwner = GetWindow(hWndSelf, GW_OWNER);
    
    if (isResizable) {
    // if we're resizable, use the icon from our owner window
	HICON	hIconOwnerBig;
	HICON	hIconOwnerSmall;
	HICON	hIconClassBig;
	HICON	hIconClassSmall;
	HICON	hIconDialog;

	// try getting owner's small icon
	hIconOwnerSmall = (HICON)::SendMessage(hWndOwner, WM_GETICON, ICON_SMALL, NULL);
        hIconOwnerBig = (HICON)::SendMessage(hWndOwner, WM_GETICON, ICON_BIG, NULL);
	hIconClassBig = (HICON)::GetClassLong(hWndOwner, GCL_HICON);
	hIconClassSmall= (HICON)::GetClassLong(hWndOwner, GCL_HICONSM);
	if (hIconOwnerSmall != NULL) {
	    hIconDialog = hIconOwnerSmall;
	} else if (hIconOwnerBig != NULL) {
	    hIconDialog = hIconOwnerBig;
	} else if (hIconClassSmall != NULL) {
	    hIconDialog = hIconClassSmall;
	} else {
	    hIconDialog = hIconClassBig;
	}
	::PostMessage(hWndSelf, WM_SETICON, ICON_SMALL, (LPARAM)hIconDialog);
	::PostMessage(hWndSelf, WM_SETICON, ICON_BIG, (LPARAM)hIconDialog);
    } else {
    // if we're not sizable remove the icon
	::PostMessage(hWndSelf, WM_SETICON, ICON_SMALL, NULL);
	::PostMessage(hWndSelf, WM_SETICON, ICON_BIG, NULL);
    }
}

// Override WmStyleChanged to adjust system menu for sizable/non-resizable dialogs
MsgRouting AwtDialog::WmStyleChanged(WORD wStyleType, LPSTYLESTRUCT lpss)
{
    UpdateSystemMenu();
    UpdateDialogIcon();
    return mrConsume;
}

MsgRouting AwtDialog::WmSize(int type, int w, int h)
{
    UpdateSystemMenu(); // adjust to reflect restored vs. maximized state
    return AwtFrame::WmSize(type, w, h);
}

LRESULT AwtDialog::WindowProc(UINT message, WPARAM wParam, LPARAM lParam)
{
    MsgRouting mr = mrDoDefault;
    LRESULT retValue = 0L;

    switch(message) {
	case WM_AWT_DLG_SHOWMODAL:
	    mr = WmShowModal();
	    break;
	case WM_AWT_DLG_ENDMODAL:
	    mr = WmEndModal();
	    break;
    }

    if (mr != mrConsume) {
        retValue = AwtFrame::WindowProc(message, wParam, lParam);
    }
    return retValue;
}

/////////////////////////////////////////////////////////////////////////////////////////////
// DialogPeer native methods

void
sun_awt_windows_WDialogPeer_create(Hsun_awt_windows_WDialogPeer* self,
				  Hsun_awt_windows_WComponentPeer* hParent)
{
    AwtToolkit::CreateComponent(
        self, hParent, (AwtToolkit::ComponentFactory)AwtDialog::Create);

    CHECK_PEER_CREATION(self);
}

void
sun_awt_windows_WDialogPeer_showModal(struct Hsun_awt_windows_WDialogPeer *self)
{
    CHECK_PEER(self);
    AwtDialog* d = (AwtDialog*)unhand(self)->pData;
    ::PostMessage( d->GetHWnd(), WM_AWT_DLG_SHOWMODAL, 0, 0L);
}

void
sun_awt_windows_WDialogPeer_endModal(struct Hsun_awt_windows_WDialogPeer *self)
{
    CHECK_PEER(self);
    AwtDialog* d = (AwtDialog*)unhand(self)->pData;
    ::PostMessage( d->GetHWnd(), WM_AWT_DLG_ENDMODAL, 0, 0L);
}
