/*
 * @(#)awt_Graphics.h	1.80 98/07/01
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
#ifndef _AWTGRAPHICS_H_
#define _AWTGRAPHICS_H_

#include "awt_image.h"
#include "awt_Brush.h"
#include "awt_Pen.h"
#include <sun_awt_windows_WGraphics.h>

#define HWinGraphics Hsun_awt_windows_WGraphics
#define HWinPrintJob Hsun_awt_windows_WPrintJob

extern int LogicalSelectClipRgn(HDC hDC, HRGN hClipRgn);

class AwtSharedDC
{
public:
    INLINE AwtSharedDC(AwtComponent* pComponent, HDC hDC) {
        m_useCount = 1;
        m_pOwner = NULL;
        m_hDC = hDC;
        m_hOldBmp = NULL;
        m_pComponent = pComponent;
        ::InitializeCriticalSection(&m_lock);
    }

    // Get the Window being drawn to
    // (NULL if this is an offscreen graphics)
    INLINE AwtComponent* GetComponent() {
        return m_pComponent;
    }
    INLINE HWND GetHWnd() {
        return (m_pComponent == NULL) ? NULL : m_pComponent->GetHWnd();
    }

    // Get the Device Context without "grabbing it"
    // (only used as a source DC for off-screen blits)
    INLINE HDC GetDC() {
	return m_hDC;
    }

    // Attach this Shared DC to some object (increment reference count)
    // Returns TRUE if successfully attached, and FALSE if the object has
    // been asynchronously destroyed.
    INLINE BOOL Attach() {
        // If the object is already invalid (countReferences is zero)
        // we can't attach to the object
        if (!m_useCount)
            return FALSE;
        // Increase the number of references
        ::InterlockedIncrement(&m_useCount);
        return TRUE;
    }

    // Detach this Shared DC from some object
    // actual detach is performed in Detach()
    void Detach(class AwtGraphics *pOldOwner);

    INLINE void SetBitmap(HBITMAP hBM) {
	m_hOldBmp = (HBITMAP) ::SelectObject(m_hDC, hBM);
    }

    INLINE void UnsetBitmap() {
	if (m_hOldBmp != NULL) {
	    ::SelectObject(m_hDC, m_hOldBmp);
	}
    }

    // Lock this shared DC for exclusive writing.  This is necessary so that one 
    // thread isn't executing GDI calls when another thread has the DC.
    // Locking always implies attaching
    INLINE void Lock() {
        // keep this object referenced while a lock is outstanding
	Attach();
        AWT_CS_TRACE((stderr, "> %2X locking  (cnt:%d,rec:%d,own:%X) [%s hwnd:%X]\n", ::GetCurrentThreadId(),
                      m_lock.LockCount, m_lock.RecursionCount, m_lock.OwningThread,
                      "AwtSharedDC", GetHWnd()));
        ::EnterCriticalSection(&m_lock); 
        AWT_CS_TRACE((stderr, "> %2X  locked  (cnt:%d,rec:%d,own:%X) [%s hwnd:%X]\n", ::GetCurrentThreadId(),
                      m_lock.LockCount, m_lock.RecursionCount, m_lock.OwningThread,
                      "AwtSharedDC", GetHWnd()));
    } 

    // Grab this Device Context and assign a new "owner"
    // must only be called after lock is obtained
    HDC GrabDC(class AwtGraphics *pNewOwner);

    // The current (grabbed) DC has become invalid -- release and re-get it (workaround for NT 4.0 bug).
    // must only be called after lock is obtained, *and* dc is grabbed
    HDC ResetDC();

    // Unlock this shared DC
    // Unlocking implies detaching, including free/delete if last reference
    INLINE void Unlock() {
        ::LeaveCriticalSection(&m_lock);
        AWT_CS_TRACE((stderr, "> %2X unlocked (cnt:%d,rec:%d,own:%X) [%s hwnd:%X]\n", ::GetCurrentThreadId(),
                      m_lock.LockCount, m_lock.RecursionCount, m_lock.OwningThread,
                      "AwtSharedDC", GetHWnd()));
        // decrement reference count and free if necessary        
        Detach();
    } 

private:
    // private to make sure we are only deleted from our own Detach() below, and
    // always from a single thread
    INLINE ~AwtSharedDC() {
        FreeDC();  // cleanup & release dc as needed
        ::DeleteCriticalSection(&m_lock);
    }

    // perform actual Detach of this Shared DC (decrement reference count and
    // free if necessary)
    INLINE void Detach() {
        if (m_useCount > 0 && ::InterlockedDecrement(&m_useCount) == 0)
            delete this;
    }

    // Get rid of the Windows DC object
    void FreeDC();

    long		 m_useCount;
    class AwtGraphics	*m_pOwner;
    HDC			 m_hDC;
    HBITMAP		 m_hOldBmp;
    AwtComponent	*m_pComponent;
    CRITICAL_SECTION     m_lock;
};

class AwtGraphics : public AwtObject
{
public:
    AwtGraphics();
    virtual ~AwtGraphics();

// Access methods
    INLINE int GetRop() { return m_rop; }

// Operations
    // Creates a graphics context on 'pWindow'.
    static AwtGraphics *Create(HWinGraphics *pJavaObject, 
                               AwtComponent* pWindow);
    static AwtGraphics *Create(HWinGraphics *pJavaObject, AwtImage *pImage);
    static AwtGraphics *CreateCopy(HWinGraphics *pJavaObject, 
                                   AwtGraphics *pSource);
    static AwtGraphics *Create(HWinGraphics *pJavaObject, HDC hdc);

    virtual void Dispose();

    INLINE AwtComponent* GetComponent() {
        if (m_pSharedDC == NULL) {return NULL;}
        return m_pSharedDC->GetComponent();
    }

    INLINE HWND GetHWnd() {
        if (m_pSharedDC == NULL) {return NULL;}
        return m_pSharedDC->GetHWnd();
    }

protected:
    INLINE HDC GetDC() {
        if (m_pSharedDC != NULL) {
            m_pSharedDC->Lock();
            if (m_hDC == NULL) {
                ValidateDC();
                if (m_hDC == NULL) {
                    m_pSharedDC->Unlock();
                }
            } else if (m_bClipping) {
		VERIFY(LogicalSelectClipRgn(m_hDC, m_hClipRgn) != ERROR);
            }
        }
        return m_hDC;
    }
    INLINE void ReleaseDC() {
        if (m_pSharedDC != NULL) {
            if (m_pSharedDC->GetComponent() != NULL) {
                // Bug #4055623 - Performance enhancements
                if (m_bClipping) {
                    LogicalSelectClipRgn(m_hDC, NULL);
                }
            } 
            m_pSharedDC->Unlock();
        }
    }

public:
    void ValidateDC();
    void InvalidateDC();

    // Sets the m_color instance variable to the absolute COLORREF
    void ChangeColor(COLORREF color);

    // Calculates the appropriate new COLORREF according to the
    // current ROP and calls ChangeColor on the new COLORREF
    void SetColor(COLORREF color);

    void SetFont(AwtFont *pFont);

    INLINE AwtFont* GetFont() { return m_pFont; }

    void SetRop(int rop, COLORREF fgcolor, COLORREF color);

    void ChangeClip(long nX, long nY, long nW, long nH, BOOL set);
    void RemoveClip();

    // Returns the clipping rectangle.  If there is none, return NULL.
    RECT* GetClipRect();

    // Draws a line using the foreground color between points (nX1, nY1)
    // and (nX2, nY2).  The end-point is painted.
    void DrawLine(long nX1, long nY1, long nX2, long nY2);

    POINT* TransformPoly(long* xpoints, long* ypoints, POINT *points,
			 int *pNpoints, BOOL close);

    BOOL Polyline(long* xpoints, long* ypoints, int nPoints, BOOL close);

    BOOL FillPolygon(long* xpoints, long* ypoints, int nPoints);

    void ClearRect(long nX, long nY, long nW, long nH);

    void FillRect(long nX, long nY, long nW, long nH);

    void DrawRect(long nX, long nY, long nW, long nH);

    void RoundRect(long nX, long nY, long nW, long nH, long nArcW, long nArcH);
	
    void FillRoundRect(long nX, long nY, long nW, long nH, long nArcW, 
                       long nArcH);
	
    void Ellipse(long nX, long nY, long nW, long nH);
	
    void FillEllipse(long nX, long nY, long nW, long nH);
	
    // Converts an angle given in 1/64ths of a degree to a pair of coordinates
    // that intersects the line produced by the angle.
    void AngleToCoords(long nAngle, long *nX, long *nY);
	
    void Arc(long nX, long nY, long nW, long nH, long nStartAngle, 
             long nEndAngle);
	
    void FillArc(long nX, long nY, long nW, long nH, long nStartAngle, 
                 long nEndAngle);
	
private:
    INLINE void SetBrushColor(BOOL bSetBrush) {
        if (bSetBrush != m_bBrushSet) {
            ChangeBrush(bSetBrush);
        }
    };
    void ChangeBrush(BOOL bSetBrush);
    INLINE void SetPenColor(BOOL bSetPen) {
        if (bSetPen != m_bPenSet) {
            ChangePen(bSetPen);
        }
    };
    void ChangePen(BOOL bSetPen);

public:
    // Draws the string at the specified point.  (nX, nY) designates
    // the reference point of the string - i.e. the top of the string
    // plus the ascent.  The width of the string is returned.
    long DrawStringWidth(wchar_t *pzString, long nX, long nY);

    // Draws a single piece of a multi-font string at the specified 
    // point.  (nX, nY) designates the reference point of the string - 
    // i.e. the top of the string plus the ascent.  The width of the 
    // string is returned.
    long DrawSubstringWidth(BYTE* bytes, long length, Hjava_awt_Font* font, 
                            Hsun_awt_FontDescriptor* des,
                            long nX, long nY, BOOL unicode);
	
    // Draws the image at ('nX', 'nY') with optional background color c.
    void DrawImage(AwtImage *pImage, long nX, long nY,
                   struct Hjava_awt_Color *c);

    // Draws the subrectangle ('nSX1', 'nSY1') => ('nSX2', 'nSY2') of the image
    // inside the rectangle ('nDX1', 'nDY1') => ('nDX2', 'nDY2') of the DC
    // with optional background color c.
    void DrawImage(AwtImage *pImage,
		   long nDX1, long nDY1, long nDX2, long nDY2,
		   long nSX1, long nSY1, long nSX2, long nSY2,
                   struct Hjava_awt_Color *c);

    void CopyArea(long nDstX, long nDstY, long nX, long nY, long nW, long nH);

    // Delete any AwtGraphics objects which were not disposed.
    static void Cleanup();

    void PrintComponent(class AwtComponent* c);
    int GetOX() { return unhand(m_pJavaObject)->originX; };
    int GetOY() { return unhand(m_pJavaObject)->originY; }

// Variables
protected:
    // TRUE iff the underlying DC was obtained from ::PrintDlg, FALSE
    // otherwise
    BOOL m_isPrinter;

    class AwtSharedDC *m_pSharedDC;
    HDC m_hDC;
    HWinGraphics *m_pJavaObject;

private:
    BOOL m_bPaletteSelected;
    RECT m_rectClip;
    HRGN m_hClipRgn;
    BOOL m_bClipping;

    AwtFont *m_pFont;

    COLORREF m_color;
    COLORREF m_xorcolor;
    BOOL m_bBrushSet;
    AwtBrush* m_hBrush;
    BOOL m_bPenSet;
    AwtPen* m_hPen;
    int m_rop;
};

class AwtMasterPrinterGraphics : public AwtGraphics
{
public:
    AwtMasterPrinterGraphics();
    virtual ~AwtMasterPrinterGraphics() {}

    BOOL StartPrintJob(Hsun_awt_windows_WPrintJob* pj);
    void StartPage();
    void EndPage();
    void EndPrintJob();

    virtual void Dispose();

    static AwtMasterPrinterGraphics *
        Create(HWinGraphics *pJavaObject, HWinPrintJob *pj);

private:
    BOOL m_printingPage;
    BOOL m_printingJob;
};

extern int LogicalSelectClipRgn(HDC hDC, HRGN hClipRgn);

#define GET_JAVAFONT(g) \
    (Hjava_awt_Font *)execute_java_dynamic_method \
        (EE(), (HObject *)(g), "getFont", "()Ljava/awt/Font;")

#endif // _AWTGRAPHICS_H_
