/*
 * @(#)awt_Toolkit.h	1.33 98/08/25
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * The Toolkit class has two functions: it instantiates the AWT 
 * ToolkitPeer's native methods, and provides the DLL's core functions.
 *
 * There are two ways this DLL can be used: either as a dynamically-
 * loaded Java native library from the interpreter, or by a Windows-
 * specific app.  The first manner requires that the Toolkit provide
 * all support needed so the app can function as a first-class Windows
 * app, while the second assumes that the app will provide that 
 * functionality.  Which mode this DLL functions in is determined by
 * which init function is called: one is an exported DLL function, the
 * other a Java native method.
 *
 * The most basic functionality needed is a Windows message pump (also
 * known as a message loop).  When an Java app is started as a console
 * app by the interpreter, the Toolkit needs to provide that message
 * pump if the AWT is dynamically loaded.  
 */

#ifndef AWT_TOOLKIT_H
#define AWT_TOOLKIT_H

#include "awt.h"
#include "awtmsg.h"

#include "sun_awt_windows_WToolkit.h"

class AwtObject;
class AwtDialog;

//
// class CriticalSection
// ~~~~~ ~~~~~~~~~~~~~~~~
// Lightweight intra-process thread synchronization. Can only be used with
// other critical sections, and only within the same process.
//
class CriticalSection {
  public:
    // ctor argument provides a static string tag for tracing when enabled,
    // unused otherwise
#ifdef CSDEBUG
    INLINE  CriticalSection(const char* name) { ::InitializeCriticalSection(&rep); this->name = name; }
#else
    INLINE  CriticalSection(const char*) { ::InitializeCriticalSection(&rep); }
#endif
    INLINE ~CriticalSection() { ::DeleteCriticalSection(&rep); }

    class Lock {
        public:
            INLINE Lock(const CriticalSection& cs) 
	        : critSec(cs)
	    {
	        const_cast<CriticalSection &>(critSec).Enter();
	    } 
	    INLINE ~Lock()
	    { 
	        const_cast<CriticalSection &>(critSec).Leave();
	    } 

        private:
	    const CriticalSection& critSec;
    };
    friend class Lock;

  private:
    CRITICAL_SECTION rep;
#ifdef CSDEBUG
    const char* name;
#endif

    CriticalSection(const CriticalSection&);
    const CriticalSection& operator =(const CriticalSection&);

  public:
    void    Enter           (void)
    {
        AWT_CS_TRACE((stderr, "> %2X locking  (cnt:%d,rec:%d,own:%X) [%s]\n", ::GetCurrentThreadId(),
		      rep.LockCount, rep.RecursionCount, rep.OwningThread, name));
        ::EnterCriticalSection(&rep);
	AWT_CS_TRACE((stderr, "> %2X  locked  (cnt:%d,rec:%d,own:%X) [%s]\n", ::GetCurrentThreadId(),
		      rep.LockCount, rep.RecursionCount, rep.OwningThread, name));
    }
    void    Leave           (void)
    {
        ::LeaveCriticalSection(&rep);
	AWT_CS_TRACE((stderr, "> %2X unlocked (cnt:%d,rec:%d,own:%X) [%s]\n", ::GetCurrentThreadId(),
		      rep.LockCount, rep.RecursionCount, rep.OwningThread, name));
    }
};

//
// Class to encapsulate the extraction of the java string contents into a buffer
// and the cleanup of the buffer
//
class JavaStringBuffer {
  public:
    JavaStringBuffer(Hjava_lang_String* s);
    INLINE ~JavaStringBuffer() { delete[] buffer; }
    INLINE operator char*() {return buffer;}
    INLINE operator LPARAM() {return (LPARAM)buffer;}  // for SendMessage

  private:
    char* buffer;
};

class AwtToolkit {
public:
    enum {
	KB_STATE_SIZE = 256
    };

    AwtToolkit();
   ~AwtToolkit();

    BOOL Initialize(BOOL localPump);
    void Finalize();

    INLINE BOOL localPump() { return m_localPump; }
    INLINE BOOL Verbose() { return m_verbose; }
    INLINE BOOL VerifyComponents() { return m_verifyComponents; }
    INLINE BOOL BreakOnError() { return m_breakOnError; }
    INLINE HWND GetHWnd() { return m_toolkitHWnd; }

    INLINE HANDLE GetModuleHandle() { return m_dllHandle; }
    INLINE void SetModuleHandle(HANDLE h) { m_dllHandle = h; }

    INLINE static DWORD MainThread() { return GetInstance().m_mainThreadId; }
    static void Cleanup();
    static void AtExit();
    static UINT GetAsyncMouseKeyState();
    static void GetAsyncKeyboardState(PBYTE keyboardState);

    static ATOM RegisterClass();
    static void UnregisterClass();
    static LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, 
                                    LPARAM lParam);
    static LRESULT CALLBACK GetMessageFilter(int code, WPARAM wParam, 
                                             LPARAM lParam);
    static LRESULT CALLBACK ForegroundIdleFilter(int code, WPARAM wParam, 
                                                 LPARAM lParam);

    INLINE static AwtToolkit& GetInstance() { return theInstance; }
    INLINE Hsun_awt_windows_WToolkit* GetPeer() { return m_peer; }
    INLINE void SetPeer(Hsun_awt_windows_WToolkit* p) { m_peer = p; }

    // Create an AwtXxxx C++ component using a given factory
    //
    typedef void (*ComponentFactory)(void*, void*);
    static void CreateComponent(void* hComponent, void* hParent, 
                                ComponentFactory compFactory);
    static void DestroyComponent(class AwtComponent* comp);

    UINT MessageLoop();
    BOOL PumpWaitingMessages();
    BOOL PreProcessMsg(MSG& msg);
    BOOL PreProcessMouseMsg(class AwtComponent* p, MSG& msg);
    BOOL PreProcessKeyMsg(class AwtComponent* p, MSG& msg);
    
    // Create an ID which maps to an AwtObject pointer, such as a menu.
    UINT CreateCmdID(AwtObject* object);

    // removes cmd id mapping
    void RemoveCmdID(UINT id);

    // Return the AwtObject associated with its ID.
    AwtObject* LookupCmdID(UINT id);

    // Return the current application icon.
    HICON GetAwtIcon();

    // Turns on/off dialog modality for the system.
    INLINE AwtDialog* SetModal(AwtDialog* frame) { 
	AwtDialog* previousDialog = m_pModalDialog;
	m_pModalDialog = frame;
	return previousDialog;
    };
    INLINE void ResetModal(AwtDialog* oldFrame) { m_pModalDialog = oldFrame; };
    INLINE BOOL IsModal() { return (m_pModalDialog != NULL); };
    INLINE AwtDialog* GetModalDialog(void) { return m_pModalDialog; };

    // Stops the current message pump (normally a modal dialog pump)
    INLINE void StopMessagePump() { m_breakOnError = TRUE; }

    // Debug settings
    INLINE void SetVerbose(long flag)   { m_verbose = (flag != 0); }
    INLINE void SetVerify(long flag)    { m_verifyComponents = (flag != 0); }
    INLINE void SetBreak(long flag)     { m_breakOnError = (flag != 0); }
    INLINE void SetHeapCheck(long flag);

private:
    HWND CreateToolkitWnd(const char* name);

    BOOL m_localPump;
    DWORD m_mainThreadId;
    HWND m_toolkitHWnd;
    BOOL m_verbose;

    BOOL m_verifyComponents;
    BOOL m_breakOnError;

    BOOL  m_breakMessageLoop;
    UINT  m_messageLoopResult;

    class AwtComponent* m_lastMouseOver;
    BOOL                m_mouseDown;

    HHOOK m_hGetMessageHook;
    UINT  m_timer;

    class AwtCmdIDList* m_cmdIDs;
    BYTE		m_lastKeyboardState[KB_STATE_SIZE];
    CriticalSection	m_lockKB;

    static AwtToolkit theInstance;

    // The current modal dialog frame (normally NULL).
    AwtDialog* m_pModalDialog;

    // The WToolkit peer instance
    Hsun_awt_windows_WToolkit* m_peer;

    HANDLE m_dllHandle;  // The module handle.
	public:
	static int sm_nComboHeightOffset;
};

#endif // AWT_TOOLKIT_H
