/*
 * @(#)awt_Window.cpp	1.54 98/09/04 Thomas Ball
 *
 * Copyright (c) 1996 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

#include "awt_Toolkit.h"
#include "awt_Window.h"
#include "awt_Frame.h"
#include "awt_Graphics.h"

#include "java_awt_Insets.h"
#include <java_awt_Container.h>
#include <java_awt_event_ComponentEvent.h>

///////////////////////////////////////////////////////////////////////////
// AwtWindow class methods

AwtWindow::AwtWindow() {
    m_iconic = FALSE;
    m_eraseBackground = FALSE;
    m_resizing = FALSE;
    m_sizePt.x = m_sizePt.y = 0;
    VERIFY(::SetRectEmpty(&m_insets));
    VERIFY(::SetRectEmpty(&m_old_insets));
    VERIFY(::SetRectEmpty(&m_warningRect));
    SetCursor(::LoadCursor(NULL, IDC_ARROW));
}

AwtWindow::~AwtWindow() {
}

const char* AwtWindow::GetClassName() {
  return "SunAwtWindow";
}

// Create a new AwtWindow object and window.  
AwtWindow* AwtWindow::Create(Hsun_awt_windows_WWindowPeer* self,
                             Hsun_awt_windows_WComponentPeer* hParent)
{
    CHECK_NULL_RETURN(hParent, "null hParent");
    AwtWindow* parent = (AwtWindow*)unhand(hParent)->pData;
    CHECK_NULL_RETURN(parent, "null parent");
    Hjava_awt_Window* target = (Hjava_awt_Window*)unhand(self)->target;
    CHECK_NULL_RETURN(target, "null target");

    AwtWindow* window = new AwtWindow();
    CHECK_NULL_RETURN(window, "AwtWindow() failed");

    DWORD style = WS_CLIPCHILDREN | WS_POPUP;
    window->CreateHWnd("",
                       style, 0,
                       0, 0, 0, 0,
                       parent->GetHWnd(),
                       NULL,
                       ::GetSysColor(COLOR_WINDOWTEXT),
                       ::GetSysColor(COLOR_WINDOW),
                       (Hsun_awt_windows_WComponentPeer*)self
                      );

    //  here instead of during create, so that a WM_NCCALCSIZE is sent.
    window->Reshape((int)unhand(target)->x,
                    (int)unhand(target)->y,                       
                    (int)unhand(target)->width,
                    (int)unhand(target)->height);

    return window;
}

//
// Override AwtComponent's Reshape to handle minimized/maximized
// windows, fixes 4065534 (robi.khan@eng).
// NOTE: See also AwtWindow::WmSize
//
void AwtWindow::Reshape(int x, int y, int width, int height) 
{
    HWND    hWndSelf = GetHWnd();

    if ( IsIconic(hWndSelf)) {
    // normal AwtComponent::Reshape will not work for iconified windows so...
	WINDOWPLACEMENT wp;
	POINT	    ptMinPosition = {x,y};
	POINT	    ptMaxPosition = {0,0};
	RECT	    rcNormalPosition = {x,y,x+width,y+height};
	RECT	    rcWorkspace;
	HWND	    hWndDesktop = GetDesktopWindow();

	// SetWindowPlacement takes workspace coordinates, but
	// if taskbar is at top of screen, workspace coords !=
	// screen coords, so offset by workspace origin
	VERIFY(::SystemParametersInfo(SPI_GETWORKAREA, 0, (PVOID)&rcWorkspace, 0));
	::OffsetRect(&rcNormalPosition, -rcWorkspace.left, -rcWorkspace.top);

	// set the window size for when it is not-iconified
	wp.length = sizeof(wp);
	wp.flags = WPF_SETMINPOSITION;
	wp.showCmd = SW_SHOWNA;
	wp.ptMinPosition = ptMinPosition;
	wp.ptMaxPosition = ptMaxPosition;
	wp.rcNormalPosition = rcNormalPosition;
	SetWindowPlacement(hWndSelf, &wp);
	return;
    }
    
    if (IsZoomed(hWndSelf)) {
    // setting size of maximized window, we remove the
    // maximized state bit (matches Motif behaviour)
    // (calling ShowWindow(SW_RESTORE) would fire an
    //  activation event which we don't want)
	LONG	style = GetStyle();
	ASSERT(style & WS_MAXIMIZE);
	style ^= WS_MAXIMIZE;
	SetStyle(style);
    }

    AwtComponent::Reshape(x, y, width, height);
}

//
// Get and return the insets for this window (container, really).
// Calculate & cache them while we're at it, for use by AwtGraphics
// return if insets have changed
//
BOOL AwtWindow::UpdateInsets(Hjava_awt_Insets* insets)
{
    ASSERT(GetPeer() != NULL);
    Hjava_awt_Window* target = (Hjava_awt_Window*)unhand(GetPeer())->target;
    CHECK_NULL_RETURN(target, "null target");

    // fix 4167248 : don't update insets when frame is iconified
    // to avoid bizarre window/client rectangles
    if (m_iconic)
	return FALSE;

    // Code to calculate insets. Stores results in frame's data
    // members, and in the peer's Inset object.
    RECT outside;
    RECT inside;
    ::GetWindowRect(GetHWnd(), &outside);
    ::GetClientRect(GetHWnd(), &inside);

    // Update our inset member
    if (outside.right - outside.left > 0 && outside.bottom - outside.top > 0) { 
        ::MapWindowPoints(GetHWnd(), 0, (LPPOINT)&inside, 2);
        m_insets.top = inside.top - outside.top;
        m_insets.bottom = outside.bottom - inside.bottom;
        m_insets.left = inside.left - outside.left;
        m_insets.right = outside.right - inside.right;
    } else {
        // This window hasn't been sized yet -- use system metrics.
        if (IsInstanceOf(GetTarget(), "java/awt/Frame") ||
            IsInstanceOf(GetTarget(), "java/awt/Dialog")) {
            // Get outer frame sizes.
            long style = GetStyle();
            if (style & WS_THICKFRAME) {
                m_insets.left = m_insets.right = 
                    ::GetSystemMetrics(SM_CXSIZEFRAME);
                m_insets.top = m_insets.bottom = 
                    ::GetSystemMetrics(SM_CYSIZEFRAME);
            } else {
                m_insets.left = m_insets.right = 
                    ::GetSystemMetrics(SM_CXDLGFRAME);
                m_insets.top = m_insets.bottom = 
                    ::GetSystemMetrics(SM_CYDLGFRAME);
            }

            // Add in title.
            m_insets.top += ::GetSystemMetrics(SM_CYCAPTION);

	    // Add in warning string, if needed
	    if (unhand(target)->warningString != NULL) {
		m_insets.bottom += ::GetSystemMetrics(SM_CYCAPTION);
	    }

            // Add in menuBar, if any.
            if (IsInstanceOf(GetTarget(), "java/awt/Frame") &&
                ((AwtFrame*)this)->GetMenuBar()) {
                m_insets.top += ::GetSystemMetrics(SM_CYCAPTION);
            }
        }
    }

    // Get insets into our peer directly
    Hjava_awt_Insets* peerInsets = unhand(GetPeer())->insets_;
    if (peerInsets != NULL) { // may have been called during creation
        unhand(peerInsets)->top = m_insets.top;
        unhand(peerInsets)->bottom = m_insets.bottom;
        unhand(peerInsets)->left = m_insets.left;
        unhand(peerInsets)->right = m_insets.right;
    }

    // Get insets into the Inset object (if any) that was passed
    if (insets != NULL) {
	unhand(insets)->top = m_insets.top;
	unhand(insets)->left = m_insets.left;
	unhand(insets)->bottom = m_insets.bottom;
	unhand(insets)->right = m_insets.right;
    }

    BOOL	insetsChanged = !::EqualRect( &m_old_insets, &m_insets );	
    ::CopyRect( &m_old_insets, &m_insets );
    return insetsChanged;
}

/*
 * Although this function sends ComponentEvents, it needs to be defined
 * here because only top-level windows need to have move and resize
 * events fired from native code.  All contained windows have these events
 * fired from common Java code.
 */
void AwtWindow::SendComponentEvent(long eventId)
{
    static ClassClass* classEvent = NULL;
    if (classEvent == NULL) {
        char* clsName = "java/awt/event/ComponentEvent";
        classEvent = FindStickySystemClass(EE(), clsName, TRUE);
        if (classEvent == NULL) {
            SignalError(0, JAVAPKG "ClassNotFoundException", clsName);
            return;
        }
        ASSERT(classEvent);
    }

    Hjava_awt_event_ComponentEvent* hEvent = (Hjava_awt_event_ComponentEvent*)
        execute_java_constructor(EE(), NULL, classEvent,
                                 "(Ljava/awt/Component;I)", GetTarget(), 
                                 eventId);
    ASSERT(!exceptionOccurred(EE()));
    ASSERT(hEvent != NULL);
    SendEvent((Hjava_awt_AWTEvent*)hEvent);
}

void AwtWindow::SendWindowEvent(long eventId)
{
    static ClassClass* classEvent = NULL;
    if (classEvent == NULL) {
        char* clsName = "java/awt/event/WindowEvent";
        classEvent = FindStickySystemClass(EE(), clsName, TRUE);
        if (classEvent == NULL) {
            SignalError(0, JAVAPKG "ClassNotFoundException", clsName);
            return;
        }
        ASSERT(classEvent);
    }

    Hjava_awt_event_WindowEvent* hEvent = (Hjava_awt_event_WindowEvent*)
        execute_java_constructor(EE(), NULL, classEvent,
                                 "(Ljava/awt/Window;I)", GetTarget(), eventId);
    ASSERT(!exceptionOccurred(EE()));
    ASSERT(hEvent != NULL);
    SendEvent((Hjava_awt_AWTEvent*)hEvent);
}

MsgRouting AwtWindow::WmCreate()
{
    return mrDoDefault;  
}

MsgRouting AwtWindow::WmClose()
{
    SendWindowEvent(java_awt_event_WindowEvent_WINDOW_CLOSING);

    // Rely on above notification to handle quitting as needed
    return mrConsume;  
}

MsgRouting AwtWindow::WmDestroy()
{
    SendWindowEvent(java_awt_event_WindowEvent_WINDOW_CLOSED);
    return AwtComponent::WmDestroy();
}

//
// Override AwtComponent's move handling to first update the java AWT target's position
// fields directly, since Windows and below can be resized from outside of java (by user)
//
MsgRouting AwtWindow::WmMove(int x, int y)
{
    if ( IsIconic(GetHWnd()) ) {
    // fixes 4065534 (robi.khan@eng)
    // if a window is iconified we don't want to update
    // it's target's position since minimized Win32 windows
    // move to -32000, -32000 for whatever reason
    // NOTE: See also AwtWindow::Reshape
	return mrDoDefault;
    }

    // Update the java AWT target component's fields directly
    //
    Classjava_awt_Component* target = (Classjava_awt_Component*)
        unhand(unhand(m_peerObject)->target);
    RECT rect;
    GetSharedDC()->Lock();
    ::GetWindowRect(GetHWnd(), &rect);
    GetSharedDC()->Unlock();

    if (target->x != rect.left || target->y != rect.top) {
        target->x = rect.left;
        target->y = rect.top;
        SendComponentEvent(java_awt_event_ComponentEvent_COMPONENT_MOVED);
    }

    return AwtComponent::WmMove(x, y);
}

//
// Override AwtComponent's size handling to first update the java AWT target's dimension
// fields directly, since Windows and below can be resized from outside of java (by user)
//
MsgRouting AwtWindow::WmSize(int type, int w, int h)
{
     if (type == SIZE_MINIMIZED) {
        SendWindowEvent(java_awt_event_WindowEvent_WINDOW_ICONIFIED);
        m_iconic = TRUE;
	ASSERT(IsIconic(GetHWnd()));
    }
    else if (type == SIZE_RESTORED && m_iconic) {
        SendWindowEvent(java_awt_event_WindowEvent_WINDOW_DEICONIFIED);
        m_iconic = FALSE;
    }
    else if (type == SIZE_MAXIMIZED && m_iconic) {
        SendWindowEvent(java_awt_event_WindowEvent_WINDOW_DEICONIFIED);
        m_iconic = FALSE;
    }
    else {
        Classjava_awt_Component* target = (Classjava_awt_Component*)
            unhand(GetTarget());
	// fix 4167248 : ensure the insets are up-to-date before using
	BOOL insetsChanged = UpdateInsets(NULL);
        int newWidth = w + m_insets.left + m_insets.right;
        int newHeight = h + m_insets.top + m_insets.bottom;
        if (target->width != newWidth || target->height != newHeight ||
	    insetsChanged) {
            target->width = newWidth;
            target->height = newHeight;
            if (!m_resizing) {
                SendComponentEvent(java_awt_event_ComponentEvent_COMPONENT_RESIZED);
            }
        }
        return AwtComponent::WmSize(type, w, h);
    }
    
    return mrDoDefault;
}

MsgRouting AwtWindow::WmPaint(HDC)
{
    GetSharedDC()->Lock();
    // Get the rectangle that covers all update regions, if any exist.
    RECT r;
    if (::GetUpdateRect(GetHWnd(), &r, FALSE)) {

        // Now remove all update regions from this window -- do it here instead
        // of after the Java upcall, in case any new updating is requested.
        // If we don't do this any components that appear off the 
		// visable screen will not be drawn, and will throw is into a loop of 
		// WmPaint.
		// paulsh@ireland Bug ID 4114208.
 
		::ValidateRgn(GetHWnd(), NULL);

		::ValidateRect(GetHWnd(), &r);

	::OffsetRect(&r, m_insets.left, m_insets.top);

	GetSharedDC()->Unlock();

        if ((r.right-r.left) > 0 && (r.bottom-r.top) > 0 &&
            m_peerObject != NULL && m_callbacksEnabled) {
            DoCallback("handleExpose", // BUGID 4070488 handlerepaint was a no-op
                       "(IIII)V", r.left, r.top, r.right-r.left, r.bottom-r.top);
        }
    } else {
        GetSharedDC()->Unlock();
    }
    
	//Netscape : Bug # 82231 : NT3.51 -When you have 2 dialogs ovelapping 
	//each other, WM_ERASEBKGND msgs are not sent to the back window when the
	//front one is dragged accross it and the front windows borders are 
	//completely within those of the back dialog.  So force "handleExpose"
	//always on versions < 4.0
 
	if (IS_WIN4X)
	{
		m_eraseBackground = FALSE;
	}
    return mrConsume;
}

MsgRouting AwtWindow::WmSysCommand(UINT uCmdType, UINT xPos, UINT yPos)
{
    if (uCmdType == SC_SIZE) {
        m_resizing = TRUE;
    }
    return mrDoDefault;
}

MsgRouting AwtWindow::WmExitSizeMove()
{
    if (m_resizing) {
        SendComponentEvent(java_awt_event_ComponentEvent_COMPONENT_RESIZED);
        m_resizing = FALSE;
    }
    return mrDoDefault;
}

MsgRouting AwtWindow::WmSettingChange(WORD wFlag, LPCTSTR pszSection)
{
    if (wFlag == SPI_SETNONCLIENTMETRICS) {
    // user changed window metrics in
    // Control Panel->Display->Appearance
    // which may cause window insets to change
	UpdateInsets(NULL);
	return mrConsume;
    }
    return mrDoDefault;
}

MsgRouting AwtWindow::WmNcCalcSize(BOOL fCalcValidRects, 
                                   LPNCCALCSIZE_PARAMS lpncsp, UINT& retVal)
{
    if (fCalcValidRects == FALSE) {
        return mrDoDefault;
    }

    Hjava_awt_Window* target = (Hjava_awt_Window*)GetTarget();
    if (unhand(target)->warningString != NULL) {
	RECT r;
	::CopyRect(&r, &lpncsp->rgrc[0]);
	retVal = DefWindowProc(WM_NCCALCSIZE, fCalcValidRects, (LPARAM)lpncsp);

	// Adjust non-client area for warning banner space.
	m_warningRect.left = lpncsp->rgrc[0].left;
	m_warningRect.right = lpncsp->rgrc[0].right;
	m_warningRect.bottom = lpncsp->rgrc[0].bottom;
	m_warningRect.top = m_warningRect.bottom - ::GetSystemMetrics(SM_CYCAPTION);
        if (GetStyle() & WS_THICKFRAME) {
	    m_warningRect.top -= 2;
        } else {
	    m_warningRect.top += 2;
        }

	lpncsp->rgrc[0].bottom = m_warningRect.top;

	// Convert to window-relative coordinates.
	::OffsetRect(&m_warningRect, -r.left, -r.top);

        // Notify target of Insets change.
        UpdateInsets(NULL);

	return mrConsume;
	// NB:  AwtDialog calls us, and relies on the fact that
	// DefWindowProc() is called if and only if mrDoDefault is
	// not returned by us.
    } else {
	// WM_NCCALCSIZE is usually in response to a resize, but
	// also can be triggered by SetWindowPos(SWP_FRAMECHANGED),
	// which means the insets will have changed - rnk 4/7/1998
	retVal = DefWindowProc(WM_NCCALCSIZE, fCalcValidRects, (LPARAM)lpncsp);
 	UpdateInsets(NULL);
	return mrConsume;
    }
}

MsgRouting AwtWindow::WmNcPaint(HRGN hrgn)
{
    DefWindowProc(WM_NCPAINT, (WPARAM)hrgn, 0);
    Hjava_awt_Window* target = (Hjava_awt_Window*)GetTarget();
    if (unhand(target)->warningString != NULL) {
	RECT r;
        ::CopyRect(&r, &m_warningRect);
	HDC hDC = ::GetWindowDC(GetHWnd());
	ASSERT(hDC);
	int iSaveDC = ::SaveDC(hDC);
	VERIFY(::SelectClipRgn(hDC, NULL) != NULL);
	VERIFY(::FillRect(hDC, &m_warningRect, ::GetStockObject(BLACK_BRUSH)));

        if (GetStyle() & WS_THICKFRAME) {
            // draw edge
            VERIFY(::DrawEdge(hDC, &r, EDGE_RAISED, BF_TOP));
            r.top += 2;
            VERIFY(::DrawEdge(hDC, &r, EDGE_SUNKEN, BF_RECT));
            ::InflateRect(&r, -2, -2);
        }

	// draw warning text
	LPWSTR text = GET_WSTRING(unhand(target)->warningString);
	VERIFY(::SetBkColor(hDC, RGB(255, 225, 0)) != CLR_INVALID);
	VERIFY(::SetTextColor(hDC, RGB(0, 0, 0)) != CLR_INVALID);
	VERIFY(::SelectObject(hDC, ::GetStockObject(DEFAULT_GUI_FONT)) != NULL);
	VERIFY(::SetTextAlign(hDC, TA_LEFT | TA_BOTTOM) != GDI_ERROR);
        VERIFY(::ExtTextOutW(hDC, r.left, r.bottom, ETO_CLIPPED | ETO_OPAQUE, 
	                     &r, text, wcslen(text), NULL));
	VERIFY(::RestoreDC(hDC, iSaveDC));
	::ReleaseDC(GetHWnd(), hDC);
    }
    return mrConsume;
}

MsgRouting AwtWindow::WmNcHitTest(UINT x, UINT y, UINT& retVal)
{
    POINT pt = { x, y };
    if (::PtInRect(&m_warningRect, pt)) {
        retVal = HTNOWHERE;
    } else {
        retVal = DefWindowProc(WM_NCHITTEST, 0, MAKELPARAM(x, y));
    }
    return mrConsume;
}

BOOL CALLBACK InvalidateChildRect(HWND hWnd, LPARAM)
{
    ::InvalidateRect(hWnd, NULL, TRUE);
    return TRUE;
}

void AwtWindow::Invalidate(RECT* r)
{
    ::InvalidateRect(GetHWnd(), NULL, TRUE);
    ::EnumChildWindows(GetHWnd(), (WNDENUMPROC)InvalidateChildRect, 0);
}

void AwtWindow::SetResizable(BOOL isResizable)
{
    if (IsEmbedded()) {
	return;
    }
    long style = GetStyle();
    static const long SizeStyles = WS_THICKFRAME|WS_MAXIMIZEBOX|WS_MINIMIZEBOX;

    if (isResizable) {
	style |= SizeStyles;
    } else {
	style &= ~SizeStyles;
    }
    SetStyle(style);
    RedrawNonClient();
}

// SetWindowPos flags to cause frame edge to be recalculated
static const UINT SwpFrameChangeFlags =
    SWP_FRAMECHANGED | /* causes WM_NCCALCSIZE to be called */
    SWP_NOMOVE | SWP_NOSIZE | SWP_NOZORDER |
    SWP_NOACTIVATE | SWP_NOCOPYBITS | 
    SWP_NOREPOSITION | SWP_NOSENDCHANGING;

//
// Forces WM_NCCALCSIZE to be called to recalculate
// window border (updates insets) without redrawing it
//
void AwtWindow::RecalcNonClient()
{
    ::SetWindowPos(GetHWnd(), (HWND) NULL, 0, 0, 0, 0, SwpFrameChangeFlags|SWP_NOREDRAW);
}

//
// Forces WM_NCCALCSIZE to be called to recalculate
// window border (updates insets) and redraws border to match
//
void AwtWindow::RedrawNonClient()
{
    ::SetWindowPos(GetHWnd(), (HWND) NULL, 0, 0, 0, 0, SwpFrameChangeFlags|SWP_ASYNCWINDOWPOS);
}

/////////////////////////////////////////////////////////////////////////////////////////////
// WindowPeer native methods

void
sun_awt_windows_WWindowPeer_toFront(Hsun_awt_windows_WWindowPeer* self)
{
    CHECK_PEER(self);
    AwtWindow* w = (AwtWindow*)unhand(self)->pData;

    ::SetWindowPos(w->GetHWnd(), HWND_TOP, 0,0,0,0,
                   SWP_NOMOVE|SWP_NOSIZE);
    ::SetForegroundWindow(w->GetHWnd()) ;
}

void
sun_awt_windows_WWindowPeer_toBack(Hsun_awt_windows_WWindowPeer* self)
{
    CHECK_PEER(self);
    AwtWindow* w = (AwtWindow*)unhand(self)->pData;

    ::SetWindowPos(w->GetHWnd(), HWND_BOTTOM, 0,0,0,0,
                   SWP_NOMOVE|SWP_NOSIZE|SWP_NOACTIVATE);
}

void
sun_awt_windows_WWindowPeer__setTitle(Hsun_awt_windows_WWindowPeer* self,
                                      Hjava_lang_String* title)
{
    CHECK_PEER(self);
    CHECK_NULL(title, "null title");

    AwtWindow* w = (AwtWindow*)unhand(self)->pData;

    if (IS_NT) {
	int length = javaStringLength(title) + 1;
	WCHAR *buffer = new WCHAR[length];
	javaString2unicode(title, buffer, length);
	buffer[length-1] = L'\0';
	VERIFY(::SetWindowTextW(w->GetHWnd(), buffer));
	delete[] buffer;
    } else {
	int length = javaStringLength(title);
	char *buffer = new char[length * 2 + 1];
	AwtFont::javaString2multibyte(title, buffer, length);
	VERIFY(::SetWindowText(w->GetHWnd(), buffer));
	delete[] buffer;
    }
}

void
sun_awt_windows_WWindowPeer__setResizable(Hsun_awt_windows_WWindowPeer* self,
                                          long resizable)
{
    CHECK_PEER(self);
    AwtWindow* w = (AwtWindow*)unhand(self)->pData;
    ASSERT(!IsBadReadPtr(w, sizeof(AwtWindow)));
    w->SetResizable(resizable != 0);
}

void
sun_awt_windows_WWindowPeer_create(Hsun_awt_windows_WWindowPeer* self, Hsun_awt_windows_WComponentPeer* hParent)
{
    CHECK_PEER(hParent);
    AwtToolkit::CreateComponent(self, hParent, (AwtToolkit::ComponentFactory)AwtWindow::Create);

    CHECK_PEER_CREATION(self);
}

void
sun_awt_windows_WWindowPeer_updateInsets(Hsun_awt_windows_WWindowPeer* self,
                                         Hjava_awt_Insets* insets)
{
    CHECK_PEER(self);
    CHECK_NULL(insets, "null insets");

    AwtWindow* w = (AwtWindow*)unhand(self)->pData;
    w->UpdateInsets(insets);
}

Hjava_awt_Component* sun_awt_windows_WWindowPeer_getContainerElement(
    Hsun_awt_windows_WWindowPeer*, Hjava_awt_Container* container, long index)
{
    CHECK_NULL_RETURN(container, "null container");
    Hjava_awt_Component* comp = NULL;
 
    if (unhand(container)->ncomponents > 0) {
        HArrayOfObject* comps = unhand(container)->component;
        comp = (Hjava_awt_Component*)unhand(comps)->body[index];
    }
 
    return comp;
}

void
sun_awt_windows_WWindowPeer_requestComponentFocusIfActive(
    Hsun_awt_windows_WWindowPeer* self,
    Hsun_awt_windows_WComponentPeer * peer)
{
    CHECK_PEER(self);
    AwtWindow * window = PDATA(AwtWindow, self);
    ASSERT(peer != NULL);
    AwtComponent * component = PDATA(AwtComponent, peer);

    // Request the focus, passing the containing window as the wParam so
    // SetFocus() will not be called (causing unwanted activation) if
    // the components ancestor top-level window is not active
    // (see AwtComponent handling of WM_AWT_COMPONENT_SETFOCUS)
    WPARAM	wParam = (WPARAM)window->GetHWnd();
    VERIFY(::PostMessage(component->GetHWnd(), WM_AWT_COMPONENT_SETFOCUS, wParam, 0));
}
